/*******************************************************************************
 * Copyright (c) 2000, 2003 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package AcmeDiagramLegend.parts;

import java.util.Iterator;
import java.util.List;

import org.eclipse.draw2d.AbstractHintLayout;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.draw2d.geometry.Transposer;

import AcmeDiagramLegend.figures.LegendFigure;
import AcmeDiagramLegend.figures.LegendItemIcon;
import AcmeDiagramLegend.figures.LegendItemLabel;
import AcmeDiagramLegend.figures.ColumnSeparatorFigure;
import AcmeDiagramLegend.model.AcmeLegend;
import AcmeDiagramLegend.model.ColumnSeparator;
import acmeLib.core.globals.CategoryT;


/**
 * Lays out children in rows or columns, wrapping when the current row/column is filled.
 * The aligment and spacing of rows in the parent can be configured.  The aligment and
 * spacing of children within a row can be configured.
 */
public class LegendLayout
extends AbstractHintLayout
{
	
	private static final int DEFAULT_COLUMN_SPACING = 15;
	/** Constant to specify components to be aligned in the center */
	public static final int ALIGN_CENTER = 0;
	/** Constant to specify components to be aligned on the left/top */
	public static final int ALIGN_LEFTTOP = 1;
	/** Constant to specify components to be aligned on the right/bottom */
	public static final int ALIGN_RIGHTBOTTOM = 2;
	
	/** Constant to specify components should be layed out horizontally */
	public static final boolean HORIZONTAL = true;
	/** Constant to specify components should be layed out vertically */
	public static final boolean VERTICAL = false;
	
	/** The horizontal property. */
	protected boolean horizontal = true;
	/** 
	 * The property that determines whether leftover space at the end of a row/column should
	 * be filled by the last item in that row/column. 
	 */
	protected boolean fill = false;
	
	/** The transposer used in converting horizontal layout to vertical. */
	protected Transposer transposer; {
		transposer = new Transposer();
		transposer.setEnabled(!horizontal);
	}
	
	/** The alignment along the major axis. */
	protected int majorAlignment = ALIGN_LEFTTOP;
	/** The alignment along the minor axis. */
	protected int minorAlignment = ALIGN_LEFTTOP;
	/** The spacing along the minor axis. */
	protected int minorSpacing = 5;
	/** The spacing along the major axis. */
	protected int majorSpacing = 5;
	private WorkingData data = null;
	
	private int _numComponents;
	private int _numConnectors;
	private int _numRoles;
	private int _numPorts;
	
	private int _componentColumn;
	private int _connectorColumn;
	private int _portColumn;
	private int _roleColumn;
	
	private int _componentIconWidth;
	private int _componentTextWidth;
	private int _connectorIconWidth;
	private int _connectorTextWidth;
	private int _portIconWidth;
	private int _portTextWidth;
	private int _roleIconWidth;
	private int _roleTextWidth;
	
	private int _componentX;
	private int _connectorX;
	private int _portX;
	private int _roleX;
	
	private int _legendWidth;
	private int _legendHeight;
	private AcmeLegend _legend;
	
	
	
	/**
	 * Holds the necessary information for layout calculations.
	 */
	class WorkingData {
		int rowHeight, rowWidth, rowCount, rowX, rowY, maxWidth;
		Rectangle bounds[], area;
		IFigure row[];
	}
	
	/**
	 * Constructs a FlowLayout with horizontal orientation.
	 * @since 2.0
	 */
	public LegendLayout(AcmeLegend legend) { setHorizontal(false);
		_legend = legend;
	}
	
	/**
	 * Constructs a FlowLayout whose orientation is given in the input.
	 * @param isHorizontal <code>true</code> if the layout should be horizontal
	 * @since 2.0
	 */
	public LegendLayout(boolean isHorizontal) {
		_numComponents=0;
		_numConnectors=0;
		_numRoles=0;
		_numPorts=0;
		_componentColumn=0;
		_connectorColumn=1;
		_portColumn=2;
		_roleColumn=3;
		_legendHeight=200;
		_legendWidth=500;
		
		setHorizontal(isHorizontal);
	}
	
	/**
	 * @see org.eclipse.draw2d.AbstractLayout#calculatePreferredSize(IFigure, int, int)
	 */
	protected Dimension calculatePreferredSize(IFigure container, int wHint, int hHint) {
		// Subtract out the insets from the hints
		if (wHint > -1)
			wHint = Math.max(0, wHint - container.getInsets().getWidth());
		if (hHint > -1)
			hHint = Math.max(0, hHint - container.getInsets().getHeight());
		
		// Figure out the new hint that we are interested in based on the orientation
		// Ignore the other hint (by setting it to -1).  NOTE: The children of the
		// parent figure will then be asked to ignore that hint as well.  
		int maxWidth;
		if (isHorizontal()) {
			maxWidth = wHint;
			hHint = -1;
		} else {
			maxWidth = hHint;
			wHint = -1;
		}
		if (maxWidth <= 0) {
			maxWidth = Integer.MAX_VALUE;
		}
		
		// The preferred dimension that is to be calculated and returned
		Dimension prefSize = new Dimension();
		
		List children = container.getChildren();
		int width = 0;
		int height = 0;
		IFigure child;
		Dimension childSize;
		
		//Build the sizes for each row, and update prefSize accordingly
		for (int i = 0; i < children.size(); i++) {
			child = (IFigure)children.get(i);
			childSize = transposer.t(getChildSize(child, wHint, hHint));
			if (i == 0) {
				width = childSize.width;
				height = childSize.height;
			} else if (width + childSize.width + getMinorSpacing() > maxWidth) {
				// The current row is full, start a new row.
				prefSize.height += height + getMajorSpacing();
				prefSize.width = Math.max(prefSize.width, width);
				width = childSize.width;
				height = childSize.height;
			} else {
				// The current row can fit another child.
				width += childSize.width + getMinorSpacing();
				height = Math.max(height, childSize.height);
			}
		}
		
		// Flush out the last row's data
		prefSize.height += height;
		prefSize.width = Math.max(prefSize.width, width);
		
		// Transpose the dimension back, and compensate for the border.
		prefSize = transposer.t(prefSize);
		prefSize.width += container.getInsets().getWidth();
		prefSize.height += container.getInsets().getHeight();
		prefSize.union(getBorderPreferredSize(container));
		
		return prefSize;
	}
	
	/**
	 * Provides the given child's preferred size.
	 * 
	 * @param child the Figure whose preferred size needs to be calculated
	 * @param wHint the width hint
	 * @param hHint the height hint
	 * @return the child's preferred size
	 */
	protected Dimension getChildSize(IFigure child, int wHint, int hHint) {
		return child.getPreferredSize(wHint, hHint);
	}
	
	
	/**
	 * Returns the alignment used for an entire row/column.
	 * <P>
	 * Possible values are :
	 * <ul>
	 *   <li>{@link #ALIGN_CENTER}
	 * 	 <li>{@link #ALIGN_LEFTTOP}
	 * 	 <li>{@link #ALIGN_RIGHTBOTTOM}
	 * </ul>
	 *
	 * @return the major alignment
	 * @since 2.0
	 */
	public int getMajorAlignment() {
		return majorAlignment;
	}
	
	/**
	 * Returns the spacing in pixels to be used between children in the direction parallel to
	 * the layout's orientation.
	 * @return the major spacing
	 */
	public int getMajorSpacing() {
		return majorSpacing;
	}
	
	/** 
	 * Returns the alignment used for children within a row/column.
	 * <P>
	 * Possible values are :
	 * <ul>
	 *   <li>{@link #ALIGN_CENTER}
	 * 	 <li>{@link #ALIGN_LEFTTOP}
	 * 	 <li>{@link #ALIGN_RIGHTBOTTOM}
	 * </ul>
	 *
	 * @return the minor alignment
	 * @since 2.0
	 */
	public int getMinorAlignment() {
		return minorAlignment;
	}
	
	/**
	 * Returns the spacing to be used between children within a row/column.
	 * @return the minor spacing
	 */
	public int getMinorSpacing() {
		return minorSpacing;
	}
	
	/**
	 * Initializes the state of row data, which is internal to the layout process. 
	 */
	private void initRow() {
		data.rowX = 0;
		data.rowHeight = 0;
		data.rowWidth = 0;
		data.rowCount = 0;
	}
	
	/**
	 * Initializes state data for laying out children, based on the Figure given as input.
	 *
	 * @param parent the parent figure
	 * @since 2.0 
	 */
	
	private void getNumbers(IFigure parent)
	{
		_numComponents=0;
		_numPorts=0;
		_numRoles=0;
		_numConnectors=0;
		_componentIconWidth=30;
		_componentTextWidth=60;
		_connectorIconWidth=30;
		_connectorTextWidth=60;
		_portIconWidth=30;
		_portTextWidth=60;
		_roleIconWidth=30;
		_roleTextWidth=60;
		
		data = new WorkingData();
		
		Iterator iterator = parent.getChildren().iterator();
		int i=0;
		int iNumColumns=0;
		int iWidth=0;
		int iComponentY=40;
		int iConnectorY=40;
		int iPortY=40;
		int iRoleY=40;
		
		_legendHeight=0;
		_legendWidth=0;
		
		//Calculate the hints to be passed to children
		
		
		while (iterator.hasNext()) {
			IFigure f = (IFigure)iterator.next();
			if (f instanceof ColumnSeparatorFigure) {
				
			}
			else if (f instanceof LegendItemIcon) {
				LegendItemIcon iconFigure = (LegendItemIcon )f;
				CategoryT type = iconFigure.getType ();
				
				if(type == CategoryT.ACME_COMPONENT_TYPE)
				{
					_numComponents++;
					if(iconFigure.getBounds()!=null)
					{	
						iWidth=iconFigure.getBounds ().width;;
					}
					if(iWidth>_componentIconWidth){_componentIconWidth=iWidth;}
					
					_componentTextWidth = getTextWidth (AcmeLegend.COMPONENT_COLUMN, _componentIconWidth, _componentTextWidth);
//					descFigure.getModel ().setWidth(_componentTextWidth);
				}
				if (type == CategoryT.ACME_CONNECTOR_TYPE)
				{
					_numConnectors++;
					if(iconFigure.getBounds()!=null)
					{	
						iWidth=iconFigure.getBounds ().width;
					}

					//System.out.println("component icon width="+iWidth);
					if(iWidth>_connectorIconWidth){_connectorIconWidth=iWidth;}
					_connectorTextWidth = getTextWidth (AcmeLegend.CONNECTOR_COLUMN, _connectorIconWidth, _connectorTextWidth);
//					descFigure.getModel().setWidth (_connectorTextWidth);
				}
				if(type == CategoryT.ACME_PORT_TYPE)
				{
					_numPorts++;
					if(iconFigure.getBounds()!=null)
					{	
						iWidth=iconFigure.getBounds ().width;
					}
					//System.out.println("component icon width="+iWidth);
					if(iWidth>_componentIconWidth){_portIconWidth=iWidth;}
					_portTextWidth = getTextWidth (AcmeLegend.PORT_COLUMN, _portIconWidth, _portTextWidth);
//					descFigure.getModel ().setWidth (_portTextWidth);
				}
				if(type == CategoryT.ACME_ROLE_TYPE)
				{
					_numRoles++;
					if(iconFigure.getBounds()!=null)
					{	
						iWidth=iconFigure.getBounds ().width;
					}
					//System.out.println("component icon width="+iWidth);
					if(iWidth>_componentIconWidth){_roleIconWidth=iWidth;}
					_roleTextWidth = getTextWidth (AcmeLegend.ROLE_COLUMN, _roleIconWidth, _roleTextWidth);
//					descFigure.getModel ().setWidth (_roleTextWidth);
				}
				i++;
			}
		}
		if(_numComponents>0) {_componentColumn=iNumColumns;iNumColumns++;}
		else {_componentIconWidth = 0; _componentTextWidth = 0;}
		if(_numConnectors>0) {_connectorColumn=iNumColumns;iNumColumns++;}
		else {_connectorIconWidth = 0; _connectorTextWidth = 0;}
		if(_numPorts>0) {_portColumn=iNumColumns;iNumColumns++;}
		else {_portIconWidth = 0; _portTextWidth = 0;}
		if(_numRoles>0) {_roleColumn=iNumColumns;iNumColumns++;}
		else {
			_roleIconWidth = 0; _roleTextWidth = 0;
		}
		if(iComponentY>_legendHeight){_legendHeight=iComponentY;}
		if(iConnectorY>_legendHeight){_legendHeight=iConnectorY;}
		if(iPortY>_legendHeight){_legendHeight=iPortY;}
		if(iRoleY>_legendHeight){_legendHeight=iRoleY;}
		
	}
	
	
	
	
	private int getTextWidth(int column, int iconWith, int initialTextWidth) {
		int ctw = initialTextWidth;
		int columnWidth = _legend.getColumnWidth(column);
		if (columnWidth != 0) {
			if (columnWidth - iconWith < 70) {
				ctw = 40;
			}
			else {
				ctw = columnWidth - (iconWith + 30);
			}
		}
		return ctw;
	}

	private void initVariables(IFigure parent) {
		data.row = new IFigure[parent.getChildren().size()];
		data.bounds = new Rectangle[data.row.length];
		data.maxWidth = data.area.width;
		_componentColumn=0;
		_connectorColumn=1;
		_portColumn=2;
		_roleColumn=3;
		getNumbers(parent);
	}
	
	/**
	 * Returns <code>true</code> if the orientation of the layout is horizontal.
	 *
	 * @return <code>true</code> if the orientation of the layout is horizontal
	 * @since 2.0
	 */
	public boolean isHorizontal() {
		return horizontal;
	}
	
	/**
	 * @see org.eclipse.draw2d.AbstractHintLayout#isSensitiveHorizontally(IFigure)
	 */
	protected boolean isSensitiveHorizontally(IFigure parent) {
		return isHorizontal();
	}
	
	/**
	 * @see org.eclipse.draw2d.AbstractHintLayout#isSensitiveVertically(IFigure)
	 */
	protected boolean isSensitiveVertically(IFigure parent) {
		return !isHorizontal();
	}
	
	/**
	 * @see org.eclipse.draw2d.LayoutManager#layout(IFigure)
	 */
	public void layout(IFigure parent) {
//		AcmeLegend legend = null;
//		if (parent instanceof LegendFigure) {
//			legend = (AcmeLegend )((LegendFigure )parent).getModel ();
//		}
		
		data = new WorkingData();
		Rectangle relativeArea = parent.getClientArea();
		data.area = transposer.t(relativeArea);
		
		Iterator iterator = parent.getChildren().iterator();
		
		initVariables(parent);
		initRow();
		int i = 0; 
		int iPorts, iConnectors, iComponents, iRoles;
		int iComponentIconX=20;
		int iComponentTextX=_componentIconWidth+iComponentIconX+((_componentIconWidth == 0)?0:DEFAULT_COLUMN_SPACING);
		int iConnectorIconX=_componentTextWidth+iComponentTextX+((_componentTextWidth==0)?0:DEFAULT_COLUMN_SPACING);
		int iConnectorTextX=_connectorIconWidth+iConnectorIconX+((_connectorIconWidth==0)?0:DEFAULT_COLUMN_SPACING);
		int iPortIconX=_connectorTextWidth+iConnectorTextX+((_connectorTextWidth==0)?0:DEFAULT_COLUMN_SPACING);
		int iPortTextX=_portIconWidth+iPortIconX+((_portIconWidth==0)?0:DEFAULT_COLUMN_SPACING);
		int iRoleIconX=_portTextWidth+iPortTextX+((_portTextWidth==0)?0:DEFAULT_COLUMN_SPACING);
		int iRoleTextX=_roleIconWidth+iRoleIconX+((_roleIconWidth==0)?0:DEFAULT_COLUMN_SPACING);
		int iLegendWidth=_roleTextWidth+iRoleTextX+((_roleTextWidth==0)?0:DEFAULT_COLUMN_SPACING);
		
		int iComponentY=40;
		int iConnectorY=40;
		int iPortY=40;
		int iRoleY=40;
		
		
		
		iPorts=0;
		iConnectors=0;
		iRoles=0;
		iComponents=0;	
		
		while (iterator.hasNext()) {
			IFigure f = (IFigure)iterator.next();
			if (f instanceof ColumnSeparatorFigure) {
				ColumnSeparatorFigure sepFig = (ColumnSeparatorFigure) f;
				Rectangle bounds = sepFig.getBounds ().getCopy ();
				ColumnSeparator model = sepFig.getModel ();
				int column = model.getColumn();
				switch (column) {
					case AcmeLegend.COMPONENT_COLUMN:
						bounds.x = iConnectorIconX - 5;
						if (_legend != null) _legend.setColumnWidth (column, getComponentWidth ());
						break;
					case AcmeLegend.CONNECTOR_COLUMN:
						bounds.x = iPortIconX - 5;
						if (_legend != null) _legend.setColumnWidth (column, getConnectorWidth ());
						break;
					case AcmeLegend.PORT_COLUMN:
						bounds.x = iRoleIconX - 5;
						if (_legend != null) _legend.setColumnWidth (column, getPortWidth ());
						break;
					case AcmeLegend.ROLE_COLUMN:
						bounds.x = iRoleTextX + getRoleTextWidth() + 5;
						if (_legend != null) _legend.setColumnWidth (column, getRoleWidth ());
						break;
				}
//				setBoundsOfChild (parent, sepFig, bounds);
			}
			else if (f instanceof LegendItemIcon) {
				LegendItemIcon iconFigure = (LegendItemIcon) f;
				LegendItemLabel descFigure = (LegendItemLabel)iterator.next();
				
				if(iconFigure.getType() == CategoryT.ACME_COMPONENT_TYPE)
				{			
					iComponents++;
					descFigure.getModel().setWidth(_componentTextWidth);
					iComponentY = calculateRowNumbers (parent, iComponentIconX, iComponentTextX, iComponentY, iconFigure, descFigure);
				}
				if(iconFigure.getType() == CategoryT.ACME_CONNECTOR_TYPE)
				{
					iConnectors++;
					descFigure.getModel ().setWidth (_connectorTextWidth);
					iConnectorY = calculateRowNumbers (parent, iConnectorIconX, iConnectorTextX, iConnectorY, iconFigure, descFigure);
				}
				if(iconFigure.getType() == CategoryT.ACME_PORT_TYPE)
				{
					iPorts++;
					descFigure.getModel ().setWidth (_portTextWidth);
					iPortY = calculateRowNumbers (parent, iPortIconX, iPortTextX, iPortY, iconFigure, descFigure);
				}
				if(iconFigure.getType() == CategoryT.ACME_ROLE_TYPE)
				{
					iRoles++;
					descFigure.getModel ().setWidth (_roleTextWidth);
					iRoleY = calculateRowNumbers (parent, iRoleIconX, iRoleTextX, iRoleY, iconFigure, descFigure);
				}
				_numComponents=iComponents;
				_numConnectors=iConnectors;
				_numPorts=iPorts;
				_numRoles=iRoles;
				_legendWidth=iLegendWidth+50;
				_componentX=iComponentIconX;
				_connectorX=iConnectorIconX;
				_portX=iPortIconX;
				_roleX=iRoleIconX;
				if(iComponentY>_legendHeight){_legendHeight=iComponentY;}
				if(iConnectorY>_legendHeight){_legendHeight=iConnectorY;}
				if(iPortY>_legendHeight){_legendHeight=iPortY;}
				if(iRoleY>_legendHeight){_legendHeight=iRoleY;}
				//_legend_height+=50;
				i++;
				
			}

		}

		data = null;
				
		if (parent.getParent ().getParent() instanceof LegendFigure) {
			LegendFigure lf = (LegendFigure) parent.getParent ().getParent();
			lf.calculateSize ();
		}
		
	}
	
	
	public int getRoleWidth() {
		int width = getRoleIconWidth () + getRoleTextWidth ();
		if (width == 0) return 0;
		return width + 30;
	}

	public int getPortWidth() {
		int width = getPortIconWidth () + getPortTextWidth ();
		if (width == 0) return 0;
		return width + 30;
	}

	public int getConnectorWidth() {
		int width = getConnectorIconWidth () + getConnectorTextWidth ();
		if (width == 0) return 0;
		return width + 30;
	}

	public int getComponentWidth() {
		
		int width = getComponentIconWidth() + getComponentTextWidth ();
		if (width == 0) return 0;
		return width + 30;
	}

	private int calculateRowNumbers(IFigure parent, int iconX, int descriptionX, int yStart,  LegendItemIcon iconFigure, LegendItemLabel descFigure) {
		Rectangle rf = iconFigure.getBounds();
		rf.x=iconX;
		rf.y=yStart;
		setBoundsOfChild(parent, iconFigure, rf);
		Rectangle rg = descFigure.getBounds();
		rg.x=descriptionX;
		rg.y=yStart;
		descFigure.invalidate ();
		int iTempY = descFigure.getPreferredSize(rg.width, -1).height;
		rg.height = iTempY;
		if(iconFigure.getBounds()!=null)
		{	
			if (iconFigure.getBounds ().height>iTempY)
			{
				iTempY=iconFigure.getBounds ().height;
			}
		}
		yStart+=iTempY+10;
		setBoundsOfChild(parent, descFigure, rg);
		descFigure.invalidate();
		return yStart;
	}

	/**
	 * Layouts one row of components. This is done based on the layout's orientation, minor 
	 * alignment and major alignment.
	 *
	 * @param parent the parent figure
	 * @since 2.0
	 */
	protected void layoutRow(IFigure parent) {
		int majorAdjustment = 0;
		int minorAdjustment = 0;
		int correctMajorAlignment = majorAlignment;
		int correctMinorAlignment = minorAlignment;
		
		majorAdjustment = data.area.width - data.rowWidth + getMinorSpacing();
		
		switch (correctMajorAlignment) {
			case ALIGN_LEFTTOP: 
				majorAdjustment = 0;
				break;
			case ALIGN_CENTER:
				majorAdjustment /= 2;
				break;
			case ALIGN_RIGHTBOTTOM:
				break;
		}
		
		for (int j = 0; j < data.rowCount; j++) {
			if (fill) {
				data.bounds[j].height = data.rowHeight;	
			} else {
				minorAdjustment = data.rowHeight - data.bounds[j].height;
				switch (correctMinorAlignment) {
					case ALIGN_LEFTTOP: 
						minorAdjustment = 0;
						break;
					case ALIGN_CENTER:
						minorAdjustment /= 2;
						break;
					case ALIGN_RIGHTBOTTOM:
						break;
				}
				data.bounds[j].y += minorAdjustment+100;
			}
			//data.bounds[j].x += majorAdjustment;
			if (j==0) data.bounds[j].x=0;
			if (j==1) data.bounds[j].x=40;
			if (j==2) data.bounds[j].x=190;
			if (j==3) data.bounds[j].x=230;
			
			if (j>3) data.bounds[j].x = j*100;
			
			setBoundsOfChild(parent, data.row[j], transposer.t(data.bounds[j]));
		}
		data.rowY += getMajorSpacing() + data.rowHeight ;
		initRow();
	}
	
	/**
	 * Sets the given bounds for the child figure input.
	 *
	 * @param parent the parent figure
	 * @param child the child figure
	 * @param bounds the size of the child to be set
	 * @since 2.0
	 */
	protected void setBoundsOfChild(IFigure parent, IFigure child, Rectangle bounds) {
		parent.getClientArea(Rectangle.SINGLETON);
		bounds.translate(Rectangle.SINGLETON.x, Rectangle.SINGLETON.y);
		child.setBounds(bounds);
	}
	
	/**
	 * Sets flag based on layout orientation. If in horizontal orientation, all figures will 
	 * have the same height. If in vertical orientation, all figures will have the same width.
	 *
	 * @param value fill state desired
	 * @since 2.0
	 */
	public void setStretchMinorAxis(boolean value) {
		fill = value;
	}
	
	/**
	 * Sets the orientation of the layout. 
	 *
	 * @param flag <code>true</code> if this layout should be horizontal
	 * @since 2.0
	 */
	public void setHorizontal(boolean flag) {
		if (horizontal == flag) return;
		invalidate();
		horizontal = flag;
		transposer.setEnabled(!horizontal);
	}
	
	/**
	 * Sets the alignment for an entire row/column within the parent figure.
	 * <P>
	 * Possible values are :
	 * <ul>
	 *   <li>{@link #ALIGN_CENTER}
	 * 	 <li>{@link #ALIGN_LEFTTOP}
	 * 	 <li>{@link #ALIGN_RIGHTBOTTOM}
	 * </ul>
	 *
	 * @param align the major alignment
	 * @since 2.0
	 */
	public void setMajorAlignment(int align) {
		majorAlignment = align;
	}
	
	/**
	 * Sets the spacing in pixels to be used between children in the direction parallel to the
	 * layout's orientation.
	 *
	 * @param n the major spacing
	 * @since 2.0
	 */
	public void setMajorSpacing(int n) {
		majorSpacing = n;
	}
	
	/**
	 * Sets the alignment to be used within a row/column.
	 * <P>
	 * Possible values are :
	 * <ul>
	 *   <li>{@link #ALIGN_CENTER}
	 * 	 <li>{@link #ALIGN_LEFTTOP}
	 * 	 <li>{@link #ALIGN_RIGHTBOTTOM}
	 * </ul>
	 *
	 * @param align the minor alignment
	 * @since 2.0
	 */
	public void setMinorAlignment(int align) {
		minorAlignment = align;
	}
	
	/**
	 * Sets the spacing to be used between children within a row/column.
	 *
	 * @param n the minor spacing
	 * @since 2.0
	 */
	public void setMinorSpacing(int n) {
		minorSpacing = n;
	}
	
	/**
	 * @return Returns the componentColumn.
	 */
	public int getComponentColumn() {
		return _componentColumn;
	}
	/**
	 * @return Returns the componentIconWidth.
	 */
	public int getComponentIconWidth() {
		return _componentIconWidth;
	}
	/**
	 * @return Returns the componentTextWidth.
	 */
	public int getComponentTextWidth() {
		return _componentTextWidth;
	}
	/**
	 * @return Returns the componentX.
	 */
	public int getComponentX() {
		return _componentX;
	}
	/**
	 * @return Returns the connectorColumn.
	 */
	public int getConnectorColumn() {
		return _connectorColumn;
	}
	/**
	 * @return Returns the connectorIconWidth.
	 */
	public int getConnectorIconWidth() {
		return _connectorIconWidth;
	}
	/**
	 * @return Returns the connectorTextWidth.
	 */
	public int getConnectorTextWidth() {
		return _connectorTextWidth;
	}
	/**
	 * @return Returns the connectorX.
	 */
	public int getConnectorX() {
		return _connectorX;
	}
	/**
	 * @return Returns the legendHeight.
	 */
	public int getLegendHeight() {
		return _legendHeight;
	}
	/**
	 * @return Returns the legendWidth.
	 */
	public int getLegendWidth() {
		return _legendWidth;
	}
	/**
	 * @return Returns the numComponents.
	 */
	public int getNumComponents() {
		return _numComponents;
	}
	/**
	 * @return Returns the numConnectors.
	 */
	public int getNumConnectors() {
		return _numConnectors;
	}
	/**
	 * @return Returns the numPorts.
	 */
	public int getNumPorts() {
		return _numPorts;
	}
	/**
	 * @return Returns the numRoles.
	 */
	public int getNumRoles() {
		return _numRoles;
	}
	/**
	 * @return Returns the portColumn.
	 */
	public int getPortColumn() {
		return _portColumn;
	}
	/**
	 * @return Returns the portIconWidth.
	 */
	public int getPortIconWidth() {
		return _portIconWidth;
	}
	/**
	 * @return Returns the portTextWidth.
	 */
	public int getPortTextWidth() {
		return _portTextWidth;
	}
	/**
	 * @return Returns the portX.
	 */
	public int getPortX() {
		return _portX;
	}
	/**
	 * @return Returns the roleColumn.
	 */
	public int getRoleColumn() {
		return _roleColumn;
	}
	/**
	 * @return Returns the roleIconWidth.
	 */
	public int getRoleIconWidth() {
		return _roleIconWidth;
	}
	/**
	 * @return Returns the roleTextWidth.
	 */
	public int getRoleTextWidth() {
		return _roleTextWidth;
	}
	/**
	 * @return Returns the roleX.
	 */
	public int getRoleX() {
		return _roleX;
	}

	/**
	 * @param legend
	 */
	public void setLegend(AcmeLegend legend) {
		_legend = legend;
	}
}