package AcmeStudio.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.IPartListener;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.views.navigator.IResourceNavigator;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.resource.ASDesign;
import AcmeStudio.ui.editors.MultiPageAcmeEditor;
import acmeLib.elements.AcmeDesign;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * Maintains the context of the workbench, i.e., current project, editor, etc.
 * This is a singleton class.
 */
public class ASContext {

    /** The singleton instance of context */
    private static ASContext _instance = null;

    /** The currently active project */
    private IProject _activeProject = null;
    /** The currently active design */
    private ASDesign _activeDesign = null;
    /** The Map of managed designs */
    private Map _managedDesigns = null;
    /** The currently active editor */
    private MultiPageAcmeEditor _activeEditor = null;
    /** The previously active project */
    private IProject _prevProject = null;
    /** The previously active design */
    private ASDesign _prevDesign = null;
    /** The previously active editor */
    private MultiPageAcmeEditor _prevEditor = null;

    /** List of listeners */
    private Collection _listeners = null;

    /** Selection listener */
    private ISelectionListener _selectionListener = new ISelectionListener () {
        /**
         * @see org.eclipse.ui.ISelectionListener#selectionChanged(IWorkbenchPart, ISelection)
         */
        public void selectionChanged(IWorkbenchPart part, ISelection selection) {
            if (part instanceof IResourceNavigator) {  // then get the selection
                if (selection instanceof IStructuredSelection && selection != null) {
                    IStructuredSelection iss = (IStructuredSelection) selection;
                    Object elem = iss.getFirstElement();
                    if (elem instanceof IResource && elem != null) {
                        setActiveProject(((IResource) elem).getProject());
                    }
                }
            }
        }
    };

    /** Part listener */
    private IPartListener _partListener = new IPartListener () {
        /**
         * @see org.eclipse.ui.IPartListener#partActivated(IWorkbenchPart)
         */
        public void partActivated(IWorkbenchPart part) {
            if (part != null && part instanceof IEditorPart) {
                editorChanged((IEditorPart) part);
            }
        }
    
        /**
         * @see org.eclipse.ui.IPartListener#partBroughtToTop(IWorkbenchPart)
         */
        public void partBroughtToTop(IWorkbenchPart part) {
            if (part != null && part instanceof IEditorPart) {
                editorChanged((IEditorPart) part);
            }
        }
    
        /**
         * @see org.eclipse.ui.IPartListener#partClosed(IWorkbenchPart)
         */
        public void partClosed(IWorkbenchPart part) {
            // we don't change context on a closed part
        }
    
        /**
         * @see org.eclipse.ui.IPartListener#partDeactivated(IWorkbenchPart)
         */
        public void partDeactivated(IWorkbenchPart part) {
            // we don't change context on a deactivated part
        }
    
        /**
         * @see org.eclipse.ui.IPartListener#partOpened(IWorkbenchPart)
         */
        public void partOpened(IWorkbenchPart part) {
            if (part != null && part instanceof IEditorPart) {
                editorChanged((IEditorPart) part);
            }
        }
    };

    /**
     * Constructor for ASContext.
     */
    protected ASContext () {
        super();
        
        _managedDesigns = new Hashtable();
        _listeners = new ArrayList();

        IWorkbenchWindow window = AcmeStudioPlugin.getActiveWindow();
        window.getSelectionService().addSelectionListener(_selectionListener);
        window.getPartService().addPartListener(_partListener);
        // cause first-time selection change based on current Navigator selection
        // this would properly set the ASContext at the onset.
        // Bug Fix #371
        final IResourceNavigator nav = AcmeStudioPlugin.findNavigator();
        if (nav != null) {
            Display.getCurrent().asyncExec(new Runnable() {
                public void run() {
                    _selectionListener.selectionChanged(nav, nav.getViewer().getSelection());
                }
            });
        }
    }

    /**
     * Returns the singleton object, instantiating one if null.
     * @return ASContext  Acme context object
     */
    public static ASContext current () {
        initialize();
        return _instance;
    }
    
    /**
     * Explicitly initialize the singleton ASContext instance, causing
     * subscription to selection and part events.
     */
    public static void initialize () {
        if (_instance == null) {
            _instance = new ASContext();
        }
    }

    /**
     * Adds listener to the list of context listeners
     * @param listener  the interested listener
     */
    public void addContextListener (IASContextListener listener) {
        _listeners.add(listener);
    }

    /**
     * Removes listener from the list of context listeners
     * @param listener  the uninterested listener
     */
    public void removeContextListener (IASContextListener listener) {
        _listeners.remove(listener);
    }

    /**
     * Returns the currently active ASDesign.  NOTE: the original return type
     * was AcmeDesign.
     * 
     * @return ASDesign  the active ASDesign
     */
    public ASDesign getActiveDesign() {
        return _activeDesign;
    }

    /**
     * Sets the currently active AcmeDesign.
     * @param design  the new active AcmeDesign
     * @deprecated use #setActiveDesign(ASDesign)
     */
    public void setActiveDesign (AcmeDesign design) {
        setActiveDesign((ASDesign )ASModel.current().getASResource(design));
    }

    /**
     * Sets the currently active ASDesign.
     * @param design  the new active ASDesign
     */
    public void setActiveDesign (ASDesign design) {
        _prevDesign = _activeDesign;
        _activeDesign = design;
    }

    /**
     * Returns the active editor.
     * @return MultiPageAcmeEditor  the active editor
     */
    public MultiPageAcmeEditor getActiveEditor() {
        return _activeEditor;
    }

    /**
     * Sets the currently active editor.
     * @param activeEditor  The new active editor to set
     */
    public void setActiveEditor(IEditorPart activeEditor) {
        if (activeEditor instanceof MultiPageAcmeEditor) {
            _prevEditor = _activeEditor;
            _activeEditor = (MultiPageAcmeEditor) activeEditor;
        }
        // no need to notify listeners, is there??
        // TODO Why setActiveEditor doesn't notify listeners?
    }

    /**
     * Adds a managed AcmeDesign to the list of managed designs.
     * TODO Why is addDesign here?
     * @param name  the name of the AcmeDesign
     * @param design  the AcmeDesign to add
     */
    public void addDesign (String name, AcmeDesign design) {
        _managedDesigns.put (name, design);
    }

    /**
     * Returns the managed AcmeDesign of the given name.
     * TODO Why is getDesign here?
     * @param name  the name of the AcmeDesign to fetch
     * @return AcmeDesign  the managed AcmeDesign
     */
    public AcmeDesign getDesign(String name) {
        return (AcmeDesign) _managedDesigns.get(name);
    }

    /**
     * Returns the currently active/selected project.
     * @return IProject  current project
     */    
    public IProject getActiveProject () {
        return _activeProject;
    }

    /**
     * Sets the currently active/selected project.
     * @param proj  the new current project
     */    
    public void setActiveProject (IProject proj) {
        _prevProject = _activeProject;
        _activeProject = proj;
        fireContextChanged();
    }

    /**
     * General method for most editor part changed event.
     * @param part  the IWorkbenchPart which changed
     */    
    private void editorChanged (IEditorPart editor) {
        setActiveEditor(editor);
        IEditorInput input = editor.getEditorInput();
        if (input instanceof IFileEditorInput) {
            setActiveProject(((IFileEditorInput) input).getFile().getProject());
        }
    }

    /**
     * Notifies all the listeners of context change.
     */
    private void fireContextChanged () {
        // at this point, we can safely assume that the context changes have
        // already been effected, so that the creation of the context event
        // will happen properly (it uses the current context during init.)
        ASContextEvent event = new ASContextEvent(this, _prevProject, _prevDesign, _prevEditor);
 
        Iterator iter = _listeners.iterator();
        while (iter.hasNext()) {
            IASContextListener listener = (IASContextListener) iter.next();
            listener.contextChanged (event);
        }
    }

}
