package AcmeStudio.core;

import java.io.IOException;
import java.util.*;

import org.eclipse.core.resources.IFile;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.resource.ASDesign;
import AcmeStudio.core.resource.ASFamily;
import AcmeStudio.core.resource.ASVisualStyle;
import AcmeStudio.core.resource.IASResource;
import AcmeStudio.core.resource.IASResourceChangeEvent;
import AcmeStudio.core.resource.IASResourceChangeListener;
import AcmeStudio.ui.editors.diagram.styles.VisualStyle;
import AcmeStudio.util.DesignEvaluationCache;
import AcmeStudio.util.Helper;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeFamily;
import acmeLib.elements.AcmeOpenedDesignRecord;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * This is a Singleton class allowing other objects in AcmeStudio to obtain
 * information about the internal model elements of AcmeStudio, including
 * designs, families, and design evaluation cache.<br>
 * <br> 
 * The list of AcmeFamilies is cached, but the family list is marked as dirty
 * whenever a new mapping is added/removed involves a family resource.
 */
public class ASModel {

///////////////////////
//  Class member data
///////////////////////
    
    /** Map of current active window to ASModel instances */
    private static Map _instances = new HashMap();

    private static DesignEvaluationCache _evalCache = new DesignEvaluationCache ();

    /** Registry of available families in the system */
    private ASRegistry _famReg = null;
    /** Map of local file (canonical path string) to IASResource */
    private Map _fileToASRes = null;
    /** Map of AcmeDesign to local file (canonical path string) */
    private Map _designToFile = null;
    /** List of AcmeFamilies in the current AcmeStudio space */
    private List _acmeFamilies = null;
    /** Flag indicating whether list of families is dirty. */
    private boolean _famListDirty = true;  // assume initially dirty
    /** Listener to help determine family list dirtiness. */
    private IASResourceChangeListener _listener = null;

///////////////////////
//  Class constructor
///////////////////////
    
    /**
     * ASModel not intended for object instantiation.
     */
    protected ASModel () {
        _fileToASRes = new HashMap();
        _designToFile = new HashMap();
        _acmeFamilies = new ArrayList();
        _listener = new IASResourceChangeListener () {
            /* (non-Javadoc)
             * @see AcmeStudio.core.resource.IASResourceChangeListener#resourceChanged(AcmeStudio.core.resource.IASResourceChangeEvent)
             */
            public void resourceChanged(IASResourceChangeEvent evt) {
                if ((evt.getType() & IASResourceChangeEvent.SELF_CHANGED) > 0) {
                    checkFamilyAffected(evt.getSource());
                    if (evt.getSource() instanceof ASDesign) {
                        // find ASDesign in map
                        Iterator i = _fileToASRes.entrySet().iterator();
                        while (i.hasNext()) {
                            Map.Entry entry = (Map.Entry )i.next();
                            if (entry.getValue() == evt.getSource()) {
                                // found path, create a new mapping  of AcmeDesign to canonical path
                                _designToFile.put(((ASDesign )evt.getSource()).getDesign(), entry.getKey());
                            }
                        }
                    }
                }
            }
        };
    }

    /* (non-Javadoc)
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        if (_fileToASRes != null) {
            _fileToASRes.clear();
            _fileToASRes = null;
        }
        if (_designToFile != null) {
            _designToFile.clear();
            _designToFile = null;
        }
        if (_acmeFamilies != null) {
            _acmeFamilies.clear();
            _acmeFamilies = null;
        }
    }

//////////////////////////
//  Class static methods
//////////////////////////
    
    /**
     * Returns the ASModel instance corresponding to the current active window.
     * @return ASModel  the appropriate ASModel instance.
     */
    public static ASModel current () {
        ASModel model = (ASModel) _instances.get(AcmeStudioPlugin.getActiveWindow());
        if (model == null) {
            model = new ASModel();
            _instances.put(AcmeStudioPlugin.getActiveWindow(), model);
        }
        return model;
    }

    /**
     * Returns the current family registry.  A shorthand for
     * ASModel.current().getFamilyRegistry();
     * @return ASRegistry  the current family registry
     */    
    public static ASRegistry familyRegistry () {
        return current().getFamilyRegistry();
    }

//////////////////////////
//  Class member methods
//////////////////////////
    
    /**
     * Returns the family registry of this ASModel.
     * @return ASRegistry  the family registry
     */    
    public ASRegistry getFamilyRegistry () {
        if (_famReg == null) {
            _famReg = new ASRegistry(ASFamily.nullInstance());
        } 
        return _famReg;
    }

    /**
     * Stores a file canonical path string to IASResource mapping.
     * @param canonPath  the canonical path string to use as key
     * @param res  the IASResource to use as value
     */
    public void addFile2ASResource (String canonPath, IASResource res) {
        checkFamilyAffected(res);
        registerChangeListener(res);
        _fileToASRes.put(canonPath, res);
        
        // add a mapping of AcmeDesign to file path
        if (res instanceof ASDesign) {
            _designToFile.put(((ASDesign )res).getDesign(), canonPath);
        }
    }

    /**
     * Returns the IASResource given a file's canonical path string.
     * @param canonPath  the canonical path string to use as key
     * @return IASResource  the IASResource object
     */
    public IASResource getASResource (String canonPath) {
        return (IASResource) _fileToASRes.get(canonPath);
    }

    /**
     * Removes a file's canonical path from the mapping.
     * @param canonPath  the canonical path string to use as key
     * @return Object  the IASResource value object of the removed key
     */
    public Object removeFile2ASResource (String canonPath) {
        if (canonPath == null) return null;

        IASResource res = (IASResource )_fileToASRes.remove(canonPath);
        checkFamilyAffected(res);
        unregisterChangeListener(res);
        return res;
    }

    /**
     * Stores an AcmeDesign to IASResource mapping.
     * @param design  the AcmeDesign to use as key
     * @param res  the IASResource to use as value
     * @deprecated use #addFile2ASResource(String, IASResource)
     */
    public void addDesign2ASResource (AcmeDesign design, IASResource res) {
        // find the AcmeOpenedDesignRecord containing the design, to get file path
        AcmeOpenedDesignRecord rec = AcmeDesign.findDesignRecordByDesign(design);
        if (rec != null) {
            addFile2ASResource(rec.getCanonicalFilename(), res);
        } else if (res.getCanonicalPath() != null) {
            addFile2ASResource(res.getCanonicalPath(), res);
        }
    }

    /**
     * Returns the IASResource given the AcmeDesign.
     * @param design  the AcmeDesign to use as key
     * @return IASResource  the IASResource object
     */
    public IASResource getASResource (AcmeDesign design) {
        IASResource res = null;

        // find the AcmeOpenedDesignRecord containing the design, to get file path
        AcmeOpenedDesignRecord rec = AcmeDesign.findDesignRecordByDesign(design);
        if (rec != null) {
            res = (IASResource )_fileToASRes.get(rec.getCanonicalFilename());
        }
        if (res == null) {  // still null, use the design to file path mapping
            res = (IASResource )_fileToASRes.get(_designToFile.get(design));
        }

        return res;
    }

    /**
     * Removes an AcmeDesign to IASResource mapping.
     * @param design  the AcmeDesign to remove
     * @return Object  the removed IASResource value
     */
    public Object removeDesign2ASResource (AcmeDesign design) {
        Object res = null;
        if (design == null) return res;

        // find the AcmeOpenedDesignRecord containing the design, to get file path
        AcmeOpenedDesignRecord rec = AcmeDesign.findDesignRecordByDesign(design);
        if (rec != null) {
            res = removeFile2ASResource(rec.getCanonicalFilename());
        } else {
            res = removeFile2ASResource((String )_designToFile.get(design));
        }

        return res;
    }

    /**
     * Returns the IASResource given the AcmeDesign.
     * @param design  the AcmeDesign to use as key
     * @return IASResource  the IASResource object
     * @deprecated
     */
    public IASResource getASResourceByDesign (AcmeDesign design) {
        return getASResource(design);
    }

    /**
     * Returns the IASResource given the local IFile.
     * @param file  the IFile to use as key
     * @return IASResource  the IASResource object
     * @deprecated
     */
    public IASResource getASResourceByFile (IFile file) {
        IASResource res = null;
        try {
            res = (IASResource )getASResource(file.getLocation().toFile().getCanonicalPath());
        } catch (IOException e) {
            // ignore, will just return null
        }
        return res;
    }

    public VisualStyle getVisualStyle (AcmeFamily fam) {
        VisualStyle style = null;
        ASVisualStyle asVStyle = getASVisualStyle(fam);
        if (asVStyle != null) {
            style = asVStyle.getVisualStyle();
        }
        return style;
    }

    public ASVisualStyle getASVisualStyle (AcmeFamily fam) { 
    	if (fam == null) return null;   
        ASVisualStyle asVStyle = null;
        IASResource asRes = getASResource(fam.getParentDesign());
        if (asRes != null && asRes instanceof ASDesign) {
            ASDesign asDes = (ASDesign) asRes;
            asVStyle = asDes.getASVisualStyle();
        }
        return asVStyle;
    }

    public List getDesignFamilies () {
        if (_famListDirty) {
            computeDesignFamilies();
        }
        return _acmeFamilies;
    }
    
    public AcmeFamily findLoadedFamily (AcmeFamily family) {
        AcmeFamily foundFam = null;
        Iterator iter = getDesignFamilies().iterator();
        while (iter.hasNext()) {
            AcmeFamily fam = (AcmeFamily) iter.next();
            if (fam.equals(family)) {
                foundFam = fam;
                break;
            }
        }
        return foundFam;
    }

    /**
     * Returns the design evaluation cache held in this ASModel.
     * @return DesignEvaluationCache  the design evaluation cache
     */
    public DesignEvaluationCache getRuleManager () {
        return _evalCache;
    }

    /**
     * Derives the list all AcmeFamilies from stored hash of AcmeDesigns.
     * @return  list of design families
     */
    protected void computeDesignFamilies () {
        List famList = new ArrayList();
        
        Iterator iter = _fileToASRes.values().iterator();
        while (iter.hasNext()) {
            IASResource res = (IASResource )iter.next();
            famList.addAll(Helper.getAllFamilies(res.getDesign()));
        }
        _acmeFamilies.clear();
        _acmeFamilies.addAll(famList);
    }
    
    /**
     * Checks whether the IASResource affects the list of families, so that
     * refresh/recompute of family list may be needed.  This method is called
     * upon addition or removal of a path to IASResource mapping, as well as
     * when a change occurs to an ASResource listened to.
     * 
     * @param res  the ASResource to check if it's AcmeFamily related.
     */
    private void checkFamilyAffected(IASResource res) {
        if (res != null &&
                (res instanceof ASFamily ||
                    (res instanceof ASDesign && res.getDesign() != null
                        && res.getDesign().getFamilies() != null
                        && res.getDesign().getFamilies().getSize() > 0))) {
            // i'm either a family or a design containing family definition
            _famListDirty = true;
            getFamilyRegistry().setRefreshNeeded(true);
        }
    }

    private void registerChangeListener (IASResource res) {
        if (res != null && _listener != null) {
            res.addASResourceChangeListener(_listener);
        }
    }

    private void unregisterChangeListener (IASResource res) {
        if (res != null && _listener != null) {
            res.removeASResourceChangeListener(_listener);
        }
    }

}
