package AcmeStudio.core;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.util.*;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.ui.actions.WorkspaceModifyOperation;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.resource.ASPath;
import AcmeStudio.core.resource.ASPathIterator;
import AcmeStudio.core.resource.IASResource;
import AcmeStudio.util.Debug;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * Provides easy management of a registry of AcmeStudio resources.
 * TODO ASRegistry: How is this different from fam list in ASModel?
 * TODO ASRegistry: Should registry be targeted specifically to family?
 * TODO ASRegistry: Is refresh something that can be delegated to another authority?  
 */
public class ASRegistry implements IASContextListener {
    
    public class String2ListMap extends TreeMap {
        public List get (String str) {
            return (List) super.get(str);
        }
        
        public void putItem (String str, Object item) {
            List list = null;
            if (containsKey(str)) {
                list = get(str); 
            } else {
                list = new ArrayList();
            }
            list.add(item);
            Collections.sort(list);
            put(str, list);
        }
    }
    
    public class Object2MapMap extends HashMap {
        public SortedMap getMap (Object obj) {
            return (SortedMap) super.get(obj);
        }
    }
    
    /** Number of work units required to update registry */
    protected static int WORK_UNITS = 20;
    
    /** Registry map for global-level paths */
    protected String2ListMap _globalASResourceMap = null;
    /** Registry map for project-level paths */
    protected Object2MapMap _proj2NameMap = null;
    /** Flag indicating whether a refresh should be done next time lazy refresh is called */
    protected boolean _needRefresh = true;
    
    /** Null instance representing the ASResource type held in registry map */
    protected IASResource _nullASRes = null;
    
    /**
     * Default constructor for ASRegistry.  Requires a call to load(IASResource)
     * to load the registry.
     */
    public ASRegistry() {
        super();
        
        _proj2NameMap = new Object2MapMap();
        // TODO ASRegistry() make sure that adding context listener here is OK.
        ASContext.current().addContextListener(this);
    }
    
    /**
     * Constructor for ASRegistry.  Loads the registry with the supplied type.
     * @param nullRes  the null IASResource to give us a type.
     */
    public ASRegistry(IASResource nullRes) {
        this();
        
        load(nullRes);
    }

    /* (non-Javadoc)
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        ASContext.current().removeContextListener(this);
        super.finalize();
    }

    /**
     * When the project changes, a lazy refresh is necessary.
     * When new families are added to the global paths, a lazy refresh is also
     * necessary, but that CAN'T BE DETECTED at this time.
     * @see AcmeStudio.core.IASContextListener#contextChanged(ASContext)
     */
    public void contextChanged(ASContextEvent event) {
        if (event.hasProjectChanged()) {
            setRefreshNeeded(true);
        }
    }
    
    /**
     * Get the registry organized by path, slightly costly operation.
     * @return SortedMap  the scoped registry
     */
    public SortedMap getByPath () {
        lazyRefresh();

        SortedMap rMap = new TreeMap();
        // add in by paths from global map
        computeMapByASPath(rMap, _globalASResourceMap);
        // add in current project map
        IProject proj = ASContext.current().getActiveProject();
        SortedMap map = _proj2NameMap.getMap(proj);
        if (map != null) {
            computeMapByASPath(rMap, map);
        }
        return rMap;
    }
    
    private void computeMapByASPath (SortedMap retMap, SortedMap srcMap) {
        for (Iterator i = srcMap.values().iterator(); i.hasNext();) {
            List list = (List) i.next();
            for (Iterator j = list.iterator(); j.hasNext();) {
                IASResource asRes = (IASResource) j.next();
                ASPath asPath = asRes.getASPath();
                List resList = null;
                if (retMap.containsKey(asPath)) {
                    resList = (List) retMap.get(asPath);
                } else {
                    resList = new ArrayList();
                    retMap.put(asPath, resList);
                }
                resList.add(asRes);
            }
        }
    }
    
    /**
     * Get the ASResource registry organized by name.
     * @return SortedMap
     */
    public SortedMap getByName () {
        lazyRefresh();

        SortedMap rMap = new TreeMap();
        // add in the global map
        computeMapByName(rMap, _globalASResourceMap);
        // add in current project map
        IProject proj = ASContext.current().getActiveProject();
        SortedMap map = _proj2NameMap.getMap(proj);
        if (map != null) {
            computeMapByName(rMap, map);
        }
        return rMap;
    }
    
    private void computeMapByName (SortedMap retMap, SortedMap srcMap) {
        for (Iterator i = srcMap.entrySet().iterator(); i.hasNext();) {
            Map.Entry entry = (Map.Entry) i.next();
            List resList = null;
            if (retMap.containsKey(entry.getKey())) {
                resList = (List) retMap.get(entry.getKey());
            } else {
                resList = new ArrayList();
                retMap.put(entry.getKey(), resList);
            }
            resList.addAll((List) entry.getValue());
        }
    }
    
    /**
     * An experimental method to allow addition of an IASResource instance into
     * the registry, using current project if path is project-level.
     * @param name  the string of name to use as key
     * @param asRes  the IASResource to map to the path
     */
    public void add (String name, IASResource asRes) {
        if (asRes.getASPath().getPathLevel() == ASPath.GLOBAL_LEVEL) {
            _globalASResourceMap.putItem(name, asRes);
        } else {
            IProject proj = ASContext.current().getActiveProject();
            ((String2ListMap) _proj2NameMap.get(proj)).putItem(name, asRes);
        }
    }
    
    /**
     * Loads the registry by searching through system for the specified type.
     * @param nullRes  the null ASResource instance with whose type to load the registry
     */
    public void load (IASResource nullRes) {
        _nullASRes = nullRes;
        setRefreshNeeded(true);
    }
    
    /**
     * Returns whether a lazy refresh is needed.
     * TODO Is there a more robust way?  YES! Make a resource listener that listens for family changes?
     * @return boolean
     */
    public boolean isRefreshNeeded () {
        return _needRefresh;
    }
    
    /**
     * Sets whether a refresh is needed.
     * @param b  boolean indicating whether refresh is needed
     */
    public void setRefreshNeeded (boolean b) {
        _needRefresh = b;
    }

    /**
     * Refreshes the registry only if so indicated.
     */
    public synchronized void lazyRefresh () {
        if (isRefreshNeeded()) {
            refresh();
        }
    }
    
    /**
     * Refreshes the registry by searching through system for the specified type.
     */
    public synchronized void refresh () {
        WorkspaceModifyOperation op = new WorkspaceModifyOperation() {
            public void execute(final IProgressMonitor monitor)
            throws CoreException {
                try {
                    monitor.beginTask("Refreshing registry for " + _nullASRes.getClass().getName() + "\n", WORK_UNITS);
                    ASRegistry.this.doRefresh(monitor);
                } finally {
                    monitor.setCanceled(true);
                    monitor.done();
                }
            }
        };
        
        try {
            new ProgressMonitorDialog(AcmeStudioPlugin.getActiveWindow().getShell())
                .run(false, true, op);
        } catch (InterruptedException e) {
        } catch (InvocationTargetException e) {
            Debug.errorln(e);
        }
    }
    
    private synchronized void doRefresh (IProgressMonitor monitor)
    throws CoreException {
        monitor.worked(1);
        // load search paths
        monitor.subTask("Loading search paths...");
        List searchPath = AcmeStudioPlugin.getSearchPaths();
        monitor.worked(1);
        // search through paths
        monitor.subTask("Searching the paths...");
        Iterator iter = searchPath.iterator();
        while (iter.hasNext()) {
            ASPath asPath = (ASPath) iter.next();
            String2ListMap workingMap = null;
            IProject proj = null;  // dual purpose:  proj & indicate not global
            if (asPath.getPathLevel() == ASPath.GLOBAL_LEVEL) {
                // check if gloabl registry needs loading
                if (_globalASResourceMap == null) {  // yes
                    _globalASResourceMap = new String2ListMap();
                    workingMap = _globalASResourceMap;
                } else {  // no, proceed with next ASPath
                    // we assume that only ONE global path exists in list
                    
                    // TODO:  Need some way to determine whether global paths need refreshing.
                    continue;
                }
            } else {
                workingMap = new String2ListMap();
            }
            ASPathIterator pathIter = asPath.getPathIterator();
            while (pathIter.hasNext()) {
                IPath path = pathIter.nextPath();
                monitor.subTask("Searching " + path.toString() + "...");
                IContainer container =
                    AcmeStudioPlugin.getRoot().getContainerForLocation(path);
                File folder = null;
                if (container == null) {  // use java File io
                    folder = path.toFile();
                    if (folder == null
                            || ! (folder.exists() && folder.isDirectory())) {
                        // not an existent directory, skip it
                        Debug.warnln("ASRegistry.refresh():  " + path.toString() + " is NOT an existent directory!");
                        monitor.worked(1);
                        continue;
                    }
                } else if (!container.exists()) {
                    continue;
                } else {  // get project
                    if (proj == null) {
                        proj = container.getProject();
                    }
                }
                
                Object[] members = null;
                // Precondition & invariant from this point on:
                //     one and only one of container OR folder should ever
                //     be non-null
                // N.B.:  assert statement only supported starting JDK 1.4.1_01
                //                assert ((container != null && folder == null)
                //                        || (container == null && folder != null));
                if (container != null) {
                    members = container.members();
                } else if (folder != null) {
                    members = folder.listFiles();
                }
                for (int k=0; k < members.length; k++) {
                    if (_nullASRes.isSameType(members[k])) {
                        // found a resource of the right type
                        String filename = null;
                        String stubName = null;  // e.g., family name
                        if (container != null) {
                            IResource resource = (IResource) members[k];
                            filename = resource.getName();
                            String ext = resource.getFileExtension();
                            stubName = filename.substring(0, filename.indexOf(ext)-1);
                        } else if (folder != null) {
                            File file = (File) members[k];
                            filename = file.getName();
                            int pos = filename.indexOf(".");
                            if (pos > -1) {
                                stubName = filename.substring(0, pos);
                            } else {
                                stubName = filename;
                            }
                        }
                        monitor.subTask("Loading " + filename + "...");
                        // generate a new instance from the null resource
                        //   is there need for a ROOT IASResource?
                        IASResource newASRes = _nullASRes.getNewInstance(null);
                        // load the ASResource from the file found
                        try {
                            if (container != null) {
                                IPath namePath = new Path(filename);
                                newASRes.loadFrom(container.getFile(namePath));
                            } else if (folder != null) {
                                newASRes.loadFrom((File) members[k]);
                            }
                        } catch (Throwable t) {
                            // output the error, but intercept these so the
                            // the rest of the files can be loaded
                            Debug.errorln("Error loading ASResource " + newASRes.getPath() +"!");
                            Debug.errorln(t);
                            monitor.worked(1);
                            continue;
                        }
                        workingMap.putItem(stubName, newASRes);
                        monitor.worked(1);
                    }
                }
            }  // end iterating thru list of ASPath expansion
            if (proj != null) {  // previously handled project
                // clear previous map of names to IASResource for the project 
                if (_proj2NameMap.containsKey(proj)) {
                    _proj2NameMap.getMap(proj).clear();
                }
                // put in the new map
                _proj2NameMap.put(proj, workingMap);
            } else if (workingMap != _globalASResourceMap){
                // Not project, and neither PROJECT_LEVEL
                // Add the family to the global families space
                for (Iterator iterator = workingMap.entrySet().iterator (); iterator.hasNext ();) {
                    Map.Entry entry = (Map.Entry )iterator.next ();
                    _globalASResourceMap.put (entry.getKey(), entry.getValue());
                }
            }
        }
        
        // TODO Make sure this is the appropriate place to set need refresh.
        setRefreshNeeded(false);
    }
}
