/*
 * Created on Jun 3, 2003
 *
 */
package AcmeStudio.core.actions;

import java.util.ArrayList;
import java.util.Iterator;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.gef.EditPart;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWindowActionDelegate;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASContext;
import AcmeStudio.core.resource.provider.ASResourceProvider;
import AcmeStudio.ui.editors.DiagramEditor;
import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.io.Region;
import acmeLib.elements.*;
import acmeLib.properties.AcmeProperty;
import acmeLib.tools.GraphWalker;
import acmeLib.tools.WalkerElementOperation;

/**
 * @author Bradley Schmerl
 * Tests whether the acme system is complete
 * TODO: Add preferences to switch what to check for, and also how to report
 */
public class TestSystemCompleteAction implements IWorkbenchWindowActionDelegate {

    /** The list of dangling roles **/
    private java.util.List _danglingRoles = new ArrayList();
    /** The list of dangling ports **/
    private java.util.List _danglingPorts = new ArrayList();
    /** The list of valueless properties **/
    private java.util.List _valuelessProperties = new ArrayList();

    /**
     * 
     * @author Bradley Schmerl
     * Checks whether an individual element is complete
     */
    class CompletionTester implements WalkerElementOperation {

        /* (non-Javadoc)
        * @see acmeLib.tools.WalkerElementOperation#doElementOp(acmeLib.elements.AcmeElement)
        */
        public void doElementOp(AcmeElement elt) {
            // Check whether all properties have a value
            AcmeEnumeration properties = elt.getProperties();
            while (properties.moreItems()) {
                AcmeProperty prop = (AcmeProperty) properties.getNextItem();
                if (prop.getValue() == null) {
                    _valuelessProperties.add(prop);
                }
            }

            // Check if components have dangling (unattached) ports
            if (elt instanceof AcmeComponent) {
                AcmeComponent comp = (AcmeComponent) elt;
                AcmeSystem sys = (AcmeSystem) comp.getParent();
                AcmeEnumeration ports = comp.getPorts();
                while (ports.moreItems()) {
                    AcmePort port = (AcmePort) ports.getNextItem();
                    if (sys.getAttachments(port).getSize() == 0) {
                        _danglingPorts.add(port);
                    }
                }
            }
            // Check if connectors have dangling (unattached) roles
            else if (elt instanceof AcmeConnector) {
                AcmeConnector conn = (AcmeConnector) elt;
                AcmeSystem sys = (AcmeSystem) conn.getParent();
                AcmeEnumeration roles = conn.getRoles();
                while (roles.moreItems()) {
                    AcmeRole role = (AcmeRole) roles.getNextItem();
                    if (sys.getAttachments(role).getSize() == 0) {
                        _danglingRoles.add(role);
                    }
                }
            }
        }

    }

    /**
     * 
     */
    public TestSystemCompleteAction() {
        super();
    }

    /* (non-Javadoc)
     * @see org.eclipse.ui.IWorkbenchWindowActionDelegate#dispose()
     */
    public void dispose() {

    }

    /* (non-Javadoc)
     * @see org.eclipse.ui.IWorkbenchWindowActionDelegate#init(org.eclipse.ui.IWorkbenchWindow)
     */
    public void init(IWorkbenchWindow window) {
    }

    /* (non-Javadoc)
     * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
     */
    public void run(IAction action) {
        // Only run if the diagram editor is active
        if (ASContext.current() == null
            || ASContext.current().getActiveEditor() == null
            || ASContext.current().getActiveEditor().getDiagramEditor() == null) {
            return;
        }
        DiagramEditor editor = ASContext.current().getActiveEditor().getDiagramEditor();

        // Get the root edit part representing the system
        EditPart systemEP = editor.getRootEditPart();
        if (systemEP != null) {
            AcmeSystem system = (AcmeSystem) systemEP.getModel();
            clear();
            // Walk over the system
            GraphWalker.WalkOverAllElements(system, new CompletionTester(), GraphWalker.WALK_TOP_DOWN);
            
            // Add the appropriate markers
            ASResourceProvider provider = ASContext.current().getActiveEditor().getResourceProvider();
            IResource res = provider.getFile();
            try {
                // First, delete existing incomplete markers
                res.deleteMarkers(AcmeStudioPlugin.getMessage("TestSystemCompleteAction.0"), false, IResource.DEPTH_INFINITE); //$NON-NLS-1$

                Iterator i = _danglingPorts.iterator();
                while (i.hasNext()) {
                    AcmePort role = (AcmePort) i.next();
                    String message = role.getQualifiedName() + AcmeStudioPlugin.getMessage("TestSystemCompleteAction.UnattachedMsg"); //$NON-NLS-1$
                    createIncompleteMarker(provider, res, role, message);

                }

                i = _danglingRoles.iterator();
                while (i.hasNext()) {
                    AcmeRole role = (AcmeRole) i.next();
                    String message = role.getQualifiedName() + AcmeStudioPlugin.getMessage("TestSystemCompleteAction.UnattachedMsg"); //$NON-NLS-1$
                    createIncompleteMarker (provider, res, role, message);

                }
                
                i = _valuelessProperties.iterator ();
                while (i.hasNext ()) {
                    AcmeProperty prop = (AcmeProperty )i.next ();
                    String message = prop.getQualifiedName() + AcmeStudioPlugin.getMessage("TestSystemCompleteAction.NoValueMsg"); //$NON-NLS-1$
                    createIncompleteMarker(provider, res, prop, message);
                }
            }
            catch (CoreException e) {
            }
        }

    }

    /**
     * Creates an incomplete marker and attaches it to the resource
     * @param provider The resource provider, to get the regions
     * @param res The IResource to associate the marker with
     * @param obj The object that was incomplete
     * @param message The message to associated with the marker
     *
     */
    private void createIncompleteMarker(ASResourceProvider provider, IResource res, AcmeObject obj, String message) throws CoreException {
        IMarker marker = res.createMarker(AcmeStudioPlugin.getMessage("TestSystemCompleteAction.4")); //$NON-NLS-1$
        marker.setAttribute(IMarker.MESSAGE, message);
        marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_WARNING);
        Region reg = provider.getRegionManager().getRegion(obj);
        if (reg != null)
            marker.setAttribute(IMarker.LINE_NUMBER, reg.getBeginLine());
        else
            marker.setAttribute(IMarker.LINE_NUMBER, 1);
    }

    /* (non-Javadoc)
     * @see org.eclipse.ui.IActionDelegate#selectionChanged(org.eclipse.jface.action.IAction, org.eclipse.jface.viewers.ISelection)
     */
    public void selectionChanged(IAction action, ISelection selection) {
        // Only activate if the selection is an edit part that has an acme element as its model
        action.setEnabled(false);
        if (selection instanceof IStructuredSelection) {
            IStructuredSelection iSel = (IStructuredSelection) selection;
            Object model = null;
            if (iSel.getFirstElement() instanceof EditPart) {
                model = ((EditPart) iSel.getFirstElement()).getModel();

            }
            else if (iSel.getFirstElement() instanceof AcmeElement) {
                model = iSel.getFirstElement();
            }
            if (model != null && !(model instanceof AcmeElementType) && model instanceof AcmeElement) {
                action.setEnabled(true);
            }
        }
    }

    private void clear() {
        _danglingRoles.clear();
        _danglingPorts.clear();
        _valuelessProperties.clear();
    }
    
}
