package AcmeStudio.core.resource;

import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.widgets.Display;
import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASModel;
import AcmeStudio.core.ASStatus;
import AcmeStudio.core.resource.provider.ASResourceProvider;
import AcmeStudio.ui.editors.text.TextRegionManager;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Debug;
import AcmeStudio.util.Helper;
import AcmeStudio.util.Utils;
import AcmeStudio.wizards.convertFromVersion1.SplitAcmeDesignWizard;
import acmeLib.armaniParser.ParseError;
import acmeLib.armaniParser.ParseException;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.*;
import acmeLib.errorManagement.*;
import acmeLib.tools.AcmeDesignChangeEvent;
import acmeLib.tools.AcmeDesignChangeNotifier;
import acmeLib.tools.IAcmeDesignChangeListener;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 * 
 * AcmeStudio Design resource type.
 */
public class ASDesign extends ASResource implements IAcmeDesignChangeListener {
    
    /** File extension of the ASFamily resource */
    public static String FILE_EXT = _resource_nls.getString("ASResource.systemExt");
    
    /** The shared null instance of this class */
    protected static IASResource _nullInstance = null;
    
    /** the AcmeDesign being managed */
    private AcmeDesign _design = null;
    /** Corresponding visual style ASResource*/
    private ASVisualStyle _asVStyle = null;
    
    private boolean _changedOnLoad = false;
    /**
     * Creates an empty ASDesign object.
     * 
     * @param parent
     *           the parent IASResource object.
     * @param provider
     *           the ASResourceProvider containing this ASResource.
     */
    public ASDesign(IASResource parent, ASResourceProvider prov) {
        super(parent, prov);
        
        registerDesignChangeListener();
    }
    
    /**
     * Returns a null instance of this type. This method is useful for passing
     * IASResource objects around to identify the resource "types," as used,
     * for example, in ASRegistry.
     * 
     * @return IASResource the null instance
     */
    public static IASResource nullInstance() {
        if (_nullInstance == null) { // lazy init
            _nullInstance = new ASDesign(null, null);
        }
        return _nullInstance;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#dispose() */
    public void dispose() {
        unregisterDesignChangeListener();
        if (_asVStyle != null) {
            _asVStyle.dispose();
            _asVStyle = null;
        }
        _design = null;
        super.dispose();
    }
    
    /**
     * @see acmeLib.tools.IAcmeDesignChangeListener#designChanged(acmeLib.tools.AcmeDesignChangeEvent)
     */
    public void designChanged(AcmeDesignChangeEvent evt) {
        if (_design == null) {  // don't respond to any changes yet
            return;
        }

        // check if event is relevant to this ASDesign
        String canonPath = evt.getCanonicalPath();
        if (canonPath != null && canonPath.equals(getCanonicalPath())) {
            // relevant!
            if (evt.getNewDesign() != null && evt.getNewDesign() != _design) {
                _design = evt.getNewDesign();
                // trigger RELOAD from ASRP
                doSelfChanged();
            }
        }
    }

    /**
     * @see AcmeStudio.core.resource.IASResource#getNewInstance(IASResource) */
    public IASResource getNewInstance(IASResource parent) throws CoreException {
        return new ASDesign(parent, null);
    }
    
    /**
     * Saves a local copy of the AcmeDesign contained in the parent
     * AcmeElement.
     * 
     * @see AcmeStudio.core.resource.IASResource#setElement(AcmeElement)
     */
    public void setElement(AcmeElement element) {
        if (element == getElement()) return;  // don't need to do any update

        ASModel asModel = ASModel.current();

        super.setElement(element);
        _design = super.getDesign();

        if (_design != null) {
            // update the ASFamily for all imported designs
            Iterator i = getImportedRecords().iterator();
            while (i.hasNext()) {
                AcmeImportedDesignRecord record = (AcmeImportedDesignRecord) i.next();
                AcmeDesign design = record.getResolvedDesign();
                IASResource asRes = asModel.getASResource(record.getCanonicalFilename());
                if (asRes != null) {
                    try {
                        asRes.reload();
                    } catch (CoreException e) {
                        Debug.diagnoseln(e);
                    } catch (ParseError e) {
                        Debug.diagnoseln(e);
                    } catch (AcmeException e) {
                        Debug.diagnoseln(e);
                    }
                } else {
                    // create an ASFamily to hold this design
                    // this will cause ASFamilies to be created recursively up
                    // the import branch
                    ASFamily asFam = new ASFamily(this);
                    IPath path = new Path(record.getCanonicalFilename());
                    try {
                        IFile ifile = AcmeStudioPlugin.getRoot().getFileForLocation(path);
                        if (ifile == null || !ifile.exists()) {
                            File file = new File(record.getCanonicalFilename());
                            asFam.loadFrom(file);
                        } else {
                            asFam.loadFrom(ifile);
                        }
                    } catch (CoreException e) {
                        Debug.diagnoseln(e);
                    } catch (ParseError e) {
                        Debug.diagnoseln(e);
                    } catch (AcmeException e) {
                        Debug.diagnoseln(e);
                    }
                }
            }
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#isSameType(Object) */
    public boolean isSameType(Object resource) {
        return false;
    }
    
    /**
     * Returns the ASVisualStyle associated with this ASDesign.
     * 
     * @return
     */
    public ASVisualStyle getASVisualStyle() {
        return _asVStyle;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getDesign() */
    public AcmeDesign getDesign() {
        AcmeDesign design = _design;
        if (_design == null) {
            design = super.getDesign();
        }
        return design;
    }
    
    /**
     * Loads acme design from the input stream, marking errors if any. If the
     * input file is not an IFile, no error marker will be produced.
     * 
     * @param monitor  the progress monitor to indicate progress
     * @param input    the input stream from which to load
     * @throws CoreException
     * @throws ParseError
     * @see AcmeStudio.core.resource.ASResource#loadStream(org.eclipse.core.runtime.IProgressMonitor, java.io.InputStream)
     */
    protected synchronized void loadStream(IProgressMonitor monitor, InputStream input)
    throws CoreException, ParseError, AcmeException {
        // remove this resource as resource listener to prevent infinite trigger
        Debug.traceln("ASDesign.loadStream...");
        unregisterResourceListener();
        unregisterASProvListener();
        unregisterDesignChangeListener();
        try {
            if (monitor != null) {
                monitor.subTask(AcmeStudioPlugin.getMessage("ASDesign.ClearingErrorsMsg")); //$NON-NLS-1$
            }
            _acmeErrorList.clear();
            _parseErrorList.clear();
            if (monitor != null) {
                monitor.worked(1);
                monitor.subTask(AcmeStudioPlugin.getMessage("ASDesign.LookupDesignMsg")); //$NON-NLS-1$
            }
            String filename = getCanonicalPath();
            AcmeOpenedDesignRecord designRec = null;
            if (filename != null) {
                designRec = AcmeDesign.getDesignRecord(filename);
            }
            AcmeDesign design = null;
            if (designRec != null) { // fetch the opened design
                design = designRec.getResolvedDesign();
            }
            
            if (design == null) {
                design = Helper.getAcmeDesignFromInputStream(monitor, input, false);
                if (design.getFamilies().getSize() + design.getSystems().getSize() > 1) {
                    if (MessageDialog.openQuestion(
                            Display.getCurrent().getActiveShell(),
                            AcmeStudioPlugin.getMessage("ASDesign.SeparateDesignTitle"), //$NON-NLS-1$
                            AcmeStudioPlugin.getMessage("ASDesign.SeparateDesignMsg"))) { //$NON-NLS-1$
                        SplitAcmeDesignWizard wizard = new SplitAcmeDesignWizard(design);
                        WizardDialog dialog = new WizardDialog(Display.getCurrent().getActiveShell(), wizard);
                        dialog.create();
                        if (dialog.open() == WizardDialog.OK) {
                            _changedOnLoad = true;
                        }
                    }
                }
                Helper.resolveAllInDesign(monitor, design);
                // add to opened design record
                designRec = addDesignRecord(filename, design);
            } else if (designRec.shouldReload() || this.shouldReload()) {
                // design should be reloaded, parse anew
                
                /* BRS 8/13/2004: BugFix#306,#310
                 * First, save all the annotations so that they can be put back later on
                 */
                List persistentData = extractPersistentData();
                // reload the design
                design = Helper.getAcmeDesignFromInputStream(monitor, input);
                // re-inject annotations
                injectPersistentData(persistentData);

                // save errors
                AcmeEnumeration errors = AcmeErrorManager.getErrors();
                while (errors.moreItems()) {
                    AcmeError error = (AcmeError) errors.getNextItem();
                    if (error instanceof AcmeParseError) {
                        _parseErrorList.add(error);
                    } else {
                        _acmeErrorList.add(error);
                    }
                    _errorMap.put(error.getSource(), error);
                }
                
                // update the opened design record with the new AcmeDesign object
                updateDesignRecord(designRec, design);
            }

            setElement(design);

            // try to obtain the relevant RegionManager
            AcmeOutputStream aos = null;
            TextRegionManager regionMgr = null;
            if (getProvider() != null) {
                regionMgr = getProvider().getRegionManager();
            }
            if (regionMgr == null) { // provider's RegionManager not set
                if (monitor != null)
                    monitor.subTask(AcmeStudioPlugin.getMessage("ASDesign.RefreshingMsg")); //$NON-NLS-1$
                aos = new AcmeOutputStream();
                try {
                    getDesign().exportArmani(aos);
                    regionMgr = new TextRegionManager(aos.getRegionManager());
                    
                    if (getProvider() != null) {
                        getProvider().setRegionManager(regionMgr);
                    }
                } catch (AcmeException e1) {
                    // ignore, just don't set region manager
                }
            }
            if (monitor != null)
                monitor.subTask(AcmeStudioPlugin.getMessage("ASDesign.ProcessingErrorsMsg")); //$NON-NLS-1$

            if (hasParseError()) {
                if (monitor != null) {
                    monitor.worked(1);
                    monitor.setCanceled(true);
                }
                throw new ParseError(AcmeStudioPlugin.getMessage("ASDesign.ParseErrorMsg")); //$NON-NLS-1$
            } else {
                // if no show-stopping parse error, then finish loading
                loadExtensions(monitor);
                loadVisualStyle();
            }

            if (AcmeErrorManager.getNumErrorsReported() != 0) {
                AcmeEnumeration errors = AcmeErrorManager.getErrors();
                boolean unresolvedErrors = false;
                while (errors.moreItems() && !unresolvedErrors) {
                    Object error = errors.getNextItem ();
                    if (/*error instanceof InstanceRefResolutionError ||*/ error instanceof TypeRefResolutionError || error instanceof TypeInstantiationError)
                        unresolvedErrors = true;
                }
                if (unresolvedErrors) {
                    throw new ParseError (AcmeStudioPlugin.getMessage ("ASDesign.ContainsUnresolvedItemsError")); //$NON-NLS-1$
                }
            }
        } catch (ParseError e) {
            if (monitor != null)
                monitor.setCanceled(true);
            if (e instanceof ParseException) {
                _parseErrorList.add(e);
            }
            
            throw e;  // debug output will occur later
        } catch (IOException e) {
            // Ignore since we're done with reading input stream
            Debug.errorln(e);
        } catch (Exception e) {
            e.printStackTrace ();
            if (monitor != null) {
                monitor.setCanceled(true);
            }
            Debug.getLog().error(AcmeStudioPlugin.getMessage("ASDesign.ResolveImportErrorMsg"), e); //$NON-NLS-1$
        } finally {
            try {
                input.close();
            } catch (IOException e) {
                // ignore
            }
            // a ParseError was thrown, so process what errors have been found
            // this need not happen in a normal parse, since error will
            // be processed after a typecheck (see
            // ASResourceProvider.forceUpdateText()).
            processErrors();
            // clear the reload bit
            setReloadState(NONE);
            // reinstate this resource as resource listener
            registerResourceListener();
            registerASProvListener();
            registerDesignChangeListener();
        }
    }
    
    /**
     * Call on data persistors to load extensions.
     * @param monitor
     */
    private void loadExtensions(IProgressMonitor monitor) {
        File fileLoaded = null;
        try {
            fileLoaded = (File) _resource;
        } catch (ClassCastException e) {
            // then it should be IFile
            fileLoaded = ((IFile) _resource).getLocation().toFile();
        }
        if (fileLoaded != null) {
            for (int i = 0; i < _dataPersist.length; i++) {
                if (monitor != null) monitor.subTask(AcmeStudioPlugin.getMessage("ASDesign.ExtensionMsg") + _persistIds [i]); //$NON-NLS-1$
                try {
                    _dataPersist[i].load(this, fileLoaded);
                } catch (Exception e) {
                    //                  e.printStackTrace();
                    Debug.getLog().error (AcmeStudioPlugin.getMessage("ASDesign.ExtensionErr") + _persistIds [i]); //$NON-NLS-1$
                    Debug.errorln(e);
                } finally {
                    if (monitor != null) monitor.worked(1);
                }
            }
        }
    }

    /**
     * Load the corresponding visual style.
     * @throws CoreException
     * @throws ParseError
     * @throws AcmeException
     */
    private void loadVisualStyle() throws CoreException, ParseError, AcmeException {
        if (_asVStyle == null) {
            _asVStyle = new ASVisualStyle(this);
            AcmeFamily fam = getFamily();
            if (fam != null && Helper.getAcmeFamilies(getDesign()).contains(fam)) {
                // this design contains a family (not imported), which is true
                // when this is an AcmeFamily, or system with inline AcmeFamily, 
                // so load a visualstyle
                String name = fam.getName();
                String filename = name + "." + ASVisualStyle.FILE_EXT;
                // try searching the AcmeStudio paths
                Object obj = AcmeStudioPlugin.findFileInFamPaths(filename);
                // BugFix #146:
                //   After failing to find vstyle is standard search paths,
                //   look for vstyle file in this design's path, then go up the
                //   dir hierarchy
                if (obj == null && getFile() instanceof IFile) {
                    IFile designFile = (IFile) getFile();
                    List folders = new ArrayList();
                    // try the path from this file's location and up
                    IFolder designDir = null;
                    try {
                        designDir = designFile.getParent().getFolder(new Path("."));
                    } catch (IllegalArgumentException e) {
                        // The family may be in the system file, so try the
                        // families directory
                        designDir = designFile.getParent().getFolder(new Path("families"));
                    }
                    if (designDir != null) {
                        folders.add(designDir);
                        // try the path from a validated directory based on
                        // this path
                        IFolder validDir = (IFolder) Utils.validateFamilyPath(designDir.getFullPath());
                        if (designDir != null
                                && validDir != null
                                && !designDir.getFullPath().equals(validDir.getFullPath())) {
                            // not the same, so append this folder as well
                            folders.add(validDir);
                        }
                        Iterator i = folders.iterator();
                        while (obj == null && i.hasNext()) {
                            IFolder dir = (IFolder) i.next();
                            if (dir != null && dir.exists()) {
                                obj = dir.getFile(filename);
                            }
                            while ((obj == null || !((IFile) obj).exists())
                                    && dir != null
                                    && dir.getFullPath().segmentCount() > 2) {
                                dir = dir.getFolder(".."); // get parent
                                // directory
                                obj = dir.getFile(filename);
                            }
                        }
                    }
                }
                if (obj != null) {
                    if (obj instanceof IFile) {
                        _asVStyle.loadFrom((IFile) obj);
                    } else if (obj instanceof File) {
                        _asVStyle.loadFrom((File) obj);
                    }
                }
            } // otherwise, this is a system design, so no visual style to load
        } else {
            if (_asVStyle.shouldReload()) {
                _asVStyle.reload();
            }
        }
    }

    private AcmeOpenedDesignRecord addDesignRecord(String filename, AcmeDesign design) {
        AcmeOpenedDesignRecord designRec = new AcmeOpenedDesignRecord();
        designRec.setCanonicalFilename(filename);
        designRec.setLastModified(getLastModified());
        designRec.setResolvedDesign(design);
        AcmeDesign.addDesignRecord(filename, designRec);
        return designRec;
    }
    
    /**
     * During such an update, the ASModel design canonical path to ASDesign
     * mapping need not be updated, just update the Design Record resolved
     * design and modified timestamp.
     * @param designRec  the opened design record to update
     * @param design     the resolved design to store in the opened design record 
     */
    private void updateDesignRecord(AcmeOpenedDesignRecord designRec, AcmeDesign design) {
        if (designRec.getResolvedDesign() == design) return;

        // update the design record
        designRec.setLastModified(getLastModified());
        designRec.setResolvedDesign(design);
    }
    
    /**
     * @see AcmeStudio.core.resource.ASResource#doSave(IProgressMonitor) */
    protected boolean doSave(IProgressMonitor monitor) throws CoreException {
        boolean successSys = false;
        
        // save the instance content; family contents in _acmeFamilies are
        // automatically written; and so are _acmeImports contents
        try {
            if (_design != null) {
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                Debug.traceln("ASDesign.doSave: injecting extensions...");
                _design.doInjectionOperations();
                
                Helper.getOutputStream(_design, out);
                File fileSaved = null;
                if (_resource instanceof IFile) {
                    Debug.traceln("ASDesign.doSave: saving design as IFile...");
                    ByteArrayInputStream bais = new ByteArrayInputStream(out.toByteArray());
                    IFile ifile = (IFile )_resource;
                    if (ifile.exists()) { // create the file
                        ifile.setContents(bais, true, true, monitor);
                    } else {
                        ifile.create(bais, true, monitor);
                    }
                    
                    fileSaved = ((IFile )_resource).getLocation().toFile();
                } else if (_resource instanceof File) {
                    Debug.traceln("ASDesign.doSave: saving design as File...");
                    PrintStream pstream = new PrintStream(new FileOutputStream((File) _resource));
                    pstream.write(out.toByteArray());
                    
                    fileSaved = (File )_resource;
                } else { // should NOT happen
                    throw new CoreException(
                        new ASStatus(ASStatus.ERROR,
                                     ASStatus.RESOURCE_TYPE_UNRECOGNIZED,
                                     AcmeStudioPlugin.getMessage("ASResource.resourceTypeUnrecognized"),
                                     null));
                }
                out.close();
                _design.doExtractionOperations();
                
                // invoke save extensions
                if (fileSaved != null) {
                    for (int i = 0; i < _dataPersist.length; i++) {
                        _dataPersist[i].save(this, fileSaved);
                    }
                    // Refresh the data directory because extensions may have changed files
                    IContainer iData = AcmeStudioPlugin.getRoot().getContainerForLocation(getDataDirectory());
                    iData.refreshLocal(IResource.DEPTH_ONE, null);
                }
                
                successSys = true;
                Debug.traceln("ASDesign.doSave: AcmeSystem file saved without error.");
            } else {
                Debug.errorln("ASDesign.doSave: Design is NULL! Unable to proceed with save.");
            }
        } catch (IOException e) {
            Debug.getLog().error("ASDesign.doSave: ", e);
        }
        
        // save separate familes will be necessary if user decides to split
        // family from instance by setting preferences.
        // save visual styles
        boolean successVis = true;
        // - iterate through each family looking for visual styles to save
        // BugFix #146: don't get all the imports, just the immediate families
        List viss = new ArrayList(); // to track what vis styles were processed
        Iterator iter = getFamilies().iterator();
        while (successSys && iter.hasNext()) {
            successVis = false;
            AcmeFamily fam = (AcmeFamily) iter.next();
            Debug.diagnoseln("ASResourceProvider.saveAcme:  saving vis of fam " + fam.getName() + "...");
            if (fam == null) { // should not happen; but if does, report and
                // continue
                Debug.warnln("ASResourceProvider.saveAcme:  unexpected null family encountered while saving visual style; ignored.");
                continue;
            }
            
            ASVisualStyle asVStyle = ASModel.current().getASVisualStyle(fam);
            if (asVStyle != null) {
                viss.add(asVStyle);
                successVis = asVStyle.save(monitor);
            }
        }
        // fix: since the modification to have ASDesign load its own vis style,
        //   we haven't saved the vis style of that file, and now that needs
        //   to be done, if the fam list hasn't already taken care of it
        ASVisualStyle asVStyle = getASVisualStyle();
        if (asVStyle != null && !viss.contains(asVStyle)) {
            Debug.diagnoseln("ASResourceProvider.saveAcme():  directly saving design's vis...");
            successVis = asVStyle.save(monitor);
        }
        
        if (successSys) {
            setReloadState(NONE);
            // Make sure that the command stack is marked where we saved
            AcmeStudioChangeManager.globalChangeManager.getCommandStack(this).markSaveLocation();
        }
        return successSys && successVis;
    }
    
    /**
     * Returns whether this design is considered to be changed upon loading.
     * It would be set to true, e.g., in the case of an Acme file being split
     * into family and system files.
     * @return
     */
    public boolean changedOnLoad() {
        return _changedOnLoad;
    }
    
    protected void registerDesignChangeListener() {
        // add as listener to AcmeDesign change
        AcmeDesignChangeNotifier.getInstance().addDesignChangeListener(this);
    }
    
    protected void unregisterDesignChangeListener() {
        // remove as listener from AcmeDesign change
        AcmeDesignChangeNotifier.getInstance().removeDesignChangeListener(this);
    }

    protected List extractPersistentData () {
        List persistentData = new LinkedList();

        if (getDesign() != null) {
            getDesign().doExtractionOperations();

            for (int i = 0; i < _dataPersist.length; i++) {
                persistentData.add(_dataPersist[i].extractData(this));
            }
        }

        return persistentData;
    }
    
    protected void injectPersistentData (List dataList) {
        if (getDesign() != null) {
            // reload annotations
            int i = 0;
            for (Iterator iter = dataList.iterator(); iter.hasNext();) {
                Object data = (Object )iter.next ();
                _dataPersist [i++].injectData(this, data);
            }

            getDesign().doInjectionOperations();

            if (_prov != null) _prov.setSynchedWithText(false);
        }
    }

}
