package AcmeStudio.core.resource;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASContext;
import AcmeStudio.core.ASContextEvent;
import AcmeStudio.core.ASStatus;
import AcmeStudio.core.IASContextListener;
import AcmeStudio.util.Debug;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * Represents AcmeStudio search paths, and is designed to keep environment
 * variables in tact, unless getLocationIterator is called, at which time an
 * iterator to the first path in the environment variable is returned.
 */
public class ASPath implements Comparable, IASContextListener {
    
    /** The Project path environment variable */
    public static String PROJECT_PATH = "AS_PROJECT_PATH";
    /** The Global path environment variable */
    public static String GLOBAL_PATH = "AS_GLOBAL_PATH";
    /** The Search path environment variable */
    public static String SEARCH_PATH = "AS_SEARCH_PATH";
    /** The constant for path of Unknown level */
    public static int UNKNOWN_LEVEL = 0;
    /** The constant for path of Project level */
    public static int PROJECT_LEVEL = 1;
    /** The constant for path of Global level */
    public static int GLOBAL_LEVEL = 2;

    private static String PSEP = AcmeStudioPlugin.getResourceString("AcmeStudio.path.separator");
    private static String FSEP = AcmeStudioPlugin.getResourceString("AcmeStudio.file.separator");

    /** The string of this path */
    private String _path = null;
    /** The level of the path */
    private int _level = UNKNOWN_LEVEL;
    /** The current project for this ASPath if this is a PROJECT_LEVEL path.
     *  Otherwise, this data member is not applicable. */
    private IProject _curProj = null;
    /** Flag indicating whether context was recently changed */
    private boolean _refreshPath = false;
    
    /** The cached list of actual paths corresponding to this ASPath */
    private List _pathList = null;
    
    /**
     * Constructor for ASPath.
     */
    public ASPath(String path) {
        super();
        
        _path = path;
        // assign path level
        if (path.startsWith(PROJECT_PATH, 1)) {
            _level = PROJECT_LEVEL;
        } else if (path.startsWith(GLOBAL_PATH, 1)) {
            _level = GLOBAL_LEVEL;
        }
        ASContext.current().addContextListener(this);
    }
    
    /**
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        ASContext.current().removeContextListener(this);
        super.finalize();
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString () {
        return getPathStr();
    }    
    
    /**
     * @see java.lang.Comparable#compareTo(Object)
     */
    public int compareTo(Object obj) {
        // cast object to ASPath, so that ClassCastException is triggered if
        // the wrong object instance is supplied.
        ASPath path2 = (ASPath) obj;
        return this.getPathStr().compareTo(path2.getPathStr());
    }
    
    /**
     * @see AcmeStudio.core.IASContextListener#contextChanged(ASContext)
     */
    public void contextChanged(ASContextEvent event) {
        if (getPathLevel() == PROJECT_LEVEL
                && event.hasProjectChanged()
                && !ASContext.current().getActiveProject().equals(_curProj)) {
            // the search list for this ASPath should be refreshed
            _curProj = ASContext.current().getActiveProject();
            _refreshPath = true;
        }
    }
    
    /**
     * Returns the ASPath representation of the IPath object.
     * @param path  the path to convert
     * @return ASPath  the resulting ASPath object
     */
    public static ASPath pathFor (IPath path) {
        return pathFor(path.toString());
    }
    
    /**
     * Returns the ASPath representation of the path String.
     * @param path  the path to convert
     * @return ASPath  the resulting ASPath object
     */
    public static ASPath pathFor (String path) {
        ASPath asPath = null;

        // Search path is at end because it contains PROJECT_ and GLOBAL_PATH,
        // and something different is done with SEARCH PATH
        String[] paths = {PROJECT_PATH, GLOBAL_PATH, SEARCH_PATH};
        // four backslash is needed because the backslash within the
        // Pattern object needs to be escaped.
        path = path.replaceAll("\\\\", FSEP);

        OUTER: for (int i=0; i < paths.length; i++) {
            String pathVal = "";
            if (paths[i].equals (SEARCH_PATH)) {
                pathVal = AcmeStudioPlugin.getSearchPathString();
            }
            else {
                pathVal = AcmeStudioPlugin.getResourceString(paths[i]);
            }
            // see if the supplied path begins with the path of an env't var
            StringTokenizer listTokens = new StringTokenizer(pathVal, PSEP);
            while (listTokens.hasMoreTokens()) {
                String pathItem = listTokens.nextToken();
                pathItem = pathItem.replaceAll("\\\\", FSEP);
                if (path.startsWith(pathItem)) {
                    // yes! compose new path using env't var name
                    IPath newPath = null;
                    if (paths[i].equals(SEARCH_PATH)) {
                        newPath = new Path(path.substring(pathItem.length()));
                    }
                    else {
                        newPath = new Path("$" + paths[i]).append(path.substring(pathItem.length()));
                    }
                    asPath = new ASPath(newPath.toString());
                    break OUTER;
                }
            }
        }
        if (asPath == null) {
            asPath = new ASPath(path);
        }
        return asPath;
    }
    
    /**
     * Returns the path level of this ASPath.
     * @return int  the path level
     */
    public int getPathLevel () {
        return _level;
    }
    
    /**
     * Returns the path string of this ASPath.
     * @return String  the path string
     */
    public String getPathStr () {
        return _path;
    }
    
    /**
     * Returns the ASPath Iterator for the possible ASPath enumerations.
     * @return ASPathIterator  the returned iterator
     */
    public ASPathIterator getPathIterator () throws CoreException {
        if (_pathList != null && !_refreshPath) {
            return new ASPathIterator(_pathList);
        }
        
        _pathList = new ArrayList();
        StringTokenizer pathTokens = new StringTokenizer(_path, FSEP);
        if (pathTokens.hasMoreTokens()) {
            String var = pathTokens.nextToken();
            IPath restPath= new Path("");
            // for windows path with single drive-letter
            if (var.endsWith (":")) {
                restPath = new Path (FSEP); 
            }
            while (pathTokens.hasMoreTokens()) {
                restPath = restPath.append(pathTokens.nextToken());
            }
            if (var.startsWith("$")) {  // get the value for the environment var
                String list = AcmeStudioPlugin.getResourceString(var.substring(1));
                if (list != var) {  // found a value
                    // parse the paths
                    StringTokenizer listTokens = new StringTokenizer(list, PSEP);
                    while (listTokens.hasMoreTokens()) {
                        IPath ipath = new Path(listTokens.nextToken());
                        // append the rest of the path to this individual path
                        ipath = ipath.append(restPath);
                        Debug.diagnoseln("ASPath.getPathIterator: path added == " + ipath.toString());
                        _pathList.addAll(traverseSubpaths(ipath));
                    }
                } else {  // something wrong...
                    throw new CoreException(
                            new ASStatus(
                                    ASStatus.ERROR,
                                    ASStatus.UNDEFINED_PROPERTY,
                                    "Cannot evaluate resource property " + var + "!",
                                    null));
                }
            } else {
                // just add the item to the path list, resulting in singleton list
                _pathList.add(new Path(_path));
            }
        } else {  // empty path supplied, just ignore
        }
        _refreshPath = false;  // reset context changed flag
        return new ASPathIterator(_pathList);
    }
    
    /**
     * Recursively traverse all subdirectories of supplied path and add the
     * paths to the list returned.  It might make sense for this method to
     * use filters that can be extended to set what directories to exclude.
     * TODO: ASPath.traverseSubpaths(IPath)--Allow use of filters to exclude paths
     * @param path  the path in which to traverse subdirectories
     * @return List  list of valid subdirectories traversed, where valid
     *               subdirectories are the ones that are directory, exist, and 
     *               are not "CVS".
     */
    private List traverseSubpaths (IPath path) {
        List pathList = new ArrayList();
        pathList.add(path);
        File filepath = path.toFile();
        if (filepath.exists() && filepath.isDirectory()) {
            // treat as java.io.File and fetch all sub paths
            File subfiles[] = filepath.listFiles();
            for (int i=0; i < subfiles.length; i++) {
                if (subfiles[i].exists() && subfiles[i].isDirectory()
                        && !subfiles[i].getName().equals("CVS")) {
                    
                    IPath newpath = new Path(subfiles[i].getPath());
                    pathList.addAll(traverseSubpaths(newpath));
                }
            }
        }
        return pathList;
    }
}
