package AcmeStudio.core.resource;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.StringTokenizer;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.text.IDocument;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASModel;
import AcmeStudio.core.ASStatus;
import AcmeStudio.core.resource.datapersistence.IDataPersistence;
import AcmeStudio.core.resource.provider.ASResourceProvider;
import AcmeStudio.util.Debug;
import AcmeStudio.util.Helper;
import acmeLib.AcmeObject;
import acmeLib.armaniParser.ParseError;
import acmeLib.armaniParser.ParseException;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.Region;
import acmeLib.designRules.DesignRule;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeFamily;
import acmeLib.elements.AcmeImportedDesignRecord;
import acmeLib.elements.AcmeOpenedDesignRecord;
import acmeLib.elements.AcmeSystem;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.AcmeParseError;
import acmeLib.errorManagement.IAcmeErrorListener;
import acmeLib.properties.AcmeProperty;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * AcmeStudio Resource default implementation and common methods.
 * Must be subclassed.
 */
public abstract class ASResource implements IASResource, IAcmeErrorListener, IResourceChangeListener {
    
    /**
     * Processes extensions of the type AcmeStudio.dataPersistence, which allows
     * additional information about a file to be saved at the point of save,
     * and loaded with the loading of the file.  That is, the extension allows
     * the plug-in of additional behavior to collect more information about the
     * file being saved, and to save that information elsewhere, as well as
     * to load additional information into AcmeStudio when the corresponding
     * Acme file gets loaded. 
     */
    public static void processExtensions () {
        List persist = new ArrayList();
        List persistIds = new ArrayList();
        
        IExtensionRegistry registry = Platform.getExtensionRegistry();
        IExtensionPoint point = registry.getExtensionPoint("AcmeStudio.dataPersistence");
        IExtension[] extensions = point.getExtensions();
        for (int i = 0; i < extensions.length; i++) {
            IConfigurationElement[] elements = extensions[i].getConfigurationElements();
            for (int j = 0; j < elements.length; j++) {
                if ("dataPersistence".equals(elements[j].getName())) {
                    try {
                        IDataPersistence object = (IDataPersistence) elements[j].createExecutableExtension("class");                      
                        persist.add(object);
                        String label = extensions [i].getLabel ();
                        if (label == null || label.equals("")) {
                            label = extensions [i].getUniqueIdentifier();
                        }
                        persistIds.add (label);
                    }
                    catch (CoreException e) {
                        Debug.errorln (e);
                    }
                }
            }
        }
        
        
        // save to static class field
        _dataPersist = (IDataPersistence[]) persist.toArray(new IDataPersistence[0]);
        _persistIds = (String []) persistIds.toArray (new String [0]);
    }
    
    protected static IDataPersistence[] _dataPersist = null;
    protected static String [] _persistIds = null;
    
    public class DeltaVisitor implements IResourceDeltaVisitor {
        public boolean visit(IResourceDelta delta) {
            IResource res = delta.getResource();
            if (!(res instanceof IFile)) {
                // workspaceroot, project, or folder, so skip
                return true;
            } else {
                if (res.getLocation().equals(getCanonicalPath())) {  // let thru
                } else if (delta.getMovedFromPath() != null
                        && delta.getMovedFromPath().lastSegment().equals(new Path(getPath()).lastSegment())) {
                    // also let thru... because most likely a rename
                } else {
                    return false;  // short-circuit visit, & don't visit children
                }
            }
            switch (delta.getKind()) {
            case IResourceDelta.ADDED : {
                Debug.traceln("DeltaVisitor.visit: " + "Resource " + res.getFullPath() + " added");
                if ((delta.getFlags() & IResourceDelta.MOVED_FROM) != 0) {
                    if (delta.getMovedFromPath() != null &&
                        getFile() != null) {
                        // now check if the moved from path matches this file's path
                        if (getFile() instanceof IFile) {  // otherwise not sensible
                            if (delta.getMovedFromPath().equals(((IFile) getFile()).getFullPath())) {
                                // a rename occurred!
                                // TODO:  ASResource.DeltaVisitor.visit(), handle rename 
                            }
                        }
                    }
                }
                break;
            }
            case IResourceDelta.REMOVED : {
                Debug.traceln("DeltaVisitor.visit: " + "Resource " + res.getFullPath() + " removed");
                break;
            }
            case IResourceDelta.CHANGED : {
                //Debug.traceln("DeltaVisitor.visit: " + "Resource " + res.getFullPath() + " changed");
                try {
                    // check if any of the changed resources is a family of this file
                    String canonFile = res.getLocation().toFile().getCanonicalPath();
                    if (canonFile.equals(getCanonicalPath())) {
                        // disabled this call because it would be redundant
                        //  with the designChanged() effect in ASDesign
                        doSelfChanged();
                    } else if (isRelevantResourceChanged(canonFile)) {
                        doAncestorChanged();
                    } else {
                        return false;
                    }
                } catch (IOException e) {
                    Debug.diagnoseln(e); // ignore error
                }
                break;
            }}
            return true;
        }
    }
    
    /** ASResource property file */
    protected static ResourceBundle _resource_nls = ResourceBundle.getBundle("AcmeStudio.core.resource.asresource");
    
    /** The parent IASResource of this IASResource */
    protected IASResource _parent = null;
    /** The containing ASResourceProvider */
    protected ASResourceProvider _prov = null;
    /** The array of listeners monitoring changes to this resource */
    //@lock
    protected List _listeners = null;
    
    /** The filesystem resource under this ASResource, either IFile or File */
    protected Object _resource = null;
    /** The ASPath to the underlying resource */
    protected ASPath _asPath = null;
    /** The primary Acme element kept by this ASResource */
    protected AcmeElement _element = null;
    /** List of AcmeErrors reported by the parser */
    protected List _acmeErrorList = null;
    /** List of AcmeParseErrors or ParseExceptions reported by the parser;
     *  these cause MPAE to load only the AcmeTextEditor */
    protected List _parseErrorList = null;
    /** Map of element to error */
    protected  Map _errorMap = null;
    /** Map of element to markers */
    private Map _markerMap = null;
    /** State for external entity to communicate a reload is needed */
    protected int _reloadState = NONE;
    
    private boolean	_firingResourceChanges;
    
    /**
     * Creates an empty ASResource object.
     * @param parent  the parent IASResource object.
     * @param provider  the ASResourceProvider containing this ASResource.
     */
    public ASResource(IASResource parent, ASResourceProvider provider) {
        super();
        
        // initialize the extension points, if necessary
        // NOTE: There is possibility to refresh extension, in which case
        //   this if statement needs to be modified to check for condition to refresh
        if (_dataPersist == null) {
            processExtensions();
        }
        setParent(parent);
        setProvider(provider);
        _listeners = new ArrayList();
        _acmeErrorList = new ArrayList();
        _parseErrorList = new ArrayList();
        _errorMap = new HashMap();
        _markerMap = new HashMap();
        // register listeners
        AcmeErrorManager.addErrorListener(this);
        registerResourceListener();
        registerASProvListener();
    }
    
    /**
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        dispose();
        super.finalize();
    }
    
    /**
     * Adds the error to list to be processed later
     * @see acmeLib.errorManagement.IAcmeErrorListener#errorDetected(acmeLib.elements.AcmeDesign, acmeLib.errorManagement.AcmeError)
     */
    public void errorDetected(AcmeDesign associatedDesign, AcmeError typecheckError) {
        // Limit the markers to "root" errors (so that we don't get multiple errors for deep errors)
        if (associatedDesign == getDesign() && typecheckError.getCauses().getSize() == 0) {
            
            _acmeErrorList.add(typecheckError);
            _errorMap.put(typecheckError.getSource(), typecheckError);
        }
    }
    
    /**
     * @see acmeLib.errorManagement.IAcmeErrorListener#clearErrors(acmeLib.elements.AcmeDesign, acmeLib.AcmeObject)
     */
    public void clearErrors(AcmeDesign associatedDesign, AcmeObject object) {
        if (associatedDesign == getDesign()) {
            AcmeError origError = (AcmeError) _errorMap.get(object);
            if (origError != null) { // remove from lists and map
                _acmeErrorList.remove(origError);
                _parseErrorList.remove(origError);
                _errorMap.remove(origError);
            }
        }
    }
    
    public void clearError (AcmeDesign associatedDesign, AcmeObject object, AcmeError error) {
        if (associatedDesign == getDesign()) {
            AcmeError origError = (AcmeError) _errorMap.get(object);
            if (origError != null && error.equals (origError)) { // remove from lists and map
                _acmeErrorList.remove(origError);
                _parseErrorList.remove(origError);
                _errorMap.remove(origError);
            }
        }
    }
    
    /**
     * @see org.eclipse.core.resources.IResourceChangeListener#resourceChanged(org.eclipse.core.resources.IResourceChangeEvent)
     */
    public void resourceChanged(IResourceChangeEvent event) {
        if (!(ASResource.this instanceof ASDesign ||
              ASResource.this instanceof ASVisualStyle)) {
            // don't bother visiting delta
            return;
        }
        switch (event.getType()) {
        case IResourceChangeEvent.POST_CHANGE : {
            try {
                event.getDelta().accept(new DeltaVisitor());
            } catch (CoreException e) {
                Debug.errorln(e);
            }
            break;
        }}
    }
    
    /**
     * @see java.lang.Comparable#compareTo(Object)
     */
    public int compareTo(Object obj) {
        // cast object to ASResource, so that ClassCastException is triggered if
        // the wrong object instance is supplied.
        ASResource res2 = (ASResource) obj;
        Object resource1 = this.getFile();
        Object resource2 = res2.getFile();
        File file1 = null;
        File file2 = null;
        String str1 = this.toString(); // used if either file1 is null
        String str2 = res2.toString();
        if (resource1 instanceof IFile) {
            file1 = ((IFile) resource1).getLocation().toFile();
        }
        else if (resource1 instanceof File) {
            file1 = (File) resource1;
        }
        if (resource2 instanceof IFile) {
            file2 = ((IFile) resource2).getLocation().toFile();
        }
        else if (resource2 instanceof File) {
            file2 = (File) resource2;
        }
        int compResult = 0;
        if (file1 == null && file2 == null) {
            // special case treat as equal
            compResult = 0;
        }
        else if (file1 == null || file2 == null) {
            // compare strings
            compResult = str1.compareTo(str2);
        }
        else {
            compResult = file1.compareTo(file2);
        }
        return compResult;
    }
    
    /**
     * @see org.eclipse.core.runtime.IAdaptable#getAdapter(Class)
     */
    public Object getAdapter(Class adapter) {
        if (adapter.equals(this.getClass())) {
            return this;
        }
        else {
            return null;
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#dispose()
     */
    public void dispose() {
        // remove this ASResource as resource change listener
        unregisterResourceListener();
        // remove _prov as listener to this
        unregisterASProvListener();
        // stop listening to errors
        AcmeErrorManager.removeErrorListener(this);
        // remove design record if it's just opened but not imported 
        if (!(AcmeDesign.getDesignRecord(getCanonicalPath()) instanceof AcmeImportedDesignRecord)) {
            AcmeDesign.removeDesignRecord(getCanonicalPath());
        }
        // remove mapping of design to this ASResource
        if (getDesign() != null) {
            ASModel.current().removeDesign2ASResource(getDesign());
        }
        if (_listeners != null) {
            synchronized (_listeners) {
                _listeners.clear(); // remove listeners
            }
            _listeners = null;
        }
        _asPath = null;
        if (_acmeErrorList != null) {
            _acmeErrorList.clear();
            _acmeErrorList = null;
        }
        if (_parseErrorList != null) {
            _parseErrorList.clear();
            _parseErrorList = null;
        }
        if (_errorMap != null) {
            _errorMap.clear();
            _errorMap = null;
        }
        if (_markerMap != null) {
            _markerMap.clear();
            _markerMap = null;
        }
        _prov = null;
        _element = null;
        if (_resource != null) {
            ASModel.current().removeFile2ASResource(getCanonicalPath());
            _resource = null;
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#loadFrom(File)
     */
    public void loadFrom(File file) throws CoreException, ParseError, AcmeException {
        setFile(file);
        
        if (file != null) {
            Debug.traceln("ASResource.loadFrom: " + "Loading file " + file.getPath());
            if (file.exists()) {
                try {
                    _asPath = ASPath.pathFor(file.getPath());
                    loadStream(null, new FileInputStream(file));
                } catch (FileNotFoundException e) {
                    throw new CoreException(
                        new ASStatus(ASStatus.ERROR,
                                     ASStatus.FILE_NOT_FOUND,
                                     "ASResource unable to load " + file.getName() + "!",
                                     e));
                }
            }
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#loadFrom(IFile)
     */
    public void loadFrom(IFile file) throws CoreException, ParseError, AcmeException {
        setFile(file);

        if (file != null) {
            Debug.traceln("ASResource.loadFrom: " + "Loading file " + file.getLocation().toOSString());
            if (!file.isSynchronized(IResource.DEPTH_ZERO)) {
                file.refreshLocal(IResource.DEPTH_ZERO, null);
            }

            if (file.exists()) {
                _asPath = ASPath.pathFor(file.getLocation());
                loadStream(null, file.getContents());
            }
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#loadStream(InputStream)
     */
    public void loadStream(InputStream input) throws CoreException, ParseError, AcmeException {
        loadStream(null, input);
    }

    /**
     * @see AcmeStudio.core.resource.IASResource#reload()
     */
    public void reload() throws CoreException, ParseError, AcmeException {
        // check to see if this ASResource was changed
        if (getProvider() != null && getProvider().isDirty()) {
            // yes, so reload from the current design's output stream
            AcmeOutputStream aos = new AcmeOutputStream();
            try {
                getDesign().exportArmani(aos);
                aos.close();
                ByteArrayInputStream bais = new ByteArrayInputStream(aos.toString().getBytes());
                loadStream(bais);
            } catch (IOException e) {
                Debug.errorln(e);
            }
        }
        else {
            // no, we can load from file
            if (_resource instanceof IFile) {
                loadFrom((IFile) _resource);
            } else if (_resource instanceof File) {
                loadFrom((File) _resource);
            }
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#reloadStream(IProgressMonitor, InputStream)
     */
    public void reloadStream(IProgressMonitor monitor, InputStream input)
    throws CoreException, ParseError, AcmeException {
        // make sure we extract persistent data out of design before loading new design from input
        List persistentData = extractPersistentData();
        try {
            // load from the input stream to get new AcmeDesign
            loadStream(monitor, input);
        } catch (ParseError e) {
            throw e;
        } finally {
            // before leaving, inject extracted data back into new design
            injectPersistentData(persistentData);
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#save(IProgressMonitor)
     */
    public boolean save(IProgressMonitor monitor) throws CoreException {
        return doSave(monitor);
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#saveTo(IFile, IProgressMonitor)
     */
    public boolean saveTo(IFile file, IProgressMonitor monitor) throws CoreException {
        setFile(file);

        ByteArrayInputStream in = null;
        if (_prov.isInTextEditor() || getDesign() == null) {
            IDocument doc = _prov.getTextDocument();
            if (doc != null) {  // no document created yet
                in = new ByteArrayInputStream(doc.get().getBytes());
            }
        } else {
            // NOT in text editor and design NOT null
            // Unparse the design into the output stream and set the file contents
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            Helper.getOutputStream(getDesign(), out);
            in = new ByteArrayInputStream(out.toByteArray());
            try {
                out.close();
            } catch (IOException e) {  // ignore this error here
            }
        }
        if (file.exists()) {
            // we can safely assume that the user has verified to replace
            // existing file (?), so we replace the content
            // Nothing to do!
        } else {
            file.create(in, true, monitor);
        }
        if (in != null) {
            try {
                in.close();
            } catch (IOException e) { // ignore this error here
            }
        }
        
        return doSave(monitor);
    }

    /**
     * @see AcmeStudio.core.resource.IASResource#getASPath()
     */
    public ASPath getASPath() {
        return _asPath;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getFile()
     */
    public Object getFile() {
        return _resource;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#setFile(Object)
     */
    public void setFile(Object resource) {
        if (resource == null || resource instanceof File || resource instanceof IFile) {
            
            if (resource != _resource) { // only set if resource is not the same
                // alter the mapping of local file to this IASResource
                ASModel.current().removeFile2ASResource(getCanonicalPath());
                // set new resource
                _resource = resource;
                // add mapping using new outcome of canonical path
                ASModel.current().addFile2ASResource(getCanonicalPath(), this);
                //fireASResourceChanged(IASResourceChangeEvent.SELF_CHANGED);
            }
        } else {
            Debug.warnln("ASResource.setFile():  Invalid resource type supplied: " + resource);
            // assert failure
            CoreException ce = new CoreException(
                new ASStatus(ASStatus.WARNING,
                             ASStatus.RESOURCE_TYPE_UNRECOGNIZED,
                             AcmeStudioPlugin.getMessage("ASResource.resourceTypeUnrecognized "),
                             null));
            throw new RuntimeException(ce);
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getPath()
     */
    public String getPath() {
        String path = null;
        
        if (_resource != null) {
            if (_resource instanceof IFile) {
                path = ((IFile) _resource).getLocation().toString();
            }
            else if (_resource instanceof File) {
                path = ((File) _resource).getPath();
            }
        }
        return path;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getCanonicalPath()
     */
    public String getCanonicalPath() {
        String path = null;
        
        if (_resource != null) {
            try {
                if (_resource instanceof IFile) {
                    // BUGFIX : File could have been moved, in which case the location could be null.
                    IPath location = ((IFile )_resource).getLocation ();
                    if (location != null)
                        path = location.toFile().getCanonicalPath();
                }
                else if (_resource instanceof File) {
                    path = ((File) _resource).getCanonicalPath();
                }
            }
            catch (IOException e) { // just get regular path then
                path = getPath();
            }
        }
        return path;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getParent()
     */
    public IASResource getParent() {
        return _parent;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#setParent(IASResource)
     */
    public void setParent(IASResource parent) {
        _parent = parent;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getProvider()
     */
    public ASResourceProvider getProvider() {
        return _prov;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#setProvider(ASResourceProvider)
     */
    public void setProvider(ASResourceProvider prov) {
        if (_prov != null && _prov != prov) {
            // deal with update of listeners during provider change here
            removeASResourceChangeListener(_prov);
            if (prov != null) {
                addASResourceChangeListener(prov);
            }
        }
        _prov = prov;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getElement()
     */
    public AcmeElement getElement() {
        return _element;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#setElement(AcmeElement)
     */
    public void setElement(AcmeElement element) {
        if (_element == element) return;  // no change!

        _element = element;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getDesign()
     */
    public AcmeDesign getDesign() {
        return (_element instanceof AcmeDesign) ? (AcmeDesign) _element : null;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getFamily()
     */
    public AcmeFamily getFamily() {
        AcmeFamily fam = null;
        if (_element instanceof AcmeDesign && _element != null) {
            fam = Helper.getAcmeFamily((AcmeDesign) _element);
            if (fam == null) {
                // Perhaps this is a system, try to get it's first family
                AcmeSystem sys = Helper.getAcmeSystem((AcmeDesign )_element);
                if (sys != null) {
                    AcmeEnumeration fams = sys.getInstantiatedTypes();
                    if (fams.getSize () > 0) 
                        fam = (AcmeFamily )fams.getNextItem();
                }
            }
        }
        else if (_element instanceof AcmeFamily && _element != null) {
            fam = (AcmeFamily) _element;
        }
        return fam;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getFamilies()
     */
    public List getFamilies() {
        List fams = new ArrayList();
        if (_element instanceof AcmeDesign && _element != null) {
            fams.addAll(Helper.getAcmeFamilies((AcmeDesign) _element));
        }
        return fams;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getImport()
     */
    public AcmeDesign getImport() {
        AcmeDesign design = null;
        if (_element instanceof AcmeDesign && _element != null) {
            design = Helper.getAcmeImport((AcmeDesign) _element);
        }
        return design;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getImports()
     */
    public List getImports() {
        List designs = new ArrayList();
        if (_element instanceof AcmeDesign && _element != null) {
            designs.addAll(Helper.getAcmeImports((AcmeDesign) _element));
        }
        return designs;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getImportRecords()
     */
    public List getImportedRecords() {
        List records = new ArrayList();
        if (_element instanceof AcmeDesign && _element != null) {
            AcmeDesign design = (AcmeDesign) _element;
            records.addAll(Helper.convertAcmeEnumerationToList(design.getImportedDesigns().enumerate()));
        }
        return records;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getSystem()
     */
    public AcmeSystem getSystem() {
        AcmeSystem sys = null;
        if (_element instanceof AcmeDesign && _element != null) {
            sys = Helper.getAcmeSystem((AcmeDesign) _element);
        }
        else if (_element instanceof AcmeSystem && _element != null) {
            // should this happen?
            sys = (AcmeSystem) _element;
        }
        return sys;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#getSystems()
     */
    public List getSystems() {
        List syss = new ArrayList();
        if (_element instanceof AcmeDesign && _element != null) {
            syss.addAll(Helper.getAcmeSystems((AcmeDesign) _element));
        }
        return syss;
    }

    /**
     * @see AcmeStudio.core.resource.IASResource#hasParseError()
     */
    public boolean hasParseError() {
        return _parseErrorList != null && !_parseErrorList.isEmpty();
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#isProjectLevel()
     */
    public boolean isProjectLevel() {
        return _asPath.getPathLevel() == ASPath.PROJECT_LEVEL;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#isGlobalLevel()
     */
    public boolean isGlobalLevel() {
        return _asPath.getPathLevel() == ASPath.GLOBAL_LEVEL;
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#shouldReload()
     */
    public boolean shouldReload() {
        return _reloadState > NONE;
    }

    /**
     * @see AcmeStudio.core.resource.IASResource#setShouldReload(boolean)
     */
    public void setReloadState (int s) {
        _reloadState = s;

        if ((s & OUT_OF_SYNC) > 0) {  // this requires immediate action
            fireASResourceChanged(IASResourceChangeEvent.SELF_CHANGED_EXTERNALLY);
        }
    }

    /**
     * Formats a string so that it is a maximum of width characters wide
     * @param str The string to format
     * @param width The maximum width
     * @return  The formatted string
     */
    public static String formatMarkerString(String str, int width) {
        String retStr = "";
        String currentLine = "";
        if (str.length() < width) {
            return str;
        }
        
        StringTokenizer tokens = new StringTokenizer(str, " ");
        while (tokens.hasMoreTokens()) {
            String token = tokens.nextToken();
            if (token.length() + currentLine.length() > width) {
                retStr += currentLine + "\n";
                currentLine = token + " ";
            }
            else {
                currentLine += token + " ";
            }
        }
        retStr += currentLine;
        return retStr;
    }
    /**
     * @see AcmeStudio.core.resource.IASResource#addMarker(IFile, AcmeError)
     */
    public void addMarker(IFile file, AcmeError error) throws CoreException {
        if (_markerMap.containsKey(error.getSource()))
            return;
        
        // remove this resource as resource listener to prevent infinite trigger
        unregisterResourceListener();
        try {
            IMarker marker = file.createMarker(IMarker.PROBLEM);
            _markerMap.put(error.getSource(), marker);
            // Process error message
            String msg = error.toString();
            if (error instanceof AcmeParseError)
                msg = error.getErrorMessage();
            
            msg = formatMarkerString(msg, 80);
            marker.setAttribute(IMarker.MESSAGE, msg);
            marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
            
            if (error instanceof AcmeParseError) {
                marker.setAttribute(IMarker.LINE_NUMBER, ((AcmeParseError) error).getLineNumber());
            }
            else {
                // query region manager for line region
                AcmeObject acmeObj = (AcmeObject) error.getSource();
                if (acmeObj instanceof AcmeElement) {
                    marker.setAttribute ("AcmeErrorSource", ((AcmeElement )acmeObj).getQualifiedName());
                }
                else if (acmeObj instanceof AcmeProperty) {
                    marker.setAttribute ("AcmeErrorSource", ((AcmeProperty )acmeObj).getQualifiedName());
                }
                else if (acmeObj instanceof DesignRule) {
                    marker.setAttribute ("AcmeErrorSource", ((AcmeElement )((DesignRule )acmeObj).getDesignRuleScope()).getQualifiedName ());
                }
                Region region = null;
                if (getProvider() != null) {
                    region = getProvider().getRegion(acmeObj);
                }
                if (region != null) {
                    marker.setAttribute(IMarker.LINE_NUMBER, region.getBeginLine());
                }
                else {
                    // no parse error line number available, do the next best thing
                    // determine the line number where the word "System" occurs
                    try {
                        BufferedReader reader = new BufferedReader(new FileReader(file.getLocation().toFile()));
                        String KEYWORD = _resource_nls.getString("ASResource.acmeSystemKeyword");
                        int lineCnt = 0;
                        String line = reader.readLine();
                        while (line != null) {
                            lineCnt++;
                            if (line.indexOf(KEYWORD) >= 0) {
                                marker.setAttribute(IMarker.LINE_NUMBER, lineCnt);
                                break;
                            }
                            line = reader.readLine();
                        }
                    }
                    catch (IOException e) {
                        Debug.errorln(e);
                    }
                }
            }
            if (marker.getAttribute(IMarker.LINE_NUMBER, 0) == 0
                    && marker.getAttribute(IMarker.CHAR_START, 0) <= 0
                    && marker.getAttribute(IMarker.CHAR_END, 0) <= 0) {
                
                // final fallback, set to line 1
                marker.setAttribute(IMarker.LINE_NUMBER, 1);
            }
        }
        finally {
            // reinstate this resource as resource listener
            registerResourceListener();
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#removeMarker(AcmeObject)
     */
    public void removeMarker(AcmeObject element) {
        // remove this resource as resource listener to prevent infinite trigger
        unregisterResourceListener();
        try {
            Object m = _markerMap.get(element);
            if (m != null) {
                IMarker marker = (IMarker) m;
                IMarker[] mArray = new IMarker[] { marker };
                try {
                    AcmeStudioPlugin.getWorkspace().deleteMarkers(mArray);
                }
                catch (CoreException e) {
                    Debug.errorln(e);
                }
                _markerMap.remove(element);
                if (_errorMap.containsKey(element)) { // also remove from error map
                    _errorMap.remove(element);
                }
            }
        }
        finally {
            // reinstate this resource as resource listener
            registerResourceListener();
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#processErrors()
     */
    public void processErrors() throws CoreException {
        if (_resource instanceof IFile) {
            IFile ifile = (IFile) _resource;
            // remove all markers from resource first
            ifile.deleteMarkers(IMarker.PROBLEM, true, IResource.DEPTH_INFINITE);
            _markerMap.clear();
            if (_acmeErrorList.isEmpty() & _parseErrorList.isEmpty()) {
                return;
            }
            ArrayList allErrors = new ArrayList();
            allErrors.addAll(_acmeErrorList);
            allErrors.addAll(_parseErrorList);
            Iterator i = allErrors.iterator();
            while (i.hasNext()) {
                Object o = i.next();
                if (o instanceof AcmeError) {
                    AcmeError error = (AcmeError) o;
                    Debug.diagnoseln("ASResource.processErrors: " + error.toString());
                    if (error.getCauses().getSize() == 0) {
                        addMarker(ifile, error);
                    }
                }
                else if (o instanceof ParseException) {
                    processParseException(ifile, (ParseException) o);
                }
            }
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#addASResourceChangeListener(IASResourceChangeListener)
     */
    public void addASResourceChangeListener(IASResourceChangeListener listener) {
        if (_listeners == null) return;

        synchronized (_listeners) {
            if (_firingResourceChanges)
                Debug.diagnoseln ("ASResource.addASResourceChangeListener: " + "must queue work");
            _listeners.add(listener);
        }
    }
    
    /**
     * @see AcmeStudio.core.resource.IASResource#removeASResourceChangeListener(IASResourceChangeListener)
     */
    public void removeASResourceChangeListener(IASResourceChangeListener listener) {
        if (_listeners == null) return;

        synchronized (_listeners) {
            if (_firingResourceChanges) 
                Debug.diagnoseln ("ASResource.removeASResourceChangeListener: " + "must queue work");
            _listeners.remove(listener);
        }
    }
    
    /**
     * Notify registered listeners of resource change.
     */
    protected void fireASResourceChanged(int type) {
        ASResourceChangeEvent evt = new ASResourceChangeEvent(this, type);
        synchronized (_listeners) {
            _firingResourceChanges = true;
            for (int i = 0; i < _listeners.size(); i++) {
                ((IASResourceChangeListener )_listeners.get (i)).resourceChanged (evt);
            }
            _firingResourceChanges = false;
        }
    }
    
    /**
     * Sets whether an ancestor resource of ASResource has changed
     */
    protected void doAncestorChanged() {
        AcmeOpenedDesignRecord rec = AcmeDesign.getDesignRecord(getCanonicalPath());
        if (rec != null) {
            // mark ASResource's design record for design reload
            rec.setAncestorChanged(true);
        }
        fireASResourceChanged(IASResourceChangeEvent.ANCESTOR_CHANGED);
    }
    
    /**
     * Notifies that this ASResource's resource was changed.  Currently, this
     * is called by the DeltaVisitor in this resource, and need not cause any action.
     */
    protected void doSelfChanged() {
        fireASResourceChanged(IASResourceChangeEvent.SELF_CHANGED);
    }

    /**
     * The work horse to load resource, subclass must implement.
     * @param monitor  the progress monitor to indicate progress
     * @param input  the input stream from which to load
     * @throws CoreException
     * @throws ParseError
     */
    protected abstract void loadStream(IProgressMonitor monitor, InputStream input)
    throws CoreException, ParseError, AcmeException;

    /* comment this out after JUnit run */
    public void _test_loadStream (IProgressMonitor monitor, InputStream input) 
    throws CoreException, ParseError, AcmeException {
        loadStream(monitor, input);
    }

    /**
     * The actual work horse to save content, by polymorphism.
     * @param monitor  the progress monitor to indicate progress
     * @return <code>True</code> if save succeeded, <code>false</code> otherwise.
     * @throws CoreException
     */
    protected abstract boolean doSave(IProgressMonitor monitor)
    throws CoreException;

    /* comment this out after JUnit run */
    public List _test_extractPersistentData () { return extractPersistentData(); }
    protected abstract List extractPersistentData ();
    protected abstract void injectPersistentData (List dataList);

    /**
     * Collects and returns a list of file names of the imported designs.
     * The list is empty if there are no imports.
     * @return List
     */
    protected List computeImportFiles(AcmeDesign design) {
        List importedFiles = new ArrayList();
        if (design != null) {
            AcmeEnumeration enum = design.getImportedDesigns().enumerate();
            Iterator iter = Helper.convertAcmeEnumerationToList(enum).iterator();
            while (iter.hasNext()) {
                AcmeImportedDesignRecord rec = (AcmeImportedDesignRecord) iter.next();
                String pathname = rec.getCanonicalFilename();
                if (pathname != null) {
                    importedFiles.add(pathname);
                    // recursively add import files of imported design
                    importedFiles.addAll(computeImportFiles(rec.getResolvedDesign()));
                }
            }
        }
        return importedFiles;
    }
    
    /**
     * Subclass (or extensions) can extend/override this method to add more
     * conditions of what relevant resource changes should have an effect on
     * whether this ASResource gets reloaded.
     * @param filepath  path of the file resource to check if relevant
     * @return boolean  whether the resource in the supplied file path is a
     *                   relevant resource
     */
    protected boolean isRelevantResourceChanged(String filepath) {
        boolean bool = false;
        List importFiles = computeImportFiles(getDesign());
        if (importFiles.contains(filepath)) {
            bool = true;
        }
        if (ASVisualStyle.nullInstance().isSameType(filepath)) {
            // path is a visual style, so let's see if it corresponds with
            // a family that is in this design's import branch
            int dotPos = filepath.lastIndexOf(".");
            String designFile = filepath.substring(0, dotPos + 1) + ASDesign.FILE_EXT;
            if (importFiles.contains(designFile)) {
                bool = true;
            }
        }
        
        return bool;
    }
    
    protected String getFileExt(Object resource) {
        String ext = null;
        
        if (resource instanceof IFile) {
            IFile file = (IFile) resource;
            ext = file.getFileExtension();
        }
        else if (resource instanceof File) {
            File file = (File) resource;
            String filename = file.getName();
            int pos = filename.lastIndexOf("."); // find position of extension
            ext = filename.substring(pos + 1);
        }
        else if (resource instanceof String) {
            String filename = (String) resource;
            int pos = filename.lastIndexOf("."); // find position of extension
            ext = filename.substring(pos + 1);
        }
        
        return ext;
    }
    
    protected String getParentDirName(Object resource) {
        String folderName = null;
        
        if (resource instanceof IFile) {
            IFile file = (IFile) resource;
            IPath folder = file.getParent().getFullPath();
            folderName = folder.segment(folder.segmentCount() - 1);
        }
        else if (resource instanceof File) {
            File file = (File) resource;
            folderName = file.getParentFile().getName();
        }
        
        return folderName;
    }
    
    /**
     * Returns the last modified time of the underlying java.io.File.
     * @return long
     */
    protected long getLastModified() {
        long lastMod = 0;
        
        if (_resource != null) {
            if (_resource instanceof IFile) {
                lastMod = ((IFile) _resource).getLocation().toFile().lastModified();
            }
            else if (_resource instanceof File) {
                lastMod = ((File) _resource).lastModified();
            }
        }
        return lastMod;
    }

    protected void processParseException(IFile file, ParseException pErr) throws CoreException {
        IMarker marker = file.createMarker(IMarker.PROBLEM /*ACME_ERROR_TAG*/
        );
        marker.setAttribute(IMarker.MESSAGE, pErr.getMessage());
        marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
        // BugFix #125 Sometimes, the parse error doesn't have a token. Make sure to
        // take this into account
        if (pErr.currentToken != null) {
            marker.setAttribute(IMarker.LINE_NUMBER, pErr.currentToken.beginLine);
        }
        else {
            marker.setAttribute(IMarker.LINE_NUMBER, 1);
        }
    }
    
    protected void registerResourceListener() {
        ResourcesPlugin.getWorkspace().addResourceChangeListener(this, IResourceChangeEvent.POST_CHANGE);
    }
    
    protected void unregisterResourceListener() {
        ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
    }
    
    protected void registerASProvListener() {
        if (_prov != null) {
            addASResourceChangeListener(_prov);
        }
    }
    
    protected void unregisterASProvListener() {
        if (_prov != null) {
            removeASResourceChangeListener(_prov);
        }
    }
    
    public IPath getDataDirectory () throws IOException {
        IPath path = null;
        if (_resource instanceof IFile) {
            IFile file = (IFile )_resource;
            path = file.getLocation();
            if (!path.toFile().isDirectory()) {
                path = path.removeLastSegments(1);
            }
        }
        else if (_resource instanceof File) {
            File file = (File )_resource;
            path = new Path (file.getCanonicalPath());
            path = path.removeLastSegments(1);
            
        }
        path = path.append (".data");
        
        File file = path.toFile ();
        if (file.exists()) {
            if (file.isDirectory()) {
                return path;
            } 
            else {
                throw new IOException (path.toOSString() + " is not a directory");
            }
        }
        else {
            file.mkdir();
        }
        return path;
    }

}
