package AcmeStudio.core.resource;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.StringWriter;
import java.util.Collections;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.preference.IPreferenceStore;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASStatus;
import AcmeStudio.core.resource.provider.ASResourceProvider;
import AcmeStudio.ui.editors.diagram.styles.IVisualStyleListener;
import AcmeStudio.ui.editors.diagram.styles.VisualStyle;
import AcmeStudio.util.Debug;
import AcmeStudio.util.Utils;
import acmeLib.elements.AcmeFamily;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * AcmeStudio Visual Style resource type.
 */
public class ASVisualStyle extends ASResource implements IVisualStyleListener, IASResource {

    /** File extension of the ASVisualStyle resource */
    public static String FILE_EXT = _resource_nls.getString("ASResource.visualStyleExt");
	public static final String SAVEGLOBAL = "SaveGlobal";
	public static final String SAVELOCAL = "SaveLocal";
	public static final String READONLY = "ReadOnly";
    public static IASResource _nullInstance = null;

    /** the AcmeFamily corresponding to the stored visual style */
    private AcmeFamily _family = null;
    /** the Visual Style being managed */
    private VisualStyle _visualStyle = null;


    /**
     * Creates an empty ASVisualStyle object.
     * @param parent  the parent IASResource object.
     * @param asFam  the associated family ASResource
     */
    public ASVisualStyle (IASResource parent) {
        super(parent, null);
        
        if (parent != null) {
            _family = parent.getFamily();
        }
    }

    /**
     * Returns a null instance of this type.  This method is useful for passing
     * IASResource objects around to identify the resource "types," as used, for
     * example, in ASRegistry.
     * @return IASResource  the null instance
     */
    public static IASResource nullInstance () {
        if (_nullInstance == null) {  // lazy init
            _nullInstance = new ASVisualStyle(null);
        }
        return _nullInstance;
    }

    /**
     * Callback by the associated VisualStyle when its content changes.
     * @see AcmeStudio.ui.editors.diagram.styles.IVisualStyleListener#changeVisualStyle(AcmeStudio.ui.editors.diagram.styles.VisualStyle)
     */
    public void changeVisualStyle(VisualStyle style) {
		if (_visualStyle != null && _visualStyle.equals(style)) {
            // find the root parent ASDesign
            IASResource root = getParent();
            while (root != null && root.getParent() != null) {
               root = root.getParent();
            }
            // then access the ASResourceProvider
            if (root != null) {
                ASResourceProvider prov = root.getProvider();
                if (prov != null) {
                    prov.setDirty(true);
                }
            }
            _visualStyle.setDirty(true);
            setReloadState(EDITED);
        }
	}

    /**
     * @see AcmeStudio.core.resource.IASResource#dispose()
     */
    public void dispose() {
        super.dispose();
        if (_visualStyle != null) {
            _visualStyle.removeVisualStyleListener(this);
            _visualStyle = null;
        }
    }

    /**
     * @see AcmeStudio.core.resource.IASResource#getNewInstance(IASResource)
     */
    public IASResource getNewInstance (IASResource parent) throws CoreException {
        return new ASVisualStyle(parent);
    }

    /**
     * @see AcmeStudio.core.resource.IASResource#isSameType(Object)
     */
    public boolean isSameType (Object resource) {
        boolean retVal = false;

        String ext = getFileExt(resource);
        if (ext != null && ext.equals(FILE_EXT)) {
            // matches
            retVal = true;
        }

        return retVal;
    }

    /**
     * Gets the AcmeFamily associated with this visual style
     * @see AcmeStudio.core.resource.IASResource#getFamily()
     */
    public AcmeFamily getFamily() {
        return _family;
    }

    /**
     * Returns the underlying VisualStyle object.
     * @return VisualStyle
     */
    public VisualStyle getVisualStyle () {
        return _visualStyle;
    }

    /**
     * Sets the underlying VisualStyle object directly.
     * @param vstyle  the new VisualStyle to set
     */
    public void setVisualStyle (VisualStyle vstyle) {
        if (_visualStyle != null) {
            _visualStyle.removeVisualStyleListener(this);
        }
        _visualStyle = vstyle;
        if (vstyle != null) {
            vstyle.addVisualStyleListener(this);
        }
    }

    /**
     * Loads Visual Style from stream.
     */
    protected synchronized void loadStream (IProgressMonitor monitor, InputStream input)
    throws CoreException {
        try {
            DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            Document doc = docBuilder.parse(input);
            setVisualStyle(VisualStyle.importVisualStyle(doc.getDocumentElement()));
        } catch (ParserConfigurationException e) {
            Debug.errorln(e);
        } catch (FactoryConfigurationError e) {
            Debug.errorln(e);
        } catch (SAXException e) {
            Debug.errorln(e);
        } catch (IOException e) {
            Debug.errorln(e);
        }
    }

    /**
     * @see AcmeStudio.core.resource.ASResource#doSave(IProgressMonitor)
     */
    protected boolean doSave (IProgressMonitor monitor) throws CoreException {
        boolean successVis = false;

        // get and export the visual style to stream
        //  - [SWC 15-Feb-2005] This shouldn't be retrieved from ASModel
        //VisualStyle vstyle = ASModel.current().getVisualStyle(getFamily());
        VisualStyle vstyle = getVisualStyle();
        if (vstyle == null) {  // nothing to save
            successVis = true;
            Debug.warnln("ASVisualStyle.doSave(): vstyle is null, nothing to save.");
        } else if (! vstyle.isDirty()) {  // no need to save
            successVis = true;
            Debug.traceln("ASVisualStyle.doSave(): vstyle is not dirty, no need to save.");
        } else {
            Debug.traceln("ASVisualStyle.doSave(): exporting vstyle...");
            StringWriter sw = null;
            ByteArrayInputStream vsIn = null;
            try {
                DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
                Document doc = docBuilder.newDocument();
                Element elem = vstyle.export(doc);
                // write XML to string, formatted
                sw = new StringWriter();
				TransformerFactory tFactory = TransformerFactory.newInstance ();
                Transformer transformer = tFactory.newTransformer ();
                DOMSource source = new DOMSource (elem);
                StreamResult result = new StreamResult (sw);
                transformer.transform (source, result);
               
                vsIn = new ByteArrayInputStream(sw.toString().getBytes());
                Debug.traceln(getFamily().getName() + " elem ==\n" + sw.toString());
            } catch (ParserConfigurationException e) {
                Debug.errorln(e);
            } catch (FactoryConfigurationError e) {
                Debug.errorln(e);
            } catch (TransformerConfigurationException e) {
				Debug.errorln (e);
			} catch (TransformerException e) {
				Debug.errorln (e);;
			}

            // save style to file
            if (vsIn == null) {  // nothing to save
                successVis = true;
                Debug.warnln("ASVisualStyle.doSave(): vsIn is null, can't save.");
            } else {
                if (_resource == null) {
                    Debug.traceln("ASVisualStyle.doSave(): Visual style is new...");
                    // TODO Need to determine if visual style file is global-level 
                    if (getParent().getFile() == null) {
                        throw new CoreException (new ASStatus(
                            ASStatus.ERROR,
                            ASStatus.NULL_ASRESOURCE_FILE,
                            AcmeStudioPlugin.getMessage("ASVisualStyle.nullASFamilyFile") +
                            " " + getParent().getASPath().getPathStr() + "!",
                            null));
                    } else if (getParent().getFile() instanceof IFile) {
                        Debug.traceln("ASVisualStyle.doSave(): Saving visual style as IFile...");
                        IFile famFile = (IFile) getParent().getFile();
                        IPath path = famFile.getParent().getFullPath();
						IFolder dir = (IFolder) Utils.validateFamilyPath(path);
                        // check for existence
                        if (! dir.exists()) {
                            Debug.traceln("ASVisualStyle.doSave(): Creating directory " + dir.getName() + "...");
                            dir.create(true, true, monitor);
                        }
                        // obtain the IFile object and create it
                        IFile vfile = null;
                        if (getFamily() != null) {
                            String name = getFamily().getName() + "." + FILE_EXT;
                            vfile = dir.getFile(name);
                        } else {
                            throw new CoreException (new ASStatus(
                                ASStatus.ERROR,
                                ASStatus.NULL_ACME_ELEMENT,
                                AcmeStudioPlugin.getMessage("ASVisualStyle.nullAcmeFamily") +
                                " " + getParent().getASPath().getPathStr() + "!",
                                null));
                        }
                        if (vfile.exists()) {
                            vfile.setContents(vsIn, true, true, monitor);
                            Debug.traceln("ASVisualStyle.doSave(): Visual style file overwritten.");
                        } else {
                            vfile.create(vsIn, true, monitor);
                            Debug.traceln("ASVisualStyle.doSave(): Visual style file created.");
                        }
                        setFile(vfile);  // store new file
                        successVis = true;
                    } else if (getParent().getFile() instanceof File) {
                        Debug.traceln("ASVisualStyle.doSave(): Saving visual style as File...");
                        File famFile = (File) getParent().getFile();
                        File parent = famFile.getParentFile();
                        if (! parent.exists()) {
                            Debug.traceln("ASVisualStyle.doSave(): Creating directory " + parent.getName() + "...");
                            if (! parent.mkdir()) {
                                throw new CoreException(new ASStatus(
                                    ASStatus.ERROR,
                                    ASStatus.DIRECTORY_CREATION_FAILED,
                                    AcmeStudioPlugin.getMessage("ASResource.directoryCreationFailed") +
                                    " " + parent.getPath() + "!",
                                    null));
                            }
                        }
                        // obtain the IFile object and create it
                        File vfile = null;
                        if (getFamily() != null) {
                            String name = getFamily().getName() + "." + FILE_EXT;
                            vfile = new File(parent, name);
                        } else {
                            throw new CoreException (new ASStatus(
                                ASStatus.ERROR,
                                ASStatus.NULL_ACME_ELEMENT,
                                AcmeStudioPlugin.getMessage("ASVisualStyle.nullAcmeFamily") +
                                " " + getParent().getASPath().getPathStr() + "!",
                                null));
                        }
                        try {
                            PrintStream pstream = new PrintStream(new FileOutputStream(vfile));
                            pstream.write(sw.toString().getBytes());
                            Debug.traceln("ASVisualStyle.doSave(): Visual style file saved.");
                        } catch (FileNotFoundException e) {
                            throw new CoreException(new ASStatus(
                                ASStatus.ERROR,
                                ASStatus.FILE_NOT_FOUND,
                                AcmeStudioPlugin.getMessage("ASResource.fileSaveFailed") +
                                " " + vfile.getPath() + "!",
                                e));
                        } catch (IOException e) {
                            throw new CoreException(new ASStatus(
                                ASStatus.ERROR,
                                ASStatus.IO_ERROR,
                                AcmeStudioPlugin.getMessage("ASResource.fileSaveFailed") +
                                " " + vfile.getPath() + "!",
                                e));
                        }
                        setFile(vfile);  // store new file
                        successVis = true;
                    } else {  // should NOT happen
                        throw new CoreException(new ASStatus(
                            ASStatus.ERROR,
                            ASStatus.RESOURCE_TYPE_UNRECOGNIZED,
                            AcmeStudioPlugin.getMessage("ASResource.resourceTypeUnrecognized"),
                            null));
                    }
                } else if (_resource instanceof IFile) {
                    // we can rest assured that this visual style resource
                    // would NOT be in the read-only global path
                    IFile ifile = (IFile) _resource;
                    if (ifile.exists()) {
                        ifile.setContents(vsIn, true, true, monitor);
                        Debug.traceln("ASVisualStyle.doSave(): Visual style IFile exists and saved...");
                    } else {
                        ifile.create(vsIn, true, monitor);
                        Debug.traceln("ASVisualStyle.doSave(): Visual style IFile did not exist; created...");
                    }
                    successVis = true;
                } else if (_resource instanceof File) {
                    // TODO Visual style file likely to be global-level
                	// Check preferences to work out whether the style should be saved at the local
                	// or global level
                	IPreferenceStore store = AcmeStudioPlugin.getDefault().getPreferenceStore();
                	if (store.getString ("SaveGlobalVisPref").equals (SAVEGLOBAL)) {
                		Debug.traceln("ASVisualStyle.doSave: " + "Saving global file.");
                		File vfile = (File) _resource;
                		try {
                			PrintStream pstream = new PrintStream(new FileOutputStream(vfile));
                			pstream.write(sw.toString().getBytes());
                			Debug.traceln("ASVisualStyle.doSave(): Visual style File exists and saved...");
	                    } catch (FileNotFoundException e) {
	                        throw new CoreException(new ASStatus(
	                            ASStatus.ERROR,
	                            ASStatus.FILE_NOT_FOUND,
	                            AcmeStudioPlugin.getMessage("ASResource.fileSaveFailed") +
	                            " " + vfile.getPath() + "!",
	                            e));
	                    } catch (IOException e) {
	                        throw new CoreException(new ASStatus(
	                            ASStatus.ERROR,
	                            ASStatus.IO_ERROR,
	                            AcmeStudioPlugin.getMessage("ASResource.fileSaveFailed") +
	                            " " + vfile.getPath() + "!",
	                            e));
	                    }
                	} else if (store.getString ("SaveGlobalVisPref").equals (SAVELOCAL)) {
                		Debug.traceln("ASVisualStyle.doSave: " + "Saving visual style locally");
						IFile famFile = (IFile) getParent().getFile();
						IPath path = famFile.getParent().getFullPath();
						IFolder dir = (IFolder) Utils.validateFamilyPath(path);
						// check for existence
						if (! dir.exists()) {
							Debug.traceln("ASVisualStyle.doSave(): Creating directory " + dir.getName() + "...");
							dir.create(true, true, monitor);
						}
						// obtain the IFile object and create it
						IFile vfile = null;
						if (getFamily() != null) {
							String name = getFamily().getName() + "." + FILE_EXT;
							vfile = dir.getFile(name);
						} else {
							throw new CoreException (new ASStatus(
								ASStatus.ERROR,
								ASStatus.NULL_ACME_ELEMENT,
								AcmeStudioPlugin.getMessage("ASVisualStyle.nullAcmeFamily") +
								" " + getParent().getASPath().getPathStr() + "!",
								null));
						}
						if (vfile.exists()) {
							vfile.setContents(vsIn, true, true, monitor);
							Debug.traceln("ASVisualStyle.doSave(): Visual style file overwritten.");
						} else {
							vfile.create(vsIn, true, monitor);
							Debug.traceln("ASVisualStyle.doSave(): Visual style file created.");
						}
						setFile(vfile);  // store new file
						successVis = true;	
                	}
                	
	                successVis = true;
                } else {  // should NOT happen
                    throw new CoreException(new ASStatus(
                        ASStatus.ERROR,
                        ASStatus.RESOURCE_TYPE_UNRECOGNIZED,
                        AcmeStudioPlugin.getMessage("ASResource.resourceTypeUnrecognized"),
                        null));
                }
            }
        }

        // clear changed bit
        if (successVis) {
            setReloadState(NONE);
        }
        return successVis;
    }

    /**
     * Store the new family object
     * @see AcmeStudio.core.resource.ASResource#doAncestorChanged()
     */
    protected void doAncestorChanged() {
        AcmeFamily newFam = getParent().getFamily(); 
        if (newFam != null && newFam != _family) {
            _family = newFam;
        }

        super.doAncestorChanged();
    }

    /**
     * Computes whether the resource indicated by the supplied path is
     * relevant to this visual style.
     * @param filepath  path of the file resource to check if relevant
     * @return boolean  whether the resource in the supplied file path is a
     *                   relevant resource
     * @see AcmeStudio.core.resource.ASResource#isRelevantResourceChanged(String)
     */
    protected boolean isRelevantResourceChanged(String filepath) {
        boolean bool = false;
        if (ASFamily.nullInstance().isSameType(filepath)) {
            // path is a family, so let's see if it corresponds with this vis style
            int dotPos = filepath.lastIndexOf(".");
            String visFile = filepath.substring(0, dotPos + 1) + ASVisualStyle.FILE_EXT;
            if (visFile.equals(getCanonicalPath())) {
                bool = true;
            }
        }
        
        return bool;
    }

    protected List extractPersistentData () {
        return Collections.EMPTY_LIST;
    }

    protected void injectPersistentData (List dataList) {
    }

}
