package AcmeStudio.core.resource;

import java.io.File;
import java.io.InputStream;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;

import AcmeStudio.core.resource.provider.ASResourceProvider;
import acmeLib.AcmeObject;
import acmeLib.armaniParser.ParseError;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeFamily;
import acmeLib.elements.AcmeSystem;
import acmeLib.errorManagement.AcmeError;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * AcmeStudio Resource interface.
 */
public interface IASResource extends IAdaptable, Comparable {

    public final static int NONE = 0x00;
    public final static int EDITED = 0x01;
    public final static int OUT_OF_SYNC = 0x02;

    /**
     * Returns a new instance of this class.  Convenience method for creating
     * new instance from a null instance.
     * @param parent  the parent of the new IASResource instance
     * @return IASResource  new instance of IASResource
     */
    public abstract IASResource getNewInstance (IASResource parent) throws CoreException;

    /**
     * Cleans up references this IASResource of other objects and unregister
     * from listened events.
     */
    public void dispose ();

    /**
     * Loads from file an AcmeStudio resource.
     * @param file  the file to load from
     */
    public void loadFrom (File file) throws CoreException, ParseError, AcmeException;

    /**
     * Loads from file an AcmeStudio resource.
     * @param file  the file to load from
     */
    public void loadFrom (IFile file) throws CoreException, ParseError, AcmeException;

    /**
     * Method loadStream without progress monitor.
     * @param input  the input stream from which to load
     */
    public void loadStream (InputStream input) throws CoreException, ParseError, AcmeException;

    /**
     * Reloads this IASResource.
     */
    public void reload () throws CoreException, ParseError, AcmeException;

    /**
     * Reloads design using the input stream.
     * This method helps keep the loadStream with param IProgressMonitor
     * non-public, and is intended for use after a loadFrom has been called,
     * although there's no assertion or other enforcement to ensure this.
     * @param monitor  progress monitor to use
     * @param input  the input stream from which to load
     */
    public void reloadStream (IProgressMonitor monitor, InputStream input) throws CoreException, ParseError, AcmeException;

    /**
     * Saves resource content to current file.
     * @param monitor  progress monitor to use
     * @return boolean  whether save succeeeded
     */
    public boolean save (IProgressMonitor monitor) throws CoreException;

    /**
     * Save resource content to designated file.
     * @param file  the file to save to
     * @param monitor  progress monitor to use
     * @return boolean  whether save succeeeded
     */
    public boolean saveTo (IFile file, IProgressMonitor monitor) throws CoreException;

    /**
     * Returns the path in terms of AcmeStudio's path variable.
     * @return ASPath  the AcmeStudio ASPath
     */
    public ASPath getASPath ();

    /**
     * Returns the file object underlying the ASResource
     * @return Object  the file object underlying the ASResouce
     */
    public Object getFile ();

    /**
     * Sets the new file object underlying the ASResource
     * @param Object  the new file object underlying the ASResouce
     */
    public void setFile (Object resource);

    /**
     * Returns the filesystem path.
     * @return String  the filesystem
     */
    public String getPath ();

    /**
     * Returns the canonical path of the underlying java.io.File.
     * @return String
     */
    public String getCanonicalPath ();

    /**
     * Returns the parent IASResource object.
     * @return IASResource  the parent IASResource object
     */
    public IASResource getParent ();

    /**
     * Sets the parent IASResource.
     * @param parent  The new parent IASResource to set
     */
    public void setParent (IASResource parent);

    /**
     * Returns the ASResourceProvider object.
     * @return ASResourceProvider  the provider object
     */
    public ASResourceProvider getProvider ();

    /**
     * Sets the ASResourceProvider object.
     * @param prov  The new ASResourceProvider to set
     */
    public void setProvider (ASResourceProvider prov);

    /**
     * Returns the internal AcmeElement.
     * @return AcmeElement  the internal AcmeElement
     */
    public AcmeElement getElement();

    /**
     * Sets the AcmeElement.
     * @param element  The new AcmeElement to set
     */
    public void setElement (AcmeElement element);

    /**
     * Returns the acme design element held in this IASResource object.
     * @return AcmeDesign  the design held
     */
    public AcmeDesign getDesign ();

    /**
     * Returns the first family in the families of the acme design element held
     * in this IASResource object.
     * @return AcmeFamily  first family of the design
     */
    public AcmeFamily getFamily ();

    /**
     * Returns the list of families of the acme design element held
     * in this IASResource object.
     * @return List  list of families of the design
     */
    public List getFamilies ();

    /**
     * Returns the first imported design in the imports of the acme design
     * element held in this IASResource object.
     * @return AcmeDesign  first imported design of the design
     */
    public AcmeDesign getImport ();

    /**
     * Returns the list of imported designs of the acme design element held
     * in this IASResource object.
     * @return List  list of imported designs of the design
     */
    public List getImports ();

    /**
     * Returns the list of imported design records of the acme design element
     * held in this IASResource object.
     * @return List  list of imported design records of the design
     */
    public List getImportedRecords ();

    /**
     * Returns the first system in the acme design element held in this 
     * IASResource object.
     * @return AcmeSystem  first system in the design
     */
    public AcmeSystem getSystem ();

    /**
     * Returns the list of systems of the acme design element held
     * in this IASResource object.
     * @return List  list of systems in the design
     */
    public List getSystems ();

    /**
     * Returns whether the acme design element held in this IASResource object
     * has any parser error.
     * @return boolean  flag indicating whether there's parse error
     */
    public boolean hasParseError ();

    /**
     * Returns whether the ASPath is a project level path.
     * @return boolean  flag indicating wither path is project-level
     */
    public boolean isProjectLevel ();

    /**
     * Returns whether the ASPath is a global level path.
     * @return boolean  flag indicating wither path is global-level
     */
    public boolean isGlobalLevel ();

    /**
     * Returns whether this ASResource should be reloaded
     * @return  <code>true</code> if this ASResource should be reloaded,
     *           <code>false</code> otherwise
     */
    public boolean shouldReload ();

    /**
     * Pushes a need-to-reload state change from the external.
     * There are currently two:
     * <ol><li>Dirty (IASResource.EDITED): content has been edited
     * <li>Out of synch (IASResource.OUT_OF_SYNC): ASResource is out of synchronization with underlying resource
     * </ol>
     * Both state changes indicates a reload should be done eventually.
     */
    public void setReloadState (int s);

    /**
     * Adds error marker to the IFile.
     * @param file  the file to add error marker to
     * @param error  the error object to use to generate error markers
     * @throws CoreException
     */
    public void addMarker (IFile file, AcmeError error) 
    throws CoreException;

    /**
     * Removes marker corresponding to the supplied AcmeObject.
     * @param element  the AcmeObject for which marker to remove.
     */
    public void removeMarker (AcmeObject element);

    /**
     * Process any errors encountered during parsing in AcmeLib.
     */
    public void processErrors () throws CoreException;

    /**
     * Returns whether the supplied IResource is of the same type as this
     * IASResource, where type is defined by the file extension and possibly
     * the path.
     * @param resource  resource to apply the test
     * @return boolean  whether the resource is the right type
     */
    public abstract boolean isSameType (Object resource);

    /**
     * Add listener of IASResource change.
     * @param listener  the listener to register
     */
    public void addASResourceChangeListener (IASResourceChangeListener listener);

    /**
     * Remove listener of IASResource change.
     * @param listener  the listener to unregister
     */
    public void removeASResourceChangeListener (IASResourceChangeListener listener);

}
