package AcmeStudio.core.resource.provider;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentPartitioner;
import org.eclipse.jface.text.ITypedRegion;
import org.eclipse.jface.text.rules.DefaultPartitioner;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IFileEditorInput;
import org.eclipse.ui.IPartListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchPartReference;
import org.eclipse.ui.actions.WorkspaceModifyOperation;
import org.eclipse.ui.editors.text.FileDocumentProvider;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.views.navigator.ResourceNavigator;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASContext;
import AcmeStudio.core.ASModel;
import AcmeStudio.core.ASStatus;
import AcmeStudio.core.resource.ASDesign;
import AcmeStudio.core.resource.IASResource;
import AcmeStudio.core.resource.IASResourceChangeEvent;
import AcmeStudio.core.resource.IASResourceChangeListener;
import AcmeStudio.ui.editors.MultiPageAcmeEditor;
import AcmeStudio.ui.editors.text.AcmePartitionScanner;
import AcmeStudio.ui.editors.text.TextRegionManager;
import AcmeStudio.ui.editors.text.filters.FilteredDesignSynchronizer;
import AcmeStudio.util.ASEditorInput;
import AcmeStudio.util.AcmeModelChangeListener;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Debug;
import AcmeStudio.util.Helper;
import AcmeStudio.util.AcmeStudioChangeManager.AcmeStudioChangeEvent;
import acmeLib.AcmeObject;
import acmeLib.armaniParser.ParseError;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.Region;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeFamily;
import acmeLib.elements.AcmeSystem;
import acmeLib.tools.GraphWalker;
import acmeLib.tools.WalkerElementOperation;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Encapsulates and centrally manages a set of Acme design files within a
 * resource provider facade to interact with the MultiPageAcmeEditor (MPAE).
 * The ASResourceProvider also incorporates the DocumentProvider used by the
 * AcmeTextEditor.<br>
 * <br>
 * To manage the Acme content, this provider keeps a set of states, listens to
 * resource and part changes, as well as document (ASDocumentListener) and
 * element changes (ASElementStateListener).
 * <br>
 * See the comments for ASResourceState for the documentation on the major
 * state machine realized to keep source editing working smoothly.
 * <br>
 * Past To-Do's and what was done:<ul>
 * <li>Eventually create ElementView inside MultiPageAcmeEditor<br>
 *     - This is already the case.</li>
 * <li>Move change manager into here so the editors listen to it.<br>
 *     - Kept the editors change listeners.</li>
 * <li>Move design rule manager into here so the editors listen to it.<br>
 *     - Actually, moved to ASModel.</li>
 * <li>Given an IFile, determine what other files are needed<br>
 *     - Not sure if this is still needed</li>
 * <li>Merge content for editor, split content for saving<br>
 *     - Editor support and saving taken care of by AcmeLib and ASResource</li>
 * </ul>
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 */
public class ASResourceProvider extends FileDocumentProvider
implements IAdaptable, AcmeModelChangeListener, IASResourceChangeListener {
    
    /**
     * Listener to part activations.
     */
    public class PartListener implements IPartListener {
        /**
         * @see org.eclipse.ui.IPartListener#partActivated(IWorkbenchPart)
         */
        public void partActivated(IWorkbenchPart part) {
            if (part.equals(_editor)) {
                setEditorActive(true);
            }
        }
        
        /**
         * @see org.eclipse.ui.IPartListener#partBroughtToTop(IWorkbenchPart)
         */
        public void partBroughtToTop(IWorkbenchPart part) {
        }
        
        /**
         * @see org.eclipse.ui.IPartListener#partClosed(IWorkbenchPart)
         */
        public void partClosed(IWorkbenchPart part) {
            if (part.equals(_editor)) {
                setEditorActive(false);
                disconnect();
            }
        }
        
        /**
         * @see org.eclipse.ui.IPartListener#partDeactivated(IWorkbenchPart)
         */
        public void partDeactivated(IWorkbenchPart part) {
            if (part.equals(_editor)) {
                setEditorActive(false);
            }
        }
        
        /**
         * @see org.eclipse.ui.IPartListener#partOpened(IWorkbenchPart)
         */
        public void partOpened(IWorkbenchPart part) {
        }
        
    }
    
    /**
     * Listener to resource changes.
     * Used to track move and delete of underlying resource.
     */
    public class ResourceChangeListener implements IResourceChangeListener {
        /* (non-Javadoc)
         * @see org.eclipse.core.resources.IResourceChangeListener#resourceChanged(org.eclipse.core.resources.IResourceChangeEvent)
         */
        public void resourceChanged(IResourceChangeEvent event) {
            switch (event.getType()) {
            case IResourceChangeEvent.POST_CHANGE:
                try {
                    event.getDelta().accept(new DeltaVisitor());
                } catch (CoreException e) {
                    Debug.errorln(e);
                }
                break;
            }
        }
    }

    /**
     * The Visitor pattern applied to handle IResource change notifications from
     * the Eclipse framework.  We're mainly interested in knowing whether file
     * got renamed, removed, or content got changed.
     */
    public class DeltaVisitor implements IResourceDeltaVisitor {
        public boolean visit(IResourceDelta delta) {
            IResource res = delta.getResource();
            if (! (res instanceof IFile)) {
                // workspaceroot, project, or folder, so skip
                return true;
            }
            switch (delta.getKind()) {
            case IResourceDelta.ADDED : {
                Debug.traceln("DeltaVisitor.visit: " + "Resource " + res.getFullPath() + " added");
                if ((delta.getFlags() & IResourceDelta.MOVED_FROM) != 0) {
                    if (delta.getMovedFromPath().equals(getFile().getFullPath())) {
                        // our file got renamed
                        // and this delta resource is the new one 
                        setFile((IFile) res);
                    }
                }
                break;
            }
            case IResourceDelta.REMOVED : {
                Debug.traceln("DeltaVisitor.visit: " + "Resource " + res.getFullPath() + " removed");
                if ((delta.getFlags() & IResourceDelta.MOVED_TO) == 0) {
                    // this resource was NOT moved, but removed
                    if (delta.getFullPath().equals(getFile().getFullPath())) {
                        // our file got removed
                        setFile(null);
                    }
                }
                break;
            } }
            return true;
        }
    }

    /**
     * This class maintains the state of resources in AcmeStudio, which may
     * be represented by a collection of boolean states.
     * <br>
     * == Significance of the various states ==
     * <ol>
     * <li>connected:  determine connectedness of resource provider to resource<ul>
     *     <li>prevent any useful activity if provider not connected</li>
     * </ul></li>
     * <li>parseError:  whether the most recent load encountered parse error(s)<ul>
     *     <li>set during ASRP.reloadOp, when a parse error exception occurs</li>
     *     <li>gets used by ASRP.canOpenDesign() to affect whether diagram editor is opened</li>
     * </ul></li>
     * <li>sourceEditor:  whether we're in the source editor<ul>
     *     <li>set if we're in the source editor, by switches in MPAE.pageChange()</li>
     *     <li>if model changed, and not in source editor, then, synch changed</li>
     *     <li>prompts reloading before saving in MPAE.doSave()</li>
     *     <li>affects MPAE.pageChange() method for sure</li>
     *     <li>prevents selection change behavior when within text editor</li>
     * </ul></li>
     * <li>synchedText:  whether source and diagram models are synchronized<ul>
     *     <li>set when source and diagram models are synch'd</li>
     *     <li>prevents ASRP.reloadFromText() from executing (or prompts executing if false)</li> 
     *     <li>prevents ASRP.updateText() from executing</li>
     *     <li>affects the enablement of the FiltereOutVisAction</li>
     * </ul></li>
     * <li>isDirty:  dirty-ness state of the resource<ul>
     *     <li>set to dirty when AcmeDesign source/diagram/style is edited, or when design is "changed on load"</li>
     *     <li>prevents resource provider from processing changeModel event</li>
     *     <li>determines whether a resource should be reloaded in ASResource.reload()</li>
     *     <li>affects the "asterick" indicator on the editor tab, MPAE.isDirty()</li>
     *     <li>cause information displayed about file needing to be saved in ASRP.loadFile()</li>
     * </ul></li>
     * <li>editable:  affects whether source editor can be edited<ul>
     *     <li>if we're using filtered design, we are NOT allowed to edit the design in source view</li>
     *     <li>N.B:  currently NOT in use yet.v
     * </ul></li>
     * <li>filteringDesign:  indicates whether design is being filtered<ul>
     *     <li>this should be contained within the "editable" state, but is being
     *      used explicitly right now due to existing code.</li>
     * </ul></li>
     * <li>regionManagerValid:  whether the region manager for model source is in a valid state<ul>
     *     <li>set to true after running ASRP.forceUpdateText(); false when document changed</li>
     *     <li>allows ASRP.getRegion() to return a region</li>
     *     <li>allows AcmeFoldingStructureProvider to compute addition</li>
     *     <li>allows AcmeTextEditor.handleCursorPositionChanged() to process cursor change and the corresponding selection change</li>
     *     <li>allows DiagramEditor.IGotoMarker.gotoMarker() to determine what diagram object to select</li>
     * </ul></li>
     * </ol>
     * 
     * == State transitions ==
     * <br>
     * Four collection of states are of interest for smooth interaction of source
     * editing:<ol>
     * <li>(A) In diagram editor (NOT in source) + parse OK + synched
     * <li>(B) In source editor + parse OK + synched
     * <li>(C) In source editor + parse-state-unknown + NOT synche
     * <li>(D) In source editor + parse FAILED + no-synch-state
     * </ol>
     * Transitions:<ul>
     * <li>A -> A:  change text/save
     * <li>A -> B:  switch to source, ok always
     * <li>B -> C:  change text
     * <li>C -> B:  save OK
     * <li>C -> A:  switch to diagram OK
     * <li>C -> D:  (1) save with error OR (2) switch to diagram with error
     * <li>D -> D:  change text/save with error
     * <li>D -> B:  save OK
     * </ul>
     * 
     * @author Shang-Wen (zensoul@cs.cmu.edu)
     */
    public class ASResourceState {

        /** Flag indicating whether this provider's resource is sync'd with text editor's doc. */
        private boolean _connected;
        /** Flag indicating whether source has parse error. */
        private boolean _parseError;
        /** Flag indicating whether managed resource is being accessed under the source editor. */
        private boolean _sourceEditor;
        /** Flag indicating whether this provider's resource is sync'd with text editor's doc. */
        private boolean _synchedText;
        /** Flag indicating whether managed resource is dirty. */
        private boolean _dirty;
        /** Flag indicating whether source editor is editable. */
        private boolean _editable;
        /** Flag indicating whether design is being filtered (i.e., removing vis properties). */
        private boolean _filteringDesign;
        /** Flag indicating whether the regions are valid **/
        private boolean _regionManagerValid;
        
        public ASResourceState () {
            // initial states
            setConnected(false);
            setParseError(false);
            setInSourceEditor(false);
            setSynchedWithText(true);
            setEditable(true);
            setFilteringDesign(false);
            setDirty (false);
            setRegionManagerValid(false);
        }
        
        public boolean isConnected () {  return _connected;  }
        public boolean isDisconnected () {  return !_connected;  }
        public void setConnected (boolean b) {  _connected = b;  }

        public boolean isEditable () {  return isConnected() && _editable;  }
        public void setEditable (boolean b) {  _editable = b;  }

        public boolean hasParseError () {  return isConnected() && _parseError;  }
        public void setParseError (boolean b) {  _parseError = b;  }

        public boolean isFilteringDesign () {
            return isConnected() && _filteredDesign != null && _filteringDesign;
        }
        public void setFilteringDesign (boolean b) {
            _filteringDesign = b;
            if (_filteringDesign) {
                setEditable(false);
            } // but don't set the opposite, since other factors can affect editability
        }

        public boolean isInSourceEditor () {
            return isConnected() && _sourceEditor;
        }
        public void setInSourceEditor (boolean b) {
            if (isDisconnected()) return;
            _sourceEditor = b;
        }

        public boolean isDirty () {
            return isConnected() && _dirty;
        }
        public void setDirty (boolean b) {
            if (isDisconnected()) return;
            if (b == isDirty() && b == _editor.isDirty()) return;
            
            _dirty = b;
            _editor.setDirty(b);
            if (b && getASDesign() != null) {
                getASDesign().setReloadState(IASResource.EDITED);
            }
        }

        public boolean isSynchedWithText () {
            return isDisconnected() || _synchedText;
        }
        public void setSynchedWithText (boolean b) {
            if (isDisconnected()) return;
            if (b == isSynchedWithText()) return;
            
            _synchedText = b;
            fireProviderChanged (ResourceProviderEvent.SYNC_STATUS_CHANGED);
        }

        public boolean isRegionManagerValid () {
            return isConnected() && _regionManager != null && _regionManagerValid;
        }
        public void setRegionManagerValid (boolean b) {
            if (isDisconnected()) return;
            if (b == isRegionManagerValid()) return;

            _regionManagerValid = b;
            fireProviderChanged(ResourceProviderEvent.REGION_MANAGER_CHANGED);
        }

    }

    private static AcmePartitionScanner _scanner = new AcmePartitionScanner ();

    /** Collection of resource states that affect resource & provider behavior */
    private ASResourceState _states = null;

    /** the design to use when displaying alternative views of the text **/
    private AcmeDesign _filteredDesign;
    private FilteredDesignSynchronizer _synchronizer;
    
    /** the corresponding editor, also the item to check to make sure connected */
    private MultiPageAcmeEditor _editor = null;
    /** initial editor input used to construct resource */
    private IEditorInput _editorInput = null;
    /** file used to construct resource */
    private IFile _file = null;
    /** IEditorInputs for families */
    private IEditorInput[] _famEditorInputs = null;
    /** IEditorInput for text file */
    private IEditorInput _textEditorInput = null;
    /** the main ASDesign resource */
    private ASDesign _asDesign = null;
    /** flag indicating whether this provider's editor is active */
    private boolean _isEditorActive = false;
    /** listener of AcmeDocumentProvider document state */
    private ASDocumentListener _documentListener = null;
    /** listener of AcmeDocumentProvider element state */
    private ASElementStateListener _elementListener = null;
    /** listener for parts, mainly to find out if this provider's editor is closed */
    private PartListener _partListener = null;
    /** listener for resource change */
    private ResourceChangeListener _resourceChangeListener = null;
    /** resource watcher */
    private ResourceWatcher _resourceWatcher = null;
    /** manager for regions that map objects to character positions */
    private TextRegionManager _regionManager = null;
    /** queue for IASResourceChangeEvents until this ASRP's editor becomes active */
    private List _eventQueue = null;

    /**
     * Creates an ASResourceProvider object ready for connection to an editor.
     */
    public ASResourceProvider () {
        super();
        _states = new ASResourceState();
        // BugFix #46
        AcmeStudioChangeManager.registerListener(this);
    }

    /**
     * @see java.lang.Object#finalize()
     */
    protected void finalize() throws Throwable {
        disconnect();
        AcmeStudioChangeManager.removeListener(this);
        _states = null;
        super.finalize();
    }
    
    /**
     * @see org.eclipse.core.runtime.IAdaptable#getAdapter(Class)
     */
    public Object getAdapter (Class adapter) {
        if (adapter == ASResourceProvider.class) {
            return this;
        } else {
            return null;
        }
    }
    
    /**
     * @see AcmeStudio.util.AcmeModelChangeListener#changeModel(AcmeStudio.util.AcmeStudioChangeManager.AcmeStudioChangeEvent)
     */
    public void changeModel(AcmeStudioChangeEvent evt) {
        // BugFix #46
        if (_states.isDirty()) return;
        if (Helper.eventRelatedToDesign(evt, getAcmeDesign()) && !_states.isFilteringDesign()) {
            if (! _states.isInSourceEditor()) {
                _states.setSynchedWithText(false);
            }
            _states.setDirty(true);   
        } else if (_filteredDesign != null && Helper.eventRelatedToDesign (evt, _filteredDesign) && _states.isFilteringDesign()) {
            try {
                forceUpdateText();
                getASDesign().processErrors();
            } catch (CoreException e) {
                Debug.errorln(e);
            }
        }
    }

    /**
     * @see AcmeStudio.util.AcmeModelChangeListener#interestedInEvent(int)
     */
    public boolean interestedInEvent (int eventType) {
        return true;
    }

    /**
     * This method responds to changes in this ASRP's associated ASDesign.
     * If change occurred from externally, the ASRP fires a "need reload"
     * provider changed message.  Otherwise, whether the ASDesign changed or
     * its ancestor changed, ASRP eventually fires off "reloaded, so just renew"
     * message.
     * 
     * @see AcmeStudio.core.resource.IASResourceChangeListener#resourceChanged(AcmeStudio.core.resource.IASResourceChangeEvent)
     */
    public void resourceChanged(IASResourceChangeEvent evt) {
        if (!_states.isConnected()) return;

        if (!isEditorActive()) {
            // queue the IASResourceChangeEvent until editor is active
            _eventQueue.add(evt);
            return;
        }

        if ((evt.getType() & IASResourceChangeEvent.SELF_CHANGED_EXTERNALLY) > 0) {
            // The ASResource content changed externally, so reconnect to source and cause renew pages
            fireProviderChanged(ResourceProviderEvent.AS_RESOURCE_NEED_RELOAD);
        } else if ((evt.getType() & IASResourceChangeEvent.SELF_CHANGED) > 0) {
            // The ASResource content got reloaded, so just cause renew pages
            fireProviderChanged(ResourceProviderEvent.AS_RESOURCE_RELOADED);
        } else if ((evt.getType() & IASResourceChangeEvent.ANCESTOR_CHANGED) > 0) {
            // Reload from text OR reload from design, depending on editor active
            if (isInTextEditor()) {
                try {
                    // at the end of this method, AS_RESOURCE_RELOADED is fired
                    reloadFromText();
                } catch (CoreException e) {
                    //ignore error
                }
            } else {
                _editor.getEditorSite().getShell().getDisplay().asyncExec(new Runnable() {
                    public void run () {
                        try {
                            AcmeOutputStream aos = new AcmeOutputStream();
                            AcmeDesign design = getASDesign().getDesign();
                            design.exportArmani(aos);
                            ByteArrayInputStream bais = new ByteArrayInputStream(aos.toString().getBytes());
                            aos.close();
                            getASDesign().reloadStream(null, bais);
                        } catch (Exception e) {
                            //ignore error
                        } finally {
                            fireProviderChanged(ResourceProviderEvent.AS_RESOURCE_RELOADED);
                        }
                    }
                });
            }
        }
    }
    
    /**
     * Connect this ASResourceProvider with the provided editor.
     * This method must be called first before other services are usable.
     * The editor part must not be null.
     * @param editor  the editor using this ASResourceProvider
     */
    public void connect (IEditorPart editor) throws CoreException, ParseError {
        if (editor == null) {
            throw new CoreException(new ASStatus(ASStatus.ERROR,
                    ASStatus.INVALID_EDITOR_INPUT, 
                    AcmeStudioPlugin.getMessage("ASResourceProvider.badEditorInput"),
                    null));
        }
        
        _editor = (MultiPageAcmeEditor) editor;
        _editorInput = _editor.getEditorInput();
        if (_editorInput instanceof IFileEditorInput) {
            
            _file = ((IFileEditorInput) _editorInput).getFile();
            
            Shell shell = _editor.getEditorSite().getShell();
            Display display = shell.getDisplay();
            // watch for synchronization with local file & provider change
            _resourceWatcher = new ResourceWatcher(this, display);
            _resourceWatcher.start();
            
            IRunnableWithProgress connectOp = new WorkspaceModifyOperation () {
                public void execute (IProgressMonitor monitor)
                throws InvocationTargetException {
                    String msg = "Loading Acme Resource '" + _editorInput.getName() + "'...";
                    monitor.beginTask(msg, 100);
                    monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.PreparingMsg")); //$NON-NLS-1$
                    monitor.worked(20);
                    try {
                        connectInternal(monitor);
                    } catch (Exception e) {
                        // just wrap the exception (potentially CoreEx or ParseError) and rethrow
                        throw new InvocationTargetException(e);
                    } finally {
                        monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.WrapUpMsg")); //$NON-NLS-1$
                        monitor.worked(50);
                        monitor.done();
                    }
                }
            };
            
            // fire off a progress monitor
            try {
                /* [SWC] As of 2004-06-15, in 3.0 RC, calling ProgressMonitorDialog
                 * in some cases (when restoring an editor after fresh Platform
                 * startup) causes infinite recursion of editor activation.
                 * Do not use this call without guard!
                 */
                //new ProgressMonitorDialog(shell).run(false, false, connectOp);
                
                if (shouldRaiseDialog()) {
                    new ProgressMonitorDialog(shell).run(false, false, connectOp);
                } else {
                    // fire off connection with status bar progress monitor
                    IProgressMonitor monitor = _editor.getEditorSite().getActionBars()
                        .getStatusLineManager().getProgressMonitor();
                        connectOp.run(monitor);
                }
            } catch (InterruptedException e) {
                // ignored
                Debug.diagnoseln(e);
            } catch (InvocationTargetException e) {
                // throw the causal exception, if the second-level throwable
                // is either core or parse error
                Throwable t = e.getCause();
                Throwable reThrow = t;
                // first, see if there's another nested level
                if (t instanceof CoreException) {
                    Throwable t2 = ((CoreException )t).getStatus().getException();
                    if (!(t2 instanceof ParseError)) {
                        // throw the wrapped exception.
                        reThrow = t2;
                    }
                }
                if (reThrow instanceof ParseError) {
                    throw (ParseError) reThrow;
                } else {
                	Debug.errorln (e);
                    String eMsg = reThrow.getMessage();
                    if (eMsg == null) {
                        eMsg = reThrow.getClass().getName();
                    }
                    throw new CoreException(new ASStatus(
                        ASStatus.ERROR,
                        ASStatus.OK,
                        "Exception occurred while connecting to resource: " + eMsg,
                        reThrow));
                }
            }
        } else {
            // stop, can't create ASResourceProvider
            // unset connect status, since all hasn't gone well.
            disconnect();
            throw new CoreException(new ASStatus(ASStatus.ERROR,
                    ASStatus.INVALID_EDITOR_INPUT, AcmeStudioPlugin
                    .getMessage("ASResourceProvider.badEditorInput"),
                    null));
        }
    }
    
    protected void connectInternal (IProgressMonitor monitor)
    throws CoreException, ParseError {
        if (_editor == null) {
            return;
        }
        
        // initialize data members
        if (monitor != null) monitor.worked(10);
        _elementListener = new ASElementStateListener(this);
        _documentListener = new ASDocumentListener(this);
        _partListener = new PartListener();
        AcmeStudioPlugin.getActiveWindow().getPartService().addPartListener(_partListener);
        _resourceChangeListener = new ResourceChangeListener();
        ResourcesPlugin.getWorkspace().addResourceChangeListener(_resourceChangeListener, IResourceChangeEvent.POST_CHANGE);
        _eventQueue = new ArrayList();
        
        _states.setConnected(true);
        // link the editor input with parent document provider
        if (monitor != null) monitor.worked(20);
        super.connect(getTextEditorInput());
        
        try {
            if (monitor != null) monitor.worked(10);
            loadFile(monitor);
        } catch (AcmeException e) {
            // There is an error, but it is still ok to update the text
            // because the error is about unresolved names, etc.
        } catch (ParseError e) {
            // at least sync with the content of the file in the text editor
            synchronize(getTextEditorInput());
            // also need to listen to document changes so dirty bit changes
            getDocument(getTextEditorInput()).addDocumentListener(_documentListener);
            addElementStateListener(_elementListener);
            throw e;
        }
        
        if (monitor != null) monitor.worked(20);
        _states.setSynchedWithText(false);
        // make sure text editor is in sync with file.
        synchronize(getTextEditorInput());
        // do update text before adding document and element listener to
        // avoid redundant dirty bit sets.
        updateText();
        getASDesign().processErrors();
        getDocument(getTextEditorInput()).addDocumentListener(_documentListener);
        addElementStateListener(_elementListener);
        // Part of BugFix# 207, prevents clearing dirty bit if the content
        // isn't being reloaded from file, and there's unsaved changes
        // setDirty(false); 
    }

    /**
     * Reload content of current resource provider's, causing firing of provider change.
     */
    public void reconnect () {
        setFile(_file);
    }
    
    /** 
     * Clears out all data members, so that this object can be cleanly finalized.
     */
    public void disconnect () {
        if (!_states.isConnected()) return;
        
        if (_resourceWatcher != null) {
            // remove resource watcher.
            _resourceWatcher.terminate();
            _resourceWatcher = null;
        }
        disconnectInternal();
        if (_asDesign != null) {
            _asDesign.dispose();
            _asDesign = null;
        }
        // clear this last
        _editor = null;
    }
    
    protected void disconnectInternal () {
        if (_editor == null) {
            return;
        }
        
        // unlink the editor input from parent document provider
        super.disconnect(getTextEditorInput());
        
        _editorInput = null;
        if (_elementListener != null) {
            removeElementStateListener(_elementListener);
            _elementListener = null;
        }
        if (getTextEditorInput() != null) {
            IDocument doc = getDocument(getTextEditorInput());
            if (doc != null) {
                if (_documentListener != null) {
                    doc.removeDocumentListener(_documentListener);
                    _documentListener = null;
                }
            }
        }
        if (_partListener != null) {
            AcmeStudioPlugin.getActiveWindow().getPartService().removePartListener(_partListener);
            _partListener = null;
        }
        if (_resourceChangeListener != null) {
            ResourcesPlugin.getWorkspace().removeResourceChangeListener(_resourceChangeListener);
            _resourceChangeListener = null;
        }
        if (_eventQueue != null) {
            _eventQueue.clear();
            _eventQueue = null;
        }
        _file = null;
        _famEditorInputs = null;
        _textEditorInput = null;
        
        // set disconnect status
        _states.setConnected(false);
    }
    
    /**
     * Adds listener to this provider's list of IResourceProviderListeners.
     * @param listener  the IResourceProviderListener to add
     */
    public void addResourceProviderListener (IResourceProviderListener listener) {
        if (!_states.isConnected()) return;
        
        _resourceWatcher.addProviderListener(listener);
    }
    
    /**
     * Removes listener from this provider's list of IResourceProviderListeners.
     * @param listener  the IResourceProviderListener to remove
     */
    public void removeResourceProviderListener (IResourceProviderListener listener) {
        if (!_states.isConnected()) return;
        
        _resourceWatcher.removeProviderListener(listener);
    }
    
    private IRunnableWithProgress reloadOp = new WorkspaceModifyOperation () {
        public void execute (IProgressMonitor monitor) throws CoreException {
            monitor.beginTask(AcmeStudioPlugin.getMessage("ASResourceProvider.ReloadingMsg"), 10); //$NON-NLS-1$
            monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.PreparingMsg")); //$NON-NLS-1$
            IDocument doc = getTextDocument();
            if (doc == null) {  // no document created yet
                monitor.done();
                return;
            }

            // keep the previous design, in case parsing goes wrong...
            AcmeDesign oldDesign = getASDesign().getDesign();
            // make sure that the design will indeed get reloaded
            if (_states.isDirty() && getASDesign() != null) {
                getASDesign().setReloadState(IASResource.EDITED);
            }
            String text = doc.get();
            ByteArrayInputStream bais = new ByteArrayInputStream(text.getBytes());
            try {
                monitor.worked(1);
                try {
                    getASDesign().reloadStream(monitor, bais);
                    _states.setParseError(false);
                } catch (AcmeException e) {
                    Debug.diagnoseln(e);
                }
                
                monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.UpdatingMsg")); //$NON-NLS-1$
                ASResourceProvider.this.updateText();
                getASDesign().processErrors();
                monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.FinishingMsg")); //$NON-NLS-1$
                _states.setSynchedWithText(true);
                monitor.worked(1);
            } catch (ParseError e) {
                // this means we disable the diagram editor and stay within the source editor
                _states.setParseError(true);
                monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.ProcessingParseErrMsg")); //$NON-NLS-1$
                // restore the doc content that had parse error
                getTextDocument().set(text);
                monitor.worked(1);
            } finally {
                // trigger MPAE renew pages, which will either
                //   1. show the updated AcmeDesign in the Diagram Editor; or
                //   2. reflect the parse error
                // (N.B. Bug #207 is fixed by making sure that MPAE indeed picks up the updated AcmeDesign)
                fireProviderChanged(ResourceProviderEvent.AS_RESOURCE_RELOADED);
                monitor.done();
            }
        }
    };
    
    public synchronized void reloadFromText () throws CoreException {
        if (_states.isSynchedWithText()) return;
        
        try {
            new ProgressMonitorDialog(_editor.getSite().getShell())
                .run(false, false, reloadOp);
        } catch (InterruptedException e) {
            // ignore
            Debug.diagnoseln(e);
        } catch (InvocationTargetException e) {
            if (e.getTargetException() instanceof CoreException) {
                throw (CoreException) e.getTargetException();
            } else {
                Debug.errorln(e);
            }
        }
    }
    
    public synchronized void updateText () throws CoreException {
        if (_states.isSynchedWithText() || _states.hasParseError()) return;

        // perform the injections
        if (getASDesign().getDesign().hasInjectionOperations ()) {
            getASDesign().getDesign().doInjectionOperations();
            setSynchedWithText(false);
        }
        forceUpdateText ();
    }
    
    protected synchronized void forceUpdateText () throws CoreException {
        IDocument doc = getTextDocument();
        if (doc == null) {  // no document created yet
            return;
        }
        
        // Debug.diagnoseln("ASResourceProvider.updateText()");
        // at this point, should be good to go.
        AcmeOutputStream aos = new AcmeOutputStream();
        try {
            AcmeDesign design = null;
            if (_states.isFilteringDesign()) {
                design = _filteredDesign;
            } else {
                design = getASDesign().getDesign ();
            }
            // [SWC] Workaround for Bug #264 to prevent save from bailing
            if (design != null) {
                //BugFix# 362: If error occurs during export, source isn't synched!
                //  this try/catch remedies the bug and cause partial output to source 
                try {
                    design.exportArmani(aos);
                } catch (AcmeException e) {
                    throw new CoreException(new ASStatus(ASStatus.ERROR,
                        ASStatus.ACME_ERROR, "ASResourceProvider.updateText() failed!", e));
                } catch (Throwable t) {
                    // process any run-time exceptions (such as index out of bound)
                    // TODO Link with eventual exception view
                    MessageDialog.openInformation(Display.getCurrent().getActiveShell(), AcmeStudioPlugin.getMessage("ASResourceProvider.ExportArmaniFailedTitle"), AcmeStudioPlugin.getMessage("ASResourceProvider.ExportArmaniFailedMsg") + t.getMessage()); //$NON-NLS-1$ //$NON-NLS-2$
                    Debug.errorln("forceUpdateText(): exportArmani failed!\n\n" + t);
                }
            } else {  // something is terribly wrong
                MessageDialog.openInformation(Display.getCurrent().getActiveShell(), AcmeStudioPlugin.getMessage("ASResourceProvider.DesignNullTitle"), AcmeStudioPlugin.getMessage("ASResourceProvider.DesignNullMsg")); //$NON-NLS-1$ //$NON-NLS-2$
                Debug.errorln("forceUpdateText(): Design object from ASDesign is null!");
            }
            aos.close ();
            boolean wasDirty = _states.isDirty();
            doc.set(aos.toString());
            _states.setDirty(wasDirty);
            _states.setSynchedWithText(true);
            if (!_states.isFilteringDesign()) {
                _regionManager = new TextRegionManager (aos.getRegionManager());
                _states.setRegionManagerValid(true);
                // Typecheck the design to get any errors
                IPreferenceStore store = AcmeStudioPlugin.getDefault().getPreferenceStore ();
                if (store.getBoolean ("TextEditor.AutoTypecheck")) {
                    try {
                        // [SWC] Workaround for Bug #264 to prevent save from bailing
                        if (design != null) {
                            design.typecheck();
                        }
                    } catch (Exception e) {  // something unexpected happened?
                        Debug.errorln(e);
                    }
                    AcmeTypeManager.typecheckingErrorStack.disconnect();
                }
            } else {
                _states.setRegionManagerValid(false);
            }
        } catch (IOException e) {
            throw new CoreException(new ASStatus(ASStatus.ERROR,
                ASStatus.IO_ERROR, "ASResourceProvider.updateText() failed!", e));
        }
    }
    
    /**
     * Saves acme design content to appropriate files.
     * @param monitor  progress monitor to use
     * @return boolean  whether save succeeeded
     */
    public boolean saveAcme (IProgressMonitor monitor) throws CoreException {
        if (!_states.isConnected()) return false;

        boolean successSys = false;
        if (_states.isInSourceEditor() && !_states.isSynchedWithText()) {
            // Save via the source editor
            if (_editor.getSourceEditor() == null) {
                throw new CoreException(new ASStatus(ASStatus.ERROR,
                    ASStatus.NULL_EDITOR, "ASResourceProvider.saveAcme() failed due to NULL source editor!", null));
            } else {
                // optimistically expect source editor save to succeed
                successSys = true;
                _editor.getSourceEditor().doSave(monitor);
                if (monitor.isCanceled()) {
                    successSys = false;
                }
            }
            reloadFromText();
        } else {
            successSys = getASDesign().save(monitor);
            if (successSys) {
                synchronize(getTextEditorInput());
                updateText();
                getASDesign().processErrors();
            }
        }

        if (successSys) {
            // when save succeeds, we clear the dirty state
            _states.setDirty(false);
            // cause a reload
            fireProviderChanged(ResourceProviderEvent.AS_RESOURCE_RELOADED);
        }
        return successSys;
    }
    
    /**
     * Saves acme design content to appropriate files, with specified name.
     */
    public boolean saveAcmeAs (IProgressMonitor monitor, IPath path)
    throws CoreException {
        if (!_states.isConnected()) return false;

        // Get the new file
        IFile ifile = AcmeStudioPlugin.getRoot().getFile(path); 
        boolean successSys = getASDesign().saveTo(ifile, monitor);
        if (successSys) {
            setFile(ifile);
            _states.setDirty(false);
        }
        
        return successSys;
    }
    
    /**
     * Takes the original design and create a new design, applying the
     * filters to the new design. Updates the text to reflect the new filtering
     * Note: original design is not touched.
     * @param filters The filters to apply
     */
    public void filterDesign (WalkerElementOperation [] filters) throws CoreException {
        _filteredDesign = (AcmeDesign )getASDesign().getDesign ().copy ();
        for (int i = 0; i < filters.length; i++) {
            GraphWalker.WalkOverAllElements (_filteredDesign, filters [i], GraphWalker.WALK_BOTTOM_UP);
        }
        _states.setFilteringDesign (true);
        if (_synchronizer != null) {
            AcmeStudioChangeManager.removeListener(_synchronizer);
        }
        _synchronizer = new FilteredDesignSynchronizer (getASDesign().getDesign(), _filteredDesign, filters); 
        AcmeStudioChangeManager.registerListener(_synchronizer);
        forceUpdateText ();
    }
    
    public void restoreDesign () throws CoreException {
        _states.setFilteringDesign (false);
        forceUpdateText ();
        if (_synchronizer != null) {
            AcmeStudioChangeManager.removeListener(_synchronizer);
            _synchronizer = null;
        }
        _filteredDesign = null;
    }

    
    /////////////////////
    //  Getter/Setters
    /////////////////////

    /**
     * Returns the RegionManager for the associated resources.
     * @return TextRegionManager The _regionManager to set
     */
    public TextRegionManager getRegionManager () {
        return _regionManager;
    }
    
    /**
     * Sets the RegionManager for the associated resources, only if no current
     * RegionManager is already set.
     * @param regionManager  the new RegionManager to set
     */
    public void setRegionManager(TextRegionManager regionManager) {
        if (_regionManager == null) {
            _regionManager = regionManager;
            _states.setRegionManagerValid(true);
        }
    }
    
    /**
     * Returns whether design can be opened, essentially equivalent to whether
     * a parse error exists.
     * @return boolean whether design can be opened
     */
    public boolean canOpenDesign() {
        return _states.isConnected() && !_states.hasParseError()
            && getASDesign () != null && !getASDesign().hasParseError();
    }

    /**
     * Returns the ASDesign of this resource provider.
     * @return ASDesign
     */
    public ASDesign getASDesign () {
        return _states.isConnected() ? _asDesign : null;
    }
    
    /**
     * Returns the AcmeDesign.
     * @return AcmeDesign the AcmeDesign
     */
    public AcmeDesign getAcmeDesign() {
        return (_states.isConnected() ?
                ((canOpenDesign()) ? getASDesign().getDesign() : null) :
                    null);
    }
    
    /**
     * Returns the AcmeFamily.
     * @return AcmeFamily an AcmeFamily
     */
    public AcmeFamily getAcmeFamily() {
        return _states.isConnected() ? getASDesign().getFamily() : null;
    }
    
    /**
     * Returns the list of Acme Families.
     * @return List a list of families
     */
    public List getAcmeFamilies() {
        return _states.isConnected() ? getASDesign().getFamilies() : null;
    }
    
    /**
     * Collects all families from family list and imported design records
     * into a single list.
     * @return List  the list of all families in design and imports
     */
    public List getAllFamilies() {
        List families = new ArrayList();
        
        if (_states.isConnected()) {
            families.addAll(getASDesign().getFamilies());
            Iterator iter = getASDesign().getImports().iterator();
            while (iter.hasNext()) {
                AcmeDesign design = (AcmeDesign) iter.next();
                families.addAll(Helper.getAcmeFamilies(design));
            }
        }
        
        return families;
    }
    
    /**
     * Returns the IEditorInput containing the given AcmeSystem in this provider.
     * @param sys  the AcmeSystem to provide IEditorInput for
     * @return IEditorInput  the IEditorInput containing AcmeSystem
     */
    public IEditorInput getSystemEditorInput (AcmeSystem sys) {
        return _states.isConnected() ? new ASEditorInput(this, sys) : null;
    }
    
    /**
     * Returns the IEditorInput containing the given AcmeFamily.
     * @param fam  the AcmeFamily to provide IEditorInput for
     * @return IEditorInput  the IEditorInput containing AcmeFamily
     */
    public IEditorInput getFamilyEditorInput (AcmeFamily fam) {
        return _states.isConnected() ? new ASEditorInput(this, fam) : null;
    }
    
    /**
     * Returns an Array of IEditorInputs containing the all AcmeFamilies in the
     * AcmeDesign held by this provider.
     * @return IEditorInput[]  the IEditorInput list
     */
    public IEditorInput[] getFamilyEditorInputs () {
        if (!_states.isConnected()) return null;
        
        if (_famEditorInputs != null) {
            return _famEditorInputs;
        }
        // otherwise, lazy-instantiate
        List fams = getAllFamilies();
        _famEditorInputs = new IEditorInput[fams.size()];
        Iterator i = fams.iterator();
        for (int j=0; i.hasNext() && j < fams.size(); j++) {
            AcmeFamily fam = (AcmeFamily) i.next();
            _famEditorInputs[j] = getFamilyEditorInput(fam);
        }
        return _famEditorInputs;
    }
    
    /**
     * Returns the IFileEditorInput containing this provider's IFile.
     * @return IFileEditorInput  the IFileEditorInput containing IFile
     */
    public IEditorInput getTextEditorInput () {
        if (!_states.isConnected()) return null;
        
        if (_textEditorInput == null) {
            // lazy-instantiate
            _textEditorInput = new FileEditorInput(_file);
        }
        return _textEditorInput;
    }
    
    /**
     * Returns the IDocument stored in the AcmeDocumentProvider.
     * @return IDocument  the doc provider's IDocument
     */
    public IDocument getTextDocument () {
        IDocument doc = null;
        if (_states.isConnected() && getTextEditorInput() != null) {
            // fetch the IDocument
            doc = getDocument(getTextEditorInput());
        }
        return doc;
    }
    
    /**
     * Returns a region corresponding an AcmeObject
     * @param o  AcmeObject to get text region of
     * @return Region  the text region which the AcmeObject spans
     */
    public Region getRegion (AcmeObject o) {
        if (_states.isRegionManagerValid()) {
            return _regionManager.getRegion(o);
        }
        return null;
    }
    
    /**
     * Returns the elementListener.
     * @return ASElementStateListener
     */
    public ASElementStateListener getElementListener() {
        return _states.isConnected() ? _elementListener : null;
    }
    
    /**
     * Returns name of the acme design file.
     */
    public IFile getFile () {
        return _states.isConnected() ? _file : null;
    }

    /**
     * Returns whether the main editor of this resource provide is active
     * @return boolean
     */
    public boolean isEditorActive() {
        return _isEditorActive;
    }
    
    /**
     * Returns whether the dirty bit is set on.
     * @return boolean  the status of dirtiness
     */
    public boolean isDirty () {
        return _states.isDirty();
    }
    
    /**
     * Returns whether the editor is a text editor.
     * @return boolean  whether text editor or not
     */
    public boolean isInTextEditor () {
        return _states.isInSourceEditor();
    }
    
    /**
     * Returns whether this provider is synched with text in text editor.
     * @return boolean   whether provider and text is synched
     */
    public boolean isSynchedWithText () {
        return _states.isSynchedWithText();
    }
    
    /**
     * Returns whether the regions manager is updated and valid.
     * @return boolean
     */
    public boolean isRegionManagerValid() {
        return _states.isRegionManagerValid();
    }
    
    /**
     * @param isActive  new state of editor activeness
     */
    protected void setEditorActive(boolean isActive) {
        if (_isEditorActive == isActive) return;  //no need to cause actions

        _isEditorActive = isActive;

        // process any queued IASResourceChangeEvents
        if (_isEditorActive) {
            Iterator i = _eventQueue.iterator();
            while (i.hasNext()) {
                IASResourceChangeEvent evt = (IASResourceChangeEvent )i.next();
                resourceChanged(evt);
                i.remove();
            }
        }
    }
    
    /**
     * Sets the new dirty status.
     * @param isDirty  the new dirty status to set
     */
    public void setDirty (boolean isDirty) {
        _states.setDirty(isDirty);
    }
    
    /**
     * Sets whether current editor is text editor
     * @param isTextEditor  the new status
     */
    public void setInTextEditor (boolean isTextEditor) {
        _states.setInSourceEditor(isTextEditor);
    }
    
    /**
     * Sets whether this provider is synched with text in text editor.
     * @param isSynchedWithText  the new sync status
     */
    public void setSynchedWithText (boolean isSynchedWithText) {
        _states.setSynchedWithText(isSynchedWithText);
    }
    
    /**
     * Sets the regionManagerValid.
     * @param regionManagerValid The regionManagerValid to set
     */
    public void setRegionManagerValid(boolean regionManagerValid) {
        _states.setRegionManagerValid(regionManagerValid);
    }
    

    ////////////////////////////////
    //  Protected/Private Members
    ////////////////////////////////

    /**
     * @see org.eclipse.ui.texteditor.AbstractDocumentProvider#createDocument(Object)
     */
    protected IDocument createDocument(Object element) throws CoreException {
        IDocument document =  super.createDocument(element);
        if (document != null) {
            IDocumentPartitioner part = new DefaultPartitioner (_scanner, new String [] {AcmePartitionScanner.ACME_MULTILINE_COMMENT}) {
                public ITypedRegion [] computePartitioning (int offset, int length) {
                    return super.computePartitioning(offset,length)   ;
                }   
            };
            part.connect (document);
            document.setDocumentPartitioner(part);
        }
        return document;
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.ui.texteditor.AbstractDocumentProvider#createAnnotationModel(java.lang.Object)
     */
    protected IAnnotationModel createAnnotationModel(Object element) throws CoreException {
        if (element instanceof IFileEditorInput) {
            return new AcmeSourceAnnotationModel (this, (IFileEditorInput )element);
        }
        
        return super.createAnnotationModel(element);
    }
    
    /**
     * Loads an input file, saves pertinent file information, and causes content
     * to be parsed.
     */
    private void loadFile (IProgressMonitor monitor)
    throws CoreException, ParseError, AcmeException {
        if (monitor != null) {
            monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.LoadingMsg")); //$NON-NLS-1$
        }

        // change context of AcmeContext
        ASContext.current().setActiveProject(_file.getProject());
        
        // only load design & visual style if _asDesign has not yet been initialized 
        if (getASDesign() == null) {
            IASResource asRes = null;
            try {
                asRes = ASModel.current().getASResource (_file.getLocation().toFile().getCanonicalPath());
            } catch (IOException e) {
                // ignore, just have null asRes
            }
            if (asRes != null && asRes instanceof ASDesign) {
                _asDesign = (ASDesign) asRes;
                if (getASDesign().getProvider() != this) {
                    getASDesign().setProvider(this);
                    // keep previous dirty state
                    // TODO loadFile() Figure out if we should preserve old dirty state for reload 
                    boolean wasDirty = _states.isDirty();
                    getASDesign().reload();
                    _states.setDirty(wasDirty);
                }
            } else {
                _asDesign = new ASDesign(null, this);
                getASDesign().loadFrom(_file);
                _states.setDirty(getASDesign ().changedOnLoad ());
                if (_states.isDirty()) {
                    MessageDialog.openInformation(Display.getCurrent().getActiveShell(), AcmeStudioPlugin.getMessage("ASResourceProvider.MustSaveTitle"), AcmeStudioPlugin.getMessage("ASResourceProvider.MustSaveMsg")); //$NON-NLS-1$ //$NON-NLS-2$
                }
                
            }
        } else {  // change the underlying file of _asDesign
            getASDesign().setFile(_file);
            if (getASDesign().shouldReload()) {
                boolean wasDirty = _states.isDirty();
                getASDesign().reload();
                _states.setDirty(wasDirty);
            }
        }
    }
    
   /**
     * Sets resource provider's file to a new file, causing new load and
     * firing of provider change.
     * @param newFile
     */
    protected void setFile (final IFile newFile) {
        if (!_states.isConnected()) {
            return;
        }
        
        if (newFile == null) {
            _file = null;
            fireProviderChanged(ResourceProviderEvent.ABOUT_TO_REMOVE);
            return;
        }
        
        IRunnableWithProgress reconn = new WorkspaceModifyOperation() {
            public void execute (IProgressMonitor monitor) throws CoreException{
                monitor.beginTask(AcmeStudioPlugin.getMessage("ASResourceProvider.ReloadingAcmeDesignMsg"), 4); //$NON-NLS-1$
                try {
                    monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.DisconnectingMsg")); //$NON-NLS-1$
                    // Bug Fix #103
                    // when we set to a new file, we should keep the current design
                    // and we can do this by preventing loadFile() from replacing
                    // the ASDesign if it's non-null, so keep _asDesign around;
                    // this way, the user has to explicitly save to get any changes
                    // to go into the new file
                    // disconnect from current file (will NOT null-out _asDesign)
                    disconnectInternal();
                    monitor.worked(1);
                    monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.ReconnectingMsg")); //$NON-NLS-1$
                    // reinstate _file
                    // BRS: Changed to preserve "additional" information like visualization, etc.
                    //if (_file != newFile) {
                    _file = newFile;
                    connectInternal(monitor);
                    //}
                    monitor.worked(2);
                } catch (CoreException e) {
                    Debug.errorln(e);
                } catch (ParseError e) {
                    Debug.errorln(e);
                } finally {
                    monitor.subTask(AcmeStudioPlugin.getMessage("ASResourceProvider.WrappingUpMsg")); //$NON-NLS-1$
                    // Fire provider changed event to cause renew pages
                    fireProviderChanged(ResourceProviderEvent.AS_RESOURCE_RELOADED);
                    monitor.done();
                }
            }
        };
        
        try {
            //if (_file != newFile) {
            new ProgressMonitorDialog(_editor.getSite().getShell())
                .run(false, false, reconn);
            //}
        } catch (InterruptedException e) {
            // ignore
            Debug.diagnoseln(e);
        } catch (InvocationTargetException e) {
            Debug.errorln(e);
        }
    }
    
    /**
     * Queues the provider's change kind for event firing.
     */
    private void fireProviderChanged (int kind) {
        if (_resourceWatcher != null) {
            _resourceWatcher.providerChangedState(kind);
        }
    }

    /**
     * The primary reason for this series of checks is prevent recursion
     * of calls to create editor in 3.0RC 1&2 when user clicks on an unopened
     * editor tab in a fresh workbench causing the editor to be restored.
     * <p>
     * To raise the dialog, we use conservative heuristics to allow only a 
     * few safe cases.
     * The heuristics are that, if the workbench is just starting (in splash),
     * then raising a dialog is just fine.  Otherwise, if the editor is being
     * restored due to a synchronized selection on the Navigator side, then
     * raising a dialog is safe to do.  Or, if the editor is being created
     * due to a double-click from the Navigator, raising a dialog is safe also.
     *  
     * @return boolean  whether or not to raise a ProgressMonitorDialog
     */
    private boolean shouldRaiseDialog() {
        boolean raiseDialog = false;
        IWorkbenchPartReference ref = _editor.getSite().getPage().getActivePartReference();
        if (ref == null) {
            // this means nothing has been started yet, i.e., in the
            // splash screen state, so bring up the explicit progress
            raiseDialog = true;
        } else {
            // see what selection is active
            // if the active selection matches what's editorInput is 
            //   being loaded, then we should be able to raise a dialog
            String inputName = _editorInput.getName();
            ISelection sel = _editor.getSite().getPage().getSelection();
            if (sel instanceof IStructuredSelection) {
                Object obj = ((IStructuredSelection )sel).getFirstElement();
                if (obj != null) {
                    if (obj.toString().endsWith(inputName)) {
                        // let's make sure the active part IS the navigator
                        // and that the synch button is toggled on
                        IWorkbenchPart part = ref.getPart(false);
                        if (part instanceof ResourceNavigator) {
                            if (((ResourceNavigator) part).isLinkingEnabled()) {
                                raiseDialog = true;
                            }
                        }
                    }
                }
            }
            // find the editor reference for this editor
            IEditorReference[] refs = _editor.getEditorSite().getPage().getEditorReferences();
            boolean foundPartRef = false;
            for (int i=0; i < refs.length; i++) {
                if (refs[i].getPartName().equals(inputName)) {
                    foundPartRef = true;
                }
            }
            if (! foundPartRef) { // no editor ref yet means a fresh load
                // definitely raise a dialog
                raiseDialog = true;
            }
        }
        return raiseDialog;
    }
    
}
