/*
 * Created on Mar 20, 2003
 */
package AcmeStudio.core.resource.provider;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.resource.IASResource;
import AcmeStudio.core.resource.IASResourceChangeEvent;



/**
 * Running Thread to do background watching of the ASResourceProvider.
 * 
 * This watcher thread does the following:
 *   - checks for content change external to Eclipse of the underlying file resource.
 *   - checks periodically to see if a provider changed event needs to be fired.
 *     Fires provider changed event, with composite 'kind' flags.
 *     This will induce certain reloading behavior in the provider.
 * 
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 */
class ResourceWatcher extends Thread {
	private final ASResourceProvider _provider;

    /** sleep time */
    public static final long SLEEP_TIME = 1000; // ms
	/** this thread object */
	private volatile Thread _watcher = null;
    /** current running watcher thread */
    private volatile Thread _thisThread = null;

	/** the default display */
	private Display _display = null;

    /** flag to indicate a current firing is occuring */
    private boolean _firingInProgress = false;
    /** changes "queued" for the next firing of provider changed event */
    private int _providerChangedState = ResourceProviderEvent.NONE;
    /** list of IResourceProviderListeners listening to this provider */
    private List _listeners = null;
    /** The provider change notification operation */
    private Runnable firingOp = new Runnable() {
        public void run() {
            if (!isWatcherAlive() || !firingInProgress()) return;
            try {
                fireProviderChanged();
            } finally {
                setFiringInProgress(false);  // finished firing
            }
        }
    };

    /** flag to indicate a current user query is occuring */
    private boolean _queryInProgress = false;
    /** the underlying file of resource being watched */
    private IFile _watchFile = null;
    /** the last modification timestamp of the watched resource */
    private long _modTimestamp = 0;
    /** The synchronization query operation */
    private Runnable querySyncOp = new Runnable() {
        public void run() {
            if (!isWatcherAlive() || !queryInProgress()) return;
            try {
                boolean doSync =
                    MessageDialog.openQuestion(_display.getActiveShell(),
                        AcmeStudioPlugin.getMessage("ResourceWatcher.syncTitle"), 
                        AcmeStudioPlugin.getMessage("ResourceWatcher.syncResource"));
                if (doSync
                    && _provider != null
                    && _provider.getASDesign() != null) {
                    // set the state and let the reload op do the actual reloading
                    _provider.getASDesign().setReloadState(IASResource.OUT_OF_SYNC);
                }
                // whether synched or not, save last modification time to
                // prevent repeated query
                if (_watchFile != null) {
                    _modTimestamp = _watchFile.getLocation().toFile().lastModified();
                }
            } finally {
                setQueryInProgress(false);  // finished querying
            }
        }
    };

	/**
	 * Constructor for ResourceWatcher.
	 */
	public ResourceWatcher(
		ASResourceProvider provider,
		Display display)
    {
		super("ResourceWatcher");
		_provider = provider;

		_watcher = this;
		_display = display;
        _listeners = new ArrayList();
	}

	/**
     * This run method checks three things every period (SLEEP_TIME):
     *     1) A provider change event needs to be fired
     *     2) The underlying IResource is out of synch and the user needs to be notified.
	 * <br>
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
        _thisThread = currentThread();
		while (isWatcherAlive()) {
			try {
				sleep(SLEEP_TIME); // every sleep period
			} catch (InterruptedException e) {
				// ignore
			}
            
            // check if provider change event needs firing
            if (!firingInProgress()
                && _providerChangedState != ResourceProviderEvent.NONE) {

                setFiringInProgress(true);  // prevent multiple async firing
                _display.asyncExec(firingOp);
            }
            
            // check if underlying resource is still in sync
            if (_provider != null) {
                _watchFile = _provider.getFile();
                if (_watchFile != null
                    && _provider.isEditorActive()
                    && !queryInProgress()
                    && !_watchFile.isSynchronized(IResource.NONE)) {

                    if (_modTimestamp >= _watchFile.getLocation().toFile().lastModified()) {
                        // don't raise any queries, because no recent 
                        // modification occurred
                        continue;
                    } else {
                        // underlying resource no longer in sync, and editor is active
                        // so prompt user for action
                        setQueryInProgress(true);  // prevent multiple async querying
                        _display.asyncExec(querySyncOp);
                    }
                }
            }
		}
	}

	/**
	 * Tells this thread to stop running, allowing GC.
	 */
	public void terminate() {
		_watcher = null;  // stop thread run
        // clean up
        _display = null;
        if (_listeners != null) {
            _listeners.clear();  // remove listeners
            _listeners = null;
        }
        _providerChangedState = IASResourceChangeEvent.NONE;
        _watchFile = null;
	}

    /**
     * Returns whether this thread is still alive.  This has a slightly
     * different semantic from Thread.isAlive(), and is determined by _watcher.
     * If _watcher is null, this method returns false.
     * @return boolean  <code>true</code> if alive; <code>false</code> if terminated
     */
    private boolean isWatcherAlive () {
        return _watcher == _thisThread;
    }

    /**
     * ORs the changed state to the current amalgam of change states.
     * @param kind  the provider state change to OR
     */
    public void providerChangedState(int kind) {
        _providerChangedState |= kind;
        _watcher.interrupt();
    }

    /**
     * Adds listener to this provider's list of IResourceProviderListeners.
     * @param listener  the IResourceProviderListener to add
     */
    public void addProviderListener(IResourceProviderListener listener) {
        _listeners.add(listener);
    }

    /**
     * Removes listener from this provider's list of IResourceProviderListeners.
     * @param listener  the IResourceProviderListener to remove
     */
    public void removeProviderListener (IResourceProviderListener listener) {
        _listeners.remove(listener);
    }

    /**
     * Actually fires off provider changed event on the queued changes so far.
     */
    private void fireProviderChanged() {
        if (_listeners == null) return;  // thread terminated, no one to fire
        ResourceProviderEvent event = new ResourceProviderEvent(_provider, _providerChangedState);
        // fetch the array of listeners to avoid running into CurrentModEx
        // when any of the intervening provider changed call causes the
        // listeners List to change
        IResourceProviderListener[] listeners = (IResourceProviderListener[])
            _listeners.toArray(new IResourceProviderListener[0]);
        for (int i=0; i < listeners.length; i++) {
            if (listeners[i] != null) {
                listeners[i].providerChanged(event);
            }
        }
        _providerChangedState = ResourceProviderEvent.NONE;
    }

    private void setFiringInProgress (boolean state) {
        _firingInProgress = state;
    }

    private boolean firingInProgress () {
        return _firingInProgress;
    }

    private void setQueryInProgress (boolean state) {
        _queryInProgress = state;
    }

    private boolean queryInProgress () {
        return _queryInProgress;
    }

}
