/*
 * Created on Jul 26, 2004
 *
 */
package AcmeStudio.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import AcmeStudio.core.ASModel;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Helper;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.elements.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;


/**
 * @author Bradley Schmerl
 * Created Jul 26, 2004
 * 
 * Given a parent and a representation, this command searches through the list of unbound ports or roles 
 * in the representation, and attempts to bind them to ports or roles in the parent
 */
public class BindingCommand extends ModelCommand {

	private final AcmeRepresentation	_rep;
	private final Object	_announcer;
	private List	_pointsToBind;
	private List	_newElements = new ArrayList ();
	private List	_newBindings = new ArrayList ();;
	private boolean	_createAll;
	private Map	_innerToOuterMap;

	/**
	 * 
	 * @param rep The representation that the command will be run against
	 * @param announcer The announcer
	 */
	public BindingCommand(AcmeRepresentation rep, Object announcer) {
		super ("Complete Binding");
		_rep = rep;
		_announcer = announcer;
	}
	
	
	/**
	 * @return true if the command can be executed
	 * @see org.eclipse.gef.commands.Command#canExecute()
	 **/
	public boolean canExecute() {
		return _rep != null && (_createAll || _innerToOuterMap != null);
	}
	
	
	/**
	 * @return
	 * @see AcmeStudio.model.ModelCommand#canUndo()
	 **/
	public boolean canUndo() {
		return canExecute ();
	}
	
	/**
	 * Executes teh command, adding the new bindings
	 * 
	 * @see org.eclipse.gef.commands.Command#execute()
	 *
	 */
	public void execute () {
		if (_pointsToBind == null)
			_pointsToBind = ModelHelper.collectUnboundBindingPointsInRepresentation(_rep);
		doExecute ();
	}


	/**
	 * Creates the bindings. 
	 */
	protected void doExecute() {
		ASModel.current().getRuleManager().suspendEvaluation();
		AcmeSystem parentSystem = Helper.getAcmeSystem(_rep.getParent ());
		boolean isConnRep = ModelHelper.isConnectorRep (_rep.getSystem());
		for (Iterator iter = _pointsToBind.iterator (); iter.hasNext ();) {
			if (isConnRep) {
				AcmeRole role = (AcmeRole )iter.next ();
				if (_createAll) {
					String name = role.getName ();
					createNewRole (parentSystem, role, name);
				}
				else {
					Object newRoleMap = _innerToOuterMap.get (role);
					if (newRoleMap == null) {
						createNewRole (parentSystem, role, role.getName ());
					}
					else if (newRoleMap instanceof String) {
						createNewRole (parentSystem, role, (String )newRoleMap);
					}
					else {
						AcmeRole roleToBindTo = (AcmeRole )newRoleMap;
						AcmeBinding binding = new AcmeBinding (roleToBindTo, role);
						((AcmeBindingsMap )_rep.getAbstractionMap()).addBinding (binding);
						binding.setParentRep(_rep);
						_newBindings.add (binding);
						AcmeStudioChangeManager.globalChangeManager.onAddBinding (_rep, binding, this);
					}
				}
			}
			else {
				AcmePort port = (AcmePort )iter.next ();
				if (_createAll) {
					String name = port.getName ();
					createNewPort (parentSystem, port, name);
				}
				else {
					Object newPortMap = _innerToOuterMap.get (port);
					if (newPortMap == null) {
						createNewPort (parentSystem, port, port.getName ());
					}
					else if (newPortMap instanceof String) {
						createNewPort (parentSystem, port, (String )newPortMap);
					}
					else {
						AcmePort portToBindTo = (AcmePort )newPortMap;
						AcmeBinding binding = new AcmeBinding (portToBindTo, port);
						((AcmeBindingsMap )_rep.getAbstractionMap()).addBinding(binding);
						_newBindings.add (binding);
						binding.setParentRep (_rep);
						AcmeStudioChangeManager.globalChangeManager.onAddBinding (_rep, binding, this);
					}
				}
			}
		}
		ASModel.current().getRuleManager().resumeEvaluation();
	}
	
	
	/**
	 * @param parentSystem
	 * @param role
	 * @param name
	 */
	private void createNewRole(AcmeSystem parentSystem, AcmeRole role, String name) {
		AcmeRole newRole = new AcmeRole (name);
		// Copy across the types
		AcmeEnumeration types = role.getDeclaredTypes();
		while (types.moreItems ()) {
			AcmeElementType type = (AcmeElementType )types.getNextItem ();
			AcmeFamily declaredIn = (AcmeFamily )((AcmeElement )type).getParent ();
			if (parentSystem.declaresType(declaredIn)) {
				newRole.addDeclaredType(type);
			}
		}
		types = role.getInstantiatedTypes();
		while (types.moreItems ()) {
			AcmeElementType type = (AcmeElementType )types.getNextItem ();
			AcmeFamily declaredIn = (AcmeFamily )((AcmeElement )type).getParent ();
			if (parentSystem.declaresType(declaredIn)) {
				newRole.addInstantiatedType(type);
			}
		}
		newRole.instantiateElements();
		
		// Add the role and the binding
		try {
			AcmeConnector parentConnector = (AcmeConnector )_rep.getParent();
			newRole.setName(Helper.getUniqueName(parentConnector, null, newRole.getName ()));
			parentConnector.addRole (newRole);
			_newElements.add (newRole);
			AcmeStudioChangeManager.globalChangeManager.onAddRole (parentConnector, newRole, this);
			
			AcmeBinding binding = new AcmeBinding (newRole, role);
			((AcmeBindingsMap )_rep.getAbstractionMap()).addBinding(binding);
			_newBindings.add (binding);
			AcmeStudioChangeManager.globalChangeManager.onAddBinding (_rep, binding, this);
		}
		catch (AcmeIllegalNameException e) {
			_errorReason = e;
		}
	}


	/**
	 * @param parentSystem
	 * @param port
	 * @param name
	 */
	private void createNewPort(AcmeSystem parentSystem, AcmePort port, String name) {
		AcmePort newPort = new AcmePort (name);
		
		// copy the types
		AcmeEnumeration types = port.getDeclaredTypes();
		while (types.moreItems ()) {
			AcmeElementType type = (AcmeElementType )types.getNextItem ();
			AcmeFamily declaredIn = (AcmeFamily )((AcmeElement )type).getParent ();
			if (parentSystem.declaresType(declaredIn)) {
				newPort.addDeclaredType(type);
			}
		}
		types = port.getInstantiatedTypes();
		while (types.moreItems ()) {
			AcmeElementType type = (AcmeElementType )types.getNextItem ();
			AcmeFamily declaredIn = (AcmeFamily )((AcmeElement )type).getParent ();
			if (parentSystem.declaresType(declaredIn)) {
				newPort.addInstantiatedType(type);
			}
		}
		newPort.instantiateElements();
		
		// add the port and binding
		try {
			AcmeComponent parentComponent = (AcmeComponent )_rep.getParent();
			newPort.setName(Helper.getUniqueName(parentComponent, newPort.getName ()));
			parentComponent.addPort (newPort);
			_newElements.add (newPort);
			AcmeStudioChangeManager.globalChangeManager.onAddPort (parentComponent, newPort, this);
			
			AcmeBinding binding = new AcmeBinding (newPort, port);
			((AcmeBindingsMap )_rep.getAbstractionMap()).addBinding(binding);
			_newBindings.add (binding);
			AcmeStudioChangeManager.globalChangeManager.onAddBinding (_rep, binding, this);
		}
		
		catch (AcmeIllegalNameException e) {
			_errorReason = e;
		}
	}


	/**
	 * 
	 * @see org.eclipse.gef.commands.Command#undo()
	 **/
	public void undo() {
		ASModel.current().getRuleManager().suspendEvaluation();
		for (Iterator iter = _newBindings.iterator (); iter.hasNext ();) {
			AcmeBinding binding = (AcmeBinding )iter.next ();
			((AcmeBindingsMap )_rep.getAbstractionMap()).removeBinding(binding);
			binding.setParentRep (null);
			AcmeStudioChangeManager.globalChangeManager.onRemoveBinding(_rep, binding, this);
		}
		
		boolean connectorRep = ModelHelper.isConnectorRep (_rep.getSystem());
		for (Iterator iter = _newElements.iterator (); iter.hasNext ();) {
			if (connectorRep) {
				AcmeRole role = (AcmeRole )iter.next ();
			    AcmeConnector connector = (AcmeConnector )_rep.getParent ();
				(connector).removeRole(role);
				AcmeStudioChangeManager.globalChangeManager.onRemoveRole(connector, role, this);
			}
			else {
				AcmePort port = (AcmePort )iter.next ();
				AcmeComponent component = (AcmeComponent )_rep.getParent ();
				component.removePort (port);
				AcmeStudioChangeManager.globalChangeManager.onRemovePort(component, port, this);
			}
		}
		ASModel.current ().getRuleManager().resumeEvaluation();
	}


	/**
	 * @param createAll Indicate whether to autogenerate all outer ports
	 */
	public void setCreateAll(boolean createAll) {
		_createAll = createAll;
	}


	/**
	 * Sets the mappings of all unbound innners to their corresponding outers. The value of the mapping can be:
	 * <ol>
	 * <li><b>AcmePort</b>: The existing outer port to bind to
	 * <li><b>String</b>: The name of a new outer port, which will be based on the inner port
	 * <li><b>null</b>: A new port will be auto generated
	 * </ol>
	 * @param innerToOuterMap
	 */
	public void setDisambiguation(Map innerToOuterMap) {
		_innerToOuterMap = innerToOuterMap;

	}
	/**
	 * @param pointsToBind The pointsToBind to set.
	 */
	public void setPointsToBind(List pointsToBind) {
		_pointsToBind = pointsToBind;
	}
}
