package AcmeStudio.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.draw2d.PositionConstants;
import org.eclipse.draw2d.geometry.Point;

import AcmeStudio.ui.editors.diagram.figures.SystemView;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Geometry;
import AcmeStudio.util.Helper;
import AcmeStudio.wizards.connection.PortConnectionRule;
import AcmeStudio.wizards.connection.PortRoleAssociation;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.elements.AcmeAttachment;
import acmeLib.elements.AcmeComponent;
import acmeLib.elements.AcmeConnector;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmePort;
import acmeLib.elements.AcmePortType;
import acmeLib.elements.AcmeRole;
import acmeLib.elements.AcmeSystem;

/**
 * @author yh
 * 
 * A ModelCommand to connect an AcmePort to another AcmePort by creating an
 * AcmeConnector with appropriate roles and attaching the roles to
 * corresponding ports. It causes the Acme change(s) to be announced.
 */
public class ConnectElementsCommand extends ModelCommand
{
	public static final PortConnectionRule DEFAULT_CONNECTION_RULE = new PortConnectionRule();
	static
	{
		DEFAULT_CONNECTION_RULE.addAssociation(new PortRoleAssociation("port", 
		"role", "r1"));
		DEFAULT_CONNECTION_RULE.addAssociation(new PortRoleAssociation("port", 
		"role", "r2"));
		DEFAULT_CONNECTION_RULE.setDirectionless();
	}

	protected PortConnectionRule _rule;
	protected String _role1Name = "r1";
	protected String _role2Name = "r2";
	protected String _connName = "conn";
	List _locations = new ArrayList(2);
	protected List _ports = new ArrayList(2);
	protected List _portTypes = new ArrayList(2);
	List _roles = new ArrayList(2);
	protected AcmeConnector _conn;
	private Point connLocation;
	protected List _components = new ArrayList(2);
	List _componentLocations = new ArrayList(2);
	//    private Point _location1, _location2;
	protected AcmeSystem _system;
	private SystemView _systemView;
	protected Object _announcer;
	protected boolean connConnection;
	protected boolean connSource;
	protected boolean reverse;
	protected boolean[] undoPortsExisting = new boolean[2];
	protected boolean undoConnExisting;
	private AcmeAttachment[] undoAttachments = new AcmeAttachment[2];
	/**
	 * Constructor for ConnectPortsCommand.
	 */
	public ConnectElementsCommand()
	{
		super("ConnectPortsCommand");
		undoPortsExisting[0] = false;
		undoPortsExisting[1] = false;
		undoConnExisting = false;
		undoAttachments[0] = null;
		undoAttachments[1] = null;
	}

	public ConnectElementsCommand(
			AcmePort port1, AcmePort port2, Point location1, Point location2, 
			Object announcer)
	{
		super("ConnectPortsCommand");
		_ports.add(0, port1);
		_ports.add(1, port2);
		_locations.add(0, location1);
		_locations.add(1, location2);
		_announcer = announcer;
		undoAttachments[0] = null;
		undoAttachments[1] = null;
	}

	public AcmeSystem getSystem()
	{
		if (_system == null)
		{
			AcmeElement holder = (getPort1() != null) ? (AcmeElement)getPort1()
													  : (getPort2() != null) ? (AcmeElement)getPort2()
													  : (getComponent1() != null) ? (AcmeElement)getComponent1()
													  : (getComponent2() != null) ? (AcmeElement)getComponent2()
													  : null;
			_system = Helper.getAcmeSystem(holder);
		}
		return _system;
	}

	/**
	 * Ensure that the ports and components we would be connecting
	 * are not the same item, they are not already connected, and they
	 * are not null
	 */
	public boolean canExecute()
	{
		// LFW: method modified to work when components and connectors can be
		//connected as well as ports

		return (
		getPort1() != null && getPort2() != null
		&& getPort1() != getPort2()
		&& !ModelHelper.arePortsAttached(getPort1(), getPort2())
		|| 
		(getComponent1() != null && getComponent2() != null
		&& getComponent1() != getComponent2())
		|| 
		(getPort1() != null && getComponent2() != null
		|| getPort2() != null && getComponent1() != null)
		|| 
		(getConnector() != null && (getPort1() != null || 
		getPort2() != null || 
		getComponent1() != null || 
		getComponent2() != null)));
	}

	/**
	 * @see org.eclipse.gef.commands.Command#canUndo()
	 */
	public boolean canUndo()
	{
		return true;
	}

	/**
	 * @see org.eclipse.gef.commands.Command#execute()
	 */
	public void execute()
	{
		createConnection();
		setLabel(getDescription());
	}

	public String getDescription()
	{
		String name1, name2;
		
		if (!connConnection)
		{
			name1 = getPort1().getQualifiedName();
			name2 = getPort2().getQualifiedName();
		}
		else
		{
			if (connSource)
			{
				name1 = getConnector().getQualifiedName();
				name2 = getPort2().getQualifiedName();
			}
			else
			{
				name1 = getPort1().getQualifiedName();
				name2 = getConnector().getQualifiedName();
			}
		}

		return "Connect " + name1 + " to " + name2;
	}

	/**
	 * @see org.eclipse.gef.commands.Command#redo()
	 */
	public void redo()
	{
		createConnection();
	}

	/**
	 * @see org.eclipse.gef.commands.Command#undo()
	 */
	public void undo()
	{
	    //remove the attachments that were made
	    removeAttachments();
		removeRoles(_roles);
		
		if (_conn != null)
		{
		    if (undoConnExisting == true)
		    {
		        getSystem().removeConnector(_conn);
		        AcmeStudioChangeManager.globalChangeManager
				.onRemoveConnector(getSystem(), _conn, _announcer);
		    }
			removePorts(_ports);
			
		}
	}

	private void removeAttachments()
	{
	    for (int i=0; i < 2; i++)
	    {
	        if (undoAttachments[i] != null)
	        {
	            _system.removeAttachment(undoAttachments[i]);
	            AcmeStudioChangeManager.globalChangeManager.onRemoveAttachment(_system, undoAttachments[i], this);
	            undoAttachments[i] = null;
	        }
	    }
	}
	
	private void removeRoles(List r)
	{
	    for (int i=0; i<r.size(); i++)
	    {
	        AcmeRole role = (AcmeRole)r.get(i);
	        _conn.removeRole(role);
	        AcmeStudioChangeManager.globalChangeManager.onRemoveRole(_conn, role, _announcer);
	    }
	}
	
	private void removePorts(List p)
	{
	    for (int i=0; i < p.size() ; i++)
	    {
	        if (undoPortsExisting[i] == true)
	        {
		        AcmePort port = (AcmePort)p.get(i);
		        
		        if (_system.getAttachments(port).getSize() == 0)
		            ((AcmeComponent)_components.get(i)).removePort(port);
		        
		        AcmeStudioChangeManager.globalChangeManager.onRemovePort(((AcmeComponent)_components.get(i)), port, _announcer);
	        }
	    }
	}
	
	/**
	 * This is the main method of the command.  It creates the needed
	 * elements to create the connection, and goes through the steps
	 * of actually connecting them
	 *
	 */
	private void createConnection()
	{
		if (_rule == DEFAULT_CONNECTION_RULE)
		{
			try
			{
			    undoConnExisting = true;
				_conn = createConnector(_connName, "connector");
				getSystem().addConnector(_conn);
				AcmeStudioChangeManager.globalChangeManager.onAddConnector(getSystem(), _conn, _announcer);
				AcmeRole role1 = createRole(_conn, "r1", "role", (Point) _locations.get(0));
				_roles.add(role1);
				AcmeRole role2 = createRole(_conn, "r2", "role",
				(Point) _locations.get(1));
				_roles.add(role2);
				getSystem().addAttachment((AcmePort) _ports.get(0), role1);
				getSystem().addAttachment((AcmePort) _ports.get(1), role2);
				AcmeStudioChangeManager.globalChangeManager.onAddAttachment(
				        getSystem(), getSystem().getAttachment((AcmePort) _ports.get(0), role1), this);
				AcmeStudioChangeManager.globalChangeManager.onAddAttachment(
				        getSystem(), getSystem().getAttachment((AcmePort) _ports.get(1), role2), this);
			}
			catch (AcmeIllegalNameException e)
			{
				e.printStackTrace();
			}
		}
		else
		{
			if (!connConnection) //the connection does not involve a connector
			{
			    //create the connector
				_conn = createConnector(_connName, _rule.getConnectorType());
				undoConnExisting = true;
				try
				{
					getSystem().addConnector(_conn);
					AcmeStudioChangeManager.globalChangeManager
					.onAddConnector(getSystem(), _conn, _announcer);
					if (_conn != null)
					{
						PortRoleAssociation[] assocs = _rule.getAssociations();
						for (int i = 0; i < _ports.size(); i++)
						{						
						    //create the roles as necessary
						    AcmePort port = (AcmePort) _ports.get(i);
						    PortRoleAssociation assoc = assocs[(reverse) ? 1 - i : i];
						    String roleName = "";
						    if (assoc != null && assoc.getRoleName() != null && !assoc.getRoleName().equals(""))
							{
								roleName = assoc.getRoleName();
							}
							else
							{
								roleName = Helper.getUniqueName(_conn, "r");
							}
							AcmeRole role = createRole(_conn, roleName, assoc
							.getRoleType(), (Point) _locations.get(i));
							_roles.add(role);
							undoAttachments[i] = new AcmeAttachment(port, role);
							getSystem().addAttachment(undoAttachments[i]);
							AcmeStudioChangeManager.globalChangeManager
							.onAddAttachment(getSystem(), getSystem()
							.getAttachment(port, role), _announcer);
						}
					}
				}
				catch (AcmeIllegalNameException e)
				{
					e.printStackTrace();
				}
			}
			else
			{
				//All we need to do is create the role between the connector and the created port
				undoConnExisting = false;
				PortRoleAssociation assoc = ((PortRoleAssociation[])_rule.getAssociations())[connSource ? 1 : 0];
				
				AcmePort port = (AcmePort)_ports.get(connSource ? 1 : 0);
				
				String roleName;
				if (assoc != null && assoc.getRoleName() != null && !assoc.getRoleName().equals(""))
					roleName = assoc.getRoleName();
				else
					roleName = Helper.getUniqueName(_conn, "r");
				
				AcmeRole role = createRole(getConnector(),  roleName, 
										   					assoc.getRoleType(), 
															(Point) _locations.get(connSource ? 1 : 0));
				_roles.add(role);
				
				undoAttachments[connSource ? 1 : 0] = new AcmeAttachment(port, role);
				getSystem().addAttachment(undoAttachments[connSource ? 1 : 0]);
				
				AcmeStudioChangeManager.globalChangeManager
				.onAddAttachment(getSystem(), getSystem()
				.getAttachment(port, role), _announcer);
			}
		}
	}

	/**
	 * This method creates a role based on the parameters pased in
	 * 
	 * @param connector The connector the role should be attached to
	 * @param name The name of the connector
	 * @param type The type the connector should have
	 * @param location Where the connector should be created
	 * 
	 * @return An AcmeRole representing the created Role
	 */
	private AcmeRole createRole(AcmeConnector connector, String name, 
	String type, Point location)
	{
		name = Helper.getUniqueName(connector, name);
		AcmeElementType acmeType = null;
		
		if (_system == null)
		    _system = getSystem();
		
		if (type != null && !type.equals("") && !"role".equals(type))
		{
		    //get the type out of the string passed in
			acmeType = (AcmeElementType) _system.lookupName(type, true);
		}
		AcmeEnumeration roles = connector.getRoles();
		AcmeRole role = null;
		//if there are any unattached roles of the correct type on this
		//connector, use them in the connection
		while (roles.moreItems() && role == null)
		{
			AcmeRole r = (AcmeRole) roles.getNextItem();
			if (r.declaresType(acmeType))
			{
				 if (_system.getAttachments (r).getSize () == 0) 
				 {
				 	role = r;
				 	SystemView.setLocation(role, location);
				 }
			}
		}
		
		if (role == null) //if there was no suitable role to use
		{
			role = new AcmeRole(name);
			SystemView.setLocation(role, location);
			try
			{
				connector.addRole(role);
				AcmeStudioChangeManager.globalChangeManager
				.onAddRole(connector, role, _announcer);
			}
			catch (AcmeIllegalNameException e)
			{
				return null;
			}
			if (acmeType != null)
			{
				//            AcmeType acmeType = getFamily().getType(type);
				role.addDeclaredType(acmeType);
				role.addInstantiatedType(acmeType);
				role.instantiateElements();
			}
		}
		role.setParentDesign(_system.getParentDesign());
		return role;
	}

	/**
	 * This method creates a port as specified in the parameters
	 * 
	 * @param name The name of the port
	 * @param component The component that should be this port's parent
	 * @param type The type of the port
	 * @param location The place to create the port
	 * @param i Whether this is the source or target item of the connection
	 * @return
	 */
	protected AcmePort createPort(String name, AcmeComponent component, 
	AcmePortType type, Point location, int i)
	{
		name = Helper.getUniqueName(component, name);
		AcmePort port = null;
		// BRS: Changed create port to handle types correctly.
		port = new AcmePort (name);
		if (type != null) {
			port.addDeclaredType(type);
			port.addInstantiatedType(type);
			port.instantiateElements();
		}
		
		port.setParentDesign(component.getParentDesign());
		
		//Figure out where on the component this port should be displayed
		
		Point compLocation = null;
		Point compLocation2 = null;
		
		compLocation = SystemView.getLocation(
		            ((AcmeElement)_components.get(0) != null) ? (AcmeElement)_components.get(0) : (AcmeElement)_conn, 
		            false);
		
		compLocation2 = SystemView.getLocation(
		            ((AcmeElement)_components.get(1) != null) ? (AcmeElement)_components.get(1) : (AcmeElement)_conn, 
		            false);
		
		int side;
		
		//Figure out whether the source is on the right of the target, or the
		//target is on the right of the source
		if ((i == 0 && compLocation.x <= compLocation2.x) || (i == 1 && compLocation.x > compLocation2.x))
		    side = Geometry.getSideOfPoint (compLocation2, compLocation);
		else
		    side = Geometry.getOppositeSide(Geometry.getSideOfPoint(compLocation2, compLocation));
		    
		Point finalP = new Point();

		if ((AcmeElement)_components.get(i) != null)
		{
		    if (i==0)
		        finalP = compLocation;
		    else
		        finalP = compLocation2;
		    
			//First do it for the first part of the connection
			if (side == PositionConstants.NORTH)
			    finalP.y = ((i==0) ? compLocation.y : compLocation2.y) - (Helper.getSize((AcmeComponent)_components.get(i)).height / 2);
			else if (side == PositionConstants.EAST)
			    finalP.x = ((i==0) ? compLocation.x : compLocation2.x) + (Helper.getSize((AcmeComponent)_components.get(i)).width / 2);
			else if (side == PositionConstants.SOUTH)
			    finalP.y = ((i==0) ? compLocation.y : compLocation2.y) + (Helper.getSize((AcmeComponent)_components.get(i)).height / 2);
			else if (side == PositionConstants.WEST)
			    finalP.x = ((i==0) ? compLocation.x : compLocation2.x) - (Helper.getSize((AcmeComponent)_components.get(i)).width / 2);
		}
		
		//finalP = finalP.translate(location);
		Point temp = new Point(finalP.x - location.x, finalP.y - location.y);
		
		SystemView.setLocation(port, temp);
		_locations.set(i, location);
		
		//With the location determined, add the port and set necessary properties
		try
		{
			component.addPort(port);
			if (type != null)
			{
				port.addDeclaredType(type);
				port.addInstantiatedType(type);
				port.instantiateElements();
			}
			AcmeStudioChangeManager.globalChangeManager.onAddPort(component, 
			port, 
			_announcer);
		}
		catch (AcmeIllegalNameException e)
		{
			return null;
		}
		return port;
	}

	/**
	 * This method creates an AcmeConnector based on the parameters
	 * 
	 * @param name The name of the connector being created
	 * @param type The type of the connector to be created
	 * 
	 * @return An AcmeConnector holding the created object
	 */
	
	private AcmeConnector createConnector(String name, String type)
	{
		name = Helper.getUniqueName(getSystem(), name);
		AcmeConnector conn = new AcmeConnector(name);
		int totalX = 0;
		int totalY = 0;
		
		//Determine the location to create the connector at by 
		//iterating over objects in the connection and trying to
		//find the middle
		for (Iterator iter = _locations.iterator(); iter.hasNext();)
		{
			Point point = (Point) iter.next();
			totalX += point.x;
			totalY += point.y;
		}
		Point location = new Point(totalX / _locations.size(), totalY
		/ _locations.size());
		SystemView.setLocation(conn, location);
		
		if (type != null && !type.equals("") && !"connector".equals(type))
		{
			AcmeElementType acmeType = (AcmeElementType) _system
			.lookupName(type, true);
			//            AcmeType acmeType = getFamily().getType(type);
			conn.addDeclaredType(acmeType);
			conn.addInstantiatedType(acmeType);
			conn.instantiateElements();
		}
		conn.setParentDesign(_system.getParentDesign());
		return conn;
	}

	/******************************************************
	 * 
	 * THIS SECTION OF THE CODE AND BELOW CONTAINS GETTER/SETTER
	 * FUNCTIONS FOR THIS CLASS AND ITS MEMBERS, AS WELL AS 
	 * UTILITY ACCESS FUNCTIONS FOR THE CLASS
	 * 
	 ******************************************************/
	/**
	 * @return SystemView
	 */
	public SystemView getSystemView()
	{
		return _systemView;
	}

	/**
	 * Sets the _systemView.
	 * 
	 * @param _systemView
	 *                 The _systemView to set
	 */
	public void setSystemView(SystemView _systemView)
	{
		this._systemView = _systemView;
	}
    /**
     * @return Returns the _rule.
     */
    public PortConnectionRule getRule()
    {
        return _rule;
    }
    /**
     * @param _rule The _rule to set.
     */
    public void setRule(PortConnectionRule _rule)
    {
        this._rule = _rule;
    }
	/**
	 * @return
	 */
	public Point getLocation1()
	{
		return (Point) getArrayValue(_locations, 0);
	}

	/**
	 * @param _location1
	 */
	public void setLocation1(Point location1)
	{
		setArrayValue(_locations, 0, location1);
	}

	/**
	 * @return
	 */
	public AcmePort getPort1()
	{
		return (AcmePort) getArrayValue(_ports, 0);
	}

	/**
	 * @param _port1
	 */
	public void setPort1(AcmePort port1)
	{
		setArrayValue(_ports, 0, port1);
		if (_ports.size() >= 1)
			_ports.set(0, port1);
		else
			_ports.add(port1);
	}

	public void setConnector(AcmeConnector conn1)
	{
		_conn = conn1;
	}

	public AcmeConnector getConnector()
	{
		return _conn;
	}

	public void setLocationConn(Point locConn)
	{
		connLocation = locConn;
	}

	public Point getLocationConn()
	{
		return connLocation;
	}

	/**
	 * This method is a safe accessor of data in a list.
	 * 
	 * @author LFW
	 * 
	 * @param list item to retrive things from
	 * @param int index of item to retrieve from list
	 * @return item at index, or null, if OB
	 */
	private Object getArrayValue(List list, int i)
	{
		if (i + 1 > list.size())
			return null;
		else
			return list.get(i);
	}

	/**
	 * @param _ports2
	 * @param i
	 * @param port1
	 */
	private void setArrayValue(List list, int i, Object o)
	{
		if (list.size() - 1 >= i)
		{
			list.set(i, o);
		}
		else
		{
			for (int index = list.size(); index < i; index++)
			{
				list.add(null);
			}
			list.add(o);
		}
	}

	/**
	 * @return
	 */
	public Point getLocation2()
	{
		return (Point) getArrayValue(_locations, 1);
	}

	/**
	 * @param _location1
	 */
	public void setLocation2(Point location2)
	{
		setArrayValue(_locations, 1, location2);
	}

	/**
	 * @return
	 */
	public AcmePort getPort2()
	{
		return (AcmePort) getArrayValue(_ports, 1);
	}

	/**
	 * @param _port1
	 */
	public void setPort2(AcmePort port2)
	{
		setArrayValue(_ports, 1, port2);
	}

	public AcmePortType getPort1Type()
	{
		/*
		 * if (getArrayValue (_portTypes, 0) == null) { setArrayValue
		 * (_portTypes, 0, getPort1 ().getDeclaredTypes().getNextItem ());
		 */
		return (AcmePortType) getArrayValue(_portTypes, 0);
	}

	public AcmePortType getPort2Type()
	{
		/*
		 * if (getArrayValue (_portTypes, 1) == null) { setArrayValue
		 * (_portTypes, 1, getPort2 ().getDeclaredTypes().getNextItem ());
		 */
		return (AcmePortType) getArrayValue(_portTypes, 1);
	}

	public Point getComponentLocation1()
	{
		return (Point) getArrayValue(_componentLocations, 0);
	}

	/**
	 * @param _Component1
	 */
	public void setComponentLocation1(Point component1)
	{
		setArrayValue(_componentLocations, 0, component1);
	}

	/**
	 * @return
	 */
	public AcmeComponent getComponent1()
	{
		return (AcmeComponent) getArrayValue(_components, 0);
	}

	/**
	 * @param _port1
	 */
	public void setComponent1(AcmeComponent component1)
	{
		setArrayValue(_components, 0, component1);

		//Commented out by LFW --
		//Made no sense in the generalized connection structure
		/*
		 * if (_components.size() >= 1) _components.set(0, component1); else
		 */
	}

	public Point getComponentLocation2()
	{
		return (Point) getArrayValue(_componentLocations, 1);
	}

	/**
	 * @param _Component1
	 */
	public void setComponentLocation2(Point component2)
	{
		setArrayValue(_componentLocations, 1, component2);
	}

	/**
	 * @return
	 */
	public AcmeComponent getComponent2()
	{
		return (AcmeComponent) getArrayValue(_components, 1);
	}

	/**
	 * @param _port1
	 */
	public void setComponent2(AcmeComponent component2)
	{
		setArrayValue(_components, 1, component2);

		//Below commented out by LFW --
		//Made no sense in the generalized connection structure
		/*
		 * if (_components.size() >= 1) _components.set(0, component2); else
		 */
	}  
	
	
}
