package AcmeStudio.model;

import java.util.Iterator;
import java.util.LinkedList;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.ui.editors.diagram.figures.AcmeVisElement;
import AcmeStudio.ui.editors.diagram.figures.SystemView;
import AcmeStudio.ui.editors.diagram.styles.ComponentVisualization;
import AcmeStudio.ui.editors.diagram.styles.ConnectorVisualization;
import AcmeStudio.ui.editors.diagram.styles.PlacementRule;
import AcmeStudio.ui.editors.diagram.styles.VisualStyleRegistry;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Helper;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.elements.AcmeAttachment;
import acmeLib.elements.AcmeComponent;
import acmeLib.elements.AcmeConnector;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmePort;
import acmeLib.elements.AcmeRole;
import acmeLib.elements.AcmeSystem;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;

/**
 * @author schmerl
 *
 * A ModelCommand to create a AcmeElement on another AcmeElement.  This
 * command can be used for creating AcmeComponent, AcmeConnector, AcmePort,
 * AcmeRole, and AcmeAttachment.  It causes the Acme change(s) to be announced.
 */
public class CreateElementCommand extends ModelCommand {
    protected Object _announcer;

    private AcmeElement _child;
    private Point _location;
    private AcmeElement _parent;
    private AcmeVisElement _parentVis;
    private AcmeSystem _system;
    private SystemView _systemView;

	private Dimension	_size;

    /**
     * Constructor for CreateElementCommand.
     * @param label
     */
    public CreateElementCommand() {
        super("CreateCommand");
    }

    public CreateElementCommand(
        AcmeElement parent,
        AcmeElement child,
        Point location,
        Object announcer) {
        super("CreateCommand");
        _parent = parent;
        if (_parent instanceof AcmeSystem)
            _system = (AcmeSystem) _parent;
        setChild(child);
        _location = location;
        _announcer = announcer;
    }

    private PlacementRule arrangePort(
        AcmePort port,
        PlacementRule rule,
        ComponentVisualization cv) {
        LinkedList ll = cv.getPortPlacementRules();
        AcmeEnumeration ae = port.getDeclaredTypes();
        boolean found = false;
        while (!found && ae.moreItems()) {
            String portType =
                ((AcmeElementType) ae.getNextItem()).getTypeName();
            Iterator iter = ll.iterator();
            while (!found && iter.hasNext()) {
                PlacementRule aRule = (PlacementRule) iter.next();
                if (aRule._type.equals(portType)
                    || aRule._type.equals("port")) {
                    found = true;
                    rule = aRule;
                }
            }
        }
        return rule;
    }

    /**
     * @see org.eclipse.gef.commands.Command#canExecute()
     */
    public boolean canExecute() {
        boolean allowedType = true;
        if (_child instanceof AcmePort && _parentVis != null) {
            Object obj = _parentVis.getVisualization();
            ComponentVisualization cv = null;
            if (obj instanceof ComponentVisualization)
                cv = (ComponentVisualization) obj;
            if (cv != null && cv.restrictPortTypes()) {
                allowedType = false;
                LinkedList ll = cv.getAllowedPortTypes();
                AcmeEnumeration ae = _child.getDeclaredTypes();
                while (ae.moreItems()) {
                    String portType =
                        ((AcmeElementType) ae.getNextItem()).getTypeName();
                    if (ll.contains(portType)) {
                        allowedType = true;
                        break;
                    }
                }
            }
        }
        return (allowedType && super.canExecute());
    }

    /**
     * @see org.eclipse.gef.commands.Command#canUndo()
     */
    public boolean canUndo() {
        return true;
    }
    
    public void setSize (Dimension size) {
    	_size = size;
    }
    /**
     * @see org.eclipse.gef.commands.Command#execute()
     */
    public void execute() {
    	if (_size == null) _size = Helper.getSize(_child);
    	else {
    		_systemView.translateToRelative (_size);
    		// The size was set by dragging to create, so let's translate the
    		// location back
    		_location.translate(_size.width/2, _size.height/2);
    	}
    	Helper.setSize (_child, _size);
        try {
            if (_child instanceof AcmeComponent) {
            	 if (!Helper.isUniqueComponentName(_child.getName(), _system)) {
                    _child.setName(
                        Helper.getUniqueName(
                            _system,
                            _child.getName()));
                }
                AcmeComponent comp = (AcmeComponent) _child;
                if (!(_parent instanceof AcmeElementType)) {

                    AcmeEnumeration aePorts = comp.getPorts();
//                    Dimension size = Helper.getSize(comp);
                    if (_systemView != null) {
                        _systemView.setElementLocation(_child, _location, false);
                    } else {
                        SystemView.setLocation(_child, _location);
                    }
                    
                    // Make sure any ports defined inside the component get
                    // set appropriate locations
                    while (aePorts.moreItems()) {
                        AcmePort port = (AcmePort) aePorts.getNextItem();
                        Point pLocation =
                            Helper.calculateRandomPortPoint(_size, port);
                        pLocation.x += _location.x;
                        pLocation.y += _location.y;
                        Object obj = VisualStyleRegistry.getVisualization(comp);
                        ComponentVisualization cv = null;
                        if (obj instanceof ComponentVisualization) {
                            cv = (ComponentVisualization) obj;
                        }
                        PlacementRule rule = null;
                        if (cv != null) {
                            if (cv.forcePortPlacement()) {
                                rule = arrangePort(port, rule, cv);
                            }
                        }
                        if (_systemView != null) {
                            _systemView.savePortPosition(
                                port,
                                comp,
                                pLocation,
                                rule);
                        }
                    }
                    //                    if (_systemView != null) {
                    //                        _systemView.setElementLocation(_child, _location, false);
                    //                    } else {
                    //                        SystemView.setLocation(_child, _location);
                    //                    }
                }
                _system.addComponent((AcmeComponent) _child);
                _child.setParentDesign(_system.getParentDesign());
                AcmeStudioChangeManager.globalChangeManager.onAddComponent(
                    _system,
                    (AcmeComponent) _child,
                    _announcer);
            } else if (_child instanceof AcmeConnector) {
                AcmeConnector conn = (AcmeConnector) _child;
                if (!Helper.isUniqueConnectorName(_child.getName(), _system)) {
                    _child.setName(
                        Helper.getUniqueName(
                            _system,
                            _child.getName()));
                }
                SystemView.setLocation(_child, _location);
                if (!(_parent instanceof AcmeElementType)) {
                    AcmeEnumeration aeRoles = conn.getRoles();
                    int roleCount = aeRoles.getSize();
                    if (roleCount == 0) {
                        Object obj =
                            VisualStyleRegistry.getVisualization(_child);
                        ConnectorVisualization cv = null;
                        int countRoleNumber = 2;
                        if (obj instanceof ConnectorVisualization) {
                            cv = (ConnectorVisualization) obj;
                        }
                        if (cv != null) {
                            countRoleNumber = cv.getDefaultRoleCount();
                        }
                        for (int i = 0; i < countRoleNumber; i++) {
                            AcmeRole role = new AcmeRole("r" + (i + 1));
                            createRole(role, i, countRoleNumber);
                            conn.addRole(role);
                        }
                    } else {
                        int i = 0;
                        int count = aeRoles.getSize();
                        while (aeRoles.moreItems()) {
                            AcmeRole role = (AcmeRole) aeRoles.getNextItem();
                            createRole(role, i++, count);
                        }
                    }
                }
                _system.addConnector((AcmeConnector) _child);
                _child.setParentDesign(_system.getParentDesign());
                AcmeStudioChangeManager.globalChangeManager.onAddConnector(
                    _system,
                    (AcmeConnector) _child,
                    _announcer);
            } else if (_child instanceof AcmePort) {
                if (_parent == null) {
                    _parent = Helper.findClosestComponent(_system, _location);
                }
                AcmeComponent parentComponent = (AcmeComponent) _parent;
                if (!Helper
                    .isUniquePortName(_child.getName(), parentComponent)) {
                    _child.setName(
                        Helper.getUniqueName(
                            parentComponent,
                            _child.getName()));
                }
                parentComponent.addPort((AcmePort) _child);
                _child.setParentDesign(parentComponent.getParentDesign());
                if (!(_parent instanceof AcmeElementType)) {
                    PlacementRule rule = null;
                    ComponentVisualization cv = null;
                    if (_parentVis != null) {
                        Object obj = _parentVis.getVisualization();
                        if (obj instanceof ComponentVisualization)
                            cv = (ComponentVisualization) obj;
                        if (cv != null && cv.forcePortPlacement()) {
                            rule = arrangePort((AcmePort) _child, rule, cv);
                        }
                    }
                    if (_systemView != null) {
                    _systemView.savePortPosition(
                        (AcmePort) _child,
                        parentComponent,
                        _location,
                        rule);
                    }
                }
                AcmeStudioChangeManager.globalChangeManager.onAddPort(
                    parentComponent,
                    (AcmePort) _child,
                    _announcer);
            } else if (_child instanceof AcmeRole) {
                if (_parent == null) {
                    _parent = Helper.findClosestConnector(_system, _location);
                }
                AcmeConnector parentConnector = (AcmeConnector) _parent;
                if (!Helper
                    .isUniqueRoleName(_child.getName(), parentConnector)) {
                    _child.setName(
                        Helper.getUniqueName(
                            parentConnector,
                            _child.getName()));
                }
                parentConnector.addRole((AcmeRole) _child);
                _child.setParentDesign(parentConnector.getParentDesign());
                //Helper.setLocation(_child, _location);
                if (!(_parent instanceof AcmeElementType)) {
                    int x = _location.x + 80;
                    int y = _location.y;
                    if (y > 60) {
                        y -= 60;
                    } else {
                        y += 60;
                    }
                    SystemView.setLocation(_child, new Point(x, y));
                }
                AcmeStudioChangeManager.globalChangeManager.onAddRole(
                    parentConnector,
                    (AcmeRole) _child,
                    _announcer);
            } else if (_child instanceof AcmeAttachment) {
                _system.addAttachment((AcmeAttachment) _child);
                _child.setParentDesign(_system.getParentDesign());
                AcmeStudioChangeManager.globalChangeManager.onAddAttachment(
                    _system,
                    (AcmeAttachment) _child,
                    _announcer);
            } else if (_child instanceof AcmeBinding) {
                AcmeBinding binding = (AcmeBinding) _child;
                AcmeRepresentation rep = binding.getParentRep();
                AcmeBindingsMap bm = (AcmeBindingsMap )rep.getAbstractionMap();
                if (bm == null) {
                    bm = new AcmeBindingsMap ();
                    rep.setAbstractionMap(bm);
                    bm.setParentRep (rep);
                }
                bm.addBinding (binding);
                AcmeStudioChangeManager.globalChangeManager.onAddBinding(
                    rep, binding, _announcer);
            }
        } catch (AcmeIllegalNameException e) {
            this._errorReason = e;
        }
        //        Helper.setLocation(_child, _location);
        setLabel(AcmeStudioPlugin.getMessage("CreateElementCommand.CreateLabel") + _child.getName()); //$NON-NLS-1$
    }

    public String getDescription() {
        String name = _child.getQualifiedName();
        return AcmeStudioPlugin.getMessage("CreateElementCommand.CreateLabel") + name; //$NON-NLS-1$
    }

    public AcmeElement getParent() {
        return _parent;
    }

    /**
     * @see org.eclipse.gef.commands.Command#redo()
     */
    public void redo() {
        try {
            if (_child instanceof AcmeComponent) {
                _system.addComponent((AcmeComponent) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddComponent(
                    _system,
                    (AcmeComponent) _child,
                    _announcer);
            } else if (_child instanceof AcmeConnector) {
                _system.addConnector((AcmeConnector) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddConnector(
                    _system,
                    (AcmeConnector) _child,
                    _announcer);
            } else if (_child instanceof AcmePort) {
                ((AcmeComponent) _parent).addPort((AcmePort) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddPort(
                    (AcmeComponent) _parent,
                    (AcmePort) _child,
                    _announcer);
            } else if (_child instanceof AcmeRole) {
                ((AcmeConnector) _parent).addRole((AcmeRole) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddRole(
                    (AcmeConnector) _parent,
                    (AcmeRole) _child,
                    _announcer);
            } else if (_child instanceof AcmeAttachment) {
                _system.addAttachment((AcmeAttachment) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddAttachment(
                    _system,
                    (AcmeAttachment) _child,
                    _announcer);
            } else if (_child instanceof AcmeBinding) {
                execute();
            }
        } catch (Exception e) {
        }
    }

    /**
     * Sets the child.
     * @param child The child to set
     */
    public void setChild(AcmeElement child) {
        this._child = child;
        if (_child instanceof AcmeComponent) {
            if (!Helper.isUniqueComponentName(_child.getName(), _system)) {
                try {
                    _child.setName(
                        Helper.getUniqueName(
                            _system,
                            _child.getName()));
                } catch (AcmeIllegalNameException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Sets the location.
     * @param location The location to set
     */
    public void setLocation(Point location) {
        this._location = location;
    }

    /**
     * Sets the parent.
     * @param parent The parent to set
     */
    public void setParent(AcmeElement parent) {
        this._parent = parent;
    }

    /**
     * Sets the parent.
     * @param parent The parent to set
     */
    public void setParentVis(AcmeVisElement parent) {
        this._parentVis = parent;
    }


    /**
     * Sets the system.
     * @param system The system to set
     */
    public void setSystem(AcmeSystem system) {
        this._system = system;
    }

    /**
     * @see org.eclipse.gef.commands.Command#undo()
     */
    public void undo() {
        try {
            if (_child instanceof AcmeComponent) {
                _system.removeComponent((AcmeComponent) _child);
                AcmeStudioChangeManager.globalChangeManager.onRemoveComponent(
                    _system,
                    (AcmeComponent) _child,
                    this);
            } else if (_child instanceof AcmeConnector) {
                _system.removeConnector((AcmeConnector) _child);
                AcmeStudioChangeManager.globalChangeManager.onRemoveComponent(
                    _system,
                    (AcmeComponent) _child,
                    this);
            } else if (_child instanceof AcmeRole) {
                ((AcmeConnector) _parent).removeRole((AcmeRole) _child);
                AcmeStudioChangeManager.globalChangeManager.onRemoveRole(
                    (AcmeConnector) _parent,
                    (AcmeRole) _child,
                    this);
            } else if (_child instanceof AcmePort) {
                ((AcmeComponent) _parent).removePort((AcmePort) _child);
                AcmeStudioChangeManager.globalChangeManager.onRemovePort(
                    (AcmeComponent) _parent,
                    (AcmePort) _child,
                    this);
            } else if (_child instanceof AcmeAttachment) {
                _system.removeAttachment((AcmeAttachment) _child);
                AcmeStudioChangeManager.globalChangeManager.onRemoveAttachment(
                    _system,
                    (AcmeAttachment) _child,
                    this);
            } else if (_child instanceof AcmeBinding) {
                AcmeBinding binding = (AcmeBinding) _child;
                AcmeRepresentation rep = binding.getParentRep();
                AcmeBindingsMap bm = (AcmeBindingsMap )rep.getAbstractionMap();
                if (bm != null) {
                    bm.removeBinding(binding);
                }
                AcmeStudioChangeManager.globalChangeManager.onRemoveBinding(rep, binding, this);
            }
        } catch (Exception e) {
        }

    }

    private void createRole(AcmeRole role, int index, int count) {
        int r = 60;
        double alpha = 2 * Math.PI / count;
        double delta = Math.PI / 6 + alpha * index;
        int x = (int) (r * Math.cos(delta));
        int y = (int) (r * Math.sin(delta));
        SystemView.setLocation(
            role,
            new Point(_location.x + x, _location.y + y));
    }
    /**
     * @return SystemView
     */
    public SystemView getSystemView() {
        return _systemView;
    }

    /**
     * Sets the _systemView.
     * @param _systemView The _systemView to set
     */
    public void setSystemView(SystemView _systemView) {
        this._systemView = _systemView;
    }

}
