package AcmeStudio.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Debug;
import AcmeStudio.util.Helper;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.elements.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;

/**
 * @author Bradley Schmerl
 * Act
 */
public class DeleteElementCommand extends ModelCommand {

    protected AcmeElement _child;
    protected AcmeElement _parent;
    private Object _announcer;
    private List _deletedAttachments = new ArrayList ();

    /**
     * Constructor for DeleteElementCommand.
     */
    public DeleteElementCommand() {
        super("Delete");
    }

    public DeleteElementCommand (AcmeElement parent, AcmeElement child, Object announcer) {
        super ("Delete");
        this._parent = parent;
        this._child = child;
        this._announcer = announcer;   
    }

    /**
     * @see org.eclipse.gef.commands.Command#canExecute()
     */
    public boolean canExecute() {
        return super.canExecute();
    }

    /**
     * @see org.eclipse.gef.commands.Command#canUndo()
     */
    public boolean canUndo() {
        return super.canUndo();
    }

    /**
     * @see org.eclipse.gef.commands.Command#execute()
     */
    public void execute() {
    	if (_child instanceof AcmeBinding) {
    		// Set name to indicate binding
    		AcmeBinding binding = (AcmeBinding )_child;
    		setLabel ("Delete Binding " + binding.getOuter().getQualifiedName() + " <->" + binding.getInner ().getQualifiedName ());
    	}
        else 
        	setLabel ("Delete " + _child.getName());
        primExecute ();
    }

    /**
     * @see org.eclipse.gef.commands.Command#undo()
     */
    public void undo() {
        try {
            AcmeSystem parentSystem = null;
            if (_child instanceof AcmeComponent) {
                parentSystem = (AcmeSystem) _parent;
                parentSystem.addComponent((AcmeComponent) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddComponent(
                    parentSystem,
                    (AcmeComponent) _child,
                    _announcer);
            }
            else if (_child instanceof AcmeConnector) {
                parentSystem = (AcmeSystem) _parent;
                parentSystem.addConnector((AcmeConnector) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddConnector(
                    parentSystem,
                    (AcmeConnector) _child,
                    _announcer);
            }
            else if (_child instanceof AcmePort) {
                ((AcmeComponent) _parent).addPort((AcmePort) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddPort(
                    (AcmeComponent) _parent,
                    (AcmePort) _child,
                    _announcer);
            }
            else if (_child instanceof AcmeRole) {
            	if (_parent != null) {
	                ((AcmeConnector) _parent).addRole((AcmeRole) _child);
	                AcmeStudioChangeManager.globalChangeManager.onAddRole(
	                    (AcmeConnector) _parent,
	                    (AcmeRole) _child,
	                    _announcer);
            	}
            }
            else if (_child instanceof AcmeAttachment) {
                parentSystem = (AcmeSystem) _parent;
                parentSystem.addAttachment((AcmeAttachment) _child);
                AcmeStudioChangeManager.globalChangeManager.onAddAttachment(
                    parentSystem,
                    (AcmeAttachment) _child,
                    _announcer);
            }
            else if (_child instanceof AcmeBinding) {
                AcmeBinding binding = (AcmeBinding) _child;
                AcmeRepresentation rep = binding.getParentRep();
                AcmeBindingsMap bm = (AcmeBindingsMap )rep.getAbstractionMap();
                if (bm == null) {
                    bm = new AcmeBindingsMap ();
                    rep.setAbstractionMap(bm);
                    bm.setParentRep (rep);
                }
                bm.addBinding (binding);
                AcmeStudioChangeManager.globalChangeManager.onAddBinding(
                    rep, binding, _announcer);
            }

            if (parentSystem != null) {
                Iterator it = _deletedAttachments.iterator();
                while (it.hasNext()) {
                    AcmeAttachment att = (AcmeAttachment) it.next();
                    parentSystem.addAttachment(att);
                    AcmeStudioChangeManager.globalChangeManager.onAddAttachment(parentSystem, att, _announcer);

                }
                _deletedAttachments.clear();
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }

    }


    public String getDescription() {
        try {
            String name = _child.getName ();
            Debug.diagnoseln("DeleteElementCommand.getDescription():  child name == " + name);
            return "Delete " + name;
        } catch (Exception e) {
            e.printStackTrace ();
        }
        return "Delete";
    }

    private void deleteAttachments (AcmeElement el, AcmeSystem sys) {
        List leafElements = new ArrayList ();
        
        if (el instanceof AcmeRole || el instanceof AcmePort) {
            leafElements.add(el);
        }
        else if (el instanceof AcmeComponent) {
            AcmeEnumeration ports = ((AcmeComponent )el).getPorts ();
            while (ports.moreItems ()) {
                leafElements.add (ports.getNextItem ());
            }
        }
        else if (el instanceof AcmeConnector) {
            AcmeEnumeration roles = ((AcmeConnector )el).getRoles ();
            while (roles.moreItems ()) {
                leafElements.add (roles.getNextItem ());
            }
        }
        
        Iterator elementIt = leafElements.iterator();
        
        AcmeEnumeration attachments = new AcmeEnumeration ();
        while (elementIt.hasNext()) {
            Object element = elementIt.next ();
            if (element instanceof AcmeRole) {
                AcmeEnumeration atts = sys.getAttachments ((AcmeRole )element);
                attachments.concatenate(atts);
            }
            else if (element instanceof AcmePort) {
                AcmeEnumeration atts = sys.getAttachments ((AcmePort )element);
                attachments.concatenate(atts);
                
            }
        }
        
        while (attachments.moreItems ()) {
            _deletedAttachments.add (attachments.getNextItem ());
        }
    }

    protected void primExecute() {
        try {
//          Helper.purgeElement(_child);
            if (_child instanceof AcmeComponent) {
                AcmeSystem sys = (AcmeSystem )_child.getParent();
                deleteAttachments (_child, sys);
                _parent = sys;
                sys.removeComponent ((AcmeComponent )_child);
                AcmeStudioChangeManager.globalChangeManager.onRemoveComponent (sys, (AcmeComponent )_child, _announcer);
            }
            else if (_child instanceof AcmeConnector) {
                AcmeSystem sys = (AcmeSystem )_child.getParent ();
                deleteAttachments (_child, sys);
                _parent = sys;
                sys.removeConnector ((AcmeConnector )_child);
                AcmeStudioChangeManager.globalChangeManager.onRemoveConnector (sys, (AcmeConnector )_child, _announcer);
            }
            else if (_child instanceof AcmePort) {
                AcmeComponent comp = (AcmeComponent )_child.getParent ();
                deleteAttachments (_child, (AcmeSystem )comp.getParent ());
                comp.removePort ((AcmePort )_child);
                AcmeStudioChangeManager.globalChangeManager.onRemovePort (comp, (AcmePort )_child, _announcer);
                _parent = comp;
            }
            else if (_child instanceof AcmeRole) {
                if (_child.getParent () == null) return;
                AcmeConnector conn = (AcmeConnector )_child.getParent ();
                AcmeSystem sys = (AcmeSystem )conn.getParent ();
                deleteAttachments (_child, sys);
                conn.removeRole ((AcmeRole )_child);
                AcmeStudioChangeManager.globalChangeManager.onRemoveRole (conn, (AcmeRole )_child, _announcer);
                _parent = conn;
            }
            else if (_child instanceof AcmeAttachment) {
                AcmeSystem sys = (AcmeSystem )_child.getParent ();
                deleteAttachments (_child, sys);
                _parent = sys;
                sys.removeAttachment ((AcmeAttachment )_child);
                AcmeStudioChangeManager.globalChangeManager.onRemoveAttachment (sys, (AcmeAttachment )_child, _announcer);
            }
            else if (_child instanceof AcmeBinding) {
                AcmeBinding binding = (AcmeBinding)_child;
                AcmeRepresentation parentRep = binding.getParentRep();
                AcmeBindingsMap bm = (AcmeBindingsMap)parentRep.getAbstractionMap();
                if (bm != null) {
                    bm.removeBinding(binding);
                }
                AcmeStudioChangeManager.globalChangeManager.onRemoveBinding(parentRep, binding, _announcer);
            }
            for (Iterator iter = _deletedAttachments.iterator (); iter.hasNext ();) {
				AcmeAttachment att = (AcmeAttachment )iter.next ();
				AcmeStudioChangeManager.globalChangeManager.onRemoveAttachment(Helper.getAcmeSystem(_parent), att, _announcer);
			}
        } catch (Exception e) {
            e.printStackTrace ();
        }
    }

    public void setChild (AcmeElement child) {
        this._child = child;
    }
}
