/*
 * Created on Jul 20, 2004
 *
 */
package AcmeStudio.model;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;
import AcmeStudio.util.Helper;
import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.designRules.Scope;
import acmeLib.elements.*;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmeFamily;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyBearer;
import acmeLib.properties.AcmePropertyType;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;


/**
 * @author Bradley Schmerl
 * Created Jul 20, 2004
 * 
 * Helper class to add functionality to AcmeLib. Methods in here should only manipulate AcmeLib 
 * structures, and will be considered for moving to AcmeLib at a later date.
 */
public class ModelHelper {

	private ModelHelper() {
		super ();
	}

	/**
	 * Gets the types of the object. If the object is a types, then all the supertypes are returned;
	 * if the object is an element, all the declared and instantiated types are returned.
	 * @param object The object to get the types of
	 * @return the types
	 */
	public static AcmeEnumeration getTypes(AcmeObject object) {
	    AcmeEnumeration types = null;
		if (object instanceof AcmeElementType) {
			types = ModelHelper.gatherSuperTypes (((AcmeElementType) object));
		}
		else if (object instanceof AcmeElement) {
			AcmeElement el = (AcmeElement) object;
			types = el.getDeclaredTypes();
			types.concatenate(el.getInstantiatedTypes());
			AcmeEnumeration superTypes = new AcmeEnumeration ();
			while (types.moreItems ()) {
				AcmeElementType type = (AcmeElementType )types.getNextItem ();
				superTypes.concatenate(ModelHelper.gatherSuperTypes(type));
			}
			types.concatenate(superTypes);
			types = types.castToAcmeSetVector().enumerate();
		}
	    return types;
	}

	/**
	 * Returns the copy of the property in the inherited type of propertyParent,
	 * @param property The unified property (that might be a result of unifying type properties)
	 * @param propertyParent The container of the property
	 * @return The property in a super type
	 */
	public static  AcmeProperty getInheritedProperty(AcmeProperty property, AcmePropertyBearer propertyParent) {
		AcmeProperty iProp = null;
		AcmeEnumeration types = getTypes((AcmeObject )propertyParent);
	
		if (types != null) {
			while (iProp == null && types.moreItems()) {
				AcmeElement sT = (AcmeElement) types.getNextItem();
				iProp = sT.getProperty(property.getName());
			}
		}
	    
	    // Check in parents
	    if (iProp == null) {
	        if (propertyParent instanceof AcmeElement) {
	            AcmeElement parent = ((AcmeElement )propertyParent).getParent ();
	            
	            types = getTypes (parent);
	            if (types != null && 
	            		types.getSize() != 0) {
	                AcmeElement sT = (AcmeElement )types.getNextItem ();
	                Object child = sT.lookupName (((AcmeElement )propertyParent).getName (), false);
	                if (child != null && child instanceof AcmeElement) {
	                    AcmeElement c = (AcmeElement )child;
	                    iProp = c.getProperty(property.getName ());
	                }
	            }
	            
	            
	        }
	    }
	    
		return iProp;
	}

	/**
	 * Deterime whether a property is inherited
	 * @param property The property that may be inherited
	 * @param bearer The container of the property
	 * @return whether the property is inherited
	 */
	public static boolean isInheritedProperty(AcmeProperty property, AcmePropertyBearer bearer) {
		return getInheritedProperty(property, bearer) != null;
	}

	/**
	 * Determine whether the value of a property is inherited 
	 * @param property the property that may be inherited
	 * @param bearer the container for the property
	 * @return whether the property's value is inherited
	 */
	public static  boolean isInheritedValue(AcmeProperty property, AcmePropertyBearer bearer) {
		AcmeProperty iProp = getInheritedProperty(property, bearer);
		return iProp != null && iProp.getValue() != null;
	}

	/**
	 * Determines whether some substructure (i.e., port or role) is inherited
	 * @param subElement The subelement that may be inherited
	 * @param parentElement the container of the sub element
	 * @return whether the subelement is inherited
	 */
	public static boolean isInheritedSubstructure(AcmeElement subElement, AcmeElement parentElement) {
		boolean inherited = false;
		AcmeEnumeration types = null;
		if (parentElement == null) return false;
		if (parentElement instanceof AcmeElementType) {
	
			types = ModelHelper.gatherSuperTypes(((AcmeElementType) parentElement));
		}
		else if (parentElement != null) {
			types = parentElement.getDeclaredTypes();
		}
		while (types.moreItems()) {
			AcmeElementType type = (AcmeElementType) types.getNextItem();
			if (((Scope) type).lookupName(subElement.getName(), false) != null) {
				inherited = true;
			}
		}
		return inherited;
	}

	/**
	 * Recursively gathers the supertypes of a type
	 * @param type The type to get the supertypes of
	 * @return the collection of supertyeps of type
	 */
	public static AcmeEnumeration gatherSuperTypes(AcmeElementType et) {
	
		AcmeEnumeration types = new AcmeEnumeration();
		AcmeEnumeration st = et.getSuperTypes().enumerate();
		types.concatenate(st);
		while (st.moreItems()) {
			AcmeElementType type = (AcmeElementType) st.getNextItem();
			types.concatenate(gatherSuperTypes(type));
		}
		return types.castToAcmeSetVector().enumerate();
	
	}

	/**
	 * Returns the subtypes that are delcared in the same scope (i.e., not in sub-families)
	 * @param type The type to get the subtypes of
	 * @return The list of subtypes
	 */
	public static List getSubtypesInSameScope(AcmeElementType targetType) {
		List subtypes = new ArrayList();
		AcmeFamily parentScope = (AcmeFamily) ((AcmeElement) targetType).getScope();
		AcmeEnumeration types = parentScope.getTypes();
		while (types.moreItems()) {
			AcmeType type = (AcmeType) types.getNextItem();
			if (type instanceof AcmeElementType) {
				AcmeElementType eType = (AcmeElementType) type;
				if (eType != targetType && eType.getSuperTypes().contains((AcmeObject) targetType)) {
					subtypes.add(eType);
					subtypes.addAll(getSubtypesInSameScope(eType));
				}
	
			}
	
		}
	
		return subtypes;
	}

/**
 * Returns the most family in which t1 and t2 are in scope
 * @param t1
 * @param t2
 * @return
 */
	public static AcmeElement getMostCommonFamily(AcmeType t1, AcmeType t2) {
		if (t1 == null || t2 == null)
			return null;
		AcmeFamily f1 = null;
		AcmeFamily f2 = null;
		if (t1 instanceof AcmeElementType) {
			f1 = (AcmeFamily) ((AcmeElement) t1).getParent();
		}
		else if (t1 instanceof AcmePropertyType) {
			f1 = (AcmeFamily) ((AcmePropertyType) t1).getDeclarationScope();
		}
		if (t2 instanceof AcmeElementType) {
			f2 = (AcmeFamily) ((AcmeElement) t2).getParent();
		}
		else if (t2 instanceof AcmePropertyType) {
			f2 = (AcmeFamily) ((AcmePropertyType) t2).getDeclarationScope();
		}
	
		if (f1 == f2)
			return f1;
	
		if (f1.lookupName(t2.getTypeName(), true) != null) {
			// If we can find t2 in f1, then f2 is a superfamily of f1
			return f2;
		}
		else {
			return f1;
		}
	
	}

	/**
	 * Given a fully qualified name, the object corresponding to that name in the design is returned
	 * @param des The design to look for the object
	 * @param name The fully qualified name of the object
	 * @return
	 */
	public static Object retrieveElementFromDesign(AcmeDesign des, String name) {
		Object curObj = des;
		StringTokenizer tokens = new StringTokenizer(name, ".");
		while (tokens.hasMoreTokens()) {
			String frag = tokens.nextToken();
			if (curObj instanceof Scope) {
				curObj = ((Scope) curObj).lookupName(frag, false);
			}
			else {
				curObj = null;
			}
			if (curObj == null) {
				return null;
			}
		}
		return curObj;
	}

	/**
	 * FInd the first attachment in which acmeRole is the role
	 * @param acmeRole
	 * @return
	 */
	public static AcmeAttachment findAttachment(AcmeRole acmeRole) {
			AcmeAttachment result = null;
			AcmeSystem acmeSystem = Helper.getRootAcmeSystem(acmeRole);
			AcmeEnumeration atts = acmeSystem.getAttachments (acmeRole);
			if (atts != null && atts.getSize() != 0) result = (AcmeAttachment )atts.getNextItem();
	//		Vector attachments = getAttachments(acmeSystem);
	//		for (int i = 0; i < attachments.size(); i++) {
	//			AcmeAttachment attachment = (AcmeAttachment) attachments.elementAt(i);
	//			if (attachment.getRole() == acmeRole) {
	//				result = attachment;
	//				break;
	//			}
	//		}
			return result;
		}

	/**
	 * Returns true of port1 and port2 are attached by one connector
	 * @param port1
	 * @param port2
	 * @return
	 */
	public static boolean arePortsAttached(AcmePort port1, AcmePort port2) {
		AcmeSystem acmeSystem = Helper.getRootAcmeSystem(port1);
		Vector attachments = Helper.getAttachments(acmeSystem);
		AcmeRole role1 = null, role2 = null;
		for (int i = 0; i < attachments.size(); i++) {
			AcmeAttachment attachment = (AcmeAttachment) attachments.elementAt(i);
			if (attachment.getPort() == port1) {
				role1 = attachment.getRole();
				break;
			}
		}
		if (role1 != null) {
			for (int i = 0; i < attachments.size(); i++) {
				AcmeAttachment attachment = (AcmeAttachment) attachments.elementAt(i);
				if (attachment.getPort() == port2) {
					role2 = attachment.getRole();
					break;
				}
			}
		}
		else {
			return false;
		}
	
		if (role2 != null) {
			return role1.getParent() == role2.getParent();
		}
	
		return false;
	}

	/**
	 * Finds the port that is attached to the role
	 * @param acmeRole
	 * @return
	 */
	public static AcmePort findAttachedPort(AcmeRole acmeRole) {
		AcmePort result = null;
		AcmeSystem acmeSystem = Helper.getRootAcmeSystem(acmeRole);
		Vector attachments = Helper.getAttachments(acmeSystem);
		for (int i = 0; i < attachments.size(); i++) {
			AcmeAttachment attachment = (AcmeAttachment) attachments.elementAt(i);
			if (attachment.getRole() == acmeRole) {
				result = attachment.getPort();
				break;
			}
		}
		return result;
	}

	/**
	 * Returns true if the role is unattached
	 * @param acmeRole
	 * @return
	 */
	public static boolean isAvailableRole(AcmeRole acmeRole) {
		boolean result = true;
		AcmeSystem acmeSystem = Helper.getRootAcmeSystem(acmeRole);
		Vector attachments = Helper.getAttachments(acmeSystem);
		for (int i = 0; i < attachments.size(); i++) {
			AcmeAttachment attachment = (AcmeAttachment) attachments.elementAt(i);
			if (attachment.getRole() == acmeRole) {
				result = false;
				break;
			}
		}
		return result;
	}

	/**
	 * Returns true if the acmePort is not attached to a role
	 * @param acmePort
	 * @return
	 */
	public static boolean isAvailablePort(AcmePort acmePort) {
		boolean result = true;
		AcmeSystem acmeSystem = Helper.getRootAcmeSystem(acmePort);
		Vector attachments = Helper.getAttachments(acmeSystem);
		for (int i = 0; i < attachments.size(); i++) {
			AcmeAttachment attachment = (AcmeAttachment) attachments.elementAt(i);
			if (attachment.getPort() == acmePort) {
				result = false;
				break;
			}
		}
		return result;
	}

	/**
	 * @param system
	 * @return
	 */
	public static boolean isConnectorRep(AcmeSystem system) {
		boolean isConnectorRep = false;
	    if (system.getParentRepresentation() != null)
	    {
	        if (system.getParentRepresentation().getParent() instanceof AcmeConnector)
	        {
	            isConnectorRep = true;
	        }
	    }
		return isConnectorRep;
	}

	/**
	 * Returns a list of ports inside a representation that are not attached or bound
	 * @param rep The representation that to check
	 * @return A list containing ports or roles that are unattached or bound
	 */
	public static List collectUnboundBindingPointsInRepresentation (AcmeRepresentation rep) {
		List unbound = new ArrayList ();
		AcmeSystem system = rep.getSystem();
		AcmeBindingsMap bindings = (AcmeBindingsMap )rep.getAbstractionMap();
		if (bindings == null) {
			bindings = new AcmeBindingsMap ();
			rep.setAbstractionMap(bindings);
		}
		if (isConnectorRep(system)) {
			AcmeEnumeration connectors = system.getConnectors ();
			while (connectors.moreItems ()) {
				AcmeConnector connector = (AcmeConnector )connectors.getNextItem ();
				AcmeEnumeration roles = connector.getRoles ();
				while (roles.moreItems ()) {
					AcmeRole role = (AcmeRole )roles.getNextItem ();
					if (system.getAttachments(role).getSize () == 0 && bindings.getBindings(role).getSize() == 0) 
						unbound.add (role);
				}
			}
		}
		else {
			AcmeEnumeration components = system.getComponents ();
			while (components.moreItems ()) {
				AcmeComponent component = (AcmeComponent )components.getNextItem (); 
				AcmeEnumeration ports = component.getPorts ();
				while (ports.moreItems ()) {
					AcmePort port = (AcmePort )ports.getNextItem ();
					if (system.getAttachments () != null && system.getAttachments(port).getSize () == 0 && bindings.getBindings (port).getSize () == 0) 
						unbound.add (port);
				}
			}
		}
		return unbound;
	}

	/**
	 * @param type
	 * @return returns the family that a particular type is declared in
	 */
	public static AcmeFamily getDeclaredIn(AcmeElementType type) {
		AcmeElement typeAsElement = (AcmeElement )type;
		AcmeElement parent = typeAsElement.getParent ();
		if (parent instanceof AcmeFamily) return (AcmeFamily )parent;
		return null;
	}
}
