package AcmeStudio.model;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.util.AcmeStudioChangeManager;
import acmeLib.AcmeObject;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyBearer;

/**
 * @author schmerl
 *
 * Command, identified by getKind(), to add or delete a property from an
 * AcmeObject.
 */
public class PropertyCommand extends ModelCommand {

    private static final String ADD_PROP_REQUEST = "$AddProp$";
    private static final String DEL_PROP_REQUEST = "$DelProp$";
    
    public static final String ADD_PROP = ADD_PROP_REQUEST;
    public static final String REMOVE_PROP = DEL_PROP_REQUEST;

    private AcmeProperty _property;
    private AcmeObject _target;
    private Object _announcer;
    
    /**
     * Constructor for AddPropertyCommand.
     */
    public PropertyCommand(String kind, Object announcer) {
        super (kind);
        setKind (kind);
        this._announcer = announcer;
    }

    /**
     * Constructor for AddPropertyCommand.
     * @param label
     */
    public PropertyCommand(String kind, AcmeObject target, AcmeProperty prop, Object announcer) {
        super (kind);
        setKind (kind);
        this._announcer = announcer;
        this._target = target;
        this._property = prop;
    }

    /**
     * Sets the newProperty.
     * @param newProperty The newProperty to set
     */
    public void setNewProperty(AcmeProperty newProperty) {
        this._property = newProperty;
    }

    /**
     * Sets the target.
     * @param target The target to set
     */
    public void setTarget(AcmeObject target) {
        this._target = target;
    }

    /**
     * @see org.eclipse.gef.commands.Command#canExecute()
     */
    public boolean canExecute() {
        return super.canExecute ();
    }

    /**
     * @see org.eclipse.gef.commands.Command#canUndo()
     */
    public boolean canUndo() {
        return super.canUndo ();
    }

    /**
     * @see org.eclipse.gef.commands.Command#execute()
     */
    public void execute() {
        if (getKind ().equals (ADD_PROP)) {
            addProperty ();
            setLabel (AcmeStudioPlugin.getMessage ("PropertyCommand.AddPropertyLabel") + _property.getName ()); //$NON-NLS-1$
        }
        else if (getKind ().equals (REMOVE_PROP)) {
            removeProperty ();
            setLabel (AcmeStudioPlugin.getMessage ("PropertyCommand.RemovePropertyLabel") + _property.getName ()); //$NON-NLS-1$
        }
    }
    
    protected void addProperty () {
        try {
            if (_target instanceof AcmePropertyBearer) {
                AcmePropertyBearer element = (AcmePropertyBearer )_target;
                element.addProperty (_property);
                AcmeStudioChangeManager.globalChangeManager.onAddProperty (element, _property, _announcer);
            }
        }
        catch (AcmeIllegalNameException e) {
            _errorReason = e;
        }
    }

    /**
     * @see org.eclipse.gef.commands.Command#redo()
     */
    public void redo() {
        execute ();
    }

    /**
     * @see org.eclipse.gef.commands.Command#undo()
     */
    public void undo() {
        if (_errorReason != null) return;
        
        if (getKind ().equals (ADD_PROP)) {
            removeProperty ();
        }
        else if (getKind ().equals (REMOVE_PROP)) {
            addProperty ();
        }        
    }
    
    protected void removeProperty () {
        if (_target instanceof AcmePropertyBearer) {
            AcmePropertyBearer element = (AcmePropertyBearer )_target;
            element.removeProperty (_property);
            AcmeStudioChangeManager.globalChangeManager.onRemoveProperty (element, _property, _announcer);          
        }        
    }

}
