package AcmeStudio.model;

import java.util.Iterator;
import java.util.List;

import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Helper;
import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.elements.*;

/**
 * @author schmerl
 *
 * Command, identified by getKind(), to assign or unassign a type or a family.
 */
public class TypeAssignmentCommand extends ModelCommand {

    private static final String ASSIGN_TYPE_REQUEST = "$AssignType$";
    private static final String UNASSIGN_TYPE_REQUEST = "$UnAssignType$";
    private static final String ASSIGN_FAMILY_REQUEST  = "$AssignFamily$";
    private static final String UNASSIGN_FAMILY_REQUEST = "$UnAssignFamily$";
    
    public static final String ASSIGN_TYPE = ASSIGN_TYPE_REQUEST;
    public static final String UNASSIGN_TYPE = UNASSIGN_TYPE_REQUEST;
    public static final String ASSIGN_FAMILY = ASSIGN_FAMILY_REQUEST;
    public static final String UNASSIGN_FAMILY = UNASSIGN_FAMILY_REQUEST;

    private AcmeElement _element;
    private AcmeElementType _type;
    private String _typeName; // the name of the type (to use if _type is null)
    private boolean _instantiate;
    private Object _announcer;
    
    public TypeAssignmentCommand (String commandType, AcmeElement element, AcmeElementType type, boolean instantiate, Object announcer) {
        super (commandType);
        setKind (commandType);
        _element = element;
        _type = type;
        _announcer = announcer;
        _instantiate = instantiate;
    }
    
    public TypeAssignmentCommand (String commandType, AcmeSystem system, AcmeFamily family, Object announcer) {
        super (commandType);
        setKind (commandType);
        _element = system;
        _type = family;
        _announcer = announcer;         
    }
    
    public TypeAssignmentCommand (String commandType, AcmeSystem system, String familyName, Object announcer) {
    	super (commandType);
        setKind (commandType);
    	_element = system;
    	_announcer = announcer;
    	_type = null;
    	_typeName = familyName;
    }
    
    /**
     * @see org.eclipse.gef.commands.Command#canExecute()
     */
    public boolean canExecute() {
        return (_element != null) && (_type != null || _typeName != null);
    }

    /**
     * @see org.eclipse.gef.commands.Command#canUndo()
     */
    public boolean canUndo() {
        return !getKind ().equals (UNASSIGN_FAMILY);
    }

    /**
     * @see org.eclipse.gef.commands.Command#execute()
     */
    public void execute() {
        if (getKind ().equals (ASSIGN_TYPE) && !(_element instanceof AcmeSystem)) {
            setLabel ("Assign type " + (_type==null?_typeName:_type.getTypeName ()));
            assignType ();
        }
        else if (getKind ().equals (UNASSIGN_TYPE) && !(_element instanceof AcmeSystem)) {
            setLabel ("Unassign type " +(_type==null?_typeName:_type.getTypeName ()));
            unassignType ();
        }
        else if (getKind ().equals (ASSIGN_FAMILY)) {
            setLabel ("Assign family " + (_type==null?_typeName:_type.getTypeName ()));
            assignFamily ();
        }
        else if (getKind ().equals (UNASSIGN_FAMILY)) {
            setLabel ("Unassign type " + (_type==null?_typeName:_type.getTypeName ()));
            unassignFamily ();
        }
        
    }
    
    protected void assignFamily () {
        AcmeSystem system = (AcmeSystem )_element;
        AcmeFamily family = null;
        if (_type == null) {
        	family = (AcmeFamily )system.lookupName(_typeName, true);
        }
        else {
          family = (AcmeFamily )_type;
        }
        
        if (family != null) {
	        system.addDeclaredType (family);
	        system.addInstantiatedType(family);
	        system.instantiateElements();
	        AcmeStudioChangeManager.globalChangeManager.onAssignFamily ((AcmeSystem )_element, (AcmeFamily )_type, _announcer);
        } 
    }
    
    protected void unassignFamily () {
        // This needs to remove all references to the types in the family
        // This has a bug - should remove types from super families too,
        // unless the superfamily is specifically mentioned in the declared
        // list
        AcmeFamily family = (AcmeFamily )_type;
        AcmeEnumeration types = family.getTypes ();
        
        ModelCompoundCommand command = new ModelCompoundCommand ("UnassignFamily");
        while (types.moreItems ()) {
            AcmeType type = (AcmeType )types.getNextItem ();
            List affectedObjects = Helper.retrieveAllReferencesToType (type, (AcmeSystem )_element);
            Iterator i = affectedObjects.iterator ();
            while (i.hasNext()) {
                AcmeObject o = (AcmeObject )i.next ();
                if (o instanceof AcmeElement && type instanceof AcmeElementType) {
                    AcmeElement elem = (AcmeElement )o;
                    TypeAssignmentCommand c = new TypeAssignmentCommand (TypeAssignmentCommand.UNASSIGN_TYPE, elem, (AcmeElementType )type, elem.instantiatesType((AcmeElementType )type), _announcer);
                    command.add(c);
                }
            }
        }
        command.execute();
        _element.removeInstantiatedType(_type);
        _element.removeDeclaredType (_type);
        
        // BUG: If a representation refers to a family, then we don't want to remove the import.
        //           This causes unresolved reference errors when the design is reloaded.
        // If the design is imported, remove the imported design
        if (family.getParentDesign() != _element.getParentDesign()) {
            AcmeDesign desToSeek = family.getParentDesign ();
            AcmeEnumeration desRecs = _element.getParentDesign ().getImportedDesigns().enumerate ();
            boolean found = false;
            AcmeImportedDesignRecord rec = null;
            while (desRecs.moreItems() && !found) {
                rec = (AcmeImportedDesignRecord )desRecs.getNextItem ();
                if (rec.getResolvedDesign() == desToSeek) {
                    found = true;
                }   
            }
            if (rec != null) {
                _element.getParentDesign ().removeImportedDesign(rec.getCanonicalFilename());
            }
        }
        
        AcmeStudioChangeManager.globalChangeManager.onUnAssignFamily ((AcmeSystem )_element, (AcmeFamily )_type, _announcer);
     }
    
    public void assignType () {     
        if (_instantiate) {
            _element.addInstantiatedType(_type);
            _element.instantiateElements();
        }
        
        _element.addDeclaredType (_type);             

         AcmeStudioChangeManager.globalChangeManager.onAssignType (_element, _type, _announcer);
    }

    /**
     * @see org.eclipse.gef.commands.Command#redo()
     */
    public void redo() {
        execute ();
    }

    /**
     * This doesn't really undo instantiations
     * @see org.eclipse.gef.commands.Command#undo()
     */
    public void undo() {
        if (getKind ().equals (ASSIGN_TYPE)) {
            unassignType ();
        }
        else if (getKind ().equals (UNASSIGN_TYPE)) {
            assignType ();
        }        
        else if (getKind ().equals (ASSIGN_FAMILY)) {
            unassignFamily ();
        }
        else if (getKind ().equals (UNASSIGN_FAMILY)) {
            assignFamily ();
        }
    }
    
    protected void unassignType () {
        if (_instantiate) {
            _element.removeInstantiatedType(_type); 
        }
        
        _element.removeDeclaredType (_type);

        AcmeStudioChangeManager.globalChangeManager.onUnAssignType (_element, _type, _announcer);
    }

}
