package AcmeStudio.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;

import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Helper;
import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmeFamily;

/**
 * @author schmerl
 *
 * Command, identified by getKind(), to add or remove type from an AcmeFamily
 * or AcmeDesign.
 */
public class TypeCommand extends ModelCommand {

    private static final String ADD_TYPE_REQUEST = "$AddType$";
    private static final String REMOVE_TYPE_REQUEST = "$RemoveType$";
    
    public static final String ADD_TYPE = ADD_TYPE_REQUEST;
    public static final String REMOVE_TYPE = REMOVE_TYPE_REQUEST;


    private AcmeType _type;
    private AcmeElement _namespace; 
    private Object _announcer;
    
    public TypeCommand (String command, AcmeType type, AcmeFamily family, Object announcer) {
        super (command);
        setKind (command);
        _type = type;
        _namespace = family;
        _announcer = announcer;
    }
    
    public TypeCommand (String command, AcmeType type, AcmeDesign design, Object announcer) {
        super (command);
        setKind (command);
        _type= type;
        _namespace = design;
        _announcer = announcer;
    }
            
    public boolean canExecute () {
        return (_type != null) && (_namespace != null);
    }
    
    public boolean canUndo () {
        return getKind ().equals (ADD_TYPE);
    }
    
    protected void addType () {
        if (_namespace instanceof AcmeFamily) {
            AcmeFamily family = (AcmeFamily )_namespace;
           try {
                 family.addType (_type);
                AcmeStudioChangeManager.globalChangeManager.onAddType (family, _type, _announcer);
            }
            catch (AcmeIllegalNameException e) {
                _errorReason = e;
            }
        }
        else if (_namespace instanceof AcmeDesign) {
            AcmeDesign design = (AcmeDesign )_namespace;
            try {
                design.addType (_type);
                AcmeStudioChangeManager.globalChangeManager.onAddType (design, _type, _announcer);
            }
            catch (AcmeIllegalNameException e) {
                _errorReason = e;
            }
        }
    }        
    
    public void removeType () {
        List affectedObjects = new ArrayList ();
        ModelCompoundCommand command = new ModelCompoundCommand ("RemoveType");
        if (_namespace instanceof AcmeFamily) {
            
            AcmeFamily family = (AcmeFamily )_namespace;
            affectedObjects = Helper.retrieveAllReferencesToType(_type, family);
            family.removeType (_type);
            AcmeStudioChangeManager.globalChangeManager.onRemoveType (family, _type, _announcer);

        }
        else if (_namespace instanceof AcmeDesign) {
            AcmeDesign design = (AcmeDesign )_namespace;
            affectedObjects = Helper.retrieveAllReferencesToType(_type, design);
            design.removeType (_type);
            AcmeStudioChangeManager.globalChangeManager.onRemoveType (design, _type, _announcer);
        }
        Iterator i = affectedObjects.iterator ();
        while (i.hasNext ()) {
            AcmeObject o = (AcmeObject )i.next ();
            if (o instanceof AcmeElement && _type instanceof AcmeElementType) {
                AcmeElement elem = (AcmeElement )o;
                TypeAssignmentCommand c = 
                    new TypeAssignmentCommand (TypeAssignmentCommand.UNASSIGN_TYPE, elem, (AcmeElementType )_type, true,_announcer);
                command.add (c);
            }  
            
        }     
        command.execute ();      
    }
    
    public void execute () {
        if (getKind ().equals (ADD_TYPE)) {
            setLabel ("Add type " + _type.getTypeName ());
            addType ();                  
        }
        else if (getKind ().equals (REMOVE_TYPE)) {
            if (MessageDialog.openConfirm (Display.getDefault ().getActiveShell(), "Remove type confirmation", "Removing " + _type.getTypeName () + " cannot be undone. Still remove?")) {
                setLabel ("Remove type " + _type.getTypeName ());
                removeType ();
            }
        }
    }
    
    public void redo () {
        execute ();
    }
    
    public void undo () {
        if (getKind ().equals (ADD_TYPE)) {
            removeTypeWithoutChecks ();                  
        }
        else if (getKind ().equals (REMOVE_TYPE)) {
            addType ();
        }        
    }
    
    protected void removeTypeWithoutChecks () {
        if (_namespace instanceof AcmeFamily) {
            AcmeFamily family = (AcmeFamily )_namespace;
            family.removeType (_type);
            AcmeStudioChangeManager.globalChangeManager.onRemoveType (family, _type, _announcer);
        }
        else if (_namespace instanceof AcmeDesign) {
            AcmeDesign design = (AcmeDesign )_namespace;
            design.removeType (_type);
            AcmeStudioChangeManager.globalChangeManager.onRemoveType (design, _type, _announcer);
        }        
    }
}
