package AcmeStudio.prefs;

import org.eclipse.draw2d.*;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.preference.PreferencePage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Label;
import org.eclipse.ui.IWorkbench;
import com.swtdesigner.SWTResourceManager;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.resource.ASVisualStyle;
import AcmeStudio.ui.editors.diagram.styles.*;
import AcmeStudio.ui.editors.diagram.styles.constants.HorizontalTextAlignmentOptions;
import AcmeStudio.ui.editors.diagram.styles.constants.LabelOptions;
import AcmeStudio.ui.editors.diagram.styles.constants.Shape;
import AcmeStudio.ui.editors.diagram.styles.constants.VerticalTextAlignmentOptions;
import AcmeStudio.util.ImageManager;
import AcmeStudio.util.Listeners;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.properties.AcmeProperty;

/**
 * @author yh
 *
 */
public class VisPreferences
    extends PreferencePage {

    /**
     * Creates a button to choose a line style, and adds the associated selection listener
     * @param group Where the button is to go.
     * @param style The style of the button
     * @param listener The selection listener to add to the button
     * @return The new button
     */
    private static Button createLineStyleButton(Composite group, int style) {
        Button b = new Button(group, SWT.RADIO);
        String key = "__linestyle_" + style;
        Image img = ImageManager.createImage(key, 50, 12, false);
        drawLineImage(img, style);
        b.setImage(img);
        b.setData(new Integer(style));
        return b;
    }

    /**
     * Draws a line image in a particular style, for use as a depiction
     * in a button.
     * @param img The image to draw a line into
     * @param style The style of the line to draw
     */
    private static void drawLineImage(Image img, int style) {
        GC gc = new GC(img);
        gc.setBackground(ColorConstants.button);
        gc.fillRectangle(img.getBounds());
        gc.setLineStyle(style);
        gc.drawLine(1, 6, 49, 6);
        gc.dispose();
    }

    private static String fromFontToString(Font font) {
        if (font == null) {
            return null;
        }
        String result = "";
        FontData fd = font.getFontData()[0];
        result += fd.getName() + " " + fd.getHeight() + " " + fd.getStyle();
        return result;
    }

    private static Font fromStringToFont(String data) {
        if (data == null) {
            return null;
        }
        String[] info = data.split(" ");
        FontData fd = new FontData();
        if (info.length > 0)
            fd.setName(info[0]);
        if (info.length > 1)
            fd.setHeight(Integer.parseInt(info[1]));
        if (info.length > 2)
            fd.setStyle(Integer.parseInt(info[2]));
        return new Font(Display.getDefault(), fd);
    }

    public static ElementVisualization loadVisFromStore(int style) {
        IPreferenceStore store =
            AcmeStudioPlugin.getDefault().getPreferenceStore();
        final String type;
        ElementVisualization vis = null;
        switch (style) {
            case ElementVisualization.DEFAULT_COMPONENT :
                type = "ComponentVisualization.";
                vis = new ComponentVisualization();
                break;
            case ElementVisualization.DEFAULT_CONNECTOR : {
                type = "ConnectorVisualization.";
                ConnectorVisualization cv = new ConnectorVisualization();
                cv.setAutoCenterBinaryConnectors(store.getBoolean (type+"AutoCenterBinary"));
				vis = cv;
                break;
            }
            case ElementVisualization.DEFAULT_PORT :
                type = "PortVisualization.";
                vis = new PortVisualization();
                break;
            case ElementVisualization.DEFAULT_ROLE :
                type = "RoleVisualization.";
                vis = new RoleVisualization();
                break;
            default :
                return vis;
        }

        vis.setShape(store.getInt(type + "Shape"));
        vis.setDirection(store.getInt(type + "Direction"));
        vis.setReplicateShape(store.getBoolean(type + "ReplicateShape"));
        vis.setSize(
            new Dimension(
                store.getInt(type + "Width"),
                store.getInt(type + "Height")));
        vis.setOutlineWidth(store.getInt(type + "OutlineWidth"));
        vis.setOutlineStyle(store.getInt(type + "OutlineStyle"));
        vis.setOutlineColor(
        		SWTResourceManager.getColor (
        				store.getInt(type + "OutlineColorR"),
						store.getInt(type + "OutlineColorG"),
						store.getInt(type + "OutlineColorB")));
        vis.setFillColor(
        		SWTResourceManager.getColor (
        				store.getInt(type + "FillColorR"),
						store.getInt(type + "FillColorG"),
						store.getInt(type + "FillColorB")));
        vis.setTextColor(
        	SWTResourceManager.getColor (
                store.getInt(type + "TextColorR"),
                store.getInt(type + "TextColorG"),
                store.getInt(type + "TextColorB")));
        vis.setFont(fromStringToFont(store.getString(type + "Font")));
        vis.setLabelVerticalAlignment(
            store.getInt(type + "LabelVerticalAlignment"));
        vis.setLabelHorizontalAlignment(
            store.getInt(type + "LabelHorizontalAlignment"));
        vis.setLabelMargin(store.getInt(type + "LabelMargin"));
        vis.setWordwrap(store.getBoolean(type + "WordWrap"));
        vis.setLabelSource(store.getInt(type + "LabelSource"));
        vis.setLabelProperty(store.getString(type + "LabelProperty"));
        vis.setLabelValue(store.getString(type + "LabelValue"));
        vis.setPreserveAspectRatio(
            store.getBoolean(type + "PreserveAspectRatio"));
        vis.setHeightResizeable(store.getBoolean(type + "HeightResizeable"));
        vis.setWidthResizeable(store.getBoolean(type + "WidthResizeable"));
        return vis;
    }

    public static void setDefaults(IPreferenceStore store) {
        setDefaults(store, ElementVisualization.DEFAULT_COMPONENT);
        setDefaults(store, ElementVisualization.DEFAULT_CONNECTOR);
        setDefaults(store, ElementVisualization.DEFAULT_PORT);
        setDefaults(store, ElementVisualization.DEFAULT_ROLE);
    }

    private static void setDefaults(IPreferenceStore store, int style) {
        ElementVisualization vis;
        String type;
		switch (style) {
            case ElementVisualization.DEFAULT_COMPONENT :
                type = "ComponentVisualization.";
                vis = new ComponentVisualization();
                break;
            case ElementVisualization.DEFAULT_CONNECTOR : {
                type = "ConnectorVisualization.";
                ConnectorVisualization cv = new ConnectorVisualization();
                store.setDefault (type+"AutoCenterBinary", cv.autoCenterBinaryConnectors());
				vis = cv;
            
                break;
            }
            case ElementVisualization.DEFAULT_PORT :
                type = "PortVisualization.";
                vis = new PortVisualization();
                break;
            case ElementVisualization.DEFAULT_ROLE : {
                type = "RoleVisualization.";
		        RoleVisualization roleVisualization = new RoleVisualization();
                vis = roleVisualization;
                break;
            }
            default :
                return;
        }

        store.setDefault(type + "Shape", vis.getShape());
        store.setDefault(type + "Direction", vis.getDirection());
        store.setDefault(type + "Width", vis.getSize().width);
        store.setDefault(type + "Height", vis.getSize().height);
        store.setDefault(type + "FillColorR", vis.getFillColor().getRed());
        store.setDefault(type + "FillColorG", vis.getFillColor().getGreen());
        store.setDefault(type + "FillColorB", vis.getFillColor().getBlue());
        store.setDefault(
            type + "OutlineColorR",
            vis.getOutlineColor().getRed());
        store.setDefault(
            type + "OutlineColorG",
            vis.getOutlineColor().getGreen());
        store.setDefault(
            type + "OutlineColorB",
            vis.getOutlineColor().getBlue());
        store.setDefault(type + "TextColorR", vis.getTextColor().getRed());
        store.setDefault(type + "TextColorG", vis.getTextColor().getGreen());
        store.setDefault(type + "TextColorB", vis.getTextColor().getBlue());
        store.setDefault(type + "OutlineStyle", vis.getOutlineStyle());
        store.setDefault(type + "OutlineWidth", vis.getOutlineWidth());
        store.setDefault(type + "ReplicateShape", vis.getReplicateShape());
        store.setDefault(type + "Font", fromFontToString(vis.getFont()));
        store.setDefault(
            type + "LabelVertialAlignment",
            vis.getLabelVerticalAlignment());
        store.setDefault(
            type + "LabelHorizontalAlignment",
            vis.getLabelHorizontalAlignment());
        store.setDefault(type + "LabelMargin", vis.getLabelMargin());
        String value = vis.getLabelValue();
        if (value != null)
            store.setDefault(type + "LabelValue", value);
        value = vis.getLabelProperty();
        if (value != null)
            store.setDefault(type + "LabelProperty", value);
        store.setDefault(type + "LabelSource", vis.getLabelSource());
        store.setDefault(type + "WordWrap", vis.getWordwrap());
        store.setDefault(type + "WidthResizeable", vis.canResizeWidth());
        store.setDefault(type + "HeightResizeable", vis.canResizeHeight());
        store.setDefault(
            type + "PreserveAspectRatio",
            vis.mustPreserveAspectRatio());
    }

    Text _defaultWidthText, _defaultHeightText;
    ColorEditor _fillColorButton, _outlineColorButton;
    Button _noLineButton;

    Button _fontButton, _fontColorButton;

    Combo _horizAlignCombo, _vertAlignCombo;
    Button[] _labelButtons;

    Text _labelText;
    Button[] _lineStyleButtons = new Button[5];
    Text _marginText;
    Combo _outlineWidthCombo;

    Button _preserveAspectRatioButton,
        _widthResizeableButton,
        _heightResizeableButton;
    Combo _propertyCombo;
    Button _replicateShapeButton;
    Combo _shapeCombo, _directionCombo;

    protected String _storeType = null;
    protected int _style;
    protected AcmeElement _type;

    /**
     * The style being edited. Note, this is a copy of an original style,
     * which should faciliate undo
     */
    protected ElementVisualization _vis;
    Button _wordwrapButton;
    ImageFigure imgFig;

    Composite previewGroup;

    FigureCanvas previewImageCanvas;
    ScrollPane previewPane;
	private Button	_autoCenterButton;

    public VisPreferences() {
        setDescription("Setup default visualizations.");
        if (!(this instanceof ComponentVisPreferences
            || this instanceof ConnectorVisPreferences
            || this instanceof PortVisPreferences
            || this instanceof RoleVisPreferences))
            noDefaultAndApplyButton();
    }
    
    private void createAlignmentOptions(Composite parent) {
        Group alignmentGroup = new Group(parent, SWT.NONE);
        alignmentGroup.setText("Alignment");
        alignmentGroup.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        GridLayout agl = new GridLayout(2, false);
        alignmentGroup.setLayout(agl);
        Label horizAlignLabel = new Label(alignmentGroup, SWT.NONE);
        horizAlignLabel.setText("Horiz. Alignment:");
        _horizAlignCombo = new Combo(alignmentGroup, SWT.DROP_DOWN);
        _horizAlignCombo.setItems(
            HorizontalTextAlignmentOptions.LABELS);
        _horizAlignCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _vis.setLabelHorizontalAlignment(
                    _horizAlignCombo.getSelectionIndex());
                updatePreview ();
            }
        });
        Label vertAlignLabel = new Label(alignmentGroup, SWT.NONE);
        vertAlignLabel.setText("Vert. Alignment:");
        _vertAlignCombo = new Combo(alignmentGroup, SWT.DROP_DOWN);
        _vertAlignCombo.setItems(
            VerticalTextAlignmentOptions.LABELS);
        _vertAlignCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _vis.setLabelVerticalAlignment(
                    _vertAlignCombo.getSelectionIndex());
                updatePreview ();
            }
        });
        Label marginLabel = new Label(alignmentGroup, SWT.NONE);
        marginLabel.setText("Margin:");
        _marginText = new Text(alignmentGroup, SWT.SINGLE | SWT.BORDER);
        _marginText.addVerifyListener(Listeners.VERIFY_NUMBER);
        _marginText.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent evt) {
            }
            // update the preview when the text is left
            public void focusLost(FocusEvent evt) {
                if (_marginText.getText().equals("")) {
                    _marginText.setText("0");
                }
                int value = -1;
                try {
                    value = Integer.parseInt(_marginText.getText());
                } catch (Exception e) {
                }
                if (value >= 0) {
                    _vis.setLabelMargin(value);
                }
                updatePreview ();
            }
        });
        _wordwrapButton = new Button(alignmentGroup, SWT.CHECK);
        _wordwrapButton.setText("Use Word Wrap");
        _wordwrapButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _vis.setWordwrap(_wordwrapButton.getSelection());
                updatePreview ();
            }
        });
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.preference.PreferencePage#createContents(org.eclipse.swt.widgets.Composite)
     */
    protected Control createContents(Composite parent) {
        return null;
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.preference.PreferencePage#createContents(org.eclipse.swt.widgets.Composite)
     */
    protected Control createContentsArea(Composite parent) {
    	parent.setLayout (new GridLayout (1, false));
        initPreview(parent);

        createFolderArea(parent);

        return parent;
    }
    private void createDetailsGroup(Composite parent) {
        Group detailsGroup = new Group(parent, SWT.NONE);
        GridData gd = new GridData(GridData.FILL_HORIZONTAL);
        detailsGroup.setLayoutData(gd);
        detailsGroup.setText("Properties");
        GridLayout rcl = new GridLayout(2, false);
        detailsGroup.setLayout(rcl);
        Label fillColorLabel = new Label(detailsGroup, SWT.NONE);
        fillColorLabel.setText("Fill Color:");
        _fillColorButton = new ColorEditor(detailsGroup);
        final Button fillColorButton = _fillColorButton.getButton();
        fillColorButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                RGB rgb = _fillColorButton.getColorValue();
                _vis.setFillColor(
                		SWTResourceManager.getColor(rgb));
                updatePreview ();
            }
        });
        Label outlineColorLabel = new Label(detailsGroup, SWT.NONE);
        outlineColorLabel.setText("Outline Color:");
        _outlineColorButton = new ColorEditor(detailsGroup);
        final Button outlineColorButton = _outlineColorButton.getButton();
        outlineColorButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                RGB rgb = _outlineColorButton.getColorValue();
                if (rgb == null)
                    return;
                _vis.setOutlineColor(
                    new Color(outlineColorButton.getDisplay(), rgb));
                updatePreview ();
            }
        });
        Label outlineStyleLabel = new Label(detailsGroup, SWT.NONE);
        outlineStyleLabel.setText("Outline style:");
        Group outlineGroup = new Group(detailsGroup, SWT.NONE);
        GridLayout ogl = new GridLayout(2, true);
        outlineGroup.setLayout(ogl);
        _noLineButton = new Button(outlineGroup, SWT.RADIO);
        _noLineButton.setText("No Line");
        GridData nlgd = new GridData();
        nlgd.horizontalSpan = 2;
        _noLineButton.setLayoutData(nlgd);
        _noLineButton.setData(new Integer(SWT.NONE));

        _lineStyleButtons[0] =
            createLineStyleButton(outlineGroup, SWT.LINE_SOLID);
        _lineStyleButtons[1] =
            createLineStyleButton(outlineGroup, SWT.LINE_DOT);
        _lineStyleButtons[2] =
            createLineStyleButton(outlineGroup, SWT.LINE_DASH);
        _lineStyleButtons[3] =
            createLineStyleButton(outlineGroup, SWT.LINE_DASHDOT);
        _lineStyleButtons[4] =
            createLineStyleButton(outlineGroup, SWT.LINE_DASHDOTDOT);
        Label outlineWidthLabel = new Label(detailsGroup, SWT.NONE);
        outlineWidthLabel.setText("Outline Width:");
        _outlineWidthCombo = new Combo(detailsGroup, SWT.DROP_DOWN);
        _outlineWidthCombo.setItems(new String[] { "1", "2", "3", "4", "5" });
        GridData owgd = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
        _outlineWidthCombo.setLayoutData(owgd);
        SelectionAdapter buttonSelected = new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                if (((Button) evt.widget).getSelection()
                    && evt.widget.getData() instanceof Integer) {
                    Integer style = (Integer) evt.widget.getData();
                    _vis.setOutlineStyle(style.intValue());
                    if (style.intValue() == SWT.NONE) {
                        _outlineWidthCombo.setEnabled(false);
                    } else {
                        _outlineWidthCombo.setEnabled(true);
                    }
                    updatePreview ();
                }
            }
        };

        _outlineWidthCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _vis.setOutlineWidth(
                    _outlineWidthCombo.getSelectionIndex() + 1);
                updatePreview ();
            }
        });

        _noLineButton.addSelectionListener(buttonSelected);
        for (int i = 0; i < _lineStyleButtons.length; i++) {
            _lineStyleButtons[i].addSelectionListener(buttonSelected);
            updatePreview ();
        }
    }

    protected void createFolderArea(Composite parent) {
        // The number of tabs in the folder 
        int tabCount = 0;
        // Create the shape page and add it to the tab
        TabFolder folder = new TabFolder(parent, SWT.NONE);
        folder.setLayoutData(new GridData(GridData.FILL_BOTH));
        TabItem shapeTab = new TabItem(folder, SWT.NONE, tabCount++);
        shapeTab.setText("Shape");
        shapeTab.setControl(createShapePage(folder));

        // Create the label page and add it to the tab
        TabItem labelTab = new TabItem(folder, SWT.NONE, tabCount++);
        labelTab.setText("Label");
        labelTab.setControl(createLabelPage(folder));
        if (_vis instanceof ConnectorVisualization) {
        	TabItem roleTab = new TabItem (folder, SWT.NONE, tabCount++);
        	roleTab.setText ("Role Policy");
        	roleTab.setControl (createRolePolicyPage (folder));
        }
        
        updateFields();
        folder.pack();

        // Make sure the preview has an image to start with
        updatePreview();
    }
    /**
	 * @param folder
	 * @return
	 */
	private Control createRolePolicyPage(Composite parent) {
		 Composite thePage = new Composite(parent, SWT.NONE);
	        GridLayout pageLayout = new GridLayout(1, true);
	        thePage.setLayout(pageLayout);
		_autoCenterButton = new Button (thePage, SWT.CHECK);
		_autoCenterButton.setText ("Automatically center connector, if it has two roles");
		_autoCenterButton.addSelectionListener(new SelectionAdapter () {
			
			/**
			 * @param e
			 * @see org.eclipse.swt.events.SelectionAdapter#widgetSelected(org.eclipse.swt.events.SelectionEvent)
			 **/
			public void widgetSelected(SelectionEvent e) {
				((ConnectorVisualization )_vis).setAutoCenterBinaryConnectors(_autoCenterButton.getSelection ());
			}
		});
		return thePage;
	}

	/**
	 * 
	 */
	private void updatePreview() {
		imgFig.setImage (_vis.getPreviewImage((AcmeElementType )_type).getImage());
	}

	private void createFontOptions(Composite parent) {
        // The font options group
        Group fontGroup = new Group(parent, SWT.NONE);
        GridLayout fl = new GridLayout(2, false);
        fontGroup.setLayout(fl);
        fontGroup.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        fontGroup.setText("Font settings");
        final FontDialog fontDialog = new FontDialog(fontGroup.getShell());
        _fontButton = new Button(fontGroup, SWT.PUSH);
        _fontButton.setText("Choose font...");
        _fontButton.setSize(100, 30);
        GridData fbd = new GridData();
        fbd.horizontalSpan = 2;
        _fontButton.setLayoutData(fbd);
        _fontButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                FontData fd = fontDialog.open();
                if (fd == null)
                    return;
                Font font = new Font(_fontButton.getDisplay(), fd);
                _vis.setFont(font);
                _fontButton.setFont(font);
                updatePreview ();
            }
        });
        Label fontColorLabel = new Label(fontGroup, SWT.NONE);
        fontColorLabel.setText("Font Color:");
        final ColorDialog fontColorDialog =
            new ColorDialog(fontColorLabel.getShell());
        _fontColorButton = new Button(fontGroup, SWT.PUSH);
        _fontColorButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                RGB rgb = fontColorDialog.open();
                if (rgb == null)
                    return;
                _vis.setTextColor(
                		SWTResourceManager.getColor  (rgb));
                Image image = _fontColorButton.getImage();
                drawImage(image, _vis.getTextColor());
                _fontColorButton.setImage(image);
                updatePreview ();
            }
        });
    }

    /** 
     * Creates the label page in the dialog box, which allows
     * the user to change the label options
     * @param parent Where the page is to be placed
     */
    private Composite createLabelPage(Composite parent) {
        Composite thePage = new Composite(parent, SWT.NONE);
        GridLayout pageLayout = new GridLayout(2, true);
        thePage.setLayout(pageLayout);
        // The left column
        Composite leftColumn = new Composite(thePage, SWT.NONE);
        leftColumn.setLayout(new GridLayout(1, false));
        leftColumn.setLayoutData(new GridData(GridData.FILL_BOTH));
        createFontOptions(leftColumn);
        createAlignmentOptions(leftColumn);

        Composite rightColumn = new Composite(thePage, SWT.NONE);
        rightColumn.setLayout(new GridLayout(1, false));
        rightColumn.setLayoutData(new GridData(GridData.FILL_BOTH));
        createLabelSourceOptions(rightColumn);

        return thePage;
    }

    /**
     * Creates buttons for the label source, and sets the data to be
     * the alignment constant corresponding to the source; also attaches
     * a selection listener
     * @param parent Where to put the button
     * @param title The title to use for the button
     * @param source The constant that indicates the source
     * @param listener What to do if the button is selected
     */
    private Button createLabelSourceButton(
        Composite parent,
        String title,
        int source) {
        Button b = new Button(parent, SWT.RADIO);
        b.setText(title);
        b.setData(new Integer(source));
        return b;
    }
    private void createLabelSourceOptions(Composite parent) {
        Group labelSource = new Group(parent, SWT.NONE);
        labelSource.setLayout(new GridLayout(1, true));
        labelSource.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        labelSource.setText("Source");
        _labelButtons = new Button[6];
        _labelButtons[0] =
            createLabelSourceButton(
                labelSource,
                "No label",
                LabelOptions.NOLABEL);
        _labelButtons[1] =
            createLabelSourceButton(
                labelSource,
                "Static label",
                LabelOptions.STATICLABEL);
        Composite labelTextGroup = new Composite(labelSource, SWT.NONE);
        //labelTextGroup.setLayout(new GridLayout(1, false));
        //labelTextGroup.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
        _labelText = new Text(labelTextGroup, SWT.SINGLE | SWT.BORDER);
        GridData gd = new GridData();
        gd.horizontalIndent = 25;
        labelTextGroup.setLayoutData(gd);
        _labelText.setSize(85, 20);
        _labelText.addModifyListener(new ModifyListener() {
            public void modifyText(ModifyEvent e) {
                _vis.setLabelValue(_labelText.getText());
                updatePreview ();
            }
        });

        _labelButtons[2] =
            createLabelSourceButton(
                labelSource,
                "Acme identifier",
                LabelOptions.ACMEID);
        _labelButtons[3] =
            createLabelSourceButton(
                labelSource,
                "Acme type",
                LabelOptions.ACMETYPE);
        _labelButtons[4] =
            createLabelSourceButton(
                labelSource,
                "Acme identifier and type",
                LabelOptions.ACMEIDANDTYPE);
        _labelButtons[5] =
            createLabelSourceButton(
                labelSource,
                "Acme property value",
                LabelOptions.ACMEPROPERTYVALUE);
        _propertyCombo = new Combo(labelSource, SWT.DROP_DOWN);
        gd = new GridData();
        gd.horizontalIndent = 25;
        _propertyCombo.setLayoutData(gd);
        // fill the property combo with the properties from the type
        if (_type != null) {
            AcmeEnumeration properties = _type.getProperties();
            while (properties.moreItems()) {
                AcmeProperty prop = (AcmeProperty) properties.getNextItem();
                _propertyCombo.add(prop.getName());
            }
        }
        _propertyCombo.addSelectionListener(new SelectionListener() {
            public void widgetDefaultSelected(SelectionEvent evt) {
            }
            public void widgetSelected(SelectionEvent evt) {
                int selection = _propertyCombo.getSelectionIndex();
                if (selection != -1) {
                    _vis.setLabelProperty(_propertyCombo.getItem(selection));
                } else {
                    _vis.setLabelProperty(_propertyCombo.getText());
                }
                updatePreview ();
            }
        });
        SelectionAdapter buttonSelected = new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                if (((Button) evt.widget).getSelection()
                    && evt.widget.getData() instanceof Integer) {
                    Integer source = (Integer) evt.widget.getData();
                    _vis.setLabelSource(source.intValue());
                    if (source.intValue()
                        == LabelOptions.ACMEPROPERTYVALUE) {
                        _propertyCombo.setEnabled(true);
                    } else {
                        _propertyCombo.setEnabled(false);
                    }
                    if (source.intValue()
                        == LabelOptions.STATICLABEL) {
                        _labelText.setEnabled(true);
                    } else {
                        _labelText.setEnabled(false);
                    }
                    updatePreview ();
                }
            }
        };
        for (int i = 0; i < _labelButtons.length; i++) {
            _labelButtons[i].addSelectionListener(buttonSelected);
        }
    }

    private void createShapeGroup(Composite parent) {
        Group shapeGroup = new Group(parent, SWT.NONE);
        GridData gd = new GridData(GridData.FILL_HORIZONTAL);
        gd.verticalAlignment = GridData.BEGINNING;
        shapeGroup.setLayoutData(gd);
        GridLayout shapeLayout = new GridLayout(2, false);
        shapeGroup.setLayout(shapeLayout);
        shapeGroup.setText("Basic Shape");
        Label shapeLabel = new Label(shapeGroup, SWT.NONE);
        shapeLabel.setText("Stock Image:");
        Composite shapeComboContainer = new Composite(shapeGroup, SWT.NONE);
        _shapeCombo = new Combo(shapeComboContainer, SWT.DROP_DOWN);
        Label directionLabel = new Label(shapeGroup, SWT.NONE);
        directionLabel.setText("Rotation:");
        GridData dld = new GridData();
        dld.horizontalAlignment = GridData.HORIZONTAL_ALIGN_END;
        directionLabel.setLayoutData(dld);
        Composite directionComboContainer = new Composite(shapeGroup, SWT.NONE);
        _directionCombo = new Combo(directionComboContainer, SWT.DROP_DOWN);
        _directionCombo.setSize(100, 20);
        _shapeCombo.setSize(100, 20);
        _shapeCombo.setItems(Shape.LABELS);
        _shapeCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                // Sets the new stock shape. Because the constants in
                // Visualization.Shapes begin at 1, we need to add one
                // to the selection
                _vis.setShape(_shapeCombo.getSelectionIndex() + 1);
                switch (_shapeCombo.getSelectionIndex() + 1) {
                    case Shape.TRIANGLE :
                    case Shape.FILTER :
                        _directionCombo.setEnabled(true);
                        break;
                    default :
                        _directionCombo.setEnabled(false);
                }
                updatePreview();
            }
        });
        final String[] directionOptions =
            new String[] { "0", "90", "180", "270" };
        _directionCombo.setItems(directionOptions);
        _directionCombo.setData("270", new Integer(PositionConstants.NORTH));
        _directionCombo.setData("90", new Integer(PositionConstants.SOUTH));
        _directionCombo.setData("0", new Integer(PositionConstants.EAST));
        _directionCombo.setData("180", new Integer(PositionConstants.WEST));
        _directionCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                Integer data =
                    (Integer) _directionCombo.getData(
                        directionOptions[_directionCombo.getSelectionIndex()]);
                _vis.setDirection(data.intValue());
                updatePreview ();
            }
        });
        _replicateShapeButton = new Button(shapeGroup, SWT.CHECK);
        _replicateShapeButton.setText("Replicate Shape");
        _replicateShapeButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _vis.setReplicateShape(_replicateShapeButton.getSelection());
                updatePreview ();
            }
        });
    }

    /**
     * Creates the shape page that contains controls to edit
     * the shape of the style.
     * @param parent Where the page will go
     */
    public Composite createShapePage(Composite parent) {
        Composite thePage = new Composite(parent, SWT.NONE);
        GridLayout pageLayout = new GridLayout(2, true);
        thePage.setLayout(pageLayout);
        // The left column
        Composite leftColumn = new Composite(thePage, SWT.NONE);
        leftColumn.setLayout(new GridLayout(1, false));
        leftColumn.setLayoutData(new GridData(GridData.FILL_BOTH));
        // The shape selection group
        createShapeGroup(leftColumn);
        createSizeGroup(leftColumn);
        // The size settings
        // The color and other options group
        Composite rightColumn = new Composite(thePage, SWT.NONE);
        GridLayout gl = new GridLayout(1, false);
        rightColumn.setLayout(gl);
        rightColumn.setLayoutData(new GridData(GridData.FILL_BOTH));
        createDetailsGroup(rightColumn);
        //createImageGroup(thePage);
        //      thePage.setSize (400, 400);
        return thePage;
    }

    private void createSizeGroup(Composite parent) {
        Group sizeGroup = new Group(parent, SWT.NONE);
        GridData sgld = new GridData();
        sgld.grabExcessHorizontalSpace = true;
        sgld.horizontalAlignment = GridData.FILL;
        sizeGroup.setLayoutData(sgld);
        sizeGroup.setText("Size settings");
        GridLayout sggl = new GridLayout(3, false);
        sizeGroup.setLayout(sggl);
        _preserveAspectRatioButton = new Button(sizeGroup, SWT.CHECK);
        _preserveAspectRatioButton.setText("Preserve Aspect Ratio");
        GridData pard = new GridData();
        pard.horizontalSpan = 3;
        _preserveAspectRatioButton.setLayoutData(pard);
        Label dwLabel = new Label(sizeGroup, SWT.NONE);
        dwLabel.setText("Default Width:");
        Composite widthTextContainer = new Composite(sizeGroup, SWT.NONE);
        _defaultWidthText =
            new Text(widthTextContainer, SWT.SINGLE | SWT.BORDER);
        _defaultWidthText.setSize(60, 20);
        //_defaultWidthText.setLayoutData(new GridData(GridData.FILL));
        // Make sure that the size is a number
        _defaultWidthText.addVerifyListener(Listeners.VERIFY_NUMBER);
        GridData dwtgd = new GridData();
        dwtgd.grabExcessHorizontalSpace = true;
        dwtgd.horizontalAlignment = GridData.FILL;
        // Reflect changes in the preview if the numbers are changed,
        // but only when we leave
        _defaultWidthText.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent evt) {
            }
            // update the preview when the text is left
            public void focusLost(FocusEvent evt) {
                if (_defaultWidthText.getText().equals("")) {
                    _defaultWidthText.setText("0");
                }
                int value = -1;
                try {
                    value = Integer.parseInt(_defaultWidthText.getText());
                } catch (Exception e) {
                }
                if (value > 0) {
                    Dimension size = _vis.getSize();
                    size.width = value;
                    _vis.setSize(size);
                }
                updatePreview ();
            }
        });
        _widthResizeableButton = new Button(sizeGroup, SWT.CHECK);
        _widthResizeableButton.setText("Resizeable");
        Label dhLabel = new Label(sizeGroup, SWT.NONE);
        dhLabel.setText("Default Height:");
        Composite heightTextContainer = new Composite(sizeGroup, SWT.NONE);
        _defaultHeightText =
            new Text(heightTextContainer, SWT.SINGLE | SWT.BORDER);
        _defaultHeightText.setSize(60, 20);
        _defaultHeightText.addVerifyListener(Listeners.VERIFY_NUMBER);
        _defaultHeightText.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent evt) {
            }
            // update the preview when the text is left
            public void focusLost(FocusEvent evt) {
                if (_defaultHeightText.getText().equals("")) {
                    _defaultHeightText.setText("0");
                }
                int value = -1;
                try {
                    value = Integer.parseInt(_defaultHeightText.getText());
                } catch (Exception e) {
                }
                if (value > 0) {
                    Dimension size = _vis.getSize();
                    size.height = value;
                    _vis.setSize(size);
                }
                updatePreview ();
            }
        });
        _heightResizeableButton = new Button(sizeGroup, SWT.CHECK);
        _heightResizeableButton.setText("Resizeable");

        _preserveAspectRatioButton
            .addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _vis.setPreserveAspectRatio(
                    _preserveAspectRatioButton.getSelection());
                updatePreview ();
            }
        });

        _widthResizeableButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _vis.setWidthResizeable(_widthResizeableButton.getSelection());
            }
        });

        _heightResizeableButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _vis.setHeightResizeable(
                    _heightResizeableButton.getSelection());
            }
        });
    }

    protected IPreferenceStore doGetPreferenceStore() {
        return AcmeStudioPlugin.getDefault().getPreferenceStore();
    }

    /**
     * Fills the image with the color
     * @param image The image to fill
     * @param color The color to fill with
     */
    private void drawImage(Image image, Color color) {
        GC gc = new GC(image);
        gc.setBackground(color);
        Rectangle bounds = image.getBounds();
        gc.fillRectangle(0, 0, bounds.width, bounds.height);
        gc.drawRectangle(0, 0, bounds.width - 1, bounds.height - 1);
        gc.dispose();
    }

    /* (non-Javadoc)
     * @see org.eclipse.ui.IWorkbenchPreferencePage#init(org.eclipse.ui.IWorkbench)
     */
    public void init(IWorkbench workbench) {
    }

    private void initPreview(Composite previewGroup) {
        previewImageCanvas = new FigureCanvas(previewGroup);
        GridData gd = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
        previewImageCanvas.setLayoutData(gd);
        previewImageCanvas.setBackground(ColorConstants.white);
        previewPane = new ScrollPane();
        FrameBorder fb = new FrameBorder("Preview");
        TitleBarBorder tbb = (TitleBarBorder) fb.getInnerBorder();
        tbb.setBackgroundColor(ColorConstants.white);
        tbb.setTextColor(ColorConstants.darkGray);
        FontData fontData = new FontData();
        fontData.setStyle(SWT.BOLD);
        tbb.setFont(new Font(null, fontData));
        previewPane.setBorder(fb);
        previewPane.setPreferredSize(510, 120);
        imgFig = new ImageFigure();
        imgFig.setBounds(previewPane.getClientArea());
        previewPane.setContents(imgFig);
        previewImageCanvas.setContents(previewPane);
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.preference.PreferencePage#performApply()
     */
    protected void performApply() {
        saveVisToStore();
        super.performApply();
    }

    protected void performDefaults() {
        super.performDefaults();
        setToDefaults();
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.preference.IPreferencePage#performOk()
     */
    public boolean performOk() {
        saveVisToStore();
        return super.performOk();
    }

    private void saveVisToStore() {
        if (_vis == null)
            return;
        IPreferenceStore store =
            AcmeStudioPlugin.getDefault().getPreferenceStore();
        store.setValue(_storeType + "Shape", _vis.getShape());
        store.setValue(_storeType + "Direction", _vis.getDirection());
        store.setValue(_storeType + "ReplicateShape", _vis.getReplicateShape());
        store.setValue(_storeType + "Width", _vis.getSize().width);
        store.setValue(_storeType + "Height", _vis.getSize().height);
        store.setValue(_storeType + "OutlineWidth", _vis.getOutlineWidth());
        store.setValue(_storeType + "OutlineStyle", _vis.getOutlineStyle());
        store.setValue(
            _storeType + "OutlineColorR",
            _vis.getOutlineColor().getRed());
        store.setValue(
            _storeType + "OutlineColorG",
            _vis.getOutlineColor().getGreen());
        store.setValue(
            _storeType + "OutlineColorB",
            _vis.getOutlineColor().getBlue());
        store.setValue(_storeType + "FillColorR", _vis.getFillColor().getRed());
        store.setValue(
            _storeType + "FillColorG",
            _vis.getFillColor().getGreen());
        store.setValue(
            _storeType + "FillColorB",
            _vis.getFillColor().getBlue());
        store.setValue(_storeType + "TextColorR", _vis.getTextColor().getRed());
        store.setValue(
            _storeType + "TextColorG",
            _vis.getTextColor().getGreen());
        store.setValue(
            _storeType + "TextColorB",
            _vis.getTextColor().getBlue());
        store.setValue(_storeType + "Font", fromFontToString(_vis.getFont()));
        store.setValue(
            _storeType + "LabelVerticalAlignment",
            _vis.getLabelVerticalAlignment());
        store.setValue(
            _storeType + "LabelHorizontalAlignment",
            _vis.getLabelHorizontalAlignment());
        store.setValue(_storeType + "LabelMargin", _vis.getLabelMargin());
        store.setValue(_storeType + "WordWrap", _vis.getWordwrap());
        store.setValue(_storeType + "LabelSource", _vis.getLabelSource());
        String value = _vis.getLabelProperty();
        if (value != null)
            store.setValue(_storeType + "LabelProperty", value);
        value = _vis.getLabelValue();
        store.setValue(_storeType + "WidthResizeable", _vis.canResizeWidth());
        store.setValue(_storeType + "HeightResizeable", _vis.canResizeHeight());
        store.setValue(
            _storeType + "PreserveAspectRatio",
            _vis.mustPreserveAspectRatio());
        if (_vis instanceof ConnectorVisualization) {
        	store.setValue (_storeType+"AutoCenterBinary", ((ConnectorVisualization )_vis).autoCenterBinaryConnectors());
        }
        if (value != null)
            store.setValue(_storeType + "LabelValue", value);
        VisualStyleRegistry.updateDefaultVisualizations(); 
        _vis.firePropertyChange(Visualization.PROP_VIS_CHANGED, _vis);
    }

    protected void setToDefaults() {
        IPreferenceStore store = getPreferenceStore();
        store.setToDefault(_storeType + "Shape");
        store.setToDefault(_storeType + "Direction");
        store.setToDefault(_storeType + "Width");
        store.setToDefault(_storeType + "Height");
        store.setToDefault(_storeType + "OutlineColorR");
        store.setToDefault(_storeType + "OutlineColorG");
        store.setToDefault(_storeType + "OutlineColorB");
        store.setToDefault(_storeType + "FillColorR");
        store.setToDefault(_storeType + "FillColorG");
        store.setToDefault(_storeType + "FillColorB");
        store.setToDefault(_storeType + "TextColorR");
        store.setToDefault(_storeType + "TextColorG");
        store.setToDefault(_storeType + "TextColorB");
        store.setToDefault(_storeType + "OutlineStyle");
        store.setToDefault(_storeType + "OutlineWidth");
        store.setToDefault(_storeType + "ReplicateShape");
        store.setToDefault(_storeType + "Font");
        store.setToDefault(_storeType + "LabelVerticalAlignment");
        store.setToDefault(_storeType + "LabelHorizontalAlignment");
        store.setToDefault(_storeType + "LabelMargin");
        store.setToDefault(_storeType + "WordWrap");
        store.setToDefault(_storeType + "LabelSource");
        store.setToDefault(_storeType + "LabelVaule");
        store.setToDefault(_storeType + "LabelProperty");
        store.setToDefault(_storeType + "HeightResizeable");
        store.setToDefault(_storeType + "WidthResizeable");
        store.setToDefault(_storeType + "PreserveAspectRatio");
        if (_vis instanceof ConnectorVisualization) store.setToDefault(_storeType+ "AutoCenterBinary");
        store.setDefault("GlobalVisStylePref", ASVisualStyle.SAVELOCAL);
        _vis = loadVisFromStore(_style);
        updateFields();
    }
    
    Image fillImage =
        ImageManager.createImage(_storeType+"_EditVis_fill", 30, 12, false);
    Image outlineImage =
        ImageManager.createImage(_storeType+"_EditVis_outline", 30, 12, false);

    private void updateFields() {
        int direction = -1;
        switch (_vis.getDirection()) {
            case PositionConstants.EAST :
                direction = 0;
                break;
            case PositionConstants.SOUTH :
                direction = 1;
                break;
            case PositionConstants.WEST :
                direction = 2;
                break;
            case PositionConstants.NORTH :
                direction = 3;
                break;
        }
        if (direction >= 0)
            _directionCombo.select(direction);

        _shapeCombo.select(_vis.getShape() - 1);
        _replicateShapeButton.setSelection(_vis.getReplicateShape());
        _defaultWidthText.setText(Integer.toString(_vis.getSize().width));
        _defaultHeightText.setText(Integer.toString(_vis.getSize().height));

        //drawImage(fillImage, _vis.getFillColor());
        //drawImage(outlineImage, _vis.getOutlineColor());
        _fillColorButton.setColorValue(_vis.getFillColor().getRGB());
        _outlineColorButton.setColorValue(_vis.getOutlineColor().getRGB());
        _noLineButton.setSelection(false);
        for (int i = 0; i < _lineStyleButtons.length; i++) {
            _lineStyleButtons[i].setSelection(false);
        }
        switch (_vis.getOutlineStyle()) {
            case SWT.NONE :
                _noLineButton.setSelection(true);
                break;
            case SWT.LINE_SOLID :
                _lineStyleButtons[0].setSelection(true);
                break;
            case SWT.LINE_DOT :
                _lineStyleButtons[1].setSelection(true);
                break;
            case SWT.LINE_DASH :
                _lineStyleButtons[2].setSelection(true);
                break;
            case SWT.LINE_DASHDOT :
                _lineStyleButtons[3].setSelection(true);
                break;
            case SWT.LINE_DASHDOTDOT :
                _lineStyleButtons[4].setSelection(true);
                break;
            default :
                _noLineButton.setSelection(true);
        }
        _outlineWidthCombo.select(_vis.getOutlineWidth() - 1);

        Font f = _vis.getFont();
        if (f != null) {
            _fontButton.setFont(f);
        }
        final Image fontImage =
            ImageManager.createImage("EditStyle_fontColor", 30, 12, false);
        drawImage(fontImage, _vis.getTextColor());
        _fontColorButton.setImage(fontImage);

        _horizAlignCombo.select(_vis.getLabelHorizontalAlignment());
        _vertAlignCombo.select(_vis.getLabelVerticalAlignment());
        _marginText.setText(new Integer(_vis.getLabelMargin()).toString());
        _wordwrapButton.setSelection(_vis.getWordwrap());

        _propertyCombo.setEnabled(false);
        _labelText.setEnabled(false);
        String labelValue = _vis.getLabelValue();
        if (labelValue != null)
            _labelText.setText(labelValue);

        for (int i = 0; i < _labelButtons.length; i++) {
            _labelButtons[i].setSelection(false);
        }
        switch (_vis.getLabelSource()) {
            case LabelOptions.NOLABEL :
                _labelButtons[0].setSelection(true);
                break;
            case LabelOptions.ACMEID :
                _labelButtons[2].setSelection(true);
                break;
            case LabelOptions.ACMEIDANDTYPE :
                _labelButtons[4].setSelection(true);
                break;
            case LabelOptions.ACMETYPE :
                _labelButtons[3].setSelection(true);
                break;
            case LabelOptions.ACMEPROPERTYVALUE :
                _labelButtons[5].setSelection(true);
                String prop = _vis.getLabelProperty();
                _propertyCombo.setText(prop != null ? prop : "");
                _propertyCombo.setEnabled(true);
                break;
            case LabelOptions.STATICLABEL :
                _labelButtons[1].setSelection(true);
                _labelText.setEnabled(true);
                String value = _vis.getLabelValue();
                _labelText.setText(value != null ? value : "");
                break;
        }
        _heightResizeableButton.setSelection(_vis.canResizeHeight());
        _widthResizeableButton.setSelection(_vis.canResizeWidth());
        _preserveAspectRatioButton.setSelection(_vis.mustPreserveAspectRatio());
        if (_vis instanceof ConnectorVisualization) 
        	_autoCenterButton.setSelection (((ConnectorVisualization )_vis).autoCenterBinaryConnectors());
    }

	/* (non-Javadoc)
	 * @see org.eclipse.jface.dialogs.IDialogPage#createControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createControl(Composite parent) {
		GridData gd;
		Composite content = new Composite (parent, SWT.NULL);
		setControl (content);
		GridLayout layout = new GridLayout ();
		layout.marginWidth = 0;
		layout.marginHeight = 0;
		content.setLayout (layout);
		applyDialogFont (content);
		initializeDialogUnits(content);
		
		Control body = createContents (content);
		if (body != null) {
			body.setLayoutData (new GridData (GridData.FILL_BOTH));
		}
		
	}
}
