package AcmeStudio.ui.editors;
import java.util.ArrayList;
import java.util.EventObject;
import java.util.Iterator;
import java.util.Map;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.draw2d.FigureCanvas;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.LightweightSystem;
import org.eclipse.draw2d.PositionConstants;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.draw2d.parts.Thumbnail;
import org.eclipse.gef.*;
import org.eclipse.gef.commands.CommandStack;
import org.eclipse.gef.commands.CommandStackListener;
import org.eclipse.gef.editparts.AbstractConnectionEditPart;
import org.eclipse.gef.editparts.ScalableFreeformRootEditPart;
import org.eclipse.gef.editparts.ZoomManager;
import org.eclipse.gef.palette.PaletteRoot;
import org.eclipse.gef.rulers.RulerProvider;
import org.eclipse.gef.tools.SelectionTool;
import org.eclipse.gef.ui.actions.*;
import org.eclipse.gef.ui.palette.PaletteViewer;
import org.eclipse.gef.ui.palette.PaletteViewerProvider;
import org.eclipse.gef.ui.palette.FlyoutPaletteComposite.FlyoutPreferences;
import org.eclipse.gef.ui.parts.GraphicalEditorWithFlyoutPalette;
import org.eclipse.gef.ui.parts.GraphicalViewerKeyHandler;
import org.eclipse.gef.ui.parts.ScrollingGraphicalViewer;
import org.eclipse.gef.ui.rulers.RulerComposite;
import org.eclipse.gef.ui.stackview.CommandStackInspectorPage;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.widgets.*;
import org.eclipse.ui.*;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.ide.IGotoMarker;
import org.eclipse.ui.part.MultiPageSelectionProvider;
import org.eclipse.ui.texteditor.ITextEditorActionDefinitionIds;
import org.eclipse.ui.texteditor.MarkerUtilities;
import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.actions.OpenTypeHierarchyAction;
import AcmeStudio.core.resource.ASDesign;
import AcmeStudio.core.resource.provider.ASResourceProvider;
import AcmeStudio.model.RepresentationCommand;
import AcmeStudio.model.TypeAssignmentCommand;
import AcmeStudio.ui.editors.actions.*;
import AcmeStudio.ui.editors.diagram.DiagramPaletteViewerProvider;
import AcmeStudio.ui.editors.diagram.actions.*;
import AcmeStudio.ui.editors.diagram.figures.AcmeVisElement;
import AcmeStudio.ui.editors.diagram.figures.SystemView;
import AcmeStudio.ui.editors.diagram.palette.AcmeStudioEditDomain;
import AcmeStudio.ui.editors.diagram.palette.PaletteManager;
import AcmeStudio.ui.editors.diagram.parts.*;
import AcmeStudio.ui.editors.diagram.ruler.DiagramRulerProvider;
import AcmeStudio.ui.editors.diagram.ruler.model.DiagramRuler;
import AcmeStudio.ui.editors.diagram.templates.SystemTemplateTransferDropTargetListener;
import AcmeStudio.util.ASEditorInput;
import AcmeStudio.util.AcmeModelChangeListener;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Helper;
import AcmeStudio.util.AcmeStudioChangeManager.AcmeStudioChangeEvent;
import acmeLib.AcmeObject;
import acmeLib.core.extensions.IElementExtension;
import acmeLib.designRules.Scope;
import acmeLib.elements.*;
import acmeLib.properties.AcmeProperty;

/**
 * @author Hong Yan (yh@cs.cmu.edu)
 *
 * Implements the editor that shows graphical representation of an acme system
 * in an Acme design. The editor should provide a visual environment for users
 * to create, edit and delete acme objects. It should also handle and report
 * events triggered by user action, such like element creation, element
 * deletion, element selection and etc.
 * The DiagramEditor supports the visualization of the following acme elements:
 *      1. AcmeComponent
 *      2. AcmeConnector
 *      3. AcmePort
 *      4. AcmeRole
 * 
 */
public class DiagramEditor
extends GraphicalEditorWithFlyoutPalette
implements CommandStackListener, AcmeModelChangeListener, ISelectionChangedListener {
    
    
public static final String	INLINEREPRESENTATION	= "inlineRepresentation";
	/** The key handler of this editor. */
    private KeyHandler _sharedKeyHandler;

    /** The ASResourceProvider providing the edited resource */
    private ASResourceProvider _resourceProv = null;
    /** The acme system maintained in this editor. */
    private AcmeSystem _acmeSystem = null;

    /** True if it is needed to save previous opened document. */
    private boolean _savePreviouslyNeeded = false;

    private boolean _showGrid = true;

    /** The selected acme object. */
    private AcmeObject _selectedObject = null;

    private MultiPageAcmeEditor _parentEditor;
    private PaletteManager _paletteManager;
    private DiagramRuler _topRuler;
    private DiagramRuler _leftRuler;
    private AcmeStudioEditDomain _editDomain;
    private Object _gotoMarkerManager;
    
    /* (non-Javadoc)
     * @see org.eclipse.gef.ui.parts.GraphicalEditor#hookGraphicalViewer()
     */
    protected void hookGraphicalViewer() {
        super.hookGraphicalViewer();
    }
    
    /**
     * Constructor for the editor. 
     * Set the default tool and edit domain.
     */
    public DiagramEditor(MultiPageAcmeEditor parentEditor, AcmeSystem system) {
        super();
        _acmeSystem = system;
        _editDomain = new AcmeStudioEditDomain (this);
        _editDomain.setCommandStack(getCommandStack());
        _editDomain.setDefaultTool(new SelectionTool());
        setEditDomain(_editDomain);
        _parentEditor = parentEditor;
        _gotoMarkerManager = new IGotoMarker () {
            
            public void gotoMarker(IMarker marker) {
                try {
                    int start = MarkerUtilities.getCharStart(marker);
                    int end = MarkerUtilities.getCharEnd(marker);
                    String acmeSource = (String )marker.getAttribute("AcmeErrorSource");
                    if (acmeSource == null) {
                        if (start < 0 || end < 0) {
                            int line = MarkerUtilities.getLineNumber(marker);
                            if (line > -1) {
                                --line;
                                start = _resourceProv.getDocument(_parentEditor.getEditorInput()).getLineOffset(line);
                                end = _resourceProv.getDocument (_parentEditor.getEditorInput ()).getLineLength(line);
                            }
                        }
                        
                        if (start >= 0 && end >= 0 && _resourceProv.isRegionManagerValid()) {
                            AcmeObject o = _resourceProv.getRegionManager().getAcmeObjectAtPos(start);
                            if (o instanceof AcmeElement) {
                                AcmeElement element = (AcmeElement )o;
                                selectInEditor (element);
                            }
                        }
                    }
                    else {
                        AcmeObject o = null;
                        if (getAcmeDesign() != null) {
                            o = getAcmeDesign().lookupQualifiedName(acmeSource);
                        }
                        AcmeElement element = null;
                        if (o != null) {
                            if (o instanceof AcmeElement) {
                                element = (AcmeElement )o;
                            }
                            else if (o instanceof AcmeProperty) {
                                
                                Scope scope = ((AcmeProperty )o).getScope();
                                while (scope instanceof AcmeProperty) {
                                    scope = ((AcmeProperty )scope).getScope();
                                }
                                element= (AcmeElement )scope;
                            }
                        }
                        if (element != null) {
                            selectInEditor(element);
                        }
                    }
                }
                catch (BadLocationException e) {
                    
                }
                catch (CoreException e) {
                    e.printStackTrace();
                }
            }
            
            /**
             * @param element
             */
            private void selectInEditor(AcmeElement element) {
                selectInViewer(new StructuredSelection (element));
                DiagramEditor.this.updateActions(DiagramEditor.this.getSelectionActions());
                getGraphicalViewer().getRootEditPart().refresh();
            }
            
        };
    }
    
    /**
     * Configure the graphical viewer of the editor.
     */
    protected void configureGraphicalViewer() {
        super.configureGraphicalViewer();
        ScrollingGraphicalViewer viewer =
            (ScrollingGraphicalViewer)getGraphicalViewer();
        
        ScalableFreeformRootEditPart root = new ScalableFreeformRootEditPart();
        java.util.List zoomLevels = new ArrayList (3);
        zoomLevels.add (ZoomManager.FIT_ALL);
        zoomLevels.add (ZoomManager.FIT_WIDTH);
        zoomLevels.add (ZoomManager.FIT_HEIGHT);
        root.getZoomManager().setZoomLevelContributions(zoomLevels);
        
        ZoomInAction zoomIn = new ZoomInAction(root.getZoomManager());
        ZoomOutAction zoomOut = new ZoomOutAction(root.getZoomManager());
        getActionRegistry().registerAction(zoomIn);
        getActionRegistry().registerAction(zoomOut);
        getSite ().getKeyBindingService().registerAction(zoomIn);
        getSite ().getKeyBindingService().registerAction(zoomOut);
        viewer.setRootEditPart(root);
        viewer.setEditPartFactory(new ElementEditPartFactory());
        
        EditorContextMenuProvider provider =
            new EditorContextMenuProvider(viewer, getActionRegistry());
        provider.setEditorPart(this);
        
        viewer.setContextMenu(provider);
        getSite ().registerContextMenu("#AcmeStudio.ui.editors.DiagramEditor.contextMenu", provider, viewer);
        
        ((FigureCanvas)viewer.getControl()).setScrollBarVisibility(
                FigureCanvas.ALWAYS);
        viewer.setKeyHandler(
                new GraphicalViewerKeyHandler(viewer).setParent(
                        getCommonKeyHandler()));
        viewer.addSelectionChangedListener(_parentEditor);
        loadProperties ();
        
        IAction showRulers = new ToggleRulerVisibilityAction(getGraphicalViewer());
        getActionRegistry().registerAction(showRulers);
        getSite().getKeyBindingService().registerAction(showRulers);
        
        IAction snapAction = new ToggleSnapToGeometryAction(getGraphicalViewer());
        getActionRegistry().registerAction(snapAction);
        getSite().getKeyBindingService().registerAction(snapAction);
        
        IAction showGrid = new ToggleGridAction (getGraphicalViewer()); 
        getActionRegistry().registerAction(showGrid);
        getSite().getKeyBindingService().registerAction(showGrid);
        
    }
    
    protected void loadProperties() {
        // Ruler properties
        DiagramRuler ruler = getRuler(PositionConstants.WEST);
        RulerProvider provider = null;
        if (ruler != null) {
            provider = new DiagramRulerProvider(ruler);
        }
        getGraphicalViewer().setProperty(RulerProvider.PROPERTY_VERTICAL_RULER, provider);
        ruler = getRuler(PositionConstants.NORTH);
        provider = null;
        if (ruler != null) {
            provider = new DiagramRulerProvider(ruler);
        }
        getGraphicalViewer().setProperty(RulerProvider.PROPERTY_HORIZONTAL_RULER, provider);
        getGraphicalViewer().setProperty(RulerProvider.PROPERTY_RULER_VISIBILITY, 
                new Boolean(getRulerVisibility()));
        
        // Snap to Geometry property
        getGraphicalViewer().setProperty(SnapToGeometry.PROPERTY_SNAP_ENABLED, 
                new Boolean(isSnapToGeometryEnabled()));
        
        // Grid properties
        getGraphicalViewer().setProperty(SnapToGrid.PROPERTY_GRID_ENABLED, 
                new Boolean(isGridEnabled()));
        Dimension spacing = getGridSpacing();
        if (spacing == null)
            spacing = new Dimension(SnapToGrid.DEFAULT_GRID_SIZE, SnapToGrid.DEFAULT_GRID_SIZE);
        getGraphicalViewer().setProperty(SnapToGrid.PROPERTY_GRID_SPACING, spacing);
        Point origin = getGridOrigin();
        if (origin == null)
            origin = new Point();
        getGraphicalViewer().setProperty(SnapToGrid.PROPERTY_GRID_ORIGIN, origin);
        
        // Zoom
        ZoomManager manager = (ZoomManager)getGraphicalViewer()
        .getProperty(ZoomManager.class.toString());
        if (manager != null)
            manager.setZoom(getZoom());
    }
    
    /**
     * @return
     */
    private double getZoom() {
        return 100.0;
    }
    
    /**
     * @return
     */
    private Point getGridOrigin() {
        return new Point ();
    }
    
    /**
     * @return
     */
    private Dimension getGridSpacing() {
        IPreferenceStore store = AcmeStudioPlugin.getDefault ().getPreferenceStore();
        int gridX = store.getInt("DiagramEditor.gridXSpacing");
        int gridY = store.getInt ("DiagramEditor.gridYSpacing");
        return new Dimension (gridX, gridY);
    }
    
    /**
     * @return
     */
    private boolean isGridEnabled() {
        return AcmeStudioPlugin.getDefault ().getPreferenceStore().getBoolean ("DiagramEditor.gridEnabled");
    }
    
    /**
     * @return
     */
    private boolean isSnapToGeometryEnabled() {
        return AcmeStudioPlugin.getDefault ().getPreferenceStore().getBoolean ("DiagramEditor.snapToGeometry");
    }
    
    /**
     * @return
     */
    private boolean getRulerVisibility() {
        return AcmeStudioPlugin.getDefault ().getPreferenceStore().getBoolean ("DiagramEditor.rulersEnabled");
    }
    
    /**
     * 
     */
    private void loadRulers() {
        DiagramRuler ruler = getRuler (PositionConstants.WEST);
        RulerProvider provider = null;
        if (ruler != null) {
            provider = new DiagramRulerProvider (ruler);
        }
        getGraphicalViewer ().setProperty(RulerProvider.PROPERTY_VERTICAL_RULER, provider);
        
        ruler = getRuler (PositionConstants.NORTH);
        provider = null;
        if (ruler != null) {
            provider = new DiagramRulerProvider (ruler);
        }
        getGraphicalViewer ().setProperty (RulerProvider.PROPERTY_HORIZONTAL_RULER, provider);
    }
    
    /**
     * @param i
     * @return
     */
    private DiagramRuler getRuler(int orientation) {
        DiagramRuler result = null;
        switch (orientation) {
        case PositionConstants.NORTH :
            if (_topRuler == null) {
                _topRuler = new DiagramRuler (true);
            }
        result = _topRuler;
        break;
        case PositionConstants.WEST :
            if (_leftRuler == null) {
                _leftRuler = new DiagramRuler (false);
            }
        result = _leftRuler;
        break;
        }
        return result;
    }
    
    /**
     * Initialize the graphical viewer of the editor.
     */
    protected void initializeGraphicalViewer() {
        super.initializeGraphicalViewer();
        ScrollingGraphicalViewer viewer =
            (ScrollingGraphicalViewer)getGraphicalViewer();
        viewer.setContents(getAcmeSystem());
        getGraphicalViewer().addDropTargetListener(
                new SystemTemplateTransferDropTargetListener(getGraphicalViewer()));
        final FigureCanvas canvas = (FigureCanvas)viewer.getControl();
        canvas.getHorizontalBar().addSelectionListener (new SelectionAdapter () {
            public void widgetSelected (SelectionEvent evt) {
                SystemView view = (SystemView )((SystemEditPart )DiagramEditor.this.getEditPart (getAcmeSystem ())).getFigure ();
                view.adjustErrorFigure (canvas.getViewport());
            }
        });
        canvas.getVerticalBar().addSelectionListener (new SelectionAdapter () {
            public void widgetSelected (SelectionEvent evt) {
                SystemView view = (SystemView )((SystemEditPart )DiagramEditor.this.getEditPart (getAcmeSystem ())).getFigure ();
                view.adjustErrorFigure(canvas.getViewport());
            }
        });
        if (AcmeStudioPlugin.getDefault ().getPreferenceStore().getBoolean("DiagramEditor.zoomToFitOnOpen")) {
            
            final Listener zoomToFitListener = new Listener () {
                
                public void handleEvent(Event event) {
                    final Listener ftl = this;
                    Display.getCurrent().syncExec(new Runnable () {
                        
                        public void run () {
                            Rectangle systemBounds = DiagramEditor.this.getSystemView().getSystemBounds();
                            if (systemBounds.width > 200 || systemBounds.height > 200)
                                DiagramEditor.this.getActionRegistry().getAction (ZoomToFitAction.ID).run ();
                            if (getZoomManager ().getZoom() > 3) getZoomManager ().setZoom(2);
                            canvas.removeListener(SWT.Resize, ftl);
                        }
                    });
                }
            };
            canvas.addListener(SWT.Resize, zoomToFitListener);
        }
        //		getActionRegistry().getAction (ZoomToFitAction.ID).run ();
        viewer.addSelectionChangedListener(this);
        
        ((ScalableFreeformRootEditPart )viewer.getRootEditPart()).getZoomManager().setZoomLevels(new double [] {0.1, 0.25, 0.5, 0.75, 1, 1.5, 2, 3, 4});
    }
    
    /**
     * Returns the KeyHandler with common bindings for both the Outline and Graphical Views.
     * For example, delete is a common action.
     */
    protected KeyHandler getCommonKeyHandler() {
        if (_sharedKeyHandler == null) {
            _sharedKeyHandler = new KeyHandler();
            //            	public boolean keyPressed (KeyEvent evt) {
            //            		evt.keyCode = 0;
            //            		return super.keyPressed(evt);
            //            	}
            //            };
            //            
            _sharedKeyHandler.put(
                    KeyStroke.getPressed(SWT.DEL, 127, 0),
                    getActionRegistry().getAction(ActionFactory.DELETE.getId()));
            // Maintain compability with 2.0.1
            _sharedKeyHandler.put(
                    KeyStroke.getPressed(SWT.DEL, 0, 0),
                    getActionRegistry().getAction(ActionFactory.DELETE.getId ()));
            // Editing the name is now handled by direct editing of the element, so
            // we don't need to add this to the key handler
            //            for ( char c = 'a'; c <= 'z'; c++) {
            //              	EditNamePopupAction action = new EditNamePopupAction (this);
            //            	action.setInitialChar(c);
            //            	action.setId (EditNamePopupAction.ID+c);
            //            	getActionRegistry ().registerAction (action);
            //            	getSelectionActions().add(action.getId ());
            //            	_sharedKeyHandler.put (KeyStroke.getPressed(c, 0), action);
            //            }
            //            for ( char c = 'A'; c <= 'Z'; c++) {
            //				EditNamePopupAction action = new EditNamePopupAction (this);
            //				action.setInitialChar(c);
            //				action.setId (EditNamePopupAction.ID+c);
            //				getActionRegistry ().registerAction(action);
            //				getSelectionActions ().add (action.getId ());
            //				_sharedKeyHandler.put (KeyStroke.getPressed(c,  0), action);
            //            }
        }
        return _sharedKeyHandler;
    }
    
    /**
     * Returns the current acme system.
     * @return AcmeSystem
     */
    public AcmeSystem getAcmeSystem() {
        return _acmeSystem;
    }

    /**
     * Returns the current acme design.
     * [PUZZLE] Should this be derived from parent design of AcmeSystem, or the
     *   AcmeDesign of the ASDesign object managed by the supplied ASRP?
     * @return AcmeDesign
     * @deprecated use instead #getDesign() which returns ASDesign
     */
    protected AcmeDesign getAcmeDesign() {
        return (getASDesign() == null) ? null : getASDesign().getDesign();
    }
    
    /**
     * Returns the current ASDesign.
     * @return ASDesign
     */
    protected ASDesign getASDesign() {
        return (_resourceProv == null) ? null : _resourceProv.getASDesign();
    }
    
    /**
     * Get the root of the toolkit palette.
     */
    protected PaletteRoot getPaletteRoot() {
        if (_paletteManager != null) _paletteManager.dispose ();
        _paletteManager = new PaletteManager(getAcmeSystem(), this);
        //        Point size = getPaletteViewer().getControl().getSize();
        //        size.x = 400;
        //        getPaletteViewer().getControl().setSize(size);
        return _paletteManager.getPaletteRoot();
    }
    
    /**
     * Initialize the action registry and register the following actions:
     *      1. Delete
     *      2. New Component
     *      3. New Connector
     *      4. New Role
     *      5. New Port
     *      6. New Binding
     *      7. Alignment
     *      8. Save
     */
    protected void createActions() {
        //        super.createActions();
        ActionRegistry registry = getActionRegistry();
        IAction action;
        
        IWorkbenchPart wpart = this;
        
        action = new UndoAction(this);
        registry.registerAction(action);
        getStackActions().add(action.getId());
        
        action = new RedoAction(this);
        registry.registerAction(action);
        getStackActions().add(action.getId());
        
        action = new DiagramSelectAllAction (this);
        registry.registerAction (action);
        
        action = new DiagramDeleteAction(this);
        registry.registerAction(action);
        getSelectionActions().add(action.getId());
        
        action = new SaveAction(this);
        registry.registerAction(action);
        getPropertyActions().add(action.getId());
        
        action = new PrintAction(this);
        registry.registerAction(action);
        
        //		IAction deleteAction = new DeleteAction (this);
        //        registry.registerAction (deleteAction);
        
        //        setAction(ZoomAction.ZOOM_IN, new ZoomAction(this, true));
        //        setAction(ZoomAction.ZOOM_OUT, new ZoomAction(this, false));
        
        createSelectionAction(
                EditTypeVisualizationAction.ID,
                new EditTypeVisualizationAction(this));
        createSelectionAction(
                TypeAssignmentCommand.ASSIGN_TYPE,
                new TypeAssignmentMenuAction(
                        this,
                        TypeAssignmentCommand.ASSIGN_TYPE));
        createSelectionAction(
                TypeAssignmentCommand.UNASSIGN_TYPE,
                new TypeAssignmentMenuAction(
                        this,
                        TypeAssignmentCommand.UNASSIGN_TYPE));
        createSelectionAction(
                TypeAssignmentCommand.ASSIGN_FAMILY,
                new TypeAssignmentMenuAction(
                        this,
                        TypeAssignmentCommand.ASSIGN_FAMILY));
        createSelectionAction(
                TypeAssignmentCommand.UNASSIGN_FAMILY,
                new TypeAssignmentMenuAction(
                        this,
                        TypeAssignmentCommand.UNASSIGN_FAMILY));
        createSelectionAction(NavigateUpAction.ID, new NavigateUpAction(this));
        //       registry.registerAction(new AssignTypeAction(this));
        createSelectionAction(
                RepresentationCommand.ADD_REP,
                new RepresentationAction(RepresentationCommand.ADD_REP, this));
        createSelectionAction(
                RepresentationCommand.REMOVE_REP,
                new RepresentationAction(RepresentationCommand.REMOVE_REP, this));
        createSelectionAction (
                ExportUnboundToParentAction.ID,
                new ExportUnboundToParentAction (this));
        createSelectionAction(
                NewElementAction.NEW_COMPONENT,
                new NewElementAction(this, NewElementAction.NEW_COMPONENT));
        createSelectionAction(
                NewElementAction.NEW_CONNECTOR,
                new NewElementAction(this, NewElementAction.NEW_CONNECTOR));
        createSelectionAction(
                NewElementAction.NEW_ROLE,
                new NewElementAction(this, NewElementAction.NEW_ROLE));
        action =
            new NewConnectionPatternAction(this, NewConnectionPatternAction.NEW_CONNECTION_RULE);
        registry.registerAction(action);
        action = new BrowsePatternsAction (this, BrowsePatternsAction.BROWSE_PATTERNS);
        registry.registerAction (action);
        createSelectionAction(
                NewElementAction.NEW_PORT,
                new NewElementAction(this, NewElementAction.NEW_PORT));
        createSelectionAction (EditNamePopupAction.ID, new EditNamePopupAction (this));
        createSelectionAction (ChangeOrderAction.MOVE_BACK, new ChangeOrderAction (this, ChangeOrderAction.MOVE_BACK));
        createSelectionAction (ChangeOrderAction.MOVE_FORWARD, new ChangeOrderAction (this, ChangeOrderAction.MOVE_FORWARD));
        createSelectionAction(ChangeOrderAction.MOVE_TO_BACK, new ChangeOrderAction (this, ChangeOrderAction.MOVE_TO_BACK));
        createSelectionAction(ChangeOrderAction.MOVE_TO_FRONT, new ChangeOrderAction (this, ChangeOrderAction.MOVE_TO_FRONT));
        
        getActionRegistry().registerAction(new ShowGridAction(this));
        getActionRegistry().registerAction(new ZoomToFitAction (this));
        createSelectionAction(
                NewBindingAction.NEW_BINDING,
                new NewBindingAction(this));
        SelectionAction selAction = new DirectEditAction (wpart);
        createSelectionAction(selAction.getId (), selAction);
        selAction = new AlignmentAction (wpart, PositionConstants.LEFT);
        createSelectionAction (selAction.getId (), selAction);
        selAction = new AlignmentAction(wpart, PositionConstants.RIGHT);
        createSelectionAction(selAction.getId (), selAction);
        selAction = new AlignmentAction(wpart, PositionConstants.TOP);
        createSelectionAction(selAction.getId (), selAction);
        selAction = new AlignmentAction (wpart, PositionConstants.BOTTOM);
        createSelectionAction(selAction.getId (), selAction);
        selAction = new AlignmentAction(wpart, PositionConstants.CENTER);
        createSelectionAction(selAction.getId (), selAction);
        selAction = new AlignmentAction(wpart, PositionConstants.MIDDLE);
        createSelectionAction(selAction.getId (), selAction);
        createSelectionAction(
                ActionFactory.COPY.getId (),
                new CopyAcmeElementAction(this));
        createSelectionAction (ActionFactory.CUT.getId (), new CutElementsAction (this));
        createSelectionAction(
                ActionFactory.PASTE.getId (),
                new PasteAcmeElementAction(this));
        createSelectionAction(NavigateUpAction.ID, new NavigateUpAction(this));
        createSelectionAction(
                ResetSystemLayoutCommand.RESET_SYSTEMLAYOUT,
                new ResetSystemLayoutAction(
                        this,
                        ResetSystemLayoutCommand.RESET_SYSTEMLAYOUT));
        createSelectionAction(
                ResetPortAlignmentCommand.RESET_ALIGNMENT,
                new ResetPortAlignmentAction(
                        this,
                        ResetPortAlignmentCommand.RESET_ALIGNMENT));
        createSelectionAction(
                AdjustPortAlignmentCommand.ADJUST,
                new AdjustPortAlignmentAction(
                        this,
                        AdjustPortAlignmentCommand.ADJUST));
        createSelectionAction(
                ResetRoleRouteCommand.RESET_ROUTE,
                new ResetRoleRouteAction(
                        this,
                        ResetRoleRouteCommand.RESET_ROUTE));
        createSelectionAction (OpenTypeHierarchyAction.ID, 
                new SelectionAction (wpart) {
            protected OpenTypeHierarchyAction _action = new OpenTypeHierarchyAction ();
            
            public boolean calculateEnabled () {
                return _action.isEnabled();
            }
            
            public void run () {
                Iterator i = getSelectedObjects ().iterator();
                while (i.hasNext()) {
                    Object o = i.next ();
                    if (o instanceof EditPart) {
                        if (((EditPart )o).getModel() instanceof AcmeElement) {
                            _action.setElement((AcmeElement )((EditPart )o).getModel());
                            _action.run ();
                        }
                    }   
                }
            }      
            
            public String getText () {return _action.getText();}
            
            public void setSelection (ISelection selection) {
                super.setSelection (selection);
            }
        });
        createSelectionAction(NavigateToImportedFamilyAction.ID, new NavigateToImportedFamilyAction (this));
        createSelectionAction (WrapRepresentationAction.WRAPREP_ID, new WrapRepresentationAction (this, WrapRepresentationAction.WRAPREP_ID));
        createSelectionAction (HideSelectionAction.ID, new HideSelectionAction (this));
        registry.registerAction(new ShowFiguresAction (this));
        
        createSelectionAction (HighlightObjectsInDiagramAction.ID, new HighlightObjectsInDiagramAction (this));
        
        registry.registerAction(new UnhighlightAllAction (this));
        
        createSelectionAction (GEFActionConstants.MATCH_HEIGHT, new MatchHeightAction (this));
        createSelectionAction (GEFActionConstants.MATCH_WIDTH, new MatchWidthAction (this));
        //        createSelectionAction (GEFActionConstants.MATCH_SIZE, new MatchSizeAction (this));
        //        IAction showHRuler = new ToggleRulerVisibilityAction (_rulerComp, PositionConstants.NORTH, getGraphicalViewer ());
        //        IAction showVRuler = new ToggleRulerVisibilityAction (_rulerComp, PositionConstants.BOTTOM, getGraphicalViewer ());
        //        registry.registerAction (showHRuler);
        //        registry.registerAction (showVRuler);
        //        getSite ().getKeyBindingService().registerAction (showHRuler);
        //        getSite ().getKeyBindingService().registerAction (showVRuler);
        //        
        // Hook in global actions
        //        getActionRegistry()
        //            .getAction(IWorkbenchActionConstants.UNDO)
        //            .setActionDefinitionId(IWorkbenchActionDefinitionIds.UNDO);
        //        setAction(new SaveAction(this));
        //		getActionRegistry().getAction (ZoomToFitAction.ID).run ();
        
    }
    
    protected void createSelectionAction(String id, SelectionAction a) {
        a.setId(id);
        getActionRegistry().registerAction(a);
        getSelectionActions().add(a.getId());
    }
    
    /**
     * @return True if it is needed to save previously edited document.
     */
    protected boolean savePreviouslyNeeded() {
        return _savePreviouslyNeeded;
    }
    
    /**
     * Set savePreviouslyNeeded parameter.
     * @param value True if it is needed to save previously edited document.
     */
    protected void setSavePreviouslyNeeded(boolean value) {
        _savePreviouslyNeeded = value;
    }
    
    /**
     * Set the site of the editor.
     * @see org.eclipse.gef.ui.parts.GraphicalEditor#setSite(IWorkbenchPartSite).
     */
    protected void setSite(IWorkbenchPartSite site) {
        super.setSite(site);
        
    }
    
    /**
     * Handler of command stack change event.
     * @param e Command stack change event.
     */
    public void commandStackChanged(EventObject e) {
        if (isDirty() || isSaveOnCloseNeeded()) {
            if (!savePreviouslyNeeded()) {
                setSavePreviouslyNeeded(true);
                setDirty(true);
            }
        } else {
            setSavePreviouslyNeeded(false);
            setDirty(((CommandStack)e.getSource()).isDirty());
        }
        super.commandStackChanged(e);
    }
    
    /**
     * Perform the SAVE action.
     * @param progressMonitor The progress monitor.
     */
    public void doSave(IProgressMonitor progressMonitor) {
        setDirty(true);
    }
    
    /**
     * Get the adapter of a certain type.
     * @param type The type of the adapter.
     */
    public Object getAdapter(Class type) {
        if (type == CommandStackInspectorPage.class)
            return new CommandStackInspectorPage(getCommandStack());
        if (type == CommandStack.class) {
            return getCommandStack();
        }
        if (type == ZoomManager.class)
            return (
                    (ScalableFreeformRootEditPart)getGraphicalViewer()
                    .getRootEditPart())
                    .getZoomManager();
        
        if (type == IGotoMarker.class) {
            return _gotoMarkerManager;
        }
        return super.getAdapter(type);
    }
    
    /**
     * @see org.eclipse.ui.part.EditorPart#gotoMarker(IMarker).
     */
    public void gotoMarker(IMarker marker) {
    }
    
    /**
     * Get the root edit part of the editor.
     * @return The root edit part.
     */
    public SystemEditPart getRootEditPart() {
        return (SystemEditPart)getGraphicalViewer().getEditPartRegistry().get(getAcmeSystem());
    }
    
    /**
     * @return True if it is needed to save the document when close it.
     */
    public boolean isDirty() {
        if (_resourceProv != null) {
            return _resourceProv.isDirty();
        } else {
            return false;
        }
    }
    
    /**
     * Indicates that SAVEAS action is allowed.
     * @return True
     */
    public boolean isSaveAsAllowed() {
        return true;
    }
    
    /**
     * @return True if it is needed to save the document when close it.
     */
    public boolean isSaveOnCloseNeeded() {
        return getCommandStack().getUndoCommand() != null &&
        getCommandStack().isDirty();
    }
    
    /**
     * Sets the dirty status.
     * @param dirty  the new dirty status to set to
     */
    public void setDirty(boolean dirty) {
        if (_resourceProv != null && _resourceProv.isDirty() != dirty) {
            
            // this prevents circular invocation
            _resourceProv.setDirty(dirty);
            firePropertyChange(IEditorPart.PROP_DIRTY);
        }
    }
    
    /**
     * Set the input of the editor.
     * @param input Editor input.
     */
    public void setInput(IEditorInput input) {
        if (getGraphicalViewer() != null) {
            getGraphicalViewer().deselectAll();
            getGraphicalViewer().getVisualPartMap().clear();
            getGraphicalViewer().getRootEditPart().deactivate();
            
            // AcmeStudioChangeManager.globalChangeManager.unregisterListener (this);
        }
        if (input instanceof ASEditorInput) {
            super.setInput(input);
            ASEditorInput model = (ASEditorInput)input;
            _resourceProv = model.getResourceProvider();
            _acmeSystem = (AcmeSystem)model.getElement();
            super.setInput(input);
            if (getGraphicalViewer() != null) { // already have a viewer
                this.configureGraphicalViewer();
                this.initializeGraphicalViewer();
            }
        }
        AcmeStudioChangeManager.registerListener(this);
    }
    
    public void refreshInput() {
        if (getEditorInput() != null) {
            setInput(getEditorInput());
        }
    }
    
    /**
     * Set the current acme system.
     * @param acmeSystem The current acme system.
     * @deprecated never been used, nor should be anymore 
     */
    public void setAcmeSystem(AcmeSystem acmeSystem) {
        _acmeSystem = acmeSystem;
    }
    
    /**
     * @see AcmeStudio.model.AcmeFocusListener#setFocus(AcmeObject)
     */
    public void setFocus(AcmeObject object) {
        if (_selectedObject == object)
            return;
        boolean interestedIn = false;
        if (object instanceof AcmeElement) {
            AcmeElement element = (AcmeElement)object;
            interestedIn = (element.getParentDesign() == getAcmeDesign());
        }
        
        if (interestedIn) {
            //  set the selection to be this
            _selectedObject = object;
            getGraphicalViewer().deselectAll();
            //          getGraphicalViewer ().select ((ElementEditPart )getGraphicalViewer ().getEditPartRegistry().get(object));
        }
        
    }
    
    /**
     * @see AcmeStudio.model.AcmeModelChangeListener#changeModel(AcmeStudioChangeEvent)
     */
    public void changeModel(
            AcmeStudioChangeManager.AcmeStudioChangeEvent evt) {
        if (evt.getData() != null) {
            if (Helper.eventRelatedToDesign(evt, getAcmeDesign())) {
                switch (evt.getEventType()) {
                case AcmeStudioChangeManager.ASSIGN_TYPE :
                case AcmeStudioChangeManager.UNASSIGN_TYPE :
                    if (!(evt.getTarget() instanceof AcmeElementType)) {
                        AcmeElement element = (AcmeElement) evt.getTarget();
                        GraphicalEditPart p =
                            (GraphicalEditPart)getEditPart(element);
                        if (p != null) {
                            IFigure figure = p.getFigure();
                            if (figure instanceof AcmeVisElement) {
                                ((AcmeVisElement)figure)
                                .resetVisualization();
                            }
                        }
                    }
                break;
                case AcmeStudioChangeManager.SET_PROPERTY :
                    // If the visualization has been updated
                    // TODO: Check to see if this needs to happen on SET_PROPERTY, or whether
                    // 		 that can just be removed.
                case AcmeStudioChangeManager.UPDATE_VIS_INFO:
                    if (!(evt.getTarget () instanceof AcmeElementType)) {
                        AcmeElement element = (AcmeElement )evt.getTarget ();
                        if (element instanceof AcmePort) element = element.getParent ();
                        GraphicalEditPart p = (GraphicalEditPart )getEditPart (element);
    						if (p != null && !(p instanceof ConnectorEditPart))
                            p.refresh ();
                    }
                break;
                case AcmeStudioChangeManager.SET_NAME :
                    if (!(evt.getTarget() instanceof AcmeElementType)) {
                        AcmeElement element = (AcmeElement)evt.getTarget();
                        GraphicalEditPart p =
                            (GraphicalEditPart)getEditPart(element);
                        if (p != null) {
                            if (p.getFigure().getParent() != null) {
                                p.refresh();
                                p.getFigure().getParent().repaint();
                            } else {
                                p.getFigure().repaint();
                            }
                        }
                    }
                break;
                case AcmeStudioChangeManager.ASSIGN_FAMILY :
                case AcmeStudioChangeManager.UNASSIGN_FAMILY :
                    //  					_paletteManager.reset();
                    //  					configurePaletteViewer();
                    //  					hookPaletteViewer();
                    //  					initializePaletteViewer();
                    this.refreshPalette(true);
                break;
                case AcmeStudioChangeManager.ALIGN_PORTS :
                    if (getEditPart ((AcmeElement )evt.getTarget()) != null)
                        ((ComponentEditPart) getEditPart((AcmeElement) evt
                                .getTarget()))
                                .refresh();
                break;
                case AcmeStudioChangeManager.ROUTE_ROLES :
                    if (getEditPart ((AcmeElement )evt.getTarget()) != null)
                        ((ConnectorEditPart) getEditPart((AcmeElement) evt
                                .getTarget()))
                                .refresh();
                break;
                case AcmeStudioChangeManager.REMOVE_ATTACHMENT:
                case AcmeStudioChangeManager.ADD_ATTACHMENT: {
                    AcmeAttachment attachment = (AcmeAttachment )evt.getSource();
                    PortEditPart pep = (PortEditPart )getEditPart (attachment.getPort());
                    if (pep != null) {
                        pep.refresh();
                    }
                }
                break;
                case AcmeStudioChangeManager.ADD_TYPE:
                case AcmeStudioChangeManager.REMOVE_TYPE: {
                    _paletteManager.updateFamily((AcmeFamily )evt.getTarget());
                    this.refreshPalette (false);
                    break;
                }
                case AcmeStudioChangeManager.REMOVE_COMPONENT:
                case AcmeStudioChangeManager.REMOVE_CONNECTOR:
                case AcmeStudioChangeManager.REMOVE_PORT:
                case AcmeStudioChangeManager.REMOVE_ROLE: {
                    
                    AcmeElement acmeElement = (AcmeElement )evt.getSource();
                    EditPart part = getEditPart (acmeElement);
                    if (getGraphicalViewer().getSelectedEditParts().contains (part)) {
                        getGraphicalViewer ().deselect(part);
                    }
                }
                // No break here because we should fall through
                default :
                    
                    if (getRootEditPart() != null) {
                        EditPartViewer epv = this.getRootEditPart().getViewer();
                        if (epv != evt.getAnnouncer()) {
                            getRootEditPart().refresh();
                        }
                    }
                    // New intersection calculation
                    /*				IntersectionCalculationThread _thread = new IntersectionCalculationThread ((SystemEditPart )getEditPart(_acmeSystem));
                     _thread.start ();*/
                }
                if (AcmeStudioChangeManager.isPluginEvent(evt)) {
                    if (evt.getTarget() instanceof AcmeElement) {
                        AcmeElement refreshElement = (AcmeElement )evt.getTarget ();
                        if (getEditPart (refreshElement) != null) {
                            getEditPart (refreshElement).refresh();
                        }
                    }
                    getRootEditPart ().refresh ();
                }
            }
        }
    }
    
    public boolean interestedInEvent (int et) {
        return et != AcmeStudioChangeManager.ADD_GROUP && et != AcmeStudioChangeManager.REMOVE_GROUP && et != AcmeStudioChangeManager.ADD_MEMBER && et != AcmeStudioChangeManager.REMOVE_MEMBER;
    }
    /**
     * @see org.eclipse.ui.IWorkbenchPart#getTitle()
     */
    public String getTitle() {
        String systemName = "???";
        if (getAcmeSystem() != null) {
            systemName = getAcmeSystem().getName();
        }
        return "System " + systemName;
    }
    
    /**
     * @see org.eclipse.gef.ui.parts.GraphicalEditorWithPalette#initializePaletteViewer()
     */
    /*   protected void initializePaletteViewer() {
     //    	if (_paletteManager != null) _paletteManager.reset();
      try {
      super.initializePaletteViewer();
      }
      catch (Exception e) {}
      getPaletteViewer ().enableVerticalScrollbar(true);
      getPaletteViewer().addDragSourceListener(
      new TemplateTransferDragSourceListener(getPaletteViewer()));
      }
      */
    public boolean getRefreshNeeded () {
        return _needsRefreshing;
    }
    
    protected boolean _needsRefreshing = false;
    
    public void refreshPalette(boolean resetPalette) {
        if (getPaletteViewerProvider() != null
                && getPaletteViewerProvider().getEditDomain().getPaletteViewer().getControl() != null
                && !getPaletteViewerProvider().getEditDomain ().getPaletteViewer ().getControl().isDisposed()) {
            //            getPaletteViewer ().setContents (getPaletteViewer ().getRootEditPart());
            if (resetPalette && _paletteManager != null) 
            	_paletteManager.reset ();
            //    		if (_paletteManager != null)
            //    			_paletteManager.reset();
            //    		else {
            //    		  _paletteManager = new PaletteManager (_acmeSystem, this);
            //    		}
            PaletteViewer paletteViewer = getEditDomain ().getPaletteViewer ();
            //			paletteViewer.setPaletteRoot(getPaletteManager ().getPaletteRoot());
            /*_paletteManager.dispose ();
             _paletteViewerProvider = null;
             _editDomain = new DefaultEditDomain(this);
             _editDomain.setCommandStack(getCommandStack());
             _editDomain.setDefaultTool(new SelectionTool());
             setEditDomain(_editDomain);
             PaletteViewer paletteViewer = getEditDomain().getPaletteViewer();
             DiagramPaletteViewerProvider paletteViewerProvider = (DiagramPaletteViewerProvider )getPaletteViewerProvider ();
             paletteViewer = paletteViewerProvider.createPaletteViewer();
             (paletteViewerProvider).configurePaletteViewer(paletteViewer);
             (paletteViewerProvider).hookPaletteViewer(paletteViewer);*/
            paletteViewer.getControl().redraw();
            paletteViewer.getControl().update();
            getGraphicalViewer().getRootEditPart().refresh();
        }
    }
    /**
     * @see org.eclipse.gef.ui.parts.GraphicalEditor#getCommandStack()
     */
    protected CommandStack getCommandStack() {
        CommandStack commandStack =
            AcmeStudioChangeManager.globalChangeManager.getCommandStack(getASDesign());
        if (this.getEditDomain() != null) {
            this.getEditDomain().setCommandStack(commandStack);
        }
        return commandStack;
    }
    
    /**
     * Get the edit part representing the model
     * @param model The model of the edit part
     * @return The Edit part of the model
     */
    public EditPart getEditPart(AcmeElement model) {
        EditPart editPart = null;
        if (getGraphicalViewer() != null
                && getGraphicalViewer().getEditPartRegistry() != null) {
            Map map = this.getGraphicalViewer().getEditPartRegistry();
            Object o = map.get (model);
            if (o == null) {
                Iterator iter = map.keySet().iterator();
                while (iter.hasNext()) {
                    Object obj = iter.next();
                    if (obj instanceof AcmeElement) {
                        String objName = ((AcmeElement)obj).getQualifiedName();
                        if (model.getQualifiedName().equals(objName)) {
                            editPart = (EditPart)map.get(obj);
                            break;
                        }
                    }
                }
            }
            else {
                editPart = (EditPart )o;
            }
        }
        return editPart;
    }
    
    /**
     * Get the system view
     * @return The system view
     */
    public SystemView getSystemView() {
        return (SystemView) getFigure(getAcmeSystem());
    }
    
    /**
     * Get the figure representing the model
     * @param model The model of the figure
     * @return The figure of the model
     */
    public IFigure getFigure(AcmeElement model) {
        EditPart editPart = getEditPart(model);
        IFigure fig = null;
        if (editPart != null) {
            if (editPart instanceof GraphicalEditPart) {
                GraphicalEditPart ep = (GraphicalEditPart) editPart;
                fig = ep.getFigure ();
            }
            else if (editPart instanceof AbstractConnectionEditPart) {
                AbstractConnectionEditPart ep = (AbstractConnectionEditPart) editPart;
                fig = ep.getFigure();
            }
        }
        return fig;
    }
    /**
     * @see org.eclipse.ui.IEditorPart#doSaveAs()
     */
    public void doSaveAs() {
    }
    
    /**
     * @see org.eclipse.ui.IWorkbenchPart#dispose()
     */
    public void dispose() {
        AcmeStudioChangeManager.removeListener(this);
        if (getGraphicalViewer() != null) {
            getGraphicalViewer().removeSelectionChangedListener(_parentEditor);
            getGraphicalViewer().removeSelectionChangedListener(this);
        }
        if (_paletteManager != null) {
            _paletteManager.dispose();
            _paletteManager = null;
        }
        _resourceProv = null;
        _parentEditor = null;
        _parent = null;
        
        super.dispose();
    }
    
    public ZoomManager getZoomManager() {
        return (
                (ScalableFreeformRootEditPart)getGraphicalViewer()
                .getRootEditPart()).getZoomManager();
    }
    
    /**
     * @see org.eclipse.jface.viewers.ISelectionChangedListener#selectionChanged(SelectionChangedEvent)
     */
    public void selectionChanged(SelectionChangedEvent event) {
        
        if (event.getSelectionProvider() != this.getGraphicalViewer()
                && !(event.getSelectionProvider()
                        instanceof MultiPageSelectionProvider)
                        && !(event.getSelectionProvider()
                                instanceof ScrollingGraphicalViewer)) {
            ISelection selection = event.getSelection();
            selectInViewer (selection);
        }
        this.updateActions(this.getSelectionActions());
        getGraphicalViewer().getRootEditPart().refresh();
        
    }
    
    /**
     * @param selection
     */
    private void selectInViewer(ISelection selection) {
        if (selection instanceof IStructuredSelection) {
            IStructuredSelection sels = (IStructuredSelection) selection;
            if (sels.isEmpty())
                return;
            Iterator i = sels.iterator();
            java.util.List selectedEditParts = new ArrayList();
            
            while (i.hasNext()) {
                Object sel = i.next();
                
                if (sel instanceof AcmeElement
                        && !(sel instanceof AcmeElementType)
                        && !(sel instanceof AcmeSystem)) {
                    Object ep =
                        getGraphicalViewer().getEditPartRegistry().get(sel);
                    if (ep instanceof EditPart) {
                        //                        getGraphicalViewer().setSelection(
                        //                            new StructuredSelection(ep));
                        selectedEditParts.add(ep);
                    }
                }
            }
            if (!selectedEditParts.isEmpty()) {
                
                StructuredSelection newSelections =
                    new StructuredSelection(selectedEditParts);
                getGraphicalViewer().setSelection(newSelections);
            }
        }
    }
    
    /**
     * @see org.eclipse.gef.ui.parts.GraphicalEditor#getGraphicalViewer()
     */
    public GraphicalViewer getGraphicalViewer() {
        return super.getGraphicalViewer();
    }
    
    /**
     * @see org.eclipse.ui.IWorkbenchPart#createPartControl(Composite)
     */
    
    Composite _parent;
    
    private java.util.List _hiddenFigureGroups = new ArrayList ();
    
    private boolean _highlightMode;
    
    private RulerComposite _rulerComp;
    
    private Canvas _overviewParent;
    
    private PaletteViewerProvider	_paletteViewerProvider;
    
    public static final String	PALETTE_DOCK_LOCATION	=  "Dock Location";
    
    public static final String	PALETTE_STATE	= "Palette state";
    
    public static final String	PALETTE_SIZE	= "Palette size";
    
    private static final double	DEFAULT_PALETTE_SIZE	= 130;
	private IElementExtension	_inlineRepData = new IElementExtension () {
		
		public boolean isPersistent() {
			return false;
		}

		public void injectIntoElement(AcmeElement element) {
		}

		public void extractFromElement(AcmeElement element) {
		}

		public boolean isInjected(AcmeElement element) {
			return false;
		}
		
	};
    
    static {
        AcmeStudioPlugin.getDefault().getPreferenceStore().setDefault (PALETTE_SIZE, DEFAULT_PALETTE_SIZE);
    }
    
    public void createPartControl(Composite parent) {
        _parent = parent;
        
        //        if (getAcmeSystem ().getParentRepresentation() == null) {
        super.createPartControl(parent);
    }
    /*       else {
     PublicSplitter hsplitter = new PublicSplitter (parent, SWT.HORIZONTAL);
     SashForm c = new SashForm (hsplitter, SWT.VERTICAL);
     //        	Composite c = new Composite (hsplitter, SWT.NONE);
      c.setLayout(new GridLayout (1, false));
      createPaletteViewer (c);
      createParentOverview (c);
      c.setLayout(new GridLayout (1, false));
      getPaletteViewer().getControl().setLayoutData(new GridData (GridData.FILL_VERTICAL));
      _overviewParent.setLayoutData(new GridData (GridData.FILL_BOTH));
      c.setWeights(new int [] {67, 33});
      createGraphicalViewer (hsplitter);
      
      hsplitter.maintainSize (c);
      hsplitter.setFixedSize (getInitialPaletteSize());
      hsplitter.addFixedSizeChangeListener(new PropertyChangeListener () {
      public void propertyChange (PropertyChangeEvent evt) {
      handlePaletteResized (((PublicSplitter )evt.getSource ()).getFixedSize ());
      }
      });
      }*/
    //		getActionRegistry().getAction (ZoomToFitAction.ID).run ();
    
    //    }
    
    /**
     * Places an overview of the parent of a representation
     * @param parent the composite to place the overview
     */
    public void createParentOverview(Composite parent) {
        
        _overviewParent = new Canvas (parent, SWT.BORDER);
        configureOverview ();
        
        // Double-click navigates up
        _overviewParent.addMouseListener (new MouseAdapter () {
            
            public void mouseDoubleClick(MouseEvent e) {
                IAction action = getActionRegistry().getAction(NavigateUpAction.ID);
                action.run ();
            }
        });
    }
    
    /**
     * Finds the right thumbnail and places it in the overview
     *
     */
    public void configureOverview() {
        RootEditPart rep = null;
        AcmeSystem parentSystem = (AcmeSystem) getAcmeSystem ().getParentRepresentation().getParent().getParent ();
        DiagramEditor ded = Helper.getDiagramEditorForSystem(parentSystem);	 
        if (ded == null) return;
        rep = ded.getGraphicalViewer().getRootEditPart();
        if (rep != null) {
            LightweightSystem lws = new LightweightSystem (_overviewParent);
            ScalableFreeformRootEditPart root = (ScalableFreeformRootEditPart )rep;
            Thumbnail tn = new Thumbnail ((root).getFigure ());
            tn.setSource (root.getLayer (LayerConstants.PRINTABLE_LAYERS));
            lws.setContents (tn);
            tn.setVisible(true);
        }
        
    }
    
    public boolean requiresOverview () {
        return _overviewParent != null;
    }
    
    public void registerKeyBindings() {
        IKeyBindingService kbs = getEditorSite().getKeyBindingService();
        
        IAction action;
        
        action = getActionRegistry().getAction(ActionFactory.UNDO.getId ());
        action.setActionDefinitionId(ITextEditorActionDefinitionIds.UNDO);
        kbs.registerAction(action);
        
        action = getActionRegistry().getAction(ActionFactory.REDO.getId ());
        action.setActionDefinitionId(ITextEditorActionDefinitionIds.REDO);
        kbs.registerAction(action);
        
        action = getActionRegistry().getAction(ActionFactory.COPY.getId ());
        action.setActionDefinitionId(ITextEditorActionDefinitionIds.COPY);
        kbs.registerAction(action);
        
        action = getActionRegistry().getAction(ActionFactory.PASTE.getId ());
        action.setActionDefinitionId(ITextEditorActionDefinitionIds.PASTE);
        kbs.registerAction(action);
        
        action = getActionRegistry().getAction (ActionFactory.SELECT_ALL.getId ());
        action.setActionDefinitionId(ITextEditorActionDefinitionIds.SELECT_ALL);
        kbs.registerAction(action);
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.gef.ui.parts.GraphicalEditor#getActionRegistry()
     */
    public ActionRegistry getActionRegistry() {
        return super.getActionRegistry();
    }
    
    public Canvas getOverview (FigureCanvas overview) {
        LightweightSystem lws = new LightweightSystem (overview);
        if (getGraphicalViewer () != null) {
            RootEditPart rep = getGraphicalViewer ().getRootEditPart();
            if (rep instanceof ScalableFreeformRootEditPart) {
                ScalableFreeformRootEditPart root = (ScalableFreeformRootEditPart )rep;
                Thumbnail thumbnail = new Thumbnail (root.getFigure());
                lws.setContents (thumbnail);  
            }   
        }
        return overview;
    }
    
    public ISelection getSelection () {
        return getGraphicalViewer ().getSelection();
    }
    
    /**
     * @return boolean
     */
    public boolean isShowGrid() {
        return _showGrid;
    }
    
    /**
     * Sets the _showGrid.
     * @param _showGrid The _showGrid to set
     */
    public void setShowGrid(boolean _showGrid) {
        this._showGrid = _showGrid;
    }
    
    
    //	/* (non-Javadoc)
    //	 * @see org.eclipse.gef.ui.parts.GraphicalEditorWithPalette#createPaletteViewer(org.eclipse.swt.widgets.Composite)
    //	 */
    //	protected void createPaletteViewer(Composite parent) {
    //		PaletteViewer viewer = new AcmeStudioResilientPaletteViewer ();
    //		setPaletteViewer (viewer);
    //		viewer.createControl (parent);
    //		configurePaletteViewer ();
    //		hookPaletteViewer ();
    //		initializePaletteViewer ();
    //	}
    
    /**
     * @param figuresToHide
     */
    public void addHiddentFigureGroup(AcmeGroup hiddenGroup) {
        _hiddenFigureGroups.add (hiddenGroup);
        
        
    }
    
    /**
     * @return Returns the hiddenFigureGroups.
     */
    public java.util.List getHiddenFigureGroups() {
        return _hiddenFigureGroups;
    }
    
    /**
     * @param b
     */
    public void setHighlightMode(boolean b) {
        _highlightMode = b;
        getRootEditPart().refresh();
    }
    
    /**
     * @return
     */
    public boolean getHighlightMode() {
        return _highlightMode;
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.gef.ui.parts.GraphicalEditor#createGraphicalViewer(org.eclipse.swt.widgets.Composite)
     */
    protected void createGraphicalViewer(Composite parent) {
        _rulerComp = new RulerComposite (parent, SWT.NONE);
        super.createGraphicalViewer(_rulerComp);
        _rulerComp.setGraphicalViewer((ScrollingGraphicalViewer )getGraphicalViewer());
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.gef.ui.parts.GraphicalEditorWithFlyoutPalette#getPalettePreferences()
     */
    protected FlyoutPreferences getPalettePreferences() {
        return new FlyoutPreferences () {
            public int getDockLocation () {
                return AcmeStudioPlugin.getDefault().getPreferenceStore ().getInt (PALETTE_DOCK_LOCATION);
            }
            public int getPaletteState () {
                return AcmeStudioPlugin.getDefault ().getPreferenceStore().getInt (PALETTE_STATE);
            }
            public int getPaletteWidth () {
                int size = AcmeStudioPlugin.getDefault().getPreferenceStore().getInt (PALETTE_SIZE);
                return (int )((size==0)?DEFAULT_PALETTE_SIZE:size);
            }
            public void setDockLocation (int location) {
                AcmeStudioPlugin.getDefault().getPreferenceStore().setValue (PALETTE_DOCK_LOCATION, location);
            }
            public void setPaletteState (int state) {
                AcmeStudioPlugin.getDefault().getPreferenceStore ().setValue (PALETTE_STATE, state);
            }
            public void setPaletteWidth (int width) {
                AcmeStudioPlugin.getDefault().getPreferenceStore().setValue (PALETTE_SIZE, width);
            }
        };
    }
    
    protected PaletteViewerProvider createPaletteViewerProvider () {
        if (_paletteViewerProvider == null) 
            _paletteViewerProvider =  new DiagramPaletteViewerProvider (getEditDomain(), this);
        return _paletteViewerProvider;
    }
    
    /**
     * @return
     */
    public PaletteManager getPaletteManager() {
        return _paletteManager;
    }
    /* (non-Javadoc)
     * @see org.eclipse.gef.ui.parts.GraphicalEditorWithFlyoutPalette#getGraphicalControl()
     */
    protected Control getGraphicalControl() {
        return _rulerComp;
    }
    /**
     * @return Returns the overviewParent.
     */
    public Canvas getOverviewParent() {
        return _overviewParent;
    }
    /**
     * @return Returns the editDomain.
     */
    public DefaultEditDomain getEditDomain() {
        return _editDomain;
    }

    /**
     * 
     */
    public void injectOrderInformation() {
        SystemEditPart sep = (SystemEditPart )getEditPart(getAcmeSystem());
        sep.injectOrderInformation ();
    }

	/**
	 * 
	 */
	public void hideRepresentationBorders() {
		_acmeSystem.setUserData (INLINEREPRESENTATION, _inlineRepData);
		getRootEditPart().refresh ();
	}
	
	public void showRepresentationBorders () {
		_acmeSystem.setUserData(INLINEREPRESENTATION, null);
		getRootEditPart().refresh ();
	}
    
}