package AcmeStudio.ui.editors.actions;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.gef.EditPart;
import org.eclipse.gef.Request;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.editparts.AbstractEditPart;
import org.eclipse.gef.ui.actions.SelectionAction;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchPart;

import AcmeStudio.model.ModelCompoundCommand;
import AcmeStudio.model.TypeAssignmentCommand;
import AcmeStudio.ui.editors.DiagramEditor;
import AcmeStudio.ui.editors.diagram.parts.ElementEditPart;
import AcmeStudio.ui.editors.diagram.parts.RoleEditPart;
import AcmeStudio.ui.editors.diagram.parts.SystemEditPart;
import AcmeStudio.util.IMenuBuilder;
import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.filters.AcmeObjectFilter;
import acmeLib.core.filters.ComponentTypeEnumFilter;
import acmeLib.core.filters.ConnectorTypeEnumFilter;
import acmeLib.core.filters.PortTypeEnumFilter;
import acmeLib.core.filters.RoleTypeEnumFilter;
import acmeLib.elements.AcmeComponent;
import acmeLib.elements.AcmeConnector;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmeFamily;
import acmeLib.elements.AcmePort;
import acmeLib.elements.AcmeRole;
import acmeLib.elements.AcmeSystem;

/**
 * @author Bradley Schmerl (schmerl@cs.cmu.edu)
 *
 * Action for assigning types and families. Builds a submenu
 * with the type names in them.
 * 
 */
public class TypeAssignmentMenuAction
    extends SelectionAction
    implements IMenuBuilder {

    private Map _assignActionCache = new Hashtable ();
    private Map _unassignActionCache = new Hashtable ();

    private Request _request;

    private boolean isGroupAction = true;
    
    private AcmeElementType _targetType;
    private List _targetElements;

    public TypeAssignmentMenuAction (IWorkbenchPart editor, String type) {
        super (editor);
        if (type.equals (TypeAssignmentCommand.ASSIGN_TYPE)) {
            setText ("Assign Type");
        }
        else if (type.equals (TypeAssignmentCommand.UNASSIGN_TYPE)) {
            setText ("Unassign Type");   
        }   
        else if (type.equals (TypeAssignmentCommand.UNASSIGN_FAMILY)) {
            setText ("Unassign Family");
        }
        else if (type.equals (TypeAssignmentCommand.ASSIGN_FAMILY)) {
            setText ("Assign Family");
        }
        setId (type);    
        _request = new Request (type);
    }

    /**
     * @see org.eclipse.gef.ui.actions.EditorPartAction#calculateEnabled()
     */
    protected boolean calculateEnabled() {
        if (!isGroupAction) return true;

        
        return calculateFamilyEnabled () || calculateElementEnabled ();
    }
    
    protected boolean calculateFamilyEnabled () {
        if (getSelectedObjects ().size () == 0 || (getSelectedObjects ().size () == 1 && getSelectedObjects ().get(0) instanceof SystemEditPart)) {
            if (isGroupAction && _request.getType ().equals (TypeAssignmentCommand.ASSIGN_FAMILY)) return true; // Can always import
            IWorkbenchPart ep = getWorkbenchPart();
            if (ep instanceof DiagramEditor && ((DiagramEditor )ep).getRootEditPart () != null) {
                DiagramEditor de = (DiagramEditor )ep;
                AcmeSystem system = (AcmeSystem )de.getRootEditPart().getModel();
                if (_request.getType ().equals (TypeAssignmentCommand.ASSIGN_FAMILY)) {
                    return getAvailableTypes (system).getSize() != 0;
                }
                else if (_request.getType ().equals (TypeAssignmentCommand.UNASSIGN_FAMILY)) {
                    return system.getDeclaredTypes().getSize() != 0;
                }
            }
            
        }
        return false;
    }
    
    protected boolean calculateElementEnabled () {
        if (getSelectedObjects ().size () > 0) {
            Object selObj = getSelectedObjects ().get (0);
            AcmeElement selection = null;
            
            if (selObj instanceof ElementEditPart || selObj instanceof RoleEditPart) {
                Object obj = ((EditPart )selObj).getModel();
                if (obj instanceof AcmeElement && !(obj instanceof AcmeSystem)) {
                    selection = (AcmeElement )obj;
                } else {
                    return false;
                }
            }
            else {
                return false;
            }
            
            if (getAvailableTypes (selection).getSize () == 0) {
                return false;
            }
            
            if (selection instanceof AcmeComponent ||
                selection instanceof AcmeConnector ||
                selection instanceof AcmePort ||
                selection instanceof AcmeRole) {
            
                if (_request.getType().equals (TypeAssignmentCommand.ASSIGN_TYPE)){
                    if (getSelectedObjects ().size () > 1){
                        EditPart ep = (EditPart )getSelectedObjects ().get (0);
                        if (ep.getModel() instanceof AcmeElement) {
                            AcmeElement same = (AcmeElement )ep.getModel ();
                            boolean ok = true;
                            for (int i = 1; i < getSelectedObjects ().size(); i++) {
                                ep = (EditPart )getSelectedObjects ().get (i);
                                if (!ep.getModel ().getClass().equals (same.getClass ())) {
                                    ok = false;
                                }
                            
                            }
                        
                            return ok;
                        }
                    }
                    else {
                        return true;
                    }
                }
                if (_request.getType ().equals (TypeAssignmentCommand.UNASSIGN_TYPE) &&
                    getSharedTypes (getSelectedObjects ()).size() > 0) {
                        return true;
                }
            }
        }
        return false;
    }
    
    protected void setIndividualAction () {
        isGroupAction = false;   
    }
    
    protected AcmeEnumeration getAvailableTypes (AcmeElement element) {
        AcmeEnumeration types = element.getAvailableTypes ();
           
        if (element instanceof AcmeComponent) {
            types = types.castToAcmeVector().enumerate(new ComponentTypeEnumFilter ());
        }
        else if (element instanceof AcmeConnector) {
            types = types.castToAcmeVector().enumerate (new ConnectorTypeEnumFilter ());
        }
        else if (element instanceof AcmePort) {
            types = types.castToAcmeVector().enumerate(new PortTypeEnumFilter ());
        }
        else if (element instanceof AcmeRole) {
            types = types.castToAcmeVector ().enumerate (new RoleTypeEnumFilter ());
        }
        else if (element instanceof AcmeSystem) {
            types = types.castToAcmeVector ().enumerate (new acmeLib.core.filters.AcmeObjectFilter () {
                public boolean test (AcmeObject o) {
                    return  o instanceof AcmeFamily;
                }
            });   
        }
        else {
            types = new AcmeEnumeration ();   
        }
        return types.castToAcmeSetVector ().enumerate ();
        
    }
    
    protected Set getSharedTypes (List selections) {
        Set sharedTypes = new HashSet ();
        if (selections.size() > 0) {
            AcmeElement baseElement = (AcmeElement )((EditPart )selections.get (0)).getModel ();
            sharedTypes = convertAcmeEnumerationToSet (baseElement.getDeclaredTypes());
            sharedTypes.addAll (convertAcmeEnumerationToSet(baseElement.getInstantiatedTypes()));
            
            for (int i = 1; i < selections.size(); i++) {
                Object model = null;
                
                if (selections.get (i) instanceof AbstractEditPart) {
                    AbstractEditPart ep = (AbstractEditPart )selections.get (i);
                    model = ep.getModel ();
                    if (model != null && model instanceof AcmeElement) {                
                        AcmeElement elem = (AcmeElement )model;
                        Set types = convertAcmeEnumerationToSet(elem.getDeclaredTypes ());
                        types.addAll (convertAcmeEnumerationToSet(elem.getInstantiatedTypes()));
                        sharedTypes = intersect (sharedTypes, types);
                    }
                }
            }
        }
        return sharedTypes;
    }
    
    private static Set intersect (Set set1, Set set2) {
        Set ret = new HashSet ();
        Iterator i = set1.iterator();
        while (i.hasNext ()) {
            Object o = i.next ();
            if (set2.contains(o)) {
                ret.add (o);
            }
        }
        return ret;
    }
    
    private static Set convertAcmeEnumerationToSet (AcmeEnumeration enum) {
        Set ret = new HashSet ();
        while (enum.moreItems ()) {
            ret.add (enum.getNextItem ());
        }
        return ret;
    }
    
    public IMenuManager buildMenu () {
        MenuManager manager = new MenuManager (getText ());
        if (getSelectedObjects ().size () == 0) return manager;
        if (getSelectedObjects ().get (0) instanceof ElementEditPart || getSelectedObjects ().get (0) instanceof RoleEditPart) {
            EditPart ep = (EditPart )getSelectedObjects ().get (0);
            if (ep.getModel () instanceof AcmeElement) {
            	List actions = new ArrayList ();
                AcmeElement element = (AcmeElement )ep.getModel();
          
                if (_request.getType().equals (TypeAssignmentCommand.UNASSIGN_TYPE) ||
                    _request.getType ().equals (TypeAssignmentCommand.UNASSIGN_FAMILY)) {
                    Set types = getSharedTypes (getSelectedObjects());
                    Iterator i = types.iterator ();
                    while (i.hasNext ()) {
                        AcmeElementType type = (AcmeElementType )i.next ();
                        TypeAssignmentMenuAction action = null;
                        if (_unassignActionCache.containsKey (type)) {
                            action = (TypeAssignmentMenuAction )_unassignActionCache.get (type);       
                        }
                        else {
                            action = new TypeAssignmentMenuAction (getWorkbenchPart (), (String )_request.getType ());
                            action.setIndividualAction();
                            action._targetType = type; 
                            _unassignActionCache.put (type, action);                     
                        }
                        action._targetElements = getSelectedObjects ();
                        action.setText (type.getTypeName ());
                        actions.add (action);
//                        manager.add (action);                      
                    }
                }
                else if (_request.getType ().equals (TypeAssignmentCommand.ASSIGN_TYPE) ||
                          _request.getType ().equals (TypeAssignmentCommand.ASSIGN_FAMILY)) {
                    AcmeEnumeration types = getAvailableTypes (element);
                    final Set sharedTypes = getSharedTypes (getSelectedObjects ());
                    types = types.castToAcmeVector().enumerate(new AcmeObjectFilter () {

						public boolean test(AcmeObject obj) {
							return !sharedTypes.contains (obj);
						}
                    	
                    });
                    while (types.moreItems ()) {
                        AcmeElementType type = (AcmeElementType )types.getNextItem ();
//                        if (getSelectedObjects().size() == 1) {
//                        	if (element.declaresType (type)) continue;
//                        }
                        TypeAssignmentMenuAction action = null;
                        if (_assignActionCache.containsKey (type)) {
                            action = (TypeAssignmentMenuAction )_assignActionCache.get (type);
                        }
                        else {
                            action = new TypeAssignmentMenuAction (getWorkbenchPart(), (String )_request.getType ());
                            action.setIndividualAction ();
                            action._targetType = type;
                            _assignActionCache.put (type, action);
                        }
                        action._targetElements = getSelectedObjects ();
                        action.setText (type.getTypeName ());
                        actions.add (action);
//                        manager.add (action);
                    }
                    if (_request.getType ().equals (TypeAssignmentCommand.ASSIGN_FAMILY)) {
                        ImportFamilyAction action = new ImportFamilyAction (getWorkbenchPart ());
                        manager.add (action);   
                    }
                }
                Collections.sort(actions, new Comparator () {

					public int compare(Object o1, Object o2) {
						if (o1 instanceof IAction && o2 instanceof IAction) {
							return ((IAction )o1).getText().compareTo(((IAction )o2).getText ());
						}
						return 0;
					}
                	
                });
                for (Iterator iter = actions.iterator(); iter.hasNext();) {
					IAction action = (IAction) iter.next();
					manager.add (action);
				}
            }
        }
        return manager;
    }   
            
               
         
     
    /**
     * @see org.eclipse.jface.action.IAction#run()
     */
    public void run() {
        if (!isGroupAction) {
            if (_request.getType ().equals (TypeAssignmentCommand.UNASSIGN_FAMILY)) {
                if (MessageDialog.openConfirm(Display.getDefault ().getActiveShell (), "Command confirm", "This command cannot be undone")) {
                    execute (getCommand ());
                }
            }
            else {
                execute(getCommand ());
            }
        }
    }
    
    protected Command getCommand () {
        if (!isGroupAction) {
            if (_targetElements.size() == 1) {
                Command command = 
                    new TypeAssignmentCommand ((String )_request.getType (), 
                                               (AcmeElement )((EditPart )_targetElements.get(0)).getModel(),
                                               _targetType, true, null);
                return command;
            }      
            else {
                ModelCompoundCommand command = new ModelCompoundCommand ((String )_request.getType());
                Iterator i = _targetElements.iterator ();
                while (i.hasNext ()) {
                    AcmeElement element = (AcmeElement )((EditPart )i.next ()).getModel ();
                    command.add (new TypeAssignmentCommand ((String )_request.getType (),
                                                             element, _targetType, true, null));   
                }          
                return command;
            }
        }
        return null;
    }

}
