package AcmeStudio.ui.editors.diagram;

import java.util.ArrayList;
import java.util.List;
import org.eclipse.draw2d.PositionConstants;
import org.eclipse.gef.internal.GEFMessages;
import org.eclipse.gef.ui.actions.*;
import org.eclipse.jface.action.*;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.actions.RetargetAction;
import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.actions.AcmeStudioRetargetAction;
import AcmeStudio.model.RepresentationCommand;
import AcmeStudio.model.TypeAssignmentCommand;
import AcmeStudio.prefs.DiagramToolbarPreferences;
import AcmeStudio.ui.editors.DiagramEditor;
import AcmeStudio.ui.editors.MultiPageAcmeEditor;
import AcmeStudio.ui.editors.actions.BrowsePatternsAction;
import AcmeStudio.ui.editors.actions.CopyAcmeElementAction;
import AcmeStudio.ui.editors.actions.NewConnectionPatternAction;
import AcmeStudio.ui.editors.actions.NewElementAction;
import AcmeStudio.ui.editors.diagram.actions.*;
import AcmeStudio.util.IMenuBuilder;
import AcmeStudio.util.ImageManager;

/**
 * @author Bradley Schmerl
 * 
 * Provides the tool bar and menu items for the Diagram editor
 *
 */
public class DiagramActionBarContributor extends ActionBarContributor {

	/* Actions for creating and deleting new elements */
    private AcmeStudioRetargetAction _newComponentAction;
    private AcmeStudioRetargetAction _newConnectorAction;
    private AcmeStudioRetargetAction _newPortAction;
    private AcmeStudioRetargetAction _newRoleAction;
    private AcmeStudioRetargetAction _newConnectionRuleAction;
    private AcmeStudioRetargetAction _browsePatternsAction;
    
    // Actions for manipulating representations
    private AcmeStudioRetargetAction _newRepresentationAction;
    private GlobalSubMenu _removeRepresentationMenu;
    private AcmeStudioRetargetAction _navigateUpRepresentationAction;
	private AcmeStudioRetargetAction _wrapRepAction;

	// Actions for manipulating types
    private GlobalSubMenu _assignTypeMenu;
    private GlobalSubMenu _unassignTypeMenu;
    private GlobalSubMenu _assignFamilyMenu;
    private GlobalSubMenu _unassignFamilyMenu;

	// Actions for manipulating the diagram
    private GlobalSubMenu _resetPortAlignmentMenu;
    private GlobalSubMenu _adjustPortAlignmentMenu;
    private GlobalSubMenu _resetRoleRouteMenu;
    private GlobalSubMenu _resetSystemLayoutMenu;
//    private AcmeStudioRetargetAction _showGridAction;
    private AcmeStudioRetargetAction _zoomToFitAction;
    private AcmeStudioRetargetAction _moveBackAction;
    private AcmeStudioRetargetAction _moveForwardAction;
    private AcmeStudioRetargetAction _move2FrontAction;
    private AcmeStudioRetargetAction _move2BackAction;
	private ZoomComboContributionItem	_zoomCombo;
    
    // A class that allows an action to be a submenu
    private class GlobalSubMenu
        extends MenuManager
        implements ISelectionChangedListener {

		/** The builder used to construct the submenu **/
        public IMenuBuilder _builder;
        /** The diagram editor to listen to for selection changes**/
        protected DiagramEditor _listeningTo;
        /** The parent menu that the submenu will be added to **/
        protected IMenuManager _parent;
        /** The group to add the menu item to **/
        protected String _group;

		/** Constructs a new GlobalSubMenu **/
        public GlobalSubMenu(String id, String label) {
            super(label, id);
        }

		/**
		 * Sets the menu information that this submenu will be added to
		 * @param menu The parent menu
		 * @param group The group to add the submenu to
		 */
        public void setMenu(IMenuManager menu, String group) {
            _group = group;
            _parent = menu;
        }
		
		/**
		 * Sets the builder used to construct the submenu
		 * @param builder The builder to construct the submenu
		 */
        public void setBuilder(IMenuBuilder builder) {
            _builder = builder;
            update();
        }
        /**
        * @see org.eclipse.jface.viewers.ISelectionChangedListener#selectionChanged(SelectionChangedEvent)
        */
        public void selectionChanged(SelectionChangedEvent event) {
            update();
        }

		/**
		 * Updates the submenu by rebuilding the menu and getting the
		 * subitems of the submenu. This allows dynamic enabling/disabling
		 * of actions.
		 */
        public void update() {
            this.removeAll();
            if (_builder != null) {
                IMenuManager manager = _builder.buildMenu();
                IContributionItem[] items = manager.getItems();
                for (int i = 0; i < items.length; i++) {
                    this.add(items[i]);
                }
            }

            if (_parent != null) {
                _parent.remove(this);
                _parent.appendToGroup(_group, this);
            }
            

        }

		/**
		 * Hooks this submenu to listen to selection changes in a particular editor
		 * @param editor
		 */
        public void hook(DiagramEditor editor) {
            unhook();
            _listeningTo = editor;
            editor.getGraphicalViewer().addSelectionChangedListener(this);
        }

		/**
		 * Unhooks the submenu from the editor it is listening to
		 *
		 */
        public void unhook() {
            if (_listeningTo != null) {
                _listeningTo
                    .getGraphicalViewer()
                    .removeSelectionChangedListener(
                    this);
            }
        }

    }

    /**
     * Constructor for DiagramActionBarContributor.
     */
    public DiagramActionBarContributor() {
        super();

        _newComponentAction =
            new AcmeStudioRetargetAction(
                "AcmeStudio.design.newComponent",
                "New Component");
        _newConnectorAction =
            new AcmeStudioRetargetAction(
                "AcmeStudio.design.newConnector",
                "New Connector");
        _newPortAction =
            new AcmeStudioRetargetAction(NewElementAction.NEW_PORT, "New Port");
        _newRoleAction =
            new AcmeStudioRetargetAction(NewElementAction.NEW_ROLE, "New Role");
        _newConnectionRuleAction =
            new AcmeStudioRetargetAction(NewConnectionPatternAction.NEW_CONNECTION_RULE, "New Connection Rule");
            
        _browsePatternsAction = 
            new AcmeStudioRetargetAction(BrowsePatternsAction.BROWSE_PATTERNS, "Browse Connection Rules");
        
//        _showGridAction = new AcmeStudioRetargetAction (ShowGridAction.ID, "Shows grid", true);
        _zoomToFitAction = new AcmeStudioRetargetAction (ZoomToFitAction.ID, "Zoom to fit");
        _zoomToFitAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_ZOOM_TO_FIT));
//        _zoomCombo =  new ZoomComboContributionItem(getPage());
        _newRepresentationAction =
            new AcmeStudioRetargetAction(
                RepresentationCommand.ADD_REP,
                "Add Representation");
        _removeRepresentationMenu =
            new GlobalSubMenu(
                RepresentationCommand.REMOVE_REP,
                "Remove Representation");
        _navigateUpRepresentationAction =
            new AcmeStudioRetargetAction(
                NavigateUpAction.ID,
                "Navigate Up Representation");
        _navigateUpRepresentationAction.setImageDescriptor (ImageManager.getImageDescriptor(ImageManager.ID_NAV_UP));
		_wrapRepAction = new AcmeStudioRetargetAction (WrapRepresentationAction.WRAPREP_ID, "Encapsulate as Component");

        _assignTypeMenu =
            new GlobalSubMenu(TypeAssignmentCommand.ASSIGN_TYPE, "Assign Type");
        _unassignTypeMenu =
            new GlobalSubMenu(
                TypeAssignmentCommand.UNASSIGN_TYPE,
                "Unassign Type");
        _assignFamilyMenu =
            new GlobalSubMenu(
                TypeAssignmentCommand.ASSIGN_FAMILY,
                "Assign Family");
        _unassignFamilyMenu =
            new GlobalSubMenu(
                TypeAssignmentCommand.UNASSIGN_FAMILY,
                "Unassign Family");
        _resetSystemLayoutMenu =
            new GlobalSubMenu(ResetSystemLayoutCommand.RESET_SYSTEMLAYOUT, "Reset Port Alignment");
        _resetPortAlignmentMenu =
            new GlobalSubMenu(ResetPortAlignmentCommand.RESET_ALIGNMENT, "Reset Port Alignment");
        _adjustPortAlignmentMenu =
            new GlobalSubMenu(AdjustPortAlignmentCommand.ADJUST, "Adjust Port Alignment");
        _resetRoleRouteMenu =
            new GlobalSubMenu(ResetRoleRouteCommand.RESET_ROUTE, "Reset Role Route");
        _move2BackAction = new AcmeStudioRetargetAction (ChangeOrderAction.MOVE_TO_BACK, "Move to back");
        _move2BackAction.setImageDescriptor(ImageManager.getImageDescriptor (ImageManager.ID_MOVE_TO_BACK));
        _move2FrontAction = new AcmeStudioRetargetAction (ChangeOrderAction.MOVE_TO_FRONT, "Move to front");
        _move2FrontAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_MOVE_TO_FRONT));
        _moveBackAction = new AcmeStudioRetargetAction (ChangeOrderAction.MOVE_BACK, "Move back");
        _moveBackAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_MOVE_BACK));
        _moveForwardAction = new AcmeStudioRetargetAction (ChangeOrderAction.MOVE_FORWARD, "Move forward");
        _moveForwardAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_MOVE_FORWARD));
    }

    /**
     * @see org.eclipse.gef.ui.actions.ActionBarContributor#declareActions()
     */
    protected void buildActions() {

        addRetargetAction(new AlignmentRetargetAction(PositionConstants.LEFT));
        addRetargetAction(
            new AlignmentRetargetAction(PositionConstants.CENTER));
        addRetargetAction(new AlignmentRetargetAction(PositionConstants.RIGHT));
        addRetargetAction(new AlignmentRetargetAction(PositionConstants.TOP));
        addRetargetAction(
            new AlignmentRetargetAction(PositionConstants.MIDDLE));
        addRetargetAction(
            new AlignmentRetargetAction(PositionConstants.BOTTOM));
 
        addRetargetAction(new UndoRetargetAction());
        addRetargetAction(new RedoRetargetAction());
        addRetargetAction(new DeleteRetargetAction());

        addRetargetAction(new ZoomInRetargetAction());
        addRetargetAction(new ZoomOutRetargetAction());
        
//        addRetargetAction (new RetargetAction (GEFActionConstants.TOGGLE_HORIZONTAL_RULER_VISIBILITY, GEFMessages.ToggleHRulerVisibility_Label));
//        addRetargetAction (new RetargetAction (GEFActionConstants.TOGGLE_VERTICAL_RULER_VISIBILITY, GEFMessages.ToggleVRulerVisibility_Label));

        AcmeStudioRetargetAction asAction =
            new AcmeStudioRetargetAction(
                ActionFactory.COPY.getId(),
                ActionFactory.COPY.getId());
        asAction.setAction(new CopyAcmeElementAction(null));
        addRetargetAction(asAction);

        asAction =
            new AcmeStudioRetargetAction(
                ActionFactory.PASTE.getId(),
                ActionFactory.PASTE.getId());
        asAction.setAction(new PasteAcmeElementAction(null));
        addRetargetAction(asAction);

        asAction =
            new AcmeStudioRetargetAction(
                NavigateUpAction.ID,
                "Up");
        asAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_NAV_UP));
        asAction.setAction(new NavigateUpAction(null));
        addRetargetAction(asAction);
        
        addRetargetAction (new RetargetAction (GEFActionConstants.TOGGLE_RULER_VISIBILITY, GEFMessages.ToggleRulerVisibility_Label, IAction.AS_CHECK_BOX));
        addRetargetAction (new RetargetAction (GEFActionConstants.TOGGLE_SNAP_TO_GEOMETRY, GEFMessages.ToggleSnapToGeometry_Label, IAction.AS_CHECK_BOX));
        addRetargetAction (new RetargetAction (GEFActionConstants.TOGGLE_GRID_VISIBILITY, GEFMessages.ToggleGrid_Label, IAction.AS_CHECK_BOX));
        
    	addRetargetAction(new RetargetAction(GEFActionConstants.MATCH_WIDTH, 
    			GEFMessages.MatchWidthAction_Label));
    	addRetargetAction(new RetargetAction(GEFActionConstants.MATCH_HEIGHT,
    			GEFMessages.MatchHeightAction_Label));
  /*  	addRetargetAction(new RetargetAction(GEFActionConstants.MATCH_SIZE,
    			GEFMessages.MatchSizeAction_Label));*/

    }

    /**
     * @see org.eclipse.ui.IEditorActionBarContributor#setActiveEditor(IEditorPart)
     */
    public void setActiveEditor(IEditorPart targetEditor) {
        if (targetEditor instanceof DiagramEditor) {
        	// Reset all the retarget actions to point to the actions
        	// registered for this diagram
            DiagramEditor editor = (DiagramEditor) targetEditor;
            super.setActiveEditor(targetEditor);

            _newComponentAction.setAction(
                editor.getActionRegistry().getAction(
                    NewElementAction.NEW_COMPONENT));
            _newConnectorAction.setAction(
                editor.getActionRegistry().getAction(
                    NewElementAction.NEW_CONNECTOR));
            _newPortAction.setAction(
                editor.getActionRegistry().getAction(
                    NewElementAction.NEW_PORT));
            _newRoleAction.setAction(
                editor.getActionRegistry().getAction(
                    NewElementAction.NEW_ROLE));
            _newConnectionRuleAction.setAction(
                    editor.getActionRegistry().getAction(
                            NewConnectionPatternAction.NEW_CONNECTION_RULE));
            
            _browsePatternsAction.setAction(
                    editor.getActionRegistry().getAction(
                            BrowsePatternsAction.BROWSE_PATTERNS));
            
            
//            _showGridAction.setAction(
//                editor.getActionRegistry().getAction(
//                    ShowGridAction.ID));
			_zoomToFitAction.setAction (editor.getActionRegistry().getAction(ZoomToFitAction.ID));
			_zoomCombo.setZoomManager(editor.getZoomManager());
            _newRepresentationAction.setAction(
                editor.getActionRegistry().getAction(
                    RepresentationCommand.ADD_REP));
            _removeRepresentationMenu.setBuilder(
                (IMenuBuilder) editor.getActionRegistry().getAction(
                    RepresentationCommand.REMOVE_REP));
            _removeRepresentationMenu.hook(editor);
            _navigateUpRepresentationAction.setAction(
                editor.getActionRegistry().getAction(NavigateUpAction.ID));
			_wrapRepAction.setAction (editor.getActionRegistry().getAction(WrapRepresentationAction.WRAPREP_ID));
            _assignTypeMenu.setBuilder(
                (IMenuBuilder) editor.getActionRegistry().getAction(
                    ResetPortAlignmentCommand.RESET_ALIGNMENT));
            _assignTypeMenu.hook(editor);

            _unassignTypeMenu.setBuilder(
                (IMenuBuilder) editor.getActionRegistry().getAction(
                    TypeAssignmentCommand.UNASSIGN_TYPE));
            _unassignTypeMenu.hook(editor);

            _assignTypeMenu.hook(editor);

            _assignFamilyMenu.setBuilder(
                (IMenuBuilder) editor.getActionRegistry().getAction(
                    TypeAssignmentCommand.ASSIGN_FAMILY));
            _assignFamilyMenu.hook(editor);

            _unassignFamilyMenu.setBuilder(
                (IMenuBuilder) editor.getActionRegistry().getAction(
                    TypeAssignmentCommand.UNASSIGN_FAMILY));
            _unassignFamilyMenu.hook(editor);

            _resetPortAlignmentMenu.setBuilder(
                (IMenuBuilder) editor.getActionRegistry().getAction(
                    ResetPortAlignmentCommand.RESET_ALIGNMENT));
            _adjustPortAlignmentMenu.setBuilder(
                (IMenuBuilder) editor.getActionRegistry().getAction(
                    AdjustPortAlignmentCommand.ADJUST));
            _resetRoleRouteMenu.setBuilder(
                (IMenuBuilder) editor.getActionRegistry().getAction(
                    ResetRoleRouteCommand.RESET_ROUTE));
            _move2BackAction.setAction (editor.getActionRegistry().getAction(_move2BackAction.getId()));
            _move2FrontAction.setAction (editor.getActionRegistry ().getAction (_move2FrontAction.getId ()));
            _moveBackAction.setAction (editor.getActionRegistry().getAction (_moveBackAction.getId ()));
            _moveForwardAction.setAction (editor.getActionRegistry().getAction (_moveForwardAction.getId ()));
            editor.registerKeyBindings();
        } else {
			// Set all the actions to point to null
            _newComponentAction.setAction(null);
            _newConnectorAction.setAction(null);
            _newPortAction.setAction(null);
            _newConnectionRuleAction.setAction(null);
            _browsePatternsAction.setAction(null);
            
//            _showGridAction.setAction(null);
            _zoomToFitAction.setAction (null);
            
            _newRepresentationAction.setAction(null);
            _removeRepresentationMenu.setBuilder(null);
            _removeRepresentationMenu.unhook();

            _navigateUpRepresentationAction.setAction(null);
            _assignTypeMenu.setBuilder(null);
            _assignTypeMenu.unhook();

            _unassignTypeMenu.setBuilder(null);
            _unassignTypeMenu.unhook();

            _assignFamilyMenu.setBuilder(null);
            _assignFamilyMenu.unhook();

            _unassignFamilyMenu.setBuilder(null);
            _unassignFamilyMenu.unhook();

            _resetPortAlignmentMenu.setBuilder(null);
            _resetPortAlignmentMenu.unhook();

            _adjustPortAlignmentMenu.setBuilder(null);
            _adjustPortAlignmentMenu.unhook();

            _resetRoleRouteMenu.setBuilder(null);
            _resetRoleRouteMenu.unhook();
            
            _move2BackAction.setAction (null);
            _move2FrontAction.setAction (null);
            _moveBackAction.setAction (null);
            _moveForwardAction.setAction (null);
        }

    }

    /**
     * @see org.eclipse.ui.part.EditorActionBarContributor#contributeToToolBar(IToolBarManager)
     */
    public void contributeToToolBar(IToolBarManager tbm) {
        if (!(tbm instanceof SubToolBarManager))
            return;
        tbm.add(getAction(NavigateUpAction.ID));
        tbm.add(getAction(ActionFactory.UNDO.getId()));
        tbm.add(getAction(ActionFactory.REDO.getId()));
        if (AcmeStudioPlugin.getDefault().getPreferenceStore().getBoolean (DiagramToolbarPreferences.ALIGNMENT_TOOLBAR)) {
	        contributeAlignmentTools (tbm);
        }

        if (AcmeStudioPlugin.getDefault().getPreferenceStore().getBoolean (DiagramToolbarPreferences.ORDERING_TOOLBAR)) {
        	contributeOrderTools (tbm);
        }
        
        if (AcmeStudioPlugin.getDefault().getPreferenceStore().getBoolean (DiagramToolbarPreferences.ZOOM_TOOLBAR)) {
	        contributeZoomTools (tbm);
        }
        
//        tbm.add(new Separator());        
//        _showGridAction.setImageDescriptor (ImageManager.getImageDescriptor (ImageManager.ID_GRID));
//        _showGridAction.setToolTipText("Shows grid");
//        tbm.add (_showGridAction);

    }

    public void removeZoomTools (IToolBarManager tbm) {
    	tbm.remove("zoom-sep");
    	tbm.remove(GEFActionConstants.ZOOM_IN);
    	tbm.remove(GEFActionConstants.ZOOM_OUT);
    	tbm.remove(_zoomToFitAction.getId());
    	tbm.remove(GEFActionConstants.ZOOM_TOOLBAR_WIDGET);
    }
    
    /**
	 * @param tbm
	 */
	public void contributeZoomTools(IToolBarManager tbm) {
		tbm.add(new Separator("zoom-sep"));
		tbm.add(getAction(GEFActionConstants.ZOOM_IN));
		tbm.add(getAction(GEFActionConstants.ZOOM_OUT));
		tbm.add(_zoomToFitAction);
		try {
			if (_zoomCombo == null) _zoomCombo = new ZoomComboContributionItem (getPage());
			tbm.add(_zoomCombo);
			_zoomCombo.setZoomManager(((MultiPageAcmeEditor )getPage().getActiveEditor()).getDiagramEditor().getZoomManager());
		}
		catch (Exception e) {}
	}

	/**
	 * @param tbm
	 */
	public void contributeOrderTools(IToolBarManager tbm) {
		tbm.add (new Separator ("order-sep"));
		tbm.add(_moveBackAction);
		tbm.add (_moveForwardAction);
		tbm.add (_move2BackAction);
		tbm.add (_move2FrontAction);
	}
	
	public void removeOrderTools (IToolBarManager tbm) {
		tbm.remove("order-sep");
		tbm.remove(_moveBackAction.getId ());
		tbm.remove (_moveForwardAction.getId ());
		tbm.remove (_move2BackAction.getId ());
		tbm.remove (_move2FrontAction.getId ());
	}

	/**
	 * @param tbm
	 */
	public void contributeAlignmentTools(IToolBarManager tbm) {
		tbm.add(new Separator("align-sep"));
		tbm.add(getAction(GEFActionConstants.ALIGN_LEFT));
		tbm.add(getAction(GEFActionConstants.ALIGN_CENTER));
		tbm.add(getAction(GEFActionConstants.ALIGN_RIGHT));
		tbm.add(getAction(GEFActionConstants.ALIGN_TOP));
		tbm.add(getAction(GEFActionConstants.ALIGN_MIDDLE));
		tbm.add(getAction(GEFActionConstants.ALIGN_BOTTOM));
	}
	
	public void removeAlignmentTools (IToolBarManager tbm) {
		tbm.remove ("align-sep");
		tbm.remove (GEFActionConstants.ALIGN_LEFT);
		tbm.remove (GEFActionConstants.ALIGN_CENTER);
		tbm.remove (GEFActionConstants.ALIGN_BOTTOM);
		tbm.remove (GEFActionConstants.ALIGN_MIDDLE);
		tbm.remove (GEFActionConstants.ALIGN_RIGHT);
		tbm.remove (GEFActionConstants.ALIGN_TOP);
	}

	/**
     * @see org.eclipse.ui.IEditorActionBarContributor#init(IActionBars, IWorkbenchPage)
     */
    public void init(IActionBars bars, IWorkbenchPage page) {
        super.init(bars, page);

    }

    /* (non-Javadoc)
     * @see org.eclipse.gef.ui.actions.ActionBarContributor#declareGlobalActionKeys()
     */
    protected void declareGlobalActionKeys() {
        addGlobalActionKey(ActionFactory.PRINT.getId ());
        addGlobalActionKey(ActionFactory.COPY.getId ());
        addGlobalActionKey(ActionFactory.CUT.getId ());
        addGlobalActionKey(ActionFactory.PASTE.getId ());
        addGlobalActionKey(ActionFactory.UNDO.getId ());
        addGlobalActionKey(ActionFactory.REDO.getId ());
        addGlobalActionKey(ActionFactory.SELECT_ALL.getId ());
    }
    /* (non-Javadoc)
     * @see org.eclipse.ui.part.EditorActionBarContributor#contributeToMenu(org.eclipse.jface.action.IMenuManager)
     */
    public void contributeToMenu(IMenuManager menuManager) {
        super.contributeToMenu(menuManager);
        IMenuManager viewMenu = getActionBars ().getMenuManager ().findMenuUsingPath("AcmeStudio.view");
        List actionIds = getContributionIds (viewMenu);
        if (!actionIds.contains(GEFActionConstants.ZOOM_IN))
            viewMenu.appendToGroup("zooming",getAction(GEFActionConstants.ZOOM_IN));
        if (!actionIds.contains (GEFActionConstants.ZOOM_OUT))
            viewMenu.appendToGroup("zooming", getAction(GEFActionConstants.ZOOM_OUT));
        if (!actionIds.contains (GEFActionConstants.TOGGLE_RULER_VISIBILITY))
            viewMenu.appendToGroup ("options", getAction (GEFActionConstants.TOGGLE_RULER_VISIBILITY));
        if (!actionIds.contains (GEFActionConstants.TOGGLE_GRID_VISIBILITY)) 
            viewMenu.appendToGroup ("options", getAction (GEFActionConstants.TOGGLE_GRID_VISIBILITY));
        if (!actionIds.contains (GEFActionConstants.TOGGLE_SNAP_TO_GEOMETRY))
            viewMenu.appendToGroup ("options", getAction (GEFActionConstants.TOGGLE_SNAP_TO_GEOMETRY));
 /*       if (!actionIds.contains (GEFActionConstants.MATCH_SIZE)) {
        	viewMenu.appendToGroup ("options", new Separator ());
        	viewMenu.appendToGroup ("options", getAction (GEFActionConstants.MATCH_SIZE));
        }*/
        if (!actionIds.contains (GEFActionConstants.MATCH_WIDTH)) {
        	viewMenu.appendToGroup ("options", getAction (GEFActionConstants.MATCH_WIDTH));
        }
        if (!actionIds.contains (GEFActionConstants.MATCH_HEIGHT)) {
        	viewMenu.appendToGroup ("options", getAction (GEFActionConstants.MATCH_HEIGHT));
        }
        viewMenu.appendToGroup("options", new Separator ());
        buildOrderMenu (viewMenu);
        buildAlignMenu (viewMenu);
        viewMenu.appendToGroup ("options", new Separator ());
        if (!actionIds.contains (_resetSystemLayoutMenu.getId()))
            viewMenu.appendToGroup ("visualizations", _resetSystemLayoutMenu);
        if (!actionIds.contains (_resetPortAlignmentMenu.getId ()))
            viewMenu.appendToGroup ("visualizations", _resetPortAlignmentMenu);
        if (!actionIds.contains (_adjustPortAlignmentMenu.getId ()))
            viewMenu.appendToGroup ("visualizations", _adjustPortAlignmentMenu);
        if (!actionIds.contains (_resetRoleRouteMenu.getId ()))
            viewMenu.appendToGroup ("visualizations", _resetRoleRouteMenu);
        
        IMenuManager designMenu =
            getActionBars().getMenuManager().findMenuUsingPath(
                "AcmeStudio.design");
        
        actionIds = getContributionIds(designMenu);

        if (!actionIds.contains (_newComponentAction.getId ()))
            designMenu.appendToGroup("elements", _newComponentAction);
        if (!actionIds.contains (_newConnectorAction.getId ()))
            designMenu.appendToGroup("elements", _newConnectorAction);
        if (!actionIds.contains (_newPortAction.getId ()))
            designMenu.appendToGroup("elements", _newPortAction);
        if (!actionIds.contains (_newRoleAction.getId ()))
            designMenu.appendToGroup("elements", _newRoleAction);
        if (!actionIds.contains (_newConnectionRuleAction.getId()))
            designMenu.appendToGroup("elements", _newConnectionRuleAction);
        if (!actionIds.contains (_browsePatternsAction.getId ())) 
        	designMenu.appendToGroup ("elements", _browsePatternsAction);
        if (!actionIds.contains (_newRepresentationAction.getId ()))
            designMenu.appendToGroup("representations", _newRepresentationAction);
        if (!actionIds.contains (_removeRepresentationMenu.getId ()))
            designMenu.appendToGroup("representations", _removeRepresentationMenu);
        if (!actionIds.contains (_wrapRepAction.getId ()))
            designMenu.appendToGroup ("representations", _wrapRepAction);
        if (!actionIds.contains (_assignTypeMenu.getId ()))
            designMenu.appendToGroup("types", _assignTypeMenu);
        if (!actionIds.contains (_unassignTypeMenu.getId ()))
            designMenu.appendToGroup("types", _unassignTypeMenu);
        _removeRepresentationMenu.setMenu(designMenu, "representations");
        _assignTypeMenu.setMenu(designMenu, "types");
        _unassignTypeMenu.setMenu(designMenu, "types");
        _resetSystemLayoutMenu.setMenu(viewMenu, "visualizations");
        _resetPortAlignmentMenu.setMenu(viewMenu, "visualizations");
        _adjustPortAlignmentMenu.setMenu(viewMenu, "visualizations");
        _resetRoleRouteMenu.setMenu(viewMenu, "visualizations");
        
        IMenuManager familyMenu =
            getActionBars().getMenuManager().findMenuUsingPath(
                "AcmeStudio.family");
        actionIds = getContributionIds(familyMenu);
        if (!actionIds.contains (_assignFamilyMenu.getId ()))
            familyMenu.appendToGroup("assignment", _assignFamilyMenu);
        if (!actionIds.contains (_unassignFamilyMenu.getId ()))
            familyMenu.appendToGroup("assignment", _unassignFamilyMenu);
        _assignFamilyMenu.setMenu(familyMenu, "assignment");
        _unassignFamilyMenu.setMenu(familyMenu, "assignment");

        //        menuManager.insertBefore (IWorkbenchActionConstants.M_PROJECT, familyMenu);
    }

    /**
	 * @param viewMenu
	 */
	private void buildAlignMenu(IMenuManager viewMenu) {
		IMenuManager orderMenu = viewMenu.findMenuUsingPath("align");
        if (orderMenu == null) {
        	orderMenu = new MenuManager ("Align", "align");
        	orderMenu.add(new GroupMarker ("options"));
        	viewMenu.appendToGroup("options", orderMenu);
        }
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", getAction (GEFActionConstants.ALIGN_LEFT));
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", getAction (GEFActionConstants.ALIGN_CENTER));
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", getAction (GEFActionConstants.ALIGN_RIGHT));
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", getAction (GEFActionConstants.ALIGN_TOP));
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", getAction (GEFActionConstants.ALIGN_MIDDLE));
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", getAction (GEFActionConstants.ALIGN_BOTTOM));
	}

	/**
	 * @param viewMenu
	 */
	private void buildOrderMenu(IMenuManager viewMenu) {
		IMenuManager orderMenu = viewMenu.findMenuUsingPath("order");
        if (orderMenu == null) {
        	orderMenu = new MenuManager ("Order", "order");
        	orderMenu.add(new GroupMarker ("options"));
        	viewMenu.appendToGroup("options", orderMenu);
        }
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", _move2FrontAction);
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", _move2BackAction);
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", _moveForwardAction);
        appendToGroup (getContributionIds(orderMenu), orderMenu, "options", _moveBackAction);
	}

	/**
	 * @param actionIds
     * @param menu
     * @param group
	 * @param action
	 */
	private void appendToGroup(List actionIds, IMenuManager menu, String group, IAction action) {
		if (!actionIds.contains (action.getId())) menu.appendToGroup (group, action);
	}

	/**
     * @param viewMenu
     * @return
     */
    private List getContributionIds(IMenuManager viewMenu) {
        IContributionItem [] items = viewMenu.getItems();
        List l = new ArrayList ();
        for (int i = 0; i < items.length; i++) {
            l.add(items [i].getId());
        }
        return l;
    }


}
