package AcmeStudio.ui.editors.diagram.figures;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import org.eclipse.draw2d.ColorConstants;
import org.eclipse.draw2d.Graphics;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.SWTGraphics;
import org.eclipse.draw2d.Shape;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.gef.EditPart;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.widgets.Display;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASModel;
import AcmeStudio.ui.editors.diagram.styles.ElementVisualization;
import AcmeStudio.ui.editors.diagram.styles.VisualStyleRegistry;
import AcmeStudio.util.DesignEvaluationCache;
import AcmeStudio.util.Helper;
import AcmeStudio.util.ImageManager;
import acmeLib.elements.AcmeElement;

/**
 * @author Bradley Schmerl (schmerl@cs.cmu.edu)
 *
 * Root class for the of figures for Acme elements. Manages visualization
 * management and error views.
 */
public abstract class AcmeVisElement
	extends Shape
	implements IAcmeElementFigure, PropertyChangeListener{

    protected static final Point ORIGIN = new Point (0,0);
	/** The visualization for this element**/
	protected ElementVisualization _vis;

	/** The evaluation cache that is used for getting errors **/
	protected DesignEvaluationCache _evalCache;

	/** The model being depicted **/
	protected AcmeElement _model;

	/** The edit part for this figure **/
	protected EditPart _editPart;

	Dimension origSize = null;

	/**
	 * Returns the visualization of this figure
	 * @return The visualization for this figure
	 */
	public ElementVisualization getVisualization() {
		return _vis;
	}

	/**
	 * Resets the visualization, by relooking it up and changing the
	 * registration of change to the new visualization.
	 */
	public void resetVisualization() {
		ElementVisualization newVis =
			VisualStyleRegistry.getVisualization(_model);
		if (newVis != _vis) {
			if (_vis != null) _vis.removePropertyChangeListener(_model.getQualifiedName(), this);
			_vis = newVis;
			_vis.addPropertyChangeListener(_model.getQualifiedName(), this);
			if (_editPart != null)
				_editPart.refresh();
			this.repaint();
		}
	}

	/**
	 * Draws the filled part of the shape, including the error
	 * @param graphics The graphics object to use for the figure
	 * @see org.eclipse.draw2d.Shape#fillShape(Graphics)
	 */
	protected void fillShape(Graphics graphics) {
		Point location = getLocation();
		Dimension size = Helper.getSize(_model);
        if (_transparent) location = ORIGIN.getCopy ();
		_vis.getVariantToApply(_model).fillShape(
			graphics,
			_model,
			location,
			size);
		if (hasErrors
			&& AcmeStudioPlugin.getDefault().getPreferenceStore().getBoolean(
				"DiagramEditor.DisplayElementError")) {
			_vis.drawError(graphics, _model, location);
		}

		if (_model.getRepresentations().getSize() > 0 && AcmeStudioPlugin.getDefault().getPreferenceStore().getBoolean("DiagramEditor.DisplayRepresentationIcon")) {
			Point l = location.getCopy ();
			l.x += getSize().width - 18;
			l.y += 2;
			graphics.drawImage (ImageManager.getIcon(ImageManager.ID_HASREP), l);
		}
	}



	/**
	 * Draws the outline part of the shape
	 * @param graphics The graphics object to use for the figure
	 * @see org.eclipse.draw2d.Shape#outlineShape(Graphics)
	 */
	protected void outlineShape(Graphics graphics) {
       
		Point location = getLocation();
		Dimension size = Helper.getSize (_model);
        if (_transparent) location = ORIGIN.getCopy ();
        _vis.getVariantToApply(_model).outlineShape(
			graphics,
			_model,
			location,
			size);
	}

	/**
	 * Sets the model for this figure. The model is used to find the
	 * correct visualization, the evaluation cache for finding out about
	 * errors and the size of the figure.
	 * @param model The model that this figure is to depict
	 * @see AcmeStudio.ui.editors.diagram.figures.ElementView#setModel(AcmeElement)
	 */
	public void setModel(AcmeElement model) {
		_model = model;

		// Find the visualization and register with it
		_vis = VisualStyleRegistry.getVisualization(model);
		if (_vis != null) _vis.addPropertyChangeListener(model.getQualifiedName(), this);

		// Find the size
		Dimension size = Helper.getSize(_model);
		if (size == null) {
			size = this.getSize();
		}
		setSize(size);
		adjustSizeToHoldErrorFigure();
		
		// Find the eval cache, to find out about errors
		_evalCache = ASModel.current().getRuleManager();
		_evalCache.addPropertyChangeListener(this);
		int error = _evalCache.typecheck(model);
		setErrorFigure(error == DesignEvaluationCache.FALSE);
	}

	/**
	 * Sets the edit part that is the controller for this figure
	 * @param editPart The editPart
	 */
	public void setEditPart(EditPart editPart) {
		_editPart = editPart;
	}

	/**
	 * @return The model being managed by this figure
	 * @see AcmeStudio.ui.editors.diagram.figures.ElementView#getModel()
	 */
	public AcmeElement getModel() {
		return _model;
	}

	/**
	 * Is called when a property changes, either to do with the visualization
	 * or to do with errors.
	 * @param evt The PropertyChangeEvent
	 * @see java.beans.PropertyChangeListener#propertyChange(PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		// Has the visualization changed?
		if (evt
			.getPropertyName()
			.equals(ElementVisualization.PROP_VIS_CHANGED)) {
			// It has changed, and the visualization is part of the event
			_vis = (ElementVisualization) evt.getNewValue();
			if (_editPart != null) {
				_editPart.refresh();
			}
			this.repaint();
		} else if (
			evt.getPropertyName().equals(
				ElementVisualization.PROP_VIS_RESET)) {
			// Completely reset the visualization
			resetVisualization();
		} else if (
			evt.getPropertyName().equals(
				DesignEvaluationCache.PROP_ELEMENT_TYPECHECK)
				&& evt.getNewValue() == _model) {
			// A typecheck result changed.
			setErrorFigure(evt.getOldValue() != null);
		}

	}

	/**
	 * Set up visualization listening
	 * @see org.eclipse.draw2d.IFigure#addNotify()
	 */
	public void addNotify() {
		super.addNotify();
		if (_model != null && _vis == null) {
			_vis = VisualStyleRegistry.getVisualization(_model);
		}
		if (_vis != null) {
			_vis.addPropertyChangeListener(_model.getQualifiedName(), this);
		}
	}

	/** 
	 * Remove myself from listening to visualizations
	 * @see org.eclipse.draw2d.IFigure#removeNotify()
	 */
	public void removeNotify() {
		super.removeNotify();
		if (_vis != null) {
			_vis.removePropertyChangeListener(_model.getQualifiedName(), this);
			_vis = null;
		}
	}

	/**
	 * Works out if there is an error, and sets the flag
	 */
	protected void setErrorFigure(boolean error) {
		if (_evalCache == null)
			return;

		// Got notified of a type check, so find out what the result was
		// (it should be cached).
		int typechecks = _evalCache.typecheck(_model);
		if ((DesignEvaluationCache.FALSE == typechecks) != hasErrors) {
			hasErrors = DesignEvaluationCache.FALSE == typechecks;
			adjustSizeToHoldErrorFigure();
			IFigure parent = getParent();
			if (parent != null) {
				parent.repaint();
			} else {
				repaint();
			}
		}

	}

	public void adjustSizeToHoldErrorFigure() {
		if (hasErrors) {
			// Make the figure's size big enough to fit the error
			origSize = this.getSize();
			Dimension oldD = null;
			if (this.getSize().height < 22) {
				oldD = this.getSize().getCopy();
				oldD.height = 22;
			}
			if (this.getSize().width < 22) {
				if (oldD == null) {
					oldD = this.getSize().getCopy();
				}
				oldD.width = 22;
			}
			if (oldD != null)
				setSize(oldD);
		} else if (origSize != null) {
			setSize(origSize);
			origSize = null;
		}
	}

	/**
	 * Indicates whether there are typechecking errors
	 */
	private boolean hasErrors = false;

    private boolean _transparent;

    private Image _transparentImage;
    private int _transparencyVal;
	private boolean	_inlineRep;

	/**
	 * Returns whether errors exist
	 * @return whether errors exist
	 */
	protected boolean hasError() {
		return hasErrors;
	}

    /* (non-Javadoc)
     * @see org.eclipse.draw2d.IFigure#paint(org.eclipse.draw2d.Graphics)
     */
    public void paint(Graphics graphics) {
        if (_transparent) {
            if (_transparentImage != null && !_transparentImage.isDisposed ()) _transparentImage.dispose ();
 
            PaletteData pData = new PaletteData (0xFF, 0xFF00, 0xFF0000);
            ImageData iData = new ImageData (getSize ().width,  getSize ().height, 24, pData);
            iData.alpha = _transparencyVal;          
            Display display = Display.getCurrent ();
            _transparentImage = new Image (display, iData);
            GC gc = new GC (_transparentImage);
            Graphics g = new SWTGraphics (gc);
            g.pushState ();
            g.setBackgroundColor(ColorConstants.white);
            g.fillRectangle(0,0,iData.width, iData.height);
            g.popState ();
            super.paint (g);
            g.dispose ();
            gc.dispose();
            graphics.drawImage(_transparentImage, getLocation ().x, getLocation ().y);       
       
        }
       else {
           if (_transparentImage != null && _transparentImage.isDisposed ()) _transparentImage.dispose ();
           _transparentImage = null;
           if (!_inlineRep) {
           		super.paint(graphics);
           }
       }
    }


    
    /**
     * @param transparent The transparent to set.
     */
    public void setTransparent(int transparency) {
        _transparent = (transparency == 100?false:true);
        _transparencyVal = transparency;
        this.repaint ();
//        this.repaint();
    }

    /**
     * @return Returns the transparent.
     */
    public boolean isTransparent() {
        return _transparent;
    }

    /**
     * @return
     */
    public int getTransparency() {
       return _transparencyVal;
    }

	/**
	 * @param b
	 */
	public void setInlineRep(boolean inline) {
		_inlineRep = inline;
	}

}
