package AcmeStudio.ui.editors.diagram.figures;

import java.util.List;

import org.eclipse.draw2d.*;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;

import AcmeStudio.util.Helper;
import AcmeStudio.util.ImageManager;
import acmeLib.AcmeObject;
import acmeLib.elements.AcmeComponent;
import acmeLib.representations.AcmeRepresentation;

/**
 * @author Hong Yan (yh@cs.cmu.edu)
 *
 * Provides the figure for AcmeSystem.
 * 
 */
public class RepresentationView extends Figure implements UpdateListener {

    private SystemView sourceFigure;
    private boolean isDirty;
    private float scaleX;
    private float scaleY;
    private Dimension targetSize = new Dimension(0, 0);
    private Image thumbnailImage;
    private Image overview;
    private Dimension thumbnailImageSize;
    private ThumbnailUpdater updater = new ThumbnailUpdater();

    class ThumbnailUpdater implements Runnable {
        protected int MAX_BUFFER_SIZE = 256;

        private boolean isRunning = false;
        private boolean isActive = true;
        private int hTiles, vTiles;
        private Dimension tileSize;
        private Image sourceBuffer;
        private Dimension sourceBufferSize;
        private GC sourceGC, thumbnailGC;
        private Graphics sourceGraphics;
        private int currentHTile, currentVTile;

        public void deactivate() {
            setActive(false);
            stop();
            if (thumbnailImage != null) {
                ImageManager.imageDisposed("RepresentationView.deactivate",thumbnailImage);
                thumbnailImage.dispose();
                thumbnailImage = null;
                thumbnailImageSize = null;
            }
        }

        protected int getCurrentHTile() {
            return currentHTile;
        }

        protected int getCurrentVTile() {
            return currentVTile;
        }

        public boolean isActive() {
            return isActive;
        }

        public boolean isRunning() {
            return isRunning;
        }

        public void resetTileValues() {
            hTiles =
                (int) Math.ceil(
                    (float) sourceFigure.getSize().width
                        / (float) MAX_BUFFER_SIZE);
            vTiles =
                (int) Math.ceil(
                    (float) sourceFigure.getSize().height
                        / (float) MAX_BUFFER_SIZE);

            tileSize =
                new Dimension(
                    (int) Math.ceil(
                        (float) sourceFigure.getSize().width / (float) hTiles),
                    (int) Math.ceil(
                        (float) sourceFigure.getSize().height
                            / (float) vTiles));

            currentHTile = 0;
            currentVTile = 0;
        }

        public void restart() {
            stop();
            start();
        }

        public void run() {
            if (!isActive())
                return;
            int v = getCurrentVTile();
            int sy1 = v * tileSize.height;
            int sy2 =
                Math.min(
                    (v + 1) * tileSize.height,
                    sourceFigure.getSize().height);

            int ty1 = Math.round(sy1 * getScaleY());
            int ty2 =
                Math.min(
                    Math.round(sy2 * getScaleY()),
                    targetSize.height);

            int h = getCurrentHTile();
            int sx1 = h * tileSize.width;
            int sx2 =
                Math.min(
                    (h + 1) * tileSize.width,
                    sourceFigure.getSize().width);
            int tx1 = Math.round(sx1 * getScaleX());
            int tx2 =
                Math.min(Math.round(sx2 * getScaleX()), targetSize.width);
            org.eclipse.draw2d.geometry.Point p =
                sourceFigure.getBounds().getLocation();
            Rectangle rect =
                new Rectangle(sx1 + p.x, sy1 + p.y, sx2 - sx1, sy2 - sy1);
            sourceGraphics.pushState();
            sourceGraphics.translate(-sx1 - p.x, -sy1 - p.y);
            sourceGraphics.setClip(rect);
            sourceGraphics.fillRectangle(rect);
            sourceFigure.paintGraphics(sourceGraphics);
            if ((tx2 - tx1 > 0) && (ty2 - ty1 > 0))
                thumbnailGC.drawImage(
                    sourceBuffer,
                    0,
                    0,
                    sx2 - sx1,
                    sy2 - sy1,
                    tx1,
                    ty1,
                    tx2 - tx1,
                    ty2 - ty1);
            sourceGraphics.popState();
            repaint();

            if (getCurrentHTile() < (hTiles - 1))
                setCurrentHTile(getCurrentHTile() + 1);
            else {
                setCurrentHTile(0);
                if (getCurrentVTile() < (vTiles - 1))
                    setCurrentVTile(getCurrentVTile() + 1);
                else
                    setCurrentVTile(0);
            }

            if (getCurrentHTile() != 0 || getCurrentVTile() != 0)
                Display.getCurrent().timerExec(10, this);
            else if (isDirty()) {
                setDirty(false);
                Display.getCurrent().timerExec(10, this);
            } else
                stop();
        }

        public void setActive(boolean value) {
            isActive = value;
        }

        protected void setCurrentHTile(int count) {
            currentHTile = count;
        }

        protected void setCurrentVTile(int count) {
            currentVTile = count;
        }

        public void start() {
            if (!isActive())
                return;
            isRunning = true;
            setDirty(false);
            resetTileValues();

            if (!tileSize.equals(sourceBufferSize)) {
                if (sourceBuffer != null) {
                 
                    ImageManager.imageDisposed("RepresentationView.ThumbnailUpdater.start.sourceBuffer", sourceBuffer);
                    sourceBuffer.dispose();
                }
                sourceBuffer =
                    new Image(
                        Display.getDefault(),
                        tileSize.width,
                        tileSize.height);
                ImageManager.imageCreated("RepresentationView.ThumbnailUpdater.start.sourceBuffer", sourceBuffer);
                sourceBufferSize = new Dimension(tileSize);
            }
            if (sourceGC != null) sourceGC.dispose ();
            if (sourceGraphics != null) sourceGraphics.dispose ();
            sourceGC = new GC(sourceBuffer);
            sourceGraphics = new SWTGraphics(sourceGC);
            Color color = sourceFigure.getForegroundColor();
            if (color != null)
                sourceGraphics.setForegroundColor(color);
            color = sourceFigure.getBackgroundColor();
            if (color != null)
                sourceGraphics.setBackgroundColor(color);
            sourceGraphics.setFont(sourceFigure.getFont());

            setScales(
                targetSize.width / (float) sourceFigure.getSize().width,
                targetSize.height / (float) sourceFigure.getSize().height);

            if (!targetSize.equals(thumbnailImageSize)) {
                if (thumbnailImage != null) {
                    ImageManager.imageDisposed("RepresentationView.ThumbnailUpdater.start.thumbnailImage", thumbnailImage);
                    thumbnailImage.dispose();
                }
                thumbnailImage =
                    new Image(
                        Display.getDefault(),
                        targetSize.width,
                        targetSize.height);
                ImageManager.imageCreated ("RepresentationView.ThumbnailUpdater.start.thumbnailImage", thumbnailImage);
                thumbnailImageSize = new Dimension(targetSize);
            }
            if (thumbnailGC != null) {
                thumbnailGC.dispose ();
            }
            thumbnailGC = new GC(thumbnailImage);
            Display.getCurrent().asyncExec(this);
        }

        public void stop() {
            if (sourceGC != null) {
                sourceGC.dispose();
                sourceGC = null;
            }
            if (thumbnailGC != null) {
                thumbnailGC.dispose();
                thumbnailGC = null;
            }
            sourceGraphics = null;
            if (sourceBuffer != null) {
                ImageManager.imageDisposed("RepresentationView.ThumbnailUpdater.stop.sourceBuffer", sourceBuffer);
                sourceBuffer.dispose();
                sourceBuffer = null;
            }
            sourceBufferSize = null;
            // Don't dispose of the thumbnail image since it is needed to paint the 
            // figure when the source is not dirty (i.e. showing/hiding the dock).
            isRunning = false;
        }
    }

    private Dimension adjustToAspectRatio(
        Dimension size,
        boolean adjustToMaxDimension) {
        Dimension sourceSize = sourceFigure.getSize();
        Dimension borderSize =
            new Dimension(getInsets().getWidth(), getInsets().getHeight());
        size.expand(borderSize.getNegated());
        int width, height;
        if (adjustToMaxDimension) {
            width =
                Math.max(
                    size.width,
                    (int) (size.height
                        * sourceSize.width
                        / (float) sourceSize.height
                        + 0.5));
            height =
                Math.max(
                    size.height,
                    (int) (size.width
                        * sourceSize.height
                        / (float) sourceSize.width
                        + 0.5));
        } else {
            width =
                Math.min(
                    size.width,
                    (int) (size.height
                        * sourceSize.width
                        / (float) sourceSize.height
                        + 0.5));
            height =
                Math.min(
                    size.height,
                    (int) (size.width
                        * sourceSize.height
                        / (float) sourceSize.width
                        + 0.5));
        }
        size.width = width;
        size.height = height;
        return size.expand(borderSize);
    }

    /**
     * Deactivates this Thumbnail.
     */
    public void deactivate() {
        updater.deactivate();
        sourceFigure.getUpdateManager().removeUpdateListener(this);
    }

    /**
     * Returns the preferred size of this Thumbnail.  The preferred size will be 
     * calculated in a way that maintains the source Figure's aspect ratio.
     * 
     * @param wHint The width hint
     * @param hHint The height hint
     * @return The preferred size
     */
    public Dimension getPreferredSize(int wHint, int hHint) {
        if (prefSize == null)
            return adjustToAspectRatio(getBounds().getSize(), false);

        Dimension preferredSize = adjustToAspectRatio(prefSize.getCopy(), true);

        if (maxSize == null)
            return preferredSize;

        Dimension maximumSize = adjustToAspectRatio(maxSize.getCopy(), true);
        if (preferredSize.contains(maximumSize))
            return maximumSize;
        else
            return preferredSize;
    }

    public void setImage(Image img)
    {
        this.overview = img;
    }
    
    /**
     * Returns the scale factor on the X-axis.
     * @return X scale
     */
    protected float getScaleX() {
        return scaleX;
    }

    /**
     * Returns the scale factor on the Y-axis.
     * @return Y scale
     */
    protected float getScaleY() {
        return scaleY;
    }

    /**
     * Returns the scaled Image of the source Figure.  If the Image needs to be 
     * updated, the ThumbnailUpdater will notified.
     * 
     * @return The thumbnail image
     */
    public Image getThumbnailImage() {
        Dimension oldSize = targetSize;
        targetSize = getPreferredSize();
        targetSize.expand(
            new Dimension(getInsets().getWidth(), getInsets().getHeight())
                .negate());

        if ((isDirty()) && !updater.isRunning())
            updater.start();
        else if (oldSize != null && !targetSize.equals(oldSize)) {
            revalidate();
            updater.restart();
        }

        return thumbnailImage;
    }

    /**
     * Returns <code>true</code> if the source figure has changed.
     * 
     * @return <code>true</code> if the source figure has changed
     */
    protected boolean isDirty() {
        return isDirty;
    }

    /**
     * @see org.eclipse.draw2d.UpdateListener#notifyPainting(Rectangle)
     */
    public void notifyPainting(Rectangle damage) {
        setDirty(true);
        repaint();
    }

    /**
     * @see org.eclipse.draw2d.UpdateListener#notifyValidating()
     */
    public void notifyValidating() {
        setDirty(true);
        revalidate();
    }

    /**
     * @see org.eclipse.draw2d.Figure#paintFigure(Graphics)
     */
    protected void paintFigure(Graphics graphics) {
        Image thumbnail = getThumbnailImage();
        if (thumbnail == null)
            return;
        Rectangle r = getClientArea();
        graphics.drawImage(thumbnail, r.getLocation());
        graphics.drawImage(overview, r.getLocation().translate(0, -220));        
        /*Point rightTop = r.getTopRight().translate(-5, 0);
        Point rightBottom = r.getBottomRight().translate(-5, -10);
        Point slotRight = r.getBottomLeft().translate(30, -10);
        Point slotLeft = r.getBottomLeft().translate(20, -10);
        Point slotBottom = r.getBottomLeft().translate(25, 0);
        Point leftBottom = r.getBottomLeft().translate(0, -10);
        graphics.drawLine(r.getTopLeft(), rightTop);
        graphics.drawLine(rightTop, rightBottom);
        graphics.drawLine(rightBottom, slotRight);
        graphics.drawLine(slotRight, slotBottom);
        graphics.drawLine(slotBottom, slotLeft);
        graphics.drawLine(slotLeft, leftBottom);
        graphics.drawLine(leftBottom, r.getTopLeft());*/
    }

    public void setDirty(boolean value) {
        isDirty = value;
    }

    protected void setScales(float x, float y) {
        scaleX = x;
        scaleY = y;
    }

    public void setSource(IFigure fig) {
        if (sourceFigure == fig)
            return;
        if (sourceFigure != null)
            sourceFigure.getUpdateManager().removeUpdateListener(this);
        sourceFigure = (SystemView)fig;
        if (sourceFigure != null) {
            setScales(
                (float) getSize().width / (float) sourceFigure.getSize().width,
                (float) getSize().height
                    / (float) sourceFigure.getSize().height);
            sourceFigure.getUpdateManager().addUpdateListener(this);
            repaint();
        }
    }

    /**
     * The AcmeSystem represented by this figure
     */
    private AcmeRepresentation _model = null;

    /**
     * The representation border
     */
    private IFigure _representationBorder;

    /**
     * The scale of the figure
     */
    private double _scale = 1.0;

    /**
     * Constructor
     */
    public RepresentationView() {
        //setVisible(false);
        //setSize(200, 100);
        //setLocation(new Point(200, 200));
    }

    /**
     * Set the model represented by this figure
     * @param model The model to be represented
     */
    public void setModel(AcmeObject model) {
        this._model = (AcmeRepresentation) model;
        AcmeComponent acmeComponent =
            (AcmeComponent) ((AcmeRepresentation) model).getParent();
        Point location = SystemView.getLocation(acmeComponent, true);
        Dimension size = Helper.getSize(acmeComponent);
        setSize(size.width * 3, size.height * 3);
        setLocation(location.translate(-size.width / 2, -size.height * 7 / 2));
    }

    /**
     * Get the model this figure is representing
     * @return The AcmeSystem model
     */
    public AcmeObject getModel() {
        return this._model;
    }

    /**
     * Returns the scale.
     * @return double
     */
    public double getScale() {
        return _scale;
    }

    /**
     * Sets the scale.
     * @param scale The scale to set
     */
    public void setScale(double scale) {
        this._scale = scale;
    }

    public RepresentationBorderView getComponentRepresentationBorder() {
        return (RepresentationBorderView) _representationBorder;
    }

    public void setBindingPoints(List bp) {
    }

    public void paint(Graphics graphics) {
        super.paint(graphics);
    }

    /**
     * @see org.eclipse.draw2d.UpdateListener#notifyValidating()
     */
    public void notifyPainting(Rectangle damage, java.util.Map dirtyRegions) { }

}