package AcmeStudio.ui.editors.diagram.figures;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import org.eclipse.draw2d.*;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.PointList;
import org.eclipse.swt.SWT;

import AcmeStudio.ui.editors.diagram.styles.ElementVisualization;
import AcmeStudio.ui.editors.diagram.styles.VisualStyleRegistry;
import acmeLib.elements.AcmeConnector;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeRole;

/**
 * @author Hong Yan (yh@cs.cmu.edu)
 *
 * The visual representation of AcmeRole.
 * 
 */
public class RoleView extends RoundedJumpingRoleConnection /*PolylineConnection*/ implements IAcmeElementFigure, PropertyChangeListener {

	private ElementVisualization _vis;
    
    private boolean _transparent = false;

	/**
	 * The AcmeRole described by this figure
	 */
	private AcmeRole _model = null;

	/**
	 * The canvas this figure is drawn on
	 */
	private FigureCanvas _figureCanvas = null;

	/**
	 * The decoration of the role figure
	 */
	private RoleDecoration _roleDecoration;
	
	/**
	 * The distance from the selection point to the end of the role
	 */
	private double _selectionDistance = -1.0;
	private PointList _bendpoints = null;
	public Point _endBeforeMoving = null;

	public boolean _isBendpointMoving = false;
	public boolean _isBendpointBeingCreated = false;

	/**
	 * Constructor
	 */
	public RoleView() {
		super();
	}

	/**
	 * Get the AcmeRole represented by this figure
	 * @return The AcmeRole
	 */
	public AcmeElement getModel() {
		return this._model;
	}

	/**
	 * Set the AcmeRole to be represented by this figure
	 * @param model The AcmeRole
	 */
	public void setModel(AcmeElement model) {
		this._model = (AcmeRole) model;
		_vis = VisualStyleRegistry.getVisualization(model);
		_vis.addPropertyChangeListener(model.getQualifiedName() + "r", this);
		this.setForegroundColor(getVisualization().getFillColor());
		scale();

	}

	/**
	 * Set the canvas this figure is drawn on
	 * @param figureCanvas The figure canvas
	 */
	public void setFigureCanvas(FigureCanvas figureCanvas) {
		this._figureCanvas = figureCanvas;
	}

	/**
	 * Get the canvas this figure is drawn on
	 * @return The canvas
	 */
	public FigureCanvas getFigureCanvas() {
		return this._figureCanvas;
	}

	/**
	 * Set the selection distance
	 * @param selectionDistance The distance to the selection point
	 */
	public void setSelectionDistance(double selectionDistance) {
		this._selectionDistance = selectionDistance;
	}

	/**
	 * Get the selection distance
	 * @return The distance to the selection point
	 */
	public double getSelectionDistance() {
		return this._selectionDistance;
	}

	/**
	 * Set the end of the role figure
	 * @param point The end position
	 */
	public void setEnd(Point point) {
		SystemView.setLocation(this._model, point);
		if (_model != null) {
			if (_roleDecoration == null) {
				_roleDecoration = new RoleDecoration(_model, this);
			}
			else {
				_roleDecoration.setModel(_model);
			}
			this.setTargetDecoration(_roleDecoration);
//			if (_nameLabel == null) {
//				ConnectionEndpointLocator targetEndpointLocator = new ConnectionEndpointLocator(this, true);
//				targetEndpointLocator.setVDistance(15);
//				_nameLabel = new Label(_model.getName());
//				this.add(_nameLabel, targetEndpointLocator);
//			}
//			else {
//				_nameLabel.setText(_model.getName ());
//			}
		}
		super.setEnd(point);
	}

	/**
	 * Get the offset of the client area
	 * @return the offset
	 */
	public Point getOffset() {
		int offsetY = 0, offsetX = 0;
		if (this._figureCanvas != null) {
			offsetY = this._figureCanvas.getViewport().getClientArea().y;
			offsetX = this._figureCanvas.getViewport().getClientArea().x;
		}
		return new Point(offsetX, offsetY);
	}

	protected void scale() {
		AcmeElement e = this._model.getParent();
		if (e instanceof AcmeConnector) {
			AcmeConnector acmeConnector = (AcmeConnector) e;
			Point location = SystemView.getLocation(acmeConnector, true);
			Point endPoint = SystemView.getLocation(this._model, true);

			this.setStart(new Point(location.x, location.y));
			this.setEnd(endPoint);
		}
		else {
			int y = (int) (Math.random() * 100) + 50;
			this.setStart(new Point(100, y));
			this.setEnd(new Point(200, y));
		}
	}

	/**
	 * @see org.eclipse.draw2d.PolylineConnection#layout()
	 */
	public void layout() {
		ConnectionRouter cr = getConnectionRouter();
		if (cr != null) {
			cr.route(this);
		}

		bounds = null;
		repaint();
		fireMoved();
	}

	/**
	 * @see org.eclipse.draw2d.IFigure#addNotify()
	 */
	public void addNotify() {
		super.addNotify();
		if (_model != null && _vis == null) {
			_vis = VisualStyleRegistry.getVisualization(_model);
		}
		if (_vis != null) {
			_vis.addPropertyChangeListener(_model.getQualifiedName() + "r", this);
		}
	}

	/**
	 * @see org.eclipse.draw2d.Shape#outlineShape(Graphics)
	 */
	protected void outlineShape(Graphics graphics) {
		graphics.pushState();
		ElementVisualization vis = _vis.getVariantToApply(_model);
		this.setForegroundColor(vis.getOutlineColor());
        if (_transparent) {
            this.setLineStyle (SWT.LINE_DOT);
        }
        else {
            this.setLineStyle(vis.getOutlineStyle());
        }
		//this.setLineWidth(vis.getOutlineWidth());
		super.outlineShape(graphics);
		graphics.popState();
	}

	public ElementVisualization getVisualization() {
		return _vis;
	}

	public void resetVisualization() {
		ElementVisualization newVis = VisualStyleRegistry.getVisualization(_model);
		if (newVis != _vis) {
			_vis.removePropertyChangeListener(_model.getQualifiedName(), this);
			_vis = newVis;
			_vis.addPropertyChangeListener(_model.getQualifiedName(), this);
			this.repaint();
		}
	}

	/**
	 * @see java.beans.PropertyChangeListener#propertyChange(PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		if (evt.getPropertyName().equals(ElementVisualization.PROP_VIS_CHANGED)) {
			_vis = (ElementVisualization) evt.getNewValue();
			//            if (_editPart != null) {
			//                _editPart.refresh();
			//            }
			this.repaint();
		}
		else if (evt.getPropertyName().equals(ElementVisualization.PROP_VIS_RESET)) {
			resetVisualization();
		}
	}

	/**
	 * @see org.eclipse.draw2d.IFigure#removeNotify()
	 */
	public void removeNotify() {
		super.removeNotify();
		if (_vis != null) {
			_vis.removePropertyChangeListener(_model.getQualifiedName() + "r", this);
			_vis = null;
		}
	}

	public RoleDecoration getDecoration() {
		return _roleDecoration;
	}

	public Point getLastBendpoint() {
		if (_bendpoints == null || _bendpoints.size() == 0)
			return getStart();
		return _bendpoints.getLastPoint();
	}

	public PointList getBendpoints() {
		return _bendpoints;
	}

	/**
	 * @param bendPoints
	 */
	public void setBendpoints(PointList bendPoints) {
		_bendpoints = bendPoints;
	}

    /**
     * @param transparent The transparent to set.
     */
    public void setTransparent(int transparency) {
        _transparent = transparency < 100;
        _roleDecoration.setTransparent(transparency);
    }

}