/*
 * Created on Jun 15, 2004
 *
 */
package AcmeStudio.ui.editors.diagram.palette.extension;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.eclipse.gef.palette.PaletteContainer;
import org.eclipse.gef.palette.PaletteEntry;
import org.eclipse.gef.palette.ToolEntry;
import org.eclipse.gef.ui.actions.SetActivePaletteToolAction;
import org.eclipse.gef.ui.palette.PaletteViewer;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.resource.ImageDescriptor;

import AcmeStudio.ui.editors.diagram.palette.connection.RuleBasedConnectionCreationToolEntry;


/**
 * @author Bradley Schmerl
 * Created Jun 15, 2004
 * 
 * The model object for a PaletteTree - A tree where the leaves are tools.
 */
public class PaletteTree extends PaletteContainer {

	private List subtree = new ArrayList ();
	
	/** Type identifier **/
	public static final String PALETTE_TYPE_TREE = "$PaletteTree";
	
	/** Property name for the active entry **/
	public static final String PROPERTY_ACTIVE_ENTRY = "Active Entry";
	
	private PaletteEntry activeEntry;
	
	/**
	 * Creates a new PaletteTree with the given name, description, and icon. 
	 * @see PaletteEntry#PaletteEntry(String, String, ImageDescriptor, String)
	 */
	public PaletteTree(String label, String shortDescription, ImageDescriptor icon) {
		super (label, shortDescription, icon, PALETTE_TYPE_TREE);
		setUserModificationPermission(PERMISSION_LIMITED_MODIFICATION);
	}
	
	/**
	 * Returns trye if this type can be a child of this container
	 * Only accepts ToolEntry's and PaletteTree's
	 * 
	 * @param type the type being requested
	 * @return true if this can be a child of this container
	 */
	public boolean acceptsType (Object type) {
		if (!type.equals (ToolEntry.PALETTE_TYPE_TOOL) && !type.equals (PALETTE_TYPE_TREE))
			return false;
		return super.acceptsType (type);
	}
	
	
	
	/**
	 * @param index
	 * @param entry
	 * @see org.eclipse.gef.palette.PaletteContainer#add(int, org.eclipse.gef.palette.PaletteEntry)
	 **/
	public void add(int index, PaletteEntry entry) {
		super.add(index, entry);
//		checkActiveEntry ();
	}
	/**
	 * Checks to make sure the active entry is up-to-date and sets it to the first child if it is null
	 */
	public void checkActiveEntry() {
		// Check active entyr on the root
		PaletteTree root = this;
		while (root.getParent () instanceof PaletteTree) {
			root = (PaletteTree )root.getParent ();
		}
		
		PaletteEntry currEntry = activeEntry;
		if (!root.treeContainsEntry (activeEntry)) activeEntry = null;
//		if (!getChildren ().contains (activeEntry)) activeEntry = null;
		if (activeEntry == null && root.getChildren ().size () > 0)
			activeEntry = root.getFirstNonTreeEntry ();
		listeners.firePropertyChange (PROPERTY_ACTIVE_ENTRY, currEntry, activeEntry);
	}

	private PaletteEntry getFirstNonTreeEntry () {
		PaletteEntry entry = null;
		Iterator iter = getChildren ().iterator ();
		while (iter.hasNext () && entry == null) {
			Object child = (Object )iter.next ();
			if (child instanceof PaletteTree) {
				PaletteTree tree = (PaletteTree )child;
				entry = tree.getFirstNonTreeEntry();
			}
			else if (child instanceof ToolEntry) {
				ToolEntry tool = (ToolEntry )child;
				entry = tool;
			}
		}
		return entry;
	}
	/**
	 * @param list
	 * @see org.eclipse.gef.palette.PaletteContainer#addAll(java.util.List)
	 **/
	public void addAll(List list) {
		super.addAll (list);
		checkActiveEntry ();
	}
	
	/**
	 * Returns the PaletteEntry referring to the active entry that shoul be shown in the palette
	 * @return active entry to be shown in the palette
	 */
	public PaletteEntry getActiveEntry () {
		checkActiveEntry ();
		return activeEntry;
	}
	
	
	/**
	 * @param entry
	 * @see org.eclipse.gef.palette.PaletteContainer#remove(org.eclipse.gef.palette.PaletteEntry)
	 **/
	public void remove(PaletteEntry entry) {
		super.remove (entry);
		checkActiveEntry ();
	}
	
	public void setActiveEntry (PaletteEntry entry) {
		PaletteEntry oldEntry = activeEntry;
		if (activeEntry != null && (activeEntry.equals (entry))) return; // || !treeContainsEntry (entry))) return;
		activeEntry = entry;
		listeners.firePropertyChange (PROPERTY_ACTIVE_ENTRY, oldEntry, activeEntry);
		for (Iterator iter = getChildren ().iterator (); iter.hasNext ();) {
			PaletteEntry nt = (PaletteEntry )iter.next ();
			if (nt instanceof PaletteTree) {
				PaletteTree tree = (PaletteTree )nt;
				tree.setActiveEntry (entry);
			}
		}
	}

	/**
	 * Searches recursively through the tree of children and sees whether the entry is contained therein
	 * @param entry The entry to search for
	 * @return whether the entry is in the tree of children
	 */
	public boolean treeContainsEntry(PaletteEntry entry) {
		boolean found = false;
		Iterator iter = getChildren ().iterator ();
		while (iter.hasNext () && !found) {
			Object object = (Object )iter.next ();
			if (object instanceof PaletteTree) {
				PaletteTree tree = (PaletteTree )object;
				found = tree.treeContainsEntry(entry);
			}
			else if (object.equals (entry)) {
				found = true;
			}
		}
		return found;
	}
	
	/**
	 * Returns a menu containing entries for each of the tools, organized according to the PaletteTree structure
	 * @param viewer
	 * @return the menu containing all the entries in the PaletteTree
	 */
	public MenuManager getSubMenu (PaletteViewer viewer) {
		MenuManager menuManager = new MenuManager (getLabel ());
		Iterator children = getChildren().iterator ();
		PaletteEntry entry = null;
		while (children.hasNext ()) {
			entry = (PaletteEntry )children.next ();
			if (entry instanceof PaletteTree) {
				PaletteTree tree = (PaletteTree )entry;
				MenuManager subMenu = tree.getSubMenu (viewer);
				menuManager.add (subMenu);
			}
			else {
				boolean active = getActiveEntry ().equals (entry);
				if (entry instanceof RuleBasedConnectionCreationToolEntry) {
					RuleBasedConnectionCreationToolEntry rbte = (RuleBasedConnectionCreationToolEntry) entry;
					menuManager.add(new SetActivePaletteToolAction (viewer, rbte.getDescription(), entry.getSmallIcon (), active, rbte));
				}
				else 
					menuManager.add (new SetActivePaletteToolAction (viewer, entry.getLabel (), entry.getSmallIcon (), active, (ToolEntry )entry));
			}
		}
		return menuManager;
	}
}
