package AcmeStudio.ui.editors.diagram.parts;

import java.util.Iterator;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.EditPartFactory;
import org.eclipse.gef.EditPartViewer;
import AcmeStudio.ui.editors.diagram.IElementFactory;
import AcmeStudio.ui.editors.diagram.model.AcmeRepresentationModel;
import acmeLib.elements.*;
import acmeLib.representations.AcmeRepresentation;

/**
 * @author Hong Yan (yh@cs.cmu.edu)
 *
 * A factory for creating new EditParts for acme elements.
 * {@link EditPartViewer EditPartViewers} can be
 * configured with an <code>EditPartFactory</code>. Whenever an
 * <code>EditPart</code> in that viewer needs to create another
 * EditPart, it can use the Viewer's factory. The factory is also
 * used by the viewer whenever {@link EditPartViewer#setContents(Object)}
 * is called.
 */
public class ElementEditPartFactory implements EditPartFactory {

    /**
     * Creates a new EditPart given the specified <i>context</i> and <i>model</i>.
     * @param context The context in which the EditPart is being created,
     * such as its parent.
     * @param model the model of the EditPart being created
     * @return EditPart the new EditPart
     */
    public EditPart createEditPart(EditPart context, Object model) {
        EditPart child = null;
        if (model instanceof AcmeComponent) {
            child = new ComponentEditPart();
        }
        else if (model instanceof AcmeConnector) {
            child = new ConnectorEditPart();
        }
        else if (model instanceof AcmeSystem) {
            child = new SystemEditPart();
        }
        else if (model instanceof AcmePort) {
            child = new PortEditPart();
        }
        else if (model instanceof AcmeRole) {
            child = new RoleEditPart();
        }
        else if (model instanceof AcmeAttachment) {
            child = new AttachmentEditPart();
        }
        else if (model instanceof SystemEditPart.Binding) {           
            child = new BindingEditPart();
        } 
        else if (model instanceof SystemEditPart.BindingPoint) {
            child = new BindingPointEditPart ();
        }
        else if (model instanceof AcmeRepresentation) {
        	child = new AcmeRepresentationEditPart ();
        }
        else if (model instanceof AcmeRepresentationModel) {
        	child = new RepresentationBorderEditPart ();
        }
        else {
            Iterator iter = SystemEditPart.factories.iterator();
            while (iter.hasNext()) {
                IElementFactory factory = (IElementFactory) iter.next();
                if ((child = factory.getEditPart(model)) != null) {
                    break;               
                }
            }
        }
        
        if (child == null) return null;
        child.setModel(model);
        return child;
    }

}
