package AcmeStudio.ui.editors.diagram.parts;
/*
 * Licensed Material - Property of IBM
 * (C) Copyright IBM Corp. 2001, 2002 - All Rights Reserved.
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */

import org.eclipse.draw2d.AbstractLocator;
import org.eclipse.draw2d.Connection;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.PointList;

/**
 * Repositions a {@link Figure Figure} attached to a 
 * {@link Connection Connection} when the Connection is moved. 
 * Provides for alignment at the start, middle, or end of the
 * Connection.
 */
public class RoleLocator extends AbstractLocator {

    private Connection connection;
    private int distance;

    /**
     * Constructs a ConnectionLocator with the passed Connection
     * and alignment.
     * 
     * @param connection Connection that ConnectionLocator is
     *                    associated with.
     * @param align Alignment of ConnectionLocator. 
     *               Valid values are integer constants START, MIDDLE, or END
     * @since 2.0
     */
    public RoleLocator(Connection connection, int distance) {
        setConnection(connection);
        setDistance(distance);
    }

    /**
     * Returns alignment of ConnectionLocator.
     * 
     * @since 2.0
     */
    public int getDistance() {
        return distance;
    }

    /**
     * Returns connection associated with ConnectionLocator.
     * 
     * @since 2.0
     */
    protected Connection getConnection() {
        return connection;
    }

    /**
     * Returns ConnectionLocator's reference point.
     * 
     * @since 2.0
     */
    protected Point getReferencePoint() {
        Point p = getLocation(getConnection().getPoints());
        getConnection().translateToAbsolute(p);
        return p;
    }

    /**
     * Returns a point from the passed PointList
     * dependent on ConnectionLocator's alignment.
     * If START, returns first point in points
     * If END, returns last point in points
     * If MIDDLE, returns middle of line represented by points.
     * 
     * @since 2.0
     */
    protected Point getLocation(PointList points) {
        int i = points.size() - 1;
        int j = i - 1;
        Point p1 = points.getPoint(j);
        Point p2 = points.getPoint(i);
        Dimension d = p2.getDifference(p1);
        double t =
            ((double) distance * distance)
                / (d.width * d.width + d.height * d.height);
        t = Math.sqrt(t);
        return Point.SINGLETON.setLocation(
            p2.x - (int) (d.width * t),
            p2.y - (int) (d.height * t));
    }

    /**
     * Sets alignment to passed value.
     * 
     * @since 2.0
     */
    protected void setDistance(int distance) {
        this.distance = distance;
    }

    /**
     * Sets Connection associated with ConnectionLocator to 
     * passed Connection.
     * 
     * @since 2.0
     */
    protected void setConnection(Connection connection) {
        this.connection = connection;
    }

}