package AcmeStudio.ui.editors.diagram.styles;
import java.io.ByteArrayInputStream;
import java.util.*;
import org.eclipse.draw2d.*;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CCombo;
import org.eclipse.swt.custom.CTabFolder;
import org.eclipse.swt.custom.CTabItem;
import org.eclipse.swt.events.*;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.ui.forms.events.ExpansionAdapter;
import org.eclipse.ui.forms.events.ExpansionEvent;
import org.eclipse.ui.forms.events.IExpansionListener;
import org.eclipse.ui.forms.widgets.ExpandableComposite;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.eclipse.ui.forms.widgets.Section;
import AcmeStudio.ui.editors.diagram.styles.constants.*;
import AcmeStudio.ui.editors.diagram.styles.constants.Shape;
import AcmeStudio.ui.widgets.ImageCombo;
import AcmeStudio.util.*;
import acmeLib.armaniParser.ArmaniParser;
import acmeLib.armaniParser.ParseException;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.designRules.AcmeResolutionException;
import acmeLib.designRules.DesignRule;
import acmeLib.elements.*;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.properties.AcmeProperty;
import com.swtdesigner.SWTResourceManager;
/**
 * @author Bradley Schmerl (schmerl@cs.cmu.edu)
 * Implements the dialog for editing or creating a new visual style
 * for depicting elements in a diagram.
 * 
 */
public class EditVisualizationDialog extends Dialog implements ImagePreviewer {
    private ImageSelectionGroup _imageSelectionGroup;
    private Button _modifyVariantButton;
    private static Image NO_VARIANT = null;
    // Armani rules corresponding to allowable types
    protected static final String ALLOWABLEPORTSRULE =
        "invariant forall p in self.ports | exists t in {%} | declaresType (p, t) <<vis-ports : boolean = true;>>;";
    protected static final String ALLOWABLEROLESRULE =
        "invariant forall r in self.roles | exists t in {%} | declaresType (r, t) <<vis-roles : boolean = true;>>;";
    /**
     * The location of the dialog box
     */
    protected Composite _parent;
    /**
     * The style being edited. Note, this is a copy of an original style,
     * which should faciliate undo
     */
    protected ElementVisualization _style;
    /**
     * The type information for the style (this is the type that the 
     * style will be associated with)
     */
    protected AcmeElementType _type;
    protected Composite _previewComp;
    /**
     * The label that contains an image depicting a preview of what the style
     * will look like.
     */
    //	protected Label _previewImageCanvas;
    protected FigureCanvas _previewImageCanvas;
    protected ImageFigure _imgFig;
    protected ScrollPane _previewPanel;
    protected ImageFigure _variantImage;
    /**
     * A combo box that holds all the types that a new style can be 
     * created for
     */
    protected Combo _typeCombo;
    /**
     * Indicates whether we are creating a new visual style or editing
     * an existing style.
     */
    protected boolean _newVisualization;
    /** 
     * The folder containing all the pages for the different options.
     */
    protected CTabFolder _folder;
    /**
     * The composite containing all the information to do with choosing
     * the shape in the style.
     */
    protected Composite _shapePage;
    /**
     * The composite containing all the information to do with setting
     * type specific styles.
     */
    protected Composite _specialPage;
    /**
     * The composite containing all the widgets to do with choosing label
     * options for the style.
     */
    protected Composite _labelPage;
    // Some of thest to be used later
    //	protected Composite _detailsPage;
    protected Composite _variantsPage;
    //	protected Composite _portAlignmentPage;

    /**
     * The choice of options for stock shapes. The order of these should
     * correspond to the order in VisStyles.Shapes.
     */
    private String[] _stockShapes = Shape.LABELS;

    /**
     * The strings for horizontal alignment options. The order of these should
     * correspond to the order of the constants in Visualization.HorizontalTextAlignmentOptions
     * for the combo boxes to work.
     */
    private String[] _horizAlignmentOptions =
        HorizontalTextAlignmentOptions.LABELS;
    /**
     * The strings for vertical alignment options. The order of these should
     * correspond to the order of the constants in Visualization.VerticalTextAlignmentOptions
     * for the combo boxes to work.
     */
    private String[] _vertAlignmentOptions =
        VerticalTextAlignmentOptions.LABELS;

    private String[] _directionOptions =
        //new String[] { "North", "South", "West", "East" };
    new String[] { "0", "90", "180", "270" };

    /**
     * The choice of options for port/role orientations.
     */
    private String[] _orientations = Orientation.LABELS;

    /**
     * The button used for selecting stock shapes (as opposed to images)
     */
    protected Button _stockShapeButton;
    /**
     * The combo that contains all the valid stock shapes (entries from _stockShapes)
     */
    protected ImageCombo _shapeCombo;
    protected Combo _directionCombo;
    protected CCombo _orientationCombo;

    /**
     * The button indicating that a shape should be replicated.
     */
    protected Button _replicateShapeButton;
    /**
     * The button indicating that a port should rotate.
     */
    protected Button _rotatePortButton;
    /**
     * The button indicating that a port should align with role.
     */
    protected Button _alignPortButton;
    /**
     * The button indicating that a role should rotate.
     */
    protected Button _rotateRoleButton;
    /**
     * The button indicating that word wrap should be used.
     */
    protected Button _wordwrapButton;
    protected Composite _restrictTypeContainer;
    protected Group _portAlignmentGroup;
    protected Composite _alignPorts;
    protected Label _allowableLabel;
    protected Label _availableLabel;
    protected Button _deleteFromAllowableButton;
    protected Button _addToAllowableButton;
    protected Button _deleteFromAlignPortButton;
    protected Button _addToAlignPortButton;
    protected Button _deleteFromRouteRoleButton;
    protected Button _addToRouteRoleButton;
    protected Button _onlyButton;
    protected Button _allButton;
    protected LinkedList _portAlignmentRules = null;
    protected LinkedList _roleRouteRules = null;
    /**
     * The button to set the mode to clear decorations
     */
    protected Button _clearDecorationButton;

    /**
     * THe button to allow selection of an image to use (not a decoration)
     */
    protected Button _clearImageButton;

    protected Button _chooseImageButton;

    /**
     * A check button for indicating whether to preserve aspect ratio
     * when resizing a figure.
     */
    protected Button _preserveAspectRatioButton;
    /**
     * The text box to allow entering of the default width
     */
    protected Text _defaultWidthText;
    /**
     * The text box to allowing entering of the default height
     */
    protected Text _defaultHeightText;
    /** 
     * A check box that tells whether the user is allowed to resize 
     * the width
     */
    protected Button _widthResizeableButton;
    /** 
     * A check box that tells whether the user is allowed to resize 
     * the height
     */
    protected Button _heightResizeableButton;
    /**
     * Button to allow the user to choose the outline color
     */
    protected Button _outlineColorButton;
    /**
     * Button to allow the user to choose the fill color
     */
    protected Button _fillColorButton;
    protected Label _imageLabel;
    /**
     * Label containing the title
     */
    protected Label _titleLabel;
    /**
     *  Buttons to allow the user to select the outline line style
     */
    protected Button _noLineButton;
    protected Button _solidLineButton;
    protected Button _dotLineButton;
    protected Button _dashLineButton;
    protected Button _dashDotLineButton;
    protected Button _dashDotDotLineButton;
    /**
     * Allows the user to specify the width of the outline
     */
    protected Combo _outlineWidthCombo;
    /**
     * Allows the user to select the label font
     */
    protected Button _fontButton;
    /** 
     * Allows the user to select the label color
     */
    protected Button _fontColorButton;
    /**
     * Buttons to allow the user to choose where the label
     * comes from
     */
    protected Button _noLabelButton;
    protected Button _staticLabelButton;
    protected Button _identifierButton;
    protected Button _idAndTypeButton;
    protected Button _typeButton;
    protected Button _propertyButton;
    /** 
     * If the label can come from a property, then this allows the
     * user to select the property in the type.
     */
    protected CCombo _propertyCombo;
    /** 
     * If the label is a static one, then this allows the
     * user to set the static label value.
     */
    protected Text _labelText;
    /** 
     * Allows the user to select the horizontal and vertical alignments
     */
    protected Combo _horizAlignCombo;
    protected CCombo _vertAlignCombo;
    protected CCombo _roleNumber;
    protected List _variantsList;
    protected List _allowableTypeList;
    protected List _availableTypeList;
    protected PortAlignmentList _alignPortTypeTable;
    protected RoleRouteList _routeRoleTypeTable;
    protected List _alignPortTypeList;
    protected List _routeRoleTypeList;
    protected Button _newVariantButton;
    protected Button _deleteVariantButton;
    protected Button _orderUpButton;
    protected Button _orderDownButton;
    protected Label _variantPreviewLabel;
    /**
     * The text containing how big the margin should be
     */
    protected Text _marginText;
    private Button _addAllToAllowableButton;
    private Button _deleteAllFromAllowableButton;
	private boolean	_askAboutElement;
	private Button	_btnOverrideTypeVis;
	protected boolean	_overrideTypeVis;
	private AcmeElement	_element;
	private Button	_resizeToFitButton;
	private boolean	_canOverrideTypeVis;
    /**
     * Constructor for EditVisualizationDialog.
     * This takes a copy of the visual style so that it can be changed
     * directly and then not copied over if the action is cancelled
     * @param parentShell The shell where the dialog will be raised
     * @param vs The visual style to edit or create
     */
    public EditVisualizationDialog(
        Shell parentShell,
        ElementVisualization vs) {
        super(parentShell);
        _style = vs.copy();
        initializeStaticImage();
    }
    
    public EditVisualizationDialog (Shell parentShell, ElementVisualization vis, AcmeElement element) {
    	super (parentShell);
    	_style = vis.copy ();
    	_element = element;
    	_askAboutElement = true;
    	_overrideTypeVis = true;
    	initializeStaticImage();
    }
    /**
     * Constructor for EditVisualizationDialog, which creates a new visualization
     * Finds the best fitting style from the registry, and then copies the style
     * @param parentShell The Shell where the dialog will be raised
     * @param type The type that will be associated with the new style
     */
    public EditVisualizationDialog(
        Shell parentShell,
        AcmeElementType type,
        boolean isCreating) {
        super(parentShell);
        _type = type;
        _style =
            VisualStyleRegistry.getVisualization(type).copy((AcmeElement) type);
        specializeVisualization ();
        
        
        _newVisualization = isCreating;
        initializeStaticImage();
        updateTitle();
    }
    
    private void specializeVisualization() {
		// Convert erroneous generic visualizations to their proper type
        if (_style instanceof GenericVisualization) {
        	if (_type instanceof AcmePortType) {
        		PortVisualization pVis = new PortVisualization ();
        		pVis.copyInto (_style);
        		_style.dispose ();
        		_style = pVis;
        	}
        	else if (_type instanceof AcmeComponentType) {
        		ComponentVisualization cVis = new ComponentVisualization ();
        		cVis.copyInto(_style);
        		_style.dispose ();
        		_style = cVis;
        	}
        	else if (_type instanceof AcmeConnectorType) {
           		ConnectorVisualization cVis = new ConnectorVisualization ();
        		cVis.copyInto(_style);
        		_style.dispose ();
        		_style = cVis;        		
        	}
        	else if (_type instanceof AcmeRoleType) {
           		RoleVisualization cVis = new RoleVisualization ();
        		cVis.copyInto(_style);
        		_style.dispose ();
        		_style = cVis;        		
        	}
        	
        }
	}
	public EditVisualizationDialog (Shell parentShell, AcmeElement element, AcmeElementType type) {
    	super (parentShell);
    	_style = VisualStyleRegistry.getVisualization (element).copy();
    	specializeVisualization ();
    	_newVisualization = false;
    	_askAboutElement = true;
    	_overrideTypeVis = VisualStyleRegistry.hasSpecificVisualization (element);
    	_type = type;
    	_element = element;
    	initializeStaticImage();
    	updateTitle();
    }
    
    private void updateTitle() {
        if (_titleLabel != null) {
            if (_newVisualization) {
            	if (_element != null && isOverrideTypeVis()) {
            		_titleLabel.setText ("Creating visualization for " + _element.getName ());
            	}
            	else if (_type != null)
	                _titleLabel.setText(
	                    "Creating visualization for " + _type.getTypeName());
            		
            } else {
                // If editing an existing visualization, don't let the 
                // user change the associated type
               	if (_element != null && isOverrideTypeVis()) 
            		_titleLabel.setText ("Creating visualization for " + _element.getName ());
            	else if (_type != null)
	                _titleLabel.setText(
	                    "Modifying visualization"
	                        + " for "
	                        + (_type == null ? "<none>" : _type.getTypeName()));
            }
        }
    }
    private Image initializeStaticImage() {
        if (NO_VARIANT == null) {
            NO_VARIANT = new Image(Display.getDefault(), 100, 100);
            GC gc = new GC(NO_VARIANT);
            String noImage = "No Variant";
            Point p = gc.textExtent(noImage);
            gc.drawText(noImage, 50 - p.x / 2, 50 - p.y / 2);
            gc.dispose();
        }
        return NO_VARIANT;
    }
    
    private Control superCreateDialogArea (Composite parent) {
//    	 create a composite with standard margins and spacing
		Composite composite = new Composite(parent, SWT.V_SCROLL);
		GridLayout layout = new GridLayout();
		layout.marginHeight = convertVerticalDLUsToPixels(IDialogConstants.VERTICAL_MARGIN);
		layout.marginWidth = convertHorizontalDLUsToPixels(IDialogConstants.HORIZONTAL_MARGIN);
		layout.verticalSpacing = convertVerticalDLUsToPixels(IDialogConstants.VERTICAL_SPACING);
		layout.horizontalSpacing = convertHorizontalDLUsToPixels(IDialogConstants.HORIZONTAL_SPACING);
		composite.setLayout(layout);
		composite.setLayoutData(new GridData(GridData.FILL_BOTH));
		applyDialogFont(composite);
		return composite;
    }
    /**
     * Creates the dialog area, which has a preview, the type the 
     * visualization is associated with, and a tabbed folder containing
     * pages that allow various options of the style to be edited.
     * @param parent The place where all the controls go
     * @see org.eclipse.jface.dialogs.Dialog#createDialogArea(Composite)
     */
    protected Control createDialogArea(Composite parent) {
        _parent = (Composite) super.createDialogArea(parent);
        FormToolkit toolkit = new FormToolkit (Display.getCurrent());
        toolkit.adapt (_parent);
        toolkit.paintBordersFor (_parent);
        GridLayout gl = (GridLayout) _parent.getLayout();
        gl.verticalSpacing = 10;
        //        _parent.setSize (500, 500);
        // If this is a new visualization, then make a combo
        // box that lets the user choose which type the visualization is
        // for.
        _titleLabel = toolkit.createLabel(_parent, "");
        updateTitle();
        // Create the preview area
        _previewComp = toolkit.createComposite(_parent);
        gl = new GridLayout(1, false);
        gl.marginWidth = 10;
        _previewComp.setLayout(gl);
        _previewComp.setLayoutData(
            new GridData(
                GridData.VERTICAL_ALIGN_BEGINNING
                    | GridData.HORIZONTAL_ALIGN_CENTER));
        //        Label previewLabel = new Label(_previewComp, SWT.NONE);
        //        previewLabel.setLayoutData(new GridData(GridData.BEGINNING));
        // 		previewLabel.setText ("Preview:");
        // This label will contain an image depicting what the
        // visualization will look like.
        // 		_previewImageCanvas = new Label (_previewComp, SWT.SHADOW_IN);		
        initPreview();
        // The number of tabs in the folder	
        int tabCount = 0;
        if (_askAboutElement) {
        	_btnOverrideTypeVis = toolkit.createButton(_parent, "Apply visualization to this element only (other elements of the same type will not change)",SWT.CHECK);
        	_btnOverrideTypeVis.addSelectionListener(new SelectionAdapter () {
     
				public void widgetSelected (SelectionEvent evt) {
        			if (_btnOverrideTypeVis.getSelection()) {
        				_overrideTypeVis = true;
        				updateTitle();
        			}
        			else {
        				_overrideTypeVis = false;
        				updateTitle ();
        			}
        		}
        	});
        	_btnOverrideTypeVis.setSelection(isOverrideTypeVis());
        	_btnOverrideTypeVis.setEnabled(_canOverrideTypeVis);
        }
        // Create the shape page and add it to the tab
        _folder = new  CTabFolder(_parent, SWT.NONE);
//        toolkit.paintBordersFor(_folder);
        _folder.setBackground(ColorConstants.white);
        CTabItem shapeTab = new CTabItem(_folder, SWT.NONE, tabCount++);
        shapeTab.setText("Shape");
        _shapePage = createShapePage(_folder, toolkit);
        shapeTab.setControl(_shapePage);
        // Create the label page and add it to the tab
        CTabItem labelTab = new CTabItem(_folder, SWT.NONE, tabCount++);
        labelTab.setText("Label");
        _labelPage = createLabelPage(_folder, toolkit);
        labelTab.setControl(_labelPage);
        CTabItem variantTab = new CTabItem(_folder, SWT.NONE, tabCount++);
        variantTab.setText("Variants");
        _variantsPage = createVariantsPage(_folder, toolkit);
        variantTab.setControl(_variantsPage);
        createSpecialPage(tabCount, toolkit);
        _folder.setSelection(shapeTab);
      
        // Set all the controls to contain information based
        // on the visual style
        importStyle();
//        _folder.pack();
        // Make sure the preview has an image to start with
        updatePreview();
        
//                _parent.pack();
        return _parent;
    }
    /**
     * Creates a title for the page. Assumes that parent has GridLayout
     * @param parent The composite in which to place the title and description
     * @param title The title of the page
     * @param description The description (or help message) for the page
     */
    private void createPageTitle(
        Composite parent,
        String title,
        String description) {
        
        Label descLabel = new Label(parent, SWT.LEFT);
        descLabel.setText(description);
        GridData descData = new GridData();
        descData.horizontalSpan = GridData.FILL_HORIZONTAL;
        descData.horizontalIndent = 25;
        descLabel.setLayoutData(descData);
    }
    /**
     * Creates the group to allow choosing of shapes. Includes components to:
     * a) choose a stock shape
     * b) choose an image to display
     * c) indicate whether the shape should be replicated
     * @param parent the element to put these components into
     * @param toolkit
     * @param expansionAdapter
     */
    private void createShapeGroup(final Composite parent, FormToolkit toolkit, ExpansionAdapter expansionAdapter) {
    	Section expandableComposite = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
    	toolkit.createCompositeSeparator(expandableComposite);
        expandableComposite.setText("Basic Shape");
        expandableComposite.setExpanded (false);
        expandableComposite.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	Composite shapeGroup = toolkit.createComposite (expandableComposite, SWT.NONE);
    	 shapeGroup.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	expandableComposite.setClient(shapeGroup);
    	
		expandableComposite.addExpansionListener(expansionAdapter);
        GridLayout shapeLayout = new GridLayout(2, false);
        shapeGroup.setLayout(shapeLayout);
        toolkit.createLabel(shapeGroup, "Stock Image:", SWT.NONE);
//        Composite shapeComboContainer = new Composite(shapeGroup, SWT.NONE);
        _shapeCombo = new ImageCombo (shapeGroup, SWT.DROP_DOWN);
        _shapeCombo.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
        _shapeCombo.setSize(150, 32);
        Object [] imgs = Shape.SHAPEIMAGES.toArray ();
        Image [] simgs = new Image [imgs.length-1];
        for (int i = 0; i < simgs.length; i++) {
			simgs [i] = (Image )imgs [i+1];
		}
        _shapeCombo.setItems(_stockShapes, simgs);
        _shapeCombo.select(_style.getShape());
        _shapeCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                // Sets the new stock shape. Because the constants in
                // Visualization.Shapes begin at 1, we need to add one
                // to the selection
                _style.setShape(_shapeCombo.getSelectionIndex() + 1);
                switch (_shapeCombo.getSelectionIndex() + 1) {
                    case Shape.TRIANGLE :
                    case Shape.FILTER :
                        _directionCombo.setEnabled(true);
                        break;
                    default :
                        _directionCombo.setEnabled(false);
                }
                updatePreview();
            }
        });
        toolkit.createLabel (shapeGroup, "Rotation");
        _directionCombo = new Combo(shapeGroup, SWT.DROP_DOWN);
        _directionCombo.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
        _directionCombo.setBackground(ColorConstants.white);
        _directionCombo.setSize(150, 32);
        _directionCombo.setItems(_directionOptions);
        _directionCombo.setData("270", new Integer(PositionConstants.NORTH));
        _directionCombo.setData("90", new Integer(PositionConstants.SOUTH));
        _directionCombo.setData("0", new Integer(PositionConstants.EAST));
        _directionCombo.setData("180", new Integer(PositionConstants.WEST));
        int direction = -1;
        switch (_style.getDirection()) {
            case PositionConstants.EAST :
                direction = 0;
                break;
            case PositionConstants.SOUTH :
                direction = 1;
                break;
            case PositionConstants.WEST :
                direction = 2;
                break;
            case PositionConstants.NORTH :
                direction = 3;
                break;
        }
        if (direction >= 0)
            _directionCombo.select(direction);
        _directionCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                Integer data =
                    (Integer) _directionCombo.getData(
                        _directionOptions[_directionCombo.getSelectionIndex()]);
                _style.setDirection(data.intValue());
                updatePreview();
            }
        });
        _replicateShapeButton = toolkit.createButton(shapeGroup, "Replicate shape", SWT.CHECK);
        _replicateShapeButton.setSelection(_style.getReplicateShape());
        _replicateShapeButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _style.setReplicateShape(_replicateShapeButton.getSelection());
                updatePreview();
            }
        });
    }
    /**
     * Creates the size options for the shape, allowing the user to indicate:
     * a) Whether the aspect ratio should be preserved
     * b) The default width, and if it is resizeable
     * c) The defualt height, and if it is resizeable
     * @param parent The element to put the controls
     * @param toolkit
     * @param expansionAdapter
     */
    private void createSizeGroup(Composite parent, FormToolkit toolkit, ExpansionAdapter expansionAdapter) {
    	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
    	expandComp.setExpanded (false);
    	toolkit.createCompositeSeparator(expandComp);
    	expandComp.setExpanded (false);
    	expandComp.addExpansionListener(expansionAdapter);
        expandComp.setText("Size settings");
        Composite sizeGroup = toolkit.createComposite (expandComp);
    	expandComp.setClient(sizeGroup);
        GridData sgld = new GridData();
        sgld.grabExcessHorizontalSpace = true;
        sgld.horizontalAlignment = GridData.FILL;
        sizeGroup.setLayoutData(sgld);
        GridLayout sggl = new GridLayout(3, false);
        sizeGroup.setLayout(sggl);
        _preserveAspectRatioButton = toolkit.createButton (sizeGroup, "Preserve Aspect Ratio", SWT.CHECK);
        GridData pard = new GridData();
        pard.horizontalSpan = 3;
        _preserveAspectRatioButton.setLayoutData(pard);
        toolkit.createLabel(sizeGroup, "Default Width:");
        _defaultWidthText = toolkit.createText (sizeGroup, "",SWT.SINGLE | SWT.BORDER);
        _defaultWidthText.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
//        _defaultWidthText.setSize(60, 20);
        //_defaultWidthText.setLayoutData(new GridData(GridData.FILL));
        // Make sure that the size is a number
        _defaultWidthText.addVerifyListener(Listeners.VERIFY_NUMBER);
    
        // Reflect changes in the preview if the numbers are changed,
        // but only when we leave
        _defaultWidthText.addFocusListener(new FocusListener() {
            // update the preview when the text is left
            public void focusLost(FocusEvent evt) {
                if (_defaultWidthText.getText().equals("")) {
                    _defaultWidthText.setText("0");
                }
                int value = -1;
                try {
                    value = Integer.parseInt(_defaultWidthText.getText());
                } catch (Exception e) {
                }
                if (value > 0) {
                    Dimension size = _style.getSize().getCopy ();
                    size.width = value;
                    _style.setSize(size);
                    updatePreview();
                }
            }
            public void focusGained(FocusEvent evt) {
            }
        });
        _widthResizeableButton = toolkit.createButton(sizeGroup, "Resizeable", SWT.CHECK);
        toolkit.createLabel (sizeGroup, "Default Height:");
        _defaultHeightText = toolkit.createText (sizeGroup, "",SWT.SINGLE | SWT.BORDER);
        _defaultHeightText.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
        _defaultHeightText.addVerifyListener(Listeners.VERIFY_NUMBER);
        _defaultHeightText.addFocusListener(new FocusListener() {
            // update the preview when the text is left
            public void focusLost(FocusEvent evt) {
                if (_defaultHeightText.getText().equals("")) {
                    _defaultHeightText.setText("0");
                }
                int value = -1;
                try {
                    value = Integer.parseInt(_defaultHeightText.getText());
                } catch (Exception e) {
                }
                if (value > 0) {
                    Dimension size = _style.getSize().getCopy ();
                    size.height = value;
                    _style.setSize(size);
                    updatePreview();
                }
            }
            public void focusGained(FocusEvent evt) {
            }
        });
        _heightResizeableButton = toolkit.createButton(sizeGroup, "Resizeable", SWT.CHECK);
        
        _resizeToFitButton = toolkit.createButton (sizeGroup, "Automatically resize to fit label", SWT.CHECK);
        
        GridData	rgd	= new GridData (GridData.FILL_HORIZONTAL);
        rgd.horizontalSpan = GridData.FILL;
		_resizeToFitButton.setLayoutData(rgd);
    }
    private void setRestrictionGroupEnabled(boolean enabled) {
        _restrictTypeContainer.setEnabled(enabled);
        _allowableLabel.setEnabled(enabled);
        _availableLabel.setEnabled(enabled);
        _allowableTypeList.setEnabled(enabled);
        _availableTypeList.setEnabled(enabled);
        _deleteFromAllowableButton.setEnabled(enabled);
        _addToAllowableButton.setEnabled(enabled);
    }
    private void createPortRestrictionGroup(
        Composite parent,
        final AcmeEnumeration ports, FormToolkit toolkit, ExpansionAdapter expansionAdapter) {
        final Button resizeButton = toolkit.createButton(parent, "Automatically resize component to fit ports", SWT.CHECK);
       
        boolean shouldResize =
            ((ComponentVisualization) _style).resizeToFitPorts();
        resizeButton.setSelection(shouldResize);
        resizeButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                boolean selection = resizeButton.getSelection();
                ((ComponentVisualization) _style).setResizeToFitPorts(
                    selection);
            }
        });
        Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
        expandComp.setExpanded (false);
       	toolkit.createCompositeSeparator(expandComp);
        expandComp.setText("Port Type Restrictions");
        expandComp.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	Composite portRestrictionGroup = toolkit.createComposite (expandComp, SWT.NONE);
    	portRestrictionGroup.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	expandComp.setClient(portRestrictionGroup);
    	
        GridLayout sggl = new GridLayout(1, false);
        portRestrictionGroup.setLayout(sggl);
        final Button onButton = toolkit.createButton(portRestrictionGroup, "Only allow ports of the following types:", SWT.CHECK);
        GridData pard = new GridData();
        pard.horizontalSpan = 1;
        onButton.setLayoutData(pard);
        boolean shouldRestrict =
            ((ComponentVisualization) _style).restrictPortTypes();
        onButton.setSelection(shouldRestrict);
        onButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                boolean selection = onButton.getSelection();
                ((ComponentVisualization) _style).setRestrictPortTypes(
                    selection);
                setRestrictionGroupEnabled(selection);
                refreshPortsInAlignmentGroup(!selection);
                updatePreview();
            }
        });
        _restrictTypeContainer = toolkit.createComposite (portRestrictionGroup, SWT.NONE);
        GridData sgld = new GridData(GridData.FILL_HORIZONTAL);
        _restrictTypeContainer.setLayoutData(sgld);
        sggl = new GridLayout(3, false);
        _restrictTypeContainer.setLayout(sggl);
        _allowableLabel = toolkit.createLabel (_restrictTypeContainer, "Allowable Types:", SWT.NONE);
        pard = new GridData();
        pard.horizontalSpan = 2;
        _allowableLabel.setLayoutData(pard);
        _availableLabel = toolkit.createLabel(_restrictTypeContainer, "Available Types:", SWT.NONE);
        _allowableTypeList =
            new List(
                _restrictTypeContainer,
                SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);
        pard = new GridData(GridData.FILL_BOTH);
        pard.grabExcessHorizontalSpace = true;
        _allowableTypeList.setLayoutData(pard);
        refreshAllowedPorts(true);
        Composite buttons = toolkit.createComposite(_restrictTypeContainer, SWT.NONE);
        buttons.setLayout(new GridLayout(1, true));
        createTypeMoveButtons(buttons, toolkit);
        Composite availablePortTypeContainer = toolkit.createComposite (_restrictTypeContainer, SWT.NONE);
        pard = new GridData(GridData.VERTICAL_ALIGN_FILL | GridData.HORIZONTAL_ALIGN_FILL);
        pard.grabExcessHorizontalSpace = true;
        availablePortTypeContainer.setLayoutData(pard);
        availablePortTypeContainer.setLayout(new GridLayout (1, false));
        _availableTypeList =
            new List(
                availablePortTypeContainer,
                SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);
        pard = new GridData(GridData.VERTICAL_ALIGN_FILL | GridData.HORIZONTAL_ALIGN_FILL);
        pard.grabExcessHorizontalSpace = true;
        pard.grabExcessVerticalSpace = true;
        _availableTypeList.setLayoutData(pard);
        _availableTypeList.setSize(200, 80);
        refreshAvailableTypes(ports, true);
        setRestrictionGroupEnabled(shouldRestrict);
    }
    private void createRoleRestrictionGroup(
        Composite parent,
        final AcmeEnumeration roles, FormToolkit toolkit, ExpansionAdapter expansionAdapter) {
    	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
        expandComp.setExpanded (false);
       	toolkit.createCompositeSeparator(expandComp);
        expandComp.setText("Role Type Restrictions");
        expandComp.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	Composite roleRestrictionGroup = toolkit.createComposite (expandComp, SWT.NONE);
    	roleRestrictionGroup.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	expandComp.setClient(roleRestrictionGroup);
    	
        GridLayout sggl = new GridLayout(1, false);
        roleRestrictionGroup.setLayout(sggl);
        final Button onButton = toolkit.createButton(roleRestrictionGroup, "Only allow roles of the following types:", SWT.CHECK);
        GridData pard = new GridData();
        pard.horizontalSpan = 1;
        onButton.setLayoutData(pard);
        boolean shouldRestrict =
            ((ConnectorVisualization) _style).restrictRoleTypes();
        onButton.setSelection(shouldRestrict);
        onButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                boolean selection = onButton.getSelection();
                ((ConnectorVisualization) _style).setRestrictRoleTypes(
                    selection);
                refreshRolesInRouteGroup(!selection);
                setRestrictionGroupEnabled(selection);
            }
        });
        _restrictTypeContainer = toolkit.createComposite(roleRestrictionGroup, SWT.NONE);
        GridData sgld = new GridData(GridData.FILL_HORIZONTAL);
        _restrictTypeContainer.setLayoutData(sgld);
        sggl = new GridLayout(3, false);
        _restrictTypeContainer.setLayout(sggl);
        _allowableLabel = toolkit.createLabel(_restrictTypeContainer, "Allowable Types:", SWT.NONE);
        pard = new GridData();
        pard.horizontalSpan = 2;
        _allowableLabel.setLayoutData(pard);
        _availableLabel = toolkit.createLabel(_restrictTypeContainer, "Available Types:", SWT.NONE);
        _allowableTypeList =
            new List(
                _restrictTypeContainer,
                SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);
        pard = new GridData(GridData.FILL_BOTH);
        pard.grabExcessHorizontalSpace = true;
        _allowableTypeList.setLayoutData(pard);
        refreshAllowedRoles(true);
        Composite buttons = toolkit.createComposite(_restrictTypeContainer, SWT.NONE);
        buttons.setLayout(new FillLayout (SWT.VERTICAL));
        
        createTypeMoveButtons(buttons, toolkit);
        Composite availableRoleTypeContainer =
            toolkit.createComposite(_restrictTypeContainer, SWT.NONE);
        pard = new GridData(GridData.VERTICAL_ALIGN_FILL | GridData.HORIZONTAL_ALIGN_FILL);
        pard.grabExcessHorizontalSpace = true;
        pard.grabExcessVerticalSpace = true;
        availableRoleTypeContainer.setLayoutData(pard);
        availableRoleTypeContainer.setLayout (new GridLayout (1, false));
        _availableTypeList =
            new List(
                availableRoleTypeContainer,
                SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);
        pard = new GridData(GridData.VERTICAL_ALIGN_FILL | GridData.HORIZONTAL_ALIGN_FILL);
        pard.grabExcessHorizontalSpace = true;
        pard.grabExcessVerticalSpace = true;
        _availableTypeList.setLayoutData(pard);
//        _availableTypeList.set
//        _availableTypeList.setSize(200, 80);
        refreshAvailableTypes(roles, true);
        setRestrictionGroupEnabled(shouldRestrict);
    }

    private void createTypeMoveButtons(Composite buttons, FormToolkit toolkit) {
        _addToAllowableButton = toolkit.createButton(buttons, "< Add", SWT.PUSH);
        _addToAllowableButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                String[] selections = _availableTypeList.getSelection();
                for (int i = 0; i < selections.length; i++) {
                    String sel = selections[i];
                    _allowableTypeList.add(sel);
                    _availableTypeList.remove(sel);
                    sortList (_allowableTypeList);
                }
                if (_onlyButton.getSelection())
                    refreshRolesInRouteGroup(false);
            }
        });
        
        _addAllToAllowableButton = toolkit.createButton (buttons, "<< Add all", SWT.PUSH);
        _addAllToAllowableButton.addSelectionListener(new SelectionAdapter () {
            public void widgetSelected (SelectionEvent evt) {
                String [] items= _availableTypeList.getItems ();
                for (int i = 0; i < items.length; i++) {
                    String item = items [i];
                    _allowableTypeList.add (item);
                    _availableTypeList.remove(item);
                }
                sortList(_allowableTypeList);
            }
        });
        _deleteFromAllowableButton = toolkit.createButton(buttons, "Remove >", SWT.PUSH);
        _deleteFromAllowableButton
            .addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                String[] selections = _allowableTypeList.getSelection();
                for (int i = 0; i < selections.length; i++) {
                    String sel = selections[i];
                    _allowableTypeList.remove(sel);
                    _availableTypeList.add(sel);
                }
                sortList(_availableTypeList);
                if (_onlyButton.getSelection())
                    refreshRolesInRouteGroup(false);
            }
        });
        _deleteAllFromAllowableButton = toolkit.createButton (buttons, "Remove All >>", SWT.PUSH);
        _deleteAllFromAllowableButton.addSelectionListener(new SelectionAdapter () {
           public void widgetSelected (SelectionEvent evt) {
               String [] items = _allowableTypeList.getItems();
               for (int i = 0; i < items.length; i++) {
                   String item = items [i];
                   _availableTypeList.add(item);
                   _allowableTypeList.remove (item);
               }
               sortList (_availableTypeList);
           }
        });
    }
    private void refreshPortsInAlignmentGroup(boolean all) {
        String[] items1, items2, items;
        items1 = _availableTypeList.getItems();
        items2 = _allowableTypeList.getItems();
        if (all) {
            items = new String[items1.length + items2.length];
            System.arraycopy(items1, 0, items, 0, items1.length);
            System.arraycopy(items2, 0, items, items1.length, items2.length);
        } else {
            items = items2;
        }
        for (int i = 0; i < items.length; i++) {
            PlacementRule rule = new PlacementRule();
            rule._type = items[i];
            if (_portAlignmentRules.contains(rule)) {
                items[i] = "";
            }
        }
        java.util.List ports = Arrays.asList(items);
        Collections.sort(ports);
        Iterator i = ports.iterator();
        if (all) {
            _alignPortTypeList.add("port");
        }
        _alignPortTypeList.removeAll();
        while (i.hasNext()) {
            String port = (String) i.next();
            if (port.length() > 0)
                _alignPortTypeList.add(port);
        }
    }

    public static void sortList(List list) {
        java.util.List items = Arrays.asList(list.getItems());
        list.removeAll();
        Collections.sort(items);
        Iterator i = items.iterator();
        while (i.hasNext()) {
            list.add((String) i.next());
        }
    }

    private void refreshRolesInRouteGroup(boolean all) {
        String[] items1, items2, items;
        items1 = _availableTypeList.getItems();
        items2 = _allowableTypeList.getItems();
        if (all) {
            items = new String[items1.length + items2.length];
            System.arraycopy(items1, 0, items, 0, items1.length);
            System.arraycopy(items2, 0, items, items1.length, items2.length);
        } else {
            items = items2;
        }
        for (int i = 0; i < items.length; i++) {
            ConnectorVisualization.RoleRouteRule rule =
                new ConnectorVisualization.RoleRouteRule();
            rule._roleType = items[i];
            if (_roleRouteRules.contains(rule)) {
                items[i] = "";
            }
        }
        java.util.List roles = Arrays.asList(items);
        Collections.sort(roles);
        Iterator i = roles.iterator();
        if (all) {
            _routeRoleTypeList.add("role");
        }
        _routeRoleTypeList.removeAll();
        while (i.hasNext()) {
            String role = (String) i.next();
            if (role.length() > 0)
                _routeRoleTypeList.add(role);
        }
    }

    private void refreshAvailableTypes(AcmeEnumeration types, boolean all) {
        _availableTypeList.removeAll();
        if (all) {
            java.util.List typeList =
                Helper.convertAcmeEnumerationToList(types);
            Collections.sort(typeList, new Utils.AcmeElementNameComparator());
            Iterator i = typeList.iterator();
            java.util.List names = Arrays.asList(_allowableTypeList.getItems());
            while (i.hasNext()) {
                String name = ((AcmeElement) i.next()).getName();
                if (!names.contains(name))
                    _availableTypeList.add(name);
            }
        }
    }
    private void refreshAllowedPorts(boolean all) {
        _allowableTypeList.removeAll();
        if (all) {
            LinkedList allowablePorts =
                ((ComponentVisualization) _style).getAllowedPortTypes();
            Collections.sort(allowablePorts);
            Iterator iter = allowablePorts.iterator();
            while (iter.hasNext()) {
                String portType = (String) iter.next();
                if (portType.length() > 0)
                    _allowableTypeList.add(portType);
            }
        }
    }
    private void refreshAllowedRoles(boolean all) {
        _allowableTypeList.removeAll();
        if (all) {
            LinkedList allowableRoles =
                ((ConnectorVisualization) _style).getAllowedRoleTypes();
            Collections.sort(allowableRoles);
            Iterator iter = allowableRoles.iterator();
            while (iter.hasNext()) {
                String roleType = (String) iter.next();
                if (roleType.length() > 0)
                    _allowableTypeList.add(roleType);
            }
        }
    }

    private void createPortAlignmentGroup(Composite parent, FormToolkit toolkit, ExpansionAdapter expansionAdapter) {
    	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
        expandComp.setExpanded (false);
       	toolkit.createCompositeSeparator(expandComp);
        expandComp.setText("PortAlignment");
        expandComp.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	Composite _portAlignmentGroup = toolkit.createComposite (expandComp, SWT.NONE);
    	_portAlignmentGroup.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	expandComp.setClient(_portAlignmentGroup);
    	
       
        GridLayout sggl = new GridLayout(3, false);
        _portAlignmentGroup.setLayout(sggl);
        _alignPortTypeTable =
            new PortAlignmentList(_portAlignmentGroup, SWT.NONE);
        GridData pard = new GridData(GridData.FILL_BOTH);
        pard.grabExcessHorizontalSpace = true;
        _alignPortTypeTable.setLayoutData(pard);
        _alignPortTypeTable.setSize(new Point(300, 80));
        _alignPortTypeTable.setInput(_portAlignmentRules);
        Composite buttons = toolkit.createComposite(_portAlignmentGroup, SWT.NONE);
        buttons.setLayout(new GridLayout(1, true));
        _deleteFromAlignPortButton = toolkit.createButton(buttons, ">>", SWT.PUSH | SWT.ARROW_RIGHT);
        _deleteFromAlignPortButton
            .addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                java.util.List deleted =
                    _alignPortTypeTable.deleteSelectedRules();
                for (Iterator iter = deleted.iterator(); iter.hasNext();) {
                    String sel = (String) iter.next();
                    _alignPortTypeList.add(sel);
                }
                sortList(_alignPortTypeList);
            }
        });
        _addToAlignPortButton = toolkit.createButton(buttons, "<<", SWT.PUSH | SWT.ARROW_LEFT);
        _addToAlignPortButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                String[] selections = _alignPortTypeList.getSelection();
                for (int i = 0; i < selections.length; i++) {
                    String sel = selections[i];
                    PlacementRule rule = new PlacementRule();
                    rule._type = sel;
                    if (!_portAlignmentRules.contains(rule)) {
                        _portAlignmentRules.add(rule);
                        _alignPortTypeList.remove(sel);
                    }
                }
                _alignPortTypeTable.setInput(_portAlignmentRules);
            }
        });
        _alignPorts = toolkit.createComposite(_portAlignmentGroup, SWT.NONE);
        _alignPorts.setLayout(new GridLayout(1, false));
        Label availablePortLabel = toolkit.createLabel(_alignPorts, "Available Types:", SWT.NONE);
        pard = new GridData(GridData.FILL_HORIZONTAL);
        availablePortLabel.setLayoutData(pard);
        Composite alignPortContainer = toolkit.createComposite(_alignPorts, SWT.NONE);
        _alignPortTypeList =
            new List(
                alignPortContainer,
                SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.MULTI);
        pard = new GridData(GridData.FILL_BOTH);
        _alignPortTypeList.setLayoutData(pard);
        _alignPortTypeList.setSize(150, 80);

        _onlyButton = toolkit.createButton(_alignPorts, "Show allowable ports only", SWT.RADIO);
        pard = new GridData(GridData.VERTICAL_ALIGN_END);
        _onlyButton.setLayoutData(pard);
        _onlyButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                if (_onlyButton.getSelection())
                    refreshPortsInAlignmentGroup(
                        !((ComponentVisualization) _style).restrictPortTypes());
            }
        });
        _allButton = toolkit.createButton(_alignPorts, "Show all port types", SWT.RADIO);
        pard = new GridData(GridData.VERTICAL_ALIGN_END);
        _allButton.setLayoutData(pard);
        _allButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                if (_allButton.getSelection())
                    refreshPortsInAlignmentGroup(true);
            }
        });
        refreshPortsInAlignmentGroup(true);
        _allButton.setSelection(true);
    }

    private void createRoleRouteGroup(
        final AcmeEnumeration roles,
        Composite parent, FormToolkit toolkit, ExpansionAdapter expansionAdapter) {
    	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
        expandComp.setExpanded (false);
       	toolkit.createCompositeSeparator(expandComp);
        expandComp.setText("Role Route");
        expandComp.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	Composite roleRouteGroup = toolkit.createComposite (expandComp, SWT.NONE);
    	roleRouteGroup.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	expandComp.setClient(roleRouteGroup);
    	
        
        GridLayout sggl = new GridLayout(3, false);
        roleRouteGroup.setLayout(sggl);
        _routeRoleTypeTable = new RoleRouteList(roleRouteGroup, SWT.NONE);
        GridData pard = new GridData(GridData.FILL_BOTH);
        pard.grabExcessHorizontalSpace = true;
        pard.grabExcessVerticalSpace = true;
        _routeRoleTypeTable.setLayoutData(pard);
//        _routeRoleTypeTable.setSize(new Point(200, 200));
        _routeRoleTypeTable.setInput(_roleRouteRules);
        Composite centerPanel = toolkit.createComposite(roleRouteGroup, SWT.NONE);
        GridLayout gl = new GridLayout(1, false);
        gl.verticalSpacing = 20;
        centerPanel.setLayout(gl);
        //        Composite upperPadding = new Composite(centerPanel, SWT.BORDER);
        //        upperPadding.setSize(20, 0);
        //        Composite topButton = new Composite(centerPanel, SWT.BORDER);
        //        topButton.setLayoutData(new GridData(GridData.FILL_BOTH));
        //        Composite bottomButton = new Composite(centerPanel, SWT.BORDER);
        //        bottomButton.setLayoutData(new GridData(GridData.FILL_BOTH));
        //        Composite downPadding = new Composite(centerPanel, SWT.BORDER);
        //        topButton.setLayout(new GridLayout(1, true));
        //        bottomButton.setLayout(new GridLayout(1, true));
        _deleteFromRouteRoleButton = toolkit.createButton(centerPanel, ">>", SWT.PUSH);
        _deleteFromRouteRoleButton.setLayoutData(
            new GridData(
                GridData.HORIZONTAL_ALIGN_CENTER
                    | GridData.VERTICAL_ALIGN_CENTER));
        _deleteFromRouteRoleButton
            .addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                java.util.List deleted =
                    _routeRoleTypeTable.deleteSelectedRules();
                for (Iterator iter = deleted.iterator(); iter.hasNext();) {
                    String sel = (String) iter.next();
                    _routeRoleTypeList.add(sel);
                }
                sortList(_routeRoleTypeList);
            }
        });
        _addToRouteRoleButton = toolkit.createButton(centerPanel, "<<", SWT.PUSH);
        _addToRouteRoleButton.setLayoutData(
            new GridData(
                GridData.HORIZONTAL_ALIGN_CENTER
                    | GridData.VERTICAL_ALIGN_CENTER));
        _addToRouteRoleButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                String[] selections = _routeRoleTypeList.getSelection();
                for (int i = 0; i < selections.length; i++) {
                    String sel = selections[i];
                    ConnectorVisualization.RoleRouteRule rule =
                        new ConnectorVisualization.RoleRouteRule();
                    rule._roleType = sel;
                    if (!_roleRouteRules.contains(rule)) {
                        _roleRouteRules.add(rule);
                        _routeRoleTypeList.remove(sel);
                    }
                }
                _routeRoleTypeTable.setInput(_roleRouteRules);
            }
        });
        Composite routeRoles = toolkit.createComposite(roleRouteGroup, SWT.NONE);
        gl = new GridLayout(1, false);
        routeRoles.setLayout(gl);
        routeRoles.setLayoutData(new GridData(GridData.FILL_BOTH));
        Label availableRoleLabel = toolkit.createLabel(routeRoles, "Available Types:", SWT.NONE);
        // Don't put a List in GridLayout
        Composite roleTypeContainer = toolkit.createComposite(routeRoles, SWT.NONE);
        _routeRoleTypeList =
            new List(
                roleTypeContainer,
                SWT.BORDER | SWT.V_SCROLL | SWT.H_SCROLL | SWT.MULTI);
        pard = new GridData(GridData.FILL_BOTH);
        _routeRoleTypeList.setLayoutData(pard);
        _routeRoleTypeList.setSize(150, 150);
        _onlyButton = toolkit.createButton(routeRoles, "Show allowable roles only", SWT.RADIO);
        pard = new GridData(GridData.VERTICAL_ALIGN_END);
        _onlyButton.setLayoutData(pard);
        _onlyButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                if (_onlyButton.getSelection())
                    refreshRolesInRouteGroup(
                        !((ConnectorVisualization) _style).restrictRoleTypes());
            }
        });
        _allButton = toolkit.createButton(routeRoles, "Show all role types", SWT.RADIO);
        _allButton.setText("Show all role types");
        pard = new GridData(GridData.VERTICAL_ALIGN_END);
        _allButton.setLayoutData(pard);
        _allButton.setSelection(true);
        _allButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                if (_allButton.getSelection())
                    refreshRolesInRouteGroup(true);
            }
        });
        refreshRolesInRouteGroup(true);
        _allButton.setSelection(true);
    }
    private void createRoleNumberGroup(
        AcmeConnectorType type,
        Composite parent, FormToolkit toolkit, ExpansionAdapter expansionAdapter) {
    	 Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
         expandComp.setExpanded (false);
        	toolkit.createCompositeSeparator(expandComp);
         expandComp.setText("Connector options");
         expandComp.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
     	Composite roleNumberGroup = toolkit.createComposite (expandComp, SWT.NONE);
     	roleNumberGroup.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
     	expandComp.setClient(roleNumberGroup);
        GridData sgld = new GridData(GridData.FILL_HORIZONTAL);
        roleNumberGroup.setLayoutData(sgld);
        roleNumberGroup.setLayout(new GridLayout(2, false));
        Label roleNumberLabel = toolkit.createLabel(roleNumberGroup, "Number of default generic roles:", SWT.NONE);
        _roleNumber = new CCombo(roleNumberGroup, SWT.DROP_DOWN | SWT.BORDER) ;
        if (type != null) {
            AcmeEnumeration roles = type.getRoles();
            if (roles != null && roles.moreItems()) {
                _roleNumber.setEnabled(false);
                roleNumberLabel.setEnabled(false);
                roleNumberGroup.setEnabled(false);
                return;
            }
        }
        final int MAX_ROLE_NUMBER = 100;
        String[] roleNumbers = new String[MAX_ROLE_NUMBER + 1];
        for (int i = 0; i <= MAX_ROLE_NUMBER; i++) {
            roleNumbers[i] = Integer.toString(i);
        }
        _roleNumber.setItems(roleNumbers);
        _roleNumber.select(
            ((ConnectorVisualization) _style).getDefaultRoleCount());
        _roleNumber.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                ((ConnectorVisualization) _style).setDefaultRoleCount(
                    _roleNumber.getSelectionIndex());
            }
        });
        
    	final Button _autoCenterButton = toolkit.createButton (roleNumberGroup, "If a connector has two roles, automatically straighten connector", SWT.CHECK);
    	sgld = new GridData ();
    	sgld.horizontalSpan = 2;
    	_autoCenterButton.setLayoutData(sgld);
    	boolean autoCenter = ((ConnectorVisualization )_style).autoCenterBinaryConnectors();
    	_autoCenterButton.setSelection (autoCenter);
    	_autoCenterButton.addSelectionListener (new SelectionAdapter () {
    		
			/**
			 * @param e
			 * @see org.eclipse.swt.events.SelectionAdapter#widgetSelected(org.eclipse.swt.events.SelectionEvent)
			 **/
			public void widgetSelected(SelectionEvent e) {
				boolean autoCenter = _autoCenterButton.getSelection();
				((ConnectorVisualization )_style).setAutoCenterBinaryConnectors (autoCenter);
			}
    	});

    }
    private void createImageGroup(Composite parent, FormToolkit toolkit, IExpansionListener expansionAdapter) {
    	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE);
    	toolkit.createCompositeSeparator(expandComp);
       	expandComp.addExpansionListener(expansionAdapter);
    	expandComp.setExpanded(false);
    	expandComp.setText ("Image");
        Composite imageGroup = toolkit.createComposite(expandComp, SWT.NONE);
        expandComp.setClient(imageGroup);
        imageGroup.setLayout(new GridLayout(2, false));
        GridData igld = new GridData();
//        igld.horizontalSpan = GridData.FILL_HORIZONTAL;
        igld.horizontalAlignment = GridData.FILL;
        imageGroup.setLayoutData(igld);

        _chooseImageButton = toolkit.createButton (imageGroup, "Choose image...", SWT.PUSH);
        _chooseImageButton.setLayoutData(
            new GridData(GridData.HORIZONTAL_ALIGN_FILL));
        _chooseImageButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                chooseImage();
            }
        });
        _imageLabel = toolkit.createLabel(imageGroup, "", SWT.BORDER);
        _imageLabel.setSize(100, 100);
        GridData ilgd = new GridData();
        ilgd.verticalSpan = GridData.FILL;
        _imageLabel.setLayoutData(ilgd);
        Image image = _style.getIcon();
        if (image == null) {
            _imageLabel.setImage(new Image(Display.getCurrent(), 100, 100));
        }
        if (image != null) {
            _imageLabel.setImage(image);
        }
        _clearImageButton = toolkit.createButton (imageGroup, "Clear Image", SWT.PUSH);
        _clearImageButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
                if (!_clearDecorationButton.getSelection()) {
                    _style.setIcon((Image) null);
                    _imageLabel.setImage(null);
                    updatePreview();
                }
            }
        });
        Button setSizeToIcon = toolkit.createButton(imageGroup, "Set size to icon", SWT.PUSH);
        setSizeToIcon.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
                if (_style.getIcon() != null) {

                    int width = _style.getIcon().getBounds().width;
                    _defaultWidthText.setText(new Integer(width).toString());
                    int height = _style.getIcon().getBounds().height;
                    _defaultHeightText.setText(new Integer(height).toString());
                    _style.setSize(new Dimension(width, height));
                    updatePreview();
                }
            }
        });

        expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE);
        toolkit.createCompositeSeparator(expandComp);
      	expandComp.addExpansionListener(expansionAdapter);
        expandComp.setExpanded(false);
        expandComp.setText ("Decorations");
        Composite decorationGroup = toolkit.createComposite (expandComp, SWT.NONE);
        expandComp.setClient(decorationGroup);
        decorationGroup.setLayout(new FillLayout());
        GridData sgld = new GridData();
//        sgld.horizontalSpan = GridData.FILL_HORIZONTAL;
        sgld.horizontalAlignment = GridData.FILL;
        decorationGroup.setLayoutData(sgld);
        _imageSelectionGroup =
            new ImageSelectionGroup(decorationGroup, _style, this, toolkit);
        _clearDecorationButton = toolkit.createButton(decorationGroup, "Clear images", SWT.CHECK);
        _clearDecorationButton.setToolTipText(
            "Selecting this means pressing the image buttons clears the image");
        _clearDecorationButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
                _imageSelectionGroup.setClearMode(
                    _clearDecorationButton.getSelection());

            }
        });
        //     ;   _imageLabel = new Label(imageGroup, SWT.NONE);
        //        _imageLabel.setSize(100, 100);
        //        Image image = _style.getImage(1,1);
        //        if (image != null) {
        //            _imageLabel.setImage(image);
        //        }

    }

    protected void chooseImage() {
        FileDialog dialog =
            new FileDialog(Display.getDefault().getActiveShell(), SWT.OPEN);
        dialog.setFilterExtensions(new String[] { "*.gif", "*.jpg", "*.png" });
        String filename = dialog.open();
        if (filename != null) {
            _style.setIcon(filename);
            _imageLabel.setImage(_style.getIcon());
            updatePreview();
        }
    }

    /**
     * Creates the shape page that contains controls to edit
     * the shape of the style.
     * @param parent Where the page will go
     * @param toolkit
     */
    public Composite createShapePage(Composite parent, FormToolkit toolkit) {
    	final ScrolledForm thePage = toolkit.createScrolledForm(parent);
    	thePage.setText("Visualization Shape");
        Composite thePageComposite = thePage.getBody();
        GridLayout pageLayout = new GridLayout(1, false);
//        RowLayout pageLayout = new RowLayout (SWT.VERTICAL);
        thePageComposite.setLayout(pageLayout);
        Label descLabel = toolkit.createLabel(thePageComposite, "Specify the shape and display properties to use.",SWT.LEFT);
        GridData descData = new GridData();
        descData.horizontalSpan = GridData.FILL_HORIZONTAL;
        descData.horizontalIndent = 25;
        descLabel.setLayoutData(descData);
        ExpansionAdapter expansionAdapter = new ExpansionAdapter() {
    		public void expansionStateChanged(ExpansionEvent e) {
    			thePage.reflow(true);
    		}
    	};
        // The shape selection group
        createShapeGroup(thePageComposite, toolkit, expansionAdapter);
        createSizeGroup(thePageComposite, toolkit, expansionAdapter);
        // The size settings
        // The color and other options group
        createDetailsGroup(thePageComposite, toolkit, expansionAdapter);
        createImageGroup(thePageComposite, toolkit, expansionAdapter);
        //	    thePage.setSize (400, 400);
        return thePage;
    }
    /**
     * Creates the port page that contains controls to edit
     * the port style.
     * @param parent Where the page will go
     * @param toolkit
     */
    private Composite createPortPage(Composite parent, FormToolkit toolkit) {
    	final ScrolledForm thePage = toolkit.createScrolledForm(parent);
    	thePage.setText("Port Layout");
        Composite thePageComposite = thePage.getBody();
        GridLayout pageLayout = new GridLayout(1, false);
//        RowLayout pageLayout = new RowLayout (SWT.VERTICAL);
        thePageComposite.setLayout(pageLayout);
        Label descLabel = toolkit.createLabel(thePageComposite, "Specify how a port lays itself out",SWT.LEFT);
        GridData descData = new GridData();
        descData.horizontalSpan = GridData.FILL_HORIZONTAL;
        descData.horizontalIndent = 25;
        descLabel.setLayoutData(descData);
      
       
        // The rotation and alignment selection group
        createPortGroup(thePageComposite, toolkit);
        return thePage;
    }
    /**
     * Creates the port page that contains controls to edit
     * the port style.
     * @param parent Where the page will go
     * @param toolkit
     */
    private Composite createComponentPage(
        Composite parent,
        AcmeComponentType type, FormToolkit toolkit) {
    	final ScrolledForm thePage = toolkit.createScrolledForm(parent);
    	thePage.setText("Port Layout Policy");
        Composite thePageComposite = thePage.getBody();
        GridLayout pageLayout = new GridLayout(1, false);
//        RowLayout pageLayout = new RowLayout (SWT.VERTICAL);
        thePageComposite.setLayout(pageLayout);
        Label descLabel = toolkit.createLabel(thePageComposite, "Specify how components should layout ports.",SWT.LEFT);
        GridData descData = new GridData();
        descData.horizontalSpan = GridData.FILL_HORIZONTAL;
        descData.horizontalIndent = 25;
        descLabel.setLayoutData(descData);
        ExpansionAdapter expansionAdapter = new ExpansionAdapter() {
    		public void expansionStateChanged(ExpansionEvent e) {
    			thePage.reflow(true);
    		}
    	};
        
        AcmeEnumeration ports =
            ((AcmeFamily) type.getParent())
                .getAvailableTypes()
                .castToAcmeSetVector()
                .enumerate(new acmeLib.core.filters.PortTypeEnumFilter());
        //Enumeration ports = type.getPorts();
        createPortRestrictionGroup(thePageComposite, ports, toolkit, expansionAdapter);
        this._portAlignmentRules =
            ((ComponentVisualization) _style).getPortPlacementRules();
        createPortAlignmentGroup(thePageComposite, toolkit, expansionAdapter);
        return thePage;
    }
    /**
     * Creates the port page that contains controls to edit
     * the port style.
     * @param parent Where the page will go
     * @param toolkit
     */
    private Composite createConnectorPage(
        Composite parent,
        AcmeConnectorType type, FormToolkit toolkit) {
    	final ScrolledForm thePage = toolkit.createScrolledForm(parent);
    	thePage.setText("Role Layout Policy");
        Composite thePageComposite = thePage.getBody();
        GridLayout pageLayout = new GridLayout(1, false);
//        RowLayout pageLayout = new RowLayout (SWT.VERTICAL);
        thePageComposite.setLayout(pageLayout);
        Label descLabel = toolkit.createLabel(thePageComposite, "Specify how connectors should layout roles.",SWT.LEFT);
        GridData descData = new GridData();
        descData.horizontalSpan = GridData.FILL_HORIZONTAL;
        descData.horizontalIndent = 25;
        descLabel.setLayoutData(descData);
        ExpansionAdapter expansionAdapter = new ExpansionAdapter() {
    		public void expansionStateChanged(ExpansionEvent e) {
    			thePage.reflow(true);
    		}
    	};
        
        // The left column
        
        AcmeEnumeration roles = new AcmeEnumeration ();
        if (type != null) {
        	roles =
	            ((AcmeFamily) type.getParent())
	                .getAvailableTypes()
	                .castToAcmeSetVector()
	                .enumerate(new acmeLib.core.filters.RoleTypeEnumFilter());
        }
        else {
        	java.util.List r = new ArrayList ();
        	r.add (ConstantTypes.ROLE_BASE_TYPE);
        	roles = Helper.convertCollectionToAcmeEnumeration(r);
        }
        this._roleRouteRules =
            ((ConnectorVisualization) _style).getRoleRouteRules();
        createRoleNumberGroup(type, thePageComposite, toolkit, expansionAdapter);
        createRoleRestrictionGroup(thePageComposite, roles, toolkit, expansionAdapter);
        createRoleRouteGroup(roles, thePageComposite, toolkit, expansionAdapter);
        return thePage;
    }
    /**
     * Creates the role page that contains controls to edit
     * the role style.
     * @param parent Where the page will go
     * @param toolkit
     */
    private Composite createRolePage(Composite parent, FormToolkit toolkit) {
    	final ScrolledForm thePage = toolkit.createScrolledForm(parent);
    	thePage.setText("Role Layout");
        Composite thePageComposite = thePage.getBody();
        GridLayout pageLayout = new GridLayout(1, false);
//        RowLayout pageLayout = new RowLayout (SWT.VERTICAL);
        thePageComposite.setLayout(pageLayout);
        Label descLabel = toolkit.createLabel(thePageComposite, "Specify how a role lays itself out",SWT.LEFT);
        GridData descData = new GridData();
        descData.horizontalSpan = GridData.FILL_HORIZONTAL;
        descData.horizontalIndent = 25;
        descLabel.setLayoutData(descData);
        // The rotation selection group
        createRoleGroup(thePageComposite, toolkit);
        return thePage;
    }
    /**
     * Creates the group to allow choosing of port features. Includes components to:
     * a) indicate whether the port should rotate around the component center
     * b) indicate whether the port should align with the attached role
     * @param parent the element to put these components into
     * @param toolkit
     */
    private void createPortGroup(Composite parent, FormToolkit toolkit) {
       
        _rotatePortButton = toolkit.createButton(parent, "Rotate Port around Component Center", SWT.CHECK);
        boolean shouldRotatePort = ((PortVisualization) _style).rotatePort();
        _rotatePortButton.setSelection(shouldRotatePort);
        _rotatePortButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                boolean selection = _rotatePortButton.getSelection();
                ((PortVisualization) _style).setRotatePort(selection);
                _orientationCombo.setEnabled(selection);
                updatePreview();
            }
        });

        Composite oComp = toolkit.createComposite (parent);
        oComp.setLayout (new GridLayout (2, false));
        Label orientationLabel = toolkit.createLabel(oComp, "Orientation:", SWT.NONE);
        _orientationCombo = new CCombo(oComp, SWT.DROP_DOWN | SWT.BORDER);
        _orientationCombo.setBackground(toolkit.getColors().getBackground());
        _orientationCombo.setItems(_orientations);
        _orientationCombo.select(((PortVisualization) _style).getOrientation());
        _orientationCombo.setEnabled(shouldRotatePort);

        _alignPortButton = toolkit.createButton(parent, "Align Port with Attached Role", SWT.CHECK);
        _alignPortButton.setSelection(((PortVisualization) _style).alignPort());
        _alignPortButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                ((PortVisualization) _style).setAlignPort(
                    _alignPortButton.getSelection());
                updatePreview();
            }
        });
    }
    /**
     * Creates the group to allow choosing of role features. Includes components to:
     * a) indicate whether the role should rotate around the component center
     * @param parent the element to put these components into
     * @param toolkit
     */
    private void createRoleGroup(Composite parent, FormToolkit toolkit) {
    	
      
        _rotateRoleButton = toolkit.createButton(parent, "Rotate Role around Connector Center", SWT.CHECK);
        boolean shouldRotate = ((RoleVisualization) _style).rotateRole();
        _rotateRoleButton.setSelection(shouldRotate);
        _rotateRoleButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                boolean selection = _rotateRoleButton.getSelection();
                ((RoleVisualization) _style).setRotateRole(selection);
                _orientationCombo.setEnabled(selection);
                updatePreview();
            }
        });

        Composite oComp = toolkit.createComposite (parent);
        oComp.setLayout(new GridLayout (2, false));
        Label orientationLabel = toolkit.createLabel(oComp, "Orientation:", SWT.NONE);
        _orientationCombo = new CCombo(oComp, SWT.DROP_DOWN | SWT.BORDER);
        _orientationCombo.setBackground(toolkit.getColors().getBackground());
        _orientationCombo.setItems(_orientations);
        _orientationCombo.select(((RoleVisualization) _style).getOrientation());
        _orientationCombo.setEnabled(shouldRotate);
    }
    private Composite createVariantsPage(Composite parent, FormToolkit toolkit) {
    	final ScrolledForm thePage = toolkit.createScrolledForm(parent);
    	thePage.setText("Visualization Variants");
        Composite thePageComposite = thePage.getBody();
        GridLayout pageLayout = new GridLayout(3, false);
//        RowLayout pageLayout = new RowLayout (SWT.VERTICAL);
        thePageComposite.setLayout(pageLayout);
        Label descLabel = toolkit.createLabel(thePageComposite, "Specify the variants for this visualization.",SWT.LEFT);
        GridData descData = new GridData();
        descData.horizontalSpan = GridData.FILL_HORIZONTAL;
        descData.horizontalIndent = 25;
        descLabel.setLayoutData(descData);
        ExpansionAdapter expansionAdapter = new ExpansionAdapter() {
    		public void expansionStateChanged(ExpansionEvent e) {
    			thePage.reflow(true);
    		}
    	};
       
        Label vLabel = toolkit.createLabel(thePageComposite, "Variants:", SWT.NONE);
        Label l2 = toolkit.createLabel(thePageComposite, "", SWT.NONE);
        Label pLabel = toolkit.createLabel(thePageComposite, "Preview:", SWT.NONE);
        _variantsList = new List(thePageComposite, SWT.NONE | SWT.H_SCROLL | SWT.BORDER | SWT.FLAT);
        //        _variantsList.setSize (100, 200);
        GridData vlgd =
            new GridData(GridData.GRAB_HORIZONTAL | GridData.GRAB_VERTICAL);
        vlgd.grabExcessHorizontalSpace = true;
        vlgd.horizontalAlignment = GridData.FILL;
        vlgd.verticalAlignment = GridData.FILL;
        _variantsList.setLayoutData(vlgd);
        _variantsList.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                if (_variantsList.getSelectionCount() == 1) {
                    _modifyVariantButton.setEnabled(true);
                    updateVariantsPreview();
                } else {
                    _modifyVariantButton.setEnabled(false);
                    updateVariantsPreview();
                }
            }
        });
        Composite buttons = toolkit.createComposite(thePageComposite, SWT.NONE);
        buttons.setLayout(new FillLayout(SWT.VERTICAL));
        _newVariantButton = toolkit.createButton(buttons, "New...", SWT.PUSH);
        _newVariantButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                EditVariantDialog dialog =
                    new EditVariantDialog(
                        Display.getDefault().getActiveShell(),
                        _style,
                        (AcmeElement) _type);
                if (dialog.open() == IDialogConstants.OK_ID) {
                    VisVariant var = dialog.getVariant();
                    _variantsList.add(var._name);
                    _variantsList.setData(var._name, var);
                }
            }
        });
        _modifyVariantButton = toolkit.createButton(buttons, "Modify...", SWT.PUSH);
        _modifyVariantButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                if (_variantsList.getSelection().length == 1) {
                    String oldName = _variantsList.getSelection()[0];
                    VisVariant variant =
                        (VisVariant) _variantsList.getData(
                            _variantsList.getSelection()[0]);
                    EditVariantDialog dialog =
                        new EditVariantDialog(
                            Display.getDefault().getActiveShell(),
                            variant,
                            (AcmeElement) _type);
                    if (dialog.open() == IDialogConstants.OK_ID) {
                        if (!variant._name.equals(oldName)) {
                            int index = _variantsList.indexOf(oldName);
                            _variantsList.remove(index);
                            _variantsList.add(variant._name, index);
                            _variantsList.setData(variant._name, variant);
                        }
                    }
                }
            }
        });
        _modifyVariantButton.setEnabled(false);
        _deleteVariantButton = toolkit.createButton(buttons, "Delete", SWT.PUSH);
        _deleteVariantButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                String[] selections = _variantsList.getSelection();
                for (int i = 0; i < selections.length; i++) {
                    // Actual variant will be changed in exportStyle
                    _variantsList.remove(selections[i]);
                }
            }
        });
        _orderUpButton = toolkit.createButton(buttons, "Up", SWT.PUSH | SWT.ARROW_UP);
        //       _orderUpButton.setImage(ImageManager.getImage (ImageManager.ID_UP));
        _orderUpButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                int selection = _variantsList.getSelectionIndex();
                if (selection <= 0)
                    return;
                String name = _variantsList.getItem(selection);
                VisVariant var = (VisVariant) _variantsList.getData(name);
                _variantsList.remove(selection);
                _variantsList.add(name, selection - 1);
                _variantsList.setData(name, var);
            }
        });
        _orderDownButton = toolkit.createButton(buttons, "Down", SWT.PUSH | SWT.ARROW_DOWN);
        //        _orderDownButton.setImage (ImageManager.getImage (ImageManager.ID_DOWN));
        _orderDownButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                int selection = _variantsList.getSelectionIndex();
                if (selection == -1
                    || selection == _variantsList.getItemCount() - 1)
                    return;
                String name = _variantsList.getItem(selection);
                _variantsList.remove(selection);
                _variantsList.add(name, selection + 1);
            }
        });
        //        _variantPreviewLabel = new Label (thePage, SWT.NONE);
        initVariantPreview(thePageComposite);
        updateVariantsPreview();
        return thePage;
    }
    /**
     * Fills the image with the color
     * @param image The image to fill
     * @param color The color to fill with
     */
    private void drawImage(Image image, Color color) {
        GC gc = new GC(image);
        gc.setBackground(color);
        Rectangle bounds = image.getBounds();
        gc.fillRectangle(0, 0, bounds.width, bounds.height);
        gc.drawRectangle(0, 0, bounds.width - 1, bounds.height - 1);
        gc.dispose();
    }
    /**
     * Sets all the defaults for the color and line
     * options, based on _style.
     */
    private void setDetailsDefaults() {
        final Image fillImage =
            ImageManager.createImage("EditVis_fill", 30, 12, false);
        final Image outlineImage =
            ImageManager.createImage("EditVis_outline", 30, 12, false);
        drawImage(fillImage, _style.getFillColor());
        drawImage(outlineImage, _style.getOutlineColor());
        _fillColorButton.setImage(fillImage);
        _outlineColorButton.setImage(outlineImage);
        switch (_style.getOutlineStyle()) {
            case SWT.NONE :
                _noLineButton.setSelection(true);
                break;
            case SWT.LINE_SOLID :
                _solidLineButton.setSelection(true);
                break;
            case SWT.LINE_DOT :
                _dotLineButton.setSelection(true);
                break;
            case SWT.LINE_DASH :
                _dashLineButton.setSelection(true);
                break;
            case SWT.LINE_DASHDOT :
                _dashDotLineButton.setSelection(true);
                break;
            case SWT.LINE_DASHDOTDOT :
                _dashDotDotLineButton.setSelection(true);
                break;
            default :
                _noLineButton.setSelection(true);
        }
        _outlineWidthCombo.select(_style.getOutlineWidth() - 1);
    }
    /**
     * Creates the group for entering the color and line details of the shape,
     * allowing the user to:
     * a) Select the fill color
     * b) Select the outline color
     * c) Select the outline style
     * d) Select the width of the line
     * TODO: Fill pattern & role edge and style
     * @param parent The place to put these controls
     * @param toolkit
     * @param expansionAdapter
     */
    private void createDetailsGroup(Composite parent, FormToolkit toolkit, IExpansionListener expansionAdapter) {
    	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
    	toolkit.createCompositeSeparator(expandComp);
    	expandComp.setExpanded (false);
       	expandComp.addExpansionListener(expansionAdapter);
        expandComp.setText("Color/Line Properties");
        Composite detailsGroup = toolkit.createComposite(expandComp, SWT.NONE);
        expandComp.setClient(detailsGroup);
        GridData dgd = new GridData();
        dgd.horizontalAlignment = GridData.FILL;
        detailsGroup.setLayoutData(dgd);
        GridLayout rcl = new GridLayout(2, false);
        detailsGroup.setLayout(rcl);
        Label fillColorLabel = toolkit.createLabel(detailsGroup, "Fill Color:", SWT.NONE);
        Shell shell = fillColorLabel.getShell();
        final ColorDialog fillColorDialog = new ColorDialog(shell);
        final ColorDialog outlineColorDialog = new ColorDialog(shell);
        _fillColorButton = toolkit.createButton (detailsGroup, "", SWT.PUSH);
        _fillColorButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
            	// BugFix#249: Set the selected color of the dialog
            	fillColorDialog.setRGB(_style.getFillColor().getRGB());
                RGB rgb = fillColorDialog.open();
                if (rgb == null)
                    return;
                _style.setFillColor(
                		SWTResourceManager.getColor (rgb));
                Image image = _fillColorButton.getImage();
                drawImage(image, _style.getFillColor());
                _fillColorButton.setImage(image);
                updatePreview();
            }
        });
        Label outlineColorLabel = toolkit.createLabel (detailsGroup, "Outline Color:", SWT.NONE);
        _outlineColorButton = toolkit.createButton (detailsGroup, "", SWT.PUSH);
        _outlineColorButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
            	// BugFix#249:  Set the selected color of the dialog
            	outlineColorDialog.setRGB(_style.getOutlineColor().getRGB());
                RGB rgb = outlineColorDialog.open();
                if (rgb == null)
                    return;
                _style.setOutlineColor(
                		SWTResourceManager.getColor (rgb));
                Image image = _outlineColorButton.getImage();
                drawImage(image, _style.getOutlineColor());
                _outlineColorButton.setImage(image);
                updatePreview();
            }
        });
        Label outlineStyleLabel = toolkit.createLabel (detailsGroup, "Outline style:", SWT.NONE);
        SelectionAdapter buttonSelected = new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                if (((Button) evt.widget).getSelection()
                    && evt.widget.getData() instanceof Integer) {
                    Integer style = (Integer) evt.widget.getData();
                    _style.setOutlineStyle(style.intValue());
                    updatePreview();
                    if (style.intValue() == SWT.NONE) {
                        _outlineWidthCombo.setEnabled(false);
                    } else {
                        _outlineWidthCombo.setEnabled(true);
                    }
                }
            }
        };
        Group outlineGroup = new Group(detailsGroup, SWT.NONE);
        outlineGroup.setBackground(toolkit.getColors().getBackground());
        GridLayout ogl = new GridLayout(2, true);
        outlineGroup.setLayout(new RowLayout (SWT.HORIZONTAL));
        _noLineButton = toolkit.createButton (outlineGroup, "No Line", SWT.RADIO);
//        GridData nlgd = new GridData();
//        nlgd.horizontalSpan = 2;
//        _noLineButton.setLayoutData(nlgd);
        _noLineButton.setData(new Integer(SWT.NONE));
        _noLineButton.addSelectionListener(buttonSelected);
        _solidLineButton =
            createLineStyleButton(toolkit, outlineGroup, SWT.LINE_SOLID, buttonSelected);
        _dotLineButton =
            createLineStyleButton(toolkit,outlineGroup, SWT.LINE_DOT, buttonSelected);
        _dashLineButton =
            createLineStyleButton(toolkit,outlineGroup, SWT.LINE_DASH, buttonSelected);
        _dashDotLineButton =
            createLineStyleButton(toolkit,
                outlineGroup,
                SWT.LINE_DASHDOT,
                buttonSelected);
        _dashDotDotLineButton =
            createLineStyleButton(toolkit,
                outlineGroup,
                SWT.LINE_DASHDOTDOT,
                buttonSelected);
        Label outlineWidthLabel = toolkit.createLabel (detailsGroup, "Outline Width:", SWT.NONE);
        _outlineWidthCombo = new Combo(detailsGroup, SWT.DROP_DOWN);
        _outlineWidthCombo.setBackground(toolkit.getColors().getBackground());
        _outlineWidthCombo.setItems(new String[] { "1", "2", "3", "4", "5" });
        GridData owgd = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
        _outlineWidthCombo.setLayoutData(owgd);
        _outlineWidthCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _style.setOutlineWidth(
                    _outlineWidthCombo.getSelectionIndex() + 1);
                updatePreview();
            }
        });
    }
    /**
     * Draws a line image in a particular style, for use as a depiction
     * in a button.
     * @param img The image to draw a line into
     * @param style The style of the line to draw
     * @param color
     */
    private static void drawLineImage(Image img, int style, Color color) {
        GC gc = new GC(img);
        gc.setBackground(color);
        gc.fillRectangle(img.getBounds());
        gc.setLineStyle(style);
        gc.drawLine(1, 6, 49, 6);
        gc.dispose();
    }
    /**
     * Creates a button to choose a line style, and adds the associated selection listener
     * @param toolkit
     * @param group Where the button is to go.
     * @param style The style of the button
     * @param listener The selection listener to add to the button
     * @return The new button
     */
    public static Button createLineStyleButton(
        FormToolkit toolkit, Composite group,
        int style,
        SelectionListener listener) {
        Button b = toolkit.createButton(group, "", SWT.RADIO);
        String key = "__linestyle_" + style;
        Image img = ImageManager.createImage(key, 50, 12, false);
        drawLineImage(img, style, toolkit.getColors().getBackground());
        b.setImage(img);
        b.setData(new Integer(style));
        b.addSelectionListener(listener);
        return b;
    }
    private void initPreview() {
        _previewImageCanvas = new FigureCanvas(_previewComp);
        GridData gd = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
        _previewImageCanvas.setLayoutData(gd);
        _previewImageCanvas.setBackground(ColorConstants.white);
        _previewPanel = new ScrollPane();
        FrameBorder fb = new FrameBorder("Preview");
        TitleBarBorder tbb = (TitleBarBorder) fb.getInnerBorder();
        tbb.setBackgroundColor(ColorConstants.white);
        tbb.setTextColor(ColorConstants.darkGray);
        FontData fontData = new FontData();
        fontData.setStyle(SWT.BOLD);
        tbb.setFont(new Font(null, fontData));
        _previewPanel.setBorder(fb);
        _previewPanel.setPreferredSize(510, 120);
        _imgFig = new ImageFigure();
        _imgFig.setBounds(_previewPanel.getClientArea());
        _previewPanel.setContents(_imgFig);
        _previewImageCanvas.setContents(_previewPanel);
    }
    private void initVariantPreview(Composite parent) {
        FigureCanvas variantPreviewCanvas = new FigureCanvas(parent);
        variantPreviewCanvas.setBackground(ColorConstants.white);
        ScrollPane variantPane = new ScrollPane();
        variantPane.setPreferredSize(200, 100);
        _variantImage = new ImageFigure();
        _variantImage.setBounds(variantPane.getBounds());
        variantPane.setContents(_variantImage);
        variantPreviewCanvas.setContents(variantPane);
    }
    /**
     * Updates the preview image label based on the visual style.
     */
    public void updatePreview() {
        Image thePreview = _style.getPreviewImage(_type).getImage();
        _imgFig.setImage(thePreview);
        _previewImageCanvas.redraw();
    }
    private void updateVariantsPreview() {
        String[] selections = _variantsList.getSelection();
        if (selections.length != 1) {
            _variantImage.setImage(NO_VARIANT);
        } else {
            VisVariant var = (VisVariant) _variantsList.getData(selections[0]);
            ElementVisualization vis = var._style.copy();
            vis = vis.unify(_style);
            Image preview = vis.getPreviewImage(_type).getImage ();
            _variantImage.setImage(preview);
            //             _variantPreviewLabel.setImage (preview);      
        }
    }
    /**
     * Updates each of the controls in the dialog box to reflect
     * the values in the style being managed by this class
     */
    private void importStyle() {
        // Set shape defaults
        _shapeCombo.select(_style.getShape() - 1);
        //		_stockShapeButton.setSelection (true);
        _replicateShapeButton.setSelection(_style.getReplicateShape());
        //		_imageShapeButton.setSelection (false);
        //		_chooseImageButton.setEnabled (false);
        //		_stockShapeButton.setSelection (true);
        //        if (_style.getImage(1,1) != null) {
        //            _imageLabel.setImage(_style.getImage(1,1));
        //        }
        switch (_style.getShape()) {
            case Shape.TRIANGLE :
            case Shape.FILTER :
                _directionCombo.setEnabled(true);
                break;
            default :
                _directionCombo.setEnabled(false);
        }
        // Size defaults
        _preserveAspectRatioButton.setSelection(
            _style.mustPreserveAspectRatio());
        _defaultWidthText.setText(
            new Integer(_style.getSize().width).toString());
        _defaultHeightText.setText(
            new Integer(_style.getSize().height).toString());
        _widthResizeableButton.setSelection(_style.canResizeWidth());
        _heightResizeableButton.setSelection(_style.canResizeHeight());
        _resizeToFitButton.setSelection (_style.shouldResizeToFitLabel());
        setDetailsDefaults();
        Font f = _style.getFont();
        if (f != null) {
            _fontButton.setFont(_style.getFont());
        }
        final Image fontImage =
            ImageManager.createImage("EditStyle_fontColor", 30, 12, false);
        drawImage(fontImage, _style.getTextColor());
        _fontColorButton.setImage(fontImage);
        _propertyCombo.setEnabled(false);
        switch (_style.getLabelSource()) {
            case LabelOptions.NOLABEL :
                _noLabelButton.setSelection(true);
                break;
            case LabelOptions.ACMEID :
                _identifierButton.setSelection(true);
                break;
            case LabelOptions.ACMEIDANDTYPE :
                _idAndTypeButton.setSelection(true);
                break;
            case LabelOptions.ACMETYPE :
                _typeButton.setSelection(true);
                break;
            case LabelOptions.ACMEPROPERTYVALUE :
                _propertyButton.setSelection(true);
                _propertyCombo.setText(_style.getLabelProperty());
                _propertyCombo.setEnabled(true);
                break;
            case LabelOptions.STATICLABEL :
                _staticLabelButton.setSelection(true);
                _labelText.setEnabled(true);
                _labelText.setText(_style.getLabelValue());
                break;
        }
        _horizAlignCombo.select(_style.getLabelHorizontalAlignment());
        _vertAlignCombo.select(_style.getLabelVerticalAlignment());
        _marginText.setText(new Integer(_style.getLabelMargin()).toString());

        _imageSelectionGroup.importStyle();
        Iterator i = _style.getVariants().iterator();
        while (i.hasNext()) {
            VisVariant v = (VisVariant) i.next();
            _variantsList.add(v._name);
            _variantsList.setData(v._name, v);
        }
    }

    /**
     * Changes the visual style member to reflect the data
    * in the controls in the dialog box.
    */
    private ElementVisualization exportStyle() {
        _style.beginChangeBlock();
        _style.setShape(_shapeCombo.getSelectionIndex() + 1);
        _style.setReplicateShape(_replicateShapeButton.getSelection());
        _style.setPreserveAspectRatio(
            _preserveAspectRatioButton.getSelection());
        Dimension size = _style.getSize();
        int value = -1;
        try {
            value = Integer.parseInt(_defaultWidthText.getText());
        } catch (Exception e) {
        }
        if (value > 0) {
            size.width = value;
        }
        value = -1;
        try {
            value = Integer.parseInt(_defaultHeightText.getText());
        } catch (Exception e) {
        }
        if (value > 0) {
            size.height = value;
        }
        _style.setSize(size);
        _style.setHeightResizeable(_heightResizeableButton.getSelection());
        _style.setWidthResizeable(_widthResizeableButton.getSelection());
        _style.setResizeToFitLabel(_resizeToFitButton.getSelection ());
        if (_noLineButton.getSelection()) {
            _style.setOutlineStyle(SWT.NONE);
        } else if (_solidLineButton.getSelection()) {
            _style.setOutlineStyle(SWT.LINE_SOLID);
        } else if (_dotLineButton.getSelection()) {
            _style.setOutlineStyle(SWT.LINE_DOT);
        } else if (_dashLineButton.getSelection()) {
            _style.setOutlineStyle(SWT.LINE_DASH);
        } else if (_dashDotLineButton.getSelection()) {
            _style.setOutlineStyle(SWT.LINE_DASHDOT);
        } else if (_dashDotDotLineButton.getSelection()) {
            _style.setOutlineStyle(SWT.LINE_DASHDOTDOT);
        }
        if (_staticLabelButton.getSelection()) {
            _style.setLabelValue(_labelText.getText());
        }
        _style.setOutlineWidth(_outlineWidthCombo.getSelectionIndex() + 1);
        _style.clearVariants();
        for (int i = 0; i < _variantsList.getItemCount(); i++) {
            String name = _variantsList.getItem(i);
            VisVariant var = (VisVariant) _variantsList.getData(name);
            _style.addVariant(var);
        }
        if (_style instanceof ComponentVisualization) {
            ComponentVisualization styleComponent =
                (ComponentVisualization) _style;
            if (styleComponent.restrictPortTypes()) {
                LinkedList allowablePortTypes = new LinkedList();
                String[] portTypes = _allowableTypeList.getItems();
                for (int i = 0; i < portTypes.length; i++) {
                    String type = portTypes[i];
                    if (type.length() > 0)
                        allowablePortTypes.add(type);
                }
                styleComponent.setAllowedPortTypes(allowablePortTypes);
            }
            if (styleComponent.forcePortPlacement()) {
                styleComponent.setPortPlacementRules(this._portAlignmentRules);
            }
        } else if (_style instanceof ConnectorVisualization) {
            ConnectorVisualization styleConnector =
                (ConnectorVisualization) _style;
            if (styleConnector.restrictRoleTypes()) {
                LinkedList allowableRoleTypes = new LinkedList();
                String[] roleTypes = _allowableTypeList.getItems();
                for (int i = 0; i < roleTypes.length; i++) {
                    allowableRoleTypes.add(roleTypes[i]);
                }
                styleConnector.setAllowedRoleTypes(allowableRoleTypes);
            }
            if (this._roleRouteRules != null
                && this._roleRouteRules.size() > 0) {
                styleConnector.setRoleRouteRules(this._roleRouteRules);
            }
        } else if (_style instanceof PortVisualization) {
            PortVisualization stylePort = (PortVisualization) _style;
            if (stylePort.rotatePort()) {
                stylePort.setOrientation(_orientationCombo.getSelectionIndex());
            }
        } else if (_style instanceof RoleVisualization) {
            RoleVisualization styleRole = (RoleVisualization) _style;
            if (styleRole.rotateRole()) {
                styleRole.setOrientation(_orientationCombo.getSelectionIndex());
            }
        }

        _style.endChangeBlock();
        return _style;
    }
    /**
     * Creates the controls concerned with setting the font options
     * @param parent Where to put the controls
     * @param expansionAdapter
     * @param toolkit
     */
    private void createFontOptions(Composite parent, FormToolkit toolkit, IExpansionListener expansionAdapter) {
    	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
    	toolkit.createCompositeSeparator(expandComp);
    	expandComp.setExpanded (false);
        expandComp.setText("Font Settings");
        expandComp.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	Composite fontGroup = toolkit.createComposite (expandComp, SWT.NONE);
    	 fontGroup.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	expandComp.setClient(fontGroup);
    	fontGroup.setLayout(new GridLayout (2, false));
		expandComp.addExpansionListener(expansionAdapter);

        final FontDialog fontDialog = new FontDialog(fontGroup.getShell());
        _fontButton = toolkit.createButton(fontGroup, "Choose font...", SWT.PUSH);
        _fontButton.setSize(100, 30);
        GridData fbd = new GridData();
        fbd.horizontalSpan = 2;
        _fontButton.setLayoutData(fbd);
        _fontButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                FontData fd = fontDialog.open();
                if (fd == null)
                    return;
                Font font = new Font(_fontButton.getDisplay(), fd);
                _style.setFont(font);
                _fontButton.setFont(font);
                updatePreview();
            }
        });
        Label fontColorLabel = toolkit.createLabel(fontGroup, "Font Color:", SWT.NONE);
        final ColorDialog fontColorDialog =
            new ColorDialog(fontColorLabel.getShell());
        _fontColorButton = toolkit.createButton(fontGroup, "", SWT.PUSH);
        _fontColorButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent event) {
                RGB rgb = fontColorDialog.open();
                if (rgb == null)
                    return;
                _style.setTextColor(
                		SWTResourceManager.getColor (rgb));
                Image image = _fontColorButton.getImage();
                // Bug Fix# #354, should grab text color, not outline.
                drawImage(image, _style.getTextColor());
                _fontColorButton.setImage(image);
                updatePreview();
            }
        });
    }
    /**
     * Creates the controls to set the label source of the style
     * @param parent Where to put the controls
     * @param expansionAdapter
     * @param toolkit
     */
    private void createLabelSourceOptions(Composite parent, FormToolkit toolkit, IExpansionListener expansionAdapter) {
       	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
       	toolkit.createCompositeSeparator(expandComp);
       	expandComp.setExpanded (false);
        expandComp.setText("Label Source");
        expandComp.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	Composite labelSource = toolkit.createComposite (expandComp, SWT.NONE);
    	 labelSource.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	expandComp.setClient(labelSource);
    	
		expandComp.addExpansionListener(expansionAdapter);
        labelSource.setLayout(new RowLayout (SWT.HORIZONTAL));
        SelectionAdapter buttonSelected = new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                Button selectedButton = ((Button) evt.widget);
				if (selectedButton.getSelection()
                    && evt.widget.getData() instanceof Integer) {
                    Integer source = (Integer) evt.widget.getData();
                    _style.setLabelSource(source.intValue());
                    updatePreview();
                    if (source.intValue()
                        == LabelOptions.ACMEPROPERTYVALUE) {
                        _propertyCombo.setEnabled(true);
                    } else {
                        _propertyCombo.setEnabled(false);
                    }
                    if (source.intValue()
                        == LabelOptions.STATICLABEL) {
                        _labelText.setEnabled(true);
                    } else {
                        _labelText.setEnabled(false);
                    }
                }
				manageSelection (selectedButton, _staticLabelButton);
				manageSelection(selectedButton, _propertyButton);
				manageSelection(selectedButton, _noLabelButton);
				manageSelection(selectedButton, _identifierButton);
				manageSelection(selectedButton, _idAndTypeButton);
				manageSelection(selectedButton, _typeButton);
				
            }

			/**
			 * @param selectedButton
			 */
			private void manageSelection(Button selectedButton, Button other) {
				if (selectedButton.getSelection() && selectedButton != other) other.setSelection(false);
			}
        };
        _noLabelButton =
            createLabelSourceButton(
                labelSource,
                "No label",
                LabelOptions.NOLABEL,
                buttonSelected, toolkit);
        Composite staticComp = new Composite (labelSource, SWT.NONE);
        staticComp.setBackground(toolkit.getColors().getBackground());
        GridLayout gridLayout = new GridLayout (1, false);
        gridLayout.marginHeight = 0;
		staticComp.setLayout(gridLayout);
        _staticLabelButton =
            createLabelSourceButton(
                staticComp,
                "Static label",
                LabelOptions.STATICLABEL,
                buttonSelected, toolkit);
        _labelText = toolkit.createText(staticComp, "", SWT.SINGLE | SWT.BORDER);
        GridData gd = new GridData();
        gd.horizontalIndent = 25;
        _labelText.setLayoutData(gd);
        _labelText.setEnabled(false);
        // fill the value text with the data from the type
        String labelValue = _style.getLabelValue();
        if (labelValue != null)
            _labelText.setText(labelValue);
        _labelText.addModifyListener(new ModifyListener() {
            public void modifyText(ModifyEvent e) {
                _style.setLabelValue(_labelText.getText());
                updatePreview();
            }
        });

        _identifierButton =
            createLabelSourceButton(
                labelSource,
                "Acme identifier",
                LabelOptions.ACMEID,
                buttonSelected, toolkit);
        _typeButton =
            createLabelSourceButton(
                labelSource,
                "Acme type",
                LabelOptions.ACMETYPE,
                buttonSelected, toolkit);
        _idAndTypeButton =
            createLabelSourceButton(
                labelSource,
                "Acme identifier and type",
                LabelOptions.ACMEIDANDTYPE,
                buttonSelected, toolkit);
        Composite propComp = new Composite (labelSource, SWT.NONE);
        propComp.setBackground (toolkit.getColors().getBackground ());
        gridLayout = new GridLayout (1, false);
        gridLayout.marginHeight = 0;
        propComp.setLayout (gridLayout);
        _propertyButton =
            createLabelSourceButton(
                propComp,
                "Acme property value",
                LabelOptions.ACMEPROPERTYVALUE,
                buttonSelected, toolkit);
        _propertyCombo = new CCombo(propComp, SWT.DROP_DOWN | SWT.BORDER);
        _propertyCombo.setBackground(toolkit.getColors ().getBackground ());
        gd = new GridData();
        gd.horizontalIndent = 25;
        _propertyCombo.setLayoutData(gd);
        // fill the property combo with the properties from the type
        if (_type != null) {
        	fillPropertyCombo (_propertyCombo, _type);
            /*AcmeEnumeration properties = ((AcmeElement) _type).getProperties();
            while (properties.moreItems()) {
                AcmeProperty prop = (AcmeProperty) properties.getNextItem();
                _propertyCombo.add(prop.getName());
            }*/
        }
        _propertyCombo.addSelectionListener(new SelectionListener() {
            public void widgetSelected(SelectionEvent evt) {
                int selection = _propertyCombo.getSelectionIndex();
                if (selection != -1) {
                    _style.setLabelProperty(_propertyCombo.getItem(selection));
                } else {
                    _style.setLabelProperty(_propertyCombo.getText());
                }
            }
            public void widgetDefaultSelected(SelectionEvent evt) {
            }
        });
    }

    Set gatherAllProperties(AcmeElementType et) {
		Set s = new TreeSet(new Utils.AcmePropertyNameComparator());
		s.addAll(Helper.convertAcmeEnumerationToList(((AcmeElement) et).getProperties()));
		AcmeEnumeration types = et.getSuperTypes().enumerate();
		while (types.moreItems()) {
			AcmeElementType type = (AcmeElementType) types.getNextItem();
			s.addAll(gatherAllProperties(type));
		}
		return s;
	}
    /**
	 * @param combo
	 * @param _type2
	 */
	private void fillPropertyCombo(CCombo combo, AcmeElementType et) {
		Set s = gatherAllProperties(et);
		for (Iterator iter = s.iterator (); iter.hasNext ();) {
			AcmeProperty prop = (AcmeProperty )iter.next ();
			combo.add(prop.getName ());
		}
	}
	/**
     * Creates the controls to set the alignment options
     * @param parent Where to put the controls
	 * @param expansionAdapter
	 * @param toolkit
     */
    private void createAlignmentOptions(Composite parent, FormToolkit toolkit, IExpansionListener expansionAdapter) {
    	Section expandComp = toolkit.createSection(parent, ExpandableComposite.TWISTIE | ExpandableComposite.EXPANDED);
    	toolkit.createCompositeSeparator(expandComp);
        expandComp.setText("Label Alignment");
        expandComp.setExpanded (false);
        GridData gridData = new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL);
        gridData.horizontalIndent = 25;
		expandComp.setLayoutData(new GridData (GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL));
    	Composite alignmentGroup = toolkit.createComposite (expandComp, SWT.NONE);
    	 alignmentGroup.setLayoutData(gridData);
    	expandComp.setClient(alignmentGroup);
    	
		expandComp.addExpansionListener(expansionAdapter);
        GridLayout agl = new GridLayout(2, false);
        alignmentGroup.setLayout(agl);
        toolkit.createLabel(alignmentGroup, "Horiz. Alignment", SWT.NONE);
        _horizAlignCombo = new Combo(alignmentGroup, SWT.DROP_DOWN | SWT.BORDER);
        _horizAlignCombo.setItems(_horizAlignmentOptions);
        _horizAlignCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _style.setLabelHorizontalAlignment(
                    _horizAlignCombo.getSelectionIndex());
                updatePreview();
            }
        });
        Label vertAlignLabel = toolkit.createLabel (alignmentGroup, "Vert. Alignment:", SWT.NONE);
        _vertAlignCombo = new CCombo(alignmentGroup, SWT.DROP_DOWN);
        _vertAlignCombo.setItems(_vertAlignmentOptions);
        _vertAlignCombo.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _style.setLabelVerticalAlignment(
                    _vertAlignCombo.getSelectionIndex());
                updatePreview();
            }
        });
        Label marginLabel = toolkit.createLabel (alignmentGroup, "Margin:", SWT.NONE);
        _marginText = toolkit.createText(alignmentGroup, "", SWT.SINGLE | SWT.BORDER);
        _marginText.addVerifyListener(Listeners.VERIFY_NUMBER);
        _marginText.addFocusListener(new FocusListener() {
            // update the preview when the text is left
            public void focusLost(FocusEvent evt) {
                if (_marginText.getText().equals("")) {
                    _marginText.setText("0");
                }
                int value = -1;
                try {
                    value = Integer.parseInt(_marginText.getText());
                } catch (Exception e) {
                }
                if (value >= 0) {
                    _style.setLabelMargin(value);
                    updatePreview();
                }
            }
            public void focusGained(FocusEvent evt) {
            }
        });
        _wordwrapButton = toolkit.createButton (alignmentGroup, "Use Word Wrap", SWT.CHECK);
        boolean wordwrap = _style.getWordwrap();
        _wordwrapButton.setSelection(wordwrap);
        _wordwrapButton.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent evt) {
                _style.setWordwrap(_wordwrapButton.getSelection());
                updatePreview();
            }
        });
    }
    /** 
     * Creates the label page in the dialog box, which allows
     * the user to change the label options
     * @param parent Where the page is to be placed
     * @param toolkit
     */
    private Composite createLabelPage(Composite parent, FormToolkit toolkit) {
    	final ScrolledForm thePage = toolkit.createScrolledForm(parent);
    	thePage.setText("Visualization Label Preferences");
        Composite thePageComposite = thePage.getBody();
        GridLayout pageLayout = new GridLayout(1, false);
//        RowLayout pageLayout = new RowLayout (SWT.VERTICAL);
        thePageComposite.setLayout(pageLayout);
        Label descLabel = toolkit.createLabel(thePageComposite, "Specify the shape and display properties to use.",SWT.LEFT);
        GridData descData = new GridData();
        descData.horizontalSpan = GridData.FILL_HORIZONTAL;
        descData.horizontalIndent = 25;
        descLabel.setLayoutData(descData);
        ExpansionAdapter expansionAdapter = new ExpansionAdapter() {
    		public void expansionStateChanged(ExpansionEvent e) {
    			thePage.reflow(true);
    		}
    	};
        
        createFontOptions(thePageComposite, toolkit, expansionAdapter);
        createLabelSourceOptions(thePageComposite, toolkit, expansionAdapter);
        createAlignmentOptions(thePageComposite, toolkit, expansionAdapter);
        return thePage;
    }
    /**
     * Creates buttons for the label source, and sets the data to be
     * the alignment constant corresponding to the source; also attaches
     * a selection listener
     * @param parent Where to put the button
     * @param title The title to use for the button
     * @param source The constant that indicates the source
     * @param listener What to do if the button is selected
     * @param toolkit
     */
    private Button createLabelSourceButton(
        Composite parent,
        String title,
        int source,
        SelectionListener listener, FormToolkit toolkit) {
        Button b = toolkit.createButton(parent, title, SWT.RADIO);
        b.setData(new Integer(source));
        b.addSelectionListener(listener);
        return b;
    }
    /**
     * @see org.eclipse.jface.dialogs.Dialog#okPressed()
     */
    protected void okPressed() {
        exportStyle();
        processArmaniRules();
        ImageManager.imageDisposed(
            "EditVisDialog.okPressed",
            _imgFig.getImage());
        _imgFig.getImage().dispose();
        _imageSelectionGroup.dispose();
        super.okPressed();
    }
    protected void cancelPressed() {
        super.cancelPressed();
    }
    /**
     * For each allowable role or port type, this functions generates an Armani rule that
     * can be checked to ensure that only roles of the allowable type are permitted.
     *
     */
    protected void processArmaniRules() {
        if (_type == null)
            return;
        if (_type instanceof AcmeComponent || _type instanceof AcmeConnector) {
            AcmeElement element = (AcmeElement) _type;
            AcmeEnumeration invariants = element.getInvariants(false);
            while (invariants.moreItems()) {
                DesignRule rule = (DesignRule) invariants.getNextItem();
                if (rule.getProperty("vis-ports") != null
                    || rule.getProperty("vis-roles") != null) {
                    element.removeInvariant(rule);
                    AcmeStudioChangeManager
                        .globalChangeManager
                        .onRemoveArmaniRule(
                        element,
                        rule,
                        this);
                }
            }
        }
        String rule = null;
        if (_type instanceof AcmeComponent) {
            // Only generate the rule if we are restricting ports (BugFix #113)
            if (((ComponentVisualization) _style).restrictPortTypes()) {
                String[] allowablePorts = _allowableTypeList.getItems();
                if (allowablePorts.length > 0
                    && !allowablePorts[0].equals("")) {
                    String aPs = "";
                    for (int i = 0; i < allowablePorts.length; i++) {
                        aPs += allowablePorts[i];
                        if (i < allowablePorts.length - 1) {
                            aPs += ",";
                        }
                    }
                    int listPos = ALLOWABLEPORTSRULE.indexOf('%');
                    rule = ALLOWABLEPORTSRULE.substring(0, listPos);
                    rule += aPs;
                    rule += ALLOWABLEPORTSRULE.substring(listPos + 1);
                }
            }
        } else if (_type instanceof AcmeConnector) {
            // Only generate a rule if we are restricting roles (BugFix #113)
            if (((ConnectorVisualization) _style).restrictRoleTypes()) {
                String[] allowableRoles = _allowableTypeList.getItems();
                if (allowableRoles.length > 0
                    && !allowableRoles[0].equals("")) {
                    String aRs = "";
                    for (int i = 0; i < allowableRoles.length; i++) {
                        aRs += allowableRoles[i];
                        if (i < allowableRoles.length - 1) {
                            aRs += ",";
                        }
                    }
                    int listPos = ALLOWABLEROLESRULE.indexOf('%');
                    rule = ALLOWABLEROLESRULE.substring(0, listPos);
                    rule += aRs;
                    rule += ALLOWABLEROLESRULE.substring(listPos + 1);
                }
            }
        }
        if (rule != null) {
            // try to parse the rule and add it to the type
            try {
                ByteArrayInputStream bis =
                    new ByteArrayInputStream(rule.getBytes());
                ArmaniParser.ReInit(bis);
                AcmeErrorManager.clearAllErrors();
                DesignRule parsedRule = ArmaniParser.parse_DesignRule(_type);
                // There should be no errors
                AcmeElement typeAsEl = (AcmeElement) _type;
                typeAsEl.addInvariant(parsedRule);
                parsedRule.resolveTypeReferences();
                parsedRule.resolveInstanceReferences();
                parsedRule.typecheck();

                AcmeStudioChangeManager.globalChangeManager.onAddArmaniRule(
                    typeAsEl,
                    parsedRule,
                    this);
            } catch (ParseException e) {
            } catch (AcmeResolutionException e) {

            }
        }
    }
    /**
     * Returns the _style.
     * @return Visualization
     */
    public ElementVisualization getVisualization() {
        return _style;
    }
    protected void enableSizing(boolean enablement) {
        _heightResizeableButton.setEnabled(enablement);
        _widthResizeableButton.setEnabled(enablement);
        _defaultHeightText.setEnabled(enablement);
        _defaultWidthText.setEnabled(enablement);
    }

    private void createSpecialPage(int tabCount, FormToolkit toolkit) {
        String title = null;
        if (_style instanceof PortVisualization) {
            title = "Layout";
            _specialPage = createPortPage(_folder, toolkit);
        } else if (_style instanceof RoleVisualization) {
            title = "Layout";
            _specialPage = createRolePage(_folder, toolkit);
        } else if (
            _style instanceof ComponentVisualization
                && _type instanceof AcmeComponentType) {
            title = "Port Policy";
            _specialPage =
                createComponentPage(_folder, (AcmeComponentType) _type, toolkit);
        } else if (_style instanceof ConnectorVisualization) {
            title = "Role Policy";
            _specialPage =
                createConnectorPage(_folder, (AcmeConnectorType) _type, toolkit);
        }
        if (title != null) {
            CTabItem specialTab = new CTabItem(_folder, SWT.NONE, tabCount);
            specialTab.setText(title);
            specialTab.setControl(_specialPage);
        }
    }
	/**
	 * @return Returns the overrideTypeVis.
	 */
	public boolean isOverrideTypeVis() {
		return _overrideTypeVis;
	}
	public boolean canOverrideTypeVis() {
		return _canOverrideTypeVis;
	}
	public void setCanOverrideTypeVis(boolean canOverrideTypeVis) {
		_canOverrideTypeVis = canOverrideTypeVis;
	}
}
