package AcmeStudio.ui.editors.diagram.styles;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.jface.action.*;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.*;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.internal.WorkbenchImages;
import org.eclipse.ui.texteditor.StatusLineContributionItem;

import AcmeStudio.model.TypeAssignmentCommand;
import AcmeStudio.ui.editors.AcmeStyleEditor;
import AcmeStudio.ui.editors.actions.SuperTypeAssignmentMenuAction;
import AcmeStudio.ui.editors.diagram.styles.constants.SideOptions;
import AcmeStudio.util.HoverHelp;
import acmeLib.AcmeObject;

/**
 * @author Yan Hong (yh@cs.cmu.edu)
 *
 * Lists all the element styles a table displaying the
 * type, visualization and description. Allows creating
 * and modifying visual styles.
 */
public class ElementAlignmentList extends Composite {

    /** 
     * Holder of the element type.
     */
    private class AlignmentRuleHolder {

        private PlacementRule _rule = null;

        /**
         * Returns the element type held by this holder
         * @return the element type
         */
        public PlacementRule getRule() {
            return _rule;
        }

        /**
         * Returns the element type held by this holder
         * @return the element type
         */
        public void setRule(PlacementRule rule) {
            _rule = rule;
        }
    }

    // Constants for organizing the table
    private static final int TYPECOL = 0;
    private static final int TYPEWIDTH = 200;
    private static final int REGIONCOL = 1;
    private static final int REGIONWIDTH = 60;
    private static final int SIDECOL = 2;
    private static final int SIDEWIDTH = 100;
    private static final int ALIGNMENTCOL = 3;
    private static final int ALIGNMENTWIDTH = 80;
    private static final String[] PROPERTIES =
        { "Type", "Region", "Side", "Alignment" };
    private IContributionItem[] _myContributions;

    /** The table viewer used to view the properties **/
    private TableViewer _viewer = null;

    /** The object whose properties are being viewed **/
    private int _currentSide = SideOptions.DEFAULT;

    /** Indicates if the table is being refreshed **/
    private boolean _beingRefreshed = false;

    private AcmeStyleEditor _parentEditor = null;

    /** The action performed on double click **/
    private Action _doubleClickAction;

    /** Action to create a new type in this list **/
    private IAction _newTypeAction;

    private List _input = null;
    private CellEditor[] _cellEditors = new CellEditor[4];

    private Action _deleteRuleAction;

    private boolean _editable = true;

    private int _regions = 1;
	/*
     * The content provider class is responsible for
     * providing objects to the view. It returns all the 
     * properties to be displayed in the viewer.
     */
    private class ViewContentProvider implements IStructuredContentProvider {

        public void inputChanged(Viewer v, Object oldInput, Object newInput) {
        }
        public void dispose() {
        }

        /**
         * Returns all the types of the family.
         * @param parent The AcmeFamily
         * @return The types
         */
        public Object[] getElements(Object parent) {
            List ll = (List) parent;
            if (ll == null)
                return null;
            Object[] elements = new Object[ll.size()];
            int i = 0;
            Iterator iter = ll.iterator();
            while (iter.hasNext()) {
                PlacementRule rule = (PlacementRule) iter.next();
                AlignmentRuleHolder holder = new AlignmentRuleHolder();
                holder._rule = rule;
                elements[i++] = holder;
            }
            return elements;
        }
    }

    /** 
     * Provides routines for getting the text in each of the columns of the table
     */
    private class ViewLabelProvider
        extends LabelProvider
        implements ITableLabelProvider {
        /**
         * Returns the text associated with a particular column in the table
         * @param obj the type to get the information from
         * @param index the column to return the string for
         * @return the text to be placed in the column
         */
        public String getColumnText(Object obj, int index) {
            AlignmentRuleHolder ruleHolder = (AlignmentRuleHolder) obj;
            String retStr = null;
            switch (index) {
                case TYPECOL :
                    retStr = ruleHolder.getRule()._type;
                    if (retStr == null) {
                        retStr = "<<No Name>>";
                    }
                    break;
                case REGIONCOL :
                    retStr = Integer.toString(ruleHolder.getRule().getRegion());
                    break;
                case SIDECOL :
                    int side = ruleHolder.getRule()._side;
                    retStr = (String) SideOptions.SIDENAMES.get(side);
                    break;
                case ALIGNMENTCOL :
                    int alignment = ruleHolder.getRule()._alignment;
                    retStr =
                        (String) AlignmentOptions.ALIGNMENTNAMES.get(alignment);
                    break;
            }
            return retStr;
        }

        /** 
         * As there are no images associated with the table, just
         * return null in these routines.
         */
        public Image getColumnImage(Object obj, int index) {
            return null;
        }

        public Image getImage(Object obj) {
            return null;
        }
    }

    /**
     * Constructor for ElementTypeList. Constructs the Style table and adds
     * it to the appropriate editor part
     * @param parent the Composite in which to place the table
     * @param style the style of the table
     * @param category the category of this list
     * @param parentEditror the parent style editor
     */
    public ElementAlignmentList(Composite parent, int style) {
        super(parent, style);
        //_parentEditor = parentEditor;
        //_category = category;

        FillLayout tableLayout = new FillLayout();
        this.setLayout(tableLayout);
        this.setLayoutData(
            new GridData(
                GridData.FILL_HORIZONTAL | GridData.VERTICAL_ALIGN_FILL));

        // Create the table and set it's contents to the active design by default
        _viewer =
            new TableViewer(
                this,
                SWT.BORDER
                    | SWT.MULTI
                    | SWT.H_SCROLL
                    | SWT.V_SCROLL
                    | SWT.FULL_SELECTION);
        _viewer.setContentProvider(new ViewContentProvider());
        _viewer.setLabelProvider(new ViewLabelProvider());

        // Set the column header and widthe information
        Table table = _viewer.getTable();
        table.addKeyListener(new KeyListener() {
            public void keyPressed(KeyEvent evt) {
                if (evt.character == SWT.DEL) {
                    _deleteRuleAction.run();
                }
            }

            public void keyReleased(KeyEvent evt) {
            }
        });

        table.setHeaderVisible(true);
        table.setLinesVisible(true);

        TableColumn typeCol = new TableColumn(table, SWT.LEFT, TYPECOL);
        TableColumn regionCol = new TableColumn(table, SWT.LEFT, REGIONCOL);
        TableColumn sideCol = new TableColumn(table, SWT.LEFT, SIDECOL);
        TableColumn alignmentCol =
            new TableColumn(table, SWT.LEFT, ALIGNMENTCOL);

        typeCol.setText("Type");
        regionCol.setText("Region");
        sideCol.setText("Side");
        alignmentCol.setText("Alignment");

        ColumnLayoutData typeLayout = new ColumnPixelData(TYPEWIDTH, true);
        ColumnLayoutData regionLayout = new ColumnPixelData(REGIONWIDTH, true);
        ColumnLayoutData visLayout = new ColumnPixelData(SIDEWIDTH, true);
        ColumnLayoutData descLayout = new ColumnPixelData(ALIGNMENTWIDTH, true);

        // set columns in Table layout
        TableLayout layout = new TableLayout();
        layout.addColumnData(typeLayout);
        layout.addColumnData(regionLayout);
        layout.addColumnData(visLayout);
        layout.addColumnData(descLayout);
        table.setLayout(layout);

        // Add actions
        _doubleClickAction = new Action() {
            public void run() {
                handleDoubleClick();
            }
        };
        hookDoubleClickAction();

        makeActions();
        //        _myContributions =
        //            new ActionContributionItem[] {
        //                new ActionContributionItem(_newTypeAction),
        //                new ActionContributionItem(_deleteTypeAction)};

        hookContextMenu();

        // Start a helper to report errors in the properties
        HoverHelp help = new HoverHelp();
        help.activateHoverHelp(_viewer.getTable());
    }

    /**
     * Returns the TableView that contains the styles
     * @return TableViewer
     */
    public TableViewer getViewer() {
        return _viewer;
    }

    /** Modifies the selected style on a double click
     */
    public void handleDoubleClick() {
        

    }

    /**
     * Set the selected object to be the new object passed in
     * @param o the object to view
     */
    public void setInput(List ll) {
        if (ll == null)
            return;

        _input = ll;
        //_selectedObject = o;
        _viewer.setInput(ll);

        // Assign the cell modifiers -- this is heavy weight
        updateCellEditors();
        _viewer.setCellEditors(_cellEditors);
        _viewer.setColumnProperties(PROPERTIES);
        _viewer.setCellModifier(new StyleCellModifier());
    }

    /**
     * @see org.eclipse.swt.widgets.Control#setFocus()
     */
    public boolean setFocus() {
        return _viewer.getControl().setFocus();
    }

    /**
     * Adds the double click action to the table viewer
     */
    private void hookDoubleClickAction() {
        _viewer.addDoubleClickListener(new IDoubleClickListener() {
            public void doubleClick(DoubleClickEvent event) {
                _doubleClickAction.run();
            }
        });
    }

    /**
     * Makes all the actions for the property list
     */
    private void makeActions() {

        // The delete action, deletes the currently selected properties
        _deleteRuleAction = new Action() {
            public void run() {
                deleteSelectedRules();
            }
        };
        _deleteRuleAction.setText("Delete Rule");
        _deleteRuleAction.setToolTipText("Deletes selected rules");
        // As of Eclipse 3.0, IMG_TOOL_DELETE_HOVER == IMG_TOOL_DELETE
        _deleteRuleAction.setImageDescriptor(
            WorkbenchImages.getImageDescriptor(
                ISharedImages.IMG_TOOL_DELETE));
    }

    /**
     * Creates the context menu for this view, and hooks it into the framework
     */
    private void hookContextMenu() {
        MenuManager menuMgr = new MenuManager("StyleEditorPopupMenu");
        menuMgr.setRemoveAllWhenShown(true);
        menuMgr.addMenuListener(new IMenuListener() {
            public void menuAboutToShow(IMenuManager manager) {
                fillContextMenu(manager);
            }
        });

        Menu menu = menuMgr.createContextMenu(_viewer.getControl());
        _viewer.getControl().setMenu(menu);
        if (_parentEditor != null) {
            _parentEditor.getSite().registerContextMenu(menuMgr, _viewer);
        }
    }

    /**
     * Adds the contents of the menu
     */
    private void fillContextMenu(IMenuManager manager) {
        if (_editable) {
            SuperTypeAssignmentMenuAction assFamAction =
                (SuperTypeAssignmentMenuAction) _parentEditor.getAction(
                    TypeAssignmentCommand.ASSIGN_TYPE);
            SuperTypeAssignmentMenuAction unassFamAction =
                (SuperTypeAssignmentMenuAction) _parentEditor.getAction(
                    TypeAssignmentCommand.UNASSIGN_TYPE);
            manager.add(assFamAction.buildMenu());
            manager.add(unassFamAction.buildMenu());
            manager.add(new Separator());
            manager.add(_deleteRuleAction);
            manager.add(_newTypeAction);
            // Other plug-ins can contribute there actions here
            manager.add(new Separator("Additions"));
        } else {
            StatusLineContributionItem item =
                new StatusLineContributionItem("$XXX$");
            item.setText(
                "You must open the family file directly to edit this family");
            manager.add(item);
            manager.add(new Separator("Additions"));
        }
    }

    /**
     * Should be fired when this style bar is selected
     * @see org.eclipse.ui.ISelectionListener#selectionChanged(IWorkbenchPart, ISelection)
     */
    public void selectionChanged(IWorkbenchPart part, ISelection selection) {
    }

    public IContributionItem[] getContributions() {
        return _myContributions;
    }

    /**
     * Sent when the location (x, y) of a control changes relative
     * to its parent (or relative to the display, for <code>Shell</code>s).
     * @see org.eclipse.swt.events.ControlListener
     *
     * @param e an event containing information about the move
     */
    public void controlMoved(ControlEvent e) {
    }

    /**
     * Sent when the size (width, height) of a control changes.
     * @see org.eclipse.swt.events.ControlListener
     *
     * @param e an event containing information about the resize
     */
    public void controlResized(ControlEvent e) {
        Object obj = e.getSource();
        if (obj instanceof TableColumn) {
            if (!_beingRefreshed) {
                //_parentEditor.refreshStyleLists(this);
            }
        }
    }

    public Table getTable() {
        return _viewer.getTable();
    }

    public void refresh(ElementAlignmentList source) {
        _beingRefreshed = true;
        Table table = getTable();
        Table sourceTable = source.getTable();

        int size = table.getColumnCount();
        for (int i = 0; i < size; i++) {
            TableColumn tc = table.getColumn(i);
            TableColumn sourceTc = sourceTable.getColumn(i);
            tc.setWidth(sourceTc.getWidth());
        }
        this.redraw();
        _beingRefreshed = false;
    }

    public void deselectAll() {
        Table table = getTable();
        table.deselectAll();
    }

    /**
     * @see AcmeStudio.ui.views.parts.ElementViewPart#isValidFor(AcmeObject)
     */
    public boolean isValidFor(AcmeObject o) {
        return true;
    }

    private class StyleCellModifier implements ICellModifier {

        public boolean canModify(Object element, String property) {
            if (property.equals("Type")) {
                return false;
            }
            if (element instanceof AlignmentRuleHolder) {
                AlignmentRuleHolder ruleHolder = (AlignmentRuleHolder) element;
                _currentSide = ruleHolder.getRule()._side;
//                updateCellEditors();
            }

            return true;
        }

        public Object getValue(Object element, String property) {
            Object value = null;
            if ((element instanceof AlignmentRuleHolder)
                && canModify(element, property)) {
                AlignmentRuleHolder ruleHolder = (AlignmentRuleHolder) element;
                int side = ruleHolder.getRule()._side;
                //_currentSide = side;
//                updateCellEditors();
                if (property.equals("Type")) {
                    value = ruleHolder.getRule()._type;
                } else if (property.equals("Side")) {
                    value = new Integer(side);
                } else if (property.equals("Region")) {
                    value = new Integer(ruleHolder.getRule().getRegion() - 1);
                } else if (property.equals("Alignment")) {
                    int alignment = ruleHolder.getRule()._alignment;
                    value = new Integer(translateFromAlignment(alignment));
                }
            }
            return value;
        }

        public void modify(Object element, String property, Object value) {
            if (element instanceof TableItem) {
//                updateCellEditors();
                TableItem item = (TableItem) element;
                AlignmentRuleHolder ruleHolder =
                    (AlignmentRuleHolder) item.getData();
                if (property.equals("Side")) {
                    int test = ((Integer) value).intValue();
                    if (test >= 0)
                        ruleHolder.getRule()._side = test;
                } else if (property.equals("Alignment")) {
                    int test = ((Integer) value).intValue();
                    if (test >= 0)
                        ruleHolder.getRule()._alignment =
                            translateToAlignment(test);
                } else if (property.equals("Region")) {
                    int test = ((Integer) value).intValue();
                    if (test >= 0)
                        ruleHolder.getRule().setRegion(test + 1);
                }
                _viewer.refresh ();
//                setInput(_input);
            }
        }

        private int translateToAlignment(int index) {
            if (index == 0)
                return index;
            switch (_currentSide) {
                case SideOptions.TOP :
                case SideOptions.BOTTOM :
                case SideOptions.VCENTER :
                    switch (index) {
                        case 1 :
                        case 2 :
                            return index;
                        case 3 :
                            return 5;
                    }

                case SideOptions.LEFT :
                case SideOptions.RIGHT :
                case SideOptions.HCENTER :
                    return index + 2;
            }
            return index;
        }

        private int translateFromAlignment(int index) {
            if (index == 0)
                return index;
            switch (_currentSide) {
                case SideOptions.TOP :
                case SideOptions.BOTTOM :
                case SideOptions.VCENTER :
                    switch (index) {
                        case 1 :
                        case 2 :
                            return index;
                        case 5 :
                            return 3;
                    }

                case SideOptions.LEFT :
                case SideOptions.RIGHT :
                case SideOptions.HCENTER :
                    return index - 2;
            }
            return index;
        }

    }

    protected void updateCellEditors() {
        _cellEditors[SIDECOL] =
            new ComboBoxCellEditor(
                _viewer.getTable(),
                SideOptions.LABELGROUP2);

        String[] regionNumbers = new String[_regions];
        for (int i = 1; i <= _regions; i++) {
            regionNumbers[i - 1] = Integer.toString(i);
        }
        _cellEditors[REGIONCOL] =
            new ComboBoxCellEditor(_viewer.getTable(), regionNumbers);

        ArrayList items = new ArrayList(AlignmentOptions.ALIGNMENTNAMES);
        switch (_currentSide) {
            case SideOptions.TOP :
            case SideOptions.BOTTOM :
            case SideOptions.VCENTER :
                items.remove(
                    AlignmentOptions.ALIGNMENTNAMES.get(AlignmentOptions.TOP));
                items.remove(
                    AlignmentOptions.ALIGNMENTNAMES.get(
                        AlignmentOptions.BOTTOM));
                break;
            case SideOptions.LEFT :
            case SideOptions.RIGHT :
            case SideOptions.HCENTER :
                items.remove(
                    AlignmentOptions.ALIGNMENTNAMES.get(AlignmentOptions.LEFT));
                items.remove(
                    AlignmentOptions.ALIGNMENTNAMES.get(
                        AlignmentOptions.RIGHT));
                break;
            case SideOptions.DEFAULT :
                items.remove(
                    AlignmentOptions.ALIGNMENTNAMES.get(
                        AlignmentOptions.CENTER));
                break;
        }
        String[] labels = new String[items.size()];
        Iterator iter = items.iterator();
        int i = 0;
        while (iter.hasNext()) {
            String label = (String) iter.next();
            labels[i++] = label;
        }
        _cellEditors[ALIGNMENTCOL] =
            new ComboBoxCellEditor(_viewer.getTable(), labels);
    }


    /**
     * Sent when default selection occurs in the control.
     * <p>
     * For example, on some platforms default selection occurs
     * in a List when the user double-clicks an item or types
     * return in a Text.
     * </p>
     *
     * @param e an event containing information about the default selection
     */
    public void widgetDefaultSelected(SelectionEvent e) {
    }

    /**
     * Notifies that the selection has changed.
     *
     * @param event event object describing the change
     */
    public void selectionChanged(SelectionChangedEvent event) {
//        Debug.diagnoseln(
//            "ElementTypeList.selectionChanged: " + event.toString());
    }

    /**
     * Sent when a control gets focus.
     *
     * @param e an event containing information about the focus change
     */
    public void focusGained(FocusEvent e) {
    }

    /**
     * Sent when a control loses focus.
     *
     * @param e an event containing information about the focus change
     */
    public void focusLost(FocusEvent e) {
    }

    public void refresh() {
        if (!isDisposed() && !_viewer.getControl().isDisposed())
            this._viewer.refresh();
    }

    public void setEditable(boolean canEdit) {
        _editable = canEdit;
    }

    public java.util.List deleteSelectedRules() {
        java.util.List list = new java.util.LinkedList();
        // Get the selections
        ISelection isel = _viewer.getSelection();
        if (isel instanceof IStructuredSelection) {
            IStructuredSelection iss = (IStructuredSelection) isel;

            Iterator it = iss.iterator();
            // Iterate through the selections
            while (it.hasNext()) {
                Object obj = it.next();
                if (obj instanceof AlignmentRuleHolder) {
                    AlignmentRuleHolder ruleHolder = (AlignmentRuleHolder) obj;
                    PlacementRule rule = ruleHolder.getRule();
                    list.add(rule._type);
                    _input.remove(rule);
                }
            }

            setInput(_input);
        }
        return list;
    }

    public void setRegions(int regions) {
        this._regions = regions;
        updateCellEditors();
    }
}