package AcmeStudio.ui.editors.diagram.styles;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.*;
import org.eclipse.draw2d.*;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.PointList;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.Display;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.ui.editors.diagram.figures.ElementColorConstants;
import AcmeStudio.ui.editors.diagram.styles.constants.*;
import AcmeStudio.ui.editors.diagram.styles.constants.Shape;
import AcmeStudio.ui.views.parts.PropertyValueInspector;
import AcmeStudio.util.*;
import acmeLib.armaniParser.ParseException;
import acmeLib.elements.*;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyValue;
import acmeLib.properties.AcmeString;
import com.swtdesigner.SWTResourceManager;

/**
 * @author Hong Yan (yh@cs.cmu.edu)
 *
 * Provides the visual styles for figures.
 * 
 */
public abstract class ElementVisualization extends Visualization implements PropertyChangeListener, IManagedImageListener {
    public static final String VIS_TYPE_GENERAL = "GENERAL",
        VIS_TYPE_COMPONENT = "COMPONENT",
        VIS_TYPE_CONNECTOR = "CONNECTOR",
        VIS_TYPE_PORT = "PORT",
        VIS_TYPE_ROLE = "ROLE",
    	PROP_IMAGES_DISPOSED = "$ImageDisposed";
    
//    protected static final Map IMAGEREGISTRY /*<String, ManagedImage>*/= new HashMap ();
    /**
     * The shape of the visual element.
     */
    protected int _shape = DEFAULT_SHAPE;
    /** The image to use for the in place of the shape **/
    protected Image _image;
    /** The decorations to decorate the shape with **/
    protected ManagedImage[][] _decorations = new ManagedImage[3][3];

    /**
     * Whether to replicate the shape
     */
    protected boolean _replicateShape = DEFAULT_REPLICATESHAPE;
    /**
     * The color of the outline.
     */
    protected Color _outlineColor = DEFAULT_OUTLINECOLOR;
    /**
     * The style of the outline's line
     */
    protected int _outlineStyle = DEFAULT_OUTLINESTYLE;
    /**
     * The width of the outline.
     */
    protected int _outlineWidth = DEFAULT_OUTLINEWIDTH;
    /**
     * The fill color.
     */
    protected Color _fillColor = DEFAULT_FILLCOLOR;
    /**
     * The color of text if any.
     */
    protected Color _textColor = DEFAULT_TEXTCOLOR;
    /**
     * The size (in pixel) of the visual element.
     */
    protected Dimension _size = DEFAULT_SIZE;
    /**
     * Resize must preserve aspect ratio
     */
    protected boolean _preserveAspectRatio = false;
    /** 
     * Can resize in the x dimension
     */
    protected boolean _widthResizeable = true;
    /** 
     * Can resize in the y direction
     */
    protected boolean _heightResizeable = true;
    /**
     * The corner size (in pixel) of the visual element.
     */
    protected Dimension _cornerSize = DEFAULT_CORNERSIZE;
    /**
     * The direction of the visual element.
     * @see org.eclipse.draw2d.PositionConstants
     */
    protected int _direction = DEFAULT_DIRECTION;
    /**
     * The orientation of the visual element.
     * @see org.eclipse.draw2d.Orientable
     */
    protected int _internal_orientation = DEFAULT_INTERNAL_ORIENTATION;
    /**
     * The point list of the visual element.
     * This property is used to represent the figure when its shape
     * is irregular.
     * @see org.eclipse.draw2d.geometry.PointList
     */
    protected PointList _points = DEFAULT_POINTS;
    /**
     * The font to use for the label
     */
    protected Font _font = DEFAULT_FONT;
    /**
     * The word wrap setting of the label
     */
    protected boolean _labelWordwrap = DEFAULT_WORDWRAP;
    /**
     * The value of the label
     */
    protected String _labelValue = "";
    /**
     * The horizontal alignment of the label
     */
    protected int _labelHorizontalAlignment = DEFAULT_HORIZONTAL_ALIGNMENT;
    /**
     * The vertical alignment of the label
     */
    protected int _labelVerticalAlignment = DEFAULT_VERTICAL_ALIGNMENT;
    /**
     * The margin to use for the label
     */
    protected int _labelMargin = DEFAULT_LABEL_MARGIN;
    /**
     * What to base the label on
     */
    protected int _labelSource = DEFAULT_LABEL_SOURCE;
    
    /** Whether to resize the element according to the label **/
    protected boolean _resizeToFitLabel = DEFAULT_RESIZETOFIT_LABEL;
    /**
     * If the label is based on a property,
     * this is the name of the property
     */
    protected String _labelProperty = null;
    /**
     * Contains the list of variants for this type
     */
    protected List _variants = new Vector();
    /**
     * The default shape of the visual element.
     */
    public static final int DEFAULT_SHAPE = Shape.RECTANGLE;
    protected static final long FLAG_DEFAULT_SHAPE = 1;
    /**
     * The default outline color of the visual element.
     */
    public static final Color DEFAULT_OUTLINECOLOR =
        ElementColorConstants.black;
    protected static final long FLAG_DEFAULT_OUTLINECOLOR = 2;
    /**
     * The default outline line style
     */
    public static final int DEFAULT_OUTLINESTYLE = SWT.LINE_SOLID;
    protected static final long FLAG_DEFAULT_OUTLINESTYLE = 4;
    /**
     * The default outline width
     */
    public static final int DEFAULT_OUTLINEWIDTH = 1;
    protected static final long FLAG_DEFAULT_OUTLINE_WIDTH = 8;
    /**
     * The default fill color of the visual element.
     */
    public static final Color DEFAULT_FILLCOLOR =
        ElementColorConstants.backgroundColor;
    protected static final long FLAG_DEFAULT_FILLCOLOR = 16;
    /**
     * The default text color of the visual element.
     */
    public static final Color DEFAULT_TEXTCOLOR =
        ElementColorConstants.textColor;
    protected static final long FLAG_DEFAULT_TEXTCOLOR = 32;
    /**
     * The default size of the visual element.
     */
    public static final Dimension DEFAULT_SIZE = new Dimension(100, 50);
    protected static final long FLAG_DEFAULT_SIZE = 64;
    /**
     * The default corner size of the visual element.
     */
    public static final Dimension DEFAULT_CORNERSIZE = new Dimension(8, 8);
    protected static final long FLAG_DEFAULT_CORNERSIZE = 128;
    /**
     * The default direction of the visual element.
     */
    public static final int DEFAULT_DIRECTION = PositionConstants.EAST;
    protected static final long FLAG_DEFAULT_DIRECTION = 256;
    /**
     * The default orientation of the visual element.
     */
    public static final int DEFAULT_INTERNAL_ORIENTATION = Orientable.HORIZONTAL;
    protected static final long FLAG_DEFAULT_INTERNAL_ORIENTATION = 512;
    /**
     * The default point list of the visual element.
     */
    public static final PointList DEFAULT_POINTS = new PointList(4);
    
    /**
     * The default font
     */
    public static final Font DEFAULT_FONT = SWTResourceManager.getFont (AcmeStudioPlugin.getDefaultFontName(), 10, 0);
    protected static final long FLAG_DEFAULT_FONT = 1024;
    /**
     * The default word wrap configuration
     */
    public static final boolean DEFAULT_WORDWRAP = false;
    protected static final long FLAG_DEFAULT_WORDWRAP = 2048;
    /**
     * The default horiz. alignment
     */
    public static final int DEFAULT_HORIZONTAL_ALIGNMENT =
        HorizontalTextAlignmentOptions.CENTER;
    protected static final long FLAG_DEFAULT_HORIZONTAL_ALIGNMENT = 4096;
    /**
     * The default vert. alignment
     */
    public static final int DEFAULT_VERTICAL_ALIGNMENT =
        HorizontalTextAlignmentOptions.CENTER;
    protected static final long FLAG_DEFAULT_VERTICAL_ALIGNMENT = FLAG_DEFAULT_HORIZONTAL_ALIGNMENT << 2;
    /**
     * The default margin
     */
    public static final int DEFAULT_LABEL_MARGIN = 10;
    protected static final long FLAG_DEFAULT_LABEL_MARGIN = FLAG_DEFAULT_VERTICAL_ALIGNMENT << 2;
    /**
     * The default label source
     */
    public static final int DEFAULT_LABEL_SOURCE = LabelOptions.ACMEID;
    protected static final long FLAG_DEFAULT_LABEL_SOURCE = FLAG_DEFAULT_LABEL_MARGIN << 2;
    
    public static final boolean DEFAULT_REPLICATESHAPE = false;
    public static final long FLAG_DEFAULT_REPLICATESHAPE = FLAG_DEFAULT_LABEL_SOURCE << 2;
    
    public static final boolean DEFAULT_RESIZETOFIT_LABEL = false;
    public static final long FLAG_DEFAULT_RESIZETOFIT_LABEL = FLAG_DEFAULT_REPLICATESHAPE << 2;
    
    protected static final long LAST_DEFAULT_FLAG = FLAG_DEFAULT_RESIZETOFIT_LABEL;
    
    
    protected long _defaultFlags = 0;
    
    /**
     * Enumeration of default element styles.
     */
    public static final int DEFAULT_COMPONENT = 1,
        DEFAULT_CONNECTOR = 2,
        DEFAULT_PORT = 3,
        DEFAULT_ROLE = 4,
        DEFAULT_SYSTEM = 5;
    protected ManagedImage _preview = null;
    protected ManagedImage _previewNoLabel = null;

	private boolean _hideFromPalette = false;

	protected ManagedImage _smallImage;

	private boolean _isDefault = false;

	private static int	_numVisualizations = 0;

	private Set _superVisualizations = new HashSet ();

	private PropertyChangeSupport	_imageListeners;

	private String _iconFilename;

	/**
     * Constructor of Visualization.
     * @param style The defined style
     */
    public ElementVisualization(int style) {
    	_numVisualizations++;
//    	Debug.diagnoseln ("ElementVisualization.ElementVisualization: " + "there are now " + _numVisualizations + " visualizations");
        switch (style) {
            case DEFAULT_COMPONENT :
                _shape = Shape.RECTANGLE;
                _outlineColor = ElementColorConstants.componentOutlineColor;
                _fillColor = ElementColorConstants.componentFillColor;
                _textColor = ElementColorConstants.textColor;
                _size = new Dimension(100, 50);
                _isDefault = true;
                setAllDefaults ();
                break;
            case DEFAULT_CONNECTOR :
                _shape = Shape.OVAL;
                _outlineColor = ElementColorConstants.connectorOutlineColor;
                _fillColor = ElementColorConstants.connectorFillColor;
                _textColor = ElementColorConstants.textColor;
                _labelSource = LabelOptions.NOLABEL;
                _size = new Dimension(30, 30);
                _resizeToFitLabel = false;
                _isDefault = true;
                setAllDefaults ();
				break;
            case DEFAULT_PORT :
                _shape = Shape.RECTANGLE;
                _outlineColor = ElementColorConstants.portOutlineColor;
                _fillColor = ElementColorConstants.portFillColor;
                _textColor = ElementColorConstants.textColor;
                _labelSource = LabelOptions.NOLABEL;
                _size = new Dimension(10, 10);
                _isDefault = true;
                setAllDefaults ();
				break;
            case DEFAULT_ROLE :
                _shape = Shape.TRIANGLE;
                _outlineColor = ElementColorConstants.roleOutlineColor;
                _fillColor = ElementColorConstants.roleFillColor;
                _textColor = ElementColorConstants.textColor;
                _size = new Dimension(12, 12);
                _labelSource = LabelOptions.NOLABEL;
                _isDefault = true;
                setAllDefaults ();
                break;
        }
    }
    
    public void addImageDisposalListener (PropertyChangeListener l) {
    	   	_imageListeners.addPropertyChangeListener (l);
    }
    
    public void removeImageDisposalListener (PropertyChangeListener l) {
    	_imageListeners.removePropertyChangeListener (l);
    }
    /**
	 * 
	 */
    protected void setAllDefaults() {
		_defaultFlags = FLAG_DEFAULT_DIRECTION | FLAG_DEFAULT_CORNERSIZE | FLAG_DEFAULT_FILLCOLOR | FLAG_DEFAULT_FONT
				| FLAG_DEFAULT_HORIZONTAL_ALIGNMENT | FLAG_DEFAULT_LABEL_MARGIN | FLAG_DEFAULT_LABEL_SOURCE
				| FLAG_DEFAULT_INTERNAL_ORIENTATION | FLAG_DEFAULT_OUTLINE_WIDTH | FLAG_DEFAULT_OUTLINECOLOR | FLAG_DEFAULT_OUTLINESTYLE
				| FLAG_DEFAULT_SHAPE | FLAG_DEFAULT_SIZE & FLAG_DEFAULT_TEXTCOLOR | FLAG_DEFAULT_VERTICAL_ALIGNMENT
				| FLAG_DEFAULT_VERTICAL_ALIGNMENT;
	}
	
    protected boolean isDefaultPart (long part) {
    	return (_defaultFlags & part) != 0;
    }
    
    protected void forceDefault (long part) {
    	_defaultFlags |= part;
    }
    
	public ElementVisualization() {
    }
    /**
     * Set the shape
     * @param shape Defined shape constant
     */
    public void setShape(int shape) {
    	if (shape != _shape) {
	        this._shape = shape;
	        _defaultFlags &= ~FLAG_DEFAULT_SHAPE;
	        firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Get the shape
     * @return Shape constant
     */
    public int getShape() {
        return this._shape;
    }
    /**
     * Set the outline color
     * @param outlineColor The color    
     */
    public void setOutlineColor(Color outlineColor) {
    	if (!_outlineColor.equals (outlineColor)) {
    		this._outlineColor = outlineColor;
    		_defaultFlags &= ~FLAG_DEFAULT_OUTLINECOLOR;
    		firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Get the outline color
     * @return Ouline color
     */
    public Color getOutlineColor() {
        return this._outlineColor;
    }
    /**
     * Set the outline style
     * @param outlineColor The style    
     */
    public void setOutlineStyle(int style) {
    	if (style != _outlineStyle) {
    		this._outlineStyle = style;
    		_defaultFlags &= ~FLAG_DEFAULT_OUTLINESTYLE;
    		firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Get the outline style
     * @return Ouline style
     */
    public int getOutlineStyle() {
        return this._outlineStyle;
    }
    /**
     * Set the fill color
     * @param fillColor The color    
     */
    public void setFillColor(Color fillColor) {
    	if (!fillColor.equals (_fillColor)) {
    		this._fillColor = fillColor;
    		_defaultFlags &= ~FLAG_DEFAULT_FILLCOLOR;
    		firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Get the fill color
     * @return Fill color
     */
    public Color getFillColor() {
        return this._fillColor;
    }
    /**
     * Set the text color
     * @param textColor The color    
     */
    public void setTextColor(Color textColor) {
    	if (!textColor.equals (_textColor)) {
	        this._textColor = textColor;
	        _defaultFlags &= ~FLAG_DEFAULT_TEXTCOLOR;
	        firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Get the text color
     * @return Text color
     */
    public Color getTextColor() {
        return this._textColor;
    }
    /**
     * Set the size
     * @param size The size    
     */
    public void setSize(Dimension size) {
    	if (!_size.equals (size)) {
	        this._size = size;
	        _defaultFlags &= ~FLAG_DEFAULT_SIZE;
	        firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Get the size
     * @return Size
     */
    public Dimension getSize() {
        return this._size;
    }
    
    /**
     * Gets the size as a calculation, taking into account the resizeToFitLabel property
     * @param element
     * @return
     */
    public Dimension  getSizeWithLabel (AcmeElement element) {
    	Dimension origSize = this._size.getCopy();
    	Point location = new Point (0,0);
    	Rectangle origBounds = new Rectangle (location.x - _size.width/2, location.y - _size.height/2,_size.width, _size.height);
    	if (this._resizeToFitLabel) {
    			Rectangle labelBounds = getLabelBounds (element, origBounds);
    			if (labelBounds.width > origBounds.width) {
    				origBounds.width = labelBounds.width;
    			}
    			if (labelBounds.height > origBounds.height) {
    				origBounds.height = labelBounds.height;
    			}
    			return origBounds.getSize();
    	}
    	else {
			return origSize;
    	}
    	
    }
    /**
     * Set the corner size
     * @param size The corner size    
     */
    public void setCornerSize(Dimension size) {
    	if (!_cornerSize.equals (size)) {
    		this._cornerSize = size;
    		firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Get the corner size
     * @return Corner size
     */
    public Dimension getCornerSize() {
        return this._cornerSize;
    }
    /**
     * Set the direction
     * @param value The direction
     */
    public void setDirection(int value) {
    	if (_direction != value) {
    		if ((value & (PositionConstants.NORTH | PositionConstants.SOUTH)) != 0)
    			_internal_orientation = Orientable.VERTICAL;
    		else
    			_internal_orientation = Orientable.HORIZONTAL;
    		_direction = value;
    		_defaultFlags &= ~FLAG_DEFAULT_DIRECTION;
    		firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Get the direction
     * @return Direction
     */
    public int getDirection() {
        return this._direction;
    }
    /**
     * Set the orientation
     * @param value The orientation    
     */
    public void setInnerUsedOrientation(int value) {
        if (_internal_orientation == Orientable.VERTICAL
            && value == Orientable.HORIZONTAL) {
            if (_direction == PositionConstants.NORTH)
                setDirection(PositionConstants.WEST);
            else
                setDirection(PositionConstants.EAST);
        }
        if (_internal_orientation == Orientable.HORIZONTAL
            && value == Orientable.VERTICAL) {
            if (_direction == PositionConstants.WEST)
                setDirection(PositionConstants.NORTH);
            else
                setDirection(PositionConstants.SOUTH);
        }
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    /**
     * Get the orientation
     * @return Orientation
     */
    public int getInnerUsedOrientation() {
        return this._internal_orientation;
    }
    public boolean getWordwrap() {
        return _labelWordwrap;
    }
    public void setWordwrap(boolean labelWordwrap) {
    	if (_labelWordwrap != labelWordwrap) {
	        _labelWordwrap = labelWordwrap;
	        _defaultFlags &= ~FLAG_DEFAULT_WORDWRAP;
	        firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    public abstract ElementVisualization copy();
    
    /*{
        ElementVisualization cp = new ElementVisualization();
        cp.copyInto(this);
        return cp;
    }*/
    public ElementVisualization copy(AcmeElement elementType) {
        ElementVisualization cp = null;
//        beginChangeBlock();
        if (elementType instanceof AcmePort) {
            cp = new PortVisualization();
            cp.beginChangeBlock();
            cp.copyInto(this);
        } else if (elementType instanceof AcmeRole) {
            cp = new RoleVisualization();
            cp.beginChangeBlock();
            cp.copyInto(this);
        } else if (elementType instanceof AcmeComponent) {
            cp = new ComponentVisualization();
            cp.beginChangeBlock();
            cp.copyInto(this);
        } else if (elementType instanceof AcmeConnector) {
            cp = new ConnectorVisualization();
            cp.beginChangeBlock();
            cp.copyInto(this);
        }
        else {
        	cp = new GenericVisualization ();
        	cp.beginChangeBlock ();
        	cp.copyInto (this);
        }
        cp.endChangeBlock();
//        } else {
//            cp = new ElementVisualization();
//            cp.copyInto(this);
//        }
        return cp;
    }
    /**
     * Create a triangle shape based on the location, size, and direction
     * then store the shape in _points.
     * @param location The left-top point of the element.
     */
    public void createTriangle(
        org.eclipse.draw2d.geometry.Rectangle r,
        Dimension size,
        double rotation) {
    	size = r.getSize ();
        if (size == null) {
            size = r.getSize();
        }
        Point center = r.getCenter();
        Point head = null, p2 = null, p3 = null;
        switch (_direction) {
            case PositionConstants.EAST :
                head = new Point(center.x + size.width / 2, center.y);
                p2 =
                    new Point(
                        center.x - size.width / 2,
                        center.y - size.height / 2);
                p3 =
                    new Point(
                        center.x - size.width / 2,
                        center.y + size.height / 2);
                break;
            case PositionConstants.WEST :
                head = new Point(center.x - size.width / 2, center.y);
                p2 =
                    new Point(
                        center.x + size.width / 2,
                        center.y - size.height / 2);
                p3 =
                    new Point(
                        center.x + size.width / 2,
                        center.y + size.height / 2);
                break;

            case PositionConstants.NORTH :
                head = new Point(center.x, center.y - size.height / 2);
                p2 =
                    new Point(
                        center.x - size.width / 2,
                        center.y + size.height / 2);
                p3 =
                    new Point(
                        center.x + size.width / 2,
                        center.y + size.height / 2);
                break;
            case PositionConstants.SOUTH :
                head = new Point(center.x, center.y + size.height / 2);
                p2 =
                    new Point(
                        center.x + size.width / 2,
                        center.y - size.height / 2);
                p3 =
                    new Point(
                        center.x - size.width / 2,
                        center.y - size.height / 2);
                break;
        }
        _points.removeAllPoints();
        _points.addPoint(head);
        _points.addPoint(p2);
        _points.addPoint(p3);
        if (rotation != 0) {
            _points = rotate(rotation, _points);
        }
    }

    /**
     * Create a triangle shape based on the location, size, and direction
     * then store the shape in _points.
     * @param location The left-top point of the element.
     */
    public void createDiamond(
        org.eclipse.draw2d.geometry.Rectangle r,
        Dimension size,
        double rotation) {
        if (size == null) {
            size = r.getSize();
        }
        double r1 = ((double) size.width) / 2;
        double r2 = ((double) size.height) / 2;
        Point center = r.getCenter();
        double alpha1 = rotation;
        double alpha2 = Math.PI / 2 + rotation;
        double alpha3 = Math.PI + rotation;
        double alpha4 = Math.PI * 3 / 2 + rotation;
        Point p1 = null, p2 = null, p3 = null, p4 = null;
        p1 =
            new Point(
                Math.round(center.x + r1 * Math.cos(alpha1)),
                Math.round(center.y + r1 * Math.sin(alpha1)));
        p2 =
            new Point(
                Math.round(center.x + r2 * Math.cos(alpha2)),
                Math.round(center.y + r2 * Math.sin(alpha2)));
        p3 =
            new Point(
                Math.round(center.x + r1 * Math.cos(alpha3)),
                Math.round(center.y + r1 * Math.sin(alpha3)));
        p4 =
            new Point(
                Math.round(center.x + r2 * Math.cos(alpha4)),
                Math.round(center.y + r2 * Math.sin(alpha4)));
        _points.removeAllPoints();
        _points.addPoint(p1);
        _points.addPoint(p2);
        _points.addPoint(p3);
        _points.addPoint(p4);
    }
    /**
     * Create a triangle shape based on the location, size, and direction
     * then store the shape in _points.
     * @param location The left-top point of the element.
     */
    public void createRectangle(
        org.eclipse.draw2d.geometry.Rectangle r,
        Dimension size,
        double rotation) {
        if (size == null) {
            size = r.getSize();
        }
        double radius =
            Math.sqrt(
                (size.width * size.width + size.height * size.height))
                / 2;
        if (rotation != 0) {
            radius *= 0.8;
        }
        Point center = r.getCenter();
        double beta = Math.atan2(size.height, size.width);
        double alpha1 = beta + rotation;
        double alpha2 = Math.PI - beta + rotation;
        double alpha3 = Math.PI + beta + rotation;
        double alpha4 = Math.PI * 2 - beta + rotation;
        Point p1 = null, p2 = null, p3 = null, p4 = null;
        p1 =
            new Point(
                Math.round(center.x + radius * Math.cos(alpha1)),
                Math.round(center.y + radius * Math.sin(alpha1)));
        p2 =
            new Point(
                Math.round(center.x + radius * Math.cos(alpha2)),
                Math.round(center.y + radius * Math.sin(alpha2)));
        p3 =
            new Point(
                Math.round(center.x + radius * Math.cos(alpha3)),
                Math.round(center.y + radius * Math.sin(alpha3)));
        p4 =
            new Point(
                Math.round(center.x + radius * Math.cos(alpha4)),
                Math.round(center.y + radius * Math.sin(alpha4)));
        _points.removeAllPoints();
        _points.addPoint(p1);
        _points.addPoint(p2);
        _points.addPoint(p3);
        _points.addPoint(p4);
    }

    private PointList rotate(double alpha, PointList pointList) {
        PointList result = new PointList();
        Rectangle bounds = pointList.getBounds();
        Point center = bounds.getCenter();
        double r =
            Math.sqrt(
                bounds.height * bounds.height + bounds.width * bounds.width)
                / 2;
        for (int i = 0; i < pointList.size(); i++) {
            Point p = pointList.getPoint(i);
            double theta = Math.atan2(p.y - center.y, p.x - center.x) + alpha;
            Point newP =
                new Point(
                    center.x + r * Math.cos(theta),
                    center.y + r * Math.sin(theta));
            result.addPoint(newP);
        }
        return result;
    }
    /**
     * Fill the shape based on its type
     * @param graphics The graphics to draw the shape on.
     * @param location The left-top point of the element.
     */
    public void fillShape(
        Graphics graphics,
        AcmeElement element,
        org.eclipse.draw2d.geometry.Point location,
        Dimension size) {
        org.eclipse.draw2d.geometry.Rectangle r =
            new org.eclipse.draw2d.geometry.Rectangle(location, size);
        fillShape(graphics, r, null, 0);
    }
    public void drawError(
        Graphics graphics,
        AcmeElement element,
        org.eclipse.draw2d.geometry.Point location) {
        //        Dimension size = Helper.getSize (element);
        //        if (size == null) size = _size;
        //        org.eclipse.draw2d.geometry.Rectangle r = new org.eclipse.draw2d.geometry.Rectangle (location, size);
        location.x += 2;
        location.y += 2;
        graphics.drawImage(
            ImageManager.getIcon(ImageManager.ID_ERROR),
            location);
    }
    public void fillHollowShape(
        Graphics g,
        org.eclipse.draw2d.geometry.Rectangle r) {
        g.pushState();

        // Note, this causes the grid to disappear behind the border
        fillShape(g, r, null, 0);
        Color oldFillColor = _fillColor;
        _fillColor = SWTResourceManager.getColor (255, 255, 255);
        org.eclipse.draw2d.geometry.Rectangle smallRect =
            new org.eclipse.draw2d.geometry.Rectangle(r);
        smallRect.x += 20;
        smallRect.y += 20;
        smallRect.width -= 40;
        smallRect.height -= 40;
        fillShape(g, smallRect, null, 0);
        _fillColor = oldFillColor;
        g.setBackgroundColor(_fillColor);
        g.setForegroundColor(_outlineColor);
        g.popState();
    }

    public void fillShape(
        Graphics graphics,
        org.eclipse.draw2d.geometry.Rectangle rect,
        Dimension size,
        double rotation) {
        graphics.pushState();
        graphics.setBackgroundColor(_fillColor);
        org.eclipse.draw2d.geometry.Rectangle r =
            new org.eclipse.draw2d.geometry.Rectangle(rect);
       
		switch (_shape) {
            case Shape.RECTANGLE :
                //r.expand(-_outlineWidth, -_outlineWidth);
                createRectangle(r, size, rotation);
                graphics.fillPolygon(_points);
                break;
            case Shape.OVAL :
            	r.height++; r.width++;
                graphics.fillOval(r);
                break;
            case Shape.ROUNDEDRECT : {
            	//r.expand(-1, -1);
            	r.width++; r.height++;
            	int arcWidth = r.width / 6;
            	if (arcWidth <=5)
            		arcWidth = 6;
            	int arcHeight = r.height / 6;
            	if (arcHeight <=5) 
            		arcHeight = 6;
            	int arcSize = Math.max (arcWidth, arcHeight);
				graphics.fillRoundRectangle(r, arcSize, arcSize);
            }
                break;
            case Shape.TRIANGLE :
                //r.expand(-_outlineWidth, -_outlineWidth);
                createTriangle(r, size, rotation);
                graphics.fillPolygon(_points);
                break;
            case Shape.FILTER :
                {
                    //r.expand(-_outlineWidth, -_outlineWidth);
                    int northApex = 0;
                    int southApex = 0;
                    int eastApex = 0;
                    int westApex = 0;
                    switch (_direction) {
                        case PositionConstants.NORTH :
                            northApex = r.height / 15;
                        	if (northApex == 0) northApex = 2;
                            break;
                        case PositionConstants.SOUTH :
                            southApex = r.height / 15;
                        	if (southApex == 0) southApex = 2;
                            break;
                        case PositionConstants.EAST :
                            eastApex = r.width / 15;
                        	if (eastApex == 0) eastApex = 2;
                            break;
                        case PositionConstants.WEST :
                            westApex = r.width / 15;
                        	if (westApex == 0) westApex = 2;
                            break;
                    }
                    _points = new PointList(8);
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + westApex,
                            r.y + northApex));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width / 2,
                            r.y + southApex));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width - eastApex,
                            r.y + northApex));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width - westApex,
                            r.y + r.height / 2));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width - eastApex,
                            r.y + r.height - southApex));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width / 2,
                            r.y + r.height - northApex));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + westApex,
                            r.y + r.height - southApex));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + eastApex,
                            r.y + r.height / 2));
                    graphics.fillPolygon(_points);
                    break;
                }
            case Shape.FOLDER :
                {
                    //r.expand(-_outlineWidth, -_outlineWidth);
                    int apex = _size.height / 8;
                    _points = new PointList(7);
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(r.x, r.y));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width / 3,
                            r.y));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width / 3,
                            r.y + apex));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width,
                            r.y + apex));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width,
                            r.y + r.height));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x,
                            r.y + r.height));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(r.x, r.y + apex));
                    graphics.fillPolygon(_points);
                    break;
                }
            case Shape.DIAMOND :
                {
                    //r.expand(-_outlineWidth, -_outlineWidth);
                    createDiamond(r, size, rotation);
                    graphics.fillPolygon(_points);
                    break;
                }
            case Shape.PAGE :
                {
                    //r.expand(-_outlineWidth, -_outlineWidth);
                    int pageFold = Math.max((r.width) / 5, (r.height) / 5);
                    _points = new PointList(5);
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(r.x, r.y));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width - pageFold,
                            r.y));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width,
                            r.y + pageFold));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x + r.width,
                            r.y + r.height));
                    _points.addPoint(
                        new org.eclipse.draw2d.geometry.Point(
                            r.x,
                            r.y + r.height));
                    graphics.fillPolygon(_points);
                    break;
                }
            case Shape.REPOSITORY :
                {
                    int ovalHeight = r.width / 3;
                    int ovalWidth = r.width;
                    graphics.fillRectangle(
                        r.x,
                        r.y + ovalHeight / 2,
                        r.width+1,
                        r.height - ovalHeight + 1);
                    graphics.fillArc(new org.eclipse.draw2d.geometry.Rectangle (r.x, r.y, ovalWidth + 1, ovalHeight+1), 0, 180);
                    graphics.fillArc (new org.eclipse.draw2d.geometry.Rectangle (r.x, r.y + r.height - ovalHeight + 1, ovalWidth, ovalHeight), 0,-180);
                    break;
                }
        }

        if (_image != null) {
            org.eclipse.draw2d.geometry.Point imageLocation =
                new org
                    .eclipse
                    .draw2d
                    .geometry
                    .Rectangle(r.x, r.y, r.width, r.height)
                    .getCenter();
            org.eclipse.draw2d.geometry.Rectangle imageBounds =
                new org.eclipse.draw2d.geometry.Rectangle(
                    _image.getBounds().x,
                    _image.getBounds().y,
                    _image.getBounds().width,
                    _image.getBounds().height);
            imageLocation =
                imageLocation.translate(imageBounds.getCenter().negate());
            graphics.drawImage(_image, imageLocation);
        }

        for (int y = 0; y < 3; y++) {
            for (int x = 0; x < 3; x++) {
                if (_decorations[x][y] != null) {
                    org.eclipse.draw2d.geometry.Point imageLocation =
                        new org
                            .eclipse
                            .draw2d
                            .geometry
                            .Rectangle(
                                r.x + x * r.width / 3,
                                r.y + y * r.height / 3,
                                r.width / 3,
                                r.height / 3)
                            .getCenter();
                    org.eclipse.draw2d.geometry.Rectangle imageBounds =
                        new org.eclipse.draw2d.geometry.Rectangle(
                            _decorations[x][y].getBounds().x,
                            _decorations[x][y].getBounds().y,
                            _decorations[x][y].getBounds().width,
                            _decorations[x][y].getBounds().height);
                    imageLocation =
                        imageLocation.translate(
                            imageBounds.getCenter().negate());
                    graphics.drawImage(_decorations[x][y].getImage(), imageLocation);
                }
            }
        }
        graphics.popState();
    }
    /**
     * Outline the shape based on its type
     * @param graphics The graphics to draw the shape on.
     * @param location The left-top point of the element.
     */
    public void outlineShape(
        Graphics graphics,
        AcmeElement element,
        org.eclipse.draw2d.geometry.Point location,
        Dimension size) {
        if (_outlineStyle == SWT.NONE)
            return;
        org.eclipse.draw2d.geometry.Rectangle r =
            new org.eclipse.draw2d.geometry.Rectangle(location, size);
        outlineShape(graphics, r);
    }
    public void outlineShape(
        Graphics graphics,
        org.eclipse.draw2d.geometry.Rectangle rect) {
        if (_outlineStyle == SWT.NONE)
            return;
        graphics.pushState();
        graphics.setForegroundColor(_outlineColor);
        graphics.setLineStyle(_outlineStyle);
        graphics.setLineWidth(_outlineWidth);
        org.eclipse.draw2d.geometry.Rectangle r =
            new org.eclipse.draw2d.geometry.Rectangle(rect);
        //        r.expand (-_outlineWidth*2, -_outlineWidth*2);
        switch (_shape) {
            case Shape.OVAL :
                graphics.drawOval(r);
                break;
            case Shape.ROUNDEDRECT : {
            	int arcWidth = r.width / 6;
            	if (arcWidth <=5)
            		arcWidth = 6;
            	int arcHeight = r.height / 6;
            	if (arcHeight <=5) 
            		arcHeight = 6;
            	int arcSize = Math.max (arcWidth, arcHeight);
            	graphics.drawRoundRectangle(r, arcSize, arcSize);
            }
                break;
            case Shape.RECTANGLE :
            case Shape.FILTER :
            case Shape.FOLDER :
            case Shape.DIAMOND :
            case Shape.TRIANGLE :
                graphics.drawPolygon(_points);
                break;
            case Shape.PAGE :
                {
                    graphics.drawPolygon(_points);
                    org.eclipse.draw2d.geometry.Point startFold =
                        _points.getPoint(1);
                    org.eclipse.draw2d.geometry.Point endFold =
                        _points.getPoint(2);
                    org.eclipse.draw2d.geometry.Point vertex =
                        new org.eclipse.draw2d.geometry.Point(
                            startFold.x,
                            endFold.y);
                    graphics.drawLine(startFold, vertex);
                    graphics.drawLine(vertex, endFold);
                    break;
                }
            case Shape.REPOSITORY :
                {
                    int ovalHeight = r.width / 3;
                    int ovalWidth = r.width;
                    graphics.drawOval(
                        new org.eclipse.draw2d.geometry.Rectangle(
                            r.x,
                            r.y,
                            ovalWidth,
                            ovalHeight));
                    graphics.drawLine(
                        r.x,
                        r.y + ovalHeight / 2,
                        r.x,
                        r.y + r.height - ovalHeight / 2);
                    graphics.drawLine(
                        r.x + r.width,
                        r.y + ovalHeight / 2,
                        r.x + r.width,
                        r.y + r.height - ovalHeight / 2);
                    graphics.pushState();
                    graphics.clipRect(
                        new org.eclipse.draw2d.geometry.Rectangle(
                            r.x,
                            r.y + r.height - ovalHeight / 2,
                            ovalWidth,
                            ovalHeight / 2));
                    graphics.drawOval(
                        r.x,
                        r.y + r.height - ovalHeight,
                        ovalWidth,
                        ovalHeight);
                    graphics.popState();
                    break;
                }
        }
        graphics.popState();
    }
    public String buildLabel(AcmeElement element) {
        String label = "";
        switch (_labelSource) {
            case LabelOptions.ACMEID :
                label = element.getName();
                break;
            case LabelOptions.ACMEIDANDTYPE :
                label =
                    element.getName()
                        + " : "
                        + ((AcmeElementType) element
                            .getDeclaredTypes()
                            .getNextItem())
                            .getTypeName();
                break;
            case LabelOptions.ACMEPROPERTYVALUE :
                {
                    AcmeProperty property = element.getProperty(_labelProperty);
                    if (property != null) {
                        AcmePropertyValue val = property.getValue();
                        if (val != null) {
                            label = val.toString();
                            if (val instanceof AcmeString) { // strip the "
                                label = label.substring(1, label.length() - 1);
                            }
                        } else {
                            label = "";
                        }
                    }
                    break;
                }
            case LabelOptions.ACMETYPE :
                label =
                    ((AcmeElementType) element
                        .getDeclaredTypes()
                        .getNextItem())
                        .getTypeName();
                break;
            case LabelOptions.STATICLABEL :
                label = getLabelValue();
                break;
        }
        return label.trim ();
    }
    /**
     * Draw the label on the figure, using the visual styles
     * @param graphics The graphics to draw text on
     * @param location The top-left location of the element
     * @param text The text to be drawn
     */
    public void drawLabel(
        Graphics graphics,
        AcmeElement element,
        org.eclipse.draw2d.geometry.Point location,
        Dimension size) {
        String text = buildLabel(element);
        org.eclipse.draw2d.geometry.Rectangle r =
            new org.eclipse.draw2d.geometry.Rectangle(location, size);
        drawLabel(graphics, r, text, _textColor);
    }
    
    public void drawLabel (Graphics graphics, AcmeElement element, org.eclipse.draw2d.geometry.Point location, Dimension size, int transparency) {
        String text = buildLabel (element);
        org.eclipse.draw2d.geometry.Rectangle r =
             new org.eclipse.draw2d.geometry.Rectangle(location, size);        
        graphics.pushState ();
        Color color = new Color (null, lighten (transparency, _textColor.getRed ()), lighten (transparency,_textColor.getGreen ()), lighten (transparency, _textColor.getBlue()));
        graphics.setForegroundColor(color);
        drawLabel (graphics, r, text, color);
        color.dispose ();
        graphics.popState ();
    }
    
    private int lighten(int transparency, int col) {
        return col + (255 - col) * transparency/100;
    }
    
    private class NewLabelInfo {
    	public String _label;
    	public Dimension _size;
    }
    
    private NewLabelInfo getLabelInfo (String text, Rectangle r) {
    	String label = text;           
    	
    	Dimension textSize = FigureUtilities.getStringExtents(label, _font);
    	int innerWidth = r.width - 2 * _labelMargin;
    	if (textSize.width > innerWidth
    			&& _labelWordwrap
    			&& label.indexOf("_") >= 0) {
    		
    		
    		String[] textSegments = text.split("_");
    		label = "";
    		int w = 0;
    		int h = 0;
    		int wMax = 0;
    		int l = textSegments.length;
    		boolean hasWord = false;
    		for (int i = 0; i < l; i++) {
    			String seg = textSegments[i];
    			String segWithTail = new String(seg);
    			if (i < l - 1) {
    				segWithTail += "_";
    			}
    			Dimension segSize =
    			FigureUtilities.getStringExtents(segWithTail, _font);
    			w += segSize.width;
    			if (w > innerWidth) {
    				if (hasWord) {
    					w -= segSize.width;
    					label += "\n";
    					if (w > wMax) {
    						wMax = w;
    					}
    					label += segWithTail;
    					w =
    					FigureUtilities.getStringExtents(
    							segWithTail,
    							_font).width;
    				} else {
    					label += segWithTail;
    					label += "\n";
    					if (w > wMax) {
    						wMax = w;
    					}
    					w = 0;
    				}
    				h += segSize.height;
    				if (w > wMax) {
    					wMax = w;
    				}
    			} else {
    				label += segWithTail;
    				hasWord = true;
    			}
    		}
    		textSize = new Dimension(wMax, h);
    	}
    	NewLabelInfo info = new NewLabelInfo ();
    	info._label = label.trim ();
    	info._size = textSize;
    	return info;
    }
    
    public Rectangle getLabelBounds (AcmeElement element, Rectangle rect) {
    	Dimension size = getLabelInfo (buildLabel (element), rect)._size;
		return new Rectangle (getLabelLocation (rect, size), size);
    }
    
    public String getLabel (AcmeElement element, Rectangle rect) {
    	return getLabelInfo (buildLabel (element), rect)._label;
    }
    
    protected void drawLabel(
        Graphics graphics,
        org.eclipse.draw2d.geometry.Rectangle r,
        String text, Color textColor) {
        if (text != null && _labelSource != LabelOptions.NOLABEL) {
            graphics.pushState();
            if (_font != null) {
                graphics.setFont(_font);
            } else {
                _font = graphics.getFont();
            }
            graphics.setForegroundColor(textColor);
            // Calculate the position
            NewLabelInfo info = getLabelInfo (text, r);
            Dimension textSize = info._size;
            String label = info._label;
            Point p = getLabelLocation (r, textSize);
            graphics.drawText(label, p);
            graphics.popState();
        }
    }
    private Point getLabelLocation(org.eclipse.draw2d.geometry.Rectangle r, Dimension textSize) {
		//          FontMetrics fm = graphics.getFontMetrics ();
		//          Dimension textSize = new Dimension (text.length () * fm.getAverageCharWidth(),fm.getHeight());
		//getTextExtent (text);
		int locX;
		int locY;
		switch (_labelHorizontalAlignment) {
		    case HorizontalTextAlignmentOptions.LEFT :
		        locX = r.x + _labelMargin;
		        break;
		    case HorizontalTextAlignmentOptions.RIGHT :
		        locX = r.x + r.width - textSize.width - _labelMargin;
		        break;
		    default :
		    case HorizontalTextAlignmentOptions.CENTER :
		        locX = r.x + r.width / 2 - textSize.width / 2;
		        break;
		    case HorizontalTextAlignmentOptions.LEFT_OUTSIDE :
		        locX = r.x - textSize.width - _labelMargin;
		        break;
		    case HorizontalTextAlignmentOptions.RIGHT_OUTSIDE :
		        locX = r.x + r.width + _labelMargin;
		        break;
		}
		switch (_labelVerticalAlignment) {
		    default :
		    case VerticalTextAlignmentOptions.CENTER :
		        locY = r.y + r.height / 2 - textSize.height / 2;
		        break;
		    case VerticalTextAlignmentOptions.BOTTOM :
		        locY = r.y + r.height - textSize.height - _labelMargin;
		        break;
		    case VerticalTextAlignmentOptions.TOP :
		        locY = r.y + _labelMargin;
		        break;
		    case VerticalTextAlignmentOptions.ABOVE :
		        locY = r.y - textSize.height - _labelMargin;
		        break;
		    case VerticalTextAlignmentOptions.BELOW :
		        locY = r.y + r.height + _labelMargin;
		        break;
		}
		Point p = new Point (locX, locY);
		return p;
	}
	/**
     * Returns the _outlineWidth.
     * @return int
     */
    public int getOutlineWidth() {
        return _outlineWidth;
    }
    /**
     * Sets the _outlineWidth.
     * @param _outlineWidth The _outlineWidth to set
     */
    public void setOutlineWidth(int outlineWidth) {
    	if (outlineWidth != _outlineWidth) {
    		this._outlineWidth = outlineWidth;
    		_defaultFlags &= ~FLAG_DEFAULT_OUTLINE_WIDTH;
    		firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Returns the _replicateShape.
     * @return boolean
     */
    public boolean getReplicateShape() {
        return _replicateShape;
    }
    /**
     * Sets the _replicateShape.
     * @param _replicateShape The _replicateShape to set
     */
    public void setReplicateShape(boolean replicateShape) {
    	if (replicateShape != _replicateShape) {
	        this._replicateShape = replicateShape;
	        _defaultFlags &= ~FLAG_DEFAULT_REPLICATESHAPE;
	        firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Returns the _heightResizeable.
     * @return boolean
     */
    public boolean canResizeHeight() {
        return _heightResizeable;
    }
    /**
     * Returns the _preserveAspectRatio.
     * @return boolean
     */
    public boolean mustPreserveAspectRatio() {
        return _preserveAspectRatio;
    }
    /**
     * Returns the _widthResizeable.
     * @return boolean
     */
    public boolean canResizeWidth() {
        return _widthResizeable;
    }
    /**
     * Sets the _heightResizeable.
     * @param _heightResizeable The _heightResizeable to set
     */
    public void setHeightResizeable(boolean _heightResizeable) {
        this._heightResizeable = _heightResizeable;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    /**
     * Sets the _preserveAspectRatio.
     * @param _preserveAspectRatio The _preserveAspectRatio to set
     */
    public void setPreserveAspectRatio(boolean _preserveAspectRatio) {
        this._preserveAspectRatio = _preserveAspectRatio;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    /**
     * Sets the _widthResizeable.
     * @param _widthResizeable The _widthResizeable to set
     */
    public void setWidthResizeable(boolean _widthResizeable) {
        this._widthResizeable = _widthResizeable;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    /**
     * Returns the _font.
     * @return Font
     */
    public Font getFont() {
        return _font;
    }
    /**
     * Sets the _font.
     * @param _font The _font to set
     */
    public void setFont(Font font) {
    	if (!font.equals (_font)) {
	        this._font = font;
	        _defaultFlags &= ~FLAG_DEFAULT_FONT;
	        firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Returns the _labelHorizontalAlignment.
     * @return int
     */
    public int getLabelHorizontalAlignment() {
        return _labelHorizontalAlignment;
    }
    /**
     * Returns the _labelVerticalAlignment.
     * @return int
     */
    public int getLabelVerticalAlignment() {
        return _labelVerticalAlignment;
    }
    /**
     * Sets the _labelHorizontalAlignment.
     * @param _labelHorizontalAlignment The _labelHorizontalAlignment to set
     */
    public void setLabelHorizontalAlignment(int labelHorizontalAlignment) {
    	if (labelHorizontalAlignment != _labelHorizontalAlignment) {
    		this._labelHorizontalAlignment = labelHorizontalAlignment;
    		_defaultFlags &= ~FLAG_DEFAULT_HORIZONTAL_ALIGNMENT;
    		firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Sets the _labelVerticalAlignment.
     * @param _labelVerticalAlignment The _labelVerticalAlignment to set
     */
    public void setLabelVerticalAlignment(int labelVerticalAlignment) {
    	if (labelVerticalAlignment != _labelVerticalAlignment) {
	        this._labelVerticalAlignment = labelVerticalAlignment;
	        _defaultFlags &= ~FLAG_DEFAULT_VERTICAL_ALIGNMENT;
	        firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Returns the _labelSource.
     * @return int
     */
    public int getLabelSource() {
        return _labelSource;
    }
    /**
     * Sets the _labelSource.
     * @param _labelSource The _labelSource to set
     */
    public void setLabelSource(int labelSource) {
    	if (labelSource != _labelSource) {
	        this._labelSource = labelSource;
	        _defaultFlags &= ~FLAG_DEFAULT_LABEL_SOURCE;
	        firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    /**
     * Returns the _labelProperty.
     * @return String
     */
    public String getLabelProperty() {
        return _labelProperty;
    }
    /**
     * Sets the _labelProperty.
     * @param _labelProperty The _labelProperty to set
     */
    public void setLabelProperty(String _labelProperty) {
        this._labelProperty = _labelProperty;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    /**
     * Returns the _labelValue.
     * @return String
     */
    public String getLabelValue() {
        return _labelValue;
    }
    /**
     * Sets the _labelValue.
     * @param _labelValue The _labelValue to set
     */
    public void setLabelValue(String _labelValue) {
        this._labelValue = _labelValue;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    /**
     * Returns the _labelMargin.
     * @return int
     */
    public int getLabelMargin() {
        return _labelMargin;
    }
    /**
     * Sets the _labelMargin.
     * @param _labelMargin The _labelMargin to set
     */
    public void setLabelMargin(int labelMargin) {
    	if (_labelMargin != labelMargin) {
    		this._labelMargin = labelMargin;
    		_defaultFlags &= ~FLAG_DEFAULT_LABEL_MARGIN;
    		firePropertyChange(PROP_VIS_CHANGED, this);
    	}
    }
    public void setDecoration(String filename, int xRegion, int yRegion) {
        if (xRegion < 0 || xRegion > 2 || yRegion < 0 || yRegion > 2)
            return;
        //        ImageManager.disposeImage(filename);
        Image img =     ImageManager.createImage(filename, filename, false);
        _decorations[xRegion][yRegion] = new ManagedImage (img, this);
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    public void setDecoration(int xRegion, int yRegion, String text) {
        if (xRegion < 0 || xRegion > 2 || yRegion < 0 || yRegion > 2)
            return;
        String key = "__textdecoration_" + text;
        Dimension textSize = FigureUtilities.getStringExtents(text, _font);
        Image img =
            ImageManager.createImage(key, textSize.width, textSize.height, false);
        ManagedImage mimg = new ManagedImage (img, this);
        drawTextImage(img, text);

        _decorations[xRegion][yRegion] = mimg;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    /**
     * Draws a text image.
     * @param img The image to draw the text into
     * @param text The text to draw
     */
    private void drawTextImage(Image img, String text) {
        GC gc = new GC(img);
        gc.setBackground(getFillColor());
        gc.fillRectangle(img.getBounds());
        if (_font != null)
            gc.setFont(_font);
        gc.drawString(text, 0, 0);
        gc.dispose();
    }

    public void setDecoration(ManagedImage image, int xRegion, int yRegion) {
        if (xRegion < 0 || xRegion > 2 || yRegion < 0 || yRegion > 2)
            return;
        image.addListener(this);
        _decorations[xRegion][yRegion] = image;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    
    public ManagedImage getDecoration(int xRegion, int yRegion) {
        if (xRegion < 0 || xRegion > 2 || yRegion < 0 || yRegion > 2)
            return null;
        return _decorations[xRegion][yRegion];
    }

    public Image getIcon() {
        return _image;
    }

    public void setIcon(String filename) {
    	String key = generateKey ();
        ImageManager.disposeImage(key+filename);
        _image = ImageManager.createImage(key+filename, filename, false);
        _iconFilename = filename;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    public void setIcon(Image img) {
        _image = img;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    public void addVariant(VisVariant v) {
        int replacingIndex = -1;
        if (_variants.contains(v)) {
            replacingIndex = _variants.indexOf(v);
            _variants.remove(v);
        }
        if (replacingIndex != -1) {
            _variants.add(replacingIndex, v);
        } else {
            _variants.add(v);
        }
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    public void clearVariants() {
        _variants.clear();
    }
    public List getVariants() {
        return _variants;
    }
    public ElementVisualization getVariantToApply(AcmeElement e) {
        ElementVisualization appliedVariant = null;
        Iterator i = _variants.iterator();
        while (i.hasNext()) {
            VisVariant v = (VisVariant) i.next();
            if (v.appliesTo(e)) {
                if (appliedVariant == null) {
                    appliedVariant = v._style;
                } else {
                    appliedVariant = appliedVariant.unify(v._style);
                }
            }
        }
        if (appliedVariant != null) {
            appliedVariant = appliedVariant.unify(this);
        } else {
            appliedVariant = this;
        }
        return appliedVariant;
    }
    public ManagedImage getPreviewImage(AcmeElementType type) {
        if (_preview == null) {
            resetPreview(type);
        }
        return _preview;
    }
    
    public ManagedImage getLabellessPreviewImage (AcmeElementType type) {
    	if (_previewNoLabel == null) {
    		resetPreview (type);
    	}
    	return _previewNoLabel;
    }
    
    public ManagedImage getSmallImage (AcmeElementType type) {
    	if (_smallImage == null || _smallImage.isDisposed ()) {
    		resetSmallImage (type);
    	}
    	if (_smallImage.isDisposed ()) {
    		Debug.errorln("ElementVisualization.getSmallImage: " + getSmallImageLabel(type) + " is disposed");
    	}
    	return _smallImage;
    }
    
    /**
	 * 
	 */
	protected void resetSmallImage(AcmeElementType type) {
		String key = generateKey ()+"sm";
        if (ManagedImage.containsKey(key)) {
        	ManagedImage mip = ManagedImage.get (key);
        	_smallImage = mip;
        	mip.reference(this);
        	return;
        }
		int imageHeight = getSize ().height;
		int imageWidth = getSize ().width;
		float aspectRatio = (float )imageHeight/(float )imageWidth;
		if (imageHeight > 12) {
			imageHeight = 12;
			imageWidth = Utils.zeroGuard((int )(12/aspectRatio));
			if (imageWidth < 3) imageWidth = 20;
		}
		if (imageWidth > 20) {
			imageWidth = 20;
			imageHeight = Utils.zeroGuard ((int )(20*aspectRatio));
		}
		
		Image tempImage = new Image (null, imageWidth, imageHeight);
		GC gc = new GC (tempImage);
		SWTGraphics graphics = new SWTGraphics (gc);
		graphics.pushState ();
		graphics.setForegroundColor(Constants.TRANSPARENT_COLOR);
		graphics.setBackgroundColor(Constants.TRANSPARENT_COLOR);
		graphics.fillRectangle (0, 0, imageWidth, imageHeight);
		graphics.popState ();
		Rectangle rectangle = new Rectangle (0, 0, imageWidth-1, imageHeight-1);
		fillShape (graphics, rectangle, null, 0);
		outlineShape (graphics, rectangle);
		
		ImageData data = tempImage.getImageData();
		data.transparentPixel = data.palette.getPixel(Constants.TRANSPARENT_COLOR.getRGB ());
		Image smi = new Image (null, data);
		ImageManager.imageCreated(getSmallImageLabel(type), smi);
		_smallImage = new ManagedImage (smi, this);
		ManagedImage.put(key, _smallImage);
		tempImage.dispose ();
		gc.dispose ();
		graphics.dispose ();
	}
	
	
	protected void resetPreview(AcmeElementType type) {
        cleanUpPreviewImages (type);
        if (_size.width <= 0 || _size.height <= 0) {
            return;
        }
        String key = generateKey ();
        if (ManagedImage.containsKey(key)) {
        	ManagedImage mip = ManagedImage.get (key);
        	_preview = mip;
        	mip.reference(this);
        	mip = ManagedImage.get (key+"nl");
        	_previewNoLabel = mip;
        	if (mip != null) {
        		mip.reference (this);
        		return;
        	}
        }
        String theLabel = generatePreviewLabel (type);
        Dimension size = new Dimension(_size);
        org.eclipse.draw2d.geometry.Point shapePoint =
            new org.eclipse.draw2d.geometry.Point(0, 0);
        if (theLabel != "") {
            Font font = getFont();
            Dimension textSize =
                FigureUtilities.getStringExtents(theLabel, font);
            int min = textSize.width + 20;
            switch (_labelHorizontalAlignment) {
                case HorizontalTextAlignmentOptions.LEFT_OUTSIDE :
                    size.width = _size.width + min;
                    shapePoint = shapePoint.translate(min, 0);
                    break;
                case HorizontalTextAlignmentOptions.RIGHT_OUTSIDE :
                    size.width = _size.width + min;
                    break;
                default :
                    size.width =
                        _size.width + 20 > min ? _size.width + 20 : min;
                    shapePoint =
                        shapePoint.translate((size.width - _size.width) / 2, 0);
            }
            switch (_labelVerticalAlignment) {
                case VerticalTextAlignmentOptions.ABOVE :
                    size.height = _size.height + textSize.height + 20;
                    shapePoint = shapePoint.translate(0, textSize.height + 20);
                    break;
                case VerticalTextAlignmentOptions.BELOW :
                    size.height = _size.height + textSize.height + 20;
                    break;
                default :
                    size.height =
                        _size.height > textSize.height + 20
                            ? _size.height
                            : textSize.height + 20;
                    shapePoint =
                        shapePoint.translate(
                            0,
                            (size.height - _size.height) / 2);
            }
        }
        if (getReplicateShape()) {
            size = size.expand(8, 8);
        }
        Image pim_preview = new Image(Display.getDefault(), size.width, size.height);
        Image pim_previewNoLabel = new Image (Display.getDefault (), size.width, size.height);
        _preview = new ManagedImage (pim_preview, this);
        ManagedImage.put(key,_preview);
        _previewNoLabel = new ManagedImage (pim_previewNoLabel, this);
        ManagedImage.put (key+"nl", _previewNoLabel);
        ImageManager.imageCreated(
           key,
            _preview.getImage ());
        ImageManager.imageCreated (key+"nl",
        			_previewNoLabel.getImage ());
        generatePreview(pim_preview, true, theLabel, shapePoint);
        generatePreview(pim_previewNoLabel, false, theLabel, shapePoint);
        
    }
    protected String generatePreviewLabel(AcmeElementType type) {
		String theLabel = "";
        switch (getLabelSource()) {
            case LabelOptions.ACMEID :
                theLabel = "Identifier";
                break;
            case LabelOptions.STATICLABEL :
                theLabel = getLabelValue();
                break;
            case LabelOptions.ACMETYPE :
                theLabel = "<<";
                if (type == null) {
                    theLabel += "type";
                } else {
                    theLabel += type.getTypeName();
                }
                theLabel += ">>";
                break;
            case LabelOptions.ACMEIDANDTYPE :
                theLabel = "id : ";
                if (type == null)
                    theLabel += "type";
                else
                    theLabel += type.getTypeName();
                break;
            case LabelOptions.ACMEPROPERTYVALUE :
                theLabel = "(";
                if (getLabelProperty() == null)
                    theLabel += "value";
                else
                    theLabel += getLabelProperty();
                theLabel += ")";
                break;
        }
		return theLabel;
	}
	protected void cleanUpPreviewImages(AcmeElementType type) {
		String key = generateKey ();
		if (_preview != null) {
			ManagedImage mip = ManagedImage.get (key);
			ImageManager.imageDisposed(
                key,
                _preview.getImage ());
			_preview.dispose (this);
//            _preview.dispose();
            _preview = null;
        }
        if (_previewNoLabel != null) {
        	ManagedImage mip = ManagedImage.get (key+"nl");
        	ImageManager.imageDisposed(key+"nl",
        			_previewNoLabel.getImage ());
        	_previewNoLabel.dispose(this);
//        	_previewNoLabel.dispose ();
        	_previewNoLabel = null;
        }
        if (_smallImage != null /*&& !_isDefault*/) {
        	ManagedImage mip = ManagedImage.get (key+"sm");
        	ImageManager.imageDisposed (key+"sm", _smallImage.getImage ());
        	_smallImage.dispose (this);
//        	mip.dispose(this);
        	_smallImage = null;
        }
//        _imageListeners.firePropertyChange(PROP_IMAGES_DISPOSED, null, this);
	}
	protected String getSmallImageLabel(AcmeElementType type) {
		return (type==null?"":type.getTypeName () + "Visualization.smallImage" + this);
	}
	protected void generatePreview(Image preview, boolean doLabel, String theLabel, org.eclipse.draw2d.geometry.Point shapePoint) {
		// Don't forget to change in RoleVisualization
		GC previewGC = new GC(preview);
        SWTGraphics previewGraphics = new SWTGraphics(previewGC);
        previewGraphics.setBackgroundColor(ElementColorConstants.white);
        Rectangle labelRect =
            new Rectangle(
                shapePoint.x,
                shapePoint.y,
                _size.width,
                _size.height);
        previewGraphics.fillRectangle(labelRect);
        org.eclipse.draw2d.geometry.Rectangle r =
            new org.eclipse.draw2d.geometry.Rectangle(shapePoint, _size);
        if (getReplicateShape()) {
            org.eclipse.draw2d.geometry.Point shadowPoint =
                shapePoint.getCopy().translate(6, 6);
            org.eclipse.draw2d.geometry.Rectangle rectShadow =
                new org.eclipse.draw2d.geometry.Rectangle(shadowPoint, _size);
            fillShape(previewGraphics, rectShadow, null, 0);
            outlineShape(previewGraphics, rectShadow);
            shadowPoint = shapePoint.getCopy().translate(3, 3);
            rectShadow =
                new org.eclipse.draw2d.geometry.Rectangle(shadowPoint, _size);
            fillShape(previewGraphics, rectShadow, null, 0);
            outlineShape(previewGraphics, rectShadow);
        }
        fillShape(previewGraphics, r, null, 0);
        outlineShape(previewGraphics, r);
        if (doLabel)
        	drawLabel(previewGraphics, labelRect, theLabel, _textColor);
        previewGraphics.dispose();
        previewGC.dispose();
	}
	
	public abstract ElementVisualization unify (ElementVisualization from );
	
	protected void unify(ElementVisualization from, ElementVisualization to) {
    

        to._cornerSize = from._cornerSize;
        to._direction =
            (this._direction == PositionConstants.NONE)
                ? from._direction
                : this._direction;
        to._fillColor =
            (this._fillColor == DEFAULT_FILLCOLOR)
                ? from._fillColor
                : this._fillColor;
        to._font = (this._font == DEFAULT_FONT) ? from._font : this._font;
        to._heightResizeable =
            (this._heightResizeable || from._heightResizeable);
        to._labelHorizontalAlignment =
            (this._labelHorizontalAlignment
                == HorizontalTextAlignmentOptions.DEFAULT)
                ? from._labelHorizontalAlignment
                : this._labelHorizontalAlignment;
        to._labelMargin =
            (this._labelMargin == DEFAULT_LABEL_MARGIN)
                ? from._labelMargin
                : this._labelMargin;
        to._labelProperty =
            (this._labelProperty == null)
                ? from._labelProperty
                : this._labelProperty;
        to._labelSource =
            (this._labelSource == LabelOptions.DEFAULT)
                ? from._labelSource
                : this._labelSource;
        to._labelVerticalAlignment =
            (this._labelVerticalAlignment
                == VerticalTextAlignmentOptions.DEFAULT)
                ? from._labelVerticalAlignment
                : this._labelVerticalAlignment;
        to._internal_orientation =
            (this._internal_orientation == DEFAULT_INTERNAL_ORIENTATION)
                ? from._internal_orientation
                : this._internal_orientation;
        to._outlineColor =
            (this._outlineColor == DEFAULT_OUTLINECOLOR)
                ? from._outlineColor
                : this._outlineColor;
        to._outlineStyle =
            (this._outlineStyle == DEFAULT_OUTLINESTYLE)
                ? from._outlineStyle
                : this._outlineStyle;
        to._outlineWidth =
            (this._outlineWidth == DEFAULT_OUTLINEWIDTH)
                ? from._outlineWidth
                : this._outlineWidth;
        to._points = (this._points == null) ? from._points : this._points;
        to._preserveAspectRatio =
            this._preserveAspectRatio || from._preserveAspectRatio;
        to._replicateShape = this._replicateShape || from._replicateShape;
        to._shape = (this._shape == Shape.DEFAULT) ? from._shape : this._shape;
        to._size = (this._size == DEFAULT_SIZE) ? from._size : this._size;
        to._textColor =
            (this._textColor == DEFAULT_TEXTCOLOR)
                ? from._textColor
                : this._textColor;
        to._widthResizeable = this._widthResizeable || from._widthResizeable;
        for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                if (from._decorations[i][j] != null) {
                    to._decorations[i][j] = from._decorations[i][j];
                } else {
                    to._decorations[i][j] = this._decorations[i][j];
                }
            }
        }
      
    }
	
	protected boolean shouldBlendPart (ElementVisualization vis, long part) {
		boolean isDefault = isDefaultPart (part);
		boolean isVisDefault = vis.isDefaultPart (part);
		return isDefaultPart (part) && !vis.isDefaultPart(part);
	}
	/**
	 * @param vis
	 */
	public void blendVisualization(ElementVisualization vis, boolean treatVisAsSuper) {
		_copying = true;
		if (shouldBlendPart (vis, FLAG_DEFAULT_CORNERSIZE)) {
			setCornerSize (vis.getCornerSize());
			forceDefault(FLAG_DEFAULT_CORNERSIZE);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_DIRECTION)) {
			setDirection (vis.getDirection ());
			forceDefault (FLAG_DEFAULT_DIRECTION);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_FILLCOLOR)) {
			setFillColor(vis.getFillColor());
			forceDefault (FLAG_DEFAULT_FILLCOLOR);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_FONT)) {
			setFont (vis.getFont ());
			forceDefault (FLAG_DEFAULT_FONT);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_HORIZONTAL_ALIGNMENT)) {
			setLabelHorizontalAlignment(vis.getLabelHorizontalAlignment());
			forceDefault (FLAG_DEFAULT_HORIZONTAL_ALIGNMENT);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_LABEL_MARGIN)) {
			setLabelMargin(vis.getLabelMargin ());
			forceDefault (FLAG_DEFAULT_LABEL_MARGIN);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_LABEL_SOURCE)) {
			setLabelSource(vis.getLabelSource());
			forceDefault (FLAG_DEFAULT_LABEL_SOURCE);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_OUTLINE_WIDTH) ) {
			setOutlineWidth(vis.getOutlineWidth());
			forceDefault (FLAG_DEFAULT_OUTLINE_WIDTH);
		}
		if (shouldBlendPart(vis, FLAG_DEFAULT_OUTLINECOLOR)) {
			setOutlineColor(vis.getOutlineColor());
			forceDefault (FLAG_DEFAULT_OUTLINECOLOR);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_OUTLINESTYLE)) {
			setOutlineStyle(vis.getOutlineStyle());
			forceDefault (FLAG_DEFAULT_OUTLINESTYLE);
		}
		if (shouldBlendPart(vis, FLAG_DEFAULT_REPLICATESHAPE)) {
			setReplicateShape(vis.getReplicateShape());
			forceDefault (FLAG_DEFAULT_REPLICATESHAPE);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_RESIZETOFIT_LABEL)) {
			setResizeToFitLabel(vis.shouldResizeToFitLabel());
			forceDefault(FLAG_DEFAULT_RESIZETOFIT_LABEL);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_SHAPE)) {
			setShape(vis.getShape ());
			forceDefault (FLAG_DEFAULT_SHAPE);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_SIZE)) {
			setSize (vis.getSize());
			forceDefault (FLAG_DEFAULT_SIZE);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_TEXTCOLOR)) {
			setTextColor(vis.getTextColor ());
			forceDefault (FLAG_DEFAULT_TEXTCOLOR);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_VERTICAL_ALIGNMENT)) {
			setLabelVerticalAlignment(vis.getLabelVerticalAlignment());
			forceDefault (FLAG_DEFAULT_VERTICAL_ALIGNMENT);
		}
		if (shouldBlendPart (vis, FLAG_DEFAULT_WORDWRAP)) {
			setWordwrap(vis.getWordwrap());
			forceDefault (FLAG_DEFAULT_WORDWRAP);
		}
		_copying = false;
		
	}
	
    public void copyInto(ElementVisualization from) {
        _copying = true;
        this._cornerSize = from._cornerSize;
        this._direction = from._direction;
        this._fillColor = from._fillColor;
        this._font = from._font;
        this._heightResizeable = from._heightResizeable;
        this._labelHorizontalAlignment = from._labelHorizontalAlignment;
        this._labelMargin = from._labelMargin;
        this._labelProperty = from._labelProperty;
        this._labelSource = from._labelSource;
        this._labelValue = from._labelValue;
        this._labelVerticalAlignment = from._labelVerticalAlignment;
        this._labelWordwrap = from._labelWordwrap;
        this._internal_orientation = from._internal_orientation;
        this._outlineColor = from._outlineColor;
        this._outlineStyle = from._outlineStyle;
        this._outlineWidth = from._outlineWidth;
        this._points = from._points;
        this._preserveAspectRatio = from._preserveAspectRatio;
        this._replicateShape = from._replicateShape;
        this._resizeToFitLabel = from._resizeToFitLabel;
        this._shape = from._shape;
        this._size = from._size.getCopy();
        this._textColor = from._textColor;
        this._widthResizeable = from._widthResizeable;
        this._defaultFlags &= from._defaultFlags;
        //        this._images = from._images;
        for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                if (from._decorations[i][j] != null) {
                    this._decorations[i][j] = from._decorations[i][j];
                }

            }
        }
        this._variants.clear();
        Iterator i = from._variants.iterator();
        while (i.hasNext()) {
            VisVariant vis =
                (VisVariant) i.next();
            _variants.add(vis);
        }
        _copying = false;
        firePropertyChange(PROP_VIS_CHANGED, this);
        this.setDirty(from.isDirty());
    }
    private static void exportColor(Color col, Element el) {
        el.setAttribute("red", new Integer(col.getRed()).toString());
        el.setAttribute("green", new Integer(col.getGreen()).toString());
        el.setAttribute("blue", new Integer(col.getBlue()).toString());
    }
    private static Color importColor(Element el) {
        if (el == null) { // return right away
            return null;
        }
        try {
            int r = Integer.parseInt(el.getAttribute("red"));
            int g = Integer.parseInt(el.getAttribute("green"));
            int b = Integer.parseInt(el.getAttribute("blue"));
            return SWTResourceManager.getColor (r, g, b);
        } catch (NumberFormatException e) {
            return SWTResourceManager.getColor (0, 0, 0);
        }
    }
    private static void exportSize(Dimension size, Element el) {
        el.setAttribute("width", new Integer(size.width).toString());
        el.setAttribute("height", new Integer(size.height).toString());
    }
    private static Dimension importSize(Element el) {
        if (el == null) {
            return null;
        }
        try {
            return new Dimension(
                Integer.parseInt(el.getAttribute("width")),
                Integer.parseInt(el.getAttribute("height")));
        } catch (NumberFormatException e) {
            return new Dimension(0, 0);
        }
    }
    private static void exportFont(Font font, Element el) {
        if (font == null) {
            return;
        }
        FontData fd = font.getFontData()[0];
        el.setAttribute("name", fd.getName());
        el.setAttribute("height", new Integer(fd.getHeight()).toString());
        el.setAttribute(
            "bold",
            ((fd.getStyle() & SWT.BOLD) == 1) ? "true" : "false");
        el.setAttribute(
            "italics",
            ((fd.getStyle() & SWT.ITALIC) == 1) ? "true" : "false");
    }
    private static Font importFont(Element el) {
        if (el == null) { // return right away
            return null;
        }
        String name = el.getAttribute("name");
        if (name.equals(""))
            name = AcmeStudioPlugin.getDefaultFontName();
        int height = 10;
        try {
            height = Integer.parseInt(el.getAttribute("height"));
        } catch (NumberFormatException e) {
            // set to default
            height = 10;
        }
        int style = 0;
        if (el.getAttribute("bold").equals("true"))
            style |= SWT.BOLD;
        if (el.getAttribute("italics").equals("true"))
            style |= SWT.ITALIC;
        if (style == 0)
            style = SWT.NORMAL;
        return SWTResourceManager.getFont  (name, height, style);
    }
    static boolean importOn(Element el) {
        if (el != null && el.getAttribute("on").equalsIgnoreCase("true")) {
            return true;
        }
        return false;
    }

    static int importOrientation(Element el) {
        if (el != null
            && el.getAttribute("orientation").equalsIgnoreCase("in")) {
            return Orientation.IN;
        } else if (
            el != null
                && el.getAttribute("orientation").equalsIgnoreCase("out")) {
            return Orientation.OUT;
        }
        return -1;
    }

    static LinkedList importAllowedTypes(Element el) {
        LinkedList result = null;
        if (el != null) {
            String allowedTypes = el.getAttribute("allowedTypes");
            String[] typeList = allowedTypes.split(";");
            result = new LinkedList();
            for (int i = 0; i < typeList.length; i++) {
                result.add(typeList[i]);
            }
        }
        return result;
    }

    static void exportAllowedTypes(Element el, LinkedList typeList) {
        if (el != null && typeList != null) {
            Iterator iter = typeList.iterator();
            String types = "";
            while (iter.hasNext()) {
                String type = (String) iter.next();
                types += types.equals("") ? type : ";" + type;
            }
            el.setAttribute("allowedTypes", types);
        }
    }
    public static String directionToString(int direction) {
        String label = "None";
        switch (direction) {
            case PositionConstants.NORTH :
                label = "North";
                break;
            case PositionConstants.SOUTH :
                label = "South";
                break;
            case PositionConstants.EAST :
                label = "East";
                break;
            case PositionConstants.WEST :
                label = "West";
                break;
        }
        return label;
    }
    public static int labelToDirection(String label) {
        String l = label.toLowerCase();
        if (l.equals("north"))
            return PositionConstants.NORTH;
        else if (l.equals("south"))
            return PositionConstants.SOUTH;
        else if (l.equals("east"))
            return PositionConstants.EAST;
        else if (l.equals("west"))
            return PositionConstants.WEST;
        else
            return SWT.NONE;
    }
    public static String lineStyleToString(int style) {
        String label = "None";
        switch (style) {
            case SWT.LINE_SOLID :
                label = "Solid";
                break;
            case SWT.LINE_DASH :
                label = "Dash";
                break;
            case SWT.LINE_DASHDOT :
                label = "DashDot";
                break;
            case SWT.LINE_DASHDOTDOT :
                label = "DashDotDot";
                break;
            case SWT.LINE_DOT :
                label = "Dot";
                break;
        }
        return label;
    }
    public static int labelToLineStyle(String label) {
        String l = label.toLowerCase();
        if (l.equals("solid"))
            return SWT.LINE_SOLID;
        else if (l.equals("dash"))
            return SWT.LINE_DASH;
        else if (l.equals("dot"))
            return SWT.LINE_DOT;
        else if (l.equals("dashdot"))
            return SWT.LINE_DASHDOT;
        else if (l.equals("dashdotdot"))
            return SWT.LINE_DASHDOTDOT;
        else
            return SWT.NONE;
    }
    public Element export(Document doc) {
        Element root = doc.createElement("visualization");
        // Needs to be done above. root.setAttribute ("name", "
        Element shapeNode = doc.createElement("shape");
        shapeNode.setAttribute("type", (String) Shape.SHAPENAMES.get(_shape));
        shapeNode.setAttribute("direction", directionToString(_direction));
        shapeNode.setAttribute("replicate", _replicateShape ? "true" : "false");
        shapeNode.setAttribute(
            "outlineStyle",
            lineStyleToString(_outlineStyle));
        shapeNode.setAttribute(
            "outlineWidth",
            new Integer(_outlineWidth).toString());

        if (_image != null) {
            ImageLoader iLoader = new ImageLoader();
            iLoader.data = new ImageData[] { _image.getImageData()};
            ByteArrayOutputStream imageStream = new ByteArrayOutputStream();
            iLoader.save(imageStream, SWT.IMAGE_GIF);
            byte[] bytes = imageStream.toByteArray();
            String theBytes = new Integer(bytes.length).toString() + ",";
            for (int k = 0; k < bytes.length; k++)
                theBytes += bytes[k] + ",";
            CDATASection text = doc.createCDATASection(theBytes);
            Element imageNode = doc.createElement("image");
            imageNode.appendChild(text);
            imageNode.setAttribute("type", "icon");
            shapeNode.appendChild(imageNode);
        }

        for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                if (_decorations[i][j] != null) {

                    ImageLoader iLoader = new ImageLoader();
                    iLoader.data =
                        new ImageData[] { _decorations[i][j].getImageData()};
                    ByteArrayOutputStream imageStream =
                        new ByteArrayOutputStream();
                    try {
                        iLoader.save(imageStream, SWT.IMAGE_GIF);
                    } catch (Exception e) {
                        iLoader.save(imageStream, SWT.IMAGE_BMP);
                    }

                    byte[] bytes = imageStream.toByteArray();
                    String theBytes =
                        new Integer(bytes.length).toString() + ",";
                    for (int k = 0; k < bytes.length; k++) {
                        byte b = bytes[k];
                        theBytes += b + ",";
                    }
                    CDATASection text = doc.createCDATASection(theBytes);
                    Element imageNode = doc.createElement("image");
                    imageNode.appendChild(text);
                    imageNode.setAttribute("type", "decoration");
                    imageNode.setAttribute(
                        "xRegion",
                        new Integer(i).toString());
                    imageNode.setAttribute(
                        "yRegion",
                        new Integer(j).toString());
                    shapeNode.appendChild(imageNode);
                }
            }
        }
        Element fillColorNode = doc.createElement("fillColor");
        exportColor(_fillColor, fillColorNode);
        shapeNode.appendChild(fillColorNode);
        Element outlineColorNode = doc.createElement("outlineColor");
        exportColor(_outlineColor, outlineColorNode);
        shapeNode.appendChild(outlineColorNode);
        Element sizeNode = doc.createElement("size");
        exportSize(_size, sizeNode);
        sizeNode.setAttribute(
            "heightResizeable",
            _heightResizeable ? "true" : "false");
        sizeNode.setAttribute(
            "widthResizeable",
            _widthResizeable ? "true" : "false");
        sizeNode.setAttribute(
            "preserveAspectRatio",
            _preserveAspectRatio ? "true" : "false");
        shapeNode.appendChild(sizeNode);
        root.appendChild(shapeNode);
        Element labelNode = doc.createElement("label");
        labelNode.setAttribute(
            "source",
            (String) LabelOptions.LABELNAMES.get(_labelSource));
        labelNode.setAttribute(
            "horizAlign",
            (String) HorizontalTextAlignmentOptions.HORIZALIGNNAMES.get(
                _labelHorizontalAlignment));
        labelNode.setAttribute(
            "vertAlign",
            (String) VerticalTextAlignmentOptions.VERTALIGNNAMES.get(
                _labelVerticalAlignment));
        labelNode.setAttribute("wordwrap", getWordwrap() ? "true" : "false");
        labelNode.setAttribute("margin", new Integer(_labelMargin).toString());
        if (_labelSource == LabelOptions.ACMEPROPERTYVALUE) {
            labelNode.setAttribute("property", _labelProperty);
        } else if (_labelSource == LabelOptions.STATICLABEL) {
            labelNode.setAttribute("value", _labelValue);
        }
        Element labelColNode = doc.createElement("labelColor");
        exportColor(_textColor, labelColNode);
        labelNode.appendChild(labelColNode);
        root.appendChild(labelNode);
        Element fontNode = doc.createElement("font");
        exportFont(_font, fontNode);
        root.appendChild(fontNode);
        Iterator i = _variants.iterator();
        while (i.hasNext()) {
            VisVariant v = (VisVariant) i.next();
            Element variantNode = doc.createElement("variant");
            variantNode.setAttribute("name", v._name);
            variantNode.setAttribute(
                "exclusive",
                v._exclusive ? "true" : "false");
            Element conditionNode = doc.createElement("condition");
            if (v._type == VisVariant.PROPERTY_TYPE) {
                conditionNode.setAttribute("property", v._property);
                conditionNode.setAttribute(
                    "operator",
                    (String) VisVariant.OPNAMES.get(v._operator));
                conditionNode.setAttribute("value", v._value.toString());
            }
            else if (v._type == VisVariant.CONTEXT_TYPE) {
                conditionNode.setAttribute("containedIn", v._contextType);
                conditionNode.setAttribute("type", "contextual");
            }
            variantNode.appendChild(conditionNode);
            variantNode.appendChild(v._style.export(doc));
            root.appendChild(variantNode);
        }
        exportTypeSpecificVisualization(doc, root);
        root.setAttribute ("hideFromPalette", _hideFromPalette?"true":"false");
        root.appendChild(exportDefaults (doc));
        
        
        return root;
    }
    
    /**
	 * @param defaults
	 * @param string
	 * @return
	 */
	protected static boolean getBooleanAttribute(Element defaults, String string) {
		String b = defaults.getAttribute (string);
		if (b == "true") return true;
		else return false;
	}
	
	protected  void setImportedDefault (ElementVisualization vis, Element node, String attribute, long flag) {
		if (getBooleanAttribute (node, attribute)) vis.forceDefault (flag);
	}
	
    protected  void importDefaults (ElementVisualization vis, Element visElement) {
    	NodeList dfs = visElement.getElementsByTagName("defaults");
    	if (dfs.getLength() > 0) {
    		Element defaults = (Element )dfs.item (0);
    		setImportedDefault (vis, defaults, "direction", FLAG_DEFAULT_DIRECTION);
    		setImportedDefault (vis, defaults, "fillcolor", FLAG_DEFAULT_FILLCOLOR);
    		setImportedDefault (vis, defaults, "font", FLAG_DEFAULT_FONT);
    		setImportedDefault (vis, defaults, "horizAlign", FLAG_DEFAULT_HORIZONTAL_ALIGNMENT);
    		setImportedDefault (vis, defaults, "labelMargin", FLAG_DEFAULT_LABEL_MARGIN);
    		setImportedDefault (vis, defaults, "labelSource", FLAG_DEFAULT_LABEL_SOURCE);
    		setImportedDefault (vis, defaults, "outlineWidth", FLAG_DEFAULT_OUTLINE_WIDTH);
    		setImportedDefault (vis, defaults, "outlineColor", FLAG_DEFAULT_OUTLINECOLOR);
    		setImportedDefault (vis, defaults, "outlineStyle", FLAG_DEFAULT_OUTLINESTYLE);
    		setImportedDefault (vis, defaults, "replicateShape", FLAG_DEFAULT_REPLICATESHAPE);
    		setImportedDefault (vis, defaults, "resizeToFitLabel", FLAG_DEFAULT_RESIZETOFIT_LABEL);
    		setImportedDefault (vis, defaults, "shape", FLAG_DEFAULT_SHAPE);
    		setImportedDefault (vis, defaults, "size", FLAG_DEFAULT_SIZE);
    		setImportedDefault (vis, defaults, "textcolor", FLAG_DEFAULT_TEXTCOLOR);
    		setImportedDefault (vis, defaults, "vertAlign", FLAG_DEFAULT_VERTICAL_ALIGNMENT);
    		setImportedDefault (vis, defaults, "wordwrap", FLAG_DEFAULT_WORDWRAP);
    	}
    }
   
    protected void exportDefaultAttribute (Element defaults, String attribute, long flag) {
    	if (isDefaultPart (flag)) defaults.setAttribute (attribute, "true");
    }
    
	public Element  exportDefaults (Document doc) {
    	Element defaults = doc.createElement("defaults");
    	if (isDefaultPart(FLAG_DEFAULT_DIRECTION)) {
    		defaults.setAttribute("direction", "true");
    	}
    	if (isDefaultPart(FLAG_DEFAULT_FILLCOLOR)) {
    		defaults.setAttribute ("fillcolor", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_FONT)) {
    		defaults.setAttribute ("font", "true");
    	}
    	if (isDefaultPart(FLAG_DEFAULT_HORIZONTAL_ALIGNMENT)) {
    		defaults.setAttribute ("horizAlign", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_LABEL_MARGIN)) {
    		defaults.setAttribute ("labelMargin", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_LABEL_SOURCE)) {
    		defaults.setAttribute ("labelSource", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_OUTLINE_WIDTH)) {
    		defaults.setAttribute ("outlineWidth", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_OUTLINECOLOR)) {
    		defaults.setAttribute("outlineColor", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_OUTLINESTYLE)) {
    		defaults.setAttribute ("outlineStyle", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_REPLICATESHAPE)) {
    		defaults.setAttribute ("replicateShape", "true");
    	}
    	if (isDefaultPart(FLAG_DEFAULT_RESIZETOFIT_LABEL)) {
    		defaults.setAttribute ("resizeToFitLabel", "true");
    	}
    	if (isDefaultPart(FLAG_DEFAULT_SHAPE)) {
    		defaults.setAttribute ("shape", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_SIZE)) {
    		defaults.setAttribute ("size", "true");
    	}
    	if (isDefaultPart(FLAG_DEFAULT_TEXTCOLOR)) {
    		defaults.setAttribute("textcolor", "true");
    	}
    	if (isDefaultPart (FLAG_DEFAULT_VERTICAL_ALIGNMENT)) {
    		defaults.setAttribute ("vertAlign", "true");
    	}
    	if (isDefaultPart(FLAG_DEFAULT_WORDWRAP)) {
    		defaults.setAttribute ("wordwrap", "true");
    	}
    	return defaults;
    }
    
    public void exportTypeSpecificVisualization(Document doc, Element root) {
    }
    public static ElementVisualization importVisualization(Element visElement) {
        String visType = visElement.getAttribute("type");
        ElementVisualization vis = null;
        if (visType != null) {
            if (visType.equals(VIS_TYPE_PORT)) {
                vis = new PortVisualization();
            } else if (visType.equals(VIS_TYPE_ROLE)) {
                vis = new RoleVisualization();
            } else if (visType.equals(VIS_TYPE_COMPONENT)) {
                vis = new ComponentVisualization();
            } else if (visType.equals(VIS_TYPE_CONNECTOR)) {
                vis = new ConnectorVisualization();
            } else {
            	vis = new GenericVisualization ();
            }
        }
        String hidden = visElement.getAttribute ("hideFromPalette");
        if (hidden != null && hidden.equals ("true")) vis._hideFromPalette = true;
//        if (vis == null) {
//            vis = new ElementVisualization();
//        }
        Element shapeNode =
            (Element) visElement.getElementsByTagName("shape").item(0);
        vis.setShape(Shape.SHAPENAMES.indexOf(shapeNode.getAttribute("type")));
        vis.setDirection(
            labelToDirection(shapeNode.getAttribute("direction")));
        vis.setReplicateShape(
            Boolean
                .valueOf(shapeNode.getAttribute("replicate"))
                .booleanValue());
        Integer outlineWidth =
            Integer.getInteger(shapeNode.getAttribute("outlineWidth"));
        if (outlineWidth != null) {
            vis.setOutlineWidth(outlineWidth.intValue());
        }
        vis.setOutlineStyle(
            labelToLineStyle(shapeNode.getAttribute("outlineStyle")));
        NodeList imageNodes = shapeNode.getElementsByTagName("image");
        for (int i = 0; i < imageNodes.getLength(); i++) {

            Element imageNode = (Element) imageNodes.item(i);
            if ((imageNode.getAttribute("type") != null
                && imageNode.getAttribute("type").equals("decoration"))
                || imageNode.getAttribute("xRegion") != null) {
                int xRegion = 1, yRegion = 1;
                if (imageNode.getAttribute("xRegion") != null
                    && !"".equals(imageNode.getAttribute("xRegion"))) {
                    xRegion =
                        Integer.parseInt(imageNode.getAttribute("xRegion"));
                    yRegion =
                        Integer.parseInt(imageNode.getAttribute("yRegion"));
                }

                ByteArrayInputStream bis = importImageData(imageNode);
                ImageLoader iLoader = new ImageLoader();
                vis.setDecoration(
                    new ManagedImage (new Image(Display.getDefault(), iLoader.load(bis)[0]), vis),
                    xRegion,
                    yRegion);
//                ImageManager.imageCreated(
//                    "Visualization.importVisualization"
//                        + xRegion
//                        + ","
//                        + yRegion,
//                    vis.getDecoration(xRegion, yRegion));
            } else {
                ByteArrayInputStream bis = importImageData(imageNode);
                ImageLoader iLoader = new ImageLoader();
                vis.setIcon(
                    new Image(Display.getDefault(), iLoader.load(bis)[0]));

            }
        }
        Element fillColNode =
            (Element) shapeNode.getElementsByTagName("fillColor").item(0);
        vis.setFillColor(importColor(fillColNode));
        Element outlineColNode =
            (Element) shapeNode.getElementsByTagName("outlineColor").item(0);
        vis.setOutlineColor(importColor(outlineColNode));
        Element sizeNode =
            (Element) shapeNode.getElementsByTagName("size").item(0);
        vis.setSize(importSize(sizeNode));
        vis.setHeightResizeable(
            (sizeNode.getAttribute("heightResizeable").equals("true")
                ? true
                : false));
        String s = sizeNode.getAttribute("widtheResizeable");
        if (s != null && s.length() != 0) {
            vis.setWidthResizeable(
                sizeNode.getAttribute("widtheResizeable").equals("true")
                    ? true
                    : false);
        } else {
            vis.setWidthResizeable(
                (sizeNode.getAttribute("widthResizeable").equals("true")
                    ? true
                    : false));
        }
        vis.setPreserveAspectRatio(
            (sizeNode.getAttribute("preserveAspectRatio").equals("true")
                ? true
                : false));
        Element labelNode =
            (Element) visElement.getElementsByTagName("label").item(0);
        vis.setLabelSource(
            LabelOptions.LABELNAMES.indexOf(labelNode.getAttribute("source")));
        vis.setLabelHorizontalAlignment(
            HorizontalTextAlignmentOptions.HORIZALIGNNAMES.indexOf(
                labelNode.getAttribute("horizAlign")));
        vis.setLabelVerticalAlignment(
            VerticalTextAlignmentOptions.VERTALIGNNAMES.indexOf(
                labelNode.getAttribute("vertAlign")));
        vis.setWordwrap(
            (labelNode.getAttribute("wordwrap").equals("true") ? true : false));
        vis.setLabelMargin(Integer.parseInt(labelNode.getAttribute("margin")));
        if (vis.getLabelSource() == LabelOptions.ACMEPROPERTYVALUE) {
            vis.setLabelProperty(labelNode.getAttribute("property"));
        } else if (vis.getLabelSource() == LabelOptions.STATICLABEL) {
            vis.setLabelValue(labelNode.getAttribute("value"));
        }
        fillColNode =
            (Element) labelNode.getElementsByTagName("labelColor").item(0);
        Color color = importColor(fillColNode);
        if (color != null) {
            vis.setTextColor(color);
        }
        Element fontNode =
            (Element) visElement.getElementsByTagName("font").item(0);
        Font font = importFont(fontNode);
        if (font != null) {
            vis.setFont(font);
        }
        NodeList variants = visElement.getElementsByTagName("variant");
        for (int i = 0; i < variants.getLength(); i++) {
            Element variantNode = (Element) variants.item(i);
            String name = variantNode.getAttribute("name");
            String exclusive = variantNode.getAttribute("exclusive");

            Element condition =
                (Element) variantNode.getElementsByTagName("condition").item(0);
            VisVariant var = new VisVariant();
            var._name = name;
            var._exclusive = exclusive.equals("true") ? true : false;
            if (condition.getAttribute ("type").equals ("contextual")) {
                var._contextType = condition.getAttribute ("containedIn");
            }
            else {
                var._operator =
                VisVariant.OPNAMES.indexOf(condition.getAttribute("operator"));
                var._property = condition.getAttribute("property");
                try {
                    var._value =
                    PropertyValueInspector.getPropertyValue(
                            condition.getAttribute("value"));
                } catch (ParseException e) {
                }
            }
            var._style =
                importVisualization(
                    (Element) variantNode.getElementsByTagName(
                        "visualization").item(
                        0));
            vis.addVariant(var);
        }
        vis.importTypeSpecificVisualization(visElement);
        vis.importDefaults (vis, visElement);
        vis.firePropertyChange(PROP_VIS_CHANGED, vis);
        return vis;
    }
    
  

    private static ByteArrayInputStream importImageData(Element imageNode) {
        CDATASection imageData =
            (CDATASection) imageNode.getChildNodes().item(0);
        // replaced deprecated class
        //            StringBufferInputStream is =
        //                new StringBufferInputStream(imageData.getData());
        //            StringReader sw = new StringReader(imageData.getData());
        StringTokenizer tokens = new StringTokenizer(imageData.getData(), ",");
        int length = Integer.parseInt(tokens.nextToken());
        byte[] bytes = new byte[length];
        int index = 0;
        while (tokens.hasMoreTokens()) {
            bytes[index++] = (byte) Integer.parseInt(tokens.nextToken());
        }

        ByteArrayInputStream bis = new ByteArrayInputStream(bytes);
        return bis;
    }
    public void importTypeSpecificVisualization(Element visElement) {
    }

    public void firePropertyChange(String prop, Object newValue) {
        resetPreview(null);
        super.firePropertyChange(prop, newValue);
    }

    public void dispose() {
    	String key = generateKey ();
    	ManagedImage mip = ManagedImage.get (key);
    	if (mip != null) {
    		mip.dispose(this);
    		mip = ManagedImage.get (key+"nl");
    		if (mip != null) mip.dispose (this);
    		mip = ManagedImage.get (key+"sm");
    		if (mip != null) mip.dispose (this);
    	}
    	_preview = null;
    	_previewNoLabel = null;
    	_smallImage = null;
//        if (_preview != null)
//            _preview.dispose();
//        if (_previewNoLabel != null) {
//        	_previewNoLabel.dispose();
//        }
//        if (_smallImage != null) {
//        	_smallImage.dispose ();
//        }
        for (int i = 0; i < _decorations.length; i++) {
			for (int j = 0; j < _decorations [i].length; j++) {
				if (_decorations [i][j] != null) {
					_decorations [i][j].removeListener(this);
				}
			}
		}
    }
	/**
	 * @return Whether the palette should include the type
	 */
	public boolean isHiddenFromPalette() {
		return _hideFromPalette;
	}
	/**
	 * Sets whether to hide this type from the palette in the
	 * diagram editor
	 * @param hide true if the type is to be hidden
	 */
	public void setHiddenFromPalette(boolean hide) {
		_hideFromPalette = hide;
		firePropertyChange(PROP_VIS_CHANGED, this);		
	}
	/**
	 * @return Returns the resizeToFitLabel.
	 */
	public boolean shouldResizeToFitLabel() {
		return _resizeToFitLabel && getLabelSource() != LabelOptions.NOLABEL && !(getLabelHorizontalAlignment() == HorizontalTextAlignmentOptions.LEFT_OUTSIDE || getLabelHorizontalAlignment() == HorizontalTextAlignmentOptions.RIGHT_OUTSIDE)
			&& !(getLabelVerticalAlignment() == VerticalTextAlignmentOptions.ABOVE || getLabelVerticalAlignment() == VerticalTextAlignmentOptions.BELOW);
	}
	/**
	 * @param resizeToFitLabel The resizeToFitLabel to set.
	 */
	public void setResizeToFitLabel(boolean resizeToFitLabel) {
		_resizeToFitLabel = resizeToFitLabel;
	}

	/* (non-Javadoc)
	 * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		if (evt.getPropertyName().equals (ElementVisualization.PROP_VIS_CHANGED)) {
			if (_superVisualizations.contains(evt.getSource ())) {
				blendVisualization((ElementVisualization )evt.getSource (), true);
				firePropertyChange(PROP_VIS_CHANGED, this);
			}
		}
	}
	/**
	 * @return Returns the isDefault.
	 */
	public boolean isDefault() {
		return _isDefault;
	}
	
	public void addSuperVisualization (ElementVisualization vis) {
		_superVisualizations.add(vis);
		vis.addPropertyChangeListener("sub", this);
	}
	
	public void removeSuperVisualization (ElementVisualization vis) {
		_superVisualizations.remove(vis);
		vis.removePropertyChangeListener("sub", this);
	}
	
	protected String generateKey () {
		StringBuffer key = new StringBuffer ("vis");
		key.append(_cornerSize);
		key.append(_direction);
		key.append(_fillColor.toString());
		key.append(_font.toString());
		key.append(_labelHorizontalAlignment);
		key.append(_labelMargin);
		key.append(_labelSource);
		key.append(_labelVerticalAlignment);
		key.append(_outlineColor.toString ());
		key.append(_outlineStyle);
		key.append(_outlineWidth);
		key.append(_size.toString ());
		key.append(_shape);
		key.append(_iconFilename);
		return key.toString ();
	}
	
	
	/**
	 * @param image
	 * @see AcmeStudio.util.IManagedImageListener#imageDisposed(AcmeStudio.util.ManagedImage)
	 **/
	public void imageDisposed(ManagedImage image) {
		if (image == _preview) _preview = null;
		if (image == _previewNoLabel) _previewNoLabel = null;
		if (image == _smallImage) _smallImage = null;
		for (int i = 0; i < 3; i++)
			for (int j = 0; j < 3; j++) 
				if (image == _decorations[i][j]) _decorations [i][j] = null;
	}
}
