/*
 * Created on May 29, 2003
 *
 */
package AcmeStudio.ui.editors.diagram.styles;

import org.eclipse.draw2d.Graphics;
import org.eclipse.draw2d.SWTGraphics;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;
import org.eclipse.ui.forms.widgets.FormToolkit;
import AcmeStudio.ui.editors.style.parts.DecorationPrefDialog;
import AcmeStudio.util.IManagedImageListener;
import AcmeStudio.util.ManagedImage;

/**
 * @author Bradley Schmerl
 * Implements a grid of buttons that allow style region images to be selected
 */
public class ImageSelectionGroup implements IManagedImageListener {
    /**
     * The button that will raise a file browser to locate the image
     */
    protected Button[][] _chooseImageButton = new Button[3][3];

    /**
     * The images corresponding to the buttons
     */
    protected ManagedImage[][] _imageButtons = new ManagedImage[3][3];

    /**
     * The style to be modified
     */
    protected ElementVisualization _style = null;

    /**
     * Whether button selection clears the images
     */
    protected boolean _clearMode = false;

    /**
     * What to do when the image changes
     */
    protected ImagePreviewer _previewer;

	private final FormToolkit	_toolkit;
    /**
     * Sets the clear mode
     * @param mode Whether the clearMode should be set
     */
    public void setClearMode(boolean mode) {
        _clearMode = mode;
    }

    /**
     * Constructs a grid of buttons
     * @param parent Where the buttons should go
     * @param style The style being modified
     * @param previewer How to report updates
     */
    public ImageSelectionGroup(
        Composite parent,
        ElementVisualization style,
        ImagePreviewer previewer) {
        super();
        _style = style;
        _previewer = previewer;
		_toolkit = null;;
        createControl(parent);
    }

    /**
	 * @param decorationGroup
	 * @param style
	 * @param dialog
	 * @param toolkit
	 */
	public ImageSelectionGroup(Composite parent, ElementVisualization style, ImagePreviewer previewer, FormToolkit toolkit) {
		super ();
		 _style = style;
		_previewer = previewer;
		_toolkit = toolkit;
		createControl (parent);
	}

	/**
	 * Creates the control for this widget
	 * 
	 * @param parent
	 *                    where the buttons should go
	 */
    public void createControl(final Composite parent) {
        Group images = new Group(parent, SWT.NONE);
        if (_toolkit != null) images.setBackground(_toolkit.getColors().getBackground());
        images.setLayout(new GridLayout(3, true));
        for (int y = 0; y < 3; y++) {
            for (int x = 0; x < 3; x++) {
                final int xRegion = x;
                final int yRegion = y;
                if (_toolkit == null) _chooseImageButton[x][y] = new Button(images, SWT.PUSH);
                else _chooseImageButton [x][y] = _toolkit.createButton(images, "", SWT.PUSH);

                int iWidth = _style.getSize().width / 3;
                int iHeight = _style.getSize().height / 3;
                iWidth = iWidth == 0 ? 1 : iWidth;
                iHeight = iHeight == 0 ? 1 : iHeight;
                Image img = new Image (Display.getDefault (), iWidth, iHeight);
                _imageButtons [x][y] = new ManagedImage (img, this);
                resetButtonImage(x, y, true);
                _chooseImageButton[x][y]
                    .addSelectionListener(new SelectionAdapter() {
                    public void widgetSelected(SelectionEvent evt) {
                        if (_clearMode) {
                            resetButtonImage(xRegion, yRegion, false);
                            _previewer.updatePreview();
                        } else {
                            int selection = -1;
                            DecorationPrefDialog dpd =
                                new DecorationPrefDialog(parent.getShell());
                            dpd.open();
                            if (IDialogConstants.CANCEL_ID
                                != dpd.getReturnCode()) {
                                selection = dpd.getSelection();
                                if (selection == DecorationPrefDialog.IMAGE) {
                                    chooseImage(xRegion, yRegion);
                                } else {
                                    chooseText(xRegion, yRegion, dpd.getText());
                                }
                            }
                        }
                    }
                });
            }
        }
    }

    /**
     * Resets the button image
     * @param x The x position of the grid
     * @param y The y position of the grid
     * @param useImage Whether to use the _imageButtons image, or clear
     */
    private void resetButtonImage(int x, int y, boolean useImage) {
        GC gc = new GC(_imageButtons[x][y].getImage ());
        Graphics g = new SWTGraphics(gc);
        g.setBackgroundColor(_style.getFillColor());
        Dimension oD =
            new Dimension(
                _imageButtons[x][y].getBounds().width,
                _imageButtons[x][y].getBounds().height);
        g.fillRectangle(0, 0, oD.width, oD.height);
        if (useImage) {
            ManagedImage decoration = _style.getDecoration(x, y);
            if (decoration != null) {
                Rectangle r = _style.getDecoration(x, y).getBounds();
                Dimension iD = new Dimension(r.width, r.height);
                if (decoration != null)
                    g.drawImage(
                        decoration.getImage (),
                        oD.width / 2 - iD.width / 2,
                        oD.height / 2 - iD.height / 2);
            }
        }
        g.dispose();
        gc.dispose();
        _chooseImageButton[x][y].setImage(_imageButtons[x][y].getImage ());
                if (!useImage)
                    _style.setDecoration(
                        (ManagedImage )null,
                        x,
                        y);
    }

    /**
     * Chooses an image to put in the grid
     * @param xRegion The x coord of the grid
     * @param yRegion The y coord of the grid
     */
    protected void chooseImage(final int xRegion, final int yRegion) {
        FileDialog dialog =
            new FileDialog(Display.getDefault().getActiveShell(), SWT.OPEN);
        dialog.setFilterExtensions(new String[] { "*.gif", "*.jpg", "*.png" });
        String filename = dialog.open();
        if (filename != null) {
            _style.setDecoration(filename, xRegion, yRegion);
            resetButtonImage(xRegion, yRegion, true);
            //            _imageLabel.setImage(_style.getImage(1,1));
            _previewer.updatePreview();
        }
    }

    /**
     * Input a label to put in the grid
     * @param xRegion The x coord of the grid
     * @param yRegion The y coord of the grid
     */
    protected void chooseText(
        final int xRegion,
        final int yRegion,
        final String text) {
        _style.setDecoration(xRegion, yRegion, text);
        resetButtonImage(xRegion, yRegion, true);
        _previewer.updatePreview();
    }

    /**
     * Imports the style's images into the grid
     */
    public void importStyle() {
        for (int y = 0; y < 3; y++) {
            for (int x = 0; x < 3; x++) {
                if (_style.getDecoration(x, y) != null) {
                    resetButtonImage(x, y, true);
                }
            }
        }
    }

    /**
     * Disposes of all the temporary images for the buttons
     */
    public void dispose() {
        for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                if (_imageButtons[i][j] != null) {
                    _imageButtons[i][j].removeListener(this);
                }
            }
        }
    }

	/**
	 * @param image
	 * @see AcmeStudio.util.IManagedImageListener#imageDisposed(AcmeStudio.util.ManagedImage)
	 **/
	public void imageDisposed(ManagedImage image) {
		for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                if (_imageButtons[i][j] == image) {
                    _imageButtons [i][j] = null;
                }
            }
        }
	}
}
