package AcmeStudio.ui.editors.diagram.styles;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.jface.action.*;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.*;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.internal.WorkbenchImages;
import org.eclipse.ui.texteditor.StatusLineContributionItem;

import AcmeStudio.model.TypeAssignmentCommand;
import AcmeStudio.ui.editors.AcmeStyleEditor;
import AcmeStudio.ui.editors.actions.SuperTypeAssignmentMenuAction;
import AcmeStudio.ui.editors.diagram.styles.ConnectorVisualization.RoleRouteRule;
import AcmeStudio.util.AutoApplyEditableComboBoxCellEditor;
import AcmeStudio.util.HoverHelp;
import AcmeStudio.util.ImageManager;
import acmeLib.AcmeObject;

/**
 * @author Yan Hong (yh@cs.cmu.edu)
 *
 * Lists all the element styles a table displaying the
 * type, visualization and description. Allows creating
 * and modifying visual styles.
 */
public class RoleRouteList extends Composite {

	/** 
     * Holder of the element type.
     */
    private class RoleRouteRuleHolder {

        private ConnectorVisualization.RoleRouteRule _rule = null;

        /**
         * Returns the element type held by this holder
         * @return the element type
         */
        public ConnectorVisualization.RoleRouteRule getRule() {
            return _rule;
        }

        /**
         * Returns the element type held by this holder
         * @return the element type
         */
        public void setRule(ConnectorVisualization.RoleRouteRule rule) {
            _rule = rule;
        }
    }

    // Constants for organizing the table
    private static final int TYPECOL = 0;
    private static final int TYPEWIDTH = 200;
    private static final int ROUTECOL = 1;
    private static final int ROUTEWIDTH = 160;
    private static final String[] PROPERTIES = { "Type", "Route" };
    private IContributionItem[] _myContributions;

    /** The table viewer used to view the properties **/
    private TableViewer _viewer = null;

    /** Indicates if the table is being refreshed **/
    private boolean _beingRefreshed = false;

    private AcmeStyleEditor _parentEditor = null;

    /** The action performed on double click **/
    private Action _doubleClickAction;

    /** Action to create a new type in this list **/
    private IAction _newTypeAction;

    private LinkedList _input = null;
    private CellEditor[] _cellEditors = new CellEditor[3];

    private Action _deleteRuleAction;

    private boolean _editable = true;

    /*
     * The content provider class is responsible for
     * providing objects to the view. It returns all the 
     * properties to be displayed in the viewer.
     */
    private class ViewContentProvider implements IStructuredContentProvider {

        public void inputChanged(Viewer v, Object oldInput, Object newInput) {
        }
        public void dispose() {
        }

        /**
         * Returns all the types of the family.
         * @param parent The AcmeFamily
         * @return The types
         */
        public Object[] getElements(Object parent) {
            LinkedList ll = (LinkedList) parent;
            if (ll == null)
                return null;
            Object[] elements = new Object[ll.size()];
            int i = 0;
            Iterator iter = ll.iterator();
            while (iter.hasNext()) {
                ConnectorVisualization.RoleRouteRule rule =
                    (ConnectorVisualization.RoleRouteRule) iter.next();
                RoleRouteRuleHolder holder = new RoleRouteRuleHolder();
                holder._rule = rule;
                elements[i++] = holder;
            }
            return elements;
        }
    }

    /** 
     * Provides routines for getting the text in each of the columns of the table
     */
    private class ViewLabelProvider
        extends LabelProvider
        implements ITableLabelProvider {
        /**
         * Returns the text associated with a particular column in the table
         * @param obj the type to get the information from
         * @param index the column to return the string for
         * @return the text to be placed in the column
         */
        public String getColumnText(Object obj, int index) {
            RoleRouteRuleHolder ruleHolder = (RoleRouteRuleHolder) obj;
            String retStr = null;
            switch (index) {
                case TYPECOL :
                    retStr = ruleHolder.getRule()._roleType;
                    if (retStr == null) {
                        retStr = "<<No Name>>";
                    }
                    break;
                case ROUTECOL :
                    int route = ruleHolder.getRule()._roleRoute;
                    retStr =
                        (
                            String) ConnectorVisualization
                                .RoleRouteOptions
                                .ROUTENAMES
                                .get(
                            route);
                    break;
            }
            return retStr;
        }

        /** 
         * As there are no images associated with the table, just
         * return null in these routines.
         */
        public Image getColumnImage(Object obj, int index) {
            RoleRouteRuleHolder ruleHolder = (RoleRouteRuleHolder) obj;
            switch (index) {
                case ROUTECOL :
                    int route = ruleHolder.getRule()._roleRoute;
                    switch (route) {
                        case ConnectorVisualization.RoleRouteOptions.STRAIGHT :
                            return ImageManager.getIcon(
                                ImageManager.ID_STRAIGHT_ROLE);
                        case ConnectorVisualization.RoleRouteOptions.HBUS :
                            return ImageManager.getIcon(
                                ImageManager.ID_HBUS_ROLE);
                        case ConnectorVisualization.RoleRouteOptions.VBUS :
                            return ImageManager.getIcon(
                                ImageManager.ID_VBUS_ROLE);
                        case ConnectorVisualization.RoleRouteOptions.ELBOW :
                            return ImageManager.getIcon(
                                ImageManager.ID_ELBOW_ROLE);
                    }
            }
            return null;

        }

        public Image getImage(Object obj) {
            return null;
        }
    }

    /**
     * Constructor for ElementTypeList. Constructs the Style table and adds
     * it to the appropriate editor part
     * @param parent the Composite in which to place the table
     * @param style the style of the table
     * @param category the category of this list
     * @param parentEditror the parent style editor
     */
    public RoleRouteList(Composite parent, int style) {
        super(parent, style);
        //_parentEditor = parentEditor;
        //_category = category;

        FillLayout tableLayout = new FillLayout();
        this.setLayout(tableLayout);
        this.setLayoutData(
            new GridData(
                GridData.FILL_HORIZONTAL | GridData.VERTICAL_ALIGN_FILL));

        // Create the table and set it's contents to the active design by default
        _viewer =
            new TableViewer(
                this,
                SWT.BORDER
                    | SWT.MULTI
                    | SWT.H_SCROLL
                    | SWT.V_SCROLL
                    | SWT.FULL_SELECTION);
        _viewer.setContentProvider(new ViewContentProvider());
        _viewer.setLabelProvider(new ViewLabelProvider());

        // Set the column header and widthe information
        Table table = _viewer.getTable();
        table.addKeyListener(new KeyListener() {
            public void keyPressed(KeyEvent evt) {
                if (evt.character == SWT.DEL) {
                    _deleteRuleAction.run();
                }
            }

            public void keyReleased(KeyEvent evt) {
            }
        });

        table.setHeaderVisible(true);
        table.setLinesVisible(true);

        TableColumn typeCol = new TableColumn(table, SWT.LEFT, TYPECOL);
        TableColumn sideCol = new TableColumn(table, SWT.LEFT, ROUTECOL);

        typeCol.setText("Type");
        sideCol.setText("Route");

        ColumnLayoutData typeLayout = new ColumnPixelData(TYPEWIDTH, true);
        ColumnLayoutData routeLayout = new ColumnPixelData(ROUTEWIDTH, true);

        // set columns in Table layout
        TableLayout layout = new TableLayout();
        layout.addColumnData(typeLayout);
        layout.addColumnData(routeLayout);
        table.setLayout(layout);

        // Add actions
        _doubleClickAction = new Action() {
            public void run() {
                handleDoubleClick();
            }
        };
        hookDoubleClickAction();

        makeActions();
        //        _myContributions =
        //            new ActionContributionItem[] {
        //                new ActionContributionItem(_newTypeAction),
        //                new ActionContributionItem(_deleteTypeAction)};

        hookContextMenu();

        // Start a helper to report errors in the properties
        HoverHelp help = new HoverHelp();
        help.activateHoverHelp(_viewer.getTable());
        
        // Assign the cell modifiers -- this is heavy weight
        updateCellEditors();
        _viewer.setCellEditors(_cellEditors);
        _viewer.setColumnProperties(PROPERTIES);
        _viewer.setCellModifier(new StyleCellModifier());
    }

    /**
     * Returns the TableView that contains the styles
     * @return TableViewer
     */
    public TableViewer getViewer() {
        return _viewer;
    }

    /** Modifies the selected style on a double click
     */
    public void handleDoubleClick() {


    }

    /**
     * Set the selected object to be the new object passed in
     * @param o the object to view
     */
    public void setInput(LinkedList ll) {
        if (ll == null)
            return;

        _input = ll;
        //_selectedObject = o;
        _viewer.setInput(ll);

//        // Assign the cell modifiers -- this is heavy weight
//        updateCellEditors();
//        _viewer.setCellEditors(_cellEditors);
//        _viewer.setColumnProperties(PROPERTIES);
//        _viewer.setCellModifier(new StyleCellModifier());
    }

    /**
     * @see org.eclipse.swt.widgets.Control#setFocus()
     */
    public boolean setFocus() {
        return _viewer.getControl().setFocus();
    }

    /**
     * Adds the double click action to the table viewer
     */
    private void hookDoubleClickAction() {
        _viewer.addDoubleClickListener(new IDoubleClickListener() {
            public void doubleClick(DoubleClickEvent event) {
                _doubleClickAction.run();
            }
        });
    }

    /**
     * Makes all the actions for the property list
     */
    private void makeActions() {

        // The delete action, deletes the currently selected properties
        _deleteRuleAction = new Action() {
            public void run() {
                deleteSelectedRules();
            }
        };
        _deleteRuleAction.setText("Delete Rule");
        _deleteRuleAction.setToolTipText("Deletes selected rules");
        // As of Eclipse 3.0, IMG_TOOL_DELETE_HOVER == IMG_TOOL_DELETE
        _deleteRuleAction.setImageDescriptor(
            WorkbenchImages.getImageDescriptor(
                ISharedImages.IMG_TOOL_DELETE));
    }

    public List deleteSelectedRules() {
        List result = new LinkedList();
        // Get the selections
        ISelection isel = _viewer.getSelection();
        if (isel instanceof IStructuredSelection) {
            IStructuredSelection iss = (IStructuredSelection) isel;

            Iterator it = iss.iterator();
            // Iterate through the selections
            while (it.hasNext()) {
                Object obj = it.next();
                if (obj instanceof RoleRouteRuleHolder) {
                    RoleRouteRuleHolder ruleHolder = (RoleRouteRuleHolder) obj;
                    RoleRouteRule rule = ruleHolder.getRule();
                    result.add(rule._roleType);
                    _input.remove(rule);
                }
            }

            setInput(_input);
        }
        return result;
    }

    /**
     * Creates the context menu for this view, and hooks it into the framework
     */
    private void hookContextMenu() {
        MenuManager menuMgr = new MenuManager("StyleEditorPopupMenu");
        menuMgr.setRemoveAllWhenShown(true);
        menuMgr.addMenuListener(new IMenuListener() {
            public void menuAboutToShow(IMenuManager manager) {
                fillContextMenu(manager);
            }
        });

        Menu menu = menuMgr.createContextMenu(_viewer.getControl());
        _viewer.getControl().setMenu(menu);
        if (_parentEditor != null) {
            _parentEditor.getSite().registerContextMenu(menuMgr, _viewer);
        }
    }

    /**
     * Adds the contents of the menu
     */
    private void fillContextMenu(IMenuManager manager) {
        if (_editable && _parentEditor != null) {
            SuperTypeAssignmentMenuAction assFamAction =
                (SuperTypeAssignmentMenuAction) _parentEditor.getAction(
                    TypeAssignmentCommand.ASSIGN_TYPE);
            SuperTypeAssignmentMenuAction unassFamAction =
                (SuperTypeAssignmentMenuAction) _parentEditor.getAction(
                    TypeAssignmentCommand.UNASSIGN_TYPE);
            manager.add(assFamAction.buildMenu());
            manager.add(unassFamAction.buildMenu());
            manager.add(new Separator());
            manager.add(_deleteRuleAction);
            manager.add(_newTypeAction);
            // Other plug-ins can contribute there actions here
            manager.add(new Separator("Additions"));
        } else {
            StatusLineContributionItem item =
                new StatusLineContributionItem("$XXX$");
            item.setText(
                "You must open the family file directly to edit this family");
            manager.add(item);
            manager.add(new Separator("Additions"));
        }
    }

    /**
     * Should be fired when this style bar is selected
     * @see org.eclipse.ui.ISelectionListener#selectionChanged(IWorkbenchPart, ISelection)
     */
    public void selectionChanged(IWorkbenchPart part, ISelection selection) {
    }

    public IContributionItem[] getContributions() {
        return _myContributions;
    }

    /**
     * Sent when the location (x, y) of a control changes relative
     * to its parent (or relative to the display, for <code>Shell</code>s).
     * @see org.eclipse.swt.events.ControlListener
     *
     * @param e an event containing information about the move
     */
    public void controlMoved(ControlEvent e) {
    }

    /**
     * Sent when the size (width, height) of a control changes.
     * @see org.eclipse.swt.events.ControlListener
     *
     * @param e an event containing information about the resize
     */
    public void controlResized(ControlEvent e) {
        Object obj = e.getSource();
        if (obj instanceof TableColumn) {
            if (!_beingRefreshed) {
                //_parentEditor.refreshStyleLists(this);
            }
        }
    }

    public Table getTable() {
        return _viewer.getTable();
    }

    public void refresh(RoleRouteList source) {
        _beingRefreshed = true;
        Table table = getTable();
        Table sourceTable = source.getTable();

        int size = table.getColumnCount();
        for (int i = 0; i < size; i++) {
            TableColumn tc = table.getColumn(i);
            TableColumn sourceTc = sourceTable.getColumn(i);
            tc.setWidth(sourceTc.getWidth());
        }
        this.redraw();
        _beingRefreshed = false;
    }

    public void deselectAll() {
        Table table = getTable();
        table.deselectAll();
    }

    /**
     * @see AcmeStudio.ui.views.parts.ElementViewPart#isValidFor(AcmeObject)
     */
    public boolean isValidFor(AcmeObject o) {
        return true;
    }

    private class StyleCellModifier implements ICellModifier {

        public boolean canModify(Object element, String property) {
            if (property.equals("Type")) {
                return false;
            }
//            if (element instanceof RoleRouteRuleHolder) {
//                updateCellEditors();
//            }

            return true;
        }

        public Object getValue(Object element, String property) {
            Object value = null;
            if ((element instanceof RoleRouteRuleHolder)
                && canModify(element, property)) {
                RoleRouteRuleHolder ruleHolder = (RoleRouteRuleHolder) element;
                int route = ruleHolder.getRule()._roleRoute;
                //_currentSide = side;
//                updateCellEditors();
                if (property.equals("Type")) {
                    value = ruleHolder.getRule()._roleType;
                } else if (property.equals("Route")) {
                    value = new Integer(route);
                }
            }
            return value;
        }

        public void modify(Object element, String property, Object value) {
            if (element instanceof TableItem) {
//                updateCellEditors();
                TableItem item = (TableItem) element;
                RoleRouteRuleHolder ruleHolder =
                    (RoleRouteRuleHolder) item.getData();
                if (property.equals("Route")) {
                    ruleHolder.getRule()._roleRoute =
                        ((Integer) value).intValue();
//                 if (ruleHolder.getRule()._roleRoute == -1) {
//                 	String text = ((CCombo )_cellEditors [ROUTECOL].getControl()).getText();
//                 	Integer route = (Integer )ConnectorVisualization.RoleRouteOptions.ROUTES.get (text.toLowerCase());
//					if (route != null) {
//						ruleHolder.getRule ()._roleRoute = route.intValue();
//                 	}
//                 }
                }
                setInput(_input);
            }
        }

    }

    protected void updateCellEditors() {
    	
    	
    	
        AutoApplyEditableComboBoxCellEditor routeEditor = new AutoApplyEditableComboBoxCellEditor(
                _viewer.getTable(),
                ConnectorVisualization.RoleRouteOptions.LABELS);
        
            	
            
		_cellEditors[ROUTECOL] =
            routeEditor;
        
        _cellEditors [ROUTECOL].getControl ().addFocusListener (new FocusAdapter () {
        	public void focusLost (FocusEvent e) {
        		((AutoApplyEditableComboBoxCellEditor )_cellEditors [ROUTECOL]).fireApplyEditorValue();
        	}
        });
    }



    /**
     * Sent when default selection occurs in the control.
     * <p>
     * For example, on some platforms default selection occurs
     * in a List when the user double-clicks an item or types
     * return in a Text.
     * </p>
     *
     * @param e an event containing information about the default selection
     */
    public void widgetDefaultSelected(SelectionEvent e) {
    }

    /**
     * Notifies that the selection has changed.
     *
     * @param event event object describing the change
     */
    public void selectionChanged(SelectionChangedEvent event) {
//        Debug.diagnoseln(
//            "ElementTypeList.selectionChanged: " + event.toString());
    }

    /**
     * Sent when a control gets focus.
     *
     * @param e an event containing information about the focus change
     */
    public void focusGained(FocusEvent e) {
    }

    /**
     * Sent when a control loses focus.
     *
     * @param e an event containing information about the focus change
     */
    public void focusLost(FocusEvent e) {
    }

    public void refresh() {
        if (!isDisposed() && !_viewer.getControl().isDisposed())
            this._viewer.refresh();
    }

    public void setEditable(boolean canEdit) {
        _editable = canEdit;
    }

}