/*
 * Created on Feb 23, 2003
 *  
 */
package AcmeStudio.ui.editors.diagram.styles;

import java.util.*;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import AcmeStudio.model.ModelHelper;
import AcmeStudio.ui.editors.diagram.styles.constants.SideOptions;
import AcmeStudio.ui.editors.diagram.visualInfo.ConnectionPatternInformation;
import AcmeStudio.util.Debug;
import AcmeStudio.util.Utils;
import AcmeStudio.wizards.connection.PortConnectionRule;
import AcmeStudio.wizards.connection.PortConnectionRuleKey;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.elements.*;

/**
 * @author schmerl
 */
public class SystemVisualization extends Visualization
{
    /**
     * Enumeration of layouts.
     */
    public static class Layout
    {
        public static final int LAYERED = 0, GRID = 1, NONE = 2;
        public static final String[] LABELS = new String[] { "Layered", "Grid" };
        public static final List LAYOUTNAMES = new ArrayList();
        static
        {
            LAYOUTNAMES.add(LAYERED, "Layered");
            LAYOUTNAMES.add(GRID, "Grid");
            LAYOUTNAMES.add(NONE, "[No Layout]");
        }
        public static final Hashtable LAYOUTTABLE = new Hashtable();
        static
        {
            LAYOUTTABLE.put("Layered", new Integer(LAYERED));
            LAYOUTTABLE.put("Grid", new Integer(GRID));
            LAYOUTTABLE.put("[No Layout]", new Integer(NONE));
        }
    }

    /**
     * The default layout of the system.
     */
    private static final int DEFAULT_LAYOUT = Layout.GRID;
    protected int _layout = DEFAULT_LAYOUT;
    protected int _rows = 1;
    protected int _cols = 1;
    protected List _placementRules = new LinkedList();
    protected List _portConnectionRules = new ArrayList();
    private boolean allowNewRule;

    public SystemVisualization()
    {
    }

    public int getLayout()
    {
        return _layout;
    }

    public void setLayout(int layout)
    {
        this._layout = layout;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    public int getRows()
    {
        return _rows;
    }

    public void setRows(int rows)
    {
        this._rows = rows;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    public int getCols()
    {
        return _cols;
    }

    public void setCols(int cols)
    {
        this._cols = cols;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    public List getPlacementRules()
    {
        return _placementRules;
    }

    public void setPlacementRules(List placementRules)
    {
        this._placementRules = placementRules;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    /**
     * @return
     */
    public List getPortConnectionRules()
    {
        return _portConnectionRules;
    }

    /**
     * @param connectionRules
     */
    private void setPortConnectionRules(List map)
    {
        _portConnectionRules = map;
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    private PlacementRule getPlacementRule(AcmeElementType e)
    {
        Iterator iter = _placementRules.iterator();
        while (iter.hasNext())
        {
            PlacementRule rule = (PlacementRule) iter.next();
            if (rule._type.equals(e.getTypeName()))
            {
                return rule;
            }
        }
        return null;
    }

    /**
     * This method is similar to getPortConnectionRule, however, it returns back all
     * possibly matching rules so the user may choose between them
     * 
     * @author Larry Waldman Created on 5/24/04
     * 
     * @param PortConnectionRuleKey
     *                 the key to return matching rules against
     * @return PortConnectionRule[] an array containing all matching rules
     */
    public List getAllPortConnectionRules(PortConnectionRuleKey key)
    {
        return getMatchingRulesInList(key, _portConnectionRules);
    }

    /**
     * This function takes in the key and a list of rules, and returns back a List
     * containing all items in the list that match the key
     * 
     * @param key
     * @param rule
     */
    public static List getMatchingRulesInList(PortConnectionRuleKey key,
            List rules)
    {
        List ret = new ArrayList();
        Iterator iter = rules.iterator();
        while (iter.hasNext())
        {
            PortConnectionRule element = (PortConnectionRule) iter.next();
            if (key == null || key.matches(element.generateKey()))
            {
                ret.add(element);
            }
        }
        return ret;
    }

    /**
     * This function takes in the key and a list of rules, and returns back a List
     * containing all items in the list that match the key
     * 
     * @param key
     * @param rule
     */
    public static List getEqualsRulesInList(PortConnectionRuleKey key,
            List rules)
    {
        List ret = new ArrayList();
        Iterator iter = rules.iterator();
        while (iter.hasNext())
        {
            PortConnectionRule element = (PortConnectionRule) iter.next();
            if (key == null || key.equals(element.generateKey()))
            {
                ret.add(element);
            }
        }
        return ret;
    }
    
    public PlacementRule getPlacementRule(AcmeElement e)
    {
        if (e instanceof AcmeElementType)
            return getPlacementRule((AcmeElementType) e);
        AcmeEnumeration ae = e.getDeclaredTypes();
        PlacementRule rule = null;
        while (ae.moreItems())
        {
            AcmeElementType type = (AcmeElementType) ae.getNextItem();
            if ((rule = getPlacementRule(type)) != null)
                break;
        }
        return rule;
    }

    public void copyInto(SystemVisualization from)
    {
        setLayout(from.getLayout());
        setRows(from.getRows());
        setCols(from.getCols());
        copyPlacementRules(from.getPlacementRules());
        copyPortConnectionRules(from.getPortConnectionRules());
    }

    private void copyPlacementRules(List placementRules)
    {
        _placementRules.clear();
        _placementRules.addAll(placementRules);
    }

    private void copyPortConnectionRules(List map)
    {
        if (_portConnectionRules.size() != 0)
        {
            _portConnectionRules.clear();
        }
        if (map != null)
            _portConnectionRules.addAll(map);

    }

    public SystemVisualization copy()
    {
        SystemVisualization sv = new SystemVisualization();
        sv.copyInto(this);
        return sv;
    }

    public void importVisualization(Element layoutNode)
    {
        setLayout(importLayout(layoutNode));
        setRows(importRows(layoutNode));
        setCols(importCols(layoutNode));
        setPlacementRules(importPlacementRules(layoutNode));
        setPortConnectionRules(importPortConnectionRules(layoutNode));
    }

    private int importLayout(Element el)
    {
        if (el == null)
        { // return right away
            return Layout.NONE;
        }
        String name = el.getAttribute("name");

        //LFW: also set whether the user can define new rules
        Debug.diagnoseln("SystemVisualization.importLayout(Element):  allowNewRules == "
            + el.getAttribute("allowNewRules"));
        if (el.getAttribute("allowNewRules").equals("true"))
            setAllowNewRule(true);
        else
            setAllowNewRule(false);

        if (name.equals(""))
            return Layout.NONE;
        int layout = ((Integer) Layout.LAYOUTTABLE.get(name)).intValue();
        return layout;
    }

    private static int importRows(Element el)
    {
        if (el == null)
        { // return right away
            return 0;
        }
        String value = el.getAttribute("rows");
        if (value.equals(""))
            return 0;
        return Integer.parseInt(value);
    }

    private static int importCols(Element el)
    {
        if (el == null)
        { // return right away
            return 0;
        }
        String value = el.getAttribute("cols");
        if (value.equals(""))
            return 0;
        return Integer.parseInt(value);
    }

    private static List importPlacementRules(Element el)
    {
        LinkedList result = null;
        if (el != null)
        {
            NodeList elements = el.getElementsByTagName("placementRule");
            result = new LinkedList();
            for (int i = 0; i < elements.getLength(); i++)
            {
                Element e = (Element) elements.item(i);
                PlacementRule rule = new PlacementRule();
                rule.setType(e.getAttribute("type"));
                rule.setRegion(Integer.parseInt(e.getAttribute("region")));
                rule.setSide(((Integer) SideOptions.SIDES.get(e
                        .getAttribute("side"))).intValue());
                rule.setAlignment(((Integer) AlignmentOptions.ALIGNMENTS.get(e
                        .getAttribute("alignment"))).intValue());
                result.add(rule);
            }
        }
        return result;
    }

    /*
     * XML Format change LFW: modified 6/3/04
     * 
     * The format for all future XML .style documents (to be backward compatible) will be
     * such that every association will have a componentType attribute. While this
     * attribute will not need a definition for each association, it is possible to use it
     * to define a rule with a component.
     * 
     * ComponentTypes and PortTypes will not be mutually exclusive. At least one must have
     * a value, as can both, if the user wishes the rule to be more specific.
     */
    private static List importPortConnectionRules(Element el)
    {
        List result = null;
        if (el != null)
        {
            NodeList elements = el.getElementsByTagName("portConnectionRule");
            if (elements.getLength() > 0)
            {
                result = new ArrayList();
                for (int i = 0; i < elements.getLength(); i++)
                {
                    Element e = (Element) elements.item(i);
                    PortConnectionRule rule = PortConnectionRule
                            .importFromXMLElement(e);
                    boolean exists = result.contains(rule);
                    if (exists)
                    {
                        PortConnectionRule ex = (PortConnectionRule) result
                                .get(result.indexOf(rule));
                        ex.setDirectionless();
                    }
                    else
                    {
                        result.add(rule);
                    }
                }
            }
        }
        return result;
    }

    private static void exportPlacementRules(Document doc, Element el,
            List rules)
    {
        if (rules != null)
        {
            Iterator iter = rules.iterator();
            while (iter.hasNext())
            {
                PlacementRule rule = (PlacementRule) iter.next();
                Element e = doc.createElement("placementRule");
                e.setAttribute("type", rule.getType());
                e.setAttribute("region", Integer.toString(rule.getRegion()));
                e.setAttribute("side", (String) SideOptions.SIDENAMES.get(rule
                        .getSide()));
                e.setAttribute("alignment",
                        (String) AlignmentOptions.ALIGNMENTNAMES.get(rule
                                .getAlignment()));
                el.appendChild(e);
            }
        }
    }

    private static void safeSetAttribute(Element e, String name, String value)
    {
        if (value != null)
            e.setAttribute(name, value);
        else
            e.setAttribute(name, "");
    }

    private static void exportPortConnectionRules(Document doc, Element el,
            List rules)
    {
        if (rules != null)
        {
            Iterator iter = rules.iterator();
            while (iter.hasNext())
            {
                PortConnectionRule rule = (PortConnectionRule) iter.next();
                Element e = rule.exportAsXMLElement(doc);
                el.appendChild(e);
            }
        }
    }

    private void exportLayout(int layout, Element el)
    {
        el.setAttribute("name", (String) Layout.LAYOUTNAMES.get(layout));
        el.setAttribute("allowNewRules", (String) Boolean
                .toString(allowNewRule));
    }

    private static void exportRows(int rows, Element el)
    {
        el.setAttribute("rows", Integer.toString(rows));
    }

    private static void exportCols(int cols, Element el)
    {
        el.setAttribute("cols", Integer.toString(cols));
    }

    public Element export(Document doc)
    {
        Element layoutNode = doc.createElement("layout");
        exportLayout(_layout, layoutNode);
        exportRows(_rows, layoutNode);
        exportCols(_cols, layoutNode);
        exportPlacementRules(doc, layoutNode, _placementRules);
        exportPortConnectionRules(doc, layoutNode, _portConnectionRules);
        return layoutNode;
    }

    /**
     * @param pair
     * @param _rule
     */
    public void addPortConnectionRule(PortConnectionRule rule)
    {
        getPortConnectionRules().add(rule);
        firePropertyChange(PROP_VIS_CHANGED, this);
    }

    public void removePortConnectionRule(PortConnectionRule rule)
    {
        getPortConnectionRules().remove(rule);
        firePropertyChange(PROP_VIS_CHANGED, this);
    }
    
    /**
     * @return Returns the allowNewRule.
     */
    public boolean getAllowNewRule()
    {
        return allowNewRule;
    }

    /**
     * @param allowNewRule
     *                 The allowNewRule to set.
     */
    public void setAllowNewRule(boolean allowNewRule)
    {
        this.allowNewRule = allowNewRule;
    }

    /**
     * Method to fetch all relevant portConnectionRules that match the parameter
     * information passed in
     * 
     * @author Larry Waldman created on 5/24/04
     * @param AcmePortType[]
     *                 portTypes This parameter are the ports which should comprise the rule
     *                 being searched for
     * @param AcmeComponentType[]
     *                 compTypes This parameter contains the components which should be in the
     *                 rules being searched for
     * @param AcmeConnectorType
     *                 The connector type that should match in the rules being searched for
     * @param system
     * @return PortConnectionRule[] an array which contains the needed data
     */
    public static List checkExistingRules(AcmePortType[] portTypes,
            AcmeComponentType[] compTypes, AcmeConnectorType connType,
            AcmeSystem system)
    {
        SystemVisualization vis;
        AcmeElement el = null;

        if (portTypes == null && compTypes == null)
            return null;

        //LFW: Assume that the ports and components passed in must be of the
        // same family. As such, the call to getMostCommonFamily need not be
        // changed to include the components as paramters

        //check to see if we are in a connector connection
        boolean connectorConn = ((Utils.nullToInt(portTypes[0])
                + Utils.nullToInt(compTypes[0]) == 0) || (Utils
                .nullToInt(portTypes[1])
                + Utils.nullToInt(compTypes[1]) == 0));
        AcmeFamily mostCommonFamily = null;
        if (!connectorConn)
        {
        	
        	
        	
             mostCommonFamily = (AcmeFamily )ModelHelper
                    .getMostCommonFamily(
                            portTypes[0] != null ? (AcmeType) portTypes[0]
                                    : (AcmeType) compTypes[0],
                            portTypes[1] != null ? (AcmeType) portTypes[1]
                                    : (AcmeType) compTypes[1]);
			vis = VisualStyleRegistry.getSysVis(mostCommonFamily);
            
            if (vis == null)
            {
                //it was probably because this rule is completely made up of 
                //generic objects, so return back the generic system Vis
                vis = VisualStyleRegistry.DEFAULT_SYSTEM_VISUALIZATION;
            }
        }
        else
        {
            if (Utils.nullToInt(portTypes[0]) == 1)
                el = portTypes[0];
            else if (Utils.nullToInt(portTypes[1]) == 1)
                el = portTypes[1];
            else if (Utils.nullToInt(compTypes[0]) == 1)
                el = compTypes[0];
            else if (Utils.nullToInt(compTypes[1]) == 1)
                el = compTypes[1];

            if (el != null && !el.getName().equals("component") && !el.getName().equals("port"))
                vis = VisualStyleRegistry.getSysVis(el);
            else
                vis = VisualStyleRegistry.DEFAULT_SYSTEM_VISUALIZATION;
        }

        Set ret = new HashSet ();
        PortConnectionRuleKey portConnectionRuleKey = new PortConnectionRuleKey(
                portTypes, compTypes, connType);
        if (vis != null)
        {
        	
//        	recurseGetConnectionRules (ret, portConnectionRuleKey, vis, mostCommonFamily);
        	
        	ret.addAll (vis.getAllPortConnectionRules(portConnectionRuleKey)); //Get the rules based on the constructed key
        }

        // Get the system-specific patterns as well, and add those to the ones already obtained
        ConnectionPatternInformation info = (ConnectionPatternInformation) system
                .getUserData(ConnectionPatternInformation.CONNECTION_PATTERNS_KEY);
        if (info != null)
        {
            if (ret == null)
            {
            	ret.addAll (
            			getMatchingRulesInList(portConnectionRuleKey, info
            					.getConnectionPatterns()));
            }
            else
            {
                ret.addAll(getMatchingRulesInList(portConnectionRuleKey, info
                        .getConnectionPatterns()));
            }
        }
        return Collections.list (Collections.enumeration(ret));
    }

	/**
	 * @param ret
	 * @param portConnectionRuleKey
	 * @param vis
	 * @param mostCommonFamily
	 */
	private static void recurseGetConnectionRules(Set ret, PortConnectionRuleKey portConnectionRuleKey, SystemVisualization vis, AcmeFamily family) {
		ret.addAll (vis.getAllPortConnectionRules (portConnectionRuleKey));
		AcmeEnumeration supers = family.getSuperTypes().enumerate();
		while (supers.moreItems ()) {
			AcmeFamily fam = (AcmeFamily )supers.getNextItem ();
			SystemVisualization sVis = VisualStyleRegistry.getSysVis(fam);
			recurseGetConnectionRules(ret, portConnectionRuleKey, sVis, fam);
		}
	}

}