package AcmeStudio.ui.editors.diagram.styles;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.*;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import AcmeStudio.util.Debug;
import acmeLib.elements.*;

/**
 * @author schmerl
 *
 */
public class VisualStyle implements PropertyChangeListener {

    private Map _styles = new Hashtable();
    private Map _sysVisMap = new Hashtable();
    private boolean _isDirty = false;
    private List _listeners = new ArrayList();

    /**
     * Constructor for VisualStyle.
     */
    public VisualStyle() {
        super();
    }

    /* (non-Javadoc)
     * @see java.beans.PropertyChangeListener#propertyChange(java.beans.PropertyChangeEvent)
     */
    public void propertyChange(PropertyChangeEvent event) {
        if (event.getSource() instanceof ElementVisualization) {
            fireChangeVisualStyle((ElementVisualization) event.getSource());
        }
        else if (event.getSource () instanceof SystemVisualization) {
        	fireChangeVisualStyle ((SystemVisualization )event.getSource ());
        }
    }

    public ElementVisualization getVisualization(String id) {
        return (ElementVisualization) _styles.get(id);
    }

    public ElementVisualization getVisualization(AcmeElementType type) {
    	ElementVisualization vis = (ElementVisualization )_styles.get (type.getTypeName ());
    	vis = ensureVisIsRightClass (type, vis);
        return vis;
    }

    /**
	 * @param type
	 * @param vis
	 * @return
	 */
	public static ElementVisualization ensureVisIsRightClass(AcmeElementType type, ElementVisualization vis) {
		if (vis instanceof GenericVisualization) {
    		if (type instanceof AcmeComponentType) {
    			ComponentVisualization cVis = new ComponentVisualization ();
    			cVis.copyInto (vis);
    			vis = cVis;
    		}
    		else if (type instanceof AcmeConnectorType) {
    			ConnectorVisualization cVis = new ConnectorVisualization ();
    			cVis.copyInto (vis);
    			vis = cVis;
    		}
    		else if (type instanceof AcmeRoleType) {
    			RoleVisualization rVis = new RoleVisualization ();
    			rVis.copyInto (vis);
    			vis = rVis;
    		}
    		else if (type instanceof AcmePortType) {
    			PortVisualization pVis = new PortVisualization ();
    			pVis.copyInto (vis);
    			vis = pVis;
    		}
    	}
		return vis;
	}

	public SystemVisualization getSysVis(String id) {
        return (SystemVisualization) _sysVisMap.get(id);
    }

    public SystemVisualization getSysVis(AcmeFamily type) {
        return (SystemVisualization) _sysVisMap.get(type.getTypeName());
    }

    public void add(String id, ElementVisualization vis) {
        _isDirty = true;
        _styles.put(id, vis);
        fireChangeVisualStyle (vis);
        vis.addPropertyChangeListener(id, this);
    }
    
    public void addSysVis(String id, SystemVisualization sv) {
        _isDirty = true;
        _sysVisMap.put(id, sv);
        fireChangeVisualStyle (sv);
        sv.addPropertyChangeListener(id, this);
    }

    public void add(AcmeElementType type, ElementVisualization vis) {
        _isDirty = true;
        ElementVisualization oldVis = VisualStyleRegistry.getVisualization(type);
        _styles.put(type.getTypeName(), vis);
       fireChangeVisualStyle ( vis);
        oldVis.resetVisualization();
        vis.addPropertyChangeListener(type.getTypeName(), this);
    }

    public void addSysVis(AcmeFamily family, SystemVisualization vis) {
        _isDirty = true;
        _sysVisMap.put(family.getTypeName(), vis);
        fireChangeVisualStyle (vis);
        vis.addPropertyChangeListener(family.getTypeName(), this);
    }

    public Element export(Document doc) {
        Element styleNode = doc.createElement("visstyle");

        Iterator i = _sysVisMap.keySet().iterator();
        while (i.hasNext()) {
            String name = (String) i.next();
            SystemVisualization vis = (SystemVisualization) _sysVisMap.get(name);
            Element visNode = vis.export(doc);
            visNode.setAttribute("association", name);
            styleNode.appendChild(visNode);
        }
        
        i = _styles.keySet().iterator();
        while (i.hasNext()) {
            String name = (String) i.next();
            ElementVisualization vis = (ElementVisualization) _styles.get(name);
            Element visNode = vis.export(doc);
            visNode.setAttribute("association", name);
            if (vis instanceof PortVisualization) {
                visNode.setAttribute("type", "PORT");
            } else if (vis instanceof RoleVisualization) {
                visNode.setAttribute("type", "ROLE");
            } else if (vis instanceof ComponentVisualization) {
                visNode.setAttribute("type", "COMPONENT");
            } else if (vis instanceof ConnectorVisualization) {
                visNode.setAttribute("type", "CONNECTOR");
            }
            else {
            	Debug.errorln ("VisualStyle.export: visualization does not have a type");
            }
            styleNode.appendChild(visNode);
        }

        return styleNode;

    }

    /**
     * Takes a root DOM element and extracts the visualization attributes out
     * of it, saving the visualizations.  Clear the dirtiness state after import.
     * @param style  the element containing the visual style to extract
     * @return VisualStyle  the final visual style object
     */
    public static VisualStyle importVisualStyle(Element style) {
        VisualStyle vs = new VisualStyle();
        
        NodeList layoutNodes = style.getElementsByTagName("layout");
        for (int i = 0; i < layoutNodes.getLength(); i++) {
            Element layoutNode = (Element) layoutNodes.item(i);
            SystemVisualization sv = new SystemVisualization();
            sv.importVisualization(layoutNode);
            vs.addSysVis(layoutNode.getAttribute("association"), sv);
        }

        NodeList viss = style.getElementsByTagName("visualization");
        for (int i = 0; i < viss.getLength(); i++) {
            Element visNode = (Element) viss.item(i);
            ElementVisualization vis = ElementVisualization.importVisualization(visNode);
            vs.add(visNode.getAttribute("association"), vis);
        }

        vs.setDirty(false);
        return vs;
    }

    /**
     * Return whether style itself has changes, and scans all visualizations for 
     * changes.
     * @return boolean  whether change has been made
     */
    public boolean isDirty() {
        boolean dirty = _isDirty;

        if (!dirty) { // see if any visualization elem in style is dirty
            Iterator i = _styles.keySet().iterator();
            while (i.hasNext()) {
                String name = (String) i.next();
                ElementVisualization vis = (ElementVisualization) _styles.get(name);
                if (vis.isDirty()) {
                    dirty = true;
                    break;
                }
            }
        }
        return dirty;
    }

    /**
     * Sets the dirtiness of this style and all its visualizations
     * @param dirty  dirtiness state to set to
     */
    public void setDirty(boolean dirty) {
        Iterator i = _styles.keySet().iterator();
        while (i.hasNext()) {
            String name = (String) i.next();
            ElementVisualization vis = (ElementVisualization) _styles.get(name);
            vis.setDirty(dirty);
        }
        _isDirty = dirty;
    }

    public void addVisualStyleListener (IVisualStyleListener listener) {
        _listeners.add(listener);
    }

    public void removeVisualStyleListener (IVisualStyleListener listener) {
        _listeners.remove(listener);
    }

    protected void fireChangeVisualStyle (ElementVisualization vis) {
        for (Iterator i = _listeners.iterator(); i.hasNext();) {
            IVisualStyleListener listener = (IVisualStyleListener) i.next();
            listener.changeVisualStyle(this);
        }
        VisualStyleRegistry._cache.clear();
    }
    
    protected void fireChangeVisualStyle (SystemVisualization vis) {
    	for (Iterator iter = _listeners.iterator(); iter.hasNext(); ) {
			IVisualStyleListener listener = (IVisualStyleListener) iter.next();
			listener.changeVisualStyle (this);
		}
    	VisualStyleRegistry._cache.clear ();
    }
}
