/*
 * Created on Jun 9, 2003
 *
 */
package AcmeStudio.ui.editors.style.parts;

import java.util.Iterator;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.viewers.ColumnLayoutData;
import org.eclipse.jface.viewers.ColumnPixelData;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TableLayout;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.internal.WorkbenchImages;

import AcmeStudio.model.ModelCompoundCommand;
import AcmeStudio.model.TypeCommand;
import AcmeStudio.ui.editors.AcmeStyleEditor;
import AcmeStudio.ui.editors.actions.NewPropertyTypeAction;
import AcmeStudio.ui.views.parts.NewPropertyTypeDialog;
import AcmeStudio.util.AcmeStudioChangeManager;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.filters.PropertyTypeEnumFilter;
import acmeLib.properties.AcmePropertyType;

/**
 * @author schmerl
 */
public class PropertyStyleList extends StyleList {

    // Constants for organizing the table
    private static final int TYPECOL = 0;
    private static final int TYPEWIDTH = 120;
    private static final int DEFNCOL = 1;
    private static final int DEFNWIDTH = 120;
    private static final int DESCCOL = 2;
    private static final int DESCWIDTH = 360;
    private static final String[] PROPERTIES = { "Type", "Definition", "Description" };
    private Action _doubleClickAction;
    private IAction _newTypeAction;
    private Action _deleteTypeAction;

    /** Indicates if the table is being refreshed **/
    /*
       * The content provider class is responsible for
       * providing objects to the view. It returns all the 
       * properties to be displayed in the viewer.
       */
    private class ViewContentProvider implements IStructuredContentProvider {
        public void inputChanged(Viewer v, Object oldInput, Object newInput) {
        }
        public void dispose() {
        }
        /**
         * Returns all the types of the family.
         * @param parent The AcmeFamily
         * @return The types
         */
        public Object[] getElements(Object parent) {
            Object[] types = null;
            AcmeEnumeration pTypes =
                _acmeFamily.getTypes().castToAcmeSetVector().enumerate(new PropertyTypeEnumFilter());
            types = new Object[pTypes.getSize()];
            int index = 0;
            while (pTypes.moreItems()) {
                types[index++] = pTypes.getNextItem();
            }
            return types;
        }
    }
    /** 
     * Provides routines for getting the text in each of the columns of the table
     */
    private class ViewLabelProvider extends LabelProvider implements ITableLabelProvider {
        /**
         * Returns the text associated with a particular column in the table
         * @param obj the type to get the information from
         * @param index the column to return the string for
         * @return the text to be placed in the column
         */
        public String getColumnText(Object obj, int index) {
            AcmePropertyType type = (AcmePropertyType)obj;
            String retStr = "";
            switch (index) {
                case TYPECOL :
                    retStr = type.getTypeName();
                    break;
                case DEFNCOL :
                    {
                        retStr = type.getDefinitionAsString();
                        break;
                    }
                case DESCCOL :
                    retStr = type.getComment();

                    break;
            }
            if (retStr == null)
                retStr = "";
            return retStr;
        }
        /** 
         * As there are no images associated with the table, just
         * return null in these routines.
         */
        public Image getColumnImage(Object obj, int index) {
            return null;
        }
        public Image getImage(Object obj) {
            return null;
        }
    }

    /**
     * @param parent
     * @param style
     */
    public PropertyStyleList(Composite parent, int style, AcmeStyleEditor parentEditor) {
        super(parent, style, parentEditor);

        FillLayout tableLayout = new FillLayout();
        setLayout(tableLayout);
        setLayoutData(new GridData(GridData.FILL_HORIZONTAL | GridData.VERTICAL_ALIGN_FILL));
        _viewer = new TableViewer(this, SWT.BORDER | SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL | SWT.FULL_SELECTION);
        _viewer.setContentProvider(new ViewContentProvider());
        _viewer.setLabelProvider(new ViewLabelProvider());
        Table table = _viewer.getTable();
        table.addKeyListener(new KeyListener() {
            public void keyPressed(KeyEvent evt) {
                if (evt.character == SWT.DEL) {
                }
            }
            public void keyReleased(KeyEvent evt) {
            }
        });

        table.setHeaderVisible(true);
        table.setLinesVisible(true);
        TableColumn typeCol = new TableColumn(table, SWT.LEFT, TYPECOL);
        TableColumn defnCol = new TableColumn(table, SWT.LEFT, DEFNCOL);
        TableColumn descCol = new TableColumn(table, SWT.LEFT, DESCCOL);
        typeCol.setText(PROPERTIES[TYPECOL]);
        defnCol.setText(PROPERTIES[DEFNCOL]);
        descCol.setText(PROPERTIES[DESCCOL]);
        typeCol.addControlListener(this);
        defnCol.addControlListener(this);
        descCol.addControlListener(this);
        ColumnLayoutData typeLayout = new ColumnPixelData(TYPEWIDTH, true);
        ColumnLayoutData defnLayout = new ColumnPixelData(DEFNWIDTH, true);
        ColumnLayoutData descLayout = new ColumnPixelData(DESCWIDTH, true);
        TableLayout layout = new TableLayout();
        layout.addColumnData(typeLayout);
        layout.addColumnData(defnLayout);
        layout.addColumnData(descLayout);
        table.setLayout(layout);

        AcmeStudioChangeManager.registerListener(this);
        table.addSelectionListener(this);
        table.addFocusListener(this);

        _doubleClickAction = new Action() {
            public void run() {
                handleDoubleClick();
            }
        };
        hookDoubleClickAction();
        /* BugFix#308. NewPropetyAction should point to action in editor */
        _newTypeAction = _parentEditor.getAction(NewPropertyTypeAction.ID);
        makeActions();
        hookContextMenu();

    }
    /**
     * 
     */
    private void hookDoubleClickAction() {
        _viewer.addDoubleClickListener(new IDoubleClickListener() {
            public void doubleClick(DoubleClickEvent evt) {
                _doubleClickAction.run();
            }
        });
    }
    /**
     * 
     */
    protected void handleDoubleClick() {
        ISelection selection = _viewer.getSelection();

        AcmePropertyType pType = (AcmePropertyType) ((IStructuredSelection)selection).getFirstElement();
        NewPropertyTypeDialog dialog =
            new NewPropertyTypeDialog(Display.getCurrent().getActiveShell(), _acmeFamily, pType);
        if (dialog.open() == IDialogConstants.OK_ID) {
            ModelCompoundCommand command = new ModelCompoundCommand("Change type");
            TypeCommand c1 = new TypeCommand(TypeCommand.REMOVE_TYPE, pType, _acmeFamily, _parentEditor);
            TypeCommand c2 = new TypeCommand(TypeCommand.ADD_TYPE, dialog.getNewType(), _acmeFamily, _parentEditor);
            command.add(c1);
            command.add(c2);
            command.execute(_acmeFamily.getParentDesign());
        }
    }
    /**
     * 
     */
    private void hookContextMenu() {
        MenuManager menuMgr = new MenuManager("#PropertyTypePopupMenu");
        menuMgr.setRemoveAllWhenShown(true);
        menuMgr.addMenuListener(new IMenuListener() {
            public void menuAboutToShow(IMenuManager manager) {
                fillContextMenu(manager);
            }
        });
        Menu menu = menuMgr.createContextMenu(_viewer.getControl());
        _viewer.getControl().setMenu(menu);
        if (_parentEditor != null) {
            _parentEditor.getSite().registerContextMenu(menuMgr, _viewer);
        }
    }
    /**
     * @param manager
     */
    protected void fillContextMenu(IMenuManager manager) {
        if (_editable) {
            manager.add(_newTypeAction);
            manager.add(_deleteTypeAction);
            manager.add(new Separator("Additions"));
        }
    }
    /**
     * 
     */
    private void makeActions() {
        _deleteTypeAction = new Action() {
            public void run() {
                ISelection iSel = _viewer.getSelection();
                if (iSel instanceof IStructuredSelection) {
                    IStructuredSelection iss = (IStructuredSelection)iSel;
                    Iterator it = iss.iterator();
                    while (it.hasNext()) {
                        Object o = it.next();
                        if (o instanceof AcmePropertyType) {
                            AcmePropertyType acmeType = (AcmePropertyType)o;
                            TypeCommand command =
                                new TypeCommand(TypeCommand.REMOVE_TYPE, acmeType, _acmeFamily, _parentEditor);
                            command.execute();

                        }

                    }
                    setInput(_acmeFamily);
                }
            }

        };
        _deleteTypeAction.setText("Delete type");
        _deleteTypeAction.setToolTipText("Deletes selected types");
        // As of Eclipse 3.0, IMG_TOOL_DELETE_HOVER == IMG_TOOL_DELETE
        _deleteTypeAction.setImageDescriptor(WorkbenchImages.getImageDescriptor(ISharedImages.IMG_TOOL_DELETE));
    }

}
