/*
 * Created on Jun 29, 2004
 *
 */
package AcmeStudio.ui.editors.text;

import java.util.HashMap;
import java.util.Map;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.projection.IProjectionListener;
import org.eclipse.jface.text.source.projection.ProjectionAnnotation;
import org.eclipse.jface.text.source.projection.ProjectionAnnotationModel;
import org.eclipse.jface.text.source.projection.ProjectionViewer;
import org.eclipse.ui.texteditor.ITextEditor;
import AcmeStudio.core.resource.provider.ASResourceProvider;
import AcmeStudio.core.resource.provider.IResourceProviderListener;
import AcmeStudio.core.resource.provider.ResourceProviderEvent;
import AcmeStudio.ui.editors.AcmeTextEditor;
import acmeLib.core.io.Region;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeSystem;
import acmeLib.tools.GraphWalker;
import acmeLib.tools.WalkerElementOperation;


/**
 * @author Bradley Schmerl
 * Created Jun 29, 2004
 */
public class AcmeFoldingStructureProvider implements IProjectionListener, IResourceProviderListener {

	private static class AcmeProjectionAnnotation extends ProjectionAnnotation {
	
		private  AcmeElement	_element;
		private  boolean	_isComment;

		public AcmeProjectionAnnotation (AcmeElement element, boolean isCollapsed, boolean isComment) {
			super (isCollapsed);
			_element = element;
			_isComment = isComment;
		}
		
		/**
		 * @return Returns the element.
		 */
		public AcmeElement getElement() {
			return _element;
		}
		/**
		 * @param element The element to set.
		 */
		public void setElement(AcmeElement element) {
			_element = element;
		}
		/**
		 * @return Returns the isComment.
		 */
		public boolean isComment() {
			return _isComment;
		}
		/**
		 * @param isComment The isComment to set.
		 */
		public void setComment(boolean isComment) {
			_isComment = isComment;
		}
	}
	
	private IDocument _cachedDocument;
	private ITextEditor _editor;
	private ProjectionViewer _viewer;
	private AcmeDesign _design;
	
	private boolean _allowCollapsing = false;
	private boolean _collapseRepresentations = true;
	private ASResourceProvider	_provider;
	
	/**
	 * 
	 */
	public AcmeFoldingStructureProvider() {
		super ();
	}

	/**
	 * 
	 * @see org.eclipse.jface.text.source.projection.IProjectionListener#projectionEnabled()
	 **/
	public void projectionEnabled() {
		projectionDisabled ();
		if (_editor != null) {
			initialize ();
		}
	}

	protected void initializePreferences () {
	}
	/**
	 * 
	 */
	private void initialize() {
		if (!isInstalled ()) return;
		
		initializePreferences ();
		
		try {
			_cachedDocument = _provider.getDocument (_editor.getEditorInput());
			_allowCollapsing = true;
			
			
			ProjectionAnnotationModel model = (ProjectionAnnotationModel )_editor.getAdapter (ProjectionAnnotationModel.class);
			if (model != null) {
				
				Map additions = computeAdditions (_provider);
				model.removeAllAnnotations();
				model.replaceAnnotations(null, additions);
				
			}
			
			
		}
		finally {
			_cachedDocument = null;
			_allowCollapsing = false;
		}
	}

	private Map computeAdditions (final ASResourceProvider provider) {
		if (provider.getRegionManager() != null && provider.isRegionManagerValid()) {
			final Map annotations = new HashMap ();
			WalkerElementOperation createAnnotations = new WalkerElementOperation () {

				public void doElementOp(AcmeElement elt) {
					if (isCollapsable (elt)) {
						Region region = provider.getRegionManager().getRegion(elt);
						Position position = createProjectionPosition (region);
						if (position != null) {
							annotations.put (new AcmeProjectionAnnotation (elt,  _allowCollapsing && collapseByDefault (elt), false), position);
						}
					}
					
				}

				private boolean collapseByDefault(AcmeElement elt) {
					return elt instanceof AcmeSystem && elt.getParentRepresentation() != null;
				}

				private boolean isCollapsable(AcmeElement elt) {
					return true;
				}
				
			};
			GraphWalker.WalkOverAllElements(_provider.getAcmeDesign(), createAnnotations, GraphWalker.WALK_TOP_DOWN);
			
			return annotations;
		}
		return null;
	}
	
	/**
	 * @param region
	 * @return
	 */
	protected Position createProjectionPosition(Region region) {
		if (_cachedDocument == null) return null;
		try {
			int start = _cachedDocument.getLineOfOffset(region.getBeginIndex());
			int end = _cachedDocument.getLineOfOffset(region.getEndIndex());
			if (start != end) {
				int offset = _cachedDocument.getLineOffset(start);
				int endOffset = _cachedDocument.getLineOffset(end + 1);
				return new Position (offset, endOffset - offset);
			}
		}
		catch (BadLocationException x) {
		}
		catch (Exception  x) {
		}
		return null;
	}

	/**
	 * 
	 * @see org.eclipse.jface.text.source.projection.IProjectionListener#projectionDisabled()
	 **/
	public void projectionDisabled() {
		_cachedDocument = null;
	}

	/**
	 * @param editor
	 * @param projectionViewer
	 */
	public void install(AcmeTextEditor editor, ProjectionViewer projectionViewer) {
		_editor = editor;
		if (_editor != null) {
			_provider = (ASResourceProvider )_editor.getDocumentProvider();
			_provider.addResourceProviderListener(this);
		}
		_viewer = projectionViewer;
		_viewer.addProjectionListener(this);
	}
	
	public void uninstall () {
		if (isInstalled ()) {
			projectionDisabled();
			if (_provider != null) {
				_provider.removeResourceProviderListener(this);
			}
			_viewer.removeProjectionListener(this);
			_viewer = null;
			_editor = null;
			_provider = null;
		}
	}

	/**
	 * @return
	 */
	private boolean isInstalled() {
		return _editor != null;
	}

	/**
	 * @param event
	 * @see AcmeStudio.core.resource.provider.IResourceProviderListener#providerChanged(AcmeStudio.core.resource.provider.ResourceProviderEvent)
	 **/
	public void providerChanged(ResourceProviderEvent event) {
//		if (event.getKind() == ResourceProviderEvent.REGION_MANAGER_CHANGED) {
			if (isInstalled ()) {
				projectionDisabled ();
				projectionEnabled ();
//			}
		}
	}
}
