/*
 * Created on Apr 18, 2003
 *
 * Synchronizes between a whole design and one that has filters applied to it.
 * Changes in the whole design are mapped to changes in the filtered design.
 */
package AcmeStudio.ui.editors.text.filters;

import AcmeStudio.model.ModelHelper;
import AcmeStudio.util.AcmeModelChangeListener;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Helper;
import AcmeStudio.util.AcmeStudioChangeManager.AcmeStudioChangeEvent;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.elements.*;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyBearer;
import acmeLib.properties.AcmePropertyType;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.tools.GraphWalker;
import acmeLib.tools.WalkerElementOperation;

/**
 * @author schmerl
 * Keeps track of the original design and an alternative design, and maps changes
 * in the original design to changes in the alternative design, subject to filters
 */
public class FilteredDesignSynchronizer implements AcmeModelChangeListener {

	/** The filters that are being applied to _mirroredDesign **/
	private WalkerElementOperation[] _filters;
	/** The original, whole design **/
	private AcmeDesign _origDesign;
	/** The filtered design **/
	private AcmeDesign _mirroredDesign;

	public FilteredDesignSynchronizer(AcmeDesign orig, AcmeDesign mirror, WalkerElementOperation[] filters) {
		_filters = filters;
		_origDesign = orig;
		_mirroredDesign = mirror;
	}

	/**
	 * Applies each filter to the _mirrored design. This is done at the design level, because
	 * the WalkOverAllElements operation doesn't take an AcmeElement.
	 */
	protected void applyFilters() {
		for (int i = 0; i < _filters.length; i++) {
			GraphWalker.WalkOverAllElements(_mirroredDesign, _filters[i], GraphWalker.WALK_BOTTOM_UP);
		}
	}

	/**
	 * Given a change to the original design, map this change to the _mirroredDesign, and
	 * announce the change to the mirrored design.
	 * @param evt The event for the change
	 * @see AcmeStudio.util.AcmeModelChangeListener#changeModel(AcmeStudio.util.AcmeStudioChangeManager.AcmeStudioChangeEvent)
	 **/
	public void changeModel(AcmeStudioChangeEvent evt) {
		if (Helper.eventRelatedToDesign(evt, _origDesign)) {
			// Theres a change to the orig design
			try {
				switch (evt.getEventType()) {
					case AcmeStudioChangeManager.ADD_ARMANI_RULE :
					case AcmeStudioChangeManager.REMOVE_ARMANI_RULE :
						break;
					case AcmeStudioChangeManager.ADD_ATTACHMENT :
					case AcmeStudioChangeManager.REMOVE_ATTACHMENT :
						{
							AcmeSystem sys = (AcmeSystem) evt.getTarget();
							AcmeAttachment att = (AcmeAttachment) evt.getSource();
							AcmeSystem mSys =
								(AcmeSystem) ModelHelper.retrieveElementFromDesign(_mirroredDesign, sys.getQualifiedName());
							AcmeRole role =
								(AcmeRole) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									sys.getQualifiedName() + "." + att.getRoleNameAsString());
							AcmePort port =
								(AcmePort) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									sys.getQualifiedName() + "." + att.getPortNameAsString());
							if (evt.getEventType() == AcmeStudioChangeManager.ADD_ATTACHMENT) {
								mSys.addAttachment(port, role);
								AcmeStudioChangeManager.globalChangeManager.onAddAttachment(
									mSys,
									new AcmeAttachment(port, role),
									this);
							}
							else if (evt.getEventType() == AcmeStudioChangeManager.REMOVE_ATTACHMENT) {
								mSys.removeAttachment(new AcmeAttachment(port, role));
								AcmeStudioChangeManager.globalChangeManager.onRemoveAttachment(
									mSys,
									mSys.getAttachment(port, role),
									this);
							}

						}
						break;
					case AcmeStudioChangeManager.ADD_BINDING :
					case AcmeStudioChangeManager.REMOVE_BINDING :
						{
							AcmeRepresentation rep = (AcmeRepresentation) evt.getTarget();
							AcmeBinding binding = (AcmeBinding) evt.getSource();

							AcmeRepresentation mRep =
								(AcmeRepresentation) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									rep.getQualifiedName());
							AcmeElement outer =
								(AcmeElement) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									binding.getOuter().getQualifiedName());
							AcmeElement inner =
								(AcmeElement) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									binding.getInner().getQualifiedName());
							if (evt.getEventType() == AcmeStudioChangeManager.ADD_BINDING) {
								AcmeBinding newB = new AcmeBinding(outer, inner);
								((AcmeBindingsMap) rep.getAbstractionMap()).addBinding(newB);
								AcmeStudioChangeManager.globalChangeManager.onAddBinding(mRep, newB, this);
							}
							else if (evt.getEventType() == AcmeStudioChangeManager.REMOVE_BINDING) {
								AcmeEnumeration e = ((AcmeBindingsMap) mRep.getAbstractionMap()).getBindings();
								while (e.moreItems()) {
									AcmeBinding b = (AcmeBinding) e.getNextItem();
									if (b.getInner().equals(inner) && b.getOuter().equals(outer)) {
										((AcmeBindingsMap) mRep.getAbstractionMap()).removeBinding(b);
										AcmeStudioChangeManager.globalChangeManager.onRemoveBinding(mRep, b, this);
									}
								}
							}

							break;
						}

					case AcmeStudioChangeManager.ADD_COMPONENT :
					case AcmeStudioChangeManager.REMOVE_COMPONENT :
						{
							AcmeSystem origSys = (AcmeSystem) evt.getTarget();
							AcmeComponent origComp = (AcmeComponent) evt.getSource();

							AcmeSystem mSys =
								(AcmeSystem) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origSys.getQualifiedName());
							if (evt.getEventType() == AcmeStudioChangeManager.REMOVE_COMPONENT) {
								AcmeComponent mComp =
									(AcmeComponent) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										origSys.getQualifiedName() + "." + origComp.getName());
								mSys.removeComponent(mComp);
								AcmeStudioChangeManager.globalChangeManager.onRemoveComponent(mSys, mComp, this);
							}
							else if (evt.getEventType() == AcmeStudioChangeManager.ADD_COMPONENT) {
								AcmeComponent mComp = (AcmeComponent) origComp.copy();
								mSys.addComponent(mComp);
								AcmeStudioChangeManager.globalChangeManager.onAddComponent(mSys, mComp, this);

							}

							break;
						}
					case AcmeStudioChangeManager.ADD_COMPONENT_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmeElementType origCT = (AcmeElementType) evt.getSource();

							AcmeElementType mCT = origCT.copyType();
							AcmeFamily mFamily =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());
							mFamily.addType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onAddComponentType(
								mFamily,
								(AcmeComponentType) mCT,
								this);

							break;
						}
					case AcmeStudioChangeManager.REMOVE_COMPONENT_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmeElementType origCT = (AcmeElementType) evt.getSource();

							AcmeElementType mCT =
								(AcmeElementType) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									((AcmeElement) origCT).getQualifiedName());
							AcmeFamily mFam =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());

							mFam.removeType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onRemoveComponentType(
								mFam,
								(AcmeComponentType) mCT,
								this);
							break;
						}
					case AcmeStudioChangeManager.ADD_CONNECTOR :
					case AcmeStudioChangeManager.REMOVE_CONNECTOR :
						{
							AcmeSystem origSys = (AcmeSystem) evt.getTarget();
							AcmeConnector origComp = (AcmeConnector) evt.getSource();

							AcmeSystem mSys =
								(AcmeSystem) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origSys.getQualifiedName());
							if (evt.getEventType() == AcmeStudioChangeManager.REMOVE_CONNECTOR) {
								AcmeConnector mComp =
									(AcmeConnector) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										origSys.getQualifiedName() + "." + origComp.getName());
								mSys.removeConnector(mComp);
								AcmeStudioChangeManager.globalChangeManager.onRemoveConnector(mSys, mComp, this);
							}
							else if (evt.getEventType() == AcmeStudioChangeManager.ADD_CONNECTOR) {
								AcmeConnector mComp = (AcmeConnector) origComp.copy();
								mSys.addConnector(mComp);
								AcmeStudioChangeManager.globalChangeManager.onAddConnector(mSys, mComp, this);

							}
							break;
						}
					case AcmeStudioChangeManager.ADD_CONNECTOR_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmeElementType origCT = (AcmeElementType) evt.getSource();

							AcmeElementType mCT = origCT.copyType();
							AcmeFamily mFamily =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());
							mFamily.addType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onAddConnectorType(
								mFamily,
								(AcmeConnectorType) mCT,
								this);

							break;
						}
					case AcmeStudioChangeManager.REMOVE_CONNECTOR_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmeElementType origCT = (AcmeElementType) evt.getSource();

							AcmeElementType mCT =
								(AcmeElementType) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									((AcmeElement) origCT).getQualifiedName());
							AcmeFamily mFam =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());

							mFam.removeType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onRemoveConnectorType(
								mFam,
								(AcmeConnectorType) mCT,
								this);
							break;
						}
					case AcmeStudioChangeManager.ADD_FAMILY :
						{
							AcmeFamily family = (AcmeFamily) evt.getSource();
							AcmeFamily mFam = (AcmeFamily) family.copyType();
							_mirroredDesign.addFamily(mFam);
							AcmeStudioChangeManager.globalChangeManager.onAddFamily(_mirroredDesign, mFam, this);
							break;
						}
					case AcmeStudioChangeManager.REMOVE_FAMILY :
						{
							AcmeFamily family = (AcmeFamily) evt.getSource();
							AcmeFamily mFam = _mirroredDesign.getFamily(family.getName());
							_mirroredDesign.removeFamily(mFam.getName());
							AcmeStudioChangeManager.globalChangeManager.onRemoveFamily(_mirroredDesign, mFam, this);
							break;
						}
					case AcmeStudioChangeManager.ADD_GROUP :
					case AcmeStudioChangeManager.REMOVE_GROUP :
						{
							AcmeSystem origSys = (AcmeSystem) evt.getTarget();
							AcmeGroup origComp = (AcmeGroup) evt.getSource();

							AcmeSystem mSys =
								(AcmeSystem) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origSys.getQualifiedName());
							if (evt.getEventType() == AcmeStudioChangeManager.REMOVE_GROUP) {
								AcmeGroup mComp =
									(AcmeGroup) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										origSys.getQualifiedName() + "." + origComp.getName());
								mSys.removeGroup(mComp);
								AcmeStudioChangeManager.globalChangeManager.onRemoveGroup(mSys, mComp, this);
							}
							else if (evt.getEventType() == AcmeStudioChangeManager.ADD_GROUP) {
								AcmeGroup mComp = (AcmeGroup) origComp.copy();
								mSys.addGroup(mComp);
								AcmeStudioChangeManager.globalChangeManager.onAddGroup(mSys, mComp, this);

							}
							break;
						}
					case AcmeStudioChangeManager.ADD_MEMBER :
						{
							AcmeGroup group = (AcmeGroup) evt.getTarget();
							AcmeElement member = (AcmeElement) evt.getSource();

							AcmeGroup mGroup =
								(AcmeGroup) ModelHelper.retrieveElementFromDesign(_mirroredDesign, group.getQualifiedName());
							AcmeElement mElement =
								(AcmeElement) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									member.getQualifiedName());
							mGroup.addMember(mElement);
							AcmeStudioChangeManager.globalChangeManager.onAddMember(mGroup, mElement, this);
							break;
						}
					case AcmeStudioChangeManager.REMOVE_MEMBER :
						{
							AcmeGroup group = (AcmeGroup) evt.getTarget();
							AcmeElement element = (AcmeElement) evt.getSource();

							AcmeGroup mGroup =
								(AcmeGroup) ModelHelper.retrieveElementFromDesign(_mirroredDesign, group.getQualifiedName());
							AcmeElement mElement =
								(AcmeGroup) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									element.getQualifiedName());
							mGroup.removeMember(mElement);
							AcmeStudioChangeManager.globalChangeManager.onRemoveMember(mGroup, mElement, this);
							break;
						}
					case AcmeStudioChangeManager.ADD_PORT :
					case AcmeStudioChangeManager.REMOVE_PORT :
						{
							AcmeComponent origSys = (AcmeComponent) evt.getTarget();
							AcmePort origComp = (AcmePort) evt.getSource();

							AcmeComponent mSys =
								(AcmeComponent) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origSys.getQualifiedName());
							if (evt.getEventType() == AcmeStudioChangeManager.REMOVE_PORT) {
								AcmePort mComp =
									(AcmePort) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										origSys.getQualifiedName() + "." + origComp.getName());
								mSys.removePort(mComp);
								AcmeStudioChangeManager.globalChangeManager.onRemovePort(mSys, mComp, this);
							}
							else if (evt.getEventType() == AcmeStudioChangeManager.ADD_PORT) {
								AcmePort mComp = (AcmePort) origComp.copy();
								mSys.addPort(mComp);
								AcmeStudioChangeManager.globalChangeManager.onAddPort(mSys, mComp, this);

							}
							break;
						}
					case AcmeStudioChangeManager.ADD_PORT_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmeElementType origCT = (AcmeElementType) evt.getSource();

							AcmeElementType mCT = origCT.copyType();
							AcmeFamily mFamily =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());
							mFamily.addType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onAddPortType(
								mFamily,
								(AcmePortType) mCT,
								this);

							break;
						}
					case AcmeStudioChangeManager.REMOVE_PORT_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmeElementType origCT = (AcmeElementType) evt.getSource();

							AcmeElementType mCT =
								(AcmeElementType) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									((AcmeElement) origCT).getQualifiedName());
							AcmeFamily mFam =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());

							mFam.removeType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onRemovePortType(
								mFam,
								(AcmePortType) mCT,
								this);
							break;
						}
					case AcmeStudioChangeManager.ADD_PROPERTY :
					case AcmeStudioChangeManager.REMOVE_PROPERTY :
						{
							AcmePropertyBearer parent = (AcmePropertyBearer) evt.getTarget();
							AcmeProperty prop = (AcmeProperty) evt.getSource();
							String name = "";
							if (parent instanceof AcmeProperty) {
								name = ((AcmeProperty) parent).getQualifiedName();
							}
							else if (parent instanceof AcmeElement) {
								name = ((AcmeElement) parent).getQualifiedName();
							}
							AcmePropertyBearer mParent =
								(AcmePropertyBearer) ModelHelper.retrieveElementFromDesign(_mirroredDesign, name);
							if (evt.getEventType() == AcmeStudioChangeManager.REMOVE_PROPERTY) {

								AcmeProperty mProp = mParent.getProperty(prop.getName());
								mParent.removeProperty(prop.getName());

								AcmeStudioChangeManager.globalChangeManager.onRemoveProperty(mParent, mProp, this);

							}
							else if (evt.getEventType() == AcmeStudioChangeManager.ADD_PROPERTY) {
								AcmeProperty mProp = prop.copy();
								mParent.addProperty(mProp);

								AcmeStudioChangeManager.globalChangeManager.onAddProperty(mParent, mProp, this);

							}
							break;
						}
					case AcmeStudioChangeManager.REMOVE_PROPERTY_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmePropertyType origCT = (AcmePropertyType) evt.getSource();

							AcmePropertyType mCT =
								(AcmePropertyType) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origCT.getQualifiedName());
							AcmeFamily mFam =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());

							mFam.removeType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onRemovePropertyType(
								mFam,
								mCT,
								this);
							break;
						}
					case AcmeStudioChangeManager.ADD_PROPERTY_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmePropertyType origCT = (AcmePropertyType) evt.getSource();

							AcmePropertyType mCT = origCT.copy();
							AcmeFamily mFamily =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());
							mFamily.addType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onAddPropertyType(
								mFamily,
								mCT,
								this);

							break;
						}
					case AcmeStudioChangeManager.ADD_REPRESENTATION :
						{
							AcmeElement element = (AcmeElement) evt.getTarget();
							AcmeRepresentation rep = (AcmeRepresentation) evt.getSource();

							AcmeElement mElement =
								(AcmeElement) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									element.getQualifiedName());
							AcmeRepresentation mRep = rep.copy();
							mElement.addRepresentation(mRep);
							AcmeStudioChangeManager.globalChangeManager.onAddRepresentation(mElement, mRep, this);
							break;
						}
					case AcmeStudioChangeManager.REMOVE_REPRESENTATION :
						{
							AcmeElement element = (AcmeElement) evt.getTarget();
							AcmeRepresentation rep = (AcmeRepresentation) evt.getSource();

							AcmeElement mElement =
								(AcmeElement) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									element.getQualifiedName());
							AcmeRepresentation mRep = mElement.getRepresentation(rep.getName());
							mElement.removeRepresentation(rep.getName());
							AcmeStudioChangeManager.globalChangeManager.onRemoveRepresentation(mElement, mRep, this);
							break;
						}
					case AcmeStudioChangeManager.ADD_ROLE :
					case AcmeStudioChangeManager.REMOVE_ROLE :
						{
							AcmeConnector origSys = (AcmeConnector) evt.getTarget();
							AcmeRole origComp = (AcmeRole) evt.getSource();

							AcmeConnector mSys =
								(AcmeConnector) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origSys.getQualifiedName());
							if (evt.getEventType() == AcmeStudioChangeManager.REMOVE_PORT) {
								AcmeRole mComp =
									(AcmeRole) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										origSys.getQualifiedName() + "." + origComp.getName());
								mSys.removeRole(mComp);
								AcmeStudioChangeManager.globalChangeManager.onRemoveRole(mSys, mComp, this);
							}
							else if (evt.getEventType() == AcmeStudioChangeManager.ADD_PORT) {
								AcmeRole mComp = (AcmeRole) origComp.copy();
								mSys.addRole(mComp);
								AcmeStudioChangeManager.globalChangeManager.onAddRole(mSys, mComp, this);
							}
							break;
						}
					case AcmeStudioChangeManager.ADD_ROLE_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmeElementType origCT = (AcmeElementType) evt.getSource();

							AcmeElementType mCT = origCT.copyType();
							AcmeFamily mFamily =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());
							mFamily.addType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onAddRoleType(
								mFamily,
								(AcmeRoleType) mCT,
								this);

							break;
						}
					case AcmeStudioChangeManager.REMOVE_ROLE_TYPE :
						{
							AcmeFamily origFamily = (AcmeFamily) evt.getTarget();
							AcmeElementType origCT = (AcmeElementType) evt.getSource();

							AcmeElementType mCT =
								(AcmeElementType) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									((AcmeElement) origCT).getQualifiedName());
							AcmeFamily mFam =
								(AcmeFamily) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									origFamily.getQualifiedName());

							mFam.removeType(mCT);
							AcmeStudioChangeManager.globalChangeManager.onRemoveRoleType(
								mFam,
								(AcmeRoleType) mCT,
								this);
							break;
						}
					case AcmeStudioChangeManager.ADD_SYSTEM :
						{
							AcmeSystem system = (AcmeSystem) evt.getSource();

							AcmeSystem mSys = (AcmeSystem) system.copy();
							_mirroredDesign.addSystem(mSys);
							AcmeStudioChangeManager.globalChangeManager.onAddSystem(_mirroredDesign, mSys, this);
							break;
						}
					case AcmeStudioChangeManager.REMOVE_SYSTEM :
						{
							AcmeSystem system = (AcmeSystem) evt.getSource();
							AcmeSystem mSys = _mirroredDesign.getSystem(system.getName());
							_mirroredDesign.removeSystem(mSys);
							AcmeStudioChangeManager.globalChangeManager.onRemoveSystem(_mirroredDesign, mSys, this);
							break;
						}
					case AcmeStudioChangeManager.ADD_TYPE :
						{
							AcmeType type = (AcmeType) evt.getSource();

							AcmeType mType = null;
							if (type instanceof AcmePropertyType) {
								mType = ((AcmePropertyType) type).copy();
							}
							else if (type instanceof AcmeElementType) {
								mType = ((AcmeElementType) type).copyType();
							}
							if (mType == null)
								break;

							if (evt.getTarget() instanceof AcmeDesign) {

								_mirroredDesign.addType(mType);
								AcmeStudioChangeManager.globalChangeManager.onAddType(_mirroredDesign, mType, this);
							}
							else {
								AcmeFamily fam = (AcmeFamily) evt.getTarget();
								AcmeFamily mFam = _mirroredDesign.getFamily(fam.getTypeName());

								mFam.addType(mType);
								AcmeStudioChangeManager.globalChangeManager.onAddType(mFam, mType, this);
							}

							break;
						}
					case AcmeStudioChangeManager.REMOVE_TYPE :
						{
							AcmeType type = (AcmeType) evt.getSource();
							if (evt.getTarget() instanceof AcmeDesign) {
								AcmeType mType = _mirroredDesign.getType(type.getTypeName());
								_mirroredDesign.removeType(mType);
								AcmeStudioChangeManager.globalChangeManager.onRemoveType(_mirroredDesign, mType, this);
							}
							else {
								AcmeFamily mFam = _mirroredDesign.getFamily(((AcmeFamily) evt.getTarget()).getName());
								AcmeType mType = mFam.getType(type.getTypeName());
								mFam.removeType(mType);
								AcmeStudioChangeManager.globalChangeManager.onRemoveType(mFam, mType, this);
							}
							break;
						}
					case AcmeStudioChangeManager.ASSIGN_FAMILY :
						{
							AcmeSystem system = (AcmeSystem) evt.getTarget();
							AcmeFamily family = (AcmeFamily) evt.getSource();

							AcmeSystem mSys =
								(AcmeSystem) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									system.getQualifiedName());
							AcmeFamily mFam = _mirroredDesign.getFamily(family.getName());
							mSys.addDeclaredType(mFam);
							AcmeStudioChangeManager.globalChangeManager.onAssignFamily(mSys, mFam, this);

							break;
						}
					case AcmeStudioChangeManager.UNASSIGN_FAMILY :
						{
							AcmeSystem system = (AcmeSystem) evt.getTarget();
							AcmeFamily family = (AcmeFamily) evt.getSource();

							AcmeSystem mSys =
								(AcmeSystem) ModelHelper.retrieveElementFromDesign(
									_mirroredDesign,
									system.getQualifiedName());
							AcmeFamily mFam = _mirroredDesign.getFamily(family.getName());
							mSys.removeDeclaredType(mFam);
							AcmeStudioChangeManager.globalChangeManager.onUnAssignFamily(mSys, mFam, this);
							break;
						}
					case AcmeStudioChangeManager.ASSIGN_TYPE :
						{
							AcmeElementType type = (AcmeElementType) evt.getSource();

							if (evt.getTarget() instanceof AcmeElementType) {
								AcmeElementType target = (AcmeElementType) evt.getTarget();
								AcmeElementType mTarget =
									(AcmeElementType) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										((AcmeElement) target).getQualifiedName());
								AcmeElementType mType = (AcmeElementType) mTarget.lookupName(type.getTypeName(), true);
								mTarget.addSuperType(mType);
								AcmeStudioChangeManager.globalChangeManager.onAssignType(mTarget, mType, this);

							}
							else {
								AcmeElement target = (AcmeElement) evt.getTarget();
								AcmeElement mTarget =
									(AcmeElement) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										target.getQualifiedName());
								AcmeElementType mType = (AcmeElementType) mTarget.lookupName(type.getTypeName(), true);
								if (target.declaresType(type))
									mTarget.addDeclaredType(mType);
								if (target.instantiatesType(type))
									mTarget.addInstantiatedType(mType);
								AcmeStudioChangeManager.globalChangeManager.onAssignType(mTarget, mType, this);
							}
							break;
						}
					case AcmeStudioChangeManager.UNASSIGN_TYPE :
						{
							AcmeType type = (AcmeType) evt.getSource();
							if (evt.getTarget() instanceof AcmeElementType) {
								AcmeElementType target = (AcmeElementType) evt.getTarget();
								AcmeElementType mTarget =
									(AcmeElementType) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										((AcmeElement) target).getQualifiedName());
								AcmeElementType mType = (AcmeElementType) mTarget.lookupName(type.getTypeName(), true);
								mTarget.removeSuperType(mType);
								AcmeStudioChangeManager.globalChangeManager.onUnAssignType(mTarget, mType, this);
							}
							else {
								AcmeElement target = (AcmeElement) evt.getTarget();
								AcmeElement mTarget =
									(AcmeElement) ModelHelper.retrieveElementFromDesign(
										_mirroredDesign,
										target.getQualifiedName());
								AcmeElementType mType = (AcmeElementType) mTarget.lookupName(type.getTypeName(), true);
								mTarget.removeDeclaredType(mType);
								mTarget.removeInstantiatedType(mType);
								AcmeStudioChangeManager.globalChangeManager.onUnAssignType(mTarget, mType, this);
							}
							break;
						}
						// TODO: Finish these up
					case AcmeStudioChangeManager.IMPORT_DESIGN :
					case AcmeStudioChangeManager.SET_COMMENT :
					case AcmeStudioChangeManager.SET_NAME :
					case AcmeStudioChangeManager.SET_PROPERTY :

						}
				applyFilters();
			}
			catch (AcmeIllegalNameException e) {
				e.printStackTrace();
			}
		}

	}

	/* (non-Javadoc)
	 * @see AcmeStudio.util.AcmeModelChangeListener#interestedInEvent(int)
	 */
	public boolean interestedInEvent(int eventType) {
		return true;
	}

}
