// $Header$
package AcmeStudio.ui.views;

import org.eclipse.gef.EditPart;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IStatusLineManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerSorter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.PageBook;
import org.eclipse.ui.views.contentoutline.ContentOutlinePage;
import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASContext;
import AcmeStudio.core.ASModel;
import AcmeStudio.core.resource.ASDesign;
import AcmeStudio.ui.editors.DiagramEditor;
import AcmeStudio.ui.editors.MultiPageAcmeEditor;
import AcmeStudio.ui.editors.diagram.parts.ElementEditPart;
import AcmeStudio.ui.views.parts.DesignBrowserPreferencesDialog;
import AcmeStudio.ui.views.providers.AcmeViewContentProvider;
import AcmeStudio.ui.views.providers.AcmeViewLabelProvider;
import AcmeStudio.util.AcmeModelChangeListener;
import AcmeStudio.util.AcmeStudioChangeManager;
import AcmeStudio.util.Helper;
import AcmeStudio.util.ImageManager;
import AcmeStudio.util.AcmeStudioChangeManager.AcmeStudioChangeEvent;
import acmeLib.AcmeObject;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.elements.AcmeComponent;
import acmeLib.elements.AcmeConnector;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmePort;
import acmeLib.elements.AcmeRole;
import acmeLib.elements.AcmeSystem;
import acmeLib.representations.AcmeRepresentation;

/**
 * @author Bradley Schmerl (schmerl@cs.cmu.edu)
 *
 * DesignBrowser implements a hierarchical, tree like view of an
 * Acme resource file, displaying types, families and systems.
 * The view provides the following funcionality:
 *    1. Ability to drill down into the design
 *    2. Show or hide properties
 *    3. Allow automated change of the selection
 *    4. Update the model to changes made by other editors
 * 
 */

public class DesignBrowser extends ContentOutlinePage
    implements ISelectionChangedListener, AcmeModelChangeListener
{
	
    private AcmeViewContentProvider _contentProvider;
    // Constants that define the preferences for design browser options
	public static final String SHOW_PROPERTIES_OPTION = "DesignBrowser.showProperties";
	public static final String SHOW_INSTANCES_ONLY_OPTION = "DesignBrowser.showInstancesOnly";
	public static final String SHOW_HIDDEN_PROPERTIES_OPTION = "DesignBrowser.showHiddenProperties";
	public static final String HIDE_REPRESENTATIONS_OPTION = "DesignBrowser.hideRepresentations";
	public static final String HIDE_PORTS_OPTION = "DesignBrowser.hidePorts";
	public static final String HIDE_COMPONENTS_OPTION = "DesignBrowser.hideComponents";
	public static final String HIDE_CONNECTORS_OPTION = "DesignBrowser.hideConnectors";
	public static final String HIDE_ROLES_OPTION = "DesignBrowser.hideRoles";
	
	
	/** The tree viewer displaying the Acme design **/
	private TreeViewer _viewer;
	/** The action that toggles display of element properties in the view **/
	private Action _showPropertiesAction;
	/** The action that provides allows the user to set preferences of the view (not yet implemented) **/
	private Action _preferencesAction;
	/** The action performed when the user double clicks on an item **/
	private Action _doubleClickAction;
    /** The action performed when the user decides to show the overview **/
    private Action _showOverviewAction;
    /** The action performed when the user decides to show the outline **/
    private Action _showOutlineAction;

	/** Indicates whether the view will display properties **/
	private boolean _showProperties = false;
	private boolean _partCreated = false;

	private Object _selection = null;
	

	
	/**
	 * Provides the sorting operation for ordering names in tree
	 */ 
	class NameSorter extends ViewerSorter {
		
		private static final int CAT_COMP = 0;
		private static final int CAT_CONN = 1;
		private static final int CAT_ROLE = 2;
		private static final int CAT_PORT = 3;
		private static final int CAT_ANAL = 4;
		private static final int CAT_SYS = 5;
		private static final int CAT_NONE = 6;
		
//		/**
//		 * Don't do any sorting.
//		 * @see org.eclipse.jface.viewers.ViewerSorter#isSorterProperty(Object, String)
//		 */
//		public boolean isSorterProperty(Object element, String property) {
//			return false;
//		}

		public int category (Object element) {
			if (element instanceof DesignAnalysis) return CAT_ANAL;
			else if (element instanceof AcmeComponent) return CAT_COMP;
			else if (element instanceof AcmeConnector) return CAT_CONN;
			else if (element instanceof AcmePort) return CAT_PORT;
			else if (element instanceof AcmeRole) return CAT_ROLE;
			else if (element instanceof AcmeSystem) return CAT_SYS;
			return CAT_NONE;
		}

		public int compare (Viewer viewer, Object e1, Object e2) {
			if (category (e1) < category (e2)) {
				return -1;
			}
			else if (category (e1) > category (e2)) {
				return 1;
			}
			
			return AcmeViewLabelProvider.getLabel(e1).toLowerCase().compareTo (AcmeViewLabelProvider.getLabel (e2).toLowerCase ());
		}
	}

	/**
     * @deprecated  use instead #DesignBrowser(ASDesign)
	 */
	public DesignBrowser (AcmeDesign design) {
		super ();
        setAcmeDesign(design);
	}

    /**
     */
    public DesignBrowser (ASDesign design) {
        super ();
        setDesign(design);
    }

    PageBook _pageBook = null;
    Canvas _overview = null;
    int currentPage = 0;
	private Action	_synchronizeSelectionWithDiagramAction;
	protected boolean	_synchronizeDiagram;
 	/**
	 * This is a callback that will allow us
	 * to create the viewer and initialize it.
	 * @param parent the place to put all the parts
	 */
	public void createControl(Composite parent) {
        if (parent instanceof PageBook) {
            _pageBook = (PageBook )parent;
        }
        
		super.createControl (parent);
		// Make sure the icons are loaded
		ImageManager.loadIcons ();
		

		// Create a new tree viewer, with all the scroll bars and multiple selection
		_viewer = this.getTreeViewer(); //new TreeViewer(parent, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL);
		// Set the various providers
        _contentProvider = new AcmeViewContentProvider();
		_viewer.setContentProvider(_contentProvider);
		_viewer.setLabelProvider(new AcmeViewLabelProvider());
		_viewer.setSorter(new NameSorter ());
		// Set the viewer to display the active design
		_viewer.setInput(ASContext.current ().getActiveDesign ().getDesign ());
		_viewer.getTree().addSelectionListener(new SelectionAdapter () {
			public void widgetSelected (SelectionEvent evt) {
				if (_synchronizeDiagram) {
					bringSelectionIntoFocus();
				}
			}
		});
		// Hook in all the actions and menus and toolbar stuff
		makeActions();
//		hookContextMenu();
		hookDoubleClickAction();
//		contributeToActionBars();
		resetPreferences ();	
//		_pageBook.showPage (_viewer.getControl ());
				
		// Add the design browser as a listener to selection changed events in the tree viewer
		// BRS: This is already added in ContentOutlinePage
		//_viewer.addSelectionChangedListener(this);
		
        
        // Make a page for the overview
        _overview = new Canvas (parent, SWT.NONE);
        MultiPageAcmeEditor editor = ASContext.current().getActiveEditor();
        if (editor != null)
            editor.getOverview(_overview);
		// Add the design browser as a listener to changes in the model
		AcmeStudioChangeManager.registerListener(this);
		_partCreated = true;

	}

	private void resetPreferences () {
		// Set all the settings for this view
		IDialogSettings settings = AcmeStudioPlugin.getDefault ().getDialogSettings();
		_contentProvider.setShowProperties(settings.getBoolean (SHOW_PROPERTIES_OPTION));
		_contentProvider.setShowInstancesOnly(settings.getBoolean (SHOW_INSTANCES_ONLY_OPTION));
		_contentProvider.setShowHiddenProperties(settings.getBoolean (SHOW_HIDDEN_PROPERTIES_OPTION));
		_contentProvider.setHideComponents(settings.getBoolean (HIDE_COMPONENTS_OPTION));
		_contentProvider.setHideConnectors(settings.getBoolean (HIDE_CONNECTORS_OPTION));
		_contentProvider.setHidePorts(settings.getBoolean (HIDE_PORTS_OPTION));
		_contentProvider.setHideRoles(settings.getBoolean (HIDE_ROLES_OPTION));
		_contentProvider.setHideRepresentations(settings.getBoolean (HIDE_REPRESENTATIONS_OPTION));
		_showPropertiesAction.setChecked (_showProperties);
		
	}

	/**
	 * Adds the actions to the context menu in the tool bar
	 */
	private void fillLocalPullDown(IMenuManager manager) {
        manager.add(_showOutlineAction);
        manager.add (_showOverviewAction);
        manager.add (new Separator ());
		manager.add(_showPropertiesAction);
		manager.add (_synchronizeSelectionWithDiagramAction);
		manager.add(new Separator());
		manager.add(_preferencesAction);
	}

	/**
	 * Adds the actions to the pull down context menu
	 */
	private void fillContextMenu(IMenuManager manager) {
        manager.add (_showOutlineAction);
        manager.add (_showOverviewAction);
        manager.add (new Separator ());
		manager.add(_showPropertiesAction);
		manager.add (_synchronizeSelectionWithDiagramAction);
		manager.add (new Separator ());
		manager.add(_preferencesAction);
 		manager.add(new Separator("Additions"));
	}

	/** Fills the tool bar
	 */	
	private void fillLocalToolBar(IToolBarManager manager) {
		
        manager.add (_showOutlineAction);
        manager.add (_showOverviewAction);
        manager.add (new Separator ());
		manager.add(_showPropertiesAction);
		manager.add (_synchronizeSelectionWithDiagramAction);
		manager.add(new Separator());
		//drillDownAdapter.addNavigationActions(manager);
	}

	/**
	 * Defines all the actions for this view
	 */
	private void makeActions() {
		// Toggles whether to show properties
		
		_showPropertiesAction = new Action() {
			public void run() {
				IDialogSettings settings = AcmeStudioPlugin.getDefault ().getDialogSettings();
				_showProperties = !_showProperties;
				settings.put ("DesignBrowser.showProperties", _showProperties);
				
				_showPropertiesAction.setChecked (_showProperties);
				if (_design != null) {
					_viewer.refresh (_design.getDesign(), true);
				}
			}
			
			public int getStyle () {
				return AS_CHECK_BOX;
			}
		};
		_showPropertiesAction.setText("Display Properties");
		_showPropertiesAction.setToolTipText("Displays non-hidden properties in the browser");
		_showPropertiesAction.setImageDescriptor(ImageManager.getSmallImageDescriptor (ImageManager.ID_PROPERTY));
		_showPropertiesAction.setChecked (_showProperties);
		// Shows the preferences dialog (not implemented)
		_preferencesAction = new Action() {
			public void run() {
				Shell shell = AcmeStudioPlugin.getDefault().getWorkbench().getActiveWorkbenchWindow().getShell();
				DesignBrowserPreferencesDialog dbpd = new DesignBrowserPreferencesDialog (shell);
				if (dbpd.open () == IDialogConstants.OK_ID) {
					resetPreferences ();
					_viewer.refresh (_design.getDesign(), true);
				}
			}
		};
		_preferencesAction.setText("Preferences...");
		_preferencesAction.setToolTipText("Changes the preferences for this view");
		_preferencesAction.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
			getImageDescriptor(org.eclipse.ui.ide.IDE.SharedImages.IMG_OBJS_TASK_TSK));
            
        _showOverviewAction = new Action () {
            public void run () {
                _showOverviewAction.setChecked (true);
                _showOutlineAction.setChecked (false);
                MultiPageAcmeEditor editor = ASContext.current().getActiveEditor();

                if (editor != null)
                    editor.getOverview(_overview);
                _pageBook.showPage (_overview);
                ASContext.current().getActiveEditor().setThumbnailVisible(true);  

            }
        };
        _showOverviewAction.setText ("Show Overview");
        _showOverviewAction.setToolTipText("Displays the overview of the diagram");
        _showOverviewAction.setImageDescriptor (ImageManager.getImageDescriptor(ImageManager.ID_OVERVIEW));
        _showOverviewAction.setChecked (false);
        
        _showOutlineAction = new Action () {
            public void run () {
                _showOverviewAction.setChecked (false);
                _showOutlineAction.setChecked (true);
                _pageBook.showPage (DesignBrowser.this.getTreeViewer().getControl ());
               ASContext.current().getActiveEditor().setThumbnailVisible(false);  
            }
        };
        _showOutlineAction.setText ("Show outline");
        _showOutlineAction.setToolTipText("Displays the outline of the design");
        _showOutlineAction.setImageDescriptor (ImageManager.getImageDescriptor (ImageManager.ID_OUTLINE));
        _showOutlineAction.setChecked (true);

        _synchronizeSelectionWithDiagramAction = new Action () {
        	public void run () {
        		_synchronizeDiagram = this.isChecked();
        		bringSelectionIntoFocus ();
        	}
        };
        _synchronizeSelectionWithDiagramAction.setText ("Synchronize with editor");
        _synchronizeSelectionWithDiagramAction.setToolTipText("Synchronizes selections in the design browser with selections in the diagram");
        _synchronizeSelectionWithDiagramAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_SYNCHED));
        _synchronizeSelectionWithDiagramAction.setChecked (_synchronizeDiagram);
        
        _doubleClickAction = new Action () {
        	public void run () {
        		bringSelectionIntoFocus ();
        	}
        };
	}

	/** Provides the hook for double click
	 */
	private void hookDoubleClickAction() {
		_viewer.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent event) {
                if (_doubleClickAction != null) {
                    _doubleClickAction.run();
                }
			}
		});
	}
	
	/**
	 * Passing the focus request to the viewer's control.
	 */
	public void setFocus() {
		if (_viewer == null) _viewer = this.getTreeViewer();
		if (_viewer != null) _viewer.getControl().setFocus();
	}
	
	private AcmeObject _selectedObject = null;
	private ASDesign _design;
	
	public void setFocus (AcmeObject object) {
		if (!_partCreated) return;
		if (object instanceof AcmeDesign) {
            // was previously _design, corrected(?) to use "object"
			_viewer.setInput (object);
		}	
		else {
			_viewer.setInput (_design.getDesign());
			if (_selectedObject != object) {
				_selectedObject = object;
				_viewer.setSelection (new StructuredSelection (object), true);
			}
		}
	}

    /**
     * Sets the design used by this view. 
     * @param design
     * @deprecated  use #setDesign(ASDesign) instead to ensure proper behavior
     */
    public void setAcmeDesign (AcmeDesign design) {
        setDesign((ASDesign )ASModel.current().getASResource(design));
    }
    
    /**
     * Sets the design used by this view (originally setAcmeDesign)
     * @param design  the ASDesign object
     */
    public void setDesign (ASDesign design) {
        if (_design != design) {
        ASContext.current().setActiveDesign(design);
        _design = design;
        }
        if (_viewer != null) {
            _viewer.setInput(design.getDesign());
        }
    }
	
	/**
	 * @see org.eclipse.jface.viewers.ISelectionChangedListener#selectionChanged(SelectionChangedEvent)
	 */
	public void selectionChanged(SelectionChangedEvent event) {
		if (event.getSelection() instanceof IStructuredSelection && event.getSource() != _viewer && !(event.getSource() instanceof DesignBrowser)) {
			IStructuredSelection iss = (IStructuredSelection )event.getSelection ();
			if (iss.size () == 1) {
				AcmeElement selection = null;
				if (iss.getFirstElement () instanceof AcmeElement) selection = (AcmeElement )iss.getFirstElement ();
				if (iss.getFirstElement () instanceof ElementEditPart) {
					
					ElementEditPart elementEditPart = ((ElementEditPart )iss.getFirstElement ());
					if (elementEditPart.getModel () instanceof AcmeElement)
						selection = (AcmeElement )elementEditPart.getModel ();
				}
				if (selection != null && 
	                event.getSelection () != null &&
	                _viewer != null &&
	                !event.getSelection ().equals (_viewer.getSelection ()) && _selection != selection) {
					_selection = selection;
					this.setFocus(selection)	;
				}
			}
		} 
	}

	/**
	 * @see AcmeStudio.model.AcmeModelChangeListener#changeModel(AcmeStudioChangeEvent)
	 */
	public void changeModel(
		AcmeStudioChangeManager.AcmeStudioChangeEvent evt) {
			// Only interested if the event originated from some other view
			if (evt.getAnnouncer () != this && !_viewer.getControl ().isDisposed()) {
				switch (evt.getEventType()) {
					case AcmeStudioChangeManager.SET_NAME:
						_viewer.refresh (evt.getTarget (), true);
						break;
					case AcmeStudioChangeManager.ADD_COMPONENT:
					case AcmeStudioChangeManager.ADD_CONNECTOR:
					case AcmeStudioChangeManager.ADD_ROLE:
					case AcmeStudioChangeManager.ADD_PORT:
                    case AcmeStudioChangeManager.ADD_SYSTEM:
                    case AcmeStudioChangeManager.REMOVE_SYSTEM:
					case AcmeStudioChangeManager.REMOVE_COMPONENT:
					case AcmeStudioChangeManager.REMOVE_CONNECTOR:
					case AcmeStudioChangeManager.REMOVE_ROLE:
					case AcmeStudioChangeManager.REMOVE_PORT:
					case AcmeStudioChangeManager.REMOVE_PROPERTY:
					case AcmeStudioChangeManager.ADD_PROPERTY:
                    case AcmeStudioChangeManager.ADD_REPRESENTATION:
                    case AcmeStudioChangeManager.REMOVE_REPRESENTATION:
                    case AcmeStudioChangeManager.ADD_FAMILY:
                    case AcmeStudioChangeManager.REMOVE_FAMILY:
                    default:
                        _viewer.refresh (evt.getTarget (), true);
//						_viewer.refresh();
						break;
				}
				
			}
		
	}
	
	public boolean interestedInEvent (int et) {
		return et != AcmeStudioChangeManager.UPDATE_VIS_INFO;
	}
	
	
	/**
	 * @see org.eclipse.ui.part.Page#makeContributions(IMenuManager, IToolBarManager, IStatusLineManager)
	 */
	public void makeContributions(
		IMenuManager menuManager,
		IToolBarManager toolBarManager,
		IStatusLineManager statusLineManager) {
		super.makeContributions(menuManager, toolBarManager, statusLineManager);
		fillLocalPullDown(menuManager);
		fillLocalToolBar(toolBarManager);
		fillContextMenu(menuManager);
//		makeActions (toolBarManager);
	}


	/**
	 * @see org.eclipse.jface.viewers.ISelectionProvider#addSelectionChangedListener(ISelectionChangedListener)
	 */
	public void addSelectionChangedListener(ISelectionChangedListener listener) {
		super.addSelectionChangedListener(listener);
        if (_viewer != null)
            _viewer.addSelectionChangedListener(listener);
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.part.IPage#dispose()
	 */
	public void dispose() {
        AcmeStudioChangeManager.removeListener(this);
        if (_viewer != null) {
            _viewer.removeSelectionChangedListener(this);
            _viewer = null;
        }
        if (_overview != null) {
            _overview.dispose();
            _overview = null;
        }
        _pageBook = null;

        super.dispose();
	}

	/**
	 * @param element
	 * @param ed
	 */
	protected void revealElement(AcmeElement element, DiagramEditor ed) {
		// BugFix#240: Only reveal the element if it exists
		// It may not exist if an element in a representsion is double-clicked, for example
		EditPart editPart = ed.getEditPart(element);
		if (editPart != null) 
			ed.getGraphicalViewer().reveal(editPart);
	}

	/**
	 * @return AcmeDesign
     * @deprecated use instead #getDesign() which returns ASDesign
	 */
	public AcmeDesign getAcmeDesign() {
		return (_design == null) ? null : _design.getDesign();
	}

    /**
     * @return ASDesign  the AcmeDesign-encapsulating ASResource
     */
    public ASDesign getDesign() {
		return _design;
	}

	/**
	 * 
	 */
	protected void bringSelectionIntoFocus() {
		Object sel = ((IStructuredSelection )_viewer.getSelection()).getFirstElement();
		// Changed to allow switching between elements in representations
		if (sel instanceof AcmeRepresentation) {
			AcmeRepresentation rep = (AcmeRepresentation )sel;
			sel = rep.getSystem();
		}
		if (sel instanceof AcmeElement) {
			AcmeElement element = (AcmeElement )sel;
			MultiPageAcmeEditor mpe = ASContext.current ().getActiveEditor();
			if (mpe.getActiveEditor() instanceof DiagramEditor) {
				DiagramEditor ed = (DiagramEditor )mpe.getActiveEditor ();
				AcmeSystem targetSystem = Helper.getAcmeSystem(element);
				if (ed.getAcmeSystem() == targetSystem) { // The system is the one currently displayed
					revealElement (element, ed);
				}
				else {
					// Open the appropriate representation
					mpe.swapDiagramEditor(targetSystem, ed.getAcmeSystem ());
					if (mpe.getActiveEditor () instanceof DiagramEditor) {
     	
						// Make sure that the element is selected
						fireSelectionChanged(new StructuredSelection (ed.getAcmeSystem ()));
						StructuredSelection elementAsSelection = new StructuredSelection (element);
						fireSelectionChanged(elementAsSelection);
					
						ed = (DiagramEditor )mpe.getActiveEditor ();
						revealElement (element, ed);
					}
				}
			}
		}
	}


}