/*
 * Created on May 11, 2003
 *
 */
package AcmeStudio.ui.views;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.*;
import org.eclipse.ui.part.ViewPart;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.ui.views.providers.AcmeViewContentProvider;
import AcmeStudio.ui.views.providers.AcmeViewLabelProvider;
import AcmeStudio.util.ImageManager;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.filters.NoBaseClassesEnumFilter;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;

/**
 * @author schmerl
 * Provides a tree viewer to browse hierarchies of Acme types. Can take either
 * a type or an element.
 */
public class TypeHierarchyView extends ViewPart {
	/** The viewer for the hierarchy **/
	private TreeViewer _viewer;

	/** Indicates an empty list **/
	public static Object[] EMPTY = new Object[0];

	/** The objects whose type hierarchies are being browsed **/
	private List _rootTypes = new ArrayList();

	/** An action to erase all the contents of the browser **/
	private Action _eraseContentsAction;

	/** Show the details? **/
	private boolean _showDetails = true;

	/** The content provider **/
	class ViewContentProvider implements IStructuredContentProvider, ITreeContentProvider {

		AcmeViewContentProvider _detailProvider = new AcmeViewContentProvider();

		/* (non-Javadoc)
		* @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
		*/
		public Object[] getElements(Object inputElement) {
			return getChildren(inputElement);
		}

		class Details {
			List _details;
		}

		public ViewContentProvider() {
			_detailProvider.setShowProperties(true);
		}

		protected Details getDetails(AcmeElement element) {
			Details d = null;
			if (_showDetails) {
				d = new Details();
				d._details = Arrays.asList(_detailProvider.getChildren(element));
			}
			return d;
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.ITreeContentProvider#getChildren(java.lang.Object)
		 */
		public Object[] getChildren(Object parentElement) {
			Object[] elements = EMPTY;

			if (parentElement instanceof List) {
				// Allows multiple elements to be browsed
				return ((ArrayList) parentElement).toArray();
			}
			else if (parentElement instanceof AcmeElementType) {
				// Get all the supertypes
				AcmeElementType type = (AcmeElementType) parentElement;
				AcmeEnumeration supers = type.getSuperTypes().enumerate(new NoBaseClassesEnumFilter());
				int dSize = 0;
				Details details = getDetails((AcmeElement) type);
				if (details != null) {
					dSize = 1;
				}
				elements = new Object[supers.getSize() + dSize];
				int index = 0;
				while (supers.moreItems()) {
					AcmeElementType st = (AcmeElementType) supers.getNextItem();
					elements[index++] = st;
				}
				if (details != null) {
					elements[index] = details;
				}
			}
			else if (parentElement instanceof AcmeElement) {
				// Return the types of the element
				AcmeElement element = (AcmeElement) parentElement;
				AcmeEnumeration types = element.getDeclaredTypes();
				types.concatenate(element.getInstantiatedTypes());
				types = types.castToAcmeSetVector().enumerate();

				int dSize = 0;
				Details d = getDetails(element);
				if (d != null)
					dSize = 1;
				elements = new Object[types.getSize() + dSize];
				int index = 0;
				while (types.moreItems()) {
					elements[index++] = types.getNextItem();
				}

				if (d != null) {
					elements[index] = d;
				}
			}
			else if (parentElement instanceof Details) {
				elements = ((Details) parentElement)._details.toArray();
			}

			return elements;
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.ITreeContentProvider#getParent(java.lang.Object)
		 */
		public Object getParent(Object element) {
			return null;
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.ITreeContentProvider#hasChildren(java.lang.Object)
		 */
		public boolean hasChildren(Object element) {
			if (element instanceof AcmeElementType) {
				return ((AcmeElementType) element).getSuperTypes().getSize() != 0
					|| (getDetails((AcmeElement) element) != null
						&& !getDetails((AcmeElement) element)._details.isEmpty());
			}
			else if (element instanceof AcmeElement) {
				AcmeElement e = (AcmeElement) element;
				return (e.getDeclaredTypes().getSize() > 0 || e.getInstantiatedTypes().getSize() > 0)
					|| (getDetails((AcmeElement) element) != null
						&& !getDetails((AcmeElement) element)._details.isEmpty());
			}
			else if (element instanceof Details) {
				return true;
			}
			return false;
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.IContentProvider#dispose()
		 */
		public void dispose() {

		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.IContentProvider#inputChanged(org.eclipse.jface.viewers.Viewer, java.lang.Object, java.lang.Object)
		 */
		public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {

		}

	}

	class ViewLabelProvider extends LabelProvider {

		AcmeViewLabelProvider _labelProvider = new AcmeViewLabelProvider();

		/* (non-Javadoc)
		* @see org.eclipse.jface.viewers.ILabelProvider#getText(java.lang.Object)
		*/
		public String getText(Object element) {
			String name = "";
			if (element instanceof AcmeElementType) {
				AcmeElementType type = (AcmeElementType) element;
				name = type.getTypeName();
				if (type instanceof AcmeElement) {
					if (((AcmeElement) type).getParent() != null) {
						name += " (defined in Family ";
						name += ((AcmeElement) type).getParent().getName() + ")";
					}
				}

			}
			else if (element instanceof AcmeElement) {
				name = ((AcmeElement) element).getName();
			}
			else if (element instanceof ViewContentProvider.Details) {
				name = "Details";
			}
			else {
				name = _labelProvider.getText(element);
			}
			return name;
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.ILabelProvider#getImage(java.lang.Object)
		 */
		public Image getImage(Object element) {
			return ImageManager.getSmallImageByType(element);
		}

	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.IWorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createPartControl(Composite parent) {
		_viewer = new TreeViewer(parent);
		_viewer.setContentProvider(new ViewContentProvider());
		_viewer.setLabelProvider(new ViewLabelProvider());

		makeActions();
		fillActions();
	}

	/**
	 * Puts all the actions in the tool bar manager
	 */
	private void fillActions() {
		IToolBarManager mgr = getViewSite().getActionBars().getToolBarManager();
		mgr.add(_eraseContentsAction);
	}

	/**
	 * Make all the actions associated with this browser
	 */
	private void makeActions() {
		_eraseContentsAction = new Action() {
			public void run() {
				_rootTypes.clear();
				_viewer.refresh();
			}

		};
		_eraseContentsAction.setText("Clear view");
		_eraseContentsAction.setToolTipText("Clears all the types from the view");
		_eraseContentsAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_CLEAR));
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.IWorkbenchPart#setFocus()
	 */
	public void setFocus() {

	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.IViewPart#init(org.eclipse.ui.IViewSite)
	 */
	public void init(IViewSite site) throws PartInitException {
		super.init(site);
	}

	/**
	 * Sets the input of the browser to be this element
	 * @param element The element to be browsed
	 */
	public void browseElement(AcmeElement element) {
		if (element != null) {
			_rootTypes.add(element);
			_viewer.setInput(_rootTypes);
			getViewSite().getActionBars().updateActionBars();
		}
	}

	/**
	 * Opens the type browser in the right place
	 * @param element The element to be browsed
	 * @param window The window to show the view in
	 */
	public static void open(AcmeElement element, IWorkbenchWindow window) {
		IWorkbench workbench = AcmeStudioPlugin.getDefault().getWorkbench();
		IWorkbenchPage page;
		try {
			page = workbench.showPerspective("AcmeStudio.perspective", window);
			TypeHierarchyView part = (TypeHierarchyView) page.showView("AcmeStudio.ui.views.TypeHierarchyView");
			part.browseElement(element);
		}
		catch (WorkbenchException e) {
			e.printStackTrace();
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.ui.IWorkbenchPart#dispose()
	 */
	public void dispose() {
		super.dispose();
		_rootTypes.clear ();
	}

}
