/*
 * Created on Oct 21, 2003
 *  
 */
package AcmeStudio.ui.views.groups.wizards;

import org.eclipse.jface.wizard.IWizardPage;
import org.eclipse.jface.wizard.Wizard;

import AcmeStudio.ui.views.groups.GroupManager;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.elements.AcmeGroup;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyValue;

/**
 * @author Bradley Schmerl
 * 
 * Provides the facilities for defining membership rules for groups
 */
public class GroupRuleWizard extends Wizard {

    /** The group to define membership rules for **/
    protected AcmeGroup _group;
    /** The page to define the sorts of members **/
    private RuleKeyPage _ruleKey;
    /** The page to define rules for components **/
    private ComponentConnectorSelectionCriteriaPage _componentPage;
    /** The page to define rules for connector **/
    private ComponentConnectorSelectionCriteriaPage _connectorPage;
    /** The page to define rules for ports **/
    private PortRoleSelectionCriteriaPage _portPage;
    /** The page to define rules for roles **/
    private PortRoleSelectionCriteriaPage _rolePage;

    /**
     * Constructs the rule wizard
     * @param group The group to apply the rule to
     */
    public GroupRuleWizard(AcmeGroup group) {
        super();
        _group = group;
    }

    /**
	 * Adds the rules defined in pages as properties to the group
	 * 
	 * @see org.eclipse.jface.wizard.IWizard#performFinish()
	 */
    public boolean performFinish() {
        AcmePropertyValue compVal = _componentPage.getRuleRep ();
        AcmePropertyValue connVal = _connectorPage.getRuleRep ();
        AcmePropertyValue portVal = _portPage.getRuleRep ();
        AcmePropertyValue roleVal = _rolePage.getRuleRep ();
        
        setGroupProperty(compVal, _ruleKey.keyOnComponents(), "vis-component-rule");
        setGroupProperty(connVal, _ruleKey.keyOnConnectors(), "vis-connector-rule");
        setGroupProperty(portVal, _ruleKey.keyOnPorts (), "vis-port-rule");
        setGroupProperty(roleVal, _ruleKey.keyOnRoles (), "vis-role-rule");
        
        _group.removeAllMembers ();
        GroupManager.updateGroupMembers (_group);
        
        return true;
    }

    /**
     * Defines a property on the group
     * @param compVal The value to use
     * @param add Whether the component should be added
     * @param propertyName The name of the property
     */
    private void setGroupProperty(AcmePropertyValue val, boolean add, String propertyName)  {
        AcmeProperty compProp = _group.getProperty (propertyName);     
        if (add) {
            if (compProp == null) {
                try {
                    _group.addProperty(propertyName, ConstantTypes.ANY_RECORD_TYPE, val);
                } catch (AcmeIllegalNameException e) {
                    e.printStackTrace();
                }
            }
            else {
                compProp.setValue (val);
            }
        }
        else {
            _group.removeProperty(propertyName);
        }
    }

    /*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.wizard.IWizard#addPages()
	 */
    public void addPages() {
        _ruleKey = new RuleKeyPage(_group);
        addPage(_ruleKey);
        _componentPage = new ComponentConnectorSelectionCriteriaPage("Component", _group);
        addPage(_componentPage);
        _connectorPage = new ComponentConnectorSelectionCriteriaPage("Connector", _group);
        addPage(_connectorPage);
        _portPage = new PortRoleSelectionCriteriaPage("Port", _group);
        addPage(_portPage);
        _rolePage = new PortRoleSelectionCriteriaPage("Role", _group);
        addPage (_rolePage);
    }

    /**
     * Gets the next page of the wizard, based on how the rule is being defined
     * @param page the page to find the next of
     */
    public IWizardPage getNextPage(IWizardPage page) {
        if (page == _ruleKey) {
            if (_ruleKey.keyOnComponents()) {
                return _componentPage;
            } else if (_ruleKey.keyOnConnectors()) {
                return _connectorPage;
            } else if (_ruleKey.keyOnPorts()) {
                return _portPage;
            } else if (_ruleKey.keyOnRoles()) {
                return _rolePage;
            }
        } else if (page == _componentPage) {
            if (_ruleKey.keyOnConnectors()) {
                return _connectorPage;
            } else if (_ruleKey.keyOnPorts()) {
                return _portPage;
            } else if (_ruleKey.keyOnRoles()) {
                return _rolePage;
            }

        } else if (page == _connectorPage) {
            if (_ruleKey.keyOnPorts()) {
                return _portPage;
            } else if (_ruleKey.keyOnRoles()) {
                return _rolePage;
            }
        } else if (page == _portPage) {
            if (_ruleKey.keyOnRoles()) {
                return _rolePage;
            }
        }
        return null;
    }


}
