/*
 * Created on Mar 4, 2003
 *
 */
package AcmeStudio.ui.views.parts;

import java.io.IOException;
import java.util.*;
import org.eclipse.draw2d.ColorConstants;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.source.SourceViewer;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyleRange;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;
import AcmeStudio.ui.editors.text.AcmeSourceViewerConfiguration;
import AcmeStudio.ui.editors.text.TextRegionManager;
import acmeLib.AcmeObject;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.ArmaniExportable;
import acmeLib.core.io.Region;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.designRules.DesignAnalysisCallNode;
import acmeLib.designRules.DesignRule;
import acmeLib.designRules.Scope;

/**
 * @author schmerl
 * Implements a wrapper around a StyleText widget that provides browsing
 * of structure of design rules. Clicking on a design analysis call will
 * browse to the design analysis definition
 */
public class DesignRuleBrowser extends Composite implements ISelectionProvider {
    
    private Composite _buttonArea;

	/** The button used to browse forward **/
    private Button _forwardButton;

    /** The button used to browse backward **/
    private Button _backButton;

    /** The widget containing the design rule text **/
    private SourceViewer _browseText;
    
    /** The list of visited design rules/analyses **/
    private List _browseList = new ArrayList ();
    
    /** The current browsing position in the list **/
    private int currentPosition = -1;
    
    /** Indicates whether regions are valid (false when the editor has been modified) **/
    private boolean _validRegions = false;
    
    private ArmaniExportable currentlyDisplayed;
    
    /** The region manager for the text - maps regions <-> AcmeObjects **/
    TextRegionManager _regionManager;
    
    /** The list of selection listeners listening to this object **/
    Set _selectionListeners = new HashSet ();

    StyleRange _errorRange = null;

	private Listener	_selectionListener;

	private ISelectionChangedListener	_notifier;
	/** A way to pass in a function where we can check if browsing is valid **/
	 private interface IValidityChecker {
		 public boolean isValid ();
	 }
    
	/**
	  * 
	  * @author schmerl
	  *  Implements browsing navigation by checking to see if ctrl is pressed,
	  *  checking whether mouse moves or hovers go over design analysis calls, 
	  */    
	 private class NavigationListener implements KeyListener, MouseListener, MouseTrackListener, MouseMoveListener {
        
		 /** Indicates whether this listener is activated **/
		 boolean _listenerActivated = false;
    
		 /** The current design analysis node that we are over **/    
		 DesignAnalysisCallNode _currentNode = null;
		 /** The range of the current design analysis **/
		 StyleRange _currentRange = null;
		 /** The text widget that the listener is applied to **/
		 StyledText _text = null;
		 /** The way to check if listening is valid **/
		 IValidityChecker _validCheck;
        
		 private Cursor _handCursor;
		 private Cursor _iBeam;
		 /**
		  * Constructs a Navigation listener and adds it to the list of
		  * listeners for the text.
		  * @param text The text that this is listening to
		  * @param vc The validity checker
		  */
		 public NavigationListener (StyledText text, IValidityChecker vc) {
			 _handCursor = new Cursor (null, SWT.CURSOR_HAND);
			 _iBeam = new Cursor (null, SWT.CURSOR_IBEAM);
			 _validCheck = vc;
			 _text = text;
			 _text.addMouseMoveListener(this);
			 _text.addMouseListener (this);
			 _text.addKeyListener(this);
			 _text.addMouseTrackListener(this);
		   }
        
		 /**
		  * Checks whether browsing is valid. If it's no longer valid
		  * make sure we unset regions and deactivate the listener
		  *
		  */
		 private void checkValid () {
			 if (!_validCheck.isValid ()) {
				 _listenerActivated = false; 
				 removeCurrentStyle ();
			 }
		 }

		 /**
		  * Activates the listener when cntrl is pressed
		  */
		 public void keyPressed(KeyEvent e) {
			 if (e.keyCode == SWT.CONTROL) {
				 _listenerActivated = true;
			 }
			 checkValid ();
		 }
        

		 /**
		  * Deactivates the listener when cntrl is pressed
		  */
		 public void keyReleased(KeyEvent e) {
			 if (e.keyCode == SWT.CONTROL) {
				 _listenerActivated = false;
				 removeCurrentStyle();
			 }
		 }

		 /* (non-Javadoc)
		  * @see org.eclipse.swt.events.MouseListener#mouseDoubleClick(org.eclipse.swt.events.MouseEvent)
		  */
		 public void mouseDoubleClick(MouseEvent e) {}

		 /**
		  * Checks to see if the offset is in the current range
		  * @param offset The offset to check
		  * @return true if the offset is in the current range
		  */
		 private boolean inCurrentRange (int offset) {
			 if (_currentRange != null) {
				 if (_currentRange.start <= offset && _currentRange.start + _currentRange.length >= offset) {
					 return true;
				 }
			 }
			 return false;
		 }

		 /**
		  * Browses to the appropriate design analysis for the current call
		  */
		 public void mouseDown(MouseEvent e) {
			 try {
				 if (_errorRange != null) {
					 _errorRange.background = ColorConstants.white;
					 _browseText.getTextWidget ().setStyleRange(_errorRange);
					 _errorRange = null;
				 }
				 Point mPoint = new Point (e.x, e.y);
				 int offset = _text.getOffsetAtLocation(mPoint);
				 if (inCurrentRange (offset)) {
					 Scope scope = (Scope )_browseList.get (currentPosition);
					 Object o = scope.lookupName (_currentNode.getIdentifier (), true);
					 _listenerActivated = false;
					 if (o instanceof DesignAnalysis) {
						 _text.setCursor (_iBeam);
						 navigateTo ((DesignAnalysis )o);    
					 }
                    
                   
				 }
			 }
			 catch (IllegalArgumentException exc) {
				 removeCurrentStyle();
			 }
    
		 }

		 /* (non-Javadoc)
		  * @see org.eclipse.swt.events.MouseListener#mouseUp(org.eclipse.swt.events.MouseEvent)
		  */
		 public void mouseUp(MouseEvent e) {}

		 /* (non-Javadoc)
		  * @see org.eclipse.swt.events.MouseTrackListener#mouseEnter(org.eclipse.swt.events.MouseEvent)
		  */
		 public void mouseEnter(MouseEvent e) {}

		 /* (non-Javadoc)
		  * @see org.eclipse.swt.events.MouseTrackListener#mouseExit(org.eclipse.swt.events.MouseEvent)
		  */
		 public void mouseExit(MouseEvent e) {}

		/**
		 * Highlights a design call node if the mouse is hovered over one
		 */
		 public void mouseHover(MouseEvent e) {
			 checkValid ();
			 if (_listenerActivated) {
				 try {
					 int offset = _text.getOffsetAtLocation(new Point(e.x, e.y));
					 AcmeObject o = _regionManager.getAcmeObjectAtPos(offset);
					 if (o instanceof DesignAnalysisCallNode) {
						 removeCurrentStyle();
						 _currentNode = (DesignAnalysisCallNode) o;
						 Region dacnR = _regionManager.getRegion(o);
						 StyleRange style = new StyleRange();
						 style.foreground = ColorConstants.blue;

						 style.start = dacnR.getBeginIndex();
						 style.length = dacnR.getEndIndex() - dacnR.getBeginIndex();
						 _currentRange = style;
						 _text.setStyleRange(style);
						 _text.setCursor (_handCursor);
                        

					 }
					 else {
						 removeCurrentStyle();
					 }
				 }
				 catch (IllegalArgumentException exc) {
					 removeCurrentStyle();
				 }
			 }
		 }
        
		 /**
		  * Removes the current style (unhighlighting the region)
		  *
		  */
		 private void removeCurrentStyle () {
			 if (_currentRange != null) {
				 try {
                
				 _currentRange.foreground = ColorConstants.black;
                
				 _text.setStyleRange(_currentRange);
				 _text.setCursor (_iBeam);
				 _currentNode = null;
				 _currentRange = null;
				 }
				 catch (IllegalArgumentException e) {}
			 }
		 }

		 /**
		  * Highlights a design analysis call when the mouse move
		  * over one.
		  */
		 public void mouseMove(MouseEvent e) {
			 checkValid ();
			 if (_listenerActivated) {
				 Point mPoint = new Point(e.x, e.y);
				 try {

					 int offset = _text.getOffsetAtLocation(mPoint);
					 if (inCurrentRange (offset)) {
						 return;
					 }
					 AcmeObject o = _regionManager.getAcmeObjectAtPos(offset);
					 if (o instanceof DesignAnalysisCallNode) {
						 removeCurrentStyle();
						 _currentNode = (DesignAnalysisCallNode) o;
						 Region dacnR = _regionManager.getRegion(o);
						 StyleRange style = new StyleRange();
						 style.foreground = ColorConstants.blue;

						 style.start = dacnR.getBeginIndex();
						 style.length = dacnR.getEndIndex() - dacnR.getBeginIndex();
						 _currentRange = style;
						 _text.setStyleRange(style);
						 _text.setCursor (_handCursor);

					 }
					 else {
						 removeCurrentStyle();
					 }
				 }
				 catch (IllegalArgumentException exc) {
					 // Position is not over text
					 removeCurrentStyle();

				 }
			 }
		 }

 }
 
    /**
	 * @return
	 */
	public ArmaniExportable getCurrentlyDisplayed() {
		return currentlyDisplayed;
	}

	/**
	 * @return
	 */
	public Composite getButtonArea() {
		return _buttonArea;
	}


    /**
     * Constructs the design rule browser, with a text editing region
     * and back and forward buttons to browse
     * @param parent Where this is going to go
     */
    public DesignRuleBrowser (Composite parent) {
        super (parent, SWT.NONE);
        
//        FormLayout layout = new FormLayout ();
        
        GridLayout layout = new GridLayout (1, false);
        this.setLayout (layout);
        
        // Construct the StyledText
        _browseText = new SourceViewer (this, null, SWT.BORDER | SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL);
        _browseText.configure (new AcmeSourceViewerConfiguration ());
        
        GridData browseData = new GridData (GridData.FILL_HORIZONTAL | GridData.GRAB_VERTICAL | GridData.FILL_VERTICAL);
//        browseData.verticalAlignment = GridData.VERTICAL_ALIGN_FILL;
//        browseData.grabExcessHorizontalSpace = true;
//        browseData.grabExcessVerticalSpace = GridData.GRAB_VERTICAL;
//        browseData.widthHint = 500;
//        browseData.heightHint = 100;
        _browseText.getTextWidget ().setLayoutData (browseData);

        // Construct the button area
        _buttonArea = new Composite (this, SWT.NONE);
        GridData baData = new GridData (GridData.HORIZONTAL_ALIGN_CENTER);
//        baData.horizontalAlignment = GridData.HORIZONTAL_ALIGN_CENTER;
        _buttonArea.setLayoutData(baData);
        _buttonArea.setLayout (new FillLayout (SWT.HORIZONTAL));
        
        
        
        // Construct the back button
        _backButton = new Button (_buttonArea, SWT.PUSH);
//        _backButton.setText ("<<");
        // The following is for 2.1 forward 
        // As of Eclipse 3.0, IMG_TOOL_BACK_HOVER == IMG_TOOL_BACK
        _backButton.setImage(PlatformUI.getWorkbench().getSharedImages().getImage(ISharedImages.IMG_TOOL_BACK));
        _backButton.setEnabled (false);
        _backButton.addSelectionListener(new SelectionListener () {
            public void widgetSelected (SelectionEvent evt) {
                navigateBack ();
            }
          
            public void widgetDefaultSelected (SelectionEvent evt) {
            }
        });
        
        // Construct the forward button
        _forwardButton = new Button (_buttonArea, SWT.PUSH);
//        _forwardButton.setText (">>");       
        // The following is for 2.1 forward
        // As of Eclipse 3.0, IMG_TOOL_FORWARD_HOVER == IMG_TOOL_FORWARD
        _forwardButton.setImage(PlatformUI.getWorkbench().getSharedImages().getImage(ISharedImages.IMG_TOOL_FORWARD));
        _forwardButton.setEnabled (false);
        _forwardButton.addSelectionListener (new SelectionListener () {
            public void widgetSelected (SelectionEvent evt) {
                navigateForward ();
            }
          
            public void widgetDefaultSelected (SelectionEvent evt) {
            }
        });
        
		// Create the navigation listener and add it to the text editor
	  	new NavigationListener (_browseText.getTextWidget(), new IValidityChecker () {
		  public boolean isValid () {
			  return DesignRuleBrowser.this._validRegions;
		  }
	  });
    }

    /**
     * Navigates forward in the browse history
     */
    protected void navigateForward() {
        currentPosition++;
        ArmaniExportable o = (ArmaniExportable )_browseList.get (currentPosition);
        
        navigateTo (o, false);
        _forwardButton.setEnabled (currentPosition != _browseList.size () - 1);
        _backButton.setEnabled (currentPosition != 0);                
    }

    /**
     * Navigates backward in the browse history
     */
    protected void navigateBack() {
        currentPosition--;
        ArmaniExportable o = (ArmaniExportable )_browseList.get (currentPosition);
        navigateTo (o, false);
        _backButton.setEnabled (currentPosition != 0);                
        _forwardButton.setEnabled (currentPosition != _browseList.size () - 1);
    }
    
    /**
     * Navigates to a particular object 
     * @param o The object to put in the text editor
     * @param updateNavigation Whether this updates the browse history (false if we are called as a result of browsing)
     */
    protected void navigateTo (ArmaniExportable o, boolean updateNavigation) {
        try {
            // Output the object
            AcmeOutputStream os = new AcmeOutputStream ();
            if (o instanceof DesignRule) {
                ((DesignRule )o).exportArmani (os, false);
            }
            else {
                o.exportArmani(os);
               
             }
            os.close ();
           currentlyDisplayed = o;
            // Update the region manager and text
            _regionManager = new TextRegionManager (os.getRegionManager ());
            _browseText.setDocument(new Document (os.toString ()));
//            _browseText.setText (os.toString ());
//            _browseText.setFocus();
            _validRegions = true;
            
            if (updateNavigation) {
                // Erase the forward history
                currentPosition++;
                if (_browseList.size () > 0)
                    for (int i = currentPosition; i < _browseList.size (); ) {
                        _browseList.remove (i);  
                    
                    }
                _browseList.add (o);
                _backButton.setEnabled (currentPosition > 0);
                _forwardButton.setEnabled (currentPosition < _browseList.size () -1 );
            }
            fireSelectionChanged();
        }
        catch (IOException e) {
            _browseText.setDocument (new Document (o.toString ()));
        }
        catch (AcmeException e) {
			_browseText.setDocument (new Document (o.toString ()));
        }
                
    }
    
    /**
     * Navigates to a particular element in the browser
     * @param o The object to navigate to
     */
    public void navigateTo (ArmaniExportable o) {
        navigateTo (o, true);
    }
    
    
    /**
     * Sets whether it is possible to edit the text
     * @param editable
     */
    public void setEditable(boolean editable) {
        _browseText.setEditable(editable);
    }


    public StyledText getTextWidget () {return _browseText.getTextWidget ();}

    /**
     * Gets the text in the text editor
     * @return java.lang.String
     */
    public String getText() {
        return _browseText.getDocument ().get ();
    }

    /**
     * Returns true if there are valid regions (i.e., the text hasn't been edited)
     * @return boolean
     */
    public boolean hasValidRegions() {
        return _validRegions;
    }

    /**
     * Sets the _validRegions.
     * @param _validRegions The _validRegions to set
     */
    public void setValidRegions(boolean _validRegions) {
        this._validRegions = _validRegions;
    }


    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.ISelectionProvider#addSelectionChangedListener(org.eclipse.jface.viewers.ISelectionChangedListener)
     */
    public void addSelectionChangedListener(ISelectionChangedListener listener) {
        _selectionListeners.add (listener);
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.ISelectionProvider#getSelection()
     */
    public ISelection getSelection() {
        if (currentPosition != -1) {
            return new StructuredSelection (_browseList.get (currentPosition));
        }
        return null;
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.ISelectionProvider#removeSelectionChangedListener(org.eclipse.jface.viewers.ISelectionChangedListener)
     */
    public void removeSelectionChangedListener(ISelectionChangedListener listener) {
        _selectionListeners.remove (listener);
    }

    /* (non-Javadoc)
     * @see org.eclipse.jface.viewers.ISelectionProvider#setSelection(org.eclipse.jface.viewers.ISelection)
     */
    public void setSelection(ISelection selection) {
        if (selection instanceof IStructuredSelection) {
            Object sel = ((IStructuredSelection )selection).getFirstElement();
            if (_browseList.contains (sel)) {
                currentPosition = _browseList.indexOf (sel);
                navigateTo ((ArmaniExportable )sel, false);
            }
        }
    }

    protected void fireSelectionChanged () {
        if (currentPosition == -1) return;
        Iterator i = _selectionListeners.iterator();
        SelectionChangedEvent ce = new SelectionChangedEvent (this, new StructuredSelection (_browseList.get (currentPosition)));
        while (i.hasNext ()) {
            ISelectionChangedListener listener = (ISelectionChangedListener )i.next ();
            listener.selectionChanged(ce);
        }
    }
    
    public void setErrorHighlight (AcmeObject o) {
        if (_errorRange != null) {
            _errorRange.background = ColorConstants.white;
            try {
            
            _browseText.getTextWidget ().setStyleRange(_errorRange);
            }
            catch (IllegalArgumentException e) {}
        }
        
        Region r = _regionManager.getRegion (o);
        if (r != null) {
            _errorRange = new StyleRange ();
            _errorRange.background = ColorConstants.red;
            _errorRange.start = r.getBeginIndex ();
            _errorRange.length = r.getEndIndex () - r.getBeginIndex ();
            try {
            
            _browseText.getTextWidget ().setStyleRange (_errorRange);
            }
            catch (IllegalArgumentException e) {}
        }
    }
    
    public void setSelectionHighlight (boolean enable, final ISelectionChangedListener notifier) {
		_notifier = notifier;
    	if (!enable) {
    		if (_selectionListener != null) {
    			_browseText.getTextWidget().removeListener(SWT.KeyUp, _selectionListener);
    			_browseText.getTextWidget().removeListener(SWT.MouseUp, _selectionListener);
    		}
    	}
    	else {
    		if (_selectionListener == null) {
    			_selectionListener = new Listener  () {
    				
					public void handleEvent(Event event) {
						modifySelection();
					}
    			};
    		}
    		_browseText.getTextWidget ().addListener(SWT.KeyUp,_selectionListener);
    		_browseText.getTextWidget().addListener (SWT.MouseUp, _selectionListener);
    	}
    }

	/**
	 * 
	 */
	private void modifySelection() {
		if (_selectionRange != null) {
			_selectionRange.background = ColorConstants.white;
			try {
				_browseText.getTextWidget().setStyleRange(_selectionRange);
			}
			catch (IllegalArgumentException ex) {}
		}
		if (_validRegions) {
			AcmeObject o = _regionManager.getAcmeObjectAtPos(_browseText.getTextWidget ().getCaretOffset());
			if (o != null) {
				Region r = _regionManager.getRegion (o);
				_selectionRange = new StyleRange ();
				_selectionRange.background = ColorConstants.tooltipBackground;
				_selectionRange.start = r.getBeginIndex();
				_selectionRange.length = r.getEndIndex() - r.getBeginIndex();
				try {
					_browseText.getTextWidget().setStyleRange(_selectionRange);
				}
				catch (IllegalArgumentException ex) {}
				_notifier.selectionChanged(new SelectionChangedEvent (DesignRuleBrowser.this, new StructuredSelection (o)));
			}
		}
	}

	StyleRange	_selectionRange = null;
}
