// $Header$
package AcmeStudio.ui.views.parts;

import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import AcmeStudio.ui.widgets.*;
import AcmeStudio.util.DesignRuleHolder;
import AcmeStudio.util.Utils;
import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.designRules.AcmeResolutionException;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.designRules.DesignRule;
import acmeLib.elements.AcmeElement;
import acmeLib.errorManagement.AcmeError;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * @author Bradley Schmerl
 * Implements modification of rules and associated annotations. The user can edit
 * the rule text, labels to show instead of the rules, and browse any errors.
 */
public class ModifyRuleDialog extends Dialog implements ISelectionChangedListener {
    
    /** A check button that indicates whether the rule is to be evaluated
     * when the system changes.
     */
    private Button _btnEvaluateRule;

    /** Indicates whether the rule can be modified **/	
    private boolean _canModifyRule = true;
	
	/** The holder for the rule **/
	private DesignRuleHolder _designRule;
	
	/** The element that will host the new rule **/
	private AcmeElement _element;
    
    /** The text containing the label to print if the rule evaluates to false **/
    private Text _errMsg;
	
	/** The text editor containing any parse errors **/
	private AcmeErrorTree _errorText;
  
    /** The text containing the label to print instead of the Armani rule **/  
    private Text _label;

	/** The parent where this dialog is to be displayed **/
	private Composite _parent;
	
	/** The button to parse the rule text **/
	private Button _parseRuleButton;
    
    /** The text editor for editing the rule **/
	private DesignRuleBrowser _ruleText;
    
	/**
	 * Constructor for ModifyRuleDialog.
	 * @param parentShell
	 */
	public ModifyRuleDialog(Shell parentShell, AcmeElement element) {
		super(parentShell);
		_element = element;
	}

	/**
	 * Creates the OK and Cancel button and button area of the dialog
	 * If the rule that is being modified by the dialog has pre-existing
	 * errors, then disable the OK button.
	 * @see org.eclipse.jface.dialogs.Dialog#createButtonBar(Composite)
	 */
	protected Control createButtonBar(Composite parent) {
		Control buttonBar = super.createButtonBar(parent);
		if (_designRule != null && _designRule.hasErrors ()) {
            this.getButton (IDialogConstants.OK_ID).setEnabled (false);
		}
		return buttonBar;
	}


	/**
	 * Creates the rule dialog, which has two text editors and a 
	 * parse button (in addition to OK and Cancel buttons)
	 * @param parent the dialog's parent
	 * @return the Control for the dialog
	 * @see org.eclipse.jface.dialogs.Dialog#createDialogArea(Composite)
	 */
	protected Control createDialogArea(Composite parent) {
		_parent = (Composite) super.createDialogArea(parent);

		// Make sure that the dialog is big enough to display everything
		_parent.setSize(500, 500);
		GridLayout gl = new GridLayout(1, false);
		_parent.setLayout(gl);

		Label ruleLabel = new Label(_parent, SWT.LEFT);
		ruleLabel.setText(Messages.getString("ModifyRuleDialog.strDesignRule")); //$NON-NLS-1$
		ruleLabel.setLayoutData(new GridData());

		// Create the text editor for the rule		
		_ruleText = new DesignRuleBrowser (_parent); //new StyledText (_parent, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL);
		// Initialize the text
		if (_designRule != null) {
			if (_designRule.getDesignRule () != null) {
                setRuleText(true);

			}
			else if (_designRule.getUnparsedText() != null) {
				_ruleText.navigateTo (_designRule.getDesignRule());
			}
			else {
				
				
				_ruleText.navigateTo (Utils.EMPTY_RULE);
			}
		}
		
		GridData ruleData = new GridData (GridData.FILL_HORIZONTAL);// | GridData.GRAB_HORIZONTAL | GridData.GRAB_VERTICAL);
//		ruleData.horizontalSpan = 2;
		ruleData.heightHint = 150;
		ruleData.widthHint = 500;
		_ruleText.setLayoutData (ruleData);
		
        final ModifyListener ruleModified = new ModifyListener () {
            public void modifyText(ModifyEvent e){
                if (_canModifyRule) {
                    ModifyRuleDialog.this.getButton(IDialogConstants.OK_ID).setEnabled (false);
                      
                    _ruleText.setValidRegions(false);
                }
            }
        };
        
        
        
//		 If the rule is modified, it must first be parsed before
//		 it can be added, so disable the OK button (it will be 
//		 re-enabled if the rule parses correctly
        _ruleText.getTextWidget ().addFocusListener (new FocusListener () {
            
            public void focusGained (FocusEvent e) {
                ModifyRuleDialog.this._ruleText.getTextWidget ().addModifyListener (ruleModified);
            }
            public void focusLost (FocusEvent e) {
                ModifyRuleDialog.this._ruleText.getTextWidget().removeModifyListener(ruleModified);
                
            }        });

		
        _ruleText.addSelectionChangedListener(this);
        

//        TabFolder errorFolder = new TabFolder (_parent, SWT.NONE);
		Label errorLabel = new Label (_parent, SWT.LEFT);
		errorLabel.setText (Messages.getString("ModifyRuleDialog.strParseErrors")); //$NON-NLS-1$
		GridData errorData = new GridData ();
//		errorData.horizontalSpan = 2;
		errorLabel.setLayoutData (errorData);
		
		// Define the error text
		_errorText = new AcmeErrorTree (_parent);
		GridData etData = new GridData (GridData.FILL_HORIZONTAL);
//		etData.horizontalSpan = 2;
		etData.heightHint = 100;
		etData.widthHint = 300;
		_errorText.getTree ().setLayoutData(etData);
        
        _errorText.getTree ().addSelectionListener(new SelectionAdapter () {
            public void widgetSelected (SelectionEvent evt) {
                if (evt.item instanceof TreeItem) {
                    AcmeError e = (AcmeError )((TreeItem )evt.item).getData();
                    Object source = e.getSource ();
                    _ruleText.setErrorHighlight ((AcmeObject )source);
                }
            }
        });
		
		// If the rule already has errors, initialize the text		
		if (_designRule != null && _designRule.hasErrors()) {
			_errorText.setErrorSource(_designRule.getError());

		}
     
        // The button to set whether to evaluate the rule incrementally   
        _btnEvaluateRule = new Button (_parent, SWT.CHECK);
        _btnEvaluateRule.setText (Messages.getString("ModifyRuleDialog.strEvaluateIncrementally")); //$NON-NLS-1$
        if (_designRule != null) {
            _btnEvaluateRule.setSelection (_designRule.shouldEvaluate());
            
        }
        else {
            _btnEvaluateRule.setSelection (true);
        }
		
        // The label elements
        Label labelLabel = new Label (_parent, SWT.LEFT);
        labelLabel.setText (Messages.getString("ModifyRuleDialog.strRuleLabel")); //$NON-NLS-1$
        
        _label = new Text (_parent, SWT.MULTI | SWT.WRAP | SWT.V_SCROLL);
        if (_designRule != null && _designRule.hasLabel ()) {
            _label.setText (_designRule.getLabel());
        }
        else {
            _label.setText (""); //$NON-NLS-1$
        }
        GridData lData = new GridData (GridData.FILL_BOTH);
        lData.heightHint = 50;
        lData.widthHint = 300;
        _label.setLayoutData(lData);
        
        // The error messages
        Label errMsgLabel = new Label (_parent, SWT.NONE);
        errMsgLabel.setText (Messages.getString("ModifyRuleDialog.strRuleErrorMessage")); //$NON-NLS-1$
        
        _errMsg = new Text (_parent, SWT.MULTI | SWT.WRAP | SWT.V_SCROLL);
        if (_designRule != null) {
            _errMsg.setText (_designRule.getErrorMessage());
        }
        else {
            _errMsg.setText (""); //$NON-NLS-1$
        }
        GridData emData = new GridData (GridData.FILL_BOTH);
        emData.heightHint = 50;
        emData.widthHint = 300;
        _errMsg.setLayoutData(emData);
        _ruleText.setEditable(_canModifyRule);
		return _parent;

	}

    /**
     * Disables editing of the rule
     */
    public void disableEditing () {
        _canModifyRule = false;
        if (_ruleText != null) _ruleText.setEditable(false);
    }
    
	/**
	 * Returns the _designRule.
	 * @return DesignRule
	 */
	public DesignRuleHolder getDesignRule() {
		return _designRule;
	}

    /* (non-Javadoc)
     * @see org.eclipse.jface.dialogs.Dialog#okPressed()
     */
    protected void okPressed() {
        _designRule.setEvaluation (_btnEvaluateRule.getSelection());
        _designRule.setErrorMessage(_errMsg.getText());
        _designRule.setLabel (_label.getText ());
        super.okPressed();
    }

	/**
	 * Parses the rule text and tries to parse the text
	 */
	private void parseRule () {
		String ruleToParse = _ruleText.getText ();
		if (_designRule == null) {
			_designRule = new DesignRuleHolder ();
		}
		
		// try to parse the rule
		_designRule.parseRule(ruleToParse);
		
              
		// If it has errors, display the errors
		if (_designRule.hasErrors ()) {
			_errorText.setErrorSource (_designRule.getError ());	
		}
		else {
			// If no errors, then clear the previous errors and
			// enable to OK button
            DesignRule dr = _designRule.getDesignRule ();
            dr.setDesignRuleScope (_element);
            try {
                dr.resolveTypeReferences();    
                dr.resolveInstanceReferences ();
            }
            catch (AcmeResolutionException e) {
                setRuleText (false);
                AcmeVector errEnum = new AcmeVector ();
                errEnum.addItem(e.getError ());
                _errorText.setErrorSource (errEnum.enumerate());
                AcmeTypeManager.typecheckingErrorStack.disconnect ();
            }
            if (dr.typecheck ()) {
			     _errorText.setErrorSource (null);
			     getButton(IDialogConstants.OK_ID).setEnabled (true);
                 setRuleText(false);
            }
            else {
                setRuleText (false);
                AcmeVector errors = AcmeTypeManager.typecheckingErrorStack.getCausalErrors();
                AcmeEnumeration errEnum = errors.enumerate();
                _errorText.setErrorSource (errEnum);
                AcmeTypeManager.typecheckingErrorStack.disconnect();
            }                
		}
	}


    /**
     * Called when browsing occurs in the rule text editor. If we are browsing away from the
     * original rule, then we have to disable the parse button. When moving back to the rule,
     * enable the parse button and make the rule editor editable (if it was originally)
     * @see org.eclipse.jface.viewers.ISelectionChangedListener#selectionChanged(org.eclipse.jface.viewers.SelectionChangedEvent)
     */
    public void selectionChanged(SelectionChangedEvent event) {
        if (event.getSource() != this) {
            ISelection selection = event.getSelection ();
            if (selection instanceof IStructuredSelection) {
                Object sel = ((IStructuredSelection )selection).getFirstElement ();
                if (sel instanceof DesignAnalysis) {
                    // turn off editing and the parse button
                    _ruleText.setEditable (false);
                    _parseRuleButton.setEnabled(false);
                }
                else if (sel instanceof DesignRule) {
                    if (sel == _designRule.getDesignRule()) {
                        _ruleText.setEditable (_canModifyRule);
                        _parseRuleButton.setEnabled (true);
                    }
                }
            }
        }
    }
	
	/**
	 * Sets the _designRule.
	 * @param _designRule The _designRule to set
	 */
	public void setDesignRule(DesignRuleHolder _designRule) {
		this._designRule = _designRule;
	}

    /**
     * Sets the rule text
     * @param navigate Indicates whether we are navigating or not
     */
    protected void setRuleText(boolean navigate) {
        _ruleText.navigateTo (_designRule.getDesignRule (), navigate);
        
    }

	/* (non-Javadoc)
	 * @see org.eclipse.jface.dialogs.Dialog#createButtonsForButtonBar(org.eclipse.swt.widgets.Composite)
	 */
	protected void createButtonsForButtonBar(Composite parent) {
		// BugFix#237: Moved Parse Rule button to the button bar
		createButton (parent, 2, "Parse Rule", false);
		_parseRuleButton = this.getButton(2);
		_parseRuleButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				parseRule();
			}
		});
		super.createButtonsForButtonBar (parent);
	}
}
