// $Header$
package AcmeStudio.ui.views.parts;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Set;
import org.eclipse.gef.ui.actions.ActionRegistry;
import org.eclipse.jface.action.*;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.*;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.internal.WorkbenchImages;
import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASContext;
import AcmeStudio.core.ASModel;
import AcmeStudio.ui.editors.diagram.actions.HighlightObjectsInDiagramAction;
import AcmeStudio.ui.editors.diagram.actions.UnhighlightAllAction;
import AcmeStudio.ui.views.IElementViewPage;
import AcmeStudio.ui.views.ruleErrors.AcmeElementTree;
import AcmeStudio.ui.views.ruleErrors.RuleErrorDetailer;
import AcmeStudio.util.*;
import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.designRules.DesignRule;
import acmeLib.designRules.DesignRuleEvalException;
import acmeLib.designRules.Scope;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.properties.AcmeBoolean;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.BooleanType;

/**
 * @author Bradley Schmerl (schmerl@cs.cmu.edu) Implements a table view
 *         containing a list of the armani rules associated with an element.
 *         The view shows those rules that check successfully, and also if they
 *         are inherited from a super type.
 */
public class RuleList extends ElementViewPart implements PropertyChangeListener {

	// Constants for Column data
	private static final int RULECOL = 2;
	private static final int INHERITCOL = 0;
	private static final int EVALCOL = 1;
	private static final String[] PROPERTIES = { "inherit", "evaluate", "rule" };
	private static final Object[] emptyRules = new Object[0];

	/** The viewer used to display the rules */
	private TableViewer _viewer = null;
	public static int DELRULE = 1;
	public static int NEWRULE = 0;

	/** The object whose rules are being viewed */
	private AcmeObject _selectedObject = null;

	/** The rule that is currently selected */
	private DesignRuleHolder _selectedRule = null;

	/** A pointer to the ElementView containing this view */
	private IElementViewPage _parentView = null;

	// Actions to add and delete rules
	private Action _addRuleAction;
	private ViewSelectionAction _deleteRuleAction;
	private ViewSelectionAction _forceEvaluationAction;
	private Action _doubleClickAction;
	private ViewSelectionAction _enableEvaluationAction;
	private ViewSelectionAction _evaluateForElementAction;
	private ViewSelectionAction _browseRuleErrorsAction;
	private ViewSelectionAction _highlightRuleErrorsAction;
	private DesignEvaluationCache _evalCache;
	private static final String _evaluateForElementTextPrefix = "Enable all rules for ";
	// The list of contributions that this makes to the toolbar
	private IContributionItem[] _myContributions = new IContributionItem[0];
	private RuleSorter	_sorter;

	/*
	 * The content provider class is responsible for providing objects to the
	 * view. It returns all the rules to be displayed in the viewer.
	 */
	class ViewContentProvider implements IStructuredContentProvider {
		public void inputChanged(Viewer v, Object oldInput, Object newInput) {
		}
		public void dispose() {
		}


		/**
		 * Returns all the rules for the given object
		 * 
		 * @param parent
		 *           the Element to return the rules for
		 * @return the rules associated with this element
		 */
		public Object[] getElements(Object parent) {

			Object[] rules = RuleList.emptyRules;
			if (parent instanceof AcmeElement) {
				AcmeElement el = (AcmeElement) parent;
				AcmeEnumeration invariants = el.getInvariants(true).castToAcmeSetVector().enumerate();
				AcmeEnumeration heuristics = el.getHeuristics(true).castToAcmeSetVector().enumerate();

				// The children are a combination of invariants and heuristics
				// add all of them
				rules = new Object[invariants.getSize() + heuristics.getSize()];
				int index = 0;
				while (invariants.moreItems()) {
					DesignRuleHolder holder = new DesignRuleHolder();
					holder.setDesignRule((DesignRule) invariants.getNextItem());
					holder.setScope(holder.getDesignRule().getDesignRuleScope());
					holder.setEvalScope((Scope) _selectedObject);
					rules[index++] = holder;
				}

				while (heuristics.moreItems()) {
					DesignRuleHolder holder = new DesignRuleHolder();
					holder.setDesignRule((DesignRule) heuristics.getNextItem());
					holder.setScope(holder.getDesignRule().getDesignRuleScope());
					holder.setEvalScope((Scope) _selectedObject);
					rules[index++] = holder;
				}
			}
			return rules;
		}
	}

	/**
	 * Provides routines for getting the text in each of the columns of the
	 * table Column 1 = inheritance; Column 2 = evaluation result; Column 3 =
	 * the rule
	 */
	class ViewLabelProvider extends LabelProvider implements ITableLabelProvider, IFontProvider, IColorProvider {
		private Font	_italicFont;
		public String getColumnText(Object obj, int index) {
			DesignRuleHolder rule = (DesignRuleHolder) obj;

			String retStr = "";
			switch (index) {

				case RULECOL :
					{
						retStr = rule.getLabel();
						int satisfied = DesignEvaluationCache.TRUE;
						// Set the scope of the rule to be the selected element
						satisfied = rule.evaluate((Scope) RuleList.this._selectedObject);
						if (satisfied == DesignEvaluationCache.FALSE) {
							retStr = rule.getErrorMessage();
							if (retStr == null || retStr.equals("")) {
								retStr = rule.getLabel();
							}
							break;

						}
					}
					//                case EVALCOL :
					//                    if (rule.evaluate ((Scope
					// )RuleList.this._selectedObject) ==
					// DesignEvaluationCache.UNKNOWN) {
					//                        retStr = "?";
					//                    }
					//                    break;
			}

			return retStr;

		}
		public Image getColumnImage(Object obj, int index) {
			Image icon = null;
			DesignRule rule = ((DesignRuleHolder) obj).getDesignRule();
			DesignRuleHolder holder = (DesignRuleHolder) obj;
			switch (index) {
				case INHERITCOL :
					{
						if (holder.isInherited()) {
							icon = ImageManager.getIcon(ImageManager.ID_INHERITS);
						}

						break;
					}
				case EVALCOL :
					{
						icon = holder.getEvaluationImage();
						if (!getElementEvaluation())
							return ImageManager.getIcon(ImageManager.ID_CONSTRAINT_NOT_RUNNING);
						if (icon == null) {
							// BugFix#xxx: Don't evaluate constraints on types
							if (_selectedObject instanceof AcmeElementType) {
								int satisfied = holder.typecheck((Scope )RuleList.this._selectedObject);
								switch (satisfied) {
									case DesignEvaluationCache.ERROR:
										icon = ImageManager.getIcon (ImageManager.ID_TYPE_ERROR);
										break;
									case DesignEvaluationCache.TRUE:
										icon = null;
										break;
										
								}
							}
							else {
								int satisfied = DesignEvaluationCache.TRUE;
								// Set the scope of the rule to be the selected
								// element
								satisfied = holder.evaluate((Scope) RuleList.this._selectedObject);
								if (satisfied == DesignEvaluationCache.FALSE && rule.getIsInvariant()) {
									icon = ImageManager.getIcon(ImageManager.ID_ERROR);
								}
								else if (satisfied == DesignEvaluationCache.FALSE && !rule.getIsInvariant()) {
									icon = ImageManager.getIcon(ImageManager.ID_WARNING);
								}
								else if (satisfied == DesignEvaluationCache.ERROR) {
									icon = ImageManager.getIcon(ImageManager.ID_TYPE_ERROR);
								}
								else if (satisfied == DesignEvaluationCache.UNKNOWN) {
									icon = ImageManager.getIcon(ImageManager.ID_CONSTRAINT_RUNNING);
								}
								else if (satisfied == DesignEvaluationCache.TRUE) {
									icon = ImageManager.getIcon(ImageManager.ID_OK);
								}
							}
						}
					}
			}
			return icon;
		}
		public Image getImage(Object obj) {
			return PlatformUI.getWorkbench().getSharedImages().getImage(ISharedImages.IMG_OBJ_ELEMENT);
		}
		/**
		 * @param element
		 * @return
		 * @see org.eclipse.jface.viewers.IFontProvider#getFont(java.lang.Object)
		 **/
		public Font getFont(Object element) {
			if (_italicFont == null) {
				Font originalFont = _viewer.getControl().getFont();
				FontData fontData [] = originalFont.getFontData ();
				for (int i = 0; i < fontData.length; i++) {
					fontData [i].setStyle (fontData [i].getStyle () | SWT.ITALIC);
				}
				_italicFont = new Font (_viewer.getControl ().getDisplay(), fontData);
			}	
			DesignRuleHolder holder = (DesignRuleHolder )element;
			int satisfied = holder.evaluate ((Scope )RuleList.this._selectedObject);
			if (satisfied == DesignEvaluationCache.UNKNOWN) return _italicFont;
			return null;
		}
		/**
		 * @param element
		 * @return
		 * @see org.eclipse.jface.viewers.IColorProvider#getForeground(java.lang.Object)
		 **/
		public Color getForeground(Object element) {
			return null;
		}
		/**
		 * @param element
		 * @return
		 * @see org.eclipse.jface.viewers.IColorProvider#getBackground(java.lang.Object)
		 **/
		public Color getBackground(Object element) {
			DesignRuleHolder holder = (DesignRuleHolder )element;
			int satisfied = holder.evaluate((Scope )RuleList.this._selectedObject);
			if (satisfied == DesignEvaluationCache.FALSE && holder.getDesignRule().getIsInvariant()) 
				return Constants.DEFAULT_ERROR_BACKGROUND;
			else if (satisfied == DesignEvaluationCache.FALSE && !holder.getDesignRule().getIsInvariant()) 
				return Constants.DEFAULT_WARNING_BACKGROUND;
			else if (satisfied == DesignEvaluationCache.ERROR) {
				return Constants.DEFAULT_ERROR_BACKGROUND;
			}
			return null;
		}
		
		public void dispose () {
			if (_italicFont != null) _italicFont.dispose ();
		}
	}

	class RuleSorter extends ViewerSorter {

		private static final int CAT_INV = 1;
		private static final int CAT_HEUR = 0;
		private static final int CAT_UNKNOWN = -1;
		
		private static final int SORT_NONE = 0;
		private static final int SORT_A_Z = 3;
		private static final int SORT_Z_A = 4;
		private static final int SORT_CAT_UP = 1;
		private static final int SORT_CAT_DOWN = 2;

		static final int SORT_BY_NAME = 0;
		static final int SORT_BY_EVALUATION = 1;
		private int	_state = SORT_A_Z;
		private int	_category;
		/**
		 * @see org.eclipse.jface.viewers.ViewerSorter#category(Object) */
		public int category(Object element) {
			DesignRuleHolder holder = (DesignRuleHolder) element;
			if (holder.getDesignRule() == null)
				return CAT_UNKNOWN;
			if (holder.getDesignRule().getIsInvariant())
				return CAT_INV;
			else
				return CAT_HEUR;

		}

		/**
		 * @see org.eclipse.jface.viewers.ViewerSorter#compare(Viewer, Object,
		 *      Object)
		 */
		public int compare(Viewer viewer, Object e1, Object e2) {
			if (_state == SORT_NONE) return 0;
			DesignRuleHolder h1 = (DesignRuleHolder )e1;
			DesignRuleHolder h2 = (DesignRuleHolder )e2;
			int eval1 = h1.evaluate((Scope )_selectedObject);
			int eval2 = h2.evaluate((Scope )_selectedObject);
			int evalCompare = new Integer (eval1).compareTo (new Integer (eval2));
			if (evalCompare == 0) {
				int cat1 = 0, cat2 = 0;
				cat1 = category (e1);
				cat2 = category (e2);
				evalCompare = new Integer (eval1).compareTo (new Integer (eval2));
				if (evalCompare == 0) {
					String r1 = "", r2 = "";
				
						r1 =h1.getRuleText();
						r2 = h2.getRuleText();
						evalCompare = r1.compareTo (r2);
				}
				
			}
			return evalCompare;
			
//			return new Integer (eval1).compareTo (new Integer (eval2));
//			
//			
//			if (_category == SORT_BY_NAME) {
//				if (_state == SORT_CAT_UP || _state == SORT_CAT_DOWN) {
//					int cat1 = 0, cat2 = 0;
//					if (_state == SORT_CAT_UP) {
//						cat1 = category (e1);
//						cat2 = category (e2);
//					}
//					else {
//						cat2 = category (e1);
//						cat1 = category (e2);
//					}
//					return new Integer (cat1).compareTo(new Integer (cat2));
//				}
//				else if (_state == SORT_A_Z || _state == SORT_Z_A) {
//					String r1 = "", r2 = "";
//					if (_state == SORT_A_Z) {
//						r1 =h1.getRuleText();
//						r2 = h2.getRuleText();
//					}
//					else {
//						r2 = h1.getRuleText ();
//						r1 = h2.getRuleText ();
//					}
//					return r1.compareTo(r2);
//				}
//			}
//			else {
//				int eval1 = h1.evaluate((Scope )_selectedObject);
//				int eval2 = h2.evaluate((Scope )_selectedObject);
//				return new Integer (eval1).compareTo (new Integer (eval2));
//			}
//			return 0;
		}

		public void setCategory (int cat) {
			if (cat != _category) {
				_category = cat;
				_state = SORT_NONE;
			}
		}
		
		public void cycleState () {
			if (_category == SORT_BY_NAME)
				_state = (_state + 1) % 5;
			else
				_state = (_state + 1) % 3;
		}
	}

	/**
	 * Constructor for RuleList.
	 * 
	 * @param parent
	 * @param style
	 */
	public RuleList() {
		super();

	}

	/**
	 * @see AcmeStudio.ui.views.parts.ElementViewPart#dispose() */
	public void dispose() {
		if (_evalCache != null) {
			_evalCache.removePropertyChangeListener(this);
		}
		super.dispose();
	}

	/**
	 * Gets the list of editors for cells in a column
	 * 
	 * @param table
	 *           the table to get the editors for
	 * @return an array of editors for the table
	 */
	protected CellEditor[] getCellEditors(Table table) {
		final CellEditor[] editors = new CellEditor[3];
		editors[INHERITCOL] = new TextCellEditor(table);
		editors[EVALCOL] = new TextCellEditor(table);

		// Changes to this text editor are automatically triggered
		class AutoApplyTextCellEditor extends TextCellEditor {
			public AutoApplyTextCellEditor(Composite parent) {
				super(parent);
			}
			public void fireApplyEditorValue() {
				super.fireApplyEditorValue();
			}
		}

		editors[RULECOL] = new AutoApplyTextCellEditor(table);
		// make sure that the value is updated when focus is lost
		editors[RULECOL].getControl().addFocusListener(new FocusAdapter() {
			public void focusLost(FocusEvent e) {
				((AutoApplyTextCellEditor) editors[RULECOL]).fireApplyEditorValue();
			}
		});
		return editors;
	}

	/**
	 * Adds the double click action to the viewer */
	private void hookDoubleClickAction() {
		_viewer.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent event) {
				_doubleClickAction.run();
			}
		});
	}

	/**
	 * Creates all the actions for the rule list */
	private void makeActions() {
		// Add action
		_addRuleAction = new Action() {
			public void run() {
				_selectedRule = new DesignRuleHolder();
				_selectedRule.setScope((AcmeElement) _selectedObject);
				_selectedRule.setDesignRule((DesignRule) null);
				modifyRule();
			}
		};
		_addRuleAction.setText("New Armani Rule...");
		_addRuleAction.setToolTipText("Creates a new Armani rule");
		_addRuleAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_NEW_ARMANI));

		// Delete action
		_deleteRuleAction = new ViewSelectionAction() {
			public void run() {
				deleteRule();
			}

			public boolean isEnabled() {
				return _selectedRule != null && !_selectedRule.isInherited();
			}
			//            
			//            public boolean isDynamic () {
			//                return true;
			//            }

		};
		_deleteRuleAction.setText("Delete Armani rule");
		_deleteRuleAction.setToolTipText("Deletes the selected Armani rules");
		_deleteRuleAction.setImageDescriptor(WorkbenchImages.getImageDescriptor(ISharedImages.IMG_TOOL_DELETE));
		_viewer.addSelectionChangedListener(_deleteRuleAction);

		_forceEvaluationAction = new ViewSelectionAction() {
			public void run() {
				_selectedRule = (DesignRuleHolder) ((IStructuredSelection) _viewer.getSelection()).getFirstElement();
				if (_selectedRule != null) {
					if (_selectedRule.getDesignRule() != null) {
						_selectedRule.unevaluate((Scope) _selectedObject);
						_selectedRule.evaluate((Scope) _selectedObject);
						_viewer.refresh();
					}
				}
			}

		};
		_forceEvaluationAction.setText("Evaluate Armani rule");
		_forceEvaluationAction.setToolTipText("Forces evaluation of the selected Armani rules (for debug only)");
		_viewer.addSelectionChangedListener(_forceEvaluationAction);
		
		_doubleClickAction.setText("Edit/Browse...");
		_doubleClickAction.setToolTipText("Opens the design rule browser");

		_enableEvaluationAction = new ViewSelectionAction() {
			public void run() {
				_selectedRule.setEvaluation(!this.isChecked());
				this.setChecked(_selectedRule.shouldEvaluate());
				_viewer.refresh();
			}
			public boolean isEnabled() {
				return _selectedRule != null && _selectedRule.isInherited();
			}
		};
		_enableEvaluationAction.setText("Enable Evaluation of rule");
		_enableEvaluationAction.setToolTipText("Toggles enablement of rule evaluation (for non-inherited rules only)");
		_viewer.addSelectionChangedListener(_enableEvaluationAction);
		
		_evaluateForElementAction = new ViewSelectionAction() {
			public void run() {
				setElementEvaluation(!getElementEvaluation());
				this.setChecked(getElementEvaluation());
				_viewer.refresh();
			}
			public boolean isEnabled() {
				return _selectedObject != null && !(_selectedObject instanceof AcmeElementType);
			}
		};
		_evaluateForElementAction.setText(_evaluateForElementTextPrefix);
		_evaluateForElementAction.setToolTipText(
			"Enable evaluation of rules for this element (all rules will be affected)");
		_viewer.addSelectionChangedListener(_evaluateForElementAction);
		
		_browseRuleErrorsAction = new ViewSelectionAction () {
			public void run () {
				_selectedRule = (DesignRuleHolder) ((IStructuredSelection) _viewer.getSelection()).getFirstElement();
				if (_selectedRule != null) {
					if (_selectedRule.getDesignRule() != null) {
						_selectedRule.unevaluate ((AcmeElement )_selectedObject);
						DesignRule rule = _selectedRule.getDesignRule ();
						DesignRule.setRememberVisited(true);
						Scope origScope = rule.getDesignRuleScope();
						rule.setDesignRuleScope((AcmeElement )_selectedObject);
						try {
							rule.evaluate();
						}
						catch (DesignRuleEvalException e) {
							e.printStackTrace();
						}
						rule.setDesignRuleScope(origScope);
						DesignRule.setRememberVisited(false);
						RuleEvaluationBrowserDialog dialog = new RuleEvaluationBrowserDialog (Display.getCurrent().getActiveShell (), _selectedRule, (AcmeElement )_selectedObject);
						dialog.open ();
					}
				}
			}
			
			public boolean isEnabled () {
				return _selectedRule != null && _selectedRule.evaluate( (Scope )_selectedObject) == DesignEvaluationCache.FALSE && (_selectedObject instanceof AcmeElement && !(_selectedObject instanceof AcmeElementType));
			}
		};
		_browseRuleErrorsAction.setText ("View elements causing errors");
		_viewer.addSelectionChangedListener(_browseRuleErrorsAction);
		
		_highlightRuleErrorsAction = new ViewSelectionAction () {
			public void run () {
				_selectedRule = (DesignRuleHolder) ((IStructuredSelection) _viewer.getSelection()).getFirstElement();
				if (_selectedRule == null ||  _selectedRule.evaluate ((Scope )_selectedObject) != DesignEvaluationCache.FALSE) return;
					ActionRegistry actionRegistry = ASContext.current ().getActiveEditor().getDiagramEditor().getActionRegistry();
				if (this.isChecked()) {
					AcmeElementTree elementTree = RuleErrorDetailer.getErrorDetails(_selectedRule.getDesignRule(), (AcmeElement )_selectedObject);
					Set elements = elementTree.getAllElements();
					HighlightObjectsInDiagramAction hA = (HighlightObjectsInDiagramAction )actionRegistry.getAction (HighlightObjectsInDiagramAction.ID);
					if (hA != null) {
						hA.setSelection(new StructuredSelection (elements.toArray()));
						hA.run ();
					}
//					this.setChecked (true);
				}
				else {
//					this.setChecked (false);
					IAction action = actionRegistry.getAction(UnhighlightAllAction.ID);
					if (action != null) {
						action.run ();
					}
				}
			}
			
			public int getStyle() {
				return AS_CHECK_BOX;
			}
			
			public boolean isEnabled () {
				return _selectedRule != null && _selectedRule.getDesignRule() != null &&
				RuleErrorDetailer.canDetailRule (_selectedRule.getDesignRule ()) &&
				_selectedRule.evaluate((Scope) _selectedObject) == DesignEvaluationCache.FALSE;
			}
			

		};
		_highlightRuleErrorsAction.setText ("Highlight rule errors");
		_highlightRuleErrorsAction.setToolTipText("Highlights elements causing the selected rule to fail");
		_highlightRuleErrorsAction.setImageDescriptor(ImageManager.getImageDescriptor(ImageManager.ID_HIGHLIGHT_ERRORS));
		_viewer.addSelectionChangedListener(_highlightRuleErrorsAction);
	}

	private boolean getElementEvaluation() {
		if (_selectedObject != null && _selectedObject instanceof AcmeElement) {
			AcmeProperty evaluate = ((AcmeElement) _selectedObject).getProperty("vis-evaluate");
			if (evaluate != null) {
				return false;
			}
		}
		return true;
	}

	private void setElementEvaluation(boolean evaluate) {
		if (_selectedObject != null && _selectedObject instanceof AcmeElement) {
			AcmeProperty evalProp = ((AcmeElement) _selectedObject).getProperty("vis-evaluate");
			if (evalProp != null && evaluate) {
				((AcmeElement) _selectedObject).removeProperty("vis-evaluate");
			}
			else if (evalProp == null && !evaluate) {
				evalProp = new AcmeProperty("vis-evaluate", new BooleanType("boolean"), new AcmeBoolean(false));
				try {
					((AcmeElement) _selectedObject).addProperty(evalProp);
				}
				catch (AcmeIllegalNameException e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * Creates the context menu for this view */
	private void hookContextMenu() {
		MenuManager menuMgr = new MenuManager("#PopupMenu");
		menuMgr.setRemoveAllWhenShown(true);
		menuMgr.addMenuListener(new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {
				fillContextMenu(manager);
			}
		});

		Menu menu = menuMgr.createContextMenu(_viewer.getControl());
		_viewer.getControl().setMenu(menu);
		if (_parentView != null) {
			_parentView.getSite().registerContextMenu("RuleList", menuMgr, _viewer);
		}
	}

	/**
	 * Adds the actions to the context menu */
	private void fillContextMenu(IMenuManager manager) {
		if (_selectedObject != null) {
			_evaluateForElementAction.setText(
				_evaluateForElementTextPrefix + ((AcmeElement) _selectedObject).getName());
			_evaluateForElementAction.setChecked(getElementEvaluation());
		}
		if (_selectedRule != null) {
			_enableEvaluationAction.setChecked(_selectedRule.shouldEvaluate());
		}

		manager.add(_addRuleAction);

		manager.add(_deleteRuleAction);
		manager.add(new Separator("Evaluation"));
		manager.add(_enableEvaluationAction);
		manager.add(_evaluateForElementAction);
		manager.add(_forceEvaluationAction);
		manager.add (_browseRuleErrorsAction);
		manager.add(new Separator());
		manager.add(_doubleClickAction);
		manager.add(new Separator("Additions"));
	}

	/**
	 * Defines what happens when double click occurs Brings up the rule editor */
	public void handleDoubleClick() {
		ISelection selection = _viewer.getSelection();
		try {
			_selectedRule = (DesignRuleHolder) ((IStructuredSelection) selection).getFirstElement();
			if (_selectedRule.getDesignRule().getDesignRuleScope() == _selectedObject) {
				modifyRule();
			}
			else {
				ModifyRuleDialog ruleDialog =
					new ModifyRuleDialog(Display.getDefault().getActiveShell(), (AcmeElement) _selectedObject);
				ruleDialog.setDesignRule(_selectedRule);
				ruleDialog.disableEditing();
				ruleDialog.open();
				//                if (ruleDialog.open() != IDialogConstants.OK_ID) return;

			}
		}
		catch (ClassCastException e) {
			// only want to deal with Design rules
		}
		catch (NullPointerException ne) {
			// Ignore if nothing selected
		}
	}

	/**
	 * Deletes the currently selected rule */
	private void deleteRule() {
		// Should really only allow deletions if the rule is not inherited
		AcmeElement element = (AcmeElement) _selectedObject;
		if (_selectedRule != null && !_selectedRule.isInherited()) {
			DesignRule ruleToDelete = _selectedRule.getDesignRule();
			if (ruleToDelete != null) {
				if (ruleToDelete.getIsInvariant()) {
					element.removeInvariant(ruleToDelete);
				}
				else {
					element.removeHeuristic(ruleToDelete);
				}
				AcmeStudioChangeManager.globalChangeManager.onRemoveArmaniRule(element, ruleToDelete, this);
			}

		}
		_viewer.setInput(_selectedObject);
	}

	/**
	 * Modifies the currently selected rule */
	private void modifyRule() {
		Shell shell = AcmeStudioPlugin.getDefault().getWorkbench().getActiveWorkbenchWindow().getShell();
		DesignRule oldRule = _selectedRule.getDesignRule();
		// Create the rule editor
		ModifyRuleDialog ruleDialog = new ModifyRuleDialog(shell, (AcmeElement) _selectedObject);
		ruleDialog.setDesignRule(_selectedRule);
		if (ruleDialog.open() != IDialogConstants.OK_ID)
			return;

		// check to modify the rule
		AcmeElement element = (AcmeElement) _selectedObject;

		DesignRule newRule = ruleDialog.getDesignRule().getDesignRule();
		if (_selectedRule != null && oldRule != newRule) {

			_selectedRule.setDesignRule(newRule);
			if (newRule != null) {
				if (newRule.getIsInvariant()) {
					element.addInvariant(newRule);
				}
				else {
					element.addHeuristic(newRule);
				}
				AcmeStudioChangeManager.globalChangeManager.onAddArmaniRule(element, newRule, this);
			}
			if (oldRule != null) {
				if (oldRule.getIsInvariant()) {
					element.removeInvariant(oldRule);
				}
				else {
					element.removeHeuristic(oldRule);
				}
				AcmeStudioChangeManager.globalChangeManager.onRemoveArmaniRule(element, oldRule, this);
			}
		}
		_viewer.setInput(_selectedObject);
	}

	/**
	 * Sets the element that the rule list is looking at
	 * 
	 * @param o
	 *           the new object to be viewed
	 */
	public void setInput(AcmeObject o) {
		super.setInput (o);
		_selectedObject = o;
		if (o != null) {
			_viewer.setInput (o);
			if (o instanceof AcmeElement) {
				if (_evalCache != null) {
					_evalCache.removePropertyChangeListener (this);
				}
				_evalCache = ASModel.current ().getRuleManager ();
				_evalCache.addPropertyChangeListener (this);
			}
		}
	}
	/**
	 * @see org.eclipse.swt.widgets.Control#setFocus() */
	public boolean setFocus() {
		return _viewer.getControl().setFocus();
	}

	/**
	 * @see AcmeStudio.ui.views.parts.ElementViewPart#getContributions() */
	public IContributionItem[] getContributions() {
		return _myContributions;
	}

	/**
	 * @see AcmeStudio.ui.views.parts.ElementViewPart#isValidFor(AcmeObject) */
	public boolean isValidFor(AcmeObject o) {
		return true;
	}

	/**
	 * @see
	 * java.beans.PropertyChangeListener#propertyChange(PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		try {
		if (evt.getPropertyName().equals(DesignEvaluationCache.PROP_RULE_EVALUATION) || evt.getPropertyName().equals (DesignEvaluationCache.PROP_ELEMENT_TYPECHECK)) {
			// Optimization - only reset if rule is being managed by me
			_viewer.refresh();
		}
		} 
		catch (Exception e) {
			
		}
	}

	/*
	 * (non-Javadoc) @see
	 * AcmeStudio.ui.views.parts.ElementViewPart#createControl(org.eclipse.swt.widgets.Composite,
	 * int, AcmeStudio.ui.views.IElementViewPage)
	 */
	public void createControl(Composite parent, int style, IElementViewPage vp) {
		super.createControl(parent, style, vp);
		FillLayout tableLayout = new FillLayout();
		this.getComposite().setLayout(tableLayout);

		_viewer =
			new TableViewer(
				this.getComposite(),
				SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL | SWT.WRAP | SWT.FULL_SELECTION);
		_viewer.setContentProvider(new ViewContentProvider());
		_viewer.setLabelProvider(new ViewLabelProvider());
		
		_sorter	= new RuleSorter();
		_viewer.setSorter(_sorter);
		_viewer.setInput(ASContext.current().getActiveDesign().getDesign());

		// Set the editing properties of the table
		Table table = _viewer.getTable();
		table.addKeyListener(new KeyAdapter() {
			public void keyPressed(KeyEvent evt) {
				if (evt.character == SWT.DEL && _deleteRuleAction.isEnabled()) {
					_deleteRuleAction.run();
				}
			}
		});
		_viewer.setCellEditors(getCellEditors(table));
		_viewer.setColumnProperties(PROPERTIES);
		_viewer.addSelectionChangedListener(new ISelectionChangedListener () {

			public void selectionChanged(SelectionChangedEvent event) {
				_selectedRule = (DesignRuleHolder )((IStructuredSelection )_viewer.getSelection ()).getFirstElement ();
			}
			
		});
		// Set the table's appearance
		table.setHeaderVisible(false);
		table.setLinesVisible(true);

		// Set the column info
		TableColumn typechecksCol = new TableColumn(table, SWT.CENTER, INHERITCOL);
		TableColumn inheritedCol = new TableColumn(table, SWT.CENTER, EVALCOL);
		TableColumn ruleCol = new TableColumn(table, SWT.LEFT, RULECOL);
		
		typechecksCol.addSelectionListener(new SelectionAdapter () {
			public void widgetSelected (SelectionEvent evt) {
				_sorter.setCategory(RuleSorter.SORT_BY_EVALUATION);
				_sorter.cycleState();
				_viewer.refresh ();
			}
		});

		
		ruleCol.addSelectionListener(new SelectionAdapter () {
			public void widgetSelected (SelectionEvent evt) {
				_sorter.setCategory(RuleSorter.SORT_BY_NAME);
				_sorter.cycleState();
				_viewer.refresh ();
			}
		});
		ColumnLayoutData typechecksLayout = new ColumnPixelData(18, false);
		ColumnLayoutData inheritedLayout = new ColumnPixelData(18, false);
		ColumnLayoutData ruleLayout = new ColumnWeightData(90, true);

		// set columns in Table layout
		TableLayout layout = new TableLayout();
		layout.addColumnData(typechecksLayout);
		layout.addColumnData(inheritedLayout);
		layout.addColumnData(ruleLayout);
		table.setLayout(layout);

		// Initalize actions and menus
		_doubleClickAction = new Action() {
			public void run() {
				handleDoubleClick();
			}
		};

		hookDoubleClickAction();
		makeActions();
		hookContextMenu();
		table.setVisible(true);

		// Activate hover help
		HoverHelp help = new HoverHelp();
		help.activateHoverHelp(_viewer.getTable());

		// Initialize this view's contributions
		_myContributions =
			new IContributionItem[] {
				new ActionContributionItem(_highlightRuleErrorsAction),
				new ActionContributionItem(_addRuleAction),
				new ActionContributionItem(_deleteRuleAction)};

	}

	/* (non-Javadoc) @see AcmeStudio.ui.views.parts.ElementViewPart#getTitle() */
	public String getTitle() {
		return "Rules";
	}

}
