/*
 * Created on Sep 29, 2003
 *
 */
package AcmeStudio.ui.views.parts.wizards.functions;

import java.util.*;
import org.eclipse.jface.viewers.*;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;
import AcmeStudio.ui.views.parts.ModifyPropertyDialog;
import AcmeStudio.ui.views.providers.ParameterContentProvider;
import AcmeStudio.ui.views.providers.ParameterLabelProvider;
import AcmeStudio.util.AutoApplyEditableComboBoxCellEditor;
import AcmeStudio.util.Helper;
import AcmeStudio.util.Listeners;
import acmeLib.AcmeType;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.designRules.DesignRuleParameter;
import acmeLib.designRules.Scope;
import acmeLib.elements.*;
import acmeLib.properties.AcmePropertyType;

/**
 * Page to allow users to define the signature of a design analysis
 * @author Bradley Schmerl 
 * Created Sep 29, 2003 
 **/
public class NewFunctionSignaturePage extends WizardPage implements Listener {

	/** The namepsace where the new function will be defined **/
	private AcmeElement _namespace;
	/** Widget containing the name of the function **/
	private Text _txtName;
	/** Widget containing the return type of the function **/
	private Combo _cmbRetType;
	/** Widget containing the parameters of the function **/
	private TableViewer _vtblParameters;
	/** The new function **/
	private DesignAnalysis _function;
	/** Vector containing the parameters **/
	private Vector _params = new Vector();
	/** Button to add a parameter **/
	private Button _btnAddParam;
	/** Button to remove a parameter **/
	private Button _btnRemoveParam;
	/** All the availiable types for returning or parameters, currently in scope **/
	private Object[] _availableTypes;

	// Constants for the tables
	private final static int NAMECOL = 0;
	private final static int TYPECOL = 1;
	private final static String[] PROPERTIES = { "Name", "Type" };
	private static int COMP_CAT = 1;
	private static int CONN_CAT = 2;
	private static int PORT_CAT = 3;
	private static int ROLE_CAT = 4;
	private static int SYS_CAT = 5;
	private static int PROP_CAT = 6;
	private static int OTHER_CAT = 7;
	
	/** The parent wizard **/
	private NewFunctionWizard _parent;
	/** The name of the function **/
	private String _name;
	/** The return type of the function **/
	private AcmeType _returnType;
	/** Whether the analysis is external (i.e., points to a Java method call) **/
	boolean _isExternal = false;
	/**
	 * Creates the page
	 * @param parentWizard The parent wizard
	 * @param ns The namespace
	 */
	public NewFunctionSignaturePage(NewFunctionWizard parentWizard, AcmeElement ns) {
		
		super ("Define function signature");
		_parent = parentWizard;
		_namespace = ns;

		// Class to compare two types
		Comparator typeComparator = new Comparator() {

			protected int getCategory(AcmeType type) {
				int category = -1;
				if (type instanceof AcmeComponentType) {
					category = COMP_CAT;
				}
				else if (type instanceof AcmeConnectorType) {
					category = CONN_CAT;
				}
				else if (type instanceof AcmePortType) {
					category = PORT_CAT;
				}
				else if (type instanceof AcmeRoleType) {
					category = ROLE_CAT;
				}
				else if (type instanceof AcmeFamily) {
					category = SYS_CAT;
				}
				else if (type instanceof AcmePropertyType) {
					category = PROP_CAT;
				}
				else
					category = OTHER_CAT;

				return category;
			}

			public int compare(Object arg0, Object arg1) {
				if (arg0 instanceof AcmeType && arg1 instanceof AcmeType) {

					AcmeType t1 = (AcmeType) arg1;
					AcmeType t0 = (AcmeType) arg0;

					int typeCategory0 = getCategory(t0);
					int typeCategory1 = getCategory(t1);

					if (typeCategory1 == typeCategory0) {
						return String.CASE_INSENSITIVE_ORDER.compare(t0.getTypeName(), t1.getTypeName());
					}
					else if (typeCategory0 < typeCategory1) {
						return -1;
					}
					else if (typeCategory0 > typeCategory1) {
						return +1;
					}
				}
				return 0;
			}
		};
		
		// Creat the set of types
		SortedSet types = new TreeSet(typeComparator);
		types.addAll(Arrays.asList(ModifyPropertyDialog.getAvailablePropertyTypes(_namespace)));
		types.addAll(Helper.convertAcmeEnumerationToList(_namespace.getAvailableTypes()));
		types.add(ConstantTypes.ELEMENT_BASE_TYPE);
		types.add(ConstantTypes.COMPONENT_BASE_TYPE);
		types.add(ConstantTypes.CONNECTOR_BASE_TYPE);
		types.add(ConstantTypes.PORT_BASE_TYPE);
		types.add(ConstantTypes.SYSTEM_BASE_TYPE);
		types.add(ConstantTypes.ROLE_BASE_TYPE);
		types.add(ConstantTypes.TYPE_REFERENCE_TYPE);
		
		_availableTypes = types.toArray();

	}

	/*
	 * (non-Javadoc) @see
	 * org.eclipse.jface.dialogs.Dialog#createDialogArea(org.eclipse.swt.widgets.Composite)
	 */
	public void  createControl(Composite parent) {
		Composite top = new Composite (parent, SWT.NONE);
		top.setSize(500,500);
		GridLayout tgl = new GridLayout(2, false);
		top.setLayout(tgl);

		Label nameLabel = new Label(top, SWT.NONE);
		nameLabel.setText("Name:");

		_txtName = new Text(top, SWT.BORDER);
		GridData ngd = new GridData(GridData.GRAB_HORIZONTAL | GridData.HORIZONTAL_ALIGN_FILL);
		_txtName.setLayoutData(ngd);
		_txtName.addListener(SWT.Modify, this);
		_txtName.addVerifyListener(Listeners.VERIFY_ACMEID);

		Label retLabel = new Label(top, SWT.NONE);
		retLabel.setText("Returns:");

		_cmbRetType = new Combo(top, SWT.BORDER | SWT.DROP_DOWN);
		GridData rgd = new GridData(GridData.GRAB_HORIZONTAL | GridData.HORIZONTAL_ALIGN_FILL);
		_cmbRetType.setLayoutData(rgd);
		for (int i = 0; i < _availableTypes.length; i++) {
			_cmbRetType.add (((AcmeType )_availableTypes [i]).getTypeName());
		}
		_cmbRetType.addListener(SWT.Modify, this);

		Group params = new Group(top, SWT.BORDER);
		params.setText("Parameters:");
		params.setLayout(new GridLayout(2, false));
		GridData pgd = new GridData(GridData.GRAB_HORIZONTAL);
		pgd.horizontalSpan = GridData.FILL;
		pgd.horizontalAlignment = GridData.FILL;
		params.setLayoutData(pgd);

		_vtblParameters = new TableViewer(params, SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.FULL_SELECTION);
		ParameterContentProvider parameterContentProvider = new ParameterContentProvider();
		_vtblParameters.setContentProvider(parameterContentProvider);
		_vtblParameters.setLabelProvider(new ParameterLabelProvider());
		_vtblParameters.setColumnProperties(PROPERTIES);
		_vtblParameters.setInput(_params);
		pgd = new GridData(GridData.GRAB_HORIZONTAL /*| GridData.GRAB_VERTICAL */
		| GridData.HORIZONTAL_ALIGN_FILL | GridData.VERTICAL_ALIGN_FILL);
		pgd.verticalSpan = GridData.FILL;
		pgd.grabExcessVerticalSpace = true;
		pgd.grabExcessHorizontalSpace = true;
		Table table = _vtblParameters.getTable();
		table.setLayoutData(pgd);
		table.addListener(SWT.Modify, this);
		table.setHeaderVisible(true);
		table.setLinesVisible(true);

		TableColumn nameCol = new TableColumn(table, SWT.CENTER, 0);
		TableColumn typeCol = new TableColumn(table, SWT.CENTER, 1);
		nameCol.setText("Name");
		typeCol.setText("Type");
		ColumnLayoutData nld = new ColumnWeightData(50, true);
		ColumnLayoutData tld = new ColumnWeightData(50, true);
		TableLayout tald = new TableLayout();
		tald.addColumnData(nld);
		tald.addColumnData(tld);
		table.setLayout(tald);

		_vtblParameters.setCellModifier(new ParameterCellModifier());
		_vtblParameters.setCellEditors(getParamCellEditors(table));

		_btnAddParam = new Button(params, SWT.PUSH);
		_btnAddParam.setText("New");
		_btnAddParam.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent evt) {
				DesignRuleParameter newParam = new DesignRuleParameter("newParam", ConstantTypes.EXPLICIT_ANY_TYPE);
				_params.add(newParam);
				_vtblParameters.refresh();
			}
		});

		_btnRemoveParam = new Button(params, SWT.PUSH);
		_btnRemoveParam.setText("Remove");
		_btnRemoveParam.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent evt) {
				ISelection isel = _vtblParameters.getSelection();
				if (isel instanceof IStructuredSelection) {
					IStructuredSelection iss = (IStructuredSelection) isel;
					for (Iterator i = iss.iterator(); i.hasNext();) {
						DesignRuleParameter param = (DesignRuleParameter) i.next();
						_params.remove(param);

					}
					_vtblParameters.refresh();

				}
			}
		});
		
		final Button externalButton = new Button (top, SWT.CHECK);
		externalButton.setText("Function is external reference.");
		externalButton.addSelectionListener(new SelectionAdapter () {
			public void widgetSelected (SelectionEvent e) {
					_isExternal = externalButton.getSelection();
			}
		});

		validate ();
		setControl(top);
	}

	/**
	 * @return the editors to edit the parameters in the table
	 **/
	private CellEditor[] getParamCellEditors(Table table) {
		final CellEditor[] editors = new CellEditor[2];

		class AutoApplyTextCellEditor extends TextCellEditor {
			public AutoApplyTextCellEditor(Composite parent) {
				super(parent);
			}
			public void fireApplyEditorValue() {
				super.fireApplyEditorValue();
			}
		}

		

		editors[0] = new AutoApplyTextCellEditor(table);
		editors[0].getControl().addFocusListener(new FocusAdapter() {
			public void focusLost(FocusEvent e) {
				((AutoApplyTextCellEditor) editors[0]).fireApplyEditorValue();
			}
		});
		
		String [] typeNames = new String [_availableTypes.length];
		for (int i = 0; i < _availableTypes.length; i++) {
			typeNames [i] = ((AcmeType )_availableTypes [i]).getTypeName();
		}
		
		editors[1] = new AutoApplyEditableComboBoxCellEditor(table, typeNames);
		editors[1].getControl().addFocusListener(new FocusAdapter() {
			public void focusLost(FocusEvent e) {
				((AutoApplyEditableComboBoxCellEditor) editors[1]).fireApplyEditorValue();
			}
		});

		return editors;
	}

	/*
	 * (non-Javadoc) @see
	 * org.eclipse.swt.widgets.Listener#handleEvent(org.eclipse.swt.widgets.Event)
	 */
	public void handleEvent(Event event) {
		validate();
	}

	/**
	 * Checks to see if the signature is valid (i.e., the name doesn't already exist, the types are valid 
	 * @return whether the signature is valid in this context 
	 */
	private boolean validate() {
		
		_name = _txtName.getText();
		
		//		generatePrototype ();
		if (_name.equals("")) {
			setErrorMessage("Enter a name for the function.");
			setPageComplete (false);

			return false;
		}

		if (_namespace.lookupName(_name, false) != null) {
			setErrorMessage(_name + " is already defined in this scope.");
			setPageComplete (false);
		
			return false;
		}

		String type = _cmbRetType.getText();
		if (type.equals("")) {
			setErrorMessage("Enter a return type.");
			setPageComplete (false);
			return false;
		}

		Object exists = null;
		int index = 0;
		while (index < _availableTypes.length && exists == null) {
			if (((AcmeType )_availableTypes [index]).getTypeName ().equals (type)) {
				exists = _availableTypes [index];
			}
			else index++;
		}
		if (exists == null) {
			setPageComplete (false);
			setErrorMessage(type + " is not defined as a type in this scope.");
			return false;
		}
		else {
			_returnType = (AcmeType )exists;
		}
		
		this.setErrorMessage(null);
		setPageComplete (true);
		
		if (_function == null) {
			_function = new DesignAnalysis ();
		}
		
		try {
			//			 Reset the scope because setName later will 
			// add the function to the scope
			Scope oldScope = _function.getScope ();
			_function.setScope (null);
			_function.setName (_name);
			_function.setScope (oldScope);
		}
		catch (AcmeIllegalNameException e) {

		}
		_function.setReturnType(_returnType);
		_function.setFormalParams(_params);
		_parent.setPageTitles();

		return true;
	}

	/**
	 * @return The new function (without a function body) 
	 * */
	public DesignAnalysis getNewFunction() {
		if (_function == null) {
			_function = new DesignAnalysis ();
			try {
				_function.setName(_name);
			}
			catch (AcmeIllegalNameException e) {
				// Should never happen
			}
			_function.setFormalParams(_params);
			_function.setReturnType(_returnType);
		}
		_function.setScope(_namespace);
		_function.setIsExternalAnalysis(_isExternal);
		
		return _function;
	}


	/** 
	 * Class to define how to edit paramters in the param table
	 * @author Bradley Schmerl
	 * Created Oct 3, 2003
	 */
	private class ParameterCellModifier implements ICellModifier {

		public boolean canModify(Object element, String property) {
			if (element instanceof DesignRuleParameter) {
				DesignRuleParameter param = (DesignRuleParameter) element;
				return true;
			}
			return false;
		}

		/*
		 * (non-Javadoc) @see
		 * org.eclipse.jface.viewers.ICellModifier#getValue(java.lang.Object,
		 * java.lang.String)
		 */
		public Object getValue(Object element, String property) {
			Object value = null;
			if (element instanceof DesignRuleParameter) {
				DesignRuleParameter param = (DesignRuleParameter) element;
				if (property.equals(PROPERTIES[NAMECOL])) {
					value = param.getName();
				}
				else if (property.equals(PROPERTIES[TYPECOL])) {
					value = param.getType().getTypeName();
					boolean foundType = false;
					
					int index = 0;
					while (index < _availableTypes.length && !foundType) {
						if (_availableTypes[index].equals(param.getType().getTypeName())) {
							foundType = true;
						}
						else {
							index++;
						}
					}
					if (!foundType)
						value = new Integer(-1);
					else
						value = new Integer(index);
				}
				validate ();
			}
			return value;
		}

		/*
		 * (non-Javadoc) @see
		 * org.eclipse.jface.viewers.ICellModifier#modify(java.lang.Object,
		 * java.lang.String, java.lang.Object)
		 */
		public void modify(Object element, String property, Object value) {
			if (element instanceof TableItem) {
				TableItem item = (TableItem) element;
				DesignRuleParameter param = (DesignRuleParameter) item.getData();
				if (property.equals(PROPERTIES[NAMECOL])) {
					param.setName((String) value);
				}
				else if (property.equals(PROPERTIES[TYPECOL])) {
					int index = ((Integer) value).intValue();
					if (index != -1) {
						param.setType((AcmeType )_availableTypes [index]);
					}
					else {
						String val = param.getType().getTypeName ();
						AcmeType type = (AcmeType) _namespace.lookupName(val, true);
						if (type == null) {
							type = ModifyPropertyDialog.getPropertyType(val, _namespace);
						}
						if (type == null) type = ConstantTypes.EXPLICIT_ANY_TYPE;
					}
				}
				_vtblParameters.refresh ();
			}
		}
	}

}
