/*
 * Created on Oct 2, 2003
 *
 */
package AcmeStudio.ui.views.parts.wizards.newTypes;

import java.util.Iterator;
import java.util.Set;

import org.eclipse.jface.wizard.IWizardPage;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;

import AcmeStudio.util.Listeners;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.elements.*;

/**
 * Allows the user to enter the name and supertypes for a new type
 * @author Bradley Schmerl Created Oct 2, 2003 
 **/
public class NameAndTypePage extends WizardPage implements Listener {

	/** The family that the new type will be added to **/
	private AcmeFamily _family;
	/** The set of possible super types **/
	private Set _availableTypes;
	/** The widget that holds the name of the new type **/
	private Text _name;
	/** The widget that holds the supertypes of the new type **/
	private List _types;
	/** Whether supertypes should be displayed as fully qualified **/
	private boolean _qualifyTypeNames;

	/**
	 * Constructs the new page
	 * @param pageName The name of the page
	 * @param fam The scope to use for checking valid names
	 * @param types The set of valid supertypes
	 **/
	public NameAndTypePage(String pageName, AcmeFamily fam, Class cat, Set types) {
		super("Enter name and supertypes");
		_family = fam;
		_availableTypes = types;
	}

	/*
	 * (non-Javadoc) @see
	 * org.eclipse.jface.dialogs.IDialogPage#createControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createControl(Composite parent) {
		Composite top = new Composite(parent, SWT.NONE);
		top.setLayout(new GridLayout(2, false));

		// The name part
		Label nL = new Label(top, SWT.NONE);
		nL.setText("Type Name:");

		_name = new Text(top, SWT.BORDER);
		// Make sure this is going to be a valid name
		_name.addListener(SWT.Modify, this);
		_name.addVerifyListener(Listeners.VERIFY_ACMEID);
		GridData ngd = new GridData ();
		ngd.horizontalAlignment = GridData.FILL;
		_name.setLayoutData(ngd);

		// The supertypes part
		Label tL = new Label(top, SWT.NONE);
		tL.setText("Select supertypes:");
		GridData tlgd = new GridData();
		tlgd.horizontalSpan = GridData.FILL;
		tL.setLayoutData(tlgd);

		_types = new List(top, SWT.BORDER | SWT.MULTI | SWT.V_SCROLL);
		GridData tgd = new GridData(GridData.GRAB_HORIZONTAL | GridData.GRAB_VERTICAL);
		tgd.horizontalSpan = GridData.FILL;
		tgd.horizontalAlignment = GridData.FILL;
		tgd.verticalAlignment = GridData.FILL;
		_types.setLayoutData(tgd);
		_types.addSelectionListener(new SelectionAdapter () {
			public void widgetSelected (SelectionEvent e) {
				validate ();
			}
		});

		final Button qualifyButton = new Button(top, SWT.CHECK);
		qualifyButton.setText("Show qualified names in list");
		qualifyButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent evt) {
				_qualifyTypeNames = qualifyButton.getSelection();
				_types.removeAll();
				fillTypes();
			}
		});

		// Fill in proper detail
		fillTypes();
		validate();
		setTitle ("Enter the name and select super types of the new type");
		setControl(top);
	}

	/**
	 * Fills in the types in the type list
	 */
	private void fillTypes() {
		Iterator i = _availableTypes.iterator();
		while (i.hasNext()) {
			AcmeElementType et = (AcmeElementType) i.next();
			String name = null;
			if (!_qualifyTypeNames) {
				name = et.getTypeName();
			}
			else {
				name = ((AcmeElement) et).getQualifiedName();
			}
			_types.add(name);
			_types.setData(name, et);
		}
	}

	/*
	 * (non-Javadoc) @see
	 * org.eclipse.swt.widgets.Listener#handleEvent(org.eclipse.swt.widgets.Event)
	 */
	public void handleEvent(Event event) {
		validate();
	}

	/**
	 * Checks to see if the proposed names and types are valid
	 */
	private void validate() {
		// Check for no name
		if (_name.getText().equals("")) {
			setErrorMessage("Name cannot be empty");
			setPageComplete(false);
			return;
		}

        if (LanguageConfig.isArmaniKeyword(_name.getText ())) {
            setErrorMessage ("Name cannot be a keyword");
            setPageComplete (false);
            return;
        }
        
		// Check that the name isn't already in the family
		Object exists = _family.lookupName(_name.getText(), false);
		if (exists != null && exists != getPrototype()) {
			setErrorMessage("Name is already declared in this scope.");
			setPageComplete(false);
			return;
		}
		
		// If ok, then change the types name to this
		AcmeElementType proto = getPrototype();
		try {
			((AcmeElement )proto).setParent (null);
			proto.setTypeName (_name.getText ());
			((AcmeElement )proto).setParent (_family);
		}
		catch (AcmeIllegalNameException e) {
			setErrorMessage (e.getMessage ());
			setPageComplete (false);
			e.printStackTrace();
			return;
		}
		
		// Set the supertypes
		proto.getSuperTypes().removeAllItems();
		AcmeElementType[] types = getTypes();
		for (int i = 0; i < types.length; i++) {
			proto.addSuperType(types[i]);
		}
		setErrorMessage(null);
		setPageComplete(true);
	}

	/**
	 * @return the name of the new type
	 */
	public String getName() {
		return _name.getText();
	}

	/**
	 * @return the super types selected
	 */
	public AcmeElementType[] getTypes() {
		String[] selections = _types.getSelection();
		AcmeElementType[] types = new AcmeElementType[selections.length];
		for (int i = 0; i < selections.length; i++) {
			types[i] = (AcmeElementType) _types.getData(selections[i]);
		}
		return types;
	}

	/**
	 * Returns the property page if there is no allowable substructure for the type
	 * or the structure page if there is
	 */
	public IWizardPage getNextPage() {
		AcmeElementType pt = getPrototype();
		if (pt instanceof AcmeComponentType || pt instanceof AcmeConnectorType) {
			StructurePage next = ((NewElementTypeWizard )getWizard()).getStructurePage ();
			next.setProtoType(pt);
			return next;
		}
		else {
			PropertyPage pp = ((NewElementTypeWizard )getWizard()).getPropertyPage ();
			pp.setPrototype(getPrototype());
			return pp;
		}
	
	}

	/**
	 * @return The new type
	 **/
	private AcmeElementType getPrototype() {
		return ((NewElementTypeWizard) getWizard()).getNewType();
	}

}
