/*
 * Created on Oct 2, 2003
 *
 */
package AcmeStudio.ui.views.parts.wizards.newTypes;

import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.jface.wizard.Wizard;

import AcmeStudio.util.Utils;
import acmeLib.elements.*;

/**
 * Wizard for constructing new types. Allows user to specify inheritance
 * plus default properties/rules/structure.
 * @author Bradley Schmerl Created Oct 2, 2003 
 **/
public class NewElementTypeWizard extends Wizard {

	/** The family that the new type will be added to (used here for scoping names) **/
	private AcmeFamily _family;
	/** The category of the new type (i.e., component type, connector type, etc) **/
	private Class _category;
	/** The list of available types that can be a supertype of the new type **/
	private SortedSet _availableTypes;
	/** The page for entering the new name and type **/
	private NameAndTypePage _nameTypePage;
	/** The new type **/
	private AcmeElementType _newType;
	/** The page for entering property information about the type **/
	private PropertyPage _propertyPage;
	/** The page for entering rules that the type must satisify **/
	private RulesPage _rulesPage;
	/** The page for entering default structure **/
	private StructurePage _structurePage;

	/**
	 * Constructs the wizard
	 * @param fam The family that the new type will be added to
	 * @param clazz Whether the new type is a component type, connector type, etc
	 * @param availableTypes The list of possible super types
	 */
	public NewElementTypeWizard(AcmeFamily fam, Class clazz, List availableTypes) {
		super();
		_family = fam;
		_category = clazz;
		if (availableTypes != null) {
			_availableTypes = new TreeSet(new Utils.AcmeElementNameComparator());
			_availableTypes.addAll(availableTypes);
		}
	}

	/**
	 * Make sure that the new type is created and is parentless, so that it can be
	 * added later.
	 */
	public boolean performFinish() {

		if (_newType == null) constructNewType();
		((AcmeElement )_newType).setParent (null);
		return true;
	}

	/**
	 * Constructs the new type using name and super type information
	 *
	 */
	private void constructNewType() {
		String typeName = _nameTypePage.getName();
		if (_category == AcmeComponentType.class) {
			_newType = new AcmeComponentType(typeName);
		}
		else if (_category == AcmeConnectorType.class) {
			_newType = new AcmeConnectorType(typeName);
		}
		else if (_category == AcmePortType.class) {
			_newType = new AcmePortType(typeName);
		}
		else if (_category == AcmeRoleType.class) {
			_newType = new AcmeRoleType(typeName);
		}
		else if (_category == AcmeGroupType.class) {
			_newType = new AcmeGroupType(typeName);
		}
		else if (_category == AcmeFamily.class) {
			_newType = new AcmeFamily(typeName);
		}

		AcmeElementType[] types = _nameTypePage.getTypes();
		for (int i = 0; i < types.length; i++) {
			_newType.addSuperType(types[i]);
		}
		
		// tentatively place in family, but just by setting the type's parent
		((AcmeElement )_newType).setParent(_family);
	}

	/* (non-Javadoc) @see org.eclipse.jface.wizard.IWizard#addPages() */
	public void addPages() {
		_nameTypePage = new NameAndTypePage("Enter name and type", _family, _category, _availableTypes);
		addPage(_nameTypePage);
		
		_structurePage = new StructurePage (_family);
		addPage (_structurePage);
		_propertyPage = new PropertyPage (_family);
		addPage (_propertyPage);
		
		_rulesPage = new RulesPage (_family);
		addPage (_rulesPage);

	}

	/**
	 * 
	 * @return the new type constructed by this wizard 
	 **/
	public AcmeElementType getNewType() {
		if (_newType == null) constructNewType ();
		return _newType;
	}

	/**
	 * @return Returns the propertyPage.
	 */
	public PropertyPage getPropertyPage() {
		return _propertyPage;
	}

	/**
	 * @return Returns the structurePage.
	 */
	public StructurePage getStructurePage() {
		return _structurePage;
	}

}
