/*
 * Created on May 20, 2003
 *
 */
package AcmeStudio.ui.views.providers;

import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.Viewer;

import AcmeStudio.ui.views.parts.PropertyFilter;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.elements.*;
import acmeLib.properties.AcmeProperty;
import acmeLib.representations.AcmeRepresentation;

/**
 * Provides navigation routines to map the Acme design into 
 * tree operations. This is needed by the tree view to display the
 * Acme design.
 */
public class AcmeViewContentProvider implements IStructuredContentProvider, ITreeContentProvider {
    /** Indicates whether the view will display properties **/
    private boolean showProperties = false;
    private boolean showInstancesOnly = false;
    private boolean showHiddenProperties = false;
    private boolean hideRepresentations = false;
    private boolean hidePorts = false;
    private boolean hideComponents = false;
    private boolean hideConnectors = false;
    private boolean hideRoles = false;
    private boolean hideAnalyses = false;
    private boolean showSupertypes = false;

    /**
     * @see org.eclipse.jface.viewers.IContentProvider#InputChanged(Viewer,Object,Object)
     **/
    public void inputChanged(Viewer v, Object oldInput, Object newInput) {
    }

    /**
     * @see org.eclipse.jface.viewers.IContentProvider#dispose()
     **/
    public void dispose() {
    }

    /**
     * Gets the elements that are chidren of the argument
     * @param parent The parent element in the Acme design
     * @return An array of objects that contains all the children of the element
     * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(Object)
     **/
    public Object[] getElements(Object parent) {
        return getChildren(parent);
    }

    /**
     * Gets the parent of an Acme element
     * @param child The child node in the tree
     * @return The parent node in the tree
     * @see org.eclipse.jface.viewers.ITreeContentProvider@getParent(Object)
     **/
    public Object getParent(Object child) {
        Object parent = null;
        if (child instanceof AcmeElement) {
            // Get the parent of the element
            AcmeElement el = (AcmeElement) child;
            parent = el.getParent();
        }
        else if (child instanceof AcmeProperty) {
            // If it's a property, then the parent is the scope
            AcmeProperty prop = (AcmeProperty) child;
            parent = prop.getScope();
        }
        return parent;
    }

    /**
     * Gets all the children of a given parent
     * @param parent the node whose children to return
     * @return an array containing all the children
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getChildren(Object)
     **/
    public Object[] getChildren(Object parent) {
        /** Will contain all the properties (if they are required)**/
        AcmeEnumeration properties = AcmeEnumeration.EMPTY;
        AcmeEnumeration types = AcmeEnumeration.EMPTY;
        AcmeEnumeration families = AcmeEnumeration.EMPTY;
        AcmeEnumeration systems = AcmeEnumeration.EMPTY;
        AcmeEnumeration components = AcmeEnumeration.EMPTY;
        AcmeEnumeration connectors = AcmeEnumeration.EMPTY;
        AcmeEnumeration roles = AcmeEnumeration.EMPTY;
        AcmeEnumeration ports = AcmeEnumeration.EMPTY;
        AcmeEnumeration representations = AcmeEnumeration.EMPTY;
        /** The array to return **/
        Object[] children;
        /** The current index into the array, to be used when adding **/
        int index = 0;

        // Gather the properties
        if (parent instanceof AcmeElement && showProperties) {
            // if the viewer is showing properties, find all the properties of an element
            AcmeElement el = (AcmeElement) parent;

            properties = el.getProperties();
            // filter out the visual properties (used internally by AcmeStudio)

            if (!showHiddenProperties) {
                properties =
                    properties.castToAcmeSetVector().enumerate(
                        new PropertyFilter(new String[] { "vis-" }, true, false));
            }
        }

        // Get all the children of the design
        if (parent instanceof AcmeDesign) {
            // A design's children are all the types in the global scope,
            // all the families, and all the systems
            // What do we do about design analyses?
            AcmeDesign design = (AcmeDesign) parent;
            if (!showInstancesOnly) {
                types = design.getTypes();
                families = design.getFamilies();
            }
            systems = design.getSystems();

            // Missing invariants & design analyses

            int numChildren = properties.getSize() + types.getSize() + families.getSize() + systems.getSize();
            children = new Object[numChildren];

            // iterate through each enumeration and add it to the chidlren array
            while (types.moreItems()) {
                children[index++] = types.getNextItem();
            }

            while (families.moreItems()) {
                children[index++] = families.getNextItem();
            }

            while (systems.moreItems()) {
                children[index++] = systems.getNextItem();
            }

        }
        else if (parent instanceof AcmeFamily) {
            // Children of a family are it's types, its components and connectors in the 
            // default structure, and its representations
            AcmeFamily family = (AcmeFamily) parent;
            AcmeEnumeration analyses = AcmeEnumeration.EMPTY;
            if (!showInstancesOnly) {
                types = family.getTypes();
            }
            if (!hideComponents) {
                components = family.getComponents();
            }
            if (!hideConnectors) {
                connectors = family.getConnectors();
            }

            if (!hideRepresentations) {
                representations = family.getRepresentations();
            }
            if (!hideAnalyses) {
                analyses = family.getDesignAnalyses();
            }

            int numChildren =
                properties.getSize()
                    + types.getSize()
                    + components.getSize()
                    + connectors.getSize()
                    + representations.getSize()
                    + analyses.getSize();
            children = new Object[numChildren];

            // Iterate through each enumeration and add each child
            while (types.moreItems()) {
                children[index++] = types.getNextItem();
            }
            while (components.moreItems()) {
                children[index++] = components.getNextItem();
            }
            while (connectors.moreItems()) {
                children[index++] = connectors.getNextItem();
            }
            while (representations.moreItems()) {
                children[index++] = representations.getNextItem();
            }
            while (analyses.moreItems()) {
                children[index++] = analyses.getNextItem();
            }

        }
        else if (parent instanceof AcmeSystem) {
            // Children of system are components, connectors, and reps
            AcmeSystem system = (AcmeSystem) parent;
            if (!hideComponents) {
                components = system.getComponents();
            }
            if (!hideConnectors) {
                connectors = system.getConnectors();
            }
            if (!hideRepresentations) {
                representations = system.getRepresentations();
            }

            int numChildren =
                properties.getSize() + components.getSize() + connectors.getSize() + representations.getSize();
            children = new Object[numChildren];

            // Iterate through each of the enumerations and add to children
            while (components.moreItems()) {
                children[index++] = components.getNextItem();
            }

            while (connectors.moreItems()) {
                children[index++] = connectors.getNextItem();
            }

            while (representations.moreItems()) {
                children[index++] = representations.getNextItem();
            }

        }
        else if (parent instanceof AcmeComponent) {
            // Children of component are ports and reps
            AcmeComponent comp = (AcmeComponent) parent;
            if (!hidePorts) {
                ports = comp.getPorts();
            }
            if (!hideRepresentations) {
                representations = comp.getRepresentations();
            }
            int numChildren = ports.getSize() + properties.getSize() + representations.getSize();
            children = new Object[numChildren];

            while (ports.moreItems()) {
                children[index++] = ports.getNextItem();
            }

            while (representations.moreItems()) {
                children[index++] = representations.getNextItem();
            }

        }
        else if (parent instanceof AcmeConnector) {
            // Children of connectors are roles and reps
            AcmeConnector conn = (AcmeConnector) parent;
            if (!hideRoles) {
                roles = conn.getRoles();
            }
            if (!hideRepresentations) {
                representations = conn.getRepresentations();
            }
            int numChildren = roles.getSize() + properties.getSize() + representations.getSize();
            children = new Object[numChildren];

            while (roles.moreItems()) {
                children[index++] = roles.getNextItem();
            }

            while (representations.moreItems()) {
                children[index++] = representations.getNextItem();
            }

        }
        else if (parent instanceof AcmeRepresentation) {
            // Children of reps is its system
            AcmeRepresentation rep = (AcmeRepresentation) parent;
            children = new Object[1 + properties.getSize()];
            children[0] = rep.getSystem();
        }
        else {
            children = new Object[properties.getSize()];
        }

        if (showProperties) {
            // Add properties, if they are to be shown
            while (properties.moreItems()) {
                children[index++] = properties.getNextItem();
            }
        }

        return children;

    }

    /**
     * Returns whether the Acme element has any children
     * @param the element to check
     * @return true if the element has children
     * @see org.eclipse.jface.viewers.ITreeContentProvider#hasChildren(Object)
     **/
    public boolean hasChildren(Object parent) {
        AcmeEnumeration properties = new AcmeEnumeration();
        boolean children = false;
        boolean hasProperties = false;

        // Check to see if properties are being displayed, and the
        // element has them
        if (parent instanceof AcmeElement && showProperties) {
            AcmeElement el = (AcmeElement) parent;

            properties = el.getProperties();
            properties =
                properties.castToAcmeSetVector().enumerate(new PropertyFilter(new String[] { "vis-" }, true, false));
            hasProperties = properties.getSize() > 0;
        }

        if (parent instanceof AcmeDesign) {
            AcmeDesign design = (AcmeDesign) parent;
            children =
                design.getTypes().getSize() > 0
                    || design.getFamilies().getSize() > 0
                    || design.getSystems().getSize() > 0
                    || hasProperties;
        }
        else if (parent instanceof AcmeFamily) {
            AcmeFamily fam = (AcmeFamily) parent;
            children =
                fam.getTypes().getSize() > 0
                    || fam.getComponents().getSize() > 0
                    || fam.getConnectors().getSize() > 0
                    || fam.getRepresentations().getSize() > 0
                    || (!hideAnalyses && fam.getDesignAnalyses().getSize() > 0)
                    || hasProperties;
        }
        else if (parent instanceof AcmeSystem) {
            AcmeSystem sys = (AcmeSystem) parent;
            children =
                sys.getComponents().getSize() > 0
                    || sys.getConnectors().getSize() > 0
                    || sys.getRepresentations().getSize() > 0
                    || hasProperties;
        }
        else if (parent instanceof AcmeComponent) {
            AcmeComponent comp = (AcmeComponent) parent;
            children = comp.getPorts().getSize() > 0 || comp.getRepresentations().getSize() > 0 || hasProperties;
        }
        else if (parent instanceof AcmeConnector) {
            AcmeConnector conn = (AcmeConnector) parent;
            children = conn.getRoles().getSize() > 0 || conn.getRepresentations().getSize() > 0 || hasProperties;
        }
        else if (parent instanceof AcmeRepresentation) {
            children = true;
        }
        else if (showProperties && (parent instanceof AcmePort || parent instanceof AcmeRole)) {
            children = hasProperties;
        }

        return children;
    }
    /**
     * @return
     */
    public boolean isHideAnalyses() {
        return hideAnalyses;
    }

    /**
     * @return
     */
    public boolean isHideComponents() {
        return hideComponents;
    }

    /**
     * @return
     */
    public boolean isHideConnectors() {
        return hideConnectors;
    }

    /**
     * @return
     */
    public boolean isHidePorts() {
        return hidePorts;
    }

    /**
     * @return
     */
    public boolean isHideRepresentations() {
        return hideRepresentations;
    }

    /**
     * @return
     */
    public boolean isHideRoles() {
        return hideRoles;
    }

    /**
     * @return
     */
    public boolean isShowHiddenProperties() {
        return showHiddenProperties;
    }

    /**
     * @return
     */
    public boolean isShowInstancesOnly() {
        return showInstancesOnly;
    }

    /**
     * @return
     */
    public boolean isShowProperties() {
        return showProperties;
    }

    /**
     * @return
     */
    public boolean isShowSupertypes() {
        return showSupertypes;
    }

    /**
     * @param b
     */
    public void setHideAnalyses(boolean b) {
        hideAnalyses = b;
    }

    /**
     * @param b
     */
    public void setHideComponents(boolean b) {
        hideComponents = b;
    }

    /**
     * @param b
     */
    public void setHideConnectors(boolean b) {
        hideConnectors = b;
    }

    /**
     * @param b
     */
    public void setHidePorts(boolean b) {
        hidePorts = b;
    }

    /**
     * @param b
     */
    public void setHideRepresentations(boolean b) {
        hideRepresentations = b;
    }

    /**
     * @param b
     */
    public void setHideRoles(boolean b) {
        hideRoles = b;
    }

    /**
     * @param b
     */
    public void setShowHiddenProperties(boolean b) {
        showHiddenProperties = b;
    }

    /**
     * @param b
     */
    public void setShowInstancesOnly(boolean b) {
        showInstancesOnly = b;
    }

    /**
     * @param b
     */
    public void setShowProperties(boolean b) {
        showProperties = b;
    }

    /**
     * @param b
     */
    public void setShowSupertypes(boolean b) {
        showSupertypes = b;
    }

}
