/*
 * Created on Apr 20, 2004
 *
 */
package AcmeStudio.ui.widgets;

import java.util.ArrayList;

import org.eclipse.draw2d.ColorConstants;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTError;
import org.eclipse.swt.accessibility.*;
import org.eclipse.swt.custom.CLabel;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.widgets.*;


/**
 * @author Bradley Schmerl
 * Created Apr 20, 2004
 * 
 * The ImageCombo class represents a selectable user interface object 
 * that combines a label and a table and issues notification when an
 * item is selected from the list.
 * <p>
 * Note that although the class is a subclass of <code>Composite</code>, 
 * it does not make sense to add children to it, or set a layout on it.
 * </p>
 * <dl>
 * <dl><b>Styles</b>
 * <dd>BORDER, READ_ONLY, FLAT</dd>
 * <dt><b>Events</b>
 * <dd>Selection</dd>
 * </dl>
 */
final public class ImageCombo extends Composite {



	/**
	 * @author Bradley Schmerl
	 * Created Apr 20, 2004
	 */
	public class ItemLabelProvider extends LabelProvider implements ITableLabelProvider {

		/**
		 * 
		 */
		public ItemLabelProvider() {
			super ();
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.ITableLabelProvider#getColumnImage(java.lang.Object, int)
		 */
		public Image getColumnImage(Object element, int columnIndex) {
			if (element != null && columnIndex == 0) {
				return ((Item )element)._image;
			}
			return null;
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.ITableLabelProvider#getColumnText(java.lang.Object, int)
		 */
		public String getColumnText(Object element, int columnIndex) {
			if (element != null && columnIndex == 0) {
				Item item = (Item )element;
				if (item._string != null) return item._string;
			}
			return "";
		}
	}
	/**
	 * @author Bradley Schmerl
	 * Created Apr 20, 2004
	 */
	public class ItemContentProvider implements IStructuredContentProvider {

		/**
		 * 
		 */
		public ItemContentProvider() {
			super ();
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
		 */
		public Object [] getElements(Object inputElement) {
			return _elements.toArray ();
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.IContentProvider#dispose()
		 */
		public void dispose() {
		}

		/* (non-Javadoc)
		 * @see org.eclipse.jface.viewers.IContentProvider#inputChanged(org.eclipse.jface.viewers.Viewer, java.lang.Object, java.lang.Object)
		 */
		public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		}
	}
	CLabel _label;
	Table _table;
	int _maxItemCount = 5;
	Shell _popup;
	Button _arrow;
	boolean _hasFocus;
	
	private class Item {
		Image _image;
		String _string;
	}
	
	ArrayList _elements = new ArrayList ();
	private TableViewer	_viewer;
	
	/**
	 *  Constructs a new instance of this class given its parent
	 * and a style value describing its behavior and appearance.
	 * <p>
	 * The style value is either one of the style constants defined in
	 * class <code>SWT</code> which is applicable to instances of this
	 * class, or must be built by <em>bitwise OR</em>'ing together 
	 * (that is, using the <code>int</code> "|" operator) two or more
	 * of those <code>SWT</code> style constants. The class description
	 * lists the style constants that are applicable to the class.
	 * Style bits are also inherited from superclasses.
	 * </p>
	 * @param parent a widget will be the parent of the new instances (cannot be null)
	 * @param style the style of the widget to construct
	 * 
	 * @see SWT#BORDER
	 * @see SWT#FLAT
	 * @see Widget#getStyle
	 */
	public ImageCombo(Composite parent, int style) {
		super (parent, style = checkStyle (style));
		
		int labelStyle = 0;
		if ((style & SWT.FLAT) != 0) labelStyle |= SWT.FLAT;
		else labelStyle |= SWT.BORDER;
		_label = new CLabel (this, labelStyle);
		_label.setBackground(ColorConstants.white);
		_popup = new Shell (getShell (), SWT.NO_TRIM);
		
		int tableStyle = SWT.SINGLE | SWT.V_SCROLL;
		if ((style & SWT.FLAT) != 0) tableStyle |= SWT.FLAT;
		if ((style & SWT.RIGHT_TO_LEFT) != 0) tableStyle |= SWT.RIGHT_TO_LEFT;
		if ((style & SWT.LEFT_TO_RIGHT) != 0) tableStyle |= SWT.LEFT_TO_RIGHT;
		tableStyle |= SWT.FULL_SELECTION;
		
		_viewer = new TableViewer (_popup, tableStyle);
		_table = _viewer.getTable();
		TableColumn tc = new TableColumn (_table, SWT.LEFT, 0);
		TableLayout tl = new TableLayout ();
		tl.addColumnData(new ColumnWeightData (100, false));
		_table.setLayout(tl);
		
		_viewer.setContentProvider(new ItemContentProvider ());
		_viewer.setLabelProvider(new ItemLabelProvider ());
		
		int arrowStyle = SWT.ARROW | SWT.DOWN;
		if ((style & SWT.FLAT) != 0) arrowStyle |= SWT.FLAT;
		_arrow = new Button (this, arrowStyle);
		
		Listener listener = new Listener () {
			public void handleEvent (Event event) {
				if (_popup == event.widget) {
					popupEvent (event);
					return;
				}
				if (_label == event.widget) {
					labelEvent (event);
					return;
				}
				if (_table == event.widget) {
					tableEvent (event);
					return;
				}
				if (_arrow == event.widget) {
					arrowEvent (event);
					return;
				}
				if (ImageCombo.this == event.widget) {
					comboEvent (event);
					return;
				}
			}
		};
		
		addEventListener (listener, this, new int [] {SWT.Dispose, SWT.Move, SWT.Resize});
		addEventListener (listener, _popup, new int [] {SWT.Close, SWT.Paint, SWT.Deactivate});
		addEventListener (listener, _table, new int [] {SWT.MouseUp, SWT.Selection, SWT.Traverse, SWT.KeyDown, SWT.FocusIn, SWT.FocusOut});
		addEventListener (listener, _arrow, new int [] {SWT.Selection, SWT.FocusIn, SWT.FocusOut});
		addEventListener (listener, _label, new int [] {SWT.MouseDown, SWT.MouseUp, SWT.KeyDown, SWT.KeyUp, SWT.FocusIn, SWT.FocusOut, SWT.Traverse});
		initAccessible ();
		
	}
	
	private void addEventListener (Listener l, Control to, int [] events) {
		for (int i = 0; i < events.length; i++) to.addListener (events [i], l);
	}
	
	static int checkStyle (int style) {
		int mask = SWT.BORDER | SWT.FLAT | SWT.LEFT_TO_RIGHT | SWT.RIGHT_TO_LEFT;
		return style & mask;
	}
	
	/**
	* Adds an item.
	* <p>
	* The item is placed at the end of the list.
	* Indexing is zero based.
	*
	* @param string the new item
	* @param image the image to associate with the item
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_NULL_ARGUMENT)
	*	when the string is null
	* @exception SWTError(ERROR_ITEM_NOT_ADDED)
	*	when the item cannot be added
	*/
	public void add (String string, Image image) {
		checkWidget ();
		if (string == null && image == null) SWT.error (SWT.ERROR_NULL_ARGUMENT);
		Item newItem = new Item ();
		newItem._image = image;
		newItem._string = string;
		_elements.add (newItem);
		_viewer.setInput (_elements);
	}
	
	/**
	* Adds an item at an index.
	* <p>
	* The item is placed at an index in the list.
	* Indexing is zero based.
	*
	* This operation will fail when the index is
	* out of range.
	*
	* @param string the new item
	* @param index the index for the item
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_NULL_ARGUMENT)
	*	when the string is null
	* @exception SWTError(ERROR_ITEM_NOT_ADDED)
	*	when the item cannot be added
	*/
	public void add (String string, Image image, int index) {
		checkWidget ();
		if (string == null && image == null) SWT.error (SWT.ERROR_NULL_ARGUMENT);
		Item newItem = new Item ();
		newItem._image = image;
		newItem._string = string;
		_elements.add (index, newItem);
		_viewer.setInput (_elements);
	}
	
	/**	 
	* Adds the listener to receive events.
	* <p>
	*
	* @param listener the listener
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_NULL_ARGUMENT)
	*	when listener is null
	*/
	public void addModifyListener (ModifyListener listener) {
		checkWidget();
		if (listener == null) SWT.error (SWT.ERROR_NULL_ARGUMENT);
		TypedListener typedListener = new TypedListener (listener);
		addListener (SWT.Modify, typedListener);
	}
	/**	 
	* Adds the listener to receive events.
	* <p>
	*
	* @param listener the listener
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_NULL_ARGUMENT)
	*	when listener is null
	*/
	public void addSelectionListener(SelectionListener listener) {
		checkWidget();
		if (listener == null) SWT.error (SWT.ERROR_NULL_ARGUMENT);
		TypedListener typedListener = new TypedListener (listener);
		addListener (SWT.Selection,typedListener);
		addListener (SWT.DefaultSelection,typedListener);
	}
	
	void arrowEvent (Event event) {
		switch (event.type) {
			case SWT.FocusIn: {
				if (_hasFocus) return;
				_hasFocus = true;
				Event e = new Event();
				e.time = event.time;
				notifyListeners(SWT.FocusIn, e);
				break;
			}
			case SWT.FocusOut: {
				event.display.asyncExec(new Runnable() {
					public void run() {
						if (ImageCombo.this.isDisposed()) return;
						Control focusControl = getDisplay().getFocusControl();
						if (focusControl == _table || focusControl == _label) return;
						_hasFocus = false;
						Event e = new Event();
						notifyListeners(SWT.FocusOut, e);
					}
				});
				break;
			}
			case SWT.Selection: {
				dropDown (!isDropped ());
				break;
			}
		}
	}
	
	/**
	* Clears the current selection.
	* <p>
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	*/
	public void clearSelection () {
		checkWidget();
		
		_table.deselectAll ();
	}
	
	void comboEvent (Event event) {
		switch (event.type) {
			case SWT.Dispose:
				if (_popup != null && !_popup.isDisposed ()) _popup.dispose ();
				_popup = null;  
				_label = null;  
				_table = null;  
				_arrow = null;
				break;
			case SWT.Move:
				dropDown(false);
				break;
			case SWT.Resize:
				internalLayout();
				break;
		}
	}

	public Point computeSize (int wHint, int hHint, boolean changed) {
		checkWidget();
		int width = 0, height = 0;
		Point textSize = _label.computeSize (wHint, SWT.DEFAULT, changed);
		Point arrowSize = _arrow.computeSize(SWT.DEFAULT, SWT.DEFAULT, changed);
		Point listSize = _table.computeSize (wHint, SWT.DEFAULT, changed);
		int borderWidth = getBorderWidth();
		
		height = Math.max (hHint, Math.max(textSize.y, arrowSize.y)  + 2*borderWidth);
		width = Math.max (wHint, Math.max(textSize.x + arrowSize.x + 2*borderWidth, listSize.x + 2)  );
		return new Point (width, height);
	}
	
	/**
	* Deselects an item.
	* <p>
	* If the item at an index is selected, it is
	* deselected.  If the item at an index is not
	* selected, it remains deselected.  Indices
	* that are out of range are ignored.  Indexing
	* is zero based.
	*
	* @param index the index of the item
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	*/
	public void deselect (int index) {
		checkWidget();
		_table.deselect (index);
	}
	/**
	* Deselects all items.
	* <p>
	*
	* If an item is selected, it is deselected.
	* If an item is not selected, it remains unselected.
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	*/
	public void deselectAll () {
		checkWidget();
		_table.deselectAll ();
	}
	
	void dropDown (boolean drop) {
		if (drop == isDropped ()) return;
		if (!drop) {
			_popup.setVisible (false);
			_label.setFocus();
			return;
		}

		int index = _table.getSelectionIndex ();
		if (index != -1) _table.setTopIndex (index);
		Display display = getDisplay ();
		Rectangle listRect = _table.getBounds ();
		Rectangle parentRect = display.map (getParent (), null, getBounds());
		Point comboSize = getSize ();
		Rectangle displayRect = getMonitor().getClientArea();
		int width = Math.max (comboSize.x, listRect.width + 2);
		int height = listRect.height + 2;
		int x = parentRect.x;
		int y = parentRect.y + comboSize.y;
		if (y + height > displayRect.y + displayRect.height) y = parentRect.y - height;
		_popup.setBounds (x, y, width, height);
		_popup.setVisible (true);
		_table.setFocus();
	}
	public Control [] getChildren () {
		checkWidget();
		return new Control [0];
	}
	
	/**
	* Gets the number of items.
	* <p>
	* This operation will fail if the number of
	* items could not be queried from the OS.
	*
	* @return the number of items in the widget
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_CANNOT_GET_COUNT)
	*	when the operation fails
	*/
	public int getItemCount () {
		checkWidget();
		return _table.getItemCount ();
	}
	/**
	* Gets the height of one item.
	* <p>
	* This operation will fail if the height of
	* one item could not be queried from the OS.
	*
	* @return the height of one item in the widget
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_CANNOT_GET_ITEM_HEIGHT)
	*	when the operation fails
	*/
	public int getItemHeight () {
		checkWidget();
		return _table.getItemHeight ();
	}
	
	/**
	* Gets the index of the selected item.
	* <p>
	* Indexing is zero based.
	* If no item is selected -1 is returned.
	*
	* @return the index of the selected item.
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	*/
	public int getSelectionIndex () {
		checkWidget();
		return _table.getSelectionIndex ();
	}
	public int getStyle () {
		int style = super.getStyle();
		style &= ~SWT.READ_ONLY;
		return style;
	}
	
	void initAccessible() {
		getAccessible().addAccessibleListener(new AccessibleAdapter() {
			public void getHelp(AccessibleEvent e) {
				e.result = getToolTipText();
			}
		});
		

		
		getAccessible().addAccessibleControlListener(new AccessibleControlAdapter() {
			public void getChildAtPoint(AccessibleControlEvent e) {
				Point testPoint = toControl(new Point(e.x, e.y));
				if (getBounds().contains(testPoint)) {
					e.childID = ACC.CHILDID_SELF;
				}
			}
			
			public void getLocation(AccessibleControlEvent e) {
				Rectangle location = getBounds();
				Point pt = toDisplay(new Point(location.x, location.y));
				e.x = pt.x;
				e.y = pt.y;
				e.width = location.width;
				e.height = location.height;
			}
			
			public void getChildCount(AccessibleControlEvent e) {
				e.detail = 0;
			}
			
			public void getRole(AccessibleControlEvent e) {
				e.detail = ACC.ROLE_COMBOBOX;
			}
			
			public void getState(AccessibleControlEvent e) {
				e.detail = ACC.STATE_NORMAL;
			}

			public void getValue(AccessibleControlEvent e) {
				e.result = "";
			}
		});
	}
	boolean isDropped () {
		return _popup.getVisible ();
	}
	public boolean isFocusControl () {
		checkWidget();
		if (_label.isFocusControl() || _arrow.isFocusControl() || _table.isFocusControl() || _popup.isFocusControl()) {
			return true;
		} else {
			return super.isFocusControl();
		}
	}
	
	void internalLayout () {
		if (isDropped ()) dropDown (false);
		
		Rectangle rect = getClientArea();
		int width = rect.width;
		int height = rect.height;
		Point arrowSize = _arrow.computeSize(SWT.DEFAULT, height);
		_label.setBounds (0, 0, width - arrowSize.x, height);
		_arrow.setBounds (width - arrowSize.x, 0, arrowSize.x, arrowSize.y);
		
		Point size = getSize();
		int itemCount = _table.getItemCount();
		itemCount = (itemCount == 0) ? _maxItemCount : Math.min(_maxItemCount, itemCount);
		int itemHeight = _table.getItemHeight () * itemCount;
		Point listSize = _table.computeSize (SWT.DEFAULT, itemHeight);
		_table.setBounds (1, 1, Math.max (size.x - 2, listSize.x), listSize.y);
	}
	
	void tableEvent (Event event) {
		switch (event.type) {
			case SWT.FocusIn: {
				if (_hasFocus) return;
				_hasFocus = true;
				Event e = new Event();
				e.time = event.time;
				notifyListeners(SWT.FocusIn, e);
				break;
			}
			case SWT.FocusOut: {
				event.display.asyncExec(new Runnable() {
					public void run() {
						if (ImageCombo.this.isDisposed()) return;
						Control focusControl = getDisplay().getFocusControl();
						if (focusControl == _label || focusControl == _arrow) return;
						_hasFocus = false;
						Event e = new Event();
						notifyListeners(SWT.FocusOut, e);
					}
				});
				break;
			}
			case SWT.MouseUp: {
				if (event.button != 1) return;
				dropDown (false);
				break;
			}
			case SWT.Selection: {
				int index = _table.getSelectionIndex ();
				if (index == -1) return;
				
				Item item	= (Item )_table.getItem (index).getData ();
				_label.setImage(item._image);
				_label.setText (item._string);
				_table.setSelection(index);
				Event e = new Event();
				e.time = event.time;
				e.stateMask = event.stateMask;
				e.doit = event.doit;
				notifyListeners(SWT.Selection, e);
				event.doit = e.doit;
				break;
			}
			case SWT.Traverse: {
				switch (event.detail) {
					case SWT.TRAVERSE_TAB_NEXT:
					case SWT.TRAVERSE_RETURN:
					case SWT.TRAVERSE_ESCAPE:
					case SWT.TRAVERSE_ARROW_PREVIOUS:
					case SWT.TRAVERSE_ARROW_NEXT:
						event.doit = false;
						break;
				}
				Event e = new Event();
				e.time = event.time;
				e.detail = event.detail;
				e.doit = event.doit;
				e.keyCode = event.keyCode;
				notifyListeners(SWT.Traverse, e);
				event.doit = e.doit;
				break;
			}
			case SWT.KeyUp: {		
				Event e = new Event();
				e.time = event.time;
				e.character = event.character;
				e.keyCode = event.keyCode;
				e.stateMask = event.stateMask;
				notifyListeners(SWT.KeyUp, e);
				break;
			}
			case SWT.KeyDown: {
				if (event.character == SWT.ESC) { 
					// escape key cancels popup list
					dropDown (false);
				}
				if (event.character == SWT.CR || event.character == '\t') {
					// Enter and Tab cause default selection
					dropDown (false);
					Event e = new Event();
					e.time = event.time;
					e.stateMask = event.stateMask;
					notifyListeners(SWT.DefaultSelection, e);
				}
				//At this point the widget may have been disposed.
				// If so, do not continue.
				if (isDisposed()) break;
				Event e = new Event();
				e.time = event.time;
				e.character = event.character;
				e.keyCode = event.keyCode;
				e.stateMask = event.stateMask;
				notifyListeners(SWT.KeyDown, e);
				break;
				
			}
		}
	}
	void popupEvent(Event event) {
		switch (event.type) {
			case SWT.Paint:
				// draw black rectangle around list
				Rectangle listRect = _table.getBounds();
				Color black = getDisplay().getSystemColor(SWT.COLOR_BLACK);
				event.gc.setForeground(black);
				event.gc.drawRectangle(0, 0, listRect.width + 1, listRect.height + 1);
				break;
			case SWT.Close:
				event.doit = false;
				dropDown (false);
				break;
			case SWT.Deactivate:
				dropDown (false);
				break;
		}
	}
	public void redraw () {
		super.redraw();
		_label.redraw();
		_arrow.redraw();
		if (_popup.isVisible()) _table.redraw();
	}
	public void redraw (int x, int y, int width, int height, boolean all) {
		super.redraw(x, y, width, height, true);
	}
	
	/**
	* Removes an item at an index.
	* <p>
	* Indexing is zero based.
	*
	* This operation will fail when the index is out
	* of range or an item could not be removed from
	* the OS.
	*
	* @param index the index of the item
	* @return the selection state
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_ITEM_NOT_REMOVED)
	*	when the operation fails
	*/
	public void remove (int index) {
		checkWidget();
		_elements.remove(index);
		_viewer.setInput (_elements);
	}
	
	/**
	* Removes a range of items.
	* <p>
	* Indexing is zero based.  The range of items
	* is from the start index up to and including
	* the end index.
	*
	* This operation will fail when the index is out
	* of range or an item could not be removed from
	* the OS.
	*
	* @param start the start of the range
	* @param end the end of the range
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_ITEM_NOT_REMOVED)
	*	when the operation fails
	*/
	public void remove (int start, int end) {
		checkWidget();
		for (int i = end; i >= start; i--) _elements.remove (i);
		_viewer.setInput (_elements);
	}
	
	public void removeAll () {
		checkWidget ();
		_label.setText (""); _label.setImage (null);
		_elements.clear ();
		_viewer.setInput (_elements);
	}
	
	public void select (int index) {
		checkWidget ();
		if (index == -1) {
			_table.deselectAll();
			_label.setText (""); _label.setImage (null);
			return;
		}
		if (0 <= index && index < _table.getItemCount()) {
			if (index != getSelectionIndex()) {
				Item item = (Item )_table.getItem(index).getData();
				_label.setText (item._string==null?"":item._string);
				_label.setImage (item._image);
				_table.select (index);
				_table.showSelection();
			}
		}
	}
	public void setBackground (Color color) {
		super.setBackground(color);
		if (_label != null) _label.setBackground(color);
		if (_table != null) _table.setBackground(color);
		if (_arrow != null) _arrow.setBackground(color);
	}	
	
	public void setEnabled (boolean enabled) {
		super.setEnabled(enabled);
		if (_popup != null) _popup.setVisible (false);
		if (_label != null) _label.setEnabled(enabled);
		if (_arrow != null) _arrow.setEnabled(enabled);
	}
	
	public void setFont (Font font) {
		super.setFont (font);
		_label.setFont (font);
		_table.setFont (font);
		internalLayout();
	}
	public void setForeground (Color color) {
		super.setForeground(color);
		if (_label != null) _label.setForeground(color);
		if (_table != null) _table.setForeground(color);
		if (_arrow != null) _arrow.setForeground(color);
	}
	
	/**
	* Sets the text of an item; indexing is zero based.
	*
	* This operation will fail when the index is out
	* of range or an item could not be changed in
	* the OS.
	*
	* @param index the index for the item
	* @param string the item
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_NULL_ARGUMENT)
	*	when items is null
	* @exception SWTError(ERROR_ITEM_NOT_MODIFIED)
	*	when the operation fails
	*/
	public void setItem (int index, String string, Image image) {
		checkWidget();
		if (string == null && image == null) SWT.error (SWT.ERROR_NULL_ARGUMENT);
		Item newItem = new Item ();
		newItem._string = string;
		newItem._image = image;
		_elements.set(index, newItem);
		_viewer.setInput (_elements);
	}
	
	/**
	* Sets all items.
	*
	* @param items the array of items
	*
	* @exception SWTError(ERROR_THREAD_INVALID_ACCESS)
	*	when called from the wrong thread
	* @exception SWTError(ERROR_WIDGET_DISPOSED)
	*	when the widget has been disposed
	* @exception SWTError(ERROR_NULL_ARGUMENT)
	*	when items is null
	* @exception SWTError(ERROR_ITEM_NOT_ADDED)
	*	when the operation fails
	*/
	public void setItems (String [] strings, Image [] images) {
		checkWidget();
		if (strings == null || images == null) SWT.error (SWT.ERROR_NULL_ARGUMENT);
		if (strings.length != images.length) SWT.error (SWT.ERROR_INVALID_ARGUMENT);
		_label.setText (""); _label.setImage (null);
		_elements.clear ();
		for (int i = 0; i < images.length; i++) {
			Item item = new Item ();
			item._string = strings [i];
			item._image = images [i];
			_elements.add(item);
		}
		_viewer.setInput (_elements);
	}
	
	public void setToolTipText (String string) {
		checkWidget();
		super.setToolTipText(string);
		_arrow.setToolTipText (string);
		_label.setToolTipText (string);		
	}
	
	public void setVisible (boolean visible) {
		super.setVisible(visible);
		if (!visible) _popup.setVisible(false);
	}
	
	void labelEvent (Event event) {
		switch (event.type) {
			case SWT.FocusIn: {
				if (_hasFocus) return;
				_hasFocus = true;
				Event e = new Event();
				e.time = event.time;
				notifyListeners(SWT.FocusIn, e);
				break;
			}
			case SWT.FocusOut: {
				event.display.asyncExec(new Runnable() {
					public void run() {
						if (ImageCombo.this.isDisposed()) return;
						Control focusControl = getDisplay().getFocusControl();
						if (focusControl == _table || focusControl == _arrow) return;
						_hasFocus = false;
						Event e = new Event();
						notifyListeners(SWT.FocusOut, e);
					}
				});
				break;
			}
			case SWT.KeyDown: {
				
				if (event.character == SWT.ESC) { // escape key cancels popup list
					dropDown (false);
				}
				if (event.character == SWT.CR) {
					dropDown (false);
					Event e = new Event();
					e.time = event.time;
					e.stateMask = event.stateMask;
					notifyListeners(SWT.DefaultSelection, e);
				}
				//At this point the widget may have been disposed.
				// If so, do not continue.
				if (isDisposed()) break;
				
				if (event.keyCode == SWT.ARROW_UP || event.keyCode == SWT.ARROW_DOWN) {
					int oldIndex = getSelectionIndex ();
					if (event.keyCode == SWT.ARROW_UP) {
						select (Math.max (oldIndex - 1, 0));
					} else {
						select (Math.min (oldIndex + 1, getItemCount () - 1));
					}
			
					if (oldIndex != getSelectionIndex ()) {
						Event e = new Event();
						e.time = event.time;
						e.stateMask = event.stateMask;
						notifyListeners(SWT.Selection, e);
					}
					//At this point the widget may have been disposed.
					// If so, do not continue.
					if (isDisposed()) break;
				}
				
				// Further work : Need to add support for incremental search in 
				// pop up list as characters typed in text widget
							
				Event e = new Event();
				e.time = event.time;
				e.character = event.character;
				e.keyCode = event.keyCode;
				e.stateMask = event.stateMask;
				notifyListeners(SWT.KeyDown, e);
				break;
			}
			case SWT.KeyUp: {
				Event e = new Event();
				e.time = event.time;
				e.character = event.character;
				e.keyCode = event.keyCode;
				e.stateMask = event.stateMask;
				notifyListeners(SWT.KeyUp, e);
				break;
			}
			case SWT.Modify: {
				_table.deselectAll ();
				Event e = new Event();
				e.time = event.time;
				notifyListeners(SWT.Modify, e);
				break;
			}
			case SWT.MouseDown: {
				if (event.button != 1) return;
				boolean dropped = isDropped ();
				if (!dropped) setFocus ();
				dropDown (!dropped);
				break;
			}
			case SWT.MouseUp: {
				if (event.button != 1) return;
				break;
			}
			case SWT.Traverse: {		
				switch (event.detail) {
					case SWT.TRAVERSE_RETURN:
					case SWT.TRAVERSE_ARROW_PREVIOUS:
					case SWT.TRAVERSE_ARROW_NEXT:
						// The enter causes default selection and
						// the arrow keys are used to manipulate the list contents so
						// do not use them for traversal.
						event.doit = false;
						break;
				}
				
				Event e = new Event();
				e.time = event.time;
				e.detail = event.detail;
				e.doit = event.doit;
				e.keyCode = event.keyCode;
				notifyListeners(SWT.Traverse, e);
				event.doit = e.doit;
				break;
			}
		}
	}

}
