package AcmeStudio.util;

import java.util.ArrayList;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.gef.commands.Command;
import org.eclipse.gef.commands.CommandStack;

import AcmeStudio.core.ASModel;
import AcmeStudio.core.resource.ASDesign;
import AcmeStudio.ui.editors.diagram.visualInfo.VisualInformation;
import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.designRules.DesignRule;
import acmeLib.elements.AcmeAttachment;
import acmeLib.elements.AcmeComponent;
import acmeLib.elements.AcmeComponentType;
import acmeLib.elements.AcmeConnector;
import acmeLib.elements.AcmeConnectorType;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmeFamily;
import acmeLib.elements.AcmeGroup;
import acmeLib.elements.AcmePort;
import acmeLib.elements.AcmePortType;
import acmeLib.elements.AcmeRole;
import acmeLib.elements.AcmeRoleType;
import acmeLib.elements.AcmeSystem;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyBearer;
import acmeLib.properties.AcmePropertyType;
import acmeLib.properties.AcmePropertyValue;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.tools.AcmeChangeManager;

/**
 * @author schmerl
 *
 */
public class AcmeStudioChangeManager implements AcmeChangeManager {

    public static final int ADD_ATTACHMENT = 0;
    public static final int ADD_BINDING = 1;
    public static final int ADD_COMPONENT = 2;
    public static final int ADD_CONNECTOR = 3;
    public static final int ADD_GROUP = 4;
    public static final int ADD_MEMBER = 5;
    public static final int ADD_PORT = 6;
    public static final int ADD_PROPERTY = 7;
    public static final int ADD_REPRESENTATION = 8;
    public static final int ADD_ROLE = 9;
    public static final int ADD_SYSTEM = 10;
    public static final int ADD_TYPE = 11;
    public static final int ASSIGN_FAMILY = 12;
    public static final int ASSIGN_TYPE = 13;
    public static final int REMOVE_ATTACHMENT = 14;
    public static final int REMOVE_BINDING = 15;
    public static final int REMOVE_COMPONENT = 16;
    public static final int REMOVE_CONNECTOR = 17;
    public static final int REMOVE_GROUP = 18;
    public static final int REMOVE_MEMBER = 19;
    public static final int REMOVE_PORT = 20;
    public static final int REMOVE_PROPERTY = 21;
    public static final int REMOVE_REPRESENTATION = 22;
    public static final int REMOVE_ROLE = 23;
    public static final int REMOVE_SYSTEM = 24;
    public static final int REMOVE_TYPE = 25;
    public static final int SET_NAME = 26;
    public static final int SET_PROPERTY = 27;
    public static final int UNASSIGN_FAMILY = 28;
    public static final int UNASSIGN_TYPE = 29;
    public static final int ADD_ARMANI_RULE = 30;
    public static final int REMOVE_ARMANI_RULE = 31;
    public static final int ADD_COMPONENT_TYPE = 32;
    public static final int ADD_CONNECTOR_TYPE = 33;
    public static final int ADD_PORT_TYPE = 34;
    public static final int ADD_ROLE_TYPE = 35;
    public static final int ADD_PROPERTY_TYPE = 36;
    public static final int REMOVE_CONNECTOR_TYPE = 37;
    public static final int REMOVE_COMPONENT_TYPE = 38;
    public static final int REMOVE_PORT_TYPE = 39;
    public static final int REMOVE_ROLE_TYPE = 40;
    public static final int REMOVE_PROPERTY_TYPE = 41;
    public static final int ADD_FAMILY = 42;
    public static final int REMOVE_FAMILY = 43;
    public static final int IMPORT_DESIGN = 44;
    public static final int SET_COMMENT = 45;
    
    public static final int ALIGN_PORTS = 46;
    public static final int ROUTE_ROLES = 47;
    
	public static final int ADD_DESIGN_ANALYSIS = 48;
	public static final int REMOVE_DESIGN_ANALYSIS = 49;
 
	public static final int UPDATE_VIS_INFO = 50;
    
    public static final int PLUGIN_MASK = 0x8000;
    
    public static boolean announcing = false;
    public static List bufferedAnnouncements = new ArrayList ();
    
    private Map commandStacks = new HashMap ();

	public class AcmeStudioChangeEvent {
        protected int eventType;
        protected Object announcer = null;
        protected Object target;
        protected Object source;
        protected Object data = null;
        
        public AcmeStudioChangeEvent (int et, Object t, Object s) {
            eventType = et;
            target = t;
            source = s;
        }
        
        public AcmeStudioChangeEvent (int et, Object t, Object s, Object d) {
            eventType = et;
            target = t;
            source = s;
            data = d;
        }
        
        public int getEventType () {return eventType;}
        public Object getTarget () {return target;}
        public Object getSource () {return source;}
        public Object getData () {return data;}
        
        
        
        public void setAnnouncer (Object ann) {announcer = ann;}
        
        public Object getAnnouncer () {return announcer;}
        
    }


    /**
     * Constructor for AcmeStudioChangeManager.
     */
    private AcmeStudioChangeManager() {

    }

    /**
     * Returns the CommandStack mapped to the AcmeDesign.
     * 
     * @param des  the key AcmeDesign object
     * @return the matching CommandStack
     * @deprecated use instead #getCommandStack(ASDesign)
     */
    public CommandStack getCommandStack (AcmeDesign des) {
        return getCommandStack((ASDesign )ASModel.current().getASResource(des));
    }

    /**
     * Returns the CommandStack mapped to the ASDesign.
     * 
     * @param des  the key ASDesign object
     * @return the matching CommandStack
     */
    public CommandStack getCommandStack (ASDesign des) {
        Object stack = commandStacks.get (des);
        if (stack == null) {
            stack = new CommandStack () {
                public void execute (Command command) {
//                    Debug.traceln("Executing " + command.getLabel());
                    super.execute (command);
                    // TODO: Deal with errors in commands (look at the _errorReason)
                }
            };
            commandStacks.put (des, stack);

        }
        return (CommandStack )stack;
    }    

    public static AcmeStudioChangeManager globalChangeManager = new AcmeStudioChangeManager ();
    
    synchronized protected static void announceEvent(
        AcmeStudioChangeEvent evt,
        Object announcer) {
        evt.setAnnouncer(announcer);
        
        if (announcing) {
        	bufferedAnnouncements.add (evt);
        	return;
        }
        announcing = true;
	        try {
	        innerAnnounceEvent(evt);
	        
	        // Handle any announcements that came through as a result of this announcement
			while (bufferedAnnouncements.size () > 0) {
				List localAnnouncements = new ArrayList (bufferedAnnouncements);
				bufferedAnnouncements.clear ();
				int i = 0;
				while (i < localAnnouncements.size()) {
					AcmeStudioChangeEvent e = (AcmeStudioChangeEvent )localAnnouncements.get(i++);
					innerAnnounceEvent(e);
				}
			}
        }
		finally {
			announcing = false;
		}
    }
    
    private static void innerAnnounceEvent(AcmeStudioChangeEvent evt) {
    	Set localListeners = new HashSet (listeners);
    	Iterator it = localListeners.iterator();
    	
    	
    	try {
    		while (it.hasNext()) {
    			AcmeModelChangeListener l = (AcmeModelChangeListener) it.next();
    			try {
    				if (l.interestedInEvent(evt.getEventType())) {
    					l.changeModel(evt);
    				}
    			}
    			catch (ConcurrentModificationException e) {}
    			catch (Exception e) {
    				Debug.errorln(e);
    			}
    		}
    		localListeners = new HashSet (postListeners);
    		it = localListeners.iterator ();
    		while (it.hasNext ()) {
    			AcmeModelChangeListener l = (AcmeModelChangeListener )it.next ();
    			try {
    				if (l.interestedInEvent(evt.getEventType())) {
    					l.changeModel (evt);
    				}
    			}
    			catch (Exception e) {
    				Debug.diagnoseln(e);
    			}
    		}
    	}
    	catch (ConcurrentModificationException e) {
    		// Do nothing
    	}
    }

	public static Set listeners = new HashSet ();
   public static Set postListeners = new HashSet ();
	
    public static synchronized void registerListener (AcmeModelChangeListener listener) {
        synchronized (listeners) {
//        	if (!listeners.contains (listener)) 
        		listeners.add (listener);
        }
    }
    
    public static synchronized void registerPriorityListener (AcmeModelChangeListener listener) {
    	synchronized (listeners) {
//    		if (!listeners.contains (listener)) {
    			postListeners.add ( listener);
//    		}
    	}
    }

    public static synchronized void removeListener (AcmeModelChangeListener listener) {
        synchronized (listeners) {
        listeners.remove (listener);   
        }
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddAttachment(AcmeSystem, AcmeAttachment, Object)
     */
    public void onAddAttachment(
        AcmeSystem sys,
        AcmeAttachment att,
        Object announcer) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_ATTACHMENT, sys, att, sys.getParentDesign ());
            announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddBinding(AcmeRepresentation, AcmeBinding, Object)
     */
    public void onAddBinding(
        AcmeRepresentation representation,
        AcmeBinding binding,
        Object announcer) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_BINDING, representation, binding, representation.getParent().getParentDesign());
            announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddComponent(AcmeSystem, AcmeComponent, Object)
     */
    public void onAddComponent(
        AcmeSystem system,
        AcmeComponent comp,
        Object announcer) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_COMPONENT, system, comp, system.getParentDesign ());
            announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddConnector(AcmeSystem, AcmeConnector, Object)
     */
    public void onAddConnector(
        AcmeSystem system,
        AcmeConnector conn,
        Object announcer) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_CONNECTOR, system, conn, system.getParentDesign ());
            announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddGroup(AcmeSystem, AcmeGroup, Object)
     */
    public void onAddGroup(AcmeSystem system, AcmeGroup group, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_GROUP, system, group, system.getParentDesign ());
        announceEvent (evt, announcer);
        
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddMember(AcmeGroup, AcmeElement, Object)
     */
    public void onAddMember(AcmeGroup group, AcmeElement element, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_MEMBER, group, element, element.getParentDesign ());
        announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddPort(AcmeComponent, AcmePort, Object)
     */
    public void onAddPort(AcmeComponent comp, AcmePort port, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_PORT, comp, port, comp.getParentDesign ());
        announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddPort(AcmeComponent, AcmePort, Object)
     */
    public void onAlignPorts(AcmeComponent comp, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ALIGN_PORTS, comp, null, comp.getParentDesign ());
        announceEvent (evt, announcer);
    }

    public void onRouteRoles(AcmeConnector conn, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ROUTE_ROLES, conn, null, conn.getParentDesign ());
        announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddProperty(AcmeElement, AcmeProperty, Object)
     */
    public void onAddProperty(
        AcmePropertyBearer element,
        AcmeProperty property,
        Object announcer) {
            AcmeDesign design = null;
            if (element instanceof AcmeElement) design = ((AcmeElement )element).getParentDesign ();
            else if (element instanceof AcmeProperty) design = ((AcmeElement )((AcmeProperty )element).getScope ()).getParentDesign();
            else if (element instanceof DesignRule) design = ((AcmeElement )((DesignRule )element).getDesignRuleScope()).getParentDesign();
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_PROPERTY, element, property, design);
            announceEvent (evt, announcer);
    }


    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddRepresentation(AcmeElement, AcmeRepresentation, Object)
     */
    public void onAddRepresentation(
        AcmeElement element,
        AcmeRepresentation representation,
        Object announcer) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_REPRESENTATION, element, representation, element.getParentDesign ());
            announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddRole(AcmeConnector, AcmeRole, Object)
     */
    public void onAddRole(AcmeConnector conn, AcmeRole role, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_ROLE, conn, role, conn.getParentDesign ());
        announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddSystem(AcmeDesign, AcmeSystem, Object)
     */
    public void onAddSystem(AcmeDesign design, AcmeSystem system, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_SYSTEM, design, system, design);
        announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddType(AcmeDesign, AcmeType, Object)
     */
    public void onAddType(AcmeDesign design, AcmeType type, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_TYPE, design, type, design);
        announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAddType(AcmeFamily, AcmeType, Object)
     */
    public void onAddType(AcmeFamily family, AcmeType type, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_TYPE, family, type, family.getParentDesign());
        announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAssignFamily(AcmeSystem, AcmeFamily, Object)
     */
    public void onAssignFamily(AcmeSystem system, AcmeFamily family, Object announcer) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ASSIGN_FAMILY, system, family, system.getParentDesign ());
        announceEvent (evt, announcer);
        
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onAssignType(AcmeElement, AcmeElementType, Object)
     */
    public void onAssignType(
        AcmeElement element,
        AcmeElementType type,
        Object announcer) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ASSIGN_TYPE, element, type, element.getParentDesign ());
            announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveAttachment(AcmeSystem, AcmeAttachment, Object)
     */
    public void onRemoveAttachment(
        AcmeSystem system,
        AcmeAttachment attachment,
        Object announcer) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_ATTACHMENT, system, attachment, system.getParentDesign ());
            announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveBinding(AcmeRepresentation, AcmeBinding, Object)
     */
    public void onRemoveBinding(
        AcmeRepresentation representation,
        AcmeBinding binding,
        Object announcer) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_BINDING, representation, binding, representation.getParent ().getParentDesign());
            announceEvent (evt, announcer);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveComponent(AcmeSystem, AcmeComponent, Object)
     */
    public void onRemoveComponent(
        AcmeSystem system,
        AcmeComponent comp,
        Object ann) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_COMPONENT, system, comp, system.getParentDesign ());
            announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveConnector(AcmeSystem, AcmeConnector, Object)
     */
    public void onRemoveConnector(
        AcmeSystem system,
        AcmeConnector conn,
        Object ann) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_CONNECTOR, system, conn, system.getParentDesign ());
            announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveGroup(AcmeSystem, AcmeGroup, Object)
     */
    public void onRemoveGroup(AcmeSystem system, AcmeGroup group, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_GROUP, system, group, system.getParentDesign ());
        announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveMember(AcmeGroup, AcmeElement, Object)
     */
    public void onRemoveMember(AcmeGroup group, AcmeElement member, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_MEMBER, group, member, group.getParentDesign ());
            announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemovePort(AcmeComponent, AcmePort, Object)
     */
    public void onRemovePort(AcmeComponent comp, AcmePort port, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_PORT, comp, port, comp.getParentDesign ());
            announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveProperty(AcmeElement, AcmeProperty, Object)
     */
    public void onRemoveProperty(
        AcmePropertyBearer element,
        AcmeProperty property,
        Object ann) {
            AcmeDesign design = null;
            if (element instanceof AcmeElement) design = ((AcmeElement )element).getParentDesign ();
            else if (element instanceof AcmeProperty) design = ((AcmeElement )((AcmeProperty )element).getScope ()).getParentDesign();
            else if (element instanceof DesignRule) design = ((AcmeElement )((DesignRule )element).getDesignRuleScope()).getParentDesign();

            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_PROPERTY, element, property, design);
            announceEvent (evt, ann);
    }


    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveRepresentation(AcmeElement, AcmeRepresentation, Object)
     */
    public void onRemoveRepresentation(
        AcmeElement element,
        AcmeRepresentation representation,
        Object ann) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_REPRESENTATION, element, representation, element.getParentDesign ());
            announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveRole(AcmeConnector, AcmeRole, Object)
     */
    public void onRemoveRole(AcmeConnector conn, AcmeRole role, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_ROLE, conn, role,conn.getParentDesign ());
            announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveType(AcmeDesign, AcmeType, Object)
     */
    public void onRemoveType(AcmeDesign des, AcmeType type, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_TYPE, des, type, des);
            announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onRemoveType(AcmeFamily, AcmeType, Object)
     */
    public void onRemoveType(AcmeFamily family, AcmeType type, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_TYPE, family, type, family.getParentDesign ());
            announceEvent (evt, ann);
    }
    
    public void onRemoveSystem (AcmeDesign design, AcmeSystem system, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_SYSTEM, design, system, design);
        announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onSetName(AcmeElement, String, Object)
     */
    public void onSetName(AcmeElement element, String name, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (SET_NAME, element, null, name);
            announceEvent (evt, ann);
    }

    public void onSetComment (AcmeObject o, String comment, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (SET_COMMENT, o, null, comment);
            announceEvent (evt, ann);
       
    }
 
    /**
     * @see acmeLib.tools.AcmeChangeManager#onSetProperty(AcmeElement, AcmeProperty, AcmePropertyValue, Object)
     */
    public void onSetProperty(
        AcmePropertyBearer element,
        AcmeProperty property,
        AcmePropertyValue value,
        Object ann) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (SET_PROPERTY, element, property, value);
            announceEvent (evt, ann);
    }

    public void onUpdateVisInformation (AcmeElement element, VisualInformation vis, Object ann) {
    	AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (UPDATE_VIS_INFO, element, vis, element.getParentDesign());
    	announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onUnAssignFamily(AcmeSystem, AcmeFamily, Object)
     */
    public void onUnAssignFamily(
        AcmeSystem system,
        AcmeFamily family,
        Object ann) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (UNASSIGN_FAMILY, system, family, system.getParentDesign ());
            announceEvent (evt, ann);
    }

    /**
     * @see acmeLib.tools.AcmeChangeManager#onUnAssignType(AcmeElement, AcmeElementType, Object)
     */
    public void onUnAssignType(
        AcmeElement element,
        AcmeElementType type,
        Object ann) {
            AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (UNASSIGN_TYPE, element, type, element.getParentDesign ());
            announceEvent (evt, ann);
    }
    
    public void onAssignType (AcmeElementType type, AcmeElementType supertype, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ASSIGN_TYPE, type, supertype, ((AcmeElement )type).getParentDesign ());
        announceEvent (evt, ann);
    }
    
    public void onUnAssignType (AcmeElementType type, AcmeElementType supertype, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (UNASSIGN_TYPE, type, supertype, ((AcmeElement )type).getParentDesign ());
        announceEvent (evt, ann);
    }
    
    public void onAddArmaniRule (AcmeElement element, DesignRule rule, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (ADD_ARMANI_RULE, element, rule, element.getParentDesign ());
        announceEvent (evt, ann);     
            
    }
    
    public void onRemoveArmaniRule (AcmeElement element, DesignRule rule, Object ann) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (REMOVE_ARMANI_RULE, element, rule, element.getParentDesign ());
        announceEvent (evt, ann);   
    }
    
    public void onAddComponentType (AcmeFamily family, AcmeComponentType type, Object ann) {
        announce (ADD_COMPONENT_TYPE, family, type, ann, family.getParentDesign ());
    }

    public void onAddConnectorType (AcmeFamily family, AcmeConnectorType type, Object ann) {
        announce (ADD_CONNECTOR_TYPE, family, type, ann, family.getParentDesign ());
    }

    public void onAddPortType (AcmeFamily family, AcmePortType type, Object ann) {
        announce (ADD_PORT_TYPE, family, type, ann, family.getParentDesign ());
    }

    public void onAddRoleType (AcmeFamily family, AcmeRoleType type, Object ann) {
        announce (ADD_ROLE_TYPE, family, type, ann, family.getParentDesign ());
    }

    public void onAddPropertyType (AcmeFamily family, AcmePropertyType type, Object ann) {
        announce (ADD_PROPERTY_TYPE, family, type, ann, family.getParentDesign ());
    }

    public void onRemoveComponentType (AcmeFamily family, AcmeComponentType type, Object ann) {
        announce (REMOVE_COMPONENT_TYPE, family, type, ann, family.getParentDesign ());
    }
    
    public void onRemoveConnectorType (AcmeFamily family, AcmeConnectorType type, Object ann) {
        announce (REMOVE_CONNECTOR_TYPE, family, type, ann, family.getParentDesign ());
    }
    
    public void onRemovePortType (AcmeFamily family, AcmePortType type, Object ann) {
        announce (REMOVE_PORT_TYPE, family, type, ann, family.getParentDesign ());
    }
    
    public void onRemoveRoleType (AcmeFamily family, AcmeRoleType type, Object ann) {
        announce (REMOVE_ROLE_TYPE, family, type, ann, family.getParentDesign ());
    }
    
    public void onRemovePropertyType (AcmeFamily family, AcmePropertyType type, Object ann) {
        announce (REMOVE_PROPERTY_TYPE, family, type, ann, family.getParentDesign ());
    }


    public void onAddFamily (AcmeDesign design, AcmeFamily family, Object ann) {
        announce (ADD_FAMILY, design, family, ann, design);   
    }
    
    public void onRemoveFamily (AcmeDesign design, AcmeFamily family, Object ann) {
        announce (REMOVE_FAMILY, design, family, ann, design);
    }
    
    public void onImportDesign (AcmeDesign design, String filename, Object ann) {
        announce (IMPORT_DESIGN, design, filename, ann, design);
    }
    
   
    private void announce (int type, Object target, Object source, Object ann, AcmeDesign des) {
        AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (type, target, source, des);
        announceEvent (evt, ann);        
    }

	/**
	 * @param string
	 * @param gauge
	 * @param list
	 */
	public void onPluginChange(int type, Object target, Object source, Object announcer, AcmeDesign design) {
		int t = type | PLUGIN_MASK;
		AcmeStudioChangeEvent evt = new AcmeStudioChangeEvent (t, target, source, design);
		announceEvent (evt, announcer);
	}
    
    public static boolean isPluginEvent (AcmeStudioChangeEvent evt) {
    	return (evt.eventType & PLUGIN_MASK) == PLUGIN_MASK;
    }

	/**
	 * @param _function
	 * @param family
	 * @param _announcer
	 */
	public void onAddDesignAnalysis(DesignAnalysis function, AcmeFamily family, Object announcer) {
		announce (ADD_DESIGN_ANALYSIS, family, function, announcer, family.getParentDesign ());
	}

	/**
	 * @param _function
	 * @param design
	 * @param _announcer
	 */
	public void onAddDesignAnalysis(DesignAnalysis function, AcmeDesign design, Object announcer) {
		announce (ADD_DESIGN_ANALYSIS, design, function, announcer, design);
	}

	/**
	 * @param _function
	 * @param family
	 * @param _announcer
	 */
	public void onRemoveDesignAnalysis(DesignAnalysis function, AcmeFamily family, Object announcer) {
		announce (REMOVE_DESIGN_ANALYSIS, family, function, announcer, family.getParentDesign ());
	}

	/**
	 * @param _function
	 * @param design
	 * @param _announcer
	 */
	public void onRemoveDesignAnalysis(DesignAnalysis function, AcmeDesign design, Object announcer) {
		announce (REMOVE_DESIGN_ANALYSIS, design, function, announcer, design);
	}

    /**
     * @param evt
     * @return
     */
    public static AcmeSystem getSystemFromEvent(AcmeStudioChangeEvent evt) {
        switch (evt.getEventType()) {
            case ADD_COMPONENT:
            case ADD_CONNECTOR:
            case ADD_GROUP:
            case REMOVE_COMPONENT:
            case REMOVE_CONNECTOR:
            case REMOVE_GROUP:
                return (AcmeSystem )evt.getTarget();
            case ADD_PORT:
            case ADD_ROLE:
            case REMOVE_PORT:
            case REMOVE_ROLE:
                return (AcmeSystem )((AcmeElement )evt.getTarget ()).getParent ();
            case ADD_PROPERTY:
            case REMOVE_PROPERTY:
            case SET_PROPERTY: {
                if (evt.getTarget() instanceof AcmeSystem) return (AcmeSystem )evt.getTarget ();
                else if (evt.getTarget () instanceof AcmeComponent || evt.getTarget () instanceof AcmeConnector ||
                            evt.getTarget () instanceof AcmeGroup) {
                    return (AcmeSystem )((AcmeElement )evt.getTarget ()).getParent ();
                }
                else if (evt.getTarget () instanceof AcmeRole || evt.getTarget () instanceof AcmePort) {
                    return (AcmeSystem )((AcmeElement )evt.getTarget ()).getParent ().getParent ();
                }
            }
                
        }
        return null;
    }
    
    public static String eventTypeAsString (int evtType) {
    	switch (evtType) {
    		 case ADD_ATTACHMENT :
				return "ADD_ATTACHMENT";
			case ADD_BINDING :
				return "ADD_BINDING";
			case ADD_COMPONENT :
				return "ADD_COMPONENT";
			case ADD_CONNECTOR :
				return "ADD_CONNECTOR";
			case ADD_GROUP :
				return "ADD_GROUP";
			case ADD_MEMBER :
				return "ADD_MEMBER";
			case ADD_PORT :
				return "ADD_PORT";
			case ADD_PROPERTY :
				return "ADD_PROPERTY";
			case ADD_REPRESENTATION :
				return "ADD_REPRESENTATION";
			case ADD_ROLE :
				return "ADD_ROLE";
			case ADD_SYSTEM :
				return "ADD_SYSTEM";
			case ADD_TYPE :
				return "ADD_TYPE";
			case ASSIGN_FAMILY :
				return "ASSIGN_FAMILY";
			case ASSIGN_TYPE :
				return "ASSIGN_TYPE";
			case REMOVE_ATTACHMENT :
				return "REMOVE_ATTACHMENT";
			case REMOVE_BINDING :
				return "REMOVE_BINDING";
			case REMOVE_COMPONENT :
				return "REMOVE_COMPONENT";
			case REMOVE_CONNECTOR :
				return "REMOVE_CONNECTOR";
			case REMOVE_GROUP :
				return "REMOVE_GROUP";
			case REMOVE_MEMBER :
				return "REMOVE_GROUP";
			case REMOVE_PORT :
				return "REMOVE_PORT";
			case REMOVE_PROPERTY :
				return "REMOVE_PROPERTY";
			case REMOVE_REPRESENTATION :
				return "REMOVE_REPRESENTATION";
			case REMOVE_ROLE :
				return "REMOVE_ROLE";
			case REMOVE_SYSTEM :
				return "REMOVE_SYSTEM";
			case REMOVE_TYPE :
				return "REMOVE_TYPE";
			case SET_NAME :
				return "SET_NAME";
			case SET_PROPERTY :
				return "SET_PROPERTY";
			case UNASSIGN_FAMILY :
				return "UNASSIGN_FAMILY";
			case UNASSIGN_TYPE :
				return "UNASSIGN_TYPE";
			case ADD_ARMANI_RULE :
				return "ADD_ARMANI_RULE";
			case REMOVE_ARMANI_RULE :
				return "REMOVE_ARMANI_RULE";
			case ADD_COMPONENT_TYPE :
				return "ADD_COMPONENT_TYPE";
			case ADD_CONNECTOR_TYPE :
				return "ADD_CONNECTOR_TYPE";
			case ADD_PORT_TYPE :
				return "ADD_PORT_TYPE";
			case ADD_ROLE_TYPE :
				return "ADD_ROLE_TYPE";
			case ADD_PROPERTY_TYPE :
				return "ADD_PROPERTY_TYPE";
			case REMOVE_CONNECTOR_TYPE :
				return "REMOVE_CONNECTOR_TYPE";
			case REMOVE_COMPONENT_TYPE :
				return "REMOVE_COMPONENT_TYPE";
			case REMOVE_PORT_TYPE :
				return "REMOVE_PORT_TYPE";
			case REMOVE_ROLE_TYPE :
				return "REMOVE_ROLE_TYPE";
			case REMOVE_PROPERTY_TYPE :
				return "REMOVE_PROPERTY_TYPE";
			case ADD_FAMILY :
				return "ADD_FAMILY";
			case REMOVE_FAMILY :
				return "REMOVE_FAMILY";
			case IMPORT_DESIGN :
				return "IMPORT_DESIGN";
			case SET_COMMENT :
				return "SET_COMMENT";
			case ALIGN_PORTS :
				return "ALIGN_PORTS";
			case ROUTE_ROLES :
				return "ROUTE_ROLES";
			case ADD_DESIGN_ANALYSIS :
				return "ADD_DESIGN_ANALYSIS";
			case REMOVE_DESIGN_ANALYSIS :
				return "REMOVE_DESIGN_ANALYSIS";
			case UPDATE_VIS_INFO :
				return "UPDATE_VIS_INFO";
			default :
				return "UNKNOWN";
		}
    }
}
