package AcmeStudio.util;

import java.util.Properties;

import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.preference.IPreferenceStore;

import AcmeStudio.AcmeStudioPlugin;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 *
 * Debug object to handle debug, info, warn, error, and fatal level messages.
 * The messages are sent to log file via Logger.  Sending to Logger is a good
 * idea since standard stream output doesn't appear anywhere in deployed
 * version.<br>
 * <br>
 * The amount of debug is tuned via preference page:  DEBUG, INFO, WARN,
 * ERROR, or FATAL.  NONE is shorthand for no message.<br>
 * <br>
 * Expected usage:<br>
 * The static methods of this class serve most purposes.  The public methods
 * of the Debug class instance can be accessed by getting the singleton
 * instance.<br>
 * <br>
 * TODO:
 *     The Debug setting should eventually be settable as Plugin Preference.
 *     Preference to obtain:
 *         - Log file name.  Default: file in plugin's metadata area named ${AcmeStudio.debug.filename}
 *         - Trace on?  Default: off (ALPHA version is on)
 *         - Diagnostic on?  Default: off (ALPHA version is on)
 *         - Warning on?  Default: on
 *         - Error on?  Default: on
 *         - Fatal on?  Default: on
 */
public class Debug {
    // Preference setting constants
    public static final String PREFKEY_LEVEL = "Debugging.Level";
    public static final String PREFKEY_PATTERN = "Debugging.LogPattern";
    public static final String PREFKEY_MAX_SIZE = "Debugging.MaxSize";
    public static final String PREFKEY_MAX_BACKUPS = "Debugging.MaxBackups";
    public static final String PREFKEY_FILE = "Debugging.LogFile";

    public static final String LEVEL_NONE = "NONE";
    public static final String LEVEL_DEBUG = "DEBUG";
    public static final String LEVEL_INFO = "INFO";
    public static final String LEVEL_WARN = "WARN";
    public static final String LEVEL_ERROR = "ERROR";
    public static final String LEVEL_FATAL = "FATAL";

    public static final String DEFAULT_LEVEL = LEVEL_WARN;
    public static final String DEFAULT_PATTERN = "%d{yyyy-dd-MM HH:mm:ss.SSS} [%t] %p %c %x - %m%n";
    public static final int DEFAULT_MAX_SIZE = 1024;
    public static final int DEFAULT_MAX_BACKUPS = 5;
    public static final String DEFAULT_FILE = "AcmeStudio.log";
    
    /** Singleton debug object. */
    private static Debug _instance = null;
    
    /** Debug logger instance */
    private Logger _log = null;

    /** Debug output path */
    private IPath _filepath = null;

    /**
     * Protected constructor for Debug.
     */
    protected Debug () {
        super();
        
        init();
    }

/////////////////////////
// Static class methods
/////////////////////////

    /** 
     * Returns the debug object instance.
     * @return Debug  the singleton Debug instance
     */
    public static Debug getInstance () {
        lazyInit();

        return _instance;
    }

    /**
     * Logs tracing debug message.
     * @param obj  the object to print debug for
     */
    public static void traceln (Object obj) {
        lazyInit();

        _instance.printConsole(obj);
        getLog().debug(obj);
    }

    /**
     * Logs tracing debug message, using Throwable object.
     * @param obj  the object to print debug for
     */
    public static void traceln (Throwable obj) {
        lazyInit();

        _instance.printConsole(obj);
        getLog().debug(obj.getMessage(), obj);
    }

    /**
     * Logs diagnostic debug message.
     * @param obj  the object to print debug for
     */
    public static void diagnoseln (Object obj) {
        lazyInit();
        
        _instance.printConsole(obj);
        getLog().info(obj);
    }

    /**
     * Logs diagnostic debug message, using Throwable object.
     * @param obj  the Throwable object to print debug for
     */
    public static void diagnoseln (Throwable obj) {
        lazyInit();
        
        _instance.printConsole(obj);
        getLog().info(obj.getMessage(), obj);
    }

    /**
     * Logs warning debug message.
     * @param obj  the object to print debug for
     */
    public static void warnln (Object obj) {
        lazyInit();
        
        _instance.printConsole(obj);
        getLog().warn(obj);
    }

    /**
     * Logs warning debug message, using Throwable object.
     * @param obj  the object to print debug for
     */
    public static void warnln (Throwable obj) {
        lazyInit();
        
        _instance.printConsole(obj);
        getLog().warn(obj.getMessage(), obj);
    }

    /**
     * Logs error debug message.
     * @param obj  the object to print debug for
     */
    public static void errorln (Object obj) {
        lazyInit();
        
        _instance.printConsole(obj);
        getLog().error(obj);
    }

    /**
     * Logs error debug message, using Throwable object.
     * @param obj  the object to print debug for
     */
    public static void errorln (Throwable obj) {
        lazyInit();
        
        _instance.printConsole(obj);
        getLog().error(obj.getMessage(), obj);
    }

    /**
     * Logs fatal debug message.
     * @param obj  the object to print debug for
     */
    public static void fatalln (Object obj) {
        lazyInit();
        
        _instance.printConsole(obj);
        getLog().fatal(obj);
    }

    /**
     * Logs fatal debug message, using Throwable object.
     * @param obj  the object to print debug for
     */
    public static void fatalln (Throwable obj) {
        lazyInit();
        
        _instance.printConsole(obj);
        getLog().fatal(obj.getMessage(), obj);
    }


    public static Logger getLog () {
        lazyInit ();

        return _instance._log;
    }

    /**
     * Re-initializes debug object by resetting the Logger.
     */    
    public static void reset () {
        LogManager.shutdown();

        if (_instance != null) {
            _instance._log = null;
            _instance.init();
        }
    }
    
    /**
     * Lazy-initiales a debug object, should be called by any method that
     * accesses the static Debug instance.
     */
    private static void lazyInit () {
        if (_instance == null) {
            _instance = new Debug();
        }
    }

///////////////////////////
// Class instance methods
///////////////////////////

    /**
     * Returns the debug file name.
     * @return String  the debug file path and name.
     */
    public String getDebugFile () {
        return _filepath.toString();
    }

    /**
     * Returns the current debug level.
     * @return Level  the current debug level
     */
    public Level getDebugLevel () {
        return _log.getLevel();
    }

    /**
     * Returns whether DEBUG is turned on.
     * @return boolean  <code>true</code> if trace is on, <code>false</code> otherwise
     */
    public boolean isTraceOn () {
        return _log.isDebugEnabled();
    }

    /**
     * Returns whether INFO is turned on.
     * @return boolean  <code>true</code> if trace is on, <code>false</code> otherwise
     */
    public boolean isDiagnoseOn () {
        return _log.isInfoEnabled();
    }

    /**
     * Returns whether WARN is turned on.
     * @return boolean  <code>true</code> if trace is on, <code>false</code> otherwise
     */
    public boolean isWarnOn () {
        return _log.isEnabledFor(Level.WARN);
    }

    /**
     * Returns whether ERROR is turned on.
     * @return boolean  <code>true</code> if trace is on, <code>false</code> otherwise
     */
    public boolean isErrorOn () {
        return _log.isEnabledFor(Level.ERROR);
    }

    /**
     * Returns whether FATAL is turned on.
     * @return boolean  <code>true</code> if trace is on, <code>false</code> otherwise
     */
    public boolean isFatalOn () {
        return _log.isEnabledFor(Level.FATAL);
    }
    
    public void printConsole (Object o) {
        if (o instanceof Throwable) {
            ((Throwable) o).printStackTrace(System.err);
        } else {
            System.out.println(o);
        }
    }

    /**
     * Initializes the debug object, setting up the debug file writer.
     * The debug file defaults to ${AcmeStudio.debug.filename}, which is
     * specified in the plugin resource file.
     */
    protected void init () {
        if (_log != null) {  // don't need to initialize
            return;
        }

        IPreferenceStore store = AcmeStudioPlugin.getDefault().getPreferenceStore();

        // get debug log path
        IPath metadataPath = Platform.getPluginStateLocation(AcmeStudioPlugin.getDefault());
        IPath filePath = metadataPath.append(store.getString(PREFKEY_FILE));

        // setup logging
        Properties props = new Properties();
        props.setProperty("log4j.appender.FileLog", "org.apache.log4j.RollingFileAppender");
        props.setProperty("log4j.appender.FileLog.layout", "org.apache.log4j.PatternLayout");
        props.setProperty("log4j.appender.FileLog.layout.ConversionPattern", store.getString(PREFKEY_PATTERN));
        props.setProperty("log4j.appender.FileLog.MaxFileSize", store.getString(PREFKEY_MAX_SIZE) + "KB");
        props.setProperty("log4j.appender.FileLog.MaxBackupIndex", store.getString(PREFKEY_MAX_BACKUPS));
        props.setProperty("log4j.appender.FileLog.File", filePath.toOSString());
        String rootSetting = store.getString(PREFKEY_LEVEL) + ",FileLog";
        props.setProperty("log4j.rootCategory", rootSetting);
        PropertyConfigurator.configure(props);
        _log = Logger.getLogger(this.getClass());
    }

	/**
	 * Throws and catches a throwable, and then reports the stack. This is
	 * useful for finding call traces
	 * @param message The message to report
	 */
	public static void reportStack(String message) {
		Debug.diagnoseln (message);
		try {
			throw new Throwable (message);
		}
		catch (Throwable t) {
			diagnoseln (t);
		}
	}

}
