package AcmeStudio.util;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.*;

import org.eclipse.swt.widgets.Display;

import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.designRules.DesignRule;
import acmeLib.designRules.DesignRuleEvalException;
import acmeLib.designRules.Scope;
import acmeLib.elements.AcmeAttachment;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.errorManagement.IAcmeErrorListener;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * @author Bradley Schmerl
 * Class to manage reevaluation of design rules. Design rules should change if
 * properties or elements that they come across in their evaluation change.
 * This class keeps a hash table of elements to design rules and their scopes 
 * that should be reevaluated when the element changes.
 * It should also announce rule evaluation results to those interested (e.g., 
 * the rule list view, and annotations to the system view).
 *
 */
public class DesignEvaluationCache implements AcmeModelChangeListener, IAcmeErrorListener {

	public static final int TRUE = 1, FALSE = 0, UNKNOWN = -1, ERROR = -2;

	private Map _cache = new Hashtable(); // maps acmeobject (element and property) to set of rules
	private Map _eltCache = new Hashtable();
	private Map _errorCache = new Hashtable();

	Set _listeners = new HashSet(10);

	//   private PropertyChangeSupport _listeners = new PropertyChangeSupport (this);
	public static final String PROP_RULE_EVALUATION = "$RuleEval$";
	public static final String PROP_ELEMENT_TYPECHECK = "$TypeCheck$";

	private ThreadGroup _evaluationThreads = new ThreadGroup("Rule Evaluations");

	private class DesignRuleInfo {
		public DesignRule _rule;
		public Scope _scope;
		public RuleCache _result;
	}

	private class TypeCheckInfo {
		public AcmeElement _element;
		public RuleCache _result;
	}

	private class EvalBuffer {
		private List evalBuffer = new LinkedList();
		private  boolean	_allowReading = true;


		public synchronized void pushEvaluation(TypeCheckInfo o) {
			 {
				if (evalBuffer.contains(o)) return;
				evalBuffer.add(o);
				notifyAll ();
			}
		}

		public synchronized void pushEvaluation(DesignRuleInfo o) {
			 {
				if (evalBuffer.contains (o)) return;
				evalBuffer.add(o);
				notifyAll ();
			}
		}

		public  synchronized Object pullEvaluation() {
			Object o;
			try {
				while (!_allowReading) wait ();
				if (getBufferSize () == 0) {
					wait ();
					Debug.diagnoseln ("EvalBuffer.pullEvaluation: " + "getBufferSize == " + getBufferSize ());
				}
				if (getBufferSize () == 0) return null;
				o = evalBuffer.get (0);
				evalBuffer.remove (o);
				return o;
			}
			catch (InterruptedException e) {
//				e.printStackTrace ();
				return null;
			}
			catch (Exception e) {
				return null;
			}
		}

		

		public synchronized int getBufferSize() {
			int val;
			
				val = evalBuffer.size();
		
			return val;
		}

		public synchronized void clearBuffer() {
			 {
				evalBuffer.clear();
			}
		}

		/**
		 * 
		 */
		public synchronized void allowReading() {
			_allowReading = true;
			notifyAll ();
		}

		/**
		 * 
		 */
		public synchronized boolean preventReading() {
			boolean orig = _allowReading;
			_allowReading = false;
			return orig;
		}
	}
	
	private EvalBuffer evalBuffer = new EvalBuffer ();
	private static EvalThread _evalThread;
//	private static final Integer _evalLock = new Integer(0);

	private Display _display = null;

	private class EvalThread extends Thread {
		Display _display;

		public EvalThread(ThreadGroup group, String id, Display disp) {
			super(group, id);
			_display = disp;
		}

		private void evaluate(final DesignRuleInfo dr) throws DesignRuleEvalException {
			Scope origScope = dr._rule.getDesignRuleScope();
			Debug.traceln(
				"<evaluate scope=\""
					+ ((AcmeElement) dr._scope).getName()
					+ "\" rule=\""
					+ dr._rule.toString()
					+ "\">");
			AcmeError error = null;
			try {
				boolean result = false;

				dr._rule.setDesignRuleScope(dr._scope);
				AcmeTypeManager.typecheckingErrorStack.getCausalErrors().removeAllItems();
				result = dr._rule.evaluate();
				AcmeVector v = AcmeTypeManager.typecheckingErrorStack.getCausalErrors();
				if (v.getSize() != 0) {
					error = new ExpressionTypecheckError (dr._rule.getDesignRuleExpression(), "Design rule does not typecheck.");
					AcmeTypeManager.typecheckingErrorStack.packageErrorCauses(error);
					throw new DesignRuleEvalException (); 
				}
				if (result) {
					dr._result.setValue(TRUE);
				}
				else {
					dr._result.setValue(FALSE);
				}

			}
			catch (DesignRuleEvalException e) {
				dr._result.setValue(ERROR);
				if (error == null) {
					error = new AcmeError (dr._scope, "Unable to evaluate design rule, perhaps a property does not have a value");
					error.addCause(new AcmeError (dr._rule, e.getMessage()));
				}
				AcmeErrorManager.errorDetected (error);

			}
			finally {
				_display.asyncExec(new Runnable() {
					public void run() {
						fireRuleReEvaluated(dr._rule);
					}
				});
				dr._rule.setDesignRuleScope(origScope);
				Debug.traceln("  <result value=" + dr._result.getValue() + "/>");
				Debug.traceln("</evaluate>");
			}

		}

		
		private void typecheck(final TypeCheckInfo e) {
			Debug.traceln("<typecheck element=\"" + e._element.getQualifiedName() + "\">");
			AcmeElement element = e._element;
			boolean result = element.typecheck();
			if (result) {
				e._result.setValue(TRUE);
			}
			else {
				e._result.setValue(FALSE);
			}
			Debug.traceln("  <result value=" + e._result.getValue() + "/>");
			Debug.traceln("</typecheck>");
		}

		public void run() {
			if (_evalThread != Thread.currentThread ()) return;
			while (true) {
				Object o = evalBuffer.pullEvaluation ();
				try {
					if (o instanceof TypeCheckInfo) {
						typecheck ((TypeCheckInfo )o);
					}
					else if (o instanceof DesignRuleInfo) {
						evaluate ((DesignRuleInfo )o);
					}
				}
				catch (Exception e) {
					e.printStackTrace ();
				}
			}
		}
	}

	


	

	private class RuleCache {
		private int _lastEvaluation;

		public synchronized void setValue(int value) {
			_lastEvaluation = value;
		}

		public synchronized int getValue() {
			return _lastEvaluation;
		}
	}

	/**
	 * Constructor for DesignEvaluationCache.
	 */
	public DesignEvaluationCache() {
		super();
		AcmeErrorManager.addErrorListener(this);
		AcmeStudioChangeManager.registerListener(this);
		_evaluationThreads.setMaxPriority(Thread.MIN_PRIORITY);
		_evalThread = new EvalThread(_evaluationThreads, "Evaluation", Display.getCurrent());
		_evalThread.start ();
	}

	public void clearCachedResult(DesignRule rule, Scope scope) {
		if (_cache.containsKey(rule) && ((Map) _cache.get(rule)).containsKey(scope)) {
			((Map) _cache.get(rule)).remove(scope);
		}
	}

	public int evaluate(DesignRule rule, Scope scope) throws DesignRuleEvalException {
		_display = Display.getCurrent();
		RuleCache cacheResult = null;
		if (_cache.containsKey(rule) && ((Map) _cache.get(rule)).containsKey(scope)) {
			cacheResult = (RuleCache) ((Map) _cache.get(rule)).get(scope);
			
		}
		else {

			cacheResult = new RuleCache();
			cacheResult.setValue(UNKNOWN);

			DesignRuleInfo dr = new DesignRuleInfo();
			dr._rule = rule;
			dr._scope = scope;
			dr._result = cacheResult;
			evalBuffer.pushEvaluation(dr);

			_eltCache.clear();
			if (!_cache.containsKey(rule)) {
				Map scopes = new HashMap();
				scopes.put(scope, cacheResult);
				_cache.put(rule, scopes);
			}
			else {
				((Map) _cache.get(rule)).put(scope, cacheResult);
			}
		}
		return cacheResult.getValue();
	}

	public int typecheck(AcmeElement element) {
		_display = Display.getCurrent();
		RuleCache cacheResult = null;
		if (_eltCache.containsKey(element)) {
			cacheResult = (RuleCache) _eltCache.get(element);
		}
		else {
			cacheResult = new RuleCache();
			cacheResult.setValue(UNKNOWN);
			TypeCheckInfo info = new TypeCheckInfo();
			info._element = element;
			info._result = cacheResult;
			evalBuffer.pushEvaluation(info);

			_eltCache.put(element, cacheResult);
		}
		return cacheResult.getValue();
	}

	public AcmeEnumeration getErrors(AcmeElement element) {
		AcmeEnumeration enum = (AcmeEnumeration) _errorCache.get(element);
		if (enum != null)
			enum.reset();
		return enum;
	}

	/**
	 * @see AcmeStudio.util.AcmeModelChangeListener#changeModel(AcmeStudioChangeEvent)
	 */
	public void changeModel(AcmeStudioChangeManager.AcmeStudioChangeEvent evt) {
		AcmeElement parent = null;
		Object target = evt.target;
		Object source = evt.source;
		switch (evt.getEventType()) {
			case AcmeStudioChangeManager.ASSIGN_FAMILY :
			case AcmeStudioChangeManager.UNASSIGN_FAMILY :
				{
					//                List objectsInSystem = Helper.getAllObjects((AcmeSystem )evt.target);
					//                Iterator i = objectsInSystem. 
					_eltCache.clear();
					_errorCache.clear();
					_cache.clear();
				}
				break;
			case AcmeStudioChangeManager.REMOVE_PORT :
			case AcmeStudioChangeManager.REMOVE_ROLE :
			case AcmeStudioChangeManager.ADD_PORT :
			case AcmeStudioChangeManager.ADD_ROLE :
				resetEvaluationThread(target);
				resetElement(target);
				resetElement(source);
				break;
				// Bugfix #94
				// Name change might affect the type of the parent
			case AcmeStudioChangeManager.SET_NAME :
				{
					resetEvaluationThread(target);
					parent = ((AcmeElement) target).getParent();
					_eltCache.remove(parent);
					_errorCache.remove(parent);
				}

				break;
			case AcmeStudioChangeManager.REMOVE_ATTACHMENT :
			case AcmeStudioChangeManager.ADD_ATTACHMENT :
				{
					AcmeAttachment att = (AcmeAttachment) evt.source;
					resetEvaluationThread(att);
					resetElement((AcmeElement )evt.target);
					resetElement(att.getPort());
					resetElement(att.getRole());
					resetElement(att.getPort().getParent());
					resetElement(att.getRole().getParent());
				}
				break;
			case AcmeStudioChangeManager.UPDATE_VIS_INFO:
				// Do nothing, only visual info has changed
				break;
			default :
				resetEvaluationThread(target);
		}
	}

	public boolean interestedInEvent (int et) {
		return et != AcmeStudioChangeManager.UPDATE_VIS_INFO;
	}
	private void resetEvaluationThread(Object target) {
		boolean orig = evalBuffer.preventReading();
		if (_evalThread != null) {
			synchronized (_evalThread) {
				_evalThread.interrupt();
			}
		}
		evalBuffer.clearBuffer();
		resetElement(target);
		_cache.clear();
		if (orig) evalBuffer.allowReading();
	}

	private void resetElement(Object element) {
		if (element == null)
			return;
		_eltCache.remove(element);
		_errorCache.remove(element);
		if (element instanceof AcmeElement) {
			typecheck((AcmeElement) element);
		}
	}

	public void addPropertyChangeListener(PropertyChangeListener l) {
		_listeners.add(l);
	}

	public void removePropertyChangeListener(PropertyChangeListener l) {
		_listeners.remove(l);
	}

	public synchronized void fireRuleReEvaluated(DesignRule rule) {
		PropertyChangeEvent evt = new PropertyChangeEvent(this, PROP_RULE_EVALUATION, null, rule);
		Iterator i = _listeners.iterator();
		while (i.hasNext()) {
			((PropertyChangeListener) i.next()).propertyChange(evt);
		}

		//       _listeners.firePropertyChange (PROP_RULE_EVALUATION, null, rule);
	}

	public synchronized void fireTypeCheckEvaluated(Object element, AcmeEnumeration errors) {
		PropertyChangeEvent evt = new PropertyChangeEvent(this, PROP_ELEMENT_TYPECHECK, errors, element);
		Iterator i = _listeners.iterator();
		while (i.hasNext()) {
			((PropertyChangeListener) i.next()).propertyChange(evt);
		}
		//        _listeners.firePropertyChange (PROP_ELEMENT_TYPECHECK, errors, element);   
	}

	/* (non-Javadoc)
	 * @see acmeLib.errorManagement.IAcmeErrorListener#clearErrors(acmeLib.elements.AcmeDesign, acmeLib.AcmeObject)
	 */
	public void clearErrors(AcmeDesign associatedDesign, final AcmeObject object) {
		_errorCache.remove(object);

		// Fix up the cache
		RuleCache cacheResult = null;
		if (_eltCache.containsKey(object)) {
			cacheResult = (RuleCache) _eltCache.get(object);
			cacheResult._lastEvaluation = TRUE;
		}
		else {
			cacheResult = new RuleCache();
			cacheResult._lastEvaluation = TRUE;
			_eltCache.put(object, cacheResult);
		}

		getDisplay().asyncExec(new Runnable() {
			public void run() {
				fireTypeCheckEvaluated(object, null);
			}
		});
	}
	
	public void clearError (AcmeDesign associatedDesign,  AcmeObject object, AcmeError error) {
		clearErrors (associatedDesign, object);
	}

	/* (non-Javadoc)
	 * @see acmeLib.errorManagement.IAcmeErrorListener#errorDetected(acmeLib.elements.AcmeDesign, acmeLib.errorManagement.AcmeError)
	 */
	public void errorDetected(AcmeDesign associatedDesign, final AcmeError typecheckError) {
		Object o = typecheckError.getSource();
		if (o instanceof AcmeElement || o instanceof DesignRule || o instanceof DesignAnalysis) {

			AcmeVector v = new AcmeVector();
			v.addItem(typecheckError);

			boolean wasLogging = AcmeErrorManager.getLogErrors();
			AcmeErrorManager.setLogErrors(false);
			_errorCache.put(o, v.enumerate());

			// Fix up the cache
			RuleCache cacheResult = null;
			if (_eltCache.containsKey(o)) {
				cacheResult = (RuleCache) _eltCache.get(o);
				cacheResult._lastEvaluation = FALSE;
			}
			else {
				cacheResult = new RuleCache();
				cacheResult._lastEvaluation = FALSE;
				_eltCache.put(o, cacheResult);
			}
			AcmeErrorManager.setLogErrors(wasLogging);

			getDisplay().asyncExec(new Runnable() {
				public void run() {
					AcmeVector v = new AcmeVector();
					v.addItem(typecheckError);
					fireTypeCheckEvaluated(typecheckError.getSource(), v.enumerate());
				}
			});
		}
	}

	/**
	 * @return Display
	 */
	protected Display getDisplay() {
		if (_display == null) {
			_display = Display.getCurrent();
		}
		return _display;
	}

	/**
	 * 
	 */
	public  void resumeEvaluation() {
		evalBuffer.allowReading ();
	}

	public  void suspendEvaluation () {
		evalBuffer.preventReading ();
	}
}
