/*
 * Created on Apr 14, 2005
 *
 */
package AcmeStudio.util;

import java.util.*;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.Rectangle;


/**
 * @author Bradley Schmerl
 * Created Apr 14, 2005
 * Provides a way to manage images by providing a keyed image cache, reference counting of images, and notification of disposal
 */
public class ManagedImage {
	/** The image cache **/
	private static final Map IMAGES /*<String, ManagedImage>*/ = new HashMap ();

	/** The image being managed **/
	private Image _image;
	/** The number of references being held on this image **/
	private int _references = 0;
	/** The objects that are holding references to the image **/
	private Set _referencers = new HashSet ();
	/** The object that created the image **/
	private Object	_creator;
	/** The IManagedImageListener objects that should be notified when the image is disposed **/
	private Set _listeners = new HashSet ();
	
	/**
	 * Places an image in the image cache
	 * @param key The key to remember the image by
	 * @param image The image to remember
	 */
	public static void put (String key, ManagedImage image) {
		ManagedImage i = (ManagedImage )IMAGES.get (key);
		if (i != null && !i.getImage ().isDisposed()) 
			Debug.errorln ("ManagedImage.put: " + "Replacing an existing image!");
		IMAGES.put (key, image);
	}
	
	/**
	 * Retrieves an image from the image cache
	 * @param key The key that the image was remembered by
	 * @return The remembered image
	 */
	public static ManagedImage get (String key) {
		ManagedImage i = (ManagedImage )IMAGES.get (key);
		// If the image is disposed, let's not remember it and return null
		if (i== null || i.getImage ().isDisposed ()) {
			i = null;
			IMAGES.remove(key);
		}
		return i;
	}
	
	/**
	 * 
	 * @param key
	 * @return true if the cache contains the key
	 */
	public static boolean containsKey (String key) {
		ManagedImage i = (ManagedImage )IMAGES.get (key);
		// If the image has been disposed, then remove it and return false
		if (i != null && i.getImage ().isDisposed ()) {
			IMAGES.remove (key);
			return false;
		}
		return i != null;
	}
	
	/**
	 * Creates a new ManagedImage wrapped around the image 
	 * @param image The image to manage
	 * @param creator The object creating the image
	 */
	public ManagedImage (Image image, Object creator) {
		_image = image;
		_references++;
		_creator = creator;
		_referencers.add(creator);
		if (creator instanceof IManagedImageListener) _listeners.add(creator);
	}
	
	/**
	 * Disposes the image. This decreases the reference count and only disposes the image if the
	 * reference count is 0
	 * @param disposer The object requesting the disposal
	 */
	public void dispose(Object disposer) {
		_references--;
		if (_references <= 0) {
			_image.dispose ();
			// Also remove the image from the cache
			Object key = null;
			for (Iterator iter = IMAGES.entrySet().iterator (); iter.hasNext () && key == null;) {
				Map.Entry entry = (Map.Entry )iter.next ();
				if (entry.getValue() == _image) key = entry.getKey ();
			}
			if (key != null) IMAGES.remove (key);
			// Notify listeners
			fireImageDisposed ();
		}
		if (!_referencers.contains (disposer)) {
			Debug.diagnoseln ("ManagedImage.dispose: " + "image is being disposed by something that doesn't contain a reference:" + disposer);
		}
		_referencers.remove(disposer);
		if (disposer instanceof IManagedImageListener) _listeners.remove (disposer);
	}
	
	/**
	 * Adds a reference to the image. 
	 * @param referencedBy The object that will hold the reference. If this is an instance of IManagedImageListener, then the object
	 * 	  will be notified when the image is disposed
	 */
	public void reference (Object referencedBy) {
		_references++;
		if (_referencers.contains (referencedBy)) {
			Debug.errorln ("ManagedImage.reference: " + "image is already referenced by " + referencedBy);
		}
		_referencers.add (referencedBy);
		if (referencedBy instanceof IManagedImageListener) _listeners.add (referencedBy);
	}
	
	// The following methods are delegated to the image
	public Color getBackground() {
		return _image.getBackground ();
	}
	public Rectangle getBounds() {
		return _image.getBounds ();
	}
	public ImageData getImageData() {
		return _image.getImageData ();
	}
	public boolean isDisposed() {
		return _image.isDisposed ();
	}
	public void setBackground(Color color) {
		_image.setBackground (color);
	}
	public String toString() {
		return _image.toString ();
	}
	
	/**
	 * Returns the image being managed
	 * @return
	 */
	public Image getImage () {
		return _image;
	}
	
	/** 
	 * Adds a listener that wil be notified when the managed image is disposed
	 * @param l the listener
	 */
	public void addListener (IManagedImageListener l) {
		_listeners.add (l);
	}
	
	/**
	 * Removes a listener 
	 * @param l
	 */
	public void removeListener (IManagedImageListener l) {
		_listeners.remove (l);
	}
	
	protected void fireImageDisposed () {
		for (Iterator iter = _listeners.iterator (); iter.hasNext ();) {
			IManagedImageListener l = (IManagedImageListener )iter.next ();
			l.imageDisposed(this);
		}
	}
}
