package AcmeStudio.util;

import java.io.IOException;
import java.util.Comparator;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.runtime.IPath;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.widgets.Composite;
import org.w3c.dom.Element;
import AcmeStudio.AcmeStudioPlugin;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.ArmaniExportable;
import acmeLib.elements.AcmeElement;
import acmeLib.properties.AcmeProperty;

/**
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu) Collection of utility methods
 *         and classes
 */
public class Utils {

	/**
	 * @author Bradley Schmerl
	 * Created Oct 2, 2003
	 */
	public static class  AcmePropertyNameComparator implements Comparator {

		/**
		 * 
		 */
		public AcmePropertyNameComparator() {
			super();
		}

		/* (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		
			public int compare(Object arg0, Object arg1) {
			if (arg0 instanceof AcmeProperty && arg1 instanceof AcmeProperty) {
				return String.CASE_INSENSITIVE_ORDER.compare(
					((AcmeProperty) arg0).getName(),
					((AcmeProperty) arg1).getName());
			}
			return -1;
		}
		

	}
	/**
	 * Returns the null-guarded String representation of the supplied object.
	 * This means that this method will never return <code>null</code>.
	 * 
	 * @param obj
	 *           the object to convert to <code>null</code> -guarded String.
	 * @return the String representation of the obj (<code>obj.toString()</code>,
	 *         or blank if <code>obj</code> is <code>null</code>.
	 */
	public static String nullGuard(Object obj) {
		if (obj == null) {
			return "";
		}
		else {
			return obj.toString();
		}
	}

	/**
	 * Returns the zero-guarded integer of the supplied value. This method will
	 * never return 0 or negative value.
	 * 
	 * @param obj
	 *           the object to convert to <code>null</code> -guarded String.
	 * @return the value if the value is greater than zero, or 1 otherwise.
	 */
	public static int zeroGuard(int value) {
		if (value > 0) {
			return value;
		}
		else {
			return 1;
		}
	}

	/**
	 * Given a path, validate that it's in the family subdirectory of the
	 * current project (should the project be made a parameter)?
	 * 
	 * @param path
	 *           the path to validate
	 * @return the folder representing the valid family subfolder
	 */
	public static IContainer validateFamilyPath(IPath path) {
		IFolder dir = null;
		// BugFix #146:
		//   Cannot assume that family file has the right path
		//   because the visualization style could be created
		//   from the family editor of the instance file,
		//   so need to validate path to make sure it's under
		//   the family subdir.
		if (path != null) {
			String FAM_DIR = AcmeStudioPlugin.getResourceString("AcmeStudio.family.subdir");
			if (path.segmentCount() < 1) {
				path = null;
			}
			else if (path.segmentCount() == 1) {
				// append a families subdir
				path = path.append(FAM_DIR);
			}
			else {
				// segment count must be 2 or more
				if (!path.segment(1).equals(FAM_DIR)) {
					// not good, need to insert a families subdir
					IPath newPath = path.uptoSegment(1);
					path = newPath.append(FAM_DIR).append(path.removeFirstSegments(1));
				}
			}
			if (path != null) {
				dir = AcmeStudioPlugin.getRoot().getFolder(path);
			}
		}
		return dir;
	}

	/**
	 * 
	 * @author schmerl Used for sorting collections of AcmeElements */
	public static class AcmeElementNameComparator implements Comparator {

		/*
		 * (non-Javadoc) @see java.util.Comparator#compare(java.lang.Object,
		 * java.lang.Object)
		 */
		public int compare(Object arg0, Object arg1) {
			if (arg0 instanceof AcmeElement && arg1 instanceof AcmeElement) {
				return String.CASE_INSENSITIVE_ORDER.compare(
					((AcmeElement) arg0).getName(),
					((AcmeElement) arg1).getName());
			}
			return -1;
		}

		public AcmeElementNameComparator() {
		}

	}

	public static final ArmaniExportable EMPTY_RULE = new ArmaniExportable() {

		public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {

		}

		public void exportAcme(AcmeOutputStream s) throws IOException, AcmeException {
		}

	};
	
	public static class AutoApplyTextCellEditor extends TextCellEditor {
		public AutoApplyTextCellEditor(Composite parent) {
			super(parent);
		}
		public void fireApplyEditorValue() {
			super.fireApplyEditorValue();
		}
	}

	public static AutoApplyTextCellEditor generateAutoApplyTextCellEditor (Composite c) {
		final AutoApplyTextCellEditor tce = new AutoApplyTextCellEditor (c);
		tce.getControl ().addFocusListener (new FocusAdapter () {
			public void focusLost (FocusEvent e) {
				tce.fireApplyEditorValue ();
			}
		});
		return tce;
	}

	public static void safeSetXMLAttribute (Element e, String name, String value) {
		if (value != null) 
			e.setAttribute (name, value);
		else
			e.setAttribute (name, "");
	}

	public static int nullToInt(Object p)
	{
	    if (p == null)
	        return 0;
	    return 1;
	}
	
	public static ImageData rotateImage(ImageData srcData, int direction) {
		int bytesPerPixel = srcData.bytesPerLine / srcData.width;
		int destBytesPerLine = (direction == SWT.DOWN)? srcData.width * bytesPerPixel : srcData.height * bytesPerPixel;
		byte[] newData = new byte[srcData.data.length];
		int width = 0, height = 0;
		for (int srcY = 0; srcY < srcData.height; srcY++) {
			for (int srcX = 0; srcX < srcData.width; srcX++) {
				int destX = 0, destY = 0, destIndex = 0, srcIndex = 0;
				switch (direction){
					case SWT.LEFT: // left 90 degrees
						destX = srcY;
						destY = srcData.width - srcX - 1;
						width = srcData.height;
						height = srcData.width; 
						break;
					case SWT.RIGHT: // right 90 degrees
						destX = srcData.height - srcY - 1;
						destY = srcX;
						width = srcData.height;
						height = srcData.width; 
						break;
					case SWT.DOWN: // 180 degrees
						destX = srcData.width - srcX - 1;
						destY = srcData.height - srcY - 1;
						width = srcData.width;
						height = srcData.height; 
						break;
				}
				destIndex = (destY * destBytesPerLine) + (destX * bytesPerPixel);
				srcIndex = (srcY * srcData.bytesPerLine) + (srcX * bytesPerPixel);
				System.arraycopy(srcData.data, srcIndex, newData, destIndex, bytesPerPixel);
			}
		}
		// destBytesPerLine is used as scanlinePad to ensure that no padding is required
		return new ImageData(width, height, srcData.depth, srcData.palette, destBytesPerLine, newData);
	}
	
}
