package AcmeStudio.wizards;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.*;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.ui.*;
import org.eclipse.ui.actions.WorkspaceModifyOperation;
import org.eclipse.ui.part.FileEditorInput;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASContext;
import AcmeStudio.ui.editors.MultiPageAcmeEditor;
import AcmeStudio.util.Helper;

/**
 * This is a sample new wizard. Its role is to create a new file 
 * resource in the provided container. If the container resource
 * (a folder or a project) is selected in the workspace 
 * when the wizard is opened, it will accept it as the target
 * container. The wizard creates one file with the extension
 * "acme". If a sample multi-page editor (also available
 * as a template) is registered for the same extension, it will
 * be able to open it.
 */

abstract public class AcmeStudioNewWizard extends Wizard implements INewWizard {
    protected AcmeStudioNewWizardPage [] pages;
    protected ISelection selection;
    protected IProject _origProject;
    /**
     * Constructor for AcmeStudioNewWizard.
     */
    public AcmeStudioNewWizard() {
        super();
        setNeedsProgressMonitor(true);
        _origProject = ASContext.current().getActiveProject();
    }

    /**
     * Adding the page to the wizard.
     */
    abstract public void addPages();

    /**
     * This method is called when 'Finish' button is pressed in
     * the wizard. We will create an operation and run it
     * using wizard as execution context.
     */
    public boolean performFinish() {
        final String containerName = pages[0].getContainerName();
        final String fileName = pages[0].getFileName();
        WorkspaceModifyOperation op = new WorkspaceModifyOperation() {
            public void execute (final IProgressMonitor monitor)
            throws CoreException {
                try {
                    performFinish(containerName, fileName, monitor);
                } finally {
                    monitor.done();
                }
            }
        };
        try {
            getContainer().run(true, false, op);
        } catch (InterruptedException e) {
            return false;
        } catch (InvocationTargetException e) {
            Throwable realException = e.getTargetException();
            MessageDialog.openError(getShell(), "Error",
                realException.getMessage());
            return false;
        }
        return true;
    }

    /**
     * The worker method. It will find the container, create the file if missing
     * or just replace its contents, and open the editor on the newly created file
     */
    protected void performFinish (
        String containerName,
        String name,
        IProgressMonitor monitor)
    throws CoreException {
        // create the acme file
        String fileName = Helper.convertToAcmeFilename(name);
        monitor.beginTask("Creating " + fileName, 2);
        IPath contPath = new Path(containerName);
        IContainer cont = AcmeStudioPlugin.getRoot().getProject(contPath.segment(0));
        if (cont == null || !cont.exists()) {
            throwCoreException("The Project \"" + contPath.segment(0) + "\" does not exist!");
        }
        if (contPath.segmentCount() > 1) {
            IFolder folder = cont.getFolder(contPath.removeFirstSegments(1));
            if (! folder.exists()) {
                folder.create(true, true, monitor);
            }
            cont = folder;
        }
        final IFile f = cont.getFile(new Path (fileName));
        try {
            InputStream stream = generateNewFileStream(name);
            if (f.exists()) {
                f.setContents(stream, true, true, monitor);
            } else {
                f.create(stream, true, monitor);
            }
            stream.close();
        } catch (IOException e) {
        }
        final IEditorInput input = new FileEditorInput (f);
        monitor.worked(1);
        monitor.setTaskName("Opening file for editing...");
        getShell().getDisplay().asyncExec(new Runnable() {
            public void run() {
                IWorkbenchPage page =
                    PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
                try {
                    page.openEditor(input, "AcmeStudio.ui.editors.MultiPageAcmeEditor");
                    IEditorPart editor = page.getActiveEditor();
                    if (editor instanceof MultiPageAcmeEditor) {
                        MultiPageAcmeEditor acmeEditor =
                            (MultiPageAcmeEditor) editor;
                        setActivePage (acmeEditor);
//                        acmeEditor.setActiveAcmePage(acmeEditor.FAMILY_PAGE);
                    }
                } catch (PartInitException e) {
                }
            }
        });
        monitor.worked(1);
        
        doFinish(containerName, name, monitor);
    }

    /**
     * The subclass-specific worker method.
     */
    
    abstract protected void setActivePage (MultiPageAcmeEditor editor);
    
    abstract protected void doFinish(
        String containerName,
        String fileName,
        IProgressMonitor monitor)
    throws CoreException;

    abstract protected InputStream generateNewFileStream (String name);

    protected void throwCoreException(String message) throws CoreException {
        IStatus status =
            new Status(IStatus.ERROR, "AcmeStudio", IStatus.OK, message, null);
        throw new CoreException(status);
    }


    /**
     * We will accept the selection in the workbench to see if
     * we can initialize from it.
     * @see IWorkbenchWizard#init(IWorkbench, IStructuredSelection)
     */
    public void init(IWorkbench workbench, IStructuredSelection selection) {
//        if (selection == null || selection.isEmpty()) {
//            performCancel();
//            return;
//        }
        this.selection = selection;
    }
    /**
     * @see org.eclipse.jface.wizard.IWizard#performCancel()
     */
    public boolean performCancel() {
        ASContext.current().setActiveProject (_origProject);
        return super.performCancel();
    }


}
