package AcmeStudio.wizards;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.IDialogPage;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.dialogs.ContainerSelectionDialog;

import AcmeStudio.AcmeStudioPlugin;
import AcmeStudio.core.ASContext;
import AcmeStudio.core.resource.ASDesign;

/**
 * The "New" wizard page allows setting the container for
 * the new file as well as the file name. The page
 * will only accept file name without the extension OR
 * with the extension that matches the expected one (acme).
 */

abstract public class AcmeStudioNewWizardPage extends WizardPage {

    protected final String FAMDIR = AcmeStudioPlugin.getResourceString("AcmeStudio.family.subdir");

	protected Text _containerText;
	protected Text _nameText;
	protected ISelection selection;
    protected String _objToCreate = "File";
    protected String _defaultName = "";
    protected String _description = "";
    
	/**
	 * Constructor for SampleNewWizardPage.
	 * @param pageName
	 */
	public AcmeStudioNewWizardPage(ISelection selection) {
		super("wizardPage");
		this.selection = selection;
        this.updateDescription();
        this.setDescription(_description);
	}

	/**
	 * @see IDialogPage#createControl(Composite)
	 */
	public void createControl(Composite parent) {
		Composite container = new Composite(parent, SWT.NULL);

		GridLayout layout = new GridLayout(3, false);
//		layout.numColumns = 3;
		layout.verticalSpacing = 9;
        container.setLayout(layout);
        Label cLabel = new Label(container, SWT.NULL);
        cLabel.setText("&Container:");

		_containerText = new Text(container, SWT.BORDER | SWT.SINGLE);
		GridData gd1 = new GridData(GridData.FILL_HORIZONTAL);

		_containerText.setLayoutData(gd1);
        _containerText.setText("/");
        // BRS: Avoid eclipse bug that changes grid layout when something
        //      is selected in the text before the text is displayed
//        _containerText.setSelection(getContainerName().length());

		Button button = new Button(container, SWT.PUSH);
		button.setText("&Browse...");
		button.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				handleBrowse();
			}
		});
		Label oLabel = new Label(container, SWT.NULL);
		oLabel.setText("&"+_objToCreate+" name:");

		_nameText = new Text(container, SWT.BORDER | SWT.SINGLE);
		GridData gd2 = new GridData(GridData.FILL_HORIZONTAL);
		_nameText.setLayoutData(gd2);
		initialize();
        updateContainer();  // enforce any container path rule
		dialogChanged();
        
        _nameText.addModifyListener(new ModifyListener() {
            public void modifyText(ModifyEvent e) {
                if (e.widget == _nameText)
                    dialogChanged();
            }
        });

        _containerText.addModifyListener(new ModifyListener() {
            public void modifyText(ModifyEvent e) {
                if (e.widget == _containerText) {
                    // Bug fix# 364:  don't update container with every modify 
                    dialogChanged();
                }
            }
        });
        // Bug fix# 364:  update container only when lose focus
        _containerText.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
            }
            public void focusLost(FocusEvent e) {
                if (e.widget == _containerText) {
                    // upon leaving text field, enforce any container path rule
                    updateContainer();
                    dialogChanged();
                }
            }
        });

        setControl(container);
	}
	
	/**
	 * Tests if the current workbench selection is a suitable
	 * container to use.
	 */
	protected void initialize() {
        // Bug fix #108
        // Set container by default to what the ASContext object claims to
        // be the project, then let the current selection determine the project
        // to show, in case it is different from ASContext's current project
        IProject proj = ASContext.current().getActiveProject();
        if (proj != null) {
            _containerText.setText(proj.getFullPath().toString());
        }
		if (selection!=null && selection.isEmpty()==false && selection instanceof IStructuredSelection) {
			IStructuredSelection ssel = (IStructuredSelection)selection;
			if (ssel.size()>1) return;
			Object obj = ssel.getFirstElement();
			if (obj instanceof IResource) {
				IContainer container;
				if (obj instanceof IContainer)
					container = (IContainer)obj;
				else
					container = ((IResource)obj).getParent();
				_containerText.setText(container.getFullPath().toString());
			}
		}
		_nameText.setText(_defaultName);
	}
	
	/**
	 * Uses the standard container selection dialog to
	 * choose the new value for the container field.
	 */

	private void handleBrowse() {
		ContainerSelectionDialog dialog =
			new ContainerSelectionDialog(
				getShell(),
				ResourcesPlugin.getWorkspace().getRoot(),
				false,
				"Select new container");
		if (dialog.open() == ContainerSelectionDialog.OK) {
			Object[] result = dialog.getResult();
			if (result.length == 1) {
                Object obj = result[0];
                if (obj instanceof Path) {
                    Path path = (Path)obj;
                    _containerText.setText(path.toString());
                } else if (obj instanceof IContainer) {
    				IContainer container = (IContainer) result[0];
    				_containerText.setText(container.getFullPath().toString());
                }
			}
		}
	}
	
	/**
	 * Ensures that both text fields are set.
	 */
	private void dialogChanged() {
		String container = getContainerName();
		String fileName = getFileName();

		if (container.length() == 0) {
			updateStatus(_objToCreate+" container must be specified");
			return;
		}
		if (fileName.length() == 0) {
			updateStatus(_objToCreate+" name must be specified");
			return;
		}
        if (fileName.indexOf(".") >= 0) {
            updateStatus("Invalid "+_objToCreate+" name");
            return;
        }
        container = getContainerName();
        IPath contPath = new Path(getContainerName());
        if (contPath.segmentCount() == 0) {  // can't check
            return;
        }
        IContainer cont = AcmeStudioPlugin.getRoot().getProject(contPath.segment(0));
        if (cont == null || !cont.exists()) {
            updateStatus ("The Project \"" + contPath.segment(0) + "\" does not exist!");
            return;
        }
        if (contPath.segmentCount() > 1) {
            cont = cont.getFolder(contPath.removeFirstSegments(1));
        }
        if (! cont.exists()) {
            updateStatus ("The Project folder path " + container + " does not exist");
            return;
        } else {
            ASContext.current ().setActiveProject (cont.getProject());
            IFile f = cont.getFile(new Path (fileName + "." + ASDesign.FILE_EXT));
            if (f.exists ()) {
                updateStatus ("File " + f.getFullPath() + " already exists");
                return;
            }
        }

        updateDescription();
        setDescription(_description);
		updateStatus(null);
	}

    /**
     * Can be overridden by subclass.  Gets called before updateDescription().
     */
    protected void updateContainer() {
    }

    abstract protected void updateDescription();
    
	private void updateStatus(String message) {
		setErrorMessage(message);
		setPageComplete(message == null);
	}

	public String getContainerName() {
		String retVal = null;
        if (_containerText != null) {
            retVal = _containerText.getText();
        }
        return retVal;
	}

	public String getFileName() {
        String retVal = null;
        if (_nameText != null) {
            retVal = _nameText.getText();
        }
        return retVal;
	}
 

}