/*******************************************************************************
 * Copyright (c) 2004 Elias Volanakis. * All rights reserved. This program and
 * the accompanying materials * are made available under the terms of the
 * Common Public License v1.0 * which accompanies this distribution, and is
 * available at * http://www.eclipse.org/legal/cpl-v10.html * * Contributors:
 * *Elias Volanakis - initial API and implementation 
 ******************************************************************************/
package AcmeStudioShapeExtension.model;

import java.util.ArrayList;
import java.util.List;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.jface.viewers.ICellEditorValidator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.ui.views.properties.ColorPropertyDescriptor;
import org.eclipse.ui.views.properties.ComboBoxPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.TextPropertyDescriptor;
import AcmeStudioShapeExtension.AnnotationConstants;
import acmeLib.elements.AcmeSystem;
/**
 * Abstract prototype of a shape. Has a size (width and height), a location (x
 * and y position) and a list of incoming and outgoing connections. Use
 * subclasses to instantiate a specific shape.
 * 
 * @see org.eclipse.gef.examples.shapes.model.RectangleAnnotation
 * @see org.eclipse.gef.examples.shapes.model.EllipseAnnotation
 * @author Elias Volanakis
 */
public abstract class Annotation extends AnnotationElement {

	/**
	 * A static array of property descriptors. There is one IPropertyDescriptor
	 * entry per editable property.
	 * 
	 * @see #getPropertyDescriptors()
	 * @see #getPropertyValue(Object)
	 * @see #setPropertyValue(Object, Object)
	 */
	protected transient  List descriptors;
	
	//IPropertyDescriptor []	descriptors;
	/**
	 * ID for the Height property value (used for by the corresponding property
	 * descriptor).
	 */
	private static final String				HEIGHT_PROP				= "Shape.Height";
	/** Property ID to use when the location of this shape is modified. */
	public static final String				LOCATION_PROP			= "Shape.Location";
	private static final long				serialVersionUID		= 1;
	/** Property ID to use then the size of this shape is modified. */
	public static final String				SIZE_PROP				= "Shape.Size";
	/** Property ID to use when the list of outgoing connections is modified. */
	public static final String				SOURCE_CONNECTIONS_PROP	= "Shape.SourceConn";
	/** Property ID to use when the list of incoming connections is modified. */
	public static final String				TARGET_CONNECTIONS_PROP	= "Shape.TargetConn";
	/**
	 * ID for the Width property value (used for by the corresponding property
	 * descriptor).
	 */
	private static final String				WIDTH_PROP				= "Shape.Width";
	/**
	 * ID for the X property value (used for by the corresponding property
	 * descriptor).
	 */
	private static final String				XPOS_PROP				= "Shape.xPos";
	/**
	 * ID for the Y property value (used for by the corresponding property
	 * descriptor).
	 */
	private static final String				YPOS_PROP				= "Shape.yPos";
	
	private static final String				BACKGROUND_COLOR_PROP				        = "Shape.fillColor";
	
	private static final String 				USE_FILL_PROP						= "Shape.fill";
	
	private static final String				OUTLINE_STYLE_PROP	= "Shape.outlineStyle";
	
	private static final String 				OUTLINE_WIDTH_PROP   = "Shape.outlineWidth";
	
	private static final String 				FOREGROUND_COLOR_PROP = "Shape.outlineColor";
	
	
	/** Location of this shape. */
	private Point							location				= new Point (0, 0);
	/** Size of this shape. */
	protected Dimension						size					= new Dimension (50, 50);
	
	private RGB backgoundRGB = AnnotationConstants.DEFAULT_BACKGROUND_COLOR.getRGB ();
	
	private RGB foregroundRGB = AnnotationConstants.DEFAULT_FOREGROUND_COLOR.getRGB ();
	
	private Boolean useFill = new Boolean (AnnotationConstants.DEFAULT_FILL_MODE);
	
	private Integer lineWidth = new Integer (1);
	
	private Integer lineStyle = new Integer (SWT.LINE_SOLID);
	
	/** List of outgoing Connections. */
	private List							sourceConnections		= new ArrayList ();
	/** List of incoming Connections. */
	private List							targetConnections		= new ArrayList ();
	transient private AcmeSystem	_system;

	
	/**
	 * 
	 */
	public Annotation() {
		initializeDescriptors ();
		
		//		((PropertyDescriptor )descriptors [0]).setValidator(intValidator);
//		((PropertyDescriptor )descriptors [1]).setValidator(intValidator);
//		((PropertyDescriptor )descriptors [2]).setValidator(intValidator);
//		((PropertyDescriptor )descriptors [3]).setValidator(intValidator);
//		((PropertyDescriptor )descriptors [6]).setValidator(intValidator);

	}
	
	/**
	 * 
	 */
	protected void initializeDescriptors() {
		descriptors = new ArrayList (4);
		
		descriptors.add(new TextPropertyDescriptor (XPOS_PROP, "X"));
		descriptors.add (new TextPropertyDescriptor (YPOS_PROP, "Y"));
		descriptors.add (new TextPropertyDescriptor (WIDTH_PROP, "Width"));
		descriptors.add (new TextPropertyDescriptor (HEIGHT_PROP, "Height"));
		descriptors.add (new ColorPropertyDescriptor (BACKGROUND_COLOR_PROP, "Foreground Color")); 
	    descriptors.add (new ComboBoxPropertyDescriptor (USE_FILL_PROP, "Fill Shape", new String [] {"No", "Yes"}));
		descriptors.add (new ComboBoxPropertyDescriptor (OUTLINE_STYLE_PROP, "Line Style", new String [] {"Solid", "Dashed", "Dotted", "Dash-Dot", "Dash-Dot-Dot"} ));
	    descriptors.add (new TextPropertyDescriptor (OUTLINE_WIDTH_PROP, "Line Width"));
		descriptors.add (new ColorPropertyDescriptor (FOREGROUND_COLOR_PROP, "Background Color"));
			// use a custom cell editor validator for all four array entries
		ICellEditorValidator intValidator = new ICellEditorValidator () {
			
			public String isValid(Object value) {
				int intValue = -1;
				try {
					intValue = Integer.parseInt ((String )value);
				}
				catch (NumberFormatException exc) {
					return "Not a number";
				}
				return (intValue >= 0) ? null : "Value must be >=  0";
			}
		};
	}

	///**
	// * Add an incoming or outgoing connection to this shape.
	// * @param conn a non-null connection instance
	// * @throws IllegalArgumentException if the connection is null or has not
	// distinct endpoints
	// */
	//void addConnection(Connection conn) {
	//	if (conn == null || conn.getSource() == conn.getTarget()) {
	//		throw new IllegalArgumentException();
	//	}
	//	if (conn.getSource() == this) {
	//		sourceConnections.add(conn);
	//		firePropertyChange(SOURCE_CONNECTIONS_PROP, null, conn);
	//	}
	//	else if (conn.getTarget() == this) {
	//		targetConnections.add(conn);
	//		firePropertyChange(TARGET_CONNECTIONS_PROP, null, conn);
	//	}
	//}
	/**
	 * Return a pictogram (small icon) describing this model element. Children
	 * should override this method and return an appropriate Image.
	 * 
	 * @return a 16x16 Image or null
	 */
	public abstract Image getIcon();

	/**
	 * Return the Location of this shape.
	 * 
	 * @return a non-null location instance
	 */
	public Point getLocation() {
		return location.getCopy ();
	}

	/**
	 * Returns an array of IPropertyDescriptors for this shape.
	 * <p>
	 * The returned array is used to fill the property view, when the edit-part
	 * corresponding to this model element is selected.
	 * </p>
	 * 
	 * @see #descriptors
	 * @see #getPropertyValue(Object)
	 * @see #setPropertyValue(Object, Object)
	 */
	public IPropertyDescriptor [] getPropertyDescriptors() {
		if (descriptors == null) initializeDescriptors ();
		IPropertyDescriptor [] des = new IPropertyDescriptor [descriptors.size()];
		for (int i = 0; i < des.length; i++) {
			des [i] = (IPropertyDescriptor )descriptors.get (i);
		}
		return des;
	}

	/**
	 * Return the property value for the given propertyId, or null.
	 * <p>
	 * The property view uses the IDs from the IPropertyDescriptors array to
	 * obtain the value of the corresponding properties.
	 * </p>
	 * 
	 * @see #descriptors
	 * @see #getPropertyDescriptors()
	 */
	public Object getPropertyValue(Object propertyId) {
		if (XPOS_PROP.equals (propertyId)) {
			return Integer.toString (location.x);
		}
		if (YPOS_PROP.equals (propertyId)) {
			return Integer.toString (location.y);
		}
		if (HEIGHT_PROP.equals (propertyId)) {
			return Integer.toString (size.height);
		}
		if (WIDTH_PROP.equals (propertyId)) {
			return Integer.toString (size.width);
		}
		if (BACKGROUND_COLOR_PROP.equals (propertyId)) 
			return backgoundRGB;
		if (FOREGROUND_COLOR_PROP.equals(propertyId)) {
			return foregroundRGB;
		}
		if (OUTLINE_STYLE_PROP.equals (propertyId)) {
			return new Integer (lineStyle.intValue() - 1);
		}
		if (OUTLINE_WIDTH_PROP.equals (propertyId)) {
			return lineWidth.toString ();
		}
		if (USE_FILL_PROP.equals (propertyId)) {
			if (useFill.booleanValue ()) {
				return new Integer (1);
			}
			else return new Integer (0);
		}
		return super.getPropertyValue (propertyId);
	}

	/**
	 * Return the Size of this shape.
	 * 
	 * @return a non-null Dimension instance
	 */
	public Dimension getSize() {
		return size.getCopy ();
	}

	/**
	 * Return a List of outgoing Connections.
	 */
	public List getSourceConnections() {
		return new ArrayList (sourceConnections);
	}

	/**
	 * Return a List of incoming Connections.
	 */
	public List getTargetConnections() {
		return new ArrayList (targetConnections);
	}

	///**
	// * Remove an incoming or outgoing connection from this shape.
	// * @param conn a non-null connection instance
	// * @throws IllegalArgumentException if the parameter is null
	// */
	//void removeConnection(Connection conn) {
	//	if (conn == null) {
	//		throw new IllegalArgumentException();
	//	}
	//	if (conn.getSource() == this) {
	//		sourceConnections.remove(conn);
	//		firePropertyChange(SOURCE_CONNECTIONS_PROP, null, conn);
	//	}
	//	else if (conn.getTarget() == this) {
	//		targetConnections.remove(conn);
	//		firePropertyChange(TARGET_CONNECTIONS_PROP, null, conn);
	//	}
	//}
	/**
	 * Set the Location of this shape.
	 * 
	 * @param newLocation
	 *                     a non-null Point instance
	 * @throws IllegalArgumentException
	 *                     if the parameter is null
	 */
	public void setLocation(Point newLocation) {
		if (newLocation == null) {
			throw new IllegalArgumentException ();
		}
		location.setLocation (newLocation);
		firePropertyChange (LOCATION_PROP, null, location);
	}

	/**
	 * Set the property value for the given property id. If no matching id is
	 * found, the call is forwarded to the superclass.
	 * <p>
	 * The property view uses the IDs from the IPropertyDescriptors array to set
	 * the values of the corresponding properties.
	 * </p>
	 * 
	 * @see #descriptors
	 * @see #getPropertyDescriptors()
	 */
	public void setPropertyValue(Object propertyId, Object value) {
		if (XPOS_PROP.equals (propertyId)) {
			int x = Integer.parseInt ((String )value);
			setLocation (new Point (x, location.y));
		}
		else if (YPOS_PROP.equals (propertyId)) {
			int y = Integer.parseInt ((String )value);
			setLocation (new Point (location.x, y));
		}
		else if (HEIGHT_PROP.equals (propertyId)) {
			int height = Integer.parseInt ((String )value);
			setSize (new Dimension (size.width, height));
		}
		else if (WIDTH_PROP.equals (propertyId)) {
			int width = Integer.parseInt ((String )value);
			setSize (new Dimension (width, size.height));
		}
		else if (BACKGROUND_COLOR_PROP.equals (propertyId)) {
			RGB fill = (RGB )value;
			setBackgroundColor (fill);
		}
		else if (USE_FILL_PROP.equals (propertyId)) {
			int val = ((Integer )value).intValue();
			if (val == 0) {
				setUseFill(Boolean.FALSE);
			}
			else {
				setUseFill (Boolean.TRUE);
			}
		}
		else if (FOREGROUND_COLOR_PROP.equals (propertyId)) {
			RGB lineCol = (RGB )value;
			setForegroundColor (lineCol);
		}
		else if (OUTLINE_STYLE_PROP.equals (propertyId)) {
			int style = ((Integer )value).intValue()+1;
			setLineStyle(new Integer (style));
		}
		else if (OUTLINE_WIDTH_PROP.equals (propertyId)) {
			int width = Integer.parseInt ((String )value);
			setLineWidth (new Integer (width));
		}
		else {
			super.setPropertyValue (propertyId, value);
		}
	}

	/**
	 * Set the Size of this shape. Will not modify the size if newSize is null.
	 * 
	 * @param newSize
	 *                     a non-null Dimension instance or null
	 */
	public void setSize(Dimension newSize) {
		if (newSize != null) {
			size.setSize (newSize);
			firePropertyChange (SIZE_PROP, null, size);
		}
	}

	/**
	 * @return
	 */
	public AcmeSystem  getParentSystem() {
		return _system;
	}
	/**
	 * @param system The system to set.
	 */
	public  void setParentSystem(AcmeSystem system) {
		_system = system;
	}
	/**
	 * @return Returns the fillColor.
	 */
	public RGB getBackgroundColor() {
		return backgoundRGB;
	}
	/**
	 * @param fillColor The fillColor to set.
	 */
	public void setBackgroundColor(RGB fillColor) {
		if (fillColor == null) throw new IllegalArgumentException ();
		this.backgoundRGB = fillColor;
		firePropertyChange(BACKGROUND_COLOR_PROP, null, fillColor);
	}
	/**
	 * @return Returns the lineColor.
	 */
	public RGB getForegroundColor() {
		return foregroundRGB;
	}
	/**
	 * @param lineColor The lineColor to set.
	 */
	public void setForegroundColor(RGB lineColor) {
		AnnotationElement.checkArgument (lineColor);
		this.foregroundRGB = lineColor;
		firePropertyChange (FOREGROUND_COLOR_PROP, null, lineColor);
	}
	/**
	 * @return Returns the lineStyle.
	 */
	public Integer getLineStyle() {
		return lineStyle;
	}
	/**
	 * @param lineStyle The lineStyle to set.
	 */
	public void setLineStyle(Integer lineStyle) {
		AnnotationElement.checkArgument (lineStyle);
		this.lineStyle = lineStyle;
		firePropertyChange (OUTLINE_STYLE_PROP, null, lineStyle);
	}
	/**
	 * @return Returns the lineWidth.
	 */
	public  Integer getLineWidth() {
		return lineWidth;
	}
	/**
	 * @param lineWidth The lineWidth to set.
	 */
	public void setLineWidth(Integer lineWidth) {
		AnnotationElement.checkArgument(lineWidth);
		this.lineWidth = lineWidth;
		firePropertyChange(OUTLINE_WIDTH_PROP, null, lineWidth);
	}
	/**
	 * @return Returns the useFill.
	 */
	public Boolean getUseFill() {
		return useFill;
	}
	/**
	 * @param useFill The useFill to set.
	 */
	public void setUseFill(Boolean useFill) {
		AnnotationElement.checkArgument (useFill);
		this.useFill = useFill;
		firePropertyChange(USE_FILL_PROP, null, useFill);
	}
}