/*
 * Created on Oct 20, 2004
 *
 */
package AcmeStudioShapeExtension.model;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import org.eclipse.draw2d.Bendpoint;
import org.eclipse.draw2d.ColorConstants;
import org.eclipse.draw2d.Graphics;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.ui.views.properties.*;
import acmeLib.elements.AcmeSystem;


/**
 * @author Bradley Schmerl
 * Created Oct 20, 2004
 */
public class LineAnnotation extends AnnotationElement {
	public static final Integer DOTTED = new Integer (Graphics.LINE_DOT);
	public static final Integer DASHED = new Integer (Graphics.LINE_DASH);
	public static final Integer SOLID = new Integer (Graphics.LINE_SOLID);
	
	public static final String LINESTYLE_PROP = "Line.Style";
	public static final String LINECOLOR_PROP = "Line.Color";
	public static final String LINEWIDTH_PROP = "Line.Width";
	public static final String BENDPOINTS_PROP = "Line.Bendpoints";
	public static final String START_PROP = "Line.Start";
	public static final String END_PROP = "Line.End";
	
	private static final long serialVersionUID = 1;
	
	protected transient List descriptors;
	
	protected List bendpoints = new LinkedList ();
	protected Point start, end;
	private transient AcmeSystem _system;
	private RGB	lineColorRGB = ColorConstants.black.getRGB();
	private Integer	lineWidth = new Integer (1);
	private Integer	lineStyle = new Integer (Graphics.LINE_SOLID);
	
	public LineAnnotation () {
		initializeDescriptors ();
	}
	
	protected void initializeDescriptors () {
		descriptors = new ArrayList (4);
		descriptors.add (new ColorPropertyDescriptor (LINECOLOR_PROP, "Color"));
		descriptors.add (new ComboBoxPropertyDescriptor (LINESTYLE_PROP, "Style", new String [] {"Solid", "Dashed", "Dotted", "Dash-Dot", "Dash-Dot-Dot"}));
		descriptors.add (new TextPropertyDescriptor (LINEWIDTH_PROP, "Width"));
//		descriptors.add (new PropertyDescriptor (START_PROP, "Start"));
//		descriptors.add (new PropertyDescriptor (END_PROP, "End"));
//		descriptors.add (new PropertyDescriptor (POINTS_PROP, "Points"));
	}
	
	
	/**
	 * @return
	 * @see AcmeStudioShapeExtension.model.AnnotationElement#getPropertyDescriptors()
	 **/
	public IPropertyDescriptor [] getPropertyDescriptors() {
		if (descriptors == null) initializeDescriptors();
		IPropertyDescriptor [] des = new IPropertyDescriptor [descriptors.size ()];
		for (int i = 0; i < des.length; i++) {
			des [i] = (IPropertyDescriptor )descriptors.get (i);
		}
		return des;
	}
	
	public void setStart (Point point) {
		checkArgument(point);
		this.start = point;
		firePropertyChange(START_PROP, null, start);
	}
	
	public Point getStart () {
		return start;
	}
	
	public void setEnd (Point point) {
		checkArgument (point);
		this.end = point;
		firePropertyChange (END_PROP, null, end);
	}
	
	public Point getEnd () {
		return end;
	}
	
	
	/**
	 * @param id
	 * @return
	 * @see AcmeStudioShapeExtension.model.AnnotationElement#getPropertyValue(java.lang.Object)
	 **/
	public Object getPropertyValue(Object propertyId) {
		if (LINECOLOR_PROP.equals (propertyId))
			return lineColorRGB;
		else if (LINEWIDTH_PROP.equals (propertyId)) {
			return lineWidth.toString();
		}
		else if (LINESTYLE_PROP.equals (propertyId)) {
			return new Integer (lineStyle.intValue () - 1);
		}
		else if (START_PROP.equals(propertyId)) {
			return new LocationPropertySource (getStart());
		}
		else if (END_PROP.equals (propertyId)) {
			return new LocationPropertySource (getEnd ());
		}
		return super.getPropertyValue (propertyId);
	}
	
	
	/**
	 * @param id
	 * @param value
	 * @see AcmeStudioShapeExtension.model.AnnotationElement#setPropertyValue(java.lang.Object, java.lang.Object)
	 **/
	public void setPropertyValue(Object id, Object value) {
		if (LINECOLOR_PROP.equals (id))
			setLineColorRGB ((RGB )value);
		else if (LINESTYLE_PROP.equals (id)) {
			int style = ((Integer )value).intValue () + 1;
			setLineStyle (new Integer (style));
		}
		else if (LINEWIDTH_PROP.equals (id)) {
			int width = Integer.parseInt ((String )value);
			setLineWidth (new Integer (width));
		}
		else if (START_PROP.equals(id)) {
			LocationPropertySource locPS = (LocationPropertySource )value;
			setStart (new Point (locPS.getValue ()));
		}
		else if (END_PROP.equals (id)) {
			LocationPropertySource locPS = (LocationPropertySource )value;
			setEnd (new Point (locPS.getValue ()));
		}
		super.setPropertyValue (id, value);
	}
	
	/**
	 * @return Returns the lineStyle.
	 */
	public Integer getLineStyle() {
		return lineStyle;
	}
	/**
	 * @param lineStyle The lineStyle to set.
	 */
	public void setLineStyle(Integer lineStyle) {
		checkArgument (lineStyle);
		this.lineStyle = lineStyle;
		firePropertyChange(LINESTYLE_PROP, null, lineStyle);
	}
	/**
	 * @return Returns the lineWidth.
	 */
	public Integer getLineWidth() {
		return lineWidth;
	}
	/**
	 * @param lineWidth The lineWidth to set.
	 */
	public void setLineWidth(Integer lineWidth) {
		checkArgument(lineWidth);
		this.lineWidth = lineWidth;
		firePropertyChange (LINEWIDTH_PROP, null, lineWidth);
	}
	

	/**
	 * @return Returns the lineColorRGB.
	 */
	public RGB getLineColorRGB() {
		return lineColorRGB;
	}
	/**
	 * @param lineColorRGB The lineColorRGB to set.
	 */
	public void setLineColorRGB(RGB lineColorRGB) {
		checkArgument(lineColorRGB);
		this.lineColorRGB = lineColorRGB;
		firePropertyChange (LINECOLOR_PROP, null, lineColorRGB);
	}
	/**
	 * @return Returns the points.
	 */
	public List getBendpoints() {
		return bendpoints;
	}
	/**
	 * @param points The points to set.
	 */
	public void setBendpoints(List points) {
		checkArgument (points);
		this.bendpoints = points;
		firePropertyChange (BENDPOINTS_PROP, null, points);
	}
	
	public void setBendpoint (int index, Bendpoint point) {
		checkArgument(point);
		if (index >= this.bendpoints.size () || index < 0) throw new IndexOutOfBoundsException ();
		this.bendpoints.set (index, point);
		firePropertyChange(BENDPOINTS_PROP, null, bendpoints);
	}
	
	public void removeBendpoint (int index) {
		this.bendpoints.remove (index);
		firePropertyChange(BENDPOINTS_PROP, null, bendpoints);
	}

	
	/**
	 * @param system
	 */
	public void setParentSystem(AcmeSystem system) {
		_system = system;
	}

	/**
	 * @param index
	 * @param bendpoint
	 */
	public void insertBendpoint(int index, Bendpoint bendpoint) {
		checkArgument(bendpoint);
		this.bendpoints.add(index, bendpoint);
		firePropertyChange(BENDPOINTS_PROP, null, bendpoints);
	}
}
