package AcmeStudioShapeExtension.views;


import org.eclipse.draw2d.geometry.Point;
import org.eclipse.gef.dnd.TemplateTransfer;
import org.eclipse.jface.action.*;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.*;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.ViewPart;
import AcmeStudioShapeExtension.model.EllipseAnnotation;
import AcmeStudioShapeExtension.model.RectangleAnnotation;
import AcmeStudioShapeExtension.model.RoundedRectangleAnnotation;
import AcmeStudioShapeExtension.model.command.AnnotationCreateCommand;
import AcmeStudioShapeExtension.model.command.AnnotationCreateRequest;
import AcmeStudioShapeExtension.model.command.LineAnnotationCreateCommand;
import acmeLib.elements.AcmeSystem;


/**
 * This sample class demonstrates how to plug-in a new
 * workbench view. The view shows data obtained from the
 * model. The sample creates a dummy model on the fly,
 * but a real implementation would connect to the model
 * available either in this or another plug-in (e.g. the workspace).
 * The view is connected to the model using a content provider.
 * <p>
 * The view uses a label provider to define how model
 * objects should be presented in the view. Each
 * view can present the same model objects using
 * different labels and icons, if needed. Alternatively,
 * a single label provider can be shared between views
 * in order to ensure that objects of the same type are
 * presented in the same way everywhere.
 * <p>
 */

public class AnnotationPaletteView extends ViewPart {
	
	public static String [] SHAPE_CONSTANTS = {"Rectangle", "Ellipse", "RoundedRectangle","Line", "Label", "Note"};
	
	private TableViewer viewer;
	private Action action1;
	private Action action2;

	/*
	 * The content provider class is responsible for
	 * providing objects to the view. It can wrap
	 * existing objects in adapters or simply return
	 * objects as-is. These objects may be sensitive
	 * to the current input of the view, or ignore
	 * it and always show the same content 
	 * (like Task List, for example).
	 */
	 
	class ViewContentProvider implements IStructuredContentProvider {
		public void inputChanged(Viewer v, Object oldInput, Object newInput) {
		}
		public void dispose() {
		}
		public Object[] getElements(Object parent) {
			return SHAPE_CONSTANTS;
		}
	}
	class ViewLabelProvider extends LabelProvider implements ITableLabelProvider {
		public String getColumnText(Object obj, int index) {
			return getText(obj);
		}
		public Image getColumnImage(Object obj, int index) {
			return getImage(obj);
		}
		public Image getImage(Object obj) {
			if ("Rectangle".equals (obj)) {
				return RectangleAnnotation.RECTANGLE_ICON;
			}
			else if ("Ellipse".equals (obj)) {
				return EllipseAnnotation.ELLIPSE_ICON;
			}
			else if ("RoundedRectangle".equals (obj)) {
				return RoundedRectangleAnnotation.ROUNDED_RECTANGLE_ICON;
			}
			return null;
		}
		
		/**
		 * @param element
		 * @return
		 * @see org.eclipse.jface.viewers.LabelProvider#getText(java.lang.Object)
		 **/
		public String getText(Object element) {
			if (element instanceof String) {
				return (String )element;
			}
			return "";
		}
	}

	/**
	 * The constructor.
	 */
	public AnnotationPaletteView() {
	}

	/**
	 * This is a callback that will allow us
	 * to create the viewer and initialize it.
	 */
	public void createPartControl(Composite parent) {
		viewer = new TableViewer(parent, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL);
		viewer.setContentProvider(new ViewContentProvider());
		viewer.setLabelProvider(new ViewLabelProvider());
		viewer.setInput(SHAPE_CONSTANTS);
		
		initDragAndDrop ();
		
		
		makeActions();
		hookContextMenu();
		contributeToActionBars();
	}

	/**
	 * 
	 */
	private void initDragAndDrop() {
		int ops = DND.DROP_COPY | DND.DROP_MOVE;
		Transfer [] transfers = new  Transfer [1];
		TemplateTransfer tt = TemplateTransfer.getInstance ();
		transfers [0] = tt;
		
		viewer.addDragSupport(ops, transfers, new AnnotationDragListener (this));
	}

	private void hookContextMenu() {
		MenuManager menuMgr = new MenuManager("#PopupMenu");
		menuMgr.setRemoveAllWhenShown(true);
		menuMgr.addMenuListener(new IMenuListener() {
			public void menuAboutToShow(IMenuManager manager) {
				AnnotationPaletteView.this.fillContextMenu(manager);
			}
		});
		Menu menu = menuMgr.createContextMenu(viewer.getControl());
		viewer.getControl().setMenu(menu);
		getSite().registerContextMenu(menuMgr, viewer);
	}

	private void contributeToActionBars() {
		IActionBars bars = getViewSite().getActionBars();
		fillLocalPullDown(bars.getMenuManager());
	}

	private void fillLocalPullDown(IMenuManager manager) {
		manager.add(action1);
		manager.add(new Separator());
		manager.add(action2);
	}

	private void fillContextMenu(IMenuManager manager) {
		manager.add(action1);
		manager.add(action2);
		// Other plug-ins can contribute there actions here
		manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
	}
	

	private void makeActions() {
		action1 = new Action() {
			public void run() {
				showMessage("Action 1 executed");
			}
		};
		action1.setText("Action 1");
		action1.setToolTipText("Action 1 tooltip");
		action1.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
			getImageDescriptor(ISharedImages.IMG_OBJS_INFO_TSK));
		
		action2 = new Action() {
			public void run() {
				showMessage("Action 2 executed");
			}
		};
		action2.setText("Action 2");
		action2.setToolTipText("Action 2 tooltip");
		action2.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages().
				getImageDescriptor(ISharedImages.IMG_OBJS_INFO_TSK));
	}

	private void showMessage(String message) {
		MessageDialog.openInformation(
			viewer.getControl().getShell(),
			"Annotations View",
			message);
	}

	/**
	 * Passing the focus request to the viewer's control.
	 */
	public void setFocus() {
		viewer.getControl().setFocus();
	}

	/**
	 * @return
	 */
	public ISelection  getSelection() {
		return viewer.getSelection();
	}

	/**
	 * @param name
	 * @param system
	 * @param location
	 */
	public static void createAnnotation(String annotation, AcmeSystem system, Point location) {
		AnnotationCreateRequest request = new AnnotationCreateRequest (annotation);
		request.setLocation(location);
		if (annotation.equals ("Line")) {
			LineAnnotationCreateCommand command = new LineAnnotationCreateCommand (system, request);
			command.execute ();
		}
		else {
			AnnotationCreateCommand command = new AnnotationCreateCommand (system, request);
			command.execute();
		}
	}
}