/*
 * Created on Jul 8, 2003
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package AcmeTranslator.wizards;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.operation.ModalContext;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.dialogs.IOverwriteQuery;

import acme.translators.ADMLExporter;
import acme.translators.AcmeTranslator;
import acme.translators.XAcmeExporter;
import acmeLib.armaniParser.ArmaniParser;
import acmeLib.armaniParser.ParseException;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeImportedDesignRecord;

/**
 * @author Bradley Schmerl
 * 
 * Operation to perform exporting of a file to xAcme or ADML
 */
public class XMLExportOperation implements IRunnableWithProgress {
	
	/** The action to be performed on imported families **/
	private int _familyAction;
	/** Whether to format as xAcme (false -> ADML) **/
	private boolean formatAsXAcme;
	private IPath path;
	private IProgressMonitor monitor;
	private List resourcesToExport;
	private IOverwriteQuery overwriteCallback;
	private IResource resource;
	private List errorTable = new ArrayList (1);
	
	private static final int OVERWRITE_NOT_SET = 0;
	private static final int OVERWRITE_NONE = 1;
	private static final int OVERWRITE_ALL = 2;
	private int overwriteState = OVERWRITE_NOT_SET;
	
	private boolean createLeadupStructure = true;
	private boolean createContainerDirectories = true;
	
	/**
	 * Create an instance of this class. Use this constructor if you wish to
	 * export specific resource without a common parent source
	 */
	public XMLExportOperation (List resources, String destinationPath, IOverwriteQuery overwriteImplementor) {
		super ();
		Iterator elementsEnum = resources.iterator ();
		while (elementsEnum.hasNext ()) {
			IResource currentResource = (IResource )elementsEnum.next ();
			if (isDescendent (resources, currentResource)) {
				elementsEnum.remove ();
			}
		}
		
		resourcesToExport = resources;
		path = new Path (destinationPath);
		overwriteCallback = overwriteImplementor;
	}

	/**
	 *  Create an instance of this class.  Use this constructor if you wish to
	 *  recursively export a single resource
	 */
	public XMLExportOperation(
		IResource res,
		String destinationPath,
		IOverwriteQuery overwriteImplementor) {
		super();
		resource = res;
		path = new Path(destinationPath);
		overwriteCallback = overwriteImplementor;
	}
	/**
	 *  Create an instance of this class.  Use this constructor if you wish to
	 *  export specific resources with a common parent resource (affects container
	 *  directory creation)
	 */
	public XMLExportOperation(
		IResource res,
		List resources,
		String destinationPath,
		IOverwriteQuery overwriteImplementor) {
		this(res, destinationPath, overwriteImplementor);
		resourcesToExport = resources;
	}
	
	protected void addError (String message, Throwable e) {
		errorTable.add (new Status (IStatus.ERROR, PlatformUI.PLUGIN_ID, 0, message, e));
	}
	
	/**
	 *  Answer the total number of file resources that exist at or below self in the
	 *  resources hierarchy.
	 *
	 *  @return int
	 *  @param resource org.eclipse.core.resources.IResource
	 */
	protected int countChildrenOf(IResource resource) throws CoreException {
		if (resource.getType() == IResource.FILE)
			return 1;

		int count = 0;
		if (resource.isAccessible()) {
			IResource[] children = ((IContainer) resource).members();
			for (int i = 0; i < children.length; i++)
				count += countChildrenOf(children[i]);
		}

		return count;
	}
	/**
	 *	Answer a boolean indicating the number of file resources that were
	 *	specified for export
	 *
	 *	@return int
	 */
	protected int countSelectedResources() throws CoreException {
		int result = 0;
		Iterator resources = resourcesToExport.iterator();

		while (resources.hasNext())
			result += countChildrenOf((IResource) resources.next());

		return result;
	}
	/**
	 *  Create the directories required for exporting the passed resource,
	 *  based upon its container hierarchy
	 *
	 *  @param resource org.eclipse.core.resources.IResource
	 */
	protected void createLeadupDirectoriesFor(IResource resource) {
		IPath resourcePath = resource.getFullPath().removeLastSegments(1);

		for (int i = 0; i < resourcePath.segmentCount(); i++) {
			path = path.append(resourcePath.segment(i));
			new File (path.toOSString()).mkdir ();
		}
	}
	/**
	 *	Recursively export the previously-specified resource
	 */
	protected void exportAllResources() throws InterruptedException {
		if (resource.getType() == IResource.FILE)
			exportFile((IFile) resource, path);
		else {
			try {
				exportChildren(((IContainer) resource).members(), path);
			} catch (CoreException e) {
				// not safe to show a dialog
				// should never happen because the file system export wizard ensures that the
				// single resource chosen for export is both existent and accessible
				errorTable.add(e);
			}
		}
	}
	/**
	 *	Export all of the resources contained in the passed collection
	 *
	 *	@param children java.util.Enumeration
	 *	@param currentPath IPath
	 */
	protected void exportChildren(IResource[] children, IPath currentPath)
		throws InterruptedException {
		for (int i = 0; i < children.length; i++) {
			IResource child = children[i];
			if (!child.isAccessible())
				continue;

			if (child.getType() == IResource.FILE)
				exportFile((IFile) child, currentPath);
			else {
				IPath destination = currentPath.append(child.getName());
				new File (destination.toOSString()).mkdir ();
				try {
					exportChildren(((IContainer) child).members(), destination);
				} catch (CoreException e) {
					// not safe to show a dialog
					// should never happen because:
					// i. this method is called recursively iterating over the result of #members,
					//		which only answers existing children
					// ii. there is an #isAccessible check done before #members is invoked
					errorTable.add(e.getStatus());
				}
			}
		}
	}
	/**
	 *  Export the passed file to the specified location
	 *
	 *  @param file org.eclipse.core.resources.IFile
	 *  @param location org.eclipse.core.runtime.IPath
	 */
	protected void exportFile(IFile file, IPath location)
		throws InterruptedException {
			
		int index = file.getName ().lastIndexOf("."); //$NON-NLS-1$
		if (index == -1) return;
		String filename = file.getName ().substring (0, index);
		filename += ".xml"; //$NON-NLS-1$
		IPath fullPath = location.append(filename);
		monitor.subTask(file.getFullPath().toString());
		String properPathString = fullPath.toOSString();
		File targetFile = new File(properPathString);

		if (!checkOverwrite (targetFile)) return;

		try {
			
			// Open the file
			ArmaniParser.ReInit (file.getContents());
			AcmeDesign design = ArmaniParser.parse_AcmeDesign();
			design.doPostParsingProcessing();
			
			AcmeOutputStream aos = new AcmeOutputStream ();
			
			// Create the appropriate exporter and export
			if (formatAsXAcme) {
				XAcmeExporter exporter = new XAcmeExporter (aos);
				exporter.setFamilyAction(_familyAction);
				// Handle imported families
				if (_familyAction == AcmeTranslator.EXPORT_IMPORTED_FAMILIES) {
					
					if (!exportImportedDesigns(location, design, exporter)) return;
				}
				exporter.exportDesign(design);
				exporter.writeToFile(targetFile.getAbsolutePath());	
			}
			else {
				ADMLExporter exporter = new ADMLExporter (aos);
				exporter.setFamilyAction (_familyAction);
				if (_familyAction == AcmeTranslator.EXPORT_IMPORTED_FAMILIES) {
					if (!exportImportedDesigns(location, design, exporter)) return;
				
				}
				exporter.exportDesign (design);
				exporter.writeToFile (targetFile.getAbsolutePath());
//				throw new IOException (Messages.getString("XMLExportOperation.export.ADMLNotImplemented")); //$NON-NLS-1$
			}
			
		
			
		} catch (IOException e) {
			errorTable.add(new Status(IStatus.ERROR, PlatformUI.PLUGIN_ID, 0, "Error exporting " + fullPath + ": " + e.getMessage (), e)); //$NON-NLS-1$ //$NON-NLS-2$
		} catch (CoreException e) {
			errorTable.add(new Status(IStatus.ERROR, PlatformUI.PLUGIN_ID, 0, "Error exporting " + fullPath + ": " + e.getMessage (), e)); //$NON-NLS-1$ //$NON-NLS-2$
		} catch (ParseException e) {
			errorTable.add (new Status (IStatus.ERROR, PlatformUI.PLUGIN_ID, 0, "Error exporting " + fullPath + ": " + e.getMessage (), e)); //$NON-NLS-1$ //$NON-NLS-2$
		}
		
		
		monitor.worked(1);
		ModalContext.checkCanceled(monitor);
	}

	/**
	 * Handles exporting imported designs to XML
	 * @param location The location to export to
	 * @param design The design to export
	 * @param exporter The overall exporter to register files with
	 * @return whether the design was successfully exported
	 * @throws InterruptedException
	 * @throws IOException
	 */
	private boolean exportImportedDesigns(
		IPath location,
		AcmeDesign design,
		XAcmeExporter exporter)
		throws InterruptedException, IOException {
		AcmeEnumeration importedDesigns = design.getImportedDesigns().enumerate ().castToAcmeSetVector().enumerate();
		while (importedDesigns.moreItems()) {
			AcmeImportedDesignRecord rec = (AcmeImportedDesignRecord )importedDesigns.getNextItem ();
			String importFilename = XAcmeExporter.convertFilenameToXML(rec.getCanonicalFilename());
			IPath fullExportPath = location.append (importFilename);
			monitor.subTask (rec.getCanonicalFilename());
			File exportFile = new File (fullExportPath.toOSString());
			if (!checkOverwrite (exportFile)) return false;
			AcmeOutputStream eos = new AcmeOutputStream ();
			XAcmeExporter x = new XAcmeExporter (eos);
			if (!exportImportedDesigns(location, rec.getResolvedDesign(), x)) return false;
			x.setFamilyAction(_familyAction);
			x.exportDesign (rec.getResolvedDesign());
			x.writeToFile(exportFile.getAbsolutePath());
			exporter.transferExporters (x);
			exporter.addExporterForDesign(rec.getResolvedDesign(), x);
			
		}
		return true;
	}
	
	private boolean exportImportedDesigns (IPath location, AcmeDesign design, ADMLExporter exporter) throws InterruptedException, IOException {
	AcmeEnumeration importedDesigns = design.getImportedDesigns().enumerate ().castToAcmeSetVector().enumerate();
	while (importedDesigns.moreItems()) {
		AcmeImportedDesignRecord rec = (AcmeImportedDesignRecord )importedDesigns.getNextItem ();
		String importFilename = XAcmeExporter.convertFilenameToXML(rec.getCanonicalFilename());
		IPath fullExportPath = location.append (importFilename);
		monitor.subTask (rec.getCanonicalFilename());
		File exportFile = new File (fullExportPath.toOSString());
		if (!checkOverwrite (exportFile)) return false;
		AcmeOutputStream eos = new AcmeOutputStream ();
		ADMLExporter x = new ADMLExporter (eos);
		if (!exportImportedDesigns(location, rec.getResolvedDesign(), x)) return false;
		x.setFamilyAction(_familyAction);
		x.exportDesign (rec.getResolvedDesign());
		x.writeToFile(exportFile.getAbsolutePath());
		exporter.transferExporters (x);
		exporter.addExporterForDesign(rec.getResolvedDesign(), x);
			
	}
	return true;
}
		

	
	/**
	 * Checks to see if overwriting a file is ok
	 * @param targetFile The file to overwrite
	 * @return whether overwriting is ok
	 * @throws InterruptedException
	 */
	private boolean checkOverwrite (File targetFile) throws InterruptedException {
		if (targetFile.exists()) {
			if (!targetFile.canWrite()) {
				errorTable.add(new Status(IStatus.ERROR, PlatformUI.PLUGIN_ID, 0, Messages.getString("XMLExportOperation.export.overwriteError") + targetFile.getAbsolutePath(), null)); //$NON-NLS-1$
				monitor.worked(1);
				return false;
			}

			if(overwriteState == OVERWRITE_NONE)
				return false;
				
			if (overwriteState != OVERWRITE_ALL) {
				String overwriteAnswer =
					overwriteCallback.queryOverwrite(targetFile.getAbsolutePath());

				if (overwriteAnswer.equals(IOverwriteQuery.CANCEL))
					throw new InterruptedException();

				if (overwriteAnswer.equals(IOverwriteQuery.NO)) {
					monitor.worked(1);
					return false;
				}

				if (overwriteAnswer.equals(IOverwriteQuery.NO_ALL)) {
					monitor.worked(1);
					overwriteState = OVERWRITE_NONE;
					return false;
				}

				if (overwriteAnswer.equals(IOverwriteQuery.ALL))
					overwriteState = OVERWRITE_ALL;
			}
		}
		return true;
	}
	
	/**
	 *	Export the resources contained in the previously-defined
	 *	resourcesToExport collection
	 */
	protected void exportSpecifiedResources() throws InterruptedException {
		Iterator resources = resourcesToExport.iterator();
		IPath initPath = (IPath) path.clone();

		while (resources.hasNext()) {
			IResource currentResource = (IResource) resources.next();
			if (!currentResource.isAccessible())
				continue;

			path = initPath;

			if (resource == null) {
				// No root resource specified and creation of containment directories
				// is required.  Create containers from depth 2 onwards (ie.- project's
				// child inclusive) for each resource being exported.
				if (createLeadupStructure)
					createLeadupDirectoriesFor(currentResource);

			} else {
				// Root resource specified.  Must create containment directories
				// from this point onwards for each resource being exported
				IPath containersToCreate =
					currentResource
						.getFullPath()
						.removeFirstSegments(
							resource.getFullPath().segmentCount())
						.removeLastSegments(1);

				for (int i = 0; i < containersToCreate.segmentCount(); i++) {
					path = path.append(containersToCreate.segment(i));
					new File (path.toOSString()).mkdirs();
				}
			}

			if (currentResource.getType() == IResource.FILE)
				exportFile((IFile) currentResource, path);
			else {
				if (createContainerDirectories) {
					path = path.append(currentResource.getName());
					new File (path.toOSString()).mkdirs();
				}

				try {
					exportChildren(
						((IContainer) currentResource).members(),
						path);
				} catch (CoreException e) {
					// should never happen because #isAccessible is called before #members is invoked,
					// which implicitly does an existence check
					errorTable.add(e.getStatus());
				}
			}
		}
	}
	/**
	 * Returns the status of the export operation.
	 * If there were any errors, the result is a status object containing
	 * individual status objects for each error.
	 * If there were no errors, the result is a status object with error code <code>OK</code>.
	 *
	 * @return the status
	 */
	public IStatus getStatus() {
		IStatus[] errors = new IStatus[errorTable.size()];
		errorTable.toArray(errors);
		return new MultiStatus(PlatformUI.PLUGIN_ID, IStatus.OK, errors, Messages.getString("XMLExportOperation.export.problem"), null); //$NON-NLS-1$
	}
	/**
	 *  Answer a boolean indicating whether the passed child is a descendent
	 *  of one or more members of the passed resources collection
	 *
	 *  @return boolean
	 *  @param resources java.util.List
	 *  @param child org.eclipse.core.resources.IResource
	 */
	protected boolean isDescendent(List resources, IResource child) {
		if (child.getType() == IResource.PROJECT)
			return false;

		IResource parent = child.getParent();
		if (resources.contains(parent))
			return true;

		return isDescendent(resources, parent);
	}
	/**
	 *	Export the resources that were previously specified for export
	 *	(or if a single resource was specified then export it recursively)
	 */
	public void run(IProgressMonitor monitor) throws InterruptedException {
		this.monitor = monitor;

		if (resource != null) {
			if (createLeadupStructure)
				createLeadupDirectoriesFor(resource);

			if (createContainerDirectories
				&& resource.getType() != IResource.FILE) {
				// ensure it's a container
				path = path.append(resource.getName());
				new File (path.toOSString()).mkdirs();
			}
		}

		try {
			int totalWork = IProgressMonitor.UNKNOWN;
			try {
				if (resourcesToExport == null)
					totalWork = countChildrenOf(resource);
				else
					totalWork = countSelectedResources();
			} catch (CoreException e) {
				// Should not happen
				errorTable.add(e.getStatus());
			}
			monitor.beginTask("Exporting", totalWork); //$NON-NLS-1$
			if (resourcesToExport == null) {
				exportAllResources();
			} else {
				exportSpecifiedResources();
			}
		} finally {
			monitor.done();
		}
	}
	/**
	 *	Set this boolean indicating whether a directory should be created for
	 *	Folder resources that are explicitly passed for export
	 *
	 *	@param value boolean
	 */
	public void setCreateContainerDirectories(boolean value) {
		createContainerDirectories = value;
	}
	/**
	 *	Set this boolean indicating whether each exported resource's complete path should
	 *	include containment hierarchies as dictated by its parents
	 *
	 *	@param value boolean
	 */
	public void setCreateLeadupStructure(boolean value) {
		createLeadupStructure = value;
	}
	/**
	 *	Set this boolean indicating whether exported resources should automatically
	 *	overwrite existing files when a conflict occurs. If not
	 *	query the user.
	 *
	 *	@param value boolean
	 */
	public void setOverwriteFiles(boolean value) {
		if(value)	
			overwriteState = OVERWRITE_ALL;
	}

	/**
	 * Sets whether to export to xAcme
	 * @param b
	 */
	public void setExportXAcme(boolean b) {
		formatAsXAcme = b;
		
	}

	/**
	 * Sets what to do to imported families in a design
	 * @param b
	 */
	public void setImportFamilyAction(int familyAction) {
		_familyAction = familyAction;
		
	}


}
