package acme.translators;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Hashtable;
import java.util.TimeZone;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.w3c.dom.*;
import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;
/**
 * Class: ADMLExporter <br>Purpose: Takes an AcmeLib structure and converts it
 * to the <br>ADML XML format <br>Revision History: <br>Date Person Change:
 * <br>6/12/01 BRS Created initial version <br>Known bugs/omissions: <br>-
 * Should create a DOM structure and then serialize it
 * 
 * @author Bradley Schmerl
 */
public class ADMLExporter extends XMLExporter{

	AcmeOutputStream out = null;
	Hashtable elementToId = null;
	private int uniqueId = 1;
	Document doc;

	public ADMLExporter(AcmeOutputStream s) {
		out = s;
		elementToId = new Hashtable();
	}

	public void writeToFile(String path) {
		try {
			File f = new File(path);
			//	    FileOutputStream fo = new FileOutputStream (f);
			FileWriter fw = new FileWriter(f);

			TransformerFactory tFactory = TransformerFactory.newInstance ();
			Transformer transformer = tFactory.newTransformer ();
			DOMSource source = new DOMSource (doc.getDocumentElement());
			StreamResult result= new StreamResult (fw);
			transformer.transform (source, result);

			//	    out.flush ();
			//	    out.writeTo (fo);
			fw.close();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	public String writeToString() {
		try {
			StringWriter sw = new StringWriter();

			TransformerFactory tFactory = TransformerFactory.newInstance();
			Transformer transformer = tFactory.newTransformer();
			DOMSource source = new DOMSource (doc.getDocumentElement());
		    StreamResult result = new StreamResult (sw);
		    transformer.transform (source, result);
			sw.close();
			return sw.toString();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	/*
	 * Name: removeXMLSpecials @param st - the string to remove the XML special
	 * characters from @return a copy of st with the XML special characters
	 * replaced with XML entities Purpose: Filters a string so that special
	 * characters are replaced by their corresponding XML entities & --> &amp;
	 * < --> &lt; > --> &gt;
	 */
	private String removeXMLSpecials(String st) {
		StringBuffer editable = new StringBuffer();

		int i = 0;
		while (i < st.length()) {
			char c = st.charAt(i);
			if (c == '<') {
				editable.append("&lt;");
			}
			else if (c == '>') {
				editable.append("&gt;");
			}
			else if (c == '&') {
				editable.append("&amp;");
			}
			else {
				editable.append(c);
			}
			i++;
		}
		return editable.toString();
	}

	private String generateID(AcmeObject element) {
		String result = "" + uniqueId++;
		if (element != null) {
			elementToId.put(element, result);
		}
		return result;
	}

	/*
	 * Name: generateOpeningTag @param tagName -- the name of the tag to open
	 * @param object -- the Acme object to be associated with this tag @param
	 * name -- the value of the the name attribute for the tag (null => no name
	 * attribute output) @closeTag -- true => the tag is to be finished with
	 * ">", false => there are more attributes to come Purpose: Outputs an
	 * opening tag that looks like <tagName identifier='uniqueId' name='name'>
	 */
	private void generateOpeningTag(String tagName, AcmeObject object, String name, boolean closeTag)
		throws java.io.IOException {
		String id = generateID(object);
		if (object instanceof AcmeType) {
			id = generateTypeID((AcmeType )object);
		}
		out.write("<" + tagName + " identifier='" + id);
		if (name != null) {
			out.write("' name='" + name);
		}
		out.write("'");
		if (closeTag) {
			out.writeln(">");
		}
		else {
			out.writeln(" "); // leave space for additional tags
		}
	}

	private void processOpeningTag(Element elementXML, AcmeObject objectACME, String name) throws java.io.IOException {
		String identifier = generateID(objectACME);
		if (objectACME instanceof AcmeType) {
			identifier = generateTypeID((AcmeType )objectACME);
		}
		elementXML.setAttribute("identifier", identifier);
		if (name != null) {
			elementXML.setAttribute("name", name);
		}
	}

	private Element createTag(String tag, AcmeObject objectACME, String name) throws java.io.IOException {
		Element objectXML = doc.createElement(tag);
		processOpeningTag(objectXML, objectACME, name);
		return objectXML;
	}

	/*
	 * Name: export @param designACME - An Acme Design element, which contains
	 * the types, etc, to be exported Purpose: Produces the XML for the root
	 * Acme design object. This particular procedure supports the following DTD
	 * <!ELEMENT Design (ImportPathName | ComponentType | ConnectorType |
	 * PortType | RoleType | PropertyTypeDeclaration| > <!ATTLIST Design
	 * identifier ID #REQUIRED name CDATA #IMPLIED>
	 */
	public void exportDesign(AcmeDesign designACME) throws java.io.IOException {
		try {
			doc =  DocumentBuilderFactory.newInstance ().newDocumentBuilder ().newDocument ();
			Comment comm1 = doc.createComment("Automatically generated by ADMLExporter v1.0 using ADML version 1.0");
			doc.appendChild(comm1);

			GregorianCalendar cal = new GregorianCalendar(TimeZone.getDefault());
			Comment comm2 =
				doc.createComment(
					"Generated "
						+ cal.get(Calendar.MONTH)
						+ "/"
						+ cal.get(Calendar.DAY_OF_MONTH)
						+ "/"
						+ cal.get(Calendar.YEAR)
						+ " -->");
			doc.appendChild(comm2);
			Element designXML = createTag("Design", designACME, null);
			designXML.appendChild(comm1);
			designXML.appendChild(comm2);
			doc.appendChild(designXML);
			exportDesignInternals(designACME, designXML);
		}
		catch (DOMException e) {
			e.printStackTrace();
		}
		catch (ParserConfigurationException e) {
			e.printStackTrace();
		}
		catch (FactoryConfigurationError e) {
			e.printStackTrace();
		}

	}
	protected void exportDesignInternals(AcmeDesign designACME, Element designXML) throws IOException {
		exportImports (designACME, designXML);
		
		exportTypespace(designACME.getTypes(), designXML);

		AcmeEnumeration familiesACME = designACME.getFamilies();
		while (familiesACME.moreItems()) {
			Element familyXML = export((AcmeFamily) familiesACME.getNextItem());
			designXML.appendChild(familyXML);
		}

		AcmeEnumeration systemsACME = designACME.getSystems();
		while (systemsACME.moreItems()) {
			Element systemXML = export((AcmeSystem) systemsACME.getNextItem());
			designXML.appendChild(systemXML);
		}
	}

	/**
	 * @param enumeration
	 * @param designXML
	 */
	private void exportImports(AcmeDesign designACME, Element designXML) throws IOException {
		AcmeEnumeration importedDesigns = designACME.getImportedDesigns().enumerate ();
		while (importedDesigns.moreItems ()) {
			AcmeImportedDesignRecord rec = (AcmeImportedDesignRecord )importedDesigns.getNextItem ();
			if (familyAction == AcmeTranslator.REFERENCE_IMPORTED_FAMILIES) {
				Element importXML = createTag ("ImportPathname", rec, null);
				importXML.setAttribute("pathname",rec.getDesignFilename());
				designXML.appendChild(importXML);
			}
			else if (familyAction == AcmeTranslator.INLINE_IMPORTED_FAMILIES) {
				exportDesignInternals (rec.getResolvedDesign(), designXML);
			}
			else if (familyAction == AcmeTranslator.EXPORT_IMPORTED_FAMILIES) {
				Element importXML = createTag ("ImportPathname", rec, null);
				importXML.setAttribute ("pathname", XAcmeExporter.convertFilenameToXML (rec.getCanonicalFilename()));
				designXML.appendChild (importXML);
			}
		}
	}

	/***************************************************************************
	 * Name: export
	 * 
	 * @param family --
	 *           An Acme Family Purpose: Exports the family to ADML according
	 *           to the following DTD:
	 *           <!ELEMENT Family (ComponentType | ConnectorType | PortType | 
	 *           RoleType | PropertyTypeDeclaration | Component | Connector |
	 *           Port | Role | Property | Attachments | Representation |
	 *           MetaProperty )+>
	 *           <!ATTLIST Family identifier ID #REQUIRED name CDATA #IMPLIED>
	 **************************************************************************/
	private Element export(AcmeFamily familyACME) throws java.io.IOException {
		Element familyXML = createTag("Family", familyACME, familyACME.getName());
		AcmeEnumeration typesAcme = familyACME.getTypes();
		exportTypespace(typesAcme, familyXML);
		exportProperties(familyACME, familyXML);
		exportConstituents(familyACME, familyXML);
		return familyXML;
	}

	/**
	 * Name: exportConstituents
	 * 
	 * @param systemACME -
	 *           the system to export the constituents of Purpose: exports the
	 *           component, connectors, attachment and representations of the
	 *           system Used by both export (AcmeSystem) and export
	 *           (AcmeFamily).
	 */
	private void exportConstituents(AcmeSystem systemACME, Element toXML) throws java.io.IOException {
		// Get and export the components
		AcmeEnumeration componentsACME = systemACME.getComponents();
		while (componentsACME.moreItems()) {
			Element componentXML = export((AcmeComponent) componentsACME.getNextItem());
			toXML.appendChild(componentXML);
		}

		// Get and export the connectors
		AcmeEnumeration connectorsACME = systemACME.getConnectors();
		while (connectorsACME.moreItems()) {
			Element connectorXML = export((AcmeConnector) connectorsACME.getNextItem());
			toXML.appendChild(connectorXML);
		}

		Element attachmentsXML = doc.createElement("Attachments");
		attachmentsXML.setAttribute("identifier", generateID(null));
		toXML.appendChild(attachmentsXML);

		// Get and export the attachments
		AcmeEnumeration attachmentsACME = systemACME.getAttachments();
		while (attachmentsACME.moreItems()) {
			Element attachmentXML = export((AcmeAttachment) attachmentsACME.getNextItem());
			attachmentsXML.appendChild(attachmentXML);
		}

		AcmeEnumeration representationsACME = systemACME.getRepresentations();
		while (representationsACME.moreItems()) {
			Element representationXML = export((AcmeRepresentation) representationsACME.getNextItem());
			toXML.appendChild(representationXML);
		}
	}

	/***************************************************************************
	 * Name: export
	 * 
	 * @param types -
	 *           the types space that is to be exported Purpose: exports
	 *           component, connector, port, role and property types
	 **************************************************************************/
	private void exportTypespace(AcmeEnumeration typesACME, Element toXML) throws java.io.IOException {
		typesACME.sortByCreationOrder();
		typesACME.reset();
		while (typesACME.moreItems()) {
			AcmeType typeACME = (AcmeType) typesACME.getNextItem();
			if (typeACME instanceof AcmeComponentType) {
				Element componentTypeXML = export((AcmeComponentType) typeACME);
				toXML.appendChild(componentTypeXML);
			}
		}

		typesACME.reset();
		while (typesACME.moreItems()) {
			AcmeType typeACME = (AcmeType) typesACME.getNextItem();
			if (typeACME instanceof AcmeConnectorType) {
				Element connectorTypeXML = export((AcmeConnectorType) typeACME);
				toXML.appendChild(connectorTypeXML);
			}
		}

		typesACME.reset();
		while (typesACME.moreItems()) {
			AcmeType typeACME = (AcmeType) typesACME.getNextItem();
			if (typeACME instanceof AcmePortType) {
				Element portTypeXML = export((AcmePortType) typeACME);
				toXML.appendChild(portTypeXML);
			}
		}

		typesACME.reset();
		while (typesACME.moreItems()) {
			AcmeType typeACME = (AcmeType) typesACME.getNextItem();
			if (typeACME instanceof AcmeRoleType) {
				Element roleTypeXML = export((AcmeRoleType) typeACME);
				toXML.appendChild(roleTypeXML);
			}
		}

		typesACME.reset();
		while (typesACME.moreItems()) {
			AcmeType typeACME = (AcmeType) typesACME.getNextItem();
			if (typeACME instanceof AcmePropertyType) {
				Element propertyTypeXML = export((AcmePropertyType) typeACME, true);
				toXML.appendChild(propertyTypeXML);
			}
		}
	}

	private boolean hasSuperTypes(AcmeElementType elementACME) {
		return elementACME.getSuperTypes().enumerate(new acmeLib.core.filters.NoBaseClassesEnumFilter()).getSize() != 0;
	}

	private void exportSuperTypes(AcmeElementType elementACME, String type, Element toXML) throws java.io.IOException {
		AcmeEnumeration typesACME =
			elementACME.getSuperTypes().enumerate(new acmeLib.core.filters.NoBaseClassesEnumFilter());
		if (typesACME.getSize() != 0) {
			while (typesACME.moreItems()) {
				Element elIDXML = doc.createElement("ElementID");
				elIDXML.setAttribute("identifier", generateID(null));
				elIDXML.setAttribute("type", type);
				AcmeElementType t = (AcmeElementType) typesACME.getNextItem();
				Text val = doc.createTextNode (getTypeReference ((AcmeElement )elementACME, (AcmeElement )t));
//				Text val = doc.createTextNode(elementToId.get(t).toString());
				elIDXML.appendChild(val);
				toXML.appendChild(elIDXML);
			}
		}
	}

	private boolean hasDeclaredTypes(AcmeElement elementACME) {
		return elementACME.getDeclaredTypes().getSize() != 0;
	}

	private void exportDeclaredTypes(AcmeElement elementACME, String type, Element toXML) throws java.io.IOException {
		if (elementACME.getDeclaredTypes().getSize() != 0) {
			AcmeEnumeration typesACME = elementACME.getDeclaredTypes();
			while (typesACME.moreItems()) {
				Element elIDXML = doc.createElement("ElementID");
				elIDXML.setAttribute("identifier", generateID(null));
				elIDXML.setAttribute("type", type);
				AcmeElementType typeACME = (AcmeElementType) typesACME.getNextItem();
				String id = getTypeReference (elementACME, (AcmeElement )typeACME);
//				String id = (String) elementToId.get(typeACME);
				Text val = doc.createTextNode(id);
				toXML.appendChild(elIDXML);
				elIDXML.appendChild(val);
			}
		}
	}

	/***************************************************************************
	 * Name: export
	 * 
	 * @param systemACME -
	 *           the acme system to be exported Purpose: Exports the system,
	 *           according to the following DTD )>
	 *           <!ATTLIST System identifier ID #REQUIRED name CDATA #IMPLIED>
	 *           <!ElEMENT SystemDescription (ExtendedSystemType | SystemStructure)?>,
	 *           SystemStructure?)>
	 *           <!ELEMENT SystemStructure (Component | Connector | Property |  >
	 **************************************************************************/
	private Element export(AcmeSystem systemACME) throws java.io.IOException {
		Element systemXML = createTag("System", systemACME, systemACME.getName());
		Element systemDescXML = doc.createElement("SystemDescription");
		systemXML.appendChild(systemDescXML);
		Element structurePosXML = systemXML;

		if (hasDeclaredTypes(systemACME)) {
			Element extendedSystemTypeXML = doc.createElement("ExtendedSystemType");
			systemDescXML.appendChild(extendedSystemTypeXML);
			structurePosXML = extendedSystemTypeXML;
			exportDeclaredTypes(systemACME, "family", extendedSystemTypeXML);
		}

		if (systemACME.getComponents().getSize() != 0
			|| systemACME.getConnectors().getSize() != 0
			|| systemACME.getRoles().getSize() != 0
			|| systemACME.getPorts().getSize() != 0
			|| systemACME.getProperties().getSize() != 0) {
			// The system has some structure
			Element sysStructXML = doc.createElement("SystemStructure");
			structurePosXML.appendChild(sysStructXML);
			exportProperties(systemACME, sysStructXML);
			exportConstituents(systemACME, sysStructXML);
		}

		return systemXML;
	}

	/**
	 * Name: export
	 * 
	 * @param componentACME -
	 *           the component to export Purpose: exports ADML of component
	 *           according to the folllowing DTD
	 *           <!ELEMENT Component (ComponentDescription, MetaProperty)>
	 *           <!ATTLIST Component identifier ID #REQUIRED name CDATA #IMPLIED>
	 */
	private Element export(AcmeComponent componentACME) throws java.io.IOException {
		Element componentXML = createTag("Component", componentACME, componentACME.getName());

		Element compDescXML = exportComponentDescription(componentACME);
		componentXML.appendChild(compDescXML);
		return componentXML;
	}

	/*
	 * Name: export @param connectorACME - the connector to export Purpose:
	 * Outputs the connector according to the following DTD )> <!ATTLIST
	 * Connector identifier ID #REQUIRED name CDATA #IMPLIED>
	 */
	private Element export(AcmeConnector connectorACME) throws java.io.IOException {
		Element connectorXML = createTag("Connector", connectorACME, connectorACME.getName());
		Element connectorDesc = exportConnectorDescription(connectorACME);
		connectorXML.appendChild(connectorDesc);
		return connectorXML;
	}

	/*
	 * name: export @param roleACME - the role to export Purpose: Exports a
	 * role according to the following DTD: )> <!ATTLIST Role identifier ID
	 * #REQUIRED name CDATA #IMPLIED>
	 */
	private Element export(AcmeRole roleACME) throws java.io.IOException {
		Element roleXML = createTag("Role", roleACME, roleACME.getName());
		Element roleDescXML = exportRoleDescription(roleACME);
		roleXML.appendChild(roleDescXML);
		return roleXML;
	}

	/* )> <!ATTLIST Port identifier ID #REQUIRED name CDATA #IMPLIED> */
	private Element export(AcmePort portACME) throws java.io.IOException {
		Element portXML = createTag("Port", portACME, portACME.getName());
		Element portDescXML = exportPortDescription(portACME);
		portXML.appendChild(portDescXML);
		return portXML;
	}

	/*
	 * Name: exportProperties @param element -- an acme element that contains
	 * properties (e.g., System, Connector, Port) Purpose: Export the property
	 * component of an element according to the following DTD )> <!ATTLIST
	 * Property identifier ID #REQUIRED name CDATA #implied>
	 */
	private void exportProperties(AcmeElement elementACME, Element elementXML) throws java.io.IOException {
		AcmeEnumeration propertiesACME = elementACME.getProperties();
		while (propertiesACME.moreItems()) {
			AcmeProperty propertyACME = (AcmeProperty) propertiesACME.getNextItem();
			Element propertyXML = createTag("Property", propertyACME, propertyACME.getName());
			Element propertyTypeXML = export(propertyACME.getType(), false);
			// false means don't export as a type declaration
			propertyXML.appendChild(propertyTypeXML);
			// export the property's value (if it has one)
			if (propertyACME.getValue() != null) {
				Element propertyValueXML = export(propertyACME.getValue());
				propertyXML.appendChild(propertyValueXML);
			}
			elementXML.appendChild(propertyXML);
		}
	}

	/**
	 * Name: export
	 * 
	 * @param propertyValueACME --
	 *           the property value to export Purpose: Exports the ADML for a
	 *           property value according to the following DTD:
	 *           <!ELEMENT PropertyValue (PropertyLiteralValue | PropertySetValue | 
	 *           PropertySequenceValue | ProperyRecordValue)>
	 *           <!ELEMENT PropertyLiteralValue EMPTY>
	 *           <!ATTLIST PropertyLiteralValue identifier ID #REQUIRED  name
	 *           CDATA #IMPLIED value CDATA #IMPLIED> )>
	 *           <!ATTLIST PropertySetValue identifier ID #REQUIRED name CDATA #IMPLIED>)>
	 *           <!ATTLIST PropertySequenceValue identifier ID #REQUIRED  name
	 *           CDATA #IMPLIED> )>
	 *           <!ATTLIST PropertySequenceElement identifier ID #REQUIRED 
	 *           name CDATA #IMPLIED position CDATA #REQUIRED> )>
	 *           <!ATTLIST PropertyRecordValue identifier ID #REQUIRED  name
	 *           CDATA #IMPLIED>
	 *           <!ELEMENT PropertyFieldValue (PropertyType? , PropertyValue)>
	 *           <!ATTLIST PropertyFieldValue identifier ID #REQUIRED  name
	 *           CDATA #IMPLIED fieldName CDATA #REQUIRED>
	 */
	private Element export(AcmePropertyValue propertyValueACME) throws java.io.IOException {
		Element propertyValueXML = doc.createElement("PropertyValue");

		if (propertyValueACME instanceof AcmeInt
			|| propertyValueACME instanceof AcmeFloat
			|| propertyValueACME instanceof AcmeBoolean
			|| propertyValueACME instanceof AcmeEnum) {
			Element literalXML = createTag("PropertyLiteralValue", null, null);
			literalXML.setAttribute("value", propertyValueACME.toString());
			propertyValueXML.appendChild(literalXML);
		}
		else if (propertyValueACME instanceof AcmeString) {
			Element literalXML = createTag("PropertyLiteralValue", null, null);
			AcmeString strValACME = (AcmeString) propertyValueACME;
			literalXML.setAttribute("value", strValACME.getValue());
			propertyValueXML.appendChild(literalXML);
		}
		else if (propertyValueACME instanceof AcmeSet) {
			// Generates <PropertySetValue ...>...
			AcmeSet setValACME = (AcmeSet) propertyValueACME;
			Element setValXML = createTag("PropertySetValue", null, null);
			AcmeEnumeration valuesACME = setValACME.getValues();
			while (valuesACME.moreItems()) {
				Element valueXML = export((AcmePropertyValue) valuesACME.getNextItem());
				setValXML.appendChild(valueXML);
			}
			propertyValueXML.appendChild(setValXML);
		}
		else if (propertyValueACME instanceof AcmeSequence) {
			AcmeSequence seqValACME = (AcmeSequence) propertyValueACME;
			Element seqValXML = createTag("PropertySequenceValue", null, null);
			AcmeEnumeration valuesACME = seqValACME.getValues();
			while (valuesACME.moreItems()) {
				Element valueXML = export((AcmePropertyValue) valuesACME.getNextItem());
				seqValXML.appendChild(valueXML);
			}
			propertyValueXML.appendChild(seqValXML);
		}
		else if (propertyValueACME instanceof AcmeRecord) {
			AcmeRecord recValACME = (AcmeRecord) propertyValueACME;
			Element recValXML = createTag("PropertyRecordValue", null, null);
			AcmeEnumeration fieldsACME = recValACME.getFields();
			while (fieldsACME.moreItems()) {
				AcmeRecordField fieldACME = (AcmeRecordField) fieldsACME.getNextItem();
				Element fieldXML = createTag("PropertyFieldValue", null, fieldACME.getName());
				Element fieldTypeXML = export(fieldACME.getType(), false);
				fieldXML.appendChild(fieldTypeXML);
				Element fieldValueXML = export(fieldACME.getValue());
				fieldXML.appendChild(fieldValueXML);
				recValXML.appendChild(fieldXML);
			}
			propertyValueXML.appendChild(recValXML);
		}

		return propertyValueXML;
	}

	/*
	 * Name: export @param propertyType - the propertyType to export @param
	 * isDecl - true if this is a declaration of a property type, false if it
	 * is a reference )> <!ATTLIST PropertyTypeDeclaration identifier ID
	 * #REQUIRED name CDATA #IMPLIED> <!ELEMENT PropertyType (PropertyEnumType
	 * | PropertyLiteralType | PropertySetType|PropertySequenceType |
	 * PropertyRecordType | ElementID)> <!ATTLIST PropertyType identifier ID
	 * #REQUIRED name CDATA #IMPLIED> )> <!ATTLIST PropertyEnumType identifier
	 * ID #REQUIRED name CDATA #IMPLIED> <!ELEMENT PropertyLiteralType EMPTY>
	 * <!ATTLIST PropertyLiteralType identifier ID #REQUIRED name CDATA
	 * #IMPLIED type (int | float | string | boolean | any | identifier | uri |
	 * char) #REQUIRED> <!ELEMENT PropertySetType (PropertyType?)> <!ATTLIST
	 * PropertySetType identifier ID #REQUIRED name CDATA #REQUIRED> <!ELEMENT
	 * PropertySequenceType (PropertyType?)> <!ATTLIST PropertySequenceType
	 * identifier ID #REQUIRED name CDATA #REQUIRED> )> <!ATTLIST
	 * PropertyRecordType identifier ID #REQUIRED name CDATA #REQUIRED>
	 * <!ELEMENT PropertyFieldDeclaration (PropertyType?)> <!ELEMENT
	 * ProeprtyFieldDeclaration identifier ID #REQUIRED name CDATA #IMPLIED>
	 */
	private Element export(AcmePropertyType propertyTypeACME, boolean isDecl) throws java.io.IOException {
		Element propertyTypeXML = createTag("PropertyType", propertyTypeACME, null);
		Element propertyRootXML = propertyTypeXML;
		if (isDecl) {
			// Declaring a type instead of referring to it
			Element declXML = createTag("PropertyTypeDeclaration", propertyTypeACME, propertyTypeACME.getTypeName());
			declXML.appendChild(propertyRootXML);
			propertyRootXML = declXML;
		}

		if (!isDecl && propertyTypeACME.getIsAlias()) {
			// This is a reference to a type so just output the
			// ElementID reference
			Element refXML = doc.createElement("ElementID");
			refXML.setAttribute("identifier", generateID(null));
			refXML.setAttribute("type", "propertytype");
			refXML.setAttribute("name", propertyTypeACME.getTypeName());
			Text val = doc.createTextNode(elementToId.get(propertyTypeACME).toString());
			//	    Text val = doc.createTextNode (propertyTypeACME.getTypeName ());
			refXML.appendChild(val);
			propertyTypeXML.appendChild(refXML);
		}
		if (propertyTypeACME instanceof EnumType) {
			EnumType eTypeACME = (EnumType) propertyTypeACME;
			Element eTypeXML = createTag("PropertyEnumType", null, null);

			AcmeEnumeration valuesACME = eTypeACME.getValues();
			while (valuesACME.moreItems()) {
				AcmeEnum valACME = (AcmeEnum) valuesACME.getNextItem();
				Element valXML = doc.createElement("ElementID");
				valXML.setAttribute("identifier", generateID(null));
				valXML.setAttribute("name", valACME.toString());
				Text cd = doc.createTextNode(valACME.toString());
				valXML.appendChild(cd);
				eTypeXML.appendChild(valXML);
			}
			propertyTypeXML.appendChild(eTypeXML);
		}
		else if (propertyTypeACME instanceof FloatType) {
			propertyTypeXML.appendChild(generatePropertyLiteralType("float"));
		}
		else if (propertyTypeACME instanceof BooleanType) {
			propertyTypeXML.appendChild(generatePropertyLiteralType("boolean"));
		}
		else if (propertyTypeACME instanceof IntType) {
			propertyTypeXML.appendChild(generatePropertyLiteralType("int"));
		}
		else if (propertyTypeACME instanceof StringType) {
			propertyTypeXML.appendChild(generatePropertyLiteralType("string"));
		}
		else if (propertyTypeACME instanceof SetType) {
			SetType sTypeACME = (SetType) propertyTypeACME;
			Element sTypeXML = createTag("PropertySetType", null, null);
			Element innerXML = export((AcmePropertyType) sTypeACME.getSetType(), false);
			sTypeXML.appendChild(innerXML);
			propertyTypeXML.appendChild(sTypeXML);
		}
		else if (propertyTypeACME instanceof SequenceType) {
			SequenceType sTypeACME = (SequenceType) propertyTypeACME;
			Element sTypeXML = createTag("PropertySequenceType", null, null);
			Element innerXML = export((AcmePropertyType) sTypeACME.getSequenceType(), false);
			sTypeXML.appendChild(innerXML);
			propertyTypeXML.appendChild(sTypeXML);
		}
		else if (propertyTypeACME instanceof RecordType) {
			RecordType rTypeACME = (RecordType) propertyTypeACME;
			Element rTypeXML = createTag("PropertyRecordType", null, null);
			AcmeEnumeration fieldsACME = rTypeACME.getFields();
			while (fieldsACME.moreItems()) {
				AcmeRecordField fieldACME = (AcmeRecordField) fieldsACME.getNextItem();
				Element fieldXML = createTag("RecordFieldDeclaration", null, fieldACME.getName());

				Element fieldTypeXML = export(fieldACME.getType(), false);
				fieldXML.appendChild(fieldTypeXML);
				rTypeXML.appendChild(fieldXML);
			}
			propertyTypeXML.appendChild(rTypeXML);
		}
		return propertyRootXML;
	}

	private Element generatePropertyLiteralType(String type) {
		Element pltXML = doc.createElement("PropertyLiteralType");
		pltXML.setAttribute("identifier", generateID(null));
		pltXML.setAttribute("type", type);
		return pltXML;
	}

	/*
	 * Name: export @param componentTypeACME - the component type to export
	 * Purpose: exports the component type according to the following DTD: )>
	 * <!ATTLIST ComponentType identifier ID #REQUIRED name CDATA #IMPLIED>
	 */
	private Element export(AcmeComponentType componentTypeACME) throws java.io.IOException {
		Element componentTypeXML = createTag("ComponentType", componentTypeACME, componentTypeACME.getTypeName());
		Element componentDescXML = exportComponentDescription(componentTypeACME);
		componentTypeXML.appendChild(componentDescXML);
		return componentTypeXML;
	}
	/*
	 * Name: exportComponentDescription @param componentACME -- the component
	 * whose description will be exported ) and export(ComponentType*) to
	 * export a ComponentDescription, according to the following DTD: <!ELEMENT
	 * ComponentDescription (ExtendedComponentType | ComponentBody)?> ,
	 * ComponentBody?)> >
	 */
	private Element exportComponentDescription(AcmeComponent componentACME) throws java.io.IOException {
		Element componentDescXML = doc.createElement("ComponentDescription");
		Element structurePos = componentDescXML;

		if (componentACME instanceof AcmeElementType) {
			if (hasSuperTypes((AcmeElementType) componentACME)) {
				Element extendedXML = doc.createElement("ExtendedComponentType");
				componentDescXML.appendChild(extendedXML);
				structurePos = extendedXML;
				exportSuperTypes((AcmeElementType) componentACME, "componenttype", extendedXML);
			}
		}
		else if (hasDeclaredTypes(componentACME)) {
			Element extendedXML = doc.createElement("ExtendedComponentType");
			componentDescXML.appendChild(extendedXML);
			structurePos = extendedXML;
			exportDeclaredTypes(componentACME, "componenttype", extendedXML);
		}

		AcmeEnumeration portsACME = componentACME.getPorts();
		AcmeEnumeration representationsACME = componentACME.getRepresentations();

		if (portsACME.getSize() > 0
			|| componentACME.getProperties().getSize() > 0
			|| representationsACME.getSize() > 0) {
			Element compBody = doc.createElement("ComponentBody");
			structurePos.appendChild(compBody);
			exportProperties(componentACME, compBody);

			while (portsACME.moreItems()) {
				Element portXML = export((AcmePort) portsACME.getNextItem());
				compBody.appendChild(portXML);
			}

			while (representationsACME.moreItems()) {
				Element representationXML = export((AcmeRepresentation) representationsACME.getNextItem());
				compBody.appendChild(representationXML);
			}
		}

		return componentDescXML;
	}

	/*
	 * Name: export @param connectorTypeACME - the connector type to export
	 * Purpose: exports the connector type according to the following DTD: )>
	 * <!ATTLIST ConnectorType identifier ID #REQUIRED name CDATA #IMPLIED>
	 */
	private Element export(AcmeConnectorType connectorTypeACME) throws java.io.IOException {
		Element connectorTypeXML = createTag("ConnectorType", connectorTypeACME, connectorTypeACME.getTypeName());
		Element connectorDescXML = exportConnectorDescription(connectorTypeACME);
		connectorTypeXML.appendChild(connectorDescXML);
		return connectorTypeXML;
	}
	/*
	 * Name: exportConnectorDescription @param connectorACME -- the connector
	 * whose description will be exported ) and export(ConnectorType*) to
	 * export a ConnectorDescription, according to the following DTD: <!ELEMENT
	 * ConnectorDescription (ExtendedConnectorType | ConnectorBody)?> ,
	 * ConnectorBody?)> >
	 */
	private Element exportConnectorDescription(AcmeConnector connectorACME) throws java.io.IOException {
		Element connectorDescXML = doc.createElement("ConnectorDescription");
		Element structurePos = connectorDescXML;

		if (connectorACME instanceof AcmeElementType) {
			if (hasSuperTypes((AcmeElementType) connectorACME)) {
				Element extendedXML = doc.createElement("ExtendedConnectorType");
				connectorDescXML.appendChild(extendedXML);
				structurePos = extendedXML;
				exportSuperTypes((AcmeElementType) connectorACME, "connectortype", extendedXML);
			}
		}
		else if (hasDeclaredTypes(connectorACME)) {
			Element extendedXML = doc.createElement("ExtendedConnectorType");
			connectorDescXML.appendChild(extendedXML);
			structurePos = extendedXML;
			exportDeclaredTypes(connectorACME, "connectortype", extendedXML);
		}

		AcmeEnumeration rolesACME = connectorACME.getRoles();
		AcmeEnumeration representationsACME = connectorACME.getRepresentations();

		if (rolesACME.getSize() > 0
			|| connectorACME.getProperties().getSize() > 0
			|| representationsACME.getSize() > 0) {
			Element compBody = doc.createElement("ConnectorBody");
			structurePos.appendChild(compBody);
			exportProperties(connectorACME, compBody);

			while (rolesACME.moreItems()) {
				Element roleXML = export((AcmeRole) rolesACME.getNextItem());
				compBody.appendChild(roleXML);
			}

			while (representationsACME.moreItems()) {
				Element representationXML = export((AcmeRepresentation) representationsACME.getNextItem());
				compBody.appendChild(representationXML);
			}
		}

		return connectorDescXML;
	}

	/*
	 * Name: export @param portTypeACME - the port type to export Purpose:
	 * exports the port type according to the following DTD: )> <!ATTLIST
	 * PortType identifier ID #REQUIRED name CDATA #IMPLIED>
	 */
	private Element export(AcmePortType portTypeACME) throws java.io.IOException {
		Element portTypeXML = createTag("PortType", portTypeACME, portTypeACME.getTypeName());
		Element portDescXML = exportPortDescription(portTypeACME);
		portTypeXML.appendChild(portDescXML);
		return portTypeXML;
	}
	/*
	 * Name: exportPortDescription @param portACME -- the port whose
	 * description will be exported ) and export(PortType*) to export a
	 * PortDescription, according to the following DTD: <!ELEMENT
	 * PortDescription (ExtendedPortType | PortBody)?> , PortBody?)> >
	 */
	private Element exportPortDescription(AcmePort portACME) throws java.io.IOException {
		Element portDescXML = doc.createElement("PortDescription");
		Element structurePos = portDescXML;

		if (portACME instanceof AcmeElementType) {
			if (hasSuperTypes((AcmeElementType) portACME)) {
				Element extendedXML = doc.createElement("ExtendedPortType");
				portDescXML.appendChild(extendedXML);
				structurePos = extendedXML;
				exportSuperTypes((AcmeElementType) portACME, "porttype", extendedXML);
			}
		}
		else if (hasDeclaredTypes(portACME)) {
			Element extendedXML = doc.createElement("ExtendedPortType");
			portDescXML.appendChild(extendedXML);
			structurePos = extendedXML;
			exportDeclaredTypes(portACME, "porttype", extendedXML);
		}

		AcmeEnumeration representationsACME = portACME.getRepresentations();

		if (portACME.getProperties().getSize() > 0 || representationsACME.getSize() > 0) {
			Element compBody = doc.createElement("PortBody");
			structurePos.appendChild(compBody);
			exportProperties(portACME, compBody);

			while (representationsACME.moreItems()) {
				Element representationXML = export((AcmeRepresentation) representationsACME.getNextItem());
				compBody.appendChild(representationXML);
			}
		}

		return portDescXML;
	}

	/*
	 * Name: export @param roleTypeACME - the role type to export Purpose:
	 * exports the role type according to the following DTD: )> <!ATTLIST
	 * RoleType identifier ID #REQUIRED name CDATA #IMPLIED>
	 */
	private Element export(AcmeRoleType roleTypeACME) throws java.io.IOException {
		Element roleTypeXML = createTag("RoleType", roleTypeACME, roleTypeACME.getTypeName());
		Element roleDescXML = exportRoleDescription(roleTypeACME);
		roleTypeXML.appendChild(roleDescXML);
		return roleTypeXML;
	}
	/*
	 * Name: exportRoleDescription @param roleACME -- the role whose
	 * description will be exported ) and export(RoleType*) to export a
	 * RoleDescription, according to the following DTD: <!ELEMENT
	 * RoleDescription (ExtendedRoleType | RoleBody)?> , RoleBody?)> >
	 */
	private Element exportRoleDescription(AcmeRole roleACME) throws java.io.IOException {
		Element roleDescXML = doc.createElement("RoleDescription");
		Element structurePos = roleDescXML;

		if (roleACME instanceof AcmeElementType) {
			if (hasSuperTypes((AcmeElementType) roleACME)) {
				Element extendedXML = doc.createElement("ExtendedRoleType");
				roleDescXML.appendChild(extendedXML);
				structurePos = extendedXML;
				exportSuperTypes((AcmeElementType) roleACME, "roletype", extendedXML);
			}
		}
		else if (hasDeclaredTypes(roleACME)) {
			Element extendedXML = doc.createElement("ExtendedRoleType");
			roleDescXML.appendChild(extendedXML);
			structurePos = extendedXML;
			exportDeclaredTypes(roleACME, "roletype", extendedXML);
		}

		AcmeEnumeration representationsACME = roleACME.getRepresentations();

		if (roleACME.getProperties().getSize() > 0 || representationsACME.getSize() > 0) {
			Element compBody = doc.createElement("RoleBody");
			structurePos.appendChild(compBody);
			exportProperties(roleACME, compBody);

			while (representationsACME.moreItems()) {
				Element representationXML = export((AcmeRepresentation) representationsACME.getNextItem());
				compBody.appendChild(representationXML);
			}
		}

		return roleDescXML;
	}

	/*
	 * Name: export @param attachment - and attachment to be exported Purpose:
	 * export an attachment according to the following DTD: )> <!ATTLIST
	 * Attachment identifier ID #REQUIRED portID CDATA #REQUIRED roleID CDATA
	 * #REQUIRED>
	 */
	private Element export(AcmeAttachment attachmentACME) throws java.io.IOException {
		Element attachmentXML = doc.createElement("Attachment");
		attachmentXML.setAttribute("identifier", generateID(null));
		attachmentXML.setAttribute("portID", elementToId.get(attachmentACME.getPort()).toString());
		attachmentXML.setAttribute("roleID", elementToId.get(attachmentACME.getRole()).toString());
		exportProperties(attachmentACME, attachmentXML);
		return attachmentXML;
	}

	/*
	 * Name: export @param representation - a Representation to export Purpose:
	 * export and ADML representation according to the following DTD: <!ELEMENT
	 * Representation (System, BindingMap?)> <!ATTLIST Representation
	 * identifier ID #REQUIRED name CDATA #IMPLIED> )> <!ATTLIST BindingMap
	 * identifier ID #REQUIRED name CDATA #IMPLIED> )> <!ATTLIST Binding
	 * SourceID CDATA #REQUIRED TargetID CDATA #REQUIRED>
	 */
	private Element export(AcmeRepresentation representationACME) throws java.io.IOException {
		Element representationXML = createTag("Representation", representationACME, representationACME.getName());
		Element systemXML = export(representationACME.getSystem());
		representationXML.appendChild(systemXML);

		Element bindingsMapXML = doc.createElement("BindingsMap");
		bindingsMapXML.setAttribute("identifier", generateID(null));
		AcmeBindingsMap bindingMap = (AcmeBindingsMap) representationACME.getAbstractionMap();
		AcmeEnumeration bindingsACME = bindingMap.getBindings();
		while (bindingsACME.moreItems()) {
			AcmeBinding bindingACME = (AcmeBinding) bindingsACME.getNextItem();
			AcmeElement inner = bindingACME.getInner();
			AcmeElement outer = bindingACME.getOuter();
			Element bindingXML = doc.createElement("Binding");
			bindingXML.setAttribute("SourceID", elementToId.get(outer).toString());
			bindingXML.setAttribute("TargetID", elementToId.get(inner).toString());
			exportProperties(bindingACME, bindingXML);
			bindingsMapXML.appendChild(bindingXML);
		}
		representationXML.appendChild(bindingsMapXML);
		return representationXML;
	}

	public void setFamilyAction (int b) {
		familyAction = b;
	}
	
}
