package acme.translators;
import acmeLib.*;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.core.container.*;
import acmeLib.representations.*;
import java.util.*;
import java.io.*;
import acmeLib.core.io.*;
/**
 * Class: ADMLExporter<br>
 * Purpose: Takes an AcmeLib structure and converts it to the <br>
 *          ADML XML format <br>
 * Revision History:<br>
 *   Date    Person    Change:<br>
 * 6/12/01     BRS       Created initial version <br>
 *
 *
 * Known bugs/omissions:<br>
 * - Should create a DOM structure and then serialize it
 * @author Bradley Schmerl
 */
public class ADMLExporter_noDOM {


    AcmeOutputStream out = null;
    Hashtable elementToId = null;
    private int uniqueId = 1;
    
    public ADMLExporter_noDOM (AcmeOutputStream s) {
	out = s;
	elementToId = new Hashtable ();
    }
    
    public void writeToFile (String path) {
	try {
	    File f = new File (path);
	    FileOutputStream fo = new FileOutputStream (f);
	    out.flush ();
	    out.writeTo (fo);
	    fo.close ();
	}
	catch (Exception e) {
	    e.printStackTrace ();
	}
    }

    /* Name: removeXMLSpecials
       @param st - the string to remove the XML special characters from
       @return a copy of st with the XML special characters replaced with XML entities
       Purpose: Filters a string so that special characters are replaced by their
	          corresponding XML entities
						& --> &amp;
						< --> &lt;
						> --> &gt;
    */
    private String removeXMLSpecials (String st) {
	StringBuffer editable = new StringBuffer ();
	
	int i = 0;
	while (i < st.length ()) {
	    char c = st.charAt (i);
	    if (c == '<') {
		editable.append ("&lt;");
	    }
	    else if (c == '>') {
		editable.append ("&gt;");
	    }
	    else if (c == '&') {
		editable.append ("&amp;");
	    }
	    else {
		editable.append (c);
	    }
	    i++;
	}
	return editable.toString ();
    }
    
    private String generateID (AcmeObject element) {
	String result = "" + uniqueId++;
	if (element != null) {
	    elementToId.put (element, result);
	}
	return result;
    }
    
    /* Name: generateOpeningTag
       @param tagName -- the name of the tag to open
       @param object -- the Acme object to be associated with this tag
       @param name -- the value of the the name attribute for the tag 
       (null => no name attribute output)
       @closeTag -- true => the tag is to be finished with ">", 
       false => there are more attributes to come
       Purpose: Outputs an opening tag that looks like
       <tagName identifier='uniqueId' name='name'>
    */
    private void generateOpeningTag (String tagName, AcmeObject object, 
				     String name, boolean closeTag)
	throws java.io.IOException 
    {
	out.write ("<" + tagName + " identifier='" +
		   generateID (object));
	if (name != null) {
	    out.write ("' name='" + name);
	}
	out.write ("'");
	if (closeTag) {
	    out.writeln (">");
	}
	else {
	    out.writeln (" "); // leave space for additional tags
	}
    }
    
    /* Name: export
       @param designACME - An Acme Design element, which contains the types, 
       etc, to be exported
       Purpose: Produces the XML for the root Acme design object. 
       This particular procedure supports the following DTD
       <!ELEMENT Design (ImportPathName | ComponentType | ConnectorType | 
       PortType | RoleType | PropertyTypeDeclaration| 
       Family | System | MetaProperty)*>
       <!ATTLIST Design identifier ID #REQUIRED name CDATA #IMPLIED>
    */
    public void exportDesign (AcmeDesign designACME) 
	throws java.io.IOException 
    {
	out.writeln ("<?xml version=\"1.0\"?>");
	out.writeln ("<!-- Automatically generated by ADMLExporter v1.0 using ADML verions 1.0 -->");
	out.write ("<!-- Generated ");
	GregorianCalendar cal = new GregorianCalendar (TimeZone.getDefault ());
	out.writeln (cal.get (Calendar.MONTH) + "/" + 
		     cal.get (Calendar.DAY_OF_MONTH) + "/" + 
		     cal.get (Calendar.YEAR) + " -->");
	
	generateOpeningTag ("Design", designACME, null, true);
	out.indent ();
	
	exportTypespace (designACME.getTypes ());

	AcmeEnumeration familiesACME = designACME.getFamilies ();
	while (familiesACME.moreItems ()) {
	    export ((AcmeFamily )familiesACME.getNextItem ());
	}
	
	AcmeEnumeration systemsACME = designACME.getSystems ();
	while (systemsACME.moreItems ()) {
	    export((AcmeSystem )systemsACME.getNextItem ());
	}
	
	out.unIndent ();
	out.writeln ("</Design>");
    }
    
    /** Name: export
	@param family -- An Acme Family
	Purpose: Exports the family to ADML according to the following DTD:
	<!ELEMENT Family (ComponentType | ConnectorType | PortType | 
	RoleType | PropertyTypeDeclaration | Component |   
	Connector | Port | Role | Property | Attachments | 
	Representation | MetaProperty )+>
	<!ATTLIST Family identifier ID #REQUIRED name CDATA #IMPLIED>
    **/
    private void export (AcmeFamily familyACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("Family", familyACME, familyACME.getName (), true);
	out.indent ();
	AcmeEnumeration typesAcme = familyACME.getTypes ();
	exportTypespace (typesAcme);
	exportProperties (familyACME);
	exportConstituents (familyACME);
	out.unIndent ();
	out.writeln ("</Family>");
    }
    
    /** Name: exportConstituents
	@param systemACME - the system to export the constituents of
	Purpose: exports the component, connectors, attachment and representations of the system 
	Used by both export (AcmeSystem) and export (AcmeFamily).
    */
    private void exportConstituents (AcmeSystem systemACME)
	throws java.io.IOException 
    {
	// Get and export the components
	AcmeEnumeration componentsACME = systemACME.getComponents ();
	while (componentsACME.moreItems ()) {
	    export ((AcmeComponent )componentsACME.getNextItem ());
	}
	
	// Get and export the connectors
	AcmeEnumeration connectorsACME = systemACME.getConnectors ();
	while (connectorsACME.moreItems ()) {
	    export ((AcmeConnector )connectorsACME.getNextItem ());
	}
	
	out.writeln ("<Attachments identifier='" + generateID (null) + "'>");
	
	// Get and export the attachments
	AcmeEnumeration attachmentsACME = systemACME.getAttachments ();
	while (attachmentsACME.moreItems ()) {
	    export ((AcmeAttachment )attachmentsACME.getNextItem ());
	}
	out.writeln ("</Attachments>");
	AcmeEnumeration representationsACME = systemACME.getRepresentations ();
	while (representationsACME.moreItems ()) {
	    export ((AcmeRepresentation )representationsACME.getNextItem ());
	}
    }
    
    /** Name: export
	@param types - the types space that is to be exported
	Purpose: exports component, connector, port, role and property types
    **/
    private void exportTypespace (AcmeEnumeration typesACME)
	throws java.io.IOException 
    {
	typesACME.sortByCreationOrder ();
	typesACME.reset ();
	while (typesACME.moreItems ()) {
	    AcmeType typeACME = (AcmeType )typesACME.getNextItem ();
	    if (typeACME instanceof AcmeComponentType) {
		export ((AcmeComponentType )typeACME);
	    }
	}
	
	typesACME.reset ();
	while (typesACME.moreItems ()) {
	    AcmeType typeACME = (AcmeType )typesACME.getNextItem ();
	    if (typeACME instanceof AcmeConnectorType) {
		export ((AcmeConnectorType )typeACME);
	    }
	}
	
	typesACME.reset ();
	while (typesACME.moreItems ()) {
	    AcmeType typeACME = (AcmeType )typesACME.getNextItem ();
	    if (typeACME instanceof AcmePortType) {
		export ((AcmePortType )typeACME);
	    }
	}
	
	typesACME.reset ();
	while (typesACME.moreItems ()) {
	    AcmeType typeACME = (AcmeType )typesACME.getNextItem ();
	    if (typeACME instanceof AcmeRoleType) {
		export ((AcmeRoleType )typeACME);
	    }
	}
	
	typesACME.reset ();
	while (typesACME.moreItems ()) {
	    AcmeType typeACME = (AcmeType )typesACME.getNextItem ();
	    if (typeACME instanceof AcmePropertyType) {
		export ((AcmePropertyType )typeACME, true);
	    }
	}
    }
    

    private boolean exportSuperTypes (AcmeElementType elementACME,
				      String tag,
				      String type)
	throws java.io.IOException 
    {
	boolean hasTypes = false;
	AcmeEnumeration typesACME = 
	    elementACME.getSuperTypes ().enumerate (new acmeLib.core.filters.NoBaseClassesEnumFilter ());
	if (typesACME.getSize () != 0) {
	    hasTypes = true;
	    while (typesACME.moreItems ()) {
		out.writeln ("<" + tag + ">");
		out.indent ();
		out.writeln ("<ElementID identifier='" + generateID (null) +
			     "' type='" + type + "'>");
		out.indent ();
		AcmeElementType t = (AcmeElementType )typesACME.getNextItem ();
		out.writeln (elementToId.get (t).toString ());
		out.unIndent ();
		out.writeln ("</ElementID>");
	    }
	}
	return hasTypes;
    }

    private boolean exportDeclaredTypes (AcmeElement elementACME,
					 String tag,
				      String type)
	throws java.io.IOException 
    {
	boolean hasTypes = false;
	if (elementACME.getDeclaredTypes ().getSize () != 0) {
	    hasTypes = true;
	    AcmeEnumeration typesACME = elementACME.getDeclaredTypes ();
	    while (typesACME.moreItems ()) {
		out.writeln ("<" + tag + ">");
		out.indent ();
		out.writeln ("<ElementID identifier='" + generateID (null) +
			     "' type='" + type + "'>");
		out.indent ();
		out.writeln (elementToId.get (typesACME.getNextItem ()).toString ());
		out.unIndent ();
		out.writeln ("</ElementID>");
	    }
	}
	return hasTypes;
    }

    /** Name: export
	@param systemACME - the acme system to be exported
	Purpose: Exports the system, according to the following DTD
	<!ELEMENT System (SystemDescription, MetaProperty*)>
	<!ATTLIST System identifier ID #REQUIRED name CDATA #IMPLIED>
	<!ElEMENT SystemDescription (ExtendedSystemType | SystemStructure)?>
	<!ELEMENT ExtendedSystemType (ElementID*, SystemStructure?)>
	<!ELEMENT SystemStructure (Component | Connector | Property | 
	Attachments | Representation)*>
    **/
    private void export (AcmeSystem systemACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("System", systemACME, systemACME.getName (), true);
	out.indent ();
	out.writeln ("<SystemDescription>");
	out.indent ();
	boolean extendsTypes =  exportDeclaredTypes (systemACME, "ExtendedSystemType", "family");
	
	if (systemACME.getComponents ().getSize () != 0 ||
	    systemACME.getConnectors ().getSize () != 0 ||
	    systemACME.getRoles ().getSize () != 0 ||
	    systemACME.getPorts ().getSize () != 0 ||
	    systemACME.getProperties ().getSize () != 0 ) {
	    // The system has some structure
	    out.writeln ("<SystemStructure>");
	    out.indent ();
	    exportProperties (systemACME);
	    exportConstituents (systemACME);
	    out.unIndent ();
	    out.writeln ("</SystemStructure>");
	}
	
	if (extendsTypes) { // Close the ExtendedSystemType
	    out.unIndent ();
	    out.writeln ("</ExtendedSystemType>");
	}
	out.unIndent ();
	out.writeln ("</SystemDescription>");
	out.writeln ("</System>");
    }
    
    /**
       Name: export
       @param componentACME - the component to export
       Purpose: exports ADML of component according to the folllowing DTD
       <!ELEMENT Component (ComponentDescription, MetaProperty)>
       <!ATTLIST Component identifier ID #REQUIRED name CDATA #IMPLIED>
    */
    private void export (AcmeComponent componentACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("Component", componentACME, 
			    componentACME.getName (), true);
	out.indent ();
	exportComponentDescription (componentACME);
	out.unIndent ();
	out.writeln ("</Component>");
    }
 
    /* Name: export
	 @param connectorACME - the connector to export
	 Purpose: Outputs the connector according to the following DTD
	<!ELEMENT Connector (ConnectorDescription, MetaProperty*)>
	<!ATTLIST Connector identifier ID #REQUIRED name CDATA #IMPLIED>

    */
    private void export (AcmeConnector connectorACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("Connector", connectorACME, 
			    connectorACME.getName (), true);
	out.indent ();
	exportConnectorDescription (connectorACME);
	out.unIndent ();
	out.writeln ("</Connector>");
    }

    /* name: export
       @param roleACME - the role to export
       Purpose: Exports a role according to the following DTD:
	<!ELEMENT Role (RoleDescription, MetaProperty*)>
	<!ATTLIST Role identifier ID #REQUIRED name CDATA #IMPLIED>
    */
    private void export (AcmeRole roleACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("Role", roleACME, roleACME.getName (), true);
	out.indent ();
	exportRoleDescription (roleACME);
	out.unIndent ();
	out.writeln ("</Role>");

    }

    /*
      <!ELEMENT Port (PortDescription, MetaProperty*)>
      <!ATTLIST Port identifier ID #REQUIRED name CDATA #IMPLIED>

    */
    private void export (AcmePort portACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("Port", portACME, portACME.getName (), true);
	out.indent ();
	exportPortDescription (portACME);
	out.unIndent ();
	out.writeln ("</Port>");
    }

    /* Name: exportProperties
       @param element -- an acme element that contains properties (e.g., System, Connector, Port)
       Purpose: Export the property component of an element according to the following DTD
	<!ELEMENT Property (PropertyType?, PropertyValue?, MetaProperty*)>
	<!ATTLIST Property identifier ID #REQUIRED name CDATA #implied>
    */   
    private void exportProperties (AcmeElement element)
	throws java.io.IOException 
    {
	AcmeEnumeration propertiesACME = element.getProperties ();
	while (propertiesACME.moreItems ()) {
	    AcmeProperty propertyACME = 
		(AcmeProperty )propertiesACME.getNextItem ();
	    generateOpeningTag ("Property", propertyACME, 
				propertyACME.getName (), true);
	    out.indent ();
	    export (propertyACME.getType (), false); // false means don't export as a type declaration
	    // export the property's value (if it has one)
	    if (propertyACME.getValue () != null) {
		export (propertyACME.getValue ());
	    }
	    out.unIndent ();
	    out.writeln ("</Property>");
	}
    }

    /** Name: export
	@param propertyValueACME -- the property value to export
	Purpose: Exports the ADML for a property value according to the following DTD:
	<!ELEMENT PropertyValue (PropertyLiteralValue | PropertySetValue | 
	                         PropertySequenceValue | ProperyRecordValue)>
	<!ELEMENT PropertyLiteralValue EMPTY>
	<!ATTLIST PropertyLiteralValue identifier ID #REQUIRED 
	                               name CDATA #IMPLIED 
				       value CDATA #IMPLIED>
	<!ELEMENT PropertySetValue (PropertyValue*)>
	<!ATTLIST PropertySetValue identifier ID #REQUIRED name CDATA #IMPLIED>
	<!ELEMENT PropertySequenceValue (PropertySequenceElement*)>
	<!ATTLIST PropertySequenceValue identifier ID #REQUIRED 
	                                name CDATA #IMPLIED>
	<!ELEMENT PropertySequenceElement (PropertyValue*)>
	<!ATTLIST PropertySequenceElement identifier ID #REQUIRED 
	                                  name CDATA #IMPLIED 
					  position CDATA #REQUIRED>
	<!ELEMENT PropertyRecordValue (PropertyFieldValue*)>
	<!ATTLIST PropertyRecordValue identifier ID #REQUIRED 
	                              name CDATA #IMPLIED>
	<!ELEMENT PropertyFieldValue (PropertyType? , PropertyValue)>
	<!ATTLIST PropertyFieldValue identifier ID #REQUIRED 
	                             name CDATA #IMPLIED 
				     fieldName CDATA #REQUIRED>
    */
    private void export (AcmePropertyValue propertyValueACME)
	throws java.io.IOException 
    {
	out.writeln ("<PropertyValue>");
	out.indent ();
	
	if (propertyValueACME instanceof AcmeInt) {
	    generateOpeningTag ("PropertyLiteralValue", null, null, false);
	    out.writeln ("value='" + propertyValueACME.toString () +
			 "'/>");
	}
	else if (propertyValueACME instanceof AcmeFloat) {
	    generateOpeningTag ("PropertyLiteralValue", null, null, false);
	    out.writeln ("value='" + propertyValueACME.toString () +
			 "'/>");
	}
	else if (propertyValueACME instanceof AcmeString) {
	    generateOpeningTag ("PropertyLiteralValue", null, null, false);
	    out.writeln ("value='" + 
			 removeXMLSpecials (propertyValueACME.toString ()) +
			 "'/>");
	}
	else if (propertyValueACME instanceof AcmeBoolean) {
	    generateOpeningTag ("PropertyLiteralValue", null, null, false);
	    out.writeln ("value='" + propertyValueACME.toString () +
			 "'/>");
	}
	else if (propertyValueACME instanceof AcmeSet) {
	    // Generates <PropertySetValue ...>...
	    AcmeSet setValACME = (AcmeSet )propertyValueACME;
	    generateOpeningTag ("PropertySetValue", null, null, true);
	    out.indent ();
	    AcmeEnumeration valuesACME = setValACME.getValues ();
	    while (valuesACME.moreItems ()) {
		export ((AcmePropertyValue )valuesACME.getNextItem ());
	    }
	    out.unIndent ();
	    out.writeln ("</PropertySetValue>");
			
	}
	else if (propertyValueACME instanceof AcmeSequence) {
	    AcmeSequence seqValACME = (AcmeSequence )propertyValueACME;
	    generateOpeningTag ("PropertySequenceValue", null, null, true);
	    out.indent ();
	    AcmeEnumeration valuesACME = seqValACME.getValues ();
	    while (valuesACME.moreItems ()) {
		export ((AcmePropertyValue )valuesACME.getNextItem ());
	    }
	    out.unIndent ();
	    out.writeln ("</PropertySequenceValue>");
	}
	else if (propertyValueACME instanceof AcmeEnum) {
	    generateOpeningTag ("PropertyLiteralValue", null, null, false);
	    out.writeln ("value='" + propertyValueACME.toString () + "/>");
	}
	else if (propertyValueACME instanceof AcmeRecord) {
	    AcmeRecord recValACME = (AcmeRecord )propertyValueACME;
	    generateOpeningTag ("PropertyRecordValue", null, null, true);
	    out.indent ();
	    AcmeEnumeration fieldsACME = recValACME.getFields ();
	    while (fieldsACME.moreItems ()) {
		AcmeRecordField fieldACME = 
		    (AcmeRecordField )fieldsACME.getNextItem ();
		generateOpeningTag ("PropertyFieldValue", null, 
				    fieldACME.getName (), true);
		out.indent ();
		export (fieldACME.getType (), false);
		export (fieldACME.getValue ());
		out.unIndent ();
		out.writeln ("</PropertyFieldValue>");
	    }
	    out.unIndent ();
	    out.writeln ("</PropertyRecordValue>");
	}
	
	out.unIndent ();
	out.writeln ("</PropertyValue>");
    }
    
    /* Name: export
       @param propertyType - the propertyType to export
       @param isDecl - true if this is a declaration of a property type, false if it is a reference
	<!ELEMENT PropertyTypeDeclaration (PropertyType, MetaProperty*)>
	<!ATTLIST PropertyTypeDeclaration identifier ID #REQUIRED 
	                                  name CDATA #IMPLIED>

	<!ELEMENT PropertyType (PropertyEnumType | PropertyLiteralType | 
	                        PropertySetType|PropertySequenceType | 
				PropertyRecordType | ElementID)>
	<!ATTLIST PropertyType identifier ID #REQUIRED name CDATA #IMPLIED>
	<!ELEMENT PropertyEnumType (ElementID*)>
	<!ATTLIST PropertyEnumType identifier ID #REQUIRED name CDATA #IMPLIED>
	<!ELEMENT PropertyLiteralType EMPTY>
	<!ATTLIST PropertyLiteralType identifier ID #REQUIRED 
	                              name CDATA #IMPLIED
				      type (int | float | string | boolean | any | identifier | uri | char) #REQUIRED>
	<!ELEMENT PropertySetType (PropertyType?)>
	<!ATTLIST PropertySetType identifier ID #REQUIRED name CDATA #REQUIRED>
	<!ELEMENT PropertySequenceType (PropertyType?)>
	<!ATTLIST PropertySequenceType identifier ID #REQUIRED 
	                               name CDATA #REQUIRED>
	<!ELEMENT PorpertyRecordType (PropertyFieldDeclaration*)>
	<!ATTLIST PropertyRecordType identifier ID #REQUIRED 
	                             name CDATA #REQUIRED>
	<!ELEMENT PropertyFieldDeclaration (PropertyType?)>
	<!ELEMENT ProeprtyFieldDeclaration identifier ID #REQUIRED 
	                                   name CDATA #IMPLIED>
    */
    private void export (AcmePropertyType propertyTypeACME, boolean isDecl)
	throws java.io.IOException 
    {
	if (isDecl) {
	    // Declaring a type instead of referring to it
	    generateOpeningTag ("PropertyTypeDeclaration",
				propertyTypeACME, 
				propertyTypeACME.getTypeName (), true);
	    out.indent ();
	}
	generateOpeningTag ("PropertyType", null, null, true);
	out.indent ();
	
	if (!isDecl && propertyTypeACME.getTypeName () != null) {
	    // This is a reference to a type so just output the 
	    // ElementID reference
	    out.writeln ("<ElementID identifier='" + generateID (null) +
		       "' type='propertytype' name='" +
		       propertyTypeACME.getTypeName () + "'>");
	    out.writeln (propertyTypeACME.getTypeName ());
	    out.writeln ("</ElementID>");
	}
	else if (propertyTypeACME instanceof EnumType) {
	    EnumType eTypeACME = (EnumType )propertyTypeACME;
	    generateOpeningTag ("PropertyEnumType", null, null, true);
	    out.indent ();
	    
	    AcmeEnumeration valuesACME = eTypeACME.getValues ();
	    while (valuesACME.moreItems ()) {
		AcmeEnum valACME = (AcmeEnum )valuesACME.getNextItem ();
		out.writeln ("<ElementID identifier='" +
			     generateID (null) +
			     "' name='" + valACME.toString () +
			     "'>");
		out.writeln (valACME.toString ());
		out.writeln ("</ElementID>");
	    }
	    out.unIndent ();
	    out.writeln ("</PropertyEnumType>");
	}
	else if (propertyTypeACME instanceof FloatType) {
	    out.writeln ("<PropertyLiteralType type='float' identifier='" +
			 generateID (null) + "'/>");
	}
	else if (propertyTypeACME instanceof BooleanType) {
	    out.writeln ("<PropertyLiteralType type='boolean' identifier='" +
			 generateID (null) + "'/>");
	}
	else if (propertyTypeACME instanceof IntType) {
	    out.writeln ("<PropertyLiteralType type='int' identifier='" +
			 generateID (null) + "'/>");
	}
	else if (propertyTypeACME instanceof StringType) {
	    out.writeln ("<PropertyLiteralType type='string' identifier='" +
			 generateID (null) + "'/>");
	}
	else if (propertyTypeACME instanceof SetType) {
	    SetType sTypeACME = (SetType )propertyTypeACME;
	    generateOpeningTag ("PropertySetType", null, null, true);
	    out.indent ();
	    export ((AcmePropertyType )sTypeACME.getSetType (), false);
	    out.unIndent ();
	    out.writeln ("</PropertySetType>");
	}
	else if (propertyTypeACME instanceof SequenceType) {
	    SequenceType sTypeACME = (SequenceType )propertyTypeACME;
	    generateOpeningTag ("PropertySequenceType", null, null, true);
	    out.indent ();
	    export ((AcmePropertyType )sTypeACME.getSequenceType (), false);
	    out.unIndent ();
	    out.writeln ("</PropertySequenceType>");
	}
	else if (propertyTypeACME instanceof RecordType) {
	    RecordType rTypeACME = (RecordType )propertyTypeACME;
	    generateOpeningTag ("PropertyRecordType", null, null, true);
	    out.indent ();
	    AcmeEnumeration fieldsACME = rTypeACME.getFields ();
	    while (fieldsACME.moreItems ()) {
		AcmeRecordField fieldACME = 
		    (AcmeRecordField )fieldsACME.getNextItem ();
		generateOpeningTag ("RecordFieldDeclaration", null, 
				    fieldACME.getName (), true);
		out.indent ();
		export (fieldACME.getType (), false);
		out.unIndent ();
		out.writeln ("</PropertyFieldDeclaration>");
	    }
	    out.unIndent ();
	    out.writeln ("</PropertyRecordType>");
	}
	out.unIndent ();
	out.writeln ("</PropertyType>");
	
	if (isDecl) {
	    out.unIndent ();
	    out.writeln ("</PropertyTypeDeclaration>");
	}
    }

    /* Name: export
       @param componentTypeACME - the component type to export
       Purpose: exports the component type according to the following DTD:
	<!ELEMENT ComponentType (ComponentDescription, MetaProperty*)>
	<!ATTLIST ComponentType identifier ID #REQUIRED name CDATA #IMPLIED>
    */  
    private void export (AcmeComponentType componentTypeACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("ComponentType", componentTypeACME,
			    componentTypeACME.getTypeName (), true);
	out.indent ();
	exportComponentDescription (componentTypeACME);
	out.unIndent ();
	out.writeln ("</ComponentType>");
    }
    /* Name: exportComponentDescription
       @param componentACME -- the component whose description will be exported
       Purpose: Used by export(Component*) and export(ComponentType*) to export a  ComponentDescription, according to the following DTD:
       <!ELEMENT ComponentDescription (ExtendedComponentType | ComponentBody)?>
       <!ELEMENT ExtendedComponentType (ElementID*, ComponentBody?)>
       <!ELEMENT ComponentBody (Port | Property | Representation)*>
    */
    private void exportComponentDescription (AcmeComponent componentACME)
	throws java.io.IOException 
    {
	
	out.writeln ("<ComponentDescription>");
	out.indent ();
	
	boolean hasTypes = false;
	
	if (componentACME instanceof AcmeElementType) {
	    hasTypes = exportSuperTypes ((AcmeElementType )componentACME, 
					 "ExtendedComponentType",
					 "componenttype");
	}
	else {
	    hasTypes = exportDeclaredTypes (componentACME,
					    "ExtendedComponentType",
					    "componenttype");
	}
	AcmeEnumeration portsACME = componentACME.getPorts ();
	AcmeEnumeration representationsACME = 
	    componentACME.getRepresentations ();
	
	if (portsACME.getSize () > 0 || 
	    componentACME.getProperties ().getSize () > 0 ||
	    representationsACME.getSize () > 0) {
	    out.writeln ("<ComponentBody>");
	    out.indent ();
	    exportProperties (componentACME);
	    
	    while (portsACME.moreItems ()) {
		export ((AcmePort )portsACME.getNextItem ());
	    }

	    while (representationsACME.moreItems ()) {
		export ((AcmeRepresentation )representationsACME.getNextItem ());
	    }
	    out.unIndent ();
	    out.writeln ("</ComponentBody>");
	}
	
	if (hasTypes) {
	    out.unIndent ();
	    out.writeln ("</ExtendedComponentType>");
	}
	out.unIndent ();
	out.writeln ("</ComponentDescription>");
    }


    /* Name: export
       @param connectorTypeACME - the connector type to export
       Purpose: exports the connector type according to the following DTD:
	<!ELEMENT ConnectorType (ConnectorDescription, MetaProperty*)>
	<!ATTLIST ConnectorType identifier ID #REQUIRED name CDATA #IMPLIED>
    */  
    private void export (AcmeConnectorType connectorTypeACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("ConnectorType", connectorTypeACME,
			    connectorTypeACME.getTypeName (), true);
	out.indent ();
	exportConnectorDescription (connectorTypeACME);
	out.unIndent ();
	out.writeln ("</ConnectorType>");
    }
    /* Name: exportConnectorDescription
       @param connectorACME -- the connector whose description will be exported
       Purpose: Used by export(Connector*) and export(ConnectorType*) to export a  ConnectorDescription, according to the following DTD:
       <!ELEMENT ConnectorDescription (ExtendedConnectorType | ConnectorBody)?>
       <!ELEMENT ExtendedConnectorType (ElementID*, ConnectorBody?)>
       <!ELEMENT ConnectorBody (Role | Property | Representation)*>
    */
    private void exportConnectorDescription (AcmeConnector connectorACME)
	throws java.io.IOException 
    {
	
	out.writeln ("<ConnectorDescription>");
	out.indent ();
	
	boolean hasTypes = false;
	if (connectorACME instanceof AcmeElementType) {
	    hasTypes = exportSuperTypes ((AcmeElementType )connectorACME, 
					 "ExtendedConnectorType",
					 "connectortype");
	}
	else {
	    hasTypes = exportDeclaredTypes (connectorACME,
					    "ExtendedConnectorType",
					    "connectortype");
	}
	AcmeEnumeration rolesACME = connectorACME.getRoles ();
	AcmeEnumeration representationsACME = 
	    connectorACME.getRepresentations ();
	
	if (rolesACME.getSize () > 0 || 
	    connectorACME.getProperties ().getSize () > 0 ||
	    representationsACME.getSize () > 0) {
	    out.writeln ("<ConnectorBody>");
	    out.indent ();
	    exportProperties (connectorACME);
	    
	    while (rolesACME.moreItems ()) {
		export ((AcmeRole )rolesACME.getNextItem ());
	    }

	    while (representationsACME.moreItems ()) {
		export ((AcmeRepresentation )representationsACME.getNextItem ());
	    }
	    out.unIndent ();
	    out.writeln ("</ConnectorBody>");
	}
	
	if (hasTypes) {
	    out.unIndent ();
	    out.writeln ("</ExtendedConnectorType>");
	}
	out.unIndent ();
	out.writeln ("</ConnectorDescription>");
    }
    
    /* Name: export
       @param portTypeACME - the port type to export
       Purpose: exports the port type according to the following DTD:
	<!ELEMENT PortType (PortDescription, MetaProperty*)>
	<!ATTLIST PortType identifier ID #REQUIRED name CDATA #IMPLIED>
    */  
    private void export (AcmePortType portTypeACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("PortType", portTypeACME,
			    portTypeACME.getTypeName (), true);
	out.indent ();
	exportPortDescription (portTypeACME);
	out.unIndent ();
	out.writeln ("</PortType>");
    }

    /* Name: exportPortDescription
       @param portACME -- the port whose description will be exported
       Purpose: Used by export(Port*) and export(PortType*) to export a  PortDescription, according to the following DTD:
       <!ELEMENT PortDescription (ExtendedPortType | PortBody)?>
       <!ELEMENT ExtendedPortType (ElementID*, PortBody?)>
       <!ELEMENT PortBody (Port | Property | Representation)*>
    */
    private void exportPortDescription (AcmePort portACME)
	throws java.io.IOException 
    {
	
	out.writeln ("<PortDescription>");
	out.indent ();
	
	boolean hasTypes = false;
	if (portACME instanceof AcmeElementType) {
	    hasTypes = exportSuperTypes ((AcmeElementType )portACME, 
					 "ExtendedPortType",
					 "porttype");
	}
	else {
	    hasTypes = exportDeclaredTypes (portACME,
					    "ExtendedPortType",
					    "porttype");
	}
	AcmeEnumeration representationsACME = 
	    portACME.getRepresentations ();
	
	if (portACME.getProperties ().getSize () > 0 ||
	    representationsACME.getSize () > 0) {
	    out.writeln ("<PortBody>");
	    out.indent ();
	    exportProperties (portACME);
	    
	    while (representationsACME.moreItems ()) {
		export ((AcmeRepresentation )representationsACME.getNextItem ());
	    }
	    out.unIndent ();
	    out.writeln ("</PortBody>");
	}
	
	if (hasTypes) {
	    out.unIndent ();
	    out.writeln ("</ExtendedPortType>");
	}
	out.unIndent ();
	out.writeln ("</PortDescription>");
    }

    /* Name: export
       @param roleTypeACME - the role type to export
       Purpose: exports the role type according to the following DTD:
	<!ELEMENT RoleType (RoleDescription, MetaProperty*)>
	<!ATTLIST RoleType identifier ID #REQUIRED name CDATA #IMPLIED>
    */  
    private void export (AcmeRoleType roleTypeACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("RoleType", roleTypeACME,
			    roleTypeACME.getTypeName (), true);
	out.indent ();
	exportRoleDescription (roleTypeACME);
	out.unIndent ();
	out.writeln ("</RoleType>");
    }

    /* Name: exportRoleDescription
       @param roleACME -- the role whose description will be exported
       Purpose: Used by export(Role*) and export(RoleType*) to export a  RoleDescription, according to the following DTD:
       <!ELEMENT RoleDescription (ExtendedRoleType | RoleBody)?>
       <!ELEMENT ExtendedRoleType (ElementID*, RoleBody?)>
       <!ELEMENT RoleBody (Role | Property | Representation)*>
    */
    private void exportRoleDescription (AcmeRole roleACME)
	throws java.io.IOException 
    {
	
	out.writeln ("<RoleDescription>");
	out.indent ();
	
	boolean hasTypes = false;
	
	if (roleACME instanceof AcmeElementType) {
	    hasTypes = exportSuperTypes ((AcmeElementType )roleACME, 
					 "ExtendedRoleType",
					 "roletype");
	}
	else {
	    hasTypes = exportDeclaredTypes (roleACME,
					    "ExtendedRoleType",
					    "roletype");
	}
	AcmeEnumeration representationsACME = 
	    roleACME.getRepresentations ();
	
	if (roleACME.getProperties ().getSize () > 0 ||
	    representationsACME.getSize () > 0) {
	    out.writeln ("<RoleBody>");
	    out.indent ();
	    exportProperties (roleACME);
	    
	    while (representationsACME.moreItems ()) {
		export ((AcmeRepresentation )representationsACME.getNextItem ());
	    }
	    out.unIndent ();
	    out.writeln ("</RoleBody>");
	}
	
	if (hasTypes) {
	    out.unIndent ();
	    out.writeln ("</ExtendedRoleType>");
	}
	out.unIndent ();
	out.writeln ("</RoleDescription>");
    }	
	 
    /* Name: export
       @param attachment - and attachment to be exported
       Purpose: export an attachment according to the following DTD:
	<!ELEMENT Attachment (Property*)>
	<!ATTLIST Attachment identifier ID #REQUIRED portID CDATA #REQUIRED roleID CDATA #REQUIRED>
    */
    private void export (AcmeAttachment attachmentACME)
	throws java.io.IOException 
    {
	out.writeln ("<Attachment identifier='" + generateID (null) +
		     "' portID='" + 
		     elementToId.get (attachmentACME.getPort ()).toString () +
		     "' roleID='" +
		     elementToId.get (attachmentACME.getRole ()).toString () +
		     "'>");
	out.indent ();
	exportProperties (attachmentACME);
	out.unIndent ();
	out.writeln ("</Attachment>");
    }

    /* Name: export
       @param representation - a Representation to export
       Purpose: export and ADML representation according to the following DTD:
	<!ELEMENT Representation (System, BindingMap?)>
	<!ATTLIST Representation identifier ID #REQUIRED name CDATA #IMPLIED>
	<!ELEMENT BindingMap (Binding*)>
	<!ATTLIST BindingMap identifier ID #REQUIRED name CDATA #IMPLIED>
	<!ELEMENT Binding (Property*)>
	<!ATTLIST Binding SourceID CDATA #REQUIRED TargetID CDATA #REQUIRED>
    */    
    private void export (AcmeRepresentation representationACME)
	throws java.io.IOException 
    {
	generateOpeningTag ("Representation", representationACME,
			    representationACME.getName (), true);
	out.indent ();
	export (representationACME.getSystem ());
	out.write ("<BindingsMap identifier='" + generateID (null) + "'>");
	out.indent ();
	AcmeBindingsMap bindingMap = 
	    (AcmeBindingsMap )representationACME.getAbstractionMap ();
	AcmeEnumeration bindingsACME = bindingMap.getBindings ();
	while (bindingsACME.moreItems ()) {
	    AcmeBinding bindingACME = (AcmeBinding )bindingsACME.getNextItem ();
	    AcmeElement inner = bindingACME.getInner ();
	    AcmeElement outer = bindingACME.getOuter ();
	    
	    out.writeln ("<Binding SourceID='" +
			 elementToId.get (outer).toString () +
			 "' TargetID='" +
			 elementToId.get (inner).toString () +
			 "'>");
	    out.indent ();
	    exportProperties (bindingACME);
	    out.unIndent ();
	    out.writeln ("</Binding>");
	}
	out.unIndent ();
	out.writeln ("</BindingsMap>");
	out.unIndent ();
	out.writeln ("</Representation>");
    }

}




