package acme.translators;
import acmeLib.*;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.core.container.*;
import acmeLib.representations.*;

import java.util.*;

import org.w3c.dom.*;

/** 
 * Class: ADMLImporter<br>
 * Purpose: Creates an AcmeLib structure for an architecture written in <br>
 *          the ADML XML format. <br>
 * Revision History:<br>
 *   Date   Person     Change<br>
 * 6/13/01    BRS       Created inital version<br>
 * 6/15/01    BRS       Made it inherit from XMLImporter<br>
 *
 * Known bugs/omissions:<br>
 * - Does not yet support the Armani constraints importation.<br>
 * - Need to integrate error reporting with AcmeLib AcmeError for<br>
 *    reporting back to AcmeStudio<br>
 * - Still assumes the the xml document is self-contained
 * @author Bradley Schmerl
 */

public class ADMLImporter extends XMLImporter {

    protected Document designXML;
    protected AcmeDesign designACME;

    protected Hashtable nameToAcme;
    protected Hashtable nameToDom;
	private Map _xmlToImportedDesignMap = new HashMap ();

    private String getElementText(Element parent) {
        Text text = null;
        NodeList children = parent.getChildNodes();
        boolean found = false;
        int i = 0;
        while (i < children.getLength() && !found) {
            Node candidate = children.item(i);
            if (candidate.getNodeType() == Node.TEXT_NODE) {
                found = true;
                text = (Text) candidate;
            }
            i++;
        }
        if (text != null) {
            return text.getNodeValue();
        }
        else {
            return "";
        }
    }

    private void copyProperties(AcmeElement from, AcmeElement to)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        AcmeEnumeration properties = from.getProperties();
        while (properties.moreItems()) {
            AcmeProperty property = (AcmeProperty) properties.getNextItem();
            to.addProperty(property);
        }
    }

    /** 
     * Adds super types from a DOM type to an Acme type
     * @param typeDOM Element
     * @param typeACME AcmeElementType
     */
    private void addSuperTypes(Element typeDOM, AcmeElement typeACME, boolean isType)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        NodeList superTypesDOM = getElementsAtThisLevel(typeDOM, "ElementID");
        for (int i = 0; i < superTypesDOM.getLength(); i++) {
            Element superTypeDOM = (Element) superTypesDOM.item(i);
            String elementId = getElementText(superTypeDOM);

            // Presumption: type is defined before use
            AcmeElementType superTypeACME = (AcmeElementType) mapNameToElementType (elementId);
            if (isType) {
                ((AcmeElementType) typeACME).addSuperType(superTypeACME);
            }
            else {
                typeACME.addDeclaredType(superTypeACME);
            }
        }
    }

    /**
	 * @param elementId
	 * @return
	 */
	private AcmeElementType mapNameToElementType(String elementId) {
		StringTokenizer tokens = new StringTokenizer (elementId, "#");
		if (!elementId.startsWith ("#")) {
			String filename = tokens.nextToken ();
		}
		String id = tokens.nextToken ();
		StringTokenizer name = new StringTokenizer (id, "::");
		return new AcmeUnresolvedElementType (name.nextToken ());
	}
	
	private AcmeObject mapNameToPropertyType (String elementID) {
		StringTokenizer tokens = new StringTokenizer (elementID, "#");
		if (!elementID.startsWith ("#")) {
			String filename = tokens.nextToken ();
		}
		String id = tokens.nextToken ();
		StringTokenizer name = new StringTokenizer (id, "::");
		return new AcmeUnresolvedPropertyType (name.nextToken ());
	}

	private void transferProperties(Element elementXML, AcmeElement elementACME)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        NodeList propertiesXML = getElementsAtThisLevel(elementXML, "Property");
        for (int i = 0; i < propertiesXML.getLength(); i++) {
            Element propertyXML = (Element) propertiesXML.item(i);
            AcmeProperty propertyACME = translateProperty(propertyXML);
            elementACME.addProperty(propertyACME);
        }
    }

    private void transferRepresentations(Element elementXML, AcmeElement elementACME)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        NodeList representationsXML = getElementsAtThisLevel(elementXML, "Representation");
        for (int i = 0; i < representationsXML.getLength(); i++) {
            Element representationXML = (Element) representationsXML.item(i);
            AcmeRepresentation representationACME = translateRepresentation(representationXML, elementACME);
            elementACME.addRepresentation(representationACME);
        }
    }

    public ADMLImporter(Document doc) {
        designXML = doc;
        designACME = new AcmeDesign();
        nameToDom = new Hashtable();
        nameToAcme = new Hashtable();

    }

    public AcmeDesign importDesign() throws acmeLib.core.exceptions.AcmeIllegalNameException {
        NodeList propertyTypesXML = getElementsAtThisLevel(designXML.getDocumentElement(), "PropertyTypeDeclaration");
        NodeList importsXML = getElementsAtThisLevel(designXML.getDocumentElement(), "ImportPathname");
        NodeList componentTypesXML = getElementsAtThisLevel(designXML.getDocumentElement(), "ComponentType");
        NodeList connectorTypesXML = getElementsAtThisLevel(designXML.getDocumentElement(), "ConnectorType");
        NodeList portTypesXML = getElementsAtThisLevel(designXML.getDocumentElement(), "PortType");
        NodeList roleTypesXML = getElementsAtThisLevel(designXML.getDocumentElement(), "RoleType");
        NodeList familiesXML = getElementsAtThisLevel(designXML.getDocumentElement(), "Family");
        NodeList systemsXML = getElementsAtThisLevel(designXML.getDocumentElement(), "System");

        int i;
        
        for (i=0; i < importsXML.getLength (); i++) {
        	Element importXML = (Element )importsXML.item(i);
        	importXML.normalize ();
        	String importFile = this.processImportedDesign (importXML);
        	if (importFile != null) {
        		designACME.addDesignToImport(importFile);
        	}
        }
        
        for (i = 0; i < propertyTypesXML.getLength(); i++) {
            Element propertyTypeXML = (Element) propertyTypesXML.item(i);
            propertyTypeXML.normalize();
            AcmeType propertyTypeACME = translatePropertyType(propertyTypeXML);
            if (propertyTypeACME != null) {
                designACME.addType(propertyTypeACME);
                nameToAcme.put(propertyTypeXML.getAttribute("identifier"), propertyTypeACME);
                nameToDom.put(propertyTypeXML.getAttribute("identiier"), propertyTypeXML);
            }
        }

        for (i = 0; i < portTypesXML.getLength(); i++) {
            Element portTypeXML = (Element) portTypesXML.item(i);
            portTypeXML.normalize();
            AcmeType portTypeACME = translatePropertyType(portTypeXML);
            if (portTypeACME != null) {
                designACME.addType(portTypeACME);
                nameToAcme.put(portTypeXML.getAttribute("identifier"), portTypeACME);
                nameToDom.put(portTypeXML.getAttribute("identiier"), portTypeXML);
            }
        }

        for (i = 0; i < roleTypesXML.getLength(); i++) {
            Element roleTypeXML = (Element) roleTypesXML.item(i);
            roleTypeXML.normalize();
            AcmeType roleTypeACME = translatePropertyType(roleTypeXML);
            if (roleTypeACME != null) {
                designACME.addType(roleTypeACME);
                nameToAcme.put(roleTypeXML.getAttribute("identifier"), roleTypeACME);
                nameToDom.put(roleTypeXML.getAttribute("identiier"), roleTypeXML);
            }
        }

        for (i = 0; i < componentTypesXML.getLength(); i++) {
            Element componentTypeXML = (Element) componentTypesXML.item(i);
            componentTypeXML.normalize();
            AcmeType componentTypeACME = translatePropertyType(componentTypeXML);
            if (componentTypeACME != null) {
                designACME.addType(componentTypeACME);
                nameToAcme.put(componentTypeXML.getAttribute("identifier"), componentTypeACME);
                nameToDom.put(componentTypeXML.getAttribute("identiier"), componentTypeXML);
            }
        }

        for (i = 0; i < connectorTypesXML.getLength(); i++) {
            Element connectorTypeXML = (Element) connectorTypesXML.item(i);
            connectorTypeXML.normalize();
            AcmeType connectorTypeACME = translatePropertyType(connectorTypeXML);
            if (connectorTypeACME != null) {
                designACME.addType(connectorTypeACME);
                nameToAcme.put(connectorTypeXML.getAttribute("identifier"), connectorTypeACME);
                nameToDom.put(connectorTypeXML.getAttribute("identiier"), connectorTypeXML);
            }
        }

        for (i = 0; i < familiesXML.getLength(); i++) {
            Element familyXML = (Element) familiesXML.item(i);
            familyXML.normalize();
            AcmeFamily familyACME = translateFamily(familyXML);
            designACME.addFamily(familyACME);
            nameToAcme.put(familyXML.getAttribute("identifier"), familyACME);
            nameToDom.put(familyXML.getAttribute("identifier"), familyXML);
        }

        for (i = 0; i < systemsXML.getLength(); i++) {
            Element systemXML = (Element) systemsXML.item(i);
            systemXML.normalize();
            AcmeSystem systemACME = translateSystem(systemXML);
            designACME.addSystem(systemACME);
            nameToAcme.put(systemXML.getAttribute("identifier"), systemACME);
            nameToDom.put(systemXML.getAttribute("identifier"), systemXML);

        }

        return designACME;
    }

    /**
	 * @param importXML
	 * @return
	 */
	private String processImportedDesign(Element importXML) {
		String filename = importXML.getAttribute ("pathname");
		if (filename.indexOf(".") != -1) {
			filename = filename.substring(0, filename.lastIndexOf ("."));
		}
		return filename;
	}

	private AcmePropertyType translatePropertyType(Element propertyTypeXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        AcmePropertyType propertyTypeACME = null;
        boolean isDeclaration = false;
        String name = "";
        if (propertyTypeXML.getNodeName().equals("PropertyTypeDeclaration")) {
            isDeclaration = true;
            name = propertyTypeXML.getAttribute("name");
            propertyTypeXML = findFirstElement(propertyTypeXML);
        }

        Element typeDetailXML = findFirstElement(propertyTypeXML);

        if (typeDetailXML.getNodeName().equals("PropertyEnumType")) {
            EnumType etypeACME = new EnumType(name);
            NodeList valuesXML = getElementsAtThisLevel(typeDetailXML, "ElementID");
            for (int i = 0; i < valuesXML.getLength(); i++) {
                Element valueXML = (Element) valuesXML.item(i);
                String realValueXML = getElementText(valueXML);
                etypeACME.addValue(realValueXML);
            }
            propertyTypeACME = etypeACME;
        }
        else if (typeDetailXML.getNodeName().equals("PropertyLiteralType")) {
            String type = typeDetailXML.getAttribute("type").toLowerCase();
            if (type.equals("float")) {
                if (name == null || name.equals("")) {
                    propertyTypeACME = acmeLib.core.globals.ConstantTypes.FLOAT_TYPE;
                }
                else {
                    propertyTypeACME = new FloatType(name);
                }
            }
            else if (type.equals("int")) {
                if (name == null || name.equals("")) {
                    propertyTypeACME = acmeLib.core.globals.ConstantTypes.INT_TYPE;
                }
                else {
                    propertyTypeACME = new IntType(name);
                }
            }
            if (type.equals("string")) {
                if (name == null || name.equals("")) {
                    propertyTypeACME = acmeLib.core.globals.ConstantTypes.STRING_TYPE;
                }
                else {
                    propertyTypeACME = new StringType(name);
                }
            }
            else if (type.equals("boolean")) {
                if (name == null || name.equals("")) {
                    propertyTypeACME = acmeLib.core.globals.ConstantTypes.BOOLEAN_TYPE;
                }
                else {
                    propertyTypeACME = new BooleanType(name);
                }
            }
        }
        else if (typeDetailXML.getNodeName().equals("PropertySetType")) {
            Element propertySetTypeXML = findFirstElement(typeDetailXML);
            AcmePropertyType setTypeACME = translatePropertyType(propertySetTypeXML);
            propertyTypeACME = new SetType(name, setTypeACME);
        }
        else if (typeDetailXML.getNodeName().equals("PropertySequenceType")) {
            Element propertySeqTypeXML = findFirstElement(typeDetailXML);
            AcmePropertyType seqTypeACME = translatePropertyType(propertySeqTypeXML);
            propertyTypeACME = new SequenceType(name, seqTypeACME);
        }
        else if (typeDetailXML.getNodeName().equals("PropertyRecordType")) {
            NodeList fieldsXML = getElementsAtThisLevel(typeDetailXML, "PropertyFieldDeclaration");
            RecordType rtypeACME = new RecordType(name);
            for (int i = 0; i < fieldsXML.getLength(); i++) {
                Element fieldXML = (Element) fieldsXML.item(i);
                Element fieldTypeXML = findFirstElement(fieldXML);
                AcmePropertyType fieldTypeACME = translatePropertyType(fieldTypeXML);
                String fieldName = fieldXML.getAttribute("name");
                rtypeACME.addField(fieldName, fieldTypeACME, null);
            }
            propertyTypeACME = rtypeACME;
        }
        else if (typeDetailXML.getNodeName().equals("ElementID")) {
            String elementId = getElementText(typeDetailXML);
            AcmePropertyType referencedACME = new AcmeUnresolvedPropertyType("This is an error");
            propertyTypeACME = referencedACME;
        }

        return propertyTypeACME;
    }

    private AcmeComponentType translateComponentType(Element componentTypeXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        return (AcmeComponentType) translateComponentDetails(componentTypeXML, true);
    }

    private AcmeComponent translateComponentDetails(Element componentXML, boolean doType)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        AcmeComponent componentACME = null;

        String name = componentXML.getAttribute("name");
        if (doType) {
            componentACME = new AcmeComponentType(name);
        }
        else {
            componentACME = new AcmeComponent(name);
        }
        Element componentBodyXML = findFirstElement(findFirstElement(componentXML));
        if (componentBodyXML != null && componentBodyXML.getNodeName().equals("ExtendedComponentType")) {
            addSuperTypes(componentBodyXML, componentACME, componentACME instanceof AcmeComponentType);
            NodeList componentBodiesXML = getElementsAtThisLevel(componentBodyXML, "ComponentBody");
            if (componentBodiesXML.getLength() > 0) {
                componentBodyXML = (Element) componentBodiesXML.item(0);
            }
        }

        if (componentBodyXML != null) {

            int i;

            NodeList portsXML = getElementsAtThisLevel(componentBodyXML, "Port");
            for (i = 0; i < portsXML.getLength(); i++) {
                Element portXML = (Element) portsXML.item(i);
                AcmePort portACME = translatePort(portXML);
                componentACME.addPort(portACME);
                nameToAcme.put(portXML.getAttribute("identifier"), portACME);
                nameToDom.put(portXML.getAttribute("identifier"), portXML);
            }
            transferProperties(componentBodyXML, componentACME);
            transferRepresentations(componentBodyXML, componentACME);
        }
        return componentACME;
    }

    private AcmeConnectorType translateConnectorType(Element connectorTypeXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        return (AcmeConnectorType) translateConnectorDetails(connectorTypeXML, true);
    }

    private AcmeConnector translateConnectorDetails(Element connectorXML, boolean doType)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        AcmeConnector connectorACME = null;

        String name = connectorXML.getAttribute("name");
        if (doType) {
            connectorACME = new AcmeConnectorType(name);
        }
        else {
            connectorACME = new AcmeConnector(name);
        }
        Element connectorBodyXML = findFirstElement(findFirstElement(connectorXML));
        if (connectorBodyXML != null) {
            if (connectorBodyXML.getNodeName().equals("ExtendedConnectorType")) {
                addSuperTypes(connectorBodyXML, connectorACME, connectorACME instanceof AcmeConnectorType);
                NodeList connectorBodiesXML = getElementsAtThisLevel(connectorBodyXML, "ConnectorBody");

                if (connectorBodiesXML.getLength() > 0) {
                    connectorBodyXML = (Element) connectorBodiesXML.item(0);
                }
            }
        }

        if (connectorBodyXML != null) {
            NodeList rolesXML = getElementsAtThisLevel(connectorBodyXML, "Role");
            for (int i = 0; i < rolesXML.getLength(); i++) {
                Element roleXML = (Element) rolesXML.item(i);
                AcmeRole roleACME = translateRole(roleXML);
                connectorACME.addRole(roleACME);
                nameToAcme.put(roleXML.getAttribute("identifier"), roleACME);
                nameToDom.put(roleXML.getAttribute("identifier"), roleXML);
            }
            transferProperties(connectorBodyXML, connectorACME);
            transferRepresentations(connectorBodyXML, connectorACME);
        }
        return connectorACME;
    }

    private AcmePortType translatePortType(Element portTypeXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        return (AcmePortType) translatePortDetails(portTypeXML, true);
    }

    private AcmePort translatePortDetails(Element portXML, boolean doType)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        AcmePort portACME = null;

        String name = portXML.getAttribute("name");
        if (doType) {
            portACME = new AcmePortType(name);
        }
        else {
            portACME = new AcmePort(name);
        }
        Element portBodyXML = findFirstElement(findFirstElement(portXML));
        if (portBodyXML != null && portBodyXML.getNodeName().equals("ExtendedPortType")) {
            addSuperTypes(portBodyXML, portACME, portACME instanceof AcmePortType);
            NodeList portBodiesXML = getElementsAtThisLevel(portBodyXML, "PortBody");
            if (portBodiesXML.getLength() > 0) {
                portBodyXML = (Element) portBodiesXML.item(0);
            }
        }

        if (portBodyXML != null) {
            transferProperties(portBodyXML, portACME);
            transferRepresentations(portBodyXML, portACME);
        }

        return portACME;
    }

    private AcmeRoleType translateRoleType(Element roleTypeXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        return (AcmeRoleType) translateRoleDetails(roleTypeXML, true);
    }

    private AcmeRole translateRoleDetails(Element roleXML, boolean doType)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        AcmeRole roleACME = null;

        String name = roleXML.getAttribute("name");
        if (doType) {
            roleACME = new AcmeRoleType(name);
        }
        else {
            roleACME = new AcmeRole(name);
        }
        Element roleBodyXML = findFirstElement(findFirstElement(roleXML));
        if (roleBodyXML != null && roleBodyXML.getNodeName().equals("ExtendedRoleType")) {
            addSuperTypes(roleBodyXML, roleACME, roleACME instanceof AcmeRoleType);
            NodeList roleBodiesXML = getElementsAtThisLevel(roleBodyXML, "RoleBody");
            if (roleBodiesXML.getLength() > 0) {
                roleBodyXML = (Element) roleBodiesXML.item(0);
            }
        }
        if (roleBodyXML != null) {
            transferProperties(roleBodyXML, roleACME);
            transferRepresentations(roleBodyXML, roleACME);
        }
        return roleACME;
    }

    private AcmeFamily translateFamily(Element familyXML) throws acmeLib.core.exceptions.AcmeIllegalNameException {
        String name = familyXML.getAttribute("name");
        AcmeFamily familyACME = new AcmeFamily(name);

        //	addSuperTypes (familyXML, 
        int i;

        NodeList propertyTypesXML = getElementsAtThisLevel(familyXML, "PropertyTypeDeclaration");
        for (i = 0; i < propertyTypesXML.getLength(); i++) {
            Element propertyTypeXML = (Element) propertyTypesXML.item(i);
            propertyTypeXML.normalize();
            AcmeType propertyTypeACME = translatePropertyType(propertyTypeXML);
            if (propertyTypeACME != null) {
                familyACME.addType(propertyTypeACME);
                nameToAcme.put(propertyTypeXML.getAttribute("identifier"), propertyTypeACME);
                nameToDom.put(propertyTypeXML.getAttribute("identifier"), propertyTypeXML);
            }
        }

        NodeList componentTypesXML = getElementsAtThisLevel(familyXML, "ComponentType");
        for (i = 0; i < componentTypesXML.getLength(); i++) {
            Element componentTypeXML = (Element) componentTypesXML.item(i);
            componentTypeXML.normalize();
            AcmeType componentTypeACME = translateComponentType(componentTypeXML);
            if (componentTypeACME != null) {
                familyACME.addType(componentTypeACME);
                nameToAcme.put(componentTypeXML.getAttribute("identifier"), componentTypeACME);
                nameToDom.put(componentTypeXML.getAttribute("identifier"), componentTypeXML);
            }
        }

        NodeList connectorTypesXML = getElementsAtThisLevel(familyXML, "ConnectorType");
        for (i = 0; i < connectorTypesXML.getLength(); i++) {
            Element connectorTypeXML = (Element) connectorTypesXML.item(i);
            connectorTypeXML.normalize();
            AcmeType connectorTypeACME = translateConnectorType(connectorTypeXML);
            if (connectorTypeACME != null) {
                familyACME.addType(connectorTypeACME);
                nameToAcme.put(connectorTypeXML.getAttribute("identifier"), connectorTypeACME);
                nameToDom.put(connectorTypeXML.getAttribute("identifier"), connectorTypeXML);
            }
        }

        NodeList portTypesXML = getElementsAtThisLevel(familyXML, "PortType");
        for (i = 0; i < portTypesXML.getLength(); i++) {
            Element portTypeXML = (Element) portTypesXML.item(i);
            portTypeXML.normalize();
            AcmeType portTypeACME = translatePortType(portTypeXML);
            if (portTypeACME != null) {
                familyACME.addType(portTypeACME);
                nameToAcme.put(portTypeXML.getAttribute("identifier"), portTypeACME);
                nameToDom.put(portTypeXML.getAttribute("identifier"), portTypeXML);
            }
        }

        NodeList roleTypesXML = getElementsAtThisLevel(familyXML, "RoleType");
        for (i = 0; i < roleTypesXML.getLength(); i++) {
            Element roleTypeXML = (Element) roleTypesXML.item(i);
            roleTypeXML.normalize();
            AcmeType roleTypeACME = translateRoleType(roleTypeXML);
            if (roleTypeACME != null) {
                familyACME.addType(roleTypeACME);
                nameToAcme.put(roleTypeXML.getAttribute("identifier"), roleTypeACME);
                nameToDom.put(roleTypeXML.getAttribute("identifier"), roleTypeXML);
            }
        }

        AcmeSystem systemACME = translateSystemStructure(familyXML);
        transferSystem(systemACME, familyACME);
        return familyACME;
    }

    private AcmeSystem translateSystem(Element systemXML) throws acmeLib.core.exceptions.AcmeIllegalNameException {
        String name = systemXML.getAttribute("name");
        AcmeSystem systemACME = new AcmeSystem(name);

        Element systemDescriptionXML = findFirstElement(systemXML);

        Element structureXML = findFirstElement(systemDescriptionXML);
        if (structureXML.getNodeName().equals("ExtendedSystemType")) {
            transferTypes(structureXML, systemACME);
            structureXML = (Element) getElementsAtThisLevel(structureXML, "SystemStructure").item(0);
        }

        AcmeSystem systemStructureACME = translateSystemStructure(structureXML);
        transferSystem(systemStructureACME, systemACME);
        return systemACME;
    }

    private AcmeSystem translateSystemStructure(Element structureXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        AcmeSystem systemACME = new AcmeSystem();
        int i;

        transferProperties(structureXML, systemACME);

        NodeList componentsXML = getElementsAtThisLevel(structureXML, "Component");
        for (i = 0; i < componentsXML.getLength(); i++) {
            Element componentXML = (Element) componentsXML.item(i);
            AcmeComponent componentACME = translateComponent(componentXML);
            systemACME.addComponent(componentACME);
            nameToAcme.put(componentXML.getAttribute("identifier"), componentACME);
            nameToDom.put(componentXML.getAttribute("identifier"), componentXML);
        }

        NodeList connectorsXML = getElementsAtThisLevel(structureXML, "Connector");
        for (i = 0; i < connectorsXML.getLength(); i++) {
            Element connectorXML = (Element) connectorsXML.item(i);
            AcmeConnector connectorACME = translateConnector(connectorXML);
            systemACME.addConnector(connectorACME);
            nameToAcme.put(connectorXML.getAttribute("identifier"), connectorACME);
            nameToDom.put(connectorXML.getAttribute("identifier"), connectorXML);
        }

        NodeList representationsXML = getElementsAtThisLevel(structureXML, "Representation");
        for (i = 0; i < representationsXML.getLength(); i++) {
            Element representationXML = (Element) representationsXML.item(i);
            AcmeRepresentation representationACME = translateRepresentation(representationXML, systemACME);
            systemACME.addRepresentation(representationACME);
        }

        NodeList attachmentsXML = getElementsAtThisLevel(structureXML, "Attachments");
        DOMNodeList unifiedAttachments = new DOMNodeList();
        for (i = 0; i < attachmentsXML.getLength(); i++) {
            Element attachmentSectionXML = (Element) attachmentsXML.item(i);
            NodeList attachmentXML = getElementsAtThisLevel(attachmentSectionXML, "Attachment");

            for (int j = 0; j < attachmentXML.getLength(); j++) {
                unifiedAttachments.addElement((Element) attachmentXML.item(j));
            }
        }

        for (i = 0; i < unifiedAttachments.getLength(); i++) {
            Element attachmentXML = (Element) unifiedAttachments.item(i);
            String portID = attachmentXML.getAttribute("portID");
            String roleID = attachmentXML.getAttribute("roleID");

            AcmePort portACME = (AcmePort) nameToAcme.get(portID);
            AcmeRole roleACME = (AcmeRole) nameToAcme.get(roleID);

            systemACME.addAttachment(portACME, roleACME);
        }

        return systemACME;

    }

    private AcmeComponent translateComponent(Element componentXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        return translateComponentDetails(componentXML, false);
    }

    private AcmeConnector translateConnector(Element connectorXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        return translateConnectorDetails(connectorXML, false);
    }

    private AcmePort translatePort(Element portXML) throws acmeLib.core.exceptions.AcmeIllegalNameException {
        return translatePortDetails(portXML, false);
    }

    private AcmeRole translateRole(Element roleXML) throws acmeLib.core.exceptions.AcmeIllegalNameException {
        return translateRoleDetails(roleXML, false);
    }

    private AcmeProperty translateProperty(Element propertyXML)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        String name = propertyXML.getAttribute("name");
        AcmeProperty propertyACME = new AcmeProperty();
        propertyACME.setName(name);
        AcmePropertyType typeACME = null;

        NodeList typesXML = getElementsAtThisLevel(propertyXML, "PropertyType");
        if (typesXML.getLength() > 0) {
            typeACME = translatePropertyType((Element) typesXML.item(0));
            propertyACME.setType(typeACME);
        }

        NodeList valuesXML = getElementsAtThisLevel(propertyXML, "PropertyValue");
        if (valuesXML.getLength() > 0) {
            AcmePropertyValue valueACME = translatePropertyValue((Element) valuesXML.item(0), typeACME);
            propertyACME.setValue(valueACME);
        }

        return propertyACME;
    }

    private AcmePropertyValue translatePropertyValue(Element propertyValueXML, AcmeType typeACME)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        AcmePropertyValue propertyValueACME = null;

        Element actualValueXML = findFirstElement(propertyValueXML);
        if (actualValueXML.getNodeName().equals("PropertyLiteralValue")) {
            if (typeACME instanceof IntType || typeACME instanceof FloatType) {

                try {
                    String input = actualValueXML.getAttribute("value");
                    if (typeACME instanceof StringType) {
                        input = "\"" + input + "\"";
                    }
                    java.io.ByteArrayInputStream valueInputStream = new java.io.ByteArrayInputStream(input.getBytes());
                    try {

                        new acmeLib.armaniParser.ArmaniParser(valueInputStream);
                    }
                    catch (java.lang.Error e) {
                        acmeLib.armaniParser.ArmaniParser.ReInit(valueInputStream);
                    }
                    acmeLib.armaniParser.ArmaniParser.disable_tracing();
                    propertyValueACME = acmeLib.armaniParser.ArmaniParser.PropertyValueDeclaration();
                    if (propertyValueACME instanceof AcmeString) {
                        propertyValueACME = new AcmeString("\"" + propertyValueACME.toString() + "\"");
                    }
                }
                catch (acmeLib.armaniParser.ParseException err) {
                    propertyValueACME = new AcmeString("\"" + actualValueXML.getAttribute("value") + "\"");
                }
            }
            /*	    if (typeACME instanceof IntType) {
            	    propertyValueACME = 
            	    new AcmeInt (Integer.parseInt (actualValueXML.getAttribute ("value")));
            	    }
            	    else if (typeACME instanceof FloatType) {
            	    propertyValueACME =
            	    new AcmeFloat (Float.parseFloat (actualValueXML.getAttribute ("value")));
            }
            */
            else if (typeACME instanceof BooleanType) {
                String value = actualValueXML.getAttribute("value");
                if (value.toLowerCase().equals("false")) {
                    propertyValueACME = new AcmeBoolean(false);
                }
                else {
                    propertyValueACME = new AcmeBoolean(true);
                }
            }
            else if (typeACME instanceof StringType) {
                propertyValueACME = new AcmeString(actualValueXML.getAttribute("value"));
            }
            else if (typeACME instanceof EnumType) {
                propertyValueACME = new AcmeEnum(actualValueXML.getAttribute("value"));
            }
        }
        else if (actualValueXML.getNodeName().equals("PropertySetValue")) {
            NodeList valuesXML = getElementsAtThisLevel(actualValueXML, "PropertyValue");
            AcmeSet setValACME = new AcmeSet();
            for (int i = 0; i < valuesXML.getLength(); i++) {
                AcmePropertyValue itemACME =
                    translatePropertyValue((Element) valuesXML.item(i), ((SetType) typeACME).getSetType());
                setValACME.addValue(itemACME);
            }
            propertyValueACME = setValACME;

        }
        else if (actualValueXML.getNodeName().equals("PropertySequenceValue")) {
            NodeList valuesXML = getElementsAtThisLevel(actualValueXML, "PropertySequenceElement");
            AcmeSequence seqValACME = new AcmeSequence();
            for (int i = 0; i < valuesXML.getLength(); i++) {
                AcmePropertyValue itemACME =
                    translatePropertyValue(
                        findFirstElement((Element) valuesXML.item(i)),
                        ((SequenceType) typeACME).getSequenceType());
                seqValACME.addValue(itemACME);
            }
            propertyValueACME = seqValACME;

        }
        else if (actualValueXML.getNodeName().equals("PropertyRecordValue")) {
            AcmeRecord recValACME = new AcmeRecord();
            NodeList fieldsXML = getElementsAtThisLevel(actualValueXML, "PropertyFieldValue");
            for (int i = 0; i < fieldsXML.getLength(); i++) {
                Element fieldXML = (Element) fieldsXML.item(i);
                String fieldName = fieldXML.getAttribute("fieldName");
                AcmePropertyType fieldTypeACME = null;
                AcmePropertyValue fieldValueACME = null;
                NodeList typesXML = getElementsAtThisLevel(fieldXML, "PropertyType");
                if (typesXML.getLength() == 1) {
                    fieldTypeACME = translatePropertyType((Element) typesXML.item(0));
                }

                NodeList valuesXML = getElementsAtThisLevel(fieldXML, "PropertyValue");
                if (valuesXML.getLength() == 1) {
                    fieldValueACME = translatePropertyValue((Element) valuesXML.item(0), fieldTypeACME);
                }
                AcmeRecordField fieldACME = new AcmeRecordField(fieldName, fieldTypeACME, fieldValueACME);
                recValACME.addField(fieldACME);
            }
            propertyValueACME = recValACME;
        }
        return propertyValueACME;
    }

    private void transferTypes(Element elementXML, AcmeElement elementACME)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        NodeList typesXML = getElementsAtThisLevel(elementXML, "ElementID");
        for (int i = 0; i < typesXML.getLength(); i++) {
            Element declaredTypeDOM = (Element) typesXML.item(i);
            String elementId = getElementText(declaredTypeDOM);

            AcmeElementType declaredTypeACME = mapNameToElementType (elementId); //new AcmeUnresolvedElementType (elementId); //(AcmeElementType) nameToAcme.get(elementId);
            elementACME.addDeclaredType(declaredTypeACME);
        }
    }

    private AcmeRepresentation translateRepresentation(Element representationXML, AcmeElement elementACME)
        throws acmeLib.core.exceptions.AcmeIllegalNameException {
        String name = representationXML.getAttribute("name");

        Element systemXML = (Element) getElementsAtThisLevel(representationXML, "System").item(0);
        AcmeSystem systemACME = translateSystem(systemXML);

        AcmeBindingsMap bindingsACME = new AcmeBindingsMap();

        NodeList bindingMapsXML = getElementsAtThisLevel(representationXML, "BindingsMap");
        if (bindingMapsXML.getLength() > 0) {
            Element bindingsMapXML = (Element) bindingMapsXML.item(0);
            NodeList bindingsXML = getElementsAtThisLevel(bindingsMapXML, "Binding");
            for (int i = 0; i < bindingsXML.getLength(); i++) {
                Element bindingXML = (Element) bindingsXML.item(0);

                AcmeElement innerACME = (AcmeElement) nameToAcme.get(bindingXML.getAttribute("SourceID"));
                if (innerACME instanceof AcmePort) {
                    AcmePort innerPortACME = (AcmePort) innerACME;
                    AcmePort outerPortACME = (AcmePort) nameToAcme.get(bindingXML.getAttribute("TargetID"));
                    AcmeBinding bindingACME = new AcmeBinding(innerPortACME, outerPortACME);
                    bindingACME.setExportOuterQualifier(false);
                    bindingsACME.addBinding(bindingACME);
                }
                else if (innerACME instanceof AcmeRole) {
                    AcmeRole innerRoleACME = (AcmeRole) innerACME;
                    AcmeRole outerRoleACME = (AcmeRole) nameToAcme.get(bindingXML.getAttribute("TargetID"));
                    AcmeBinding bindingACME = new AcmeBinding(innerRoleACME, outerRoleACME);
                    bindingACME.setExportOuterQualifier(false);
                    bindingsACME.addBinding(bindingACME);
                }
            }
        }

        AcmeRepresentation representationACME = new AcmeRepresentation(systemACME, elementACME, bindingsACME, name);

        return representationACME;
    }

	/**
	 * @param design
	 * @return
	 */
	public static boolean referencesFamilies(Element design) {
		// TODO Auto-generated method stub
		return false;
	}

	/**
	 * @param name
	 * @param des
	 */
	public void addImportedDesign(String name, AcmeDesign des) {
		_xmlToImportedDesignMap.put (name, des);
	}

}
