package acme.translators;

import java.io.File;
import java.io.FileOutputStream;
import java.util.Enumeration;
import java.util.Hashtable;

import acme.translators.RoseRTTypes.*;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;

/**
 * Class: RoseRTExporter<br>
 * Purpose: Takes an AcmeLib structure and converts it to the <br>
 *          Rational Rose Real-Time Petal (*.rtmdl) format <br>
 * Revision History:<br>
 *   Date    Person    Change:<br>
 * 5/10/01    SWC      Created initial version <br>
 *
 * Known bugs/omissions:<br>
 *   - 
 * @author Owen Cheng
 */

public class RoseRTExporter {

    static private int X_OFFSET = 25;
    static private int Y_OFFSET = 100;
    static private int VIS_FACTOR = 2;
    static private String SYSTEM_NAME = "System";
    static private String CARDINALITY = "*";
    static private String CAPSULE_STEREOTYPE = "<<Capsule>>";
    static private String CLASS_DIAGRAM_NAME = "Main";
    static private String COLLAB_NAME = "Structure";
    static private String COLLAB_DIAGRAM_NAME = "Structure Diagram";
    static private String BINDING_NAME = "bc";
    static private String STEREOTYPE_LABEL = "stereotype";
    static private String PATH_DELIM = "::";
    static private String STRUCT_DELIM = ":";
    static private String CAPVW_PREFIX = "CapVw-";
    static private String INHVW_PREFIX = "InhVw-";
    static private String RELVW_PREFIX = "RelVw-";

    private acmeLib.core.io.AcmeOutputStream out = null;
    private acmeLib.core.io.AcmeOutputStream err = null;
    private boolean headerOutput = false;
    private boolean includeVisProperties = true;
    private Hashtable idHash = null;   // store id by Acme element
    private Hashtable tidHash = null;
    private Hashtable rtObjHash = null;  // store Rose RT obj by id
    private Hashtable rtViewHash = null;  // store Rose RT view by obj
    // Tally storage, each entry will contain two more levels of Hashtable
    //   The first level Hashtable is keyed by Component type
    //   The second level Hashtable is keyed by Component instance
    //   The third level Hashtable is keyed by Port type
    //   This arrangement makes count more accessible at the expense of
    //   more complicated implementation
    // = used in deriveClassDiag();
    private Hashtable tallyHash = null;

    // Rose RT content
    private RoseRTDesign petalDesign = null;
    private String rootCatName = null;
    private RoseRTList classList = null;
    private RoseRTList instList = null;
    private RoseRTList prtclList = null;

    int nextId = 10;  // start IDs off at 10


    private static final int version = 1;

    public RoseRTExporter(AcmeOutputStream s) {
        out = s;
        err = new acmeLib.core.io.AcmeOutputStream();
        idHash = new Hashtable();
        tidHash = new Hashtable();
        rtObjHash = new Hashtable();
        rtViewHash = new Hashtable();
        tallyHash = new Hashtable();
        petalDesign = new RoseRTDesign("Model");
        // get the root category name
        rootCatName = petalDesign.classCat.getName();
        classList = new RoseRTList();
        instList = new RoseRTList();
        prtclList = new RoseRTList();
    }

    public void writeToFile(String path) {
        try {
            File f = new File(path);
            FileOutputStream fo = new FileOutputStream(f);
            out.writeTo(fo);
            fo.close();
	    File ferr = new File(path + ".err");
	    FileOutputStream foerr = new FileOutputStream(ferr);
            err.writeTo(foerr);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    public void exportDesign(AcmeDesign designACME)
        throws java.io.IOException
    {
        exportTypespace(designACME.getTypes());
	exportSystems(designACME.getSystems());
	// debug
	debugLists();
	exportRTLogicalUnits();
	RoseRTLayout.layoutGraph();  // layout class diagram

        // Output the entire design
        out.writeln("\n(object PetalRT");
        out.indent(); out.indent();  // 2 of these to get 8 spaces
        out.writeln("version 60010");
        out.writeln("_written \"Rose RealTime 6.2.190.0\"");
/* Rose RT didn't seem to like these
        out.writeln("_generator \"Automatically generated by Acme's " +
                    "RoseRTExporter\"");
        GregorianCalendar cal = new GregorianCalendar(TimeZone.getDefault());
        DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT);
        out.writeln("_generated \"" + df.format(cal.getTime()) + "\"");
*/
        out.writeln("charSet 0)");
        out.writeln("");
        out.unIndent(); out.unIndent();

        out.writeln(petalDesign.toString());

        out.close();
        err.writeln("\n");
        err.close();
    }

    private void exportTypespace(AcmeEnumeration typespace)
        throws java.io.IOException
    {
        typespace.sortByCreationOrder();
        while (typespace.moreItems()) {
            AcmeType typeACME = (AcmeType)typespace.getNextItem();
            if (typeACME instanceof AcmePropertyType) {
                exportPropertyTypeDecl((AcmePropertyType)typeACME);
            }
        }

        typespace.reset();
        while (typespace.moreItems()) {
            AcmeType typeACME = (AcmeType)typespace.getNextItem();
            if (typeACME instanceof AcmeComponentType) {
                exportComponentType((AcmeComponentType)typeACME);
            }
        }

        typespace.reset();
        while (typespace.moreItems()) {
            AcmeType typeACME = (AcmeType)typespace.getNextItem();
            if (typeACME instanceof AcmeConnectorType) {
                exportConnectorType((AcmeConnectorType)typeACME);
            }
        }

        typespace.reset();
        while (typespace.moreItems()) {
            AcmeType typeACME = (AcmeType)typespace.getNextItem();
            if (typeACME instanceof AcmePortType) {
                exportPortType((AcmePortType)typeACME);
            } else if (typeACME instanceof AcmeRoleType) {
                exportRoleType((AcmeRoleType)typeACME);
            }
        }

        /* I don't think a Family will ever fall into a type space?
           typespace.reset();
           while (typespace.moreItems()) {
           AcmeType typeACME = (AcmeType)typespace.getNextItem();
           if (typeACME instanceof AcmeFamily) {
           exportFamily((AcmeFamily)typeACME);
           }
           }
        */
    }

    private void exportPropertyTypeDecl(AcmePropertyType propertyTypeACME)
        throws java.io.IOException
    {
        err.writeln("exportPropertyTypeDecl: name=" +
                    propertyTypeACME.getTypeName());
        exportPropertyType(propertyTypeACME);
    }

    private void exportPropertyType(AcmePropertyType propertyTypeACME)
        throws java.io.IOException
    {
        // debug output
        err.writeln("exportPropertyType:");
        err.indent();

        if (propertyTypeACME instanceof acmeLib.properties.RecordType) {
            err.writeln("<propertyRecordType>");
            err.indent();
            acmeLib.properties.RecordType recordACME =
                (acmeLib.properties.RecordType)propertyTypeACME;
            AcmeEnumeration fieldsACME = recordACME.getFields();
            while (fieldsACME.moreItems()) {
                AcmeRecordField fieldACME =
                    (AcmeRecordField)fieldsACME.getNextItem();
                err.write("<propertyFieldDeclaration name='");
                err.write(fieldACME.getName ());
                err.writeln("'>");
                err.indent();
                AcmePropertyType fieldTypeACME = fieldACME.getType();
                exportPropertyType(fieldTypeACME);
                err.unIndent();
                err.writeln("</propertyFieldDeclaration>");
            }
            err.unIndent();
            err.writeln("</propertyRecordType>");
        } else if (propertyTypeACME instanceof acmeLib.properties.SetType) {
            err.writeln("<propertySetType>");
            err.indent();
            acmeLib.properties.SetType setTypeACME =
                (acmeLib.properties.SetType)propertyTypeACME;
            AcmePropertyType innerTypeACME =
                (AcmePropertyType)setTypeACME.getSetType();
            exportPropertyType(innerTypeACME);
            err.unIndent();
            err.writeln("</propertySetType>");
        } else if (propertyTypeACME instanceof
                   acmeLib.properties.SequenceType) {
            err.writeln("<propertySequenceType>");
            err.indent();
            acmeLib.properties.SequenceType seqTypeACME =
                (acmeLib.properties.SequenceType)propertyTypeACME;
            AcmePropertyType innerTypeACME = seqTypeACME.getSequenceType();
            exportPropertyType(innerTypeACME);
            err.unIndent();
            err.writeln("</propertySequenceType>");
        } else if (propertyTypeACME instanceof EnumType) {
            err.writeln("<propertyEnumType>");
            err.indent();
            EnumType enumTypeACME =(EnumType)propertyTypeACME;
            AcmeEnumeration valuesACME = enumTypeACME.getValues();
            while(valuesACME.moreItems()) {
                EnumValue valueACME =(EnumValue)valuesACME.getNextItem();
                err.write("<value>");
                err.write(valueACME.value);
                err.writeln("</value>");
            }
            err.unIndent();
            err.writeln("</propertyEnumType>");
        } else if (propertyTypeACME instanceof StringType) {
            err.writeln("<propertyPrimitiveType type='string'/>");
        } else if (propertyTypeACME instanceof IntType) {
            err.writeln("<propertyPrimitiveType type='int'/>");
        } else if (propertyTypeACME instanceof FloatType) {
            err.writeln("<propertyPrimitiveType type='float'/>");
        } else if (propertyTypeACME instanceof BooleanType) {
            err.writeln("<propertyPrimitiveType type='boolean'/>");
        }

        err.unIndent();
    }

    private void exportSuperTypes(AcmeElementType elementTypeACME)
        throws java.io.IOException
    {
        /*
          int nTypes = elementTypeACME.getSuperTypes().getSize();
          if (nTypes != 0) {
          AcmeEnumeration supersACME =
          elementTypeACME.getSuperTypes().enumerate(new acmeLib.core.filters.NoBaseClassesEnumFilter());
          while (supersACME.moreItems()) {
          AcmeElementType superACME =
          (AcmeElementType)supersACME.getNextItem();
          err.writeln("<" + tag + " xlink:type='simple' xlink:href='#" +
          tidHash.get(superACME) + "'/>");
          }
          }
        */
    }

    // not sure what to do about the tag or tidHash yet
    private void exportTypes(AcmeElement elementACME)
        throws java.io.IOException
    {
        int nTypes = elementACME.getDeclaredTypes().getSize();
        err.writeln("exportTypes:");
        err.indent();
        if (nTypes != 0) {
            AcmeEnumeration supersACME =
                elementACME.getDeclaredTypes();
            while (supersACME.moreItems()) {
                AcmeElementType superACME =
                    (AcmeElementType)supersACME.getNextItem();
                // check whether it's a component type to determine whether
                // to generate a new capsule class
                if (superACME instanceof AcmeComponentType) {
                    // make sure we don't already have a class of same name
                    if (classList.findName(superACME.getTypeName()) < 0) {
                        // not found, add a new class
                        RoseRTCapsule rtCap = new
                            RoseRTCapsule(superACME.getTypeName());
                        classList.addLast(rtCap);
                    }
                }
                // debug
                err.writeln(tidHash.get(superACME) + " name=" +
                            superACME.getTypeName());
            }
        }
        err.unIndent();
    }

    // not sure what to do with these specific element type methods yet
    private void exportComponentType(AcmeComponentType componentTypeACME)
        throws java.io.IOException
    {
        err.writeln("exportComponentType: name=" +
                    componentTypeACME.getTypeName());
        err.indent();
        exportSuperTypes(componentTypeACME);
        exportComponent(componentTypeACME, null, null);
        err.unIndent();
    }

    private void exportConnectorType(AcmeConnectorType connectorTypeACME)
        throws java.io.IOException
    {
        err.writeln("exportConnectorType name=" +
                    connectorTypeACME.getTypeName());
        err.indent();
        exportSuperTypes(connectorTypeACME);
        exportConnector(connectorTypeACME, null, null, null);
        err.unIndent();
    }

    private void exportPortType(AcmePortType portTypeACME)
        throws java.io.IOException
    {
        err.writeln("exportPortType: name=" +
                    portTypeACME.getTypeName());
        err.indent();
        exportSuperTypes(portTypeACME);
        exportPort(portTypeACME);
        err.unIndent();
    }

    private void exportRoleType(AcmeRoleType roleTypeACME)
        throws java.io.IOException
    {
        err.writeln("exportRoleType: name=" +
                    roleTypeACME.getTypeName());
        err.indent();
        exportSuperTypes(roleTypeACME);
        exportRole(roleTypeACME);
        err.unIndent();
    }

    private void exportFamily(AcmeFamily familyACME)
        throws java.io.IOException
    {
        err.writeln("exportFamily: name=" + familyACME.getName());
        err.indent();
        exportSystem(familyACME, "", false);
        exportSuperTypes(familyACME);
        exportTypespace(familyACME.getTypes());
        err.unIndent();
    }

    private void exportSystems(AcmeEnumeration systemsACME)
	throws java.io.IOException
    {
        // Translate Acme design elements to Rose RT elements, and generate
        //   the collaboration structure(s).
        // We assume one system only
        if (systemsACME.moreItems()) {
            AcmeSystem systemACME = (AcmeSystem)systemsACME.getNextItem();
	    // generate new system name by appending "System" to current name
	    //   if it isn't already there
            String systemName = new String(systemACME.getName());
	    if (systemName.indexOf(SYSTEM_NAME) < 0) {  // append "System"
		systemName += SYSTEM_NAME;
		// modify the ACME system's name also
		try {
		    systemACME.setName(systemName);
		} catch (AcmeIllegalNameException aine) {
		    err.writeln("Name change for " + systemACME.getName() +
				" failed!");
		}
	    }
            // Get all of components, connectors, and attachments of the
            //   system, including those in the representations
            AcmeEnumeration allComponentsACME = new AcmeEnumeration();
            allComponentsACME.concatenate(systemACME.getComponents());
            AcmeEnumeration allConnectorsACME = new AcmeEnumeration();
            allConnectorsACME.concatenate(systemACME.getConnectors());
            AcmeEnumeration allAttachmentsACME = new AcmeEnumeration();
            allAttachmentsACME.concatenate(systemACME.getAttachments());
            getAllItems(systemACME.getComponents(), allComponentsACME,
                        allConnectorsACME, allAttachmentsACME);
            // Retrieve list of protocols (connectors)
            deriveProtocols(allConnectorsACME, allAttachmentsACME);
            // Derive class diagram and inheritance relationship
            deriveClassDiag(allComponentsACME);
            // Add a system capsule class
            RoseRTCapsule rtCap = new RoseRTCapsule(systemName);
            rtCap.setCollab(new RoseRTCapsuleCollab(COLLAB_NAME));
            rtCap.setHasRep(true);  // naturally, as it has internal structure
            // Add system capsule class to list
            classList.addFirst(rtCap);
            // Export the system
            exportSystem(systemACME, systemName, true);
        }
    }

    private void exportSystem(AcmeSystem systemACME, String sysCapsuleName,
                              boolean doSupers)
        throws java.io.IOException
    {
        // debug output
        err.writeln("exportSystem: name=" + systemACME.getName());
        err.indent();

        // Find the capsule class for this system (or rep)
        int idx = classList.findName(sysCapsuleName);
        if (idx < 0) {  // shouldn't happen!
            err.writeln("Fatal error: exportSystem(): could not locate capsule class");
            return;
        }
        RoseRTCapsule rtCap = (RoseRTCapsule)classList.get(idx);
        // Create new collaboration diagram
        rtCap.setCollabDiag(new RoseRTCollabDiagram(COLLAB_DIAGRAM_NAME));

        // Create a capsule perimeter view object
        RoseRTCapsulePerimeterVw rtCapPerimVw =
            new RoseRTCapsulePerimeterVw(COLLAB_NAME);
        rtCapPerimVw.setID(generateID(null, rtCapPerimVw));
	rtCapPerimVw.setPos(RoseRTLayout.getCenter());
	rtCapPerimVw.setWidth(RoseRTLayout.getWidth());
	rtCapPerimVw.setHeight(RoseRTLayout.getHeight());
        // Iterate through the ports of this capsule and add port view object
        //   for each port
        RoseRTList rtPorts = getAllRTPorts(rtCap);
        RoseRTPort rtPort = null;
        RoseRTPortVw rtPortVw = null;
        RoseRTItemLabel rtLabel = null;
        for (Enumeration iter = rtPorts.elements(); iter.hasMoreElements();) {
            rtPort = (RoseRTPort)iter.nextElement();
            // create a port view object
            rtPortVw = new RoseRTPortVw(STRUCT_DELIM + COLLAB_NAME +
                                        STRUCT_DELIM + rtPort.getName());
            // get new object ID
            rtPortVw.setID(generateID(null, rtPortVw));
            // create a label
            rtLabel = new RoseRTItemLabel();
            rtLabel.setItemLabel(rtPort.getName());
            rtPortVw.setObjLabel(rtLabel);
            // add to capsule perimeter view's list of icons
            rtCapPerimVw.getIcons().addLast(rtPortVw);
            instList.addLast(rtPortVw);
        }
        // Add capsule perimeter view object to capsule's collaboration diagram
        rtCap.getCollabDiag().getItems().addLast(rtCapPerimVw);

        // Export all the components
        RoseRTCapsuleRole rtCapRole = null;
        RoseRTCapsuleRoleVw rtCapRoleVw = null;
        AcmeComponent componentACME = null;
        AcmeEnumeration componentsACME = systemACME.getComponents();
        while (componentsACME.moreItems()) {
            componentACME = (AcmeComponent)componentsACME.getNextItem();
            // create RT capsule role instances
            rtCapRole = new RoseRTCapsuleRole(componentACME.getName());
            rtCapRoleVw = new RoseRTCapsuleRoleVw(STRUCT_DELIM + COLLAB_NAME +
                                                  STRUCT_DELIM +
                                                  componentACME.getName());
            exportComponent(componentACME, rtCapRole, rtCapRoleVw);
            // add the capsule role instances to capsule class object
            if (rtCap.hasParent()) {  // has parent, use refinement
                ((RoseRTRefinement)rtCap.getRefinements().get(0))
                    .getClassifierRoles().addLast(rtCapRole);
            } else {
                rtCap.getCollab().getClassifierRoles().addLast(rtCapRole);
            }
            rtCap.getCollabDiag().getItems().addLast(rtCapRoleVw);
        }

        // Export all the connectors
        RoseRTConnector rtConn = null;
        RoseRTConnectorVw rtConnVw = null;
        AcmeConnector connectorACME = null;
        AcmeEnumeration connectorsACME = systemACME.getConnectors();
        AcmeEnumeration attachmentsACME = systemACME.getAttachments();
        while (connectorsACME.moreItems()) {
            connectorACME = (AcmeConnector)connectorsACME.getNextItem();
            // create RT connector instances
            rtConn = new RoseRTConnector(connectorACME.getName());
            rtConnVw = new RoseRTConnectorVw(STRUCT_DELIM + COLLAB_NAME +
                                             STRUCT_DELIM +
                                             connectorACME.getName());
            exportConnector(connectorACME, attachmentsACME, rtConn, rtConnVw);
            // add the connector instances to capsule class object
            if (rtCap.hasParent()) {  // has parent, use refinement
                ((RoseRTRefinement)rtCap.getRefinements().get(0))
                    .getConnectors().addLast(rtConn);
            } else {
                rtCap.getCollab().getConnectors().addLast(rtConn);
            }
            rtCap.getCollabDiag().getItems().addLast(rtConnVw);
        }

        // Export all the attachments
        int numAttachments = 0;
        while (attachmentsACME.moreItems()) {
            exportAttachment((AcmeAttachment)attachmentsACME.getNextItem(),
                             numAttachments++);
        }

        if (doSupers) {
            this.exportTypes(systemACME);
        }

        err.unIndent();
    }

    private void exportComponent(AcmeComponent componentACME,
                                 RoseRTCapsuleRole rtCapRole,
                                 RoseRTCapsuleRoleVw rtCapRoleVw)
        throws java.io.IOException
    {
        // debug output
        err.writeln("exportComponent: name=" + componentACME.getName());
        err.indent();

        if (rtCapRole != null) {
            // populate RT capsule role instance
            rtCapRole.setClassifier(rootCatName + PATH_DELIM +
                                    getMatchingCapsuleClass(componentACME));
            instList.addLast(rtCapRole);
        }
        if (rtCapRoleVw != null) {
            // populate RT capsule view role instance
            rtCapRoleVw.setID(generateID(componentACME, rtCapRoleVw));
            RoseRTItemLabel rtLabel = new RoseRTItemLabel();
            rtLabel.setItemLabel(componentACME.getName());
            rtCapRoleVw.setObjLabel(rtLabel);
            rtCapRoleVw.setPos(0, 0);
            // add to instance list
            instList.addLast(rtCapRoleVw);
        }

        AcmeEnumeration propertiesACME = componentACME.getProperties();
	err.writeln(">>> Checking visual properties:");
	err.indent();
        while (propertiesACME.moreItems()) {
            AcmeProperty propertyACME =
                (AcmeProperty)propertiesACME.getNextItem();
            // We don't export properties to component instance at this time
            //   mainly because there's no appropriate place to put them
            //   except maybe as an annotation
            //exportProperty(propertyACME);

            // extract the visual properties
            String name = propertyACME.getName();
            if (rtCapRoleVw != null &&
                name.length() >= 4 &&
		name.substring(0,4).equalsIgnoreCase("vis-")){
		err.writeln(">> Visual property '" + name + "=" +
			    propertyACME.getValue());
                if (name.equalsIgnoreCase("vis-x")) {
                    // store x coordinate
                    AcmeFloat xVal = (AcmeFloat)propertyACME.getValue();
                    int x = RoseRTLayout.transformX(xVal.getValue());
                    rtCapRoleVw.getPos().setX(x);
                } else if (name.equalsIgnoreCase("vis-y")) {
                    // store y coordinate
                    AcmeFloat yVal = (AcmeFloat)propertyACME.getValue();
                    int y = RoseRTLayout.transformY(yVal.getValue());
                    rtCapRoleVw.getPos().setY(y);
                } else if (name.equalsIgnoreCase("vis-width")) {
                    // store width
                    AcmeFloat widthVal = (AcmeFloat)propertyACME.getValue();
		    int width =
			RoseRTLayout.transformLength(widthVal.getValue());
                    rtCapRoleVw.setWidth(width);
                } else if (name.equalsIgnoreCase("vis-height")) {
                    // store height
                    AcmeFloat heightVal = (AcmeFloat)propertyACME.getValue();
		    int height =
			RoseRTLayout.transformLength(heightVal.getValue());
                    rtCapRoleVw.setHeight(height);
                }
            }
        }
	err.unIndent();

        if (rtCapRoleVw != null) {
            // Set object label coordinate to some offset from object
            rtCapRoleVw.getObjLabel()
		.setPos(RoseRTLayout.getCapLabelPos(rtCapRoleVw.getPos()));
	    // dimension object to help calculate port label coordinate
	    RoseRTPoint dim = new RoseRTPoint(rtCapRoleVw.getWidth(),
					      rtCapRoleVw.getHeight());

            // Export component's ports
            AcmeEnumeration portsACME = componentACME.getPorts();
            while (portsACME.moreItems()) {
                AcmePort portACME = (AcmePort)portsACME.getNextItem();
                RoseRTConnectionPtVw rtConnPt = exportPort(portACME);
                // adjust x and y coordinates using capsule's coordinate
		rtConnPt.setOrigin(rtCapRoleVw.getPos());  // set capsule coord
		if (Math.abs((Math.abs(rtConnPt.getPos().getY()) -
			      rtCapRoleVw.getHeight()/2)) <= 1) {
		    // consider port to be on upper or lower border
		    rtConnPt.setUpLow(true);
		}
                rtConnPt.setPos(rtConnPt.getPos().getX() +
				rtCapRoleVw.getPos().getX(),
				rtConnPt.getPos().getY() +
				rtCapRoleVw.getPos().getY());
		// set object label coordinate to be some offset from port
		rtConnPt.getObjLabel().setPos
		    (RoseRTLayout.getLabelPos
		     (rtConnPt.getPos(), rtCapRoleVw.getPos(), dim));

                // add component name to port name label
                rtConnPt.setName(rtCapRoleVw.getName() + STRUCT_DELIM +
                                 rtConnPt.getName());
                // add connection point to capsule role
                rtCapRoleVw.getIcons().addLast(rtConnPt);
            }
        }

        // We assume one represenation
        AcmeEnumeration representationsACME =
            componentACME.getRepresentations();
        if (representationsACME.moreItems()) {
            AcmeRepresentation representationACME =
                (AcmeRepresentation)representationsACME.getNextItem();
	    String repCapName = getMatchingCapsuleClass(componentACME);
	    //debug
	    err.writeln("> repCapName = " + repCapName);

	    // adjust scale
	    RoseRTLayout.oldCenterX = RoseRTLayout.OLD_RCENTER_X;
	    RoseRTLayout.oldCenterY = RoseRTLayout.OLD_RCENTER_Y;

            exportRepresentation(representationACME, repCapName);
	    // reset scale
	    RoseRTLayout.oldCenterX = RoseRTLayout.OLD_CENTER_X;
	    RoseRTLayout.oldCenterY = RoseRTLayout.OLD_CENTER_Y;
        }

        err.unIndent();
    }

    private void exportConnector(AcmeConnector connectorACME,
                                 AcmeEnumeration attachmentsACME,
                                 RoseRTConnector rtConn,
                                 RoseRTConnectorVw rtConnVw)
        throws java.io.IOException
    {
        // debug output
        err.writeln("exportConnector: name=" + connectorACME.getName());
        err.indent();

        if (rtConn != null) {
            // populate RT connector instance
            rtConn.setCardinality(CARDINALITY);
            instList.addLast(rtConn);
        }
        if (rtConnVw != null) {
            // populate RT connector view instance
            rtConnVw.setID(generateID(connectorACME, rtConnVw));
            RoseRTSegLabel rtLabel = new RoseRTSegLabel();
            rtLabel.setID(generateID(null, rtLabel));
            rtConnVw.setObjLabel(rtLabel);
            rtConnVw.setPos(0, 0);
            // add to instance list
            instList.addLast(rtConnVw);
        }

        AcmeEnumeration propertiesACME = connectorACME.getProperties();
        while (propertiesACME.moreItems()) {
            AcmeProperty propertyACME =
                (AcmeProperty)propertiesACME.getNextItem();
            // We don't export properties to component instance at this time
            //   mainly because there's no appropriate place to put them
            //   except maybe as an annotation
            //exportProperty(propertyACME);

            // extract the visual properties
            String name = propertyACME.getName();
            if (rtConnVw != null &&
                name.length() >= 4 &&
		name.substring(0,4).equalsIgnoreCase("vis-")){
                if (name.equalsIgnoreCase("vis-x")) {
                    // store x coordinate
                    AcmeFloat xVal = (AcmeFloat)propertyACME.getValue();
                    int x = RoseRTLayout.transformX(xVal.getValue());
                    rtConnVw.getPos().setX(x);
                } else if (name.equalsIgnoreCase("vis-y")) {
                    // store y coordinate
                    AcmeFloat yVal = (AcmeFloat)propertyACME.getValue();
                    int y = RoseRTLayout.transformY(yVal.getValue());
                    rtConnVw.getPos().setY(y);
                }
            }
        }

        if (rtConnVw != null) {
            // Set object label coordinate to be (+X_OFFSET,+Y_OFFSET) from
            //   object
            rtConnVw.getObjLabel()
		.setPos(RoseRTLayout.getLabelPos(rtConnVw.getPos()));

            // Export connector's roles
            // We assume binary connector, so two roles
            // This is consistent with RoseRT's assumption
            AcmeEnumeration rolesACME = connectorACME.getRoles();
            while (rolesACME.moreItems()) {
                AcmeRole roleACME = (AcmeRole)rolesACME.getNextItem();
                // Role is not exported because it doesn't have explicit
                // target mapping in UML-RT
                //exportRole(roleACME);

                // find the port attached to the role
                AcmePort portACME = findAttachedPort(roleACME,
                                                     attachmentsACME);

                // debug
                err.writeln("Role: " + roleACME.getName() +
                            " Attached port: " +
                            ((portACME == null) ? "" : portACME.getName()));

                if (portACME == null) {
                    continue;
                }
                String portTypeName = getElementType(portACME).getTypeName();
                int id = ((Integer)idHash.get(portACME)).intValue();
                if (checkIfBasePort(portTypeName)) {
                    // associate supplier to this ID
                    rtConnVw.setSupplier(id);
                } else {  // associate client to this ID
                    if (rtConnVw.getClient() > 0) {
                        // Client already set, this means both are ports are
                        //   not base.  Use the first port as supplier, which
                        //   is the current client.
                        rtConnVw.setSupplier(rtConnVw.getClient());
                        // after this, the current port becomes the client
                    }
                    rtConnVw.setClient(id);
                }
            }

            // Get the actual supplier & client ends for the RT connector obj
	    generateRTConnEnds(rtConn, rtConnVw);
        }

        /* No translation for connector representation yet
         *
        AcmeEnumeration representationsACME =
            connectorACME.getRepresentations();
        while (representationsACME.moreItems()) {
            AcmeRepresentation representationACME =
                (AcmeRepresentation)representationsACME.getNextItem();
            exportRepresentation(representationACME);
        }
        */

        err.unIndent();
    }

    private RoseRTConnectionPtVw exportPort(AcmePort portACME)
        throws java.io.IOException
    {
        // debug output
        err.writeln("exportPort: name=" + portACME.getName());
        err.indent();

        // create a RT port (connection point) instance
        RoseRTConnectionPtVw rtConnPt =
            new RoseRTConnectionPtVw(portACME.getName());
        rtConnPt.setID(generateID(portACME, rtConnPt));
        RoseRTItemLabel rtLabel = new RoseRTItemLabel();
        rtLabel.setItemLabel(portACME.getName());
        rtConnPt.setObjLabel(rtLabel);
        rtConnPt.setPos(0, 0);
        // add to instance list
        instList.addLast(rtConnPt);

        AcmeEnumeration propertiesACME = portACME.getProperties();
        while (propertiesACME.moreItems()) {
            AcmeProperty propertyACME =
                (AcmeProperty)propertiesACME.getNextItem();
            // We don't export port properties to port instance at this time
            //   mainly because there's no appropriate place to put them
            //   except maybe as an annotation
            //exportProperty(propertyACME);

            // extract the visual properties
            String name = propertyACME.getName();
            if (name.length() >= 4 &&
		name.substring(0,4).equalsIgnoreCase("vis-")){
                if (name.equalsIgnoreCase("vis-x")) {
                    // store x coordinate relative to component (capsule)
                    AcmeFloat xVal = (AcmeFloat)propertyACME.getValue();
                    int x = RoseRTLayout.transformLength(xVal.getValue());
                    rtConnPt.getPos().setX(x);
                } else if (name.equalsIgnoreCase("vis-y")) {
                    // store y coordinate relative to component (capsule)
                    AcmeFloat yVal = (AcmeFloat)propertyACME.getValue();
                    int y = RoseRTLayout.transformLength(yVal.getValue());
                    rtConnPt.getPos().setY(y);
                }
            }
        }

        err.unIndent();
        return rtConnPt;
    }

    private void exportRole(AcmeRole roleACME)
        throws java.io.IOException
    {
        // debug output
        err.writeln("exportRole: name=" + roleACME.getName());
        err.indent();

        AcmeEnumeration propertiesACME = roleACME.getProperties();
        while (propertiesACME.moreItems()) {
            AcmeProperty propertyACME =
                (AcmeProperty)propertiesACME.getNextItem();
            exportProperty(propertyACME);
        }

        err.unIndent();
    }

    private void exportProperty(AcmeProperty propertyACME)
        throws java.io.IOException
    {
        if (includeProperty(propertyACME)) {
            err.writeln("exportProperty name=" + propertyACME.getName());
            err.indent();
            AcmePropertyValue valueACME = propertyACME.getValue();
            if (valueACME != null) {
                exportPropertyValue(valueACME);
            }
            exportPropertyType(propertyACME.getType());
            err.unIndent();
        }
    }

    private void exportPropertyValue(AcmePropertyValue valueACME)
        throws java.io.IOException
    {
        err.writeln("exportPropertyValue:");
        err.indent();
        if (valueACME instanceof AcmeString) {
            AcmeString stringValACME = (AcmeString)valueACME;
            err.writeln("AcmeString:");
            err.indent();
            err.writeln(stringValACME.getValue());
            err.unIndent();
        } else if (valueACME instanceof AcmeInt) {
            AcmeInt intValACME = (AcmeInt)valueACME;
            err.writeln("AcmeInt:");
            err.indent();
            err.writeln("" + intValACME.getValue());
            err.unIndent();

        } else if (valueACME instanceof AcmeFloat) {
            AcmeFloat floatValACME = (AcmeFloat)valueACME;
            err.writeln("AcmeFloat:");
            err.indent();
            err.writeln ("" + floatValACME.getValue ());
            err.unIndent();

        } else if (valueACME instanceof AcmeBoolean) {
            AcmeBoolean booleanValACME = (AcmeBoolean)valueACME;
            err.writeln("AcmeBoolean:");
            err.indent();
            err.writeln("" + booleanValACME.getValue());
            err.unIndent();

        } else if (valueACME instanceof AcmeSet) {
            AcmeSet setValACME = (AcmeSet)valueACME;
            err.writeln("AcmeSet:");
            err.indent();
            AcmeEnumeration valuesACME = setValACME.getValues();
            while (valuesACME.moreItems()) {
                AcmePropertyValue vACME =
                    (AcmePropertyValue)valuesACME.getNextItem();
                exportPropertyValue(vACME);
            }
            err.unIndent();
        } else if (valueACME instanceof AcmeSequence) {
            AcmeSequence sequenceValACME = (AcmeSequence)valueACME;
            err.writeln("AcmeSequence:");
            err.indent();
            AcmeEnumeration valuesACME = sequenceValACME.getValues();
            while (valuesACME.moreItems()) {
                AcmePropertyValue vACME =
                    (AcmePropertyValue)valuesACME.getNextItem();
                exportPropertyValue(vACME);
            }
            err.unIndent();

        } else if (valueACME instanceof AcmeEnum) {
        } else if (valueACME instanceof AcmeRecord) {
            AcmeRecord recACME = (AcmeRecord)valueACME;
            err.writeln("AcmeRecord:");
            err.indent();
            AcmeEnumeration fieldsACME = recACME.getFields();
            while (fieldsACME.moreItems()) {
                AcmeRecordField fieldACME =
                    (AcmeRecordField)fieldsACME.getNextItem();
                err.writeln("fieldName=" + fieldACME.getName());
                err.indent();
                exportPropertyValue(fieldACME.getValue());
                err.unIndent();
            }
            err.unIndent();
        }
        err.unIndent();
    }

    private void exportAttachment(AcmeAttachment attACME, int attachNum)
        throws java.io.IOException
    {
        err.writeln("exportAttachmentInstance:");
        err.indent();
        err.writeln("port=" + idHash.get(attACME.getPort()));
        err.writeln("role=" + idHash.get(attACME.getRole()));
        err.unIndent();
    }

    private void exportRepresentation(AcmeRepresentation repACME,
                                      String repCapsuleName)
        throws java.io.IOException
    {
        // debug output
        err.writeln("exportRepresentation:");
        err.indent();
        exportSystem(repACME.getSystem(), repCapsuleName, true);

        // Find the capsule class for this system (or rep)
        int idx = classList.findName(repCapsuleName);
        if (idx < 0) {  // shouldn't happen!
            err.writeln("Fatal error: exportRepresentation(): could not locate capsule class");
            return;
        }
        RoseRTCapsule rtCap = (RoseRTCapsule)classList.get(idx);
        // Find the capsule perimeter view object within capsule, used later
        RoseRTCapsulePerimeterVw rtCapPerimVw =
            new RoseRTCapsulePerimeterVw();  // dummy object to get name
        idx = rtCap.getCollabDiag().getItems()
            .findClass(rtCapPerimVw.getClass());
        if (idx < 0) {  // not found!
            // shouldn't happen if this capsule has a representation
            err.writeln("Fatal error: exportRepresentation(): could not locate capsule perimeter view object within capsule class");
            return;
        }
        rtCapPerimVw = (RoseRTCapsulePerimeterVw)rtCap.getCollabDiag()
            .getItems().get(idx);

        // Iterate through all bindings and create a connector for each
        AcmeEnumeration bindingsACME =
            ((AcmeBindingsMap)repACME.getAbstractionMap()).getBindings();
        int i = 1;
        while (bindingsACME.moreItems()) {
            AcmeBinding bindingACME = (AcmeBinding)bindingsACME.getNextItem();
            AcmeElement innerACME = bindingACME.getInner();
            AcmeElement outerACME = bindingACME.getOuter();
            Integer innerID = (Integer)idHash.get(innerACME);
            Integer outerID = (Integer)idHash.get(outerACME);

            // debug output
            err.writeln("exportBindingInstance:");
            err.indent();
            err.writeln("outer: " + outerID);
            err.writeln("inner: " + innerID);
            err.unIndent();

            // create new connector and connector view object
	    String rtConnName = BINDING_NAME + i++;
	    RoseRTConnector rtConn = new RoseRTConnector(rtConnName);
            rtConn.setCardinality(CARDINALITY);
            RoseRTConnectorVw rtConnVw =
                new RoseRTConnectorVw(STRUCT_DELIM + COLLAB_NAME +
                                      STRUCT_DELIM + rtConnName);
            rtConnVw.setID(generateID(null, rtConnVw));
            RoseRTSegLabel rtLabel = new RoseRTSegLabel();
            // add segment label to it
            rtLabel.setID(generateID(null, rtLabel));
            rtConnVw.setObjLabel(rtLabel);
            // set the client port ID
            rtConnVw.setClient(innerID.intValue());
            // lookup the supplier port ID
            //   - the supplier is a port view object of the enclosing
            //     capsule, which is different from the port (connection point)
            //     object as viewed from one level above the enclosing capsule
            //   - the outer ID retrieved from the hash provides the port
            //     object viewed from one level above, so the lookup basically
            //     seeks the port view object by name within the enclosing
            //     capsule's port view object list.
            RoseRTConnectionPtVw rtConnPt =
		(RoseRTConnectionPtVw)rtObjHash.get(outerID);
            // here we play with the port name to get the port view object name
            //   e.g.     port name:  :Structure:SomeFilter:portX
            //       port view name:  :Structure:portX  under SomeFilter
            String rtPortName = rtConnPt.getName();
            int beg = rtPortName.indexOf(rtCap.getName());
            int end = beg + rtCap.getName().length() + STRUCT_DELIM.length();
            String rtPortVwName = rtPortName.substring(0, beg) +
                rtPortName.substring(end, rtPortName.length());
            // search for this port view
            idx = rtCapPerimVw.getIcons().findName(rtPortVwName);
            if (idx < 0) {  // not found, report error and move on
                err.writeln("Error: exportRepresentation(): could not locate port view object '" + rtPortVwName + "' within the capsule perimeter view object");
                continue;
            }
            // otherwise, found; get its ID
            RoseRTPortVw rtPortVw =
                (RoseRTPortVw)rtCapPerimVw.getIcons().get(idx);
            // set the supplier port ID
            rtConnVw.setSupplier(rtPortVw.getID());
            // Get the actual supplier & client ends for the RT connector obj
	    generateRTConnEnds(rtConn, rtConnVw);
	    // adjust the port view object's coordinates
	    RoseRTPoint dim = new RoseRTPoint
		(rtCapPerimVw.getWidth(), rtCapPerimVw.getHeight());
	    RoseRTPoint pos = RoseRTLayout.getExtrapPos
		(rtConnPt.getPos(), rtConnPt.getOrigin(), rtConnPt.isUpLow(),
		 rtCapPerimVw.getPos(), dim);
	    rtPortVw.setPos(pos);
	    rtPortVw.getObjLabel().setPos
		(RoseRTLayout.getNearestLabelPos(rtPortVw.getPos(),
						 rtCapPerimVw.getPos()));
            // add the connector instances to capsule class object
            if (rtCap.hasParent()) {  // has parent, use refinement
                ((RoseRTRefinement)rtCap.getRefinements().get(0))
                    .getConnectors().addLast(rtConn);
            } else {
                rtCap.getCollab().getConnectors().addLast(rtConn);
            }
            rtCap.getCollabDiag().getItems().addLast(rtConnVw);
            instList.addLast(rtConn);
            instList.addLast(rtConnVw);
        }
        err.unIndent();
    }

    private void exportRTLogicalUnits()
	throws java.io.IOException
    {
        // Compose the logical models by adding the capsule objects, and
        //   generate the main class diagram structure in logical presentations
        // Inheritance relationships are generated as capsules with parents
        //   are encountered, the inheritance then stored and added together
        //   at the end
        // Same with the capsule role relationships
        RoseRTCapsule rtCap = null;
        RoseRTCapsuleVw rtCapVw = null;
        RoseRTCapsuleVw rtCapVw2 = null;  // for supplier
        RoseRTInheritVw rtInhVw = null;
        RoseRTItemLabel rtLabel = null;
        RoseRTItemLabel rtStereo = null;
        RoseRTCompartment rtCompart = null;
        RoseRTClassDiagram rtClassDiag =
            new RoseRTClassDiagram(CLASS_DIAGRAM_NAME);
        // used to find capsule view and recover IDs
        RoseRTList rtCapVws = new RoseRTList();
        RoseRTList rtInhVws = new RoseRTList();
        String rtPCapName = null;
        Enumeration iter = null;
        for (iter = classList.elements(); iter.hasMoreElements();) {
            rtCap = (RoseRTCapsule)iter.nextElement();

            // create a capsule view object for capsule class
            rtCapVw = new RoseRTCapsuleVw(rtCap.getName());
            rtCapVw.setID(generateID(null, rtCapVw));
	    rtCapVw.setRect(RoseRTLayout.addRectangle());
	    // name the node rectangle
	    rtCapVw.getRect().setName(CAPVW_PREFIX + rtCapVw.getName());
            rtLabel = new RoseRTItemLabel();
            rtLabel.setItemLabel(rtCap.getName());
	    rtLabel.setMaxWidth
		(RoseRTLayout.getPixelWidth(rtLabel.getItemLabel()));
	    rtLabel.setPos
		(rtCapVw.getRect().getCenter().getX() -
		 rtLabel.getMaxWidth() / 2,
		 rtCapVw.getRect().getCenter().getY() - RoseRTLayout.RECT_GAP -
		 RoseRTLayout.getPixelHeight(rtLabel.getItemLabel()));
            rtCapVw.setObjLabel(rtLabel);
            rtStereo = new RoseRTItemLabel();
            rtStereo.setLabel(STEREOTYPE_LABEL);
            rtStereo.setItemLabel(CAPSULE_STEREOTYPE);
	    rtStereo.setMaxWidth
		(RoseRTLayout.getPixelWidth(rtStereo.getItemLabel()));
	    rtStereo.setPos
		(rtCapVw.getRect().getCenter().getX() -
		 rtStereo.getMaxWidth() / 2,
		 rtCapVw.getRect().getCenter().getY() -
		 2 * RoseRTLayout.RECT_GAP -
		 RoseRTLayout.getPixelHeight(rtStereo.getItemLabel()));
            rtCapVw.setStereotype(rtStereo);
            rtCompart = new RoseRTCompartment();
            rtCapVw.setCompartment(rtCompart);
            // add a hash map entry linking capsule class to view
            rtViewHash.put(rtCap, rtCapVw);
            // add capsule view to class diagram
            rtClassDiag.getItems().addLast(rtCapVw);
            rtCapVws.addLast(rtCapVw);
            instList.addLast(rtCapVw);

            // see if capsule has parent, if so
            //   - set supplier structure
            //   - create an inherit view object
            if (rtCap.hasParent()) {  // has parent, use refinement
                ((RoseRTRefinement)rtCap.getRefinements().get(0))
                    .setSupplierStructure(STRUCT_DELIM + COLLAB_NAME);
                // create new inherit view object
                rtInhVw = new RoseRTInheritVw("");
                rtInhVw.setID(generateID(null, rtInhVw));
		rtInhVw.setRectilinear();
                // set client (child) and supplier (parent) ID
                rtInhVw.setClient(rtCapVw.getID());
                rtPCapName = rtCap.getParent().getName();
                int idx = rtCapVws.findName(rtPCapName);
                if (idx < 0) {  // error, complain and move on
                    err.writeln("Fatal error: exportSystem(): could not locate capsule class");
                } else {  // found it
                    rtCapVw2 = (RoseRTCapsuleVw)rtCapVws.get(idx);
                    rtInhVw.setSupplier(rtCapVw2.getID());
		    // add link of parent class rect to inherit view
		    // then from inherit view to child class rect
		    rtInhVw.getRect().setName(INHVW_PREFIX+ rtInhVw.getName());
		    rtCapVw2.getRect().getLinks().addLast(rtInhVw);
		    rtCapVw.getRect().setName(CAPVW_PREFIX+ rtCapVw.getName());
		    rtInhVw.getRect().getLinks().addLast(rtCapVw.getRect());
                }
                // add to inheritance list
                rtInhVws.addLast(rtInhVw);
            }

            // add capsule to logical model
            petalDesign.classCat.logicalModels.addLast(rtCap);
        }
        // Now that all the necessary capsule view objects have been created,
        //   we can check for capsule role relationships
        RoseRTCapsule rtCap2 = null;
        RoseRTCapsuleRole rtCapRole = null;
        RoseRTCapsuleRoleRelVw rtRelVw = null;
        RoseRTSegLabel rtSegLabel = null;
        RoseRTList rtRelVws = new RoseRTList();
        RoseRTList rtClsRoles = null;
        String rtCapRoleName = null;
        String rtClassName = null;
        Enumeration jter = null;
        for (iter = classList.elements(); iter.hasMoreElements();) {
            rtCap2 = (RoseRTCapsule)iter.nextElement();
            // fetch the corresponding capsule view object
            if (rtViewHash.containsKey(rtCap2)) {  // get the capsule view
                rtCapVw2 = (RoseRTCapsuleVw)rtViewHash.get(rtCap2);
            } else {  // report error and move on
                err.writeln("Error: exportDesign(): could not locate capsule role view for " + rtCap2.getName() + " while extracting the capsule role relation views");
                continue;
            }
            // at this point, capsule view object is found
            // get list of classifier roles
            if (rtCap2.hasParent()) {  // has parent, use refinement
                rtClsRoles = ((RoseRTRefinement)rtCap2.getRefinements().get(0))
                    .getClassifierRoles();
            } else {  // just get classifier roles from the collaboration
                rtClsRoles = rtCap2.getCollab().getClassifierRoles();
            }
            // see if capsule contains classifier roles in its structure, if so
            //   - add a capsule role relationship view object for each
            if (rtClsRoles.size() > 0) {  // has classifier roles
                for (jter = rtClsRoles.elements(); jter.hasMoreElements();) {
                    rtCapRole = (RoseRTCapsuleRole)jter.nextElement();
                    // get the role name and the classifier name
                    rtCapRoleName = rtCapRole.getName();
                    rtClassName = rtCapRole.getClassifier();

                    //debug
                    //err.writeln("capsule role: " + rtCapRoleName +
                    //          "  parent capsule name: " + rtClassName);

                    // strip the capsule name of its path
                    if (rtClassName.indexOf(rootCatName + PATH_DELIM) == 0) {
                        // only remove root cat name if it's found at beginning
                        rtPCapName =
                            rtClassName.substring(rootCatName.length() +
                                                  PATH_DELIM.length());
                    } else {
                        rtPCapName = rtClassName;
                    }
                    int idx = classList.findName(rtPCapName);
                    if (idx < 0) {  // report error and move on
                        err.writeln("Error: exportDesign(): could not locate capsule class " + rtPCapName + " while extracting the capsule role relation views");
                        continue;
                    }
                    // otherwise, we fetch capsule object
                    rtCap = (RoseRTCapsule)classList.get(idx);
                    // fetch the corresponding capsule view object
                    if (rtViewHash.containsKey(rtCap)) {
                        // get the capsule view
                        rtCapVw = (RoseRTCapsuleVw)rtViewHash.get(rtCap);
                    } else {  // report error and move on
                        err.writeln("Error: exportDesign(): could not locate capsule role view for " + rtCap.getName() + " while extracting the capsule role relation views");
                        continue;
                    }
                    // at this point, we can create the capsule role relation
                    rtRelVw = new RoseRTCapsuleRoleRelVw(STRUCT_DELIM +
                                                         COLLAB_NAME +
                                                         STRUCT_DELIM +
                                                         rtCapRoleName);
                    rtRelVw.setID(generateID(null, rtRelVw));
		    rtRelVw.setRectilinear();
                    // label for the role capsule end
                    rtSegLabel = new RoseRTSegLabel();
                    rtSegLabel.setID(generateID(null, rtSegLabel));
                    rtSegLabel.setItemLabel(rtCapRoleName);
                    rtRelVw.setObjLabel(rtSegLabel);
                    // label for the parent end
                    rtSegLabel = new RoseRTSegLabel();
                    rtSegLabel.setID(generateID(null, rtSegLabel));
                    rtSegLabel.setHidden(true);
                    rtSegLabel.setItemLabel("");
                    rtRelVw.setObjLabel2(rtSegLabel);
                    // set the client and supplier capsule view ID
                    rtRelVw.setClient(rtCapVw.getID());
                    rtRelVw.setSupplier(rtCapVw2.getID());
		    // add link of container class rect to relations view
		    // then from relations view to contained class rect
		    rtRelVw.getRect().setName(RELVW_PREFIX+ rtRelVw.getName());
		    rtCapVw2.getRect().getLinks().addLast(rtRelVw);
		    rtCapVw.getRect().setName(CAPVW_PREFIX+ rtCapVw.getName());
		    rtRelVw.getRect().getLinks().addLast(rtCapVw.getRect());

                    // add capsule role relation view to relations list
                    rtRelVws.addLast(rtRelVw);
                }
            }
        }

        // Add protocols to logical model
        RoseRTProtocol rtPrtcl = null;
        for (iter = prtclList.elements(); iter.hasMoreElements();) {
            rtPrtcl = (RoseRTProtocol)iter.nextElement();
            petalDesign.classCat.logicalModels.addLast(rtPrtcl);
        }
        // Add inheritance view objects to class diagram
        for (iter = rtInhVws.elements(); iter.hasMoreElements();) {
            rtInhVw = (RoseRTInheritVw)iter.nextElement();
            rtClassDiag.getItems().addLast(rtInhVw);
        }
        // Add capsule role relationship view objects to class diagram
        for (iter = rtRelVws.elements(); iter.hasMoreElements();) {
            rtRelVw = (RoseRTCapsuleRoleRelVw)iter.nextElement();
            rtClassDiag.getItems().addLast(rtRelVw);
        }
        // Add class diagram to logical presentation
        petalDesign.classCat.logicalPresentations.addLast(rtClassDiag);
    }

    private void layoutClassDiagram() {
	
    }


    ////////////////////
    // Utility methods
    ////////////////////

    /**
     * Checks whether the Acme port is a base port based on some criteria
     *   of the port type name.
     *
     * @return boolean
     */
    private boolean checkIfBasePort(String portTypeName) {
        // look for "out", "source", or "src" in the port type name
        if (portTypeName.indexOf("out") > -1 ||
            portTypeName.indexOf("Out") > -1 ||
            portTypeName.indexOf("OUT") > -1 ||
            portTypeName.indexOf("source") > -1 ||
            portTypeName.indexOf("Source") > -1 ||
            portTypeName.indexOf("SOURCE") > -1 ||
            portTypeName.indexOf("src") > -1 ||
            portTypeName.indexOf("Src") > -1 ||
            portTypeName.indexOf("SRC") > -1) {

            // found, treat this as the base port
            return true;
        }
        return false;
    }

    private void deriveClassDiag(AcmeEnumeration componentsACME)
        throws java.io.IOException
    {
        // key for the entry that stores the min port type count
        AcmeComponent minKey = new AcmeComponent("_Min_");
        int val, minVal;
        AcmeComponent comp = null;
        AcmePort port = null;
        AcmeEnumeration portsACME = null;
        String typeName = null;  // used for first-level key
        AcmeComponent instComp = null;  // used for second-level key
        String portTypeName = null;  // used for third-level key
        Hashtable instHash = null;  // for the second-level map
        Hashtable portTypeHash = null;  // for the third-level map
        Hashtable minHash = null;  // for the min tally map
        Enumeration iter = null;  // iterator for Set

        // compute a tally of the port type occurrences for each component
        //   instance, and categorize by component types, then component
        //   instances, at the same time keeping track of the minimum count
        while (componentsACME.moreItems()) {
            comp = (AcmeComponent)componentsACME.getNextItem();
            typeName = getElementType(comp).getTypeName();

            // debug output
            err.writeln(comp.getName() + " : " + typeName);

            // see if a 2nd-level Hashtable exists for this component type
            if (tallyHash.containsKey(typeName)) {
                // yes, so no need to create a map
            } else {
                // no, so create a Hashtable for this component type
                instHash = new Hashtable();
                tallyHash.put(typeName, instHash);
            }
            // create a Hashtable for this component instance
            //   (instance names are unique, so no check needed)
            instComp = comp;
            portTypeHash = new Hashtable();
            instHash.put(instComp, portTypeHash);
            // now count all the ports for this instance
            portsACME = comp.getPorts();
            while (portsACME.moreItems()) {
                port = (AcmePort)portsACME.getNextItem();
                portTypeName = getElementType(port).getTypeName();

                // debug output
                err.writeln("    " + portTypeName);

                if (portTypeHash.containsKey(portTypeName)) {  // increment
                    val = ((Integer)portTypeHash.get(portTypeName)).intValue();
                    val++;
                    portTypeHash.put(portTypeName, new Integer(val));
                } else {  // first count
                    portTypeHash.put(portTypeName, new Integer(1));
                }
            }
            // check if a Min port count Hashtable exists, if not, create
            if (instHash.containsKey(minKey)) {  // get the minHash
                minHash = (Hashtable)instHash.get(minKey);
            } else {  // create a minHash
                minHash = new Hashtable();
                instHash.put(minKey, minHash);
            }
            // now compare and update the current min tally
            iter = portTypeHash.keys(); //.iterator();
            while (iter.hasMoreElements()) {
                portTypeName = (String)iter.nextElement();  // get next port type
                // extract values from the port Hashtable
                val = ((Integer)portTypeHash.get(portTypeName)).intValue();
                if (minHash.containsKey(portTypeName)) {  // exists, compare
                    // extract the current min value
                    minVal = ((Integer)minHash.get(portTypeName)).intValue();
                    if (val < minVal) {  // update as min value
                        minHash.put(portTypeName, new Integer(val));
                    }
                } else {  // not exist, put value in as min value
                    minHash.put(portTypeName, new Integer(val));
                }
            }
        }
        //=> At this point, Hashtable contains the tally of all port type
        //   occurrence counts of all component instances of all compoment
        //   types in this system, and the min count for each compoment type

        // We now generate capsule classes based on the content of the Hash
        RoseRTCapsule rtParent = null;
        RoseRTCapsule rtCap = null;
        RoseRTPort rtPort = null;
        RoseRTInheritanceRel rtInhRel = null;
        String newCapName = null;
        Enumeration jter = null;  // another iterator for Set
        Enumeration kter = null;  // another iterator for Set
        iter = tallyHash.keys(); //.iterator();
        while (iter.hasMoreElements()) {
            // Generate base capsule class for the component type
            typeName = (String)iter.nextElement();  // get next component type
            // see if Capsule object of same name already exists
            int idx = classList.findName(typeName);
            if (idx < 0) {  // not found
                // create new Capsule object
                rtCap = new RoseRTCapsule(typeName);
                rtCap.setCollab(new RoseRTCapsuleCollab(COLLAB_NAME));
                // add to list of capsule classes
                classList.addLast(rtCap);
            } else {  // found, use it
                rtCap = (RoseRTCapsule)classList.get(idx);
            }
            rtParent = rtCap;
            // retrive port types from minHash
            instHash = (Hashtable)tallyHash.get(typeName);
            minHash = (Hashtable)instHash.get(minKey);
            jter = minHash.keys(); //.iterator();
            while (jter.hasMoreElements()) {
                portTypeName = (String)jter.nextElement();
                // get min tally of port type
                minVal = ((Integer)minHash.get(portTypeName)).intValue();
                // create that many ports
                for (int i=1; i <= minVal; i++) {
                    rtPort = generateRTPort(portTypeName, i);
                    // add to the capsule's port list
                    rtCap.getCollab().getPorts().addLast(rtPort);
                }
            }

            // Generate sub-capsule classes for the extended component types
            jter = instHash.keys(); //.iterator();
            while (jter.hasMoreElements()) {
                // next component instance
                instComp = (AcmeComponent)jter.nextElement();
                // determine whether this instance has different port
                //   occurrence count compared to minHash, and whether it has
                //   any representation; if either is the case, a sub-class
                //   needs to be created
                if (isHashSameCount((Hashtable)instHash.get(instComp), minHash)
		    && instComp.getRepresentations().getSize() == 0) {
                    // nope!  same count as minHash and no representation, so
                    //   don't create new sub-class for this component instance
                    continue;  // refresh loop
                }
                // otherwise, create new Capsule class using instance name
                //   appended with the parent capsule name, if the parent
                //   capsule name isn't already part of the instance name
                //   e.g.  if parent is named "Filter" and instance, "Merger",
                //         then the new name would be "MergerFilter".
                //         whereas if the instance is named "MergeFilter,"
                //         then the new name would be unchanged ("MergeFilter")
                newCapName = new String(instComp.getName());
                if (newCapName.indexOf(rtParent.getName()) < 0) {
                    // need to append parent capsule name
                    newCapName += rtParent.getName();
                    // modify the ACME component's name also
                    try {
                        instComp.setName(newCapName);
                    } catch (AcmeIllegalNameException aine) {
                        err.writeln("Name change for " + instComp.getName() +
                                       " failed!");
                    }
                }
                rtCap = new RoseRTCapsule(newCapName);
                rtInhRel = new RoseRTInheritanceRel();
                rtInhRel.setSupplier(rootCatName + PATH_DELIM +
                                     rtParent.getName());
                rtCap.getInheritance().addLast(rtInhRel);
                rtCap.getRefinements().addLast(new RoseRTRefinement());
                rtCap.setParent(rtParent);
                if (instComp.getRepresentations().getSize() > 0) {
                    rtCap.setHasRep(true);
                }
                // add to list of capsule classes
                classList.addLast(rtCap);
                // retrieve port types
                portTypeHash = (Hashtable)instHash.get(instComp);
                kter = portTypeHash.keys(); //.iterator();
                while (kter.hasMoreElements()) {
                    portTypeName = (String)kter.nextElement();
                    // get tally and min tally of port type
                    val = ((Integer)portTypeHash.get(portTypeName)).intValue();
                    minVal = ((Integer)minHash.get(portTypeName)).intValue();
                    // create the additional ports, if necessary
                    // # assert(val >= minVal)
                    for (int i=minVal+1; i <= val; i++) {
                        rtPort = generateRTPort(portTypeName, i);
                        // add to the capsule's port list
                        ((RoseRTRefinement)rtCap.getRefinements().get(0))
                            .getPorts().addLast(rtPort);
                    }
                }
            }
        }

        // Rename all ports of the component instances, according to count
        //   Since there's no exact correspondence of port instance with a
        //   particular index, we can assign any number.
        componentsACME.reset();
        while (componentsACME.moreItems()) {
            comp = (AcmeComponent)componentsACME.getNextItem();
            portsACME = comp.getPorts();
            portTypeHash = new Hashtable();
            while (portsACME.moreItems()) {
                port = (AcmePort)portsACME.getNextItem();
                portTypeName = getElementType(port).getTypeName();
                // get count, use as index
                if (portTypeHash.containsKey(portTypeName)) {  // increment
                    val = ((Integer)portTypeHash.get(portTypeName)).intValue();
                    val++;
                    portTypeHash.put(portTypeName, new Integer(val));
                } else {  // first count
                    val = 1;
                    portTypeHash.put(portTypeName, new Integer(val));
                }
                // change port name
                try {
                    //debug
                    //err.writeln(port.getName() + " name changed");

                    port.setName(generateRTPortName(portTypeName, val));
                } catch (AcmeIllegalNameException aine) {
                    err.writeln("Name change for " + port.getName() +
                                       " failed!");
                }
            }
        }
    }

    private void deriveProtocols(AcmeEnumeration connectorsACME,
                                 AcmeEnumeration attachmentsACME) {
        AcmeConnector conn = null;
        String typeName = null;
        RoseRTProtocol rtProt = null;

        // iterate through all connectors and add their type names as a
        //   protocols to the protocol list, unique only
        while (connectorsACME.moreItems()) {
            // get next connector
            conn = (AcmeConnector)connectorsACME.getNextItem();
            typeName = getElementType(conn).getTypeName();
            // see if we need to create a RT protocol object
            if (prtclList.findName(typeName) < 0) {  // create one
                rtProt = new RoseRTProtocol(typeName);
                // we assume binary protocol (i.e. binary connector)
                rtProt.makeBinary();
                prtclList.addLast(rtProt);
            } else {  // no need to do anymore
                continue;
            }
            // now determine which role is the base role and which conjugate
            //   by checking the attached port
            AcmeEnumeration rolesACME = conn.getRoles();
            String portTypeName = null;
            String basePortName = null;
            String conjPortName = null;
            while (rolesACME.moreItems()) {
                AcmeRole roleACME = (AcmeRole)rolesACME.getNextItem();
                // find the port attached to the role
                AcmePort portACME = findAttachedPort(roleACME,
                                                     attachmentsACME);
                if (portACME == null) {
                    continue;
                }
                // port found, get the port type name
                portTypeName = getElementType(portACME).getTypeName();

                // debug
                //err.writeln("protocol port: " + portTypeName);

                // see if port can be considered a base port
                if (checkIfBasePort(portTypeName)) {  // yes!
                    basePortName = portTypeName;
                } else {
                    if (! rolesACME.moreItems() &&
                        basePortName == null) {
                        // This is the last port and we haven't found a
                        //   candidate base port, use the first port, which
                        //   should currently be in "conjPortName," as base
                        basePortName = conjPortName;
                        // after this, "conjPortName" will be assigned with
                        //   the last port
                    }
                    conjPortName = portTypeName;
                }
            }

            // debug
            //err.writeln("base port: " + basePortName);
            //err.writeln("conjugate port: " + conjPortName);

            // save the base and conjugate port information
            rtProt.setBasePortName(basePortName);
            rtProt.setConjugatePortName(conjPortName);
        }
    }

    private AcmePort findAttachedPort(AcmeRole roleACME,
                                      AcmeEnumeration attachmentsACME) {
        if (attachmentsACME == null) {  // can't do much
            return null;
        }

        AcmeAttachment attACME = null;
        attachmentsACME.reset();
        while (attachmentsACME.moreItems()) {
            attACME = (AcmeAttachment)attachmentsACME.getNextItem();
            if (attACME.getRole().equals(roleACME)) {  // found it
                return attACME.getPort();
            }
        }
        //=> at this point, not found
        return null;
    }

    private int generateID(AcmeElement elementACME, RoseRTObject rtObj) {
        int id = nextId++;
        Integer idI = new Integer(id);
        if (elementACME != null) {
            idHash.put(elementACME, idI);
        }
        rtObjHash.put(idI, rtObj);
        return id;
    }

    /**
     * Given the connector view object, finds the actual supplier and client
     *   object referred to by the corresponding ID, and use their names
     *   to generate the connector object's ends.
     * Requires rtObjHash hashtable
     *
     * @param rtConn the RT Connector object to add ends to
     * @param rtConnVw the RT Connector view object to get supplier & client
     * @return RoseRTConnector the modified RT Connector object with new ends
     */
    private RoseRTConnector generateRTConnEnds(RoseRTConnector rtConn,
					       RoseRTConnectorVw rtConnVw) {
	// Get the actual supplier & client ends for the RT connector obj
	RoseRTObject rtObj = null;
	String itemName = null;
	RoseRTString endName = null;
	Integer idI = new Integer(rtConnVw.getSupplier());
	if (rtObjHash.containsKey(idI)) {
	    // get the supplier name
	    rtObj = (RoseRTObject)rtObjHash.get(idI);
	    itemName = rtObj.getName();
	    endName = new RoseRTString("ends \"" + itemName + "\"");
	    rtConn.getEnds().addLast(endName);
	}
	idI = new Integer(rtConnVw.getClient());
	if (rtObjHash.containsKey(idI)) {
	    // get the client name
	    rtObj = (RoseRTObject)rtObjHash.get(idI);
	    itemName = rtObj.getName();
	    endName = new RoseRTString("ends \"" + itemName + "\"");
	    rtConn.getEnds().addLast(endName);
	}

	return rtConn;
    }

    private RoseRTPort generateRTPort(String portTypeName, int index) {
        // create RT port object
        String name = generateRTPortName(portTypeName, index);
        RoseRTPort rtPort = new RoseRTPort(name);
        // get RT port type (protocol) and determine if port is a conjugate
        String classifierName = null;
        for (Enumeration iter = prtclList.elements();
             iter.hasMoreElements();) {
            RoseRTProtocol rtPrtcl = (RoseRTProtocol)iter.nextElement();
            classifierName = rootCatName + PATH_DELIM + rtPrtcl.getName();

            // debug
            //err.writeln(classifierName);
            rtPort.setClassifier(classifierName);
            String basePortName = rtPrtcl.getBasePortName();
            String conjugatePortName = rtPrtcl.getConjugatePortName();
            if (basePortName != null && basePortName.equalsIgnoreCase(portTypeName)) {
                // found the protocol that port belongs to, and it's base port
                rtPort.setConjugated(false);
                break;
            } else if (conjugatePortName != null && conjugatePortName.equalsIgnoreCase(portTypeName)) {
                // found the protocol that port belongs to, and its conjugate
                rtPort.setConjugated(true);
                break;
            }
        }
        instList.addLast(rtPort);

        return rtPort;
    }

    private String generateRTPortName(String portTypeName, int index) {
        // generate port instance name
        String newName = new String(portTypeName);
        // remove "Port" from name, if exist
        int pos = portTypeName.indexOf("Port");
        if (pos == -1) {  // try lower case
            pos = portTypeName.indexOf("port");
        }
        if (pos == -1) {  // try all upper case
            pos = portTypeName.indexOf("PORT");
        }
        if (pos > -1) {  // found, chop off
            newName = portTypeName.substring(0, pos);
        }
        // insert lowercase p to name at beginning, for port
        newName = "p" + newName;
        // append number, if > 1
        if (index > 1) {
            newName = newName + index;
        }
        return newName;
    }

    /**
     * Iterates through the component list and recursively calls itself on the
     *   component list of any component representation, collecting any
     *   components, connectors, and attachments.  The net effect is
     *   to flatten the components, connectors, and attachments of the entire
     *   system into single lists.
     *
     * @param componentsACME list of top-level components
     * @param allComponentsACME list of top-level components
     * @param allConnectorsACME list of top-level connectors
     * @param allAttachmentsACME list of top-level attachments
     */
    private void getAllItems(AcmeEnumeration componentsACME,
                             AcmeEnumeration allComponentsACME,
                             AcmeEnumeration allConnectorsACME,
                             AcmeEnumeration allAttachmentsACME) {
        AcmeComponent compACME = null;
        AcmeEnumeration repsACME = null;
        AcmeRepresentation repACME = null;
        AcmeSystem sysACME = null;
        while (componentsACME.moreItems()) {
            compACME = (AcmeComponent)componentsACME.getNextItem();
            repsACME = compACME.getRepresentations();
            // for simplicity, we will assume only one representation
            if (repsACME.moreItems()) {
                repACME = (AcmeRepresentation)repsACME.getNextItem();
                sysACME = repACME.getSystem();
                allComponentsACME.concatenate(sysACME.getComponents());
                allConnectorsACME.concatenate(sysACME.getConnectors());
                allAttachmentsACME.concatenate(sysACME.getAttachments());
                getAllItems(sysACME.getComponents(), allComponentsACME,
                            allConnectorsACME, allAttachmentsACME);
            }
        }
    }

    /**
     * Returns a list of all ports defined in the given capsule and inherited
     *   from ancestor capsules.
     * We made some assumptions in this method as we've been in the others:
     *   - ONE capsule refinement, if capsule has parent
     *   - ONE superclass, if capsule has parent
     * This method requires the existence of a capsule class list (classList)
     *   and a defined root category name (rootCatName).
     *
     * @param rtCap RT Capsule for which to get ports
     * @return RoseRTList list of ports
     */
    private RoseRTList getAllRTPorts(RoseRTCapsule rtCap) {
        RoseRTList rtPorts = new RoseRTList();
        if (rtCap.hasParent()) {  // has parent
            // get the ports defined in capsule first
            RoseRTRefinement rtRef =
                (RoseRTRefinement)rtCap.getRefinements().get(0);
            rtPorts.addAll(rtRef.getPorts().getList());
            // find the parent capsule (superclass)
            RoseRTInheritanceRel rtInh =
                (RoseRTInheritanceRel)rtCap.getInheritance().get(0);
            String rtPCapName = rtInh.getSupplier();
            // strip the capsule name of its path
            if (rtPCapName.indexOf(rootCatName + PATH_DELIM) == 0) {
                // only remove root cat name if it's found at the beginning
                rtPCapName = rtPCapName.substring(rootCatName.length() +
                                                  PATH_DELIM.length());
            }
            // find capsule by name in class list
            int idx = classList.findName(rtPCapName);
            if (idx >= 0) {  // found
                RoseRTCapsule rtPCap =
                    (RoseRTCapsule)classList.get(idx);
                // recursively call method on the parent capsule
                rtPorts.addAll(getAllRTPorts(rtPCap).getList());
            }  // if not found, do nothing more
        } else {  // no parent, so just return the ports defined in capsule
            rtPorts.addAll(rtCap.getCollab().getPorts().getList());
        }

        return rtPorts;  // hand over the list of ports
    }

    /**
     * Returns the Acme type of the element, selecting only the first one
     *   in the enumeration.
     *
     * @param elementACME Acme element to find the type of
     * @return AcmeElementType Acme type of element
     */
    private AcmeElementType getElementType(AcmeElement elementACME) {
        AcmeEnumeration typesACME = elementACME.getDeclaredTypes();
        // for simplicity, we will assume only one parent type,
        //   i.e. no instance has multiple types
        if (typesACME.moreItems()) {
            return (AcmeElementType)typesACME.getNextItem();
        }
        return null;
    }

    /**
     * Searches the class list first to find the capsule class with the
     *   same name as this component.  Then searches the tally hash map to
     *   find the capsule class whose port counts best fits the given Acme
     *   component's port counts.
     *
     * @param componentACME Acme component to match capsule class to
     * @return String name of matching capsule class object
     */
    private String getMatchingCapsuleClass(AcmeComponent componentACME)
	throws java.io.IOException
    {
        // Search the class list for a matching name
        int idx = classList.findName(componentACME.getName());
        if (idx > -1) {
            return ((RoseRTCapsule)classList.get(idx)).getName();
        }
        // Generate a port count hash
        int val;
        AcmeEnumeration portsACME = null;
        AcmePort portACME = null;
        String portTypeName = null;
        Hashtable cntHash = new Hashtable();
        // get the component's list of ports
        portsACME = componentACME.getPorts();
        while (portsACME.moreItems()) {
            portACME = (AcmePort)portsACME.getNextItem();
            portTypeName = getElementType(portACME).getTypeName();
            if (cntHash.containsKey(portTypeName)) {  // increment
                val = ((Integer)cntHash.get(portTypeName)).intValue();
                val++;
                cntHash.put(portTypeName, new Integer(val));
            } else {  // first count
                cntHash.put(portTypeName, new Integer(1));
            }
        }
        // Search thru tally map
        Hashtable instHash = null;
        Hashtable portTypeHash = null;
        String typeName = null;
        AcmeComponent instComp = null;
        Enumeration iter = null;
        Enumeration jter = null;  // another iterator for Set
        iter = tallyHash.keys(); //.iterator();
        while (iter.hasMoreElements()) {
            typeName = (String)iter.nextElement();  // get next component type
            instHash = (Hashtable)tallyHash.get(typeName);
            jter = instHash.keys(); //.iterator();
            while (jter.hasMoreElements()) {
                instComp = (AcmeComponent)jter.nextElement();
		//debug
	        //err.writeln(">> comparing against " + instComp.getName());
                portTypeHash = (Hashtable)instHash.get(instComp);
                if (isHashSameCount(portTypeHash, cntHash)) {
		    // matched the count!

		    //debug
		    //err.writeln("--- match found!");
                    return typeName;
                }
            }
        }
        return null;
    }

    private boolean includeProperty(AcmeProperty property) {
        boolean inc = true;
        if (!includeVisProperties) {
            String name = property.getName();
            if (name.length() >= 4 &&
		name.substring(0,4).equalsIgnoreCase("vis-")){
                inc = false;
            }
        }
        return inc;
    }

    /**
     * Implements an "equality" operator for hash tables, which should have
     *   been available with the Hashtable class, but is apparently not
     *   implemented with the expected behavior in Java 1.1 (MSJVM).
     * We take two hash tables and first compare the size.
     * Having determined that they are the same size (i.e. = cardinality),
     *   we iterate thru the keys of one table and compare the values agains
     *   those of the same key in the second table.  If all values come out
     *   equal, then we conclude that the hash tables are equal.
     * This is used in this class to compare the equality of port type counts.
     *
     * @return boolean
     */
    private boolean isHashSameCount(Hashtable ht1, Hashtable ht2) {
	if (ht1.size() != ht2.size()) {  // then they cannot be equal
	    return false;
	}
	Object keyObj = null;
	for (Enumeration iter = ht1.keys(); iter.hasMoreElements();) {
	    keyObj = iter.nextElement();
	    if (! ht1.get(keyObj).equals(ht2.get(keyObj))) {
		return false;
	    }
	}
	//=> It has iterated through all elements without failing equality
	//   test, so hashtables are equal.
	return true;
    }

    private void debugLists()
        throws java.io.IOException
    {
        // debug:  show the list of classes and instances generated
        err.writeln("Classes:");
        err.indent();
        RoseRTObject obj = null;
        for (Enumeration li = classList.elements();
             li.hasMoreElements();) {
            obj = (RoseRTObject)li.nextElement();
            err.writeln(obj.getName());
        }
        err.unIndent();
        err.writeln("Instances:");
        err.indent();
        for (Enumeration li = instList.elements();
             li.hasMoreElements();) {
            obj = (RoseRTObject)li.nextElement();
            err.writeln(obj.getName());
        }
        err.unIndent();
        err.writeln("Protocols:");
        err.indent();
        for (Enumeration li = prtclList.elements();
             li.hasMoreElements();) {
            obj = (RoseRTObject)li.nextElement();
            err.writeln(obj.getName());
        }
        err.unIndent();
    }
}
