package acme.translators.RoseRTTypes;

/**
 * Class: RoseRTCapsule <br>
 * Purpose: Provides a Rose RT capsule object. <br>
 * <br>
 * Revision History: <br>
 *   Date    Person    Change: <br>
 * 5/23/01    SWC      Created initial version <br>
 * <br>
 * Known bugs/omissions:<br>
 *   - <br>
 * @author Owen Cheng (SWC)
 */

public class RoseRTCapsule extends RoseRTObject {

    static private final String OBJ = "Capsule";
    static private final String NAM = "Capsule";

    private RoseRTCapsuleCollab collab = null;
    private RoseRTList classAttribs = null;
    private String language = "C++";
    private RoseRTCollabDiagram collabDiag = null;
    private RoseRTList inhRels = null;
    private RoseRTList refinements = null;

    // content that is not printed
    private RoseRTCapsule parent = null;
    private boolean hasRep = false;

    /**
     * Default class constructor.
     */
    public RoseRTCapsule() {
	super();

	this.setType(OBJ);
	this.setName(NAM);
	this.classAttribs = new RoseRTList();
	this.classAttribs.setLabel("class_attributes");
	this.classAttribs.setType("class_attribute_list");
	this.inhRels = new RoseRTList();
	this.inhRels.setLabel("superclasses");
	this.inhRels.setType("inheritance_relationship_list");
	this.refinements = new RoseRTList();
	this.refinements.setLabel("refinements");
	this.refinements.setType("refinementsList");
    }

    /**
     * Class constructor for initializing name.
     *
     * @param name the given name of this object
     */
    public RoseRTCapsule(String name) {
	this();

	this.setName(name);
    }

    /**
     * Returns this capsule's collaboration content.
     *
     * @return RoseRTCapsuleCollab
     */
    public RoseRTCapsuleCollab getCollab() {
	return this.collab;
    }

    /**
     * Replaces the collaboration content of this capsule.
     *
     * @param collab new RoseRTCapsuleCollab to replace with
     */
    public void setCollab(RoseRTCapsuleCollab collab) {
	this.collab = collab;
    }

    /**
     * Returns the list of class attributes of this capsule.
     *
     * @return RoseRTList
     */
    public RoseRTList getClassAttribs() {
	return this.classAttribs;
    }

    /**
     * Replaces the list of class attributes of this capsule.
     *
     * @param list new RoseRTList to replace with
     */
    public void setClassAttribs(RoseRTList list) {
	this.classAttribs = list;
    }

    /**
     * Returns this capsule's collaboration diagram.
     *
     * @return RoseRTCollabDiagram
     */
    public RoseRTCollabDiagram getCollabDiag() {
	return this.collabDiag;
    }

    /**
     * Replaces the collaboration diagram of this capsule.
     *
     * @param collabDiag new RoseRTCapsuleCollab to replace with
     */
    public void setCollabDiag(RoseRTCollabDiagram collabDiag) {
	this.collabDiag = collabDiag;
    }

    /**
     * Returns this capsule's list of inheritance objects.
     *
     * @return RoseRTList.
     */
    public RoseRTList getInheritance() {
	return this.inhRels;
    }

    /**
     * Replaces the list of inheritance objects of this capsule.
     *
     * @param list new RoseRTList to replace with
     */
    public void setInheritance(RoseRTList list) {
	this.inhRels = list;
    }

    /**
     * Returns this capsule's list of refinement objects.
     *
     * @return RoseRTList.
     */
    public RoseRTList getRefinements() {
	return this.refinements;
    }

    /**
     * Replaces the list of refinement objects of this capsule.
     *
     * @param list new RoseRTList to replace with
     */
    public void setRefinements(RoseRTList list) {
	this.refinements = list;
    }

    /**
     * Returns whether this capsule has a parent.
     *
     * @return boolean
     */
    public boolean hasParent() {
	return (this.parent != null);
    }

    /**
     * Returns this capsule's parent capsule.
     *
     * @return RoseRTCapsule
     */
    public RoseRTCapsule getParent() {
	return this.parent;
    }

    /**
     * Replaces the parent capsule of this capsule.
     *
     * @param parent new RoseRTCapsule to replace with
     */
    public void setParent(RoseRTCapsule parent) {
	this.parent = parent;
    }

    /**
     * Returns whether this capsule has an Acme representation.
     *
     * @return boolean
     */
    public boolean hasRep() {
	return this.hasRep;
    }

    /**
     * Replaces the flag of whether this capsule has an Acme representation.
     *
     * @param value boolean value to set to
     */
    public void setHasRep(boolean value) {
	this.hasRep = value;
    }

    /**
     * Returns the string representation of this object, basically generating
     * all the content.
     *
     * @param leftMarg amount of spaces to indent for all lines
     * @param indIncr amount of increment spaces to indent for subsequent lines
     * @return String
     */
    public String toString(int leftMarg, int indIncr) {
	String str = "";
	int ind = leftMarg + indIncr;
	String clsattrbStr = classAttribs.toString(ind, indIncr);
	String inhStr = inhRels.toString(ind, indIncr);
	String refStr = refinements.toString(ind, indIncr);
	str += Util.indent(leftMarg) + "(object " + this.getType()
	    + " \"" + this.getName() + "\"" +
	    ((inhStr.equals("")) ? "" : ("\n" + inhStr)) +
	    ((refStr.equals("")) ? "" : ("\n" + refStr)) +
	    ((collab != null) ?
	     ("\n" + collab.toString(ind, indIncr)) : "") +
	    (clsattrbStr.equals("") ? "" : ("\n" + clsattrbStr)) +
	    "\n" + Util.indent(ind) + "language \"" + language + "\"" +
	    ((collabDiag != null) ?
	     ("\n" + collabDiag.toString(ind, indIncr)) : "") +
	    ")";

	return str;
    }

    /**
     * Returns the string representation of this object, basically generating
     * all the content.
     *
     * @return String
     */
    public String toString() {
	return this.toString(0, 8);
    }
}
