package acme.translators.RoseRTTypes;

import java.util.Enumeration;

/**
 * Class: RoseRTLayout <br>
 * Purpose: Provides layout methods for Rose RT objects. <br>
 * <br>
 * Revision History: <br>
 *   Date    Person    Change: <br>
 * 5/23/01    SWC      Created initial version <br>
 * <br>
 * Known bugs/omissions:<br>
 *   - <br>
 * @author Owen Cheng (SWC)
 */

public class RoseRTLayout extends RoseRTObject {

    static private final int RECT_WIDTH = 400;
    static private final int RECT_HEIGHT = 450;
    static public final int RECT_GAP = 50;
    static private final int LEFT_MARGIN = 300;
    static private final int MAX_COLUMN = 5;

    static private final int FONT_SIZE = 8;

    static public final float OLD_CENTER_X = 360.0f;
    static public final float OLD_CENTER_Y = 440.0f;
    static public final float OLD_RCENTER_X = 375.0f;
    static public final float OLD_RCENTER_Y = 440.0f;
    static private final float NEW_CENTER_X = 900.0f;
    static private final float NEW_CENTER_Y = 600.0f;
    static private final int X_OFFSET_SMALL = 10;
    static private final int Y_OFFSET_SMALL = 0;
    static private final int X_OFFSET = 20;
    static private final int Y_OFFSET = 30;
    static private final int FONT_HEIGHT = 40;
    static private final float DIST_FACTOR = 2.3f;  // preferably NEW_X / OLD_X
    static private final float EXPAND_FACTOR = 3.2f;
    static private final float X_DIST_OFFSET = 100.0f;
    static private final float Y_DIST_OFFSET = 150.0f;

    static public float oldCenterX = OLD_CENTER_X;
    static public float oldCenterY = OLD_CENTER_Y;

    static public RoseRTList graph = new RoseRTList();
    static private int columnIdx = 0;
    static private int rowIdx = 0;

    /**
     * Returns a center coordinate
     *
     * @return RoseRTPoint center coordinate
     */
    static public RoseRTPoint getCenter() {
	return new RoseRTPoint((int)NEW_CENTER_X, (int)NEW_CENTER_Y);
    }

    /**
     * Returns a screen width value
     *
     * @return int width value
     */
    static public int getWidth() {
	return (int)((NEW_CENTER_X - 50) * 2);
    }

    /**
     * Returns a screen height value
     *
     * @return int height value
     */
    static public int getHeight() {
	return (int)((NEW_CENTER_Y - 50) * 2);
    }

    /**
     * Given an x coordinate, transforms it to a new X by moving it from some
     *   center coordinate by some factor.
     *
     * @param x given X coordinate
     * @return int the transformed X
     */
    static public int transformX(int x) {
	return transformX((float)x);
    }
    static public int transformX(float x) {
	float newX = (x - oldCenterX) * DIST_FACTOR;
	if (newX < (-X_DIST_OFFSET/2)) {
	    newX -= X_DIST_OFFSET;
	} else if (newX > (X_DIST_OFFSET/2)) {
	    newX += X_DIST_OFFSET;
	}
	newX += NEW_CENTER_X;
	return (int)newX;
    }

    /**
     * Given an y coordinate, transforms it to a new Y by moving it from some
     *   center coordinate by some factor.
     *
     * @param y given Y coordinate
     * @return int the transformed Y
     */
    static public int transformY(int y) {
	return transformY((float)y);
    }
    static public int transformY(float y) {
	float newY = (y - oldCenterY) * DIST_FACTOR;
	if (newY < (-Y_DIST_OFFSET/2)) {
	    newY -= Y_DIST_OFFSET;
	} else if (newY > (Y_DIST_OFFSET/2)) {
	    newY += Y_DIST_OFFSET;
	}
	newY += NEW_CENTER_Y;
	return (int)newY;
    }

    /**
     * Given some length (either X or Y), transforms it to a new length by
     *   multiplying the original by some factor.
     *
     * @param len given length
     * @return int the new length
     */
    static public int transformLength(int len) {
	return transformLength((float)len);
    }
    static public int transformLength(float len) {
	float newLen = len * EXPAND_FACTOR;
	return (int)newLen;
    }

    /**
     * Given some capsule's coordinate, returns a suitable coordinate for the
     *   capsule's label.
     *
     * @param pos capsule coordinate
     * @return RoseRTPoint the label coordinate
     */
    static public RoseRTPoint getCapLabelPos(RoseRTPoint pos) {
	return adjustPos(pos, -X_OFFSET_SMALL, -Y_OFFSET_SMALL - FONT_HEIGHT);
    }

    /**
     * Given some object's coordinate, returns a suitable coordinate for its
     *   label.
     *
     * @param pos object coordinate
     * @return RoseRTPoint the label coordinate
     */
    static public RoseRTPoint getLabelPos(RoseRTPoint pos) {
	return adjustPos(pos, X_OFFSET * 2, -Y_OFFSET);
    }

    /**
     * Given some object's coordinate and the coordinate of the object it is
     *   attached to, returns a suitable coordinate for its label.
     *
     * @param pos object coordinate
     * @param posA coordinate of the attachee object
     * @param dim dimension of the attachee object
     * @return RoseRTPoint the label coordinate
     */
    static public RoseRTPoint getLabelPos(RoseRTPoint pos, RoseRTPoint posA,
					  RoseRTPoint dim) {
	if (pos.getX() >= posA.getX() && pos.getY() >= posA.getY()) {
	    // adjust to the lower right
	    return adjustPos(posA, dim.getX()/2 - X_OFFSET,
			     dim.getY()/2 + Y_OFFSET - FONT_HEIGHT);
	} else if (pos.getX() >= posA.getX() && pos.getY() < posA.getY()) {
	    // adjust to the upper right
	    return adjustPos(posA, dim.getX()/2 - X_OFFSET,
			     - dim.getY()/2 - Y_OFFSET - FONT_HEIGHT);
	} else if (pos.getX() < posA.getX() && pos.getY() >= posA.getY()) {
	    // adjust to the lower left
	    return adjustPos(posA, - dim.getX()/2 + X_OFFSET,
			     dim.getY()/2 + Y_OFFSET - FONT_HEIGHT);
	} else {
	    // adjust to the upper left 
	    return adjustPos(posA, - dim.getX()/2 + X_OFFSET,
			     - dim.getY()/2 - Y_OFFSET - FONT_HEIGHT);
	}
    }

    /**
     * Given some object's coordinate and the coordinate of the object it is
     *   attached to, returns a suitable coordinate for its label.
     *
     * @param pos object coordinate
     * @param posA coordinate of the attachee object
     * @return RoseRTPoint the label coordinate
     */
    static public RoseRTPoint getNearestLabelPos(RoseRTPoint pos,
						 RoseRTPoint posA) {
	if (pos.getX() >= posA.getX() && pos.getY() >= posA.getY()) {
	    // adjust to inner lower right
	    return adjustPos(pos, - X_OFFSET * 4, - Y_OFFSET - FONT_HEIGHT);
	} else if (pos.getX() >= posA.getX() && pos.getY() < posA.getY()) {
	    // adjust to inner upper right
	    return adjustPos(pos, - X_OFFSET * 4, Y_OFFSET - FONT_HEIGHT);
	} else if (pos.getX() < posA.getX() && pos.getY() >= posA.getY()) {
	    // adjust to inner lower left
	    return adjustPos(pos, + X_OFFSET * 3, - Y_OFFSET - FONT_HEIGHT);
	} else {
	    // adjust to inner upper left 
	    return adjustPos(pos, + X_OFFSET * 3, Y_OFFSET - FONT_HEIGHT);
	}
    }

    /**
     * Given an old attacher object's coordinate and the attachee origin point,
     *   plus a new attachee object's origin and dimension, returns a new
     *   coordinate for attacher object on the attachee object's border.
     *
     * @param pos old attacher object coordinate
     * @param originOld old attachee object origin
     * @param originNew new attachee object origin
     * @param dim dimension of new attachee object
     * @return RoseRTPoint the attacher new coordinate
     */
    static public RoseRTPoint getExtrapPos(RoseRTPoint pos,
					   RoseRTPoint originOld,
					   boolean isUpLow,
					   RoseRTPoint originNew,
					   RoseRTPoint dim) {
	int x, y;
	if (isUpLow) {  // resided on the upper of lower border
	    if (pos.getY() >= originOld.getY()) {
		y = originNew.getY() + dim.getY()/2;
	    } else {
		y = originNew.getY() - dim.getY()/2;
	    }
	    x = (int)(((float)dim.getY() /
		       ((float)pos.getY() - originOld.getY())) *
		      (pos.getX() - originOld.getX()) + originNew.getX());
	    if (x > (originNew.getX() + dim.getX()/2)) {
		x = originNew.getX() + dim.getX()/2;
	    }
	    if (x < (originNew.getX() - dim.getX()/2)) {
		x = originNew.getX() - dim.getX()/2;
	    }
	} else {
	    if (pos.getX() >= originOld.getX()) {
		x = originNew.getX() + dim.getX()/2;
	    } else {
		x = originNew.getX() - dim.getX()/2;
	    }
	    y = (int)(((float)dim.getX() /
		       ((float)pos.getX() - originOld.getX())) *
		      (pos.getY() - originOld.getY()) + originNew.getY());
	    if (y > (originNew.getY() + dim.getY()/2)) {
		y = originNew.getY() + dim.getY()/2;
	    }
	    if (y < (originNew.getY() - dim.getY()/2)) {
		y = originNew.getY() - dim.getY()/2;
	    }
	}

	//System.err.println(pos + " " + originOld + " " + isUpLow + " " +
	//		   originNew + " " + dim + " new: (" + x + "," + y +
	//		   ")");
	return new RoseRTPoint(x, y);
    }


    /**
     * Given some point coordinate, adjusts by some offset value and return
     *   the new point coordinate.
     *
     * @param pos given coordinate
     * @return RoseRTPoint the new coorindate
     */
    static public RoseRTPoint adjustPos(RoseRTPoint pos,
					int x_off, int y_off) {
	RoseRTPoint newPt = new RoseRTPoint(pos.getX(), pos.getY());
	newPt.setX(newPt.getX() + x_off);
	newPt.setY(newPt.getY() + y_off);
	return newPt;
    }


    /**
     * Add a new rectangle to graph and return the rectangle object to be
     *   stored by the invoking object.  This allows for independent layout
     *   method to layout the objects later.
     */
    static public RoseRTRectangle addRectangle() {
	// see if graph list is empty
	RoseRTRectangle rect = null;
	if (graph.size() == 0) {  // yes, empty
	    // generate a first vertex at the top-left corner (near (0,0))
	    rect = new RoseRTRectangle(RECT_WIDTH/2 + LEFT_MARGIN,
				       RECT_HEIGHT/2,
				       RECT_WIDTH - RECT_GAP,
				       RECT_HEIGHT - RECT_GAP);
	    rowIdx = 1;
	    columnIdx = 0;
	} else {
	    rect = new RoseRTRectangle(columnIdx * RECT_WIDTH + RECT_WIDTH/2 +
				       LEFT_MARGIN,
				       rowIdx * RECT_HEIGHT + RECT_HEIGHT/2,
				       RECT_WIDTH - RECT_GAP,
				       RECT_HEIGHT - RECT_GAP);
	    columnIdx++;
	    if (columnIdx+1 >= MAX_COLUMN) {  // new row
		rowIdx++;
		columnIdx = 0;
	    }
	}
	graph.addLast(rect);
	return rect;
    }

    /**
     * Returns the estimated pixel width of the string label.
     *
     * @return int > 0 pixel width
     */
    static public int getPixelWidth(String str) {
	return (str.length() * FONT_SIZE * 2) + 20;
    }

    /*
     Returns the estimated pixel height of the string label.
     *
     * @return int > 0 pixel height
     */
    static public int getPixelHeight(String str) {
	return FONT_SIZE * 3;
    }

    /**
     * Start the recursive method to layout the class diagram.
     */
    static public void layoutGraph() {
	layoutNode((RoseRTRectangle)graph.get(0));
    }

    /**
     * Recursively traverses connected nodes of graph from the given node
     *   and layout that node and it's links.
     *
     * @param node Node from which to traverse the graph
     */
    static private int spaces = 0;
    static private void layoutNode(RoseRTRectangle node) {
	// output node name
	System.err.println(Util.indent(spaces) + node.getName());
	if (node.traversed) {  // traverse no more
	    return;
	}
	node.traversed = true;
	spaces += 4;
	//RoseRTRectangle link = null;
	RoseRTRectangle sublink = null;
	RoseRTInheritVw inhVw = null;
	RoseRTPoint pt1 = null;
	RoseRTPoint pt2 = null;
	RoseRTPoint origPt = null;
	RoseRTPoint termPt = null;
	RoseRTPoint labelPt = null;
	//RoseRTObject obj = null;
	Enumeration links = node.getLinks().elements();
	Enumeration sublinks = null;
	int linkCnt = node.getLinks().size();
	double avgCnt = (linkCnt + 1) / 2.0;
	while (links.hasMoreElements()) {
	    //obj = (RoseRTObject)links.nextElement();
	    //if (obj instanceof RoseRTRectangle) {
	    //    link = (RoseRTRectangle)obj;
	    //} else if (obj instanceof RoseRTInheritVw) {
	    //}
	    int deltaPos = (int)(RECT_GAP * 2 * (linkCnt-avgCnt));
	    int dPos = deltaPos;
	    linkCnt--;
	    inhVw = (RoseRTInheritVw)links.nextElement();
	    // get links of this connector view object
	    //   assume the connector only links to _one_ other rectangle
	    sublinks = inhVw.getRect().getLinks().elements();
	    if (sublinks.hasMoreElements()) {
		sublink = (RoseRTRectangle)sublinks.nextElement();
		int firstX, firstY, secondX, secondY;
		if (Math.abs(node.getCenter().getY() -
			     sublink.getCenter().getY()) < RECT_GAP) {
		    // treat the two rectangles as horizontally adjacent,
		    //   so connector view object must take a 'U' route
		    if (sublink.getCenter().getX() < node.getCenter().getX() &&
			dPos != -deltaPos) {
			// if the order is not client to the right of supplier
			// reverse the deltaPos, do this only the first time
			dPos = -deltaPos;
		    }
		    firstX = sublink.getCenter().getX() - dPos;
		    firstY = sublink.getCenter().getY() +
			(RECT_HEIGHT - RECT_GAP)/2;
		    secondX = node.getCenter().getX() + dPos;
		    secondY = sublink.getCenter().getY() + RECT_HEIGHT -
			deltaPos/2;
		    // set the originating vertex
		    origPt = new RoseRTPoint(firstX, firstY);
		    // set first corner vertex
		    pt1 = new RoseRTPoint(firstX, secondY);
		    inhVw.getVertices().addLast(pt1);
		    // set second corner vertex
		    pt2 = new RoseRTPoint(secondX, secondY);
		    inhVw.getVertices().addLast(pt2);
		    // set terminating vertex
		    termPt = new RoseRTPoint(secondX, firstY);
		    if (inhVw instanceof RoseRTCapsuleRoleRelVw) {
			// set label coordinate
			labelPt = new RoseRTPoint(pt2.getX(), pt2.getY());
			inhVw.getObjLabel().setPos(getLabelPos(labelPt));
		    }
		} else if (Math.abs(node.getCenter().getX() -
				    sublink.getCenter().getX()) < RECT_GAP) {
		    // treat the two rectangles as vertically adjacent,
		    //   so connector view object must take a '[' route
		    if (sublink.getCenter().getY() < node.getCenter().getY() &&
			dPos != -deltaPos) {
			// if the order is not client below the supplier
			// reverse the deltaPos, do this only the first time
			dPos = -deltaPos;
		    }
		    firstX = sublink.getCenter().getX() -
			(RECT_WIDTH - RECT_GAP)/2;
		    firstY = sublink.getCenter().getY() + dPos;
		    secondX = node.getCenter().getX() - RECT_WIDTH -
			deltaPos/2;
		    secondY = node.getCenter().getY() - dPos;
		    // set the originating vertex
		    origPt = new RoseRTPoint(firstX, firstY);
		    // set first corner vertex
		    pt1 = new RoseRTPoint(secondX, firstY);
		    inhVw.getVertices().addLast(pt1);
		    // set second corner vertex
		    pt2 = new RoseRTPoint(secondX, secondY);
		    inhVw.getVertices().addLast(pt2);
		    // set terminating vertex
		    termPt = new RoseRTPoint(firstX, secondY);
		    if (inhVw instanceof RoseRTCapsuleRoleRelVw) {
			// set label coordinate
			labelPt = new RoseRTPoint(pt2.getX(), pt2.getY());
			inhVw.getObjLabel().setPos(getLabelPos(labelPt));
		    }
		} else {
		    // assume rectangles are diagonal to each other
		    //   so connector view object can take 'L' route
		    firstX = sublink.getCenter().getX() + deltaPos;
		    firstY = sublink.getCenter().getY() -
			(RECT_HEIGHT - RECT_GAP)/2;
		    secondX = node.getCenter().getX() +
			(RECT_WIDTH - RECT_GAP)/2;
		    secondY = node.getCenter().getY() - deltaPos;
		    // set the originating vertex
		    origPt = new RoseRTPoint(firstX, firstY);
		    // set the corner vertex
		    pt1 = new RoseRTPoint(firstX, secondY);
		    inhVw.getVertices().addLast(pt1);
		    // set the terminating vertex
		    termPt = new RoseRTPoint(secondX, secondY);
		    if (inhVw instanceof RoseRTCapsuleRoleRelVw) {
			// set label coordinate
			labelPt = new RoseRTPoint(pt1.getX(), pt1.getY());
			inhVw.getObjLabel().setPos(getLabelPos(labelPt));
		    }
		}
		// add the originating vertex
		inhVw.getVertices().addFirst(origPt);
		inhVw.setOrigPos(origPt);
		// add terminating vertex
		inhVw.getVertices().addLast(termPt);
		inhVw.setTermPos(termPt);

		layoutNode(sublink);
	    }
	}
	spaces -= 4;
    }
}
