package acme.translators.RoseRTTypes;

/**
 * Class: RoseRTProtocol <br>
 * Purpose: Provides a Rose RT protocol object. <br>
 * <br>
 * Revision History: <br>
 *   Date    Person    Change: <br>
 * 5/24/01    SWC      Created initial version <br>
 * <br>
 * Known bugs/omissions:<br>
 *   - <br>
 * @author Owen Cheng (SWC)
 */

public class RoseRTProtocol extends RoseRTObject {

    static private final String OBJ = "Protocol";
    static private final String NAM = "prot";

    //private String stereotype = "protocol";
    private RoseRTList collabs = null;
    private RoseRTList protRoles = null;
    private String language = "C++";
    private boolean isBinary = false;

    // content that is not printed
    private String basePortName = null;
    private String conjPortName = null;

    /**
     * Default class constructor.
     */
    public RoseRTProtocol() {
	super();

	this.setType(OBJ);
	this.setName(NAM);
	this.collabs = new RoseRTList();
	this.collabs.setLabel("collaborations");
	this.collabs.setType("collaborationsList");
	this.protRoles = new RoseRTList();
	this.protRoles.setLabel("protocolRoles");
	this.protRoles.setType("protocolRolesList");
    }

    /**
     * Class constructor for initializing name.
     *
     * @param name the given name of this object
     */
    public RoseRTProtocol(String name) {
	this();

	this.setName(name);
    }

    /**
     * Returns this protocol's collaboration content.
     *
     * @return RoseRTList
     */
    public RoseRTList getCollabs() {
	return this.collabs;
    }

    /**
     * Replaces the list of collaborations of this protocol.
     *
     * @param list new RoseRTList to replace with
     */
    public void setCollab(RoseRTList list) {
	this.collabs = list;
    }

    /**
     * Returns the list of protocol roles of this protocol.
     *
     * @return RoseRTList
     */
    public RoseRTList getProtocolRoles() {
	return this.protRoles;
    }

    /**
     * Replaces the list of protocol roles of this protocol.
     *
     * @param list new RoseRTList to replace with
     */
    public void setProtocolRoles(RoseRTList list) {
	this.protRoles = list;
    }

    /**
     * Returns the base port type name of this protocol.
     *
     * @return String
     */
    public String getBasePortName() {
	return this.basePortName;
    }

    /**
     * Replaces the base port type name of this protocol.
     *
     * @param name new port name to replace with
     */
    public void setBasePortName(String name) {
	this.basePortName = name;
    }

    /**
     * Returns the conjugate port type name of this protocol.
     *
     * @return String
     */
    public String getConjugatePortName() {
	return this.conjPortName;
    }

    /**
     * Replaces the conjugate port type name of this protocol.
     *
     * @param name new port name to replace with
     */
    public void setConjugatePortName(String name) {
	this.conjPortName = name;
    }

    /**
     * Returns whether this protocol is binary.
     * The only way this flag can be true is by calling makeBinary().
     *
     * @return boolean
     */
    public boolean isBinary() {
	return this.isBinary;
    }
    
    /**
     * Makes this protocol a binary protocol, with a base protocol role
     * (from which the conjugate role can be derived by conjugation).
     */
    public void makeBinary() {
	RoseRTObject rtobj = new RoseRTObject(null, "Collaboration",
					      "Collaboration1", 0);
	this.collabs.addLast(rtobj);
	rtobj = new RoseRTObject(null, "ProtocolRole", "base", 0);
	this.protRoles.addLast(rtobj);
	this.isBinary = true;
    }

    /**
     * Returns the string representation of this object, basically generating
     * all the content.
     *
     * @param leftMarg amount of spaces to indent for all lines
     * @param indIncr amount of increment spaces to indent for subsequent lines
     * @return String
     */
    public String toString(int leftMarg, int indIncr) {
	String str = "";
	int ind = leftMarg + indIncr;
	String clbStr = collabs.toString(ind, indIncr);
	String prStr = protRoles.toString(ind, indIncr);
	str += Util.indent(leftMarg) + "(object " + this.getType()
	    + " \"" + this.getName() + "\"" +
	    //"\n" + Util.indent(ind) + "stereotype \"" + stereotype + "\"" +
	    (clbStr.equals("") ? "" : ("\n" + clbStr)) +
	    (prStr.equals("") ? "" : ("\n" + prStr)) +
	    "\n" + Util.indent(ind) + "language \"" + language + "\"" +
	    ")";

	return str;
    }

    /**
     * Returns the string representation of this object, basically generating
     * all the content.
     *
     * @return String
     */
    public String toString() {
	return this.toString(0, 8);
    }
}
