package acme.translators;
import acmeLib.*;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.core.container.*;
import acmeLib.representations.*;
import java.util.*;
import java.io.*;

import acmeLib.core.io.*;
/**
 * Class: XAcmeExporter<br>
 * Purpose: Takes an AcmeLib structure and converts it to the <br>
 *          xAcme XML format <br>
 * Revision History:<br>
 *   Date    Person    Change:<br>
 * 5/3/01     BRS       Created initial version <br>
 *
 *
 * Known bugs/omissions:<br>
 * - Should create a DOM structure and then serialize it
 * @author Bradley Schmerl
 */

public class XAcmeExporter extends XMLExporter {

	private acmeLib.core.io.AcmeOutputStream out = null;
	private boolean headerOutput = false;
	private boolean includeVisProperties = true;
	private Map idMap = null;
	private Map pidMap = null;
	private static final int version = 1;

	public XAcmeExporter(AcmeOutputStream s) {
		out = s;
		idMap = new HashMap();
		pidMap = new HashMap();
		tidMap = new HashMap();
	}

	private void outputHeader() throws java.io.IOException {
		if (!headerOutput) {
			headerOutput = true;
			out.writeln(
				" xmlns='http://www.ics.uci.edu/pub/arch/xArch/instance.xsd'");
			out.indent();
			out.indent();
			out.writeln(
				"xmlns:xsi='http://www.w3.org/1999/XMLSchema/instance'");
			out.writeln(
				"xmlns:instance='http://www.ics.uci.edu/pub/arch/xArch/instance.xsd'");
			out.writeln(
				"xmlns:properties='http://www.cs.cmu.edu/~acme/pub/xAcme/properties.xsd'");
			out.writeln(
				"xmlns:acmeProperties='http://www.cs.cmu.edu/~acme/pub/xAcme/acmeProperties.xsd'");
			out.writeln(
				"xmlns:families='http://www.cs.cmu.edu/~acme/pub/xAcme/families.xsd'");
			out.writeln(
				"xmlns:constraints='http://www.cs.cmu.edu/~acme/pub/xAcme/constraints.xsd'");
			out.writeln("xmlns:xlink='http://www.w3.org/1999/xlink'");
			out.writeln(
				"xsi:schemaLocation='http://www.ics.uci.edu/pub/arch/xArch/instance.xsd instance.xsd'");
			out.unIndent();
			out.unIndent();
		}
	}

	public void writeToFile(String path) {
		try {
			File f = new File(path);
			FileOutputStream fo = new FileOutputStream(f);
			out.writeTo(fo);
			fo.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void exportDesign(AcmeDesign designACME)
		throws java.io.IOException {
		out.writeln("<?xml version=\"1.0\"?>");
		out.writeln(
			"<!-- Automatically generated by XAcmeExporter "
				+ version
				+ " using xAcme version 1.0 -->");
		out.write("<!-- Generated ");
		GregorianCalendar cal = new GregorianCalendar(TimeZone.getDefault());
		out.writeln(
			cal.get(Calendar.MONTH)
				+ "/"
				+ cal.get(Calendar.DAY_OF_MONTH)
				+ "/"
				+ cal.get(Calendar.YEAR)
				+ " -->");

		out.write("<typedDesign ");
		outputHeader();
		out.writeln(">");

		exportDesignInternals (designACME);
		out.writeln ("</typedDesign>");
		out.close ();
		}
		
	protected void exportDesignInternals (AcmeDesign designACME) throws IOException {
		exportImports(designACME.getImportedDesigns().enumerate());
		exportTypespace(designACME.getTypes());
		exportTypespace(designACME.getFamilies());

		AcmeEnumeration systemsACME = designACME.getSystems();
		while (systemsACME.moreItems()) {
			AcmeSystem systemACME = (AcmeSystem) systemsACME.getNextItem();
			exportSystem(systemACME, "archInstance", true, true);
		}


	}

	/**
	* @param enumeration
	*/
	private void exportImports(AcmeEnumeration importedDesigns)
		throws IOException {
		while (importedDesigns.moreItems()) {
			AcmeImportedDesignRecord rec =
				(AcmeImportedDesignRecord) importedDesigns.getNextItem();
			if (familyAction == AcmeTranslator.REFERENCE_IMPORTED_FAMILIES) {
			
				out.write("<importDesign type='resolve-on-import' name=\"");
				out.write(rec.getDesignFilename());
				out.writeln("\"/>");
			}
			else if (familyAction == AcmeTranslator.INLINE_IMPORTED_FAMILIES) {
				exportDesignInternals (rec.getResolvedDesign());	
			}
			else if (familyAction == AcmeTranslator.EXPORT_IMPORTED_FAMILIES) {
				out.write("<importDesign type='xml' name=\"");
				out.write (convertFilenameToXML(rec.getCanonicalFilename()));
				out.writeln ("\"/>");
			}
		}

	}

	private void exportTypespace(AcmeEnumeration typespace)
		throws java.io.IOException {
		typespace.sortByCreationOrder();
		while (typespace.moreItems()) {
			AcmeType typeACME = (AcmeType) typespace.getNextItem();
			if (typeACME instanceof AcmePropertyType) {
				exportPropertyTypeDecl((AcmePropertyType) typeACME);
			}
		}

		typespace.reset();
		while (typespace.moreItems()) {
			AcmeType typeACME = (AcmeType) typespace.getNextItem();
			if (typeACME instanceof AcmePortType) {
				exportPortType((AcmePortType) typeACME);
			} else if (typeACME instanceof AcmeRoleType) {
				exportRoleType((AcmeRoleType) typeACME);
			}
		}

		typespace.reset();
		while (typespace.moreItems()) {
			AcmeType typeACME = (AcmeType) typespace.getNextItem();
			if (typeACME instanceof AcmeComponentType) {
				exportComponentType((AcmeComponentType) typeACME);
			}
		}

		typespace.reset();
		while (typespace.moreItems()) {
			AcmeType typeACME = (AcmeType) typespace.getNextItem();
			if (typeACME instanceof AcmeConnectorType) {
				exportConnectorType((AcmeConnectorType) typeACME);
			}
		}

		typespace.reset();
		while (typespace.moreItems()) {
			AcmeType typeACME = (AcmeType) typespace.getNextItem();
			if (typeACME instanceof AcmeFamily) {
				exportFamily((AcmeFamily) typeACME);
			}
		}
	}

	private void exportPropertyTypeDecl(AcmePropertyType propertyTypeACME)
		throws java.io.IOException {
		out.write("<propertyType xsi:type='acmeProperties:PropertyType' ");
		out.write("name='");
		out.write(generateTypeID(propertyTypeACME));
		out.writeln("'>");
		exportPropertyType(propertyTypeACME, false);
		out.writeln("</propertyType>");
	}

	private void exportPropertyType(
		AcmePropertyType propertyTypeACME,
		boolean displayTag)
		throws java.io.IOException {
		if (displayTag) {
			out.writeln(
				"<propertyType xsi:type='acmeProperties:PropertyType'>");
			out.indent();
		}

		if (propertyTypeACME instanceof acmeLib.properties.RecordType) {
			out.writeln("<propertyRecordType>");
			out.indent();
			acmeLib.properties.RecordType recordACME =
				(acmeLib.properties.RecordType) propertyTypeACME;
			AcmeEnumeration fieldsACME = recordACME.getFields();
			while (fieldsACME.moreItems()) {
				AcmeRecordField fieldACME =
					(AcmeRecordField) fieldsACME.getNextItem();
				out.write("<propertyFieldDeclaration name='");
				out.write(fieldACME.getName());
				out.writeln("'>");
				out.indent();
				AcmePropertyType fieldTypeACME = fieldACME.getType();
				exportPropertyType(fieldTypeACME, true);
				out.unIndent();
				out.writeln("</propertyFieldDeclaration>");
			}
			out.unIndent();
			out.writeln("</propertyRecordType>");
		} else if (propertyTypeACME instanceof acmeLib.properties.SetType) {
			out.writeln("<propertySetType>");
			out.indent();
			acmeLib.properties.SetType setTypeACME =
				(acmeLib.properties.SetType) propertyTypeACME;
			AcmePropertyType innerTypeACME =
				(AcmePropertyType) setTypeACME.getSetType();
			exportPropertyType(innerTypeACME, true);
			out.unIndent();
			out.writeln("</propertySetType>");
		} else if (
			propertyTypeACME instanceof acmeLib.properties.SequenceType) {
			out.writeln("<propertySequenceType>");
			out.indent();
			acmeLib.properties.SequenceType seqTypeACME =
				(acmeLib.properties.SequenceType) propertyTypeACME;
			AcmePropertyType innerTypeACME = seqTypeACME.getSequenceType();
			exportPropertyType(innerTypeACME, true);
			out.unIndent();
			out.writeln("</propertySequenceType>");
		} else if (propertyTypeACME instanceof EnumType) {
			out.writeln("<propertyEnumType>");
			out.indent();
			EnumType enumTypeACME = (EnumType) propertyTypeACME;
			AcmeEnumeration valuesACME = enumTypeACME.getValues();
			while (valuesACME.moreItems()) {
				EnumValue valueACME = (EnumValue) valuesACME.getNextItem();
				out.write("<value>");
				out.write(valueACME.value);
				out.writeln("</value>");
			}
			out.unIndent();
			out.writeln("</propertyEnumType>");
		} else if (propertyTypeACME instanceof StringType) {
			out.writeln("<propertyPrimitiveType type='string'/>");
		} else if (propertyTypeACME instanceof IntType) {
			out.writeln("<propertyPrimitiveType type='int'/>");

		} else if (propertyTypeACME instanceof FloatType) {
			out.writeln("<propertyPrimitiveType type='float'/>");

		} else if (propertyTypeACME instanceof BooleanType) {
			out.writeln("<propertyPrimitiveType type='boolean'/>");
		}

		if (displayTag) {
			out.unIndent();
			out.writeln("</propertyType>");
		}
	}

	private void exportSuperTypes(AcmeElementType elementTypeACME, String tag)
		throws java.io.IOException {
		int nTypes = elementTypeACME.getSuperTypes().getSize();
		if (nTypes != 0) {
			AcmeEnumeration supersACME =
				elementTypeACME.getSuperTypes().enumerate(
					new acmeLib.core.filters.NoBaseClassesEnumFilter());
			while (supersACME.moreItems()) {
				AcmeElementType superACME =
					(AcmeElementType) supersACME.getNextItem();
				out.writeln(
					"<"
						+ tag
						+ " xlink:type='simple' xlink:href='" 
						+ getTypeReference ((AcmeElement )elementTypeACME, (AcmeObject )superACME)
						+ "'/>");
			}
		}
	}

	private void exportTypes(AcmeElement elementACME, String tag)
		throws java.io.IOException {
		int nTypes = elementACME.getDeclaredTypes().getSize();
		if (nTypes != 0) {
			AcmeEnumeration supersACME = elementACME.getDeclaredTypes();
			while (supersACME.moreItems()) {
				AcmeElementType superACME =
					(AcmeElementType) supersACME.getNextItem();
				
					out.writeln ("<" + tag + " xlink:type='simple' xlink:href='" + getTypeReference(elementACME, (AcmeObject )superACME) + "'/>");
					
				
			}
		}
	}

	public void exportComponent(AcmeComponent componentACME)
		throws java.io.IOException {
		exportComponent(componentACME, "componentInstance", true);
	}

	private void exportComponentType(AcmeComponentType componentTypeACME)
		throws java.io.IOException {
		out.writeln(
			"<componentType id='" + generateTypeID(componentTypeACME) + "'>");
		out.indent();
		exportSuperTypes(componentTypeACME, "superType");
		exportComponent(componentTypeACME, "structure", false);
		out.unIndent();
		out.writeln("</componentType>");
	}

	public void exportConnector(AcmeConnector connectorACME)
		throws java.io.IOException {
		exportConnector(connectorACME, "connectorInstance", true);
	}

	private void exportConnectorType(AcmeConnectorType connectorTypeACME)
		throws java.io.IOException {
		out.writeln(
			"<connectorType id='" + generateTypeID(connectorTypeACME) + "'>");
		out.indent();
		exportSuperTypes(connectorTypeACME, "superType");
		exportConnector(connectorTypeACME, "structure", false);
		out.unIndent();
		out.writeln("</connectorType>");
	}

	public void exportPort(AcmePort portACME) throws java.io.IOException {
		exportPort(portACME, "interfaceInstance", true);
	}

	private void exportPortType(AcmePortType portTypeACME)
		throws java.io.IOException {
		out.writeln(
			"<interfaceType id='" + generateTypeID(portTypeACME) + "'>");
		out.indent();
		exportSuperTypes(portTypeACME, "superType");
		exportPort(portTypeACME, "structure", false);
		out.unIndent();
		out.writeln("</interfaceType>");
	}

	public void exportRole(AcmeRole roleACME) throws java.io.IOException {
		exportRole(roleACME, "interfaceInstance", true);
	}

	private void exportRoleType(AcmeRoleType roleTypeACME)
		throws java.io.IOException {
		out.writeln(
			"<interfaceType id='" + generateTypeID(roleTypeACME) + "'>");
		out.indent();
		exportSuperTypes(roleTypeACME, "superType");
		exportRole(roleTypeACME, "structure", false);
		out.unIndent();
		out.writeln("</interfaceType>");
	}

	private void exportFamily(AcmeFamily familyACME)
		throws java.io.IOException {
		out.writeln("<family id='" + generateTypeID(familyACME) + "'>");
		out.indent();
		exportSuperTypes(familyACME, "superType");
		exportTypespace(familyACME.getTypes());
		exportSystem(familyACME, "structure", true, false);
		out.unIndent();
		out.writeln("</family>");
	}

	private String generateID(AcmeElement elementACME) {
		String id = "";
		if (elementACME != null)
			id = generateFlatName(elementACME);
		id += "::" + nextId++;
		idMap.put(elementACME, id);
		return id;
	}

	private String generatePropertyID(AcmeProperty propertyACME) {
		String id = propertyACME.getName() + "::" + nextId++;
		pidMap.put(propertyACME, id);
		return id;
	}

	private void exportSystem(
		AcmeSystem systemACME,
		String tag,
		boolean close,
		boolean doSupers)
		throws java.io.IOException {
		out.write("<" + tag + " id='");
		if (tag.equals("family")) {
			out.write(generateTypeID((AcmeFamily) systemACME) + "'");
		} else {
			out.write(generateID(systemACME) + "'");
		}
		if (doSupers && systemACME.getDeclaredTypes().getSize() > 0) {
			out.write(" xsi:type='families:TypedArchInstance'");
		} else if (
			!tag.equals("family")
				&& systemACME.getProperties().getSize() > 0) {
			out.write(" xsi:type='properties:ArchInstance'");
		}
		out.writeln(">");
		out.indent();

		AcmeEnumeration componentsACME = systemACME.getComponents();
		while (componentsACME.moreItems()) {
			exportComponent(
				(AcmeComponent) componentsACME.getNextItem(),
				"componentInstance",
				true);
		}

		AcmeEnumeration connectorsACME = systemACME.getConnectors();
		while (connectorsACME.moreItems()) {
			exportConnector(
				(AcmeConnector) connectorsACME.getNextItem(),
				"connectorInstance",
				true);
		}

		AcmeEnumeration attachmentsACME = systemACME.getAttachments();
		int numAttachments = 0;
		while (attachmentsACME.moreItems()) {
			exportAttachment((AcmeAttachment) attachmentsACME.getNextItem());
		}

		if (doSupers) {
			this.exportTypes(systemACME, "type");
		}

		if (close) {
			out.unIndent();
			out.writeln("</" + tag + ">");
		}
	}

	private void exportComponent(
		AcmeComponent componentACME,
		String tag,
		boolean determineType)
		throws java.io.IOException {
		out.write("<" + tag + " id='" + generateID(componentACME) + "'");
		if (determineType && hasTypes(componentACME)) {
			out.write(" xsi:type='families:TypedComponentInstance'");
		} else if (
			determineType
				&& (componentACME.getProperties().getSize() > 0
					|| componentACME.getRepresentations().getSize() > 0)) {
			out.write(" xsi:type='properties:ComponentInstance'");
		}
		out.writeln(">");
		out.indent();
		out.writeln("<description/>");

		AcmeEnumeration portsACME = componentACME.getPorts();
		while (portsACME.moreItems()) {
			AcmePort portACME = (AcmePort) portsACME.getNextItem();
			exportPort(portACME, "interfaceInstance", true);
		}

		AcmeEnumeration representationsACME =
			componentACME.getRepresentations();
		while (representationsACME.moreItems()) {
			AcmeRepresentation representationACME =
				(AcmeRepresentation) representationsACME.getNextItem();
			exportRepresentation(representationACME);
		}

		AcmeEnumeration propertiesACME = componentACME.getProperties();
		while (propertiesACME.moreItems()) {
			AcmeProperty propertyACME =
				(AcmeProperty) propertiesACME.getNextItem();
			exportProperty(propertyACME);
		}

		if (hasTypes(componentACME)) {
			exportTypes(componentACME, "type");
			out.writeln(
				"<instanceOf xlink:type='simple' xlink:href='"
					+ getTypeReference (componentACME, componentACME.getInstantiatedTypes().getNextItem())
					+ "'/>");
		}

		out.unIndent();
		out.writeln("</" + tag + ">");
	}

	private void exportConnector(
		AcmeConnector connectorACME,
		String tag,
		boolean determineType)
		throws java.io.IOException {
		out.write("<" + tag + " id='" + generateID(connectorACME) + "'");
		if (determineType && hasTypes(connectorACME)) {
			out.write(" xsi:type='families:TypedConnectorInstance'");
		} else if (
			determineType
				&& (connectorACME.getProperties().getSize() > 0
					|| connectorACME.getRepresentations().getSize() > 0)) {
			out.write(" xsi:type='properties:ConnectorInstance'");
		}
		out.writeln(">");
		out.indent();
		out.writeln("<description/>");

		AcmeEnumeration rolesACME = connectorACME.getRoles();
		while (rolesACME.moreItems()) {
			AcmeRole roleACME = (AcmeRole) rolesACME.getNextItem();
			exportRole(roleACME, "interfaceInstance", true);
		}

		AcmeEnumeration representationsACME =
			connectorACME.getRepresentations();
		while (representationsACME.moreItems()) {
			AcmeRepresentation representationACME =
				(AcmeRepresentation) representationsACME.getNextItem();
			exportRepresentation(representationACME);
		}

		AcmeEnumeration propertiesACME = connectorACME.getProperties();
		while (propertiesACME.moreItems()) {
			AcmeProperty propertyACME =
				(AcmeProperty) propertiesACME.getNextItem();
			exportProperty(propertyACME);
		}

		if (hasTypes(connectorACME)) {
			exportTypes(connectorACME, "type");
			out.writeln(
				"<instanceOf xlink:type='simple' xlink:href='"
					+ getTypeReference (connectorACME, connectorACME.getInstantiatedTypes().getNextItem())
					+ "'/>");
		}

		out.unIndent();
		out.writeln("</" + tag + ">");
	}

	private void exportPort(
		AcmePort portACME,
		String tag,
		boolean determineType)
		throws java.io.IOException {
		out.write("<" + tag + " id='" + generateID(portACME) + "'");
		if (determineType && hasTypes(portACME)) {
			out.write(" xsi:type='families:TypedInterfaceInstance'");
		} else if (determineType && portACME.getProperties().getSize() > 0) {
			out.write(" xsi:type='properties:InterfaceInstance'");
		}
		out.writeln(">");
		out.indent();
		out.writeln("<description/>");

		AcmeEnumeration propertiesACME = portACME.getProperties();
		while (propertiesACME.moreItems()) {
			AcmeProperty propertyACME =
				(AcmeProperty) propertiesACME.getNextItem();
			exportProperty(propertyACME);
		}

		if (hasTypes(portACME)) {
			exportTypes(portACME, "type");
			out.writeln(
				"<instanceOf xlink:type='simple' xlink:href='"
					+ getTypeReference (portACME, portACME.getInstantiatedTypes().getNextItem())
					+ "'/>");
		}

		out.unIndent();
		out.writeln("</" + tag + ">");
	}

	private void exportRole(
		AcmeRole roleACME,
		String tag,
		boolean determineType)
		throws java.io.IOException {
		out.write("<" + tag + " id='" + generateID(roleACME) + "'");
		if (determineType && hasTypes(roleACME)) {
			out.write(" xsi:type='families:TypedInterfaceInstance'");
		} else if (determineType && roleACME.getProperties().getSize() > 0) {
			out.write(" xsi:type='properties:InterfaceInstance'");
		}
		out.writeln(">");
		out.indent();
		out.writeln("<description/>");

		AcmeEnumeration propertiesACME = roleACME.getProperties();
		while (propertiesACME.moreItems()) {
			AcmeProperty propertyACME =
				(AcmeProperty) propertiesACME.getNextItem();
			exportProperty(propertyACME);
		}

		if (hasTypes(roleACME)) {
			exportTypes(roleACME, "type");
			out.writeln(
				"<instanceOf xlink:type='simple' xlink:href='"
					+ getTypeReference (roleACME, roleACME.getInstantiatedTypes().getNextItem())
					+ "'/>");
		}

		out.unIndent();
		out.writeln("</" + tag + ">");
	}

	public void exportProperty(AcmeProperty propertyACME)
		throws java.io.IOException {
		if (includeProperty(propertyACME)) {
			out.writeln(
				"<property id='" + generatePropertyID(propertyACME) + "'>");
			out.indent();
			AcmePropertyValue valueACME = propertyACME.getValue();
			if (valueACME != null) {
				exportPropertyValue(valueACME);
			}
			exportPropertyType(propertyACME.getType(), true);
			out.unIndent();
			out.writeln("</property>");
		}
	}

	private void exportPropertyValue(AcmePropertyValue valueACME)
		throws java.io.IOException {
		out.writeln("<value xsi:type='acmeProperties:PropertyValue'>");
		out.indent();
		if (valueACME instanceof AcmeString) {
			AcmeString stringValACME = (AcmeString) valueACME;
			out.writeln("<propertyPrimitiveValue>");
			out.indent();
			out.writeln(
				"<stringValue>" + stringValACME.getValue() + "</stringValue>");
			out.unIndent();
			out.writeln("</propertyPrimitiveValue>");
		} else if (valueACME instanceof AcmeInt) {
			AcmeInt intValACME = (AcmeInt) valueACME;
			out.writeln("<propertyPrimitiveValue>");
			out.indent();
			out.writeln("<intValue>" + intValACME.getValue() + "</intValue>");
			out.unIndent();
			out.writeln("</propertyPrimitiveValue>");

		} else if (valueACME instanceof AcmeFloat) {
			AcmeFloat floatValACME = (AcmeFloat) valueACME;
			out.writeln("<propertyPrimitiveValue>");
			out.indent();
			out.writeln(
				"<floatValue>" + floatValACME.getValue() + "</floatValue>");
			out.unIndent();
			out.writeln("</propertyPrimitiveValue>");

		} else if (valueACME instanceof AcmeBoolean) {
			AcmeBoolean booleanValACME = (AcmeBoolean) valueACME;
			out.writeln("<propertyPrimitiveValue>");
			out.indent();
			out.writeln(
				"<booleanValue>"
					+ booleanValACME.getValue()
					+ "</booleanValue>");
			out.unIndent();
			out.writeln("</propertyPrimitiveValue>");

		} else if (valueACME instanceof AcmeSet) {
			AcmeSet setValACME = (AcmeSet) valueACME;
			out.writeln("<propertySetValue>");
			out.indent();
			AcmeEnumeration valuesACME = setValACME.getValues();
			while (valuesACME.moreItems()) {
				AcmePropertyValue vACME =
					(AcmePropertyValue) valuesACME.getNextItem();
				exportPropertyValue(vACME);
			}
			out.unIndent();
			out.writeln("</propertySetValue>");
		} else if (valueACME instanceof AcmeSequence) {
			AcmeSequence sequenceValACME = (AcmeSequence) valueACME;
			out.writeln("<propertySequenceValue>");
			out.indent();
			AcmeEnumeration valuesACME = sequenceValACME.getValues();
			while (valuesACME.moreItems()) {
				AcmePropertyValue vACME =
					(AcmePropertyValue) valuesACME.getNextItem();
				exportPropertyValue(vACME);
			}
			out.unIndent();
			out.writeln("</propertySequenceValue>");

		} else if (valueACME instanceof AcmeEnum) {

		} else if (valueACME instanceof AcmeRecord) {
			AcmeRecord recACME = (AcmeRecord) valueACME;
			out.writeln("<propertyRecordValue>");
			out.indent();
			AcmeEnumeration fieldsACME = recACME.getFields();
			while (fieldsACME.moreItems()) {
				AcmeRecordField fieldACME =
					(AcmeRecordField) fieldsACME.getNextItem();
				out.writeln(
					"<propertyFieldValue fieldName='"
						+ fieldACME.getName()
						+ "'>");
				out.indent();
				exportPropertyValue(fieldACME.getValue());
				out.unIndent();
				out.writeln("</propertyFieldValue>");
			}
			out.unIndent();
			out.writeln("</propertyRecordValue>");
		}
		out.unIndent();
		out.writeln("</value>");
	}

	public void exportAttachment(AcmeAttachment attACME)
		throws java.io.IOException {
		out.writeln("<linkInstance id='" + generateID(null) + "'>");
		out.indent();
		out.writeln("<description/>");
		out.writeln("<point>");
		out.indent();
		out.writeln(
			"<anchorInterface xlink:type='simple' xlink:href='#"
				+ idMap.get(attACME.getPort())
				+ "'/>");
		out.unIndent();
		out.writeln("</point>");
		out.writeln("<point>");
		out.indent();
		out.writeln(
			"<anchorInterface xlink:type='simple' xlink:href='#"
				+ idMap.get(attACME.getRole())
				+ "'/>");
		out.unIndent();
		out.writeln("</point>");
		out.unIndent();
		out.writeln("</linkInstance>");
	}

	private boolean hasTypes(AcmeElement elementACME) {
		return elementACME.getDeclaredTypes().getSize() > 0
			|| elementACME.getInstantiatedTypes().getSize() > 0;
	}

	private String generateFlatName(AcmeElement elementACME) {
		return elementACME.getQualifiedName();
	}

	private void exportRepresentation(AcmeRepresentation repACME)
		throws java.io.IOException {
		out.writeln("<subArchitecture>");
		out.indent();
		exportSystem(repACME.getSystem(), "archInstance", true, true);

		AcmeEnumeration bindingsACME =
			((AcmeBindingsMap) repACME.getAbstractionMap()).getBindings();
		while (bindingsACME.moreItems()) {
			AcmeBinding bindingACME = (AcmeBinding) bindingsACME.getNextItem();
			AcmeElement innerACME = bindingACME.getInner();
			AcmeElement outerACME = bindingACME.getOuter();
			out.writeln("<interfaceInstanceMapping>");
			out.indent();
			out.writeln(
				"<outerInterfaceInstance xlink:type='simple' "
					+ "xlink:href='#"
					+ idMap.get(outerACME)
					+ "'/>");
			out.writeln(
				"<innerInterfaceInstance xlink:type='simple' "
					+ "xlink:href='#"
					+ idMap.get(innerACME)
					+ "'/>");
			out.unIndent();
			out.writeln("</interfaceInstanceMapping>");
		}
		out.unIndent();
		out.writeln("</subArchitecture>");
	}

	private boolean includeProperty(AcmeProperty property) {
		boolean inc = true;

		return inc;
		/*	if (!includeVisProperties) {
		    String name = property.getName ();
		    if (name.length () >= 4 && name.substring (0,4).equals ("vis-")){
			inc = false;
		    }
		}
		return inc;*/
	}

}
