package acme.translators;
import acmeLib.*;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.core.container.*;
import java.util.*;
import org.w3c.dom.*;
import org.w3c.dom.Element;

/** 
 * Class: XAcmeImporter<br>
 * Purpose: Creates an AcmeLib structure for an architecture written in <br>
 *          the xAcme XML format. <br>
 * Revision History:<br>
 *   Date   Person     Change<br>
 * 5/2/01    BRS       Created inital version<br>
 * 6/15/01   BRS       Made it inherit from XMLImporter<br>
 *
 *
 * Known bugs/omissions:<br>
 * - Does not yet support the Armani constraints importation.<br>
 * - Need to integrate error reporting with AcmeLib AcmeError for<br>
 *    reporting back to AcmeStudio<br>
 * - Still assumes the the xml document is self-contained
 * @author Bradley Schmerl
 */

public class XAcmeImporter extends XMLImporter {

	private Map _xmlToImportedDesignMap = new HashMap ();
	public static String FAMILY_NS = "http://www.cs.cmu.edu/~acme/pub/xAcme/families.xsd";
	// The top level nodes of both the XML DOM tree and the AcmeLib design
	protected Document xAcmeDoc;
	protected acmeLib.elements.AcmeDesign design;

	// Tables to convert from an XML id to DOM, AcmeLib or Interfaces
	protected Hashtable nameToAcme;
	protected Hashtable nameToDom;
	protected Hashtable nameToInterfaceType;

	// Class to store whether an interfaceType in xAcme has been resolved
	// to a port type or a role type in the translation
	private class RoleOrPortType {
		// Can only know if an interface type is being used as
		// a role or port type when it is used
		// This class allows us to defer this decision
		public static final int UNKNOWN = 0;
		public static final int ROLE = 1;
		public static final int PORT = 2;

		public int type = 0;
		public org.w3c.dom.Element node = null;
		public AcmeElement context = null;
	}

	private static final String PROPERTYTYPE_TAG = "propertyType";
	private static final String COMPONENTTYPE_TAG = "componentType";
	private static final String CONNECTORTYPE_TAG = "connectorType";
	private static final String INTERFACETYPE_TAG = "interfaceType";
	private static final String FAMILY_TAG = "family";
	private static final String SYSTEM_TAG = "archInstance";
	private static final String IMPORT_TAG = "importDesign";

	/**
	 * The constructor for this class
	 * @param doc Document - the parsed DOM document to translate
	 */
	public XAcmeImporter(Document doc) {
		xAcmeDoc = doc;
		design = new acmeLib.elements.AcmeDesign();
		nameToDom = new Hashtable();
		nameToAcme = new Hashtable();
		nameToInterfaceType = new Hashtable();

	}

	/**
	 * imports the design from the DOM to AcmeLib
	 */
	public acmeLib.elements.AcmeDesign importDesign()
		throws acmeLib.core.exceptions.AcmeIllegalNameException {

		NodeList importedDesigns =
			getElementsAtThisLevel(xAcmeDoc.getDocumentElement(), IMPORT_TAG);
		// Get all the property types in the global scope
		NodeList propertyTypes =
			getElementsAtThisLevel(
				xAcmeDoc.getDocumentElement(),
				PROPERTYTYPE_TAG);
		// Get all the component types in the global scope
		NodeList componentTypes =
			getElementsAtThisLevel(
				xAcmeDoc.getDocumentElement(),
				COMPONENTTYPE_TAG);
		// Get all the connector types in the global scope
		NodeList connectorTypes =
			getElementsAtThisLevel(
				xAcmeDoc.getDocumentElement(),
				CONNECTORTYPE_TAG);
		// Get all the interface types in the global scope
		NodeList interfaceTypes =
			getElementsAtThisLevel(
				xAcmeDoc.getDocumentElement(),
				INTERFACETYPE_TAG);
		// Get all the families
		NodeList families =
			getElementsAtThisLevel(xAcmeDoc.getDocumentElement(), FAMILY_TAG);
		// Get all the systems
		NodeList systems =
			getElementsAtThisLevel(xAcmeDoc.getDocumentElement(), SYSTEM_TAG);

		for (int i = 0; i < importedDesigns.getLength(); i++) {
			Element importedDes = (Element) importedDesigns.item(i);
			importedDes.normalize();
			String importFile =
				this.processImportedDesign(importedDes);
			if (importFile != null) {
				design.addDesignToImport(importFile);
			}
		}

		int i;

		// Process the global property types
		for (i = 0; i < propertyTypes.getLength(); i++) {
			Element propertyType = (Element) propertyTypes.item(i);
			propertyType.normalize();
			AcmeType type = this.processPropertyType(propertyType);
			if (type != null) {
				design.addType(type);
				nameToAcme.put(propertyType.getAttribute("id"), type);
				nameToDom.put(propertyType.getAttribute("id"), propertyType);
			}
		}

		// Process the global component types
		for (i = 0; i < componentTypes.getLength(); i++) {
			Element componentType = (Element) componentTypes.item(i);
			componentType.normalize();
			AcmeType type = this.processComponentType(componentType);
			if (type != null) {
				design.addType(type);
				nameToAcme.put(componentType.getAttribute("id"), type);
				nameToDom.put(componentType.getAttribute("id"), componentType);

			}
		}

		// Process the gloval connector types
		for (i = 0; i < connectorTypes.getLength(); i++) {
			Element connectorType = (Element) connectorTypes.item(i);
			connectorType.normalize();
			AcmeType type = this.processConnectorType(connectorType);
			if (type != null) {
				design.addType(type);
				nameToAcme.put(connectorType.getAttribute("id"), type);
				nameToDom.put(connectorType.getAttribute("id"), connectorType);
			}
		}

		// Process the global interface types
		for (i = 0; i < interfaceTypes.getLength(); i++) {
			Element interfaceType = (Element) interfaceTypes.item(i);
			interfaceType.normalize();

			// Defer processing until we know whether it is a role
			// or a port. 
			RoleOrPortType interfce = new RoleOrPortType();
			interfce.context = design;
			interfce.node = interfaceType;
			nameToInterfaceType.put(interfaceType.getAttribute("id"), interfce);
		}

		// Process the families
		for (i = 0; i < families.getLength(); i++) {
			Element family = (Element) families.item(i);
			family.normalize();
			AcmeFamily fam = this.processFamily(family);

			design.addFamily(fam);
			nameToAcme.put(family.getAttribute("id"), fam);
			nameToDom.put(family.getAttribute("id"), fam);
		}

		// Process the systems
		for (i = 0; i < systems.getLength(); i++) {
			Element systemDOM = (Element) systems.item(i);
			systemDOM.normalize();
			AcmeSystem systemACME = this.processSystem(systemDOM);
			design.addSystem(systemACME);
			nameToAcme.put(systemDOM.getAttribute("id"), systemACME);
			nameToDom.put(systemDOM.getAttribute("id"), systemACME);
		}

		return design;

	}

	/**
	 * @param importedDes
	 * @return
	 */
	private String processImportedDesign(Element importedDes) {
		String filename = importedDes.getAttribute("name");
		String xmlAtt = importedDes.getAttribute ("type");
		if (xmlAtt != null && xmlAtt.equals ("xml")) {
			filename = filename.substring (0, filename.lastIndexOf("."));
		}
		return filename;
	}

	/**
	 * return the Acme identifier of a fully qualified unique
	 *   xml identifier
	 *     Assumed xml naming convention is:
	 *        something.something.acmeName::number
	 *     this function returns acmeName
	 * @param complexName String
	 * @return the Acme name
	 */
	String decomposeName(String complexName) {
		// Returns the Acme name of a fully qualified Acme name
		String afterLastDot;
		if (complexName.lastIndexOf('.') == -1) {
			afterLastDot = complexName;
		} else {
			afterLastDot =
				complexName.substring(
					complexName.lastIndexOf('.') + 1,
					complexName.length());
		}
		String upToColon;
		if (afterLastDot.indexOf(':') == -1) {
			upToColon = afterLastDot;
		} else {
			upToColon = afterLastDot.substring(0, afterLastDot.indexOf(':'));
		}
		return upToColon;
	}

	/**
	 * Strips the hash from the beginning of a name
	 * @param nameWithHash String
	 * @return nameWithoutHash
	 */
	private String stripHash(String nameWithHash) {
		String nameWithoutHash = nameWithHash;
		if (nameWithHash.indexOf('#') != -1) {
			nameWithoutHash =
				nameWithHash.substring(
					nameWithHash.indexOf('#') + 1,
					nameWithHash.length());
		}
		return nameWithoutHash;
	}

	/**
	 * Checks to see whether an interface type can be resolved to
	 * a role or port type yet
	 * @param roleOrPort AcmeElement using the interface types
	 * @param dom Element containing the interface type definition
	 */
	private void resolveRoleOrPortTypes(AcmeElement roleOrPort, Element dom)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		// Find the interfaceType DOM node corresponding to the id of the
		// DOM element passed in
		String name = stripHash(dom.getAttribute("id"));
		RoleOrPortType rop = (RoleOrPortType) nameToInterfaceType.get(name);

		if (rop.type == RoleOrPortType.UNKNOWN) {
			AcmeType typeACME = null;
			if (roleOrPort instanceof AcmePort) { // User is a port
				rop.type = RoleOrPortType.PORT;
				typeACME = processPortType(rop.node); // Process the port
			} else if (roleOrPort instanceof AcmeRole) { // User is a role
				rop.type = RoleOrPortType.ROLE;
				typeACME = processRoleType(rop.node); // Process as role
			}

			// Add the Acme type to the appropriate defining context
			// (Design or Family)
			if (rop.context instanceof AcmeDesign) {
				((AcmeDesign) rop.context).addType(typeACME);
			} else if (rop.context instanceof AcmeFamily) {
				((AcmeFamily) rop.context).addType(typeACME);
			}
		}
	}

	/**
	 * Translates a DOM propertyType node to an Acme propertyType
	 */
	private AcmePropertyType processPropertyType(Element propertyTypeDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmePropertyType typeACME = null;
		if (!propertyTypeDOM.getNodeName().equals(PROPERTYTYPE_TAG)) {
			System.err.println("property type tag does not match");
		} else {
			NamedNodeMap attributes = propertyTypeDOM.getAttributes();
			Attr name = (Attr) attributes.getNamedItem("name");
			NodeList details = propertyTypeDOM.getChildNodes();
			String acmeName = null;
			if (name != null) {
				acmeName = this.decomposeName(name.getValue());
			}
			Element typeDetailDOM = findFirstElement(propertyTypeDOM);

			if (typeDetailDOM.getNodeName().equals("propertyEnumType")) {
				EnumType etypeACME = new EnumType(acmeName);
				NodeList valuesDOM =
					getElementsAtThisLevel(typeDetailDOM, "value");
				for (int i = 0; i < valuesDOM.getLength(); i++) {
					CharacterData valueDOM = (CharacterData) valuesDOM.item(i);
					etypeACME.addValue(valueDOM.getData());
				}
				typeACME = etypeACME;
			} else if (
				typeDetailDOM.getNodeName().equals("propertyPrimitiveType")) {
				NamedNodeMap pType = typeDetailDOM.getAttributes();
				Attr typeAttr = (Attr) pType.getNamedItem("type");
				if (typeAttr == null) // for backward bug compatability
					typeAttr = (Attr) pType.getNamedItem("name");
				if (typeAttr.getValue().toLowerCase().equals("float")) {
					if (acmeName != null) {
						typeACME = new FloatType(acmeName);
					} else {
						typeACME =
							acmeLib.core.globals.ConstantTypes.FLOAT_TYPE;
					}
				} else if (typeAttr.getValue().toLowerCase().equals("int")) {
					if (acmeName != null) {
						typeACME = new IntType(acmeName);
					} else {
						typeACME = acmeLib.core.globals.ConstantTypes.INT_TYPE;
					}
				} else if (
					typeAttr.getValue().toLowerCase().equals("string")) {
					if (acmeName != null) {
						typeACME = new StringType(acmeName);
					} else {
						typeACME =
							acmeLib.core.globals.ConstantTypes.STRING_TYPE;
					}
				} else if (
					typeAttr.getValue().toLowerCase().equals("boolean")) {
					if (acmeName != null) {
						typeACME = new BooleanType(acmeName);
					} else {
						typeACME =
							acmeLib.core.globals.ConstantTypes.BOOLEAN_TYPE;
					}
				} else if (typeAttr.getValue().toLowerCase().equals("any")) {
					typeACME = new AnyType(acmeName);
				}
			} else if (typeDetailDOM.getNodeName().equals("propertySetType")) {
				Element propertySetTypeDOM = findFirstElement(typeDetailDOM);
				AcmePropertyType setTypeACME =
					processPropertyType(propertySetTypeDOM);
				typeACME =
					new acmeLib.properties.SetType(acmeName, setTypeACME);
			} else if (
				typeDetailDOM.getNodeName().equals("propertySequenceType")) {
				Element propertySeqTypeDOM = findFirstElement(typeDetailDOM);
				AcmePropertyType seqTypeACME =
					processPropertyType(propertySeqTypeDOM);
				typeACME =
					new acmeLib.properties.SequenceType(acmeName, seqTypeACME);
			} else if (
				typeDetailDOM.getNodeName().equals("propertyRecordType")) {
				NodeList fieldsDOM =
					getElementsAtThisLevel(
						typeDetailDOM,
						"propertyFieldDeclaration");
				acmeLib.properties.RecordType rtypeACME =
					new acmeLib.properties.RecordType(acmeName);
				for (int i = 0; i < fieldsDOM.getLength(); i++) {
					Element fieldDOM = (Element) fieldsDOM.item(i);
					Element fieldTypeDOM = findFirstElement(fieldDOM);
					AcmePropertyType fieldTypeACME =
						processPropertyType(fieldTypeDOM);
					String fieldname = fieldDOM.getAttribute("name");
					rtypeACME.addField(fieldname, fieldTypeACME, null);
				}
				typeACME = rtypeACME;
			} else if (typeDetailDOM.getNodeName().equals("typeref")) {
				String elementId = typeDetailDOM.getAttribute("xlink:href");
				AcmePropertyType referencedACME =
					(AcmePropertyType) mapNameToPropertyType (elementId);
				typeACME = new AcmeUnresolvedPropertyType(acmeName);
				typeACME.setAliasedType(referencedACME);
			} else {
				System.err.println("Error: property type has illegal contents");
			}

		}
		return typeACME;
	}

	/**
	 * Translates the DOM component type node to an AcmeLib component type
	 * @param componentType Element
	 * @return AcmeComponentType, containing the translated component type
	 */
	private AcmeComponentType processComponentType(Element componentTypeDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmeComponentType componentTypeACME = null;
		if (!componentTypeDOM.getNodeName().equals(COMPONENTTYPE_TAG)) {
			System.err.println("component type tag does not match");
		} else {
			String name =
				this.decomposeName(componentTypeDOM.getAttribute("id"));
			NodeList structureDOM =
				getElementsAtThisLevel(componentTypeDOM, "structure");
			componentTypeACME = new AcmeComponentType(name);

			// Process supertypes
			this.addSuperTypes(componentTypeDOM, componentTypeACME);

			// process structure
			AcmeComponent structureACME =
				processComponentInstance((Element) structureDOM.item(0));
			AcmeEnumeration portsACME = structureACME.getPorts();
			while (portsACME.moreItems()) {
				AcmePort portACME = (AcmePort) portsACME.getNextItem();
				componentTypeACME.addPort(portACME);
			}
			this.copyProperties(structureACME, componentTypeACME);
		}
		return componentTypeACME;
	}

	/**
	 * Translates a DOM connector type to an AcmeLib connector type
	 * @param connectorTypeDOM Element
	 * @return connectorTypeACME AcmeConnectorType
	 */
	AcmeConnectorType processConnectorType(Element connectorTypeDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmeConnectorType connectorTypeACME = null;
		if (!connectorTypeDOM.getNodeName().equals(CONNECTORTYPE_TAG)) {
			System.err.println("connector type tag does not match");
		} else {
			String name =
				this.decomposeName(connectorTypeDOM.getAttribute("id"));
			NodeList structureDOM =
				getElementsAtThisLevel(connectorTypeDOM, "structure");
			connectorTypeACME = new AcmeConnectorType(name);

			// Process super types
			this.addSuperTypes(connectorTypeDOM, connectorTypeACME);

			AcmeConnector structureACME =
				processConnectorInstance((Element) structureDOM.item(0));

			// Pull apart connector and add to type
			AcmeEnumeration rolesACME = structureACME.getRoles();
			while (rolesACME.moreItems()) {
				AcmeRole roleACME = (AcmeRole) rolesACME.getNextItem();
				connectorTypeACME.addRole(roleACME);
			}
			this.copyProperties(structureACME, connectorTypeACME);
		}
		return connectorTypeACME;
	}

	/**
	 * Copies the properties from one Acme element to another
	 * @param from AcmeElement
	 * @param to AcmeElement
	 */
	private void copyProperties(AcmeElement from, AcmeElement to)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmeEnumeration properties = from.getProperties();
		while (properties.moreItems()) {
			AcmeProperty property = (AcmeProperty) properties.getNextItem();
			to.addProperty(property);
		}
	}

	AcmeObject mapNameToElementType (String elementID) {
		StringTokenizer tokens = new StringTokenizer (elementID, "#");
		if (!elementID.startsWith("#")) {
			String filename = tokens.nextToken();
		}
		String id = tokens.nextToken ();
		StringTokenizer name = new StringTokenizer (id, "::");
		return new AcmeUnresolvedElementType (name.nextToken ());	
	}
	
	AcmeObject mapNameToPropertyType (String elementID) {
		StringTokenizer tokens = new StringTokenizer (elementID, "#");
		String filename = tokens.nextToken();
		String id = tokens.nextToken ();
		tokens = new StringTokenizer (id, "::");
		return new AcmeUnresolvedPropertyType (tokens.nextToken ());	
	}

	/** 
	 * Adds super types from a DOM type to an Acme type
	 * @param typeDOM Element
	 * @param typeACME AcmeElementType
	 */
	private void addSuperTypes(Element typeDOM, AcmeElementType typeACME)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		NodeList superTypesDOM = getElementsAtThisLevel(typeDOM, "superType");
		for (int i = 0; i < superTypesDOM.getLength(); i++) {
			Element superTypeDOM = (Element) superTypesDOM.item(i);
			String elementId = superTypeDOM.getAttribute("xlink:href");

			// Presumption: type is defined before use
			AcmeElementType superTypeACME =
				(AcmeElementType) mapNameToElementType (elementId);
			typeACME.addSuperType(superTypeACME);
		}
	}

	/**
	 * Translates a DOM port type node to AcmeLib
	 * @param portTypeDOM Element
	 * @return AcmePortType
	 */
	AcmePortType processPortType(Element portTypeDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmePortType portTypeACME = null;
		if (!portTypeDOM.getNodeName().equals(INTERFACETYPE_TAG)) {
			System.err.println("interface type does not match");
		} else {
			String name = this.decomposeName(portTypeDOM.getAttribute("id"));
			portTypeACME = new AcmePortType(name);
			this.addSuperTypes(portTypeDOM, portTypeACME);

			NodeList structureDOM =
				getElementsAtThisLevel(portTypeDOM, "structure");
			AcmePort structureACME =
				processPortInstance((Element) structureDOM.item(0));
			this.copyProperties(structureACME, portTypeACME);
			this.resolveRoleOrPortTypes(portTypeACME, portTypeDOM);

		}
		return portTypeACME;
	}

	/**
	 * Translates a DOM role type to an AcmeLib role type
	 * @param roleTypeDOM Element
	 * @return AcmeRoleType
	 */
	AcmeRoleType processRoleType(Element roleTypeDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmeRoleType roleTypeACME = null;
		if (!roleTypeDOM.getNodeName().equals(INTERFACETYPE_TAG)) {
			System.err.println("interface type does not match");
		} else {
			String name = this.decomposeName(roleTypeDOM.getAttribute("id"));
			roleTypeACME = new AcmeRoleType(name);
			this.addSuperTypes(roleTypeDOM, roleTypeACME);

			NodeList structureDOM =
				getElementsAtThisLevel(roleTypeDOM, "structure");
			AcmeRole structureACME =
				processRoleInstance((Element) structureDOM.item(0));
			this.copyProperties(structureACME, roleTypeACME);
			this.resolveRoleOrPortTypes(roleTypeACME, roleTypeDOM);
		}
		return roleTypeACME;
	}

	/**
	 * Transfers declared types from a dom element to an acme element
	 * @param dom Element
	 * @param acme AcmeElement
	 */
	private void transferTypes(Element dom, AcmeElement acme)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		NodeList typesDOM = getElementsAtThisLevel(dom, "type");
		for (int i = 0; i < typesDOM.getLength(); i++) {
			Element typeDOM = (Element) typesDOM.item(i);
			String referenceDOM = typeDOM.getAttribute("xlink:href");
			AcmeElementType typeACME =
				(AcmeElementType) mapNameToElementType(referenceDOM);
			if (typeACME != null) {
				acme.addDeclaredType(typeACME);
			}
			else {
				acme.addDeclaredType (new AcmeUnresolvedElementType (stripHash (referenceDOM)));
			}
		}
	}

	/**
	 * transfers instantiations from a dom element to an AcmeLib elemt
	 * @param dom Element
	 * @param acme AcmeElement
	 */
	private void transferInstantiation(Element dom, AcmeElement acme)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		NodeList instancesDOM = getElementsAtThisLevel(dom, "instanceOf");
		if (instancesDOM.getLength() > 0) {
			Element instanceDOM = (Element) instancesDOM.item(0);
			String reference = instanceDOM.getAttribute("xlink:href");
			AcmeElementType typeACME =
				(AcmeElementType) mapNameToElementType (reference);
			if (typeACME != null) {
				acme.addInstantiatedType(typeACME);
			}
			else {
				acme.addInstantiatedType (new AcmeUnresolvedElementType (stripHash (reference)));
			}
		}
	}

	/**
	 * translates the subArchitecture nodes of a dom element to the
	 * AcmeLib element
	 * @param dom Element
	 * @param acme AcmeElement
	 */
	private void transferRepresentations(Element dom, AcmeElement acme)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		NodeList representationsDOM =
			getElementsAtThisLevel(dom, "subArchitecture");
		for (int i = 0; i < representationsDOM.getLength(); i++) {
			Element representationDOM = (Element) representationsDOM.item(i);
			String representationName = "rep" + i;
			Element innerSystemDOM =
				(Element) getElementsAtThisLevel(
					representationDOM,
					"archInstance").item(
					0);
			AcmeSystem innerSystemACME = processSystem(innerSystemDOM);

			NodeList mappingsDOM =
				getElementsAtThisLevel(
					representationDOM,
					"interfaceInstanceMapping");
			acmeLib.representations.AcmeBindingsMap bindingsACME =
				new acmeLib.representations.AcmeBindingsMap();

			for (int j = 0; j < mappingsDOM.getLength(); j++) {
				Element mappingDOM = (Element) mappingsDOM.item(j);
				Element innerDOM =
					(Element) getElementsAtThisLevel(
						mappingDOM,
						"innerInterfaceInstance").item(
						0);
				Element outerDOM =
					(Element) getElementsAtThisLevel(
						mappingDOM,
						"outerInterfaceInstance").item(
						0);
				String innerRef = innerDOM.getAttribute("xlink:href");
				String outerRef = outerDOM.getAttribute("xlink:href");

				AcmeElement innerInterface =
					(AcmeElement) nameToAcme.get(stripHash(innerRef));
				if (innerInterface instanceof AcmePort) {
					AcmePort innerPort = (AcmePort) innerInterface;
					AcmePort outerPort =
						(AcmePort) nameToAcme.get(stripHash(outerRef));
					acmeLib.representations.AcmeBinding bindingACME =
						new acmeLib.representations.AcmeBinding(
							outerPort,
							innerPort);
					bindingACME.setExportOuterQualifier(false);
					bindingsACME.addBinding(bindingACME);
				} else if (innerInterface instanceof AcmeRole) {
					AcmeRole innerRole = (AcmeRole) innerInterface;
					AcmeRole outerRole =
						(AcmeRole) nameToAcme.get(stripHash(outerRef));
					acmeLib.representations.AcmeBinding bindingACME =
						new acmeLib.representations.AcmeBinding(
							outerRole,
							innerRole);
					bindingACME.setExportOuterQualifier(false);
					bindingsACME.addBinding(bindingACME);
				}
			}
			acmeLib.representations.AcmeRepresentation representationACME =
				new acmeLib.representations.AcmeRepresentation(
					innerSystemACME,
					acme,
					bindingsACME,
					representationName);

			acme.addRepresentation(representationACME);
		}
	}

	/**
	 * Translates a DOM component instance to an AcmeLib component
	 * @param componentDOM Element
	 * @return AcmeComponent
	 */
	AcmeComponent processComponentInstance(Element componentDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		NodeList interfacesDOM =
			getElementsAtThisLevel(componentDOM, "interfaceInstance");

		String name = this.decomposeName(componentDOM.getAttribute("id"));

		AcmeComponent componentACME = new AcmeComponent(name);
		int i;
		for (i = 0; i < interfacesDOM.getLength(); i++) {
			Element interfaceDOM = (Element) interfacesDOM.item(i);
			AcmePort portACME = processPortInstance(interfaceDOM);
			componentACME.addPort(portACME);
			nameToAcme.put(interfaceDOM.getAttribute("id"), portACME);
			nameToDom.put(interfaceDOM.getAttribute("id"), interfaceDOM);

		}

		this.transferProperties(componentDOM, componentACME);
		this.transferTypes(componentDOM, componentACME);
		this.transferInstantiation(componentDOM, componentACME);
		this.transferRepresentations(componentDOM, componentACME);
		return componentACME;
	}

	/**
	 * Translates a DOM connector instance to an AcmeLib connector
	 * @param connectorDOM Element
	 * @return AcmeConnector
	 */
	private AcmeConnector processConnectorInstance(Element connectorDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		NodeList interfacesDOM =
			getElementsAtThisLevel(connectorDOM, "interfaceInstance");
		String name = this.decomposeName(connectorDOM.getAttribute("id"));
		AcmeConnector connectorACME = new AcmeConnector(name);

		for (int i = 0; i < interfacesDOM.getLength(); i++) {
			Element interfaceDOM = (Element) interfacesDOM.item(i);
			AcmeRole roleACME = processRoleInstance(interfaceDOM);
			connectorACME.addRole(roleACME);
			nameToAcme.put(interfaceDOM.getAttribute("id"), roleACME);
			nameToDom.put(interfaceDOM.getAttribute("id"), interfaceDOM);
		}

		this.transferProperties(connectorDOM, connectorACME);
		this.transferTypes(connectorDOM, connectorACME);
		this.transferInstantiation(connectorDOM, connectorACME);
		this.transferRepresentations(connectorDOM, connectorACME);
		return connectorACME;
	}

	/**
	 * Translate a DOM interface instance node to an AcmeLib port
	 * @param portDOM Element
	 * @return AcmePort
	 */
	AcmePort processPortInstance(Element portDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		String name = this.decomposeName(portDOM.getAttribute("id"));

		AcmePort portACME = new AcmePort(name);
		this.transferProperties(portDOM, portACME);
		this.transferTypes(portDOM, portACME);
		this.transferInstantiation(portDOM, portACME);

		NodeList typesDOM = getElementsAtThisLevel(portDOM, "type");
		for (int i = 0; i < typesDOM.getLength(); i++) {
			Element typeDOM = (Element) typesDOM.item(i);
			String reference = typeDOM.getAttribute("xlink:href");
			// We now know the interface type is a port type, so resolve it
			RoleOrPortType defn =
				(RoleOrPortType) nameToInterfaceType.get(stripHash(reference));
			if (defn != null)
				this.resolveRoleOrPortTypes(portACME, defn.node);
		}
		return portACME;
	}

	/** 
	 * Translate a DOM interface instance to an AcmeLib role 
	 * @param roleDOM element
	 * @return AcmeRole
	 */
	AcmeRole processRoleInstance(Element roleDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		String name = this.decomposeName(roleDOM.getAttribute("id"));

		AcmeRole roleACME = new AcmeRole(name);
		this.transferProperties(roleDOM, roleACME);
		this.transferTypes(roleDOM, roleACME);
		this.transferInstantiation(roleDOM, roleACME);
		NodeList typesDOM = getElementsAtThisLevel(roleDOM, "type");
		for (int i = 0; i < typesDOM.getLength(); i++) {
			Element typeDOM = (Element) typesDOM.item(i);
			String reference = typeDOM.getAttribute("xlink:href");
			// We now know the interface type is a role so resolve it
			RoleOrPortType rop =
				(RoleOrPortType) nameToInterfaceType.get(stripHash(reference));
			if (rop != null)
				this.resolveRoleOrPortTypes(roleACME, rop.node);
		}
		return roleACME;
	}

	/**
	 * Translate a DOM porperty node to an ACME property
	 * @param propertyDOM Element
	 * @return AcmeProperty
	 */
	AcmeProperty processAcmeProperty(Element propertyDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		String name = this.decomposeName(propertyDOM.getAttribute("id"));
		AcmeProperty propertyACME = new AcmeProperty();
		propertyACME.setName(name);
		NodeList valuesDOM = getElementsAtThisLevel(propertyDOM, "value");
		if (valuesDOM.getLength() > 0) {
			AcmePropertyValue valueACME =
				processPropertyValue((Element) valuesDOM.item(0));
			propertyACME.setValue(valueACME);
		}

		NodeList typeDOM = getElementsAtThisLevel(propertyDOM, "propertyType");
		if (typeDOM.getLength() > 0) {
			AcmePropertyType typeACME =
				processPropertyType((Element) typeDOM.item(0));
			propertyACME.setType(typeACME);
		}
		return propertyACME;
	}

	/**
	 * Translates a DOM property value node to an AcmeLib property value
	 * @param valueDOM Element
	 * @return AcmePropertyValue
	 */
	AcmePropertyValue processPropertyValue(Element valueDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmePropertyValue valueACME = null;

		Element actualValueDOM = findFirstElement(valueDOM);
		if (actualValueDOM.getNodeName().equals("propertyPrimitiveValue")) {
			Element primValueDOM = findFirstElement(actualValueDOM);
			CharacterData cvalue = (CharacterData) primValueDOM.getFirstChild();
			if (cvalue == null) return null; // There is no value
			String unparsedValue = cvalue.getData();

			if (primValueDOM.getNodeName().equals("intValue")) {
				valueACME = new AcmeInt(Integer.parseInt(unparsedValue));
			} else if (primValueDOM.getNodeName().equals("floatValue")) {
				valueACME = new AcmeFloat(Float.parseFloat(unparsedValue));
			} else if (primValueDOM.getNodeName().equals("booleanValue")) {
				if (unparsedValue.toLowerCase().equals("false")) {
					valueACME = new AcmeBoolean(false);
				} else if (unparsedValue.toLowerCase().equals("true")) {
					valueACME = new AcmeBoolean(true);
				}
			} else if (primValueDOM.getNodeName().equals("stringValue")) {
				valueACME = new AcmeString(unparsedValue);
			}
		} else if (actualValueDOM.getNodeName().equals("propertySetValue")) {
			NodeList valuesDOM =
				getElementsAtThisLevel(actualValueDOM, "propertyValue");
			AcmeSet setValACME = new AcmeSet();
			for (int i = 0; i < valuesDOM.getLength(); i++) {
				AcmePropertyValue itemACME =
					processPropertyValue((Element) valuesDOM.item(i));
				setValACME.addValue(itemACME);
			}
			valueACME = setValACME;
		} else if (
			actualValueDOM.getNodeName().equals("propertySequenceValue")) {

			NodeList valuesDOM =
				getElementsAtThisLevel(actualValueDOM, "propertyValue");
			AcmeSequence seqValACME = new AcmeSequence();
			for (int i = 0; i < valuesDOM.getLength(); i++) {
				AcmePropertyValue itemACME =
					processPropertyValue((Element) valuesDOM.item(i));
				seqValACME.addValue(itemACME);
			}
			valueACME = seqValACME;
		} else if (
			actualValueDOM.getNodeName().equals("propertyRecordValue")) {
			AcmeRecord recValACME = new AcmeRecord();
			NodeList fieldsDOM =
				getElementsAtThisLevel(actualValueDOM, "propertyFieldValue");
			for (int i = 0; i < fieldsDOM.getLength(); i++) {
				Element fieldDOM = (Element) fieldsDOM.item(i);
				String fieldName = fieldDOM.getAttribute("fieldName");
				AcmePropertyType fieldTypeACME = null;
				AcmePropertyValue fieldValueACME = null;
				NodeList typesDOM =
					getElementsAtThisLevel(fieldDOM, "propertyType");
				if (typesDOM.getLength() > 0) {
					fieldTypeACME =
						processPropertyType((Element) typesDOM.item(0));
				}
				NodeList valuesDOM =
					getElementsAtThisLevel(fieldDOM, "propertyValue");
				if (valuesDOM.getLength() > 0) {
					fieldValueACME =
						processPropertyValue((Element) valuesDOM.item(0));
				}
				AcmeRecordField fieldACME =
					new AcmeRecordField(
						fieldName,
						fieldTypeACME,
						fieldValueACME);
				recValACME.addField(fieldACME);
			}
			valueACME = recValACME;
		}

		return valueACME;
	}

	/**
	 * Translates a DOM family node to an AcmeLib family
	 * @param familyDOM Element
	 * @return AcmeFamily
	 */
	private AcmeFamily processFamily(Element familyDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmeFamily familyACME;
		String name = this.decomposeName(familyDOM.getAttribute("id"));
		familyACME = new AcmeFamily(name);
		this.addSuperTypes(familyDOM, familyACME);
		int i;

		NodeList propertyTypesDOM =
			getElementsAtThisLevel(familyDOM, "propertyType");
		for (i = 0; i < propertyTypesDOM.getLength(); i++) {
			Element propertyTypeDOM = (Element) propertyTypesDOM.item(i);
			propertyTypeDOM.normalize();
			AcmeType propertyTypeACME =
				this.processPropertyType(propertyTypeDOM);
			if (propertyTypeACME != null) {
				familyACME.addType(propertyTypeACME);
				nameToAcme.put(
					propertyTypeDOM.getAttribute("id"),
					propertyTypeACME);
				nameToDom.put(
					propertyTypeDOM.getAttribute("id"),
					propertyTypeDOM);
			}
		}

		NodeList interfaceTypesDOM =
			getElementsAtThisLevel(familyDOM, "interfaceType");
		for (i = 0; i < interfaceTypesDOM.getLength(); i++) {
			Element interfaceTypeDOM = (Element) interfaceTypesDOM.item(i);
			interfaceTypeDOM.normalize();

			RoleOrPortType interfaceUnresolved = new RoleOrPortType();
			interfaceUnresolved.context = familyACME;
			interfaceUnresolved.node = interfaceTypeDOM;
			nameToInterfaceType.put(
				interfaceTypeDOM.getAttribute("id"),
				interfaceUnresolved);
		}

		NodeList componentTypesDOM =
			getElementsAtThisLevel(familyDOM, "componentType");
		for (i = 0; i < componentTypesDOM.getLength(); i++) {
			Element componentTypeDOM = (Element) componentTypesDOM.item(i);
			componentTypeDOM.normalize();
			AcmeType componentTypeACME =
				this.processComponentType(componentTypeDOM);
			if (componentTypeACME != null) {
				familyACME.addType(componentTypeACME);
				nameToAcme.put(
					componentTypeDOM.getAttribute("id"),
					componentTypeACME);
				nameToDom.put(
					componentTypeDOM.getAttribute("id"),
					componentTypeDOM);
			}
		}

		NodeList connectorTypesDOM =
			getElementsAtThisLevel(familyDOM, "connectorType");
		for (i = 0; i < connectorTypesDOM.getLength(); i++) {
			Element connectorTypeDOM = (Element) connectorTypesDOM.item(i);
			connectorTypeDOM.normalize();
			AcmeType connectorTypeACME =
				this.processConnectorType(connectorTypeDOM);
			if (connectorTypeACME != null) {
				familyACME.addType(connectorTypeACME);
				nameToAcme.put(
					connectorTypeDOM.getAttribute("id"),
					connectorTypeACME);
				nameToDom.put(
					connectorTypeDOM.getAttribute("id"),
					connectorTypeDOM);
			}
		}

		NodeList structuresDOM = getElementsAtThisLevel(familyDOM, "structure");
		if (structuresDOM.getLength() == 1) {
			Element structureDOM = (Element) structuresDOM.item(0);
			AcmeSystem systemACME = processSystem(structureDOM);
			transferSystem(systemACME, familyACME);
		}
		return familyACME;
	}

	/**
	 * Translates a DOM system node to an AcmeLib system
	 * @param systemDOM Element
	 * @return AcmeSystem
	 */
	private AcmeSystem processSystem(Element systemDOM)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		AcmeSystem systemACME;
		int i;

		String name = this.decomposeName(systemDOM.getAttribute("id"));
		systemACME = new AcmeSystem(name);

		this.transferProperties(systemDOM, systemACME);
		this.transferTypes(systemDOM, systemACME);

		NodeList components =
			getElementsAtThisLevel(systemDOM, "componentInstance");
		for (i = 0; i < components.getLength(); i++) {
			Element componentDOM = (Element) components.item(i);
			AcmeComponent componentACME =
				processComponentInstance(componentDOM);
			systemACME.addComponent(componentACME);
			nameToAcme.put(componentDOM.getAttribute("id"), componentACME);
			nameToDom.put(componentDOM.getAttribute("id"), componentDOM);
		}

		NodeList connectors =
			getElementsAtThisLevel(systemDOM, "connectorInstance");
		for (i = 0; i < connectors.getLength(); i++) {
			Element connectorDOM = (Element) connectors.item(i);
			AcmeConnector connectorACME =
				processConnectorInstance(connectorDOM);
			systemACME.addConnector(connectorACME);
			nameToAcme.put(connectorDOM.getAttribute("id"), connectorACME);
			nameToDom.put(connectorDOM.getAttribute("id"), connectorDOM);

		}
		//	NodeList groups = getElementsAtThisLevel (systemDOM, "groupInstance");
		NodeList attachments =
			getElementsAtThisLevel(systemDOM, "linkInstance");
		for (i = 0; i < attachments.getLength(); i++) {
			NodeList points =
				getElementsAtThisLevel(attachments.item(i), "point");
			Element lhsPoint = (Element) points.item(0);
			Element rhsPoint = (Element) points.item(1);
			Element lhsDOM =
				(Element) getElementsAtThisLevel(
					lhsPoint,
					"anchorInterface").item(
					0);
			Element rhsDOM =
				(Element) getElementsAtThisLevel(
					rhsPoint,
					"anchorInterface").item(
					0);

			AcmeElement lhsACME =
				(AcmeElement) nameToAcme.get(
					stripHash(lhsDOM.getAttribute("xlink:href")));
			if (lhsACME instanceof AcmePort) {
				AcmePort portACME = (AcmePort) lhsACME;
				AcmeRole roleACME =
					(AcmeRole) nameToAcme.get(
						stripHash(rhsDOM.getAttribute("xlink:href")));
				systemACME.addAttachment(portACME, roleACME);
			} else if (lhsACME instanceof AcmeRole) {
				AcmePort portACME =
					(AcmePort) nameToAcme.get(
						stripHash(rhsDOM.getAttribute("xlink:href")));
				AcmeRole roleACME = (AcmeRole) lhsACME;
				systemACME.addAttachment(portACME, roleACME);
			}
		}

		return systemACME;

	}

	/**
	 * translates properties defined in a DOM node to an AcmeLib node
	 * @param dom Element
	 * @param acme AcmeElement
	 */
	private void transferProperties(Element dom, AcmeElement acme)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		NodeList propertiesDOM = getElementsAtThisLevel(dom, "property");
		for (int i = 0; i < propertiesDOM.getLength(); i++) {
			Element propertyDOM = (Element) propertiesDOM.item(i);
			AcmeProperty propertyACME = processAcmeProperty(propertyDOM);
			acme.addProperty(propertyACME);
		}
	}

	/**
	 * Returns whether the design contains imported family references
	 * ASSUMPTION: The design is valid xAcme
	 * @param design
	 * @return
	 */
	public static boolean referencesFamilies(Element design) {
		NodeList imports = getElementsAtThisLevel(design, IMPORT_TAG);
		if (imports.getLength() > 0) {
			for (int i = 0; i < imports.getLength(); i++) {
				Element importElement = (Element )imports.item(i);
				String val = importElement.getAttribute ("type");
				if (val != null && val.equals ("xml")) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * @param string
	 * @param des
	 */
	public void addImportedDesign(String string, AcmeDesign des) {
		_xmlToImportedDesignMap.put (string, des);
		
	}


}
