// Class: AcmeObject
// Purpose: Base class for all ACME classes.
//
// Revision History:
//    Date:   Person    Change
//    ?/?/97    DK      Created C++ version
//    4/9/97    RTM     Translated to Java

package acmeLib;

import java.util.StringTokenizer;

import acmeLib.core.io.AcmeOutputStream;

public class AcmeObject {

    // Every AcmeObject has a unique number that describes the order
    // in which it was created, relative to all of the other AcmeObjects
    // in this JVM.

    // global id that will be assigned to the next AcmeObject created
    private static int sNextCreationNumber = 1;

    private int creationOrder;

    public AcmeObject() {
        this.creationOrder = sNextCreationNumber;
        sNextCreationNumber++;
    };

    /**
     * shallow copy
     * @param newObject new object
     */
    public void copy(AcmeObject newObject) {
        newObject = this;
    }

    public int getCreationOrder() { return this.creationOrder; };
    public void setCreationOrder(int newOrder) {
        this.creationOrder = newOrder;
    };

    private String _comment;
    
    public String getComment () {
        return _comment;
    }
    
    public void setComment (String comment) {
        _comment = comment;
    }
    
    /**
     * Word wrap the comment
     * @param maxLength The length of the line
     * @return the comment split over multiple lines
     */
    protected String formatComment (int maxLength) {
        StringBuffer formattedComment = new StringBuffer ();
        StringBuffer currentLine = new StringBuffer ();
        // if it's already less than maxLength, then just return the original
        if (_comment.length ()  < maxLength) {
            return _comment;
        }
        boolean startsWithDoubleSlash = _comment.trim ().startsWith("//");
        StringTokenizer tokens = new StringTokenizer (_comment, " \n");
        while (tokens.hasMoreTokens()) {
            String token =    tokens.nextToken ();
           
            if (token.startsWith ("//") || (token.length () + currentLine.length () > maxLength)) {
                formattedComment.append (currentLine);
                int len = currentLine.length();
                if (len > 0) {
                	formattedComment.append ((currentLine.charAt (len-1) != '\n')?"\n":"");
                }
                if (token.startsWith ("//") || currentLine.length () > 0) {
                	currentLine = new StringBuffer ();
                }
                currentLine.append ((startsWithDoubleSlash && !token.startsWith ("//")?"// ":""));
                currentLine.append (token);
                currentLine.append (" ");
            }
            
            else {
                currentLine.append (token);
                currentLine.append (" ");   
            }
        }
        formattedComment.append (currentLine);
        return formattedComment.toString ();   
    }
    
    public void exportComment (AcmeOutputStream s) throws java.io.IOException {
        final int LINE_LENGTH = 80;

        
        if (_comment != null) {
            String comment = formatComment (LINE_LENGTH);
            // Strip leading newline (BugFix #122)
            while (comment.startsWith ("\n")) {
                comment = comment.substring (1);
            }
            if (comment.startsWith("//") || comment.startsWith ("/*")) {
                // haven't processed the comment
                StringTokenizer tokens = new StringTokenizer (comment, "\n");
                    while (tokens.hasMoreTokens()) {      
                                     
                        String line = tokens.nextToken();
                        
                        if (_comment.startsWith ("//") && !line.startsWith ("//")) {
                            s.writeln("// " + line);
                        }
                        else {
                            s.writeln (line);
                        }
                    }
            }
            else {
                if (comment.indexOf ("\n") != -1) {
                    StringTokenizer tokens = new StringTokenizer (comment, "\n");
                    s.write ("/* ");
                    while (tokens.hasMoreTokens()) {
                        s.writeln (tokens.nextToken ());
                    }
                    s.writeln ("*/");
                }      
                else {
                    s.writeln ("//" + comment);
                }
            }   
            
        }   
    }

    public static int getNextCreationNumber() { return sNextCreationNumber; };
};

