/*
 * Created on Mar 4, 2004
 *  
 */
package acmeLib.armaniParser;

import java.io.*;
import java.util.HashMap;
import java.util.Map;

import acmeLib.armaniParser.ParseException;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeDesign;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.AcmeParseError;
/**
 * @author Bradley Schmerl Created Mar 4, 2004
 */
public class ParseAcme {

	String			canonicalInputFilename	= null;
	InputStream		inStream				= null;
	OutputStream	outStream				= null;
	ArmaniParser	parser;
	private boolean	detailedOutput;
	private String	parserID;
	private Map		arguments;

	/**
	 *  
	 */
	public ParseAcme() {
		super ();
	}

	public static void main(String [] args) {
		new ParseAcme ().parse (args);
	}

	/**
	 * @param args
	 */
	private void parse(String [] args) {
		parserID = new String ("Acme Parser (release 1.0) for Acme 2 (with armani constraints)");
		processArgs (args);
		if (arguments.get ("help") == Boolean.TRUE) {
			outputHelpMessage ();
			return;
		}
		if (arguments.get("detailed") == Boolean.TRUE) {
			LanguageConfig.exportChildrenFromSupertype = true;
		}
		// initialize the design and start the command interpreter
		try {
			// set the parser up to tracing and debugging options
			parser.disable_tracing ();
			// initialize the design
			ArmaniParser.initDesign ();
			ArmaniParser.getGlobalDesign ().setCanonicalFilename (canonicalInputFilename);
			//			 it's possible that the initDesign() method will ReInit the
			// parser to parse the builtInAnalysisDesign, so ReInit the
			// parser with the calculated input stream.
			ArmaniParser.ReInit (inStream);
			// clear all errors from the AcmeErrorManager so that we have
			// a clean slate from which to do processing.
			AcmeErrorManager.clearAllErrors ();
			AcmeParseError.reportLineNumber = true;
			AcmeDesign design = ArmaniParser.parse_AcmeDesign ();
			if (AcmeErrorManager.getNumErrorsReported () != 0) {
				System.err.println (" ** Errors found: **");
				AcmeErrorManager.printAllErrors (System.err);
				System.err.flush ();
				AcmeParseError.reportLineNumber = false;
				throw new ParseException ("parse error");
			}
			System.err.print ("--> Resolving Imported Designs ... ");
			design.resolveImportedDesigns ();
			if (AcmeErrorManager.getNumErrorsReported () != 0) {
				System.err.println (" ** Errors found: **");
				AcmeErrorManager.printAllErrors (System.err);
				System.err.flush ();
				throw new ParseException ("import statement error");
			}
			System.err.println (" done.");
			System.err.print ("--> Resolving Type Names ...");
			design.resolveTypeReferences ();
			if (AcmeErrorManager.getNumErrorsReported () != 0) {
				System.err.println (" ** Errors found: **");
				AcmeErrorManager.printAllErrors (System.err);
				System.err.flush ();
				throw new ParseException ("typename resolution error");
			}
			System.err.println (" done.");
			System.err.print ("--> Instantiating elements ...");
			design.instantiateElements ();
			if (AcmeErrorManager.getNumErrorsReported () != 0) {
				System.err.println (" ** Errors found: **");
				AcmeErrorManager.printAllErrors (System.err);
				System.err.flush ();
				throw new ParseException ("element instantiation error");
			}
			System.err.println (" done.");
			System.err.print ("--> Resolving instance names ...");
			design.resolveInstanceReferences ();
			if (AcmeErrorManager.getNumErrorsReported () != 0) {
				System.err.println (" ** Errors found: **");
				AcmeErrorManager.printAllErrors (System.err);
				System.err.flush ();
				throw new ParseException ("instance name resolution error");
			}
			System.err.println (" done.");
			System.err.flush ();
			System.out.println (parserID + ": Acme specification parsed succesfully.");
			ArmaniParser.export_ArmaniDesign (outStream, design);
			ArmaniParser.typecheckDesign (design);
		}
		catch (ParseException e) {
			System.out.println (parserID + ":  Encountered errors during Acme specification parse.");
			e.printStackTrace ();
			ArmaniParser.printWarningMsgs ();
			ArmaniParser.printErrorMsgs ();
			ArmaniParser.pause_for_input ();
		}
		catch (acmeLib.acmeParser.ParseException e) {
			System.out.println (parserID + ":  Encountered errors during Acme specification parse.");
			e.printStackTrace ();
			ArmaniParser.printWarningMsgs ();
			ArmaniParser.printErrorMsgs ();
			ArmaniParser.pause_for_input ();
		}
		// close the files we opened
		try {
			if (outStream != null) {
				outStream.close ();
			}
		}
		catch (java.io.IOException ioExc) {
		}
		try {
			if (inStream != null) {
				inStream.close ();
			}
		}
		catch (java.io.IOException ioExc) {
		}
	}

	

	/**
	 *  
	 */
	private void outputHelpMessage() {
		System.out.println ("Acme v3.0 Parser (release 0.1):  " + "Usage is one of:");
		System.out.println ("         java ArmaniParser");
		System.out.println ("            -- for command-line interpreter from stdin");
		System.out.println ("OR");
		System.out.println ("         java ArmaniParser < inputfile");
		System.out.println ("            -- for batch-style command-line interpretation");
		System.out.println ("OR");
		System.out.println ("         java ArmaniParser inputfile");
		System.out.println ("            -- to directly parse the Acme file inputfile");
		System.out.println ("               or for batch-style command-line interpretation");
		System.out.println ("OR");
		System.out.println ("         java ArmaniParser inputfile outputfile");
		System.out.println ("            -- to directly parse the Acme file inputfile\n"
				+ "               and export the Acme design to the file outputfile");
		System.out.println ("Options:");
		System.out.println ("         -help Prints this message");
		System.out.println ("         -detailed Outputs acme file with detailed output");
	}

	/**
	 * @param args
	 */
	private void processArgs(String [] args) {
		arguments = new HashMap ();
		arguments.put ("detailed", Boolean.FALSE);
		arguments.put ("help", Boolean.FALSE);
		int firstFileIdx = -1;
		for (int i = 0; i < args.length; i++) {
			if (!args [i].startsWith ("-")) {
				firstFileIdx = i;
				break;
			}
			if (args [i].equals ("-detailed")) {
				arguments.put ("detailed", Boolean.TRUE);
			}
			else if (args [i].equals ("-help")) {
				arguments.put ("help", Boolean.TRUE);
			}
		}
		int inputFileIdx = firstFileIdx;
		int outputFileIdx = firstFileIdx == -1 ? -1 : firstFileIdx + 1;
		if (outputFileIdx == args.length) outputFileIdx = -1;
		if (firstFileIdx == -1) {
			// there are no files
			inStream = System.in;
			parser = new ArmaniParser (inStream);
			canonicalInputFilename = "System.in";
			outStream = System.out;
			detailedOutput = false;
		}
		else {
			try {
				inStream = new FileInputStream (args [firstFileIdx]);
				parser = new ArmaniParser (inStream);
				File inputFile = new File (args [firstFileIdx]);
				try {
					canonicalInputFilename = inputFile.getCanonicalPath ();
				}
				catch (IOException ioExc) {
					canonicalInputFilename = args [firstFileIdx];
				}
			}
			catch (FileNotFoundException e) {
				System.err.println (parserID + ": File " + args [firstFileIdx] + " not found.");
				return;
			}
		}
		if (outputFileIdx == -1) {
			outStream = System.out;
		}
		else {
			try {
				outStream = new FileOutputStream (args [outputFileIdx]);
			}
			catch (FileNotFoundException e) {
				System.err.println (parserID + ": Could not open " + args [outputFileIdx] + " to write");
			}
		}
	}
}
