
package acmeLib.core.container; 
import acmeLib.AcmeObject;


/**
 *  Class:   AcmeEnumeration
<BR>
 *  Purpose: Base class for all ACME enumeration classes.  The ACME enumerations
<BR>
 *           extend generic Java enumerations by removing all "null" references
<BR>
 *           when generating an enumeration.
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK        Created C++ version
<BR>
 *    4/14/97   RTM       Translated to Java
<BR>
 *    4/24/98   GZ        Added support for sorting enumerations based
<BR>
 *                        on the creationOrder of the objects.
<BR><BR>
 *  @author Bob Monroe<BR>
 */
public class AcmeEnumeration extends acmeLib.AcmeObject {

	/** An empty enumeration **/
    public final static AcmeEnumeration EMPTY = new AcmeEnumeration ();
 //
 // protected members
 //
 protected final int INITIAL_ENUMERATION_SIZE = 10;
 protected final int MAX_ARRAY_INCREASE = 400;

  /**
   *  This method is used to expand an already existing array
   *  The algorithm we use for expanding a filled array is to create a
   *  new array that is the minimum of (doubling the size of the old array
   *  or MAX_ARRAY_INCREASE), then fill the new array with the values from
   *  the old array.  This could probably be done more efficiently.
   */
 protected void expandArray() {
    // create the new array
    int new_array_size;
    if ((items.length * 2) < MAX_ARRAY_INCREASE)
        new_array_size = items.length * 2;
    else
        new_array_size = items.length + MAX_ARRAY_INCREASE;
    AcmeObject[] new_items_array = new AcmeObject[new_array_size];

    // copy the old array into the new array
    for (int i = 0; i < items.length; i++) {
        new_items_array[i] = items[i];
    };

    // reset the items array to refer to the new array.  Hopefully this
    // should mark the old array for garbage collection.
    items = new_items_array;

 };


  /**
   * This method creates a new array that is at least addMinSlots bigger than current
   * position is, then fill the new array with the values from
   * the old array.  This could probably be done more efficiently.
   * @param addMinSlots used to determine the size of the new array
   */
 protected void expandArray(int addMinSlots) {
    // create the new array
    int new_array_size;
    if ((items.length * 2) > (num_items + addMinSlots + 1))
        new_array_size = items.length * 2;
    else
        new_array_size = num_items + addMinSlots + 1;

    AcmeObject[] new_items_array = new AcmeObject[new_array_size];

    // copy the old array into the new array
    for (int i = 0; i < this.num_items; i++) {
        new_items_array[i] = items[i];
    };

    // reset the items array to refer to the new array.  Hopefully this
    // should mark the old array for garbage collection.
    items = new_items_array;

 };



  //
  // protected members
  //
  protected AcmeObject[] items;
  protected int num_items;
  protected int next_item;


  //
  // public members
  //

  /**
   * This is a constructor for this class
   */
  public AcmeEnumeration() {
      num_items = 0;
      next_item = 0;
      items = new AcmeObject[INITIAL_ENUMERATION_SIZE] ;
  };

  /**
   * This is a constructor for this class
   * @param initialSize initial size of enumeration
   */
  public AcmeEnumeration(int initialSize) {
    num_items = 0;
    next_item = 0;
    items = new AcmeObject[initialSize] ;
  };


  /**
   * copy constructor
   * @param originalEnum The enumeration that we will create a copy of.
   * @return a copy of originalEnum
   **/
  public AcmeEnumeration(AcmeEnumeration originalEnum) {
    num_items = originalEnum.num_items;
    next_item = originalEnum.next_item;
    items = new AcmeObject[originalEnum.num_items];
    for (int i = 0; i < originalEnum.items.length; i++) {
        items[i] = originalEnum.items[i];
    };
  };


  /**
   * @return num_items
   */
  public int getSize() { return num_items; };

  /**
   * This method inserts a new item by catching an out of bounds error,
   * and dynamically resizing the array if needed.
   * @param item the new AcmeObject to be inserted
   */
  public void insertItem (AcmeObject item) {
      // catch an out of bounds error, and dynamically resize the array
      // if needed.
      if (num_items >= items.length) {
        this.expandArray();
      }

      items[num_items] = item;
      num_items++;
  };


  /**
   * This method concatenates enumToConcat to the end of this enumeration.
   * @param enumToCancat the enumeration
   */
  public void concatenate(AcmeEnumeration enumToConcat) {
    // make sure the array is large enough.
    if ((enumToConcat.getSize() + this.num_items) >= items.length) {
        expandArray(enumToConcat.getSize());
    };

    // don't roll this into the for loop because it creates an infinite
    // loop if you concatenate an enumeration with itself.
    int enumToConcatNumItems = enumToConcat.num_items;
    for (int i = 0; i < enumToConcatNumItems; i++) {
        this.items[num_items] = enumToConcat.items[i];
        num_items++;
    };
  };

  /**
   * This method returns whether the enumeration contains more elements
   * @return whether the enumeration contains more elements
   */
  public boolean moreItems() {
      if (next_item<num_items)
        return true;
      else
        return false;
  };

  /**
   * This method returns the next element in the enumeration
   * @return the next element in the enumeration
   */
  public AcmeObject getNextItem() {
      AcmeObject obj = items[next_item];
      next_item++;
      return obj;
  };

  public void reset() {
    this.next_item = 0;
  };

  /**
   * @param p the begining of the partition
   * @param r the end of the partition
   * @return where the partition occured
   */
 private int partition(int p, int r) {
     int pCreationOrder = ((AcmeObject)items[p]).getCreationOrder();
     int i = p - 1, j = r + 1;
     while (true) {
         do { j--; }
            while (((AcmeObject)items[j]).getCreationOrder() > pCreationOrder);
         do { i++; }
            while (((AcmeObject)items[i]).getCreationOrder() < pCreationOrder);
        if (i < j) {
            AcmeObject temp = items[i];
            items[i] = items[j];
            items[j] = temp;
        }
        else
            return j;
     }
 }

 /**
   * This method performs quick sort recursively
   * @param p the begining of the partition
   * @param r the end of the partition
   */
 private void quickSort(int p, int r) {
     if (p < r) {
         int q = partition(p, r);
         quickSort(p, q);
         quickSort(q+1, r);
     }
 }

 public void sortByCreationOrderFromHere() {
    if (!moreItems()) return;
    quickSort(next_item, num_items - 1);
 }

 public void sortByCreationOrder() {
    if (num_items == 0) return;
    reset();
    quickSort(next_item, num_items - 1);
 }

  /**
   * @return a String of the items
   */
 public String toString() {
    if (num_items <= 0){
        return new String("empty enum");
    }
    StringBuffer buf = new StringBuffer();
    for (int i = 0; i < num_items; i++)
        buf.append("\t" + items[i].toString() + "\n");

    // get rid of the last newline.
    buf.setLength(buf.length() - 1);
    return buf.toString();
 };



  /**
   *  This utility method casts an AcmeEnumeration to a java.util.Enumeration.
   *  starts from next_item of the internal array and fills the java Enumeration
   *  until the end of the array is reached, in the same order.
   *  @return the java Enumeration of the items
   */
  public java.util.Enumeration castToJavaEnum() {
    java.util.Vector v = new java.util.Vector(this.getSize());
    for (int i = next_item; i < num_items; i++) {
        v.addElement(items[i]);
    };
    return v.elements();
  };


  /**
   *  This utility method casts an AcmeEnumeration to a java.util.Vector.
   *  starts from next_item of the internal array and fills the java Vector
   *  until the end of the array is reached, in the same order.
   *  @return the AcmeVector of the items
   */
  public AcmeVector castToAcmeVector() {
    AcmeVector v = new AcmeVector(this.getSize());
    for (int i = next_item; i < num_items; i++) {
        v.addItem(items[i]);
    };
    return v;
  };


  /**
   *  This utility method casts an AcmeEnumeration to a java.util.Vector that
   *  has set semantics.  That is, the vector has no identical items.  It
   *  starts from next_item of the internal array and fills the java Enumeration
   *  until the end of the array is reached, in the same order.
   *  @return the AcmeVector of the items
   */
public AcmeVector castToAcmeSetVector() {
    AcmeVector v = new AcmeVector(this.getSize());
    for (int i = next_item; i < num_items; i++) {
        if (!v.contains(items[i])) {
            v.addItem(items[i]);
        };
    };
    return v;
};


};  // end AcmeEnumeration class definition

