
package acmeLib.core.container;

import java.util.Hashtable;
import java.util.Enumeration;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.filters.*;
import acmeLib.AcmeObject;


/**
 *  Class:   AcmeMap
<BR>
 *  Purpose: Base class for all ACME *Map classes.  The ACME *Map classes
<BR>
 *           provide typed maps from names to ACME objects for specific
<BR>
 *           ACME classes.
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK        Created C++ version
<BR>
 *    4/9/97    RTM       Translated to Java
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */

public class AcmeMap extends acmeLib.AcmeObject {

  //
  // private members
  //
  protected Hashtable acmeMapHashTable;

  //
  // public members
  //

  /**
   * This is the constructor for this class.
   * It uses a small hash-table by default, as most Acme and Armani maps are
   * quite small.
   */
  public AcmeMap() {
    // use a small hash-table by default, as most Acme and Armani maps are
    // quite small.
    acmeMapHashTable = new Hashtable(5);
  };

  /**
   * @param initialCapacity the initial size of the acmeMapHashTable
   */
  public AcmeMap(int initialCapacity) {
    acmeMapHashTable = new Hashtable(initialCapacity);
  };

 /**
   * @param initialCapacity the initial size of the acmeMapHashTable
   * @param loadingFactor the initial loadingFactor of the acmeMapHashTable
   */
  public AcmeMap(int initialCapacity, float loadingFactor) {
    acmeMapHashTable = new Hashtable(initialCapacity, loadingFactor);
  };

  /**
   * @return the size of the acmeMapHashTable
   */
  public int getSize() { return acmeMapHashTable.size(); };

  /**
   * This method performs a shallow-copy clone of
   * "this" object and copies the values stored in "this" and
   * places them in "targetMap".
   *
   * @param targetMap the destination of the "this" Object
   * @exception java.lang.CloneNotSupportedException thrown if clone is not supported
   */
  public void copyItems(AcmeMap targetMap) throws CloneNotSupportedException {
    targetMap = (AcmeMap) this.clone();
  };

  /**
   * This method adds new Objects to the acmeMapHashTable
   * @param name the name of the Object
   * @param obj the AcmeObject
   */
  public void addItem(String name, AcmeObject obj) {
    acmeMapHashTable.put(name, obj);
  };

  /**
   * This method removes Objects from the acmeMapHashTable
   * @param name the name of the Object
   */
  public void removeItem (String name) {
    acmeMapHashTable.remove(name);
  };

  /**
   * This method looks up an AcmeObject
   * @param name the name of the Object
   * @return the Object found in acmeMapHashTable
   */
  public AcmeObject lookupItem(String name) {
    return (AcmeObject) acmeMapHashTable.get(name);
  };


  /**
   * returns true if the AcmeMap contains the AcmeObject in question, else false
   * @param testObj Check whether this Map contains testObj
   * @return true if the AcmeMap contains the AcmeObject, else false.
   **/
  public boolean contains(AcmeObject testObj) {
    return acmeMapHashTable.contains(testObj);
  };


  /**
   * This method empties the map.
   */
  public void clear() {
    this.acmeMapHashTable.clear();
  };

  /**
   * @return an AcmeEnumeration
   */
  public AcmeEnumeration enumerate() {
    return enumerate ((AcmeObjectFilter) null);
  };

  /**
   * @param e AcmeEnumeration
   * @return an AcmeEnumeration of e
   */
  public AcmeEnumeration enumerate(AcmeEnumeration e) {
    return enumerate (e, (AcmeObjectFilter) null);
  };

 /**
   * @param filter AcmeObjectFilter
   * @return an AcmeEnumeration
   */
  public AcmeEnumeration enumerate(AcmeObjectFilter filter) {
    return enumerate (new AcmeEnumeration(), filter);
  };

  /**
   * This method returns an enumeration of the objects in the map that pass the
   * "test" provided by the filter object.
   * @param acmeEnum AcmeEnumeration
   * @param filter   AcmeObjectFilter
   * @return an AcmeEnumeration
   */
  public AcmeEnumeration enumerate(AcmeEnumeration acmeEnum,
                                   AcmeObjectFilter filter) {

    for (Enumeration e = acmeMapHashTable.elements(); e.hasMoreElements() ;)
      {
         try {
            AcmeObject acmeObj = (AcmeObject) e.nextElement();
            if (acmeObj != null) {
                if (filter != null) {
                    if (filter.test(acmeObj)) acmeEnum.insertItem(acmeObj);
                } else {
                    acmeEnum.insertItem(acmeObj);
                }
            };
         } catch (ClassCastException c) { };  // if can't cast then continue
      }

    return acmeEnum;
  };

  /**
   * This method converts an AcmeMap to a string
   * @return the String representation of AcmeMap
   */
  public String toString() { return acmeMapHashTable.toString(); };
  
  public Enumeration keys () {
     
    return acmeMapHashTable.keys ();   
  }
};

