
package acmeLib.core.container;

import java.util.Enumeration;
import java.util.Vector;

import acmeLib.AcmeObject;
import acmeLib.Debug;
import acmeLib.core.filters.AcmeObjectFilter;
import acmeLib.elements.AcmeElementType;


/**
 *  Class:  AcmeVector
<BR>
 *  Purpose: Base class for all ACME *Vector classes.  The ACME vectors wrap
<BR>
 *          the generic Java vector class and return vectors of specific
<BR>
 *          classes of objects.
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK        Created C++ version
<BR>
 *    4/9/97    RTM       Translated to Java
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */


public class AcmeVector extends acmeLib.AcmeObject {

 //
 // protected members
 //
 protected Vector internalVector;

 //
 // public members
 //

  /**
   * This is a contructor for this class
   */
 public AcmeVector () {
    internalVector = new Vector();
 };

  /**
   * This is a contructor for this class
   * @param initialCapacity the initials size of internalVector
   */
 public AcmeVector (int initialCapacity) {
    internalVector = new Vector(initialCapacity);
 };

  /**
   * This is a contructor for this class
   * @param initialCapacity the initials size of internalVector
   * @param capacityIncrement the capacity increment of internalVector
   */
 public AcmeVector (int initialCapacity, int capacityIncrement) {
    internalVector = new Vector(initialCapacity, capacityIncrement);
 };

  /**
   * This constructor converts the enumeration parameter into the vector "this"
   * @param enum to be converted into the vector "this"
   */
 public AcmeVector (AcmeEnumeration enum) {
    internalVector = new Vector(enum.getSize());
    int i = 0;
    while(enum.moreItems()) {
        AcmeObject curObj = enum.getNextItem();
        this.insertItemAt(curObj, i);
        i++;
    };
 };


  /**
   * @return a new AcmeEnumeration
   */
 public AcmeEnumeration enumerate() {
    return enumerate (new AcmeEnumeration(), null);
 };

  /**
   * @return an enumeration of items
   * @param items AcmeEnumeration
   */
 public AcmeEnumeration enumerate(AcmeEnumeration items) {
    return enumerate (items, null);
 };

  /**
   * @return a new AcmeEnumeration
   * @param items AcmeEnumeration
   * @param filter AcmeObjectFilter
   */
 public AcmeEnumeration enumerate(AcmeEnumeration items,
                                  AcmeObjectFilter filter) {

    Enumeration e = internalVector.elements();

    while (e.hasMoreElements()) {

        // check that each of the elements is not null and passes the
        // filter test.  If so then add it to the items enumeration.
        AcmeObject obj = (AcmeObject) e.nextElement();
        if (obj != null) {
            if (filter != null) {
                if (filter.test(obj)) {
                    items.insertItem(obj);
                };
            } else {
                items.insertItem(obj);
            };
        };  // end if (obj != null)

    };  // end while loop

    return items;
  };

 /**
   * @return a new AcmeEnumeration
   * @param filter AcmeObjectFilter
   */
 public AcmeEnumeration enumerate(AcmeObjectFilter filter)
 {
    Enumeration e = internalVector.elements();
    AcmeEnumeration returnItems = new AcmeEnumeration();

    while (e.hasMoreElements()) {

        // check that each of the elements is not null and passes the
        // filter test.  If so then add it to the items enumeration.
        AcmeObject obj = (AcmeObject) e.nextElement();
        if (obj != null) {
            if (filter != null) {
                if (filter.test(obj)) {
                    returnItems.insertItem(obj);
                };
            } else {
                returnItems.insertItem(obj);
            };
        };  // end if (obj != null)

    };  // end while loop

    return returnItems;
  };


  /**
   * @return the size of internalVector
   */
  public int getSize() { return internalVector.size(); };

  /**
   * @param obj AcmeObject
   * @return whether internalVector contains obj
   */
  public boolean contains (AcmeObject obj) {
    return internalVector.contains(obj);
  };

  /**
   * This method does a shallow copy of this, and puts it into AcmeVector v passed
   * as a parameter
   * @param v AcmeVector, which is to contain a copy of "this"
   */
  public void copyItems (AcmeVector v) {
    for(int i = 0; i < internalVector.size(); i++) {
        v.addItem((AcmeObject) internalVector.elementAt(i));
    };
  };

  /**
   * deep copy of all items in the vector
   * @param newVector new vector
   */
  public void deepCopyItems (AcmeVector newVector)
  {
	  for (int i = 0; i < internalVector.size(); i++)
	  {
		  AcmeObject oldItem = (AcmeObject) internalVector.get(i);
		  AcmeObject newItem = null;
		  if (oldItem instanceof AcmeElementType)
		  {
			  newItem = (AcmeObject) ((AcmeElementType) oldItem).copyType();
		  }
		  else
		  {
			  try
			  {
				  newItem = (AcmeObject) oldItem.getClass().newInstance();
				  oldItem.copy(newItem);
			  }
			  catch (Exception e)
			  {
                  Debug.out("can't deep copy: " + e.getMessage());
	//              e.printStackTrace();  //To change body of catch statement use Options | File Templates.
				  newItem = oldItem; //revert to shallow copy
			  }
		  }
		  newVector.addItem(newItem);
	  }
  }


  // RTM: add a method to do a deep copy of this vector?

  /**
   * This method returns the item at a certain positin in the vector
   * @param index the index of the element to be searched for
   * @return AcmeObject at index
   */
  public AcmeObject getItemAt (int index) {
    return (AcmeObject) internalVector.elementAt(index);
  };

  /**
   * This method removes an item at a certain position from the vector
   * @param index the index of the element to be removed
   */
  public void removeItemAt (int index) {
    internalVector.removeElementAt(index);
  };

  /**
   * This method sets a position in the vector with the value of obj
   * @param obj a new AcmeObject to go at position index
   * @param index the position where obj will go
   */
  public void setItemAt (AcmeObject obj, int index) {
    internalVector.setElementAt(obj,index);
  };

  /**
   * This method removes obj from the vector
   * @param obj the AcmeObject to be removed
   */
  public void removeItem (AcmeObject obj) {
    internalVector.removeElement(obj);
  };

  /**
   * This method removes all the item from the vector
   */
  public void removeAllItems () {
    internalVector.removeAllElements();
  };

  /**
   * This method adds an item to the vector
   * @param obj the AcmeObject to be added
   */
  public void addItem (AcmeObject obj) {
    internalVector.addElement(obj);
  };

  /**
   * This method inserts an item at a specific position in the vector
   * @param obj the AcmeObject to be added at position index
   * @param index the position where obj with be added
   */
  public void insertItemAt (AcmeObject obj, int index) {
    internalVector.insertElementAt(obj, index);
  };

  /**
   * This method returns the index of obj
   * @param obj the AcmeObject, whose index is needed
   */
  public int indexOf (AcmeObject obj) {
    return internalVector.indexOf(obj);
  };

  /**
   * @param obj the AcmeObject, whose index is needed
   * @param index the index of obj
   */
  public int indexOf (AcmeObject obj, int index) {
    return internalVector.indexOf(obj, index);
  };

  /**
   * This method returns the last indes of obj
   * @param obj the AcmeObject, whose last index is needed
   */
  public int lastIndexOf (AcmeObject obj) {
    return internalVector.lastIndexOf(obj);
  };

  /**
   * This method convers an AcmeVector to a String
   * @return the String representaion of this AcmeVector
   */
  public String toString() {
    AcmeEnumeration e = this.enumerate();
    return ("[ AcmeVector = \n" + e.toString() + " ]\n");
  };

}; // end AcmeVector class definition
