/*** ACME language configuration file ***/

package acmeLib.core.globals;

import acmeLib.AcmeObject;
import acmeLib.elements.AcmeElement;
import acmeLib.properties.AcmeProperty;

/** Class:  LanguageConfig<BR>
 *  Purpose: Class that holds a set of constant definitions to determine
<BR>
 *          how the design description should be parsed and processed.
<BR>
 *          This largely takes the places of #ifdef's that would allow
<BR>
 *          us to use the same source for both AcmeLib and ArmaniLib.
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    4/13/98    RTM      * created initial class & file
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
final public class LanguageConfig {

    // Is this an Armani or an Acme design description?
    public static boolean IS_PURE_ACME = false;
    public static boolean IS_ARMANI = true;

    public static final boolean DEBUG = false;
    public static final boolean CHECK_FOR_KEYWORDS = true;
    public static boolean exportChildrenFromSupertype = false;


    // constants for exporting Armani as Acme, and importing Acme as Armani
    // the type name's are for the type definitions, prop names are prefixes
    // for the names of the actual properties storing the Armani constructs
    // as Acme properties.
    public static final String ARMANI_INVARIANT_TYPE_NAME_AS_ACME_PROP = "Armani-InvariantT";
    public static final String ARMANI_HEURISTIC_TYPE_NAME_AS_ACME_PROP = "Armani-HeuristicT";
    public static final String ARMANI_ANALYSIS_TYPE_NAME_AS_ACME_PROP  = "Armani-AnalysisT";

    public static final String ARMANI_INVARIANT_PROP_NAME_AS_ACME_PROP = "Armani-InvariantProp";
    public static final String ARMANI_HEURISTIC_PROP_NAME_AS_ACME_PROP = "Armani-HeuristicProp";
    public static final String ARMANI_ANALYSIS_PROP_NAME_AS_ACME_PROP  = "Armani-AnalysisProp";

    // meta-property name to indicate which type declaration this design rule
    // originally comes from, for tools to decide whether to perform the conversion
    // directly in the instance or to simply refer to a type's design rule
    public static final String DECLARED_IN_TYPENAME_METAPROP  = "Armani-DeclaredInType";


  /**
   * This method returns true if name is an Acme keyword, else false.
   *  not particularly efficient ...
   * @param name tested for AcmeKeyword
   * @return whether name is an AcmeKeyword
   */
    public static final boolean isAcmeKeyword(String name) {
      if (name == null)
        return false;
      else
        return (name.equalsIgnoreCase("system")
                || name.equalsIgnoreCase("component")
                || name.equalsIgnoreCase ("group")
                || name.equalsIgnoreCase ("members")
                || name.equalsIgnoreCase("components")
                || name.equalsIgnoreCase("connector")
                || name.equalsIgnoreCase("connectors")
                || name.equalsIgnoreCase("port")
                || name.equalsIgnoreCase("ports")
                || name.equalsIgnoreCase("role")
                || name.equalsIgnoreCase("roles")
                || name.equalsIgnoreCase("attachment")
                || name.equalsIgnoreCase("attachments")
                || name.equalsIgnoreCase("property")
                || name.equalsIgnoreCase("properties")
                || name.equalsIgnoreCase("representation")
                || name.equalsIgnoreCase("representations")
                || name.equalsIgnoreCase("to")
                || name.equalsIgnoreCase("binding")
                || name.equalsIgnoreCase("bindings")
                || name.equalsIgnoreCase("type")
                || name.equalsIgnoreCase("extended")
                || name.equalsIgnoreCase("with")
                || name.equalsIgnoreCase("extends")
                || name.equalsIgnoreCase("new")
                || name.equalsIgnoreCase("family")
                || name.equalsIgnoreCase("float")
                || name.equalsIgnoreCase("int")
                || name.equalsIgnoreCase ("integer")
                || name.equalsIgnoreCase("string")
                || name.equalsIgnoreCase("boolean")
                || name.equalsIgnoreCase("bool")
                || name.equalsIgnoreCase("set")
                || name.equalsIgnoreCase("sequence")
                || name.equalsIgnoreCase("seq")
                || name.equalsIgnoreCase("record")
                || name.equalsIgnoreCase("enum")
                || name.equalsIgnoreCase("any")
                || name.equalsIgnoreCase("false")
                || name.equalsIgnoreCase("true")
               );
    };


  /**
   * true if <em> name <\em> is an Armani-specific keyword, else false.
   *  Note that this does not redundantly check for Acme keywords, so
   *  it should be used as a complement to the isAcmeKeyword() method.
   * @param name tested for AcmeKeyword
   * @return whether name is an AcmeKeyword
   */
    public static final boolean isArmaniKeyword(String name) {
      if (name == null)
        return false;
      else
        return (name.equalsIgnoreCase("analysis")
                || name.equalsIgnoreCase("invariant")
                || name.equalsIgnoreCase("heuristic")
                || name.equalsIgnoreCase("import")
                || name.equalsIgnoreCase("exists")
                || name.equalsIgnoreCase("forall")
                || name.equalsIgnoreCase("external")
                || name.equalsIgnoreCase("design")
                || name.equalsIgnoreCase("and")
                || name.equalsIgnoreCase("or")
                || name.equalsIgnoreCase("in")
                || name.equalsIgnoreCase("select")
                || name.equalsIgnoreCase("collect")
                || name.equalsIgnoreCase("attachedPorts")
                || name.equalsIgnoreCase("attachedRoles")
                || name.equalsIgnoreCase("unique")
        );
    };
    
    public static final boolean shouldExport (AcmeObject child) {
    	if (!exportChildrenFromSupertype) {
	    	if (child instanceof AcmeElement) {
				AcmeElement elt = (AcmeElement )child;
				return !elt.isSameAsInherited();    
	    	}
	    	else if (child instanceof AcmeProperty) {
				AcmeProperty prop = (AcmeProperty) child;
				return !prop.isSameAsInherited ();
				
			}
    	}
    	return true;
    }

};
