/*
* Copyright (c) 1996, 1997
* Carnegie Mellon University ABLE Project
* David Garlan, Robert Monroe, and Andrew Kompanek.
*
* All Rights Reserved.
*
* Permission to use, copy, modify and distribute this software and its
* documentation is hereby granted, provided that both the copyright
* notice and this permission notice appear in all copies of the
* software, derivative works or modified versions, and any portions
* thereof.
*
* CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
* CONDITION.  CARNEGIE MELLON AND THE ABLE PROJECT DISCLAIMS ANY
* LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM
* THE USE OF THIS SOFTWARE.
*
* The ABLE project requests users of this software to return any
* improvements or extensions that they make and grant Carnegie Mellon
* the rights to redistribute these changes.  Return such changes to:
*
*      David Garlan, Robert Monroe, or Drew Kompanek
*      School of Computer Science
*      Carnegie Mellon University
*      5000 Forbes Ave.
*      Pittsburgh PA, 15218
*
*      acme-info@cs.cmu.edu
*
*/

package acmeLib.core.io;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;

import acmeLib.AcmeObject;

/**
 *  Class: AcmeOutputStream
 <BR>
 *  Purpose:  Standard output stream for exporting ACME.  The AcmeOutputStream
<BR>
 *          encapsulates a java ByteArrayOutputStream, and provides access
<BR>
 *          to all of the methods of ByteArrayOutputStream, including the
<BR>
 *          writeTo(OutputStream s) method that dumps the contents of the
<BR>
 *          ByteArrayOutputStream object to the stream passed as a parameter.
<BR>
 *          Thus, to write an ACME description to a file, the recommended
<BR>
 *          procedure is to write to an AcmeOutputStream and then send the
<BR>
 *          array on to a FileOutputStream (or System.out) with the
<BR>
 *          writeTo(...) method.
<BR><BR>
 *
 *          e.g.    AcmeOutputStream s = new AcmeOutputStream();
<BR>
 *                  design.exportAcme(s);
<BR>
 *                  s.writeTo(System.out);   // or some other file stream.
<BR><BR>
 *
 * Notes: The write(String s) and writeln(String s) are not generic stream
<BR>
 *        writing routines.  They are designed explicitly for use with
<BR><BR>
 *        the AcmeLib's exporting routines.
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97     DK      Created C++ version
<BR>
 *    4/15/97    RTM     Translated to Java
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */


public class AcmeOutputStream extends AcmeObject {

    private class $ByteArrayOutputStream extends ByteArrayOutputStream {
        public int getCount () {
            return count;
        }
    };

    protected RegionManager regionManager;

  //
  // protected members
  //
  protected int next_level;
  protected int level;
  protected StringBuffer last_line;
  protected boolean end_of_line;
  protected final int indentDepth = 4;
  protected int currentLine = 1;

  // RTM -- found a bug that ByteArrayOutputStream seems to hurl when it gets
  //        around 60k bytes.  So, whenever a buffer fills up to MAX_BUF_SIZE
  //        put the buffer into the vector of filledBufs and create a new
  //        ByteArrayOutputStream to put subsquent info into.
  protected java.util.Vector filledBufs;
  protected final int MAX_BUF_SIZE = 50000;

  protected $ByteArrayOutputStream outStream;

  // prefix will be implemented in a subclass
  // private String prefix;

  //
  // public members
  //

  /**
   * This is the constructor for this class
   */
  public AcmeOutputStream () {
	  level = 0;
	  next_level = 0;
	  last_line = new StringBuffer();
	  end_of_line = false;

      filledBufs = new java.util.Vector();
	  outStream = new $ByteArrayOutputStream();
      regionManager = new RegionManager ();
  };


  /**
   * This method writes the last_line to outStream and calls flush() on outStream.
   * @exception java.io.IOException when error occurs when writing
   */
  public void flush() throws IOException {
      // put last_line into the stream and flush it
      for (int i=0;i<(level * indentDepth);i++) { outStream.write(' '); };
      if (end_of_line) {
	    last_line.append('\n');
        currentLine++;
 
	  };

	  // create a temp byte array to dump the string into.
      // then write it to the stream.  terribly inefficient.
      byte[] buf = new byte[last_line.length()];
      (last_line.toString()).getBytes(0,last_line.length(),buf,0);

	  outStream.write(buf);
	  last_line = new StringBuffer();
	  level = next_level;

	  outStream.flush();

	  // RTM -- deal with an apparent bug in the ByteArrayOutputStream.  It seems
	  // to die around 60k bytes.  Because we don't know how long it will be until
	  // it's flushed again, replace outStream with a new ByteArrayOutputStream and
	  // save the old version.
      if (outStream.size() > MAX_BUF_SIZE) {
        filledBufs.addElement(outStream);
        outStream = new $ByteArrayOutputStream();
	  };
  };

  /**
   * This method flushes the buffer, then closes down.
   * @exception java.io.IOException when error occurs when writing
   */
  public void close() throws IOException {
     this.flush();
     outStream.close();
  };

  public void indent() {
    next_level++;
  };

  public void unIndent() {
	  if (next_level!=0) next_level--;
  };

  /**
   * This method copies str to the buffer, followed by a newline.
   * @param str the String to be written
   * @exception java.io.IOException when error occurs when writing
   */
  public void writeln (String str) throws IOException {
	  int i;
	  if (end_of_line) {
	    last_line.append('\n');
        currentLine++;
	    // indent
	    for (i=0;i<(level * indentDepth);i++) { outStream.write(' '); };

        // create a temp byte array to dump the string into.
        // then write it to the stream.  terribly inefficient.
        byte[] buf = new byte[last_line.length()];
        (last_line.toString()).getBytes(0,last_line.length(),buf,0);

        outStream.write(buf);
	    last_line = new StringBuffer();
        level = next_level;
	  }

	  last_line.append(str);
	  end_of_line = true;

  };

  /**
   * @exception java.io.IOException when error occurs when writing
   */
  public void writeln () throws IOException { this.writeln(""); };


  /**
   * This method writes str to the buffer, properly indented.  Don't add a newline
   * @param  str the String to be written
   * @exception java.io.IOException  when error occurs when writing
   */
  public void write (String str) throws IOException {
	  int i;
	  if (end_of_line) {
        last_line.append('\n');
        currentLine++;
        // indent
        for (i=0; i < (level * indentDepth); i++) { outStream.write(' '); };

        // create a temp byte array to dump the string into.
        // then write it to the stream.  terribly inefficient.
        byte[] buf = new byte[last_line.length()];
        (last_line.toString()).getBytes(0,last_line.length(),buf,0);

        outStream.write(buf);
	    last_line = new StringBuffer();
	    level = next_level;
	    end_of_line = false;
	  }
	  last_line.append(str);
  };

  /**
   * ByteArrayOutputStream methods that the AcmeOutputStream class passes on
   */
  public synchronized void reset()  { outStream.reset(); };
  /**
   * @return the size of the output stream
   */
  public int size() { return outStream.size(); };
  /**
   * @return the output converted to byteArray
   */
  public String toString() {
    StringBuffer outBuf = new StringBuffer();
    while (filledBufs.size() > 0) {
        $ByteArrayOutputStream curOutStream =
                ($ByteArrayOutputStream) filledBufs.elementAt(0);
        outBuf.append(curOutStream.toString());
        filledBufs.removeElementAt(0);
    };

    // now append the tail of the output from the current outputStream.
    outBuf.append(outStream.toString());

    String ret = outBuf.toString();
    return ret;
  };


  /**
   * @param hibyte the integer to be converted to String
   * @return hibyte converted to String
   */
  public String toString(int hibyte) { return outStream.toString(hibyte); };

  /**
   * @param out OutputStream
   * @exception java.io.IOException thrown by writeTo
   */
  public synchronized void writeTo(OutputStream out) throws IOException {
    // write out all of the previously filled buffers.
    while (filledBufs.size() > 0) {
        $ByteArrayOutputStream curOutStream =
                ($ByteArrayOutputStream) filledBufs.elementAt(0);
        curOutStream.writeTo(out);
        filledBufs.removeElementAt(0);
    };

    // now append the tail of the output from the current outputStream.
    outStream.writeTo(out);
  };
  
  protected int getCurrentPosition () {
      
      int pos = outStream.getCount () + last_line.toString ().length() + indentDepth*level;
      Iterator i = filledBufs.iterator();
      while (i.hasNext ()) {
          $ByteArrayOutputStream s = ($ByteArrayOutputStream )i.next ();
          pos += s.getCount();
      }
      if (end_of_line) {pos++;} 
      return pos;
  }
  
  protected int getCurrentLine () {
      int pos = currentLine;
      if (end_of_line) pos++;
      return pos;
  }
  
  public void beginRegion (AcmeObject o) {
    int pos = getCurrentPosition ();
    Region r =  new Region ();;
    r.setBeginIndex (pos);
    r.setBeginLine (getCurrentLine ());
    regionManager.addRegion (o, r);
  }
 
  public void endRegion (AcmeObject o) {
      Region r = regionManager.getRegion (o); 
      r.setEndIndex(getCurrentPosition ());
  }
  
  public RegionManager getRegionManager () {
      return regionManager;
  }
  

};
