package acmeLib.designRules;

import java.util.ArrayList;
import java.util.List;
import acmeLib.AcmeType;
import acmeLib.AcmeTypeParameter;
import acmeLib.Debug;
import acmeLib.armaniParser.ArmaniParser;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExpressionOperatorT;
import acmeLib.designRules.feedback.BooleanExpressionEvaluationFeedback;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.properties.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: BooleanExpressionNode<br>
 * Purpose: root of a (possibly 1-node) tree that represents a boolean<br>
 *          expression.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    10/7/97    RTM       Created<br>
 *    10/27/97   RTM       Added support for exception handling<br>
 *    2/2/98     RTM       Modified type storage to store AcmeTypes<br>
 *                          instead of ExpressionTypeT objects.<br>
 *<br>
 * @author Bob Monroe
 */
public class BooleanExpressionNode extends ExpressionNode
{
  //
  // public members
  //
/**
 * This is the constructor for this class
 * @param id The string
 */
  public BooleanExpressionNode(int id) {
    super(id);
    this.setType(ConstantTypes.BOOLEAN_TYPE);
  }


/**
 * This is the constructor for this class
 * @param copyOf The boolean expression node
 */
  public BooleanExpressionNode(BooleanExpressionNode copyOf) {
    super(copyOf);
  }


/**
 * @param id The ID
 * @return The node
 */
  public static Node jjtCreate(int id) {
    return new BooleanExpressionNode(id);
  }


/**
 * cloneExpressionTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.
 * @return The copy of the expression tree
 */
  public ExpressionNode cloneExpressionTree() {
    BooleanExpressionNode clone = new BooleanExpressionNode(this);
    this.doCloneExpressionTree(clone);
    return clone;
  };


/**
 * Check to see if the two expression nodes are equal
 * @param test The test node
 * @return true if the test node are equal to this node
 */
  public boolean equals(Object test) {
    if (!(test instanceof ExpressionNode)) {
        if (test instanceof AcmeBoolean) {
            try {
                boolean thisVal = this.evaluateAsBoolean();
                return (thisVal == ((AcmeBoolean) test).getValue());
            } catch (Exception exc) {
            };
        }
            return false;
    }
        try {
            boolean testVal = ((ExpressionNode) test).evaluateAsBoolean();
            return (testVal == this.evaluateAsBoolean());
        } catch (Exception exc) {
        };
        return false;
  };


/**
 *
 * Methods added by bmonroe to implement boolean expression typechecking
 * and evaluation
 *
 *
 *
 * Traverses the expression tree below this node to compute the type
 * of the expression.  It sets the type for this ExpressionNode and
 * determines whether the claimed type of the node matches it's
 * subexpressions and operator.
 *
 * This method has the side-effect of setting the type of this node
 * and requesting that all of its children expression nodes do the
 * same.
 *
 * Returns true if it can infer the type of the expression and
 * the expression typechecks, else returns false.
 * @return true if it can infer the type of expression and the expression typechecks
 */
public boolean typecheck() {

    // first infer the type of all of the children nodes.  Recurse
    // until leaf nodes are found.  If it can't infer the type of
    // the children or any children don't typecheck then return false
    // to indicate a type error.
    for (int i = 0; i < this.jjtGetNumChildren(); i++) {
        if (((ExpressionNode) this.jjtGetChild(i)).typecheck() == false)
            return false;
    }

    // a BooleanExpressionNode must have a boolean operator
    if (!(this.hasBooleanOperator())) {
        ExpressionTypecheckError err =
            new ExpressionTypecheckError(this, "Expression does not contain a boolean operator", this);
        AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
        return false;
    }

    if (this.jjtGetNumChildren() == 1) {
        // then this is a unary expression, if the child is a boolean
        // expression, then set the type to the type of child node (0).
        // Unary boolean operators can't change the type of the expression.

        AcmeType childType = ((ExpressionNode) jjtGetChild(0)).getType();
        if (!(childType instanceof BooleanType)) {
            this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
            ExpressionTypecheckError err =
                new ExpressionTypecheckError(
                    (ExpressionNode) jjtGetChild(0),
                    "Operand is not a boolean expression",
                    this);
            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
            return false;
        }
        else {
            this.setType(ConstantTypes.BOOLEAN_TYPE);
            // can't prove that it does not typecheck, so assume that it does.
            return true;
        }

    }

    AcmeType lChildType, rChildType;
    if (this.jjtGetNumChildren() == 2) {
        // then it is a binary boolean expression.  check that the
        // triple (operator, lChildType, rChildType) is valid.
        lChildType = ((ExpressionNode) this.jjtGetChild(0)).getType();
        rChildType = ((ExpressionNode) this.jjtGetChild(1)).getType();

        if ((lChildType instanceof BooleanType)
            && (rChildType instanceof BooleanType)) {
                if (this.hasBooleanLogicOperator()) {
                    // then this is a boolean expression
                    this.setType(ConstantTypes.BOOLEAN_TYPE);
                    return true;
                }
                else {
                    ExpressionTypecheckError err =
                        new ExpressionTypecheckError (this, this.getOperator().toString() + " is not a boolean operator", this);
                    AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                    return false;
                }
        }
        else if (
            ((lChildType instanceof IntType) || (lChildType instanceof FloatType))
                && ((rChildType instanceof IntType) || (rChildType instanceof FloatType))) {
                    if (hasNumericComparisonOperator()) {
            // then this is a valid numeric comparison expression
            this.setType(ConstantTypes.BOOLEAN_TYPE);
            return true;
                    }
                    else {
                             ExpressionTypecheckError err =
                                 new ExpressionTypecheckError (this, this.getOperator().toString() + " is not a numeric comparison operator", this);
                             AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                             return false;
                         }
        }
        else if (
            ((lChildType instanceof StringType) || (lChildType instanceof EnumType))
                && ((rChildType instanceof StringType) || (rChildType instanceof EnumType))) {
                    if (this.hasStringLogicOperator()) {
            // then this is a string comparison expression
            this.setType(ConstantTypes.BOOLEAN_TYPE);
            return true;
                    }
                    else {
                        ExpressionTypecheckError err =
                            new ExpressionTypecheckError (this, this.getOperator().toString() + " is not a string operator", this);
                        AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                        return false;
                    }
        }
        else if (
            ((lChildType instanceof AcmeElementType)
                && (rChildType instanceof AcmeElementType)) ||
			  ((lChildType instanceof AcmeTypeParameter) && (rChildType instanceof AcmeElementType)) ||
			  ((lChildType instanceof AcmeElementType) && (rChildType instanceof AcmeTypeParameter)) ||
			  ((lChildType instanceof AcmeTypeParameter) && (lChildType instanceof AcmeTypeParameter))) {
                    if (this.hasEqualityComparisonOperator()) {
            // then this is an element comparison expression
            this.setType(ConstantTypes.BOOLEAN_TYPE);
            return true;
                    }
                    else {
                        ExpressionTypecheckError err =
                            new ExpressionTypecheckError (this, this.getOperator().toString() + " is not an element operator", this);
                        AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                        return false;
                    }
        }
        
        else if (
            (lChildType instanceof SetType)
                && (rChildType instanceof SetType)) {
                    if (this.hasEqualityComparisonOperator()) {
            // then this is an element comparison expression
            this.setType(ConstantTypes.BOOLEAN_TYPE);
            return true;
                    }
                    else {
                        ExpressionTypecheckError err =
                            new ExpressionTypecheckError (this, this.getOperator().toString() + " is not a set operator", this);
                        AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                        return false;
                    }
        }
        // else this is an invalid type for a boolean node.
        this.setType(ConstantTypes.UNRESOLVABLE_TYPE);

 
        if (lChildType != rChildType) {

            ExpressionTypecheckError err =
                new ExpressionTypecheckError((ExpressionNode) this.jjtGetChild(1), "Children must be of the same type.", this);
            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
        }


        return false;
    };

    // should never get to here.  return false if we can't show that the
    // type is legal.
    return false;
}




/**
 * override ExpressionNode's evaluateAsBoolean method to return appropriate
 * values.
 * @return boolean
 * @exception DesignRuleEvalException If it cannot be evaluated as boolean
 */
  public boolean evaluateAsBoolean() throws DesignRuleEvalException {
    return this.evaluate();
  };

/**
 * Repeat for evaluateAsString, which should return the string
 * value of the evaluation result
 * @return string
 * @exception DesignRuleEvalException If it cannot be evaluated as string
 */
  public String evaluateAsString() throws DesignRuleEvalException {
    return String.valueOf(this.evaluate());
  };

/**
 * evaluate() recursively evaluates the value of the left and right
 * children expressions, applies the operator to these values, and
 * returns that computed (boolean) value.
 * evaluate() assumes that the expression it is evaluating is type-correct
 * and that it has had all of its type and instance references properly
 * resolved.  These are preconditions for the proper functioning of the method.
 * @return The boolean value
 * @exception DesignRuleEvalException If it is type-incorrect
 */
  public boolean evaluate() throws DesignRuleEvalException {
  	Debug.trace_call(this);
  	boolean rv = false;

    // if the expression doesn't typecheck then it can't be evaluated,
    // exit now with an exception.
    if (!(this.typecheck())) {
      	Debug.trace_return(this, rv);
        throw new ExpressionTypeException(null, this,
             ConstantTypes.BOOLEAN_TYPE, this.getType(),
            "Boolean design rule expression does not typecheck");
    }

    // check if this is a unary "not" operation (not is the only unary
    // boolean operator)
    if (op == ExpressionOperatorT.NOT_OP) {
        if (jjtGetChild(0) == null) {
        	Debug.trace_return (this, rv);
            throw new DesignRuleEvalException("Attempted to apply the 'not' "
                        + "operator to a null term");
        }
            if (jjtGetChild(0) instanceof ExpressionNode) {
                if (!(((ExpressionNode) jjtGetChild(0)).getType()
                        instanceof BooleanType))
                {
                	Debug.trace_return (this, rv);
                    String opMsg = new String("Attempted to apply the "
                             + "'not' operator to a non-boolean expression");
                    throw new ExpressionTypeException(null, this,
                            ConstantTypes.BOOLEAN_TYPE, this.getType(), opMsg);
                } else {
                    boolean retVal = !((ExpressionNode) jjtGetChild(0)).evaluateAsBoolean();
                    if (DesignRule.isRememberVisited()) {
                    	List l = getEvaluationData ();
                    	l.add (new BooleanExpressionEvaluationFeedback (this,  new Boolean (retVal)));
                    }
                    Debug.trace_return (this, retVal);
					return retVal;
                }
            } else {
               Debug.trace_return (this, rv);
                throw new DesignRuleEvalException("Attempted to apply the "
                             + "'not' operator to an object that is not "
                             + "an ExpressionNode");
			}
    };

    List evaluationList = null;
    if (DesignRule.isRememberVisited ()) evaluationList = getEvaluationData();
    // else see if this is an operation that takes two boolean arguments
    // and actually has two boolean children nodes available.
    BooleanExpressionNode lBoolChild, rBoolChild;
    ReferenceNode tmpRefNode;
    QuantifiedExpressionNode tmpQuanNode;
    if (jjtGetChild(0) instanceof BooleanExpressionNode) {
        lBoolChild = (BooleanExpressionNode) jjtGetChild(0);
    } else if (jjtGetChild(0) instanceof ReferenceNode) {
        tmpRefNode = (ReferenceNode) jjtGetChild(0);
        if (tmpRefNode.getType() instanceof BooleanType) {
        	boolean leftEval = tmpRefNode.evaluateAsBoolean();
            if (leftEval == true)
                lBoolChild = BooleanLiteralNode.TRUE_NODE;
            else
                lBoolChild = BooleanLiteralNode.FALSE_NODE;
        } else {
            lBoolChild = null;
        };
    } else if (jjtGetChild(0) instanceof QuantifiedExpressionNode) {  // zensoul.ADDED (1/03/03)
        tmpQuanNode = (QuantifiedExpressionNode) jjtGetChild(0);
        boolean leftEval = tmpQuanNode.evaluateAsBoolean();
        if (leftEval== true)
            lBoolChild = BooleanLiteralNode.TRUE_NODE;
        else
            lBoolChild = BooleanLiteralNode.FALSE_NODE;
    } else {
        lBoolChild = null;
    };

    if (jjtGetChild(1) instanceof BooleanExpressionNode) {
        rBoolChild = (BooleanExpressionNode) jjtGetChild(1);
    } else if (jjtGetChild(1) instanceof ReferenceNode) {
        tmpRefNode = (ReferenceNode) jjtGetChild(1);
        if (tmpRefNode.getType() instanceof BooleanType) {
            if (tmpRefNode.evaluateAsBoolean() == true)
                rBoolChild = BooleanLiteralNode.TRUE_NODE;
            else
                rBoolChild = BooleanLiteralNode.FALSE_NODE;
        } else {
            rBoolChild = null;
        };
    } else if (jjtGetChild(1) instanceof QuantifiedExpressionNode) {  // zensoul.ADDED (1/03/03)
        tmpQuanNode = (QuantifiedExpressionNode) jjtGetChild(1);
        if (tmpQuanNode.evaluateAsBoolean() == true)
            rBoolChild = BooleanLiteralNode.TRUE_NODE;
        else
            rBoolChild = BooleanLiteralNode.FALSE_NODE;
    } else {
        rBoolChild = null;
    };

    if ( (lBoolChild != null) && (rBoolChild != null) ) {
        // then we have a binary boolean op boolean expression.  evaluate it.
        boolean lBoolChildEval = lBoolChild.evaluate();
		boolean rBoolChildEval = rBoolChild.evaluate();
		Boolean retVal = null;
		
		if (op == ExpressionOperatorT.AND_OP) {
            retVal = new Boolean ( lBoolChildEval && rBoolChildEval);
        } else if (op == ExpressionOperatorT.OR_OP) {
            retVal = new Boolean ( lBoolChildEval || rBoolChildEval);
        } else if (op == ExpressionOperatorT.IMPLIES_OP) {
            if (lBoolChildEval == false)
                retVal = Boolean.TRUE;
            else
                retVal = new Boolean (rBoolChildEval);
        } else if (op == ExpressionOperatorT.IFF_OP) {
            retVal = new Boolean ((lBoolChildEval == rBoolChildEval));
        } else if (op == ExpressionOperatorT.EQUALS_OP) {
            retVal = new Boolean ( (lBoolChildEval == rBoolChildEval));
        } else if (op == ExpressionOperatorT.NOT_EQUALS_OP) {
            retVal = new Boolean ( (lBoolChildEval != rBoolChildEval));
        }
        
        if (retVal != null) {
        	if (DesignRule.isRememberVisited()) {
    			evaluationList.add (new BooleanExpressionEvaluationFeedback (this, retVal));
    		}
        	return retVal.booleanValue();
        } else {
        // else we have an illegal operator for two booleans. throw an exception
            Debug.trace_return (this, retVal.booleanValue());
            String opMsg = new String("Illegal operator "
                + this.getOperator().toString() + " for a boolean expression.");
            throw new ExpressionTypeException(null, this,
                    ConstantTypes.BOOLEAN_TYPE, this.getType(), opMsg);
        }
    }

    // else this is an operation that takes two numerical, string, element,
    // property reference, or set arguments, compares them, and returns
    // the boolean evaluation of the elements based on the node's operator.

    // evaluate the left and right expressions, determine whether they
    // are stored as integers, floats, strings, elements, or property references
    // and then evaluate the boolean value of this node appropriately.
    boolean lChildIsInt = false;
    boolean lChildIsFloat = false;
    boolean lChildIsString = false;
    boolean lChildIsElement = false;
    boolean lChildIsPropertyRef = false;
    boolean lChildIsSet = false;
    boolean rChildIsInt = false;
    boolean rChildIsFloat = false;
    boolean rChildIsString = false;
    boolean rChildIsElement = false;
    boolean rChildIsPropertyRef = false;
    boolean rChildIsSet = false;

    int lChildInt = 0;
    int rChildInt = 0;
    float lChildFloat = (float) 0.0;
    float rChildFloat = (float) 0.0;
    String lChildString = null;
    String rChildString = null;
    AcmeElement lChildElement = null;
    AcmeElement rChildElement = null;
    AcmeProperty lChildPropertyRef = null;
    AcmeProperty rChildPropertyRef = null;
    SetExpressionNode lChildSet = null;
    SetExpressionNode rChildSet = null;

    NumericExpressionNode lNumChild, rNumChild;
    StringLiteralNode lStrChild, rStrChild;

    // [SWC] DEBUG: added trace values

    // find the value of the left child
    if (jjtGetChild(0) instanceof NumericExpressionNode) {
        try {
            lNumChild = (NumericExpressionNode) jjtGetChild(0);
            if (lNumChild.getType() instanceof IntType) {
                lChildIsInt = true;
                lChildInt = lNumChild.evaluateAsInt();
            } else if (lNumChild.getType() instanceof FloatType) {
                lChildIsFloat = true;
                lChildFloat = lNumChild.evaluateAsFloat();
            }
        }
        catch (java.lang.ArithmeticException arExc) {
            // a divide by zero exception, typically.  bail in any case.
          	Debug.trace_return(this, rv);
            throw new DesignRuleEvalException("Arithmetic error encountered: "
                        + arExc.getMessage());
        };
    } else if (jjtGetChild(0) instanceof SetExpressionNode) {
        lChildSet = (SetExpressionNode) jjtGetChild(0);
        lChildIsSet = true;
    } else if (jjtGetChild(0) instanceof ReferenceNode) {
  		//System.out.println(((ReferenceNode)jjtGetChild(0)).getIdentifier());
        tmpRefNode = (ReferenceNode) jjtGetChild(0);
        if (tmpRefNode.getType() instanceof IntType) {
            lChildIsInt = true;
            lChildInt = tmpRefNode.evaluateAsInt();
        }
        else if (tmpRefNode.getType() instanceof FloatType) {
            lChildIsFloat = true;
            lChildFloat = tmpRefNode.evaluateAsFloat();
        }
        else if (tmpRefNode.getType() instanceof StringType) {
            lChildIsString = true;
            lChildString = tmpRefNode.evaluateAsString();
        }
        else if (tmpRefNode.getType() instanceof EnumType) {
            lChildIsString = true;
            lChildString = tmpRefNode.evaluateAsString();
        }
        else if (tmpRefNode.getType() instanceof AcmeElementType) {
            lChildIsElement = true;
            lChildElement = tmpRefNode.evaluateAsElement();
        }
        else if (tmpRefNode.getType () instanceof AcmeTypeParameter) {
        	lChildIsElement = true;
        	lChildElement = (AcmeElement )tmpRefNode.evaluateAsTypeRef();
        }
        else if (tmpRefNode.getType() instanceof PropertyReferenceType) {
            lChildIsPropertyRef = true;
            lChildPropertyRef = tmpRefNode.evaluateAsPropertyRef();
        }
        else if (tmpRefNode.getType() instanceof SetType) {
            lChildIsSet = true;
            lChildSet = tmpRefNode.evaluateAsSet();
        };
    } else if (jjtGetChild(0) instanceof StringLiteralNode) {
        lStrChild = (StringLiteralNode) jjtGetChild(0);
        lChildIsString = true;
        lChildString = lStrChild.evaluateAsString();
    };


    // find the value of the right child
    if (jjtGetChild(1) instanceof NumericExpressionNode) {
        try {
            rNumChild = (NumericExpressionNode) jjtGetChild(1);
            if (rNumChild.getType() instanceof IntType) {
                rChildIsInt = true;
                rChildInt = rNumChild.evaluateAsInt();
            } else if (rNumChild.getType() instanceof FloatType) {
                rChildIsFloat = true;
                rChildFloat = rNumChild.evaluateAsFloat();
            }
        }
        catch (java.lang.ArithmeticException arExc) {
            // a divide by zero exception, typically.  bail in any case.
          	Debug.trace_return(this, rv);
            throw new DesignRuleEvalException("Arithmetic error encountered: "
                        + arExc.getMessage());
        };

    } else if (jjtGetChild(1) instanceof SetExpressionNode) {
        rChildSet = (SetExpressionNode) jjtGetChild(1);
        rChildIsSet = true;
        Debug.trace(rChildSet.toString());
    } else if (jjtGetChild(1) instanceof ReferenceNode) {
        tmpRefNode = (ReferenceNode) jjtGetChild(1);
        StringBuffer buf = new StringBuffer();
        if (ArmaniParser.isTracingEnabled())
        	buf.append(tmpRefNode.toString()).append(" == ");
        if (tmpRefNode.getType() instanceof IntType) {
            rChildIsInt = true;
            rChildInt = tmpRefNode.evaluateAsInt();
            if (ArmaniParser.isTracingEnabled()) buf.append(rChildInt);
        }
        else if (tmpRefNode.getType() instanceof FloatType) {
            rChildIsFloat = true;
            rChildFloat = tmpRefNode.evaluateAsFloat();
            if (ArmaniParser.isTracingEnabled()) buf.append(rChildFloat);
        }
        else if (tmpRefNode.getType() instanceof StringType) {
            rChildIsString = true;
            rChildString = tmpRefNode.evaluateAsString();
            if (ArmaniParser.isTracingEnabled()) buf.append(rChildString);
        }
        else if (tmpRefNode.getType() instanceof EnumType) {
            rChildIsString = true;
            rChildString = tmpRefNode.evaluateAsString();
            if (ArmaniParser.isTracingEnabled()) buf.append(rChildString);
        }
        else if (tmpRefNode.getType() instanceof AcmeElementType) {
            rChildIsElement = true;
            rChildElement = tmpRefNode.evaluateAsElement();
            if (ArmaniParser.isTracingEnabled()) buf.append(rChildElement.getName());
        }
        else if (tmpRefNode.getType () instanceof AcmeTypeParameter) {
        	rChildIsElement = true;
        	rChildElement = (AcmeElement )tmpRefNode.evaluateAsTypeRef();
            if (ArmaniParser.isTracingEnabled()) buf.append(rChildElement.getName());
        }
        else if (tmpRefNode.getType() instanceof PropertyReferenceType) {
            rChildIsPropertyRef = true;
            rChildPropertyRef = tmpRefNode.evaluateAsPropertyRef();
            if (ArmaniParser.isTracingEnabled()) buf.append(rChildPropertyRef.getName());
        }
        else if (tmpRefNode.getType() instanceof SetType) {
            rChildIsSet = true;
            rChildSet = tmpRefNode.evaluateAsSet();
            if (ArmaniParser.isTracingEnabled()) buf.append(rChildSet.toString());
        };
        Debug.trace(buf.toString());
    } else if (jjtGetChild(1) instanceof StringLiteralNode) {
        rStrChild = (StringLiteralNode) jjtGetChild(1);
        rChildIsString = true;
        rChildString = rStrChild.evaluateAsString();
        Debug.trace(rChildString);
    };

    // sanity check that we've acquired values for at least two of
    // the children.
    if (!(lChildIsFloat || lChildIsInt || lChildIsString || lChildIsElement
             || lChildIsPropertyRef || lChildIsSet)
        || !(rChildIsFloat || rChildIsInt || rChildIsString
             || rChildIsElement || rChildIsPropertyRef || rChildIsSet) )
    {
      	Debug.trace_return(this, rv);
        String msg = new String("Illegal operands for boolean expression");
        throw new ExpressionTypeException(null, this,
            ConstantTypes.BOOLEAN_TYPE, this.getType(), msg);
    };

    Boolean retVal = null;
    // perform the operation on the values of the children.
    if (op == ExpressionOperatorT.LESS_OR_EQUAL_OP) {
        if (lChildIsInt && rChildIsInt) {
            retVal = Boolean.valueOf (lChildInt <= rChildInt);
        } else if (lChildIsFloat && rChildIsFloat) {
            retVal = Boolean.valueOf (lChildFloat <= rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
            retVal = Boolean.valueOf  ( ((float) lChildInt) <= rChildFloat );
        } else if (lChildIsFloat && rChildIsInt) {
            retVal = Boolean.valueOf ( lChildFloat <= ((float) rChildInt) );
        };
    } else if (op == ExpressionOperatorT.GREATER_OR_EQUAL_OP) {
        if (lChildIsInt && rChildIsInt) {
            retVal = Boolean.valueOf (lChildInt >= rChildInt);
        } else if (lChildIsFloat && rChildIsFloat) {
            retVal = Boolean.valueOf (lChildFloat >= rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
            retVal = Boolean.valueOf ( ((float) lChildInt) >= rChildFloat );
        } else if (lChildIsFloat && rChildIsInt) {
            retVal = Boolean.valueOf ( lChildFloat >= ((float) rChildInt) );
        };
    } else if (op == ExpressionOperatorT.LESS_THAN_OP) {
        if (lChildIsInt && rChildIsInt) {
            retVal = Boolean.valueOf  (lChildInt < rChildInt);
        } else if (lChildIsFloat && rChildIsFloat) {
        	retVal = Boolean.valueOf (lChildFloat < rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
        	retVal = Boolean.valueOf ( ((float) lChildInt) < rChildFloat );
        } else if (lChildIsFloat && rChildIsInt) {
        	retVal = Boolean.valueOf ( lChildFloat < ((float) rChildInt) );
        };
    } else if (op == ExpressionOperatorT.GREATER_THAN_OP) {
        if (lChildIsInt && rChildIsInt) {
        	retVal = Boolean.valueOf (lChildInt > rChildInt);
        } else if (lChildIsFloat && rChildIsFloat) {
        	retVal = Boolean.valueOf (lChildFloat > rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
        	retVal = Boolean.valueOf ( ((float) lChildInt) > rChildFloat );
        } else if (lChildIsFloat && rChildIsInt) {
        	retVal = Boolean.valueOf ( lChildFloat > ((float) rChildInt) );
        };
    } else if (op == ExpressionOperatorT.EQUALS_OP) {
        if (lChildIsInt && rChildIsInt) {
        	retVal = Boolean.valueOf (lChildInt == rChildInt);
        } else if (lChildIsFloat && rChildIsFloat) {
        	retVal = Boolean.valueOf (lChildFloat == rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
        	retVal = Boolean.valueOf ( ((float) lChildInt) == rChildFloat );
        } else if (lChildIsFloat && rChildIsInt) {
        	retVal = Boolean.valueOf ( lChildFloat == ((float) rChildInt) );
        } else if (lChildIsString && rChildIsString) {
        	retVal = Boolean.valueOf(lChildString.equals(rChildString));
        } else if (lChildIsElement && rChildIsElement) {
        	retVal = Boolean.valueOf (lChildElement.equals(rChildElement));
        } else if (lChildIsPropertyRef && rChildIsPropertyRef) {
        	retVal = Boolean.valueOf( lChildPropertyRef.equals(rChildPropertyRef));
        } else if (lChildIsSet && rChildIsSet) {
        	retVal = Boolean.valueOf( lChildSet.equals(rChildSet));
        }
    } else if (op == ExpressionOperatorT.NOT_EQUALS_OP) {
        if (lChildIsInt && rChildIsInt) {
        	retVal = Boolean.valueOf (lChildInt != rChildInt);
        } else if (lChildIsFloat && rChildIsFloat) {
        	retVal = Boolean.valueOf (lChildFloat != rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
        	retVal = Boolean.valueOf ( ((float) lChildInt) != rChildFloat );
        } else if (lChildIsFloat && rChildIsInt) {
        	retVal = Boolean.valueOf ( lChildFloat != ((float) rChildInt) );
        } else if (lChildIsString && rChildIsString) {
        	retVal = Boolean.valueOf( !(lChildString.equals(rChildString)));
        } else if (lChildIsElement && rChildIsElement) {
        	retVal = Boolean.valueOf ( !(lChildElement.equals(rChildElement)));
        } else if (lChildIsPropertyRef && rChildIsPropertyRef) {
        	retVal = Boolean.valueOf (!(lChildPropertyRef.equals(rChildPropertyRef)));
        } else if (lChildIsSet && rChildIsSet) {
        	retVal = Boolean.valueOf (!(lChildSet.equals(rChildSet)));
        };
    }
    // else we have an illegal boolean expression with two numeric or
    // string children expressions.  Return false for now but eventually
    // throw an exception.
    else {
    	Debug.trace_return(this, retVal.booleanValue ());
    	String opMsg = new String("Illegal operands/operator combination "
    			+ " for a boolean expression.  operator = "
				+ this.getOperator().toString() );
    	throw new ExpressionTypeException(null, this,
    			ConstantTypes.BOOLEAN_TYPE, this.getType(), opMsg);
    };

    if (retVal != null) {
    	if (DesignRule.isRememberVisited()) {
			evaluationList.add (new BooleanExpressionEvaluationFeedback (this, retVal));
		}
		Debug.trace_return (this, retVal.booleanValue ());
    	return retVal.booleanValue();
    }
    Debug.trace_return (this, retVal.booleanValue ());
    return retVal.booleanValue ();
    // default return value.  If I get to here then I haven't been able
    // to evaluate the expression as a boolean value.  It should be the
    // case that there is no way to fall through to here, but the java
    // compiler seems to think that there is, so throw an exception
    // for good measure
  //    throw new DesignRuleEvalException("Unable to evaluate expression as a Boolean.");
  };


/**
 * @return
 */
private List getEvaluationData() {
	List l = (List )getUserData (getEvaluationKey());
	if (l == null) {
		l = new ArrayList (1);
		setUserData (getEvaluationKey(), l);
	}
	return l;
};


}
