package acmeLib.designRules;

import java.io.IOException;

import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.properties.AcmeBoolean;

/**
 * Class: BooleanLiteralNode<br>
 * Purpose: Simple expression tree leaf-node that stores true or false<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    10/8/97    RTM     Created<br>
 *    2/2/98     RTM     Modified type storage to store AcmeTypes<br>
 *                        instead of ExpressionTypeT objects.<br>
 *<br>
 * @author Bob Monroe
 */
public class BooleanLiteralNode extends BooleanExpressionNode
{
/**
 * stores the value of this literal.
 */
  protected boolean value;

/**
 * two static instances of this class that always evaluate to true
 * or false, respectively.  Useful in the evaluateAsBoolean method
 * of this class' superclass (BooleanExpressionNode)
 */
  public static final BooleanLiteralNode TRUE_NODE =
                        new BooleanLiteralNode(ArmaniParserTreeConstants.JJTBOOLEANLITERALNODE, true);
  public static final BooleanLiteralNode FALSE_NODE =
                        new BooleanLiteralNode(ArmaniParserTreeConstants.JJTBOOLEANLITERALNODE, false);


/**
 * JJTree-specific methods
 * @param id The ID
 * @return Node
 */
  public static Node jjtCreate(int id) {
    return new BooleanLiteralNode(id);
  }

/** 
 * This is the constructor for this class
 * @param id The string
 */
  public BooleanLiteralNode(int id) {
    super(id);
    this.setType(ConstantTypes.BOOLEAN_TYPE);
    this.setValue(false);       // default to False?
  }

  //
  // Methods added by bmonroe
  //
/**
 * This is the constructor for this class
 * @param id The string
 * @param val The boolean value
 */
  public BooleanLiteralNode(int id, boolean val) {
    super(id);
    this.setType(ConstantTypes.BOOLEAN_TYPE);
    this.setValue(val);
  }

/**
 * copy constructor
 * @param copyOf The boolean literal node to be copied
 */
  public BooleanLiteralNode(BooleanLiteralNode copyOf) {
    super(copyOf.id);
    this.setType(ConstantTypes.BOOLEAN_TYPE);
    this.setValue(copyOf.getValue());
  }


/**
 * cloneTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.  
 * @return The copy of the expression tree
 */
  public ExpressionNode cloneExpressionTree() {
    BooleanLiteralNode clone = new BooleanLiteralNode(this);
    this.doCloneExpressionTree(clone);
    clone.setType(this.getType());
    clone.setValue(this.getValue());
    return clone;
  };


/**
 * Check if the two nodes are equal
 * @param test The test node
 * @return true if the test node is equal to this node
 */
  public boolean equals(Object test) {
    if (!(test instanceof ExpressionNode)) {
        if (test instanceof AcmeBoolean) {
            return (this.getValue() == ((AcmeBoolean) test).getValue());
        }
            return false;
    }
        try {
            boolean testVal = ((ExpressionNode) test).evaluateAsBoolean();
            return (testVal == this.getValue());
        } catch (Exception exc) {
        };
        return false;
  };

/**
 * Get the value of this node
 * @return The boolean value
 */
  public boolean getValue() { return this.value; }
/**
 * Set the value of this node
 * @param val The new boolean value of this node
 */
  public void setValue(boolean val) { this.value = val; };


/**
 * because this node is a literal it must typecheck (it can't store
 * anything but a boolean).  Furthermore, it can't have any children nodes
 * so there's no need to traverse the children.
 * @return true if it is type correct
 */
  public boolean typecheck() {
    this.setType(ConstantTypes.BOOLEAN_TYPE);
    return true;
  };


/**
 * because this is always a leaf node, evaluate() just returns
 * the value of the literal.
 * @return The boolean value
 */
  public boolean evaluate() { return this.value; };


/**
 * override the generic evaluateAsBoolean and evaluateAsString methods
 * to return an appropriate value.
 * @return The evaluate boolean value
 */
  public boolean evaluateAsBoolean() {
    return this.evaluate();
  }

/**
 * override the generic evaluateAsBoolean and evaluateAsString methods
 * to return an appropriate value.
 * @return The evaluate string value
 */
  public String evaluateAsString() {
    return String.valueOf(this.evaluate());
  }


/**
 * generic exporting method
 * @return The output string
 */
  public String toString() {
    StringBuffer buf = new StringBuffer();
    if(this.getInParentheses()) {
        buf.append("(" + String.valueOf(this.value) + ")");
    } else {
        buf.append(String.valueOf(this.value));
    };

    return buf.toString();
  };

/**
 * exporting method
 * @param prefix The prefix
 * @return The output string
 */
  public String toString(String prefix) {
    return prefix + this.toString();
  };

    /* (non-Javadoc)
     * @see acmeLib.core.io.Exportable#exportAcme(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportAcme(AcmeOutputStream s) throws IOException, AcmeException {
        s.write (this.toString ());
    }

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion (this);
        s.write (this.toString ());
        s.endRegion (this);
    }

}
