package acmeLib.designRules;
import java.util.*;
import acmeLib.*;
import acmeLib.core.container.*;
import acmeLib.core.globals.*;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.typeManagement.*;


/**
 * Class: BuiltInPredicates<br>
 * Purpose: defines Armani's built-in predicates.  These are referenced from<br>
 *          ArmaniLib's constraint checker as external analyses.  The individual<br>
 *          predicates are defined as public static methods.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    3/25/98    RTM      * Created original class<br>
 *<br>
 * @author Bob Monroe
 */
public class BuiltInPredicates {


    //
    // protected methods can be defined as helpers for the public methods.
    //

  /**
   * retrieves the set expression that this node either is, or references.
   * if it is unable to retrieve set expression from the exprNode parameter
   * then it throws a DesignRuleEvalException.
   * @param exprNode The expression node
   * @return The set expression node
   * @exception DesignRuleEvalException If it cannot retrieve set expression from the exprNode
   */
    protected static SetExpressionNode getSetFromExprNode(ExpressionNode exprNode)
                                throws DesignRuleEvalException
    {
        if (!(exprNode instanceof SetExpressionNode)) {
            if (exprNode instanceof ReferenceNode) {
                return ((ReferenceNode) exprNode).evaluateAsSet();
            }
                throw new DesignRuleEvalException("Non-SetExpressionNode "
                    + "object passed as Set{}-typed argument for built-in "
                    + "predicate.  Unable to compute evaluate predicate.");
        }
            return (SetExpressionNode) exprNode;
    };


//
// public static methods.
// all external analyses need to be hooked in as public static methods.
//


/****************************************************************
 ** Topology-testing methods
 ****************************************************************/

	/**
	 * implements the dangling (element) Armani predicate. Valid elements
	 * are port or roles. Dangling means that the roles or ports are
	 * not attached to anything, or not bound if they are part of a representation.
	 * Armani signature:
	 * external analysis dangling (e : Element) : boolean =
	 * acmeLib.designRules.BuiltinPredicates.isDangling (e);
	 * @param elementNode The node to see if it is dangling
	 * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public static synchronized boolean isDangling (ExpressionNode elementNode) throws DesignRuleEvalException {
		AcmeElement elt = elementNode.evaluateAsElement();
		if (elt instanceof AcmePort) {
			AcmePort port = (AcmePort) elt;
			AcmeSystem system = (AcmeSystem )port.getParent ().getParent();
			AcmeEnumeration atts = system.getAttachments (port);
			if (atts != null && atts.getSize() != 0) {
				return false;
			}
			else {
				AcmeRepresentation rep = system.getParentRepresentation();
				if (rep == null) {
					return true;
				}
				else {
					AcmeBindingsMap map = (AcmeBindingsMap )rep.getAbstractionMap();
					if (map == null) return true;
					AcmeEnumeration bnds = map.getBindings(port);
					if (bnds != null && bnds.getSize () != 0) {
						return false;
					}
					else {
						return true;
					}	
				}
			}
		}
		else if (elt instanceof AcmeRole) {
			AcmeRole role = (AcmeRole )elt;
			AcmeSystem system = (AcmeSystem )role.getParent ().getParent ();
			AcmeEnumeration atts = system.getAttachments (role);
			if (atts != null && atts.getSize () != 0) {
				return false;
			}	
			else {
				AcmeRepresentation rep = system.getParentRepresentation();
				if (rep == null) {
					return true;
				}
				else {
					AcmeBindingsMap map = (AcmeBindingsMap )rep.getAbstractionMap();
					AcmeEnumeration bnds = map.getBindings (role);
					if (bnds != null && bnds.getSize () != 0) {
						return false;
					}
					else {
						return true;
					}
				}
			}
		}
		throw new DesignRuleEvalException ("Expression passed in is not a port or role");
	}

  /**
   * implements the attached(Element, Element) Armani predicate.  Valid
   * (Element,Element) pairs include (component, connector), (port, role),
   * (port,connector), and (role, component).
   * Armani signature:
   * external analysis attached(e1 : Element, e2 : Element) : boolean =
   * acmeLib.designRules.BuiltInPredicates.attached(e1,e2);
   * @param elementNode1 Element node 1
   * @param elementNode2 Element node 2
   * @return true if they are valid pairs
   * @exception DesignRuleEvalException If there is anevaluation error
   */
    public static synchronized boolean attached(ExpressionNode elementNode1,
                                   ExpressionNode elementNode2)
                                   throws DesignRuleEvalException
    {
        // dereference the ExpressionNode's to retrieve the port and role in question
        AcmeElement elt1 = elementNode1.evaluateAsElement();
        AcmeElement elt2 = elementNode2.evaluateAsElement();

        // There are four valid combinations of elements as arguments:
        //  - port/role (standard arguments)
        //  - component/connector
        //  - role/component
        //  - port/connector
        //
        // check for each of these in turn, then check the resulting pair
        // (if any) to see if they are attached.

        AcmePort port = null;
        AcmeRole role = null;
        AcmeComponent comp = null;
        AcmeConnector conn = null;

        boolean portRolePairFound = false;
        boolean compConnPairFound = false;
        boolean roleCompPairFound = false;
        boolean portConnPairFound = false;

        boolean attached = false;

        if (elt1 instanceof AcmePort) {
            port = (AcmePort) elt1;
            if (elt2 instanceof AcmeRole) {
                role = (AcmeRole) elt2;
                portRolePairFound = true;
            } else if (elt2 instanceof AcmeConnector) {
                conn = (AcmeConnector) elt2;
                portConnPairFound = true;
            };

        } else if (elt1 instanceof AcmeRole) {
            role = (AcmeRole) elt1;
            if (elt2 instanceof AcmePort) {
                port = (AcmePort) elt2;
                portRolePairFound = true;
            } else if (elt2 instanceof AcmeComponent) {
                comp = (AcmeComponent) elt2;
                roleCompPairFound = true;
            };

        } else if (elt1 instanceof AcmeComponent) {
            comp = (AcmeComponent) elt1;
            if (elt2 instanceof AcmeConnector) {
                conn = (AcmeConnector) elt2;
                compConnPairFound = true;
            } else if (elt2 instanceof AcmeRole) {
                role = (AcmeRole) elt2;
                roleCompPairFound = true;
            };

        } else if (elt1 instanceof AcmeConnector) {
            conn = (AcmeConnector) elt1;
            if (elt2 instanceof AcmeComponent) {
                comp = (AcmeComponent) elt2;
                compConnPairFound = true;
            } else if (elt2 instanceof AcmePort) {
                port = (AcmePort) elt2;
                portConnPairFound = true;
            };
        };


        // at this point, if the arguments are valid I will have an appropriate
        // pair of elements to check whether they are attached.
        if (portRolePairFound) {
            if (port.getParent() instanceof AcmeComponent) {
                AcmeComponent portParent = (AcmeComponent) port.getParent();
                if (portParent.getParent() instanceof AcmeSystem) {
                    AcmeSystem parentSys = (AcmeSystem) portParent.getParent();
                    if (parentSys.getAttachment(port, role) != null) {
                        attached = true;
                    };
                };
            };
        }

        else if (compConnPairFound) {
            // then check each of the possible port/role combinations of comp
            // and conn for attachment.
            if (comp.getParent() instanceof AcmeSystem) {
                AcmeSystem parentSys = (AcmeSystem) comp.getParent();
                AcmeEnumeration portsEnum = comp.getPorts();
                AcmeEnumeration rolesEnum = conn.getRoles();
                AcmePort curPort = null;
                AcmeRole curRole = null;

                while(portsEnum.moreItems() && !attached) {
                    curPort = (AcmePort) portsEnum.getNextItem();
                    while (rolesEnum.moreItems() && !attached) {
                        curRole = (AcmeRole) rolesEnum.getNextItem();
                        if (parentSys.getAttachment(curPort, curRole) != null) {
                            attached = true;
                        };
                    };
                    rolesEnum.reset();
                };
            };
        }

        else if (roleCompPairFound) {
            if (comp.getParent() instanceof AcmeSystem) {
                AcmeSystem parentSys = (AcmeSystem) comp.getParent();
                AcmeEnumeration portsEnum = comp.getPorts();
                AcmePort curPort = null;

                while(portsEnum.moreItems() && !attached) {
                    curPort = (AcmePort) portsEnum.getNextItem();
                    if (parentSys.getAttachment(curPort, role) != null) {
                        attached = true;
                    };
                };
            };
        }

        else if (portConnPairFound) {
            if (conn.getParent() instanceof AcmeSystem) {
                AcmeSystem parentSys = (AcmeSystem) conn.getParent();
                AcmeEnumeration rolesEnum = conn.getRoles();
                AcmeRole curRole = null;

                while(rolesEnum.moreItems() && !attached) {
                    curRole = (AcmeRole) rolesEnum.getNextItem();
                    if (parentSys.getAttachment(port, curRole) != null) {
                        attached = true;
                    };
                };
            };
        }

        return attached;
    };


public static synchronized SetExpressionNode connectedPorts (ExpressionNode portNode) throws DesignRuleEvalException {
	TransientSet retSet = new TransientSet ();
	retSet.setSetType(ConstantTypes.PORT_BASE_TYPE);
	AcmeElement elt1 = portNode.evaluateAsElement ();
	if (elt1 instanceof AcmePort) {
		AcmePort port = (AcmePort )elt1;
		AcmeVector ports = new AcmeVector ();
		
		AcmeSystem parentSystem = (AcmeSystem )port.getParent ().getParent ();
		AcmeEnumeration atts = parentSystem.getAttachments(port);
		while (atts.moreItems ()) {
			AcmeAttachment att = (AcmeAttachment )atts.getNextItem ();
			AcmeRole role = att.getRole ();
			AcmeConnector conn = (AcmeConnector )role.getParent ();
			AcmeEnumeration roles = conn.getRoles ();
			while (roles.moreItems ()) {
				AcmeRole endRole = (AcmeRole )roles.getNextItem ();
				if (endRole != role) {
					AcmeEnumeration endAtts = parentSystem.getAttachments(endRole);
					while (endAtts.moreItems ()) {
						AcmeAttachment endAtt = (AcmeAttachment )endAtts.getNextItem ();
						ports.addItem(endAtt.getPort ());
					}
				}
			}
		}
		retSet.setSetType(ports.enumerate().castToAcmeSetVector());
	}
	return retSet;
	
}

/**
 * implements the connected(component, component) Armani predicate
 * Armani signature:
 * external analysis connected(c1 : Component, c2 : Component) : boolean =
 * acmeLib.designRules.BuiltInPredicates.connected(c1,c2);
 * @param compNode1 Node 1
 * @param compNode2 Node 2
 * @return true if they are connected
 * @exception DesignRuleEvalException If there is an evaluation error
 */
    public static synchronized boolean connected(ExpressionNode compNode1,
                                    ExpressionNode compNode2)
                                    throws DesignRuleEvalException
    {
try {
        // dereference the ExpressionNode's to retrieve the port and role in question
        AcmeElement elt1 = compNode1.evaluateAsElement();
        AcmeElement elt2 = compNode2.evaluateAsElement();

		if (elt1 instanceof AcmeComponent && elt2 instanceof AcmeComponent) {

        	AcmeComponent comp1 = null;
        	AcmeComponent comp2 = null;

            comp1 = (AcmeComponent) elt1;
            comp2 = (AcmeComponent) elt2;
      

        	AcmeSystem parentSys = null;
        	if (comp1.getParent() instanceof AcmeSystem) {
            	parentSys = (AcmeSystem) comp1.getParent();
        	};

        	if ((parentSys == null) && comp2.getParent() instanceof AcmeSystem) {
            	parentSys = (AcmeSystem) comp2.getParent();
        	};

        	if (parentSys != null) {
            	return parentSys.areConnectedComponents(comp1, comp2);
        	}
		}
		else if (elt1 instanceof AcmePort && elt2 instanceof AcmePort) {
			AcmePort port1 = (AcmePort )elt1;
			AcmePort port2 = (AcmePort )elt2;
			
			AcmeSystem parentSys = null;
			if (port1.getParent () instanceof AcmeComponent && port1.getParent ().getParent () instanceof AcmeSystem) {
				parentSys = (AcmeSystem )port1.getParent ().getParent ();	
			}
			
			if ((parentSys == null) && (port2.getParent () instanceof AcmeComponent && port2.getParent ().getParent () instanceof AcmeSystem)) {
				parentSys = (AcmeSystem )port2.getParent ().getParent ();
			}
			
			if (parentSys != null) {
				return parentSys.areConnectedPorts(port1, port2);
			}
		}
	    else {
            throw new DesignRuleEvalException("Error! Non-Component or Not-Port type argument "
                + "passed to acmeLib.designRules.BuiltInPredicates.connected");
        };
            return false;   // if neither have a valid parent then they're
                            // not connected...
} catch (Throwable e) {
    e.printStackTrace ();
    return false;
}
    };


public static Map getConnectedComponents(AcmeComponent fromComp, AcmeSystem sys, Set usedConnectors) {
	Map compConnMap = new HashMap ();
    List connComps = new ArrayList (5);
    List throughConnectors = new ArrayList  (5);
    if (usedConnectors == null) usedConnectors = new HashSet ();
    AcmeEnumeration ports = fromComp.getPorts();
    while (ports.moreItems()) {
        AcmePort port = (AcmePort) ports.getNextItem();
        AcmeEnumeration roles = sys.getAttachments(port);
        while (roles.moreItems()) {
            AcmeRole role = ((AcmeAttachment) roles.getNextItem()).getRole();
            AcmeConnector conn = (AcmeConnector) role.getParent();
            if (!usedConnectors.contains(conn)) {
                AcmeEnumeration otherRoles = conn.getRoles();
                while (otherRoles.moreItems()) {
                    AcmeRole or = (AcmeRole) otherRoles.getNextItem();
                    if (or != role) {
                        AcmeEnumeration otherPorts = sys.getAttachments(or);
                        while (otherPorts.moreItems()) {
                            AcmePort op = ((AcmeAttachment) otherPorts.getNextItem()).getPort();
                            if (!compConnMap.containsKey (op.getParent ())) {
                            	Set connections = new HashSet ();
                            	connections.add (conn);
                            	compConnMap.put (op.getParent (), connections);
                            }
                            else {
                            	((Set )compConnMap.get (op.getParent ())).add (conn);
                            }
                            
                        }
                    }
                }
            }
        }
    }

    return compConnMap;

}


private static boolean recurseReachable (AcmeComponent sourceComp, AcmeComponent targetComp, AcmeSystem inSys, Set visitedComponents, Set usedConnectors) {
    Map connections = getConnectedComponents (sourceComp, inSys, usedConnectors);
   
   
    boolean reachable = connections.get (targetComp) != null;
    Iterator comps = connections.entrySet ().iterator();
   
    while (comps.hasNext() && !reachable) {
    	Map.Entry entry = (Map.Entry )comps.next ();
        AcmeComponent comp = (AcmeComponent )entry.getKey();
        Set visited = new HashSet (visitedComponents);
//        Set connectSet = new HashSet (usedConnectors);
//        connectSet.addAll ((Set )entry.getValue());
        if (!visited.contains (comp)) {
            visited.add (comp);
            for (Iterator iter = ((Set )entry.getValue ()).iterator (); iter.hasNext ();) {
				AcmeConnector conn = (AcmeConnector )iter.next ();
				Set connectSet = new HashSet (usedConnectors);
				connectSet.add (conn);
			
				reachable = recurseReachable (comp, targetComp, inSys, visited, connectSet);
            }
        }
    }
    
    
    return reachable;
        
}

public synchronized static boolean reachable (ExpressionNode node1, ExpressionNode node2) throws DesignRuleEvalException {
    
    AcmeElement elt1 = node1.evaluateAsElement();
    AcmeElement elt2 = node2.evaluateAsElement ();
    
    if (elt1 instanceof AcmeComponent && elt2 instanceof AcmeComponent) {
        AcmeComponent fromComp = (AcmeComponent )elt1;
        AcmeComponent toComp = (AcmeComponent )elt2;
        
        AcmeSystem parentSys = null;
        if (fromComp.getParent () instanceof AcmeSystem) {
            parentSys = (AcmeSystem )fromComp.getParent();
        }
        if (toComp.getParent () != parentSys) {
            return false;
        }
        
        if (parentSys.areConnectedComponents(fromComp, toComp)) {
            return true;
        }
        
        Set visitedComponents = new HashSet ();
        Set usedConnectors = new HashSet ();
        visitedComponents.add (fromComp);
        boolean r =  recurseReachable (fromComp, toComp, parentSys, visitedComponents, usedConnectors);
        return r;
        
    }
    else {
        DesignRuleEvalException ex = new DesignRuleEvalException ("Error! Non-Component type argument " 
                                            + "passed to acmeLib.designRules.BuiltInPredicates.reachable");
        throw ex;
    }
    
}



/**
 * calculates the parent (element) of the element passed as an argument.
 * Armani signature:
 * external analysis parent(e : Element) : Element =
 * acmeLib.designRules.BuiltInPredicates.parent(e);
 * @param node The expression node
 * @return The parent AcmeElement
 * @exception DesignRuleEvalException If unable to resolve expression
 */
    public static synchronized AcmeElement parent(ExpressionNode node)
        throws DesignRuleEvalException
    {
        AcmeElement childElt = null;
        if (node == null) {
            throw new DesignRuleEvalException("ExpressionNode passed as an "
                        + "Element e to parent(e) is null");
        }
            childElt = node.evaluateAsElement();
            if (childElt == null) {
                throw new DesignRuleEvalException("Unable to resolve Expression "
                            + "node passed to parent(e) as an Element.");
            }
                return childElt.getParent();
    };


/****************************************************************
 ** Type-based methods
 ****************************************************************/


/**
 * implements the declaresType(Any, Type) Armani predicate.
 * returns true if object "Any" declares type "Type" in the Armani spec.
 * Armani signature:
 * external analysis declaresType(a : Any, t : Type) : boolean =
 * acmeLib.designRules.BuiltInPredicates.declaresType(a,t);
 * @param eltNode Element node
 * @param typeNode Type node
 * @return true if element node declares type in the Armani spec
 * @exception DesignRuleEvalException If there is an evaluation error
 */
    public static synchronized boolean declaresType(ExpressionNode objNode,
                                       ExpressionNode typeNode)
                                       throws DesignRuleEvalException
    {
    try {
        AcmeProperty prop = null;
        AcmeElement element = null;
        AcmeType type = null;

        // figure out what sorts of entities we have.
        // first check if obj is an element or a property
        try {
            element = objNode.evaluateAsElement();
        }
        catch (DesignRuleEvalException dreExc) {
            // object isn't an element.  see if it's a property
            try {
                prop = objNode.evaluateAsPropertyRef();
            }
            catch (DesignRuleEvalException innerDrExc) {
                // object is neither a property nor an element, return false.
                return false;
            };
        };

        // then check to see if typeNode is an element type or a property type
        type = typeNode.evaluateAsTypeRef();


        // then try to figure out if the object declares the type
        if ((type != null) && (element != null)) {
            // check for a matched element, element type pair.
            if (type instanceof BaseElementParameterType) {
                return true;    // special case for "Element" type
            }
            else if (type instanceof AcmeElementType) {
                return element.declaresType((AcmeElementType) type);
            }
                return false;
        }

        else if ((prop != null) && (type != null)) {
            if (type instanceof PropertyReferenceType) {
                return true;    // special case for "Property" type
            }
                return (prop.getType() == type);
        }

            // we don't have a valid elt/eltType or prop/propType pair
            return false;
    } catch (Throwable e) {
        e.printStackTrace ();
        return false;
    }
    };

    public static synchronized SetExpressionNode declaredTypes (ExpressionNode elementExpression) throws DesignRuleEvalException {
    	TransientSet retValSet = new TransientSet ();
    	AcmeElement element = elementExpression.evaluateAsElement();
    	AcmeEnumeration declaredTypes = element.getDeclaredTypes();
    	retValSet.setSetType(declaredTypes.castToAcmeSetVector());
    	return retValSet;
    }

    public static synchronized SetExpressionNode superTypesOf (ExpressionNode typeExpr) throws DesignRuleEvalException {
    	TransientSet retValSet = new TransientSet ();
    	AcmeType type = typeExpr.evaluateAsTypeRef();
    	if (!(type instanceof AcmeElementType)) {
    		throw new DesignRuleEvalException (type.getTypeName() + " is not a valid element type");
    	}
    	else {
    		AcmeElementType t = (AcmeElementType )type;
    		retValSet.setSetType (gatherSuperTypes(t).castToAcmeSetVector());
    	}
    	return retValSet;
    }
    /**
     * implements the satisfiesType(Any, Type) Armani predicate.
     * returns true if object "Any" satisfies type "Type" in the Armani spec.
     * Armani signature:
     * external analysis satisfiesType(a : Any, t : Type) : boolean =
     * acmeLib.designRules.BuiltInPredicates.satisfiesType(a,t);
     * @param eltNode Element node
     * @param typeNode Type node
     * @return true if "Any" node satisfies "Type" in the Armani spec
     * @exception DesignRuleEvalException If there is an evaluation error
     */
    public static synchronized boolean satisfiesType(ExpressionNode objNode,
                                        ExpressionNode typeNode)
                                        throws DesignRuleEvalException
    {
    	
        AcmeProperty prop = null;
        AcmeElement element = null;
        AcmeType type = null;

        // figure out what sorts of entities we have.
        // first check if obj is an element or a property
        try {
            element = objNode.evaluateAsElement();
        }
        catch (DesignRuleEvalException dreExc) {
            // object isn't an element.  see if it's a property
            try {
                prop = objNode.evaluateAsPropertyRef();
            }
            catch (DesignRuleEvalException innerDrExc) {
                // object is neither a property nor an element, return false.
                return false;
            };
        };

        // then check to see if typeNode is an element type or a property type
        type = typeNode.evaluateAsTypeRef();

        boolean errorCheckingEnabled = AcmeTypeManager.errorReportingIsEnabled();
        AcmeTypeManager.disableErrorReporting();
        boolean retVal = false;
        // then try to figure out if the object satisfies the type
        if ((type != null) && (element != null)) {
            // check for a matched element, element type pair.
            if (type instanceof BaseElementParameterType) {
                retVal = true;    // special case for "Element" type
            }
            else if (type instanceof AcmeElementType) {
                retVal =  AcmeTypeManager.satisfiesElementType(
                                        (AcmeElementType) type, element);
            }
        }

        else if ((prop != null) && (type != null)) {
            if (type instanceof PropertyReferenceType) {
                retVal =  true;    // special case for "Property" type
            }
            else if (type instanceof AcmePropertyType) {
                retVal = AcmeTypeManager.typecheckProperty(
                            (AcmePropertyType) type, prop.getValue());
            }
        }
        if (errorCheckingEnabled) AcmeTypeManager.enableErrorReporting();
       
        return retVal;

    };


	public static synchronized boolean extendsType (ExpressionNode subTypeNode, ExpressionNode superTypeNode) throws DesignRuleEvalException {
    	AcmeType subtype = subTypeNode.evaluateAsTypeRef();
    	AcmeType supertype = superTypeNode.evaluateAsTypeRef ();
    	
    	if (subtype == null || supertype == null) {
    		return false;
    	}
    	
    	if (subtype instanceof AcmePropertyType && supertype instanceof AcmePropertyType) {
    		return subtype.equals (supertype);
    	}
    	
    	if (subtype instanceof AcmeElementType && supertype instanceof AcmeElementType) {
    		return AcmeTypeManager.isDeclaredSubtype((AcmeElementType )subtype, (AcmeElementType )supertype);
    	}
		return false;
	} 

/****************************************************************
 ** Set-based methods
 ****************************************************************/

/**
 * computes the size of the set passed as param setNode, which needs to
 * be a SetExpressionNode. Armani signature:
 * external analysis size(s : Set{}) : int =
 * acmeLib.designRules.BuiltInPredicates.setSize(s);
 * @param setExprNode The set expression node
 * @return The set size
 * @exception DesignRuleEvalException If there is an evaluation error
 */
    public static synchronized int setSize(ExpressionNode setExprNode)
                              throws DesignRuleEvalException
    {
        return getSetFromExprNode(setExprNode).size();
    };



/**
 * calculate the sum of the numeric values stored in the set passed as
 * an argument.
 * external analysis setSum(s : Set{Float}) : float =
 * acmeLib.designRules.BuiltInPredicates.setSum(s);
 * @param setNode The set node
 * @return The sum
 * @exception DesignRuleEvalException If there is an evaluation error
 */
    public static synchronized float setSum(ExpressionNode setNode)
                               throws DesignRuleEvalException
    {
        // convert the ExpressionNode into a SetExpressionNode.  If the
        // conversion doesn't work properly it will throw a DesignRuleEvalException
        // that simply propogates up.
        SetExpressionNode resolvedSet = getSetFromExprNode(setNode);
        if (!(resolvedSet.getSetType() instanceof SetType) &&
            !AcmeTypeManager.isLogicalSubtype(((SetType )resolvedSet.getSetType()).getSetType (),
                                              ConstantTypes.FLOAT_TYPE, false) )
        {
            throw new DesignRuleEvalException("Error, set passed to setSum "
                        + "method is not a subtype of Set{Float}.");
        };

        float sum = (float) 0.0;
        AcmeEnumeration e = resolvedSet.enumerate();
        while (e.moreItems()) {
            AcmeObject curObj = e.getNextItem();
            if (curObj instanceof ExpressionNode) {
                sum += ((ExpressionNode) curObj).evaluateAsFloat();
            } else if (curObj instanceof AcmeProperty) {
                AcmePropertyValue curObjVal = ((AcmeProperty) curObj).getValue();
                if (curObjVal instanceof AcmeFloat) {
                    sum += ((AcmeFloat) curObjVal).getValue();
                } else if (curObjVal instanceof AcmeInt) {
                    sum += (float) ((AcmeInt) curObjVal).getValue();
                } else {
                    throw new DesignRuleEvalException("Error, attempted to "
                        + "sum a non-numeric property value.");
                };
            } else if (curObj instanceof AcmeFloat) {
                sum += ((AcmeFloat) curObj).getValue();
            } else if (curObj instanceof AcmeInt) {
                sum += (float) ((AcmeInt) curObj).getValue();
            } else {
                throw new DesignRuleEvalException("Error, attempted to "
                    + "sum a non-numeric property value.");
            };
        }

        return sum;
    };


/**
 * RTM 3/29/98 -- Other than setSize and setSum, none of the set functions are
 *                really working properly.  They have typechecking "issues"
 *                related to working on arbitrary types, and they have an
 *                incoherent concept of set element equality that makes
 *                union, intersection, and contains rather untrustworthy.
 *                All of these things need to be fixed.
 *
 *
 * calculate the unions of the sets passed in as setNode1 and setNode2
 * be a SetExpressionNode. Armani signature:
 * external analysis union(s1, s2 : Set{}) : Set{} =
 * acmeLib.designRules.BuiltInPredicates.setUnion(s1,s2);
 * @param setNode1 The set node 1
 * @param setNode2 The set node 2
 * @return The union set expression node
 * @exception DesignRuleEvalException If there is an evaluation error
 */
    public static synchronized SetExpressionNode setUnion(ExpressionNode setNode1,
                                             ExpressionNode setNode2)
                                             throws DesignRuleEvalException
    {
        // convert the ExpressionNodes into SetExpressionNodes.  If the
        // conversion doesn't work properly it will throw a DesignRuleEvalException
        // that simply propogates up.
        SetExpressionNode resolvedSet1 = getSetFromExprNode(setNode1);
        SetExpressionNode resolvedSet2 = getSetFromExprNode(setNode2);

        // create the TransientSet to return with a vector containing all of
        // the elements of resolvedSet1.  Then repeatedly add the elements
        // of resolvedSet2 to the return set.  Because the TransientSet's
        // addElement(elt) method enforces set semantics, the element should
        // only be added if it is not already a member of the set.
        // RTM 3/27/98 -- Unfortunately, this will almost always incorrectly add
        //                multiple items, because it's doing object equality
        //                checking. rather than value comparison.  I haven't
        //                figured out how to make ExpressionNodes properly
        //                implement value equality comparisons, especially not
        //                in any reasonably efficient way.
        TransientSet retValSet =
            new TransientSet(resolvedSet1.enumerate().castToAcmeVector());
        AcmeEnumeration set2Enum = resolvedSet2.enumerate();
        while (set2Enum.moreItems()) {
            retValSet.addElement(set2Enum.getNextItem());
        };

        // figure out what the set type for the new set will be.
        AcmeVector typesVector = new AcmeVector(2);
        typesVector.addItem((AcmeObject) resolvedSet1.getSetType());
        typesVector.addItem((AcmeObject) resolvedSet2.getSetType());

        AcmeType unionedSetType =
                    AcmeTypeManager.inferLeastCommonSupertype(typesVector);
        retValSet.setType(new SetType(unionedSetType.getTypeName(),
                                      unionedSetType));
        retValSet.setSetType(unionedSetType);

        return retValSet;
    };


  

/**
 * calculate the unions of the sets passed in as setNode1 and setNode2
 * be a SetExpressionNode. Armani signature:
 * external analysis union(s1, s2 : Set{}) : Set{} =
 * acmeLib.designRules.BuiltInPredicates.setUnion(s1,s2);
 * @param setNode1 The set node 1
 * @param setNode2 The set node 2
 * @return The intersection set expression node
 * @exception DesignRuleEvalException If there is an evaluation error
 */
    public static synchronized SetExpressionNode setIntersection(ExpressionNode setNode1,
                                                    ExpressionNode setNode2)
                                                    throws DesignRuleEvalException
    {
        // convert the ExpressionNodes into SetExpressionNodes.  If the
        // conversion doesn't work properly it will throw a DesignRuleEvalException
        // that simply propogates up.
        SetExpressionNode resolvedSet1 = getSetFromExprNode(setNode1);
        SetExpressionNode resolvedSet2 = getSetFromExprNode(setNode2);

        // create the TransientSet to return with an empty vector.
        // walk down the list of elements in resolvedSet1, and add each one
        // if it is also in resolvedSet2.
        // RTM 3/27/98 -- Unfortunately, this will almost always incorrectly
        //                discard most elements because it's doing object
        //                equality checking rather than value comparison.  I
        //                haven't figured out how to make ExpressionNodes
        //                properly implement value equality comparisons,
        //                especially not in any reasonably efficient way.
        TransientSet retValSet = new TransientSet(new AcmeVector());
        AcmeVector set1Vector = resolvedSet1.enumerate().castToAcmeVector();
        AcmeVector set2Vector = resolvedSet2.enumerate().castToAcmeVector();
        for (int i = 0; i < set1Vector.getSize(); i++) {
            AcmeObject curObj = set1Vector.getItemAt(i);
            if (set2Vector.contains(curObj)) {
                retValSet.addElement(curObj);
            };
        };


        // figure out what the set type for the new set will be.
        AcmeVector typesVector = new AcmeVector(2);
        typesVector.addItem((AcmeObject) resolvedSet1.getSetType());
        typesVector.addItem((AcmeObject) resolvedSet2.getSetType());

        AcmeType intersectionSetType =
                    AcmeTypeManager.inferLeastCommonSupertype(typesVector);
        retValSet.setType(new SetType(intersectionSetType.getTypeName(),
                                      intersectionSetType));
        retValSet.setSetType(intersectionSetType);

        return retValSet;
    };




/**
 * determine whether the set passed in as subsetNode is a subset of
 * the set passed in as supersetNode.  Return true if the subset
 * relation holds, false if it does not.
 * external analysis isSubset(subset, superset : Set{}) : boolean =
 * acmeLib.designRules.BuiltInPredicates.isSubset(subset,superset);
 * @param subsetNode The subset node
 * @param supersetNode The superset node
 * @return true if subset node is subset of superset node
 * @exception DesignRuleEvalException If there is an evalutaion error
 */
    public static synchronized boolean isSubset(ExpressionNode subsetNode,
                                   ExpressionNode supersetNode)
                                   throws DesignRuleEvalException
    {
        // convert the ExpressionNodes into SetExpressionNodes.  If the
        // conversion doesn't work properly it will throw a DesignRuleEvalException
        // that simply propogates up.
        SetExpressionNode resolvedSubset = getSetFromExprNode(subsetNode);
        SetExpressionNode resolvedSuperset = getSetFromExprNode(supersetNode);

        // if the type of the subset node is not a logical subtype of the
        // supertype node then the sub/super type relationship can not hold.
        // check this first.
        if (!(AcmeTypeManager.isLogicalSubtype(resolvedSubset.getSetType(),
                                             resolvedSuperset.getSetType(), false)))
        {
            return false;
        };

        // RTM 3/27/98 -- Unfortunately, this will almost always incorrectly
        //                determine the subset/superset relationship because
        //                SetExpressionNodes are compared by object equality
        //                rather than a more sane form of value equality.
        //                This needs to be fixed...

        // check that each member of subset is a member of superset.
        boolean subsetHolds = true;
        AcmeEnumeration subsetEnum = resolvedSubset.enumerate();
        AcmeVector supersetVector =
                        resolvedSuperset.enumerate().castToAcmeVector();
        while (subsetEnum.moreItems() && subsetHolds) {
            AcmeObject nextElt = subsetEnum.getNextItem();
            if (supersetVector.contains(nextElt) == false) {
                subsetHolds = false;
            };
        };

        return subsetHolds;
    };





/**
 * determine whether the set passed in as setNode contains the element
 * the set passed in as itemNode.  Return true if the contains
 * relation holds, false if it does not.
 * external analysis contains(e : Element, s : Set{Element}) : boolean =
 * acmeLib.designRules.BuiltInPredicates.contains(e,s);
 *
 * RTM 3/30/98 -- Expand to include arbitrary objects, not just Elements.
 *                Unfortunately, I don't yet know how to handle the type
 *                "any" in Armani without messing up the rest of the type
 *                system...
 * @param elementNode The element node
 * @param setNode The set node
 * @return true if the set node contains element node
 * @exception DesignRuleEvalException If there is an evaluation error
 */
    public static synchronized boolean contains(ExpressionNode elementNode,
                                   ExpressionNode setNode)
                                   throws DesignRuleEvalException
    {
        // convert the setNode into a SetExpressionNode.  If the
        // conversion doesn't work properly it will throw a DesignRuleEvalException
        // that simply propogates up.
        SetExpressionNode resolvedSet = getSetFromExprNode(setNode);
        AcmeElement resolvedElement = elementNode.evaluateAsElement();

        // RTM 3/27/98 -- Unfortunately, this will almost always incorrectly
        //                determine the subset/superset relationship because
        //                SetExpressionNodes are compared by object equality
        //                rather than a more sane form of value equality.
        //                This needs to be fixed...

        // check whether resolvedElement is a member of resolvedSet
        boolean elementFound = false;
        AcmeEnumeration setEnum = resolvedSet.enumerate();
        while (setEnum.moreItems() && !elementFound) {
            AcmeObject nextElt = setEnum.getNextItem();
            AcmeElement curEltObject = null;
            try {
                if (nextElt instanceof ExpressionNode) {
                    curEltObject = ((ExpressionNode) nextElt).evaluateAsElement();
                    if (curEltObject == resolvedElement) {
                        elementFound = true;
                    };
                } else if (nextElt instanceof AcmeElement) {
                    if (nextElt == resolvedElement) {
                        elementFound = true;
                    };
                };
            } catch (DesignRuleEvalException dre) {
                // do nothing, the element currently tested just isn't a match...
            };
        };

        return elementFound;
    };




    //////////////////////////////////////
    //
    // element naming and dereferencing
    //
    //////////////////////////////////////


    public static synchronized String getObjectName(ExpressionNode objNode)
                                   throws DesignRuleEvalException
    {
        // figure out if the node is a property, an element, or a type
        // reference and return the appropriate string accordingly.
        // cycle through them by trying to evaluate the expression node
        // as various types things and catching any evaluation errors.
        boolean foundName = false;
        try {
        if (objNode instanceof ReferenceNode) {
            ((ReferenceNode) objNode).refreshReferentObject();
        };

        
            // first try to get the name of an element
            AcmeElement elt = objNode.evaluateAsElement();
            if (elt != null) {
                return elt.getName();
            };
        } catch (DesignRuleEvalException dreExc) {
            foundName = false;
        };

        // then try it as a property reference
        if (foundName == false) {
            try {
                AcmeProperty prop = objNode.evaluateAsPropertyRef();
                if (prop != null) {
                    return prop.getName();
                };
            } catch (DesignRuleEvalException dreExc) {
                foundName = false;
            };
        };

        // try it as a type if it's still not found...
        if (foundName == false) {
            try {
                AcmeType type = objNode.evaluateAsTypeRef();
                if (type != null) {
                    return type.getTypeName();
                };
            } catch (DesignRuleEvalException dreExc) {
                foundName = false;
            };
        };

        // if we get to here then we don't seem to have a named entity.
        // so throw an exception.
        throw new DesignRuleEvalException("Error.  Unable to determine name of "
                    + "object passed to name(Any) analysis.");
    };




    /** Returns a string containing the declarations for Armani's "built-in"
     *  design analyses.  The idea is that this string gets returned to the
     *  parser, which parses it to create a design with these default
     *  declarations, stores it as an imported design that is not written-out
     *  on export, because these are the basic methods that all designs have
     *  access to.
     *  @return String that contains the declarations of the "built-in"
     *          design analyses.
     **/
    public static String getArmaniLibDecls() {
        StringBuffer buf = new StringBuffer();

        //
        // add the built-design analysis declarations:
        //
        // Topological constraints:
        //
        buf.append("external analysis connected(c1 : Element, c2 : Element) : boolean = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.connected(c1,c2);");

        buf.append("external analysis attached(e1 : Element, e2 : Element) : boolean = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.attached(e1, e2);");

        buf.append("external analysis parent(e : Element) : Element = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.parent(e);");

        buf.append ("external analysis reachable (c1 : Component, c2 : Component) : boolean =");
        buf.append ("  acmeLib.designRules.BuiltInPredicates.reachable (c1,c2);");
        
        buf.append ("external analysis dangling (e : Element) : boolean = ");
        buf.append ("  acmeLib.designRules.BuiltInPredicates.isDangling (e);");
        
        buf.append ("external analysis connectedPorts (p : Port) : set {Port} = ");
        buf.append ("  acmeLib.designRules.BuiltInPredicates.connectedPorts (p);");
        //
        // Type-based predicates and functions:
        //
        buf.append("external analysis declaresType(a : any, t : Type) : boolean = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.declaresType(a, t);");

        buf.append("external analysis satisfiesType(a : any, t : Type) : boolean = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.satisfiesType(a, t);");

		buf.append ("external analysis subtypeOf (subtype : type, supertype : type) : boolean = ");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.extendsType (subtype, supertype);");
		
		buf.append ("external analysis typesDeclared (e : element) : set {} =");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.declaredTypes (e);");
		
		buf.append ("external analysis superTypes (t : type) : set {} = ");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.superTypesOf (t);");

        //
        // Set-based predicates and functions
        //
        buf.append("external analysis union(s1, s2 : Set{}) : Set{} = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.setUnion(s1, s2);");

        buf.append("external analysis intersection(s1, s2 : Set{}) : Set{} = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.setIntersection(s1, s2);");

        buf.append("external analysis contains(e : Element, s : Set{}) : boolean = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.contains(e,s);");

        buf.append("external analysis isSubset(subset, superset : Set{}) : boolean = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.isSubset(subset,superset); ");

        buf.append("external analysis size(s : Set{}) : int = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.setSize(s); ");

        buf.append("external analysis sum(s : Set{float}) : float = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.setSum(s); ");


        // retrieves the name of a property or an element
        buf.append("external analysis name(a : Any) : string = ");
        buf.append("  acmeLib.designRules.BuiltInPredicates.getObjectName(a); ");

		// retrieves a type based on a name
		buf.append ("external analysis lookupType (a : any, s : string) : Type = ");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.lookupType (a,s);");
		

		buf.append ("external analysis isInherited (p : Property) : boolean =");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.isInherited (p);");
		
		buf.append ("external analysis declaredIn (p : Property) : Element =");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.declaredIn (p);");
		
		buf.append ("external analysis assignedIn (p : Property) : Element =");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.assignedIn (p);");
		
		buf.append ("external analysis hasValue (p : Property) : boolean =");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.hasValue (p);");
		
		buf.append ("external analysis narrowType (e : element, t : type) : type =");
		buf.append ("  acmeLib.designRules.BuiltInPredicates.narrowType (e, t);");
		
		
        // add the built-in property types that handle operations
        buf.append("Property Type opParamT    = Record [ paramName : string;  paramType : string; ];");
        buf.append("Property Type operationT  = Record [ opName : string;");
        buf.append("                                     javaMethod : string;");
        buf.append("                                     returnType : string;");
        buf.append("                                     params : Sequence<opParamT>;");
        buf.append("                                     description : string; ];");
        buf.append("Property Type operationsT = Set{operationT};");


        // return the resulting string.
        return buf.toString();
    };

    /*
     * Dangling roles
       Isolated components
       Isolated connectors
        Unattached ports
        Empty representations
        Uninitialized properties
        Missing bindings
        Non 1-1 bindings

     */

    /** 
     * Returns the set of roles that are unattached in a particular system.
     * @param node The node containing an expression that refers to a connector
     * @return The set of roles that are not attached
     */
    public static synchronized SetExpressionNode danglingRoles (ExpressionNode node) throws DesignRuleEvalException {
        TransientSet retValSet = new TransientSet(new AcmeVector());
        
        // Get the system out of the node
        AcmeElement element = node.evaluateAsElement();
        AcmeConnector connector = null;
        if (element instanceof AcmeSystem) {
            connector = (AcmeConnector )element;
        }
        else {
            throw new DesignRuleEvalException ("Non connector element passed to acmeLib.designRules.BuiltInPredicates.danglingRoles.");
        }
        
        // Will contain all the roles, after which we will take out the connected ones
        java.util.Set roles = new java.util.HashSet (5);
        
//        // Fill the roles list
//        AcmeEnumeration sysRoles = connector.getRoles ();
//        while (sysRoles.moreItems()) {
//            AcmeRole role = (AcmeRole )sysRoles.getNextItem();
//            roles.add (role);
//        }
        
//        AcmeEnumeration connectors = connector.getConnectors ();
//        while (connectors.moreItems()) {
//            AcmeConnector conn = (AcmeConnector )connectors.getNextItem();
            AcmeEnumeration connRoles = connector.getRoles ();
            while (connRoles.moreItems ()) {
                roles.add (connRoles.getNextItem());
            }
//        }
        
        // Now remove all the roles involved in attachments 
        AcmeEnumeration attachments = ((AcmeSystem )connector.getParent ()).getAttachments();
        while (attachments.moreItems ()) {
            AcmeAttachment att = (AcmeAttachment )attachments.getNextItem ();
            roles.remove (att.getRole());   
        }
        
        // Copy across to return
        java.util.Iterator i = roles.iterator(); 
        while (i.hasNext ()) {
            AcmeRole role = (AcmeRole )i.next ();
            retValSet.addElement (role);   
        }      
        
        return retValSet;
        
    }

    /**
     * Find all the roles in the system that is attached to this port
     * @param port The port to find attached roles for
     * @param sys The system that should contain the attachements
     * @return The set of roles attached to this port
     */
    private static Set internalAttachedRoles (AcmePort port, AcmeSystem sys) {
        Set attachedRoles = new HashSet ();
        AcmeEnumeration atts = sys.getAttachments();
        while (atts.moreItems()) {
            AcmeAttachment att = (AcmeAttachment )atts.getNextItem();
            if (att.getPort().equals (port)) {
                attachedRoles.add (att.getRole());
            }
        }
        
        return attachedRoles;
    }


    /**
     * Finds all the isolated connectors in a system. i.e., all connectors that 
     * are not attached to any ports
     * @param sysNode The node containing the system in which to find the connectors
     * @return SetExpressionNode The set of isolated coneectors
     * @throws DesignRuleEvalException 
     */
    public static synchronized SetExpressionNode isolatedConnectors (ExpressionNode sysNode) throws DesignRuleEvalException {
        TransientSet retValSet = new TransientSet (new AcmeVector ());
        
        // Check to make sure node is a system
        AcmeElement el = sysNode.evaluateAsElement();
        AcmeSystem system = null;
        if (el instanceof AcmeSystem) {
            system = (AcmeSystem )el;
        }
        else {
            throw new DesignRuleEvalException ("Non system element passed to acmeLib.designRules.BuiltInPredicates.danglingRoles.");
        }
         
        // Examine each connector
        AcmeEnumeration conns = system.getComponents();
        while (conns.moreItems()) {
            AcmeConnector connector = (AcmeConnector )conns.getNextItem ();
            boolean hasAttachedRoles = false;
            AcmeEnumeration roles = connector.getRoles ();
            // see if any ports have attached roles
            while (roles.moreItems() && !hasAttachedRoles) {
                AcmeRole role = (AcmeRole )roles.getNextItem ();
                if (system.getAttachments(role).getSize () != 0) {
                    hasAttachedRoles = true;
                }
                
            }
            
            // If no attached roles, this is an isolated component
            if (!hasAttachedRoles) {
                retValSet.addElement (connector);
            }
        }
        
        
        return retValSet;
    }   
    
    public static synchronized AcmeType lookupType (ExpressionNode context, ExpressionNode typeName) throws DesignRuleEvalException {
    	AcmeElement ctxtEl = context.evaluateAsElement();
    	String typeN = typeName.evaluateAsString();
    	Object o = ctxtEl.lookupName (typeN, true);
    	if (o instanceof AcmeType) {
    		AcmeType type = (AcmeType )o;
    		return type;
    	}
    	
        throw new DesignRuleEvalException ("Error: Unable to resolve name " + typeN + " to a type");
    }
    
    public static boolean isInherited (ExpressionNode propExpr) throws DesignRuleEvalException {
    	AcmeProperty prop = propExpr.evaluateAsPropertyRef();
    	return declaredIn (propExpr) != prop.getScope ();
    }
    
    public static AcmeElement assignedIn (ExpressionNode propExpr) throws DesignRuleEvalException {
    	AcmeProperty prop = propExpr.evaluateAsPropertyRef();
    	AcmeElement el = declaredIn (propExpr);
    	
    	if (el.getProperty(prop.getName ()) != null && el.getProperty (prop.getName ()).getValue () != null) {
    		return el;
    	}
    	else {
    		return (AcmeElement )prop.getScope ();
    	}
    }
    
    public static boolean hasValue (ExpressionNode propExpr) throws DesignRuleEvalException {
    	AcmeProperty prop = propExpr.evaluateAsPropertyRef ();
    	if (prop.getValue () != null) return true;
    	else return false;
    }
    
    public static synchronized AcmeElement declaredIn (ExpressionNode propExpr) throws DesignRuleEvalException {
    	AcmeProperty prop = propExpr.evaluateAsPropertyRef();
    	if (prop.getScope () instanceof AcmeElement) {
			AcmeElement parent = (AcmeElement) prop.getScope ();
			AcmeProperty iProp = null;
			AcmeEnumeration types = getTypes (parent); 
			if (types != null) {
				while (iProp == null && types.moreItems ()) {
					AcmeElement sT = (AcmeElement )types.getNextItem ();
					iProp = sT.getProperty (prop.getName ());
				}
			}
			
			if (iProp == null) {
				AcmeElement p = parent.getParent ();
				types = getTypes (p);
				if (types != null && types.getSize () != 0) {
					AcmeElement sT = (AcmeElement )types.getNextItem ();
					Object child = sT.lookupName (((AcmeElement )parent).getName (), false);
					if (child != null && child instanceof AcmeElement) {
						AcmeElement c = (AcmeElement )child;
						iProp = c.getProperty (prop.getName ());
					}
				}
			}
			
			if (iProp.getScope () instanceof AcmeElement) {
				AcmeElement declared = (AcmeElement) iProp.getScope ();
				return declared;
			}
		}
    	throw new DesignRuleEvalException ("Error: Property is not declared in any element: " + prop.getQualifiedName());
    }


	/**
	 * @param parent
	 * @return
	 */
	private static AcmeEnumeration getTypes(AcmeElement object) {
		AcmeEnumeration types = null;
		if (object instanceof AcmeElementType) {
			types = gatherSuperTypes ((AcmeElementType )object);
		}
		else {
			types = object.getDeclaredTypes();
			types.concatenate(object.getInstantiatedTypes());
			types = types.castToAcmeSetVector().enumerate ();
		}
		return types;
	}


	/**
	 * @param type
	 * @return
	 */
	private static AcmeEnumeration gatherSuperTypes(AcmeElementType et) {
		AcmeEnumeration types = new AcmeEnumeration();
		AcmeEnumeration st = et.getSuperTypes().enumerate();
		types.concatenate(st);
		while (st.moreItems()) {
			AcmeElementType type = (AcmeElementType) st.getNextItem();
			types.concatenate(gatherSuperTypes(type));
		}
		return types.castToAcmeSetVector().enumerate();
	}
	
	/***************************************** MDS Extensions *******************************************/
	
	public static AcmeType narrowType (ExpressionNode elementNode, ExpressionNode typeNode) throws DesignRuleEvalException{
		
		AcmeElement element = elementNode.evaluateAsElement ();
		AcmeType t = typeNode.evaluateAsTypeRef ();
		if (!(t instanceof AcmeElementType)) {
			throw new DesignRuleEvalException (t.getTypeName () + " cannot be a property type in narrowType");
		}
		AcmeElementType type = (AcmeElementType )t;
		AcmeVector typesInHierarchy = new AcmeVector ();
		AcmeEnumeration declaredTypes = element.getDeclaredTypes ();
		// If this type is a subtype of many types, then go up one level so that
		// narrowType may be keyed on one of the supertypes.
		// TODO: Generalize this
		boolean wentUp = false;
		while (declaredTypes.getSize() == 1) {
			wentUp = true;
			declaredTypes = ((AcmeElementType )declaredTypes.getNextItem()).getSuperTypes().enumerate ();
		}
		if (declaredTypes.getSize () <=1 && wentUp) declaredTypes = element.getDeclaredTypes ();
		while (declaredTypes.moreItems ()) {
			AcmeElementType subType = (AcmeElementType )declaredTypes.getNextItem ();
			if (AcmeTypeManager.isDeclaredSubtype (subType, type)) {
				typesInHierarchy.addItem ((AcmeElement )subType);
			}
		}
		AcmeType lct = (AcmeType )AcmeTypeManager.inferGreatestCommonSupertype (typesInHierarchy);
		
		return lct;
	}
	

};
