package acmeLib.designRules;

import java.io.IOException;
import java.util.Vector;

import acmeLib.AcmeType;
import acmeLib.AcmeUnresolvedType;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ConversionConstants;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmeFamily;
import acmeLib.elements.AcmeUnresolvedElementType;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.DesignAnalysisParamError;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.errorManagement.TypeRefResolutionError;
import acmeLib.properties.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: DesignAnalysis<br>
 * Purpose: Stores an Armani DesignAnalysis definition.  This consists of a<br>
 *          name, a design rule expression (ExpressionNode), a vector of<br>
 *          formal parameters available to the expression,  a reference<br>
 *          to the parent object in which it was declared (its scope), and a<br>
 *          set of properties (meta-properties?) associated with the design<br>
 *          analysis.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    10/15/97   RTM      * Created prototype for this class (AbstractDesignRule)<br>
 *    10/28/97   RTM      * implemented parameter passing and evaluation<br>
 *    1/5/98     RTM      * Changed to DesignAnalysis to reflect new language<br>
 *                           feature that replaces nameable design rules with<br>
 *                           functions (analyses) that can be named and reused.<br>
 *    1/6/98     RTM      * Added support for non-boolean return values.<br>
 *                           Implemented the TypedEntity interface for typechecking.<br>
 *<br>
 *@author Bob Monroe
 */
public class DesignAnalysis
	extends acmeLib.AcmeObject
	implements
		acmeLib.designRules.Scope,
		acmeLib.designRules.TypedEntity,
		acmeLib.core.io.Exportable,
		acmeLib.core.io.ArmaniExportable {
	//
	// protected members
	//

	/**
	 * names the design analysis
	 */
	protected String name = null;

	/**
	 * stores this rule's expression tree
	 */
	protected ExpressionNode designRuleExpr = null;

	/**
	 * formal params referenced in this rule
	 */
	protected java.util.Vector formalParams = null;

	/**
	 * actual params referenced in this rule.  when a design rule invokes
	 * this analysis it provides a set of actual params that are stored in
	 * the protected variable actualParams, and the evaluation is done
	 * against those actual params.  This reference should not hold state
	 * between different evaluations.  It should be provided anew for each
	 * invocation of the design rule (although a caching scheme could be
	 * implemented later for efficiency, logically it would still be provided
	 * anew with each evaluation/invocation of the analysis).
	 */
	protected java.util.Vector actualParams = null;

	/**
	 * maps parameter names to {actual,formal}Params vector index
	 * the index is stored as an Integer object, from which the
	 * int value can be retrieved.
	 */
	protected java.util.Hashtable paramNameMap = null;

	/**
	 * scope in which this is declared
	 */
	protected acmeLib.designRules.Scope scope = null;

	/**
	 * store the return type of this analysis
	 */
	protected AcmeType returnType = null;

	/**
	 * true if this analysis has been successfully typechecked, else false.
	 */
	protected boolean isTypeCorrect;

	/**
	 * stores the properties of this design analysis.
	 */
	protected AcmeMap propMap;

	/**
	 * flag to indicate whether this is a link to an externally defined
	 * design analysis (a java method rather than an Armani expression).
	 */
	protected boolean isExternalAnalysis = false;

	//
	// public members
	//
	/**
	 * This is the constructor of this class
	 */
	public DesignAnalysis() {
		// initialize the protected members of this rule to defaults
		this.name = null;
		this.designRuleExpr = null;
		this.setFormalParams(new Vector(1, 1));
		this.setActualParams(new Vector(2, 1));
		this.scope = null;
		this.isTypeCorrect = false;
		this.isExternalAnalysis = false;
		this.returnType = acmeLib.core.globals.ConstantTypes.UNRESOLVED_TYPE;
		this.propMap = new AcmeMap();
	};

	/**
	 * This is the constructor of this class
	 * @param name The string
	 */
	public DesignAnalysis(String name) {
		// initialize the protected members of this rule to defaults
		this.name = name;
		this.designRuleExpr = null;
		this.setFormalParams(new Vector(1, 1));
		this.setActualParams(new Vector(2, 1));
		this.scope = null;
		this.isTypeCorrect = false;
		this.isExternalAnalysis = false;
		this.returnType = acmeLib.core.globals.ConstantTypes.UNRESOLVED_TYPE;
		this.propMap = new AcmeMap();
	};

	/**
	 * This is the constructor of this class
	 * @param name The name
	 * @param formalParameters The formal parameters
	 * @param scope The scope
	 * @param expr The expression node
	 */
	public DesignAnalysis(
		String name,
		java.util.Vector formalParameters,
		acmeLib.designRules.Scope scope,
		ExpressionNode expr) {
		this.name = name;
		this.designRuleExpr = expr;
		this.setFormalParams(formalParameters);
		this.setActualParams(new Vector(2, 1));
		this.scope = scope;
		this.isTypeCorrect = false;
		this.isExternalAnalysis = false;
		this.returnType = acmeLib.core.globals.ConstantTypes.UNRESOLVED_TYPE;
		this.propMap = new AcmeMap();
	};

	// accessor/settor methods
	/**
	 * Get the name
	 * @return The name
	 */
	public String getName() {
		return this.name;
	}
	/**
	 * Set the name
	 * @param newName The new name
	 * @exception acmeLib.core.exceptions.AcmeIllegalNameException If the name is illegal
	 */
	public void setName(String newName) throws acmeLib.core.exceptions.AcmeIllegalNameException {
		String oldName = this.getName();
		this.name = newName;
		if (this.getScope() instanceof AcmeElement) {
			((AcmeElement) this.getScope()).childNameChanged(oldName, this);
		};
	};

	/**
	 * Retrieves the fully qualified, period delimited, name of the design analysis.
	 * This name identifies an object in the structure tree that can be reached
	 * by taking a unique path from the root design.
	 * @return the fully qualified name of the AcmeElement
	 **/
	public String getQualifiedName() {
		StringBuffer nameBuf = new StringBuffer();

		if (this.getScope() != null) {
			String parentName = "";

			if (this.getScope() instanceof AcmeElement) {
				// all design analyses should be declared in the scope of an element...
				AcmeElement scopeElt = (AcmeElement) this.getScope();
				parentName = scopeElt.getQualifiedName();
			};

			if (parentName != null) {
				if (!(parentName.equals(""))) {
					nameBuf.append(parentName + ".");
				};
				nameBuf.append(this.getName());
			};
		}
		else {
			nameBuf.append(this.getName());
		};

		return nameBuf.toString();
	};

	/**
	 * Get the formal parameters
	 * @return the formal parameters vector
	 */
	public java.util.Vector getFormalParams() {
		return formalParams;
	};
	/**
	 * Set the formal parameters
	 * @param newParams The new formal parameters vector
	 */
	public void setFormalParams(java.util.Vector newParams) {
		this.formalParams = newParams;

		// load the paramNameMap with the references for the new set of
		// formal params.  The Vector passed as a parameter should contain
		// a vector of DesignRuleParameters.  use the name of each of these
		// parameters as the key into the Hashtable that stores their index
		// in the vector.
		this.paramNameMap = new java.util.Hashtable(formalParams.size() + 1);
		for (int i = 0; i < formalParams.size(); i++) {
			DesignRuleParameter curParam = null;
			if (this.formalParams.elementAt(i) instanceof DesignRuleParameter) {
				curParam = (DesignRuleParameter) this.formalParams.elementAt(i);
				this.paramNameMap.put(curParam.getName(), new Integer(i));
			}
			else {
				// this is an error, announce it.
				String errorMsg =
					"Attempted to add a non-DesignRuleParameter object: "
						+ this.formalParams.elementAt(i).toString()
						+ " to design analysis: "
						+ this.getName();
				DesignAnalysisParamError badParamError = new DesignAnalysisParamError(this, errorMsg);
				AcmeErrorManager.errorDetected(badParamError);
				AcmeTypeManager.typecheckingErrorStack.pushCausalError(badParamError);

			}
		};
	};

	/**
	 * Get the actual parameter
	 * @return the actual parameter vector
	 */
	public java.util.Vector getActualParams() {
		return actualParams;
	};
	/**
	 * Set the actual parameters
	 * @param newParams The new actual parameters vector
	 */
	public void setActualParams(java.util.Vector newParams) {
		this.actualParams = newParams;
	};

	/**
	 * Get the design rule expression
	 * @return the design rule expression
	 */
	public ExpressionNode getDesignRuleExpression() {
		return designRuleExpr;
	};
	/**
	 * Set the design rule expression
	 * @param newExpr The new design rule expression
	 */
	public void setDesignRuleExpression(ExpressionNode newExpr) {
		this.designRuleExpr = newExpr;
	};

	/**
	 * Get the scope
	 * @return the scope
	 */
	public acmeLib.designRules.Scope getScope() {
		return this.scope;
	};
	/**
	 * Set the scope
	 * @param newScope The new scope
	 */
	public void setScope(acmeLib.designRules.Scope newScope) {
		this.scope = newScope;
	};

	/**
	 * Get the isTypeCorrect boolean value
	 * @return the isTypeCorrect boolean value
	 */
	public boolean getIsTypeCorrect() {
		return this.isTypeCorrect;
	};
	/**
	 * Set the isTypeCorrect boolean value
	 * @param typeCorrect The new isTypeCorrect boolean value
	 */
	public void setIsTypeCorrect(boolean typeCorrect) {
		this.isTypeCorrect = typeCorrect;
	};

	/**
	 * Get the isExternalAnalysis boolean value
	 * @return the isExternalAnalysis boolean value
	 */
	public boolean getIsExternalAnalysis() {
		return this.isExternalAnalysis;
	};
	/**
	 * Set the isExternalAnalysis boolean value
	 * @param isExternal The new isExternalAnalysis boolean value
	 */
	public void setIsExternalAnalysis(boolean isExternal) {
		this.isExternalAnalysis = isExternal;
	};

	/**
	 * Get the return type
	 * @return the return AcmeType
	 */
	public AcmeType getReturnType() {
		return this.returnType;
	};
	/**
	 * Set the return AcmeType
	 * @param newType The new return AcmeType
	 */
	public void setReturnType(AcmeType newType) {
		this.returnType = newType;
	};

	/**
	 * Returns true if the variable name passed is a parameter of this
	 * design analysis, and thus volatile.  else returns false.
	 * @param variableName The variable name
	 * @return true if the variable name is volatile
	 */
	public boolean isVolatileVariable(String variableName) {
		// if variableName is stored in the paramNameMap it's volatile,
		// else it is not volatile.
		return (paramNameMap.get(variableName) != null);
	};

	/**
	 * allows the typechecking method to look up a parameter of this
	 * analysis by name.  Useful for typechecking.
	 * @param variableName The variable name
	 * @return The formal parameter
	 */
	public DesignRuleParameter lookupFormalParam(String variableName) {
		Integer paramIndex = (Integer) paramNameMap.get(variableName);
		if (paramIndex == null) {
			// then this variableName isn't a formal parameter for this
			// design analysis.
			return null;
		}
		return (DesignRuleParameter) formalParams.elementAt(paramIndex.intValue());

	};

	/**
	 * calculate the type of the expression node used in this analysis
	 * and typecheck the expression node's tree.  If the expression node does
	 * not typecheck return false.  If the expression does typecheck then see
	 * if the type of the expression matches the claimed return type
	 * of this design analysis.  return true if they do match, otherwise
	 * return false.
	 * @return true if type is correct
	 */
	public boolean typecheck() {
		boolean exprTypechecks;

		// typecheck the design analysis' expression.
		if (designRuleExpr == null) {
			this.setIsTypeCorrect(false);
			return false;
		}
		else {
			exprTypechecks = designRuleExpr.typecheck();
			this.setIsTypeCorrect(exprTypechecks);
		}

		// typecheck the formal parameters to confirm that all of the
		// types have been properly resolved.
		if (!exprTypechecks) {
			ExpressionTypecheckError err =
				new ExpressionTypecheckError(designRuleExpr, "Design rule does not typecheck", designRuleExpr);
			AcmeTypeManager.typecheckingErrorStack.packageErrorCauses(err);
			return false;
		}
		else {
			DesignRuleParameter curFormalParam = null;
			int numParams = 0;
			if (this.formalParams != null)
				numParams = this.formalParams.size();
			for (int i = 0;(i < numParams) && exprTypechecks; i++) {
				curFormalParam = (DesignRuleParameter) this.formalParams.elementAt(i);
				AcmeType curParamType = curFormalParam.getType();
				if ((curParamType == null)
					|| (curParamType instanceof AcmeUnresolvedType)
					|| (curParamType instanceof AcmeUnresolvedElementType)
					|| (curParamType instanceof AcmeUnresolvedPropertyType)) {
					exprTypechecks = false;
					this.setIsTypeCorrect(false);
					ExpressionTypecheckError err =
						new ExpressionTypecheckError(curFormalParam, "Parameter does not typecheck", designRuleExpr);
					AcmeTypeManager.typecheckingErrorStack.packageErrorCauses(err);
				};
			};
		};

		// finally, confirm that the analysis return type is a subtype of the
		// design rule expression type.
		if (!exprTypechecks) {

			this.setIsTypeCorrect(false);
			return false;
		}
		else {
			this.setIsTypeCorrect(AcmeTypeManager.isLogicalSubtype(designRuleExpr.getType(), this.returnType, false));
			if (this.getIsTypeCorrect()) {
				AcmeErrorManager.errorRepaired(this, null);
			}
			return this.getIsTypeCorrect();
		}
	};

	/**
	 * The evaluateAs* methods perform the design analysis with the actual
	 * parameters passed for evaluation
	 *
	 *
	 * Evaluate this DesignAnalysis' expression tree as a boolean expression.
	 * If it does not have the type "BOOLEAN_NODE" then throw a type exception.
	 * @param actualParams The actual parameter vector
	 * @return The evaluated boolean value
	 * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public boolean evaluateAsBoolean(Vector actualParams) throws DesignRuleEvalException {
		// make sure that there is a design rule expression to evaluate.
		if (designRuleExpr == null) {
			throw new DesignRuleEvalException(
				"Unable to evaluate design analysis as a boolean because "
					+ "the design analysis has no design rule expression to evaluate.");
		};

		// then make sure that the return type for this design analysis is boolean
		if (!(AcmeTypeManager.isLogicalSubtype(this.returnType, ConstantTypes.BOOLEAN_TYPE, false))) {
			throw new ExpressionCastException(
				null,
				this.getDesignRuleExpression(),
				ConstantTypes.BOOLEAN_TYPE,
				this.getReturnType(),
				"Unable to evaluate design analysis "
					+ this.getName()
					+ " as a boolean because the return type "
					+ "of the design analysis is not a subtype of boolean.");
		};

		// cache the actual parameters during evaluation
		java.util.Vector previousActualParams = this.actualParams;
		this.actualParams = actualParams;

		// then make sure that this expression is type correct.  If it is not
		// typecorrect, attempt to typecheck it, and if it fails to typecheck
		// throw a DesignRuleEvalException indicating the problem.
		if (this.getIsTypeCorrect() == false) {
			this.setIsTypeCorrect(designRuleExpr.typecheck());
			if (this.getIsTypeCorrect() == false)
				throw new DesignRuleEvalException(
					"Unable to evaluate design analysis as a boolean because "
						+ "the design analysis expression does not typecheck.");
		};

		// evaluate the expression with the provided actual params.
		boolean result;
		try {
			result = designRuleExpr.evaluateAsBoolean();
		}
		catch (DesignRuleEvalException dre) {
			this.actualParams = previousActualParams;
			throw dre;
		}

		// restore the params to their previous value and return the result
		this.actualParams = previousActualParams;
		return result;
	};

	/**
	 * Evaluates this analysis' expression tree as a float expression.
	 * If its return type is not a subtype of float then throw an
	 * ExpressionCastException.
	 * @param actualParams The actual parameters
	 * @return The evaluated float value
	 * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public float evaluateAsFloat(Vector actualParams) throws DesignRuleEvalException {
		// make sure that there is a design rule expression to evaluate.
		if (designRuleExpr == null) {
			throw new DesignRuleEvalException(
				"Unable to evaluate design analysis as a float because "
					+ "the design analysis has no design rule expression to evaluate.");
		};

		// then make sure that the return type for this design analysis is float
		if (!(AcmeTypeManager.isLogicalSubtype(this.returnType, ConstantTypes.FLOAT_TYPE, false))) {
			throw new ExpressionCastException(
				null,
				this.getDesignRuleExpression(),
				ConstantTypes.FLOAT_TYPE,
				this.getReturnType(),
				"Unable to evaluate design analysis "
					+ this.getName()
					+ " as a float because the return type "
					+ "of the design analysis is not a subtype of float.");
		};

		// then make sure that this expression is type correct.  If it is not
		// typecorrect, attempt to typecheck it, and if it fails to typecheck
		// throw a DesignRuleEvalException indicating the problem.
		if (this.getIsTypeCorrect() == false) {
			this.setIsTypeCorrect(designRuleExpr.typecheck());
			if (this.getIsTypeCorrect() == false)
				throw new DesignRuleEvalException(
					"Unable to evaluate design analysis as a float because "
						+ "the design analysis expression does not typecheck.");
		};

		// cache the actual parameters during evaluation
		java.util.Vector previousActualParams = this.actualParams;
		this.actualParams = actualParams;

		// evaluate the expression with the provided actual params.
		float result;
		try {
			result = designRuleExpr.evaluateAsFloat();
		}
		catch (DesignRuleEvalException dre) {
			this.actualParams = previousActualParams;
			throw dre;
		}

		// restore the params to their previous value and return the result
		this.actualParams = previousActualParams;
		return result;

	};

	/**
	 * Evaluates this analysis' expression tree as an int expression.
	 * If its return type is not a subtype of int then throw an
	 * ExpressionCastException.
	 * @param actualParams The actual parameters
	 * @return The evaluated integer
	 * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public int evaluateAsInt(Vector actualParams) throws DesignRuleEvalException {
		// make sure that there is a design rule expression to evaluate.
		if (designRuleExpr == null) {
			throw new DesignRuleEvalException(
				"Unable to evaluate design analysis as an int because "
					+ "the design analysis has no design rule expression to evaluate.");
		};

		// then make sure that the return type for this design analysis is float
		if (!(AcmeTypeManager.isLogicalSubtype(this.returnType, ConstantTypes.INT_TYPE, false))
			&& !(AcmeTypeManager.isLogicalSubtype(this.returnType, ConstantTypes.FLOAT_TYPE, false))) {
			throw new ExpressionCastException(
				null,
				this.getDesignRuleExpression(),
				ConstantTypes.INT_TYPE,
				this.getReturnType(),
				"Unable to evaluate design analysis "
					+ this.getName()
					+ " as an int because the return type "
					+ "of the design analysis is not a subtype of int.");
		};

		// then make sure that this expression is type correct.  If it is not
		// typecorrect, attempt to typecheck it, and if it fails to typecheck
		// throw a DesignRuleEvalException indicating the problem.
		if (this.getIsTypeCorrect() == false) {
			this.setIsTypeCorrect(designRuleExpr.typecheck());
			if (this.getIsTypeCorrect() == false) {
				throw new DesignRuleEvalException(
					"Unable to evaluate design analysis as an int because "
						+ "the design analysis expression does not typecheck.");
			}
		};

		// cache the actual parameters during evaluation
		java.util.Vector previousActualParams = this.actualParams;
		this.actualParams = actualParams;

		// evaluate the expression with the provided actual params.
		int result;
		try {
			result = designRuleExpr.evaluateAsInt();
		}
		catch (DesignRuleEvalException dre) {
			this.actualParams = previousActualParams;
			throw dre;
		}

		// restore the params to their previous value and return the result
		this.actualParams = previousActualParams;
		return result;
	};

	/**
	 * Evaluates this analysis' expression tree as a string expression.
	 * If its return type is not a subtype of string then throw an
	 * ExpressionCastException.
	 * @param actualParams The actual parameters
	 * @return The evaluated string
	 * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public String evaluateAsString(Vector actualParams) throws DesignRuleEvalException {
		// make sure that there is a design rule expression to evaluate.
		if (designRuleExpr == null) {
			throw new DesignRuleEvalException(
				"Unable to evaluate design analysis as a string because "
					+ "the design analysis has no design rule expression to evaluate.");
		};

		// then make sure that the return type for this design analysis is float
		if (!(AcmeTypeManager.isLogicalSubtype(this.returnType, ConstantTypes.STRING_TYPE, false))) {
			throw new ExpressionCastException(
				null,
				this.getDesignRuleExpression(),
				ConstantTypes.STRING_TYPE,
				this.getReturnType(),
				"Unable to evaluate design analysis "
					+ this.getName()
					+ " as a string because the return type "
					+ "of the design analysis is not a subtype of string.");
		};

		// then make sure that this expression is type correct.  If it is not
		// typecorrect, attempt to typecheck it, and if it fails to typecheck
		// throw a DesignRuleEvalException indicating the problem.
		if (this.getIsTypeCorrect() == false) {
			this.setIsTypeCorrect(designRuleExpr.typecheck());
			if (this.getIsTypeCorrect() == false)
				throw new DesignRuleEvalException(
					"Unable to evaluate design analysis as a string because "
						+ "the design analysis expression does not typecheck.");
		};

		// cache the actual parameters during evaluation
		java.util.Vector previousActualParams = this.actualParams;
		this.actualParams = actualParams;

		// evaluate the expression with the provided actual params.
		String result;
		try {
			result = designRuleExpr.evaluateAsString();
		}
		catch (DesignRuleEvalException dre) {
			this.actualParams = previousActualParams;
			throw dre;
		}

		// restore the params to their previous value and return the result
		this.actualParams = previousActualParams;
		return result;
	};

	/**
	 * Evaluate this DesignAnalysis' expression tree as an Element-typed
	 * expression. If it does not have an ElementType for a type then throw
	 * a type exception.
	 * @param actualParams The actual parameters
	 * @return The evaluted AcmeElement
	 * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public AcmeElement evaluateAsElement(Vector actualParams) throws DesignRuleEvalException {
		// make sure that there is a design rule expression to evaluate.
		if (designRuleExpr == null) {
			throw new DesignRuleEvalException(
				"Unable to evaluate design analysis as an Element because "
					+ "the design analysis has no design rule expression to evaluate.");
		};

		// then make sure that the return type for this design analysis is
		// an AcmeElementType
		if (!(this.returnType instanceof AcmeElementType)) {
			throw new ExpressionCastException(
				null,
				this.getDesignRuleExpression(),
				ConstantTypes.ELEMENT_BASE_TYPE,
				this.getReturnType(),
				"Unable to evaluate design analysis "
					+ this.getName()
					+ " as an Element because the return type "
					+ "of the design analysis is not a subtype of AcmeElementType.");
		};

		// then make sure that this expression is type correct.  If it is not
		// typecorrect, attempt to typecheck it, and if it fails to typecheck
		// throw a DesignRuleEvalException indicating the problem.
		if (this.getIsTypeCorrect() == false) {
			this.setIsTypeCorrect(designRuleExpr.typecheck());
			if (this.getIsTypeCorrect() == false)
				throw new DesignRuleEvalException(
					"Unable to evaluate design analysis as an Element because "
						+ "the design analysis expression does not typecheck.");
		};

		// cache the actual parameters during evaluation
		java.util.Vector previousActualParams = this.actualParams;
		this.actualParams = actualParams;

		// evaluate the expression with the provided actual params.
		AcmeElement result;
		try {
			result = designRuleExpr.evaluateAsElement();
		}
		catch (DesignRuleEvalException dre) {
			this.actualParams = previousActualParams;
			throw dre;
		}

		// restore the params to their previous value and return the result
		this.actualParams = previousActualParams;
		return result;
	};

	/**
	 * Evaluate this DesignAnalysis' expression tree as an PropertyReference-typed
	 * expression. If it does not have a PropertyReferenceType for a type then
	 * throw a type exception.
	 * @param actualParams The actual parameters
	 * @return The evaluted AcmeProperty
	 * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public AcmeProperty evaluateAsPropertyRef(Vector actualParams) throws DesignRuleEvalException {
		// make sure that there is a design rule expression to evaluate.
		if (designRuleExpr == null) {
			throw new DesignRuleEvalException(
				"Unable to evaluate design analysis as a Property Reference because "
					+ "the design analysis has no design rule expression to evaluate.");
		};

		// then make sure that the return type for this design analysis is
		// an AcmeElementType
		if (!(this.returnType instanceof PropertyReferenceType)) {
			throw new ExpressionCastException(
				null,
				this.getDesignRuleExpression(),
				ConstantTypes.PROPERTY_REFERENCE_TYPE,
				this.getReturnType(),
				"Unable to evaluate design analysis "
					+ this.getName()
					+ " as a Property Reference because the return type "
					+ "of the design analysis is not a subtype of PropertyReferenceType.");
		};

		// then make sure that this expression is type correct.  If it is not
		// typecorrect, attempt to typecheck it, and if it fails to typecheck
		// throw a DesignRuleEvalException indicating the problem.
		if (this.getIsTypeCorrect() == false) {
			this.setIsTypeCorrect(designRuleExpr.typecheck());
			if (this.getIsTypeCorrect() == false)
				throw new DesignRuleEvalException(
					"Unable to evaluate design analysis as a Property Reference because "
						+ "the design analysis expression does not typecheck.");
		};

		// cache the actual parameters during evaluation
		java.util.Vector previousActualParams = this.actualParams;
		this.actualParams = actualParams;

		// evaluate the expression with the provided actual params.
		AcmeProperty result;
		try {
			result = designRuleExpr.evaluateAsPropertyRef();
		}
		catch (DesignRuleEvalException dre) {
			this.actualParams = previousActualParams;
			throw dre;
		}

		// restore the params to their previous value and return the result
		this.actualParams = previousActualParams;
		return result;
	};

	public AcmeType evaluateAsTypeRef(Vector actualParams) throws DesignRuleEvalException {
		// make sure that there is a design rule expression to evaluate.
		if (designRuleExpr == null) {
			throw new DesignRuleEvalException(
				"Unable to evaluate design analysis as a Set because "
					+ "the design analysis has no design rule expression to evaluate.");
		};

		if (!(this.returnType instanceof AcmeType)) {

			throw new ExpressionCastException(
				null,
				this.getDesignRuleExpression(),
				ConstantTypes.TYPE_REFERENCE_TYPE,
				this.getReturnType(),
				"Unable to evaluate design analysis "
					+ this.getName()
					+ " as a type reference because the return type "
					+ "of the design analysis is not a subtype of AcmeType.");
		};

		// cache the actual parameters during evaluation
		java.util.Vector previousActualParams = this.actualParams;
		this.actualParams = actualParams;

		// evaluate the expression with the provided actual params.
		AcmeType result;
		try {
			result = designRuleExpr.evaluateAsTypeRef();
		}
		catch (DesignRuleEvalException dre) {
			this.actualParams = previousActualParams;
			throw dre;
		}

		// restore the params to their previous value and return the result
		this.actualParams = previousActualParams;
		return result;
	}

	/**
	 * Evaluate this DesignAnalysis' expression tree as a Set-typed
	 * expression. If it does not have a SetType for a type then throw
	 * a type exception.
	 * @param actualParams The actual parameters
	 * @return The evaluated set expression node
	 * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public SetExpressionNode evaluateAsSet(Vector actualParams) throws DesignRuleEvalException {
		// make sure that there is a design rule expression to evaluate.
		if (designRuleExpr == null) {
			throw new DesignRuleEvalException(
				"Unable to evaluate design analysis as a Set because "
					+ "the design analysis has no design rule expression to evaluate.");
		};

		// then make sure that the return type for this design analysis is
		// an AcmeElementType
		if (!(this.returnType instanceof SetType)) {
			throw new ExpressionCastException(
				null,
				this.getDesignRuleExpression(),
				ConstantTypes.ANY_SET_TYPE,
				this.getReturnType(),
				"Unable to evaluate design analysis "
					+ this.getName()
					+ " as a Set because the return type "
					+ "of the design analysis is not a subtype of SetType.");
		};

		// then make sure that this expression is type correct.  If it is not
		// typecorrect, attempt to typecheck it, and if it fails to typecheck
		// throw a DesignRuleEvalException indicating the problem.
		if (this.getIsTypeCorrect() == false) {
			this.setIsTypeCorrect(designRuleExpr.typecheck());
			if (this.getIsTypeCorrect() == false)
				throw new DesignRuleEvalException(
					"Unable to evaluate design analysis as a Set because "
						+ "the design analysis expression does not typecheck.");
		};

		// cache the actual parameters during evaluation
		java.util.Vector previousActualParams = this.actualParams;
		this.actualParams = actualParams;

		// evaluate the expression with the provided actual params.
		SetExpressionNode result;
		try {
			result = designRuleExpr.evaluateAsSet();
		}
		catch (DesignRuleEvalException dre) {
			this.actualParams = previousActualParams;
			throw dre;
		}

		// restore the params to their previous value and return the result
		this.actualParams = previousActualParams;
		return result;
	};

	/**
	 * DesignAnalysis implements the Scope interface so that
	 * the variables that are bound in the expression can be looked
	 * up in the vector of actual parameters first, and if it is not
	 * found there then it passes the request up to the design analysis'
	 * parent.
	 * @param name The name
	 * @param bubbleUp passes the request up
	 * @return The object
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		// the vector index of the actual parameters is stored in the
		// paramNameMap hashtable, which has to be filled in for the
		// current set of formal parameters before lookupName
		// will find the names appropriately.
		Object namedObject = null;
		Integer nameVectorLocation = (Integer) paramNameMap.get(name);

		if (nameVectorLocation == null)
			namedObject = null; // name requested for lookup is not a parameter.
		else if ((nameVectorLocation.intValue() < 0) || (nameVectorLocation.intValue() > (formalParams.size() - 1))) {
			// then our variable is out of the legal range of formal or
			// actual params.  this is an internal error. return null.
			return null;
		}
		else {
			// create a VolatileVariable object that will store the type
			// of the variable, and, if it is available, the current value
			// of the variable also.
			int paramIndex = nameVectorLocation.intValue();
			VolatileVariable var = new VolatileVariable();
			DesignRuleParameter curFormalParam = (DesignRuleParameter) this.formalParams.elementAt(paramIndex);

			var.refObjectType = curFormalParam.getType();
			var.declarationScope = this;

			if ((actualParams != null) && (actualParams.size() > 0)) {
					Object curActualParam = null;
					try {
						curActualParam = this.actualParams.elementAt(paramIndex);
						var.referentObject = curActualParam;
					}
					catch (ArrayIndexOutOfBoundsException e) {
						var.referentObject = null;
					}
					catch (NullPointerException ne) {
						var.referentObject = null;
					};
			}
			else {
				var.referentObject = null;
			};

			// we've found the named object, return the VolatileVariable
			// with the appropriate fields filled in.
			namedObject = var;
		};

		// if we haven't found anything, then bubbleUp
		if (namedObject == null) {
			namedObject = this.getScope().lookupName(name, true);
		};

		return namedObject;
	};

	/**
	 * Implement the Exportable interface -- export this DesignAnalysis
	 * @return The output string
	 */

	public String toString() {
		String ret = "";
		AcmeOutputStream s = new AcmeOutputStream();
		try {
			exportArmani(s);
			s.flush();
			ret = s.toString();
			s.close();
		}
		catch (IOException e) {
		}
		catch (AcmeException e) {
		}

		return ret;
	};

	/**
	 * export
	 * @param prefix The prefix
	 * @return The output string
	 */
	public String toString(String prefix) {
		return prefix + this.toString();
	};

	/**
	 * implement the Exportable interface
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException For I/O exception
	 * @exception acmeLib.core.exceptions.AcmeException For AcmeException
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, acmeLib.core.exceptions.AcmeException {
		StringBuffer buf = new StringBuffer("Property ");
		buf.append(ConversionConstants.ARMANI_ANALYSIS_PROP_NAME_AS_ACME_PROP);
		buf.append("-" + String.valueOf(this.getCreationOrder()) + " : ");
		buf.append(ConversionConstants.ARMANI_ANALYSIS_TYPE_NAME_AS_ACME_PROP);
		buf.append(" = " + AcmeString.massageForExport(this.toString()));

		// if this is defined in a type (family) declaration, then add a
		// meta-property to the exported method annotating as much.
		if (this.getScope() instanceof AcmeElementType) {
			String typename = ((AcmeElementType) this.getScope()).getTypeName();
			buf.append(
				" << "
					+ ConversionConstants.DECLARED_IN_TYPENAME_METAPROP
					+ " = "
					+ AcmeString.massageForExport(typename)
					+ "; >>");
		};

		buf.append(";");
		s.writeln(buf.toString());
	};

	/**
	 * implement the ArmaniExportable interface
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException For I/O exception
	 * @exception acmeLib.core.exceptions.AcmeException For AcmeException
	 */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, acmeLib.core.exceptions.AcmeException {
		exportComment(s);
		s.beginRegion(this);
		if (this.getIsExternalAnalysis()) {
			s.write("external analysis ");
		}
		else {
			s.write("analysis ");
		};

		s.write(this.getName());

		// export the formal params
		s.write("(");
		java.util.Enumeration params = this.getFormalParams().elements();
		DesignRuleParameter curParam;
		while (params.hasMoreElements()) {
			curParam = (DesignRuleParameter) params.nextElement();
			curParam.exportArmani(s);

			if (params.hasMoreElements())
				s.write(", ");
		};

		s.write(") ");

		// export the return type.
		if (this.getReturnType() != null) {
			s.write(": " + this.getReturnType().getTypeName());
		};

		// export the analysis expression
		s.writeln(" =");
		s.indent();
		if (this.designRuleExpr != null) {
			designRuleExpr.exportArmani(s);
		}
		else {
			s.write("<no expression specified>");
		};

		s.writeln(";");
		s.endRegion(this);
		s.unIndent();
	};

	/**
	 * resolve the typenames of any unresolved types in the list of formal
	 * parameters and this DesignAnalysis' return type.
	 */
	public void resolveTypeReferences() throws AcmeResolutionException {
		String errorMsg = null;
		if (this.scope == null) {
			// then we have no place to look up the typenames so don't
			// bother trying...  (announce an error first, though).
			errorMsg =
				"Design Analysis: "
					+ this.getName()
					+ " does not have a valid parent scope for resolving"
					+ " parameter type names.";
			TypeRefResolutionError badTypeRefError = new TypeRefResolutionError(this, errorMsg);
			AcmeErrorManager.errorDetected(badTypeRefError);
			AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

			throw new AcmeResolutionException(badTypeRefError);
		}
		else {
			// first resolve this design analysis' return type.
			if ((this.returnType instanceof AcmeUnresolvedType)
				|| (this.returnType instanceof AcmeUnresolvedPropertyType)
				|| (this.returnType instanceof AcmeUnresolvedElementType)) {

				Object referentTypeObj = null;
				if (this.returnType instanceof AcmeUnresolvedPropertyType) {
					AcmeUnresolvedPropertyType t = (AcmeUnresolvedPropertyType) this.returnType;
					if (t.isQualified()) {
						Object fam = scope.lookupName(t.getQualifiedName(), true);
						if (fam instanceof AcmeFamily) {
							AcmeFamily f = (AcmeFamily) fam;
							referentTypeObj = f.getType(t.getTypeName());
						}
					}
				}
				else if (this.returnType instanceof AcmeUnresolvedElementType) {
					AcmeUnresolvedElementType t = (AcmeUnresolvedElementType) this.returnType;
					if (t.isQualified()) {
						Object fam = scope.lookupName(t.getQualifier(), true);
						if (fam instanceof AcmeFamily) {
							referentTypeObj = ((AcmeFamily) fam).getType(t.getTypeName());
						}
					}
				}
				else {
					// then lookup the reference to the return type object and
					// change this' returnType to the found type object.
					referentTypeObj = this.scope.lookupName(this.returnType.getTypeName(), true);
				}

				if (referentTypeObj instanceof AcmeType) {
					this.setReturnType((AcmeType) referentTypeObj);
				}
				else {
					// Else we didn't find a legal type.
					// Announce a type resolution error.
					errorMsg =
						"Invalid return type reference in design analysis: "
							+ this.getName()
							+ "()'s return type \""
							+ this.returnType.getTypeName()
							+ "\"";
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, this.returnType.getTypeName());
					badTypeRefError.setUnresolvedRef(this.returnType);
					AcmeErrorManager.errorDetected(badTypeRefError);
					AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

					throw new AcmeResolutionException(badTypeRefError);
				};

			}
			else if (this.returnType == null) {
				// Error. Announce a type resolution error.
				errorMsg = "Invalid return type in design analysis: " + this.getName() + "(). Null type object.";
				TypeRefResolutionError badTypeRefError = new TypeRefResolutionError(this, errorMsg);
				AcmeErrorManager.errorDetected(badTypeRefError);
				AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

				// set the return type to a non-null, non-legal type
				this.returnType = ConstantTypes.UNRESOLVABLE_TYPE;
				throw new AcmeResolutionException(badTypeRefError);
			};

			// then resolve the types of any unresolved parameters.
			DesignRuleParameter curParam = null;
			int numParams = 0;
			if (this.formalParams != null)
				numParams = this.formalParams.size();
			for (int i = 0; i < numParams; i++) {
				curParam = (DesignRuleParameter) this.formalParams.elementAt(i);
				AcmeType curParamType = curParam.getType();
				if ((curParamType instanceof AcmeUnresolvedType)
					|| (curParamType instanceof AcmeUnresolvedElementType)
					|| (curParamType instanceof AcmeUnresolvedPropertyType)) {
					Object referentObj = null;
					if (curParamType instanceof AcmeUnresolvedPropertyType) {
						AcmeUnresolvedPropertyType t = (AcmeUnresolvedPropertyType) curParamType;
						if (t.isQualified()) {
							Object fam = scope.lookupName(t.getQualifiedName(), true);
							if (fam instanceof AcmeFamily) {
								AcmeFamily f = (AcmeFamily) fam;
								referentObj = f.getType(t.getTypeName());
							}
						}
					}
					else if (curParamType instanceof AcmeUnresolvedElementType) {
						AcmeUnresolvedElementType t = (AcmeUnresolvedElementType) curParamType;
						if (t.isQualified()) {
							Object fam = scope.lookupName(t.getQualifier(), true);
							if (fam instanceof AcmeFamily) {
								referentObj = ((AcmeFamily) fam).getType(t.getTypeName());
							}
						}
					}
					else {
						// then lookup the reference to that type object and
						// change the reference to the found type object.
						referentObj = this.scope.lookupName(curParamType.getTypeName(), true);
					}
					if (referentObj instanceof AcmeType)
						curParam.setType((AcmeType) referentObj);
					else {
						// Else we didn't find a legal type.
						// Announce a type resolution error.
						errorMsg =
							"Invalid parameter type reference for \""
								+ curParam.getName()
								+ " : "
								+ curParamType.getTypeName()
								+ "\" in design analysis "
								+ this.getName();

						TypeRefResolutionError badTypeRefError =
							new TypeRefResolutionError(this, errorMsg, curParamType.getTypeName());
						badTypeRefError.setUnresolvedRef(this.returnType);
						AcmeErrorManager.errorDetected(badTypeRefError);
						AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

						// if we didn't find a legal type object to refer to then
						// fall through and leave the current param's type object
						// an AcmeUnresolvedType.
						throw new AcmeResolutionException(badTypeRefError);
					};

				}
				//                else if (curParamType instanceof AcmeTypeParameter) {
				//                    Object referentObj = this.lookupName (curParam.getName(), true);
				//                    
				//                    if (referentObj instanceof AcmeType) {
				//                        curParam.setType ((AcmeType )referentObj);
				//                    }
				//                    else {
				//                        errorMsg = "Invalid parameter type reference for \""
				//                                    + curParam.getName () + " : "
				//                                    + curParamType.getTypeName () 
				//                                    + " \" in design analysis " + this.getName ();
				//                        TypeRefResolutionError badTypeRefError = 
				//                            new TypeRefResolutionError (this, errorMsg, curParamType.getTypeName());
				//                        badTypeRefError.setUnresolvedRef (this.returnType);
				//                        AcmeErrorManager.errorDetected (badTypeRefError);
				//                    }
				//                }
				else if (curParamType == null) {
					// then explicitly set the type to an unresolved type
					// object to prevent later methods from crashing when
					// they hit the null type.  First announce a type res. error.
					errorMsg =
						"Invalid parameter type in design analysis: "
							+ this.getName()
							+ "(). Parameter "
							+ curParam.getName()
							+ "'s type is null.";
					TypeRefResolutionError badTypeRefError = new TypeRefResolutionError(this, errorMsg);
					AcmeErrorManager.errorDetected(badTypeRefError);
					AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

					curParam.setType(ConstantTypes.UNRESOLVABLE_PROPERTY_TYPE);
					throw new AcmeResolutionException(badTypeRefError);
				};
			};
		};

		// then resolve the type references of any expression nodes in this
		// analysis that include types
		if (this.designRuleExpr != null)
			this.designRuleExpr.resolveTypeReferences();

	}; // end resolveTypeReferences()

	/**
	 * Traverse the expression tree setting all variable references in
	 * the tree to volatile and resolving any instance references (which
	 * should generally cause an error, as design analysis expressions
	 * can't explicitly refer to instances declared outside of the
	 * analysis.  If there is such a reference it should generate an
	 * error that is announced to the ErrorManager.
	 */
	public void resolveInstanceReferences() throws AcmeResolutionException {
		if (this.designRuleExpr != null) {
			this.designRuleExpr.resolveVolatileFlag();
			this.designRuleExpr.resolveInstanceReferences();
		};
	}

	/**
	 * @param key
	 */
	public void clearAllUserData(Object key) {
		designRuleExpr.clearAllUserData(key);
	}

	/**
	 * @param key
	 */
	public void setEvaluationKey(Object key) {
		designRuleExpr.setEvaluationKey(key);
	};

}
