package acmeLib.designRules;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;
import acmeLib.AcmeType;
import acmeLib.AcmeUnresolvedType;
import acmeLib.Debug;
import acmeLib.armaniParser.ArmaniParser;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.designRules.feedback.DesignAnalysisCallEvaluationFeedback;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AnyType;
import acmeLib.properties.PropertyReferenceType;
import acmeLib.properties.SetType;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: DesignAnalysisCallNode<br>
 * Purpose: A DesignAnalysisCallNode node is an expression node that represents<br>
 *          a call to a function (aka invocation of a DesignAnalysis).<br>
 *          This class makes use of the capabilities provided by ReferenceNode<br>
 *          (its superclass), including scope management, storing qualified<br>
 *          names, and reference lookup. It also stores a vector of actual<br>
 *          parameters.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    1/5/97     RTM      * Modified previous DesignRuleCallNode to reflect<br>
 *                           new language structuring in which all functions<br>
 *                           are DesignAnalyses.<br>
 *    1/6/03     SWC      * Modified toString() to print out the referencd<br>
 *                           object of the actual parameters.<br>
 *<br>
 * @author Bob Monroe
 */
public class DesignAnalysisCallNode extends ReferenceNode
                                    implements acmeLib.designRules.TypedEntity
{

  //
  // protected members
  //

/**
 * a list of the actual parameters used to evaluate the design rule.
 * the elements stored within this vector are ExpressionNode objects.
 */
  protected Vector actualParams;


  //
  // public members
  //
/**
 * This is the constructor for this class
 * @param id The ID
 */
  public DesignAnalysisCallNode(int id) {
    super(id);
    actualParams = new java.util.Vector(4,2);

    // designAnalysisCallNodes can not be volatile as Armani has no
    // higher-order functions.
    this.setIsVolatile(false);

    this.setType(ConstantTypes.UNRESOLVED_TYPE);
  }

/**
 * @param id The ID
 * @return The node
 */
  public static Node jjtCreate(int id) {
    return new DesignAnalysisCallNode(id);
  }

/**
 * Copy constructor
 * @param copyOf The design analysis call node
 */
  public DesignAnalysisCallNode(DesignAnalysisCallNode copyOf) {
    super(copyOf);
    this.setReferentDesignAnalysis(copyOf.getReferentDesignAnalysis());
    this.setActualParams(copyOf.getActualParams());
    this.setIsVolatile(copyOf.getIsVolatile());
  }



/** cloneExpressionTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.
 * @return The copy of the expression tree
 */
  public ExpressionNode cloneExpressionTree() {
    DesignAnalysisCallNode clone = new DesignAnalysisCallNode(this);
    this.doCloneExpressionTree(clone);
    clone.setActualParams(this.getActualParams());
    return clone;
  };


/**
 * Access the DesignAnalysis referenced by this node
 * @return The design analysis
 */
  public DesignAnalysis getReferentDesignAnalysis() {
    if (this.getReferentObject() instanceof DesignAnalysis)
        return (DesignAnalysis) this.getReferentObject();
    return null;
  };
/**
 * Set the design analysis
 * @param newAnalysis The new design analysis
 */
  public void setReferentDesignAnalysis(DesignAnalysis newAnalysis) {
    this.setReferentObject(newAnalysis);
  };


/**
 * Access this node's current vector of actual parameters
 * @return The acutal parameters vector
 */
  public Vector getActualParams() {
    return this.actualParams;
  };
/**
 * Set the new actual parameters vector
 * @param newActualParams The new actual parameters
 */
  public void setActualParams(Vector newActualParams) {
    this.actualParams = (Vector)newActualParams.clone();
  }

/**
 * load the actualParameter vector with the children of this node, which
 * should be stored as a vector of the actual arguments that will be
 * passed to the analysis being called.  If this node has no children
 * (and thus no actual parameters), then don't do anything, using only
 * the default empty vector instead.
 */
  public void loadActualParams()  {
    if (this.children != null){
        Vector temp = new Vector();
        for (int i = 0; i < this.children.length; i++){
            temp.addElement(this.children[i]);
        }
        this.setActualParams(temp);
    }
  }


/**
 * typecheck infers the type of this node based on the return type of
 * the design analysis it references.  It then calls typecheck on each
 * of its actual parameters.  Finally, it checks that the actual parameters
 * satisfy the types required by the referenced analysis' formal params.
 * If all of this is successful, then it returns true.  If it runs into
 * trouble anywhere, it returns false.
 * @return true if it is type correct
 */
  public boolean typecheck() {


    // if referentObject is null or we don't have an analysis to typecheck
    // then return false.
    if (this.getReferentDesignAnalysis() == null) {
        try {
            this.refreshReferentObject();
        }
        catch (AcmeResolutionException e) {
            return false;
        }
    };

    if (this.getReferentDesignAnalysis() == null) {
        return false;
    } else {
        this.setType(this.getReferentDesignAnalysis().getReturnType());
    };


    // assume that the actualParams typecheck, unless we can prove otherwise.
    boolean actualParamsTypecheck = true;

    // load the actualParameters vector with the dereferenced values
    // of the actual parameters stored in children nodes.
    loadActualParams();

    // get the formal params of the referent design analysis, and confirm
    // that the type of each of the actual params is a subtype of the required
    // formal param type.  The actual params are stored as children nodes
    // of this node, retrieve them using the jjtGetChild(index) method.
    java.util.Vector formalParams =
                        this.getReferentDesignAnalysis().getFormalParams();

    if (formalParams == null)
        return false;       // internal error.

    if (jjtGetNumChildren() != formalParams.size ()) {
        actualParamsTypecheck = false;
        ExpressionTypecheckError err = new ExpressionTypecheckError (this, "Expected " + formalParams.size () + " parameters, got " + this.jjtGetNumChildren(), this);
        AcmeTypeManager.typecheckingErrorStack.pushCausalError (err);
        AcmeErrorManager.errorDetected (err);
        return false;
    }

    for (int i = 0; (i < formalParams.size() /*- 1*/) /*&& actualParamsTypecheck*/; i++)
    {
        try {
        	((ExpressionNode )jjtGetChild (i)).typecheck();
            AcmeType actualParamType = ((ExpressionNode )jjtGetChild (i)).getType ();
            AcmeElement elem = null;
            boolean doLogical = true;
            if (actualParamType instanceof AcmeUnresolvedType || ConstantTypes.isBaseType(actualParamType)) {
                if (jjtGetChild (i) instanceof IdentifierNode) {
                    
                	IdentifierNode	identifierNode	= (IdentifierNode )jjtGetChild (i);
					if ((identifierNode).getReferentObject() instanceof AcmeType) {
                        actualParamType = (AcmeType )(identifierNode).getReferentObject();
                    }
                    else if ((identifierNode).getReferentObject() instanceof VolatileVariable) {
                    	// TODO: If this is the case, then it isn't good enough to juse get one type, need to get all of the types
                        actualParamType = ((VolatileVariable )(identifierNode).getReferentObject()).refObjectType;
                    }
                    else if ((identifierNode).getReferentObject() instanceof AcmeProperty) {
                    	actualParamType = ((AcmeProperty )(identifierNode).getReferentObject ()).getType();
                    }
                    // BugFix#285: If the referentObject is an AcmeElement, we don't need to rely
                    // on the referentType so just get the referentObject
                    else if (identifierNode.getReferentObject () instanceof AcmeElement) {
                    	elem = (AcmeElement )identifierNode.getReferentObject();
                    	doLogical = false;
                    }
                    else {
                        elem = (AcmeElement )this.getReferenceScope ().lookupName((identifierNode).getIdentifier(), true);
                        doLogical = false;
                    }
                }
                else if (jjtGetChild (i) instanceof VolatileVariable) {
                	// TODO: If this is the case, then it isn't good enough to juse get one type, need to get all of the types
                    if (((VolatileVariable )jjtGetChild (i)).refObjectType != null) {
                        actualParamType =    ((VolatileVariable )jjtGetChild (i)).refObjectType;
                    }
                }
                else if (jjtGetChild (i) instanceof DesignAnalysisCallNode) {
                    actualParamType = ((DesignAnalysisCallNode )jjtGetChild (i)).getReferentDesignAnalysis().getReturnType ();
                }
            }  
            boolean thisParamTypecheck = false;
            if (doLogical) {
            
                thisParamTypecheck = 
                    AcmeTypeManager.isLogicalSubtype(
                    		actualParamType,
                    ((DesignRuleParameter) formalParams.elementAt(i)).getType(), 
					false);
            }
            else {
                thisParamTypecheck = AcmeTypeManager.declaresType(((DesignRuleParameter) formalParams.elementAt(i)).getType(), elem);
            }
            actualParamsTypecheck = actualParamsTypecheck && thisParamTypecheck;
            if (!thisParamTypecheck) {
                ExpressionTypecheckError err = 
                    new ExpressionTypecheckError ((ExpressionNode )this.jjtGetChild(i), ((ExpressionNode )this.jjtGetChild (i)).toString() + " is incorrect or unknown type", 
                        this);
                AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
//            } else {
//            	// [SWC 02-July-2004] Cancel any error in referencing an AcmeRecordField
//            	if (jjtGetChild(i) instanceof IdentifierNode) {
//            		IdentifierNode idNode = (IdentifierNode )jjtGetChild(i);
//            		if (idNode.isVolatile && idNode.getReferentObject() instanceof AcmeRecordField) {
//            			AcmeEnumeration e = AcmeTypeManager.typecheckingErrorStack.getCausalErrors().enumerate();
//            			while (e.moreItems()) {
//            				AcmeError err = (AcmeError )e.getNextItem();
//            				if (err.getSource() == idNode) {
//            					AcmeTypeManager.typecheckingErrorStack.cancelError(err);
//            					break;
//            				}
//            			}
//            		}
//            	}
            }
        } catch (ClassCastException e) {
            actualParamsTypecheck = false;
        };
    };

//     use the superclass to recursively typecheck all children nodes.
    if (super.typecheck() == false) {
        this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
        return false;
    };
    
    return actualParamsTypecheck;
  }


	/**
	 * @param key
	 * @see acmeLib.designRules.ExpressionNode#clearAllUserData(java.lang.Object)
	 **/
	public void clearAllUserData(Object key) {
		super.clearAllUserData (key);
		if (this.getReferentDesignAnalysis() != null) this.getReferentDesignAnalysis().clearAllUserData (key);
	}
	/**
	 * @param key
	 * @see acmeLib.designRules.ExpressionNode#setEvaluationKey(java.lang.Object)
	 **/
	void setEvaluationKey(Object key) {
		super.setEvaluationKey (key);
		if (this.getReferentDesignAnalysis() != null) this.getReferentDesignAnalysis().setEvaluationKey (key);
	}
/**
 * EvaluateAsBoolean calculates the actual parameters for the call
 * and invokes the referred to design rule with the current set of
 * actual parameters.  The actual parameters are initially parsed
 * as children nodes of the design rule call node.
 * @return The evaluated boolean value
 * @exception DesignRuleEvalException If thereis an evaluation error
 */
  public synchronized boolean evaluateAsBoolean() throws DesignRuleEvalException {

    if (this.getReferentDesignAnalysis() == null)
        this.refreshReferentObject ();
        if (this.getReferentDesignAnalysis () == null) {
            
            throw new DesignRuleEvalException("Unable to dereference the "
                    + "identifier " + this.getIdentifier()
                    + " or one of its arguments as a design analysis." );
        }
    if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                           ConstantTypes.BOOLEAN_TYPE, false)))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate DesignAnalysis: "
                + this.getIdentifier() + "as a boolean expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.BOOLEAN_TYPE, this.getType(), opMsg);
    };

    // load the actualParameters vector with the dereferenced values
    // of the actual parameters stored in children nodes.
    loadActualParams();

    // evaluate the design analysis with the assigned actual parameters
    Debug.trace_call (this);
    boolean retVal = this.getReferentDesignAnalysis().evaluateAsBoolean(
                                                    this.getActualParams());
    if (DesignRule.isRememberVisited()) {
    	List l = (List )getUserData(getEvaluationKey());
    	if (l == null) {
    		l = new ArrayList ();
    		setUserData (getEvaluationKey(), l);
    	}
    	l.add (new DesignAnalysisCallEvaluationFeedback (this, getActualParams(), Boolean.valueOf (retVal)));
    }
    Debug.trace_return (this, retVal);
	return retVal;
  };



/**
 * EvaluateAsInt calculates the actual parameters for the call
 * and invokes the referred to design rule with the current set of
 * actual parameters.  The actual parameters are initially parsed
 * as children nodes of the design rule call node.
 * @return The evaluated integer
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public int evaluateAsInt() throws DesignRuleEvalException {
  	
    if (this.getReferentDesignAnalysis() == null)
        throw new DesignRuleEvalException("Unable to dereference the "
                + "identifier " + this.getIdentifier()
                + " or one of its arguments as a design analysis." );

    if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                           ConstantTypes.INT_TYPE, false)))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate DesignAnalysis: "
                + this.getIdentifier() + "as a boolean expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.INT_TYPE, this.getType(), opMsg);
    };

    // load the actualParameters vector with the dereferenced values
    // of the actual parameters stored in children nodes.
    loadActualParams();

    // evaluate the design analysis with the assigned actual parameters
    return this.getReferentDesignAnalysis().evaluateAsInt(this.getActualParams());
  };




/**
 * EvaluateAsFloat calculates the actual parameters for the call
 * and invokes the referred to design rule with the current set of
 * actual parameters.  The actual parameters are initially parsed
 * as children nodes of the design rule call node.
 * @return The evaluated float
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public float evaluateAsFloat() throws DesignRuleEvalException {

    if (this.getReferentDesignAnalysis() == null)
        throw new DesignRuleEvalException("Unable to dereference the "
                + "identifier " + this.getIdentifier()
                + " or one of its arguments as a design analysis." );

    if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                           ConstantTypes.FLOAT_TYPE, false)))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate DesignAnalysis: "
                + this.getIdentifier() + "as a float expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);
    };

    // load the actualParameters vector with the dereferenced values
    // of the actual parameters stored in children nodes.
    loadActualParams();

    // evaluate the design analysis with the assigned actual parameters
    return this.getReferentDesignAnalysis().evaluateAsFloat(this.getActualParams());
  };




/**
 * EvaluateAsString calculates the actual parameters for the call
 * and invokes the referred to design rule with the current set of
 * actual parameters.  The actual parameters are initially parsed
 * as children nodes of the design rule call node.
 * @return The evaluated string
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public String evaluateAsString() throws DesignRuleEvalException {

    if (this.getReferentDesignAnalysis() == null)
        throw new DesignRuleEvalException("Unable to dereference the "
                + "identifier " + this.getIdentifier()
                + " or one of its arguments as a design analysis." );

    if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                           ConstantTypes.STRING_TYPE, false)))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate DesignAnalysis: "
                + this.getIdentifier() + "as a string expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.STRING_TYPE, this.getType(), opMsg);
    };

    // load the actualParameters vector with the dereferenced values
    // of the actual parameters stored in children nodes.
    loadActualParams();

    // evaluate the design analysis with the assigned actual parameters
    return this.getReferentDesignAnalysis().evaluateAsString(this.getActualParams());
  };




/**
 * EvaluateAsElement calculates the actual parameters for the call
 * and invokes the referred to design rule with the current set of
 * actual parameters.  The actual parameters are initially parsed
 * as children nodes of the design rule call node.
 * @return The evaluated AcmeElement
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public AcmeElement evaluateAsElement() throws DesignRuleEvalException {

    if (this.getReferentDesignAnalysis() == null)
        throw new DesignRuleEvalException("Unable to dereference the "
                + "identifier " + this.getIdentifier()
                + " or one of its arguments as a design analysis." );

    if (!(this.getType() instanceof AcmeElementType)
        && !(this.getType() instanceof AnyType))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate DesignAnalysis: "
                + this.getIdentifier() + "as an Element.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.ELEMENT_BASE_TYPE, this.getType(), opMsg);
    };

    // load the actualParameters vector with the dereferenced values
    // of the actual parameters stored in children nodes.
    loadActualParams();

    // evaluate the design analysis with the assigned actual parameters
    return this.getReferentDesignAnalysis().evaluateAsElement(
                                                    this.getActualParams());
  };





/**
 * EvaluateAsPropertyRef calculates the actual parameters for the call
 * and invokes the referred to design rule with the current set of
 * actual parameters.  The actual parameters are initially parsed
 * as children nodes of the design rule call node.
 * @return The evaluated AcmeElement
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public AcmeProperty evaluateAsPropertyRef() throws DesignRuleEvalException {

    if (this.getReferentDesignAnalysis() == null)
        throw new DesignRuleEvalException("Unable to dereference the "
                + "identifier " + this.getIdentifier()
                + " or one of its arguments as a design analysis." );

    if (!(this.getType() instanceof PropertyReferenceType))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate DesignAnalysis: "
                + this.getIdentifier() + "as a Property Reference.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.PROPERTY_REFERENCE_TYPE, this.getType(), opMsg);
    };

    // load the actualParameters vector with the dereferenced values
    // of the actual parameters stored in children nodes.
    loadActualParams();

    // evaluate the design analysis with the assigned actual parameters
    return this.getReferentDesignAnalysis().evaluateAsPropertyRef(
                                                    this.getActualParams());
  };


  /* (non-Javadoc)
   * @see acmeLib.designRules.ExpressionNode#evaluateAsTypeRef()
   */
  public AcmeType evaluateAsTypeRef() throws DesignRuleEvalException {
      if (this.getReferentDesignAnalysis () == null) {
          throw new DesignRuleEvalException ("Unable to dereference the identifier " +
                this.getIdentifier () + " or one of its arguments as a design analysis.");
      }
      
      if (this.getType () != ConstantTypes.TYPE_REFERENCE_TYPE) {
          throw new ExpressionCastException(null, this,
               ConstantTypes.TYPE_REFERENCE_TYPE, this.getType(),
              "Unable to evaluate generic design rule expression node as a type object");
      }
      
      loadActualParams ();
      
      return this.getReferentDesignAnalysis ().evaluateAsTypeRef (this.getActualParams ());
  
                
  }


/**
 * EvaluateAsSet calculates the actual parameters for the call
 * and invokes the referred to design rule with the current set of
 * actual parameters.  The actual parameters are initially parsed
 * as children nodes of the design rule call node.
 * @return The evaluated set expression node
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public SetExpressionNode evaluateAsSet() throws DesignRuleEvalException {

    DesignAnalysis refDA = this.getReferentDesignAnalysis();
    if (refDA == null)
        throw new DesignRuleEvalException("Unable to dereference the "
                + "identifier " + this.getIdentifier()
                + " or one of its arguments as a design analysis." );

    if (!(refDA.getReturnType() instanceof SetType))
    {
        // error, throw an exception.
        String opMsg = new String("Type Error! Unable to evaluate DesignAnalysis: "
                + this.getIdentifier() + "as a Set.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.ANY_SET_TYPE, this.getType(), opMsg);
    };

    // load the actualParameters vector with the dereferenced values
    // of the actual parameters stored in children nodes.
    loadActualParams();

    return this.getReferentDesignAnalysis().evaluateAsSet(this.getActualParams());
  };




/**
 * write this expression out as a string in Armani syntax
 * @return The output string
 */
  public String toString() {
    String ret = "";
    try {
        AcmeOutputStream s = new AcmeOutputStream ();
        exportArmani (s);
        s.flush ();
        ret = s.toString ();
        s.close ();
    }
    catch (IOException e) {}
    catch (AcmeException e) {}
    
    return ret;

  };
  
  /* (non-Javadoc)
   * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
   */
  public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
    s.beginRegion(this);
    if (this.inParentheses) s.write ("(");  
    s.write (this.getIdentifier () + "(");
    
    // append all of the actual parameters to the stream
    // first make sure they are loaded in
    this.loadActualParams ();
    Enumeration e = actualParams.elements();
    if (e.hasMoreElements()) {
        ExpressionNode eNode = (ExpressionNode )e.nextElement ();
        eNode.exportArmani(s);
        if (ArmaniParser.isArgPrintingEnabled ()) {
            s.write (toStringRefObj (eNode));
        }
    }
    
    while (e.hasMoreElements ()) {
        ExpressionNode eNode = (ExpressionNode )e.nextElement();
        s.write (", ");
        eNode.exportArmani (s);
        if (ArmaniParser.isArgPrintingEnabled()) {
            s.write (toStringRefObj (eNode));
        }
    }
    s.write (")");
    
    if (this.getInParentheses()) s.write (")");
    s.endRegion (this);
  }
/**
 * zensoul added (01/06/2003)
 * @param the expression node whose reference object we'd like to get
 * @return the string rep of the refereenced obj, if available
 */
  private String toStringRefObj(ExpressionNode e) {
    StringBuffer buf = new StringBuffer();
    if (e instanceof ReferenceNode && e != null) {
        Object obj = ((ReferenceNode) e).getReferentObject();
        buf.append("[<");
        if (obj instanceof VolatileVariable) {
            buf.append(((VolatileVariable) obj).refObjectType.getTypeName());
            buf.append(":");
            obj = ((VolatileVariable) obj).referentObject;
        }
        if (obj instanceof IdentifierNode) {
            buf.append(((IdentifierNode) obj).getIdentifier());
        } else if (obj instanceof AcmeElement) {
            buf.append(((AcmeElement) obj).getName());
        } else if (obj instanceof AcmeType) {
            buf.append(((AcmeType) obj).getTypeName());
        } else {
            if (obj == null) {
                buf.append("null");
            } else {
                buf.append(obj.toString());
            }
        }
        buf.append(">]");
    }

    return buf.toString();
  }

/**
 * @param prefix The prefix
 * @return The output string
 */
  public String toString(String prefix) { return prefix + toString(); }




	/**
	 * @throws AcmeResolutionException
	 * @see acmeLib.designRules.ReferenceNode#refreshReferentObject()
	 **/
	public void refreshReferentObject() throws AcmeResolutionException {
		super.refreshReferentObject ();
		this.getReferentDesignAnalysis().clearAllUserData (getEvaluationKey());
		this.getReferentDesignAnalysis().setEvaluationKey (getEvaluationKey ());
	}
}

