package acmeLib.designRules;

import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.exceptions.AcmeDuplicateNameException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.exceptions.AcmeKeywordNameException;
import acmeLib.core.globals.ConversionConstants;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.ArmaniExportable;
import acmeLib.core.io.Exportable;
import acmeLib.elements.AcmeElementType;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.properties.*;
import acmeLib.properties.AcmePropertyBearer;
import acmeLib.properties.AcmeString;
import acmeLib.properties.BooleanType;
import acmeLib.typeManagement.AcmeTypeManager;
/**
 * Class: DesignRule<br>
 * Purpose: Represents the concrete declaration of a design rule.  A<br>
 *          DesignRule consists of a design rule expression<br>
 *          (ExpressionNode) and a reference to the scope (element) to<br>
 *          which the design rule is bound. A concrete binding also has a<br>
 *          set of properties (meta-properties?) associated with the design<br>
 *          rule that can override properties specified for abstract design<br>
 *          rules that this instance makes use of.<br>
 *<br>
 *          to do: describe how this class fits into the whole Armani system.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person     Change<br>
 *    1/5/98     RTM        Adapted old DesignRuleBinding class to work with<br>
 *                           updated DesignRule class structure.<br>
 *    2/3/98     RTM        Dramatically simplified class and changed typing<br>
 *                           to work with new AcmeType-based typesystem.<br>
 *<br>
 * @author Bob Monroe
 */
public class DesignRule extends AcmeObject implements Scope,
                                                      TypedEntity,
                                                      Exportable,
                                                      ArmaniExportable,
                                                      AcmePropertyBearer
{
    //
    // protected members
    //

  /**
   * stores this rule's expression tree
   */
    protected ExpressionNode designRuleExpr;

  /**
   * scope in which this design rule binding is declared scope can be
   * a component, connector, port, role, system, or a type of any of those.
   */
    protected acmeLib.designRules.Scope designRuleScope;

  /**
   * true if this rule is an invariant, else the rule is a heuristic.
   */
    protected boolean isInvariant;

  /**
   * true if this rule has been successfully typechecked, else false.
   */
    protected boolean isTypeCorrect;

  /**
   * true if this rule has been typechecked at all, else false.
   */
    protected boolean wasTypeChecked;

    /**
     * The map of properties pertaining to this design rule
     */
    protected AcmeMap propMap = new AcmeMap ();
    
    /** Whether to remember the objects visited in evaluation **/
    protected static boolean rememberVisited = false;
    //
    // public members
    //
  /**
   * This is the constructor for this class
   */
    public DesignRule() {
        // initialize the protected members of this rule to defaults
        this.isInvariant = false;
        this.wasTypeChecked = false;
        this.isTypeCorrect = false;
        this.designRuleExpr = null;
        this.designRuleScope = null;
    };

  /**
   * This is the constructor for this class
   * @param scope The scope
   * @param expr The expression node
   * @param invariant The boolean invariant
   */
    public DesignRule(Scope scope, ExpressionNode expr, boolean invariant)
    {
        this.isInvariant = invariant;
        this.wasTypeChecked = false;
        this.isTypeCorrect = false;
        this.designRuleExpr = expr;
        this.setDesignRuleScope(scope);
    };


  /**
   * copy constructor
   * @param copyOf The DesignRule
   */
    public DesignRule(DesignRule copyOf) {
        this.isInvariant = copyOf.getIsInvariant();
        this.setDesignRuleScope(copyOf.getDesignRuleScope());
        this.setIsTypeCorrect(copyOf.getIsTypeCorrect());

        // do a deep copy of the design rule expression
        this.designRuleExpr =
                copyOf.getDesignRuleExpression().cloneExpressionTree();

        // reset the scope of all nodes within the copied tree
        this.designRuleExpr.setExprTreeScope(this);
    };


    // accessor methods
  /**
   * Get the isInvariant boolean value
   * @return The isInvariant boolean value
   */
    public boolean getIsInvariant() { return this.isInvariant; };
  /**
   * Set the isInvariant boolean value
   * @param invariant The new isInvariant boolean value
   */
    public void setIsInvariant(boolean invariant) {
        this.isInvariant = invariant;
    };

  /**
   * Get the isTypeCorrect boolean value
   * @return The isTypeCorrect boolean value
   */
    public boolean getIsTypeCorrect() { return this.isTypeCorrect; };
  /**
   * Set the isTypeCorrect boolean value
   * @param typeCorrect The new isTypeCorrect boolean value
   */
    public void setIsTypeCorrect(boolean typeCorrect) {
        this.isTypeCorrect = typeCorrect;
    };


  /**
   * true if this design rule has been typechecked.  false indicates that
   * the rule needs to be typechecked.
   * @return The isTypeCorrect boolean value
   */
    public boolean getWasTypeChecked() { return this.wasTypeChecked; };

  /**
   * set to true if this design rule has been typechecked.  defaults to false.
   * @param typeCorrect The new isTypeCorrect boolean value
   */
    public void setWasTypeChecked(boolean typeChecked) {
        this.wasTypeChecked = typeChecked;
    };


  /**
   * Get the design rule expression node
   * @return The design rule expression node
   */
    public ExpressionNode getDesignRuleExpression() {
        return designRuleExpr;
    };
  /**
   * Set the design rule expression node
   * @param newExpr The new design rule expression node
   */
    public void setDesignRuleExpression(ExpressionNode newExpr) {
        this.designRuleExpr = newExpr;
    };

  /**
   * Get the design rule scope
   * @return The design rule scope
   */
    public Scope getDesignRuleScope() { return this.designRuleScope; };
  /**
   * Set the design rule scope
   * @param newScope The new design rule scope
   */
    public void setDesignRuleScope(Scope newScope) {
        this.designRuleScope = newScope;
    };


  /**
   * determine whether the expression stored by this design rule
   * typechecks. Return true if the expression node this design rule
   * stores typechecks and is a BooleanType, else return false.
   * @return true if type is correct
   */
    public boolean typecheck() {
        boolean exprTypechecks;

        // first have the expression typecheck itself.
        if (designRuleExpr == null) {
            this.setIsTypeCorrect(false);
            return false;
        } else {
            exprTypechecks = designRuleExpr.typecheck();
        };

        // then confirm that the expression is a boolean expression.
        if (exprTypechecks) {
            exprTypechecks = (designRuleExpr.getType() instanceof BooleanType);
            if (!exprTypechecks) {
                ExpressionTypecheckError err =
                    new ExpressionTypecheckError (this, "Expression is not a boolean", designRuleExpr);
                AcmeTypeManager.typecheckingErrorStack.pushCausalError(err); 
            }
        };

        if (!exprTypechecks) {
            ExpressionTypecheckError err = 
                new ExpressionTypecheckError (this, "Rule does not typecheck", designRuleExpr);
            AcmeTypeManager.typecheckingErrorStack.packageErrorCauses(err);
        }
        // cache a flag indicating whether this expression typechecks.
        this.setIsTypeCorrect(exprTypechecks);
        this.setWasTypeChecked(true);
        return exprTypechecks;
    };



  /**
   * evaluate this design rule.  True means that the design rule predicate
   * is satisfied, false indicates that design rule predicate is not
   * satisfied.
   * @return The evaluated boolean value
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public boolean evaluate() throws DesignRuleEvalException {
        if (this.getIsTypeCorrect() == true) {
        	if (isRememberVisited()) {
	        	this.designRuleExpr.clearAllUserData(designRuleScope);
	        	this.designRuleExpr.setEvaluationKey(designRuleScope);
        	}
            return this.designRuleExpr.evaluateAsBoolean();
        } else {
            if (this.typecheck() == true) {
            	if (isRememberVisited ()) {
	            	this.designRuleExpr.clearAllUserData (designRuleScope);
	            	this.designRuleExpr.setEvaluationKey (designRuleScope);
            	}
                return this.designRuleExpr.evaluateAsBoolean();
            }
            else
                return false;
                // RTM 2/11/98 -- throw or report an error/anomaly here?
        }
    };



  /**
   * since a design rule is a scope for a design rule expression, it has
   * to implement the Scope interface, which it does below.
   * @param name The name
   * @param bubbleUp Passes the request up
   * @return The object
   */
    public Object lookupName(String name, boolean bubbleUp) {
        Object namedObject = null;
        if (bubbleUp && (this.designRuleScope != null))
            namedObject = this.designRuleScope.lookupName(name, true);
        return namedObject;
    };


  /**
   * export the design rule to a string
   * @return The output string
   */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        if (this.isInvariant)
            buf.append("invariant ");
        else
            buf.append("heuristic ");

        if (this.designRuleExpr != null) {
            buf.append(designRuleExpr.toString());
        } else {
            buf.append("<no design rule expression specified>");
        };
        buf.append(";");

        return buf.toString();
    };


  /**
   * implement the Exportable interface
   * @param s AcmeOutputStream
   * @exception java.io.IOException For I/O exception
   * @exception acmeLib.core.exceptions.AcmeException For AcmeException
   */
    public void exportAcme(AcmeOutputStream s)
              throws java.io.IOException, acmeLib.core.exceptions.AcmeException
    {
        StringBuffer buf = new StringBuffer("Property ");

        // export it as a heuristic or invariant, as appropriate.
        if (this.isInvariant) {
            buf.append(ConversionConstants.ARMANI_INVARIANT_PROP_NAME_AS_ACME_PROP
                + "-" + String.valueOf(this.getCreationOrder()) + " : "
                + ConversionConstants.ARMANI_INVARIANT_TYPE_NAME_AS_ACME_PROP
                + " = " + AcmeString.massageForExport(this.toString()));
        } else {
            buf.append(ConversionConstants.ARMANI_HEURISTIC_PROP_NAME_AS_ACME_PROP
                + "-" + String.valueOf(this.getCreationOrder()) + " : "
                + ConversionConstants.ARMANI_HEURISTIC_TYPE_NAME_AS_ACME_PROP
                + " = " + AcmeString.massageForExport(this.toString()));
        };

        // if this is defined in a type declaration, then add a meta-property
        // to the exported method annotating as much.
        if (this.getDesignRuleScope() instanceof AcmeElementType) {
            String typename = ((AcmeElementType) this.getDesignRuleScope()).getTypeName();
            buf.append(" << " + ConversionConstants.DECLARED_IN_TYPENAME_METAPROP
                       + " = " + AcmeString.massageForExport (typename) + "; >>");
        };

        buf.append(";");
        s.writeln(buf.toString());
    };


  /**
   * implement the ArmaniExportable interface
   * @param s AcmeOutputStream
   * @exception java.io.IOException For I/O exception
   * @exception acmeLib.core.exceptions.AcmeException For AcmeException
   */
    public void exportArmani(AcmeOutputStream s, boolean exportProperties)
              throws java.io.IOException, acmeLib.core.exceptions.AcmeException
    {
        exportComment (s);
        s.beginRegion (this);
        if (this.isInvariant) {
            s.write ("invariant ");
        }
        else {
            s.write ("heuristic ");
        }
        if (this.designRuleExpr != null) {
            designRuleExpr.exportArmani (s);
        } else {
            s.write("<no design rule expression specified>");
        };
        if (exportProperties && propMap.getSize () > 0) {
            s.write ("<<");
            AcmeEnumeration properties = getProperties ();
            while (properties.moreItems ()) {
                AcmeProperty prop = (AcmeProperty )properties.getNextItem();
                prop.export (s, false);
                   
            }
            s.write (">>");
        }    
        s.write(";");
        
        s.endRegion (this);
        s.writeln();
//        String rule = this.toString ();
//        
//        // Process for indents
//        StringTokenizer tokenizer = new StringTokenizer (rule, "|");
//        int numIndents = 0;
//        while (tokenizer.hasMoreTokens()) {
//            String token = tokenizer.nextToken();
//            s.writeln (token + (tokenizer.hasMoreTokens ()?"|":""));
//            s.indent ();
//            numIndents++;
//        }
//        
//        for (int i = numIndents; i > 0; i--) {
//            s.unIndent();   
//        }
        
    };


    public void exportArmani (AcmeOutputStream s) throws java.io.IOException, acmeLib.core.exceptions.AcmeException
    {
        exportArmani (s, true);
    }

  /**
   * resolve all types referenced by this design rule's expression.
   */
    public void resolveTypeReferences() throws AcmeResolutionException {
      if (this.designRuleExpr != null)
          this.designRuleExpr.resolveTypeReferences();
    };


  /**
   *  Resolve all references to instances within this design rule's expression.
   */
    public void resolveInstanceReferences() throws AcmeResolutionException {
      if (this.designRuleExpr != null)
          this.designRuleExpr.resolveInstanceReferences();
    };

    /* (non-Javadoc)
     * @see acmeLib.properties.AcmePropertyBearer#addProperty(acmeLib.properties.AcmeProperty)
     */
    public void addProperty(AcmeProperty p) throws AcmeIllegalNameException {
        if (LanguageConfig.CHECK_FOR_KEYWORDS) {
            if (LanguageConfig.isAcmeKeyword(p.getName ()))
                throw new AcmeKeywordNameException ("Illegal keyword \"" + p.getName () + "\" used for property name.");
                    
        }
        
        if (p.getName () == null) {
            throw new AcmeIllegalNameException ("Error. Attempt to add a property with null name.");
            
        }
        else if (propMap.lookupItem(p.getName ()) != null) {
            throw new AcmeDuplicateNameException ("Attempted to redefine identifier " + p.getName () + " as an added property");
        }
        
        propMap.addItem (p.getName (), p);
        p.setScope (this);
    }

    /* (non-Javadoc)
     * @see acmeLib.properties.AcmePropertyBearer#addProperty(java.lang.String, acmeLib.properties.AcmePropertyType, acmeLib.properties.AcmePropertyValue)
     */
    public void addProperty(String name, AcmePropertyType t, AcmePropertyValue val) throws AcmeIllegalNameException {
        AcmeProperty prop = new AcmeProperty (name, t, val);
        addProperty (prop);
    }

    /* (non-Javadoc)
     * @see acmeLib.properties.AcmePropertyBearer#getNumberOfProperties()
     */
    public int getNumberOfProperties() {
        return propMap.getSize();
    }

    /* (non-Javadoc)
     * @see acmeLib.properties.AcmePropertyBearer#getProperties()
     */
    public AcmeEnumeration getProperties() {
        return propMap.enumerate ();
    }

    /* (non-Javadoc)
     * @see acmeLib.properties.AcmePropertyBearer#getProperty(java.lang.String)
     */
    public AcmeProperty getProperty(String name) {
        return (AcmeProperty )propMap.lookupItem(name);
    }

    /* (non-Javadoc)
     * @see acmeLib.properties.AcmePropertyBearer#removeProperty(acmeLib.properties.AcmeProperty)
     */
    public void removeProperty(AcmeProperty prop) {
        if (prop != null)
            propMap.removeItem (prop.getName ());
    }

    /* (non-Javadoc)
     * @see acmeLib.properties.AcmePropertyBearer#removeProperty(java.lang.String)
     */
    public void removeProperty(String name) {
        AcmeProperty prop = (AcmeProperty )propMap.lookupItem (name);
        if (prop != null) {
            prop.setScope (null);
        }
        propMap.removeItem (name);

    }

	/**
	 * @return Returns the rememberVisited.
	 */
	public static boolean isRememberVisited() {
		return rememberVisited;
	}
	/**
	 * @param rememberVisited The rememberVisited to set.
	 */
	public static void setRememberVisited(boolean rememberVisited) {
		DesignRule.rememberVisited = rememberVisited;
	}
}