package acmeLib.designRules;

import java.io.IOException;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.ArmaniExportable;

/**
 * Class: DesignRuleParameter<br>
 * Purpose: Base class that represents a parameter for use in a design rule.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    10/20/97    RTM      Created prototype for this class<br>
 *    2/3/98      RTM      updated to work with AcmeType-based type system.<br>
 *                          removed all ReferentObjectTypeT and ExpressionTypeT's<br>
 *    3/10/98     RTM      made class for use solely with Formal Parameters.<br>
 *                          actual params are now just trees of ExpressionNodes<br>
 *<br>
 * @author Bob Monroe
 */
public class DesignRuleParameter extends AcmeObject implements ArmaniExportable{

    //
    // protected members
    //

  /**
   * parameter name
   */
    protected String name;

  /**
   * type object of a non-primitive, named type.  null if primitive type.
   */
    protected AcmeType paramType;


    //
    // public members
    //

  /** 
   * This is the constructor for this class
   */
    public DesignRuleParameter() {
        this.name = null;
        this.paramType = null;
    };

  /**
   * This is the constructor for this class
   * @param name The name
   * @param paramType The AcmeType
   */
    public DesignRuleParameter(String name, AcmeType paramType) {
        this.name = name;
        this.paramType = paramType;
    };


  /**
   * copy constructor
   * @param copyFrom The DesignRuleParameter
   */
    public DesignRuleParameter(DesignRuleParameter copyFrom) {
        this.name = copyFrom.name;
        this.paramType = copyFrom.paramType;
    };


  /**
   * create a new copy of this DesignRuleParameter
   * @return The copy of the design rule parameter
   */
    public DesignRuleParameter copy() {
        return new DesignRuleParameter(this);
    };


    // accessor methods
  /**
   * Get the name
   * @return The name
   */
    public String getName() { return this.name; };
  /** 
   * Set the name
   * @param newName The new name
   */
    public void setName(String newName) { this.name = newName; };

  /**
   * Get the AcmeType
   * @return The AcmeType
   */
    public AcmeType getType() { return this.paramType; }
  /**
   * Set the AcmeType
   * @param newType The new AcmeType
   */
    public void setType(AcmeType newType) {
        this.paramType = newType;
    };


  /** 
   * write the parameter out to a string
   * @return The output string
   */
    public String toString() {
        String ret = "";
        try {
            AcmeOutputStream s = new AcmeOutputStream ();
            exportArmani (s);
            s.flush ();
            ret = s.toString ();
            s.close ();
        }
        catch (IOException e) {}
        catch (AcmeException e) {}
    
        return ret;    
    };

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion (this);
        s.write (this.getName ());
        if (this.getType () != null) {
            s.write (" : ");
            s.write (this.getType ().getTypeName ());
        }
        s.endRegion (this);
    }

    /* (non-Javadoc)
     * @see acmeLib.core.io.Exportable#exportAcme(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportAcme(AcmeOutputStream s) throws IOException, AcmeException {
    }

}