package acmeLib.designRules;

import acmeLib.*;
import acmeLib.core.container.*;
import acmeLib.typeManagement.*;

/**
 * Class: DrUtils<br>
 * Purpose: Stores a set of static utility methods (and values) available<br>
 *          for use by all design rule package classes.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    1/9/98     RTM       Created<br>
 *<br>
 * @author Bob Monroe
 */
public class DrUtils {

  /**
   *
   * compares two SetExpressionNodes for equality.
   * The two sets are equal iff the following conditions are met:
   * 1) The size of the sets are the same
   * 2) there is some subtype/supertype relationship between the type
   *     of set1 and the type of set2
   * 3) every item i1 in set 1 has a corresponding item i2 in set 2
   *     such that i1.equals(i2).
   * @param set1 The set expression node 1
   * @param set2 The set expression node 2
   * @return true if the two sets are equal
   */
    public static boolean equalSets(SetExpressionNode set1,
                                    SetExpressionNode set2)
    {
        if (set1 == set2) {
            return true;
        };

        if (set1.size() != set2.size()) {
            return false;
        };

        if (!AcmeTypeManager.isLogicalSubtype(set1.getSetType(),
                                              set2.getSetType(), false))
        {
            if (!AcmeTypeManager.isLogicalSubtype(set2.getSetType(),
                                                  set1.getSetType(), false))
            {
                // only the empty set will be equivalent if the types
                // have no sub-supertype relationship
                return ((set1.size() == 0) && (set2.size() == 0));
            };
        };

        boolean setsAreEqual = true;
        AcmeEnumeration set1Enum = set1.enumerate();
        AcmeEnumeration set2Enum = set2.enumerate();
        AcmeObject curObj1, curObj2;
        while (set1Enum.moreItems() && setsAreEqual) {
            curObj1 = set1Enum.getNextItem();
            set2Enum.reset();
            boolean matchingEltFound = false;
            while(set2Enum.moreItems() && (matchingEltFound == false)) {
                curObj2 = set2Enum.getNextItem();
                if(curObj1.equals(curObj2)) {
                    matchingEltFound = true;
                };
            };

            // we didn't find a match for curObj1, the sets are unequal
            if (matchingEltFound == false) {
                setsAreEqual = false;
            };
        };

        return setsAreEqual;
    };
}
