package acmeLib.designRules;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import acmeLib.AcmeType;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExpressionOperatorT;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.properties.*;
import acmeLib.typeManagement.AcmeTypeManager;

/** Expression node is the base class from which all type-specific
    expression nodes inherit. */
/**
 * Class: ExpressionNode<br>
 * Purpose: Base class from which all type-specific design rule expression<br>
 *          nodes inherit.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person   Change<br>
 *    10/7/97    RTM     Created<br>
 *    10/27/97   RTM     Added support for exception handling<br>
 *    2/2/98     RTM     Modified type storage to store AcmeTypes<br>
 *                        instead of ExpressionTypeT objects.<br>
 *<br>
 * @author Bob Monroe
 */
public abstract class ExpressionNode extends SimpleNode implements Node, TypedEntity
{

  //
  // Protected members
  //

/**
 * stores the expression type
 */
  protected acmeLib.AcmeType type;

/**
 * stores the expression operator (assumes unary or binary expression).
 */
  protected acmeLib.core.globals.ExpressionOperatorT op;

/**
 * indicates whether this should be parenthesized on export
 */
  protected boolean inParentheses;

/**
 * static reference to the global type manager
 */
  protected static AcmeTypeManager globalTypeMgr;
  
  protected Map userData;

private Object	_evalKey;

/**
 *
 * helper methods to determine whether this expression's operator
 * is a boolean, numeric, or set operator.  String ops to be added later.
 *
 * @return true if boolean operator exists
 */
  protected boolean hasBooleanOperator() {
    return ( (this.getOperator() == ExpressionOperatorT.NOT_OP)
             || (this.getOperator() == ExpressionOperatorT.EQUALS_OP)
             || (this.getOperator() == ExpressionOperatorT.NOT_EQUALS_OP)
             || (this.getOperator() == ExpressionOperatorT.LESS_OR_EQUAL_OP)
             || (this.getOperator() == ExpressionOperatorT.GREATER_OR_EQUAL_OP)
             || (this.getOperator() == ExpressionOperatorT.LESS_THAN_OP)
             || (this.getOperator() == ExpressionOperatorT.GREATER_THAN_OP)
             || (this.getOperator() == ExpressionOperatorT.OR_OP)
             || (this.getOperator() == ExpressionOperatorT.AND_OP)
             || (this.getOperator() == ExpressionOperatorT.IMPLIES_OP)
             || (this.getOperator() == ExpressionOperatorT.IFF_OP)
           );
  };

/**
 * @return true if boolean logic operator exists
 */
  protected boolean hasBooleanLogicOperator() {
    return ( (this.getOperator() == ExpressionOperatorT.NOT_OP)
             || (this.getOperator() == ExpressionOperatorT.OR_OP)
             || (this.getOperator() == ExpressionOperatorT.AND_OP)
             || (this.getOperator() == ExpressionOperatorT.IMPLIES_OP)
             || (this.getOperator() == ExpressionOperatorT.IFF_OP)
             || (this.getOperator() == ExpressionOperatorT.EQUALS_OP)
             || (this.getOperator() == ExpressionOperatorT.NOT_EQUALS_OP)
           );
  };

/**
 * @return true if string logic operator exists
 */
  protected boolean hasStringLogicOperator() {
    return ( (this.getOperator() == ExpressionOperatorT.EQUALS_OP)
          || (this.getOperator() == ExpressionOperatorT.NOT_EQUALS_OP)) ;
  };

/**
 * @return true if numeric comparison operator exists
 */
  protected boolean hasNumericComparisonOperator() {
    return ( (this.getOperator() == ExpressionOperatorT.EQUALS_OP)
             || (this.getOperator() == ExpressionOperatorT.NOT_EQUALS_OP)
             || (this.getOperator() == ExpressionOperatorT.LESS_OR_EQUAL_OP)
             || (this.getOperator() == ExpressionOperatorT.GREATER_OR_EQUAL_OP)
             || (this.getOperator() == ExpressionOperatorT.LESS_THAN_OP)
             || (this.getOperator() == ExpressionOperatorT.GREATER_THAN_OP)
           );
  };

/**
 * @return true if numeric operator exists
 */
  protected boolean hasNumericOperator() {
    return ( (this.getOperator() == ExpressionOperatorT.PLUS_OP)
             || (this.getOperator() == ExpressionOperatorT.MINUS_OP)
             || (this.getOperator() == ExpressionOperatorT.MULT_OP)
             || (this.getOperator() == ExpressionOperatorT.DIVIDE_OP)
             || (this.getOperator() == ExpressionOperatorT.MODULO_OP)
             || (this.getOperator() == ExpressionOperatorT.UNARY_MINUS_OP)
           );
  };


/**
 * @return true if set operator exists
 */
  protected boolean hasSetOperator() {
    return ( (this.getOperator() == ExpressionOperatorT.SET_UNION_OP)
             || (this.getOperator() == ExpressionOperatorT.SET_INTERSECTION_OP)
             || (this.getOperator() == ExpressionOperatorT.SET_MEMBER_OP)
             || (this.getOperator() == ExpressionOperatorT.SET_DIFFERENCE_OP)
           );
  };


/**
 * @return true if equality comparison operator exists
 */
  protected boolean hasEqualityComparisonOperator() {
    return ( (this.getOperator() == ExpressionOperatorT.EQUALS_OP)
             || (this.getOperator() == ExpressionOperatorT.NOT_EQUALS_OP) );
  };


/**
 * doCloneExpressionTree is a protected utility method that recurses through
 * "this" expression tree, copying all of the children from "this" to
 * "clone".
 * @param clone The expression tree to be copied
 */
  protected void doCloneExpressionTree(ExpressionNode clone) {
    // if there are no children then there's no need to clone them...
    if (this.children == null)
        return;

    int numChildren = this.children.length;
    Node newChildren[] = new Node[numChildren];;

    // make a copy of each of the children of this node, and add it to the
    // array of new children
    for (int i = 0; i < numChildren; i++) {
        ExpressionNode newChild =
            ((ExpressionNode) this.children[i]).cloneExpressionTree();
        newChild.jjtSetParent(clone);
        newChildren[i] = newChild;
    }
    clone.children = newChildren;
  }

  //
  // public members
  //

/**
 * Generic constructor with node id
 * @param id The ID
 */
  public ExpressionNode(int id) {
    super(id);
    this.setInParentheses(false);
    this.setType(ConstantTypes.UNRESOLVED_TYPE);
    // nodes have a no-op operator by default.
    this.setOperator(ExpressionOperatorT.NO_OP);
  }


/**
 * Copy constructor returns a new object that is a copy of an
 * existing ExpressionNode.
 * @param copyOf The expression node to be copied
 */
  public ExpressionNode(ExpressionNode copyOf) {
    super(copyOf.id);
    this.setInParentheses(copyOf.getInParentheses());
//    this.info = copyOf.getInfo();
    this.parent = copyOf.parent;
    this.setType(copyOf.getType());
    this.setOperator(copyOf.getOperator());

    // if the object we're cloning has children, shallow-copy the vector
    // of children nodes.
    if (copyOf.children != null)
        this.children = (Node[])copyOf.children.clone();
  }


/**
 * cloneExpressionTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.
 * @return The copied expression tree
 */
  public abstract ExpressionNode cloneExpressionTree();


/**
 * get or set the type of the expression
 * @return The AcmeType
 */
  public AcmeType getType() { return type; }
/**
 * @param newType The new type
 */
  public void setType(AcmeType newType) { this.type = newType; }

/**
 * get or set the operator of the expression
 * @return The expression operator
 */
  public ExpressionOperatorT getOperator() { return op; }
/**
 * @param newOp The new operator
 */
  public void setOperator(ExpressionOperatorT newOp) { this.op = newOp; }

/**
 * get or set the flag indicating whether this expression is parenthesized
 * @return The inParentheses boolean value
 */
  public boolean getInParentheses() {
    return this.inParentheses;
  };
/**
 * @param doParentheses The new inParentheses boolean value
 */
  public void setInParentheses(boolean doParentheses) {
    this.inParentheses = doParentheses;
  };

/**
 * get or set the static global type manager reference
 * @return The AcmeTypeMAnager
 */
  public AcmeTypeManager getGlobalTypeMgr() { return globalTypeMgr; }
/**
 * @param newTypeMgr The new AcmeTypeManager
 */
  public void setGlobalTypeMgr(AcmeTypeManager newTypeMgr) {
    globalTypeMgr = newTypeMgr;
  };


/**
 * ExpressionNode just recursively calls this method on all of its
 * children nodes.  Because the ExpressionNode class does not store
 * a scope reference, the only way the scope is changed is if the
 * child node is a subclass of ExpressionNode that overrides
 * the method to perform an actual modification of the scope.
 * @param newScope The new scope
 */
  public void setExprTreeScope(Scope newScope) {
    int numChildren;
    if (this.children == null)
        numChildren = 0;
    else
        numChildren = this.children.length;
    for (int i = 0; i < numChildren; i++) {
        ((ExpressionNode) this.children[i]).setExprTreeScope(newScope);
    };
  };


/**
 * Traverses the expression tree below to typecheck the expression.
 * ExpressionNode.typecheck() does no actual typechecking, it only traverses
 * the tree and requests all of its children nodes typecheck themselves.
 * Subclasses that want to do any actual typechecking should override this
 * method. It returns the result of its children's typecheck, unless it has
 * no children in which case it just returns true.
 * @return true if the type is correct
 */
  public boolean typecheck() {
    // assume that this expression typechecks, unless we find a node that
    // doesn't typecheck.
    boolean typechecks = true;
    boolean tmpBool;
    int numChildren;

    if (this.children == null)
        numChildren = 0;
    else
        numChildren = this.children.length;
    for (int i = 0; (i < numChildren) && typechecks; i++) {
        tmpBool = ((ExpressionNode) this.children[i]).typecheck();
        if (tmpBool == false)
            typechecks = false;
    };

    return typechecks;
  }


/**
 * All of the evaluateAs* methods should be overridden by subclasses
 * that return sensible values.  If a subclass can not be evaluated
 * as a specified type (e.g. evaluate a string node as boolean) then
 * the subtype should not override these methods, and an appropriate
 * exception will be thrown by this class.
 *
 * Note that none of the ExpressionNode.evaluateAs* methods actually
 * return a valid value in this base class.  They all simply throw
 * various types of exceptions based on the type and casting combination.
 *
 * Evaluate this expression tree as a boolean expression.  If it is not
 * a boolean typed-node then throw a type exception.
 * @return The evaluated boolean value
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public boolean evaluateAsBoolean() throws DesignRuleEvalException {
    // needs to be overridden by subclass that can handle boolean evaluation.
    if (!(this.getType() instanceof BooleanType)) {
        throw new ExpressionCastException(null, this,
             ConstantTypes.BOOLEAN_TYPE, this.getType(),
            "Unable to evaluate generic design rule expression node as a boolean");
    } else {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate "
            + "generic ExpressionNode's evaluateAsBoolean method,"
            + "which should be an abstract method.");
    }
  }


/**
 * Evaluate this expression tree as a float expression.  If it is not
 * a float typed-node then throw a type exception.
 * @return The evaluated float
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public float evaluateAsFloat() throws DesignRuleEvalException {
    // needs to be overridden by subclass that can handle float evaluation.
    if (!(this.getType() instanceof FloatType))
    {
        throw new ExpressionCastException(null, this,
             ConstantTypes.FLOAT_TYPE, this.getType(),
            "Unable to evaluate generic design rule expression node as a float");
    } else {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate "
            + "generic ExpressionNode's evaluateAsFloat method,"
            + "which should be an abstract method.");
    }
  }


/**
 * Evaluate this expression tree as an integer expression. If it is not
 * an int typed-node then throw a type exception.
 * @return The evaluated integer
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public int evaluateAsInt() throws DesignRuleEvalException {
    // needs to be overridden by subclass that can handle integer evaluation.
    if (!(this.getType() instanceof FloatType)
        && !(this.getType() instanceof IntType))
    {
        throw new ExpressionCastException(null, this,
             ConstantTypes.INT_TYPE, this.getType(),
            "Unable to evaluate generic design rule expression node as an int");
    } else {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate "
            + "generic ExpressionNode's evaluateAsInt method,"
            + "which should be an abstract method.");
    }
  }


/**
 * Evaluate this expression tree as a string expression.  If it is not
 * a string typed-node then throw a type exception.
 * @return The evaluated string
 * @exception DesignRuleEvalException If there is evaluation error
 */
  public String evaluateAsString() throws DesignRuleEvalException {
    // needs to be overridden by subclass that can handle integer evaluation.
    if (!(this.getType() instanceof StringType))
    {
        throw new ExpressionCastException(null, this,
             ConstantTypes.STRING_TYPE, this.getType(),
            "Unable to evaluate generic design rule expression node as a string");
    } else {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate "
            + "generic ExpressionNode's evaluateAsString method,"
            + "which should be an abstract method.");
    }
  }




/**
 * Evaluate this expression tree as an AcmeElement expression.  If it can not
 * be evaluated as an AcmeElement then throw a type exception.
 * @return The evaluated AcmeElement
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public AcmeElement evaluateAsElement() throws DesignRuleEvalException {
    // needs to be overridden by subclass that can handle element evaluation.
    if (!(this.getType() instanceof AcmeElementType))
    {
        throw new ExpressionCastException(null, this,
             ConstantTypes.ELEMENT_BASE_TYPE, this.getType(),
            "Unable to evaluate generic design rule expression node as an element");
    } else {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate "
            + "generic ExpressionNode's evaluateAsElement method,"
            + "which should be an abstract method.");
    }
  }




/**
 * Evaluate this expression tree as an AcmeProperty reference expression.
 * If it can not be evaluated as an AcmeProperty then throw a type exception.
 * @return The evaluated AcmeProperty
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public AcmeProperty evaluateAsPropertyRef() throws DesignRuleEvalException {
    // needs to be overridden by subclass that can handle element evaluation.
    if (!(this.getType() instanceof AcmePropertyType))
    {
        throw new ExpressionCastException(null, this,
             ConstantTypes.PROPERTY_REFERENCE_TYPE, this.getType(),
            "Unable to evaluate generic design rule expression node as a property reference");
    } else {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate "
            + "generic ExpressionNode's evaluateAsPropertyRef method,"
            + "which should be an abstract method.");
    }
  }




/**
 * Evaluate this expression tree as a reference to a Type object.
 * If it can not be evaluated as a Type then throw a type exception.
 * @return The evaluated AcmePropertyType
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public AcmeType evaluateAsTypeRef() throws DesignRuleEvalException {
    // needs to be overridden by subclass that can handle element evaluation.
    if (this.getType() != ConstantTypes.TYPE_REFERENCE_TYPE)
    {
        throw new ExpressionCastException(null, this,
             ConstantTypes.TYPE_REFERENCE_TYPE, this.getType(),
            "Unable to evaluate generic design rule expression node as a type object");
    } else {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate "
            + "generic ExpressionNode's evaluateAsTypeRef method,"
            + "which should be an abstract method.");
    }
  }





/**
 * Evaluate this expression tree as a SetExpressionNode.  If it can not
 * be evaluated as a SetExpressionNode then throw a type exception.
 * @return The evaluated set expression node
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public SetExpressionNode evaluateAsSet() throws DesignRuleEvalException {
    // needs to be overridden by subclass that can handle element evaluation.
    if (!(this.getType() instanceof SetType))
    {
        throw new ExpressionCastException(null, this,
             ConstantTypes.ANY_SET_TYPE, this.getType(),
            "Unable to evaluate generic design rule expression node as a set");
    } else {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate "
            + "generic ExpressionNode's evaluateAsSet method,"
            + "which should be an abstract method.");
    }
  }



/**
 * Implement a generic exporting method in toString()
 * @return The output string
 */
  public String toString() {
    StringBuffer buf = new StringBuffer();
    // put the initial parenthese in, if needed
    if (this.inParentheses) {
        buf.append("(");
    };

    if (jjtGetNumChildren() == 0) {
        buf.append("<invalid expression>");
    }

    // check if it's a unary operator.
    else if (jjtGetNumChildren() == 1) {
        buf.append(op.toString());
        buf.append(jjtGetChild(0).toString());
    }

    // else check for a binary operator
    else if (jjtGetNumChildren() == 2) {
        buf.append(jjtGetChild(0).toString());
        buf.append(" " + this.getOperator().toString() + " ");
        buf.append(jjtGetChild(1).toString());
    }

    else if (jjtGetNumChildren() > 2) {
        // error, this is a binary expression
        buf.append(" <error, too many children of ExpressionNode> ");
    };

    // put the trailing parenthese in, if needed
    if (this.inParentheses) {
        buf.append(")");
    };

    return buf.toString();
  };

  public void exportArmani (AcmeOutputStream s) throws IOException, AcmeException {
      s.beginRegion (this);
      if (this.inParentheses) s.write ("(");
      
      if (jjtGetNumChildren () == 0) {
          s.write ("<invalid expression>");
      } 
      else if (jjtGetNumChildren () == 1) {
          s.write (op.toString());
          jjtGetChild (0).exportArmani (s);
      }
      else if (jjtGetNumChildren () == 2) {
          jjtGetChild (0).exportArmani (s);
          s.write (" " + this.getOperator ().toString () + " ");
          jjtGetChild (1).exportArmani (s);
      }
      else if (jjtGetNumChildren () > 2) {
          s.write (" <error, too many children of ExpressionNode> ");
      }
      
      if (this.inParentheses) s.write (")");
      s.endRegion (this);
      
  }

/**
 * do all of the post-parsing processing of the design that is supposed
 * to happen after parsing, in preparation for, but not including,
 * typechecking.
 */
  public void doPostParsingProcessing()
  {
    try {  
        this.resolveTypeReferences();
        this.resolveInstanceReferences();
    }
    catch (AcmeResolutionException e) {
        e.printStackTrace();
    }
  };



/**
 * resolve all types referenced by this design rule's expression.
 * ExpressionNode.resolveTypeReferences() does no typename resolution
 * because it is a quasi-abstract superclass.  It simply implements
 * the recursive traversal of the tree that requests interested
 * nodes to resolve their types.
 */
  public void resolveTypeReferences() throws AcmeResolutionException{
    int numChildren;
    if (this.children == null)
        numChildren = 0;
    else
        numChildren = this.children.length;
    for (int i = 0; i < numChildren; i++) {
        ((ExpressionNode) this.children[i]).resolveTypeReferences();
    };
  };


/**
 * recursively resolve all references to entity instances in this expression
 * tree.  resolveInstanceReferences does nothing in the generic ExpressionNode
 * class other than recursively calling itself on all of its children.
 * Subclasses that can have instance references that need to be resolved
 * should override this method.
 */
  public void resolveInstanceReferences() throws AcmeResolutionException{
    int numChildren;
    if (this.children == null)
        numChildren = 0;
    else
        numChildren = this.children.length;
    for (int i = 0; i < numChildren; i++) {
        ((ExpressionNode) this.children[i]).resolveInstanceReferences();
    };
  }



/**
 * resolveVolatileFlag() recursively traverses the expression, setting
 * the IsVolatile flag to true for all reference nodes that refer to a
 * volatile variable (e.g. a quantified variable or a parameter).  The
 * implementation in ExpressionNode only traverses the tree.  IdentifierNodes
 * override this method to actually check whether they should be volatile.
 */
  public void resolveVolatileFlag() {
    int numChildren;
    if (this.children == null)
        numChildren = 0;
    else
        numChildren = this.children.length;
    for (int i = 0; i < numChildren; i++) {
        ((ExpressionNode) this.children[i]).resolveVolatileFlag();
    };
  }
  
  public void setUserData (Object key, Object data) {
  	if (userData == null) userData = new HashMap ();
  	userData.put(key, data);
  }
  
  public Object getUserData (Object key) {
  	if (userData == null) return null;
  	return userData.get (key);
  }
  
  /**
   * Recursively removes user data from this and all children, for a particular key
   * @param key The key to clear
   */
  public void clearAllUserData (Object key) {
  	if (userData != null) userData.remove(key);
  	int numChildren;
  	if (this.children == null) numChildren = 0;
  	else numChildren = this.children.length;
  	for (int i = 0; i < numChildren; i++) {
  		((ExpressionNode )this.children [i]).clearAllUserData(key);
  	}
  }
  
  /*package*/ void setEvaluationKey (Object key) {
  	_evalKey = key;
 	int numChildren;
  	if (this.children == null) numChildren = 0;
  	else numChildren = this.children.length;
  	for (int i = 0; i < numChildren; i++) {
  		((ExpressionNode )this.children [i]).setEvaluationKey(key);;
  	}
  }

    /* (non-Javadoc)
     * @see acmeLib.core.io.Exportable#exportAcme(acmeLib.core.io.AcmeOutputStream)
     */
    public  void exportAcme(AcmeOutputStream s) throws IOException, AcmeException {
    	
    }

/**
 * @return Returns the evalKey.
 */
public Object getEvaluationKey() {
	return _evalKey;
}
}
