package acmeLib.designRules;

import java.io.IOException;

import acmeLib.AcmeType;
import acmeLib.armaniParser.ArmaniParser;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: IdentifierNode<br>
 * Purpose: expression node that represents an expression tree leaf<br>
 *          that stores an identifier reference.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person  Change<br>
 *    10/7/97    RTM     * Created original<br>
 *    10/20/97   RTM     * moved name lookup functionality to ReferenceNode<br>
 *                          superclass<br>
 *    1/27/98    RTM     * replaced unify() and inferType() with<br>
 *                          resolveInstanceNames() and typecheck().<br>
 *<br>
 * @author Bob Monroe
 */
public class IdentifierNode extends ReferenceNode
{
  //
  // public members
  //
/**
 * This is the constructor for this class
 * @param id The ID
 */
 public IdentifierNode(int id) {
    super(id);
    this.referentObject = null;
  }

/**
 * Copy constructor
 * @param copyOf The IdentifierNode to be copied
 */
  public IdentifierNode(IdentifierNode copyOf) {
    super(copyOf);
    this.referentObject = copyOf.getReferentObject();
  }

/**
 * cloneExpressionTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.
 * @return The copy of the expression node
 */
  public ExpressionNode cloneExpressionTree() {
    IdentifierNode clone = new IdentifierNode(this);
    this.doCloneExpressionTree(clone);
    return clone;
  };


/**
 * JJTree-specific methods
 * @param id The ID
 * @return The node
 */
  public static Node jjtCreate(int id) {
    return new IdentifierNode(id);
  }



/**
 * Dereference this identifier and attempt to evaluate it as a boolean
 * value.
 * @return The evaluated boolean value
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public boolean evaluateAsBoolean() throws DesignRuleEvalException {
    // recalculate the referentObject but leave the type signature of
    // the reference untouched.
    this.refreshReferentObject();

    if (this.getReferentObject() == null) {
        // error, throw an exception.
        throw new ReferenceNotFoundException(this.getReferenceScope(), this,
            "Error! Unable to resolve identifier: " + this.getIdentifier()
            + " while evaluating expression as boolean.");
    };


    if (!(AcmeTypeManager.isLogicalSubtype(
            this.getType(), ConstantTypes.BOOLEAN_TYPE, false)))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate identifier: "
                + this.getIdentifier() + "as a boolean expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.BOOLEAN_TYPE, this.getType(), opMsg);
    }
    else if (referentObject instanceof AcmeProperty) {
        AcmeProperty prop = (AcmeProperty) referentObject;
        // RTM 2/10/98 -- known bug.  elsewhere, it's not guaranteed that this
        //                property will have a value, so this can cause a
        //                null pointer exception.  This is not, however,
        //                a valid Armani construct, it's a vestigial effect
        //                from Acme's loose-ness.
        // BRS 3/7/03 -- But Armani allows rules at the family level,
        //               which may refer to properties that don't have a
        //               value.
        if (prop.getValue () == null) {
            return true;
        }
        return ((AcmeBoolean) prop.getValue()).getValue();
    }
    else if (referentObject instanceof AcmePropertyValue) {
        return ((AcmeBoolean) referentObject).getValue();
    }
    else if (referentObject instanceof AcmeRecordField) {
        return ((AcmeBoolean) ((AcmeRecordField) referentObject).getValue()).getValue();
    }
    else if (referentObject instanceof VolatileVariable) {
        VolatileVariable tmpVar = (VolatileVariable) referentObject;
        if (tmpVar.referentObject == null) {
            String opMsg = new String("Error! Identifier: "
                    + this.getIdentifier()
                    + " does not have a valid referent object");
            throw new ReferenceNotFoundException(this.getReferenceScope(),
                    this, opMsg);
        } else {
            // figure out what kind of object the volatileVariable's
            // referent object is, and evaluate it as a boolean accordingly.
            if (tmpVar.referentObject instanceof ExpressionNode) {
                return ((ExpressionNode) tmpVar.referentObject).evaluateAsBoolean();
            }
            else if (tmpVar.referentObject instanceof AcmeProperty) {
                AcmeProperty prop = (AcmeProperty) tmpVar.referentObject;
                // RTM 2/10/98 -- known bug.  elsewhere, it's not guaranteed that this
                //                property will have a value, so this can cause a
                //                null pointer exception.  This is not, however,
                //                a valid Armani construct, it's a vestigial effect
                //                from Acme's loose-ness.
                return ((AcmeBoolean) prop.getValue()).getValue();
            }
            else if (tmpVar.referentObject instanceof AcmePropertyValue) {
                return ((AcmeBoolean) tmpVar.referentObject).getValue();
            }
                // throw an exception. referentObject is not a recognized
                // type for evaluation as a boolean.
                String opMsg = new String("Error! Identifier: "
                                + this.getIdentifier()
                                + " can not be evaluated as a boolean.");
                throw new ExpressionCastException(this.getReferenceScope(),
                            this, ConstantTypes.BOOLEAN_TYPE,
                            tmpVar.refObjectType, opMsg);
        }
    }
        // throw an exception. referentObject is not a recognized type for
        // evaluation as a boolean.
        String opMsg = new String("Error! Identifier: " + this.getIdentifier()
                                  + " can not be evaluated as a boolean.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.BOOLEAN_TYPE, this.getType(), opMsg);
  };



/**
 * Dereference this identifier and attempt to evaluate it as an integer
 * value
 * @return The evaluated integer
 * @exception DesignRuleEvalException If there is an evalution error
 */
  public int evaluateAsInt() throws DesignRuleEvalException {
    // recalculate the referentObject but leave the type signature of
    // the reference untouched.
    this.refreshReferentObject();

    if (this.getReferentObject() == null) {
        // error, throw an exception.
        throw new ReferenceNotFoundException(this.getReferenceScope(), this,
            "Error! Unable to resolve identifier: " + this.getIdentifier()
            + " while evaluating expression as an integer.");
    };


    if (!(AcmeTypeManager.isLogicalSubtype(
            this.getType(), ConstantTypes.INT_TYPE, false)))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate identifier: "
                + this.getIdentifier() + "as an integer expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.INT_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof AcmeProperty) {
        AcmeProperty prop = (AcmeProperty) referentObject;
        // RTM 2/10/98 -- known bug.  elsewhere, it's not guaranteed that this
        //                property will have a value, so this can cause a
        //                null pointer exception.  This is not, however,
        //                a valid Armani construct, it's a vestigial effect
        //                from Acme's loose-ness.
        if (prop.getValue() instanceof AcmeFloat)
            return (int) ((AcmeFloat) prop.getValue()).getValue();
        else if (prop.getValue() instanceof AcmeInt)
            return ((AcmeInt) prop.getValue()).getValue();
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate identifier: "
                    + this.getIdentifier() + "as an integer expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.INT_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof AcmeFloat) {
        return (int) ((AcmeFloat) referentObject).getValue();
    }

    else if (referentObject instanceof AcmeInt) {
        return ((AcmeInt) referentObject).getValue();
    }

    else if (referentObject instanceof AcmeRecordField) {
        AcmeRecordField recField = (AcmeRecordField) referentObject;
        if (recField.getValue() instanceof AcmeFloat)
            return (int) ((AcmeFloat) recField.getValue()).getValue();
        else if (recField.getValue() instanceof AcmeInt)
            return ((AcmeInt) recField.getValue()).getValue();
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate identifier: "
                    + this.getIdentifier() + "as an integer expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.INT_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof VolatileVariable) {
        VolatileVariable tmpVar = (VolatileVariable) referentObject;
        if (tmpVar.referentObject == null) {
            String opMsg = new String("Error! Identifier: "
                    + this.getIdentifier()
                    + " does not have a valid referent object");
            throw new ReferenceNotFoundException(this.getReferenceScope(),
                    this, opMsg);
        } else {
            // figure out what kind of object the volatileVariable's
            // referent object is, and evaluate it as an int accordingly.
            if (tmpVar.referentObject instanceof ExpressionNode) {
                return ((ExpressionNode) tmpVar.referentObject).evaluateAsInt();
            }
            else if (tmpVar.referentObject instanceof AcmeProperty) {
                AcmeProperty prop = (AcmeProperty) tmpVar.referentObject;
                // RTM 2/10/98 -- known bug.  elsewhere, it's not guaranteed that this
                //                property will have a value, so this can cause a
                //                null pointer exception.  This is not, however,
                //                a valid Armani construct, it's a vestigial effect
                //                from Acme's loose-ness.
                return ((AcmeInt) prop.getValue()).getValue();
            }
            else if (tmpVar.referentObject instanceof AcmePropertyValue) {
                return ((AcmeInt) tmpVar.referentObject).getValue();
            }
                // throw an exception. referentObject is not a recognized
                // type for evaluation as an integer.
                String opMsg = new String("Error! Identifier: "
                                + this.getIdentifier()
                                + " can not be evaluated as an integer.");
                throw new ExpressionCastException(this.getReferenceScope(),
                            this, ConstantTypes.INT_TYPE,
                            tmpVar.refObjectType, opMsg);
        }
    }
        // throw an exception. referentObject is not a recognized type for
        // evaluation as an integer.
        String opMsg = new String("Error! Identifier: " + this.getIdentifier()
                                  + " can not be evaluated as an integer.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.INT_TYPE, this.getType(), opMsg);

  };



/**
 * Dereference this identifier and attempt to evaluate it as a float
 * value
 * @return The evaluated float
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public float evaluateAsFloat() throws DesignRuleEvalException {
    // recalculate the referentObject but leave the type signature of
    // the reference untouched.
    this.refreshReferentObject();

    if (this.getReferentObject() == null) {
        // error, throw an exception.
        throw new ReferenceNotFoundException(this.getReferenceScope(), this,
            "Error! Unable to resolve identifier: " + this.getIdentifier()
            + " while evaluating expression as a float.");
    };


    if (!(AcmeTypeManager.isLogicalSubtype(
            this.getType(), ConstantTypes.FLOAT_TYPE, false)))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate identifier: "
                + this.getIdentifier() + "as a float expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);
    }
    else if (referentObject instanceof AcmeProperty) {
        AcmeProperty prop = (AcmeProperty) referentObject;
        // RTM 2/10/98 -- known bug.  elsewhere, it's not guaranteed that this
        //                property will have a value, so this can cause a
        //                null pointer exception.  This is not, however,
        //                a valid Armani construct, it's a vestigial effect
        //                from Acme's loose-ness.
        if (prop.getValue() instanceof AcmeFloat)
            return ((AcmeFloat) prop.getValue()).getValue();
        else if (prop.getValue() instanceof AcmeInt)
            return (float) ((AcmeInt) prop.getValue()).getValue();
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate identifier: "
                    + this.getIdentifier() + "as a float expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof AcmeFloat) {
        return ((AcmeFloat) referentObject).getValue();
    }

    else if (referentObject instanceof AcmeInt) {
        return (float) ((AcmeInt) referentObject).getValue();
    }

    else if (referentObject instanceof AcmeRecordField) {
        AcmeRecordField recField = (AcmeRecordField) referentObject;
        if (recField.getValue() instanceof AcmeFloat)
            return ((AcmeFloat) recField.getValue()).getValue();
        else if (recField.getValue() instanceof AcmeInt)
            return (float) ((AcmeInt) recField.getValue()).getValue();
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate identifier: "
                    + this.getIdentifier() + "as a float expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof VolatileVariable) {
        VolatileVariable tmpVar = (VolatileVariable) referentObject;
        if (tmpVar.referentObject == null) {
            String opMsg = new String("Error! Identifier: "
                    + this.getIdentifier()
                    + " does not have a valid referent object");
            throw new ReferenceNotFoundException(this.getReferenceScope(),
                    this, opMsg);
        }
            // figure out what kind of object the volatileVariable's
            // referent object is, and evaluate it as a float accordingly.
            if (tmpVar.referentObject instanceof ExpressionNode) {
                return ((ExpressionNode) tmpVar.referentObject).evaluateAsFloat();
            }
            else if (tmpVar.referentObject instanceof AcmeProperty) {
                AcmeProperty prop = (AcmeProperty) tmpVar.referentObject;
                // RTM 2/10/98 -- known bug.  elsewhere, it's not guaranteed that this
                //                property will have a value, so this can cause a
                //                null pointer exception.  This is not, however,
                //                a valid Armani construct, it's a vestigial effect
                //                from Acme's loose-ness.
                return ((AcmeFloat) prop.getValue()).getValue();
            }
            else if (tmpVar.referentObject instanceof AcmePropertyValue) {
                return ((AcmeFloat) tmpVar.referentObject).getValue();
            }
                // throw an exception. referentObject is not a recognized
                // type for evaluation as a float.
                String opMsg = new String("Error! Identifier: "
                                + this.getIdentifier()
                                + " can not be evaluated as a float.");
                throw new ExpressionCastException(this.getReferenceScope(),
                            this, ConstantTypes.FLOAT_TYPE,
                            tmpVar.refObjectType, opMsg);
    }
        // throw an exception. referentObject is not a recognized type for
        // evaluation as a float.
        String opMsg = new String("Error! Identifier: " + this.getIdentifier()
                                  + " can not be evaluated as a float.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);


  };


/**
 * Dereference this identifier and attempt to evaluate it as a string
 * value
 * @return The evaluated string
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public String evaluateAsString() throws DesignRuleEvalException {
    // recalculate the referentObject but leave the type signature of
    // the reference untouched.
    this.refreshReferentObject();

    if (this.getReferentObject() == null) {
        // error, throw an exception.
        throw new ReferenceNotFoundException(this.getReferenceScope(), this,
            "Error! Unable to resolve identifier: " + this.getIdentifier()
            + " while evaluating expression as a string.");
    };


    if (!(AcmeTypeManager.isLogicalSubtype(this.getType(), ConstantTypes.STRING_TYPE, false))
        && !(this.getType()instanceof EnumType))
    {
        if (referentObject instanceof AcmeEnum) {
            return ((AcmeEnum) referentObject).getValue();
        };

        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate identifier: "
                + this.getIdentifier() + "as a string expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.STRING_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof AcmeProperty) {
        AcmeProperty prop = (AcmeProperty) referentObject;
        // RTM 2/10/98 -- known bug.  elsewhere, it's not guaranteed that this
        //                property will have a value, so this can cause a
        //                null pointer exception.  This is not, however,
        //                a valid Armani construct, it's a vestigial effect
        //                from Acme's loose-ness.
        if (prop.getValue() instanceof AcmeString) {
            return ((AcmeString) prop.getValue()).getValue();
        }
        else if (prop.getValue() instanceof AcmeEnum) {
            return ((AcmeEnum) prop.getValue()).getValue();
        }
        else if (prop.getValue () == null) {
        	return "";
        }
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate identifier: "
                    + this.getIdentifier() + "as a string expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.STRING_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof AcmePropertyValue) {
        if (referentObject instanceof AcmeString) {
            return ((AcmeString) referentObject).getValue();
        }
        else if (referentObject instanceof AcmeEnum) {
            return ((AcmeEnum) referentObject).getValue();
        }
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate identifier: "
                    + this.getIdentifier() + "as a string expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.STRING_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof AcmeRecordField) {
        AcmeRecordField recField = (AcmeRecordField) referentObject;
        if (recField.getValue() instanceof AcmeString)
            return ((AcmeString) recField.getValue()).getValue();
        else if (recField.getValue() instanceof AcmeEnum)
            return ((AcmeEnum) recField.getValue()).getValue();
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate identifier: "
                    + this.getIdentifier() + "as a string expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.STRING_TYPE, this.getType(), opMsg);
    }

    else if (referentObject instanceof VolatileVariable) {
        VolatileVariable tmpVar = (VolatileVariable) referentObject;
        if (tmpVar.referentObject == null) {
            String opMsg = new String("Error! Identifier: "
                    + this.getIdentifier()
                    + " does not have a valid referent object");
            throw new ReferenceNotFoundException(this.getReferenceScope(),
                    this, opMsg);
        } else {
            // figure out what kind of object the volatileVariable's
            // referent object is, and evaluate it as a string accordingly.
            if (tmpVar.referentObject instanceof ExpressionNode) {
                return ((ExpressionNode) tmpVar.referentObject).evaluateAsString();
            }
            else if (tmpVar.referentObject instanceof AcmeProperty) {
                AcmeProperty prop = (AcmeProperty) tmpVar.referentObject;
                // RTM 2/10/98 -- known bug.  elsewhere, it's not guaranteed that this
                //                property will have a value, so this can cause a
                //                null pointer exception.  This is not, however,
                //                a valid Armani construct, it's a vestigial effect
                //                from Acme's loose-ness.
                return ((AcmeString) prop.getValue()).getValue();
            }
            else if (tmpVar.referentObject instanceof AcmePropertyValue) {
                return ((AcmeString) tmpVar.referentObject).getValue();
            }
                // throw an exception. referentObject is not a recognized
                // type for evaluation as a string.
                String opMsg = new String("Error! Identifier: "
                                + this.getIdentifier()
                                + " can not be evaluated as a float.");
                throw new ExpressionCastException(this.getReferenceScope(),
                            this, ConstantTypes.STRING_TYPE,
                            tmpVar.refObjectType, opMsg);
        }
    }

        // throw an exception. referentObject is not a recognized type for
        // evaluation as a string.
        String opMsg = new String("Error! Identifier: " + this.getIdentifier()
                                  + " can not be evaluated as a string.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.STRING_TYPE, this.getType(), opMsg);

  }
;




/**
 * Dereference this identifier and attempt to resolve it as an AcmeElement.
 * @return The evaluated AcmeElement
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public AcmeElement evaluateAsElement() throws DesignRuleEvalException {
    // recalculate the referentObject but leave the type signature of the
    // reference untouched.
    this.refreshReferentObject();

    if (this.getReferentObject() == null) {
        // error, throw an exception.
        throw new ReferenceNotFoundException(this.getReferenceScope(), this,
            "Error! Unable to resolve identifier: " + this.getIdentifier()
            + " while evaluating expression as an element.");
    };


    if (!(this.getType() instanceof AcmeElementType)
        && !(this.getType() instanceof AnyType))
    {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate identifier: "
                + this.getIdentifier() + "as an Element-typed expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.ELEMENT_BASE_TYPE, this.getType(), opMsg);
    }
    else if (referentObject instanceof AcmeElement) {
        return (AcmeElement) referentObject;
    }
    else if (referentObject instanceof VolatileVariable) {
        VolatileVariable tmpVar = (VolatileVariable) referentObject;

        // if it's still null then we can't resolve it.
        if (tmpVar.referentObject == null) {
            String opMsg = new String("Error! Identifier: "
                    + this.getIdentifier()
                    + " does not have a valid referent object");
            throw new ReferenceNotFoundException(this.getReferenceScope(),
                    this, opMsg);
        } else {
            // figure out what kind of object the volatileVariable's
            // referent object is, and evaluate it as an element.
            if (tmpVar.referentObject instanceof ExpressionNode) {
                return ((ExpressionNode) tmpVar.referentObject).evaluateAsElement();
            }
            else if (tmpVar.referentObject instanceof AcmeElement) {
                return (AcmeElement) tmpVar.referentObject;
            }
                // throw an exception. referentObject is not a recognized
                // type for evaluation as a boolean.
                String opMsg = new String("Error! Identifier: "
                                + this.getIdentifier()
                                + " can not be evaluated as an Element.");
                throw new ExpressionCastException(this.getReferenceScope(),
                            this, ConstantTypes.ELEMENT_BASE_TYPE,
                            tmpVar.refObjectType, opMsg);
        }
    }
        // throw an exception. referentObject is not a recognized type for
        // evaluation as a boolean.
        String opMsg = new String("Error! Identifier: " + this.getIdentifier()
                                  + " can not be evaluated as an element.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.ELEMENT_BASE_TYPE, this.getType(), opMsg);
  };





/**
 * Dereference this identifier and attempt to resolve it as an AcmeProperty
 * object (that is, a reference to a property).
 * @return The evaluated AcmeElement
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public AcmeProperty evaluateAsPropertyRef() throws DesignRuleEvalException {
    // recalculate the referentObject but leave the type signature of the
    // reference untouched.
    this.refreshReferentObject();

    if (this.getReferentObject() == null) {
        // error, throw an exception.
        throw new ReferenceNotFoundException(this.getReferenceScope(), this,
            "Error! Unable to resolve identifier: " + this.getIdentifier()
            + " while evaluating expression as a property ref.");
    };


    if (!(this.getType() instanceof AcmePropertyType)) {
        // error, throw an exception.
        String opMsg = new String("Error! Unable to evaluate identifier: "
                + this.getIdentifier() + "as an Property Reference expression.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.PROPERTY_REFERENCE_TYPE, this.getType(), opMsg);
    }
    else if (referentObject instanceof AcmeProperty) {
        return (AcmeProperty) referentObject;
    }
    else if (referentObject instanceof VolatileVariable) {
        VolatileVariable tmpVar = (VolatileVariable) referentObject;

        // if it's still null then we can't resolve it.
        if (tmpVar.referentObject == null) {
            String opMsg = new String("Error! Identifier: "
                    + this.getIdentifier()
                    + " does not have a valid referent object");
            throw new ReferenceNotFoundException(this.getReferenceScope(),
                    this, opMsg);
        } else {
            // figure out what kind of object the volatileVariable's
            // referent object is, and evaluate it as a property object.
            if (tmpVar.referentObject instanceof ExpressionNode) {
                return ((ExpressionNode) tmpVar.referentObject).evaluateAsPropertyRef();
            }
            else if (tmpVar.referentObject instanceof AcmeProperty) {
                return (AcmeProperty) tmpVar.referentObject;
            }
                // throw an exception. referentObject is not a recognized
                // type for evaluation as a boolean.
                String opMsg = new String("Error! Identifier: "
                                + this.getIdentifier()
                                + " can not be evaluated as a property reference.");
                throw new ExpressionCastException(this.getReferenceScope(),
                            this, ConstantTypes.PROPERTY_REFERENCE_TYPE,
                            tmpVar.refObjectType, opMsg);
        }
    }
        // throw an exception. referentObject is not a recognized type for
        // evaluation as a boolean.
        String opMsg = new String("Error! Identifier: " + this.getIdentifier()
                                  + " can not be evaluated as a property reference.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.PROPERTY_REFERENCE_TYPE, this.getType(), opMsg);
  };





/**
 * Evaluate this expression tree as a reference to a Type object.
 * If it can not be evaluated as a Type then throw a type exception.
 * @return The evaluated AcmePropertyType
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public AcmeType evaluateAsTypeRef() throws DesignRuleEvalException {

    // recalculate the referentObject but leave the type signature of
    // the reference untouched.
    this.refreshReferentObject();

    if (this.getReferentObject() == null) {
        // error, throw an exception.
        throw new ReferenceNotFoundException(this.getReferenceScope(), this,
            "Error! Unable to resolve identifier: " + this.getIdentifier()
            + " while evaluating expression as a type reference");
    };


    if (referentObject instanceof AcmeType) {
        return (AcmeType) referentObject;
    }

    else if (referentObject instanceof VolatileVariable) {
        VolatileVariable tmpVar = (VolatileVariable) referentObject;

        // if it's still null then we can't resolve it.
        if (tmpVar.referentObject == null) {
            String opMsg = new String("Error! Identifier: "
                    + this.getIdentifier()
                    + " does not have a valid referent object");
            throw new ReferenceNotFoundException(this.getReferenceScope(),
                    this, opMsg);
        } else {
            // figure out what kind of object the volatileVariable's
            // referent object is, and evaluate it as a type reference.
            if (tmpVar.referentObject instanceof ExpressionNode) {
                return ((ExpressionNode) tmpVar.referentObject).evaluateAsTypeRef();
            }
            else if (tmpVar.referentObject instanceof AcmeType) {
                return (AcmeType) tmpVar.referentObject;
            }
                // throw an exception. referentObject is not a recognized
                // type for evaluation as a boolean.
                String opMsg = new String("Error! Identifier: "
                                + this.getIdentifier()
                                + " can not be evaluated as a type reference.");
                throw new ExpressionCastException(this.getReferenceScope(),
                            this, ConstantTypes.TYPE_REFERENCE_TYPE,
                            tmpVar.refObjectType, opMsg);
        }
    }
        // throw an exception. referentObject is not a recognized type for
        // evaluation as a boolean.
        String opMsg = new String("Error! Identifier: " + this.getIdentifier()
                                  + " can not be evaluated as a type reference.");
        throw new ExpressionCastException(this.getReferenceScope(), this,
                ConstantTypes.TYPE_REFERENCE_TYPE, this.getType(), opMsg);

  }





/**
 * Dereference this identifier and attempt to resolve it as a
 * SetExpressionNode.
 * @return The evaluated set expression node
 * @exception DesignRuleEvalException If there is an evaluation error
 */
public SetExpressionNode evaluateAsSet() throws DesignRuleEvalException {
    // recalculate the referentObject but leave the type signature of the
    // reference untouched.
    this.refreshReferentObject();

    if (this.getReferentObject() == null) {
        // error, throw an exception.
        throw new ReferenceNotFoundException(
            this.getReferenceScope(),
            this,
            "Error! Unable to resolve identifier: " + this.getIdentifier() + " while evaluating expression as a set.");
    };

    if (!(this.getType() instanceof SetType)) {
        // error, throw an exception.
        String opMsg =
            new String("Error! Unable to evaluate identifier: " + this.getIdentifier() + "as a Set-typed expression.");
        throw new ExpressionCastException(
            this.getReferenceScope(),
            this,
            ConstantTypes.ANY_SET_TYPE,
            this.getType(),
            opMsg);
    }

    else if (referentObject instanceof AcmeProperty) {
        AcmeProperty refProp = (AcmeProperty) referentObject;
        if (refProp.getValue() instanceof SetExpressionNode) {
            return (SetExpressionNode) refProp.getValue();
        }
        String opMsg =
            new String("Error! Unable to cast property: " + this.getIdentifier() + "to a Set-typed expression.");
        throw new ExpressionCastException(
            this.getReferenceScope(),
            this,
            ConstantTypes.ANY_SET_TYPE,
            this.getType(),
            opMsg);
    }

    else if (referentObject instanceof AcmeRecordField) {
        AcmeRecordField recField = (AcmeRecordField) referentObject;
        if (recField.getValue() instanceof SetExpressionNode) {
            return (SetExpressionNode) recField.getValue();
        }
        // error, throw an exception.
        String opMsg =
            new String("Error! Unable to evaluate identifier: " + this.getIdentifier() + "as a set-typed expression.");
        throw new ExpressionCastException(
            this.getReferenceScope(),
            this,
            ConstantTypes.ANY_SET_TYPE,
            this.getType(),
            opMsg);
    }

    else if (referentObject instanceof VolatileVariable) {
        VolatileVariable tmpVar = (VolatileVariable) referentObject;
        if (tmpVar.referentObject == null && tmpVar.refObjectType == null) {
            String opMsg =
                new String("Error! Identifier: " + this.getIdentifier() + " does not have a valid referent object");
            throw new ReferenceNotFoundException(this.getReferenceScope(), this, opMsg);

        }
        else if (tmpVar.referentObject != null) {
            // figure out what kind of object the volatileVariable's
            // referent object is, and evaluate it as an element.
            if (tmpVar.referentObject instanceof ExpressionNode) {
                return ((ExpressionNode) tmpVar.referentObject).evaluateAsSet();
            }
            else if (tmpVar.referentObject instanceof SetExpressionNode) {
                return (SetExpressionNode) tmpVar.referentObject;
            }
            else if (tmpVar.referentObject instanceof AcmeProperty) {
                AcmeProperty refProp = (AcmeProperty) tmpVar.referentObject;
                if (refProp.getValue() instanceof SetExpressionNode) {
                    return (SetExpressionNode) refProp.getValue();
                }
                String opMsg =
                    new String(
                        "Error! Unable to cast property: "
                            + ((AcmeProperty) tmpVar.referentObject).getName()
                            + " referenced by identifier "
                            + this.getIdentifier()
                            + " to a Set-typed expression.");
                throw new ExpressionCastException(
                    this.getReferenceScope(),
                    this,
                    ConstantTypes.ANY_SET_TYPE,
                    this.getType(),
                    opMsg);
            }
        }
        else if (tmpVar.refObjectType != null && tmpVar.refObjectType instanceof SetType) {
            return new TransientSet(((SetType) tmpVar.refObjectType).getSetType());
        }
        // throw an exception. referentObject is not a recognized
        // type for evaluation as a set.
        String opMsg = new String("Error! Identifier: " + this.getIdentifier() + " can not be evaluated as a Set.");
        throw new ExpressionCastException(
            this.getReferenceScope(),
            this,
            ConstantTypes.ANY_SET_TYPE,
            tmpVar.refObjectType,
            opMsg);

    }
    // throw an exception. referentObject is not a recognized type for
    // evaluation as a boolean.
    String opMsg = new String("Error! Identifier: " + this.getIdentifier() + " can not be evaluated as an element.");
    throw new ExpressionCastException(
        this.getReferenceScope(),
        this,
        ConstantTypes.ELEMENT_BASE_TYPE,
        this.getType(),
        opMsg);
};






/**
 * Look up the entity that this identifier references, and set the
 * identifier's type as appropriate.  Return true if looking up and
 * setting the id's type fields was successful, return false if
 * difficulties are encountered, or the type can't be resolved.
 * @return true if type is correct
 */
  public boolean typecheck() {

    // first use this method's supertype to recurse and typecheck'
    // all of this node's children.
    if (super.typecheck() == false)
        return false;

    // first refresh the reference to make sure we're pointing to the
    // object we are identifying.
    try {
        super.refreshReferentObject();
    }
    catch (AcmeResolutionException e) {
        return false;
    }

    if (this.getReferentObject() == null) {
        // no object to check the type of!
        this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
        return false;
    };

    // handle the case where the referent object is a property
    if (this.getReferentObject() instanceof AcmeProperty) {
        AcmeProperty refProp = ((AcmeProperty) this.getReferentObject());
        this.setType(refProp.getType());
        return true;
    }

    else if (this.getReferentObject() instanceof AcmeRecordField) {
        // then set this node's type to the type of the record field
        this.setType(((AcmeRecordField) this.getReferentObject()).getType());
        return true;
    }

    else if (this.getReferentObject() instanceof AcmePropertyType) {
        // then this refers to a property type, set its type appropriately.
        this.setType((AcmePropertyType) this.getReferentObject());
        return true;
    }

    else if (this.getReferentObject() instanceof AcmeEnum) {
        // then this refers to a value of an enumerated type, set its
        // type appropriately.
        this.setType(((AcmeEnum) this.getReferentObject()).getType());
        return true;
    }

    // RTM 2/4/98 -- This isn't quite correct.  We should note that this
    //               identifier refers to an ElementType, rather than an
    //               element of this type.  The same problem appears with
    //               property types above.  We need to be able to name
    //               types directly for the analyses "declaresType(elt, eltType)"
    //               and "satisfiesType(elt, eltType)"  (perhaps these
    //               also work for property types).
    else if (this.getReferentObject() instanceof AcmeElementType) {
        // then this refers to an element type, set its type appropriately.
        AcmeElementType eltType = ((AcmeElementType) this.getReferentObject());
        this.setType(eltType);
        return true;
    }

    else if (this.getReferentObject() instanceof AcmeElement) {
        // then this refers to an element, set its type appropriately.
        AcmeElement eltType = ((AcmeElement) this.getReferentObject());

        // set the type of this expression to one of the constant base
        // element types.  If further typechecking needs to determine that
        // the element satisfies a richer type (e.g. FilterT instead of
        // Component), then that method can dererefence the actual
        // referent object to see if it satisfies the desired type.
        if (eltType instanceof AcmeSystem)
            this.setType(ConstantTypes.SYSTEM_BASE_TYPE);
        else if (eltType instanceof AcmeComponent)
            this.setType(ConstantTypes.COMPONENT_BASE_TYPE);
        else if (eltType instanceof AcmeConnector)
            this.setType(ConstantTypes.CONNECTOR_BASE_TYPE);
        else if (eltType instanceof AcmePort)
            this.setType(ConstantTypes.PORT_BASE_TYPE);
        else if (eltType instanceof AcmeRole)
            this.setType(ConstantTypes.ROLE_BASE_TYPE);
        else {
            // can't be an element that isn't a CCRP or S, right?
            this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
            return false;
        };

        return true;
    }

    else if (this.getReferentObject() instanceof VolatileVariable) {
        VolatileVariable var = ((VolatileVariable) this.getReferentObject());
        this.setType(var.refObjectType);
        return true;
    }

        // if we get to this point then we couldn't resolve the type properly.
        this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
        return false;

  };  // end typecheck() method



/**
 * export method
 * @return The output string
 */
  public String toString() {
      String ret = "";
      try {
          AcmeOutputStream s = new AcmeOutputStream ();
          exportArmani (s);
          s.flush ();
          ret = s.toString ();
          s.close ();
      }
      catch (IOException e) {}
      catch (AcmeException e) {}
    
      return ret;
  };

/**
 * export method
 * @param prefix The prefix
 * @return The output string
 */
  public String toString(String prefix) {
    return prefix + this.toString();
  };

  /**
   * zensoul added (01/06/2003)
   * @param the expression node whose reference object we'd like to get
   * @return the string rep of the refereenced obj, if available
   */
  private String toStringRefObj(ExpressionNode e) {
  	StringBuffer buf = new StringBuffer();
  	if (e instanceof ReferenceNode && e != null) {
  		Object obj = ((ReferenceNode) e).getReferentObject();
  		buf.append("[");
  		if (obj instanceof VolatileVariable) {
  			buf.append(((VolatileVariable) obj).refObjectType.getTypeName());
  			buf.append(":");
  			obj = ((VolatileVariable) obj).referentObject;
  		}
  		if (obj instanceof IdentifierNode) {
  			buf.append(((IdentifierNode) obj).getIdentifier());
  		} else if (obj instanceof AcmeElement) {
  			buf.append(((AcmeElement) obj).getName());
  		} else if (obj instanceof AcmeProperty) {
  			buf.append(((AcmeProperty) obj).getValue());
  		} else if (obj instanceof AcmeEnumeration) {
  			buf.append("{");
  			AcmeEnumeration enum = (AcmeEnumeration) obj;
  			while (enum.moreItems()) {
  				Object eObj = enum.getNextItem();
  				if (eObj instanceof AcmeElement) {
  	  	  			buf.append(((AcmeElement) eObj).getName());
  				} else {
  	  				buf.append(eObj.toString());
  				}
  				if (enum.moreItems()) buf.append(", ");
  			}
  			buf.append("}");
  		} else if (obj instanceof AcmeType) {
  			buf.append(((AcmeType) obj).getTypeName());
  		} else {
  			if (obj == null) {
  				buf.append("null");
  			} else {
  				buf.append(obj.toString());
  			}
  		}
//  		if (obj != null) {
//  			buf.append("::");
//  			buf.append(obj.getClass());
//  		}
  		buf.append("]");
  	}
  	
  	return buf.toString();
  }


/**
 * resolveVolatileFlag overrides (and makes use of) the ExpressionNode
 * implementation to both continue traversing the expression tree (though there
 * shouldn't be any children of an identifier node now, that could change...).
 * It also queries its scope to ask whether it's identifier is a volatile
 * variable, and sets its isVolatile flag appropriately.
 */
  public void resolveVolatileFlag() {
    String variableName = null;
    if (this.qualifiedNames != null) {
        if (this.qualifiedNames.size() > 0)
            // we're only interested in the outermost variable name
            // for volatility purposes
            variableName = (String) this.qualifiedNames.elementAt(0);
    };

    if (this.referenceScope instanceof DesignAnalysis) {
        this.setIsVolatile(
            ((DesignAnalysis) this.referenceScope).isVolatileVariable(variableName));
    } else if (this.referenceScope instanceof QuantifiedExpressionNode) {
        this.setIsVolatile(
            ((QuantifiedExpressionNode) this.referenceScope).isVolatileVariable(variableName));
    };

    super.resolveVolatileFlag();
  }

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion (this);
        if (this.getInParentheses()) {
            s.write ("(");
        }
        s.write (this.getIdentifier());
        if (ArmaniParser.isArgPrintingEnabled()) {
            s.write (toStringRefObj (this));
        }
        if (this.getInParentheses()) {
        	s.write (")");
        }
        s.endRegion (this);
    }

}
