package acmeLib.designRules;

import java.io.IOException;

import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.properties.AcmeFloat;
import acmeLib.properties.AcmeInt;
import acmeLib.properties.FloatType;

/**
 * Class: IntLiteralNode<br>
 * Purpose: Simple expression tree leaf-node that stores a float value<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    10/8/97    RTM     Created<br>
 *    2/2/98     RTM     Modified type storage to store AcmePropertyTypes<br>
 *                        instead of ExpressionTypeT objects.<br>
 *<br>
 * @author Bob Monroe
 */
public class IntLiteralNode extends NumericExpressionNode
{
/**
 * stores the value of this literal.
 */
  protected int value;

/**
 * JJTree-specific methods
 * @param id The ID
 * @return The node
 */
  public static Node jjtCreate(int id) {
    return new IntLiteralNode(id);
  }

/**
 * This is the constructor for this class
 * @param id The ID
 */
  public IntLiteralNode(int id) {
    super(id);
    this.setType(ConstantTypes.INT_TYPE);
    this.setValue(0);       // default to False?
  }

  //
  // Methods added by bmonroe
  //

/**
 * This is the constructor for this class
 * @param id The ID
 * @param val The value
 */
  public IntLiteralNode(int id, int val) {
    super(id);
    this.setType(ConstantTypes.INT_TYPE);
    this.setValue(val);
  }

/**
 * Copy constructor
 * @param copyOf The IntLiteralNode to be copied
 */
  public IntLiteralNode(IntLiteralNode copyOf) {
    super(copyOf);
    this.setValue(copyOf.getValue());
  }

/** 
 * cloneExpressionTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.  
 * @return The copy of the expression node
 */
  public ExpressionNode cloneExpressionTree() {
    IntLiteralNode clone = new IntLiteralNode(this);
    this.doCloneExpressionTree(clone);
    clone.setType(this.getType());
    clone.setValue(this.getValue());
    return clone;
  };


/**
 * Check to see if two nodes are equal
 * @param test The tested node
 * @return true if the test node is equal to this node
 */
  public boolean equals(Object test) {
    if (!(test instanceof ExpressionNode)) {
        if (test instanceof AcmeInt) {
            return (this.getValue() == ((AcmeInt) test).getValue());
        } else if (test instanceof AcmeFloat) {
            return (((float) this.getValue()) == ((AcmeFloat) test).getValue());
        }
            return false;
    }
        try {
            ExpressionNode testNode = (ExpressionNode) test;
            if (testNode.getType() instanceof FloatType) {
                return (((float) this.getValue()) == testNode.evaluateAsFloat());
            }
                return (this.getValue() == testNode.evaluateAsInt());
        } catch (Exception exc) {
        };
        return false;
  };


/**
 * Get the value
 * @return The value
 */
  public int getValue() { return this.value; }
/**
 * Set the value
 * @param val The new value
 */
  public void setValue(int val) { this.value = val; };


/**
 * because this node is a literal it must typecheck (it can't store
 * anything but a float).  Furthermore, it can't have any children nodes
 * so there's no need to traverse the children.
 * @return true if type is correct
 */
  public boolean typecheck() {
    this.setType(ConstantTypes.INT_TYPE);
    return true;
  };



/**
 * because this is always a leaf node, evaluate() just returns
 * the value of the literal.
 * @return The value of the literal
 */
  public int evaluate() { return this.value; };
/**
 * @return The value of the literal as float
 */
  public int evaluateAsInt() { return this.value; };
/**
 * @return The value of the literal as integer
 */
  public float evaluateAsFloat() { return (float) this.value; };
/**
 * @return The value of the literal as string
 */
  public String evaluateAsString() { return String.valueOf(this.value); };


/**
 * generic exporting method
 * @return The output string
 */
  public String toString() {
      String ret = "";
      try {
          AcmeOutputStream s = new AcmeOutputStream ();
          exportArmani (s);
          s.flush ();
          ret = s.toString ();
          s.close ();
      }
      catch (IOException e) {}
      catch (AcmeException e) {}
    
      return ret;
  };


/**
 * export method
 * @param prefix The prefix
 * @return The output string
 */
  public String toString(String prefix) {
    return prefix + this.toString();
  };

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion (this);
        if (this.getInParentheses()) {
            s.write ("(" + String.valueOf (value) + ")");
        }
        else {
            s.write (String.valueOf (value));
        }
        s.endRegion (this);
    }

}
