package acmeLib.designRules;

import java.io.File;
import java.lang.reflect.Method;
import java.util.StringTokenizer;

import acmeLib.AcmeType;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.elements.AcmeComponent;
import acmeLib.elements.AcmeConnector;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmePort;
import acmeLib.elements.AcmeRole;
import acmeLib.elements.AcmeSystem;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.properties.AcmeProperty;
import acmeLib.tools.ExternalAnalysesClassLoader;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: JavaMethodCallNode<br>
 * Purpose: A JavaMethodCallNode node is an expression node that represents<br>
 *          a call to a java method.  As of 3/24/98 this is only used for<br>
 *          the body of an ExternalDesignAnalysis expression.  This adds caching<br>
 *          for the Java class and method objects referenced, mechanisms for<br>
 *          resolving and dereferencing the class and method, and typechecking<br>
 *          methods.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *   3/24/98    RTM      * created initial class<br>
 *<br>
 * @author Bob Monroe
 */
public class JavaMethodCallNode extends DesignAnalysisCallNode implements Scope
{

    //
    // protected members
    //

  /**
   * cache the string name of the class and method
   */
    protected String className = null;
    protected String methodName = null;

  /**
   * cache the Class and Method objects to invoke when the design analysis
   * is analyzed.
   */
    protected Class javaClass = null;
    protected Method javaMethod = null;

  /**
   * stores the array of ExpressionNode parameters to pass to the
   * java method.
   */
    ExpressionNode javaMethodParams[] = null;

  /** Class loader for specific external design analysis path */
    ExternalAnalysesClassLoader loader = null;

  /**
   * Constant Class objects to speed up typechecking
   */
    protected static Class acmeElementClass = AcmeElement.class;
                                
    protected static Class acmeSystemClass = AcmeSystem.class;
                               
    protected static Class acmeComponentClass = AcmeComponent.class;
                                
    protected static Class acmeConnectorClass = AcmeConnector.class;
                                
    protected static Class acmePortClass = AcmePort.class;
                                
    protected static Class acmeRoleClass = AcmeRole.class;
                                

  /**
   * This Class object is dependent on the fact that the first Interface
   * declared for TransientSet is SetExpressionNode.  If this changes in
   * TransientSet.java, it needs to be updated here as well.  This is a
   * bit goofy because the return type for set-valued external analysis
   * methods is the SetExpressionNode _interface_ rather than a class.
   * I haven't figured out how to
   */
    protected static Class setExpressionNodeInterface = null;
                            // (new TransientSet()).getClass().getInterfaces()[0];
    protected static boolean staticsInitialized = false;



  /**
   * make sure the static members that need special initialization
   * have been initialized.  This is a kludge to deal with the fact
   * that calculating the appropriate initialization value for some
   * of the Class objects requires throwing and catching exceptions,
   * which can't be done in an initialization statement outside of
   * a method.
   */
    protected static void initStatics() {
        if (!staticsInitialized) {
            try {
                setExpressionNodeInterface =
                    Class.forName("acmeLib.designRules.SetExpressionNode");
            } catch (ClassNotFoundException cnfExc) {
                System.err.println("Warning: unable to load class "
                        + "\"acmeLib.designRules.SetExpressionNode\"");
            };
            staticsInitialized = true;
        };
    };


  /**
   * helper method for resolveInstanceReferences() and other methods that
   * looks up the referent object, which is a java method in this class.
   */
    public void refreshReferentObject() {
    	// parse the qualified identifier to get the class and method names.
        this.setMethodName(this.getUnqualifiedIdentifier());
        int lastIndex = this.getIdentifier().lastIndexOf(".");
        if (lastIndex == -1) {
            // we don't have an expression to dereference. announce an error
            String errMsg = "Error, no class name associated with method name "
                            + this.methodName + " in JavaMethodCallNode "
                            + this.getIdentifier();
            AcmeError newError = new AcmeError(this, errMsg);
            AcmeErrorManager.errorDetected(newError);
			AcmeTypeManager.typecheckingErrorStack.pushCausalError(newError);

            return;
        } else {
            // find the hack-ish triple underscore to determine a path
            final String HACK_SEP = "___";
            int pathIndex = getIdentifier().indexOf(HACK_SEP);
            String rootPath = ".";
            if (pathIndex > -1) {
                // found! extract path and create class loader
                rootPath = getIdentifier().substring(0, pathIndex);
                setClassName(getIdentifier().substring(pathIndex + HACK_SEP.length(), lastIndex));
                // process the substitution of any env't var
                if (rootPath.startsWith("_")) {
                    final String PSEP = System.getProperty("path.separator");
                    final String FSEP = "__";
                    final String REPL_FSEP = "/";
                    int end = rootPath.indexOf(FSEP);
                    if (end == -1) {  // no file sep, use the length of the path
                        end = rootPath.length();
                    }
                    String pathVar = rootPath.substring(1, end);  // don't include the "_"
                    StringTokenizer pathTokens = new StringTokenizer(System.getProperty(pathVar), PSEP);
                    while (pathTokens.hasMoreTokens()) {
                        String pathValue = pathTokens.nextToken();
                        String newPath = pathValue;
                        if (end < rootPath.length()) {
                            newPath += rootPath.substring(end).replaceAll("__", REPL_FSEP);
                        }
                        File testFile = new File(newPath);
                        if (testFile.exists()) {
                         	rootPath = newPath;
                            break;  // found one, stop searching
                        }
                    }
                }
                if (loader == null) {
                	loader = new ExternalAnalysesClassLoader(getClass().getClassLoader());
                }
                loader.setPath(rootPath);
            } else {
            	setClassName(getIdentifier().substring(0, lastIndex));
            }
        };

        // if we get to here then we should have the className and methodName
        // parsed.  look them up and cache them.  Throw an error if we can't
        // look them up properly.
        try {
            if (loader != null) {
                setJavaClass(loader.loadClass(getClassName()));
            } else {
            	setJavaClass(Class.forName(getClassName()));
            }
            if (this.getJavaClass() == null) {
                throw new ClassNotFoundException();
            };
        } catch (ClassNotFoundException exc) {
            // else can't recognize what the design rule expression refers to...
            String errMsg = "Error, unable to load class " + this.getClassName()
                            + " referenced as: " + this.getClassName();
            AcmeError newError = new AcmeError(this, errMsg);
            AcmeErrorManager.errorDetected(newError);
			AcmeTypeManager.typecheckingErrorStack.pushCausalError(newError);

            return;
        };

        // RTM 3/23/98 -- For the time being, all arguments are passed as
        //                ExpressionNodes that the native implementing method
        //                has to decode properly.  Make this richer if I can
        //                come up with a good way to make the implementations
        //                of the type systems for Java and Armani work more
        //                closely together.
        int numArgs = this.jjtGetNumChildren();
        Class args[] = new Class[numArgs];
   
        for (int i = 0; i < numArgs; i++) {
            args[i] = ExpressionNode.class;
        };

        try {
            this.setJavaMethod(
                    this.getJavaClass().getMethod(this.getMethodName(),args));
        }

        catch (NoSuchMethodException nsmExc) {
            // requested method doesn't exist, announce an error
            String errMsg = "Error, unable to load method " + this.getMethodName()
                            + " of class " + this.getClassName()
                            + " with args: " + args.toString();
            AcmeError newError = new AcmeError(this, errMsg);
            AcmeErrorManager.errorDetected(newError);
			AcmeTypeManager.typecheckingErrorStack.pushCausalError(newError);

            return;
        }

        catch (SecurityException secExc) {
            // requested method doesn't exist, announce an error
            String errMsg = "Error, security exception raised while"
                            + " loading method " + this.getMethodName()
                            + " of class " + this.getClassName();
            AcmeError newError = new AcmeError(this, errMsg);
			AcmeTypeManager.typecheckingErrorStack.pushCausalError(newError);

            AcmeErrorManager.errorDetected(newError);
            return;
        };

        // if we get to here then we've successfully resolved the name of the
        // java class and method, and looked up the methods and classes.
        this.setReferentObject(this.getJavaMethod());
    }


    //
    // public members
    //
  /**
   * This is the constructor of this class
   * @param id The ID
   */
    public JavaMethodCallNode(int id) {
        super(id);
        // initialize the protected members of this rule to defaults
        this.setClassName(null);
        this.setMethodName(null);
        this.setJavaClass(null);
        this.setJavaMethod(null);

        // make sure the static members that need special initialization
        // have been initialized.  This is a kludge to deal with the fact
        // that calculating the appropriate initialization value for some
        // of the Class objects requires throwing and catching exceptions,
        // which can't be done in an initialization statement outside of
        // a method.
        initStatics();
    }

  /**
   * Copy constructor
   * @param copyOf The node to be copied
   */
    public JavaMethodCallNode(JavaMethodCallNode copyOf) {
        super(copyOf);
        // initialize the protected members of this rule to defaults
        this.setClassName(null);
        this.setMethodName(null);
        this.setJavaClass(null);
        this.setJavaMethod(null);
        initStatics();
    }

  /**
   * JJTree-specific methods
   * @param id The ID
   * @return The node
   */
    public static Node jjtCreate(int id) {
        return new JavaMethodCallNode(id);
    }


  /**
   * cloneExpressionTree recursively deep-copies an entire expression tree,
   * creating deep copies of all of it's children nodes in addition
   * to its own fields.
   * @return The copy of the expression node
   */
    public ExpressionNode cloneExpressionTree() {
        JavaMethodCallNode clone = new JavaMethodCallNode(this);
        this.doCloneExpressionTree(clone);
        clone.setClassName(this.className);
        clone.setMethodName(this.methodName);
        clone.setJavaClass(this.javaClass);
        clone.setJavaMethod(this.javaMethod);
        return clone;
    };


    // accessor/settor methods
  /**
   * Set the class name
   * @param newName The new class name
   */
    public void setClassName(String newName) { this.className = newName; };
  /**
   * Get the class name
   * @return the class name
   */
    public String getClassName() { return this.className; };

  /**
   * Set the method name
   * @param newName The new method name
   */
    public void setMethodName(String newName) { this.methodName = newName; };
  /**
   * Get the method name
   * @return the method name
   */
    public String getMethodName() { return this.methodName; };

  /**
   * Set the Java class name
   * @param newClass The new Java class name
   */
    public void setJavaClass(Class newClass) { this.javaClass = newClass; };
  /**
   * Get the Java class name
   * @return the Java class name
   */
    public Class getJavaClass() { return this.javaClass; };

  /**
   * Set the Java method name
   * @param newMethod The new Java method name
   */
    public void setJavaMethod(Method newMethod) { this.javaMethod = newMethod; };
  /**
   * Get the Java method name
   * @return the Java method name
   */
    public Method getJavaMethod() { return this.javaMethod; };

  /**
   * Set the Java method parameters
   * @param params The new Java method parameters
   */
    public void setJavaMethodParams(ExpressionNode[] params) {
        this.javaMethodParams = params;
    };
  /**
   * Get the Java method parameters
   * @return the Java method parameters
   */
    public ExpressionNode[] getJavaMethodParams() {
        return this.javaMethodParams;
    };


  /**
   * load the javaParams array with the ExpressionNodes passed to
   * this JavaMethodCallNode's parent DesignAnalysis as actual parameters.
   * Identifier children of this node are dereferenced to retrieve the
   * ExpressionNode passed to the parent as parameters, Literal
   * ExpressionNodes are are put directly into the array of actual parameters.
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public void loadJavaMethodParams() throws DesignRuleEvalException
    {
        this.loadActualParams();
        if (this.getActualParams() == null) {
            this.setJavaMethodParams(new ExpressionNode[0]);
            return;
        };

        int numParams = this.getActualParams().size();
        ExpressionNode newJavaParams[] = new ExpressionNode[numParams];
        for (int i = 0; i < numParams; i++) {
            ExpressionNode curChild =
                        (ExpressionNode) this.getActualParams().elementAt(i);
            if (curChild instanceof IdentifierNode) {
                String curParamName =
                        ((IdentifierNode) curChild).getUnqualifiedIdentifier();
                Scope refScope = ((IdentifierNode) curChild).getReferenceScope();
                Object daParam = refScope.lookupName(curParamName, true);

                if (daParam instanceof VolatileVariable) {
                    VolatileVariable tmpDaParam = (VolatileVariable) daParam;
                    if (tmpDaParam.referentObject instanceof ExpressionNode) {
                        newJavaParams[i] = (ExpressionNode) tmpDaParam.referentObject;
                    } else {
                        String errMsg = "JavaMethodCall parameter: "
                            + curParamName
                            + " does not refer to a valid ExpressionNode.";
                        throw new DesignRuleEvalException(errMsg);
                    };
                }

                else if (daParam instanceof ExpressionNode) {
                    newJavaParams[i] = (ExpressionNode) daParam;
                }

                else {  // else something is really messed up, throw an exception.
                    String errMsg = "Unable to properly dereference "
                        + "JavaMethodCall parameter: " + curParamName;
                    throw new DesignRuleEvalException(errMsg);
                };
            }

            else if (curChild instanceof ExpressionNode) {
                newJavaParams[i] = (ExpressionNode) curChild;
            }

            else {
                String errMsg = "Unrecognized Parameter number(" + i + ")"
                    + " of JavaMethodCall: " + this.getIdentifier();
                throw new DesignRuleEvalException(errMsg);
            };
        };

        this.setJavaMethodParams(newJavaParams);
    };


  /**
   * typecheck infers the type of this expression by converting the return
   * type of the referenced Java method to an appropriate Armani type.  It
   * returns true if it is successful in infering the type, else returns false.
   * @return true if type is correct
   */
    public boolean typecheck() {
        // RTM 3/24/98 -- currently implemented only for primitive property
        //                types.  Extend in the near-future as I figure out
        //                how to do so...

        // assume that this does not typecheck unless I can show that it does.
        boolean typechecks = false;
        if (this.getJavaMethod() == null) {
            this.refreshReferentObject();
            // if it's still null it's an error
            if (this.getJavaMethod() == null) {
                // RTM 3/24/98 -- add appropriate exception/error throwing here.
                return false;
            };
        };

        // look up the return type of the method.
        Class returnClass = this.getJavaMethod().getReturnType();

        // check for primitive types
        if (returnClass == Boolean.TYPE) {
            this.setType(ConstantTypes.BOOLEAN_TYPE);
            typechecks = true;
        } else if (returnClass == Integer.TYPE) {
            this.setType(ConstantTypes.INT_TYPE);
            typechecks = true;
        } else if (returnClass == Float.TYPE) {
            this.setType(ConstantTypes.FLOAT_TYPE);
            typechecks = true;
        } else if (returnClass == (new String()).getClass()) {
            this.setType(ConstantTypes.STRING_TYPE);
            typechecks = true;
        }

        // check for basic element types
        else if (returnClass == JavaMethodCallNode.acmeElementClass) {
            this.setType(ConstantTypes.ELEMENT_BASE_TYPE);
            typechecks = true;
        } else if (returnClass == JavaMethodCallNode.acmeSystemClass) {
            this.setType(ConstantTypes.SYSTEM_BASE_TYPE);
            typechecks = true;
        } else if (returnClass == JavaMethodCallNode.acmeComponentClass) {
            this.setType(ConstantTypes.COMPONENT_BASE_TYPE);
            typechecks = true;
        } else if (returnClass == JavaMethodCallNode.acmeConnectorClass) {
            this.setType(ConstantTypes.CONNECTOR_BASE_TYPE);
            typechecks = true;
        } else if (returnClass == JavaMethodCallNode.acmePortClass) {
            this.setType(ConstantTypes.PORT_BASE_TYPE);
            typechecks = true;
        } else if (returnClass == JavaMethodCallNode.acmeRoleClass) {
            this.setType(ConstantTypes.ROLE_BASE_TYPE);
            typechecks = true;
        }

        // check for set types.
        else if (returnClass == JavaMethodCallNode.setExpressionNodeInterface) {
            this.setType(ConstantTypes.ANY_SET_TYPE);
            typechecks = true;
        }

        // RTM 3/24/98 -- don't quite know how to check for other compound
        //                property types yet...

        return typechecks;
    }




  /**
   * EvaluateAsBoolean calculates the actual parameters for the Java
   * method call and invokes the method referred to by this node
   * with the set of Armani actual parameters (in the form of ExpressionNode's)
   * as Java parameters.  The actual parameters are initially parsed
   * as children nodes of the Java method call node.
   * @return The evaluated boolean value
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public boolean evaluateAsBoolean() throws DesignRuleEvalException {

        if (this.getJavaMethod() == null) {
            throw new DesignRuleEvalException("The Java method "
                    + this.getIdentifier() + " is not properly dereferenced "
                    + "and therefore can not be evaluated as a boolean.");
        };

        if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.BOOLEAN_TYPE, false)))
        {
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate Java Method: "
                    + this.getIdentifier() + " as a boolean expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.BOOLEAN_TYPE, this.getType(), opMsg);
        };

        // load the javaMethodParams vector with the dereferenced values
        // of the actual parameters passed to the DesignAnalysis.
        this.loadJavaMethodParams();

        // call the static method with the args ExpressionNodes
        Object retObj = null;
        try {
            retObj = this.getJavaMethod().invoke(null, this.getJavaMethodParams());
        } catch (Throwable t) {
            t.printStackTrace();
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a boolean "
                + "because external method invocation failed: "
                + t.getMessage());
        };

        boolean result = false;
        if (retObj instanceof Boolean) {
            result = ((Boolean) retObj).booleanValue();
        } else {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a boolean "
                + "because external method does not return a Boolean object.");
        };

        return result;
    };




  /**
   * EvaluateAsInt calculates the actual parameters for the Java
   * method call and invokes the method referred to by this node
   * with the set of Armani actual parameters (in the form of ExpressionNode's)
   * as Java parameters.  The actual parameters are initially parsed
   * as children nodes of the Java method call node.
   * @return The evaluated integer value
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public int evaluateAsInt() throws DesignRuleEvalException {

        if (this.getJavaMethod() == null) {
            throw new DesignRuleEvalException("The Java method "
                    + this.getIdentifier() + " is not properly dereferenced "
                    + "and therefore can not be evaluated as an integer.");
        };

        if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.INT_TYPE, false)))
        {
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate Java Method: "
                    + this.getIdentifier() + " as an integer expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.INT_TYPE, this.getType(), opMsg);
        };

        // load the javaMethodParams vector with the dereferenced values
        // of the actual parameters passed to the DesignAnalysis.
        this.loadJavaMethodParams();

        // call the static method with the args ExpressionNodes
        Object retObj = null;
        try {
            retObj = this.getJavaMethod().invoke(null, this.getJavaMethodParams());
        } catch (Throwable t) {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as an integer "
                + "because external method invocation failed: "
                + t.getMessage());
        };

        int result;
        if (retObj instanceof Integer) {
            result = ((Integer) retObj).intValue();
        } else if (retObj instanceof Float) {
            result = ((Float) retObj).intValue();
        } else {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as an integer "
                + "because external method does not return an Integer or "
                + "Float object.");
        };

        return result;
    };



  /**
   * EvaluateAsFloat calculates the actual parameters for the Java
   * method call and invokes the method referred to by this node
   * with the set of Armani actual parameters (in the form of ExpressionNode's)
   * as Java parameters.  The actual parameters are initially parsed
   * as children nodes of the Java method call node.
   * @return The evaluated float value
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public float evaluateAsFloat() throws DesignRuleEvalException {

        if (this.getJavaMethod() == null) {
            throw new DesignRuleEvalException("The Java method "
                    + this.getIdentifier() + " is not properly dereferenced "
                    + "and therefore can not be evaluated as a float.");
        };

        if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.FLOAT_TYPE, false)))
        {
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate Java Method: "
                    + this.getIdentifier() + " as a float expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);
        };

        // load the javaMethodParams vector with the dereferenced values
        // of the actual parameters passed to the DesignAnalysis.
        this.loadJavaMethodParams();

        // call the static method with the args ExpressionNodes
        Object retObj = null;
        try {
            retObj = this.getJavaMethod().invoke(null, this.getJavaMethodParams());
        } catch (Throwable t) {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a float "
                + "because external method invocation failed: "
                + t.getMessage());
        };

        float result;
        if (retObj instanceof Integer) {
            result = ((Integer) retObj).floatValue();
        } else if (retObj instanceof Float) {
            result = ((Float) retObj).floatValue();
        } else {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a float "
                + "because external method does not return an Integer or "
                + "Float object.");
        };

        return result;
    };



  /**
   * EvaluateAsString calculates the actual parameters for the Java
   * method call and invokes the method referred to by this node
   * with the set of Armani actual parameters (in the form of ExpressionNode's)
   * as Java parameters.  The actual parameters are initially parsed
   * as children nodes of the Java method call node.
   * @return The evaluated string value
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public String evaluateAsString() throws DesignRuleEvalException {

        if (this.getJavaMethod() == null) {
            throw new DesignRuleEvalException("The Java method "
                    + this.getIdentifier() + " is not properly dereferenced "
                    + "and therefore can not be evaluated as a string.");
        };

        if ( (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.STRING_TYPE, false)))
            && (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.FLOAT_TYPE, false)))
            && (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.INT_TYPE, false)))
            && (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.BOOLEAN_TYPE, false))) )
        {
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate Java Method: "
                    + this.getIdentifier() + " as a string expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.STRING_TYPE, this.getType(), opMsg);
        };

        // load the javaMethodParams vector with the dereferenced values
        // of the actual parameters passed to the DesignAnalysis.
        this.loadJavaMethodParams();

        // call the static method with the args ExpressionNodes
        Object retObj = null;
        try {
            retObj = this.getJavaMethod().invoke(null, this.getJavaMethodParams());
        } catch (Throwable t) {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a string "
                + "because external method invocation failed: "
                + t.getMessage());
        };

        String result;
        if (retObj instanceof String) {
            result = (String) retObj;
        } else {
            result = retObj.toString();
        };

        return result;
    };




  /**
   * EvaluateAsElement calculates the actual parameters for the Java
   * method call and invokes the method referred to by this node
   * with the set of Armani actual parameters (in the form of ExpressionNode's)
   * as Java parameters.  The actual parameters are initially parsed
   * as children nodes of the Java method call node.
   * @return The evaluated AcmeElement
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public AcmeElement evaluateAsElement() throws DesignRuleEvalException {

        if (this.getJavaMethod() == null) {
            throw new DesignRuleEvalException("The Java method "
                    + this.getIdentifier() + " is not properly dereferenced "
                    + "and therefore can not be evaluated as an element.");
        };

        if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.ELEMENT_BASE_TYPE, false)))
        {
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate Java Method: "
                    + this.getIdentifier() + " as an element expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.ELEMENT_BASE_TYPE, this.getType(), opMsg);
        };

        // load the javaMethodParams vector with the dereferenced values
        // of the actual parameters passed to the DesignAnalysis.
        this.loadJavaMethodParams();

        // call the static method with the args ExpressionNodes
        Object retObj = null;
        try {
            retObj = this.getJavaMethod().invoke(null, this.getJavaMethodParams());
        } catch (Throwable t) {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as an element "
                + "because external method invocation failed: "
                + t.getMessage());
        };

        AcmeElement result = null;
        if (retObj instanceof AcmeElement) {
            result = (AcmeElement) retObj;
        } else {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as an element "
                + "because external method does not return an AcmeElement-typed "
                + " object.");
        };

        return result;
    };




  /**
   * EvaluateAsPropertyRef calculates the actual parameters for the Java
   * method call and invokes the method referred to by this node
   * with the set of Armani actual parameters (in the form of ExpressionNode's)
   * as Java parameters.  The actual parameters are initially parsed
   * as children nodes of the Java method call node.
   * @return The evaluated AcmeProperty
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public AcmeProperty evaluateAsPropertyRef() throws DesignRuleEvalException {

        if (this.getJavaMethod() == null) {
            throw new DesignRuleEvalException("The Java method "
                    + this.getIdentifier() + " is not properly dereferenced "
                    + "and therefore can not be evaluated as a Property Reference.");
        };

        if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.PROPERTY_REFERENCE_TYPE, false)))
        {
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate Java Method: "
                    + this.getIdentifier() + " as a property reference expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.PROPERTY_REFERENCE_TYPE, this.getType(), opMsg);
        };

        // load the javaMethodParams vector with the dereferenced values
        // of the actual parameters passed to the DesignAnalysis.
        this.loadJavaMethodParams();

        // call the static method with the args ExpressionNodes
        Object retObj = null;
        try {
            retObj = this.getJavaMethod().invoke(null, this.getJavaMethodParams());
        } catch (Throwable t) {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a property "
                + "reference because external method invocation failed: "
                + t.getMessage());
        };

        AcmeProperty result = null;
        if (retObj instanceof AcmeProperty) {
            result = (AcmeProperty) retObj;
        } else {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a property reference "
                + "because external method does not return an AcmeProperty-typed "
                + " object.");
        };

        return result;
    };


	/**
	 * EvaluateAsTypeRef calculates the actual parameters for the Java
   * method call and invokes the method referred to by this node
   * with the set of Armani actual parameters (in the form of ExpressionNode's)
   * as Java parameters.  The actual parameters are initially parsed
   * as children nodes of the Java method call node.
   * @return The evaluated AcmeType
   * @exception DesignRuleEvalException If there is an evaluation error
	 */
	public AcmeType evaluateAsTypeRef() throws DesignRuleEvalException {
		if (this.getJavaMethod() == null) {
			 throw new DesignRuleEvalException("The Java method "
					 + this.getIdentifier() + " is not properly dereferenced "
					 + "and therefore can not be evaluated as a Type Reference.");
		 };

		 if (!this.getType().equals (ConstantTypes.UNRESOLVED_TYPE) && 
		 		!(AcmeTypeManager.isLogicalSubtype(this.getType(),
												ConstantTypes.TYPE_REFERENCE_TYPE, false)))
		 {
			 // error, throw an exception.
			 String opMsg = new String("Error! Unable to evaluate Java Method: "
					 + this.getIdentifier() + " as a type reference expression.");
			 throw new ExpressionCastException(this.getReferenceScope(), this,
					 ConstantTypes.TYPE_REFERENCE_TYPE, this.getType(), opMsg);
		 };

		 // load the javaMethodParams vector with the dereferenced values
		 // of the actual parameters passed to the DesignAnalysis.
		 this.loadJavaMethodParams();

		 // call the static method with the args ExpressionNodes
		 Object retObj = null;
		 try {
			 retObj = this.getJavaMethod().invoke(null, this.getJavaMethodParams());
		 } catch (Throwable t) {
			 throw new DesignRuleEvalException(
				 "Unable to evaluate external design analysis as a type "
				 + "reference because external method invocation failed: "
				 + t.getMessage());
		 };

		 AcmeType result = null;
		 if (retObj instanceof AcmeType) {
			 result = (AcmeType) retObj;
		 } else {
			 throw new DesignRuleEvalException(
				 "Unable to evaluate external design analysis as a type reference "
				 + "because external method does not return a type");
		 };

		 return result;		
	}


  /**
   * EvaluateAsSet calculates the actual parameters for the Java
   * method call and invokes the method referred to by this node
   * with the set of Armani actual parameters (in the form of ExpressionNode's)
   * as Java parameters.  The actual parameters are initially parsed
   * as children nodes of the Java method call node.
   * @return The evaluated set
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public SetExpressionNode evaluateAsSet() throws DesignRuleEvalException {

        if (this.getJavaMethod() == null) {
            throw new DesignRuleEvalException("The Java method "
                    + this.getIdentifier() + " is not properly dereferenced "
                    + "and therefore can not be evaluated as a set.");
        };

        if (!(AcmeTypeManager.isLogicalSubtype(this.getType(),
                                               ConstantTypes.ANY_SET_TYPE, false)))
        {
            // error, throw an exception.
            String opMsg = new String("Error! Unable to evaluate Java Method: "
                    + this.getIdentifier() + " as a set expression.");
            throw new ExpressionCastException(this.getReferenceScope(), this,
                    ConstantTypes.ANY_SET_TYPE, this.getType(), opMsg);
        };

        // load the javaMethodParams vector with the dereferenced values
        // of the actual parameters passed to the DesignAnalysis.
        this.loadJavaMethodParams();

        // call the static method with the args ExpressionNodes
        Object retObj = null;
        try {
            retObj = this.getJavaMethod().invoke(null, this.getJavaMethodParams());
        } catch (Throwable t) {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a set "
                + "because external method invocation failed: "
                + t.getMessage());
        };

        SetExpressionNode result = null;
        if (retObj instanceof SetExpressionNode) {
            result = (SetExpressionNode) retObj;
        } else {
            throw new DesignRuleEvalException(
                "Unable to evaluate external design analysis as a set"
                + "because external method does not return an AcmeElement-typed "
                + " object.");
        };

        return result;
    };






  /**
   * Implements the Scope interface.  Only the actual parameters
   * of the JavaMethodCallNode can be defined within the scope of this node,
   * and the useful referent object for those are stored in this node's
   * parent -- a DesignAnalysis object.  So simply pass requests to
   * lookup a name up to the DesignAnalysis parent object.
   * @param name The name
   * @param bubbleUp pass the request up
   * @return The object
   */
    public Object lookupName(String name, boolean bubbleUp) {
        if (this.getReferenceScope() instanceof Scope) {
            return this.getReferenceScope().lookupName(name, bubbleUp);
        }
            return null;
    };
};