package acmeLib.designRules;

import acmeLib.AcmeType;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExpressionOperatorT;
import acmeLib.properties.*;

/**
 * Class: NumericExpressionNode<br>
 * Purpose: root of a (possibly 1-node) tree that represents a numeric<br>
 *          expression.  This can store either a floating point or integer<br>
 *          representation.  Likewise it can be viewed as having either a<br>
 *          floating point or integer value.  Merging integer and<br>
 *          floating point expressions simplifies evaluation and storage<br>
 *          in dealing with the Java type system and JavaCC.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    10/9/97    RTM       Created<br>
 *    10/28/97   RTM       Added support for exception handling<br>
 *    2/2/98     RTM       Modified type storage to store AcmeTypes<br>
 *                          instead of ExpressionTypeT objects.<br>
 *<br>
 * @author Bob Monroe
 */
public class NumericExpressionNode extends ExpressionNode
{
  //
  // public members
  //

/**
 * This is the constructor for this class
 * @param id The ID
 */
  public NumericExpressionNode(int id) {
    super(id);
    this.setType(ConstantTypes.FLOAT_TYPE);      // default to float type
  }


/**
 * Copy constructor
 * @param copyOf The NumericExpressionNode to be copied
 */
  public NumericExpressionNode(NumericExpressionNode copyOf) {
    super(copyOf);
  }


/**
 * JJTree-specific methods
 * @param id The ID
 * @return The node
 */
 public static Node jjtCreate(int id) {
    return new NumericExpressionNode(id);
  }


/** 
 * cloneExpressionTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.  
 * @return The copy of the expression node
 */
  public ExpressionNode cloneExpressionTree() {
    NumericExpressionNode clone = new NumericExpressionNode(this);
    this.doCloneExpressionTree(clone);
    return clone;
  };



/**
 * Check to see if two nodes are equal
 * @param test The tested node
 * @return true if the test node is equal to this node
 */
  public boolean equals(Object test) {
    float testFloatVal;
    if (!(test instanceof ExpressionNode)) {
        if (test instanceof AcmeFloat) {
            testFloatVal = ((AcmeFloat) test).getValue();
        } else if (test instanceof AcmeInt) {
            testFloatVal = (float) ((AcmeInt) test).getValue();
        } else {
            return false;
        };

        try {
            return (testFloatVal == this.evaluateAsFloat());
        } catch (Exception exc) {
        };

    }
        try {
            testFloatVal = ((ExpressionNode) test).evaluateAsFloat();
            return (testFloatVal == this.evaluateAsFloat());
        } catch (Exception exc) {
        };
        return false;
  };



  //
  // Methods added by bmonroe to implement numeric expression typechecking
  // and evaluation
  //


/** 
 * Traverses the expression tree below this node to compute the type
 * of the expression.  It sets the type for this ExpressionNode and
 * determines whether the claimed type of the node matches it's
 * subexpressions and operator.
 *
 * This method has the side-effect of setting the type of this node
 * and requesting that all of its children expression nodes do the
 * same.
 *
 * Returns true if it can infer the type of the expression and
 * the expression typechecks, else returns false.
 * @return true if it can infer the type of the expression and the expression typechecks
 */
  public boolean typecheck() {
    // first infer the type of all of the children nodes.  Recurse
    // until leaf nodes are found.  If it can't infer the type of
    // the children or any children don't typecheck then return false
    // to indicate a type error.
    for (int i = 0; i < this.jjtGetNumChildren(); i++) {
        if ( ((ExpressionNode) this.jjtGetChild(i)).typecheck() == false)
            return false;
    }

    // a NumericExpressionNode must have a numeric operator
    if (!(this.hasNumericOperator()))
        return false;

    // check for a unary expression.
    if (this.jjtGetNumChildren() == 1) {
        // set this node's type to the type of child node (0).
        // Unary operators can't change the type of the expression.

        // the child expression must be a numeric type (because this
        // is a NumericExpressionNode object.
        AcmeType childType = ((ExpressionNode) jjtGetChild(0)).getType();
        if ( !(childType instanceof FloatType)
             && !(childType instanceof IntType) )
        {
            return false;
        }

        this.setType( ((ExpressionNode) jjtGetChild(0)).getType());

        // can't prove that it does not typecheck, so assume that it does.
        return true;
    }

    AcmeType lChildType, rChildType;
    if (this.jjtGetNumChildren() == 2) {
        // then it is a binary expression, if both of the children
        // are integers, then it is an integer expression, otherwise
        // it is a floating point expression.
        lChildType = ((ExpressionNode) this.jjtGetChild(0)).getType();
        rChildType = ((ExpressionNode) this.jjtGetChild(1)).getType();

        if ((lChildType instanceof IntType) && (rChildType instanceof IntType))
        {
            this.setType(ConstantTypes.INT_TYPE);
            return true;
        }
        else if ( ( ((lChildType instanceof IntType) &&
                     (rChildType instanceof FloatType))
                  ||
                    ((lChildType instanceof FloatType) &&
                     (rChildType instanceof IntType))
                  ||
                    ((lChildType instanceof FloatType) &&
                     (rChildType instanceof FloatType))
                  )
                &&
                  (this.getOperator() != ExpressionOperatorT.MODULO_OP)
                )
        {
            this.setType(ConstantTypes.FLOAT_TYPE);
            return true;
        }
            // else this is an invalid type for a numeric node.
            this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
            return false;
    };

    // should never get to here.  return false for now.  probably throw
    // an exception when the exception handling is added.
    return false;
  }



/**
 * evaluate*() recursively evaluates the value of the left and right
 * children expressions, applies the operator to these values, and
 * returns that computed (numeric) value.
 * - evaluateAsFloat returns the numeric value of this expression as a float.
 * - evaluateAsInt returns the numeric value of this expression as an integer.
 * - evaluateAsString returns a string representation of this expression
 *    if the expression is a numeric value.
 *
 * @return The evaluated float value
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public float evaluateAsFloat() throws DesignRuleEvalException {
    // check if this is a unary "negation" operation
    // (negation is the only numeric unary operator)
    if ( (op == ExpressionOperatorT.UNARY_MINUS_OP) &&
         (jjtGetNumChildren() == 1))
    {
        if (jjtGetChild(0) instanceof NumericExpressionNode) {
            return -((NumericExpressionNode) jjtGetChild(0)).evaluateAsFloat();
        }
            String opMsg = new String("Attempted to apply the "
                             + "'negation' operator to an object that is "
                             + "not a NumericExpressionNode");
            throw new ExpressionTypeException(null, this,
                         ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);
    };

    // else this is an operation that takes two numeric arguments,
    // operates on them, and returns the float value of the evaluation.

    // evaluate the left and right expressions, casting to floats as
    // needed.  Then evaluate the value of this expression appropriately.
    // (support for strings to be added later)

    boolean lChildValid = false;   // flag true when lChild is valid float
    boolean rChildValid = false;   // flag true when rChild is valid float

    float lChildFloat = (float) 0.0;
    float rChildFloat = (float) 0.0;

    NumericExpressionNode lNumChild, rNumChild;
    ReferenceNode tmpRefNode;

    // find the value of the left child
    if (jjtGetChild(0) instanceof NumericExpressionNode) {
        lNumChild = (NumericExpressionNode) jjtGetChild(0);
        if (lNumChild.getType() instanceof IntType) {
            lChildFloat = (float) lNumChild.evaluateAsInt();
            lChildValid = true;
        } else if (lNumChild.getType() instanceof FloatType) {
            lChildFloat = lNumChild.evaluateAsFloat();
            lChildValid = true;
        }
    } else if (jjtGetChild(0) instanceof ReferenceNode) {
        tmpRefNode = (ReferenceNode) jjtGetChild(0);
        if (tmpRefNode.getType() instanceof IntType) {
            lChildFloat = (float) tmpRefNode.evaluateAsInt();
            lChildValid = true;
        }
        else if (tmpRefNode.getType() instanceof FloatType) {
            lChildFloat = tmpRefNode.evaluateAsFloat();
            lChildValid = true;
        }
    } else if (jjtGetChild(0) instanceof IntLiteralNode) {
        lChildFloat = (float) ((IntLiteralNode) jjtGetChild(0)).evaluateAsInt();
        lChildValid = true;
    } else if (jjtGetChild(0) instanceof FloatLiteralNode) {
        lChildFloat = ((FloatLiteralNode) jjtGetChild(0)).evaluateAsFloat();
        lChildValid = true;
    };

    // find the value of the right child
    if (jjtGetChild(1) instanceof NumericExpressionNode) {
        rNumChild = (NumericExpressionNode) jjtGetChild(1);
        if (rNumChild.getType() instanceof IntType) {
            rChildFloat = (float) rNumChild.evaluateAsInt();
            rChildValid = true;
        } else if (rNumChild.getType() instanceof FloatType) {
            rChildFloat = rNumChild.evaluateAsFloat();
            rChildValid = true;
        }
    } else if (jjtGetChild(1) instanceof ReferenceNode) {
        tmpRefNode = (ReferenceNode) jjtGetChild(1);
        if (tmpRefNode.getType() instanceof IntType) {
            rChildFloat = (float) tmpRefNode.evaluateAsInt();
            rChildValid = true;
        }
        else if (tmpRefNode.getType() instanceof FloatType) {
            rChildFloat = tmpRefNode.evaluateAsFloat();
            rChildValid = true;
        }
    } else if (jjtGetChild(1) instanceof IntLiteralNode) {
        rChildFloat = (float) ((IntLiteralNode) jjtGetChild(1)).evaluateAsInt();
        rChildValid = true;
    } else if (jjtGetChild(0) instanceof FloatLiteralNode) {
        rChildFloat = ((FloatLiteralNode) jjtGetChild(1)).evaluateAsFloat();
        rChildValid = true;
    };


    // sanity check that we've acquired values for at least two of
    // the children.
    if (!(lChildValid && rChildValid)) {
        String msg = new String("Illegal operands for numeric expression");
        throw new ExpressionTypeException(null, this,
            ConstantTypes.FLOAT_TYPE, this.getType(), msg);
    };

    // perform the operation on the values of the children.
    if (op == ExpressionOperatorT.PLUS_OP)
        return lChildFloat + rChildFloat;
    else if (op == ExpressionOperatorT.MINUS_OP)
        return lChildFloat - rChildFloat;
    else if (op == ExpressionOperatorT.MULT_OP)
        return lChildFloat * rChildFloat;
    else if (op == ExpressionOperatorT.DIVIDE_OP)
        return lChildFloat / rChildFloat;
    else if (op == ExpressionOperatorT.MODULO_OP)
        return (float) -123456.789;  // throw a type mismatch exception

    // if it gets to here then throw an exception because the expression
    // does not have a valid numeric operator
    String opMsg = new String("Illegal operator "
                    + this.getOperator().toString()
                    + " for a numeric expression.");
    throw new ExpressionTypeException(null, this,
                    ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);

  };




/**
 * @return The evaluated integer value
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public int evaluateAsInt() throws DesignRuleEvalException {
    // check if this is a unary "negation" operation
    // (negation is the only numeric unary operator)
    if ( (op == ExpressionOperatorT.UNARY_MINUS_OP) &&
         (jjtGetNumChildren() == 1))
    {
        if (jjtGetChild(0) instanceof NumericExpressionNode) {
            return -((NumericExpressionNode) jjtGetChild(0)).evaluateAsInt();
        }
            String opMsg = new String("Attempted to apply the "
                             + "'negation' operator to an object that is "
                             + "not a NumericExpressionNode");
            throw new ExpressionTypeException(null, this,
                         ConstantTypes.INT_TYPE, this.getType(), opMsg);

    };

    // else this is an operation that takes two numerical or string arguments,
    // operates on them, and returns the integer value of the evaluation.

    // evaluate the left and right expressions, determine whether they
    // are stored as integers or floats, and then evaluate the value of
    // this node appropriately.  (support for strings to be added later)
    boolean lChildIsInt = false;
    boolean lChildIsFloat = false;
    boolean rChildIsInt = false;
    boolean rChildIsFloat = false;

    int lChildInt = 0;
    int rChildInt = 0;
    float lChildFloat = (float) 0.0;
    float rChildFloat = (float) 0.0;

    NumericExpressionNode lNumChild, rNumChild;
    ReferenceNode tmpRefNode;

    // find the value of the left child
    if (jjtGetChild(0) instanceof NumericExpressionNode) {
        lNumChild = (NumericExpressionNode) jjtGetChild(0);
        if (lNumChild.getType() instanceof IntType) {
            lChildIsInt = true;
            lChildInt = lNumChild.evaluateAsInt();
        } else if (lNumChild.getType() instanceof FloatType) {
            lChildIsFloat = true;
            lChildFloat = lNumChild.evaluateAsFloat();
        }
    } else if (jjtGetChild(0) instanceof ReferenceNode) {
        tmpRefNode = (ReferenceNode) jjtGetChild(0);
        if (tmpRefNode.getType() instanceof IntType) {
            lChildIsInt = true;
            lChildInt = tmpRefNode.evaluateAsInt();
        }
        else if (tmpRefNode.getType() instanceof FloatType) {
            lChildIsFloat = true;
            lChildFloat = tmpRefNode.evaluateAsFloat();
        };
    } else if (jjtGetChild(0) instanceof IntLiteralNode) {
        lChildIsInt = true;
        lChildInt = ((IntLiteralNode) jjtGetChild(0)).evaluateAsInt();
    } else if (jjtGetChild(0) instanceof FloatLiteralNode) {
        lChildIsFloat = true;
        lChildFloat = ((FloatLiteralNode) jjtGetChild(0)).evaluateAsFloat();
    };

    // find the value of the right child
    if (jjtGetChild(1) instanceof NumericExpressionNode) {
        rNumChild = (NumericExpressionNode) jjtGetChild(1);
        if (rNumChild.getType() instanceof IntType) {
            rChildIsInt = true;
            rChildInt = rNumChild.evaluateAsInt();
        } else if (rNumChild.getType() instanceof FloatType) {
            rChildIsFloat = true;
            rChildFloat = rNumChild.evaluateAsFloat();
        }
    } else if (jjtGetChild(1) instanceof ReferenceNode) {
        tmpRefNode = (ReferenceNode) jjtGetChild(1);
        if (tmpRefNode.getType() instanceof IntType)
        {
            rChildIsInt = true;
            rChildInt = tmpRefNode.evaluateAsInt();
        }
        else if (tmpRefNode.getType() instanceof FloatType) {
            rChildIsFloat = true;
            rChildFloat = tmpRefNode.evaluateAsFloat();
        };
    };

    // sanity check that we've acquired values for at least two of
    // the children.
    if (!(lChildIsFloat || lChildIsInt) || !(rChildIsFloat || rChildIsInt)) {
        String msg = new String("Illegal operands for numeric integer expression");
        throw new ExpressionTypeException(null, this,
            ConstantTypes.FLOAT_TYPE, this.getType(), msg);
    };

    // perform the operation on the values of the children.
    if (op == ExpressionOperatorT.PLUS_OP) {
        if (lChildIsInt && rChildIsInt) {
            return lChildInt + rChildInt;
        } else if (lChildIsFloat && rChildIsFloat) {
            return (int) (lChildFloat + rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
            return (int) (lChildInt + rChildFloat);
        } else if (lChildIsFloat && rChildIsInt) {
            return (int) (lChildFloat + rChildInt);
        };
    } else if (op == ExpressionOperatorT.MINUS_OP) {
        if (lChildIsInt && rChildIsInt) {
            return lChildInt - rChildInt;
        } else if (lChildIsFloat && rChildIsFloat) {
            return (int) (lChildFloat - rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
            return (int) (lChildInt - rChildFloat);
        } else if (lChildIsFloat && rChildIsInt) {
            return (int) (lChildFloat - rChildInt);
        };
    } else if (op == ExpressionOperatorT.MULT_OP) {
        if (lChildIsInt && rChildIsInt) {
            return lChildInt * rChildInt;
        } else if (lChildIsFloat && rChildIsFloat) {
            return (int) (lChildFloat * rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
            return (int) (lChildInt * rChildFloat);
        } else if (lChildIsFloat && rChildIsInt) {
            return (int) (lChildFloat * rChildInt);
        };
    } else if (op == ExpressionOperatorT.DIVIDE_OP) {
        if (lChildIsInt && rChildIsInt) {
            return lChildInt / rChildInt;
        } else if (lChildIsFloat && rChildIsFloat) {
            return (int) (lChildFloat / rChildFloat);
        } else if (lChildIsInt && rChildIsFloat) {
            return (int) (lChildInt / rChildFloat);
        } else if (lChildIsFloat && rChildIsInt) {
            return (int) (lChildFloat / rChildInt);
        };
    } else if (op == ExpressionOperatorT.MODULO_OP) {
        if (lChildIsInt && rChildIsInt) {
            return (lChildInt % rChildInt);
        }
            // throw an exception.  MODULO_OP requires that both
            // children are integers.
            String opMsg = new String("Illegal operands.  Both operands "
                    + "must be integers in a valid modulo expression.");
            throw new ExpressionTypeException(null, this,
                    ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);

    }

    // if it gets to here then throw an exception because the expression
    // does not have a valid integer operator
    String opMsg = new String("Illegal operator "
                    + this.getOperator().toString()
                    + " for an integer expression.");
    throw new ExpressionTypeException(null, this,
                    ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);
  };


/**
 * @return The evaluated string
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public String evaluateAsString() throws DesignRuleEvalException {
    if (this.getType() instanceof IntType)
        return String.valueOf(this.evaluateAsInt());
    else if (this.getType() instanceof FloatType)
        return String.valueOf(this.evaluateAsFloat());
    else
        return "<unable to evaluate node as a string>";
  };

}
