package acmeLib.designRules;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import acmeLib.AcmeType;
import acmeLib.AcmeTypeParameter;
import acmeLib.AcmeUnresolvedType;
import acmeLib.Debug;
import acmeLib.armaniParser.ArmaniParser;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.designRules.feedback.ElementEvaluationFeedback;
import acmeLib.elements.*;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.errorManagement.TypeRefResolutionError;
import acmeLib.properties.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: QuantifiedExpressionNode<br>
 * Purpose: Stores the representation of a quantified expression in the Armani<br>
 *          design rule/constraint language.  A QuantifiedExpression in Armani<br>
 *          is a boolean-typed construct.  This class can store both universal<br>
 *          and existential quantifications.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    1/8/98     RTM       Created<br>
 *    2/25/98    RTM       Adapted class to store only a single variable<br>
 *                          and a single set over which it was quantified.<br>
 *    1/06/03    SWC       Added debugging in evaluateAsBoolean() for universal<br>
 *                          and existential quantifications, and error output.<br>
 *                          Source was also reformatted in Eclipse.<br>
 *<br>
 * @author Bob Monroe
 */
public class QuantifiedExpressionNode extends ExpressionNode implements Scope
{

    //
    // Protected members
    //

  /**
   * Flag that indicates what type of quantification operation to perform.
   * (default to Universal)
   */
    protected int quantOp = UNIVERSAL_QUANTIFICATION;

  /**
   * Set over which bound variables are quantified.  A setReference will
   * not be initially resolved to a SetExpressionNode by the parser because
   * it looks to the parser like any other reference.  As a result, the
   * actual set that an identifier or method is referring to may have
   * to be resolved during the resolveInstanceReferences stage (which
   * will be taken care of by the ExpressionNode superclass method).  The
   * getSetExpression() method will then convert the generic reference
   * node appropriately if it can, and cache the result in cachedSetExpr.
   * cachedSetExpr is initially null. cachedExprDirty flags whether the
   * set expression has been changed, thus (potentially) invalidating
   * the cached set expression.
   */
    protected ExpressionNode setExpr = null;
    protected SetExpressionNode cachedSetExpr = null;
    boolean cachedSetExprDirty = true;

  /**
   * Expression over which quantified variables are evaluated.
   * This expression must be evaluable as a boolean, but it isn't
   * stored as a BooleanExpressionNode so that individual identifiers
   * and functions with boolean return values can be used without an
   * annoying check that the result is true (ala C).
   * E.g. forall c in {self.components} | c.typechecks;
   *  instead of forall c in {self.components} | c.typechecks == true;
   */
    protected ExpressionNode evalExpr;

  /**
   * Store the name of the quantified variable
   */
    protected String varName = null;

  /**
   * Stores the type of the quantified variable
   */
    protected AcmeType varType = ConstantTypes.UNSPECIFIED_TYPE;
    
    protected VolatileVariable varTypeAsVolatile = null;
    
    /**
     * Flags whether to use satisfiesType or declaresType in the
     * set expression,
     */
    protected boolean checkSatisfies = true;

  /**
   * true if the variable's type was inferred, and therefore should
   * not be exported.
   */
    protected boolean varTypeInferred = false;

  /**
   * Stores the current value of the referenced variable
   */
    protected Object curVarValue = null;

  /**
   * Reference to the scope in which this expression lives.
   */
    protected Scope parentScope = null;


    //
    // public members
    //

  /**
   * constants for determining this object's quantification operation.
   */
    public static final int UNIVERSAL_QUANTIFICATION = 1;
    public static final int EXISTENTIAL_QUANTIFICATION = 2;
    public static final int UNIQUE_EXISTENTIAL_QUANTIFICATION = 3;

    public QuantifiedExpressionNode (QuantifiedExpressionNode copyOf) {
        super (copyOf.id);
        this.setType (copyOf.getType ());
        this.setQuantOp(copyOf.getQuantOp());
        this.setExpr = copyOf.setExpr;
    }

  /**
   * This is the constructor of this class
   */
    public QuantifiedExpressionNode() {
        super(ArmaniParserTreeConstants.JJTQUANTIFIEDEXPRESSIONNODE);

        // the type of a base QuantififiedExpressionNode is boolean.
        // subclasses can override this as necessary.
        this.setType(ConstantTypes.BOOLEAN_TYPE);
    }

  /**
   * This is the constructor of this class
   * @param id The ID
   */
    public QuantifiedExpressionNode(int id) {
        super(id);

        // the type of a base QuantififiedExpressionNode is boolean.
        // subclasses can override this as necessary.
        this.setType(ConstantTypes.BOOLEAN_TYPE);
    }

  /**
   * This is the constructor of this class
   * @param id The ID
   * @param newEvalExpr The new eval expression
   */
    public QuantifiedExpressionNode(int id, ExpressionNode newEvalExpr) {
        super(id);
        this.evalExpr = newEvalExpr;

        // the type of a base QuantififiedExpressionNode is boolean.
        // subclasses can override this as necessary.
        this.setType(ConstantTypes.BOOLEAN_TYPE);
    }

  /**
   * JJTree-specific methods
   * @param id The ID
   * @return The node
   */
    public static Node jjtCreate(int id) {
        return new QuantifiedExpressionNode(id);
    }


    // accessor/settor methods
  /**
   * Get the quantified operator
   * @return the quantified operator
   */
    public int getQuantOp() { return this.quantOp; }
  /**
   * Set the quantified operator
   * @param newQuanOp The new quantified operator
   */
    public void setQuantOp(int newQuantOp) {
        this.quantOp = newQuantOp;
    }

  /**
   * Get the variable name
   * @return the variable name
   */
    public String getVarName() { return this.varName; }
  /**
   * Set the variable name
   * @param newVarName The new variable name
   */
    public void setVarName(String newVarName) { this.varName = newVarName; }

  /**
   * Get the variable type
   * @return the variable type
   */
    public AcmeType getVarType() { 
        if (this.varType != ConstantTypes.UNSPECIFIED_TYPE) {
            return varType;
        }
//        else if (this.varTypeAsVolatile != null) {
//            VolatileVariable v = getParentScope ().lookupName ();
//            return v;
//        }
        
        return this.varType; 
    }
  /**
   * Set the variable type
   * @param newVarTyp The new variable type
   */
   public void setVarType(AcmeType newVarType) { 

       this.varType = newVarType; 
   }


    public void setVarType (VolatileVariable newVarType) {
        this.varTypeAsVolatile = newVarType;
    }
  /**
   * Get the varTypeInferred boolean value
   * @return the varTypeInferred boolean value
   */
    public boolean getVarTypeInferred() { return this.varTypeInferred; }
  /**
   * Set the varTypeInferred boolean value
   * @param isInferred The new varTypeInferred boolean value
   */
    public void setVarTypeInferred(boolean isInferred) {
        this.varTypeInferred = isInferred;
    };

  /**
   * Get the current variable value
   * @return the current variable value
   */
    public Object getCurVarValue() { return this.curVarValue; }
  /**
   * Set the current variable value
   * @param newVarValue The new current variable value
   */
    public void setCurVarValue(Object newVarValue) {
        this.curVarValue = newVarValue;
    }

  /**
   * Get the eval expression node
   * @return the eval expression node
   */
    public ExpressionNode getEvalExpression() { return this.evalExpr; };
  /**
   * Set the eval expression node
   * @param expr The new eval expression node
   */
    public void setEvalExpression(ExpressionNode expr) {
        this.evalExpr = expr;
    }


  /**
   * Get the parent scope
   * @return the parent scope
   */
    public Scope getParentScope() { return this.parentScope; };
  /**
   * Set the parent scope
   * @param newScope The new parent scope
   */
    public void setParentScope(Scope newScope) { this.parentScope = newScope; };


  /**
   * the settor/accessor methods for dealing with this node's set expression
   * pull the appropriate SetExpressionNode from the expression that was
   * passed as an argument, and attempt to extract the SetExpressionNode
   * referred to on a lazy basis (that is, only when it is requested, which
   * should be after all references have been properly resolved).
   * @param expr The expression node
   */
    public void setSetExpression(ExpressionNode expr) {
        this.setExpr = expr;
        this.cachedSetExpr = null;
        this.cachedSetExprDirty = true;
    };
  /**
   * @return The expression node
   */
    public SetExpressionNode getSetExpression() {
    	// [SWC 02-July-2004] determine whether to use cached set or re-evaluate
    	if (setExpr instanceof IdentifierNode) {
    		Object refObj = ((IdentifierNode) setExpr).getReferentObject();
    		if (refObj instanceof AcmeProperty) {
    			// for the case of an AcmeProperty, consider cache dirty if
    			// the value object and the cached set are not the same object
    			if (cachedSetExpr != ((AcmeProperty )refObj).getValue()) {
    				cachedSetExprDirty = true;
    			}
    		}
    	}
        if ((this.cachedSetExprDirty == false) && (this.cachedSetExpr != null)) {
            return this.cachedSetExpr;
        }
            try {
                this.cachedSetExpr = setExpr.evaluateAsSet();
                this.cachedSetExprDirty = false;
                return this.cachedSetExpr;
            } catch (DesignRuleEvalException drExc) {
                this.cachedSetExpr = null;
                this.cachedSetExprDirty = true;
        };
        return null;
    };



  /**
   * Returns true if the variable name passed as a parameter is the variable
   * quantified, else returns false.
   * @param variableName The variable name
   * @return true if it is volatile variable
   */
    public boolean isVolatileVariable(String variableName) {
        if (variableName == null)
            return false;
        else
            return variableName.equalsIgnoreCase(varName);
    };


  /**
   * typecheck recursively infers the type of its evaluation expression
   * and checks that all of the members of its set expression satisfy the
   * type declared for the quantified variable.  It then confirms that the
   * evaluation expression is a boolean expression.
   * returns true if all of these conditions are satisfied, else returns false.
   * @return true if type is correct
   */
    public boolean typecheck() {
        boolean evalExprTypechecks, setExprTypechecks, varMatchesSetType;
        try {
        	// Make sure that the type of the quantified
        	// variable is resolved
			this.setExpr.resolveTypeReferences();
		}
		catch (AcmeResolutionException e1) {
			// THis should be picked up later on
		}
        // first have the setExpr node resolve its type, before asking
        // for the full expansion of the set.
        setExprTypechecks = this.setExpr.typecheck();
        if (!setExprTypechecks) {
            return false;
        };

        // get the resolved set expression, if it's available.  If we don't
        // have a valid SetExpressionNode to quantify over then return false.
        SetExpressionNode resolvedSetExpr = this.getSetExpression();
        if (resolvedSetExpr == null) {
            ExpressionTypecheckError err = 
                new ExpressionTypecheckError (setExpr, "Could not resolve to set", this);
            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
            return false;
        };

        // check to see that the Set Expression typechecks, infer the
        // type of the set expression in the process, and then check that
        // the set expression's type is a subtype of the variable's type.
        if (resolvedSetExpr instanceof ExpressionNode) {
            setExprTypechecks = ((ExpressionNode) resolvedSetExpr).typecheck();
        } else if (resolvedSetExpr instanceof AcmeProperty) {
//            setExprTypechecks = AcmeTypeManager.typecheckProperty(
//                                            (AcmeProperty) resolvedSetExpr);
            setExprTypechecks = true;
        }

        // if the type of the quantified variable is unspecified, then
        // "infer" it to be the type of the set contents and flag the type
        // as inferred.
        if (this.getVarType() == ConstantTypes.UNSPECIFIED_TYPE) {
            if (resolvedSetExpr.getSetType() != null) {
                this.setVarType(resolvedSetExpr.getSetType());
                this.setVarTypeInferred(true);
            };
            varMatchesSetType = true;
        }
        /* BRS: Need some way to get at types that are passed in as params */
        else if (this.getVarType () == ConstantTypes.TYPE_REFERENCE_TYPE) {
            varMatchesSetType = true;        
        
        } else {
            // else the variable has a type that was explicitly declared.
            // make sure that there is a sub/super type relationship between
            // the declared type of the variable and the type of the set.
            // if they are fundamentally mismatched (e.g. vartype is a
            // property type and the set contains elements) then declare
            // a typechecking error.
            if (this.getVarType() instanceof AnyType) {
                varMatchesSetType = true;
            } else if (
                (resolvedSetExpr.getSetType() instanceof AcmePropertyType)
                    && (this.getVarType() instanceof AcmePropertyType)) {
                varMatchesSetType =
                    (AcmeTypeManager
                        .isLogicalSubtype(resolvedSetExpr.getSetType(), this.getVarType(), true)
                        || AcmeTypeManager.isLogicalSubtype(
                            this.getVarType(),
                            resolvedSetExpr.getSetType(), true));
            }
            /** BRS: Added type checking for representation set **/
            else if (
                (resolvedSetExpr.getSetType() instanceof AcmeFamily)
                    && (this.getVarType() instanceof AcmeFamily)) {
                varMatchesSetType = true;
            } else if (
                (resolvedSetExpr.getSetType() instanceof AcmeElementType)
                    && (this.getVarType() instanceof AcmeElementType)) {
                varMatchesSetType =
                    (AcmeTypeManager
                        .isDeclaredSubtype(
                            (AcmeElementType) this.getVarType(),
                            (AcmeElementType) resolvedSetExpr.getSetType())
                        || AcmeTypeManager.isDeclaredSubtype(
                            (AcmeElementType) resolvedSetExpr.getSetType(),
                            (AcmeElementType) this.getVarType())
                        || (this.getVarType() instanceof AcmeElementProtoType)
                        || (resolvedSetExpr.getSetType() instanceof AcmeElementProtoType));
            /** SWC 01/03/03: Added type checking for element TYPE **/
            /** This is not right, because we should resolve this to
             *  do proper type checking!
             */
            } else if (
                (resolvedSetExpr.getSetType() instanceof AcmeElementType)
                    && (this.getVarType() instanceof AcmeTypeParameter)) {
                Object resolvedType = lookupName (this.getVarType ().getTypeName(), true);
                if (resolvedType instanceof VolatileVariable) {
                    VolatileVariable v = (VolatileVariable )resolvedType;
                    if (v.referentObject instanceof IdentifierNode) {
                        try {
                        
                            AcmeType type = ((IdentifierNode )v.referentObject).evaluateAsTypeRef();
                            varMatchesSetType = AcmeTypeManager.isLogicalSubtype(type, resolvedSetExpr.getSetType (), true);
                            this.varType = type;
                       }
                       catch (DesignRuleEvalException e) {
                           e.printStackTrace ();
                           varMatchesSetType = false;
                       }
                    }
                    else {
                        varMatchesSetType = true;
                    }
                }
                else {
                    varMatchesSetType = true;
                }
                
            } else if (resolvedSetExpr.size () == 0) {
            		varMatchesSetType = true;
            }
            else {
                // else there is no recognizable sub/supertype relationship
                // between the quantified variable type and the set contents,
                // so flag the mismatch.
                varMatchesSetType = false;
            };

        }; // end if type is not inferred.

        if (!varMatchesSetType) {            
            String type = resolvedSetExpr.getSetType().getTypeName ();
            
            ExpressionTypecheckError err = 
                new ExpressionTypecheckError (this, "\"" + varName + "\" does not match the set type: " + type, this);
            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
        }

        // RTM 4/26/98 -- the evaluation expression has to wait to resolve
        //                its references until the type of the quantified
        //                variable has been inferred.
        
        try {
            evalExpr.resolveTypeReferences();
            evalExpr.resolveInstanceReferences();
        }
        catch (AcmeResolutionException e) {
            AcmeTypeManager.typecheckingErrorStack.pushCausalError(e.getError ());
            return false;
        }
        evalExprTypechecks = evalExpr.typecheck();
        if (!evalExprTypechecks) {
            evalExprTypechecks &= (evalExpr.getType() instanceof BooleanType);
            ExpressionTypecheckError err = new ExpressionTypecheckError (evalExpr, "Not a boolean expression", this);
            AcmeTypeManager.typecheckingErrorStack.pushCausalError (err);
            
        };

        // this only typechecks if all of the checks returned true.
        return evalExprTypechecks && setExprTypechecks && varMatchesSetType;
    };


    

  /**
   * Returns true iff the evaluation expression is true for all values v
   * (universal quantification), or returns true iff the evaluation
   * expression is true for at least one value of v (existential
   * quantification) where v is quantified over setExpr.
   * @return The evaluated boolean value
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public boolean evaluateAsBoolean() throws DesignRuleEvalException {
        if (parentScope instanceof DesignAnalysis) {
            // The set expression could be passed in as a parameter
            // in which case the set should be refreshed, and 
            // the var type also, if it is inferred
            cachedSetExprDirty = true; 
            if (this.varTypeInferred) {
            
                setVarType (ConstantTypes.UNSPECIFIED_TYPE);
                this.typecheck ();
            }
        }
        
        // first check that we have non-empty set and evaluation expressions.
        if ((this.getSetExpression() == null) ||
            (this.getEvalExpression() == null))
        {
            // the expression is nonsensical
            return false;
        }

        // refresh the set over which we are quantifying.
        this.getSetExpression().refreshSet();

      	Debug.trace_call(this);

      	// make sure that the expression is true for _all_ values of the
        // set if this is a universal quantification.  conversely, make sure
        // that the expression is true for at least one of the values of
        // the set if this is an existential quantification.
        AcmeEnumeration e = this.getSetExpression().enumerate();
        boolean retVal;

        if (this.getQuantOp() == UNIVERSAL_QUANTIFICATION) {
            retVal = true;
            while(e.moreItems() && (retVal == true)) {
                // set the cur var value to the appropriate entity.
                // if it is a literal just set the literal value.
                // if, otoh, it is a ReferenceNode then set the value
                // to the referenceNode's referent object.
                Object curValObj = e.getNextItem();
                if (curValObj instanceof ReferenceNode) {
                    curValObj = ((ReferenceNode) curValObj).getReferentObject();
                };

                boolean curValSatisfiesType = isTypeCompliant(curValObj);


                if (curValSatisfiesType)
                {
                    this.setCurVarValue(curValObj);
                    Debug.trace(curValObj);

                    try {
                        retVal = this.getEvalExpression().evaluateAsBoolean();
                      
                        if (DesignRule.isRememberVisited()) {
                        	List l = (List )getUserData (getEvaluationKey());
                        	if (l == null) {
                        		l = new ArrayList ();
                        		setUserData(getEvaluationKey(), l);
                        	}
                        	l.add(new ElementEvaluationFeedback (curValObj, new Boolean (retVal)));
                        }
                        //if(this.getEvalExpression().evaluateAsBoolean() == false)
                            //retVal = false;
                            // RTM 2/25/98 -- add more detailed error flagging for
                            //                false evaluations here.  (along the lines
                            //                of the typechecking error stack).
                    } catch (DesignRuleEvalException drExc) {
                        // zensoul.DEBUG
                        //System.err.println("BAD: " + this.getEvalExpression().toString());
                        //System.err.println("     " + drExc.getMessage());
                        //drExc.printStackTrace(System.err);

                        retVal = false;
                        // RTM 2/28/98 -- internal error or type error?
                        //                either way it blows the expression.
                    };
                };
            };
        }
        else if (this.getQuantOp() == EXISTENTIAL_QUANTIFICATION) {
            // else this is an existential quantification, check that the
            // eval expression holds for _at_least_one_ value of the set.
            retVal = false;
            while(e.moreItems() && (retVal == false)) {
                // set the cur var value to the appropriate entity.
                // if it is a literal just set the literal value.
                // if, otoh, it is a ReferenceNode then set the value
                // to the referenceNode's referent object.
                Object curValObj = e.getNextItem();
                if (curValObj instanceof ReferenceNode) {
                    curValObj = ((ReferenceNode) curValObj).getReferentObject();
                };
                
                
                if (isTypeCompliant(curValObj)) {
                    this.setCurVarValue(curValObj);
                    Debug.trace(curValObj);

                    try {
                        retVal = this.getEvalExpression().evaluateAsBoolean();
                        if (DesignRule.isRememberVisited()) {
                        	List l = (List )getUserData (getEvaluationKey());
                        	if (l == null) {
                        		l = new ArrayList ();
                        		setUserData(getEvaluationKey(), l);
                        	}
                        	l.add(new ElementEvaluationFeedback (curValObj, Boolean.valueOf (retVal)));
                        }
                        //if(this.getEvalExpression().evaluateAsBoolean() == true)
                        //retVal = true;
                        // RTM 2/25/98 -- add more detailed error flagging for
                        //                false evaluations here.  (along the lines
                        //                of the typechecking error stack).
                    }
                    catch (DesignRuleEvalException drExc) {
                        // zensoul.DEBUG
                        //System.err.println("BAD: " + this.getEvalExpression().toString());
                        //System.err.println("     " + drExc.getMessage());
                        //drExc.printStackTrace(System.err);

                        // invalid expression eval, don't change retVal
                        // RTM 2/28/98 -- internal error or type error?
                    };
                }
            };
        } 
        else if (this.getQuantOp () == UNIQUE_EXISTENTIAL_QUANTIFICATION) {
            // else this is a unique existential quantification, check that
            // the eval expression holds for _exactly_one value of the set
            retVal = false;
            int numTrue = 0;
            while (e.moreItems ()) {
                // set the cur var value to the appropriate entity.
                // if it is a literal just set the literal value.
                // if, otoh, it is a ReferenceNode then set the value
                // to the referenceNode's referent object.
                Object curValObj = e.getNextItem ();
                if (curValObj instanceof ReferenceNode) {
                    curValObj = ((ReferenceNode )curValObj).getReferentObject();   
                };
                
                if (isTypeCompliant(curValObj)) {
                    this.setCurVarValue(curValObj);
                    Debug.trace(curValObj);

                    try {
                        retVal = this.getEvalExpression().evaluateAsBoolean();
                        if (retVal) {
                            numTrue++;
                        }
                        if (DesignRule.isRememberVisited()) {
                        	List l = (List )getUserData (getEvaluationKey());
                        	if (l == null) {
                        		l = new ArrayList ();
                        		setUserData(getEvaluationKey(), l);
                        	}
                        	l.add(new ElementEvaluationFeedback (curValObj, Boolean.valueOf(retVal)));
                        }
                    }
                    catch (DesignRuleEvalException drExec) {
                        //System.err.println("BAD: " + this.getEvalExpression().toString());
                        //System.err.println("     " + drExec.getMessage());
                    }
                }
            }   
            // The retVal should only be true if the expression was true
            // for exactly one element
            retVal = numTrue == 1;
        }
        
        
        else {
            throw new ExpressionCastException(this, this,
                    ConstantTypes.BOOLEAN_TYPE, this.getType(),
                    "Unrecognized quantification operator found in "
                    + "quantified expression: "+ this.toString());
        };

      	Debug.trace_return(this, retVal);
        return retVal;
    }

    public boolean isTypeCompliant(Object curValObj) {
        // if there is a variable type specified use the varType
        // specification as a filter.  If the curValObj satisfies
        // varType then evaluate it, otherwise ignore it and go to
        // the next item in the set.
        boolean curValSatisfiesType = false;
        boolean previousLoggingStatus = AcmeTypeManager.errorReportingIsEnabled();
        
        try {
            // turn error logging off temporarily so any error messages
            // generated by this typecheck will not be permanently
            // recorded.  Be sure to set it back as it was before
            // proceeding.
            AcmeTypeManager.disableErrorReporting();
            
            // BRS: Work out whether the current value is valid (i.e.,
            // whether it satisifies or declares the type (depending
            // on the form of the expression)
            curValSatisfiesType = false;
            // Commented out, because satisfiesType can cause infinite looping
            /*if (isSatisfies()) {
            	if (checkSatisfies) {
            		checkSatisfies = false;
            		curValSatisfiesType = AcmeTypeManager.satisfiesType(this.getVarType(), curValObj);
            		checkSatisfies = true;
            	}            		
            }
            else {*/
            if (getVarType () instanceof AcmeTypeParameter && curValObj instanceof AcmeType) {
            	curValSatisfiesType = true;
            }
            else {
                curValSatisfiesType = AcmeTypeManager.declaresType (this.getVarType (), curValObj);
            }
            if (previousLoggingStatus == true) {
                AcmeTypeManager.enableErrorReporting();
            } else {
                AcmeTypeManager.disableErrorReporting();
            };
        }
        catch (Throwable t) {
            // zensoul.DEBUG
            System.err.println(this.toString() + " NOT SATISFY TYPE due to " +
                               curValObj.toString());
            // whatever is thrown, just figure that curVal doesn't
            // satisfy the type and reset the error logging
            curValSatisfiesType = false;
            if (previousLoggingStatus == true) {
                AcmeTypeManager.enableErrorReporting();
            } else {
                AcmeTypeManager.disableErrorReporting();
            };
        }
        finally {
            // be extra sure...
            if (previousLoggingStatus == true) {
                AcmeTypeManager.enableErrorReporting();
            } else {
                AcmeTypeManager.disableErrorReporting();
            };
            checkSatisfies = true;
        };
        return curValSatisfiesType;
    };


  /**
   * helper method to convert the quantOp to a string for exporting, etc.
   * potentially also useful to the world at large, so it's public.
   * @return The string for the quantOp
   */
    public String getQuantOpAsString() {
        if (this.getQuantOp() == UNIVERSAL_QUANTIFICATION) {
            return "Forall";
        } else if (this.getQuantOp() == EXISTENTIAL_QUANTIFICATION) {
            return "Exists";
        }
        else if (this.getQuantOp () == UNIQUE_EXISTENTIAL_QUANTIFICATION) {
            return "Exists Unique";
        }
        return "<unrecognized quantification operator: " + this.getQuantOp();
    };

  /**
   * extensions to dump more interesting expression output messages.
   * @return The output string
   */
    public String toString() {
        String ret = "";
        try {
            AcmeOutputStream s = new AcmeOutputStream ();
            exportArmani (s);
            s.flush ();
            ret = s.toString ();
            s.close ();
        }
        catch (IOException e) {}
        catch (AcmeException e) {}
    
        return ret;
    };

  /**
   * export method
   * @param prefix The prefix
   * @return The output string
   */
    public String toString(String prefix) { return prefix + toString(); }




  /**
   * resolve the instances of the node's set expression.
   *
   * RTM 4/26/98 -- do _not_ call resolveInstanceReferences on
   *                the evaluation expression, even though it is
   *                a child node.  resolveInstanceReferences will
   *                be called on the expression node by the
   *                typecheck() method, after it has properly inferred
   *                the type of the set and quantified variable.
   *                This is a bit of a crufty hack to allow quantified
   *                expressions to have an inferred type for the
   *                quantified variable.
   */
    public void resolveInstanceReferences() throws AcmeResolutionException {
        SetExpressionNode curSetExpr = this.getSetExpression();
        if (curSetExpr instanceof ExpressionNode) {
            ((ExpressionNode) curSetExpr).resolveInstanceReferences();
        };
    };


  /**
   * resolve the typename of the variable's type, if necessary
   */
    public void resolveTypeReferences() throws AcmeResolutionException {
        // resolve the type of the quantified variable.
        String errorMsg = null;
        AcmeType foundType = null;
        VolatileVariable v = null;
        Object foundObj = null;
        if ((this.getVarType() instanceof AcmeUnresolvedType)
            || (this.getVarType() instanceof AcmeUnresolvedPropertyType)
            || (this.getVarType() instanceof AcmeUnresolvedElementType)) {
            if (this.getParentScope() == null) {
                // Error. Announce a type resolution error.
                errorMsg =
                    "Quantified Expression: "
                        + this.toString()
                        + " does not have a valid parent scope for resolving"
                        + " type name: "
                        + ((AcmeType) this.getVarType()).getTypeName();
                TypeRefResolutionError badTypeRefError =
                    new TypeRefResolutionError(this, errorMsg, this.getVarType().getTypeName());
                badTypeRefError.setUnresolvedRef(this.getVarType());
                AcmeErrorManager.errorDetected(badTypeRefError);
				AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);


            } else {
                java.util.StringTokenizer types =
                    new java.util.StringTokenizer(this.getVarType().getTypeName(), ".");
                Scope curScope = this.getParentScope();
                String curName = types.nextToken();
//                 (01/03/03 zensoul) modified to allow VolatileVariable as lookupName result.
                Object namedObj = curScope.lookupName(curName, true);
                if (namedObj instanceof Scope) {
                    curScope = (Scope) namedObj;
                    if (curScope instanceof AcmeType) {
                        foundType = (AcmeType) namedObj;
                    } else {
                        foundType = null;
                    }
                } else if (namedObj instanceof VolatileVariable) {
                    VolatileVariable var = (VolatileVariable )namedObj;
                    v = var;
                    curScope = var.declarationScope;
                    foundType = new AcmeTypeParameter (curName);
                } else {
                    // SWC 01/03/03:  Not sure what to do in this case.
                }
//                 END (01/03/03 zensoul)
//                curScope = (Scope)curScope.lookupName(curName,true);
//		        if (curScope instanceof AcmeType){
//		            foundType = (AcmeType)curScope;
//	  	      	} else {
//	    	        foundType = null;
//	      	  	}
                while (types.hasMoreElements() && curScope != null) {
                    curName = types.nextToken();
                    curScope = (AcmeElement) curScope.lookupName(curName, false);
                    if (curScope instanceof AcmeType) {
                        foundType = (AcmeType) curScope;
                    } else {
                        foundType = null;
                    }
                }

                if (foundType != null) {
                    this.setVarType(foundType);
                    this.setVarType (v);
                } else {
                    // Error. Announce a type resolution error.
                    errorMsg =
                        "Variable "
                            + this.getVarName()
                            + " declares unresolvable type: "
                            + ((AcmeType) this.getVarType()).getTypeName()
                            + " in quantified expression: "
                            + this.toString();
                    TypeRefResolutionError badTypeRefError =
                        new TypeRefResolutionError(this, errorMsg, this.getVarType().getTypeName());
                    badTypeRefError.setUnresolvedRef(this.getVarType());
                    AcmeErrorManager.errorDetected(badTypeRefError);
					AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

                }
                /*String varTypeName = this.getVarType().getTypeName();
                foundObj = this.getParentScope().lookupName(varTypeName, true);
                if (foundObj instanceof AcmeType) {
                    this.setVarType((AcmeType) foundObj);
                } else {
                    // Error. Announce a type resolution error.
                    errorMsg = "Variable " + this.getVarName()
                               + " declares unresolvable type: "
                               + ((AcmeType) this.getVarType()).getTypeName()
                               + " in quantified expression: " + this.toString();
                    TypeRefResolutionError badTypeRefError =
                        new TypeRefResolutionError(this, errorMsg,
                                               this.getVarType().getTypeName());
                    badTypeRefError.setUnresolvedRef(this.getVarType());
                    AcmeErrorManager.errorDetected(badTypeRefError);
                };*/
            };

        } else if (this.getVarType() == null) {
            // Error. Announce a type resolution error.
            errorMsg =
                "Variable "
                    + this.getVarName()
                    + " must declare a type but it does not."
                    + " (in quantified expression: "
                    + this.toString()
                    + ")";
            TypeRefResolutionError badTypeRefError = new TypeRefResolutionError(this, errorMsg);
            AcmeErrorManager.errorDetected(badTypeRefError);
			AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

            // set the var's type to a non-null, non-legal type
            this.setVarType(ConstantTypes.UNRESOLVABLE_TYPE);
        };
        // resolve the types for all children nodes.  This should resolve
        // any other types in the set expression or the evaluation expression
        // because they are both stored as children of this node.  If for
        // some reason the expression representation is changed so that they
        // are no longer stored as children, then their type refs will have
        // to be resolved explicitly.  (RTM 3/5/98).
        super.resolveTypeReferences();


    };


 /**
  * implement the Scope interface.  Check whether name is a quantified
  * variable in this expression.  If it is then return the ExpressionNode
  * that represents the value of this variable.
  * If this expression does not define the variable name then pass the
  * request up to this expression's scope, if the bubbleUp parameter is
  * true.
  * @param name The name
  * @param bubbleUp pass the request up
  * @return The object
  */
  public Object lookupName(String name, boolean bubbleUp) {
    Object retObj = null;
    if (name == null)
        return null;
    else if(name.equals(varName)) {
        // create an appropriate volatile variable and return it.
        VolatileVariable var = new VolatileVariable();
        var.refObjectType = this.getVarType();
        var.declarationScope = this;

        var.referentObject = this.getCurVarValue();
        retObj = var;
    }

    if ((retObj == null) && (bubbleUp))
        retObj = this.getParentScope().lookupName(name, bubbleUp);

    return retObj;
  };

    /**
     * @see acmeLib.designRules.ExpressionNode#cloneExpressionTree()
     */
    public ExpressionNode cloneExpressionTree() {
        QuantifiedExpressionNode clone = new QuantifiedExpressionNode (this);
        doCloneExpressionTree (clone);
        return clone;
    }

    /**
	 * @param clone
	 */
	protected void doCloneExpressionTree(ExpressionNode cl) {
		QuantifiedExpressionNode clone = (QuantifiedExpressionNode )cl;
		super.doCloneExpressionTree (clone);
        clone.varTypeAsVolatile = this.varTypeAsVolatile;
        clone.varType = this.varType;
       	clone.setExpr = this.setExpr.cloneExpressionTree();
       	clone.varName = this.varName;
       	clone.quantOp = this.quantOp;
       	clone.evalExpr = this.evalExpr.cloneExpressionTree();
       	clone.inParentheses = this.inParentheses;
	}

	/**
     * Returns the satisfies.
     * @return boolean
     */
    public boolean isSatisfies() {
        return checkSatisfies;
    }

    /**
     * Sets the satisfies.
     * @param satisfies The satisfies to set
     */
    public void setSatisfies(boolean satisfies) {
        this.checkSatisfies = satisfies;
    }

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion (this);
       if (this.getInParentheses()) {
            s.write("(");
        };

        s.write(this.getQuantOpAsString() + " ");
        s.write(this.getVarName());
        if ((this.getVarType() != null)
             && (this.getVarType() != ConstantTypes.UNSPECIFIED_TYPE)
             && !this.getVarTypeInferred())
        {
            if (this.checkSatisfies) {
                s.write(" : ");
            }
            else {
                s.write (" :! ");
            }
            s.write(this.getVarType().getTypeName());
        };
        s.write(" in ");
        if (setExpr != null) {
            setExpr.exportArmani (s);
        } else {
            s.write("{ <null set reference> }");
        }

        s.writeln(" | ");
        s.indent ();
        if (this.getEvalExpression() != null) {
            this.getEvalExpression ().exportArmani(s);
        } else {
            s.write("{ <null expression> }");
        };

        if (this.getInParentheses())
            s.write(")");
        s.unIndent ();
        s.write(" ");
        s.endRegion (this);
    }

}
