package acmeLib.designRules;

import java.io.IOException;

import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.InstanceRefResolutionError;
import acmeLib.properties.AcmeBoolean;
import acmeLib.properties.AcmeFloat;
import acmeLib.properties.AcmeInt;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmeString;
import acmeLib.properties.PropertyReferenceType;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: ReferenceNode<br>
 * Purpose: A reference node is an expression node that represents an<br>
 *          expression tree leaf that stores a reference to either an<br>
 *          identifier or a function.  This class supplies reference<br>
 *          resolution and stores qualified identifiers.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    10/20/97   RTM      * Created original class<br>
 *    10/28/97   RTM      * Added exception handling.<br>
 *     2/2/98    RTM      * updated typechecking and refreshReferentObject to<br>
 *                           deal with VolatileVariables and AcmeType's,<br>
 *                           removing all references to ExpressionTypeT's and<br>
 *                           ReferentObjectTypeT's.<br>
 *<br>
 * @author Bob Monroe
 */
public class ReferenceNode extends ExpressionNode
                           implements acmeLib.designRules.TypedEntity
{

  //
  // protected members
  //

/**
 * stores the references qualifiers and individual name as
 * a vector of strings, ie, the qualified identifier s1.s2.s3
 * is stored as qualifiedNames(0) = s1, qualifiedNames(1) = s2, ...
 * the last element of the qualifiedNames vector is the non-qualified
 * name of the reference.
 */
  protected java.util.Vector qualifiedNames;

/**
 * stores this reference as a single fully qualified string.
 */
  protected String identifierName;

/**
 * stores the object that the identifier refers to.  Subclasses
 * make sense of the type of this object.  The generic reference
 * node really doesn't care about the object's type or value.
 */
  protected Object referentObject;

/**
 * reference to the scope in which this expression should initially
 * attempt to look up the name stored as its referrent.
 * This scope should generally be a DesignRule object or
 * a DesignAnalysis, though it need be neither.  Eventually, properties
 * will be able to store derived expressions, so this may end up being
 * able to be a property value as well.
 */
  protected Scope referenceScope = null;

/**
 * determines whether this reference is volatile and should therefore
 * be looked up each time it is evaluated.  In general, the two kinds
 * of references that will be listed as volatile are quantified variables
 * and parameter names used in design analysis expressions.
 * volatile always defaults to false unless it is explicitly set.
 * In general, explicit element instances, type definitions, and properties
 * are not volatile.  Though these non-volatile items can be changed,
 * they are not expected to change every time the expression is evaluated.
 * Parameters for design analyses and quantified variables, on the other
 * hand, will likely change every time the expression is evaluated.
 */
  protected boolean isVolatile = false;

/**
 * the neverRefresh flag is more strict than the isVolatile flag.  If the
 * neverRefresh flag is set to true, then the reference will never be
 * refreshed.  This allows extensions to Armani and tools to set a value
 * for an identifier that may not change throughout the lifetime of the
 * session.  the neverRefresh flag is not stored persistently, so it is
 * only used to prevent unnecessary refreshes during a design's in-core
 * lifetime.
 */
  protected boolean neverRefresh = false;


/**
 * helper method for resolveInstanceReferences() and other methods to
 * be implemented in subclasses that looks up the referent object in
 * the current scope.
 */
  public void refreshReferentObject() throws AcmeResolutionException {
    // do not refresh referent object if explicitly asked not to...
    if (neverRefresh) {
        return;
    };

    // find the object that this node references.
    Scope curScope = this.getReferenceScope();

    // RTM 1/27/98 -- wherever lookupError is set I need to add
    //                appropriate error handling.
    String lookupError = null;

    // drill down through the string of qualified names looking for
    // the referent object
    if (qualifiedNames.size() == 0) {
        // then we don't have an identifier to try to resolve.
        this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
        InstanceRefResolutionError noRefNameError =
            new InstanceRefResolutionError(this,
                    "Reference node has no identifier to dereference."
                    + " (internal error?)", "\"\"");
        noRefNameError.setUnresolvedRef(this);
        AcmeErrorManager.errorDetected(noRefNameError);
        AcmeTypeManager.typecheckingErrorStack.pushCausalError(noRefNameError);
        throw new AcmeResolutionException (noRefNameError);
    }

    else if (qualifiedNames.size() > 1) {
        for (int i = 0; i < (qualifiedNames.size() - 1); i++) {
            // retrieve the scope for the i'th element in the qualifier vector
            String curScopeName = (String) qualifiedNames.elementAt(i);
            Object curFoundObj =
                curScope.lookupName(curScopeName, true);

            if (curFoundObj instanceof Scope) {
                // set the scope to that of the returned object
                curScope = (Scope) curFoundObj;
            } else if (curFoundObj instanceof VolatileVariable) {
                VolatileVariable curVolVar = (VolatileVariable) curFoundObj;
                if (curVolVar.referentObject instanceof Scope) {
                    curScope = (Scope) curVolVar.referentObject;

                } else if (curVolVar.referentObject instanceof ReferenceNode) {
                    ReferenceNode curVolVarRef = (ReferenceNode) curVolVar.referentObject;
                    if (curVolVarRef.getType() instanceof AcmeElement) {
                        try {
                            curScope = curVolVarRef.evaluateAsElement();
                        } catch (DesignRuleEvalException drExc) {
                            curScope = null;
                            referentObject = null;
                            this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
                            lookupError = "Instance Reference Lookup Error:  "
                                          + "Attempted to lookup the name: "
                                          + curScopeName +  " of the identifier: "
                                          + this.getIdentifier()
                                          + " in the non-scope object: "
                                          + curFoundObj.toString() + " (1)";
                            InstanceRefResolutionError notScopeError =
                                new InstanceRefResolutionError(this, lookupError,
                                                               this.getIdentifier());
                            notScopeError.setUnresolvedRef(this);
                            AcmeErrorManager.errorDetected(notScopeError);
							AcmeTypeManager.typecheckingErrorStack.pushCausalError(notScopeError);

                            throw new AcmeResolutionException (notScopeError);
                        };
                    }

                    else if (curVolVarRef.getType() instanceof PropertyReferenceType)
                    {
                        try {
                            curScope = curVolVarRef.evaluateAsPropertyRef();
                        } catch (DesignRuleEvalException drExc) {
                            curScope = null;
                            referentObject = null;
                            this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
                            lookupError = "Instance Reference Lookup Error:  "
                                          + "Attempted to lookup the name: "
                                          + curScopeName +  " of the identifier: "
                                          + this.getIdentifier()
                                          + " in the non-scope object: "
                                          + curFoundObj.toString() + " (1)";
                            InstanceRefResolutionError notScopeError =
                                new InstanceRefResolutionError(this, lookupError,
                                                               this.getIdentifier());
                            notScopeError.setUnresolvedRef(this);
                            AcmeErrorManager.errorDetected(notScopeError);
							AcmeTypeManager.typecheckingErrorStack.pushCausalError(notScopeError);

                            throw new AcmeResolutionException (notScopeError);
                           
                        };
                    };

                } else if (curVolVar.referentObject == null) {
                    // then try to resolve the type to see if it is a scope
                    if (curVolVar.refObjectType instanceof Scope) {
                        curScope = (Scope) curVolVar.refObjectType;
                    // if not, then we have an error
                    } else {
                        curScope = null;
                        referentObject = null;
                        this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
                        lookupError = "Instance Reference Lookup Error:  "
                                      + "Attempted to lookup the name: "
                                      + curScopeName +  " of the identifier: "
                                      + this.getIdentifier()
                                      + " in the non-scope object: "
                                      + curFoundObj.toString() + " (2)";
                        InstanceRefResolutionError notScopeError =
                            new InstanceRefResolutionError(this, lookupError,
                                                           this.getIdentifier());
                        notScopeError.setUnresolvedRef(this);
                        AcmeErrorManager.errorDetected(notScopeError);
						AcmeTypeManager.typecheckingErrorStack.pushCausalError(notScopeError);
                        
                        throw new AcmeResolutionException (notScopeError);
                    };
                } else {
                    // the i'th thing retrieved was not a valid scope or volatile
                    // variable so the identifier can't live within it legally.
                    // set the fields of this node appropriately and return.
                    curScope = null;
                    referentObject = null;
                    this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
                    lookupError = "Instance Reference Lookup Error:  "
                                  + "Attempted to lookup the name: "
                                  + curScopeName +  " of the identifier: "
                                  + this.getIdentifier()
                                  + " in the non-scope object: "
                                  + curFoundObj.toString() + " (3)";
                    InstanceRefResolutionError notScopeError =
                        new InstanceRefResolutionError(this, lookupError,
                                                       this.getIdentifier());
                    notScopeError.setUnresolvedRef(this);
                    AcmeErrorManager.errorDetected(notScopeError);
					AcmeTypeManager.typecheckingErrorStack.pushCausalError(notScopeError);
                    throw new AcmeResolutionException (notScopeError);
                };

            } else {
                // the i'th thing retrieved was not a valid scope or volatile
                // variable so the identifier can't live within it legally.
                // set the fields of this node appropriately and return
                curScope = null;
                referentObject = null;
                String curFoundObjString;
                if (curFoundObj == null) {
                    curFoundObjString = "<null found object>";
                } else {
                    curFoundObjString = curFoundObj.toString();
                };
                this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
                lookupError = "Instance Reference Lookup Error:  "
                              + "Attempted to lookup the name: "
                              + curScopeName +  " of the identifier: "
                              + this.getIdentifier()
                              + " in the non-scope object: "
                              + curFoundObjString + " (4)";
                InstanceRefResolutionError notScopeError =
                    new InstanceRefResolutionError(this, lookupError,
                                                   this.getIdentifier());
                notScopeError.setUnresolvedRef(this);
                AcmeErrorManager.errorDetected(notScopeError);
				AcmeTypeManager.typecheckingErrorStack.pushCausalError(notScopeError);
                throw new AcmeResolutionException (notScopeError);
            }
        };
    };

    // if we get to here, then curScope points to the scope of the
    // object that (should) contain the final identifier segment.
    int lastName = qualifiedNames.size() - 1;
    referentObject =
        curScope.lookupName((String) qualifiedNames.elementAt(lastName), true);

    // if we were unable to resolve the object that this references
    // leave the referentObject null and return.
    if (referentObject == null ) {
        lookupError = "Identifier \"" + this.getIdentifier()
                      + "\" could not be dereferenced to a valid object.";
        InstanceRefResolutionError refNotFoundError =
            new InstanceRefResolutionError(this, lookupError,
                                           this.getIdentifier());
        refNotFoundError.setUnresolvedRef(this);
        AcmeErrorManager.errorDetected(refNotFoundError);
		AcmeTypeManager.typecheckingErrorStack.pushCausalError(refNotFoundError);
        throw new AcmeResolutionException (refNotFoundError);
    };
  };


  //
  // public members
  //
/**
 * This is the constructor for this class
 * @param id The ID
 */
  public ReferenceNode(int id) {
    super(id);
    qualifiedNames = new java.util.Vector(4,4);
    referentObject = null;
  }

/**
 * JJTree-specific methods
 * @param id The ID
 * @return The node
 */
  public static Node jjtCreate(int id) {
    return new ReferenceNode(id);
  }

/**
 * Copy constructor
 * @param copyOf The ReferenceNode to be copied
 */
  public ReferenceNode(ReferenceNode copyOf) {
    super(copyOf);
    this.setReferenceScope(copyOf.getReferenceScope());
    this.setReferentObject(copyOf.getReferentObject());
    this.setIsVolatile(copyOf.getIsVolatile());

    int vectorSize = 4;
    if (copyOf.qualifiedNames != null)
        vectorSize = copyOf.qualifiedNames.size();
    this.qualifiedNames = new java.util.Vector(vectorSize, vectorSize);
    this.setIdentifier(copyOf.getIdentifier());
  }


/**
 * cloneExpressionTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.
 * @return The copy of the expression node
 */
  public ExpressionNode cloneExpressionTree() {
    ReferenceNode clone = new ReferenceNode(this);
    this.doCloneExpressionTree(clone);
    clone.identifierName = this.identifierName;
    clone.isVolatile = this.isVolatile;
    clone.referenceScope = this.referenceScope;
    return clone;
  };


/**
 * Check to see if two nodes are equal
 * @param test The tested node
 * @return true if the test node is equal to this node
 */
  public boolean equals(Object test) {
    /** basic algorithm:
      - refresh the referent object referred to (if needed).
      - based on the type of this referent element, evaluate test
          as the appropriate type, and then apply the equals method
          to the result.  put a wrapper of an AcmePropertyValue
          object around the actual AcmePropertyValue if needed.
     */
    if (test == null) {
        return false;
    };

    if (this.neverRefresh == false) {
        try {
            this.refreshReferentObject();
        }
        catch (AcmeResolutionException e) {
            return false;
        }
    };

    ExpressionNode testExprNode = null;
    if (test instanceof ExpressionNode) {
        testExprNode = (ExpressionNode) test;
    };

    // Check for (and then compare) string values.
    if (AcmeTypeManager.isLogicalSubtype(
            this.getType(), ConstantTypes.STRING_TYPE, false))
    {
        String thisVal = null;
        try {
            thisVal = this.evaluateAsString();
        } catch (Exception exc) {
            return false;
        };

        if (testExprNode != null) {
            try {
                String testVal = testExprNode.evaluateAsString();
                if (thisVal == null) {
                    return (testVal == null);
                }
                    return thisVal.equals(testVal);
            }
            catch (Exception exc) {
            };
        }

        else {
            // else check if test is a string literal or an AcmeString
            if (test instanceof AcmeString) {
                if (thisVal == null) {
                    return ((AcmeString) test).getValue() == null;
                }
                    return thisVal.equals(((AcmeString) test).getValue());
            }
            else if (test instanceof String) {
                if (thisVal == null) {
                    return ((String) test) == null;
                }
                    return thisVal.equals((String) test);
            }
                return false;   // can't find a match for it...
        };
    }

    // Check for (and then compare) boolean values.
    else if (AcmeTypeManager.isLogicalSubtype(
               this.getType(), ConstantTypes.BOOLEAN_TYPE, false))
    {
        boolean thisVal;
        try {
            thisVal = this.evaluateAsBoolean();
        } catch (Exception exc) {
            return false;
        };

        if (testExprNode != null) {
            try {
                return (thisVal == testExprNode.evaluateAsBoolean());
            }
            catch (Exception exc) {
            };
        }

            // else check if test is a boolean literal or an AcmeBoolean
            if (test instanceof AcmeBoolean) {
                return thisVal == ((AcmeBoolean) test).getValue();
            }
            else if (test instanceof Boolean) {
                return ((Boolean) test).booleanValue() == thisVal;
            }
                return false;   // can't find a match for it...
    }

    // Check for (and then compare) numeric values (all numeric types can
    // be evaluated as floats).
    else if (AcmeTypeManager.isLogicalSubtype(
               this.getType(), ConstantTypes.FLOAT_TYPE, false))
    {
        float thisVal;
        try {
            thisVal = this.evaluateAsFloat();
        } catch (Exception exc) {
            return false;
        };

        if (testExprNode != null) {
            try {
                return (thisVal == testExprNode.evaluateAsFloat());
            }
            catch (Exception exc) {
            };
        }

            // else check if test is a Java Float or Int object literal
            // or an AcmeFloat or AcmeInt
            if (test instanceof AcmeFloat) {
                return thisVal == ((AcmeFloat) test).getValue();
            }
            else if (test instanceof Float) {
                return ((Float) test).floatValue() == thisVal;
            }
            else if (test instanceof AcmeInt) {
                return thisVal == (float) ((AcmeInt) test).getValue();
            }
            else if (test instanceof Integer) {
                return ((float) ((Integer) test).floatValue()) == thisVal;
            }
                return false;   // can't find a match for it...
    }

    // Check for (and then compare) Element values.
    else if (this.getType() instanceof AcmeElementType)
    {
        AcmeElement thisVal;
        try {
            thisVal = this.evaluateAsElement();
        } catch (Exception exc) {
            return false;
        };

        if (testExprNode != null) {
            try {
                return (thisVal == testExprNode.evaluateAsElement());
            }
            catch (Exception exc) {
            };
        }

        else {
            return thisVal.equals(test);
        };
    }


    // Check for (and then compare) Type references.
    else if ((this.getType() == ConstantTypes.TYPE_REFERENCE_TYPE)
              && (testExprNode.getType() == ConstantTypes.TYPE_REFERENCE_TYPE))
    {
        if (testExprNode != null) {
            try {
                return (this.evaluateAsTypeRef() == testExprNode.evaluateAsTypeRef());
            }
            catch (Exception exc) {
            };
        }
            return false;
    }

    // Check for (and then compare) Property references.
    else if (this.getType() instanceof PropertyReferenceType)
    {
        AcmeProperty thisVal;
        try {
            thisVal = this.evaluateAsPropertyRef();
        } catch (Exception exc) {
            return false;
        };

        if (testExprNode != null) {
            try {
                return (thisVal == testExprNode.evaluateAsPropertyRef());
            }
            catch (Exception exc) {
            };
        }

            return thisVal.equals(test);
    }



    // Check for (and then compare) Set values.
    else if (AcmeTypeManager.isLogicalSubtype(
               this.getType(), ConstantTypes.ANY_SET_TYPE, false))
    {
        SetExpressionNode thisVal;
        try {
            thisVal = this.evaluateAsSet();
        } catch (Exception exc) {
            return false;
        };

        if (testExprNode != null) {
            try {
                if (thisVal != null) {
                    return thisVal.equals(testExprNode.evaluateAsSet());
                }
                    return false;
            }
            catch (Exception exc) {
            };
        }

        else {
            return thisVal.equals(test);
        };
    }

    else {
        // else this doesn't evaluate to a well understood type
        // so return false.
        return false;

        // RTM 4/22/98 -- at least enums are not properly dealt with here.
        //                figure out how to make those compare properly.
    };
    return false;
  };


  // referenceScope accessors/settors
/**
 * Get the reference scope
 * @return The reference scope
 */
  public Scope getReferenceScope() { return this.referenceScope; };
/**
 * Set the reference scope
 * @param newScope The new scope
 */
  public void setReferenceScope(Scope newScope) {
    this.referenceScope = newScope;
  };

/**
 * Get the reference object
 * @return The reference object
 */
  public Object getReferentObject() { return this.referentObject; };
/**
 * Set the reference object
 * @param newRefObj The new object
 */
  public void setReferentObject(Object newRefObj) {
    this.referentObject = newRefObj;
  };

/**
 * Check if it is volatile
 * @return true if it is volatile
 */
  public boolean getIsVolatile() { return this.isVolatile; }
/**
 * Set the isVolatile boolean value
 * @param isVolatile The new isVolatile boolean value
 */
  public void setIsVolatile(boolean isVolatile) { this.isVolatile = isVolatile; };

/**
 * Check if it never refreshes
 * @return true if it never refreshes
 */
  public boolean getNeverRefresh() { return this.neverRefresh; };
/**
 * Set the neverRefresh boolean value
 * @param newNeverRefresh The new neverRefresh boolean value
 */
  public void setNeverRefresh(boolean newNeverRefresh) {
    this.neverRefresh = newNeverRefresh;
  };

/**
 * Set this node's scope and set all children's scope by calling the
 * recursive method in this class' superclass.
 * @param newScope The new scope
 */
  public void setExprTreeScope(Scope newScope) {
    super.setExprTreeScope(newScope);
    this.setReferenceScope(newScope);
  };


/**
 * get/setIdentifier manipulates the identifier string being stored,
 * not the type of node it is, which is stored in the supertype
 * @return The identifier string
 */
  public String getIdentifier() { return this.identifierName; };
/**
 * @return The identifier string
 */
  public String getUnqualifiedIdentifier() {
    int lastIdIndex = this.qualifiedNames.size() - 1;
    if ((lastIdIndex >= 0) && (qualifiedNames != null))
        return (String) this.qualifiedNames.elementAt(lastIdIndex);
    else
        return null;
  };


/**
 * @param identifierName The identifier string
 */
  public void setIdentifier(String identifierName) {
    this.identifierName = identifierName;

    if ((identifierName == null) || identifierName.equals("")) {
        // don't hold anything for the empty or null string
        qualifiedNames.removeAllElements();
        this.identifierName = identifierName;
        return;
    };

    // parse identifierName to fill in the qualifiedNames vector
    String curToken;
    int curStart = 0;
    int curEnd = 0;
    int vectorIndex = 0;
    boolean endReached = false;
    qualifiedNames.removeAllElements();

    while (!endReached) {
        curEnd = identifierName.indexOf('.', curStart);
        if (curEnd == -1) {
            // then this is the last segment of the name
            endReached = true;
            curEnd = identifierName.length();
        }

        qualifiedNames.insertElementAt(
                identifierName.substring(curStart, curEnd),
                vectorIndex);
        curEnd++;
        curStart = curEnd;
        vectorIndex++;
    };
  };


/**
 * extensions to dump more interesting expression output messages.
 * @return The output string
 */
  public String toString() {
      String ret = "";
      try {
          AcmeOutputStream s = new AcmeOutputStream ();
          exportArmani (s);
          s.flush ();
          ret = s.toString ();
          s.close ();
      }
      catch (IOException e) {}
      catch (AcmeException e) {}
    
      return ret;
  };

/**
 * export method
 * @param prefix The prefix
 * @return The output string
 */
  public String toString(String prefix) { return prefix + toString(); }


/**
 * resolveInstanceReferences() looks up the entity that identifier
 * references.  It sets the referent object to the object that returned
 * from the lookup.
 */
  public void resolveInstanceReferences() throws AcmeResolutionException {
    // first, recursively call this on all of this referenceNode's
    // children, as implemented in ExpressionNode.resolveInstanceReferences()
    super.resolveInstanceReferences();

    // look up this reference to set this node's referentObject, but
    // _not_ the node's type (the node's type is set in the typecheck()
    // method.
    this.refreshReferentObject();

    if (this.getReferentObject() instanceof VolatileVariable) {
        this.setIsVolatile(true);
    };
  };

  
/**
 * @throws AcmeResolutionException
 * @see acmeLib.designRules.ExpressionNode#resolveTypeReferences()
 **/
public void resolveTypeReferences() throws AcmeResolutionException {
	super.resolveTypeReferences ();
	
	try {
	// Identifier could refer to a type, so refresh the object
	this.refreshReferentObject();
	if (this.getReferentObject() instanceof VolatileVariable) {
		this.setIsVolatile(true);
	}
	}
	catch (AcmeResolutionException e) {
		AcmeTypeManager.typecheckingErrorStack.popCausalError();
	}
}

/**
 * typecheck just delegates all requests to the superclass.  True
 * typechecking should be implemented in the subclasses.
 * This is a quasi-abstract method that should be implemented in
 * all subclasses.
 * @return true if type is correct
 */
  public boolean typecheck() {
    return super.typecheck();
  };



/**
 * Because ReferenceNode is an "unofficially" abstract base class,
 * All of the evaluateAs*() methods throw an exception if they are
 * actually called from this class rather than a subclass.  They
 * are implemented properly by subclasses that can return meaningful
 * values of the type requested.
 * @return The evaluated boolean value
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public boolean evaluateAsBoolean() throws DesignRuleEvalException {
    String opMsg = new String("Unable to evaluate ReferenceNode as a boolean");
    throw new ExpressionCastException(null, this,
            ConstantTypes.BOOLEAN_TYPE, this.getType(), opMsg);
  }

/**
 * @return The evaluated integer value
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public int evaluateAsInt() throws DesignRuleEvalException {
    String opMsg = new String("Unable to evaluate ReferenceNode as an int");
    throw new ExpressionCastException(null, this,
            ConstantTypes.INT_TYPE, this.getType(), opMsg);
  }

/**
 * @return The evaluated float value
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public float evaluateAsFloat() throws DesignRuleEvalException {
    String opMsg = new String("Unable to evaluate ReferenceNode as a float");
    throw new ExpressionCastException(null, this,
            ConstantTypes.FLOAT_TYPE, this.getType(), opMsg);
  }

/**
 * @return The evaluated string value
 * @exception DesignRuleEvalException If there is an evaluation error
 */
  public String evaluateAsString() throws DesignRuleEvalException {
    String opMsg = new String("Unable to evaluate ReferenceNode as a string");
    throw new ExpressionCastException(null, this,
            ConstantTypes.STRING_TYPE, this.getType(), opMsg);
  }

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion (this);
        java.util.Enumeration e = qualifiedNames.elements();
        if (this.getInParentheses()) s.write ("(");
        if (e.hasMoreElements())
            s.write((String) e.nextElement());

        while (e.hasMoreElements()) {
            s.write(".");
            s.write((String) e.nextElement());
        };

        if (this.getInParentheses())
           s.write (")");
        s.endRegion (this);
    }

}


