package acmeLib.designRules;

import java.io.IOException;
import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.AcmeTypeParameter;
import acmeLib.AcmeUnresolvedType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.filters.AcmeObjectFilter;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeUnresolvedElementType;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.errorManagement.TypeRefResolutionError;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyType;
import acmeLib.properties.AcmeUnresolvedPropertyType;
import acmeLib.properties.SetType;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: SetConstructorNode<br>
 * Purpose: Stores the representation of a set constructor in the Armani<br>
 *          design rule/constraint language.  A SetConstructor in Armani<br>
 *          is an expression that constructs a set S' from an existing<br>
 *          set S.  This class is a subclass of QuantifiedExpressionNode<br>
 *          because it has fundamentally the same structure in the Armani<br>
 *          language and the same typechecking semantics.  The primary<br>
 *          difference between the SetConstructorNode and the base<br>
 *          QuantifiedExpressionNode is that the SetConstructorNode evaluates<br>
 *          to a set, and the QuantifiedExpressionNode evaluates to a<br>
 *          boolean value.<br>
 *<br>
 *          SetConstructorNode implements SetExpressionNode.<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    3/9/98     RTM       Created<br>
 *<br>
 * @author Bob Monroe
 */
public class SetConstructorNode extends QuantifiedExpressionNode implements SetExpressionNode, Scope {
    //
    // Protected members
    //

    private VolatileVariable subpartTypeAsVolatile;

    /**
     * stores the elements of this set (not visible outside of this class)
     */
    protected AcmeVector constructedSetVector;

    /** Store the name of the subpart of a variable for COLLECT evaluation **/
    protected String subpartName = null;

    /** Stores the type of the elements to go in the set **/
    protected AcmeType subpartType = ConstantTypes.UNSPECIFIED_TYPE;

    /** Store whether the type was inferred **/
    protected boolean subpartTypeInferred = false;
    
    protected ExpressionNode varIdentifier = null;
    //
    // public members
    //

    /**
     * constants for determining this object's quantification operation.
     */
    public static final int SELECT = 100;
    public static final int COLLECT = 101;
    

    /**
     * @return
     */
    public boolean isSubpartTypeInferred() {
        return subpartTypeInferred;
    }

    /**
     * @param inferred
     */
    public void setSubpartTypeInferred(boolean inferred) {
        subpartTypeInferred = inferred;
    }

    /**
     * This is the constructor for this class
     */
    public SetConstructorNode() {
        super(ArmaniParserTreeConstants.JJTSETCONSTRUCTORNODE);

        // initially set the type and quantOp to default values.
        this.setType(new SetType("anonymous-set-type", ConstantTypes.UNRESOLVED_TYPE));
        this.setQuantOp(SELECT);
        this.constructedSetVector = new AcmeVector();
    }

    /**
     * This is the constructor for this class
     * @param id The ID
     */
    public SetConstructorNode(int id) {
        super(id);

        // initially set the type and quantOp to default values.
        this.setType(new SetType("anonymous-set-type", ConstantTypes.UNRESOLVED_TYPE));
        this.setQuantOp(SELECT);
        this.constructedSetVector = new AcmeVector();
    }
    
    public SetConstructorNode (SetConstructorNode copyOf) {
    	super (copyOf);
    }

    /**
     * This is the constructor for this class
     * @param id The ID
     * @param newEvalExpr The new eval expression
     */
    public SetConstructorNode(int id, ExpressionNode newEvalExpr) {
        super(id);
        this.evalExpr = newEvalExpr;

        // initially set the type and quantOp to default values.
        this.setType(new SetType("anonymous-set-type", ConstantTypes.UNRESOLVED_TYPE));
        this.setQuantOp(SELECT);
        this.constructedSetVector = new AcmeVector();
    }

    /**
     * JJTree-specific methods
     * @param id The ID
     * @return The node
     */
    public static Node jjtCreate(int id) {
        return new SetConstructorNode(id);
    }

    /**
     * compares for set equality (value equality)
     * @param test The tested object
     * @return true if the tested object is equal to this object
     */
    public boolean equals(Object test) {
        if (test instanceof SetExpressionNode) {
            return DrUtils.equalSets(this, (SetExpressionNode)test);
        }
        return false;
    };

    /**
     * typecheck requests that its supertype do the following:
     *  - recursively infer the type of its evaluation expression
     *  - check that all of the members of its set expression satisfy the
     *    type declared for the quantified variable.
     *  - confirm that the evaluation expression is a boolean expression.
     * if all of these checks are true then super.typecheck() returns true
     * and this method sets the type of this expression to a set containing
     * instances of the quantified variable's type and returns true.
     * if super.typecheck() returns false it sets the type to UNRESOLVABLE_TYPE
     * and returns false.
     * @return true if type is correct
     */
    public boolean typecheck() {
        boolean typechecks = super.typecheck();
        if (typechecks == false) {
            this.setType(ConstantTypes.UNRESOLVED_TYPE);
            return false;
        }
        if (this.getQuantOp() == SELECT) {
            // attempt to set the type of this node appropriately.
            AcmeType thisSetsType = this.getType();
            if (thisSetsType instanceof SetType) {
                ((SetType)thisSetsType).setSetType(this.getVarType());
            }
            else {
                thisSetsType = new SetType("anonymous-set-type", this.getVarType());
            };

            return true;
        }
        else {
            AcmeType thisSetsType = this.getType();
            if (thisSetsType instanceof SetType) {
                ((SetType)thisSetsType).setSetType(this.subpartType);
            }
            else {
                thisSetsType = new SetType("anonymous-set-type", this.subpartType);
            }
            // Check to see that the part.subpart actually typechecks
            String name = varIdentifier.toString();
//            String name = this.getVarName() + "." + this.subpartName;
            Object obj = this.getType();
            if (obj instanceof Scope) {
                Object subpart = ((Scope)obj).lookupName(this.subpartName, false);
                if (subpart != null) {
                    if (subpart instanceof AcmeProperty) {
                        AcmePropertyType propType = ((AcmeProperty)subpart).getType();
                        if (!AcmeTypeManager.isLogicalSubtype(propType, this.subpartType, false)) {
                            ExpressionTypecheckError err =
                                new ExpressionTypecheckError(
                                    this,
                                    propType.getTypeName()
                                        + " is not a logical subtype of "
                                        + subpartType.getTypeName()
                                        + " for "
                                        + name);
                            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                            return false;
                        }
                    }
                    else if (subpart instanceof AcmeElement) {
                        if (!AcmeTypeManager.declaresType(this.subpartType, (AcmeElement)subpart)) {
                            ExpressionTypecheckError err =
                                new ExpressionTypecheckError(
                                    this,
                                    name + " does not declare the type " + subpartType.getTypeName());
                            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                            return false;
                        }
                    }
                    else if (subpart instanceof AcmeEnumeration && !(this.subpartType instanceof SetType)) {
                        ExpressionTypecheckError err =
                            new ExpressionTypecheckError(this, "Expecting " + name + " to be a set");
                        AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                        return false;
                    }
                }
                else {
                    ExpressionTypecheckError err = new ExpressionTypecheckError(this, "Could not resolve " + name);
                    AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
                    return false;
                }
            }
            else {
                ExpressionTypecheckError err = new ExpressionTypecheckError(this, "Could not resolve " + name);
                AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);

                return false;
            }
            return true;
        }
    };

    /**
     * Return "this" node when it is asked to evaluate itself as a set.
     * @return The set expression node
     * @exception DesignRuleEvalException If there is an evaluation error
     */
    public SetExpressionNode evaluateAsSet() throws DesignRuleEvalException {
        // first refresh the set
        this.refreshSet();

        // then return a transient set that contains the current contents of
        // the constructed set.  This deals with "issues" relating to multiple
        // refreshes that are potentially both inside and outside of the scope
        // of a design analysis call
        TransientSet returnSet = new TransientSet(this.constructedSetVector);
        returnSet.setType(this.getType());

        return returnSet;

        // return this;
    };

    /** SetConstructors can not be evaluated as booleans.  throw an appropriate
     * exception.
     * RTM 3/9/98 -- When I get the "evaluate as set" methods working properly
     *               I should add one to this class.
     * @return The evaluated boolean value
     * @exception DesignRuleEvalException If there is an evaluation error
     */
    public boolean evaluateAsBoolean() throws DesignRuleEvalException {
        // should always call a subclass's implementation of this method
        throw new DesignRuleEvalException("Attempted to evaluate " + "a SetConstructorNode as a Boolean");
    };

    /**
     * toString() and exportAsAcme are taken care of primarily by the parent
     * class QuantifiedExpressionNode.  getQuantOpAsString is the only method
     * that needs to be overridden for this to work.
     *
     * convert the quantOp to a string for exporting, etc.
     * potentially useful to the world at large, so it's public.
     * @return The string
     */
    public String getQuantOpAsString() {
        if (this.getQuantOp() == SELECT) {
            return "Select";
        }
        return super.getQuantOpAsString();

    };

    /**
     * export method
     * @return The output string
     */
    public String toString() {
        return super.toString();
    };

    //
    // implement the SetExpressionNode interface.
    //

    /** accessors/settors for set expression scope. */
    /**
     * Get the set scope
     * @return The set scope
     */
    public Scope getSetScope() {
        return this.getParentScope();
    }
    /**
     * Set the set scope
     * @param newScope The new scope
     */
    public void setSetScope(Scope newScope) {
        this.setParentScope(newScope);
    }

    /** accessors/settors for set type. */
    /**
     * Get the set type
     * @return The set type
     */
    public AcmeType getSetType() {
        return this.getVarType();
    }
    /**
     * Set the set type
     * @param newType The new set type
     */
    public void setSetType(AcmeType newType) {
        this.setVarType(newType);
    }

    /**
     * returns the size of the current set
     * @return The size
     */
    public int size() {
        return this.enumerate().getSize();
    };

    /**
     * contains(Object elt) returns true iff elt is a member of this set
     * @return true if elt is a member of this set
     */
    public boolean contains(AcmeObject elt) {
        this.refreshSet();
        return this.constructedSetVector.contains(elt);
    };

    /**
     * enumerate() returns an enumeration of all of the elements of the set
     * @return an enumeration of all elements in the set
     */
    public AcmeEnumeration enumerate() {
        this.refreshSet();
        return this.constructedSetVector.enumerate();
    }

    /**
     * refreshSet() recomputes the value of the set based on the source set
     * and the evaluation expression.  This node's set is first emptied and
     * then all elements of the source set that satisfy the boolean evaluation
     * expression are added to the set.
     */
    public void refreshSet() {

        // empty the vector that represents the set
        this.constructedSetVector.removeAllItems();

        // check that we have non-empty set and evaluation expressions.
        if ((this.getSetExpression() == null) || (this.getEvalExpression() == null)) {
            // then the set will be empty. just return with it empty.
            return;
        }

        // refresh the set over which we are quantifying.
        this.getSetExpression().refreshSet();

        // assign each of the values in the source set to the quantified
        // variable.  if the eval expression evaluates to true with this
        // variable bound to the current element, then add the current
        // element to the constructed set.
        AcmeEnumeration e = this.getSetExpression().enumerate().castToAcmeSetVector().enumerate(new AcmeObjectFilter () {

			public boolean test(AcmeObject obj) {
				return isTypeCompliant(obj);
			}
        	
        });
        if (this.getQuantOp() == SELECT) {
            while (e.moreItems()) {

                Object curValObj = e.getNextItem();
                if (curValObj instanceof ReferenceNode) {
                    curValObj = ((ReferenceNode)curValObj).getReferentObject();
                };

                this.setCurVarValue(curValObj);
                try {
                    if (this.getEvalExpression().evaluateAsBoolean() == true) {
                        if ((curValObj) instanceof AcmeObject) {
                            this.constructedSetVector.addItem((AcmeObject)curValObj);
                        }
                        else {
                            // else I don't know what to do.  flag an (internal) error?
                            System.err.println(
                                "Warning.  internal discrpancy."
                                    + " Attempted to add a non-AcmeObject"
                                    + " to a constructed set.");
                        };
                    }
                }
                catch (DesignRuleEvalException drExc) {
                    // RTM 2/28/98 -- internal error or type error?
                    //                either way it blows the expression.
                    //                and we shouldn't add the item.
                };
            };
        }
        else if (this.getQuantOp() == COLLECT) {
            while (e.moreItems()) {
                Object curValObj = e.getNextItem();
                if (curValObj instanceof ReferenceNode) {
                    curValObj = ((ReferenceNode)curValObj).getReferentObject();
                }
                this.setCurVarValue(curValObj);
                try {
                    if (this.getEvalExpression().evaluateAsBoolean() == true) {
                        if (curValObj instanceof AcmeObject) {
                            if (curValObj instanceof Scope) {
                                Scope obj = (Scope)curValObj;
                                Object item = obj.lookupName(subpartName, false);
                                if (item == null) {
                                    System.err.println(
                                        "Warning. internal discrepency. Cannot access "
                                            + subpartName
                                            + " in "
                                            + varName);
                                }
                                else {
                                    if (item instanceof AcmeProperty) {
                                        this.constructedSetVector.addItem(((AcmeProperty)item).getValue());
                                    }
                                    else {
                                        this.constructedSetVector.addItem((AcmeObject)curValObj);
                                    }
                                }
                            }
                        }
                        else {
                            // else I don't know what to do.  flag an (internal) error?
                            System.err.println(
                                "Warning.  internal discrpancy."
                                    + " Attempted to add a non-AcmeObject"
                                    + " to a constructed set.");
                        }
                    }
                }
                catch (DesignRuleEvalException drExc) {
                }
            }
        }
        else {
            // else this is an unrecognized quantification, simply
            // return an empty set.

            // RTM 3/9/98 -- If this happens it is an internal error.
            //               flag it appropriately (how?)
            System.err.println(
                "Warning.  internal discrepancy.  unrecognized" + " quantification operator: " + this.getQuantOp());
        };

    };

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        if (this.getQuantOp() == SELECT) {
            s.beginRegion(this);
            s.write("{");
            super.exportArmani(s);
            s.write("}");
            s.endRegion(this);
        }
        else {
            s.beginRegion(this);
            s.write("{");
            if (this.getInParentheses())
                s.write("(");
            s.write("collect ");
            s.write(this.getVarName() + "." + subpartName);
            s.write(" : ");
            s.write(this.getVarType().getTypeName());
            s.write(".");
            s.write(this.subpartType.getTypeName());
            s.write(" in ");

            if (setExpr != null) {
                setExpr.exportArmani(s);
            }
            else {
                s.write("{}");
            }
            s.writeln(" | ");
            s.indent();
            if (this.getEvalExpression() != null) {
                this.getEvalExpression().exportArmani(s);
            }
            else {
                s.write("<null expression>");
            }

            if (this.getInParentheses())
                s.write(")");
            s.unIndent();
            s.write("}");
            s.endRegion(this);
        }
    }

    /**
     * @param string
     */
    public void setSubpartName(String string) {
        subpartName = string;
    }

    /**
     * @param subpartType
     */
    public void setSubpartType(AcmeType subpartType) {
        this.subpartType = subpartType;
    }
    
    public void setIdentifier (ExpressionNode id) {
        varIdentifier = id;
    }

    /* (non-Javadoc)
     * @see acmeLib.designRules.ExpressionNode#resolveInstanceReferences()
     */
    public void resolveInstanceReferences() throws AcmeResolutionException {
        //        if (getQuantOp () == SELECT) {
        super.resolveInstanceReferences();
        //        } 
        //        else {
        //            
        //        }
    }

    /* (non-Javadoc)
     * @see acmeLib.designRules.ExpressionNode#resolveTypeReferences()
     */
    public void resolveTypeReferences() throws AcmeResolutionException {
        if (getQuantOp() == SELECT) {
            super.resolveTypeReferences();
        }
        else {
            String errorMsg = null;
            AcmeType foundType = null;
            VolatileVariable v = null;
            Object foundObj = null;
            AcmeType foundSubpartType = null;
            VolatileVariable spv = null;
            if (this.subpartType instanceof AcmeUnresolvedType
                || this.subpartType instanceof AcmeUnresolvedPropertyType
                || this.subpartType instanceof AcmeUnresolvedElementType) {
                if (this.parentScope == null) {
                    errorMsg =
                        "Quantified Expression: "
                            + this.toString()
                            + " does not have a valid parent scope for resolving "
                            + "type name: "
                            + ((AcmeType)this.getVarType()).getTypeName();
                    TypeRefResolutionError badTypeRefError =
                        new TypeRefResolutionError(this, errorMsg, this.getVarType().getTypeName());
                    badTypeRefError.setUnresolvedRef(this.getVarType());
                    AcmeErrorManager.errorDetected(badTypeRefError);
					AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

                }

                else {
                    Scope curScope = this.getParentScope();
                    Object namedObj = curScope.lookupName(subpartName, true);
                    if (namedObj instanceof Scope) {
                        curScope = (Scope)namedObj;
                        if (curScope instanceof AcmeType) {
                            foundSubpartType = (AcmeType)namedObj;
                        }
                        else {
                            foundSubpartType = null;
                        }
                    }
                    else if (namedObj instanceof VolatileVariable) {
                        VolatileVariable var = (VolatileVariable)namedObj;
                        spv = var;
                        curScope = var.declarationScope;
                        foundSubpartType = new AcmeTypeParameter(subpartName);
                    }

                    if (foundSubpartType != null) {
                        this.subpartType = foundSubpartType;
                        this.subpartTypeAsVolatile = spv;
                    }
                    else {
                        // Error. Announce a type resolution error.
                        errorMsg =
                            "Variable "
                                + this.getVarName()
                                + " declares unresolvable type: "
                                + ((AcmeType)this.getVarType()).getTypeName()
                                + " in quantified expression: "
                                + this.toString();
                        TypeRefResolutionError badTypeRefError =
                            new TypeRefResolutionError(this, errorMsg, this.getVarType().getTypeName());
                        badTypeRefError.setUnresolvedRef(this.getVarType());
                        AcmeErrorManager.errorDetected(badTypeRefError);
						AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

                    }
                }
            }
            if ((this.getVarType() instanceof AcmeUnresolvedType)
                || (this.getVarType() instanceof AcmeUnresolvedPropertyType)
                || (this.getVarType() instanceof AcmeUnresolvedElementType)) {
                if (this.getParentScope() == null) {
                    // Error. Announce a type resolution error.
                    errorMsg =
                        "Quantified Expression: "
                            + this.toString()
                            + " does not have a valid parent scope for resolving"
                            + " type name: "
                            + ((AcmeType)this.getVarType()).getTypeName();
                    TypeRefResolutionError badTypeRefError =
                        new TypeRefResolutionError(this, errorMsg, this.getVarType().getTypeName());
                    badTypeRefError.setUnresolvedRef(this.getVarType());
					AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

                    AcmeErrorManager.errorDetected(badTypeRefError);

                }
                else {
                    java.util.StringTokenizer types =
                        new java.util.StringTokenizer(this.getVarType().getTypeName(), ".");
                    Scope curScope = this.getParentScope();
                    String curName = types.nextToken();
                    //                     (01/03/03 zensoul) modified to allow VolatileVariable as lookupName result.
                    Object namedObj = curScope.lookupName(curName, true);
                    if (namedObj instanceof Scope) {
                        curScope = (Scope)namedObj;
                        if (curScope instanceof AcmeType) {
                            foundType = (AcmeType)namedObj;
                        }
                        else {
                            foundType = null;
                        }
                    }
                    else if (namedObj instanceof VolatileVariable) {
                        VolatileVariable var = (VolatileVariable)namedObj;
                        v = var;
                        curScope = var.declarationScope;
                        foundType = new AcmeTypeParameter(curName);
                    }
                    else {
                        // SWC 01/03/03:  Not sure what to do in this case.
                    }
                    //                     END (01/03/03 zensoul)
                    //                    curScope = (Scope)curScope.lookupName(curName,true);
                    //                  if (curScope instanceof AcmeType){
                    //                      foundType = (AcmeType)curScope;
                    //                  } else {
                    //                      foundType = null;
                    //                  }
                    while (types.hasMoreElements() && curScope != null) {
                        curName = types.nextToken();
                        curScope = (AcmeElement)curScope.lookupName(curName, false);
                        if (curScope instanceof AcmeType) {
                            foundType = (AcmeType)curScope;
                        }
                        else {
                            foundType = null;
                        }
                    }

                    if (foundType != null) {
                        this.setVarType(foundType);
                        this.setVarType(v);
                    }
                    else {
                        // Error. Announce a type resolution error.
                        errorMsg =
                            "Variable "
                                + this.getVarName()
                                + " declares unresolvable type: "
                                + ((AcmeType)this.getVarType()).getTypeName()
                                + " in quantified expression: "
                                + this.toString();
                        TypeRefResolutionError badTypeRefError =
                            new TypeRefResolutionError(this, errorMsg, this.getVarType().getTypeName());
                        badTypeRefError.setUnresolvedRef(this.getVarType());
                        AcmeErrorManager.errorDetected(badTypeRefError);
						AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);

                    }
                    /*String varTypeName = this.getVarType().getTypeName();
                    foundObj = this.getParentScope().lookupName(varTypeName, true);
                    if (foundObj instanceof AcmeType) {
                        this.setVarType((AcmeType) foundObj);
                    } else {
                        // Error. Announce a type resolution error.
                        errorMsg = "Variable " + this.getVarName()
                                   + " declares unresolvable type: "
                                   + ((AcmeType) this.getVarType()).getTypeName()
                                   + " in quantified expression: " + this.toString();
                        TypeRefResolutionError badTypeRefError =
                            new TypeRefResolutionError(this, errorMsg,
                                                   this.getVarType().getTypeName());
                        badTypeRefError.setUnresolvedRef(this.getVarType());
                        AcmeErrorManager.errorDetected(badTypeRefError);
                    };*/
                };

            }
            else if (this.getVarType() == null) {
                // Error. Announce a type resolution error.
                errorMsg =
                    "Variable "
                        + this.getVarName()
                        + " must declare a type but it does not."
                        + " (in quantified expression: "
                        + this.toString()
                        + ")";
                TypeRefResolutionError badTypeRefError = new TypeRefResolutionError(this, errorMsg);
                AcmeErrorManager.errorDetected(badTypeRefError);
				AcmeTypeManager.typecheckingErrorStack.pushCausalError(badTypeRefError);


                // set the var's type to a non-null, non-legal type
                this.setVarType(ConstantTypes.UNRESOLVABLE_TYPE);
            }

        }

    }
    
    
	/**
	 * @return
	 * @see acmeLib.designRules.QuantifiedExpressionNode#cloneExpressionTree()
	 **/
	public ExpressionNode cloneExpressionTree() {
		SetConstructorNode clone = new SetConstructorNode (this);
		doCloneExpressionTree(clone);
		return clone;
	}
	
	
	/**
	 * @param cl
	 * @see acmeLib.designRules.QuantifiedExpressionNode#doCloneExpressionTree(acmeLib.designRules.ExpressionNode)
	 **/
	protected void doCloneExpressionTree(ExpressionNode cl) {
		super.doCloneExpressionTree (cl);
		SetConstructorNode clone = (SetConstructorNode )cl;
		clone.subpartName = subpartName;
		clone.subpartType = subpartType;
		clone.subpartTypeInferred = subpartTypeInferred;
		clone.varIdentifier = varIdentifier;
	}
}