package acmeLib.designRules;

import java.io.IOException;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.ArmaniExportable;
import acmeLib.properties.AcmeUnresolvedPropertyType;
import acmeLib.properties.SetType;
import acmeLib.typeManagement.AcmeTypeManager;


/**
 * Class: SetLiteralNode<br>
 * Purpose: Stores the representation of an Armani set expression that is<br>
 *          given as a series of literals.  That is, all members of the<br>
 *          set are explicitly specified in the set expression (though<br>
 *          the entities specified may be references themselves)<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    1/9/98     RTM       Created<br>
 *<br>
 * @author Bob Monroe
 */
public class SetLiteralNode extends ExpressionNode implements SetExpressionNode
{

    //
    // protected members
    //

  /**
   * stores the elements of this set
   */
    protected AcmeVector setVector;

  /**
   * stores the scope of this set.
   */
    protected Scope setScope = null;

  /**
   * stores the type of entities stored within this set.
   */
    protected AcmeType setType;


    //
    // public members
    //

  /**
   * constructors
   * @param id The ID
   */
    public SetLiteralNode(int id) {
        super(id);
        this.setType(new SetType("set-literal",
                     new AcmeUnresolvedPropertyType("unresolved")));
        this.setVector = new AcmeVector(1,1);
    }


  /**
   * Copy constructor
   * @param copyOf The FloatLiteralNode to be copied
   */
    public SetLiteralNode(SetLiteralNode copyOf) {
        super(copyOf);
        this.setType(copyOf.getType());
        this.setVector = new AcmeVector(1,1);
    }


  /**
   * JJTree-specific methods
   * @param id The ID
   * @return The node
   */
    public static Node jjtCreate(int id) {
        return new SetLiteralNode(id);
    }


    /** accessors/settors for set expression scope. */
  /**
   * Get the set scope
   * @return The set scope
   */
    public Scope getSetScope() { return this.setScope; }
  /**
   * Set the set scope
   * @param newScope The new scope
   */
    public void setSetScope(Scope newScope) { this.setScope = newScope; }


    /** accessors/settors for set type. */
  /**
   * Get the set type
   * @return The set type
   */
    public AcmeType getSetType() { return this.setType; };
  /**
   * Set the set type
   * @param newType The new set type
   */
    public void setSetType(AcmeType newType) { this.setType = newType; };


  /**
   * compares for set equality (value equality)
   * @param test The tested node
   * @return true if the test node is equal to this node
   */
    public boolean equals(Object test) {
        if (test instanceof SetExpressionNode) {
            return DrUtils.equalSets(this, (SetExpressionNode) test);
        }
            return false;

    };


  /**
   * returns the size of the current set
   * @return the size
   */
    public int size() { return setVector.getSize(); }


  /**
   * refreshSet() recomputes the value of the set based on the expression
   * tree stored in the children of this node.  The value of a SetLiteral
   * shouldn't change over time (unlike a reference, set constructor, or
   * set function), so SetLiteralNode.refresh() is primarily used to
   * initalize the internal node storage vector (that is used for
   * enumerations).
   */
    public void refreshSet() {
        this.setVector.removeAllItems();
        int numChildren = this.jjtGetNumChildren();
        for (int i = 0; i < numChildren; i++) {
            this.setVector.addItem((AcmeObject) this.jjtGetChild(i));
        }
    }


  /**
   * contains(Object elt) returns true iff elt is a member of this set
   * @param elt The element
   * @return true if the element if a member of this set
   */
    public boolean contains(AcmeObject elt) {
        this.refreshSet();
        return this.setVector.contains(elt);
    };



  /**
   * enumerate() returns an enumeration of all of the elements of the set
   * @return The enumeration of all elements in the set
   */
    public AcmeEnumeration enumerate() {
        this.refreshSet();
        return this.setVector.enumerate();
    }


  /**
   * return a string representing this set of literals
   * (for exporting/unparsing)
   * @return The output string
   */
    public String toString() {
        String ret = "";
        try {
            AcmeOutputStream s = new AcmeOutputStream ();
            exportArmani (s);
            s.flush ();
            ret = s.toString ();
            s.close ();
        }
        catch (IOException e) {}
        catch (AcmeException e) {}
    
        return ret;
    }

  /**
   * return a string representing this set of literals
   * @param prefix The prefix
   * @return The output string
   */
    public String toString(String prefix) {
        return prefix + this.toString();
    };


  /**
   * SetLiteralNode calls typecheck on all of its children and then builds
   * a "greatest common supertype (gcs)" by comparing the types of entities
   * with AcmeTypeManager.isLogicalSubtype().  It then assigns itself the
   * type of the gcs.  If it has no gcs then it assigns itself
   * ConstantTypes.UNRESOLVABLE_TYPE and returns false, else returns true.
   * @return true if type is correct
   */
    public boolean typecheck() {
        // put all of the children literal nodes into the setVector
        this.refreshSet();

        // typecheck all of this node's children to calculate their types and
        // figure out whether they each typecheck.
        boolean typechecks = true;
        if (super.typecheck() == false) {
            this.setType(ConstantTypes.UNRESOLVABLE_TYPE);
            return false;

        } else {
            // make a vector of the types of all of this set's elements.
            int vectorSize = this.setVector.getSize();
            AcmeVector typeVector = new AcmeVector(vectorSize);
            for (int i = 0; i < vectorSize; i++) {
                AcmeType eltType =
                    ((ExpressionNode) this.setVector.getItemAt(i)).getType();
                typeVector.insertItemAt((AcmeObject) eltType, i);
            };

            // infer the type of the set from the type vector
            AcmeType setContentsType =
                        AcmeTypeManager.inferLeastCommonSupertype(typeVector);
            this.setSetType(setContentsType);
            if ((setContentsType == ConstantTypes.UNSPECIFIED_TYPE)
                || (setContentsType == ConstantTypes.EXPLICIT_ANY_TYPE))
            {
                this.setType(ConstantTypes.ANY_SET_TYPE);
            }
                this.setType(new SetType("anonymous-set-type", this.getSetType()));
            };

            return true;
    }



  /**
   * Return "this" node when it is asked to evaluate itself as a set.
   * @return The set expression node
   * @exception DesignRuleEvalException IF there is an evaluation error
   */
    public SetExpressionNode evaluateAsSet() throws DesignRuleEvalException {
        return this;
    };

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion (this);
        s.write("{");
        AcmeEnumeration e = this.enumerate();
        while (e.moreItems()) {
            Object o = e.getNextItem ();
            if (o instanceof ArmaniExportable) {
                ((ArmaniExportable )o).exportArmani (s);
            }
            else {
                s.write(o.toString());
            }
            
            if (e.moreItems()) {
                s.write(", ");
            };
        }
        s.write("}");
        s.endRegion (this);
    }

	/* (non-Javadoc)
	 * @see acmeLib.designRules.ExpressionNode#cloneExpressionTree()
	 */
	public ExpressionNode cloneExpressionTree() {
		SetLiteralNode clone = new SetLiteralNode (this);
		return clone;
	}
}
