package acmeLib.designRules;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmeSetType;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.properties.*;
import acmeLib.typeManagement.AcmeTypeManager;


/**
 * Class: SetReferenceNode<br>
 * Purpose: Stores a reference to an Armani set expression or property.<br>
 *          This is divided into a reference to the containing element<br>
 *          and the operation that is performed on that element to retrieve<br>
 *          a set.  For example self.ports identifies an element (self)<br>
 *          and the operation to perform on that element (ports), which<br>
 *          returns "self's" set of ports.<br>
 *<br>
 *<br>
 * Revision History:<br>
 *    Date:     Person    Change<br>
 *    1/12/98    RTM       Created<br>
 *    2/2/98     RTM       Modified type storage to store AcmePropertyTypes<br>
 *                          instead of ExpressionTypeT objects.<br>
 *    3/7/98     RTM       Updated set to properly retrieve referenced sets<br>
 *                          such as "comp.ports" by retrieving them as<br>
 *                          enumerations.  Adapted the rest of the class to<br>
 *                          deal with not necessarily having an AcmeSet as<br>
 *                          a reference.<br>
 *<br>
 * @author Bob Monroe
 */
public class SetReferenceNode extends IdentifierNode implements SetExpressionNode
{

    //
    // protected members
    //

  /**
   * stores the type of the entities stored within this set.
   */
    protected AcmeType setType = null;

    //
    // public members
    //

  /**
   * constructors
   * @param id The ID
   */
    public SetReferenceNode(int id) {
        super(id);
        this.setType(new SetType("set-reference", ConstantTypes.UNRESOLVED_TYPE));
    }


  /**
   * Copy constructor
   * @param copyOf The ReferenceNode to be copied
   */
    public SetReferenceNode(SetReferenceNode copyOf) {
        super(copyOf);
        this.setType(copyOf.getType());
        this.setSetType (copyOf.getSetType ());
    }


  /**
   * JJTree-specific methods
   * @param id The ID
   * @return The node
   */
    public static Node jjtCreate(int id) {
        return new SetReferenceNode(id);
    }


    /** accessors/settors for set expression scope. */
  /**
   * Get the set scope
   * @return The set scope
   */
    public Scope getSetScope() { return this.getReferenceScope(); }
  /**
   * Set the set scope
   * @param newScope The new scope
   */
    public void setSetScope(Scope newScope) {
        this.setReferenceScope(newScope);
    }

    /** accessors/settors for set type.  This is the type of entities stored
        in the set, not the type of the set itself. */
  /**
   * Get the set type
   * @return The AcmeType
   */
    public AcmeType getSetType() { return this.setType; }
  /**
   * Set the set type
   * @param newType The new AcmeType
   */
    public void setSetType(AcmeType newType) {
        this.setType = newType;
    }


  /**
   * compares for set equality (value equality)
   * @param test The tested object
   * @return true if the tested object if equal to this object
   */
    public boolean equals(Object test) {
        if (test instanceof SetExpressionNode)
            return DrUtils.equalSets(this, (SetExpressionNode) test);
        return false;
    };


    //
    // implementation of SetExpressionNode interface methods
    //

  /**
   * returns the size of the current set
   * @return The size
   */
    public int size() {
        // RTM 1/12/98 -- horribly inefficient but general method that allows
        //                for various set representation (e.g. sets and enums).
        int setSize = 0;
        AcmeEnumeration e = this.enumerate();
        while (e.moreItems()) { e.getNextItem(); setSize++; }
        return setSize;
    };


  /**
   * contains(Object elt) returns true iff elt is a member of this set
   * @param elt The element
   * @return true if elt is a member of this set
   */
    public boolean contains(AcmeObject elt) {
        AcmeEnumeration e = this.enumerate();
        while (e.moreItems()) {
            if (elt == e.getNextItem())
                return true;
        };
        return false;
    };


  /**
   * refreshSet() recomputes the value of the set based on the value of
   * the set to which this node refers.
   */
    public void refreshSet() {
        try {
            this.refreshReferentObject();
        }
        catch (AcmeResolutionException e) {
        }
    }


  /**
   * enumerate() returns an enumeration of all of the elements of the set
   * @return The enumeration of all elements in the set
   */
    public AcmeEnumeration enumerate() {
        // look up the entity that identifier refers to.  if the
        // referentObject is null,  then return a null enumeration for now.
        this.refreshSet();
        acmeLib.core.container.AcmeEnumeration returnEnum = null;

        if (this.referentObject instanceof AcmeEnumeration) {
            // RTM 3/7/98 -- eventually, this should be checked for
            //               semantics, but perhaps that should be done
            //               in the refreshSet phase...  For now the check
            //               is definitely not done here.
            returnEnum = (AcmeEnumeration) this.referentObject;
        } else if (this.referentObject instanceof AcmeSet) {
            returnEnum = ((AcmeSet) this.referentObject).getValues();
        } else if (this.referentObject instanceof AcmeSequence) {
            // then convert the sequence to a set and return an
            // enumeration of its entities.
            // RTM 2/25/98 -- not yet fully implemented, just return
            //                the sequence, regardless of set-semantics
            //                for now.  Eventually, do a check to convert
            //                it to a set before returning the enumeration.
            returnEnum = ((AcmeSequence) this.referentObject).getValues();
        } else {
            // else we have an error of some sort. return an empty enum.
            // RTM 2/25/98 -- Should this be an error or what?
            returnEnum = new AcmeEnumeration();
        }

        return returnEnum;
    }


/**
 * typecheck attempts to infer the type of the set (or checks if it
 * has been explicitly set), and then checks that all of the entities
 * in the set satisfy that type.
 * @return true if the type is correct
 */
public boolean typecheck() {
    // look up the entity that identifier refers to and infer the setType
    // for this object appropriately.  return false if the types within
    // the set are inconsistent with that type.
    this.refreshSet();
    boolean setTypechecks = false;
    if (this.referentObject instanceof AcmeEnumeration) {
        // then the parser or the calling program that created this
        // ReferenceNode should have set the setType.
        setTypechecks =
            ((this.getSetType() != ConstantTypes.UNRESOLVABLE_TYPE)
                && (this.getSetType() != ConstantTypes.UNRESOLVED_TYPE)
                && (!(this.getSetType() instanceof AcmeUnresolvedPropertyType))
                && (this.getSetType() != null));
        if (!setTypechecks) {   
            ExpressionTypecheckError err = new ExpressionTypecheckError(this, "Type is unresolved", this);
            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
        }
    }

    else if (this.referentObject instanceof AcmeProperty) {
        AcmeProperty refProp = (AcmeProperty) this.referentObject;
        this.setType(refProp.getType());
        if (refProp.getValue() instanceof AcmeSet) {
            this.setSetType(((AcmeSet) refProp.getValue()).getSetType());
            setTypechecks = AcmeTypeManager.typecheckProperty(refProp);
        }
        else {
            ExpressionTypecheckError err = new ExpressionTypecheckError(this, "Reference is not a set", this);
            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);

            setTypechecks = false;
        };
    }

    else if (this.referentObject instanceof AcmeSet) {
        AcmeSet refSet = (AcmeSet) this.referentObject;
        this.setType(refSet.getType());
        this.setSetType(refSet.getSetType());
        setTypechecks = AcmeTypeManager.typecheckProperty(refSet.getType(), refSet);
        if (!setTypechecks) {
            ExpressionTypecheckError err = new ExpressionTypecheckError(this, "Reference is not a set", this);
            AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
        }

    }
    else if (this.referentObject instanceof AcmeSequence) {
        // then convert the sequence to a set and return an
        // enumeration of its entities.
        // RTM 4/25/98 -- for now just return false.  you can't yet
        //                treat a sequence like a set.
        this.setSetType(ConstantTypes.UNRESOLVABLE_TYPE);
        setTypechecks = false;
        ExpressionTypecheckError err = new ExpressionTypecheckError(this, "Reference is not a set", this);
        AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);

    }
    else {
        // else we have an error of some sort. The referent object is
        // not of a recognized type
        this.setSetType(ConstantTypes.UNRESOLVABLE_TYPE);
        ExpressionTypecheckError err = new ExpressionTypecheckError(this, "Reference is not a set", this);
        AcmeTypeManager.typecheckingErrorStack.pushCausalError(err);
        setTypechecks = false;
    };
    if (setTypechecks && getSetType () instanceof AcmeElementType) {
    	setType (new SetType ("anonymous-set-type", (AcmeElementType )getSetType()));
    }
    
    return setTypechecks;
};


  /**
   * Return "this" node when it is asked to evaluate itself as a set.
   * @return The set expression node
   * @exception DesignRuleEvalException If there is an evaluation error
   */
    public SetExpressionNode evaluateAsSet() throws DesignRuleEvalException {
        return this;
    };

    /**
     * @see acmeLib.designRules.ExpressionNode#cloneExpressionTree()
     */
    public ExpressionNode cloneExpressionTree() {
        SetReferenceNode clone = new SetReferenceNode (this);
        clone.setType = this.setType;
        this.doCloneExpressionTree (clone);
        return clone;
    }

}
