package acmeLib.designRules;

import java.io.IOException;

import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.properties.AcmeString;

/**
 * Class: StringLiteralNode<br>
 * Purpose: Simple expression tree leaf-node that stores a string value<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    10/8/97    RTM     Created<br>
 *    2/2/98     RTM     Modified type storage to store AcmePropertyTypes<br>
 *                        instead of ExpressionTypeT objects.<br>
 *<br>
 */
public class StringLiteralNode extends ExpressionNode
{
/**
 * stores the value of this literal.
 */
  protected String value;

/**
 * JJTree-specific methods
 * @param id The ID
 * @return The node
 */
  public static Node jjtCreate(int id) {
    return new StringLiteralNode(id);
  }

/** 
 * constructors 
 * @param id The ID
 */
  public StringLiteralNode(int id) {
    super(id);
    this.setType(ConstantTypes.STRING_TYPE);
    this.setValue("");       // default to empty string?
  }

  //
  // Methods added by bmonroe
  //

/** 
 * constructors 
 * @param id The ID
 * @param val The value
 */
  public StringLiteralNode(int id, String val) {
    super(id);
    this.setType(ConstantTypes.STRING_TYPE);
    this.setValue(AcmeString.massageForInternalStorage(val));
  }

  /**
   constructors
   @param val the value
  */
  public StringLiteralNode(String val) {
  	super(ArmaniParserTreeConstants.JJTSTRINGLITERALNODE);
    this.setType(ConstantTypes.STRING_TYPE);
    this.setValue(AcmeString.massageForInternalStorage(val));
  }


 /**
 * Copy constructor
 * @param copyOf The StringLiteralNode to be copied
 */
  public StringLiteralNode(StringLiteralNode copyOf) {
    super(copyOf);
    this.setType(copyOf.getType());
    this.setValue(copyOf.getValue());
  }
  

/**
 * cloneTree recursively deep-copies an entire expression tree,
 * creating deep copies of all of it's children nodes in addition
 * to its own fields.  
 * @return The expression node
 */
  public ExpressionNode cloneExpressionTree() {
    StringLiteralNode clone = new StringLiteralNode(this);
    this.doCloneExpressionTree(clone);
    clone.setType(this.getType());
    clone.setValue(this.getValue());
    return clone;
  };


/** 
 * compares for set equality (value equality) 
 * @param test The tested node
 * @return true if the test node is equal to this node
 */
  public boolean equals(Object test) {
    if (!(test instanceof ExpressionNode)) {
        if (test instanceof AcmeString) {
            if (this.getValue() == null) {
                return (test == null);
            }
                return (this.getValue().equals(((AcmeString) test).getValue()));
        }
            return false;
    }
        try {
            String testVal = ((ExpressionNode) test).evaluateAsString();
            if (testVal == null) {
                return (this.getValue() == null);
            }
                return (testVal.equals(this.getValue()));
        } catch (Exception exc) {

        };
            return false;
  };


/**
 * Get the value
 * @return The value
 */
  public String getValue() { return this.value; }
/**
 * Set the value
 * @param newScope The new value
 */
  public void setValue(String val) { this.value = val; };

/**
 * because this node is a literal it must typecheck (it can't store
 * anything but a string).  Furthermore, it can't have any children nodes
 * so there's no need to traverse the children.
 * @return true if type is correct
 */
  public boolean typecheck() {
    this.setType(ConstantTypes.STRING_TYPE);
    return true;
  };


/**
 * because this is always a leaf node, evaluate() just returns
 * the value of the literal.
 * @return The evaluated string
 */
  public String evaluate() { return this.value; };
/**
 * @return The evaluated string
 */
  public String evaluateAsString() { return this.value; };


/**
 * generic exporting method. express this string literal expression as a string.
 * @return The output string
 */
  public String toString() {
      if(this.getInParentheses())
        return "( " + AcmeString.massageForExport(this.getValue()) + " )";
      else
        return AcmeString.massageForExport(this.getValue());
  };


/**
 * @param prefix The prefix
 * @return The output string
 */
  public String toString(String prefix) {
    return prefix + this.toString();
  };

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion (this);
        if (this.getInParentheses()) {
            s.write ("( " + AcmeString.massageForExport(getValue ()) + " )");
        }
        else {
            s.write (AcmeString.massageForExport(getValue ()));
        }
        s.endRegion (this);
    }

}