package acmeLib.designRules;

import java.io.IOException;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.AcmeUnresolvedType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.ArmaniExportable;
import acmeLib.properties.SetType;

// RTM 3/27/98 -- Typechecking of external analysis Set{...} return types
// depends on SetExpressionNode being the first interface declared for
// TransientSet If this changes here then acmeLib.designRules.JavaMethodCallNode
// will need to be updated appropriately (look for similar comment in that file).
/**
 * Class: TransientSet
 * Purpose: Works as a helper class that stores a temporary set of
 *          elements that has been either constructed or created by
 *          application of some set-typed functions.  A TransientSet
 *          can also be an object generated to be the referentObject
 *          of a set reference node.  In general, this class will be
 *          created by a SetExpressionNode to store and manipulate
 *          sets that are derived, generated, or otherwise created.
 *
 * Revision History:
 *    Date:      Person    Change
 *    2/25/98     RTM       Created
 *
 * @author Bob Monroe
 */
public class TransientSet extends ExpressionNode implements SetExpressionNode {

    //
    // protected members
    //

    /**
     * stores the elements of this set
     */
    protected AcmeVector setVector;

    /**
     * stores the type of this entity (ie the entire type of "Set{Foo}")
     */
    protected AcmeType type;

    /**
     * stores the type of entities stored within this set.
     * (ie the "foo" part of "Set{foo}")
     */
    protected AcmeType setType;

    /**
     * stores the set's Scope.  Not guaranteed to be set properly, in general.
     */
    protected acmeLib.designRules.Scope scope = null;

    //
    // public members
    //

    /**
     * constructors
     */
    public TransientSet() {
        super(ArmaniParserTreeConstants.JJTRANSIENTSET);
        this.setVector = new AcmeVector(2, 1);
        this.setSetType(new SetType("tempSet", new AcmeUnresolvedType("unresolved")));
    }

    /**
     * @param newSetVector The new set vector
     */
    public TransientSet(AcmeVector newSetVector) {
        super(ArmaniParserTreeConstants.JJTRANSIENTSET);
        this.setVector = newSetVector;
        this.setSetType(new SetType("tempSet", new AcmeUnresolvedType("unresolved")));
    }

    /**
     * @param newSetType The new set type
     */
    public TransientSet(AcmeType newSetType) {
        super(ArmaniParserTreeConstants.JJTRANSIENTSET);
        this.setVector = new AcmeVector(2, 1);
        this.setSetType(newSetType);
    }

    /**
     * compares for set equality (value equality)
     * @param test The tested object
     * @return true if the tested object is equal to this object
     */
    public boolean equals(Object test) {
        if (test instanceof SetExpressionNode) {
            return DrUtils.equalSets(this, (SetExpressionNode) test);
        }
        return false;
    };

    /** accessors/settors for set expression scope. */
    /**
     * Get the set scope
     * @return The set scope
     */
    public Scope getSetScope() {
        return this.scope;
    };
    /**
     * Set the set scope
     * @param newScope The new scope
     */
    public void setSetScope(Scope newScope) {
        this.scope = newScope;
    };

    /** accessors/settors for overal set type. */
    /**
     * Get the type
     * @return The AcmeType
     */
    public AcmeType getType() {
        return this.type;
    };
    /**
     * Set the type
     * @param newType The new AcmeType
     */
    public void setType(AcmeType newType) {
        this.type = newType;
    };

    /** accessors/settors for type of set contents. */
    /**
     * Get the set type
     * @return The AcmeType
     */
    public AcmeType getSetType() {
        return this.setType;
    };
    /**
     * Set the set type
     * @param newType The new AcmeType
     */
    public void setSetType(AcmeType newSetType) {
        this.setType = newSetType;
    };

    /** accessors/settors for set vector. */
    /**
     * Get the set vector
     * @return The AcmeVector
     */
    public AcmeVector getSetVector() {
        return this.setVector;
    };
    /**
     * Set the set vector
     * @param newVector The new AcmeVector
     */
    public void setSetType(AcmeVector newVector) {
        this.setVector = newVector;
    };

    /**
     * returns the size of the current set
     * @return The size
     */
    public int size() {
        return setVector.getSize();
    }

    /**
     * transient sets can't really be refreshed, so this implements the
     * interface by doing nothing for refreshSet()
     */
    public void refreshSet() {
    };

    /**
     * remove all of the elements from this set
     */
    public void emptySet() {
        this.setVector.removeAllItems();
    };

    /**
     * contains(Object elt) returns true iff elt is a member of this set
     * @param elt The element
     * @return true if element is a member of this set
     */
    public boolean contains(AcmeObject elt) {
        return this.setVector.contains(elt);
    };

    /**
     * addElement(Object elt) adds elt to this set, using set addition
     * semantics.  That is, if elt is already a member of this set then
     * elt will not be re-added.
     * @param elt The element
     */
    public void addElement(AcmeObject elt) {
        // RTM 1/12/98 -- this doesn't do an "equals" comparison, only
        //                a reference comparison (that is, is this exact
        //                object already stored in the vector, not does this
        //                vector contain an equivalent object.)
        if (!(setVector.contains(elt)))
            setVector.addItem(elt);
        // else quietly do nothing, as elt is already in the set.
    };

    /**
     * removeElement(Object elt) removes elt from this set. If elt is
     * not a member of this set then the set will not be changed.
     * @param elt The element
     */
    public void removeElement(AcmeObject elt) {
        // RTM 1/12/98 -- this doesn't do an "equals" comparison, only
        //                a reference comparison (that is, is this exact
        //                object already stored in the vector, not does this
        //                vector contain an equivalent object.)
        setVector.removeItem(elt);
    };

    /**
     * enumerate() returns an enumeration of all of the elements of the set
     * @return The enumeration of all elements in the set
     */
    public AcmeEnumeration enumerate() {
        return this.setVector.enumerate();
    }

    /**
     * return the string representing this set (for exporting/unparsing)
     * @return The output string
     */
    public String toString() {
        String ret = "";
        try {
            AcmeOutputStream s = new AcmeOutputStream();
            exportArmani(s);
            s.flush();
            ret = s.toString();
            s.close();
        }
        catch (IOException e) {
        }
        catch (AcmeException e) {
        }

        return ret;
    };

    /**
     * return the string representing this set
     * @param prefix The prefix
     * @return The output string
     */
    public String toString(String prefix) {
        return prefix + this.toString();
    };

    /**
     * when evaluated as a set, this node returns itself.
     * @return The set expression node
     */
    public SetExpressionNode evaluateAsSet() {
        return this;
    };

    /* (non-Javadoc)
     * @see acmeLib.core.io.ArmaniExportable#exportArmani(acmeLib.core.io.AcmeOutputStream)
     */
    public void exportArmani(AcmeOutputStream s) throws IOException, AcmeException {
        s.beginRegion(this);
        s.write("{");
        AcmeEnumeration e = this.enumerate();
        while (e.moreItems()) {
            AcmeObject nextItem = e.getNextItem();
            if (nextItem instanceof ArmaniExportable) {

                ((ArmaniExportable) nextItem).exportArmani(s);

            }
            else {
                s.write(nextItem.toString());
            }
            if (e.moreItems()) {
                s.write(", ");
            };
        }
        s.write("}");

        s.endRegion(this);
    }

	/**
	 * @return
	 * @see acmeLib.designRules.ExpressionNode#cloneExpressionTree()
	 **/
	public ExpressionNode cloneExpressionTree() {
		TransientSet ts = new TransientSet ();
		doCloneExpressionTree(ts);
		return ts;
	}

}
