
package acmeLib.elements;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.properties.AcmeProperty;

// class Attachment
//
//   The Attachment class defines an association between a Port and
//   Role which are logically attached in a System.  Valid attachments
//   are always a member of an AttachmentMap within a System.


/**
 *  Class:    AcmeAttachment
<BR>
 *  Purpose:  Class that represents an attachment within an ACME system
 <BR>
 *  Description: The Attachment class defines an association between a Port and
<BR>
 *            Role which are logically attached in a System.  Valid attachments
<BR>
 *            are always a member of an AttachmentMap within a System.
<BR><BR>
 *
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    4/3/97    RTM     Translated to Java
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeAttachment extends AcmeElement {

  //
  // Protected members
  //
  protected AcmePort attachedPort;
  protected AcmeRole attachedRole;

  //
  // Public members
  //
  /**
   * This is the constructor for this class
   * @param p AcmePort
   * @param r AcmeRole
   */
  public AcmeAttachment (AcmePort p, AcmeRole r) {
     attachedPort = p;
     attachedRole = r;
  };

  /**
   * @return null
   */
  public acmeLib.designRules.Scope getScope() { return null; };

  /**
   * @return a copy of an AcmeElement
   */
  public AcmeElement copy() {
    AcmeAttachment new_attachment =
        new AcmeAttachment(this.getPort(), this.getRole());
    this.copyProperties(new_attachment);
    this.copyRepresentations(new_attachment);
    return new_attachment;
  };


  /**
   * This method tests for equality of this and test.  Because an attachment is
   * just a relation, if the (port,role) pairs of this and test
   * refer to the same ports and roles, then test == this.
   * @param test an Object to be compared with "this'
   * @return whether "this" and test are equal
   */
  public boolean equals(Object test) {
    if (!(test instanceof AcmeAttachment)) {
        return false;
    }
        AcmeAttachment testA = (AcmeAttachment) test;
        return (this.getPort() == testA.getPort())
                && (this.getRole() == testA.getRole());
  };


  /**
   * This method returns the value of attachedPort
   * @return attachedPort
   */
  public AcmePort getPort() { return attachedPort; };
  /**
   * This method sets the value of attachedPort
   * @param p the new value of attachedPort
   */
  public void setPort (AcmePort p) { attachedPort = p; };
  /**
   * This method returns the value of attachedRole
   * @return attachedRole
   */
  public AcmeRole getRole() { return attachedRole; };
  /**
   * This method sets the value of attachedRole
   * @param r the new value of attachedRole
   */
  public void setRole (AcmeRole r) { attachedRole = r; };


  /**
   * @return the name of the attached port
   */
  public String getPortNameAsString() {
    StringBuffer nameBuf = new StringBuffer();
    if (attachedPort != null) {
        if (attachedPort.getParent() != null) {
            nameBuf.append(attachedPort.getParent().getName() + ".");
        }
        nameBuf.append(attachedPort.getName());
    }
    return nameBuf.toString();
  };


 /**
   * @return the name of the attached role
   */
  public String getRoleNameAsString() {
    StringBuffer nameBuf = new StringBuffer();
    if (attachedRole != null) {
        if (attachedRole.getParent() != null) {
            nameBuf.append(attachedRole.getParent().getName() + ".");
        }
        nameBuf.append(attachedRole.getName());
    }
    return nameBuf.toString();
  };


  //
  // exporting
  //

/**
 * @param s AcmeOutputStream
 * @param exportLang ExportLanguage
 * @exception java.io.IOException when error occurs when writing
 */
protected void export(AcmeOutputStream s, ExportLanguage exportLang) throws java.io.IOException {
    s.beginRegion(this);
    s.write("Attachment ");
    s.write(attachedPort.getParent().getName());
    s.write(".");
    s.write(attachedPort.getName());
    s.write(" to ");

    s.write(attachedRole.getParent().getName());
    s.write(".");
    s.write(attachedRole.getName());
    if (this.getNumberOfProperties() > 0) {
        s.writeln("");
        s.indent();
        s.writeln("{ ");
        AcmeEnumeration propsEnum = this.getProperties();
        propsEnum.sortByCreationOrder();
        while (propsEnum.moreItems()) {
            AcmeProperty curProp = (AcmeProperty) propsEnum.getNextItem();
            if (exportLang == ExportLanguage.ARMANI) {
                curProp.exportArmani(s);
            }
            else {
                curProp.exportAcme(s);
            };
        };
        s.write("}");
        s.unIndent();
    };
    s.writeln(";");
    s.endRegion(this);
};

  /**
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs when writing
   * @exception AcmeException for acme exception
   */
  public void exportAcme(AcmeOutputStream s)
    throws java.io.IOException, AcmeException
  {
    this.export(s, ExportLanguage.ACME);
  };


  /**
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs when writing
   * @exception AcmeException for acme exception
   */
  public void exportArmani(AcmeOutputStream s)
     throws java.io.IOException, AcmeException
  {
    exportComment (s);
    this.export(s, ExportLanguage.ARMANI);
  }



};
