package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeDuplicateNameException;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.exceptions.AcmeKeywordNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.ElementTypecheckError;
import acmeLib.errorManagement.TypeInstantiationError;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 *  Class:    AcmeComponent
<BR>
 *  Purpose:  Stores the structure of an ACME component
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    4/24/97   RTM     Translated to Java
<BR>
 *    5/30/97   RTM     All stubbed methods filled in
<BR>
 *    6/13/97   RTM     Modified to work with second generation object model
<BR>
 *    1/14/98   RTM     added support for multiple declared types.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeComponent extends AcmeElement {

	//
	// Private members
	//
	private AcmeMap portMap;

	//
	// Protected members
	//

	/**
	 * This method does a deep copy of the ports in this component and puts
	 * them into the ports list of destComp.  This removes
	 * any ports that are currently in destComp
	 * This method also
	 * - gets the properties of this element,
	 * - makes a new property map in dest_obj
	 * - for each property in this elt, puts a deep copy of it in dest_obj.
	 * @param destComp AcmeComponent
	 */
	protected void copyPorts(AcmeComponent destComp) {
		// - get the properties of this element,
		// - make a new property map in dest_obj
		// - foreach property in this elt, put a deep copy of it in dest_obj.
		AcmeEnumeration ports_enum = this.getPorts();
		destComp.portMap = new AcmeMap();

		AcmePort cur_port;
		while (ports_enum.moreItems()) {
			cur_port = (AcmePort) ports_enum.getNextItem();
			try {
				destComp.addPort((AcmePort) cur_port.copy());
			}
			catch (AcmeIllegalNameException ainExc) {
				// if we're copying a port that has been legally added elsewhere
				// into a fresh portmap then it won't have an illegal name.
			};
		}
	};

	//
	// Public members
	//
	/**
	 * This is a constructor for this class
	 * @param name used to initialize this.name
	 */
	public AcmeComponent(String name) {
		portMap = new AcmeMap();
		this.name = name;
	};

	/**
	  * This is a constructor for this class
	  * @param name used to initialize this.name
	  * @param type AcmeComponentType
	  */
	public AcmeComponent(String name, AcmeComponentType type) {
		portMap = new AcmeMap();
		this.name = name;
		addDeclaredType(type);
	};

	/**
	 * This method does a deep copy of "this" component
	 * @return the copy of the AcmeElement
	 */
	public AcmeElement copy() {
		AcmeComponent new_comp = new AcmeComponent(this.getName());
		this.copyPorts(new_comp);
		this.copyCommonEltStructure(new_comp);
		return new_comp;
	};

	/**
	 * This method uses this element instance as a prototype to create a new type.
	 * @param newTypeName used to create a new  AcmeComponentType
	 * @return the new AcmeComponentType, created by this function
	 */
	public AcmeElementType copyAsPrototype(String newTypeName) {
		AcmeComponentType new_comp_type = new AcmeComponentType(newTypeName);
		this.copyPorts(new_comp_type);
		this.copyCommonEltStructure(new_comp_type);

		// copyCommonEltStructure overwrites the name of the new type, so
		// reset it properly.  Don't check for legal names here, that will
		// happen if/when we try to add it to anything.
		new_comp_type.name = newTypeName;

		// add all of the declared types of this instance as supertypes of
		// the new type.
		AcmeEnumeration e = this.getDeclaredTypes();
		while (e.moreItems()) {
			AcmeObject curType = e.getNextItem();

			if (curType instanceof AcmeComponentType) {
				new_comp_type.addSuperType((AcmeComponentType) curType);
			}
			else if (curType instanceof AcmeElementProtoType) {
				new_comp_type.addSuperType((AcmeElementProtoType) curType);
			}
			else if (curType instanceof AcmeElementTypeRef) {
				AcmeElementTypeRef t = (AcmeElementTypeRef) curType;
				if (t.getReferentType() instanceof AcmeComponentType
					|| t.getReferentType() instanceof AcmeElementProtoType) {
					new_comp_type.addSuperType((AcmeElementTypeRef) curType);
				}

			}

		};

		return new_comp_type;
	};

	/**
	 * This method updates the port name in portMap, if it's a port, otherwise just
	 * lets the superclass take care of it.
	 * @param oldName the name of the old port
	 * @param obj AcmeObject
	 * @exception AcmeIllegalNameException thrown when the name of the port is not valid
	 */
	public void childNameChanged(String oldName, AcmeObject obj) throws AcmeIllegalNameException {
		if (obj instanceof AcmePort) {
			String newName = ((AcmePort) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a " + "port name to null.");
			}
			else if (!newName.equals(oldName)) {
				// remove the oldName from the name map, readd object with new name
				AcmePort tmp_port = this.getPort(oldName);
				if (tmp_port != null) {
					eltNamespace.removeName(oldName);
					tmp_port.setParent(null);
					portMap.removeItem(oldName);
				};

				this.addPort((AcmePort) obj);
			};
			// else the name hasn't really changed, just return.
		}
		else {
			super.childNameChanged(oldName, obj);
		};
	};

	//
	// Port manipulations
	//
	/**
	 * This method adds a port to the component
	 * @param p AcmePort to be added
	 * @exception AcmeIllegalNameException thrown when the name of the port is illegal
	 */
	public void addPort(AcmePort p) throws AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isAcmeKeyword(p.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + p.getName() + "\" used for port name.");
		};

		if (p.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a port with null name.");
		}

		else if (eltNamespace.containsName(p.getName())) {
			throw new AcmeDuplicateNameException(
				"Attempted to redefine identifier " + p.getName() + " as an added port.");
		};

		eltNamespace.addName(p.getName());
		portMap.addItem(p.getName(), p);
		p.setParent(this);
	};

	/**
	  * This method removes a port from the component
	  * @param p AcmePort to be removed
	  */
	public void removePort(AcmePort p) {
		removePort(p.getName());
	};

	/**
	 * This method removes a port from the component
	 * @param name the name of the port to be removed
	 */
	public void removePort(String name) {
		AcmePort p = (AcmePort) portMap.lookupItem(name);
		if (p == null)
			return; // there's no port called "name" to remove
		else {
			this.detachPort(p);
			this.unbindPort(p);
			portMap.removeItem(name);
			eltNamespace.removeName(name);
			p.setParent(null);
		};
	};

	/**
	 * This method returns an enumeration of all ports
	 * @return an AcmeEnumeration of the portMap
	 */
	public AcmeEnumeration getPorts() {
		return portMap.enumerate();
	};

	/**
	 * This method returns a port with a specific name
	 * @param name the name of the port
	 * @return AcmePort corresponding to name
	 */
	public AcmePort getPort(String name) {
		return (AcmePort) portMap.lookupItem(name);
	};

	/**
	 * Detach this port from all other entities in this system.  This method
	 * removes all of the attachments that this port is associated with.
	 * @param port The port to detach from the rest of the system
	 */
	public void detachPort(AcmePort port) {
		if (this.getParent() instanceof AcmeSystem) {
			AcmeSystem parentSys = (AcmeSystem) this.getParent();
			AcmeEnumeration attachments = parentSys.getAttachments(port);
			while (attachments.moreItems()) {
				AcmeAttachment curAttachment = (AcmeAttachment) attachments.getNextItem();
				parentSys.removeAttachment(curAttachment);
			};
		};
	};

	/**
	 * Remove all bindings that this port participates in as a precursor to
	 * deleting the port.
	 * @param port The port to unbind
	 */
	public void unbindPort(AcmePort port) {
		AcmeEnumeration reps = this.getRepresentations();
		while (reps.moreItems()) {
			AcmeRepresentation curRep = (AcmeRepresentation) reps.getNextItem();
			if (curRep.getAbstractionMap() instanceof AcmeBindingsMap) {
				AcmeBindingsMap curMap = (AcmeBindingsMap) curRep.getAbstractionMap();
				AcmeEnumeration bindings = curMap.getBindings();
				while (bindings.moreItems()) {
					AcmeBinding curBinding = (AcmeBinding) bindings.getNextItem();
					// HACK: Need to fix copy so that it copies references properly
					if (curBinding.getOuter() == port) {
						//                if (curBinding.getOuter().getQualifiedName().equals (port.getQualifiedName ())) {
						curMap.removeBinding(curBinding);
					};
				};
			};
		};
	};

	/**
	 * All AcmeComponents satisfy the type ConstantTypes.COMPONENT_BASE_TYPE.
	 * If the generic supertype method doesn't find the type declared, check
	 * whether the type is the COMPONENT_BASE_TYPE.
	 * @param type the type of the AcmeElement
	 * @return whether the type is COMPONENT_BASE_TYPE
	 */
	public boolean declaresType(AcmeElementType type) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType ();
		}
		if (type == ConstantTypes.COMPONENT_BASE_TYPE) {
			return true;
		}
		return super.declaresType(type);
	};

	//
	// exporting
	//

	/**
	 * @return  AcmeEnumeration
	 */
	protected AcmeEnumeration getChildrenToExport() {
		AcmeEnumeration retEnum = super.getChildrenToExport();
		retEnum.concatenate(this.getPorts());
		return retEnum;
	}

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Component", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception  AcmeException for acme exception
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Component", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this component as a string with the
	 *  form "Component <name> [ : <Typename>+ ]".  For example:
	 *  "Component foo : compT".
	 * @return string containing this element's type signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Component " + this.getName());
		AcmeEnumeration typeEnum = this.getDeclaredTypes();
		if (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" : <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" : " + eltType.getTypeName());
		}

		while (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * This method implements the acmeLib.designRules.Scope interface to map names
	 * to elements or properties.
	 * It looks up the child object called 'name'.
	 * First checks for ports, then has AcmeElement check for properties,
	 * and representations.
	 * @param name compared with 'ports'
	 * @param bubbleUp boolean
	 * @return the Object corresponding to name
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		// lookup the child object called 'name'.
		// First check for ports, then have AcmeElement check for properties,
		// and representations.
		Object retObj = null;

		// check for the special case of a request for all of this component's ports.
		if (name != null) {
			if (name.equalsIgnoreCase("ports"))
				return this.getPorts();
			else if (name.equalsIgnoreCase("representations")) {
				// BRS: Representations is a bit of a tricky thing. Since each representation can only
				// have one system, I will form an enumeration out of the systems and return these...
				AcmeEnumeration reps = this.getRepresentations();
				AcmeVector sys = new AcmeVector();
				while (reps.moreItems()) {
					sys.addItem(((AcmeRepresentation) reps.getNextItem()).getSystem());
				}

				return sys.enumerate();
			}

		};

		if (retObj == null) {
			retObj = this.getPort(name);
		}

		if (retObj == null) {
			// just check for a property or representation, bubbling up occurs in
			// the generic element lookupName method.
			retObj = super.lookupName(name, bubbleUp);
		}

		return retObj;
	};

	/**
	 *  AcmeComponent.resolveTypeReferences() relies primarily on it's superclass'
	 *  implementation to do type name resolution.  This subclass adds a
	 *  recursive call to resolveTypeReferences() for each of this Component's ports.
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
		AcmeEnumeration e = this.getPorts();
		while (e.moreItems()) {
			((AcmePort) e.getNextItem()).resolveTypeReferences();
		};
	};

	/**
	 * Recursively resolve all of the instance references of this
	 * system's substructure
	 */
	public void resolveInstanceReferences() {
		super.resolveInstanceReferences();
		AcmeEnumeration e = this.getPorts();
		while (e.moreItems()) {
			((AcmePort) e.getNextItem()).resolveInstanceReferences();
		};
	};

	/**
	 * instantiateElements() walks over the vector of instantiatedTypes of this
	 * element and adds the minimal structure declared in the type description
	 * to this element.  It delegates much of the work to the AcmeTypeManager
	 * class, so that we can change type instantiation policies relatively
	 * centrally (ie by changing the AcmeTypeManager methods) for all elements
	 * at once.  This method recursively calls the same method on all of the
	 * substructure of this element instance.
	 */
	public void instantiateElements() {

		// unify this element with each of the types it instantiates.
		for (int i = 0; i < this.instantiatedTypes.getSize(); i++) {
			// check that we have a prototype component or component type
			// to instantiate.  If we don't then announce an error and move
			// to the next type.
			AcmeObject iType = this.instantiatedTypes.getItemAt(i);
			if (iType instanceof AcmeElementTypeRef) {
				iType = (AcmeObject) ((AcmeElementTypeRef) iType).getReferentType();
			}
			if (iType instanceof AcmeComponentType) {
				AcmeTypeManager.unifyComponent((AcmeComponentType) iType, this);

			}
			else if (iType instanceof AcmeElementProtoType) {
				AcmeTypeManager.unifyElementWithPrototype((AcmeElementProtoType) iType, this);

			}
			else if (iType instanceof AcmeType) {
				String errorMsg = null;
				if (iType instanceof AcmeUnresolvedElementType) {
					errorMsg =
						"Instantiation Error.  Unable to instantiate "
							+ " unresolved element type: "
							+ ((AcmeUnresolvedElementType) iType).getTypeName()
							+ " in component: "
							+ this.getName();
				}
				else {
					errorMsg =
						"Instantiation Error.  Type: "
							+ ((AcmeType) iType).getTypeName()
							+ " is neither a component nor an element type.  Therefore it "
							+ "can not be used to instantiate component: "
							+ this.getName();
				};

				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg, (AcmeType) iType);
				AcmeErrorManager.errorDetected(instError);

			}
			else {
				String errorMsg =
					"Instantiation Error.  Attempted to instantiate a "
						+ "completely unrecognizable type in component: "
						+ this.getName();
				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg);
				AcmeErrorManager.errorDetected(instError);
			};
		}

		// then have each of its ports instantiate themselves.
		AcmeEnumeration e;
		e = this.getPorts();
		while (e.moreItems()) {
			((AcmePort) e.getNextItem()).instantiateElements();
		};

		// finally, call AcmeElement to have it instantiate representations and any
		// other standard element structure.
		super.instantiateElements();
	};

	/**
	 * This method confirms that this component satisfies all of its declared types and
	 * design rules, and that it's properties, representations, and ports
	 * all typecheck also.
	 * @return whether the component satisfies all of its declared types
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckComponent(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			AcmeErrorManager.errorRepaired(this, null);
			return true;
		}
		// component doesn't typecheck.  There should be an error at the
		// top of the AcmeTypeManager's typecheckingErrorStack.  If there
		// is such an error then announce it to the AcmeErrorManager and
		// return false.
		AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
		if (eltTypeError != null) {
			AcmeErrorManager.errorDetected(eltTypeError);

			return false;
		}
		// no error was reported.  something went wrong?
		// announce a generic error.
		ElementTypecheckError err =
			new ElementTypecheckError(this, "Component " + this.getName() + " does not typecheck.", this);
		AcmeErrorManager.errorDetected(err);
		return false;
	};

	/**
	 * Creates a workshop for editing this element.
	 * implements the acmeLib.Editable interface.
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editComponent(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

	public void view() {
		System.out.println("Attempted to view: " + this.getName());
		System.out.println("  but AcmeComponent.view() is not yet implemented.");
	}

};
