package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.*;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 *  Class:  AcmeConnector
 <BR>
 *  Purpose: Stores the structure of an ACME connector
 <BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    4/24/97   RTM     Translated to Java
<BR>
 *    5/30/97   RTM     All stubbed methods filled in
<BR>
 *    6/16/97   RTM     Modified to work with second generation object model
<BR>
 *    1/14/98   RTM     added support for multiple declared types.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeConnector extends AcmeElement {

	//
	// Private members
	//
	private AcmeMap roleMap;

	//
	// Protected members
	//

	/**
	 * This method does a deep copy of the roles in this connector and put
	 * them into the roles list of destConn.  This removes
	 * any roles that are currently in destConn
	 * @param destConn AcmeConnector
	 */
	protected void copyRoles(AcmeConnector destConn) {
		AcmeEnumeration roles_enum = this.getRoles();
		destConn.roleMap = new AcmeMap();

		AcmeRole cur_role;
		while (roles_enum.moreItems()) {
			cur_role = (AcmeRole) roles_enum.getNextItem();
			try {
				destConn.addRole((AcmeRole) cur_role.copy());
			}
			catch (AcmeIllegalNameException ainExc) {
				// if we're copying a port that has been legally added elsewhere
				// into a fresh portmap then it won't have an illegal name.
			};

		}
	};

	//
	// Public members
	//
	/**
	 * This is a constructor for this class
	 * @param name for the AcmeElement
	 */
	public AcmeConnector(String name) {
		roleMap = new AcmeMap();
		this.name = name;
	};

	/**
	 * This is a constructor for this class
	 * @param name for the AcmeElement
	 * @param type AcmeConnectorType
	 */
	public AcmeConnector(String name, AcmeConnectorType type) {
		roleMap = new AcmeMap();
		this.name = name;
		addDeclaredType(type);
	};

	/**
	 * This method does a deep copy of "this" connector
	 * @return a copy of "this" connector
	 */
	public AcmeElement copy() {
		AcmeConnector new_conn = new AcmeConnector(this.getName());
		this.copyRoles(new_conn);
		this.copyCommonEltStructure(new_conn);
		return new_conn;
	};

	/**
	 * This method uses this element instance as a prototype to create a new type.
	 * newTypeName the name of the AcmeConnectorType
	 * @param newTypeName the name of the new copy
	 * @return the copy of the connector
	 */
	public AcmeElementType copyAsPrototype(String newTypeName) {
		AcmeVector super_types;
		AcmeConnectorType new_conn_type = new AcmeConnectorType(newTypeName);
		this.copyRoles(new_conn_type);
		this.copyCommonEltStructure(new_conn_type);

		// copyCommonEltStructure overwrites the name of the new type, so
		// reset it properly.  Don't check for legal names here, that will
		// happen if/when we try to add it to anything.
		new_conn_type.name = newTypeName;

		// add all of the declared types of this instance as supertypes of
		// the new type.
		AcmeEnumeration e = this.getDeclaredTypes();
		while (e.moreItems()) {
			AcmeObject curType = e.getNextItem();
			if (curType instanceof AcmeConnectorType) {
				new_conn_type.addSuperType((AcmeConnectorType) curType);
			}
			else if (curType instanceof AcmeElementProtoType) {
				new_conn_type.addSuperType((AcmeElementProtoType) curType);
			}
			else if (curType instanceof AcmeElementTypeRef) {
				AcmeElementTypeRef t = (AcmeElementTypeRef) curType;
				if (t.getReferentType() instanceof AcmeConnectorType
					|| t.getReferentType() instanceof AcmeElementProtoType) {
					new_conn_type.addSuperType((AcmeElementTypeRef) curType);
				}

			}
		};

		return new_conn_type;
	};

	/**
	 * This method updates the role name in roleMap, if it's a role, otherwise just
	 * lets the superclass take care of it.
	 * @param oldName the name of the role to be removed
	 * @param obj AcmeObject, which is the object with the new name
	 * @exception AcmeIllegalNameException connector name is illegal
	 */
	public void childNameChanged(String oldName, AcmeObject obj) throws AcmeIllegalNameException {
		if (obj instanceof AcmeRole) {
			String newName = ((AcmeRole) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a role name to null.");
			}
			else if (!newName.equals(oldName)) {
				// remove the oldName from the name map, readd object with new name
				AcmeRole tmp_role = this.getRole(oldName);
				if (tmp_role != null) {
					eltNamespace.removeName(oldName);
					tmp_role.setParent(null);
					roleMap.removeItem(oldName);
				};
				this.addRole((AcmeRole) obj);
			};
			// else the name hasn't really changed, just return.
		}
		else {
			super.childNameChanged(oldName, obj);
		};
	};

	//
	// Role manipulations
	//
	/**
	 * @param r AcmeRole to be added
	 * @exception AcmeIllegalNameException role name is illegal
	 */
	public void addRole(AcmeRole r) throws AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isAcmeKeyword(r.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + r.getName() + "\" used for role name.");
		};

		if (r.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a role with null name.");
		}

		else if (eltNamespace.containsName(r.getName())) {
			throw new AcmeDuplicateNameException(
				"Attempted to redefine identifier " + r.getName() + " as an added role.");
		};

		eltNamespace.addName(r.getName());
		roleMap.addItem(r.getName(), r);
		r.setParent(this);
	};

	/**
	 * This method removes a role from this connector
	 * @param r AcmeRole to be removed
	 */
	public void removeRole(AcmeRole r) {
		removeRole(r.getName());
	};

	/**
	 * This method removes a role from this connector
	 * @param name name of AcmeRole to be removed
	 */
	public void removeRole(String name) {
		AcmeRole r = (AcmeRole) roleMap.lookupItem(name);
		if (r == null)
			return; // there's no role called "name" to remove
		else {
			this.detachRole(r);
			this.unbindRole(r);
			roleMap.removeItem(name);
			eltNamespace.removeName(name);
			r.setParent(null);
		};
	};

	/**
	 * This method returns an enumeration of the roles
	 * @return AcmeEnumeration of the roleMap
	 */
	public AcmeEnumeration getRoles() {
		return roleMap.enumerate();
	};

	/**
	 * This method returns a role with a specific name
	 * @param name name of the AcmeRole item
	 * @return AcmeRole corresponding to name
	 */
	public AcmeRole getRole(String name) {
		return (AcmeRole) roleMap.lookupItem(name);
	};

	/**
	 * remove all of the attachments and binding objects that this role
	 * is associated with.
	 * @param role The role to detach from the rest of the world
	 */
	public void detachRole(AcmeRole role) {
		if (this.getParent() instanceof AcmeSystem) {
			AcmeSystem parentSys = (AcmeSystem) this.getParent();
			AcmeEnumeration attachments = parentSys.getAttachments(role);
			while (attachments.moreItems()) {
				AcmeAttachment curAttachment = (AcmeAttachment) attachments.getNextItem();
				parentSys.removeAttachment(curAttachment);
			};
		};
	};

	/**
	 * Remove all bindings that this role participates in as a precursor to
	 * deleting the role.
	 * @param role The role to unbind
	 */
	public void unbindRole(AcmeRole role) {
		AcmeEnumeration reps = this.getRepresentations();
		while (reps.moreItems()) {
			AcmeRepresentation curRep = (AcmeRepresentation) reps.getNextItem();
			if (curRep.getAbstractionMap() instanceof AcmeBindingsMap) {
				AcmeBindingsMap curMap = (AcmeBindingsMap) curRep.getAbstractionMap();
				AcmeEnumeration bindings = curMap.getBindings();
				while (bindings.moreItems()) {
					AcmeBinding curBinding = (AcmeBinding) bindings.getNextItem();
					if (curBinding.getOuter() == role) {
						curMap.removeBinding(curBinding);
					};
				};
			};
		};
	};

	/** All AcmeConnectors satisfy the type ConstantTypes.CONNECTOR_BASE_TYPE.
	 *  if the generic supertype method doesn't find the type declared, check
	 *  whether the type is the CONNECTOR_BASE_TYPEe.
	 *  @param type the type of an AcmeElement
	 *  @return whether type is CONNECTOR_BASE_TYPE
	 */
	public boolean declaresType(AcmeElementType type) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType ();
		}
		if (type == ConstantTypes.CONNECTOR_BASE_TYPE) {
			return true;
		}
		return super.declaresType(type);
	};

	//
	// exporting
	//

	/**
	 * @return AcmeEnumeration
	 */
	protected AcmeEnumeration getChildrenToExport() {
		AcmeEnumeration retEnum = super.getChildrenToExport();
		retEnum.concatenate(this.getRoles());
		return retEnum;
	}

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Connector", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Connector", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this connector as a string with the
	 *  form "Connector <name> [ : <Typename>+ ]".  For example:
	 *  "Connector foo : portT".
	 * @return string containing this connector's type signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Connector " + this.getName());
		AcmeEnumeration typeEnum = this.getDeclaredTypes();
		if (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" : <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" : " + eltType.getTypeName());
		}

		while (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * This method implements the acmeLib.designRules.Scope interface to map names
	 * to elements or properties
	 * It  lookups the child object called 'name'.
	 * First it checks for roles, then has AcmeElement check for properties,
	 * and representations.
	 * @param name is compared with "roles"
	 * @param bubbleUp boolean
	 * @return Object corresponding to name
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		// lookup the child object called 'name'.
		// First check for roles, then have AcmeElement check for properties,
		// and representations.
		Object retObj = null;

		// check for the special case of a request for all of this connector's roles
		if (name != null) {
			if (name.equalsIgnoreCase("roles")) {
				return this.getRoles();
			}
			else if (name.equalsIgnoreCase("representations")) {
				// BRS: Representations is a bit of a tricky thing. Since each representation can only
				// have one system, I will form an enumeration out of the systems and return these...
				AcmeEnumeration reps = this.getRepresentations();
				AcmeVector sys = new AcmeVector();
				while (reps.moreItems()) {
					sys.addItem(((AcmeRepresentation) reps.getNextItem()).getSystem());
				}

				return sys.enumerate();
			}
		};

		if (retObj == null) {
			retObj = this.getRole(name);
		}

		if (retObj == null)
			// just check for a property or representation, don't bubble up in
			// the generic element lookupName method.
			retObj = super.lookupName(name, bubbleUp);

		return retObj;
	};

	/**
	 *  AcmeConnector.resolveTypeReferences() relies primarily on it's superclass'
	 *  implementation to do type name resolution.  This subclass adds a recursive
	 *  call to resolveTypeReferences() for each of this Connector's roles.
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
		AcmeEnumeration e = this.getRoles();
		while (e.moreItems()) {
			((AcmeRole) e.getNextItem()).resolveTypeReferences();
		};
	};

	/**
	 * Recursively resolve all of the instance references of this
	 * system's substructure
	 */
	public void resolveInstanceReferences() {
		AcmeEnumeration e = this.getRoles();
		while (e.moreItems()) {
			((AcmeRole) e.getNextItem()).resolveInstanceReferences();
		};
		super.resolveInstanceReferences();
	};

	/**
	 *  instantiateElements() walks over the vector of instantiatedTypes of this
	 *  element and adds the minimal structure declared in the type description
	 *  to this element.  It delegates much of the work to the AcmeTypeManager
	 *  class, so that we can change type instantiation policies relatively
	 *  centrally (ie by changing the AcmeTypeManager methods) for all elements
	 *  at once.  This method recursively calls the same method on all of the
	 *  substructure of this element instance.
	 */
	public void instantiateElements() {

		// unify this element with each of the types it instantiates.
		for (int i = 0; i < this.instantiatedTypes.getSize(); i++) {
			// check that we have a prototype connector or connector type
			// to instantiate.  If we don't then announce an error and move
			// to the next type.
			AcmeObject iType = this.instantiatedTypes.getItemAt(i);
			if (iType instanceof AcmeElementTypeRef) {
				iType = (AcmeObject) ((AcmeElementTypeRef) iType).getReferentType();
			}
			if (iType instanceof AcmeConnectorType) {
				AcmeTypeManager.unifyConnector((AcmeConnectorType) iType, this);

			}
			else if (iType instanceof AcmeElementProtoType) {
				AcmeTypeManager.unifyElementWithPrototype((AcmeElementProtoType) iType, this);

			}
			else if (iType instanceof AcmeType) {
				String errorMsg = null;
				if (iType instanceof AcmeUnresolvedElementType) {
					errorMsg =
						"Instantiation Error.  Unable to instantiate "
							+ " unresolved element type: "
							+ ((AcmeUnresolvedElementType) iType).getTypeName()
							+ " in connector: "
							+ this.getName();
				}
				else {
					errorMsg =
						"Instantiation Error.  Type: "
							+ ((AcmeType) iType).getTypeName()
							+ " is not a connector type.  Therefore it can not be "
							+ "used to instantiate connector: "
							+ this.getName();
				};

				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg, (AcmeType) iType);
				AcmeErrorManager.errorDetected(instError);

			}
			else {
				String errorMsg =
					"Instantiation Error.  Attempted to instantiate a "
						+ "completely unrecognizable type in connector: "
						+ this.getName();
				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg);
				AcmeErrorManager.errorDetected(instError);
			};
		}

		// then have each of its roles instantiate themselves.
		AcmeEnumeration e;
		e = this.getRoles();
		while (e.moreItems()) {
			((AcmeRole) e.getNextItem()).instantiateElements();
		};

		// finally, call AcmeElement to have it instantiate representations and any
		// other standard element structure.
		super.instantiateElements();
	};

	/**
	 * This method determines whether this connector satisfies all of its declared types and
	 * design rules, and that it's properties, representations, and roles
	 * all typecheck also.
	 * @return whether connector typechecks
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckConnector(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			AcmeErrorManager.errorRepaired(this, null);
			return true;
		}
		// connector doesn't typecheck.  There should be an error at the
		// top of the AcmeTypeManager's typecheckingErrorStack.  If there
		// is such an error then announce it to the AcmeErrorManager and
		// return false.
		AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
		if (eltTypeError != null) {
			AcmeErrorManager.errorDetected(eltTypeError);
			return false;
		}
		// no error was reported.  something went wrong?
		// announce a generic error.
		AcmeErrorManager.errorDetected(
			new ElementTypecheckError(this, "Connector " + this.getName() + " does not typecheck.", this));
		return false;
	};

	/**
	 * This method implements the acmeLib.Editable interface.
	 * It also creates a workshop for editing this element
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editConnector(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

	public void view() {
		System.out.println("Attempted to view: " + this.getName());
		System.out.println("  but AcmeConnector.view() is not yet implemented.");
	}

};
