package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 *  Class:  AcmeConnectorType
<BR>
 *  Purpose: Stores the type specification for a connector.  This is
<BR>
 *          a subclass of an AcmeConnector because the Acme type system
<BR>
 *          is prototype-based.  As a result, a type is just a prototypical
<BR>
 *          instance of a connector that specifies the aspects that an
<BR>
 *          instance of that type must posses.  Typechecking is done
<BR>
 *          by an external class, AcmeTypeManager, that understands the
<BR>
 *          rules of typechecking
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    6/16/97   RTM     Created class
<BR>
 *    1/14/98   RTM     added support for multiple declared types.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeConnectorType extends AcmeConnector implements AcmeElementType {

	//
	// private members
	//
	AcmeVector superTypes;

	//
	// public members
	//
	/**
	 * This method  creates the prototype object in the superclass constructor
	 * @param typeName the name of the AcmeConnector
	 */
	public AcmeConnectorType(String typeName) {
		// create the prototype object in the superclass constructor
		super(typeName, null);
		this.superTypes = new AcmeVector();
	};

	/**
	 * This method returns the name of the connector
	 * @return the name of the AcmeConnector
	 */
	public String getTypeName() {
		return this.getName();
	};
	/**
	 * This method sets the connector's name
	 * @param typeName used to sed the name of "this"
	 * @exception AcmeIllegalNameException thrown when typeName is illegal
	 */
	public void setTypeName(String typeName) throws AcmeIllegalNameException {
		this.setName(typeName);
	};

	// methods for adding supertypes to a type description and retrieving
	// a type's set of supertypes
	/**
	 * This method is used for adding supertypes to a type description
	 * @param superType AcmeElementType
	 */
	public void addSuperType(AcmeElementType superType) {
		if (superType instanceof AcmeElementTypeRef) {
			superType = ((AcmeElementTypeRef) superType).getReferentType();
		}
		if ((superType instanceof AcmeConnectorType)
			|| (superType instanceof AcmeUnresolvedElementType)
			|| (superType instanceof AcmeElementProtoType)) {
			superTypes.addItem((AcmeObject) superType);
		}
		// else throw an exception?
	};

	/**
	 * This method is used for removing supertypes from a type description
	 * @param superType AcmeElementType
	 */
	public void removeSuperType(AcmeElementType superType) {
		superTypes.removeItem((AcmeObject) superType);
		if (superType instanceof AcmeElementTypeRef) {
			declaredTypes.removeItem((AcmeObject) ((AcmeElementTypeRef) superType).getReferentType());
		}
	};

	/**
	 * This method is used for retrieving a type's set of supertypes
	 * @return AcmeVector of the superTypes
	 */
	public AcmeVector getSuperTypes() {
		return this.superTypes;
	};

	/**
	 * This method makes a new connector type object, with a deep copy of the old
	 * connector's properties, representations, declared, instantiated,
	 * and super-type vectors.
	 * @return a copy of an AcmeElementType
	 */
	public AcmeElementType copyType() {
		AcmeConnectorType new_conn_type = new AcmeConnectorType(this.getTypeName());
		this.copyRoles(new_conn_type);
		this.copyCommonEltStructure(new_conn_type);
		this.superTypes.copyItems(new_conn_type.superTypes);
		return new_conn_type;
	};

	/**
	 * resolveTypeReferences searches through this type object's vector of supertypes
	 * for UnresolvedElementType objects, looks up the actual type that is
	 * referred to by each UnresolvedElementType object, and replaces that
	 * object with the actual type object that should be referred to.
	 * it then calls resolveTypeReferences in its superclass to resolve the type
	 *  names of all of its substructure (roles and representations).
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
		AcmeElementType curType = null;
		AcmeType resolvedType = null;
		AcmeUnresolvedElementType curUnresolvedType;
		for (int i = 0; i < this.superTypes.getSize(); i++) {
			curType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (curType instanceof AcmeUnresolvedElementType) {
				curUnresolvedType = (AcmeUnresolvedElementType) curType;

				if (curUnresolvedType.isQualified()) {
					if (this.getParentDesign() != null) {
						AcmeFamily qFam = this.getParentDesign().getFamily(curUnresolvedType.getQualifier());
						if (qFam != null) {
							AcmeType refType = qFam.getType(curUnresolvedType.getTypeName());
							if (refType != null && refType instanceof AcmeElementType) {
								resolvedType = new AcmeElementTypeRef(qFam, (AcmeElementType) refType);
							}
						}
					}
				}
				else { // first check to see if the type is declared in the same
					// family that this type is declared in.  If an AcmeConnectorType
					// object's parent field is not null then parent must be a family.
					if (this.getParent() instanceof AcmeFamily) {
						resolvedType = ((AcmeFamily) this.getParent()).getType(curUnresolvedType.getTypeName());
					}

					if (this.getParent() instanceof AcmeDesign) {
						resolvedType = ((AcmeDesign) this.getParent()).getType(curUnresolvedType.getTypeName());
					}
					else if ((resolvedType == null) && (this.getParentDesign() != null)) {
						// then check to see if the supertype is declared in the same
						// design as this type.
						resolvedType = this.getParentDesign().getType(curUnresolvedType.getTypeName());
					}
				}
				if (resolvedType == null) {
					String errorMsg =
						"Unable to resolve supertype \""
							+ curUnresolvedType.getTypeName()
							+ "\" of connector type "
							+ this.getTypeName();
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

				}
				else if (resolvedType instanceof AcmeElementTypeRef) {
					AcmeElementTypeRef ref = (AcmeElementTypeRef) resolvedType;
					if (!(ref.getReferentType() instanceof AcmeConnectorType)
						&& !(ref.getReferentType() instanceof AcmeElementProtoType)) {
						String errorMsg =
							"Connector type \""
								+ this.getTypeName()
								+ "\"'s supertype \""
								+ curUnresolvedType.getTypeName()
								+ "\" is not a connector or element type.";
						TypeRefResolutionError badTypeRefError =
							new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
						badTypeRefError.setUnresolvedRef(curUnresolvedType);
						AcmeErrorManager.errorDetected(badTypeRefError);

						resolvedType = null;
					}
				}
				else if (
					!(resolvedType instanceof AcmeConnectorType) && !(resolvedType instanceof AcmeElementProtoType)) {
					String errorMsg =
						"Connector type \""
							+ this.getTypeName()
							+ "\"'s supertype \""
							+ curUnresolvedType.getTypeName()
							+ "\" is not a connector or element type.";
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

					resolvedType = null;
				}
				else if (resolvedType == this) {
					// Is there an immediate recursive resolution
					String errorMsg = "Connector type \"" + this.getTypeName () + "\" cannot have itself as a supertype.";
					TypeRefResolutionError badTypeError = new TypeRefResolutionError (this, errorMsg, resolvedType.getTypeName());
					badTypeError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected (badTypeError);
					
					resolvedType = null;
				}

				if (resolvedType != null) {
					this.superTypes.setItemAt((AcmeObject) resolvedType, i);
				};
			};
		}
	}

	/**
	 *  lookupName implements the acmeLib.designRules.Scope interface.  It
	 *  extends the AcmeConnector.lookupName method by searching
	 *  through all of its supertypes for a named entity before calling
	 *  its supertype (AcmeConnector)'s lookupName method to search for
	 *  ports, properties, and representations contained in this type's
	 *  prototype.
	 *  @param name is compared with "self"
	 *  @param bubbleUp boolean
	 *  @return Object corresponding to name
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;

		// check first whether this is referring to "this" type.
		if (name == null) {
			return null;
		}
		else if (name.equalsIgnoreCase("self")) {
			return this;
		};

		// recursively walk up the tree of supertypes hunting for "name"
		for (int i = 0;(i < this.superTypes.getSize()) && (retObj == null); i++) {
			AcmeElementType curSuperType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (!(curSuperType instanceof AcmeUnresolvedElementType))
				retObj = curSuperType.lookupName(name, false);
		};

		// if we didn't find it in a supertype then check for it in this type's
		// prototype object.
		if (retObj == null)
			retObj = super.lookupName(name, bubbleUp);

		return retObj;
	};

	/**
	 *  This method determines whether this connector type's properties, representations,
	 *  design rules, and role declarations are all type-consistent, and that
	 *  all declared supertypes are AcmeConnectorType's also.  This method
	 *  delegates all of the actual checking to the AcmeTypeManager class.
	 *  @return whether connector type does typecheck
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckConnectorType(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			AcmeErrorManager.errorRepaired(this, null);
			return true;
		}
		// connector type doesn't typecheck.  There should be an error at the
		// top of the AcmeTypeManager's typecheckingErrorStack.  If there
		// is such an error then announce it to the AcmeErrorManager and
		// return false.
		AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
		if (eltTypeError != null) {
			AcmeErrorManager.errorDetected(eltTypeError);
			return false;
		}
		// no error was reported.  something went wrong?
		// announce a generic error.
		AcmeErrorManager.errorDetected(
			new ElementTypecheckError(this, "Connector Type " + this.getTypeName() + " does not typecheck.", this));
		return false;
	};

	//
	// Exporting methods
	//

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltType(s, "Connector Type", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltType(s, "Connector Type", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this connector type as a string of the
	 *  form "Connector Type <name> [ extends <Typename>+ ]".  For example:
	 *  "Connector Type fooT : connT, superConnT".
	 * @return string containing this component type's signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Connector Type " + this.getTypeName());
		AcmeEnumeration superTypesEnum =
			this.getSuperTypes().enumerate(new acmeLib.core.filters.NoBaseClassesEnumFilter());
		if (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" extends <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" extends " + eltType.getTypeName());
		}

		while (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * Creates a workshop for editing this connector type.
	 * implements the acmeLib.Editable interface.
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editConnectorType(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

}
