
package acmeLib.elements;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.StringTokenizer;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.Debug;
import acmeLib.acmeParser.AcmeParser;
import acmeLib.armaniParser.ArmaniParser;
import acmeLib.armaniParser.ParseException;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeDuplicateNameException;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.exceptions.AcmeKeywordNameException;
import acmeLib.core.filters.AcmeObjectFilter;
import acmeLib.core.filters.NoAcmeFamiliesFilter;
import acmeLib.core.globals.ConversionConstants;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.ArmaniExportable;
import acmeLib.core.io.Exportable;
import acmeLib.designRules.AcmeResolutionException;
import acmeLib.designRules.BuiltInPredicates;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.designRules.Scope;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.ExpressionTypecheckError;
import acmeLib.errorManagement.TypeRefResolutionError;
import acmeLib.properties.*;
import acmeLib.tools.GraphWalker;
import acmeLib.tools.WalkerElementOperation;

/**
 *  Class:  AcmeDesign<BR>
 *  Purpose:Representation of an ACME Design.  An ACME Design is
 *          the root of a design in ACME.  A Design contains a set
 *          of global AcmeFamily definitions, a global Typespace
 *          which contains the global types defined for the design,
 *          and an instance of the AcmeSystem class which describes
 *          the topology of the toplevel system.<BR>
 *
 *          Applications can sub-class the Design class to include
 *          application specific design-level interfaces.<BR>
 *
 *
 *  Revision History:<BR>
 *    Date:     Person    Change<BR>
 *    4/22/97   RTM       Translated to Java<BR>
 *    6/16/97   RTM       Modified to work with second generation object model<BR>
 *    1/5/98    RTM       Added support for storing Design Analyses<BR>
 *    4/29/98   RTM       Changed AcmeDesign to be a subclass of AcmeElement
 *                         because a design can now support properties and, in
 *                         Armani, design rules.  We are moving towards making a
 *                         design a toplevel entity in its own right.  At this
 *                         point, though, an AcmeDesign can not be named, nor
 *                         can it have representations, because I have not yet
 *                         figured out what it means to have either, though it
 *                         can have design rules that refer to the toplevel
 *                         entities.  For now, get/set{Representation,Name}
 *                         are stubbed out.  They will be returned when I
 *                         understand what they mean.  Likewise, getting and
 *                         setting parents, scopes, and parentDesigns are
 *                         stubbed out to return null.
 *
 *    12/22/98  RTM       Added the static builtInsDesign to store all built-in
 *                         analyses, property types, etc.
 *    01/16/03  SWC       Eliminated extraneous addDesignToImport() method.
 *                        Modified addDesignToImport to seek familes within
 *                         paths that are passed in via an environment variable
 *                         called FAMILY_PATH.
 *
 *  @author Bob Monroe
 */

public class AcmeDesign extends AcmeElement implements acmeLib.designRules.Scope
{

  //
  // Protected members
  //
  
static final DesignEnumTypeFilter	ENUMFILTER	= new DesignEnumTypeFilter();
protected AcmeMap typeMap;
  protected AcmeMap familyMap;
  protected AcmeMap systemMap;
  protected AcmeVector importedDesigns;
  protected AcmeMap designAnalysisMap;
  
  protected Set injectionOperations = null;
  protected Set extractionOperations = null;

  /**
   * The filename of "this" design
   */
   protected String canonicalFilename;

  /**
   * stores a map of all designs that have been parsed to avoid infinite
   * recursion of circular includes.  openedDesignsMap stores
   * AcmeImportedDesignRecord's referenced by their canonical filename.
   */
   protected static AcmeMap openedDesignsMap = new AcmeMap();


   /**
    * stores all of the built-in analyses, properties, etc. available for
    * reference by all designs.
    **/
   protected static AcmeDesign builtInsDesign = null;


   /**
    * Flag to indicate whether the builtInsDesign has been initialized yet.
    **/
   protected static boolean builtInsDesignInitialized = false;


  /**
   * retrieve the enumerated types declared in this design.
   * @return AcmeEnumeration of typeMap
   */
  protected AcmeEnumeration getEnumTypes()  {
	return typeMap.enumerate(new AcmeEnumeration(),
	                         ENUMFILTER);
  };


    /**
     * initialize the builtInsDesign by requesting that the BuiltInPredicates
     * class send a string of builtIns declarations to the parser, that will
     * be added to the builtInsDesign as appropriate.
     * if the builtInsDesign has already been initialized then just return
     * quietly.
     **/
    protected void initBuiltInsDesign() {
        // quietly return if this has already been initialized. otherwise make sure
        // that it is not initialized again.
        if (builtInsDesignInitialized == true) {
            return;
        } else {
            builtInsDesignInitialized = true;
            builtInsDesign = new AcmeDesign();
        }

        // else initialize the builtInsDesign by retrieving the declarations of
        // the built-in analyses as a stream.
        Debug.out ("BuiltInPredicates.getArmaniLibDecls()");
        String builtInStr = BuiltInPredicates.getArmaniLibDecls();
        Debug.out ("new java.io.ByteArrayInputStream(builtInStr.getBytes())");
        java.io.ByteArrayInputStream builtInInputStream =
                new java.io.ByteArrayInputStream(builtInStr.getBytes());
        Debug.out ("ArmaniParser.getGlobalDesign()");
        AcmeDesign trueGlobalDesign = ArmaniParser.getGlobalDesign();
        if (trueGlobalDesign == null) { // Could be using AcmeParser, 
                                        // in which case, builtin predicates 
                                        // aren't going to be used.
            trueGlobalDesign = AcmeParser.getGlobalDesign ();
        }
        //if (trueGlobalDesign == null) return;
        try {
            Debug.out ("ArmaniParser.ReInit(builtInInputStream)");
            if (ArmaniParser.token_source == null) {
                new ArmaniParser (builtInInputStream);
            }
            ArmaniParser.ReInit (builtInInputStream);
            Debug.out ("ArmaniParser.setGlobalDesign(builtInsDesign)");
            ArmaniParser.setGlobalDesign(builtInsDesign);
            Debug.out ("ArmaniParser.parse_AcmeDesign()");
            builtInsDesign = ArmaniParser.parse_AcmeDesign();
            Debug.out ("builtInsDesign.doPostParsingProcessing()");
            builtInsDesign.doPostParsingProcessing();

            // reset the global_design to whatever it was previously.
            ArmaniParser.setGlobalDesign(trueGlobalDesign);

        } catch (acmeLib.armaniParser.ParseException parseErr) {
            parseErr.printStackTrace();
            // there should be no parse errors
            System.err.println("Error.  caught error parsing built-in declarations: "
                               + parseErr.getMessage());
            builtInsDesign = null;
        }
        finally {
            // reset the global_design to whatever it was previously and,
            // if the built-ins were parsed successfully then add them as
            // an imported design to the previous global design.
            ArmaniParser.setGlobalDesign(trueGlobalDesign);
        };
    };




  //
  // Public members
  //

  /**
   * Base constructor initializes all name-maps in the design and all of Armani's
   * built-in predicates and functions.
   */
  public AcmeDesign() {
    this.typeMap = new AcmeMap();
	this.familyMap = new AcmeMap();
	this.systemMap = new AcmeMap();
	this.designAnalysisMap = new AcmeMap();
    this.importedDesigns = new AcmeVector(1,1);
    this.canonicalFilename = null;

    initBuiltInsDesign();
  };



  /**
   * @return canonicalFilename
   */
  public String getCanonicalFilename() { return this.canonicalFilename; };

  /**
   * @param newCanonicalFilename used to set the canonicalFilename
   */
  public void setCanonicalFilename(String newCanonicalFilename) {
    this.canonicalFilename = newCanonicalFilename;
    // when the canonicalFilename is set, register "this" design
    // as the design that corresponds to that filename in the designAnalysisMap
    AcmeOpenedDesignRecord thisDesignRec = getDesignRecord(this.canonicalFilename);
    if (thisDesignRec == null) {
        thisDesignRec = new AcmeOpenedDesignRecord();
        thisDesignRec.setCanonicalFilename(this.canonicalFilename);
        thisDesignRec.setResolvedDesign(this);
        addDesignRecord(this.canonicalFilename, thisDesignRec);
    } else {
        thisDesignRec.setResolvedDesign(this);
    }
  };



  /**
   * updates the child name if it is a type, system, or design analysis
   * declared in this design, otherwise it lets the superclass take care of it.
   * @param oldName the old name of the child object
   * @param obj the child object with a changed name
   * @exception AcmeIllegalNameException thrown when the new name is illegal
   */

  public void childNameChanged(String oldName, AcmeObject obj)
    throws AcmeIllegalNameException
  {
    if (obj instanceof AcmeType) {
        this.removeType(oldName);
        this.addType((AcmeType) obj);

    } else if (obj instanceof DesignAnalysis) {
        String newName = ((DesignAnalysis) obj).getName();
        if (newName == null) {
            throw new AcmeIllegalNameException(
                    "Error.  Attempted to set a design analysis name to null.");
        } else if (!newName.equals(oldName)) {
            this.removeDesignAnalysis(oldName);
            this.addDesignAnalysis((DesignAnalysis) obj);
        };
        // else the name hasn't really changed, just return.

    } else if (obj instanceof AcmeSystem) {
        String newName = ((AcmeSystem) obj).getName();
        if (newName == null) {
            throw new AcmeIllegalNameException(
                    "Error.  Attempted to set a system name to null.");
        } else if (!newName.equals(oldName)) {
            this.removeSystem(oldName);
            this.addSystem((AcmeSystem) obj);
        };
        // else the name hasn't really changed, just return.

    } else {
        super.childNameChanged(oldName, obj);
    }
  };




  /**
   * Retrieves the fully qualified, period delimited, name of the AcmeElement.
   * This name identifies an object in the structure tree that can be reached
   * by taking a unique path from the root design.  AcmeDesign overrides this
   * method to return an empty string, indicating that it is at the root of the
   * structure tree.
   * RTM 12/7/98 -- Note that we might need and/or want to extend this to deal
   *                explicitly with imported designs and the types and families
   *                stored in those designs.
   * @return the empty string ""
   **/
  public String getQualifiedName() { return ""; };


  /**
   * @return a reference to the design that stores all of the "built-in"
   * analyses and types.
   */
  public AcmeDesign getBuiltInsDesign() { return AcmeDesign.builtInsDesign; };


  //
  // add, remove, and retrieve systems declared in this design.
  //
  /**
   * returns an enumeration of systems declared in this design
   * @return systems declared in this design
   */
  public AcmeEnumeration getSystems() { return this.systemMap.enumerate(); };

  /**
   * returns a systems declared in this design
   * @return system declared in this design
   */
  public AcmeSystem getSystem(String name) {
	AcmeSystem sys = (AcmeSystem) systemMap.lookupItem(name);
    return sys;
  };


 /**
   * adds a system to  this design
   * @param sys the AcmeSystem to be added
   * @exception AcmeIllegalNameException thrown when the system name is illegal
   */
  public void addSystem(AcmeSystem sys) throws AcmeIllegalNameException {
    if (LanguageConfig.CHECK_FOR_KEYWORDS) {
        if (LanguageConfig.isAcmeKeyword(sys.getName()))
            throw new AcmeKeywordNameException("Illegal keyword \""
                    + sys.getName() + "\" used for system name.");
    };

    if (sys.getName() == null) {
        throw new AcmeIllegalNameException("Error.  Attempted to add a "
                                           + "system with null name.");
    }

    else if (eltNamespace.containsName(sys.getName())) {
        throw new AcmeDuplicateNameException("Attempted to redefine identifier "
                + sys.getName() + " as an added system.");
    };

    eltNamespace.addName(sys.getName());
	systemMap.addItem (sys.getName(), sys);
	sys.setParent(this);
	sys.setParentDesign(this);
  };


  /**
   * removes a system from this design
   * @param sys the AcmeSystem to be removed
   */
  public void removeSystem(AcmeSystem sys) {
    this.removeSystem(sys.getName());
  };


  /**
   * removes a system from this design
   * @param name the name of the AcmeSystem to be removed
   */
  public void removeSystem(String name) {
    AcmeSystem sys = (AcmeSystem) systemMap.lookupItem(name);
    if (sys != null) {
        sys.setParent(null);
    };
    systemMap.removeItem(name);
    eltNamespace.removeName(name);
  };


  //
  // add, remove, and retrieve types declared in this design.
  //
  /**
   * add a type to this design
   * @param type the AcmeType to be added to this design
   * @exception AcmeIllegalNameException thrown when the type name is illegal
   */
  public void addType(AcmeType type) throws AcmeIllegalNameException {
    if (type == null) {
        return;
    };

    if (LanguageConfig.CHECK_FOR_KEYWORDS)
        if (LanguageConfig.isAcmeKeyword(type.getTypeName()))
            throw new AcmeKeywordNameException("Illegal keyword \""
                + type.getTypeName() + "\" used for typename.");

    if (type.getTypeName() == null) {
        throw new AcmeIllegalNameException(
            "Error.  Attempted to add a type declaration with null typename.");
    }

    else if (eltNamespace.containsName(type.getTypeName()))
        throw new AcmeDuplicateNameException("Attempted to redefine "
            + "identifier " + type.getTypeName());

    eltNamespace.addName(type.getTypeName());
    this.typeMap.addItem(type.getTypeName(), (AcmeObject) type);
    if (type instanceof AcmeElement) {
        ((AcmeElement) type).setParent(this);
        ((AcmeElement) type).setParentDesign(this.getParentDesign());
    } else if (type instanceof AcmePropertyType) {
        ((AcmePropertyType) type).setDeclarationScope(this);
    };
  };



  /**
   * retrieves all of the types defined in this design.
   * @return AcmeEnumeration of the types
   */
  public AcmeEnumeration getTypes() {
    return this.typeMap.enumerate(new NoAcmeFamiliesFilter ());
  };



  /**
   * retrieves the type <em>name</em> if it is declared in this design
   * @param name the name of the AcmeType
   * @return AcmeType of this design
   */
  public AcmeType getType(String name) {
    return (AcmeType) this.typeMap.lookupItem(name);
  };


  /**
   * removes the AcmeType <em>name</em> from this design
   * @param name the name of the AcmeType to be removed
   */
  public void removeType(String name) {
    this.typeMap.removeItem(name);
    eltNamespace.removeName(name);
  };


  /**
   * removes the AcmeType <em>type</em> from this design
   * @param name the name of the AcmeType to be removed
   */
  public void removeType(AcmeType type) {
    if (type instanceof AcmeFamily) {
        eltNamespace.removeName(type.getTypeName());
        this.removeFamily(type.getTypeName());
    } else if (type != null) {
        eltNamespace.removeName(type.getTypeName());
        this.typeMap.removeItem(type.getTypeName());
    };
  };


  //
  // imported designs manipulations.
  //

  /**
   * tries to find the requested filename and convert it to it's canonical form
   * It checks to see if this design's file has been imported elsewhere.
   * @param filenameStr the name of the file
   */
  public void addDesignToImport(String filenameStr)
  {
    // try to find the requested filename and convert it to it's canonical form
    String canonicalFilename = null;
    java.io.File designFile = new File(filenameStr);

    // first, treat the import name as an explicit path.
    // test to see if this file exists.  If it doesn't exist, massage the
    // string to remove unnecessary backslashes and quotes and try again.
    if (designFile.exists() == false) {
        designFile = new File(AcmeImportedDesignRecord.massageFileImportString(filenameStr));
    };

    // SWC (01/16/2003) added to seek for families in AS_SEARCH_PATH
    if (designFile.exists() && designFile.isFile()) {
        try {
            canonicalFilename = designFile.getCanonicalPath();
        } catch (IOException ioExc) {
            canonicalFilename = filenameStr;
        };
    } else if (designFile.getPath().startsWith("$")) {  // substitute env't var
        String PSEP = System.getProperty("path.separator");
        String FSEP = System.getProperty("file.separator");
        String path = designFile.getPath();
        int end = path.indexOf(FSEP);
        if (end == -1) {  // use the length of the path
            end = path.length();
        }
        String pathVar = path.substring(1, end);  // don't include the "$"
        StringTokenizer pathTokens = new StringTokenizer(System.getProperty(pathVar), PSEP);
        while (pathTokens.hasMoreTokens()) {
            String pathValue = pathTokens.nextToken();
            String newPath = pathValue;
            if (end < path.length()) {
                newPath += path.substring(end);
                // BugFix #153
                if (!newPath.endsWith(".acme")) newPath += ".acme";
            }
            designFile = new File(newPath);
            if (designFile.exists()) {
                try {
                    canonicalFilename = designFile.getCanonicalPath();
                    break;  // found one, stop searching
                } catch (IOException e) {
                    // failed to find name, try next one; no need to treat exception
                }
            }
        }
    } else {
        // finally, treat the import name as a Family name and search for the
        // corresponding file in the list of paths supplied in AS_SEARCH_PATH
        String familyPaths = System.getProperty("AS_SEARCH_PATH");
        String PSEP = System.getProperty("path.separator");
        String FSEP = System.getProperty("file.separator");
        StringTokenizer pathTokens = new StringTokenizer(familyPaths, PSEP);
OUTER:  while (pathTokens.hasMoreElements()) {  // get each path element
            StringBuffer pathBuf = new StringBuffer(pathTokens.nextToken());
            pathBuf.append(FSEP + filenameStr);
            if (!filenameStr.endsWith (".acme"))
            	pathBuf.append(".acme");
            canonicalFilename = pathBuf.toString();
            if (canonicalFilename.startsWith("$")) {
                int end = canonicalFilename.indexOf(FSEP);
                if (end == -1) {  // use the length of the path
                    end = canonicalFilename.length();
                }
                String pathVar = canonicalFilename.substring(1, end);  // don't include the "$"
                StringTokenizer listTokens = new StringTokenizer(System.getProperty(pathVar), PSEP);
                while (listTokens.hasMoreTokens()) {
                    String pathValue = listTokens.nextToken();
                    String newPath = pathValue;
                    if (end < canonicalFilename.length()) {
                        newPath += canonicalFilename.substring(end);
                    }
                    designFile = new File(newPath);
                    if (designFile.exists()) {
                        try {
                            canonicalFilename = designFile.getCanonicalPath();
                            break OUTER;  // found one, stop searching
                        } catch (IOException e) {
                            // failed to find name, try next one; no need to treat exception
                        }
                    }
                }
            } else {
                designFile = new File(canonicalFilename);
                if (designFile.exists()) {
                    try {
                        canonicalFilename = designFile.getCanonicalPath();
                        break;  // found one, stop searching
                    } catch (IOException e) {
                        // failed to find name, try next one; no need to treat exception
                    }
                }
            }
        }
    }

    // see if this design's file has been imported elsewhere.
    AcmeImportedDesignRecord designRecord = null;
    AcmeOpenedDesignRecord openedRecord = getDesignRecord(canonicalFilename);
    if (openedRecord != null
        && openedRecord instanceof AcmeImportedDesignRecord) {
        // use the imported design record
        designRecord = (AcmeImportedDesignRecord) openedRecord;
        // check if design should be reloaded
        if (designRecord.shouldReload()) {
            designRecord.setLastModified(designFile.lastModified());
            designRecord.setResolvedDesign(null);
        }
    } else if (openedRecord == null
                || openedRecord.shouldReload())
    {
        // no previous import, or imported file was modified
        designRecord = new AcmeImportedDesignRecord();
        designRecord.setDesignFilename(filenameStr);
        designRecord.setCanonicalFilename(canonicalFilename);
        designRecord.setLastModified(designFile.lastModified());
        designRecord.setResolvedDesign(null);
        addDesignRecord(canonicalFilename, designRecord);
    } else {
        // transfer opened design record over to imported design record
        designRecord = new AcmeImportedDesignRecord();
        designRecord.setDesignFilename(filenameStr);
        designRecord.setCanonicalFilename(openedRecord.getCanonicalFilename());
        designRecord.setAncestorChanged(openedRecord.isAncestorChanged());
        if (openedRecord.shouldReload()) {
            designRecord.setLastModified(designFile.lastModified());
            designRecord.setResolvedDesign(null);
        } else {
            designRecord.setLastModified(openedRecord.getLastModified());
            designRecord.setResolvedDesign(openedRecord.getResolvedDesign());
        }
        addDesignRecord(canonicalFilename, designRecord);
    }

    designRecord.addImporter(this);  // zensoul 2003-06-02
    this.importedDesigns.addItem(designRecord);
  };


  /**
   * returns a vector of this design's imported design record's
   * @return importedDesigns importedDesigns
   */
  public AcmeVector getImportedDesigns() {
    return this.importedDesigns;
  };




  /**
   * @return the design object that this importing filename references.
   */
  public AcmeDesign getImportedDesign(String filenameStr) {
    // try to find the requested filename and convert it to its canonical form
    String canonicalFilename = null;
    java.io.File designFile = new File(filenameStr);

    // test to see if this file exists.  If it doesn't exist, massage the
    // string to remove unnecessary backslashes and quotes and try again.
    if (designFile.exists() == false) {
        designFile = new File(AcmeImportedDesignRecord.massageFileImportString(filenameStr));
    };

    try {
        canonicalFilename = designFile.getCanonicalPath();
    } catch (IOException ioExc) {
        canonicalFilename = filenameStr;
    };


    // see if this design's file has been imported elsewhere.
    try {
        AcmeImportedDesignRecord designRecord = (AcmeImportedDesignRecord)
                        getDesignRecord(canonicalFilename);
        if (designRecord != null) {
            return designRecord.getResolvedDesign();
        } else {
            return null;  // no resolved design to return
        }
    } catch (ClassCastException e) {
        // not an import design record, so no resolved design to return
        return null;
    }
  };




  /**
   * remove the imported design identified by designFilename
   * @param designFilename the canonical filename of the imported design
   *                       to be removed.
   */
  public void removeImportedDesign(String designFilename) {
    // remove the record from the importedDesigns vector, but keep it in
    // the importedDesignMap in case it is either imported by another design
    // or if it is later re-imported then it won't need to be reparsed
    // because it will still be in memory.

    try {
        AcmeImportedDesignRecord designRecord = (AcmeImportedDesignRecord)
                        getDesignRecord(designFilename);
        if (designRecord != null) {
            this.importedDesigns.removeItem(designRecord);
        };
    } catch (ClassCastException e) {
        // not imported, ignore
    }
  };


  /**
   * Adds a canonical filename to design record mapping to the opened design
   * record map.
   * @param filename  the canonical filename of the design record
   * @param rec  the design record
   */
  public static void addDesignRecord (String filename, AcmeOpenedDesignRecord rec) {
    if (filename != null) {
        openedDesignsMap.addItem(filename, rec);
    }
  }

  /**
   * Returns the design record given the canonical filename.
   * @param filename  the canonical filename of the design record
   * @return AcmeOpenedDesignRecord  the design record
   */
  public static AcmeOpenedDesignRecord getDesignRecord (String filename) {
    AcmeOpenedDesignRecord designRec = null;
    if (filename != null) {
        designRec = (AcmeOpenedDesignRecord) openedDesignsMap.lookupItem(filename);
    }
    return designRec;
  }

  /**
   * Returns the design record given the imported design.
   */
  public static AcmeOpenedDesignRecord findDesignRecordByDesign (AcmeDesign design) {
    AcmeOpenedDesignRecord rec = null;
      
    AcmeEnumeration elements = openedDesignsMap.enumerate();
    while (elements.moreItems()) {
        rec = (AcmeOpenedDesignRecord) elements.getNextItem();
        if (rec.getResolvedDesign().equals(design)) {  // found
            break;
        }
    }
    return rec;
  }

  /**
   * Removes a canonical filename to design record mapping from the opened 
   * design record map.
   * @param filename  the canonical filename of the design record to remove
   */
  public static void removeDesignRecord (String filename) {
    if (filename != null) {
        openedDesignsMap.removeItem(filename);
    }
  }

  /**
   * clears the DesignsMap
   */
  public static void clearImportedDesignsMap() {
    openedDesignsMap.clear();
  };





  /**
   * parses all of the imported designs and adds them to the list of
   * available designs to search for names of types, families, and
   * analyses.  Although this method is public, it should generally
   * only be called by other Acme/Armani Lib methods, as it's use
   * can be dangerous.   You've been forewarned...
   * @exception acmeLib.acmeParser.ParseException error while parsing
   * @exception acmeLib.armaniParser.ParseException error while parsing
   *
   */
  public void resolveImportedDesigns()
        throws acmeLib.acmeParser.ParseException, acmeLib.armaniParser.ParseException
  {
    java.io.FileInputStream input_stream = null;
    AcmeDesign curDesign = null;
    AcmeImportedDesignRecord curDesignRecord = null;

    // for each of the designs imported by this design, have the parser
    // parse the requested file and return the resulting design.  Then
    // add the new design to the appropriate AcmeImportedDesignRecord.
    // Finally, recursively resolve the imported designs of all of the
    // just-resolved designs.
    int numImports = this.importedDesigns.getSize();
    for (int i = 0; i < numImports; i++) {
        curDesignRecord =
                (AcmeImportedDesignRecord) this.importedDesigns.getItemAt(i);
        if (curDesignRecord.getResolvedDesign() == null) {
            // then parse the design in the named file.
            try {
            	
                input_stream =
                    new java.io.FileInputStream(curDesignRecord.getCanonicalFilename());

                // need to be a bit tricky here to make this work for both
                // Armani and Acme designs...
                if (acmeLib.core.globals.LanguageConfig.IS_PURE_ACME) {
                    AcmeParser.ReInit(input_stream);
                    AcmeParser.initDesign();
                    curDesignRecord.setResolvedDesign(AcmeParser.parse_AcmeDesign());
                } else if (acmeLib.core.globals.LanguageConfig.IS_ARMANI) {
                    ArmaniParser.ReInit(input_stream);
                    ArmaniParser.initDesign();
                    curDesignRecord.setResolvedDesign(ArmaniParser.parse_AcmeDesign());
                };
            }
            // FIX ME: Should be fixed when I get a chance to fix the AcmeParser
            catch (acmeLib.acmeParser.ParseError acmeParseExc) {
                AcmeParser.getErrorStack().push(new String(
                    "Error in design import file: "
                     + curDesignRecord.getDesignFilename()
                     + " -- " + acmeParseExc.getMessage()));
//                throw acmeParseExc;
            }
            catch (acmeLib.armaniParser.ParseException armaniParseExc) {
                ArmaniParser.getErrorStack().push(new String(
                    "Error in design import file: "
                     + curDesignRecord.getDesignFilename()
                     + " -- " + armaniParseExc.getMessage()));
                throw armaniParseExc;
            }
            catch (FileNotFoundException fnfExc) {
                String errMsg = new String( "Error unable to open design import file: "
                     + curDesignRecord.getDesignFilename()
                     + " -- " + fnfExc.getMessage());
                ArmaniParser.getErrorStack().push(errMsg);
                ParseException newParseException =
                                            new ParseException(errMsg);
                throw newParseException;
            }
            catch (SecurityException secExc) {
                String errMsg = new String( "Error: SecurityException thrown "
                    + "opening design import file: "
                    + curDesignRecord.getDesignFilename()
                    + " -- " + secExc.getMessage());
                ArmaniParser.getErrorStack().push(errMsg);
                ParseException newParseException =
                                            new ParseException(errMsg);
                throw newParseException;
            }
            catch (Exception e) {
            	String errMsg = new String ("Error: unable to open the design import file: " +
            			curDesignRecord.getDesignFilename() + " -- " + e.getMessage());
            	ArmaniParser.getErrorStack ().push (errMsg);
            	ParseException pE = new ParseException (errMsg);
            	throw pE;
            }

            // if we just parsed the design, then recursively call this method
            // on the new design.
            if (curDesignRecord.getResolvedDesign() != null) {
                curDesignRecord.getResolvedDesign().resolveImportedDesigns();
                curDesignRecord.getResolvedDesign ().resolveTypeReferences();
                curDesignRecord.getResolvedDesign ().resolveInstanceReferences();
            };

        };
    };

  };



  ////////////////////////
  // family manipulations
  ////////////////////////
  /**
   * adds a family to the design
   * @param fam the AcmeFamily to be added to this design
   * @exception AcmeIllegalNameException family name is illegal
   */
  public void addFamily (AcmeFamily fam) throws AcmeIllegalNameException
  {
    if (LanguageConfig.CHECK_FOR_KEYWORDS)
        if (LanguageConfig.isAcmeKeyword(fam.getName()))
            throw new AcmeKeywordNameException("Illegal keyword \""
                    + fam.getName() + "\" used for family name.");
    if (getFamily(fam.getName()) != null)
        throw new AcmeDuplicateNameException("Attempted to redefine "
            + "family " + fam.getName());

    else if (eltNamespace.containsName(fam.getTypeName()))
        throw new AcmeDuplicateNameException("Attempted to redefine "
            + "identifier " + fam.getTypeName());

    addType(fam);
    familyMap.addItem (fam.getName(), fam);
  };

  /**
   * returns a family corresponding to name
   * @param name the name of the AcmeFamily
   * @return AcmeFamily corresponding to name
   */
  public AcmeFamily getFamily (String name) {
    return (AcmeFamily) familyMap.lookupItem (name);
  };

  /**
   * returns an enumeration of all the families in this design
   * @return AcmeEnumeration of the familyMap
   */
  public AcmeEnumeration getFamilies() {
	return familyMap.enumerate();
  };


  /**
   * removes a family from this design
   * @param name the name of the family to be removed from the design
   */
  public void removeFamily(String name) {
    if (getFamily(name) != null){
        this.familyMap.removeItem(name);
        removeType(name);
    }
  };




  /////////////////////////
  // DesignAnalyses
  /////////////////////////

  /**
   * returns this design's design analises called <em>name</em>
   * @param name the name of the DesignAnalysis to be returned
   * @return DesignAnalisis corresponding to name
   */
  public DesignAnalysis getDesignAnalysis (String name) {
    return (DesignAnalysis) designAnalysisMap.lookupItem (name);
  };


  /**
   * returns an enumeration of all the DesignAnalysis
   * @return AcmeEnumeration of designAnalisisMap
   */
  public AcmeEnumeration getDesignAnalyses() {
	return designAnalysisMap.enumerate();
  };


  /**
   * adds a design analysis to this design
   * @param newAnalysis DesignAnalysis to be added to this design
   * @exception AcmeIllegalNameException design analysis name is illegal
   */
  public void addDesignAnalysis (DesignAnalysis newAnalysis)
    throws AcmeIllegalNameException
  {
    if (newAnalysis == null)
        return;     // there is no analysis to add, return gracefully
    if (LanguageConfig.CHECK_FOR_KEYWORDS)
        if (LanguageConfig.isAcmeKeyword(newAnalysis.getName()))
            throw new AcmeKeywordNameException("Illegal keyword \""
                + newAnalysis.getName() + "\" used for design analysis name.");

    if (newAnalysis.getName() == null) {
        throw new AcmeIllegalNameException(
            "Error.  Attempted to add a design analysis with null name.");
    }

    else if (eltNamespace.containsName(newAnalysis.getName()))
        throw new AcmeDuplicateNameException("Attempted to redefine "
            + "identifier " + newAnalysis.getName());

    eltNamespace.addName(newAnalysis.getName());
	newAnalysis.setScope(this);
	designAnalysisMap.addItem(newAnalysis.getName(), newAnalysis);
  };


  /**
   * removes a design analysis from this design
   * @param name the name of the DesignAnalysis to be removed from this design
   */
  public void removeDesignAnalysis (String name) {
    DesignAnalysis tmpRule = getDesignAnalysis(name);
    if (tmpRule == null)
        return;  // there's no such role, return gracefully
    else {
        eltNamespace.removeName(tmpRule.getName());
        removeDesignAnalysis(tmpRule);
    };
  };

  /**
   * removes a design analysis from this design
   * @param rule DesignAnalysis to be removed from this design
   */
  public void removeDesignAnalysis (DesignAnalysis rule) {
    if (rule == null) return;
	eltNamespace.removeName(rule.getName());
	designAnalysisMap.removeItem (rule.getName());
  };






  /////////////////////////////////////////////////////////////////////////////
  // Operations  (operations are represented as properties of type operationT)
  /////////////////////////////////////////////////////////////////////////////

  /**
   * @return an enumeration containing all of the operations declared in this
   *         design.  These operations can be represented as properties of type
   *         operationT or as property value objects that satisfy type
   *         operationsT.
   */
  public AcmeEnumeration getOperations() {
    AcmeEnumeration ops = new AcmeEnumeration();
    if (builtInsDesign == null) {
        return ops;     // we don't have an operationsType to check against...
    };

    Object opTypeObj = builtInsDesign.lookupName("operationT", false);
    Object opSetTypeObj = builtInsDesign.lookupName("operationsT", false);

    AcmePropertyType opPropType, opSetPropType;

    if (opTypeObj instanceof AcmePropertyType) {
        opPropType = (AcmePropertyType) opTypeObj;
    } else {
        return ops;     // again, we don't have an operationsType to check against...
    };

    if (opSetTypeObj instanceof AcmePropertyType) {
        opSetPropType = (AcmePropertyType) opSetTypeObj;
    } else {
        return ops;     // again, we don't have an operationSetType to check against...
    };

    // walk through all of the properties, looking for the operations-typed
    // properties.  add them to ops as they are found.
    AcmeEnumeration props = this.getProperties();

    while (props.moreItems()) {
        AcmeProperty curProp = (AcmeProperty) props.getNextItem();
        if (curProp.getType() == opPropType) {
            ops.insertItem(curProp);
        } else if (curProp.getType() == opSetPropType) {
            AcmeSet opSet = (AcmeSet) curProp.getValue();
            AcmeEnumeration setVals = opSet.getValues();
            while(setVals.moreItems()) {
                ops.insertItem(setVals.getNextItem());
            };
        };
    };

    return ops;
  };




  ///////////////////////
  // Exporting methods
  ///////////////////////

  /**
   * protected helper method that exports both Armani and Acme.
   * @param s AcmeOutputStream
   * @param exportLang ExportLanguage
   * @exception java.io.IOException when error occurs when writing
   * @exception AcmeException for acme exception
   */
  protected void export(AcmeOutputStream s, ExportLanguage exportLang)
    throws java.io.IOException, AcmeException
  {
    // export this designs "imports"
    int numImports = this.importedDesigns.getSize();
    for (int i = 0; i < numImports; i++) {
        AcmeImportedDesignRecord curDesignRecord =
                (AcmeImportedDesignRecord) this.importedDesigns.getItemAt(i);
        if (curDesignRecord.getDesignFilename() != null) {
            s.writeln("import "
                      + AcmeImportedDesignRecord.massageForExport(curDesignRecord.getDesignFilename())
                      + ";");
        } else if (curDesignRecord.getCanonicalFilename() != null) {
            s.writeln("import "
                      + AcmeImportedDesignRecord.massageForExport(curDesignRecord.getCanonicalFilename())
                      + ";");
        };
    };

    // if we are exporting Acme from an Armani description, then
    // export property types for invariants, heuristics, and analyses.
    if ((exportLang == ExportLanguage.ACME) && LanguageConfig.IS_ARMANI) {
        Object foundObj = null;
        s.writeln("");
        foundObj = this.lookupName(ConversionConstants.ARMANI_ANALYSIS_TYPE_NAME_AS_ACME_PROP,
                                   true);
        if (foundObj == null) {
            s.writeln("Property Type "
                      + ConversionConstants.ARMANI_ANALYSIS_TYPE_NAME_AS_ACME_PROP
                      + " = String;");
            foundObj = null;
        };

        foundObj = this.lookupName(ConversionConstants.ARMANI_INVARIANT_TYPE_NAME_AS_ACME_PROP,
                                   true);
        if (foundObj == null) {
            s.writeln("Property Type "
                      + ConversionConstants.ARMANI_INVARIANT_TYPE_NAME_AS_ACME_PROP
                      + " = String;");
            foundObj = null;
        };

        foundObj = this.lookupName(ConversionConstants.ARMANI_HEURISTIC_TYPE_NAME_AS_ACME_PROP,
                                   true);
        if (foundObj == null) {
            s.writeln("Property Type "
                      + ConversionConstants.ARMANI_HEURISTIC_TYPE_NAME_AS_ACME_PROP
                      + " = String;");
            foundObj = null;
        };

        s.writeln("");
    };

    // order and export the children of this design
    AcmeEnumeration childEnum = new AcmeEnumeration();
    childEnum.concatenate(this.getTypes());
    childEnum.concatenate(this.getFamilies());
    childEnum.concatenate(this.getProperties());
    childEnum.concatenate(this.getInvariants(false));
    childEnum.concatenate(this.getHeuristics(false));
    childEnum.concatenate(this.getDesignAnalyses());
    childEnum.concatenate(this.getSystems());
    childEnum.concatenate(this.getMaps());

    childEnum.sortByCreationOrder();
    while(childEnum.moreItems()) {
        AcmeObject nextItem = childEnum.getNextItem();
        if (exportLang == ExportLanguage.ARMANI) {
            if (nextItem instanceof ArmaniExportable) {
                ((ArmaniExportable) nextItem).exportArmani(s);
            }
            else {
                s.writeln("");
                s.writeln("/* ERROR.  Unable to export as Armani: "
                            + nextItem.toString() + " */");
            };
        }
        else {
            // default is to export as Acme.
            if (nextItem instanceof Exportable) {
                ((Exportable) nextItem).exportAcme(s);
            }
            else {
                s.writeln("");
                s.writeln("/* ERROR.  Unable to export as Acme: "
                            + nextItem.toString() + " */");
            };
        };

        // space the toplevel entities out a bit.
        s.writeln("");
    };
  };


  /**
   * Export this design as Acme
   * @param s AcmeOutputStream  Stream to send the exported design to.
   * @exception java.io.IOException when error occurs when writing
   * @exception AcmeException for acme exception
   */
  public void exportAcme(AcmeOutputStream s)
    throws java.io.IOException, AcmeException
  {
    this.export(s, ExportLanguage.ACME);
  };


  /**
   * Export this design as Armani
   * @param s AcmeOutputStream  Stream to send the exported design to.
   * @exception java.io.IOException when error occurs when writing
   * @exception AcmeException for acme exception
   */

  public void exportArmani(AcmeOutputStream s)
    throws java.io.IOException, AcmeException
  {
//  	doInjectionOperations();
    this.export(s, ExportLanguage.ARMANI);
  };


  /**
   * implements the acmeLib.designRules.Scope interface to map names
   * to types, families, systems, or design analyses.
   * It also looks up whatever is bound to 'name' in this scope.
   * Searches in the following order:
   * - global typenames
   * - family names
   * - system names
   * - map names
   * - design analyses
   * - imported designs
   * - builtInsDesign
   * @param name the name of the object to be looked up
   * @param bubbleUp boolean
   * @return an object corresponding to name
   */
  public Object lookupName(String name, boolean bubbleUp) {


    if (name == null)
        return null;
    Object retObj = this.getType(name);
    if (retObj == null) {
        retObj = this.getFamily(name);
    };

    if (retObj == null) {
        retObj = this.getDesignAnalysis(name);
    };

    if (retObj == null) {
        retObj = this.getSystem(name);
    };
    
    if (retObj == null) {
        retObj = this.getMap(name);
    }


    // check to see if it's a valid enumerated type value name.
    if (retObj == null) {
        AcmeEnumeration enumTypes = this.getEnumTypes();
        while(enumTypes.moreItems() && (retObj == null)) {
            EnumType curEnumType = (EnumType) enumTypes.getNextItem();
            AcmeEnum curVal = (AcmeEnum) curEnumType.lookupName(name, false);
            if (curVal != null) {
                retObj = curVal;
            };
        };
    };


    // If nothing is found in this design, search through all imported
    // designs until name is matched or the supply of designs is exhausted.
    // always bubbleUp, as imported designs are logically just an expansion
    // of "this" design.
    int numDesigns = this.getImportedDesigns().getSize();
    for (int i = 0; i < numDesigns && (retObj == null); i++) {
        AcmeImportedDesignRecord curDesignRec =
            (AcmeImportedDesignRecord) this.getImportedDesigns().getItemAt(i);
        if (curDesignRec.getResolvedDesign() != null) {
            AcmeDesign curDesign = curDesignRec.getResolvedDesign();
            retObj = curDesign.lookupName(name, bubbleUp);
        };
    };

    // finally, check to see if it is available in the builtIns Design
    // (note that this should be used for Armani only, and not for PURE_ACME)
    if (retObj == null) {
        // make sure that builtInsDesign has (a) been initialized and
        // (b) we're not already inside it (to avoid infinite loops 'an 'at).
        if ((builtInsDesign != null) && (builtInsDesign != this)) {
            retObj = builtInsDesign.lookupName(name, bubbleUp);
        };
    };

    return retObj;
  };


  /**
   * Searches in this design for the AcmeObject identified by the qualified name
   * given in <code>qname</code>, by breaking up the constituent delimited
   * parts, and traversing down the element hiearchy.
   * @param qname  The qualified name of the AcmeObject to lookup.
   * @return AcmeObject  the AcmeObject corresponding to the qualified name,
   * or <code>null</code> if none is found.
   */
  public AcmeObject lookupQualifiedName (String qname) {
      AcmeObject retObj = null;

      // the scope recursively searched
      Scope searchScope = this;
      // divide qualified name into constituent, period-delimited, parts
      StringTokenizer tokens = new StringTokenizer(qname, ".");
      while (tokens.hasMoreTokens()) {
          String name = tokens.nextToken();
          // look this name up in the current level of scope
          Object obj = searchScope.lookupName(name, true);
          if (obj instanceof Scope) {
              searchScope = (Scope) obj;
          } else {
              // not a Scope?  cannot continue
              break;
          }
          if (obj instanceof AcmeObject) {
              // save what we've got so far as element to return
              retObj = (AcmeObject) obj;
          }
      }

      return retObj;
  }

  /**
   * Do all of the post-parsing processing of the design that is supposed
   * to happen after parsing, in preparation for, but not including,
   * typechecking.  Specifically, this calls:<BR>
   * <UL>
   * <LI> resolveImportedDesigns()
   * <LI> resolveTypeReferences()
   * <LI> instantiateElements()
   * <LI> resolveInstanceReferences()
   * </UL>
   */
  public void doPostParsingProcessing()
  {
    try {
        this.resolveImportedDesigns();
    }
    catch (acmeLib.acmeParser.ParseException parseExc) {
        String errMsg = "Error found resolving imported designs in "
            + "AcmeDesign.doPostParsingProcessing()";
        AcmeErrorManager.errorDetected(
            new AcmeError(this, errMsg));
    }
    catch (acmeLib.armaniParser.ParseException parseExc) {
        String errMsg = "Error found resolving imported designs in "
            + "AcmeDesign.doPostParsingProcessing()";
        AcmeErrorManager.errorDetected(
            new AcmeError(this, errMsg));
    }

    // RTM 5/4/98 -- make this just call super.doPostParsingProcessing()
    //               when I make AcmeDesign an AcmeElement.
    this.resolveTypeReferences();
    this.instantiateElements();
    this.resolveInstanceReferences();
  };


  /**
   * resolveTypeReferences walks over this design's types, design analyses,
   * and systems, resolving any unresolved typenames and replacing
   * the unresolved typenames with fully resolved type objects.
   */
  public void resolveTypeReferences() {

    super.resolveTypeReferences();

    // resolve all type references in all of the imported designs first
    int numDesigns = this.getImportedDesigns().getSize();
    for (int i = 0; i < numDesigns; i++) {
        AcmeImportedDesignRecord curDesignRec =
            (AcmeImportedDesignRecord) this.getImportedDesigns().getItemAt(i);
        if (curDesignRec.getResolvedDesign() != null) {
            curDesignRec.getResolvedDesign().resolveTypeReferences();
        } else {
            String errMsg = "Error, unable to resolve types in imported file: "
                + curDesignRec.getDesignFilename() + " because the file's design "
                + "is null...";
            AcmeErrorManager.errorDetected(
                new TypeRefResolutionError(curDesignRec.getDesignFilename(), errMsg));
        };
    };

    // call resolveTypeReferences() on all of the types declared in this design
    // to resolve references to supertypes from within all element types
    AcmeEnumeration e = this.getTypes();

    while (e.moreItems()) {
        AcmeObject curType = e.getNextItem();
        if (curType instanceof AcmeElementProtoType) {
            ((AcmeElementProtoType) curType).resolveTypeReferences();
        }
        else if (curType instanceof AcmeComponentType) {
            ((AcmeComponentType) curType).resolveTypeReferences();
        }
        else if (curType instanceof AcmeConnectorType) {
            ((AcmeConnectorType) curType).resolveTypeReferences();
        }
        else if (curType instanceof AcmePortType) {
            ((AcmePortType) curType).resolveTypeReferences();
        }
        else if (curType instanceof AcmeRoleType) {
            ((AcmeRoleType) curType).resolveTypeReferences();
        }
        else if (curType instanceof AcmePropertyType) {
            ((AcmePropertyType) curType).resolveTypeReferences(this);
        }
        else if (curType instanceof AcmeMapType) {
            ((AcmeMapType) curType).resolveTypeReferences();
        }
    }

    // call resolveTypeReferences on all families
    e = this.getFamilies();
    while (e.moreItems()) {
        ((AcmeFamily) e.getNextItem()).resolveTypeReferences();
    };

    // resolveTypeReferences on all global design analyses
    e = this.getDesignAnalyses();
    while (e.moreItems()) {
        try {
            ((DesignAnalysis) e.getNextItem()).resolveTypeReferences();
        }
        catch (AcmeResolutionException e1) {
//            e1.printStackTrace();
        }
    };

    // call resolveTypeReferences recursively on all systems defined in this design.
    e = this.getSystems();
    while (e.moreItems()) {
        ((AcmeSystem) e.getNextItem()).resolveTypeReferences();
    }
  }





  /**
   * recursively resolves instance references in all of this design's
   * types, families and systems.
   */
  public void resolveInstanceReferences() {

    // resolve all type references in all of the imported designs first.
    int numDesigns = this.getImportedDesigns().getSize();
    for (int i = 0; i < numDesigns; i++) {
        AcmeImportedDesignRecord curDesignRec =
            (AcmeImportedDesignRecord) this.getImportedDesigns().getItemAt(i);
        if (curDesignRec.getResolvedDesign() != null) {
            curDesignRec.getResolvedDesign().resolveInstanceReferences();
        } else {
            String errMsg = "Error, unable to resolve instance references in "
                + "imported file: " + curDesignRec.getDesignFilename()
                + " because the file's design is null...";
            AcmeErrorManager.errorDetected(
                new TypeRefResolutionError(curDesignRec.getDesignFilename(), errMsg));
        };
    };

    // resolve instance references in "this" design.
    AcmeEnumeration e = this.getTypes();

    while (e.moreItems()) {
        AcmeObject curType = e.getNextItem();
        if (curType instanceof AcmeElementProtoType) {
            ((AcmeElementProtoType) curType).resolveInstanceReferences();
        }
        else if (curType instanceof AcmeComponentType) {
            ((AcmeComponentType) curType).resolveInstanceReferences();
        }
        else if (curType instanceof AcmeConnectorType) {
            ((AcmeConnectorType) curType).resolveInstanceReferences();
        }
        else if (curType instanceof AcmePortType) {
            ((AcmePortType) curType).resolveInstanceReferences();
        }
        else if (curType instanceof AcmeRoleType) {
            ((AcmeRoleType) curType).resolveInstanceReferences();
        }
        else if (curType instanceof AcmePropertyType) {
            ((AcmePropertyType) curType).resolveTypeReferences(this);
        }
    };

    // RTM 1/16/98 -- No need to resolve property instances, yet, because
    //                properties can't currently reference other things.
    //                (though that may change in the future).

    // call resolveInstanceReferences on all families
    e = this.getFamilies();
    while (e.moreItems()) {
        ((AcmeFamily) e.getNextItem()).resolveInstanceReferences();
    };

    // call resolveInstanceReferences on all global design analyses
    e = this.getDesignAnalyses();
    while (e.moreItems()) {
        try {
            ((DesignAnalysis) e.getNextItem()).resolveInstanceReferences();
        }
        catch (AcmeResolutionException e1) {
            e1.printStackTrace();
        }
    };

    // call resolveInstanceReferences recursively on all systems defined in
    // this design.
    e = this.getSystems();
    while (e.moreItems()) {
        ((AcmeSystem) e.getNextItem()).resolveInstanceReferences();
    }

    // call resolveInstanceReferences recursively on all maps defined in
    // this design.
    e = this.getMaps();
    while (e.moreItems()) {
        ((AcmeMapInstance) e.getNextItem()).resolveInstanceReferences();
    };
  };




  /**
   * instantiateElement() walks over the structures declared in this design,
   * calling instantiateElements() on the entities of this design in the
   * following order:
   *  1) All types declared in the global design space.
   *  2) All families declared in the global design space.
   *  3) All systems declared in the global design space.
   *
   * Each of these calls is recursively sent down through the substructure
   * of each of the toplevel entities on which it is called.
   */
  public void instantiateElements()
  {
    // instantiate all elements in all of the imported designs first.
    int numDesigns = this.getImportedDesigns().getSize();
    for (int i = 0; i < numDesigns; i++) {
        AcmeImportedDesignRecord curDesignRec =
            (AcmeImportedDesignRecord) this.getImportedDesigns().getItemAt(i);
        if (curDesignRec.getResolvedDesign() != null) {
            curDesignRec.getResolvedDesign().instantiateElements();
        } else {
            String errMsg = "Error, unable to instantiate elements in "
                + "imported file: " + curDesignRec.getDesignFilename()
                + " because the file's design is null...";
            AcmeErrorManager.errorDetected(
                new TypeRefResolutionError(curDesignRec.getDesignFilename(), errMsg));
        };
    };


    AcmeEnumeration e = this.getTypes();
    while(e.moreItems()) {
        AcmeObject curType = e.getNextItem();
        if (curType instanceof AcmeElement) {
            ((AcmeElement) curType).instantiateElements();
        };
    };

    e = this.getFamilies();
    while (e.moreItems()) {
        ((AcmeFamily) e.getNextItem()).instantiateElements();
    }

    e = this.getSystems();
    while (e.moreItems()) {
        ((AcmeSystem) e.getNextItem()).instantiateElements();
    };
  };


  /**
   * walks over all of the types and design analyses, checking
   * their expressions for type-correctness.  It then asks its families
   * and systems to typecheck themselves.  It returns true if everything
   * in the design typechecks, else it returns false.
   * @return true if everything in the design typechecks, else false
   */
  public boolean typecheck() {
    boolean designTypechecks = true;
    boolean curItemTypechecks = true;
    AcmeEnumeration e = null;

    // typecheck all of the imported designs first.
    int numDesigns = this.getImportedDesigns().getSize();
    for (int i = 0; i < numDesigns; i++) {
        AcmeImportedDesignRecord curDesignRec =
            (AcmeImportedDesignRecord) this.getImportedDesigns().getItemAt(i);
        if (curDesignRec.getResolvedDesign() != null) {
            if (curDesignRec.getResolvedDesign().typecheck() == false) {
                designTypechecks = false;
                String errMsg = "Typechecking error found in "
                    + "imported file: " + curDesignRec.getDesignFilename();
                AcmeErrorManager.errorDetected(
                    new AcmeError(curDesignRec.getDesignFilename(), errMsg));
            };
        } else {
            String errMsg = "Error, unable to instantiate elements in "
                + "imported file: " + curDesignRec.getDesignFilename()
                + " because the file's design is null...";
            AcmeErrorManager.errorDetected(
                new TypeRefResolutionError(curDesignRec.getDesignFilename(), errMsg));
        }
    }

    // check that all of the globally-declared types are type-correct.
    e = this.getTypes();
    while (e.moreItems()) {
        AcmeObject curType = e.getNextItem();
        if (curType instanceof AcmeMapType) {
            if (((AcmeMapType)curType).typecheck() == false){
                designTypechecks = false;
                break;
            }
        } else if (curType instanceof AcmeElementType) {
            if (((AcmeElementType) curType).typecheck() == false) {
                designTypechecks = false;
            }
        }
    }
    
    // check that all of the globally-declared maps are type-correct.
    e = this.getMaps();
    while (e.moreItems()) {
        AcmeObject curMap = e.getNextItem();
        if (curMap instanceof AcmeMapInstance){
            if (((AcmeMapInstance)curMap).typecheck() == false) {
                designTypechecks = false;
                break;
            }
        }
    }

    // typecheck the family definitions.
    e = this.getFamilies();
    while (e.moreItems()) {
        AcmeFamily curType = (AcmeFamily) e.getNextItem();
        if (!curType.typecheck())
            designTypechecks = false;
    };


    // typecheck all global design analyses.  announce any errors found
    // to the global error manager (this isn't done by the DesignAnalysis class)
    e = this.getDesignAnalyses();
    while (e.moreItems()) {
        DesignAnalysis curAnalysis = (DesignAnalysis) e.getNextItem();
        curItemTypechecks = curAnalysis.typecheck();
        if (!curItemTypechecks) {
            designTypechecks = false;
            ExpressionTypecheckError exprTypecheckError =
                new ExpressionTypecheckError(curAnalysis,
                        "Global design analysis declaration for \""
                        + curAnalysis.getName() + "\" does not typecheck. ",
                        curAnalysis.getDesignRuleExpression());
            AcmeErrorManager.errorDetected(exprTypecheckError);
        };
    };

    // typecheck all of the design rules and analyses declared in all
    // of this design's systems
    e = this.getSystems();
    while (e.moreItems()) {
        AcmeSystem curSys = (AcmeSystem) e.getNextItem();
        curItemTypechecks = curSys.typecheck();
        if (!curItemTypechecks) {
            designTypechecks = false;
            // AcmeSystem.typecheck() takes care of announcing any errors
            // it finds.
        };
    };

    if (designTypechecks) {
        AcmeErrorManager.errorRepaired (this, null);
    }
    return designTypechecks;
  };


  /**
   * retrieves all of the types that are declared in this design and this
   * design's imported designs.  This includes element types, property types,
   * and families.
   */
  public AcmeEnumeration getAvailableTypes() {
    AcmeEnumeration allAvailableTypes = this.getTypes();
    allAvailableTypes.concatenate(this.getFamilies());

    int numImports = this.importedDesigns.getSize();
    for (int i = 0; i < numImports; i++) {
        AcmeImportedDesignRecord curDesignRecord =
                (AcmeImportedDesignRecord) this.importedDesigns.getItemAt(i);
        if (curDesignRecord.getResolvedDesign() != null) {
            allAvailableTypes.concatenate(
                        curDesignRecord.getResolvedDesign().getAvailableTypes());
        };
    };

    return allAvailableTypes;
  };


  /**
   * implement the acmeLib.Editable interface.
   * It also creates a workshop for editing this element
   */
  public void edit() {
//    if (this.editor == null) {
//        this.editor = new ElementWorkshopFrame();
//        editor.editDesign(this);
//    };
//    editor.setVisible(true);
//    editor.toFront();
  };


  public void view() {
    System.out.println("Attempted to view: " + this.getName());
    System.out.println("  but AcmeElement.view() is not yet implemented.");
  };

	/* (non-Javadoc)
	 * @see acmeLib.elements.AcmeElement#copy()
	 */
	public AcmeElement copy() {
		try {
			
			AcmeDesign designCopy = new AcmeDesign ();
			this.copyProperties(designCopy);
			
			AcmeEnumeration imports = importedDesigns.enumerate();
			while (imports.moreItems ()) {
				AcmeImportedDesignRecord rec = (AcmeImportedDesignRecord )imports.getNextItem ();
				AcmeImportedDesignRecord newRec = new AcmeImportedDesignRecord();
                rec.copy(newRec);
                designCopy.importedDesigns.addItem (newRec);
			}
			
			AcmeEnumeration analyses = getDesignAnalyses();
			while (analyses.moreItems ()) {
				DesignAnalysis an = (DesignAnalysis )analyses.getNextItem();
				designCopy.addDesignAnalysis(an);
			}
			
			AcmeEnumeration types = getTypes ();
			while (types.moreItems ()) {
				AcmeType type = (AcmeType )types.getNextItem ();
				designCopy.addType(type);
			}
			
			AcmeEnumeration families = getFamilies ();
			while (families.moreItems()) {
				AcmeFamily fam = (AcmeFamily )families.getNextItem ();
				designCopy.addFamily ((AcmeFamily )fam.copyType());
			}
			
			AcmeEnumeration systems= getSystems ();
			while (systems.moreItems ()) {
				AcmeSystem sys = (AcmeSystem )systems.getNextItem ();
				designCopy.addSystem ((AcmeSystem )sys.copy());
			}
//			// Need to stitch in references that point to the old design to
//			// point to the new design
			GraphWalker.WalkOverAllElements(designCopy, new StitchElementOperation (designCopy, this), GraphWalker.WALK_TOP_DOWN);
			return designCopy;
		}
		catch (AcmeException e) {
			return null;
		}
	}

	public void registerInjectionOperation (WalkerElementOperation op)  {
		if (injectionOperations == null)  {
			injectionOperations = new HashSet ();
		}
		injectionOperations.add(op);
	}
	
	public void registerExtractionOperation (WalkerElementOperation op)  {
		if (extractionOperations == null)  {
			extractionOperations = new HashSet ();
		}
		extractionOperations.add (op);
	}
	
	public void unregisterInjectionOperation (WalkerElementOperation op)  {
		if (injectionOperations != null)  {
			injectionOperations.remove(op);
			if (injectionOperations.isEmpty()) injectionOperations = null;
		}
	}
	public void unregisterExtractionOperation (WalkerElementOperation op)  {
		if (extractionOperations != null)  {
			extractionOperations.remove(op);
			if (extractionOperations.isEmpty()) extractionOperations = null;
		}
	}
	
	public void doExtractionOperations ()  {
		if (extractionOperations == null) return;
		Iterator i = extractionOperations.iterator();
		while (i.hasNext ())  {
			WalkerElementOperation op = (WalkerElementOperation )i.next ();
			GraphWalker.WalkOverAllElements(this, op, GraphWalker.WALK_TOP_DOWN);
		}
	}
	
	public void doInjectionOperations ()  {
		if (injectionOperations == null) return;
		Iterator i = injectionOperations.iterator();
		while (i.hasNext ())  {
			WalkerElementOperation op = (WalkerElementOperation )i.next ();
			GraphWalker.WalkOverAllElements(this, op, GraphWalker.WALK_TOP_DOWN);
		}		
	}


	/**
	 * @return
	 */
	public boolean hasInjectionOperations() {
		return injectionOperations != null && !injectionOperations.isEmpty();
	}
	
	public boolean hasExtractionOperations () {
		return extractionOperations != null && !extractionOperations.isEmpty ();
	}


	/* (non-Javadoc)
	 * @see acmeLib.elements.AcmeElement#sameAsInherited()
	 */
	public boolean isSameAsInherited() {
		return false;
	}
};





/**
 * filter class to return only enum filters.
 * This class extends AcmeObjectFilter
 */
class DesignEnumTypeFilter extends AcmeObjectFilter {
  /**
   * @param obj an AcmeObject
   * @return whether obj is an instance of EnumType
   */
    public boolean test(AcmeObject obj) {
        return ( obj instanceof EnumType );
    };
};


