package acmeLib.elements;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeDuplicateNameException;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.exceptions.AcmeKeywordNameException;
import acmeLib.core.extensions.IElementExtension;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.core.io.ArmaniExportable;
import acmeLib.core.io.Exportable;
import acmeLib.designRules.AcmeResolutionException;
import acmeLib.designRules.DesignRule;
import acmeLib.designRules.DesignRuleEvalException;
import acmeLib.designRules.ExpressionNode;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.TypeRefResolutionError;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyType;
import acmeLib.properties.AcmePropertyValue;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.typeManagement.AcmeNamespace;
import acmeLib.typeManagement.AcmeTypeManager;
import acmeLib.typeManagement.IInheritable;

/**
 *  Class: AcmeElement<BR><BR>
 *
 *  Purpose: base class ACME Systems, components, connectors, ports & roles.<BR><BR>
 *
 *  Description: AcmeElement is a base class that corresponds to the basic<BR>
 *  building blocks of a design; an AcmeElement belongs to<BR>
 *  the category corresponding to System, Component, Connector, Port,<BR>
 *  or Role.  Every AcmeElement has a set of instances of the AcmeProperty<BR>
 *  class and a set of Representations. <BR><BR>
 *
 *
 *  Revision History:<BR>
 *    Date:     Person    Change<BR>
 *    ?/?/97    DK      Created C++ version<BR>
 *    4/22/97   RTM     Translated to Java<BR>
 *    5/30/97   RTM     All stubbed methods filled in<BR>
 *    6/13/97   RTM     revised for second generation object model.<BR>
 *    8/6/97    RTM     removed deprecated getCategoryName method<BR>
 *    1/14/98   RTM     removed InstantiationMethod data and methods<BR>
 *    1/14/98   RTM     added support for declarations of multiple types including<BR>
 *                      methods add/removeDeclaredType() and getDeclaredTypes()<BR><BR>
 *    11/5/99  JH      added support for map
 *
 *  @author Bob Monroe<BR>
 */
public abstract class AcmeElement
	extends acmeLib.AcmeObject
	implements
		acmeLib.designRules.Scope,
		acmeLib.Editable,
		acmeLib.core.io.Exportable,
		acmeLib.core.io.ArmaniExportable,
		acmeLib.properties.AcmePropertyBearer,
		IInheritable {
	//
	// Protected class members
	//
	protected AcmeMap propMap;
	protected AcmeMap repMap;
	protected AcmeMap mapMap;

	/**
	 * parentDesign is the design in which this element is
	 * defined.  the parent is the AcmeElement that directly contains
	 * this element
	 */
	protected AcmeDesign parentDesign;
	protected AcmeElement parent;

	//
	// Protected class members
	//
	protected String name;
	protected AcmeNamespace eltNamespace;

	/**
	 *  vector of ElementType objects that this element claims to satisfy
	 */
	protected AcmeVector declaredTypes;

	/**
	 * vector of ElementType objects that this element instantiates
	 * (with the new <TypeName> Armani textual statement).
	 */
	protected AcmeVector instantiatedTypes;

	protected AcmeVector inheritedFrom;

	/**
	 *  storage for the design rules applied to this element.
	 */
	protected AcmeVector invariants;
	protected AcmeVector heuristics;

	/**
	 * for information annotations not part of the acme definition
	 */
	protected Map userData = null;
	protected static final int CANDIDATE_EXTENDS_WITH = 1;
	protected static final int WROTE_OPENING_BRACE = 2;

	/* (non-Javadoc)
	 * @see acmeLib.typeManagement.IInheritable#getInheritedSources()
	 */
	public AcmeEnumeration getInheritedSources() {
		if (inheritedFrom == null)
			return new AcmeEnumeration();
		else
			return inheritedFrom.enumerate();
	}

	/**
	 * @return
	 */
	public IElementExtension getUserData(Object key) {
		if (userData == null) {
			return null;
		}
		else {
			return (IElementExtension) userData.get(key);
		}
	}

	/**
	 * @param data
	 */
	public void setUserData(Object key, IElementExtension data) {
		if (userData == null) {
			userData = new HashMap();
		}
        if (data == null) {
            userData.remove(key);
        }
		else {
            userData.put(key, data);
        }
	}

	//
	// protected instantiation methods
	//

	/**
	 *  Most of the specifics of instantiating elements is implemented in
	 *  AcmeElement's subclasses.  AcmeElement.instantiateElements only
	 *  instantiates the representations of the element, because that is the
	 *  only instantiation step that is completely common amongst all
	 *  of the AcmeElement subclasses.  Thus it is more of a helper routine
	 */
	protected void instantiateRepresentations() {
		AcmeEnumeration e;
		AcmeSystem curSys = null;
		e = this.getRepresentations();
		while (e.moreItems()) {
			curSys = ((AcmeRepresentation) e.getNextItem()).getSystem();
			curSys.instantiateElements();
		}
	}

	//
	// protected copy methods
	//

	/**
	 * This method does a deep copy of this element's representations to destObj
	 * @param destObj AcmeElement which gets the element's representaions
	 */
	protected void copyRepresentations(AcmeElement destObj) {
		AcmeEnumeration reps_enum = this.getRepresentations();
		destObj.repMap = new AcmeMap();

		AcmeRepresentation cur_rep, new_rep;
		while (reps_enum.moreItems()) {
			cur_rep = (AcmeRepresentation) reps_enum.getNextItem();
			new_rep = ((AcmeRepresentation) cur_rep).copy();

			new_rep.setParent(destObj);
			AcmeSystem newRepSys = new_rep.getSystem();
			if (newRepSys != null) {
				newRepSys.setParent(destObj);
				newRepSys.setParentRepresentation(new_rep);
				// walk over all of the copied attachments making sure that
				// their parent is set properly to the newly copied system
				AcmeEnumeration e = newRepSys.getAttachments();
				while (e.moreItems()) {
					AcmeAttachment curAttach = (AcmeAttachment) e.getNextItem();
					curAttach.setParent(newRepSys);
				}
			}

			try {
				destObj.addRepresentation(new_rep);
			}
			catch (AcmeDuplicateNameException dupExc) {
				// if it is a duplicate name, then we just replace the
				// old rep with the new rep directly.  This shouldn't be an
				// issue, because we are supposed to be only creating a
				// new copy of dest_obj.  We're not doing unification.
				new_rep.setParent(destObj);
				destObj.repMap.addItem(new_rep.getName(), new_rep);
			}
			catch (acmeLib.core.exceptions.AcmeIllegalNameException ainExc) {
				// assume that if we're copying a representation that has been
				// legally added elsewhere then it won't have an illegal name.
			}
		}
	}

	/**
	 * This method performs a deep copy of the properties in this object and puts
	 * them into the properties list of destObj.  It removes
	 * any properties that are currently in destObj
	 *
	 * @param destObj AcmeElement, which gets the properties in this object
	 */

	public void copyProperties(AcmeElement destObj) {
		AcmeEnumeration props_enum = this.getProperties();
		destObj.propMap = new AcmeMap();

		AcmeProperty cur_prop;
		while (props_enum.moreItems()) {
			cur_prop = (AcmeProperty) props_enum.getNextItem();
			try {
				destObj.addProperty(cur_prop.copy());
			}
			catch (AcmeException e) {
			}
			// if the property is unnamed then simply don't copy this property
		}
	}

	/**
	 * This method performs a deep copy of the map instances in this object 
	 * and puts them into the map instances list of destObj, overwriting
	 * existing map instances of desObj
	 * 
	 * @param destObj AcmeElement, which gets the map instances in this object
	 */
	public void copyMaps(AcmeElement destObj) {
		AcmeEnumeration maps_enum = this.getMaps();
		destObj.mapMap = new AcmeMap();

		AcmeMapInstance cur_map = null;
		while (maps_enum.moreItems()) {
			cur_map = (AcmeMapInstance) maps_enum.getNextItem();
			try {
				destObj.addMap((AcmeMapInstance) cur_map.copy());
			}
			catch (AcmeIllegalNameException e) {
			}
		}
	}

	/**
	 * This is a helper method that can be used by element.copy and all of its subclass'
	 * implementation of copy to do a shallow-copy of all of the declared and
	 * instantiated types from this to the destObj, a deep copy of the vectors
	 * of invariants and heuristic expressions, the element's parent and parent
	 * design, and the element's name, as well as a deep-copy of all of the
	 * properties and representations.  In general, <element-subclass>.copy()
	 * should call this helper method within its copy() method.
	 *
	 * @param destObj AcmeElement, which gets the Common Element Structure of this object
	 */
	protected void copyCommonEltStructure(AcmeElement destObj) {
		destObj.name = this.getName(); // this.getName() shouldn't return
		// an illegal name, so just set new
		// name directly.
		destObj.setParent(this.getParent());
		destObj.setParentDesign(this.getParentDesign());
		this.copyProperties(destObj);
		this.copyRepresentations(destObj);
		this.copyMaps(destObj);
		this.declaredTypes.deepCopyItems(destObj.declaredTypes);
		this.instantiatedTypes.deepCopyItems(destObj.instantiatedTypes);

		// deep-copy the invariants of this and put the new invariants into destObj
		if (destObj.invariants == null)
			destObj.invariants = new AcmeVector();
		AcmeEnumeration e = this.getInvariants(false);
		while (e.moreItems()) {
			DesignRule oldDR = (DesignRule) e.getNextItem();
			DesignRule newDR = new DesignRule(oldDR);
			newDR.setDesignRuleScope(destObj);
			destObj.invariants.addItem(newDR);
		}

		// deep-copy the heuristics of this and put the new invariants into destObj
		if (destObj.heuristics == null)
			destObj.heuristics = new AcmeVector();
		e = this.getHeuristics(false);
		while (e.moreItems()) {
			DesignRule oldDR = (DesignRule) e.getNextItem();
			DesignRule newDR = new DesignRule(oldDR);
			newDR.setDesignRuleScope(destObj);
			destObj.heuristics.addItem(newDR);
		}
		
		// copy the user-data items
		if (userData != null) {
			for (Iterator iter = userData.entrySet().iterator (); iter.hasNext ();) {
				Map.Entry entry = (Map.Entry )iter.next ();
				destObj.setUserData(entry.getKey(), (IElementExtension )entry.getValue());
			}
		}
	}

	//
	// Public class members
	//
	/**
	 * This is a constructor for this class
	 */
	public AcmeElement() {
		parentDesign = null;
		parent = null;
		name = null;
		// instantiate and initialize property and rep maps and design rule storage
		propMap = new AcmeMap();
		repMap = new AcmeMap();
		mapMap = new AcmeMap();
		eltNamespace = new AcmeNamespace();
		invariants = new AcmeVector();
		heuristics = new AcmeVector();
		declaredTypes = new AcmeVector();
		instantiatedTypes = new AcmeVector();
	}

	/**
	 * This is a constructor for this class
	 * @param name the name of the AcmeElement to be created
	 */
	public AcmeElement(String name) {
		this.parentDesign = null;
		this.parent = null;
		this.name = name;
		// instantiate and initialize child maps
		propMap = new AcmeMap();
		repMap = new AcmeMap();
		mapMap = new AcmeMap();
		eltNamespace = new AcmeNamespace();
		invariants = new AcmeVector();
		heuristics = new AcmeVector();
		declaredTypes = new AcmeVector();
		instantiatedTypes = new AcmeVector();
	}

	/**
	 * This method returns a new AcmeElement object that is a deep copy of "this"
	 */
	public abstract AcmeElement copy();

	/*{
	  AcmeElement new_elt = new AcmeElement(this.getName());
	  new_elt.setParent(this.getParent());
	  this.copyCommonEltStructure(new_elt);
	  return new_elt;
	}*/

	/**
	 * This method checks for equality of elements by object identity equality.
	 * @param test_value the Object to be compared with "this"
	 * @return whether test_value and "this" are equal
	 */
	public boolean equals(Object test_value) {
		if (test_value == this) {
			return true;
		}
		else if (test_value instanceof ExpressionNode) {
			try {
				return this == ((ExpressionNode) test_value).evaluateAsElement();
			}
			catch (DesignRuleEvalException drExc) {
				return false;
			}
		}
		else if (test_value instanceof AcmeElementType && this instanceof AcmeElementType) {
			AcmeElementType thisType = (AcmeElementType) this;
			AcmeElementType testType = (AcmeElementType) test_value;
			if (thisType instanceof AcmeElementTypeRef) {
				thisType = ((AcmeElementTypeRef) thisType).getReferentType();
			}
			if (testType instanceof AcmeElementTypeRef) {
				testType = ((AcmeElementTypeRef) testType).getReferentType ();
			}
			return (thisType.getTypeName().equals(testType.getTypeName()))
				&& ((AcmeElement) thisType).getParent().equals(((AcmeElement) testType).getParent());
		}
		else {

			return false;
		}
	}

	//
	// Map manipulations
	//
	/**
	 * This method adds a map to the component
	 * @param map AcmeMap to be added
	 * @exception acmeLib.core.exceptions.AcmeIllegalNameException thrown when the name of the map is illegal
	 */
	public void addMap(AcmeMapInstance map) throws acmeLib.core.exceptions.AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isAcmeKeyword(map.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + map.getName() + "\" used for map name.");
		}

		if (map.getName() == null) {
			throw new acmeLib.core.exceptions.AcmeIllegalNameException(
				"Error.  Attempted to add a map with null name.");
		}

		else if (eltNamespace.containsName(map.getName())) {
			throw new AcmeDuplicateNameException(
				"Attempted to redefine identifier " + map.getName() + " as an added map.");
		}

		eltNamespace.addName(map.getName());
		mapMap.addItem(map.getName(), map);
		map.setParent(this);
	}

	/**
	  * This method removes a map from the component
	  * @param map AcmeMap to be removed
	  */
	public void removeMap(AcmeMapInstance map) {
		removeMap(map.getName());
	}

	/**
	 * This method removes a map from the component
	 * @param name the name of the map to be removed
	 */
	public void removeMap(String name) {
		AcmeMapInstance map = (AcmeMapInstance) mapMap.lookupItem(name);
		if (map == null)
			return; // there's no map called "name" to remove
		else {
			mapMap.removeItem(name);
			eltNamespace.removeName(name);
			map.setParent(null);
		}
	}

	/**
	 * This method returns an enumeration of all maps
	 * @return an AcmeEnumeration of the mapMap
	 */
	public AcmeEnumeration getMaps() {
		return mapMap.enumerate();
	}

	/**
	 * This method returns a map with a specific name
	 * @param name the name of the map
	 * @return AcmeMap corresponding to name
	 */
	public AcmeMapInstance getMap(String name) {
		return (AcmeMapInstance) mapMap.lookupItem(name);
	}

	/**
	 * Returns the name of the AcmeElement
	 * @return the name of the AcmeElement
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Retrieves the fully qualified, period delimited, name of the AcmeElement.
	 * This name identifies an object in the structure tree that can be reached
	 * by taking a unique path from the root design.
	 * @return the fully qualified name of the AcmeElement
	 **/
	public String getQualifiedName() {
		StringBuffer nameBuf = new StringBuffer();

		if (this.getParent() != null) {
			String parentName = this.getParent().getQualifiedName();
			if (parentName != null) {
				if (!(parentName.equals(""))) {
					nameBuf.append(parentName + ".");
				}
				nameBuf.append(this.getName());
			}
		}
		else {
			nameBuf.append(this.getName());
		}

		return nameBuf.toString();
	}

	/**
	 * Sets the name of the AcmeElement
	 * @param newName the name being assigned to the AcmeElement
	 * @exception acmeLib.core.exceptions.AcmeIllegalNameException thrown when newName is illegal
	 */
	public void setName(String newName) throws acmeLib.core.exceptions.AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isArmaniKeyword(newName) || LanguageConfig.isAcmeKeyword(newName)) {
				throw new AcmeKeywordNameException("Illegal keyword \"" + newName + "\" used for element name");
			}
		}
		String oldName = this.name;
		this.name = newName;
		if (this.getParent() != null) {
			this.getParent().childNameChanged(oldName, this);
		}
	}

	/**
	 * This method returns the AcmeDesign of this element
	 * @return AcmeDesign of this element
	 */
	public AcmeDesign getParentDesign() {
		return this.parentDesign;
	}
	/**
	 * This method sets the ParentDesign of the AcmeElement
	 * @param newParent the ParentDesign being assigned to the AcmeElement
	 */
	public void setParentDesign(AcmeDesign newParent) {
		this.parentDesign = newParent;
	}

	/**
	 * This method returns the element or design in which this element is declared.
	 * @return the element or design in which this element is declared
	 */
	public acmeLib.designRules.Scope getScope() {
		if (this.getParent() != null)
			return this.getParent();
		else
			return this.getParentDesign();
	}

	/**
	 * @return eltNamespace of AcmeElement
	 */
	public AcmeNamespace getNamespace() {
		return eltNamespace;
	}

	/**
	 * @return parent of AcmeElement
	 */
	public AcmeElement getParent() {
		return parent;
	}
	/**
	 * @param parent AcmeElement being assigned to parent
	 */
	public void setParent(AcmeElement parent) {
		this.parent = parent;
	}

	/**
	 *  An AcmeElement (and its subclasses) store almost all of their
	 *  children entities in maps that are keyed by the name of those
	 *  entities.  If the name of a child changes then the parent element
	 *  needs to update its map accordingly so that the new name is
	 *  keyed to the appropriate child object and the old name is no
	 *  longer keyed to that object.
	 *  All AcmeElements that change the name of a child element, a
	 *  property, representation, design analysis, or declared type should
	 *  call this method.  Subclasses should override it as appropriate.
	 *  @param oldName old name of AcmeProperty or AcmeRepresentation
	 *  @param obj the new AcmeObject
	 *  @exception acmeLib.core.exceptions.AcmeIllegalNameException thrown when an element name is illegal
	 */
	public void childNameChanged(String oldName, AcmeObject obj)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		// check for oldName in the following order: property, representation,
		if (obj instanceof AcmeProperty) {
			String newName = ((AcmeProperty) obj).getName();
			if (newName == null) {
				throw new acmeLib.core.exceptions.AcmeIllegalNameException(
					"Error.  Attempted to set a " + "property name to null.");
			}
			else if (!newName.equals(oldName)) {
				this.removeProperty(oldName);
				this.addProperty((AcmeProperty) obj);
			}
			// else the name hasn't really changed, just return.
		}

		else if (obj instanceof AcmeRepresentation) {
			String newName = ((AcmeRepresentation) obj).getName();
			if (newName == null) {
				throw new acmeLib.core.exceptions.AcmeIllegalNameException(
					"Error.  Attempted to set a " + "representation name to null.");
			}
			else if (!newName.equals(oldName)) {
				this.removeRepresentation(oldName);
				this.addRepresentation((AcmeRepresentation) obj);
			}
			// else the name hasn't really changed, just return.
		}
	}

	/**
	 * This method uses this element instance as a prototype to create a new type.
	 * @param newTypeName the name of the new type
	 * @return null
	 */

	public AcmeElementType copyAsPrototype(String newTypeName) {
		// quasi-abstract method to be redefined by subclasses
		return null;
	}

	/**
	 * This method goes recursively until it finds the AcmeSystem that it lives
	 * in, which then returns it's scope (AcmeSystem overrides the method)
	 * @return the parent of this element
	 */
	public AcmeRepresentation getParentRepresentation() {
		return parent.getParentRepresentation();
	}

	/**
	 * This method returns an enumeration of the element's properties
	 * @return AcmeEnumeration of this AcmeElement
	 */
	public AcmeEnumeration getProperties() {
		return propMap.enumerate();
	}

	/**
	 * This method returns the number of properties of this element
	 * @return number of properties of this element
	 */
	public int getNumberOfProperties() {
		return propMap.getSize();
	}

	/**
	 * This method returns the property corresponding to name
	 * @return AcmeProperty corresponding to name
	 */
	public AcmeProperty getProperty(String name) {
		AcmeProperty p = (AcmeProperty) propMap.lookupItem(name);
		return p;
	}

	//
	// property manipulation methods
	//
	/**
	 * This method adds an AcmeProperty to this AcmeElement
	 * @param p AcmeProperty to be added to this AcmeElement
	 * @exception  acmeLib.core.exceptions.AcmeIllegalNameException thrown when the property name is illegal
	 */
	public void addProperty(AcmeProperty p) throws acmeLib.core.exceptions.AcmeIllegalNameException {
		this.doAddProperty(p);
	}

	/**
	  * This method adds an AcmeProperty to this AcmeElement
	  * @param name the name of the property to be added to this element
	  * @param t the type of the property to be added to this element
	  * @param val the property value to be added to this element
	  *
	  * @exception  acmeLib.core.exceptions.AcmeIllegalNameException thrown when the property name is illegal
	  */
	public void addProperty(String name, AcmePropertyType t, AcmePropertyValue val)
		throws acmeLib.core.exceptions.AcmeIllegalNameException {
		this.doAddProperty(new AcmeProperty(name, t, val));
	}

	/**
	 * This method removes an AcmeProperty from this element
	 * @param prop the AcmeProperty to be removed
	 */
	public void removeProperty(AcmeProperty prop) {
		this.removeProperty(prop.getName());
	}

	/**
	 * This method removes an AcmeProperty from this element
	 * @param name the name of the AcmeProperty to be removed
	 */
	public void removeProperty(String name) {
		AcmeProperty prop = (AcmeProperty) propMap.lookupItem(name);
		if (prop != null) {
			prop.setScope(null);
		}
		propMap.removeItem(name);
		eltNamespace.removeName(name);
	}

	/**
	 * This method adds an AcmeProperty to this element
	 * @param p AcmeProperty to be added
	 * @exception acmeLib.core.exceptions.AcmeIllegalNameException thrown when the property name is illegal
	 */
	protected void doAddProperty(AcmeProperty p) throws acmeLib.core.exceptions.AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isAcmeKeyword(p.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + p.getName() + "\" used for property name.");
		}

		if (p.getName() == null) {
			throw new acmeLib.core.exceptions.AcmeIllegalNameException(
				"Error.  Attempted to add a " + "property with null name.");
		}

		else if (eltNamespace.containsName(p.getName())) {
			throw new AcmeDuplicateNameException(
				"Attempted to redefine identifier " + p.getName() + " as an added property.");
		}

		eltNamespace.addName(p.getName());
		propMap.addItem(p.getName(), p);
		p.setScope(this);
	}

	//
	// representations manipulation methods
	//

	/**
	 * This method returns the representations of this element
	 * @return AcmeEnumeration of the element's representations
	 */
	public AcmeEnumeration getRepresentations() {
		return repMap.enumerate();
	}

	/**
	 * This method returns the number of representaions
	 * @return the number of representations
	 */
	public int getNumberOfReps() {
		return repMap.getSize();
	}

	/**
	 * @param name the name of the AcmeRepresentation to be returned
	 * @return AcmeRepresentation corresponding to name
	 */
	public AcmeRepresentation getRepresentation(String name) {
		return (AcmeRepresentation) repMap.lookupItem(name);
	}

	/**
	 * @param rep AcmeRepresentation to be added to this element
	 * @exception acmeLib.core.exceptions.AcmeIllegalNameException thrown when the representation name is illegal
	 */
	public void addRepresentation(AcmeRepresentation rep) throws acmeLib.core.exceptions.AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isAcmeKeyword(rep.getName()))
				throw new AcmeKeywordNameException(
					"Illegal keyword \"" + rep.getName() + "\" used for representation name.");
		}

		if (rep.getName() == null) {
			throw new acmeLib.core.exceptions.AcmeIllegalNameException(
				"Error.  Attempted to add a " + "representation with null name.");
		}

		else if (eltNamespace.containsName(rep.getName())) {
			throw new AcmeDuplicateNameException(
				"Attempted to redefine identifier " + rep.getName() + " as an added representation.");
		}

		// RTM 10/97: currently not allowing duplicate names interferes with the
		//            approach that I take to instantiate types.
		eltNamespace.addName(rep.getName());
		rep.setParent(this);
		repMap.addItem(rep.getName(), rep);
	}

	/**
	 * Remove representation <em>name</em> from this element
	 * @param name the name of the representation to be removed
	 */
	public void removeRepresentation(String name) {
		AcmeRepresentation rep = (AcmeRepresentation) repMap.lookupItem(name);
		if (rep != null) {
			rep.setParent(null);
		}
		eltNamespace.removeName(name);
		repMap.removeItem(name);
	}

	/**
	 * Remove representation <em>rep</em> from this element
	 * @param rep the name of the representation to be removed
	 */
	public void removeRepresentation(AcmeRepresentation rep) {
		if (rep == null) {
			return;
		}
		String repName = rep.getName();
		rep.setParent(null);
		eltNamespace.removeName(repName);
		repMap.removeItem(repName);
	}

	//
	// DeclaredType manipulation methods
	//

	/**
	 * @return AcmeEnumeration of the element's declaredTypes
	 */
	public AcmeEnumeration getDeclaredTypes() {
		return this.declaredTypes.enumerate();
	}

	/**
	 * @param newType AcmeElementType to be added to the element's declaredTypes
	 */
	public void addDeclaredType(AcmeElementType newType) {
		if (!(this.declaredTypes.contains((AcmeObject) newType)))
			this.declaredTypes.addItem((AcmeObject) newType);
	}

	/**
	 * @param typeToRemove AcmeElementType to be removed from the element's declaredTypes
	 */
	public void removeDeclaredType(AcmeElementType typeToRemove) {
		this.declaredTypes.removeItem((AcmeObject) typeToRemove);
		if (typeToRemove instanceof AcmeElementTypeRef) {
			declaredTypes.removeItem((AcmeObject) ((AcmeElementTypeRef) typeToRemove).getReferentType());
		}
	}

	/**
	 * This method removes all declaredTypes of the element
	 */
	public void removeAllDeclaredTypes() {
		this.declaredTypes.removeAllItems();
	}

	/**
	 * This method returns the number of declaredTypes
	 * @return number of the element's declaredTypes
	 */
	public int numberOfDeclaredTypes() {
		return this.declaredTypes.getSize();
	}

	/**
	 * This method checks whether the element's declaredTypes contain type
	 * @param type AcmeElementType to be searched for in declaredTypes
	 * @return whether the element's declaredTypes contain type
	 */
	public boolean declaresType(AcmeElementType type) {
		boolean retVal = false;
		// BugFix#285: Every element declares the element base type
		if (type == ConstantTypes.ELEMENT_BASE_TYPE) return true;
		
		if (this.declaredTypes.contains((AcmeObject) type)) {
			return true;
		}
		else {
			// ask each of the declared types for this element whether
			// they are subtypes of the queried type.  Because the element
			// indirectly declares that it satisfies the types of all of
			// its declared types' supertypes, if one of it's declared
			// types is a subtype of the requested type, then the element
			// declares the requested type as well.
			AcmeEnumeration e = this.getDeclaredTypes();
			while (e.moreItems() && (retVal == false)) {
				AcmeElementType curEltType = (AcmeElementType) e.getNextItem();
				if (AcmeTypeManager.isDeclaredSubtype(curEltType, type)) {
					retVal = true;
				}
			}
		}

		return retVal;
	}

	/**
	 * Instantiated Type manipulation methods.
	 * InstantiatedTypes are the type objects that this element instantiates
	 * with the "new <TypeName>" in a textual Armani description.
	 * @return AcmeEnumeration of the element's instantiatedTypes
	 */

	public AcmeEnumeration getInstantiatedTypes() {
		return this.instantiatedTypes.enumerate();
	}

	/**
	 * @param newType AcmeElementType to be added to the element's instantiatedTypes
	 */
	public void addInstantiatedType(AcmeElementType newType) {
		if (!(this.instantiatedTypes.contains((AcmeObject) newType)))
			this.instantiatedTypes.addItem((AcmeObject) newType);
	}

	/**
	 * @param typeToRemove AcmeElementType to be removed from the element's instantiatedTypes
	 */
	public void removeInstantiatedType(AcmeElementType typeToRemove) {
		this.instantiatedTypes.removeItem((AcmeObject) typeToRemove);
		if (typeToRemove instanceof AcmeElementTypeRef) {
			declaredTypes.removeItem((AcmeObject) ((AcmeElementTypeRef) typeToRemove).getReferentType());
		}
	}

	/**
	 * This method removes all instatiatedTypes
	 */
	public void removeAllInstantiatedTypes() {
		this.instantiatedTypes.removeAllItems();
	}

	/**
	 * @return number of the element's instantiatedTypes
	 */
	public int numberOfInstantiatedTypes() {
		return this.instantiatedTypes.getSize();
	}

	/**
	  * @param type AcmeElementType to be searched for in the element's instantiatedTypes
	  * @return whether type is in the element's instantiatedTypes
	  */
	public boolean instantiatesType(AcmeElementType type) {
		return this.instantiatedTypes.contains((AcmeObject) type)
			|| (type instanceof AcmeElementTypeRef
				&& instantiatedTypes.contains((AcmeObject) ((AcmeElementTypeRef) type).getReferentType()));
	}

	//
	// Design Rule manipulation methods
	//

	/**
	 * retrieves all of the invariants declared in this element and, if
	 * the includeInherited parameter is true then it also retrieves all
	 * of the invariants that this element (or element type) inherits.
	 * @param  includeInherited if true then traverse the inheritance tree
	 *          (either super types or declared types) and retrieve all of
	 *          the invariants that this element inherits in addition to
	 *          its locally declared invariants.  If false then it only returns
	 *          the invariants declared locally for this element.
	 * @return AcmeEnumeration the invariants that this element claims
	 *          to satisfy.
	 */
	public AcmeEnumeration getInvariants(boolean includeInherited) {
		if (!includeInherited) {
			return this.invariants.enumerate();
		}
		else {
			AcmeEnumeration allInvariants = new AcmeEnumeration();
			if (this instanceof AcmeElementType) {
				AcmeEnumeration supTypes = ((AcmeElementType) this).getSuperTypes().enumerate();
				while (supTypes.moreItems()) {
					AcmeElementType curSuperType;
					AcmeElement curSupElt;
					curSuperType = (AcmeElementType) supTypes.getNextItem();
					if (curSuperType instanceof AcmeElement) {
						curSupElt = (AcmeElement) curSuperType;
						allInvariants.concatenate(curSupElt.getInvariants(includeInherited));
					}
				}
			}
			else {
				// else this is an instance, recurse up the declared types.
				AcmeEnumeration declTypes = this.getDeclaredTypes();
				while (declTypes.moreItems()) {
					AcmeElementType curDeclType;
					AcmeElement curDeclElt;
					curDeclType = (AcmeElementType) declTypes.getNextItem();
					if (curDeclType instanceof AcmeElement) {
						curDeclElt = (AcmeElement) curDeclType;
						allInvariants.concatenate(curDeclElt.getInvariants(includeInherited));
					}
				}
			}

			// add this element's local invariants to the list
			allInvariants.concatenate(this.invariants.enumerate());
			return allInvariants;
		}
	}

	/**
	 * This method adds an invariant to this element
	 * @param newInvariant DesignRule to be added to the element's invariants
	 */
	public void addInvariant(DesignRule newInvariant) {
		this.invariants.addItem(newInvariant);
	}

	/**
	 * This method removes an invariant from this element
	  * @param invariantToRemove DesignRule to be removed from the element's invariants
	  */
	public void removeInvariant(DesignRule invariantToRemove) {
		this.invariants.removeItem(invariantToRemove);
	}

	/**
	 * This method removes all invariants
	 **/
	void removeAllInvariants() {
		this.invariants.removeAllItems();
	}

	//
	// heuristics
	//

	/**
	 * retrieves all of the heuristics declared in this element and, if
	 * the includeInherited parameter is true then it also retrieves all
	 * of the heuristics that this element (or element type) inherits.
	 * @param  includeInherited if true then traverse the inheritance tree
	 *          (either super types or declared types) and retrieve all of
	 *          the heuristics that this element inherits in addition to
	 *          its locally declared heuristics.  If false then it only returns
	 *          the heuristics declared locally for this element.
	 * @return AcmeEnumeration the heuristics that this element claims
	 *          to satisfy.
	 */
	public AcmeEnumeration getHeuristics(boolean includeInherited) {
		if (!includeInherited) {
			return this.heuristics.enumerate();
		}
		else {
			AcmeEnumeration allHeuristics = new AcmeEnumeration();
			if (this instanceof AcmeElementType) {
				AcmeEnumeration supTypes = ((AcmeElementType) this).getSuperTypes().enumerate();
				while (supTypes.moreItems()) {
					AcmeElementType curSuperType;
					AcmeElement curSupElt;
					curSuperType = (AcmeElementType) supTypes.getNextItem();
					if (curSuperType instanceof AcmeElement) {
						curSupElt = (AcmeElement) curSuperType;
						allHeuristics.concatenate(curSupElt.getHeuristics(includeInherited));
					}
				}
			}
			else {
				// else this is an instance, recurse up the declared types.
				AcmeEnumeration declTypes = this.getDeclaredTypes();
				while (declTypes.moreItems()) {
					AcmeElementType curDeclType;
					AcmeElement curDeclElt;
					curDeclType = (AcmeElementType) declTypes.getNextItem();
					if (curDeclType instanceof AcmeElement) {
						curDeclElt = (AcmeElement) curDeclType;
						allHeuristics.concatenate(curDeclElt.getHeuristics(includeInherited));
					}
				}
			}

			// add this element's local heuristics to the list
			allHeuristics.concatenate(this.heuristics.enumerate());
			return allHeuristics;
		}
	}

	/**
	 * @param newHeuristic DesignRule to be added to the element's heuristics
	 */
	public void addHeuristic(DesignRule newHeuristic) {
		this.heuristics.addItem(newHeuristic);
	}
	/**
	 * @param heuristicToRemove DesignRule to be removed from the element's heuristics
	 */

	public void removeHeuristic(DesignRule heuristicToRemove) {
		this.heuristics.removeItem(heuristicToRemove);
	}

	/**
	 * This method removes all heuristics
	 */
	void removeAllHeuristics() {
		this.heuristics.removeAllItems();
	}

	//
	// Exporting methods
	//

	/**
	 * This method puts all of the children entities that should be exported
	 * into an (unsorted) enumeration and returns it.  Subclasses
	 * can add children as desired.
	 * @return AcmeEnumeration of all the children entities of the element
	 */
	protected AcmeEnumeration getChildrenToExport() {
		// all elements have properties, representations, invariants, and heuristics.
		AcmeEnumeration retEnum = new AcmeEnumeration();
		retEnum.concatenate(this.getProperties());
		retEnum.concatenate(this.getRepresentations());
		retEnum.concatenate(this.getMaps());
		retEnum.concatenate(this.getInvariants(false));
		retEnum.concatenate(this.getHeuristics(false));
		return retEnum;
	}

	/**
	 * Export this element as an instance.  protected helper method
	 * @param s AcmeOutputStream
	 * @param categoryName the name of the category
	 * @param children AcmeEnumeration
	 * @param exportLanguage ExportLanguage
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	protected void exportEltInstance(
		AcmeOutputStream s,
		String categoryName,
		AcmeEnumeration children,
		ExportLanguage exportLang)
		throws java.io.IOException, AcmeException {
		exportComment(s);
		s.beginRegion(this);
		int headerFlag = writeElementHeader(s, categoryName, children);

		headerFlag = writeEltChildren(s, children, exportLang, headerFlag);
		if ((headerFlag & WROTE_OPENING_BRACE) == WROTE_OPENING_BRACE) 
		    s.write("}");

		s.writeln(";");
		s.endRegion(this);
		s.flush();
	}

	protected int writeEltChildren(AcmeOutputStream s, AcmeEnumeration children, ExportLanguage exportLang, int headerFlag) throws IOException, AcmeException {
		if (children.getSize() > 0) {
			children.sortByCreationOrder();
			s.indent();
			while (children.moreItems()) {
				AcmeObject nextChild = children.getNextItem();
				if (LanguageConfig.shouldExport(nextChild)) {
					if ((headerFlag & CANDIDATE_EXTENDS_WITH) == CANDIDATE_EXTENDS_WITH && !((headerFlag & WROTE_OPENING_BRACE) == WROTE_OPENING_BRACE)) {
						s.writeln(" extended with {");
						headerFlag |= WROTE_OPENING_BRACE;
					}
					if (exportLang == ExportLanguage.ARMANI) {
						if (nextChild instanceof ArmaniExportable) {
							((ArmaniExportable) nextChild).exportArmani(s);
						}
						else {
							s.writeln("/* unable to export object in Armani */");
						}
					}
					else {
						// export Acme by default
						if (nextChild instanceof Exportable) {
							((Exportable) nextChild).exportAcme(s);
						}
						else {
							s.writeln("/* unable to export object in Acme */");
						}
					}
				}
			}
			s.unIndent();


		}
		return headerFlag;
	}

	protected int writeElementHeader(AcmeOutputStream s, String categoryName, AcmeEnumeration children) throws IOException {
		// first export the appropriate declaration header, followed by the
		// body containing all children to export.
		s.write(categoryName + " " + this.getName());

		// export explicitly declared and instantiated typenames.
		AcmeEnumeration typeEnum = this.getDeclaredTypes();
		if (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            s.write(" : <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			s.write(" : " + eltType.getTypeName());
		}

		while (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			s.write(", " + eltType.getTypeName());
		}

		int headerFlag = 0;
		// write out the instantiated types, if there are any
		typeEnum = this.getInstantiatedTypes();
		if (typeEnum.moreItems()) {
			s.write(" = ");
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType) {
			//            s.write(" new <unresolved type name: " + eltType.getTypeName() + "> ");
			//        } else {
			s.write(" new " + eltType.getTypeName());
			//        }

			while (typeEnum.moreItems()) {
				eltType = (AcmeElementType) typeEnum.getNextItem();
				s.write(", " + eltType.getTypeName());
			}

			if (children.getSize() > 0) {
				headerFlag |= CANDIDATE_EXTENDS_WITH;
			}
		}

		else if (children.getSize() > 0) {
			s.writeln(" = {");
			headerFlag |= WROTE_OPENING_BRACE;
		}
		return headerFlag;
	}

	/**
	 * Export this element as an instance.  protected helper method
	 * @param s AcmeOutputStream
	 * @param categoryName the name of the category
	 * @param children AcmeEnumeration
	 * @param exportLanguage ExportLanguage
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	protected void exportEltType(
		AcmeOutputStream s,
		String categoryName,
		AcmeEnumeration children,
		ExportLanguage exportLang)
		throws java.io.IOException, AcmeException {
		exportComment(s);
		s.beginRegion(this);
		AcmeElementType thisAsType = null;
		if (!(this instanceof AcmeElementType)) {
			// can't export it as a type, throw an exception.
			throw new AcmeException(
				"Error, unable to export element "
					+ this.getName()
					+ " as an ElementType because "
					+ "is is an Element, but not an ElementType.");
		}
		else {
			thisAsType = (AcmeElementType) this;
		}

		AcmeEnumeration superTypesEnum =
			thisAsType.getSuperTypes().enumerate(new acmeLib.core.filters.NoBaseClassesEnumFilter());

		s.write(categoryName + " " + thisAsType.getTypeName());

		if (superTypesEnum.getSize() == 0) {
			s.writeln(" = {");
		}
		else {
			// the type has supertypes, export those also
			s.write(" extends ");
			while (superTypesEnum.moreItems()) {
				String superTypeName =
					new String((String) ((AcmeElementType) superTypesEnum.getNextItem()).getTypeName());
				s.write(superTypeName);
				if (superTypesEnum.moreItems())
					s.write(", ");
			}
			s.writeln(" with {");
		}

		// write out the structure of this type or extension by exporting its children
		s.indent();
		children.sortByCreationOrder();
		while (children.moreItems()) {
			AcmeObject nextChild = children.getNextItem();
			if (exportLang == ExportLanguage.ARMANI) {
				if (nextChild instanceof ArmaniExportable) {
					((ArmaniExportable) nextChild).exportArmani(s);
					s.writeln();
				}
				else {
					s.writeln("/* unable to export object in Armani */");
				}
			}
			else {
				// export Acme by default
				if (nextChild instanceof Exportable) {
					((Exportable) nextChild).exportAcme(s);
				}
				else {
					s.writeln("/* unable to export object in Acme */");
				}
			}
		}
		s.unIndent();
		s.writeln("}");
		s.endRegion(this);
		//    s.writeln ();
		s.flush();
	}

	/**
	 *  exportAcme should be filled in by subclasses.
	 *  this base method is just a stub.
	 *  @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws AcmeException, java.io.IOException {
	}
	/**
	 * exportArmani should be filled in by subclasses.
	 * this base method is just a stub.
	 *  @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportArmani(AcmeOutputStream s) throws AcmeException, java.io.IOException {
		AcmeEnumeration retEnum = new AcmeEnumeration();
		this.exportEltInstance(s, "AcmeElement", retEnum, ExportLanguage.ARMANI);
	}

	/**
	 * Returns the type signature for this element as a string with the
	 *  form "<Category> <name> [ : <Typename>+ ]".  For example:
	 *  "Component foo : compT".  The type signature for ElementTypes are
	 *  of the form "<Category> Type <name> [ extends <Typename>+ ]"  For example:
	 *  "Component Type compT extends superCompT1, superCompT2".
	 * <br>Note that this is an abstract method that should be filled
	 * in by subclasses.
	 * @return string containing this element's type signature.
	 **/
	public String getTypeSignature() {
		return this.getName();
	}

	/**
	 * This method implements the acmeLib.designRules.Scope interface to map names
	 * to elements or properties
	 * @param name the object's name
	 * @param bubbleUp boolean
	 * @return an object corrresponding to name
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		// lookup the child object called 'name'.
		// First check if this is asking for "self" in which case we have
		// a match, then check for properties named "name", then check
		// representations named "name".
		Object retObj = null;
		if (name == null) {
			return null;
		}

		// special case request for this element
		if (name.equalsIgnoreCase("self")) {
			return this;

			// special case request for all properties
		}
		else if (name.equalsIgnoreCase("properties")) {
			return this.getProperties();

			// special case request for all representations
		}
		else if (name.equalsIgnoreCase("representations")) {
			// BRS: Representations is a bit of a tricky thing. Since each representation can only
			// have one system, I will form an enumeration out of the systems and return these...
			AcmeEnumeration reps = this.getRepresentations();
			AcmeVector sys = new AcmeVector();
			while (reps.moreItems()) {
				sys.addItem(((AcmeRepresentation) reps.getNextItem()).getSystem());
			}

			return sys.enumerate();

		}
		else if (name.equalsIgnoreCase("maps")) {
			return this.getMaps();
		}
		else {
			retObj = this.getProperty(name);
			if (retObj == null)
				retObj = this.getRepresentation(name);
		}

		if ((retObj == null) && (name.equals(this.getName()))) {
			retObj = this;
		}

		if (retObj == null) {
			if ((bubbleUp) && (this.getParent() != null))
				retObj = this.getParent().lookupName(name, bubbleUp);
			else if ((bubbleUp) && (this.getParentDesign() != null))
				retObj = this.getParentDesign().lookupName(name, bubbleUp);
		}

		return retObj;
	}

	/**
	 * This method does all of the post-parsing processing of the design that is supposed
	 * to happen after parsing, in preparation for, but not including,
	 * typechecking.
	 */
	public void doPostParsingProcessing() {
		this.resolveTypeReferences();
		this.instantiateElements();
		this.resolveInstanceReferences();
	}

	/**
	 * resolveTypeReferences walks over this element's types, it's properties, and
	 * its representations, resolving any unresolved typenames and replacing
	 * the unresolved typenames with fully resolved type objects.
	 */
	public void resolveTypeReferences() {

		// resolve the names of all types this element declares it satisfies
		String lookupError = null;
		AcmeElementType curType;
		Object resolvedType;
		AcmeUnresolvedElementType curUnresolvedType;
		for (int i = 0; i < this.declaredTypes.getSize(); i++) {
			curType = (AcmeElementType) this.declaredTypes.getItemAt(i);
			if (curType instanceof AcmeUnresolvedElementType) {
				resolvedType = null;
				curUnresolvedType = (AcmeUnresolvedElementType) curType;
				if (curUnresolvedType.isQualified()) {
					Object o = this.lookupName(curUnresolvedType.getQualifier(), true);
					if (o instanceof AcmeFamily) {
						AcmeFamily fam = (AcmeFamily) o;
						AcmeType ref = fam.getType(curUnresolvedType.getTypeName());
						if (ref instanceof AcmeElementType) {
							resolvedType = new AcmeElementTypeRef(fam, (AcmeElementType) ref);
						}
					}
				}
				else {
					resolvedType = this.lookupName(curUnresolvedType.getTypeName(), true);
				}
				if (resolvedType == null) {
					//                try{throw new Exception();}catch(Exception e){e.printStackTrace();}
					lookupError =
						"Unable to resolve declared element typename: "
							+ curUnresolvedType.getTypeName()
							+ " in element "
							+ this.getName();
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, lookupError, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(this);
					AcmeErrorManager.errorDetected(badTypeRefError);

				}
				else if (!(resolvedType instanceof AcmeElementType)) {
					lookupError =
						"Declared typename: "
							+ curUnresolvedType.getTypeName()
							+ " in element "
							+ this.getName()
							+ " is not an ElementType";
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, lookupError, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(this);
					AcmeErrorManager.errorDetected(badTypeRefError);
					resolvedType = null;
				}

				// if the name of the unresolved type object was resolved properly
				// then replace the unresolvedElementType object in the declaredTypes
				// vector with the newly resolved type object.  Otherwise keep the
				// unresolvedElementType in the vector of declared types.
				if ((resolvedType != null) && !(resolvedType instanceof AcmeUnresolvedElementType)) {
					this.declaredTypes.setItemAt((AcmeObject) resolvedType, i);
				}
			}
		}

		// resolve the names of all types this element instantiates.
		for (int i = 0; i < this.instantiatedTypes.getSize(); i++) {
			curType = (AcmeElementType) this.instantiatedTypes.getItemAt(i);
			if (curType instanceof AcmeUnresolvedElementType) {
				resolvedType = null;
				curUnresolvedType = (AcmeUnresolvedElementType) curType;
				if (curUnresolvedType.isQualified()) {
					Object o = this.lookupName(curUnresolvedType.getQualifier(), true);
					if (o instanceof AcmeFamily) {
						AcmeFamily fam = (AcmeFamily) o;
						AcmeType ref = fam.getType(curUnresolvedType.getTypeName());
						if (ref instanceof AcmeElementType) {
							resolvedType = new AcmeElementTypeRef(fam, (AcmeElementType) ref);
						}
					}
				}
				else {
					resolvedType = this.lookupName(curUnresolvedType.getTypeName(), true);
				}
				if (resolvedType == null) {
					lookupError =
						"Unable to resolve instantiated element typename: "
							+ curUnresolvedType.getTypeName()
							+ " in element "
							+ this.getName();
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, lookupError, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(this);
					AcmeErrorManager.errorDetected(badTypeRefError);

				}
				else if (!(resolvedType instanceof AcmeElementType)) {
					lookupError =
						"Instantiated typename: "
							+ curUnresolvedType.getTypeName()
							+ " in element "
							+ this.getName()
							+ " is not an ElementType";
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, lookupError, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(this);
					AcmeErrorManager.errorDetected(badTypeRefError);
					resolvedType = null;
				}

				// if the name of the unresolved type object was resolved properly
				// then replace the unresolvedElementType object in the declaredTypes
				// vector with the newly resolved type object.  Otherwise keep the
				// unresolvedElementType in the vector of declared types.
				if ((resolvedType != null) && !(resolvedType instanceof AcmeUnresolvedElementType)) {
					this.instantiatedTypes.setItemAt((AcmeObject) resolvedType, i);
				}

			}
		}

		// Resolves the typenames used in this element's invariants.
		AcmeEnumeration rules = this.getInvariants(false);
		while (rules.moreItems()) {
			try {
				((DesignRule) rules.getNextItem()).resolveTypeReferences();
			}
			catch (AcmeResolutionException e) {
				//            e.printStackTrace();
			}
		}

		// resolve the typenames used in this element's heuristics.
		rules = this.getHeuristics(false);
		while (rules.moreItems()) {
			try {
				((DesignRule) rules.getNextItem()).resolveTypeReferences();
			}
			catch (AcmeResolutionException e) {
				//            e.printStackTrace();
			}
		}

		// resolve the names of all types referenced by properties of this element
		AcmeEnumeration props = this.getProperties();
		while (props.moreItems()) {
			((AcmeProperty) props.getNextItem()).resolveTypeReferences();
		}

		// resolve the typenames of all of this element's representations.
		AcmeEnumeration reps = this.getRepresentations();
		while (reps.moreItems()) {
			((AcmeRepresentation) reps.getNextItem()).resolveTypeReferences();
		}

		//resolve the typenames of all of this element's maps
		AcmeEnumeration maps = this.getMaps();
		while (maps.moreItems()) {
			((AcmeMapInstance) maps.getNextItem()).resolveTypeReferences();
		}

	}

	/**
	 * Resolves the names of all references to instances in this element and return
	 * the resolved element.
	 */
	public AcmeElement getResolvedInstanceReference() {
		System.err.println(this +" AcmeElement.getResolvedInstanceReference not implemented yet");
		try {
			throw new Exception();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	/**
	 * Resolves the names of all references to instances in this element and
	 *  all of this element's children.  This resolves instance references used
	 *  in attachments, design rules, substructure (e.g. ports and roles) and the
	 *  bindings of representations (though bindings are resolved in
	 *  AcmeRepresentation.resolveInstanceReferences()).  If properties are later
	 *  modified to support holding references to elements then those references
	 *  would be resolved in this method also.
	 */
	public void resolveInstanceReferences() {
		// resolve the instance names in all of the design rules in this element
		AcmeEnumeration e;
		e = this.getInvariants(false);
		while (e.moreItems()) {
			try {
				((DesignRule) e.getNextItem()).resolveInstanceReferences();
			}
			catch (AcmeResolutionException e1) {
				//            e1.printStackTrace();
			}
		}

		e = this.getHeuristics(false);
		while (e.moreItems()) {
			try {
				((DesignRule) e.getNextItem()).resolveInstanceReferences();
			}
			catch (AcmeResolutionException e1) {
				//            e1.printStackTrace();
			}
		}

		// walk through this elements substructure by calling
		// resolveInstanceReferences() on each of its representations.
		e = this.getRepresentations();
		while (e.moreItems()) {
			((AcmeRepresentation) e.getNextItem()).resolveInstanceReferences();
		}

		//resolveInstanceReferences on maps
		e = this.getMaps();
		while (e.moreItems()) {
			((AcmeMapInstance) e.getNextItem()).resolveInstanceReferences();
		}
	}

	/**
	 *  instantiateElements() walks over the vector of instantiatedTypes of this
	 *  element and adds the minimal structure declared in the type description
	 *  to this element.  It delegates much of the work to the AcmeTypeManager
	 *  class, so that we can change type instantiation policies relatively
	 *  centrally (ie by changing the AcmeTypeManager methods) for all elements
	 *  at once.  This method recursively calls the same method on all of the
	 *  substructure of this element instance.
	 *
	 *  Details of the instantiation are implemented by subclasses.
	 */
	public void instantiateElements() {
		this.instantiateRepresentations();
	}

	/**
	 *  typecheck() is a quasi-abstract method that requests that the
	 *  AcmeTypeManager check whether this element satisfies all of the
	 *  types that it declares it satisfies, all of the instance-specific
	 *  design rules that it specifies, and that all of its substructure
	 *  and properties are type-correct.  This method should be overriden
	 *  by all of the subclasses of AcmeElement.  Further, it does _not_
	 *  need to be called as super.typecheck() by its subclasses.
	 *  The AcmeTypeManager subclass-specific methods handle all of the
	 *  common element typechecking on their own.  They do _not_ need
	 *  this method to do so.  Typechecking is almost completely handled
	 *  in the AcmeTypeManager, and the AcmeElement.typecheck() methods
	 *  should delegate all typechecking to the AcmeTypeManager.
	 *
	 *  @return whether all components typecheck
	 *
	 */
	public boolean typecheck() {
		return AcmeTypeManager.typecheckCommonElementItems(this);
	}

	/**
	 * implement the acmeLib.Editable interface.
	 * It also creates a workshop for editing this element
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editElement(this);
		//    }
		//    editor.setVisible(true);
		//    editor.toFront();
	}

	public void view() {
		System.out.println("Attempted to view: " + this.getName());
		System.out.println("  but AcmeElement.view() is not yet implemented.");
	}

	/**
	 * retrieves all of the types that are available/visible to this element,
	 * based on the Acme/Armani type scoping rules.  Currently, that means
	 * that an element looks for the system in which it is declared, retrieves
	 * the union of all of the types declared in that system's families (and
	 * their super-types) and the design in which it is declared and all
	 * of that design's imported designs.
	 * This method should be overridden by any subclasses that allow types
	 * to be declared in their scopes (ie AcmeDesigns and AcmeFamilies).
	 */
	public AcmeEnumeration getAvailableTypes() {
		AcmeEnumeration allAvailableTypes = null;
		if (this.getParent() != null) {
			allAvailableTypes = this.getParent().getAvailableTypes();
		}
		else {
			allAvailableTypes = new AcmeEnumeration();
		}

		return allAvailableTypes;
	}

	public void addInheritedSource(IInheritable element) {
		if (element instanceof AcmeElement) {
			AcmeElement elt = (AcmeElement) element;

			if (inheritedFrom == null)
				inheritedFrom = new AcmeVector(1);
			if (!inheritedFrom.contains(elt)) {
				inheritedFrom.addItem(elt);
			}
		}
	}

	public boolean isSameAsInherited() {
		boolean same = false;
		if (inheritedFrom != null) {
			AcmeEnumeration inhs = inheritedFrom.enumerate();
			while (inhs.moreItems() && !same) {
				AcmeElement element = (AcmeElement) inhs.getNextItem();
				same =
					name.equals(element.getName())
						&& declaresSameTypes(element)
						&& instantiatesSameTypes(element)
						&& hasSameProperties(element);
			}
		}
		return same;
	}

	public boolean isInherited() {
		return inheritedFrom != null && inheritedFrom.getSize() != 0;
	}

	/**
	 * @param element
	 * @return
	 */
	private boolean hasSameProperties(AcmeElement element) {
		boolean same = true;
		AcmeEnumeration properties = this.getProperties();
		while (properties.moreItems() && same) {
			AcmeProperty prop = (AcmeProperty) properties.getNextItem();
			AcmeProperty eProp = element.getProperty(prop.getName());
			if (eProp == null || !eProp.equals(prop)) {
				same = false;
			}
		}
		return same;
	}

	/**
	 * @param element
	 * @return
	 */
	private boolean instantiatesSameTypes(AcmeElement element) {
		boolean same = true;
		AcmeEnumeration types = this.getInstantiatedTypes();
		while (types.moreItems() && same) {
			AcmeElementType type = (AcmeElementType) types.getNextItem();
			same = element.instantiatesType(type);
		}
		return same;
	}

	/**
	 * @param element
	 * @return
	 */
	private boolean declaresSameTypes(AcmeElement element) {
		boolean same = true;
		AcmeEnumeration types = this.getDeclaredTypes();
		while (types.moreItems() && same) {
			AcmeElementType type = (AcmeElementType) types.getNextItem();
			same = element.declaresType(type);
		}
		return same;
	}

}
