package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: AcmeElementProtoType<BR>
 * Purpose: Stores the type specification for a generic element.  That is,
 *          it stores a set of properties, invariants, heuristics, and
 *          representations that can be required of any category of element.
 *          An AcmeElementProtoType is a prototypical instance of an AcmeElement.
 *          Unlike an AcmeElement, though, and AcmeElementProtoType can be
 *          instantiated in Armani as a first-class entity.  It is not just an
 *          abstract base type.
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    11/30/98  RTM     Created class
<BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeElementProtoType extends AcmeElement implements AcmeElementType {

	//
	// protected members
	//
	AcmeVector superTypes;

	//
	// public members

	/**
	 * This is the constructor for this class
	 * It creates the prototype object in the superclass constructor
	 * @param typeName the name of the new port type
	 */
	public AcmeElementProtoType(String typeName) {
		// create the prototype object in the superclass constructor
		super(typeName);
		this.superTypes = new AcmeVector();
	};

	/**
	 * @return the name of this element type
	 */
	public String getTypeName() {
		return this.getName();
	};
	/**
	 * Sets the name of this element type
	 * @param typeName the name for this port type
	 * @exception AcmeIllegalNameException The name is illegal
	 */
	public void setTypeName(String typeName) throws AcmeIllegalNameException {
		this.setName(typeName);
	};

	/**
	 * This method adds a supertype to a type description
	 * @param superType the supertype to be added to the type description
	 */
	public void addSuperType(AcmeElementType superType) {
		if (superType instanceof AcmeElementTypeRef) {
			superType = ((AcmeElementTypeRef) superType).getReferentType();
		}
		if ((superType instanceof AcmeElementProtoType) || (superType instanceof AcmeUnresolvedElementType)) {
			superTypes.addItem((AcmeObject) superType);
		}
		// else throw an exception?
	};

	/**
	 * @return the supertypes of this ElementPrototype
	 */
	public AcmeVector getSuperTypes() {
		return this.superTypes;
	};

	/**
	 * Removes a supertype from a type description
	 * @param superType the supertype to be removed from the type description
	 */
	public void removeSuperType(AcmeElementType superType) {
		superTypes.removeItem((AcmeObject) superType);
		if (superType instanceof AcmeElementTypeRef) {
			declaredTypes.removeItem((AcmeObject) ((AcmeElementTypeRef) superType).getReferentType());
		}
	};

	/**
	 * Makes a new element prototype object, with a deep copy of the old
	 * element prototype's properties, representations, declared,
	 * instantiated, and super-type vectors.
	 * @return a copy of this element prototype object
	 */
	public AcmeElementType copyType() {
		AcmeElementProtoType new_elt_type = new AcmeElementProtoType(this.getTypeName());
		this.copyCommonEltStructure(new_elt_type);
		this.superTypes.copyItems(new_elt_type.superTypes);
		return new_elt_type;
	};

	/**
	 * resolveTypeReferences searches through this type object's vector of supertypes
	 * for UnresolvedElementType objects, looks up the actual type that is
	 * referred to by each UnresolvedElementType object, and replaces that
	 * object with the actual type object that should be referred to.
	 * it then calls resolveTypeReferences in its superclass to resolve the name of
	 * all of its substructure (just representations).
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
		AcmeElementType curType = null;
		AcmeType resolvedType = null;
		AcmeUnresolvedElementType curUnresolvedType;
		for (int i = 0; i < this.superTypes.getSize(); i++) {
			curType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (curType instanceof AcmeUnresolvedElementType) {
				curUnresolvedType = (AcmeUnresolvedElementType) curType;

				// first check to see if the type is declared in the same
				// family that this type is declared in.  If an AcmeElementProtoType
				// object's parent field is not null then parent must be a family.
				if (this.getParent() instanceof AcmeFamily) {
					resolvedType = ((AcmeFamily) this.getParent()).getType(curUnresolvedType.getTypeName());
				}

				if (this.getParent() instanceof AcmeDesign) {
					resolvedType = ((AcmeDesign) this.getParent()).getType(curUnresolvedType.getTypeName());
				}
				else if ((resolvedType == null) && (this.getParentDesign() != null)) {
					// then check to see if the supertype is declared in the same
					// design as this type.
					resolvedType = this.getParentDesign().getType(curUnresolvedType.getTypeName());
				}

				if (resolvedType == null) {
					String errorMsg =
						"Unable to resolve supertype \""
							+ curUnresolvedType.getTypeName()
							+ "\" of Element Type "
							+ this.getTypeName();
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

				}
				else if (!(resolvedType instanceof AcmeElementProtoType)) {
					String errorMsg =
						"Element Type \""
							+ this.getTypeName()
							+ "\"'s supertype \""
							+ curUnresolvedType.getTypeName()
							+ "\" is not an Element Type.";
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

					resolvedType = null;
				};

				if (resolvedType != null) {
					this.superTypes.setItemAt((AcmeObject) resolvedType, i);
				};
			};
		}
	}

	/**
	 *  lookupName implements the acmeLib.designRules.Scope interface.  It
	 *  extends the AcmeElement.lookupName method by searching
	 *  through all of its supertypes for a named entity before calling
	 *  its supertype (AcmeElement)'s lookupName method to search for
	 *  ports, properties, and representations contained in this type's
	 *  prototype.
	 *  @param name the name to be mapped to a type or design analyses
	 *  @param bubbleUp boolean, if falso don't check superclass
	 *  @return Object corresponding to "name"
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;

		// check first whether this is referring to "this" type.
		if (name == null) {
			return null;
		}
		else if (name.equalsIgnoreCase("self")) {
			return this;
		};

		// recursively walk up the tree of supertypes hunting for "name"
		for (int i = 0;(i < this.superTypes.getSize()) && (retObj == null); i++) {
			AcmeElementType curSuperType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (!(curSuperType instanceof AcmeUnresolvedElementType))
				retObj = curSuperType.lookupName(name, false);
		};

		// if we didn't find it in a supertype then check for it in this type's
		// prototype object.
		if (retObj == null)
			retObj = super.lookupName(name, bubbleUp);

		return retObj;
	};

	/**
	 *  This method determines whether this Element type's properties, representations,
	 *  and design rule declarations are all type-consistent, and that
	 *  all declared supertypes are AcmeElementProtoType's also.  This method
	 *  delegates all of the actual checking to the AcmeTypeManager class.
	 *  @return whether all types do typecheck
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckElementProtoType(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			return true;
		}
		// port type doesn't typecheck.  There should be an error at the
		// top of the AcmeTypeManager's typecheckingErrorStack.  If there
		// is such an error then announce it to the AcmeErrorManager and
		// return false.
		AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
		if (eltTypeError != null) {
			AcmeErrorManager.errorDetected(eltTypeError);
			return false;
		}
		// no error was reported.  something went wrong?
		// announce a generic error.
		AcmeErrorManager.errorDetected(
			new ElementTypecheckError(this, "Element Type " + this.getTypeName() + " does not typecheck.", this));
		return false;

	};

	//
	// Exporting methods
	//

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltType(s, "Element Type", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltType(s, "Element Type", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this port type as a string of the
	 *  form "Element Type <name> [ extends <Typename>+ ]".  For example:
	 *  "Port Type fooT : portT, superPortT".
	 * @return string containing this component type's signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Element Type " + this.getTypeName());
		AcmeEnumeration superTypesEnum =
			this.getSuperTypes().enumerate(new acmeLib.core.filters.NoBaseClassesEnumFilter());
		if (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" extends <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" extends " + eltType.getTypeName());
		}

		while (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * Creates a workshop for editing this port type.
	 * implements the acmeLib.Editable interface.
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editElementProtoType(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	}

	/* (non-Javadoc)
	 * @see acmeLib.elements.AcmeElement#copy()
	 */
	public AcmeElement copy() {
		return (AcmeElement) this.copyType();
	}

	/* (non-Javadoc)
	 * @see acmeLib.elements.AcmeElement#sameAsInherited()
	 */
	public boolean isSameAsInherited() {
		return false;
	};

}
