
package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.*;
import acmeLib.core.filters.AcmeObjectFilter;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.*;
import acmeLib.properties.AcmeEnum;
import acmeLib.properties.AcmePropertyType;
import acmeLib.properties.EnumType;
import acmeLib.typeManagement.AcmeTypeManager;


/**
 *  Class:  AcmeFamily
<BR>
 *  Purpose:Represents an Acme family description.  A Family is a system type
<BR>
 *          that can store types defined in this family, a set of design
<BR>
 *          analyses available for use in instances of this system type, and
<BR>
 *          a prototype system instance that defines the default and required
<BR>
 *          structure of an instance of this type.
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *     ?/?/97    DK       Created C++ version
<BR>
 *    4/25/97   RTM      Translated to Java
<BR>
 *    6/16/97   RTM      moved from elementTypes package to elements package
<BR>
 *   12/??/97   RTM      converted from standalone class to subclass of
<BR>
 *                        AcmeSystem for type system consistency.
<BR>
 *    1/23/98   RTM      brought into synch with other element types so that
<BR>
 *                        a family is now just a system type (with a typespace).
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeFamily extends AcmeSystem implements AcmeElementType
{
	protected static final FamilyEnumTypeFilter ENUMFILTER = new FamilyEnumTypeFilter ();
  //
  // protected members
  //
  protected AcmeMap typeMap;

  /**
   * super-families
   */
  protected AcmeVector superTypes;


  /**
   * Protected helper method to retrieve the enumerated types declared
   * in this design.
   * @return AcmeEnumeration of the types declared in this design
   */
  protected AcmeEnumeration getEnumTypes ()  {
	return typeMap.enumerate(new AcmeEnumeration(),ENUMFILTER);
  };


  //
  // public members
  //

  /**
   * This is a constructor for this class
   */
  public AcmeFamily () {
    super();
    this.typeMap = new AcmeMap();
    this.name = null;
    this.superTypes = new AcmeVector();
  };

  /**
   * This is a constructor for this class
   * @param name the name of the AcmeFamily to be created
   */
  public AcmeFamily (String name) {
    super();
    this.name = name;
    this.typeMap = new AcmeMap();
    this.superTypes = new AcmeVector();
  };


  //
  // add, remove, and retrieve types declared in this family.
  //
  /**
   * This method adds a type to this family
   * @param type AcmeType to be added to this family
   * @exception  AcmeIllegalNameException thrown when the type's name is illegal
   */
  public void addType(AcmeType type) throws AcmeIllegalNameException {
    if (type == null) {
        return;
    };

    if (LanguageConfig.CHECK_FOR_KEYWORDS)
        if (LanguageConfig.isAcmeKeyword(type.getTypeName()))
            throw new AcmeKeywordNameException("Illegal keyword \""
                + type.getTypeName() + "\" used for typename.");

    if (type.getTypeName() == null) {
        throw new AcmeIllegalNameException(
            "Error.  Attempted to add a type declaration with null typename.");
    }

    else if (eltNamespace.containsName(type.getTypeName()))
        throw new AcmeDuplicateNameException("Attempted to redefine "
            + "identifier " + type.getTypeName());

    eltNamespace.addName(type.getTypeName());
    this.typeMap.addItem(type.getTypeName(), (AcmeObject) type);
    if (type instanceof AcmeElement) {
        ((AcmeElement) type).setParent(this);
        ((AcmeElement) type).setParentDesign(this.getParentDesign());
    } else if (type instanceof AcmePropertyType) {
        ((AcmePropertyType) type).setDeclarationScope(this);
    };
  };


  /**
   * This method retrieves all of the types defined in this family.
   * @return AcmeEnumeration of the types
   */
  public AcmeEnumeration getTypes() {
    return this.typeMap.enumerate();
  };


  /**
   * This method retrieves the type "name" that is declared in this family or one
   * of it's supertype's.
   * @param name the name of the type to be returned
   * @return AcmeType corresponding to "name"
   */
  public AcmeType getType(String name) {
    // first check for the type in this family.
    AcmeType retType = null;
    retType = (AcmeType) this.typeMap.lookupItem(name);
    if (retType != null)
        return retType;

    // else, if it's not found then look for it in each of this family's
    // supertypes.
    AcmeFamily curSuperFam = null;
    for (int i = 0; i < this.superTypes.getSize(); i++) {
        if (this.superTypes.getItemAt(i) instanceof AcmeFamily) {
            curSuperFam = (AcmeFamily) this.superTypes.getItemAt(i);
            retType = curSuperFam.getType(name);
        };
        if (retType != null)
            return retType;
    };

    // if we get to here then the type called "name" was not found, return null.
    return retType;
  };


  /**
   * This method removes a type from a family
   * @param name the name of the type to be removed
   */
  public void removeType(String name) {
    this.typeMap.removeItem(name);
    eltNamespace.removeName(name);
  };


  /**
   * removes the AcmeType <em>type</em> from this design
   * @param type the name of the AcmeType to be removed
   */
  public void removeType(AcmeType type) {
    if (type != null) {
        eltNamespace.removeName(type.getTypeName());
        this.typeMap.removeItem(type.getTypeName());
    };
  };



  /**
   * This method returns the ParentDesign of this family
   * @return the ParentDesign of this family
   */
  public acmeLib.designRules.Scope getScope() { return this.getParentDesign(); };


  /**
   * This method returns the name of this AcmeType
   * @return the name of this type
   */
  public String getTypeName() { return this.getName(); };
 /**
   * This method sets the name of this AcmeType
   * @param newTypeName the new name for the AcmeType
   * @exception AcmeIllegalNameException thrown when the type's name is illegal
   */
  public void setTypeName(String newTypeName) throws AcmeIllegalNameException
  {
    this.setName(newTypeName);
  };



  /**
   * This method returns the supertype of this type description
   * @return  the supertype of this type description
   */
  public AcmeVector getSuperTypes() { return this.superTypes; };


  /**
   * This method adds a supertype to this type description
   * @param superType the supertype to be added to this description
   */
  public void addSuperType(AcmeElementType superType) {
    if ((superType instanceof AcmeFamily)
          || (superType instanceof AcmeUnresolvedElementType)
          || (superType instanceof AcmeElementProtoType))
    {
        superTypes.addItem((AcmeObject) superType);
    };
    // else throw an exception?
  };


  /**
   * This method removes a supertype from this type description
   * @param superType the supertype to be removed
   */
  public void removeSuperType(AcmeElementType superType) {
    superTypes.removeItem((AcmeObject) superType);
  };


  public void copy(AcmeObject newObject)
  {
      newObject = (AcmeObject) copyType();
  }

  /**
   * This method returns a new AcmeFamily that is a deep-copy of this family.
   * @return a new AcmeFamilty which is a copy of this family
   */
  public AcmeElementType copyType() {
    AcmeFamily new_fam = new AcmeFamily(this.getTypeName());

    // walk through the children, adding copies of them to the new system
      //TODO implement walking through children?
	AcmeEnumeration child_enumeration = new AcmeEnumeration();
	AcmeEnumeration current;

    // this try block should not encounter any double names or keywords,
    // as there should be no way for them to be added to the original
    // system that is being copied, so just ignore either of the
    // Acme*NameExceptions.
    try {
     	current = portMap.enumerate();
    	while (current.moreItems())
    	   new_fam.addPort( (AcmePort) (((AcmePort) current.getNextItem())).copy());

    	current = roleMap.enumerate();
    	while (current.moreItems())
    	   new_fam.addRole( (AcmeRole) (((AcmeRole) current.getNextItem())).copy() );

    	current = componentMap.enumerate();
    	while (current.moreItems())
    	   new_fam.addComponent( (AcmeComponent) (((AcmeComponent) current.getNextItem())).copy() );

    	current = connectorMap.enumerate();
    	while (current.moreItems())
    	   new_fam.addConnector( (AcmeConnector) (((AcmeConnector) current.getNextItem())).copy() );

    	current = attachments.enumerate();
    	while (current.moreItems())
    	   new_fam.addAttachment( (AcmeAttachment) (((AcmeAttachment) current.getNextItem())).copy() );

        // RTM 1/15/98 -- Currently, just shallow copy the map of design analyses
        //                because none of the design rule/analyses/expressions
        //                have deep-copy methods implemented
        new_fam.designAnalysisMap = this.designAnalysisMap;

        this.copyCommonEltStructure(new_fam);
        this.superTypes.deepCopyItems(new_fam.superTypes);

    }
    catch (AcmeIllegalNameException ain_ex) {
        // if this is copying an existing family, assume that all checks
        // were enforced in copying the previous family, so this should
        // cause no exceptions.
    };

    return new_fam;
  };


  /**
   * This method updates the child type name if it's a type declared in this family,
   * otherwise let the superclass take care of it.
   * @param oldName the name of the old child type
   * @param obj an Object, checked for an instance of AcmeType
   * @exception AcmeIllegalNameException thrown when the type's name is illegal
   */
  public void childNameChanged(String oldName, AcmeObject obj)
    throws AcmeIllegalNameException
  {
    if (obj instanceof AcmeType) {
        this.typeMap.removeItem(oldName);
        this.addType((AcmeType) obj);
    } else {
        super.childNameChanged(oldName, obj);
    }
  };


  //
  // exporting
  //

  /**
   * This method adds the types declared in this family to the other structure, properties
   * and rules retreived in the superClass' method.
   * @return AcmeEnumeration of all types decladed in this family and all other structures,properties
   * and rules
   */
  protected AcmeEnumeration getChildrenToExport() {
    AcmeEnumeration retEnum = super.getChildrenToExport();
    retEnum.concatenate(this.getTypes());
    return retEnum;
  };


  /**
   * This method exports an ElementType using an AcmeOutputStream
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs when writing
   * @exception AcmeException for acme exception
   */
  public void exportAcme(AcmeOutputStream s)
     throws java.io.IOException, AcmeException
  {
    this.exportEltType(s, "Family", this.getChildrenToExport(),
                           ExportLanguage.ACME);
  };


  /**
   * This method exports an ElementType using an AcmeOutputStream
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs when writing
   * @exception AcmeException for acme exception
   */
  public void exportArmani(AcmeOutputStream s)
     throws java.io.IOException, AcmeException
  {
    this.exportEltType(s, "Family", this.getChildrenToExport(),
                           ExportLanguage.ARMANI);
  };



  /**
   * Returns the type signature for this family as a string of the
   *  form "Family <name> [ extends <Typename>+ ]".  For example:
   *  "Family fooT : famT, superFamT".
   * @return string containing this family's type's signature.
   **/
  public String getTypeSignature() {
    StringBuffer buf = new StringBuffer();
    if (acmeLib.core.globals.LanguageConfig.IS_ARMANI == true) {
        buf.append("Style " + this.getTypeName());
    } else {
        buf.append("Family " + this.getTypeName());
    };

    AcmeEnumeration superTypesEnum =
            this.getSuperTypes().enumerate(
                    new acmeLib.core.filters.NoBaseClassesEnumFilter());
    if (superTypesEnum.moreItems()) {
        AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
//        if (eltType instanceof AcmeUnresolvedElementType)
//            buf.append(" extends <unresolved type name: " + eltType.getTypeName() + "> ");
//        else
            buf.append(" extends " + eltType.getTypeName());
    }

    while (superTypesEnum.moreItems()) {
        AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
        buf.append(", " + eltType.getTypeName());
    }

    return buf.toString();
  };


  /**
   *  This method implements the acmeLib.designRules.Scope interface to map names
   *  to types or design analyses specified in the family.  The family
   *  lookup just tries to retrieve type and design analysis definitions,
   *  not substructure declared in a family, this is deferred to the supertype
   *  AcmeSystem's lookupName method if bubbleUp is true.
   *  @param name the name to be mapped to a type or design analyses
   *  @param bubbleUp boolean, if false don't check superclass
   *  @return Object corresponding to "name"
   */
  public Object lookupName(String name, boolean bubbleUp) {

    // first check to see if name refers to "this" family
    if (name == null) {
        return null;
    } else if (name.equalsIgnoreCase("self")) {
        return this;
    };

    // lookup the child object called 'name'.
    Object retObj = null;
    retObj = this.getType(name);
    if (retObj == null) {
        retObj = this.getDesignAnalysis(name);
    };

    // check to see if it's a valid enumerated type value name.
    if (retObj == null) {
        AcmeEnumeration enumTypes = this.getEnumTypes();
        while(enumTypes.moreItems() && (retObj == null)) {
            EnumType curEnumType = (EnumType) enumTypes.getNextItem();
            AcmeEnum curVal = (AcmeEnum) curEnumType.lookupName(name, false);
            if (curVal != null) {
                retObj = curVal;
            };
        };
    };


    // if we've found what we're looking for then return it immediately...
    if (retObj != null) {
        return retObj;
    };


    // if the named entity has not yet been found, repeat for all of this
    // family's supertypes (super-families).
    for(int i = 0; i < this.superTypes.getSize(); i++) {
        try {
            retObj = ((AcmeFamily) superTypes.getItemAt(i)).lookupName(name,bubbleUp);
        } catch (ClassCastException ccExc) {
            // if it's not an AcmeFamily then we can't look into it, carry on.
        };
        if (retObj != null)
            return retObj;
    };

    // finally, if we haven't found anything in the family-specific checks
    // look for family-default structure in the System superclass.  If
    // "bubbleUp" is set to false then don't check the superclass to prevent
    // infinite loops.
    if ((bubbleUp) && (retObj == null)) {
        retObj = super.lookupName(name, false);
    };

    // finally, if we can't find anything in the family check for types
    // and design analyses in the design enclosing the family.
    if ((retObj == null) && bubbleUp && (this.getParentDesign() != null)) {
        retObj = this.getParentDesign().lookupName(name, false);
    }

    return retObj;
  };



  /**
   *  instantiateElement() calls instantiateElements() on all of the types
   *  defined in this family, and then calls its super.instantiateElements()
   *  to instantiate all of the default structure in the family via the
   *  AcmeSystem superclass.
   */
  public void instantiateElements()
  {
    AcmeEnumeration e = this.getTypes();
    while (e.moreItems()) {
        AcmeObject nextObj = e.getNextItem();
        if (nextObj instanceof AcmeElement) {
            ((AcmeElement) nextObj).instantiateElements();
        };
    };
    super.instantiateElements();
  };




  /**
   * This method resolves the names of all of this family's supertypes, and the types
   * referred to by all of the substructure, types, design analyses, and
   * design rules of this family.
   */
  public void resolveTypeReferences() {

    // first resolve the names of any unresolved family supertypes
    //  (super-families)
    AcmeElementType curType = null;
    Object resolvedType = null;
    AcmeUnresolvedElementType curUnresolvedType;
    for (int i = 0; i < this.superTypes.getSize(); i++) {
        curType = (AcmeElementType) this.superTypes.getItemAt(i);
        if (curType instanceof AcmeUnresolvedElementType) {
            curUnresolvedType = (AcmeUnresolvedElementType) curType;
            // dereference this family super-type's object in this' design.
             resolvedType = this.lookupName (curUnresolvedType.getTypeName (), true);
             if (resolvedType == null) {
                String errorMsg = "Unable to resolve family supertype \""
                                  + curUnresolvedType.getTypeName()
                                  + "\" of family type "
                                  + this.getTypeName();
                TypeRefResolutionError badTypeRefError =
                    new TypeRefResolutionError(this, errorMsg,
                                           curUnresolvedType.getTypeName());
                badTypeRefError.setUnresolvedRef(curUnresolvedType);
                AcmeErrorManager.errorDetected(badTypeRefError);

            } else if (!(resolvedType instanceof AcmeFamily)
                        && !(resolvedType instanceof AcmeElementProtoType))
            {
                String errorMsg = "Family \"" + this.getTypeName()
                                  + "\"'s supertype \""
                                  + curUnresolvedType.getTypeName()
                                  + "\" is neither a family nor an element type.";
                TypeRefResolutionError badTypeRefError =
                    new TypeRefResolutionError(this, errorMsg,
                                           curUnresolvedType.getTypeName());
                badTypeRefError.setUnresolvedRef(curUnresolvedType);
                AcmeErrorManager.errorDetected(badTypeRefError);

                resolvedType = null;
            };

            if (resolvedType != null) {
                this.superTypes.setItemAt((AcmeObject) resolvedType, i);
            };
        };
    };

    // then walk over all of the types declared in this system and resolve
    // their type names
    AcmeEnumeration e = this.getTypes();
    while (e.moreItems()) {
        AcmeObject curObj = e.getNextItem();
        if (curObj instanceof AcmeElementProtoType) {
            ((AcmeElementProtoType) curObj).resolveTypeReferences();
        }
        else if (curObj instanceof AcmeComponentType) {
            ((AcmeComponentType) curObj).resolveTypeReferences();
        }
        else if (curObj instanceof AcmeConnectorType) {
            ((AcmeConnectorType) curObj).resolveTypeReferences();
        }
        else if (curObj instanceof AcmePortType) {
            ((AcmePortType) curObj).resolveTypeReferences();
        }
        else if (curObj instanceof AcmeRoleType) {
            ((AcmeRoleType) curObj).resolveTypeReferences();
        }
        else if (curObj instanceof AcmePropertyType) {
            ((AcmePropertyType) curObj).resolveTypeReferences(this);
        }
    };

    // finally, resolve the names of all substructure and design analyses
    // declared in the substructure.  This is all already implemented in
    // AcmeSystem.resolveTypeReferences(), so use it.
    super.resolveTypeReferences();
  };


  /**
   * This method resolves instance references in all of this system's attachments,
   * and recursively resolves all of the instance references of this
   * system's substructure (components, connectors, ports, and roles).
   */
  public void resolveInstanceReferences() {

    // walk over all of the types declared in this system and resolve
    // any instance references made within them.
    AcmeEnumeration e = this.getTypes();
    while (e.moreItems()) {
        AcmeObject nextObj = e.getNextItem();
        if (nextObj instanceof AcmeElement) {
            ((AcmeElement) nextObj).resolveInstanceReferences();
        }
    }

    // finally, resolve the names of all substructure and design analyses
    // declared in the substructure.  This is all already implemented in
    // AcmeSystem.resolveTypeReferences(), so use it.
    super.resolveInstanceReferences();
  };



  /**
   * This method determines whether this family's properties, representations,
   * design rules, and substructure declarations are all type-consistent, and
   * that all declared supertypes are AcmeConnectorType's also.  This method
   * delegates all of the actual checking to the AcmeTypeManager class.
   * @return whether all types do typecheck
   */
  public boolean typecheck() {
    boolean typechecks = AcmeTypeManager.typecheckFamily(this);
    if (typechecks) {
        // things are hunky-dory, return true.
        AcmeErrorManager.errorRepaired (this, null);
        return true;
    }
        // connector type doesn't typecheck.  There should be an error at the
        // top of the AcmeTypeManager's typecheckingErrorStack.  If there
        // is such an error then announce it to the AcmeErrorManager and
        // return false.
        AcmeError eltTypeError =
            AcmeTypeManager.typecheckingErrorStack.popCausalError();
        if (eltTypeError != null) {
            AcmeErrorManager.errorDetected(eltTypeError);
            return false;
        }
            // no error was reported.  something went wrong?
            // announce a generic error.
            AcmeErrorManager.errorDetected(
                new ElementTypecheckError(this, "Family "
                        + this.getTypeName() + " does not typecheck.", this));
            return false;
  };


  /**
   * retrieves all of the types that are declared in this family, this
   * family's supertypes, and the design in which this family is declared.
   */
  public AcmeEnumeration getAvailableTypes() {
    AcmeEnumeration allAvailableTypes = this.getTypes();
    for (int i = 0; i < this.superTypes.getSize(); i++) {
        AcmeObject curObj = this.superTypes.getItemAt(i);
        if (curObj instanceof AcmeElement) {
            allAvailableTypes.concatenate(
                                ((AcmeElement) curObj).getAvailableTypes());
        };
    };

    if (this.getParentDesign() != null) {
        allAvailableTypes.concatenate(this.getParentDesign().getAvailableTypes());
    };

    return allAvailableTypes;
  };


  /**
   * implement the acmeLib.Editable interface.
   * It also creates a workshop for editing this element
   */
  public void edit() {
//    if (this.editor == null) {
//        this.editor = new ElementWorkshopFrame();
//        editor.editFamily(this);
//    };
//    editor.setVisible(true);
//    editor.toFront();
  };

  public void view() {
    System.out.println("Attempted to view: " + this.getName());
    System.out.println("  but AcmeElement.view() is not yet implemented.");
  };

};



/**
 * This class is an enumeration filter to remove the "SYSTEM_BASE_TYPE"
 */
class NoBaseSystemClassEnumFilter extends AcmeObjectFilter {
  /**
   * This method returns whether obj is of SYSTEM_BASE_TYPE
   * @param obj the AcmeObject, whose type is to be tested
   * @return  whether obj is of SYSTEM_BASE_TYPE
   */
    public boolean test(AcmeObject obj) {
        return ( obj != ConstantTypes.SYSTEM_BASE_TYPE );
    };
};



/**
 * This class is a filter class to return only enum filters.
 */
class FamilyEnumTypeFilter extends AcmeObjectFilter {
  /**
   * This method returns whether obj is an instance of EnumType
   * @param obj the AcmeObject, which is to be tested
   * @return  whether obj is an instance of EnumType
   */
    public boolean test(AcmeObject obj) {
        return ( obj instanceof EnumType );
    };
};

