package acmeLib.elements;

import java.util.Enumeration;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.*;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.TypeInstantiationError;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 *  Class:    AcmeGroup
<BR>
 *  Purpose:  Stores the structure of an ACME group
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    4/24/97   RTM     Translated to Java
<BR>
 *    5/30/97   RTM     All stubbed methods filled in
<BR>
 *    6/13/97   RTM     Modified to work with second generation object model
<BR>
 *    1/14/98   RTM     added support for multiple declared types.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeGroup extends AcmeElement {

	//
	// Private members
	//
	private AcmeMap memberMap;
    private AcmeVector memberSet;
	//
	// Protected members
	//

	//
	// Public members
	//
	/**
	 * This is a constructor for this class
	 * @param name used to initialize this.name
	 */
	public AcmeGroup(String name) {
		memberMap = new AcmeMap();
        memberSet = new AcmeVector ();
		this.name = name;
	};

	/**
	  * This is a constructor for this class
	  * @param name used to initialize this.name
	  * @param type AcmeGroupType
	  */
	public AcmeGroup(String name, AcmeGroupType type) {
		memberMap = new AcmeMap();
        memberSet = new AcmeVector ();
		this.name = name;
		addDeclaredType(type);
	};

	/**
	 * This method does a deep copy of "this" group
	 * @return the copy of the AcmeElement
	 */
	public AcmeElement copy()  {
		AcmeGroup new_group = new AcmeGroup(this.getName());
		this.copyCommonEltStructure(new_group);
        try {
            memberMap.copyItems(new_group.memberMap);
        } catch (CloneNotSupportedException e) {
            Enumeration enum = memberMap.keys();
            while (enum.hasMoreElements()) {
                String key = (String) enum.nextElement();
                new_group.memberMap.addItem(key, this.memberMap.lookupItem(key));
            }
        }
        memberSet.copyItems(new_group.memberSet);
		return new_group;
	};

	/**
	 * This method uses this element instance as a prototype to create a new type.
	 * @param newTypeName used to create a new  AcmeGroupType
	 * @return the new AcmeGroupType, created by this function
	 */
	public AcmeElementType copyAsPrototype(String newTypeName) {
		AcmeGroupType new_group_type = new AcmeGroupType(newTypeName);
		this.copyCommonEltStructure(new_group_type);

		// copyCommonEltStructure overwrites the name of the new type, so
		// reset it properly.  Don't check for legal names here, that will
		// happen if/when we try to add it to anything.
		new_group_type.name = newTypeName;

		// add all of the declared types of this instance as supertypes of
		// the new type.
		AcmeEnumeration e = this.getDeclaredTypes();
		while (e.moreItems()) {
			AcmeObject curType = e.getNextItem();
			if (curType instanceof AcmeGroupType) {
				new_group_type.addSuperType((AcmeGroupType) curType);
			}
			else if (curType instanceof AcmeElementProtoType) {
				new_group_type.addSuperType((AcmeElementProtoType) curType);
			}
			else if (curType instanceof AcmeElementTypeRef) {
				AcmeElementTypeRef t = (AcmeElementTypeRef) curType;
				if (t.getReferentType() instanceof AcmeGroupType
					|| t.getReferentType() instanceof AcmeElementProtoType) {
					new_group_type.addSuperType((AcmeElementTypeRef) curType);
				}

			}
		};

		return new_group_type;
	};

	/**
	 * All AcmeGroups satisfy the type ConstantTypes.COMPONENT_BASE_TYPE.
	 * If the generic supertype method doesn't find the type declared, check
	 * whether the type is the COMPONENT_BASE_TYPE.
	 * @param type the type of the AcmeElement
	 * @return whether the type is COMPONENT_BASE_TYPE
	 */
	public boolean declaresType(AcmeElementType type) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType ();
		}
		if (type == ConstantTypes.GROUP_BASE_TYPE) {
			return true;
		}
		return super.declaresType(type);
	};

	//
	// exporting
	//

	/**
	 * @return  AcmeEnumeration
	 */
	protected AcmeEnumeration getChildrenToExport() {
		AcmeEnumeration retEnum = super.getChildrenToExport();
//		retEnum.concatenate(this.getMembers());
		return retEnum;
	}

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		exportComment (s);
		s.beginRegion (this);
		AcmeEnumeration children = this.getChildrenToExport ();
		int headerFlag = writeElementHeader (s, "Group", children);
		headerFlag = writeEltChildren (s, children, ExportLanguage.ACME, headerFlag);
		AcmeEnumeration members = this.getMembers ();
		if (members.moreItems ()) {
			if ((headerFlag & CANDIDATE_EXTENDS_WITH) == CANDIDATE_EXTENDS_WITH && !((headerFlag & WROTE_OPENING_BRACE) == WROTE_OPENING_BRACE)) {
				s.writeln (" extended with {");
				headerFlag |= WROTE_OPENING_BRACE;
			}
			if ((headerFlag & WROTE_OPENING_BRACE) != WROTE_OPENING_BRACE) {
				s.writeln (" = {");
				headerFlag |= WROTE_OPENING_BRACE;
			}
			s.indent ();
			s.writeln ("Members {");
			s.indent ();
			AcmeElement context = this.getParent();
			String contextName = context.getQualifiedName ();
			while (members.moreItems()) {
				AcmeElement element = (AcmeElement) members.getNextItem();
				String elementName = element.getQualifiedName ();
				String name = elementName.substring (contextName.length () + 1);
				s.write (name);
				s.write("; ");
			}
			s.unIndent();
			s.writeln ();
			s.writeln ("};");
			s.unIndent ();
		}
		s.endRegion (this);
		if ((headerFlag & WROTE_OPENING_BRACE) == WROTE_OPENING_BRACE) {
			s.write ("}");
		}
		s.writeln (";");
		s.endRegion (this);
		s.flush ();
	};

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception  AcmeException for acme exception
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		exportAcme (s);
	};

	/**
	 * Returns the type signature for this group as a string with the
	 *  form "Group <name> [ : <Typename>+ ]".  For example:
	 *  "Group foo : compT".
	 * @return string containing this element's type signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Group " + this.getName());
		AcmeEnumeration typeEnum = this.getDeclaredTypes();
		if (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" : <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" : " + eltType.getTypeName());
		}

		while (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * This method implements the acmeLib.designRules.Scope interface to map names
	 * to elements or properties
	 * It  lookups the child object called 'name'.
	 * First it checks for roles, then has AcmeElement check for properties,
	 * and representations.
	 * @param name is compared with "roles"
	 * @param bubbleUp boolean
	 * @return Object corresponding to name
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		// lookup the child object called 'name'.
		// First check for roles, then have AcmeElement check for properties,
		// and representations.
		Object retObj = null;

		// check for the special case of a request for all of this connector's roles
		if (name != null) {
			if (name.equalsIgnoreCase("members"))
				return this.getMembers();
		};

		if (retObj == null) {
			retObj = this.getMember(name);
		}

		if (retObj == null)
			// just check for a property or representation, don't bubble up in
			// the generic element lookupName method.
			retObj = super.lookupName(name, bubbleUp);

		return retObj;
	};

	/**
	 *  AcmeGroup.resolveTypeReferences() relies primarily on it's superclass'
	 *  implementation to do type name resolution.  This subclass adds a
	 *  recursive call to resolveTypeReferences() for each of this Group's ports.
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
	};

	/**
	 * Recursively resolve all of the instance references of this
	 * system's substructure
	 */
	public void resolveInstanceReferences() {
		super.resolveInstanceReferences();
		Enumeration e = this.memberMap.keys();
		while (e.hasMoreElements()) {
			String key = (String) e.nextElement();
			AcmeElement el = (AcmeElement) this.memberMap.lookupItem(key);
			if (el instanceof AcmeUnresolvedElement) {
                if (el.getParent() == null) {
                    el.setParent(this.getParent ());
                }
				AcmeElement newEl = ((AcmeUnresolvedElement) el).getResolvedInstanceReference();
				if (newEl != null) {
					this.memberMap.removeItem(key);
					this.memberMap.addItem(key, newEl);
				}
			}
		}

	};

	/**
	 * instantiateElements() walks over the vector of instantiatedTypes of this
	 * element and adds the minimal structure declared in the type description
	 * to this element.  It delegates much of the work to the AcmeTypeManager
	 * class, so that we can change type instantiation policies relatively
	 * centrally (ie by changing the AcmeTypeManager methods) for all elements
	 * at once.  This method recursively calls the same method on all of the
	 * substructure of this element instance.
	 */
	public void instantiateElements() {

		// unify this element with each of the types it instantiates.
		for (int i = 0; i < this.instantiatedTypes.getSize(); i++) {
			// check that we have a prototype group or group type
			// to instantiate.  If we don't then announce an error and move
			// to the next type.
			AcmeObject iType = this.instantiatedTypes.getItemAt(i);
			if (iType instanceof AcmeElementTypeRef) {
				iType = (AcmeObject) ((AcmeElementTypeRef) iType).getReferentType();
			}
			if (iType instanceof AcmeGroupType) {
				AcmeTypeManager.unifyGroup((AcmeGroupType) iType, this);

			}
			else if (iType instanceof AcmeElementProtoType) {
				AcmeTypeManager.unifyElementWithPrototype((AcmeElementProtoType) iType, this);

			}
			else if (iType instanceof AcmeType) {
				String errorMsg = null;
				if (iType instanceof AcmeUnresolvedElementType) {
					errorMsg =
						"Instantiation Error.  Unable to instantiate "
							+ " unresolved element type: "
							+ ((AcmeUnresolvedElementType) iType).getTypeName()
							+ " in group: "
							+ this.getName();
				}
				else {
					errorMsg =
						"Instantiation Error.  Type: "
							+ ((AcmeType) iType).getTypeName()
							+ " is neither a group nor an element type.  Therefore it "
							+ "can not be used to instantiate group: "
							+ this.getName();
				};

				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg, (AcmeType) iType);
				AcmeErrorManager.errorDetected(instError);

			}
			else {
				String errorMsg =
					"Instantiation Error.  Attempted to instantiate a "
						+ "completely unrecognizable type in group: "
						+ this.getName();
				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg);
				AcmeErrorManager.errorDetected(instError);
			};
		}

		// finally, call AcmeElement to have it instantiate representations and any
		// other standard element structure.
		super.instantiateElements();
	};

	/**
	 * This method confirms that this group satisfies all of its declared types and
	 * design rules, and that it's properties, representations, and ports
	 * all typecheck also.
	 * @return whether the group satisfies all of its declared types
	 */
	public boolean typecheck() {
		return true;
	};

	/**
	 * Creates a workshop for editing this element.
	 * implements the acmeLib.Editable interface.
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editGroup(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

	public void view() {
		System.out.println("Attempted to view: " + this.getName());
		System.out.println("  but AcmeGroup.view() is not yet implemented.");
	};

	//
	// Member manipulations
	//
	/**
	 * @param r AcmeElement to be added
	 * @exception AcmeIllegalNameException member name is illegal
	 */
	public void addMember(AcmeElement r) throws AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isAcmeKeyword(r.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + r.getName() + "\" used for member name.");
		};

		if (r.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a member with null name.");
		}

//		eltNamespace.addName(r.getName());
		memberMap.addItem(r.getQualifiedName(), r);
        memberSet.addItem(r);
	};

	/**
	 * This method removes a member from this connector
	 * @param r AcmeElement to be removed
	 */
	public void removeMember(AcmeElement r) {
		removeMember(r.getQualifiedName());
        if (containsMember(r)) {
            // The name didn't match anything, so we should go through and find the member key
            String realKey = null;
            Enumeration enum = memberMap.keys();
            while (realKey == null && enum.hasMoreElements()) {
                String key = (String )enum.nextElement();
                if (memberMap.lookupItem (key) == r) {
                    realKey = key;
                }
            }
            if (realKey != null) {
                removeMember (realKey);
            }
        }
	};

	/**
	 * This method removes a member from this connector
	 * @param name name of AcmeElement to be removed
	 */
	public void removeMember(String name) {
		AcmeElement r = (AcmeElement) memberMap.lookupItem(name);
		if (r == null)
			return; // there's no member called "name" to remove
		else {
			memberMap.removeItem(name);
			eltNamespace.removeName(name);
            memberSet.removeItem (r);
		};
	};

	/**
	 * This method returns an enumeration of the members
	 * @return AcmeEnumeration of the memberMap
	 */
	public AcmeEnumeration getMembers() {
		return memberMap.enumerate();
	};

	/**
	 * This method returns a member with a specific name
	 * @param name name of the AcmeElement item
	 * @return AcmeElement corresponding to name
	 */
	public AcmeElement getMember(String name) {
		return (AcmeElement) memberMap.lookupItem(name);
	};

	/**
	 * This method does a deep copy of the roles in this connector and put
	 * them into the roles list of destConn.  This removes
	 * any roles that are currently in destConn
	 * @param destConn AcmeConnector
	 */
	protected void copyMembers(AcmeGroup destGroup) {
		AcmeEnumeration members_enum = this.getMembers();
		destGroup.memberMap = new AcmeMap();

		AcmeElement cur_member;
		while (members_enum.moreItems()) {
			cur_member = (AcmeElement) members_enum.getNextItem();
			try {
				destGroup.addMember((AcmeElement) cur_member.copy());
			}
			catch (AcmeIllegalNameException ainExc) {
				// if we're copying a port that has been legally added elsewhere
				// into a fresh portmap then it won't have an illegal name.
			};

		}
	}

    /**
     * 
     */
    public void removeAllMembers() {
        memberMap.clear();
        memberSet.removeAllItems();
    }
    
    public boolean containsMember (AcmeElement member) {
        return memberSet.contains(member);
    }

};
