package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.TypeRefResolutionError;

/**
 *  Class:   AcmeGroupType
<BR>
 *  Purpose:  Stores the type specification for a group.  This is
<BR>
 *          a subclass of an AcmeGroup because the Acme type system
<BR>
 *          is prototype-based.  As a result, a type is just a prototypical
<BR>
 *          instance of a group that specifies the aspects that an
<BR>
 *          instance of that type must posses.  Typechecking is done
<BR>
 *          by an external class, AcmeTypeManager, that understands the
<BR>
 *          rules of typechecking
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *   6/16/97   RTM     Created class
<BR>
 *   1/14/98   RTM     added support for multiple declared types.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeGroupType extends AcmeGroup implements AcmeElementType {

	//
	// protected members
	//
	protected AcmeVector superTypes;

	//
	// public members
	/**
	 * This is the constructor for this class.
	 * It creates the prototype object in the superclass constructor
	 * @param typeName the name of the componend type
	 */
	public AcmeGroupType(String typeName) {
		// create the prototype object in the superclass constructor
		super(typeName, null);
		this.superTypes = new AcmeVector();
	};

	/**
	 * @return the name of the type
	 */
	public String getTypeName() {
		return this.getName();
	};
	/**
	  * @param typeName used to set the name of the type
	  * @exception  AcmeIllegalNameException type name is illegal
	  */
	public void setTypeName(String typeName) throws AcmeIllegalNameException {
		this.setName(typeName);
	};

	// methods for adding supertypes to a type description and retrieving
	// a type's set of supertypes
	/**
	 * @param superType AcmeElementType, which is to be added to superTypes
	 */
	public void addSuperType(AcmeElementType superType) {
		if (superType instanceof AcmeElementTypeRef) {
			superType = ((AcmeElementTypeRef) superType).getReferentType();
		}
		if ((superType instanceof AcmeGroupType)
			|| (superType instanceof AcmeUnresolvedElementType)
			|| (superType instanceof AcmeElementProtoType)) {
			superTypes.addItem((AcmeObject) superType);
		}
		// else throw an exception?
	};

	/**
	 * @param superTypes AcmeElementType, which is to be removed from superTypes
	 */
	public void removeSuperType(AcmeElementType superType) {
		superTypes.removeItem((AcmeObject) superType);
		if (superType instanceof AcmeElementTypeRef) {
			declaredTypes.removeItem((AcmeObject) ((AcmeElementTypeRef) superType).getReferentType());
		}
	};

	/**
	 * @return superTypes
	 */
	public AcmeVector getSuperTypes() {
		return this.superTypes;
	};

	/**
	 * This method makes a new group type object, with a deep copy of the old group's
	 * properties, representations, declared, instantiated, and super-type
	 * vectors.
	 * @return AcmeElementType
	 */
	public AcmeElementType copyType() {
		AcmeGroupType new_group_type = new AcmeGroupType(this.getTypeName());
		this.copyCommonEltStructure(new_group_type);
		this.superTypes.copyItems(new_group_type.superTypes);
		return new_group_type;
	};

	/** ResolveTypeReferences searches through this type object's vector of supertypes
	 *  for UnresolvedElementType objects, looks up the actual type that is
	 *  referred to by each UnresolvedElementType object, and replaces that
	 *  object with the actual type object that should be referred to.
	 *  it then calls resolveTypeReferences in its superclass to resolve the type names
	 *  of all of its substructure (ports and representations).
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
		AcmeElementType curType = null;
		AcmeType resolvedType = null;
		AcmeUnresolvedElementType curUnresolvedType;
		for (int i = 0; i < this.superTypes.getSize(); i++) {
			curType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (curType instanceof AcmeUnresolvedElementType) {
				curUnresolvedType = (AcmeUnresolvedElementType) curType;
				if (curUnresolvedType.isQualified()) {
					if (this.getParentDesign() != null) {
						AcmeFamily qFam = this.getParentDesign().getFamily(curUnresolvedType.getQualifier());
						if (qFam != null) {
							AcmeType refType = qFam.getType(curUnresolvedType.getTypeName());
							if (refType != null && refType instanceof AcmeElementType) {
								resolvedType = new AcmeElementTypeRef(qFam, (AcmeElementType) refType);
							}
						}
					}
				}
				else {
					// first check to see if the supertype is declared in the same
					// family that this type is declared in.  If an AcmeGroupType
					// object's parent field is not null then parent must be a family.
					if (this.getParent() instanceof AcmeFamily) {
						resolvedType = ((AcmeFamily) this.getParent()).getType(curUnresolvedType.getTypeName());
					}

					if (this.getParent() instanceof AcmeDesign) {
						resolvedType = ((AcmeDesign) this.getParent()).getType(curUnresolvedType.getTypeName());
					}
					else if ((resolvedType == null) && (this.getParentDesign() != null)) {
						// then check to see if the supertype is declared in the same
						// design as this type.
						resolvedType = this.getParentDesign().getType(curUnresolvedType.getTypeName());
					}
				}
				if (resolvedType == null) {
					String errorMsg =
						"Unable to resolve supertype \""
							+ curUnresolvedType.getTypeName()
							+ "\" of group type "
							+ this.getTypeName();
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

				}
				else if (resolvedType instanceof AcmeElementTypeRef) {
					AcmeElementTypeRef ref = (AcmeElementTypeRef) resolvedType;
					if (!(ref.getReferentType() instanceof AcmeGroupType)
						&& !(ref.getReferentType() instanceof AcmeElementProtoType)) {
						String errorMsg =
							"Group type \""
								+ this.getTypeName()
								+ "\"'s supertype \""
								+ curUnresolvedType.getTypeName()
								+ "\" is not a group or element type.";
						TypeRefResolutionError badTypeRefError =
							new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
						badTypeRefError.setUnresolvedRef(curUnresolvedType);
						AcmeErrorManager.errorDetected(badTypeRefError);

						resolvedType = null;
					}
				}
				else if (!(resolvedType instanceof AcmeGroupType) && !(resolvedType instanceof AcmeElementProtoType)) {
					String errorMsg =
						"Group type \""
							+ this.getTypeName()
							+ "\"'s supertype \""
							+ curUnresolvedType.getTypeName()
							+ "\" is not a group or element type.";
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

					resolvedType = null;
				}
				else if (resolvedType == this) {
					// Is there an immediate recursive resolution
					String errorMsg = "Group type \"" + this.getTypeName () + "\" cannot have itself as a supertype.";
					TypeRefResolutionError badTypeError = new TypeRefResolutionError (this, errorMsg, resolvedType.getTypeName());
					badTypeError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected (badTypeError);
					
					resolvedType = null;
				}

				if (resolvedType != null) {
					this.superTypes.setItemAt((AcmeObject) resolvedType, i);
				};

			};
		}
	}

	/**
	 *  lookupName implements the acmeLib.designRules.Scope interface.  It
	 *  extends the AcmeGroup.lookupName method by searching
	 *  through all of its supertypes for a named entity before calling
	 *  its supertype (AcmeGroup)'s lookupName method to search for
	 *  ports, properties, and representations contained in this type's
	 *  prototype.
	 *  @param name compared with "self"
	 *  @param bubbleUp used in the super.lookupName method
	 *  @return an Object corresponding to name
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;

		// check first whether this is referring to "this" type.
		if (name == null) {
			return null;
		}
		else if (name.equalsIgnoreCase("self")) {
			return this;
		};

		// recursively walk up the tree of supertypes hunting for "name"
		for (int i = 0;(i < this.superTypes.getSize()) && (retObj == null); i++) {
			AcmeElementType curSuperType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (!(curSuperType instanceof AcmeUnresolvedElementType))
				retObj = curSuperType.lookupName(name, false);
		};

		// if we didn't find it in a supertype then check for it in this type's
		// prototype object.
		if (retObj == null)
			retObj = super.lookupName(name, bubbleUp);

		return retObj;
	};

	/**
	 *  This method determines whether this group type's properties, representations,
	 *  design rules, and port declarations are all type-consistent, and that
	 *  all declared supertypes are AcmeGroupType's also.  This method
	 *  delegates all of the actual checking to the AcmeTypeManager class.
	 *  @return whether group type typechecks.
	 */
	public boolean typecheck() {
		return true;
	};

	//
	// Exporting methods
	//

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltType(s, "Group Type", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltType(s, "Group Type", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this group type as a string of the
	 *  form "Group Type <name> [ extends <Typename>+ ]".  For example:
	 *  "Group Type fooT : compT, superCompT".
	 * @return string containing this group type's signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Group Type " + this.getTypeName());
		AcmeEnumeration superTypesEnum =
			this.getSuperTypes().enumerate(new acmeLib.core.filters.NoBaseClassesEnumFilter());
		if (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" extends <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" extends " + eltType.getTypeName());
		}

		while (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * Creates a workshop for editing this group type.
	 * implements the acmeLib.Editable interface.
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editGroupType(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

}
