// AcmeMapInstance.java
package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.exceptions.AcmeDuplicateNameException;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeKeywordNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.designRules.DesignRule;
import acmeLib.errorManagement.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: AcmeMapInstance
 * Purpose: Stores the structure of an Acme mapping instance
 * 
 * Revision History:
 * Data:      Person    Change
 * 11/14/99    JH        Created class
 * @author Jianing Hu
 */

/**
 * This is a constructor for this class
 * @param name used to initialize this.name
 */
public class AcmeMapInstance extends AcmeElement {
	/**
	 * The list of maplets
	 * Though maplets are nameless, we'll give each a "inner" name for convenience.
	 * A maplet will be named source-target, while both source and target will be
	 *  named by the concatenation of their elements, namely e1_e2_...
	 */
	private AcmeMap mapletMap;

	public AcmeMapInstance() {
		super();
		mapletMap = new AcmeMap();
	}

	public AcmeMapInstance(String name) {
		super(name);
		mapletMap = new AcmeMap();
		this.name = name;
	}

	/** 
	 * This method does a deep copy of the maplets in this map and puts 
	 * them into the maplets list of destMap. This removes any maplets
	 * that are currently in destMap.
	 * @param destMap AcmeMapInstance
	 */
	protected void copyMaplets(AcmeMapInstance destMap) {
		AcmeEnumeration maplet_enum = this.getMaplets();
		destMap.mapletMap = new AcmeMap();

		AcmeMaplet cur_maplet;
		while (maplet_enum.moreItems()) {
			cur_maplet = (AcmeMaplet) maplet_enum.getNextItem();
			try {
				destMap.addMaplet((AcmeMaplet) cur_maplet.copy());
			}
			catch (Exception e) {
				// This should never happen, since all maplets are fine here.
			}
		}
	}

	/**
	 * This method does a deep copy of "this" map instance
	 * @return the copy of the AcmeElement
	 */
	public AcmeElement copy() {
		AcmeMapInstance new_map = new AcmeMapInstance(this.getName());
		this.copyMaplets(new_map);
		this.copyCommonEltStructure(new_map);
		return new_map;
	}

	/**
	 * This method uses this element instance as a prototype to create a new type.
	 * @param newTypeName used to create a new AcmeMapType
	 * @return the new AcmeMapType, created by this function
	 * Notice the different between map and other Acme elements:
	 * A map type doesn't indicate the complete structures of its instances.
	 * That is because that a map type doesn't consist of maplets.
	 */
	public AcmeElementType copyAsPrototype(String newTypeName) {
		AcmeMapType new_map_type = new AcmeMapType(newTypeName);
		this.copyCommonEltStructure(new_map_type);
		// copyCommonEltStructure overwrites the name of the new type, so
		// reset it properly.  Don't check for legal names here, that will
		// happen if/when we try to add it to anything.
		new_map_type.name = newTypeName;

		// add all of the declared types of this instance as supertypes of
		// the new type.
		AcmeEnumeration e = this.getDeclaredTypes();
		while (e.moreItems()) {
			AcmeObject curType = e.getNextItem();
			if (curType instanceof AcmeMapType) {
				new_map_type.addSuperType((AcmeMapType) curType);
			}
			else if (curType instanceof AcmeElementProtoType) {
				new_map_type.addSuperType((AcmeElementProtoType) curType);
			}
			else if (curType instanceof AcmeElementTypeRef) {
				AcmeElementTypeRef t = (AcmeElementTypeRef) curType;
				if (t.getReferentType() instanceof AcmeMapType
					|| t.getReferentType() instanceof AcmeElementProtoType) {
					new_map_type.addSuperType((AcmeElementTypeRef) curType);
				}

			}
		}

		return new_map_type;
	}

	/**
	 * This method adds a maplet to the map
	 * @param maplet AcmeMaplet to be added
	 */
	public void addMaplet(AcmeMaplet maplet) throws acmeLib.core.exceptions.AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isAcmeKeyword(maplet.getName())) {
				//This should never happen for current Acme keywords.
				throw new AcmeKeywordNameException("Illegal keyword \"" + maplet.getName() + "\"used for maplet name.");
			}
		}

		if (maplet.getName() == null) {
			throw new acmeLib.core.exceptions.AcmeIllegalNameException(
				"Error. Attempted to add a maplet with null name");
		}

		if (eltNamespace.containsName(maplet.getName())) {
			throw new AcmeDuplicateNameException(
				"Attempted to redefine identifier " + maplet.getName() + "as an added maplet.");
		}

		eltNamespace.addName(maplet.getName());
		mapletMap.addItem(maplet.getName(), maplet);
		maplet.setParent(this);
		//		addDesignRules(maplet);
	}

	/**
	 * This method removes a mapelt from the map
	 * @param maplet AcmeMaplet to be removed
	 */
	public void removeMaplet(AcmeMaplet maplet) {
		removeMaplet(maplet.getName());
	}

	/**
	 * This method removes aa maplet from the map
	 * @param name the name of the maplet to be removed
	 */
	public void removeMaplet(String name) {
		AcmeMaplet maplet = (AcmeMaplet) mapletMap.lookupItem(name);
		if (maplet == null) {
			return;
		}
		else {
			mapletMap.removeItem(name);
			eltNamespace.removeName(name);
			maplet.setParent(null);
		}
	}

	/**
	 * This method return an enumeration of all maplets
	 * @return an AcmeEnumeration of the mapletMap
	 */
	public AcmeEnumeration getMaplets() {
		return mapletMap.enumerate();
	}

	/**
	 * This method returns a maplet with a specific name
	 * @param name the name of the maplet
	 * @return AcmeMaplet corresponding to name
	 */
	public AcmeMaplet getMaplet(String name) {
		return (AcmeMaplet) mapletMap.lookupItem(name);
	}

	/**
	 * All AcmeMapInstances satisfy the type ConstantTypes.MAP_BASE_TYPE.
	 * If the generic supertype method doesn't find the type declared, check
	 * whether the type is the MAP_BASE_TYPE.
	 * @param type the type of the AcmeElement
	 * @return whether the type is MAP_BASE_TYPE
	 */
	public boolean declaresType(AcmeElementType type) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType ();
		}
		if (type == ConstantTypes.MAP_BASE_TYPE) {
			return true;
		}
		else {
			return super.declaresType(type);
		}
	}

	//exporting...
	protected AcmeEnumeration getChildrenToExport() {
		AcmeEnumeration retEnum = super.getChildrenToExport();
		retEnum.concatenate(this.getMaplets());
		return retEnum;
	}

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Map ", this.getChildrenToExport(), ExportLanguage.ACME);
	}

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Map ", this.getChildrenToExport(), ExportLanguage.ARMANI);
	}

	/**
	 * Returns the type signature for this map as a string with the 
	 * form "Map <name> [ : <Typename>+ ]"
	 * @return string containing this elements's type signature
	 */
	public String getTypeSignature() {
		if (true)
			return "";
		StringBuffer buf = new StringBuffer("Map " + this.getName());
		AcmeEnumeration typeEnum = this.getDeclaredTypes();
		if (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			////			if (eltType instanceof AcmeUnresolvedElementType)
			////				buf.append(" : <unresolved type name: " + eltType.getTypeName() + ">");
			//			else
			buf.append(" : " + eltType.getTypeName());
		}

		while (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//			if (eltType instanceof AcmeUnresolvedElementType)
			//				buf.append(" , <unresolved type name: " + eltType.getTypeName() + ">");
			//			else
			buf.append(" , " + eltType.getTypeName());
		}

		return buf.toString();
	}

	/**
	 * This method implements the acmeLib.designRules.Scope interface to map names
	 * to elements or properties.
	 * It looks up the child object called 'name'.
	 * First checks for maplets, then has AcmeElement check for properties,
	 * and representations.
	 * @param name compared with maplets
	 * @param bubbleUp boolean
	 * @return the Object corresponding to name
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;
		if (name.equalsIgnoreCase("maps")) {
			retObj = this.getMaps();
			return retObj;
		}
		else if (name.equalsIgnoreCase("maplets")) {
			retObj = this.getMaplets();
			return retObj;
		}
		AcmeEnumeration e = this.getDeclaredTypes();
		while (e.moreItems()) {
			AcmeObject curType = e.getNextItem();
			if (curType instanceof AcmeMapType) {
				AcmeMapType curMapType = (AcmeMapType) curType;
				retObj = curMapType.lookupName(name, false);
				if (retObj != null) {
					return retObj;
				}
			}
		}
		if (bubbleUp) {
			e = this.getDeclaredTypes();
			while (e.moreItems()) {
				AcmeObject curType = e.getNextItem();
				if (curType instanceof AcmeMapType) {
					AcmeMapType curMapType = (AcmeMapType) curType;
					retObj = curMapType.lookupName(name, true);
					if (retObj != null) {
						return retObj;
					}
				}
			}
		}
		if (retObj == null) {
			// just check for a property or representation, bubbling up occurs in
			// the generic element lookupName method, though the current syntax 
			// doesn't allow a map to have such children.
			retObj = super.lookupName(name, bubbleUp);
		}

		return retObj;
	}

	/**
	 *  AcmeMapInstance.resolveTypeReferences() relies primarily on it's superclass'
	 *  implementation to do type name resolution.  This subclass adds a
	 *  recursive call to resolveTypeReferences() for each of this MapInstace's maplets.
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
		AcmeEnumeration e = this.getMaplets();
		while (e.moreItems()) {
			AcmeMaplet maplet = (AcmeMaplet) e.getNextItem();
			addDesignRules(maplet);
			addTypeMaps(maplet);
			maplet.resolveTypeReferences();
		}
	}

	// Add to the maplet the design rules declared in the types.
	protected void addDesignRules(AcmeMaplet maplet) {
		AcmeEnumeration types = getDeclaredTypes();
		while (types.moreItems()) {
			AcmeMapType type = (AcmeMapType) types.getNextItem();
			AcmeEnumeration invariants = type.getInvariants(true);
			while (invariants.moreItems()) {
				DesignRule invariant = new DesignRule((DesignRule) invariants.getNextItem());
				invariant.setDesignRuleScope(maplet);
				maplet.addInvariant(invariant);
			}
			AcmeEnumeration heuristics = type.getHeuristics(true);
			while (heuristics.moreItems()) {
				DesignRule heuristic = new DesignRule((DesignRule) heuristics.getNextItem());
				heuristic.setDesignRuleScope(maplet);
				maplet.addHeuristic(heuristic);
			}
		}
	}

	// Add to the maplet the maps declared in the types
	protected void addTypeMaps(AcmeMaplet maplet) {
		AcmeEnumeration types = getDeclaredTypes();
		while (types.moreItems()) {
			AcmeMapType type = (AcmeMapType) types.getNextItem();
			AcmeEnumeration maps = type.getMaps();
			while (maps.moreItems()) {
				AcmeMapInstance map = (AcmeMapInstance) ((AcmeMapInstance) maps.getNextItem()).copy();
				map.setParent(maplet);
				try {
					maplet.addMap(map);
				}
				catch (Exception e) {
				}
			}
		}
	}

	/**
	 * Recursively resolve all of the instance references of this
	 * system's substructure
	 */
	public void resolveInstanceReferences() {
		super.resolveInstanceReferences();
		AcmeEnumeration e = this.getMaplets();
		while (e.moreItems()) {
			((AcmeMaplet) e.getNextItem()).resolveInstanceReferences();
		}
	}

	/**
	 * instantiateElements() walks over the vector of instantiatedTypes of this
	 * element and adds the minimal structure declared in the type description
	 * to this element.  It delegates much of the work to the AcmeTypeManager
	 * class, so that we can change type instantiation policies relatively
	 * centrally (ie by changing the AcmeTypeManager methods) for all elements
	 * at once.  This method recursively calls the same method on all of the
	 * substructure of this element instance.
	 */
	public void instantiateElements() {

		// unify this element with each of the types it instantiates.
		for (int i = 0; i < this.instantiatedTypes.getSize(); i++) {
			// check that we have a prototype component or component type
			// to instantiate.  If we don't then announce an error and move
			// to the next type.
			if (this.instantiatedTypes.getItemAt(i) instanceof AcmeMapType) {
				AcmeTypeManager.unifyMaps((AcmeMapInstance) this.instantiatedTypes.getItemAt(i), this);

			}
			else if (this.instantiatedTypes.getItemAt(i) instanceof AcmeType) {
				String errorMsg = null;
				if (this.instantiatedTypes.getItemAt(i) instanceof AcmeUnresolvedElementType) {
					errorMsg =
						"Instantiation Error.  Unable to instantiate "
							+ " unresolved element type: "
							+ ((AcmeUnresolvedElementType) this.instantiatedTypes.getItemAt(i)).getTypeName()
							+ " in map: "
							+ this.getName();
				}
				else {
					errorMsg =
						"Instantiation Error.  Type: "
							+ ((AcmeType) this.instantiatedTypes.getItemAt(i)).getTypeName()
							+ " is not a map.  Therefore it "
							+ "can not be used to instantiate map: "
							+ this.getName();
				}

				TypeInstantiationError instError =
					new TypeInstantiationError(this, errorMsg, (AcmeType) this.instantiatedTypes.getItemAt(i));
				AcmeErrorManager.errorDetected(instError);

			}
			else {
				String errorMsg =
					"Instantiation Error.  Attempted to instantiate a "
						+ "completely unrecognizable type in map: "
						+ this.getName();
				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg);
				AcmeErrorManager.errorDetected(instError);
			}
		}

		// then have each of its maplets instantiate themselves.
		AcmeEnumeration e;
		e = this.getMaplets();
		while (e.moreItems()) {
			((AcmeMaplet) e.getNextItem()).instantiateElements();
		}

		// finally, call AcmeElement to have it instantiate representations and any
		// other standard element structure.
		super.instantiateElements();
	}

	/**
	 * This method confirms that this map satisfies all of its declared types and
	 * design rules, and that it's maplets all typecheck as well.
	 * @return whether the map satisfies all of its declared types
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckMap(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			AcmeErrorManager.errorRepaired(this, null);

			return true;
		}
		else {
			// map doesn't typecheck.  There should be an error at the
			// top of the AcmeTypeManager's typecheckingErrorStack.  If there
			// is such an error then announce it to the AcmeErrorManager and
			// return false.
			AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
			if (eltTypeError != null) {
				AcmeErrorManager.errorDetected(eltTypeError);
				return false;
			}
			else {
				// no error was reported.  something went wrong?
				// announce a generic error.
				AcmeErrorManager.errorDetected(
					new ElementTypecheckError(this, "Map " + this.getName() + " does not typecheck.", this));
				return false;
			}
		}
	}

	/**
	 * implements the acmeLib.Editable interface.
	 * Nothing available yet.
	 */
	public void edit() {
		System.out.println("Attempted to edit: " + this.getName());
		System.out.println("  but AcmeMapInstance.edit() is not yet implemented.");
	}

	public void view() {
		System.out.println("Attempted to view: " + this.getName());
		System.out.println("  but AcmeMapInstance.view() is not yet implemented.");
	}

}