
package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.*;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.*;
import acmeLib.typeManagement.AcmeTypeManager;


/**
 *  Class:   AcmeMapType
 <BR>
 *  Purpose:  Stores the type specification for a map.  This is
 <BR>
 *          a subclass of an AcmeMapInstance because the Acme type system
 <BR>
 *          is prototype-based.  As a result, a type is just a prototypical
 <BR>
 *          instance of an element that specifies the aspects that an
 <BR>
 *          instance of that type must posses.  Typechecking is done
 <BR>
 *          by an external class, AcmeTypeManager, that understands the
 <BR>
 *          rules of typechecking
 <BR><BR>
 *
 *  Revision History:
 <BR>
 *    Date:     Person    Change
 <BR>
 *   11/4/99     JH     Created class
 <BR><BR>
 *
 *  @author Jianing Hu<BR>
 */
public class AcmeMapType extends AcmeMapInstance
	implements AcmeElementType {

	//
	// protected members
	//
	protected AcmeVector superTypes;
	protected AcmeMap typeMap;
	protected AcmeMapSignature signature;

	//
	// public members
	/**
	 * This is the constructor for this class.
	 * It creates the prototype object in the superclass constructor
	 * @param typeName the name of the map type
	 */
	public AcmeMapType(String typeName) {
		// create the prototype object in the superclass constructor
		super(typeName);
		this.superTypes = new AcmeVector();
		this.typeMap = new AcmeMap();
	}

    public AcmeMapType(String typeName, AcmeMapSignature signature){
        super(typeName);
        this.superTypes = new AcmeVector();
        this.typeMap = new AcmeMap();
        this.signature = signature;
    }

	/**
	 * @return the name of the type
	 */
	public String getTypeName() { return this.getName(); }
	/**
	 * @param typeName used to set the name of the type
	 * @exception  AcmeIllegalNameException type name is illegal
	 */
	 
	public void setTypeName(String typeName) throws AcmeIllegalNameException
	{
		this.setName(typeName);
	}

    /**
     * @return the signature of the type
     */
    public AcmeMapSignature getSignature(){
        return this.signature;
    }
    
    /**
     * @param signature used to set the signature of the map type
     */
    public void setSignature(AcmeMapSignature signature){
        this.signature = signature;
    }

	/**
	 * This method confirms that this maptype's design rules are consistent with its scope
	 * @return whether the maptype type checks
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckMapType(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			return true;
		} else {
			// map doesn't typecheck.  There should be an error at the
			// top of the AcmeTypeManager's typecheckingErrorStack.  If there
			// is such an error then announce it to the AcmeErrorManager and
			// return false.
			AcmeError eltTypeError =
									AcmeTypeManager.typecheckingErrorStack.popCausalError();
			if (eltTypeError != null) {
				AcmeErrorManager.errorDetected(eltTypeError);
				return false;
			} else {
				// no error was reported.  something went wrong?
				// announce a generic error.
				AcmeErrorManager.errorDetected(
											   new ElementTypecheckError(this, "MapType "
																			   + this.getName() + " does not typecheck.", this));
				return false;
			}
		}
	}

	// methods for adding supertypes to a type description and retrieving
	// a type's set of supertypes
	/**
	 * @param superType AcmeElementType, which is to be added to superTypes
	 */
	public void addSuperType(AcmeElementType superType) {
		if (superType instanceof AcmeElementTypeRef) {
			superType = ((AcmeElementTypeRef) superType).getReferentType();
		}
		if ((superType instanceof AcmeMapType)
			|| (superType instanceof AcmeUnresolvedElementType)
			|| (superType instanceof AcmeElementProtoType))
		{
			superTypes.addItem((AcmeObject) superType);
		}
		// else throw an exception?
	}


	/**
	 * @param superTypes AcmeElementType, which is to be removed from superTypes
	 */
	public void removeSuperType(AcmeElementType superType) {
		superTypes.removeItem((AcmeObject) superType);
		if (superType instanceof AcmeElementTypeRef) {
			declaredTypes.removeItem((AcmeObject) ((AcmeElementTypeRef) superType).getReferentType());
		}
	}


	/**
	 * @return superTypes
	 */
	public AcmeVector getSuperTypes() { return this.superTypes; }


	/**
	 * This method makes a new map type object, with a deep copy of the old type
	 * @return AcmeElementType
	 */
	public AcmeElementType copyType() {
		AcmeMapType new_map_type =
										new AcmeMapType (this.getTypeName());
		this.copyMaplets(new_map_type);
		this.copyCommonEltStructure(new_map_type);
		this.superTypes.copyItems(new_map_type.superTypes);
		return new_map_type;
	}

  //
  // add, remove, and retrieve types declared in this map type.
  //
  /**
   * This method adds a type to this type
   * @param type AcmeType to be added to this map type. @param could be a maptype or a compound element type
   * @exception  AcmeIllegalNameException thrown when the type's name is illegal
   */
  public void addType(AcmeType type) throws AcmeIllegalNameException {
    if (type == null) {
        return;
    }

    if (LanguageConfig.CHECK_FOR_KEYWORDS)
        if (LanguageConfig.isAcmeKeyword(type.getTypeName()))
            throw new AcmeKeywordNameException("Illegal keyword \""
                + type.getTypeName() + "\" used for typename.");

    if (type.getTypeName() == null) {
        throw new AcmeIllegalNameException(
            "Error.  Attempted to add a type declaration with null typename.");
    }

    else if (eltNamespace.containsName(type.getTypeName()))
        throw new AcmeDuplicateNameException("Attempted to redefine "
            + "identifier " + type.getTypeName());

    eltNamespace.addName(type.getTypeName());
    this.typeMap.addItem(type.getTypeName(), (AcmeObject) type);
    if (type instanceof AcmeElement) {
        ((AcmeElement) type).setParent(this);
        ((AcmeElement) type).setParentDesign(this.getParentDesign());
    }
  }


  /**
   * This method retrieves all of the types defined in this type.
   * @return AcmeEnumeration of the types
   */
  public AcmeEnumeration getTypes() {
    return this.typeMap.enumerate();
  }


  /**
   * This method retrieves the type "name" that is declared in this type or one
   * of it's supertype's.
   * @param name the name of the type to be returned
   * @return AcmeType corresponding to "name"
   */
  public AcmeType getType(String name) {
    // first check for the type in this family.
    AcmeType retType = null;
    retType = (AcmeType) this.typeMap.lookupItem(name);
    if (retType != null)
        return retType;

    // else, if it's not found then look for it in each of this type's
    // supertypes.
    AcmeMapType curSuperType = null;
    for (int i = 0; i < this.superTypes.getSize(); i++) {
        if (this.superTypes.getItemAt(i) instanceof AcmeMapType) {
            curSuperType = (AcmeMapType) this.superTypes.getItemAt(i);
            retType = curSuperType.getType(name);
        }
        if (retType != null)
            return retType;
    }

    // if we get to here then the type called "name" was not found, return null.
    return retType;
  }


  /**
   * This method removes a type from a map type
   * @param name the name of the type to be removed
   */
  public void removeType(String name) {
    this.typeMap.removeItem(name);
    eltNamespace.removeName(name);
  }


  /**
   * removes the AcmeType <em>type</em> from this design
   * @param name the name of the AcmeType to be removed
   */
  public void removeType(AcmeType type) {
    if (type != null) {
        eltNamespace.removeName(type.getTypeName());
        this.typeMap.removeItem(type.getTypeName());
    }
  }

    protected void resolveSourceTypeReferences(){
		AcmeElementType curType = null;
		AcmeType resolvedType = null;
		AcmeUnresolvedElementType curUnresolvedType = null;
		curType = signature.getSource();
		if (curType instanceof AcmeUnresolvedElementType){
		    curUnresolvedType = (AcmeUnresolvedElementType) curType;
            java.util.StringTokenizer types = 
                new java.util.StringTokenizer(curUnresolvedType.getTypeName(),".");
		    AcmeElement curScope = this.getParent();
		    String curName = types.nextToken();
		    curScope = (AcmeElement)curScope.lookupName(curName,true);
	        if (curScope instanceof AcmeType){
	            resolvedType = (AcmeType)curScope;
	        }else{
	            resolvedType = null;
	        }
		    while (types.hasMoreElements() && curScope != null){
		        curName = types.nextToken();
		        curScope = (AcmeElement)curScope.lookupName(curName,false);
		        if (curScope instanceof AcmeType){
		            resolvedType = (AcmeType)curScope;
		        }else{
		            resolvedType = null;
		        }
			}
			if (resolvedType == null) {
				String errorMsg = "Unable to resolve source type \""
								  + curUnresolvedType.getTypeName()
								  + "\" of map type "
								  + this.getTypeName();
				TypeRefResolutionError badTypeRefError =
														new TypeRefResolutionError(this, errorMsg,
																				   curUnresolvedType.getTypeName());
				badTypeRefError.setUnresolvedRef(curUnresolvedType);
				AcmeErrorManager.errorDetected(badTypeRefError);
			} else if (!(resolvedType instanceof AcmeElementType)){
				String errorMsg = "Map type \"" + this.getTypeName()
							  + "\"'s source type \""
							  + curUnresolvedType.getTypeName()
							  + "\" is not an element type.";
   				TypeRefResolutionError badTypeRefError =
														new TypeRefResolutionError(this, errorMsg,
																				   curUnresolvedType.getTypeName());
				badTypeRefError.setUnresolvedRef(curUnresolvedType);
				AcmeErrorManager.errorDetected(badTypeRefError);
				resolvedType = null;
			}
		    if (resolvedType != null) {
			    signature.setSource((AcmeElementType) resolvedType);
			}
		}
    }

    protected void resolveDomainTypeReferences(){
		AcmeElementType curType = null;
		AcmeType resolvedType = null;
		AcmeUnresolvedElementType curUnresolvedType = null;
		curType = signature.getDomain();
		if (curType instanceof AcmeUnresolvedElementType){
		    curUnresolvedType = (AcmeUnresolvedElementType) curType;
            java.util.StringTokenizer types = 
                new java.util.StringTokenizer(curUnresolvedType.getTypeName(),".");
		    AcmeElement curScope = this.getParent();
		    if (curScope == null){
    		    System.out.println("curscope = NULL");
    		}
		    String curName = types.nextToken();
		    curScope = (AcmeElement)curScope.lookupName(curName,true);
	        if (curScope instanceof AcmeType){
	            resolvedType = (AcmeType)curScope;
	        }else{
	            resolvedType = null;
	        }
		    while (types.hasMoreElements() && curScope != null){
		        curName = types.nextToken();
		        curScope = (AcmeElement)curScope.lookupName(curName,false);
		        if (curScope instanceof AcmeType){
		            resolvedType = (AcmeType)curScope;
		        }else{
		            resolvedType = null;
		        }
			}
			if (resolvedType == null) {
				String errorMsg = "Unable to resolve domain type \""
								  + curUnresolvedType.getTypeName()
								  + "\" of map type "
								  + this.getTypeName();
				TypeRefResolutionError badTypeRefError =
														new TypeRefResolutionError(this, errorMsg,
																				   curUnresolvedType.getTypeName());
				badTypeRefError.setUnresolvedRef(curUnresolvedType);
				AcmeErrorManager.errorDetected(badTypeRefError);
			} else if (!(resolvedType instanceof AcmeElementType)){
				String errorMsg = "Map type \"" + this.getTypeName()
							  + "\"'s domain type \""
							  + curUnresolvedType.getTypeName()
							  + "\" is not an element type.";
   				TypeRefResolutionError badTypeRefError =
														new TypeRefResolutionError(this, errorMsg,
																				   curUnresolvedType.getTypeName());
				badTypeRefError.setUnresolvedRef(curUnresolvedType);
				AcmeErrorManager.errorDetected(badTypeRefError);
				resolvedType = null;
			}
		    if (resolvedType != null) {
			    signature.setDomain((AcmeElementType) resolvedType);
			}
		}
    }

	/** ResolveTypeReferences searches through this type object's vector of supertypes
	*  for UnresolvedElementType objects, looks up the actual type that is
	*  referred to by each UnresolvedElementType object, and replaces that
	*  object with the actual type object that should be referred to.
	*  it then calls resolveTypeReferences in its superclass to resolve the type names
	*  of all of its substructure (ports and representations).
	*/
	// Copied from AcmeComponentType. Not sure if it works for AcmeMapType
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
		AcmeElementType curType = null;
		AcmeType resolvedType = null;
		AcmeUnresolvedElementType curUnresolvedType;
		resolveSourceTypeReferences();
		resolveDomainTypeReferences();
		for (int i = 0; i < this.superTypes.getSize(); i++) {
			curType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (curType instanceof AcmeUnresolvedElementType) {
				curUnresolvedType = (AcmeUnresolvedElementType) curType;

				// first check to see if the supertype is declared in the same
				// family that this type is declared in.  If an AcmeComponentType
				// object's parent field is not null then parent must be a family.
				if (this.getParent() instanceof AcmeFamily) {
					resolvedType = ((AcmeFamily) this.getParent()).getType(
																		   curUnresolvedType.getTypeName());
				}

				if (this.getParent() instanceof AcmeDesign) {
					resolvedType = ((AcmeDesign) this.getParent()).getType(
																		   curUnresolvedType.getTypeName());
				}
				else if ((resolvedType == null) && (this.getParentDesign() != null)) {
					// then check to see if the supertype is declared in the same
					// design as this type.
					resolvedType = this.getParentDesign().getType(
																  curUnresolvedType.getTypeName());
				}

				if (resolvedType == null) {
					String errorMsg = "Unable to resolve supertype \""
									  + curUnresolvedType.getTypeName()
									  + "\" of map type "
									  + this.getTypeName();
					TypeRefResolutionError badTypeRefError =
															new TypeRefResolutionError(this, errorMsg,
																					   curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

				} else if (!(resolvedType instanceof AcmeMapType))
				{
					String errorMsg = "Map type \"" + this.getTypeName()
									  + "\"'s supertype \""
									  + curUnresolvedType.getTypeName()
									  + "\" is not a map type.";
					TypeRefResolutionError badTypeRefError =
															new TypeRefResolutionError(this, errorMsg,
																					   curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

					resolvedType = null;
				}

				 if (resolvedType != null) {
					 this.superTypes.setItemAt((AcmeObject) resolvedType, i);
				 }

			}
		}
		AcmeEnumeration eTypes = getTypes();
		while (eTypes.moreItems()){
		    ((AcmeMapType)eTypes.getNextItem()).resolveTypeReferences();
		}
	}


	/**
	*  lookupName implements the acmeLib.designRules.Scope interface.
	*  @param name compared with "self"
	*  @param bubbleUp used in the super.lookupName method
	*  @return an Object corresponding to name
	*/
	// Copied from AcmeComponent. Not sure if it works
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;

		// check first whether this is referring to "this" type.
		if (name == null) {
			return null;
		} else if (name.equalsIgnoreCase("self")) {
			return this;
		}
        
        retObj = typeMap.lookupItem(name);
        if (retObj != null){
            return retObj;
        }

		// recursively walk up the tree of supertypes hunting for "name"
		for (int i = 0; (i < this.superTypes.getSize()) && (retObj == null); i++) {
			AcmeElementType curSuperType =
										  (AcmeElementType) this.superTypes.getItemAt(i);
			if (!(curSuperType instanceof AcmeUnresolvedElementType))
				retObj = curSuperType.lookupName(name, false);
		}

		// if we didn't find it in a supertype then check for it in this type's
		// prototype object.
		if (retObj == null)
			retObj = super.lookupName(name, bubbleUp);

		return retObj;
	}


	//
	// Exporting methods
	//

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s)
		throws java.io.IOException, AcmeException
	{
		this.exportEltType(s, "Map Type", this.getChildrenToExport(),
						   ExportLanguage.ACME);
	}


	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportArmani(AcmeOutputStream s)
		throws java.io.IOException, AcmeException
	{
		this.exportEltType(s, "Map Type", this.getChildrenToExport(),
						   ExportLanguage.ARMANI);
	}


	/**
	 * Returns the type signature for this map type as a string of the
	 *  form "Map Type <name> : source<->domain[ extends <Typename>+ ]".
	 * @return string containing this map type's signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Map Type " + this.getTypeName());
		buf.append( " : " + signature.toString());
		AcmeEnumeration superTypesEnum =
			this.getSuperTypes().enumerate(
			   new acmeLib.core.filters.NoBaseClassesEnumFilter());
		if (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
//			if (eltType instanceof AcmeUnresolvedElementType)
//				buf.append(" extends <unresolved type name: " + eltType.getTypeName() + "> ");
//			else
				buf.append(" extends " + eltType.getTypeName());
		}

		while (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	}
}


