
package acmeLib.elements;

import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.ElementTypecheckError;
import acmeLib.typeManagement.AcmeTypeManager;


/**
 *  Class:  AcmeMaplet
 <BR>
 *  Purpose:  Stores the structure of an ACME Maplet
 <BR><BR>
 *
 *  Revision History:
 <BR>
 *    Date:     Person    Change
 <BR>
 11/5/99		JH		  Create class
 <BR><BR>
 *
 *  @author Jianing Hu<BR>
 */
public class AcmeMaplet extends AcmeElement {
	AcmeElement source, target;
	
	public AcmeMaplet(){System.err.println("this constructor of AcmeMaplet should never be called");}

	public AcmeMaplet(AcmeElement source, AcmeElement target){
		this.source = source;
		if (source instanceof AcmeUnresolvedElement){
		    source.setParent(this);
		}
		this.target = target;
		if (target instanceof AcmeUnresolvedElement){
		    target.setParent(this);
		}
		String name = source.getName() + "-" + target.getName();
		try{
    		this.setName(name);
    	}catch (AcmeIllegalNameException e){}
	}
	
	public String getName(){
	    return source.getName() + "-" + target.getName();
	}

	public AcmeElement getSource(){
		return this.source;
	}
	
	public AcmeElement getTarget(){
		return this.target;
	}
	
	public String toString(){
	    return source.getName()+" to "+target.getName();
	}
	
	//
	// exporting
	//
	/**
	 * This method exprts a maplet
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s)
		throws java.io.IOException, AcmeException
	{
		this.exportEltInstance(s, "Maplet", this.getChildrenToExport(),
							   ExportLanguage.ACME);
	}

	/**
	 * This method exports a maplet
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportArmani(AcmeOutputStream s)
		throws java.io.IOException, AcmeException
	{
		this.exportEltInstance(s, "Maplet", this.getChildrenToExport(),
							   ExportLanguage.ARMANI);
	}

	/**
	 * This method makes a new maplet object, with a deep copy of the old maplet.
	 * In the present code, the new maplet does not inherit the parent of the
	 * original port.
	 * @return a copy of the AcmeMaplet
	 */
	public AcmeElement copy() {
		AcmeMaplet new_maplet = new AcmeMaplet (this.getSource(),this.getTarget());
		this.copyCommonEltStructure(new_maplet);
		return new_maplet;
	}

    public void resolveInstanceReferences(){
        // very bad hacking
        // I am going to fix this.
        // JH
        if (source instanceof AcmeUnresolvedElement){
            AcmeElement tempSource;
            System.out.println(source.getName());
            if ((source.getName().length() >= 6) && (!source.getName().substring(0,6).equalsIgnoreCase("source"))){
                tempSource = source.getResolvedInstanceReference();
                if (tempSource != null){
                    source = tempSource;
                }
            } else {
                source.setParent(this.getParent().getParent());
                tempSource = source.getResolvedInstanceReference();
                if (tempSource != null){
                    source = tempSource;
                }
                source.setParent(this);
            }
        } else {
            source.resolveInstanceReferences();
        }
        if (target instanceof AcmeUnresolvedElement){
            AcmeElement tempTarget;
            if ((target.getName().length() >= 6) && (!target.getName().substring(0,6).equalsIgnoreCase("target"))){
                tempTarget = target.getResolvedInstanceReference();
                if (tempTarget != null){
                    target = tempTarget;
                }
            } else {
                target.setParent(this.getParent().getParent());
                tempTarget = target.getResolvedInstanceReference();
                if (tempTarget != null){
                    target = tempTarget;
                }
                target.setParent(this);
            }
        } else {
            target.resolveInstanceReferences();
        }
        super.resolveInstanceReferences();
    }
    
    /**
	 *  AcmeMaplet.resolveTypeReferences() simply passes the request to its supertype
	 *  AcmeElement.  This method is explicitly implemented because so that
	 *  when it's subclasses call super.resolveTypeReferences() it will be propogated
	 *  up the inheritance tree properly.
	 */
	public void resolveTypeReferences() { super.resolveTypeReferences(); }

	public Object lookupName(String name, boolean bubbleUp) {
	    if (name.equalsIgnoreCase("source")){
	        return getSource();
	    }
	    if (name.equalsIgnoreCase("target")){
	        return getTarget();
	    }
	    return super.lookupName(name,bubbleUp);
	}

	/**
	 * This method determines whether this maplet satisfies the type of the map
	 * instance in which it is declared
	 * @return whether this port typeckecks
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckMaplet(this);
		if (typechecks) {
			// things are hunky-dory, return true.
            AcmeErrorManager.errorRepaired (this, null);

			return true;
		} else {
			// maplet doesn't typecheck.  There should be an error at the
			// top of the AcmeTypeManager's typecheckingErrorStack.  If there
			// is such an error then announce it to the AcmeErrorManager and
			// return false.
			AcmeError eltTypeError =
									AcmeTypeManager.typecheckingErrorStack.popCausalError();
			if (eltTypeError != null) {
				AcmeErrorManager.errorDetected(eltTypeError);
				return false;
			} else {
				// no error was reported.  something went wrong?
				// announce a generic error.
				AcmeErrorManager.errorDetected(
											   new ElementTypecheckError(this, "Maplet "
																			   + this.getName() + " does not typecheck.", this));
				return false;
			}
		}
	}


}


