/*
 * Created on Jun 2, 2003
 */
package acmeLib.elements;

import java.io.File;

import acmeLib.AcmeObject;
import acmeLib.tools.AcmeDesignChangeNotifier;

/**
 * Purpose: Stores the record of a design that has been opened.
 *
 * TODO:
 *     Generated comment.
 */
public class AcmeOpenedDesignRecord extends AcmeObject {
    /** The fully qualified and resolved filename for the imported design's file. */
    private String canonicalFilename = null;

    /** The last modified time in milliseconds in since Jan 01, 1970, 00:00:00 GMT. */
    private long lastModified = 0;

    /** Stores a reference to the resolved design object. */
    private AcmeDesign resolvedDesign = null;
    
    /** Indicates whether the ancestor design of this opened design has changed. */
    private boolean ancestorChanged = false;

    public void copy (AcmeObject newObject) {
        AcmeOpenedDesignRecord newRec = (AcmeOpenedDesignRecord) newObject;
        newRec.canonicalFilename = new String(this.canonicalFilename);
        newRec.lastModified = this.lastModified;
        newRec.resolvedDesign = this.resolvedDesign; //shallow ok?
        newRec.ancestorChanged = this.ancestorChanged;
    }

	public void setCanonicalFilename(String canonicalFilename) {
		this.canonicalFilename = canonicalFilename;
	}

	public String getCanonicalFilename() {
		return canonicalFilename;
	}

	public void setLastModified(long lastModified) {
		this.lastModified = lastModified;
	}

	public long getLastModified() {
		return lastModified;
	}

    /**
     * Sets the new resolved AcmeDesign, and also unsets the ancestor changed
     * state, if the supplied resolved design is non-null.
     * @param resolvedDesign
     */
	public void setResolvedDesign(AcmeDesign resolvedDesign) {
        if (resolvedDesign != null && this.resolvedDesign != resolvedDesign) {
            setAncestorChanged(false);
        }
        AcmeDesign oldDesign = this.resolvedDesign;
		this.resolvedDesign = resolvedDesign;
        AcmeDesignChangeNotifier.getInstance().onDesignChange(
            getCanonicalFilename(), oldDesign, resolvedDesign);
	}

	public AcmeDesign getResolvedDesign() {
		return resolvedDesign;
	}

    /**
     * @return boolean <code>True</code> if an ancestor design has changed,
     *                  <code>false</code> otherwise
     */
    public boolean isAncestorChanged() {
        return ancestorChanged;
    }

    /**
     * @param changed  whether ancestor has changed
     */
    public void setAncestorChanged(boolean changed) {
        ancestorChanged = changed;
    }

    /**
     * Returns whether the opened design should be reloaded.  This is determined
     * by either of two factors:<br>
     * <ul>
     * <li>If <code>isAncestorChanged()</code> returns <code>true</code>; or
     * <li>If the modified time of the file given by <code>getCanonicalFilename()</code>
     *     is newer than <code>getLastModified()</code>.
     * </ul>
     * @return boolean <code>True</code> if the design should be reloaded,
     *                  <code>false</code> otherwise.
     */
    public boolean shouldReload () {
        File designFile = new File(getCanonicalFilename());
        long modTime = designFile.lastModified();
        return isAncestorChanged() || (modTime > getLastModified());
    }

    /**
     * helper method that massages an imported filename into a format
     * that Win* is likely to be able to conver to a file.  First it
     * removes any leading and trailing quotes.  It then converts all
     * double backslashes in the string into single backslashes.  This
     * @param s the string to be manipulated containing the filename
     * @return the changed string
     */
    public static String massageFileImportString(String s) {

        // first trim the leading and trailing quotes, if they exist
        int start_index = 0;
        int end_index = s.length();

        if ( s.startsWith("\"") && s.endsWith("\"") && !s.endsWith("\\\"") ) {
            start_index = 1;
            end_index = s.length() - 1;
        };

        String prepped_str = new String (s.substring(start_index, end_index));

        // now that we have prepared the string, convert all backslashed
        // quotes to non-backslashed quotes.

        StringBuffer result_buf = new StringBuffer();
        int str_end = prepped_str.length();
        int cur_index = 0;
        int next_slash_quote_index = 0;
        boolean end_reached = false;

        while (!end_reached) {
            // find the next occurence of the substring \"
            next_slash_quote_index = prepped_str.indexOf("\\\\", cur_index);

            if (next_slash_quote_index == -1) {
                // there were no more double backslashes found
                result_buf.append(prepped_str.substring(cur_index, str_end));
                end_reached = true;
            } else {
                // append everything up to the double backslash to the results
                // buffer, append a single backslash, and move the cur_index
                // to the following double backslash.
                result_buf.append(prepped_str.substring(cur_index,
                                            next_slash_quote_index));
                cur_index = next_slash_quote_index + 1;
            };

            if (cur_index > str_end)
                end_reached = true;
        }

        return result_buf.toString();
    };


    /**
     * massageForExport checks if string s has any whitespace, if it does not
     * then it simply returns the entire string.  If, otoh, it does have
     * whitespace it wraps the filename in quotes and double-backslashes all
     * of the backslashes found in the filename.
     *
     * NOTE: this is a function.  it does not adjust the internal representation
     *       of any AcmeImportedDesignRecord fields It simply massages the
     *       string passed in as an argument
     * @param s the string to be manipulated containing the filename
     * @return the changed string
    */
    public static String massageForExport(String s) {

        if (((s.indexOf(' ') == -1) && (s.indexOf('\t') == -1)
              && (s.indexOf('\n') == -1)) || s.startsWith("\""))
        {
            return s;   // no whitespace so no need to massage it
        };

        // otherwise wrap the string in quotes and double-backslash all of
        // the backslashes in it.
        StringBuffer result_buf = new StringBuffer();
        // add the initial quote to the result string
        result_buf.append("\"");

        // search the string s for double backslashes, when found replace with
        // a single backslash
        int cur_index = 0;
        int next_backslash_index = 0;
        int str_end = s.length();
        boolean end_reached = false;

        while (!end_reached) {
            next_backslash_index = s.indexOf('\\', cur_index);

            if (next_backslash_index == -1) {
                // there were no more backslashes found
                result_buf.append(s.substring(cur_index,str_end));
                end_reached = true;
            } else if (next_backslash_index == 0) {
                result_buf.append("\\\\");
                cur_index = next_backslash_index + 1;
            } else {
                result_buf.append(s.substring(cur_index, next_backslash_index));
                result_buf.append("\\\\");
                cur_index = next_backslash_index + 1;
            };

            if (cur_index > str_end)
                end_reached = true;
        }

        // add the final quote to the result string
        result_buf.append("\"");

        return result_buf.toString();
    };

}
