package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 *  Class:  AcmePort
<BR>
 *  Purpose:  Stores the structure of an ACME Port
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    4/3/97    RTM     Translated to Java
<BR>
 *    5/30/97   RTM     All stubbed methods filled in
<BR>
 *    6/16/97   RTM     Revised to work with 2nd generation type system
<BR>
 *    1/14/98   RTM     added support for multiple declared types.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmePort extends AcmeElement {

	//
	// Public Members
	//
	/**
	 * This is a constructor for this class
	 * @param name the name of the port to be created
	 */
	public AcmePort(String name) {
		this.name = name;
	};

	/**
	 * This is a constructor for this class
	 * @param name the name of the port to be created
	 * @param type the type of the port to be created
	 */
	public AcmePort(String name, AcmePortType t) {
		this.name = name;
		addDeclaredType(t);
	};

	//
	// exporting
	//
	/**
	 * This method exprts a port
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Port", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	  * This method exprts a port
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Port", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this port as a string with the
	 *  form "Port <name> [ : <Typename>+ ]".  For example:
	 *  "Port foo : portT".
	 * @return string containing this port's type signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Port " + this.getName());
		AcmeEnumeration typeEnum = this.getDeclaredTypes();
		if (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" : <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" : " + eltType.getTypeName());
		}

		while (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * This method makes a new port object, with a deep copy of the old port's
	 * properties, representations, declared and instantiated type vectors.
	 * In the present code, the new port does not inherit the parent of the
	 * original port.
	 * @return a copy of the AcmePort
	 */
	public AcmeElement copy() {
		AcmePort new_port = new AcmePort(this.getName());
		this.copyCommonEltStructure(new_port);
		return new_port;
	};

	/**
	 * This method uses this element instance as a prototype to create a new type.
	 * @param newTypeName the name of hte new type
	 * @return AcmeElementType corresponding to newTypeName
	 */
	public AcmeElementType copyAsPrototype(String newTypeName) {
		AcmeVector super_types;
		AcmePortType new_port_type = new AcmePortType(newTypeName);
		this.copyCommonEltStructure(new_port_type);

		// copyCommonEltStructure overwrites the name of the new type, so
		// reset it properly.  Don't check for legal names here, that will
		// happen if/when we try to add it to anything.
		new_port_type.name = newTypeName;

		// add all of the declared types of this instance as supertypes of
		// the new type.
		AcmeEnumeration e = this.getDeclaredTypes();

		while (e.moreItems()) {
			AcmeObject curType = e.getNextItem();
			if (curType instanceof AcmePortType) {
				new_port_type.addSuperType((AcmePortType) curType);
			}
			else if (curType instanceof AcmeElementProtoType) {
				new_port_type.addSuperType((AcmeElementProtoType) curType);
			}
			else if (curType instanceof AcmeElementTypeRef) {
				AcmeElementTypeRef t = (AcmeElementTypeRef) curType;
				if (t.getReferentType() instanceof AcmePortType
					|| t.getReferentType() instanceof AcmeElementProtoType) {
					new_port_type.addSuperType((AcmeElementTypeRef) curType);
				}

			}
		};

		return new_port_type;
	};

	/**
	 *  All AcmePorts satisfy the type ConstantTypes.PORT_BASE_TYPE.
	 *  if the generic supertype method doesn't find the type declared, check
	 *  whether the type is the PORT_BASE_TYPE.
	 *  @param type the AcmeElementType compared to PORT_BASE_TYPE
	 *  @return whether the type is the PORT_BASE_TYPE
	 */
	public boolean declaresType(AcmeElementType type) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType ();
		}
		if (type == ConstantTypes.PORT_BASE_TYPE) {
			return true;
		}
		return super.declaresType(type);
	};

	/**
	 *  AcmePort.resolveTypeReferences() simply passes the request to its supertype
	 *  AcmeElement.  This method is explicitly implemented because so that
	 *  when it's subclasses call super.resolveTypeReferences() it will be propogated
	 *  up the inheritance tree properly.
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
	}

	/**
	 * instantiateElements() walks over the vector of instantiatedTypes of this
	 * element and adds the minimal structure declared in the type description
	 * to this element.  It delegates much of the work to the AcmeTypeManager
	 * class, so that we can change type instantiation policies relatively
	 * centrally (ie by changing the AcmeTypeManager methods) for all elements
	 * at once.  This method recursively calls the same method on all of the
	 * substructure of this element instance.
	 */
	public void instantiateElements() {

		// unify this element with each of the types it instantiates.
		for (int i = 0; i < this.instantiatedTypes.getSize(); i++) {
			// check that we have a prototype port or port type
			// to instantiate.  If we don't then announce an error and move
			// to the next type.
			AcmeObject iType = this.instantiatedTypes.getItemAt(i);
			if (iType instanceof AcmeElementTypeRef) {
				iType = (AcmeObject) ((AcmeElementTypeRef) iType).getReferentType();
			}
			if (iType instanceof AcmePortType) {
				AcmeTypeManager.unifyPort((AcmePortType) iType, this);

			}
			else if (iType instanceof AcmeElementProtoType) {
				AcmeTypeManager.unifyElementWithPrototype((AcmeElementProtoType) iType, this);

			}
			else if (iType instanceof AcmeType) {
				String errorMsg = null;
				if (iType instanceof AcmeUnresolvedElementType) {
					errorMsg =
						"Instantiation Error.  Unable to instantiate "
							+ " unresolved element type: "
							+ ((AcmeUnresolvedElementType) iType).getTypeName()
							+ " in port: "
							+ this.getName();
				}
				else {
					errorMsg =
						"Instantiation Error.  Type: "
							+ ((AcmeType) iType).getTypeName()
							+ " is not a port type.  Therefore it can not be "
							+ "used to instantiate port: "
							+ this.getName();
				};

				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg, (AcmeType) iType);
				AcmeErrorManager.errorDetected(instError);

			}
			else {
				String errorMsg =
					"Instantiation Error.  Attempted to instantiate a "
						+ "completely unrecognizable type in port: "
						+ this.getName();
				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg);
				AcmeErrorManager.errorDetected(instError);
			};
		}

		// finally, call AcmeElement to have it instantiate representations and any
		// other standard element structure.
		super.instantiateElements();

	};

	/**
	 * This method determines whether this port satisfies all of its declared types and
	 * design rules, and that it's properties, and representations all
	 * typecheck also.
	 * @return whether this port typeckecks
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckPort(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			AcmeErrorManager.errorRepaired(this, null);

			return true;
		}
		// port doesn't typecheck.  There should be an error at the
		// top of the AcmeTypeManager's typecheckingErrorStack.  If there
		// is such an error then announce it to the AcmeErrorManager and
		// return false.
		AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
		if (eltTypeError != null) {
			AcmeErrorManager.errorDetected(eltTypeError);
			return false;
		}
		// no error was reported.  something went wrong?
		// announce a generic error.
		AcmeErrorManager.errorDetected(
			new ElementTypecheckError(this, "Port " + this.getName() + " does not typecheck.", this));
		return false;
	};

	/**
	 * Creates a workshop for editing this element.
	 * implements the acmeLib.Editable interface.
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editPort(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

	/**
	 * This method is used to implement the acmeLib.Editable interface
	 */
	public void view() {
		System.out.println("Attempted to view: " + this.getName());
		System.out.println("  but AcmePort.view() is not yet implemented.");
	};

	/**
	 * This method implements looking up names. It looks up the child object
	 * called 'name'. It checks for attachedRoles.
	 * @param name compared with 'attachedRoles'
	 * @param bubbleUp whether to look in surrounding scopes
	 * @return the Object corresponding to name
	 * @see acmeLib.designRules.Scope#lookupName(String, boolean)
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;

		if (name != null) {
			if (name.equalsIgnoreCase("attachedroles")) {
				// BRS: Need to find the system, traverse through the attachments
				// and return ports that are attached to these roles

				// Find the parent system
				AcmeElement parent = this.getParent();
				if (parent instanceof AcmeComponent) {
					parent = parent.getParent();
				}

				if (parent instanceof AcmeSystem) {
					AcmeSystem sys = (AcmeSystem) parent;

					// Iterate through the attachments, finding ones in which
					// the port is involved.
					AcmeEnumeration attachments = sys.getAttachments();
					AcmeVector roles = new AcmeVector();
					while (attachments.moreItems()) {
						AcmeAttachment att = (AcmeAttachment) attachments.getNextItem();
						if (!(att instanceof AcmeUnresolvedAttachment) && att.attachedPort.equals(this)) {
							roles.addItem(att.attachedRole);
						}
					}
					return roles.enumerate();
				}

			}
		}
		return super.lookupName(name, bubbleUp);
	}

}; // end AcmePort class
