package acmeLib.elements;

import java.io.IOException;

import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;


/**
 *  Class: AcmeRecordFieldValue
 *
 *  Purpose: a description for a record field
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
 *	 11/5/99	JH			Created
 * 
 * @author Jianing Hu
 */
public class AcmeRecordFieldValue extends acmeLib.AcmeObject {

    //
    // protected members
    //
    protected String name;
	protected AcmeElementType type;
	protected AcmeElement value;
	protected AcmeRecordValue parent;

    //
    // public members
    //
  /**
   * This is a constructor for this class
   * @param name is the value to be assigned to the record field's name
   * @param t the value to be assigned to the record field's type
   * @param val the value to be assigned to the record field's value
   */
   public AcmeRecordFieldValue (String name, AcmeElementType t, AcmeElement val) {
        this.name = name;
        this.setType(t);
        value = val;
        parent = null;
    }

  /**
   * This method sets the name of this record field and alerts its parent
   * record that the name has been changed.
   * @param name value to be assigned to the name of the record field
   * @exception acmeLib.core.exceptions.AcmeIllegalNameException if the
   * parameter "name" is illegal
   **/
	public void setName (String name) throws AcmeIllegalNameException
	{
	    String oldName = this.name;
	    this.name = name;
	    if (this.getParent() != null) {
	        this.getParent().childNameChanged(oldName, this);
	    }
    }


  /**
   * This method returns the name of this record field
   * @return the name of this record field
   */
	public String getName () { return this.name; };

 /**
   * This method sets the type of this record field
   * @param t the value to be assigned to type
   */
	public void setType (AcmeElementType t) { type = t; };
 /**
   * This method returns the type of this record field
   * @return the type of this record field
   */
	public AcmeElementType getType () { return this.type; };

  /**
   * This method returns the value of this record field
   * @return the value of this record field
   */
	public AcmeElement getValue () { return value; };
  /**
   * This method sets the value of this record field
   * @param val the value of this record field
   */
	public void setValue(AcmeElement val) { this.value = val; };


  /**
   * This method returns the record that contains this record field
   * @return the parent record
   */
	public AcmeRecordValue getParent() { return parent; };

  /**
   * This method sets the record that contains this record field
   * @param val the parent of this record field
   */
	public void setParent(AcmeRecordValue newParent) { this.parent = newParent; };


  /**
   * This method returns a deep-copy of "this" RecordField
   * @return a copy of this record field
   */

    public AcmeRecordFieldValue copy() {
        AcmeElementType type;
        AcmeElement val;

        if (this.getType() == null)
            type = null;
        else
            type = this.getType();

        if (this.getValue() == null)
            val = null;
        else
            val = this.getValue().copy();

        AcmeRecordFieldValue newField =
                new AcmeRecordFieldValue(this.getName(), type, val);

        return newField;
    }


  /**
   * This method converts AcmeRecordFieldValue to String
   * @return the String value of this record field
   */
	public String toString() {
        StringBuffer buf = new StringBuffer();
	    buf.append(this.getName());
        if (this.getType() != null) {
	        if (this.getType() != ConstantTypes.UNSPECIFIED_TYPE) {
        	    buf.append(" : " + this.getType().getTypeName());
    	    }
    	    // else the type is undeclared so don't declare one...
	    }

        if (getValue() != null) {
		    buf.append(" = " + this.getValue().toString());
	    }

	    return buf.toString();
    }

    public void export (AcmeOutputStream s) throws IOException {
        s.beginRegion (this);
        s.write (toString ());
        s.endRegion (this);
    }

    /**
     * This method resolves the property type object that this record field type
     * stores (e.g. resolves "foo" and bar of a [ x : foo; y : bar]
     * record type definition).  Recursively calls resolveTypeReferences on the
     * field value object also.
     * @param scope acmeLib.designRules.Scope
     */
/*    public void resolveTypeReferences(acmeLib.designRules.Scope scope) {
        String lookupError = null;
        if (this.getType() instanceof AcmeUnresolvedElementType) {
            Object foundObject = null;
            if (scope == null) {
                // if scope is null then we can't resolve names,
                // announce an error and return.
                lookupError = "Unable to resolve record field typename: "
                                   + this.getType().getTypeName();
                TypeRefResolutionError badScopeError =
                    new TypeRefResolutionError(this, lookupError,
                                               this.getType().getTypeName());
                badScopeError.setUnresolvedRef(this);
                AcmeErrorManager.errorDetected(badScopeError);
                return;
            } else {
                foundObject =
                    scope.lookupName(this.getType().getTypeName(), true);
           }

            if (foundObject instanceof AcmeElementType) {
                this.setType((AcmeElementType) foundObject);
            } else {
                // Announce an error and return
                lookupError = "Unable to resolve record field typename: "
                                   + this.getType().getTypeName();
                TypeRefResolutionError notPropertyTypeError =
                    new TypeRefResolutionError(this, lookupError,
                                               this.getType().getTypeName());
                notPropertyTypeError.setUnresolvedRef(this);
                AcmeErrorManager.errorDetected(notPropertyTypeError);
//                this.setType(ConstantTypes.UNRESOLVABLE_PROPERTY_TYPE);
                return;
            }
        }*

        // now that we have a resolved  type, resolve the typenames
        // on the type object, to resolve any unresolved names
        // in complex property types.
        if (this.getType() != null) {
            this.getType().resolveTypeReferences(scope);
        } else {
            // Announce an error and return
            lookupError = "Unable to resolve record field typename: "
                               + this.getType().getTypeName();
            TypeRefResolutionError notPropertyTypeError =
                new TypeRefResolutionError(this, lookupError,
                                           this.getType().getTypeName());
            notPropertyTypeError.setUnresolvedRef(this);
            AcmeErrorManager.errorDetected(notPropertyTypeError);
            this.setType(ConstantTypes.UNRESOLVABLE_PROPERTY_TYPE);
            return;
        }

        // recursively resolve all types that might be referenced in the
        // property value.
        if (this.getValue() != null) {
            this.getValue().resolveTypeReferences(scope);
        }
    }*/
}
