package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 *  Class:  AcmeRole
<BR>
 *  Purpose: Stores the structure of an ACME Role
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    4/3/97    RTM     Translated to Java
<BR>
 *    5/30/97   RTM     All stubbed methods filled in
<BR>
 *    6/16/97   RTM     Revised to work with 2nd generation type system
<BR>
 *    1/14/98   RTM     added support for multiple declared types.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeRole extends AcmeElement {

	//
	// Public members
	//

	/**
	 * This is a constructor for this class
	 * @param name the name of the role to be created
	 */
	public AcmeRole(String name) {
		this.name = name;
	};

	/**
	  * This is a constructor for this class
	  * @param name the name of the role to be created
	  * @param t the type of the role to be created
	  */
	public AcmeRole(String name, AcmeRoleType t) {
		this.name = name;
		this.addDeclaredType(t);
	};

	//
	// exporting
	//

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Role", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs when writing
	  * @exception AcmeException for acme exception
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "Role", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this role as a string with the
	 *  form "Role <name> [ : <Typename>+ ]".  For example:
	 *  "Role foo : roleT".
	 * @return string containing this role's type signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Role " + this.getName());
		AcmeEnumeration typeEnum = this.getDeclaredTypes();
		if (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" : <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" : " + eltType.getTypeName());
		}

		while (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * This method makes a new role object, with a deep copy of the old role's
	 * properties and representations.  In the present code, the
	 * new role does not inherit the parent of the original role.
	 * @return a copy of this role object
	 */
	public AcmeElement copy() {
		AcmeRole new_role = new AcmeRole(this.getName());
		this.copyCommonEltStructure(new_role);
		return new_role;
	};

	/**
	 * This method uses this element instance as a prototype to create a new type.
	 * @param newTypeName the name of the AcmeRoleType to be copied
	 * @return a copy of this AcmeRoleType
	 */

	public AcmeElementType copyAsPrototype(String newTypeName) {
		AcmeVector super_types;
		AcmeRoleType new_role_type = new AcmeRoleType(newTypeName);
		this.copyCommonEltStructure(new_role_type);

		// copyCommonEltStructure overwrites the name of the new type, so
		// reset it properly.  Don't check for legal names here, that will
		// happen if/when we try to add it to anything.
		new_role_type.name = newTypeName;

		// add all of the declared types of this instance as supertypes of
		// the new type.
		AcmeEnumeration e = this.getDeclaredTypes();
		while (e.moreItems()) {
			AcmeObject curType = e.getNextItem();
			if (curType instanceof AcmeRoleType) {
				new_role_type.addSuperType((AcmeRoleType) curType);
			}
			else if (curType instanceof AcmeElementProtoType) {
				new_role_type.addSuperType((AcmeElementProtoType) curType);
			}
			else if (curType instanceof AcmeElementTypeRef) {
				AcmeElementTypeRef t = (AcmeElementTypeRef) curType;
				if (t.getReferentType() instanceof AcmePortType
					|| t.getReferentType() instanceof AcmeElementProtoType) {
					new_role_type.addSuperType((AcmeElementTypeRef) curType);
				}

			}
		};

		return new_role_type;
	};

	/**
	 * all AcmeRoles satisfy the type ConstantTypes.ROLE_BASE_TYPE.
	 * if the generic supertype method doesn't find the type declared, check
	 * whether the type is the ROLE_BASE_TYPE.
	 * @param type the type to be compared to ROLE_BASE_TYPE
	 * @return whether the type is the ROLE_BASE_TYPE
	 */
	public boolean declaresType(AcmeElementType type) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType ();
		}
		if (type == ConstantTypes.ROLE_BASE_TYPE) {
			return true;
		}
		return super.declaresType(type);
	};

	/**
	 *  AcmeRole.resolveTypeReferences() simply passes the request to its supertype
	 *  AcmeElement.  This method is explicitly implemented because so that
	 *  when it's subclasses call super.resolveTypeReferences() it will be propogated
	 *  up the inheritance tree properly.
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
	}

	/**
	 * instantiateElements() walks over the vector of instantiatedTypes of this
	 * element and adds the minimal structure declared in the type description
	 * to this element.  It delegates much of the work to the AcmeTypeManager
	 * class, so that we can change type instantiation policies relatively
	 * centrally (ie by changing the AcmeTypeManager methods) for all elements
	 * at once.  This method recursively calls the same method on all of the
	 * substructure of this element instance.
	 */
	public void instantiateElements() {

		// unify this element with each of the types it instantiates.
		for (int i = 0; i < this.instantiatedTypes.getSize(); i++) {
			// check that we have a prototype role or role type
			// to instantiate.  If we don't then announce an error and move
			// to the next type.
			AcmeObject iType = this.instantiatedTypes.getItemAt(i);
			if (iType instanceof AcmeElementTypeRef) {
				iType = (AcmeObject) ((AcmeElementTypeRef) iType).getReferentType();
			}
			if (iType instanceof AcmeRoleType) {
				AcmeTypeManager.unifyRole((AcmeRoleType) iType, this);

			}
			else if (iType instanceof AcmeElementProtoType) {
				AcmeTypeManager.unifyElementWithPrototype((AcmeElementProtoType) iType, this);

			}
			else if (iType instanceof AcmeType) {
				String errorMsg = null;
				if (iType instanceof AcmeUnresolvedElementType) {
					errorMsg =
						"Instantiation Error.  Unable to instantiate "
							+ " unresolved element type: "
							+ ((AcmeUnresolvedElementType) iType).getTypeName()
							+ " in role: "
							+ this.getName();
				}
				else {
					errorMsg =
						"Instantiation Error.  Type: "
							+ ((AcmeType) iType).getTypeName()
							+ " is not a role type.  Therefore it can not be "
							+ "used to instantiate role: "
							+ this.getName();
				};

				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg, (AcmeType) iType);
				AcmeErrorManager.errorDetected(instError);

			}
			else {
				String errorMsg =
					"Instantiation Error.  Attempted to instantiate a "
						+ "completely unrecognizable type in role: "
						+ this.getName();
				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg);
				AcmeErrorManager.errorDetected(instError);
			};
		}

		// finally, call AcmeElement to have it instantiate representations and any
		// other standard element structure.
		super.instantiateElements();

	};

	/**
	 * This method determine whethers this role satisfies all of its declared types and
	 * design rules, and that it's properties, and representations all
	 * typecheck also.
	 * @return whether this role does typecheck
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckRole(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			AcmeErrorManager.errorRepaired(this,null);

			return true;
		}
		// role doesn't typecheck.  There should be an error at the
		// top of the AcmeTypeManager's typecheckingErrorStack.  If there
		// is such an error then announce it to the AcmeErrorManager and
		// return false.
		AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
		if (eltTypeError != null) {
			AcmeErrorManager.errorDetected(eltTypeError);
			return false;
		}
		// no error was reported.  something went wrong?
		// announce a generic error.
		AcmeErrorManager.errorDetected(
			new ElementTypecheckError(this, "Role " + this.getName() + " does not typecheck.", this));
		return false;
	};

	/**
	 * Creates a workshop for editing this element.
	 * implements the acmeLib.Editable interface.
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editRole(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

	/**
	 * This method is used to implement the acmeLib.Editable interface.
	 */
	public void view() {
		System.out.println("Attempted to view: " + this.getName());
		System.out.println("  but AcmeRole.view() is not yet implemented.");
	};

	/**
	 * This method implements looking up names. It looks up the child object
	 * called 'name'. It checks for attachedPorts.
	 * @param name compared with 'attachedPorts'
	 * @param bubbleUp whether to look in surrounding scopes
	 * @return the Object corresponding to name
	 * @see acmeLib.designRules.Scope#lookupName(String, boolean)
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;

		if (name != null) {
			if (name.equalsIgnoreCase("attachedports")) {
				// BRS: Need to find the system, traverse through the attachments
				// and return ports that are attached to these roles
				AcmeElement parent = this.getParent();
				if (parent instanceof AcmeConnector) {
					parent = parent.getParent();
				}

				if (parent instanceof AcmeSystem) {
					AcmeSystem sys = (AcmeSystem) parent;

					// Iterate through attachments involving the role
					AcmeEnumeration attachments = sys.getAttachments();
					AcmeVector ports = new AcmeVector();
					while (attachments.moreItems()) {
						AcmeAttachment att = (AcmeAttachment) attachments.getNextItem();
						if (!(att instanceof AcmeUnresolvedAttachment) && att.attachedRole.equals(this)) {
							ports.addItem(att.attachedPort);
						}
					}
					return ports.enumerate();
				}
				else {
					return new AcmeEnumeration();
				}

			}
		}

		return super.lookupName(name, bubbleUp);
	}

};
