package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.*;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 *  Class:  AcmeRoleType
<BR>
 *  Purpose:  Stores the type specification for a role.  This is
<BR>
 *          a subclass of an AcmeRole because the Acme type system
<BR>
 *          is prototype-based.  As a result, a type is just a prototypical
<BR>
 *          instance of a role that specifies the aspects that an
<BR>
 *          instance of that type must posses.  Typechecking is done
<BR>
 *          by an external class, AcmeTypeManager, that understands the
<BR>
 *          rules of typechecking
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    6/16/97   RTM     Created class
<BR>
 *    1/14/98   RTM     added support for multiple declared types.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeRoleType extends AcmeRole implements AcmeElementType {

	//
	// private members
	//
	AcmeVector superTypes;

	//
	// public members
	//
	/**
	  * This is a constructor for this class
	  * @param typeName the name for the role type to created
	  */
	public AcmeRoleType(String typeName) {
		// create the prototype object in the superclass constructor
		super(typeName, null);
		this.superTypes = new AcmeVector();
	};

	/**
	  * This method returns the name of this role type
	  * @return the name of this role type
	  */
	public String getTypeName() {
		return this.getName();
	};
	/**
	  * This method sets the name of this role type
	  * @param typeName the name for this role type
	  * @exception AcmeIllegalNameException The name is illegal
	  */
	public void setTypeName(String typeName) throws AcmeIllegalNameException {
		this.setName(typeName);
	};

	/**
	 * This method adds a supertype to a type description
	 * @param superType the supertype to be added to the type description
	 */
	public void addSuperType(AcmeElementType superType) {
		if (superType instanceof AcmeElementTypeRef) {
			superType = ((AcmeElementTypeRef) superType).getReferentType();
		}
		if ((superType instanceof AcmeRoleType)
			|| (superType instanceof AcmeUnresolvedElementType)
			|| (superType instanceof AcmeElementProtoType) 
			) {
			superTypes.addItem((AcmeObject) superType);
		}
		// else throw an exception?
	};

	/**
	  * This method gets the supertypes of a type description
	  * @return the supertypes of a type description
	  */
	public AcmeVector getSuperTypes() {
		return this.superTypes;
	};

	/**
	  * This method removes a supertype from a type description
	  * @param superType the supertype to be removed from the type description
	  */
	public void removeSuperType(AcmeElementType superType) {
		superTypes.removeItem((AcmeObject) superType);
		if (superType instanceof AcmeElementTypeRef) {
			declaredTypes.removeItem((AcmeObject) ((AcmeElementTypeRef) superType).getReferentType());
		}
	};

	/**
	 * This method makes a new role type object, with a deep copy of the old role's
	 * properties, representations, declared, instantiated, and super-type
	 * vectors.
	 * @return a copy of this role type object
	 */
	public AcmeElementType copyType() {
		AcmeRoleType new_role_type = new AcmeRoleType(this.getTypeName());
		this.copyCommonEltStructure(new_role_type);
		this.superTypes.copyItems(new_role_type.superTypes);
		return new_role_type;
	};

	/**
	 * resolveTypeReferences searches through this type object's vector of supertypes
	 * for UnresolvedElementType objects, looks up the actual type that is
	 * referred to by each UnresolvedElementType object, and replaces that
	 * object with the actual type object that should be referred to.
	 * it then calls resolveTypeReferences in its superclass to resolve the name of
	 * all of its substructure (just representations).
	 */
	public void resolveTypeReferences() {
		super.resolveTypeReferences();
		AcmeElementType curType = null;
		AcmeType resolvedType = null;
		AcmeUnresolvedElementType curUnresolvedType;
		for (int i = 0; i < this.superTypes.getSize(); i++) {
			curType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (curType instanceof AcmeUnresolvedElementType) {
				curUnresolvedType = (AcmeUnresolvedElementType) curType;
				if (curUnresolvedType.isQualified()) {
					if (this.getParentDesign() != null) {
						AcmeFamily qFam = this.getParentDesign().getFamily(curUnresolvedType.getQualifier());
						if (qFam != null) {
							AcmeType refType = qFam.getType(curUnresolvedType.getTypeName());
							if (refType != null && refType instanceof AcmeElementType) {
								resolvedType = new AcmeElementTypeRef(qFam, (AcmeElementType) refType);
							}
						}
					}
				}
				else {
					// first check to see if the type is declared in the same
					// family that this type is declared in.  If an AcmeRoleType
					// object's parent field is not null then parent must be a family.
					if (this.getParent() instanceof AcmeFamily) {
						resolvedType = ((AcmeFamily) this.getParent()).getType(curUnresolvedType.getTypeName());
					}

					if (this.getParent() instanceof AcmeDesign) {
						resolvedType = ((AcmeDesign) this.getParent()).getType(curUnresolvedType.getTypeName());
					}
					else if ((resolvedType == null) && (this.getParentDesign() != null)) {
						// then check to see if the supertype is declared in the same
						// design as this type.
						resolvedType = this.getParentDesign().getType(curUnresolvedType.getTypeName());
					}
				}
				if (resolvedType == null) {
					String errorMsg =
						"Unable to resolve supertype \""
							+ curUnresolvedType.getTypeName()
							+ "\" of role type "
							+ this.getTypeName();
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

				}
				else if (resolvedType instanceof AcmeElementTypeRef) {
					AcmeElementTypeRef ref = (AcmeElementTypeRef) resolvedType;
					if (!(ref.getReferentType() instanceof AcmeRoleType)
						&& !(ref.getReferentType() instanceof AcmeElementProtoType)) {
						String errorMsg =
							"Role type \""
								+ this.getTypeName()
								+ "\"'s supertype \""
								+ curUnresolvedType.getTypeName()
								+ "\" is not a role or element type.";
						TypeRefResolutionError badTypeRefError =
							new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
						badTypeRefError.setUnresolvedRef(curUnresolvedType);
						AcmeErrorManager.errorDetected(badTypeRefError);

						resolvedType = null;
					}
				}
				else if (!(resolvedType instanceof AcmeRoleType) && !(resolvedType instanceof AcmeElementProtoType)) {
					String errorMsg =
						"Role type \""
							+ this.getTypeName()
							+ "\"'s supertype \""
							+ curUnresolvedType.getTypeName()
							+ "\" is neither a role nor an element type.";
					TypeRefResolutionError badTypeRefError =
						new TypeRefResolutionError(this, errorMsg, curUnresolvedType.getTypeName());
					badTypeRefError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected(badTypeRefError);

					resolvedType = null;
				}
				else if (resolvedType == this) {
					// Is there an immediate recursive resolution
					String errorMsg = "Component type \"" + this.getTypeName () + "\" cannot have itself as a supertype.";
					TypeRefResolutionError badTypeError = new TypeRefResolutionError (this, errorMsg, resolvedType.getTypeName());
					badTypeError.setUnresolvedRef(curUnresolvedType);
					AcmeErrorManager.errorDetected (badTypeError);
					
					resolvedType = null;
				}

				if (resolvedType != null) {
					this.superTypes.setItemAt((AcmeObject) resolvedType, i);
				};
			};
		}
	}

	/**
	 *  lookupName implements the acmeLib.designRules.Scope interface.  It
	 *  extends the AcmeRole.lookupName method by searching
	 *  through all of its supertypes for a named entity before calling
	 *  its supertype (AcmeRole)'s lookupName method to search for
	 *  properties and representations contained in this type's prototype.
	 *  @param name the name to be mapped to a type or design analyses
	 *  @param bubbleUp boolean, if falso don't check superclass
	 *  @return Object corresponding to "name"
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;

		// check first whether this is referring to "this" type.
		if (name == null) {
			return null;
		}
		else if (name.equalsIgnoreCase("self")) {
			return this;
		}
		else if (name.equalsIgnoreCase("attachedports")) {
			return new AcmeEnumeration();
		}

		// recursively walk up the tree of supertypes hunting for "name"
		for (int i = 0;(i < this.superTypes.getSize()) && (retObj == null); i++) {
			AcmeElementType curSuperType = (AcmeElementType) this.superTypes.getItemAt(i);
			if (!(curSuperType instanceof AcmeUnresolvedElementType))
				retObj = curSuperType.lookupName(name, false);
		};

		// if we didn't find it in a supertype then check for it in this type's
		// prototype object.
		if (retObj == null)
			retObj = super.lookupName(name, bubbleUp);

		return retObj;
	};

	/**
	 * This method determines whether this role type's properties, representations, and
	 * design rule declarations are all type-consistent, and that
	 * all declared supertypes are AcmeRoleType's also.  This method
	 * delegates all of the actual checking to the AcmeTypeManager class.
	 *  @return whether all types do typecheck
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckRoleType(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			AcmeErrorManager.errorRepaired(this, null);

			return true;
		}
		// role type doesn't typecheck.  There should be an error at the
		// top of the AcmeTypeManager's typecheckingErrorStack.  If there
		// is such an error then announce it to the AcmeErrorManager and
		// return false.
		AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
		if (eltTypeError != null) {
			AcmeErrorManager.errorDetected(eltTypeError);
			return false;
		}
		// no error was reported.  something went wrong?
		// announce a generic error.
		AcmeErrorManager.errorDetected(
			new ElementTypecheckError(this, "Role Type " + this.getTypeName() + " does not typecheck.", this));
		return false;
	};

	//
	// Exporting methods
	//

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltType(s, "Role Type", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltType(s, "Role Type", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this role type as a string of the
	 *  form "Role Type <name> [ extends <Typename>+ ]".  For example:
	 *  "Role Type fooT : roleT, superRoleT".
	 * @return string containing this component type's signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("Role Type " + this.getTypeName());
		AcmeEnumeration superTypesEnum =
			this.getSuperTypes().enumerate(new acmeLib.core.filters.NoBaseClassesEnumFilter());
		if (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" extends <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" extends " + eltType.getTypeName());
		}

		while (superTypesEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) superTypesEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 * Creates a workshop for editing this role type.
	 * implements the acmeLib.Editable interface.
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editRoleType(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

}
