
package acmeLib.elements;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;


/** 
 *  Class: AcmeSequenceValue
 *
 *  Purpose:The Sequence class implements an ACME sequence.
 *
 *  Revision History:
 *    Date:     Person    Change
 *	11/5/99		JH		Created
 * 
 * @author Jianing Hu
 */
public class AcmeSequenceValue extends AcmeCompoundElement {

	//
	// protected members
	//
	protected AcmeSequenceType type;
	protected AcmeVector sequenceValues;   // Vector of AcmePropertyValues

	//
	// public members
	//
	/**
	 * This is a constructor for this class 
	 */
	public AcmeSequenceValue() {
		this.setType(null);
		sequenceValues = new AcmeVector();
	}

	/**
	 * This is a constructor for this class 
	 * @param t the value to be assigned to the sequence's type       
	 */
	public AcmeSequenceValue (AcmeSequenceType t) {
		this.setType(t);
		sequenceValues = new AcmeVector();
	}

	/**
	 * This method performs a deep copy -- copy all of the values from this sequence
	 * into new objects in the new sequence
	 * @return a copy of this sequence
	 */
	public AcmeElement copy () {
		AcmeSequenceValue new_seq = new AcmeSequenceValue(this.getType());
		
		copyElements(new_seq);
		
		return new_seq;
	}
	
	/**
	 * This method copies the sub-elements of this sequence to another,
	 * overwriting the sub-elements of the other sequence
	 * @param destSeq the sequence to copy to
	 */
	public void copyElements(AcmeSequenceValue destSeq){
		destSeq.sequenceValues = new AcmeVector();

		// walk over the elements in this sequence, creating a new AcmeVector
		AcmeEnumeration e = sequenceValues.enumerate();
		while (e.moreItems()) {
			try {
				destSeq.addValue( ((AcmeElement) e.getNextItem()).copy() );
			} catch (ClassCastException exception) {
				// can't copy this, just continue
			}
		}
	}

	/** 
	 * This method checks whether two sequences are equal
	 * @param test_value the sequence to be compared
	 * @return whether "this" and test_value are equal
	 */
	public boolean equals(Object test_value) {
		// first check whether test_value and this are the same types of sequences
		if (!(test_value instanceof AcmeSequenceValue))
			return false;

		// else test_value and "this" are the same types of sequences,
		// by comparing their values.
		AcmeEnumeration these_values = this.getValues();
		AcmeEnumeration test_values = ((AcmeSequenceValue) test_value).getValues();
		AcmeElement this_prop_val, test_prop_val;
		while (these_values.moreItems()) {
			if (!(test_values.moreItems()))
				return false;   // this has more values than test
			else {
				this_prop_val = (AcmeElement) these_values.getNextItem();
				test_prop_val = (AcmeElement) test_values.getNextItem();
				if (!(this_prop_val.equals(test_prop_val)))
					return false;
			}
		}

		if (test_values.moreItems())
			return false;   // test has more values than this.

		return true;    // if it gets here then test_value and this are equal
	}


	/**
	 * This method sets the type of this sequence
	 * @param t the value to be assigned to type
	 */
	public void setType (AcmeSequenceType t) { this.type = t; };
	/**
	 * This method returns the type of this sequence
	 * @return the type of this sequence
	 */
	public AcmeSequenceType getType() { return this.type; };

	/**
	 * This method returns the size of the sequence
	 * @return the size of the sequence
	 */
	public int getSize() { return sequenceValues.getSize(); };

	/**
	 * This method returns the value of this sequence
	 * @param index the property value at position index
	 * @return the value of this sequence
	 */
	public AcmeElement getValue (int index) {
		return (AcmeElement) sequenceValues.getItemAt(index);
	}
	/**
	 * This method adds value to this sequence 
	 * @param value the value to be added to this sequence
	 */ 
	public void addValue (AcmeElement value) {
		sequenceValues.addItem(value);
	}

	/**
	 * This method inserts a value in the sequence at a given index
	 * @param value the value to be inserted into the sequence
	 * @param index the position at which value will be inserted
	 */
	public void insertValueAt(AcmeElement value, int index) {
		sequenceValues.insertItemAt(value,index);
	}

	/**
	 * This method returns an enumeration of all the values in this sequence
	 * @return an enumeration of all the values in this sequence
	 */
	public AcmeEnumeration getValues () {
		return sequenceValues.enumerate();
	}

	/**
	 * This method converts AcmeSequence to String
	 * @return the String value of this sequence
	 */
	public String toString() {
		StringBuffer buf = new StringBuffer();
		AcmeEnumeration e = sequenceValues.enumerate();
		buf.append("<");
		while (e.moreItems()) {
			buf.append(e.getNextItem().toString());
			if (e.moreItems())
				buf.append(", ");
		}
		buf.append(">");
		return buf.toString();
	}
	
	/**
	 * Get the name of this sequence.
	 * A sequence doesn't have a "declared" name.
	 * This method returns an "inner" name, which is the concatenation
	 * of the sub-element names in the forms of n1_n2...
	 */
	public String getName(){
		StringBuffer buf = new StringBuffer();
		AcmeEnumeration e = this.getValues();
		AcmeElement f = null;
		while (e.moreItems()){
			f = (AcmeElement)e.getNextItem();
			buf.append(f.getName());
			if (e.moreItems()){
				buf.append("_");
			}
		}
		return buf.toString();
	}
}

