
package acmeLib.elements;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.designRules.Scope;
import acmeLib.designRules.SetExpressionNode;


/**
 *  Class: AcmeSetValue
 *
 *  Purpose:The AcmeSetValue class implements an Acme set.  It includes
 *
 *  Revision History:
 *    Date:     Person    Change
 *	11/5/99		JH		Created
 *
 *  @author Jianing Hu
 */
public class AcmeSetValue extends AcmeCompoundElement
                     implements acmeLib.designRules.SetExpressionNode
{

  //
  // protected members
  //
  protected AcmeSetType type;
  protected AcmeType contentsType = ConstantTypes.UNSPECIFIED_TYPE;
  protected AcmeVector setValues;   // Vector of AcmeElements

  /**
   *  holds the scope for/if this is used as a SetExpressionNode.  This
   *  field has no meaning when used as a regular Acme set.
   */
  protected acmeLib.designRules.Scope scope = null;



  // Acme and Armani have different equals methods, so the generic
  // AcmeSetValue.equals(Object test) method calls the appropriate value
  // depending on whether we're dealing with pure Acme or Armani.

  /**
   * Armani set equality check
   * This method checks whether two sets are equal
   * @param test the set to be compared
   * @return whether "this" and test are equal
   */
  protected boolean equalsInArmani(Object test) {
    if (test instanceof SetExpressionNode) {
        return acmeLib.designRules.DrUtils.equalSets(this, (SetExpressionNode) test);
    } else {
        return false;
    }
  }


  /**
   * "Pure-Acme" equality check
   * This method checks whether two sets are equal
   * @param test_value the set to be compared
   * @return whether "this" and test_value are equal
   */

  protected boolean equalsInAcme(Object test_value) {
    // first check whether test_value and this are the same types of sets
    if (!(test_value instanceof AcmeSetValue))
        return false;   // test_valuse isn't an AcmeSetValue
    else if ( this.getSize() != ((AcmeSetValue) test_value).getSize() )
        return false;   // the sets have a different number of elements

    // else test_value and "this" are the same types and size of sets,
    // now compare their values.
    // RTM: this is a really slow comparison algorithm (O(n^2)), which may
    //      or may not prove to be problematic for reasonably sized
    //      (<200 element)sets.  I'll try to implement a more efficient
    //      algorithm later if necessary.
    AcmeEnumeration these_values = this.getValues();
    AcmeEnumeration test_values;

    // the sets have the same number of elements, compare them
    AcmeElement this_prop_val, test_prop_val;
    while (these_values.moreItems()) {
        boolean found_match = false;
        // check whether the current value in "this" set exists in the "test" set
        test_values = ((AcmeSetValue) test_value).getValues();
        this_prop_val = (AcmeElement) these_values.getNextItem();
        while (test_values.moreItems()) {
            test_prop_val = (AcmeElement) test_values.getNextItem();
            if (test_prop_val.equals(this_prop_val)) {
                found_match = true;
                break;
            }
        }
        test_values = null;  // mark for gc
        if (found_match == false) {
            return false;   // couldn't find a match for this value in test set.
        }
    }

    // if it gets here then test_value and "this" are equal
    return true;
  }



  //
  // public members
  //
  /**
   * This is a constructor for this class
   */
  public AcmeSetValue() {
    this.setType(null);
    setValues = new AcmeVector();
  }

  /**
   * This is a constructor for this class
   * @param t the value to be assigned to the set's type
   */
  public AcmeSetValue (AcmeSetType t) {
    this.setType(t);
    setValues = new AcmeVector();
  }

  /**
   * This method performs a deep copy -- copy all of the values from this set
   * into new objects in the new set
   * @return a copy of this set
   */
  public AcmeElement copy () {
	AcmeSetValue new_set = new AcmeSetValue(this.getType());
    new_set.setSetType(this.getSetType());
	copyValues(new_set);

    return new_set;
  }
  
  public void copyValues(AcmeSetValue new_set){
	  new_set.setValues = new AcmeVector();
    // walk over the elements in this sequence, creating a new AcmeVector
    AcmeEnumeration e = setValues.enumerate();
	while (e.moreItems()) {
	    try {
	        new_set.addValue( ((AcmeElement)e.getNextItem()).copy() );
	    } catch (ClassCastException exception) {
	        // can't copy this, just continue
	    }
	}
  }

 /**
   * This method checks whether two sets are equal
   * @param test the set to be compared
   * @return whether "this" and test are equal
   */
  public boolean equals(Object test) {
    if (acmeLib.core.globals.LanguageConfig.IS_ARMANI) {
        return this.equalsInArmani(test);
    } else {
        return this.equalsInAcme(test);
    }
  }

 /**
   * This method sets the type of this set
   * @param t the value to be assigned to type
   */
  public void setType (AcmeSetType t) {
    this.type = t;
    // set the contents type for the set to the contents type of the
    // new set.
    if (this.type != null) {
        this.setSetType(this.type.getSetType());
    }
  }
 /**
   * This method returns the type of this set
   * @return the type of this set
   */
  public AcmeSetType getType() { return this.type; };

 /**
   * This method returns the size of the set
   * @return the size of the set
   */
  public int getSize() { return setValues.getSize(); };

 /**
   * This method returns the value of this set
   * @param index the position of the value to be returned
   * @return the value of this set
   */
  public AcmeElement getValue (int index) {
    return (AcmeElement) setValues.getItemAt(index);
  }

 /**
   * This method adds value to this set
   * @param value the value to be added to this set
   */
  public void addValue (AcmeElement value) {
    if (value == null) {
        return;     // nothing to add...
    }

    // check for repeat value
    boolean repeatVal = false;
    int setSize = this.setValues.getSize();
    for (int i = 0; (i < setSize) && (repeatVal == false); i++) {
        if (value.equals(this.setValues.getItemAt(i))) {
            repeatVal = true;
        }
    }

    // add value if it's not already in the set.
    if (!repeatVal)
        setValues.addItem(value);
  }

 /**
   * This method returns an enumeration of all the values in this set
   * @return an enumeration of all the values in this set
   */
  public AcmeEnumeration getValues () {
  	return setValues.enumerate();
  };


  //
  // implement the SetExpressionNode interface by binding all of the interface's
  // methods to appropriate other methods.

  /**
   * This method returns the scope of this set
   * @return scope of this set
   */
    public Scope getSetScope() { return this.scope; };
  /**
   * This method sets the scope of this set
   * @param newScope the value to be assigned to the set's scope
   */
    public void setSetScope(Scope newScope) { this.scope = newScope; };

  /**
   * This method returns this set's contentsType
   * @return this set's contentsType
   */
    public AcmeType getSetType() { return this.contentsType; }
  /**
   * This method sets the contentsType for this set
   * @param newType the contentsType of this set
   */
    public void setSetType(AcmeType newType) { this.contentsType = newType; };

    /**
     * This method returns the size of the current set
     * @return the size of the current set
     */
    public int size() { return this.setValues.getSize(); };


    /**
     * refreshSet() recomputes the value of the set.  An AcmeSet value should
     *  not change, as it can not be dependent on other referenced values,
     *  so do nothing.
     */
    public void refreshSet() {};

    /**
     * contains(Object elt) returns true iff elt is a member of this set
     * @return whether elt is a member of this set
     */
    public boolean contains(AcmeObject elt) {
        return this.setValues.contains(elt);
    }

    /**
     * enumerate() returns an enumeration of all of the elements of the set
     * @return an enumeration of all of the elements of the set
     */
    public AcmeEnumeration enumerate() { return this.getValues(); }

    /**
     * This method returns the string representing this set (for exporting/unparsing)
     * @return the String value of this set
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        AcmeEnumeration e = setValues.enumerate();
    	buf.append("{");
    	while (e.moreItems()) {
    	    buf.append(e.getNextItem().toString());
    		if (e.moreItems())
    		    buf.append(", ");
    	}
    	buf.append("}");
    	return buf.toString();
    }

   /**
     * This method returns the string representing this set following a prefix
     * @param prefix the prefix of the string to be returned
     * @return the String value of this set
     */

    public String toString(String prefix) { return prefix + this.toString(); }
}
