package acmeLib.elements;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeDuplicateNameException;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.exceptions.AcmeKeywordNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.designRules.AcmeResolutionException;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.designRules.Scope;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.ElementTypecheckError;
import acmeLib.errorManagement.TypeInstantiationError;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmePropertyType;
import acmeLib.properties.AcmeSet;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 *  Class: AcmeSystem <BR>
 *  Purpose: Class that represents an Acme system. <BR>
 *  Description:  An AcmeSystem defines a configuration of Acme elements.  It <BR>
 *  contains instances of AcmeElements: AcmeComponents, <BR>
 *  AcmeConnectors, AcmePorts and AcmeRoles.  It also contains a <BR>
 *  set of AcmeAttachments and Properties. <BR> <BR>
 *  Revision History: <BR>
 *    Date:     Person    Change <BR>
 *    ?/?/97    DK      Created C++ version <BR>
 *    4/3/97    RTM     Translated to Java <BR>
 *    1/5/98    RTM     Added support for storing Design Analyses <BR> <BR>
 *    1/8/03    SWC     Fixed NullPointerException in areConnectedPorts()<br>
 *                       method, which occurs when port1 has attachment, but<br>
 *                       port2 doesn't have an attachment.
 *
 *  @author Bob Monroe <BR>
 */
public class AcmeSystem extends AcmeElement {

	//
	// Protected members
	//
	protected AcmeRepresentation parent_rep;
	protected AcmeMap groupMap;
	protected AcmeMap componentMap;
	protected AcmeMap connectorMap;
	protected AcmeMap portMap;
	protected AcmeMap roleMap;
	protected AcmeMap designAnalysisMap;
	protected AcmeVector attachments;
	protected Map port2Attachments;
	protected Map role2Attachments;

	/**
	 * This method does a deep copy of the ports in this component and puts
	 * them into the ports list of destComp.  This removes
	 * any ports that are currently in destComp
	 * @param destSys the AcmeSystem that gets a copy of the ports in this component
	 */
	protected void copyPorts(AcmeSystem destSys) {
		// - get the properties of this element,
		// - make a new property map in dest_obj
		// - foreach property in this elt, put a deep copy of it in dest_obj.
		AcmeEnumeration ports_enum = this.getPorts();
		destSys.portMap = new AcmeMap();

		AcmePort cur_port;
		while (ports_enum.moreItems()) {
			cur_port = (AcmePort) ports_enum.getNextItem();
			try {
				destSys.addPort((AcmePort) cur_port.copy());
			}
			// there shouldn't be a name error here, since we're copying something
			// that's already been added...
			catch (AcmeIllegalNameException keyException) {
			}
		}
	};

	/**
	 * This method does a deep copy of the roles in this system and puts
	 * them into the roles list of destSys.  This removes
	 * any ports that are currently in destSys
	 * @param destSys the AcmeSystem that gets a copy of the ports in this component
	 */
	protected void copyRoles(AcmeSystem destSys) {
		// - get the properties of this element,
		// - make a new property map in dest_obj
		// - foreach property in this elt, put a deep copy of it in dest_obj.
		AcmeEnumeration roles_enum = this.getRoles();
		destSys.roleMap = new AcmeMap();

		AcmeRole cur_role;
		while (roles_enum.moreItems()) {
			cur_role = (AcmeRole) roles_enum.getNextItem();
			try {
				destSys.addRole((AcmeRole) cur_role.copy());
			}
			// there shouldn't be a name error here, since we're copying something
			// that's already been added...
			catch (AcmeIllegalNameException keyException) {
			}
		}
	};

	/**
	 * This method does a deep copy of the components in this system and puts
	 * them into the components list of destSys.  This removes
	 * any components that are currently in destSys
	 *  @param destSys the AcmeSystem that gets a copy of the ports in this component
	 */
	protected void copyComponents(AcmeSystem destSys) {
		// - get the properties of this element,
		// - make a new property map in dest_obj
		// - foreach property in this elt, put a deep copy of it in dest_obj.
		AcmeEnumeration comps_enum = this.getComponents();
		destSys.componentMap = new AcmeMap();

		AcmeComponent cur_comp;
		while (comps_enum.moreItems()) {
			cur_comp = (AcmeComponent) comps_enum.getNextItem();
			try {
				destSys.addComponent((AcmeComponent) cur_comp.copy());
			}
			// there shouldn't be a name error here, since we're copying something
			// that's already been added...
			catch (AcmeIllegalNameException keyException) {
			}
		}
	};

	/**
	 * This method does a deep copy of the components in this system and puts
	 * them into the components list of destSys.  This removes
	 * any components that are currently in destSys
	 *  @param destSys the AcmeSystem that gets a copy of the ports in this component
	 */
	protected void copyGroups(AcmeSystem destSys) {
		// - get the properties of this element,
		// - make a new property map in dest_obj
		// - foreach property in this elt, put a deep copy of it in dest_obj.
		AcmeEnumeration groups_enum = this.getGroups();
		destSys.groupMap = new AcmeMap();

		AcmeGroup cur_group;
		while (groups_enum.moreItems()) {
			cur_group = (AcmeGroup) groups_enum.getNextItem();
			try {
				destSys.addGroup((AcmeGroup) cur_group.copy());
			}
			// there shouldn't be a name error here, since we're copying something
			// that's already been added...
			catch (AcmeIllegalNameException keyException) {
			}
		}
	};

	/**
	 * This method does a deep copy of the connectors in this system and puts
	 * them into the connectors list of destSys.  This removes
	 * any connectors that are currently in destSys
	 * @param destSys the AcmeSystem that gets a copy of the ports in this component
	 */
	protected void copyConnectors(AcmeSystem destSys) {
		// - get the properties of this element,
		// - make a new property map in dest_obj
		// - foreach property in this elt, put a deep copy of it in dest_obj.
		AcmeEnumeration conns_enum = this.getConnectors();
		destSys.connectorMap = new AcmeMap();

		AcmeConnector cur_conn;
		while (conns_enum.moreItems()) {
			cur_conn = (AcmeConnector) conns_enum.getNextItem();
			try {
				destSys.addConnector((AcmeConnector) cur_conn.copy());
			}
			// there shouldn't be a name error here, since we're copying something
			// that's already been added...
			catch (AcmeIllegalNameException keyException) {
			}
		}
	};

	//
	// Public members
	//

	// Constructors

	/**
	 * This is a constructor for this class
	 */
	public AcmeSystem() {
		groupMap = new AcmeMap();
		componentMap = new AcmeMap();
		connectorMap = new AcmeMap();
		portMap = new AcmeMap();
		roleMap = new AcmeMap();
		attachments = new AcmeVector(4, 4);
		port2Attachments = new HashMap (4,4);
		role2Attachments = new HashMap (4,4);
		designAnalysisMap = new AcmeMap();

		this.name = null;
		this.setParentRepresentation(null);
	};

	/**
	 * This is a constructor for this class
	 * @param name the name of the system being created
	 */
	public AcmeSystem(String name) {
		componentMap = new AcmeMap();
		groupMap = new AcmeMap();
		connectorMap = new AcmeMap();
		portMap = new AcmeMap();
		roleMap = new AcmeMap();
		attachments = new AcmeVector(4, 4);
		port2Attachments = new HashMap (4,4);
		role2Attachments = new HashMap (4,4);
		designAnalysisMap = new AcmeMap();

		this.name = name;
		this.setParentRepresentation(null);
	};

	/**
	 * This method returns parent_rep
	 * @return parent_rep
	 */
	public AcmeRepresentation getParentRepresentation() {
		return this.parent_rep;
	};
	/**
	 * This method sets the parent representation
	 * @param rep the value to be assigned to parent_rep
	 */
	public void setParentRepresentation(AcmeRepresentation rep) {
		this.parent_rep = rep;
	};

	/**
	 * Retrieves the fully qualified, period delimited, name of the AcmeElement.
	 * This name identifies an object in the structure tree that can be reached
	 * by taking a unique path from the root design.  AcmeSystem overrides this
	 * method to deal properly with the case where it's parent is a representation
	 * rather than another AcmeElement.
	 * @return the fully qualified name of the AcmeElement
	 **/
	public String getQualifiedName() {

		StringBuffer nameBuf = new StringBuffer();

		if (this.getParentRepresentation() != null) {
			// then this system is declared in the scope of a representation.
			// set its name accordingly.
			String parentName = this.getParentRepresentation().getQualifiedName();
			if (parentName != null) {
				if (!(parentName.equals(""))) {
					nameBuf.append(parentName + ".");
				};
				nameBuf.append(this.getName());
			};
		}
		else if (this.getParent() != null) {
			// then this system is NOT declared in the scope of a representation.
			// set its name accordingly.
			String parentName = this.getParent().getQualifiedName();
			if (parentName != null) {
				if (!(parentName.equals(""))) {
					nameBuf.append(parentName + ".");
				};
				nameBuf.append(this.getName());
			};
		}
		else {
			nameBuf.append(this.getName());
		};

		return nameBuf.toString();
	};

	/**
	 * This method uses this system instance as a prototype to create a new family.
	 * @param newFamName the name of the new familty
	 * @return a copy of this family
	 */
	public AcmeElementType copyAsPrototype(String newFamName) {
		AcmeFamily new_fam = new AcmeFamily(newFamName);
		this.copyComponents(new_fam);
		this.copyConnectors(new_fam);
		this.copyPorts(new_fam);
		this.copyRoles(new_fam);
		this.copyCommonEltStructure(new_fam);

		// add all of the declared types of this instance as supertypes of
		// the new type.
		AcmeEnumeration e = this.getDeclaredTypes();
		while (e.moreItems()) {
			AcmeObject curType = e.getNextItem();
			if (curType instanceof AcmeFamily) {
				new_fam.addSuperType((AcmeFamily) curType);
			}
			else if (curType instanceof AcmeElementProtoType) {
				new_fam.addSuperType((AcmeElementProtoType) curType);
			};
		};

		return new_fam;
	};

	/**
	 * This method returns a new AcmeSystem that is a deep-copy of this system.
	 * @return a copy of this system
	 */
	public AcmeElement copy() {
		AcmeSystem newSys = new AcmeSystem(this.getName());

		// walk through the children, adding copies of them to the new system
		AcmeEnumeration child_enumeration = new AcmeEnumeration();
		AcmeEnumeration currentEnum;

		// this try block should not encounter any double names or keywords,
		// as there should be no way for them to be added to the original
		// system that is being copied, so just ignore either of the
		// Acme*NameExceptions.
		try {
			currentEnum = portMap.enumerate();
			while (currentEnum.moreItems()) {
				AcmePort newPort = (AcmePort) ((AcmePort) currentEnum.getNextItem()).copy();
				newPort.setParent(newSys);
				newSys.addPort(newPort);
			};

			currentEnum = roleMap.enumerate();
			while (currentEnum.moreItems()) {
				AcmeRole newRole = (AcmeRole) ((AcmeRole) currentEnum.getNextItem()).copy();
				newRole.setParent(newSys);
				newSys.addRole(newRole);
			};

			currentEnum = componentMap.enumerate();
			while (currentEnum.moreItems()) {
				AcmeComponent newComp = (AcmeComponent) ((AcmeComponent) currentEnum.getNextItem()).copy();
				newComp.setParent(newSys);
				newSys.addComponent(newComp);
			};

			currentEnum = groupMap.enumerate();
			while (currentEnum.moreItems()) {
				AcmeGroup newGroup = (AcmeGroup) ((AcmeGroup) currentEnum.getNextItem()).copy();
				newGroup.setParent(newSys);
				newSys.addGroup(newGroup);
			};

			currentEnum = connectorMap.enumerate();
			while (currentEnum.moreItems()) {
				AcmeConnector newConn = (AcmeConnector) ((AcmeConnector) currentEnum.getNextItem()).copy();
				newConn.setParent(newSys);
				newSys.addConnector(newConn);
			};

			currentEnum = attachments.enumerate();
			// Need to construct new attachments because
			// attachment.copy returns references in the old design
			while (currentEnum.moreItems()) {
				AcmeAttachment curAttachment = (AcmeAttachment) ((AcmeAttachment) currentEnum.getNextItem()).copy();
				AcmeRole oldRole = curAttachment.getRole();
				AcmePort oldPort = curAttachment.getPort();

				AcmePort newPort =
					newSys.getComponent(((AcmeComponent) oldPort.getParent()).getName()).getPort(oldPort.getName());
				AcmeRole newRole =
					newSys.getConnector(((AcmeConnector) oldRole.getParent()).getName()).getRole(oldRole.getName());
				curAttachment.setPort(newPort);
				curAttachment.setRole(newRole);

				curAttachment.setParent(newSys);
				newSys.addAttachment(curAttachment);
			}

			// RTM 1/15/98 -- Currently, just shallow copy the map of design analyses
			//                because none of the design rule/analyses/expressions
			//                have deep-copy methods implemented
			newSys.designAnalysisMap = this.designAnalysisMap;

			this.copyCommonEltStructure(newSys);
		}
		catch (AcmeIllegalNameException dup_ex) {
		};

		return newSys;
	};

	/**
	 * This method updates the role, port, component, connectors, or analysis name in
	 * the appropriate map.  If obj is not a port, role, component, or connector then
	 * it delegates the operation to its superclass.
	 * @param oldName the name of the old child type
	 * @param obj an Object, checked for an instance of AcmeType
	 * @exception AcmeIllegalNameException thrown when the type's name is illegal
	 */
	public void childNameChanged(String oldName, AcmeObject obj) throws AcmeIllegalNameException {

		if (obj instanceof AcmeRole) {
			String newName = ((AcmeRole) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a role name to null.");
			}
			else if (!newName.equals(oldName)) {
				// remove the oldName from the name map, read object with new name
				AcmeRole tmp_role = getRole(oldName);
				if (tmp_role != null) {
					eltNamespace.removeName(oldName);
					tmp_role.setParent(null);
					roleMap.removeItem(oldName);
				};

				this.addRole((AcmeRole) obj);
			};
			// else the name hasn't really changed, just return.
		}

		else if (obj instanceof AcmePort) {
			String newName = ((AcmePort) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a port name to null.");
			}
			else if (!newName.equals(oldName)) {
				// remove the oldName from the name map, readd object with new name
				AcmePort tmp_port = getPort(oldName);
				if (tmp_port != null) {
					eltNamespace.removeName(oldName);
					tmp_port.setParent(null);
					portMap.removeItem(oldName);
				};

				this.addPort((AcmePort) obj);
			};
			// else the name hasn't really changed, just return.
		}

		else if (obj instanceof AcmeComponent) {
			String newName = ((AcmeComponent) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a component name to null.");
			}
			else if (!newName.equals(oldName)) {
				// remove the oldName from the name map, re-add object with new name
				AcmeComponent tmp_comp = getComponent(oldName);
				if (tmp_comp != null) {
					eltNamespace.removeName(oldName);
					tmp_comp.setParent(null);
					componentMap.removeItem(oldName);
				};

				this.addComponent((AcmeComponent) obj);
			};
			// else the name hasn't really changed, just return.
		}

		else if (obj instanceof AcmeGroup) {
			String newName = ((AcmeGroup) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a group name to null.");
			}
			else if (!newName.equals(oldName)) {
				// remove the oldName from the name map, re-add object with new name
				AcmeGroup tmp_group = getGroup(oldName);
				if (tmp_group != null) {
					eltNamespace.removeName(oldName);
					tmp_group.setParent(null);
					groupMap.removeItem(oldName);
				};

				this.addGroup((AcmeGroup) obj);
			};
			// else the name hasn't really changed, just return.
		}

		else if (obj instanceof AcmeConnector) {
			String newName = ((AcmeConnector) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a connector name to null.");
			}
			else if (!newName.equals(oldName)) {
				// remove the oldName from the name map, readd object with new name
				AcmeConnector tmp_conn = getConnector(oldName);
				if (tmp_conn != null) {
					eltNamespace.removeName(oldName);
					tmp_conn.setParent(null);
					connectorMap.removeItem(oldName);
				};
				this.addConnector((AcmeConnector) obj);
			};
			// else the name hasn't really changed, just return.
		}

		else if (obj instanceof DesignAnalysis) {
			String newName = ((DesignAnalysis) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a design analysis name to null.");
			}
			else if (!newName.equals(oldName)) {
				this.removeDesignAnalysis(oldName);
				this.addDesignAnalysis((DesignAnalysis) obj);
			};
			// else the name hasn't really changed, just return.
		}

		else {
			super.childNameChanged(oldName, obj);
		};

	};

	/**
	 * sets the name of the AcmeElement, and updates it's containers where appropriate
	 * @param newName the name being assigned to the AcmeElement
	 * @exception AcmeIllegalNameException thrown when newName is illegal
	 */
	public void setName(String newName) throws AcmeIllegalNameException {
		String oldName = this.name;
		super.setName(newName);
		if (this.getParentRepresentation() != null) {
			this.getParentRepresentation().childNameChanged(oldName, this);
		};
	};

	///////////////
	// Scope
	///////////////

	/**
	 * This method returns the scope of the system
	 * @return the scope of the system
	 */
	public acmeLib.designRules.Scope getScope() {
		AcmeRepresentation parentRep = getParentRepresentation();
		if (parentRep != null) {
			return parentRep;
		}
		// else this is a top-level system, return the design it lives in.
		return this.getParentDesign();
	}

	///////////////
	// Groups
	///////////////

	/**
	 * add a group to the system
	 * @param group group to be added to the system
	 * @exception AcmeIllegalNameException thrown when the component's name is illegal
	 */
	public void addGroup(AcmeGroup group) throws AcmeIllegalNameException {
		if (group == null)
			return; // there is no component, return gracefully
		if (LanguageConfig.CHECK_FOR_KEYWORDS)
			if (LanguageConfig.isAcmeKeyword(group.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + group.getName() + "\" used for group name.");

		if (group.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a group with null name.");
		}

		else if (eltNamespace.containsName(group.getName())) {
			throw new AcmeDuplicateNameException("Attempted to redefine " + "identifier " + group.getName());
		};

		eltNamespace.addName(group.getName());
		group.setParent(this);
		if (!groupMap.contains(group)) {
			groupMap.addItem(group.getName(), group);
		};
	};

	/**
	 * This method returns a group in this system
	 * @param group_name the name of the group to be returned
	 * @return the group corresponding to group_name
	 */
	public AcmeGroup getGroup(String group_name) {
		return (AcmeGroup) groupMap.lookupItem(group_name);
	};

	/**
	 * This method removes a group from this system
	 * @param group_name the name of the group to be removed
	 */
	public void removeGroup(String group_name) {
		AcmeGroup tmp_group = getGroup(group_name);
		if (tmp_group == null)
			return; // there is no such component to remove, return gracefully
		else {
			eltNamespace.removeName(tmp_group.getName());
			removeGroup(tmp_group);
		};
	};

	/**
	  * This method removes a group from this system
	  * @param group the group to be removed
	  */
	public void removeGroup(AcmeGroup group) {
		if (group == null)
			return;
		group.setParent(null);
		eltNamespace.removeName(group.getName());
		groupMap.removeItem(group.getName());
	};

	/**
	 * This method returns all groups in this system
	 * @return an enumeration of all groups in this system
	 */
	public AcmeEnumeration getGroups() {
		return groupMap.enumerate();
	};

	///////////////
	// Components
	///////////////

	/**
	 * add a component to the system
	 * @param comp component to be added to the system
	 * @exception AcmeIllegalNameException thrown when the component's name is illegal
	 */
	public void addComponent(AcmeComponent comp) throws AcmeIllegalNameException {
		if (comp == null)
			return; // there is no component, return gracefully
		if (LanguageConfig.CHECK_FOR_KEYWORDS)
			if (LanguageConfig.isAcmeKeyword(comp.getName()))
				throw new AcmeKeywordNameException(
					"Illegal keyword \"" + comp.getName() + "\" used for component name.");

		if (comp.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a component with null name.");
		}

		else if (eltNamespace.containsName(comp.getName())) {
			throw new AcmeDuplicateNameException("Attempted to redefine " + "identifier " + comp.getName());
		};

		eltNamespace.addName(comp.getName());
		comp.setParent(this);
		if (!componentMap.contains(comp)) {
			componentMap.addItem(comp.getName(), comp);
		};
	};

	/**
	 * This method returns a component in this system
	 * @param comp_name the name of the component to be returned
	 * @return the component corresponding to comp_name
	 */
	public AcmeComponent getComponent(String comp_name) {
		return (AcmeComponent) componentMap.lookupItem(comp_name);
	};

	/**
	 * This method removes a component from this system
	 * @param comp_name the name of the component to be removed
	 */
	public void removeComponent(String comp_name) {
		AcmeComponent tmp_comp = getComponent(comp_name);
		if (tmp_comp == null)
			return; // there is no such component to remove, return gracefully
		else {
			eltNamespace.removeName(tmp_comp.getName());
			removeComponent(tmp_comp);
		};
	};

	/**
	  * This method removes a component from this system
	  * @param comp the component to be removed
	  */
	public void removeComponent(AcmeComponent comp) {
		if (comp == null)
			return;
		detachComponent(comp);
		comp.setParent(null);
		eltNamespace.removeName(comp.getName());
		componentMap.removeItem(comp.getName());
	};

	/**
	 * This method returns all component in this system
	 * @return an enumeration of all component in this system
	 */
	public AcmeEnumeration getComponents() {
		return componentMap.enumerate();
	};

	/**
	 * Remove all of the attachments and bindings that this component's
	 * ports take part in.  This method should be called before removing
	 * a component from a system with the AcmeSystem.removeComponent(...)
	 * methods.  It should <em> not </em> be called when simply renaming
	 * a component, as that should preserve connections...
	 * @param comp The component to detach from the rest of the system
	 */
	public void detachComponent(AcmeComponent comp) {
		AcmeEnumeration ports = comp.getPorts();
		AcmeEnumeration attachments = null;
		while (ports.moreItems()) {
			AcmePort curPort = (AcmePort) ports.getNextItem();
			attachments = this.getAttachments(curPort);
			while (attachments.moreItems()) {
				AcmeAttachment curAttachment = (AcmeAttachment) attachments.getNextItem();
				this.removeAttachment(curAttachment);
			};
		};
	};

	///////////////
	// Connectors
	///////////////

	/**
	 * This method adds a connector to the system
	 * @param conn the connector to be added to the system
	 * @exception AcmeIllegalNameException connector name is illegal
	 */
	public void addConnector(AcmeConnector conn) throws AcmeIllegalNameException {
		if (conn == null)
			return; // there is no connector, return gracefully
		if (LanguageConfig.CHECK_FOR_KEYWORDS)
			if (LanguageConfig.isAcmeKeyword(conn.getName()))
				throw new AcmeKeywordNameException(
					"Illegal keyword \"" + conn.getName() + "\" used for connector name.");

		if (conn.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a connector with null name.");
		}

		else if (eltNamespace.containsName(conn.getName()))
			throw new AcmeDuplicateNameException("Attempted to redefine " + "identifier " + conn.getName());

		eltNamespace.addName(conn.getName());

		conn.setParent(this);
		connectorMap.addItem(conn.getName(), conn);
	};

	/**
	 * This method returns a connector corresponding to conn_name
	 * @param conn_name the name of the connector to be returned
	 * @return the connector corresponding to conn_name
	 */
	public AcmeConnector getConnector(String conn_name) {
		return (AcmeConnector) connectorMap.lookupItem(conn_name);
	};

	/**
	 * This method removes a connector from the system
	 * @param name the name of the connector to be removed
	 */
	public void removeConnector(String name) {
		AcmeConnector tmp_conn = getConnector(name);
		if (tmp_conn == null)
			return; // there is no such connector to remove, return gracefully
		else {
			eltNamespace.removeName(tmp_conn.getName());
			removeConnector(tmp_conn);
		};
	};

	/**
	 * remove a connector from the system and deleting all attachments and
	 * bindings between this connector and other entities.
	 * @param conn the connector to be removed from the system
	 */
	public void removeConnector(AcmeConnector conn) {
		if (conn == null)
			return;
		detachConnector(conn);
		eltNamespace.removeName(conn.getName());
		conn.setParent(null);
		connectorMap.removeItem(conn.getName());
	};

	/**
	 * This method returns an enumeration of all the connectors in this system
	 * @return an enumeration of all the connectors in this system
	 */
	public AcmeEnumeration getConnectors() {
		return connectorMap.enumerate();
	};

	/**
	 * Detach this connector from all other entities in this system.  This
	 * removes all of the attachments and bindings objects that this
	 * connector's roles are associated with.
	 * @param conn The connector to detach from the rest of the system
	 */
	public void detachConnector(AcmeConnector conn) {
		AcmeEnumeration roles = conn.getRoles();
		AcmeEnumeration attachments = null;
		while (roles.moreItems()) {
			AcmeRole curRole = (AcmeRole) roles.getNextItem();
			attachments = this.getAttachments(curRole);
			while (attachments.moreItems()) {
				AcmeAttachment curAttachment = (AcmeAttachment) attachments.getNextItem();
				this.removeAttachment(curAttachment);
			};
		};
	};

	///////////////
	// Ports
	///////////////

	/**
	 * This method adds a port to the system
	 * @param port the port to be added to the system
	 * @exception AcmeIllegalNameException thrown when the port's name is illegal
	 */
	public void addPort(AcmePort port) throws AcmeIllegalNameException {
		if (port == null)
			return; // there is no port, return gracefully
		if (LanguageConfig.CHECK_FOR_KEYWORDS)
			if (LanguageConfig.isAcmeKeyword(port.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + port.getName() + "\" used for port name.");

		if (port.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a port with null name.");
		}

		else if (eltNamespace.containsName(port.getName()))
			throw new AcmeDuplicateNameException("Attempted to redefine " + "identifier " + port.getName());

		eltNamespace.addName(port.getName());

		port.setParent(this);
		portMap.addItem(port.getName(), port);
	};

	/**
	 * This method returns a port corresponding to port_name
	 * @param port_name the name of the port to be returned
	 * @return the port corresponding to port_name
	 */
	public AcmePort getPort(String port_name) {
		return (AcmePort) portMap.lookupItem(port_name);
	};

	/**
	 * This method removes a port from the system
	 * @param port_name the name of the port to be removed from the system
	 */
	public void removePort(String port_name) {
		AcmePort tmp_port = getPort(port_name);
		if (tmp_port == null)
			return; // there's no such port, return gracefully
		else {
			eltNamespace.removeName(tmp_port.getName());
			removePort(tmp_port);
		};
	};

	/**
	 * This method removes a port from the system
	 * @param port the port to be removed  from the system
	 */
	public void removePort(AcmePort port) {
		if (port == null)
			return;
		detachPort(port);
		eltNamespace.removeName(port.getName());
		port.setParent(null);
		portMap.removeItem(port.getName());
	};

	/**
	 * This method returns an enumeration of all the ports in the system
	 * @return an enumeration of all the ports in the system
	 */
	public AcmeEnumeration getPorts() {
		return portMap.enumerate();
	};

	/**
	 * Detach this port from all other entities in this system.  This method
	 * removes all of the attachments and binding objects that this port
	 * is associated with.
	 * @param port The port to detach from the rest of the system
	 */
	public void detachPort(AcmePort port) {
		AcmeEnumeration attachments = this.getAttachments(port);
		while (attachments.moreItems()) {
			AcmeAttachment curAttachment = (AcmeAttachment) attachments.getNextItem();
			this.removeAttachment(curAttachment);
		};
	};

	///////////////
	// Roles
	///////////////

	/**
	   * This method adds a role to the system
	   * @param role the role to be added to the system
	   * @exception AcmeIllegalNameException thrown when the role's name is illegal
	   */
	public void addRole(AcmeRole role) throws AcmeIllegalNameException {
		if (role == null)
			return; // there is no role, return gracefully
		if (LanguageConfig.CHECK_FOR_KEYWORDS)
			if (LanguageConfig.isAcmeKeyword(role.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + role.getName() + "\" used for role name.");

		if (role.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a role with null name.");
		}

		else if (eltNamespace.containsName(role.getName()))
			throw new AcmeDuplicateNameException("Attempted to redefine " + "identifier " + role.getName());

		eltNamespace.addName(role.getName());

		role.setParent(this);
		roleMap.addItem(role.getName(), role);
	};

	/**
	  * This method returns a role corresponding to role_name
	  * @param role_name the name of the role to be returned
	  * @return the role corresponding to role_name
	  */
	public AcmeRole getRole(String role_name) {
		return (AcmeRole) roleMap.lookupItem(role_name);
	};

	/**
	  * This method removes a role from the system
	  * @param role_name the name of the role to be removed from the system
	  */
	public void removeRole(String role_name) {
		AcmeRole tmp_role = getRole(role_name);
		if (tmp_role == null)
			return; // there's no such role, return gracefully
		else {
			eltNamespace.removeName(tmp_role.getName());
			removeRole(tmp_role);
		};
	};

	/**
	  * This method removes a role from the system
	  * @param role the role to be removed  from the system
	  */
	public void removeRole(AcmeRole role) {
		if (role == null)
			return;
		detachRole(role);
		eltNamespace.removeName(role.getName());
		role.setParent(null);
		roleMap.removeItem(role.getName());
	};

	/**
	 * This method returns an enumeration of all the roles in the system
	 * @return an enumeration of all the roles in the system
	 */
	public AcmeEnumeration getRoles() {
		return roleMap.enumerate();
	};

	/**
	 * Detach this role from all other entities in this system.  This method
	 * removes all of the attachments and binding objects that this role
	 * is associated with.
	 * @param role The role to detach from the rest of the system
	 */
	public void detachRole(AcmeRole role) {
		AcmeEnumeration attachments = this.getAttachments(role);
		while (attachments.moreItems()) {
			AcmeAttachment curAttachment = (AcmeAttachment) attachments.getNextItem();
			this.removeAttachment(curAttachment);
		};
	};

	///////////////
	// Attachments
	///////////////

	/**
	 * This method returns the name of the attachment
	 * @return the name of the attachment
	 */
	protected String genSymAttachmentName(AcmeAttachment a) {
		String attachment_name =
			new String(
				a.getPort().getParent().getName()
					+ a.getPort().getName()
					+ a.getRole().getParent().getName()
					+ a.getRole().getName());
		return attachment_name;
	};

	/**
	   * This method adds an attachment to the system
	   * @param a the attachment to be added to the system
	    */
	public void addAttachment(AcmeAttachment a) {

		// set the name of the attachment for looking up in the table
		// directly to a legal name, then check for duplicates later.
		if (a instanceof AcmeUnresolvedAttachment) {
			a.name = "unresolved-attachment" + a.getCreationOrder();
		}
		else if (a.getName() == null) {
			a.name = genSymAttachmentName(a);
		};

		// a system's attachments are logically a set of relations, so
		// insure set semantics by not allowing double entries.
		AcmeEnumeration e = this.getAttachments();
		AcmeAttachment foundDouble = null;
		while (e.moreItems()) {
			AcmeAttachment curAttachment = (AcmeAttachment) e.getNextItem();
			if (curAttachment.equals(a)) {
				foundDouble = curAttachment;
				break;
			};
		};

		if (foundDouble == null) {
			attachments.addItem(a);
			List pAtts = (List )port2Attachments.get (a.getPort ());
			if (pAtts == null) {
				pAtts = new ArrayList (1);
				port2Attachments.put (a.getPort (), pAtts);
			}
			pAtts.add(a);
			
			List rAtts = (List )role2Attachments.get (a.getRole ());
			if (rAtts == null) {
				rAtts = new ArrayList (1);
				role2Attachments.put (a.getRole (), rAtts);
			}
			rAtts.add (a);
			
		};
	};

	/**
	 * This method adds an attachment to the system using a port and a role
	 * @param p the port of the new attachment
	 * @param r the role of the new attachment
	 */
	public void addAttachment(AcmePort p, AcmeRole r) {
		this.addAttachment(new AcmeAttachment(p, r));
	};

	/**
	 * This method checks whether this system contains attachment a
	 * @param a the attachments searched for in the system
	 * @return whether this system contains attachment a
	 */
	public boolean containsAttachment(AcmeAttachment a) {
		return this.attachments.contains(a);
	};

	/**
	   * This method returns a attachment corresponding to attachmentName
	   * @param attachmentName the name of the attachment to be returned
	   * @return the attachment corresponding to attachmentName
	   */
	public AcmeAttachment getAttachment(String attachmentName) {
		AcmeEnumeration e = this.getAttachments();
		AcmeAttachment returnVal = null;
		while (e.moreItems()) {
			AcmeAttachment curAttachment = (AcmeAttachment) e.getNextItem();
			if (curAttachment.getName() != null) {
				if (curAttachment.getName().equals(attachmentName)) {
					returnVal = curAttachment;
					break;
				};
			};
		};
		return returnVal;
	};

	/**
	 * This  method returns a attachment corresponding to a port and a role
	 * @param p the port of the attachment to be returned
	 * @param r the role of the attachment to be returned
	 * @return the attachment corresponding to p and r
	 */
	public AcmeAttachment getAttachment(AcmePort p, AcmeRole r) {
		AcmeEnumeration e = this.getAttachments();
		AcmeAttachment returnVal = null;
		while (e.moreItems()) {
			AcmeAttachment curAttachment = (AcmeAttachment) e.getNextItem();
			if ((curAttachment.getPort() == p) && (curAttachment.getRole() == r)) {
				returnVal = curAttachment;
				break;
			};
		};
		return returnVal;
	};

	/**
	 * This method returns all the attachments of this system
	 * @return an enumeration of all the attachments of this system
	 */
	public AcmeEnumeration getAttachments() {
		return attachments.enumerate();
	};

	/**
	 * This method returns all the attachments of this system with role r
	 * @param r the role of the attachments to be returned
	 * @return an enumeration of all the attachments of this system with role r
	 */
	public AcmeEnumeration getAttachments(AcmeRole r) {
		
//		AcmeEnumeration all_attachments_enum = getAttachments();
		AcmeEnumeration this_roles_attachments_enum = new AcmeEnumeration();

		List rAtts = (List) role2Attachments.get (r);
		if (rAtts != null) {
			for (Iterator iter = rAtts.iterator(); iter.hasNext();) {
				AcmeAttachment att = (AcmeAttachment) iter.next();
				this_roles_attachments_enum.insertItem (att);
			}
		}
		
		/*for (int i = 0; i < attachments.getSize(); i++) {
			AcmeAttachment a = (AcmeAttachment )attachments.getItemAt(i);
			if (a.getRole () == r) this_roles_attachments_enum.insertItem (a);
		}*/
		
		/*while (all_attachments_enum.moreItems()) {
			AcmeAttachment a = (AcmeAttachment) all_attachments_enum.getNextItem();
			if (a.getRole() == r)
				this_roles_attachments_enum.insertItem(a);
		}*/

		return this_roles_attachments_enum;
	};

	/**
	  * This method returns all the attachments of this system with port p
	  * @param p the port of the attachments to be returned
	  * @return an enumeration of all the attachments of this system with port
	  */
	public AcmeEnumeration getAttachments(AcmePort p) {
//		AcmeEnumeration all_attachments_enum = getAttachments();
		AcmeEnumeration this_ports_attachments_enum = new AcmeEnumeration();

		List pAtts = (List) port2Attachments.get (p);
		if (pAtts != null) {
			for (Iterator iter = pAtts.iterator(); iter.hasNext();) {
				AcmeAttachment att = (AcmeAttachment) iter.next();
				this_ports_attachments_enum.insertItem (att);
			}
		}
//		while (all_attachments_enum.moreItems()) {
//			AcmeAttachment a = (AcmeAttachment) all_attachments_enum.getNextItem();
//			if (a.getPort() == p)
//				this_ports_attachments_enum.insertItem(a);
//		}

		return this_ports_attachments_enum;
	};

	/**
	 * This method removes an attachment from this system
	 * @param attachmentToRemove the attachment to be removed
	 */
	public void removeAttachment(AcmeAttachment attachmentToRemove) {
		this.attachments.removeItem(attachmentToRemove);
		List rAtts = (List )role2Attachments.get (attachmentToRemove.getRole());
		if (rAtts != null) {
			rAtts.remove(attachmentToRemove);
			if (rAtts.isEmpty()) role2Attachments.remove(attachmentToRemove.getRole ());
		}
		
		List pAtts = (List )port2Attachments.get (attachmentToRemove.getPort());
		if (pAtts != null) {
			pAtts.remove(attachmentToRemove);
			if (pAtts.isEmpty()) port2Attachments.remove(attachmentToRemove.getPort ());
		}
	};

	/**
	 * This method returns the size of this attachment
	 * @return the size of this attachment
	 */
	public int attachmentsSize() {
		return this.attachments.getSize();
	};

	/**
	 * This method is a predicate to check whether there is a connection
	 * between two ports in the system. Returns true if they share a common
	 * connector via attachments, else false
	 * @param port1 the first port
	 * @param port2 the second port
	 * @return whether port1 and port2 are connected
	 * 
	 */
	public boolean areConnectedPorts(AcmePort port1, AcmePort port2) {
		boolean areConnected = false;
		if ((port1 == null) || (port2 == null)) {
			return false;
		}

		AcmeComponent comp1 = (AcmeComponent) port1.getParent();
		AcmeComponent comp2 = (AcmeComponent) port2.getParent();
		// basic algorithm:
		//  - for each port p on comp1:
		//      - foreach role r attached to p.
		//          - if parent(r) has another role that is attached to a port
		//              of comp2, areConnected = true;
		AcmePort curPort = port1;
		AcmeRole curRole = null;
		AcmeRole curTargetRole = null;
		AcmeConnector curConn = null;
		AcmeAttachment curAttachment = null;
		AcmeAttachment curTargetAttachment = null;
		AcmeEnumeration portsEnum = comp1.getPorts();
		AcmeEnumeration portAttachmentsEnum = this.getAttachments(curPort);
		while (portAttachmentsEnum.moreItems() && (areConnected == false)) {
			curAttachment = (AcmeAttachment) portAttachmentsEnum.getNextItem();
			curRole = curAttachment.getRole();
			if (curRole != null) {
				curConn = (AcmeConnector) curRole.getParent();
				if (curConn != null) {
					AcmeEnumeration connRoles = curConn.getRoles();
					while (connRoles.moreItems() && (areConnected == false)) {
						curTargetRole = (AcmeRole) connRoles.getNextItem();
						if (curTargetRole != curRole) {
							AcmeEnumeration targetAttachments = this.getAttachments(curTargetRole);
							if (targetAttachments.moreItems()) {
								curTargetAttachment = (AcmeAttachment) targetAttachments.getNextItem();
								if (curTargetAttachment.getPort() != null) {
									if (curTargetAttachment.getPort() == port2) {
										areConnected = true;
									};
								};
							};
						};
					};
				};
			};
		};

		return areConnected;
	}

	/**
	 *  This method is a predicate to check whether two components in this
	 *  system are connected. returns true if they share a common connector
	 *  via attachments, else false.
	 *  RTM 3/30/98 -- This seems _really_ inefficient, see if I can speed it
	 *                 up if performance becomes an issue.
	 * @param comp1 the first component
	 * @param comp2 the second component
	 * @return whether comp1 and comp2 are connected
	 */
	public boolean areConnectedComponents(AcmeComponent comp1, AcmeComponent comp2) {
		boolean areConnected = false;
		if ((comp1 == null) || (comp2 == null)) {
			return false;
		};

		// basic algorithm:
		//  - for each port p on comp1:
		//      - foreach role r attached to p.
		//          - if parent(r) has another role that is attached to a port
		//              of comp2, areConnected = true;
		AcmePort curPort = null;
		AcmeRole curRole = null;
		AcmeRole curTargetRole = null;
		AcmeConnector curConn = null;
		AcmeAttachment curAttachment = null;
		AcmeAttachment curTargetAttachment = null;
		AcmeEnumeration portsEnum = comp1.getPorts();
		while (portsEnum.moreItems() && (areConnected == false)) {
			curPort = (AcmePort) portsEnum.getNextItem();
			AcmeEnumeration portAttachmentsEnum = this.getAttachments(curPort);
			while (portAttachmentsEnum.moreItems() && (areConnected == false)) {
				curAttachment = (AcmeAttachment) portAttachmentsEnum.getNextItem();
				curRole = curAttachment.getRole();
				if (curRole != null) {
					curConn = (AcmeConnector) curRole.getParent();
					if (curConn != null) {
						AcmeEnumeration connRoles = curConn.getRoles();
						while (connRoles.moreItems() && (areConnected == false)) {
							curTargetRole = (AcmeRole) connRoles.getNextItem();
							if (curTargetRole != curRole) {
								AcmeEnumeration targetAttachments = this.getAttachments(curTargetRole);
								curTargetAttachment = (AcmeAttachment) targetAttachments.getNextItem();
								if (curTargetAttachment != null && curTargetAttachment.getPort() != null) {
									if (curTargetAttachment.getPort().getParent() == comp2) {
										areConnected = true;
									};
								};
							};
						};
					};
				};
			};
		};
		return areConnected;
	};

	/////////////////////////
	// DesignAnalyses
	/////////////////////////

	// design analysis storage and manipulation
	/**
	 * This method returns a design analysis corresponding to name
	 * @param name the name of the design analysis to be returned
	 * @return the design analysis corresponding to name
	 */
	public DesignAnalysis getDesignAnalysis(String name) {
		return (DesignAnalysis) designAnalysisMap.lookupItem(name);
	};

	/**
	 * This method returns an enumeration of all design anayses in this system
	 * @return an enumeration of all design anayses in this system
	 */
	public AcmeEnumeration getDesignAnalyses() {
		return designAnalysisMap.enumerate();
	};

	/**
	 * This method adds a design analysis to the system
	 * @param newAnalysis the design analysis to the added to the system
	 * @exception AcmeIllegalNameException design analysis name is illegal
	 */
	public void addDesignAnalysis(DesignAnalysis newAnalysis) throws AcmeIllegalNameException {
		if (newAnalysis == null)
			return; // there is no analysis to add, return gracefully
		if (LanguageConfig.CHECK_FOR_KEYWORDS)
			if (LanguageConfig.isAcmeKeyword(newAnalysis.getName()))
				throw new AcmeKeywordNameException(
					"Illegal keyword \"" + newAnalysis.getName() + "\" used for design analysis name.");

		if (newAnalysis.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a design analysis with null name.");
		}

		else if (eltNamespace.containsName(newAnalysis.getName()))
			throw new AcmeDuplicateNameException("Attempted to redefine " + "identifier " + newAnalysis.getName());

		eltNamespace.addName(newAnalysis.getName());
		newAnalysis.setScope(this);
		designAnalysisMap.addItem(newAnalysis.getName(), newAnalysis);
	};

	/**
	 * This method removes a design analysis from the system
	 * @param name the name of the disign analysis to be removed
	 */
	public void removeDesignAnalysis(String name) {
		DesignAnalysis tmpRule = getDesignAnalysis(name);
		if (tmpRule == null)
			return; // there's no such role, return gracefully
		else {
			eltNamespace.removeName(tmpRule.getName());
			removeDesignAnalysis(tmpRule);
		};
	};

	/**
	 * This method removes a design analysis from the system
	 * @param rule the design analysis to be removed
	 */
	public void removeDesignAnalysis(DesignAnalysis rule) {
		if (rule == null)
			return;
		eltNamespace.removeName(rule.getName());
		designAnalysisMap.removeItem(rule.getName());
	};

	///////////////
	// Exporting
	///////////////

	/**
	 * @return an enumeration of all the system's children
	 */
	protected AcmeEnumeration getChildrenToExport() {
		AcmeEnumeration retEnum = super.getChildrenToExport();
		retEnum.concatenate(this.getComponents());
		retEnum.concatenate(this.getGroups());
		retEnum.concatenate(this.getConnectors());
		retEnum.concatenate(this.getPorts());
		retEnum.concatenate(this.getRoles());
		retEnum.concatenate(this.getAttachments());
		retEnum.concatenate(this.getDesignAnalyses());
		return retEnum;
	}

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "System", this.getChildrenToExport(), ExportLanguage.ACME);
	};

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs when writing
	 * @exception AcmeException for acme exception
	 */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException, AcmeException {
		this.exportEltInstance(s, "System", this.getChildrenToExport(), ExportLanguage.ARMANI);
	};

	/**
	 * Returns the type signature for this system as a string with the
	 *  form "System <name> [ : <Typename>+ ]".  For example:
	 *  "System foo : portT".
	 * @return string containing this system's type signature.
	 **/
	public String getTypeSignature() {
		StringBuffer buf = new StringBuffer("System " + this.getName());
		AcmeEnumeration typeEnum = this.getDeclaredTypes();
		if (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			//        if (eltType instanceof AcmeUnresolvedElementType)
			//            buf.append(" : <unresolved type name: " + eltType.getTypeName() + "> ");
			//        else
			buf.append(" : " + eltType.getTypeName());
		}

		while (typeEnum.moreItems()) {
			AcmeElementType eltType = (AcmeElementType) typeEnum.getNextItem();
			buf.append(", " + eltType.getTypeName());
		}

		return buf.toString();
	};

	/**
	 *  All AcmeSystems satisfy the type ConstantTypes.SYSTEM_BASE_TYPE.
	 *  if the generic supertype method doesn't find the type declared, check
	 *  whether the type is the SYSTEM_BASE_TYPE.
	 *  @param type the type to be compared to SYSTEM_BASE_TYPE
	 *  @return whether  the type is the SYSTEM_BASE_TYPE
	 */
	public boolean declaresType(AcmeElementType type) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType ();
		}
		if (type == ConstantTypes.SYSTEM_BASE_TYPE) {
			return true;
		}
		return super.declaresType(type);
	};

	/**
	 * implement the acmeLib.designRules.Scope interface to map names
	 * to elements or properties.  First try to find the name by looking
	 * for an Element with the requested name, then look in the set of
	 * attachments, then look in the properties and reps by calling
	 * lookupName on this's superclass.
	 *  @param name the name to be mapped to a type or design analyses
	 *  @param bubbleUp boolean, if falso don't check superclass
	 *  @return Object corresponding to "name"
	 */

	public Object lookupName(String name, boolean bubbleUp) {
		// lookup the child object called 'name'.
		// First check for elements, then have AcmeElement check for properties,
		// and representations, then look for a design analysis
		Object retObj = null;

		// check for the special cases requesting all of this system's
		// components or connectors
		if (name != null) {
			if (name.equalsIgnoreCase("components")) {
				retObj = this.getComponents();
			}
			else if (name.equalsIgnoreCase("connectors")) {
				retObj = this.getConnectors();
			}
			else if (name.equalsIgnoreCase("representations")) {
				// BRS: Representations is a bit of a tricky thing. Since each representation can only
				// have one system, I will form an enumeration out of the systems and return these...
				AcmeEnumeration reps = this.getRepresentations();
				AcmeVector sys = new AcmeVector();
				while (reps.moreItems()) {
					sys.addItem(((AcmeRepresentation) reps.getNextItem()).getSystem());
				}

				retObj = sys.enumerate();
			}
		};

		if (retObj == null)
			retObj = this.getComponent(name);
		if (retObj == null)
			retObj = this.getGroup(name);
		if (retObj == null)
			retObj = this.getConnector(name);
		if (retObj == null)
			retObj = this.getPort(name);
		if (retObj == null)
			retObj = this.getRole(name);
		if (retObj == null)
			retObj = this.getAttachment(name);
		if (retObj == null)
			// just check for a property or representation, don't bubble up in
			// the generic element lookupName method.  bubbleUp explicitly
			// after all families have been checked.
			retObj = super.lookupName(name, false);
		if (retObj == null)
			retObj = this.getDesignAnalysis(name);

		if ((retObj == null) && bubbleUp) {
			// first check all of this system's families
			AcmeEnumeration famEnum = this.getDeclaredTypes();
			while (famEnum.moreItems() && (retObj == null)) {
				AcmeElementType eltType = (AcmeElementType) famEnum.getNextItem();
				if (eltType instanceof AcmeFamily)
					retObj = ((AcmeFamily) eltType).lookupName(name, false);
			}

			// if the name still hasn't been found, then bubble up to this
			// system's parent.
			if (retObj == null)
				retObj = ((Scope) this.getScope()).lookupName(name, bubbleUp);
		}

		return retObj;
	};

	/**
	 * AcmeSystem.resolveTypeReferences() relies primarily on it's superclass' implementation
	 * to do type name resolution.  This subclass adds a recursive call to resolveTypeReferences()
	 * for each of this System's substructure elements (components, connectors,
	 * ports and roles).
	 */
	public void resolveTypeReferences() {
		AcmeEnumeration e = null;

		// resolve all named entities in this system's substructure and the
		// names of all of the families that this system declares.
		// (If this is an instance of a system type (family/style), then the
		//  vector of declared types should be empty).
		super.resolveTypeReferences();

		e = this.getComponents();
		while (e.moreItems()) {
			((AcmeComponent) e.getNextItem()).resolveTypeReferences();
		};

		e = this.getGroups();
		while (e.moreItems()) {
			((AcmeGroup) e.getNextItem()).resolveTypeReferences();
		};

		e = this.getConnectors();
		while (e.moreItems()) {
			((AcmeConnector) e.getNextItem()).resolveTypeReferences();
		};

		e = this.getPorts();
		while (e.moreItems()) {
			((AcmePort) e.getNextItem()).resolveTypeReferences();
		};

		e = this.getRoles();
		while (e.moreItems()) {
			((AcmeRole) e.getNextItem()).resolveTypeReferences();
		};

		e = this.getDesignAnalyses();
		while (e.moreItems()) {
			try {
				((DesignAnalysis) e.getNextItem()).resolveTypeReferences();
			}
			catch (AcmeResolutionException e1) {
				//            e1.printStackTrace();
			}
		};

	};

	/**
	 *  Resolve instance references in all of this system's attachments,
	 *  and recursively resolve all of the instance references of this
	 *  system's substructure (components, connectors, ports, and roles)
	 *  and the design analyses declared within the system (or family)
	 *  definition.
	 */
	public void resolveInstanceReferences() {
		AcmeEnumeration e = null;
		super.resolveInstanceReferences();

		e = this.getComponents();
		while (e.moreItems()) {
			((AcmeComponent) e.getNextItem()).resolveInstanceReferences();
		};

		e = this.getGroups();
		while (e.moreItems()) {
			System.out.println("resolving group instances");
			((AcmeGroup) e.getNextItem()).resolveInstanceReferences();
		};

		e = this.getConnectors();
		while (e.moreItems()) {
			((AcmeConnector) e.getNextItem()).resolveInstanceReferences();
		};

		e = this.getPorts();
		while (e.moreItems()) {
			((AcmePort) e.getNextItem()).resolveInstanceReferences();
		};

		e = this.getRoles();
		while (e.moreItems()) {
			((AcmeRole) e.getNextItem()).resolveInstanceReferences();
		};

		e = this.getAttachments();
		AcmeAttachment curAttachment = null;
		while (e.moreItems()) {
			curAttachment = (AcmeAttachment) e.getNextItem();
			if (curAttachment instanceof AcmeUnresolvedAttachment)
				 ((AcmeUnresolvedAttachment) curAttachment).resolveInstanceReferences();
		};

		e = this.getDesignAnalyses();
		while (e.moreItems()) {
			try {
				((DesignAnalysis) e.getNextItem()).resolveInstanceReferences();
			}
			catch (AcmeResolutionException e1) {
				//            e1.printStackTrace();
			}
		};

	};

	/**
	 * instantiateElements() walks over the vector of instantiatedTypes of this
	 * element and adds the minimal structure declared in the type description
	 * to this element.  It delegates much of the work to the AcmeTypeManager
	 * class, so that we can change type instantiation policies relatively
	 * centrally (ie by changing the AcmeTypeManager methods) for all elements
	 * at once.  This method recursively calls the same method on all of the
	 * substructure of this element instance.
	 */
	public void instantiateElements() {

		// unify this element with each of the types it instantiates.
		for (int i = 0; i < this.instantiatedTypes.getSize(); i++) {
			// check that we have a prototype system or family to instantiate.
			// If we don't then announce an error and move to the next type.
			if (this.instantiatedTypes.getItemAt(i) instanceof AcmeFamily) {
				AcmeTypeManager.unifySystem((AcmeFamily) this.instantiatedTypes.getItemAt(i), this);

			}
			else if (this.instantiatedTypes.getItemAt(i) instanceof AcmeElementProtoType) {
				AcmeTypeManager.unifyElementWithPrototype(
					(AcmeElementProtoType) this.instantiatedTypes.getItemAt(i),
					this);

			}
			else if (this.instantiatedTypes.getItemAt(i) instanceof AcmeType) {
				String errorMsg = null;
				if (this.instantiatedTypes.getItemAt(i) instanceof AcmeUnresolvedElementType) {
					errorMsg =
						"Instantiation Error.  Unable to instantiate "
							+ " unresolved element type: "
							+ ((AcmeUnresolvedElementType) this.instantiatedTypes.getItemAt(i)).getTypeName()
							+ " in system: "
							+ this.getName();
				}
				else {
					errorMsg =
						"Instantiation Error.  Type: "
							+ ((AcmeType) this.instantiatedTypes.getItemAt(i)).getTypeName()
							+ " is not a family.  Therefore it can not be "
							+ "used to instantiate system: "
							+ this.getName();
				};

				TypeInstantiationError instError =
					new TypeInstantiationError(this, errorMsg, (AcmeType) this.instantiatedTypes.getItemAt(i));
				AcmeErrorManager.errorDetected(instError);

			}
			else {
				String errorMsg =
					"Instantiation Error.  Attempted to instantiate a "
						+ "completely unrecognizable family in system: "
						+ this.getName();
				TypeInstantiationError instError = new TypeInstantiationError(this, errorMsg);
				AcmeErrorManager.errorDetected(instError);
			};
		}

		// then repeat the instantiation process for all of the substructure
		// of this system (components, connectors, ports, roles, reps, and
		// attachments).
		AcmeEnumeration e;
		e = this.getComponents();
		while (e.moreItems()) {
			((AcmeComponent) e.getNextItem()).instantiateElements();
		};

		e = this.getGroups();
		while (e.moreItems()) {
			((AcmeGroup) e.getNextItem()).instantiateElements();
		};

		e = this.getConnectors();
		while (e.moreItems()) {
			((AcmeConnector) e.getNextItem()).instantiateElements();
		};

		e = this.getPorts();
		while (e.moreItems()) {
			((AcmePort) e.getNextItem()).instantiateElements();
		};

		e = this.getRoles();
		while (e.moreItems()) {
			((AcmeRole) e.getNextItem()).instantiateElements();
		};

		// finally, call AcmeElement to have it instantiate representations and any
		// other standard element structure.
		super.instantiateElements();
	};

	/**
	 *  This method determines whether this system's properties, representations,
	 *  design rules, and substructure declarations are all type-consistent,
	 *  and that all declared supertypes are AcmeFamily's also.  This method
	 *  delegates all of the actual checking to the AcmeTypeManager class.
	 *  @return whether all types do typecheck
	 */
	public boolean typecheck() {
		boolean typechecks = AcmeTypeManager.typecheckSystem(this);
		if (typechecks) {
			// things are hunky-dory, return true.
			return true;
		}
		// system doesn't typecheck.  There should be an error at the
		// top of the AcmeTypeManager's typecheckingErrorStack.  If there
		// is such an error then announce it to the AcmeErrorManager and
		// return false.
		AcmeError eltTypeError = AcmeTypeManager.typecheckingErrorStack.popCausalError();
		if (eltTypeError != null) {
			AcmeErrorManager.errorDetected(eltTypeError);
			return false;
		}
		// no error was reported.  something went wrong?
		// announce a generic error.
		AcmeErrorManager.errorDetected(
			new ElementTypecheckError(this, "System " + this.getName() + " does not typecheck.", this));
		return false;
	};

	/**
	 * retrieves all of the types that are available/visible to this element,
	 * based on the Acme/Armani type scoping rules.  Currently, that means
	 * that an element looks for the system in which it is declared, retrieves
	 * the union of all of the types declared in that system's families (and
	 * their super-types) and the design in which it is declared and all
	 * of that design's imported designs.
	 * This method should be overridden by any subclasses that allow types
	 * to be declared in their scopes (ie AcmeDesigns and AcmeFamilies).
	 */
	public AcmeEnumeration getAvailableTypes() {
		AcmeEnumeration allAvailableTypes = new AcmeEnumeration();
		AcmeEnumeration fams = this.getDeclaredTypes();
		while (fams.moreItems()) {
			AcmeFamily curFam = (AcmeFamily) fams.getNextItem();
			allAvailableTypes.concatenate(curFam.getAvailableTypes());
		};

		if (this.getParentDesign() != null) {
			allAvailableTypes.concatenate(this.getParentDesign().getAvailableTypes());
		};

		return allAvailableTypes;
	};

	/////////////////////////////////////////////////////////////////////////////
	// Operations  (operations are represented as properties of type operationT)
	/////////////////////////////////////////////////////////////////////////////

	/**
	 * @return an enumeration containing all of the operations declared in this
	 *         System or style.  These operations can be represented as properties
	 *         of type operationT or as property value objects that satisfy type
	 *         operationsT.
	 */
	public AcmeEnumeration getOperations() {
		AcmeEnumeration ops = new AcmeEnumeration();
		AcmeDesign builtInsDesign = null;
		if (this.getParentDesign() != null) {
			builtInsDesign = this.getParentDesign().getBuiltInsDesign();
		}
		else if (this.getParent() instanceof AcmeDesign) {
			builtInsDesign = ((AcmeDesign) this.getParent()).getBuiltInsDesign();
		};

		if (builtInsDesign == null) {
			return ops; // we don't have an operationsType to check against...
		};

		Object opTypeObj = builtInsDesign.lookupName("operationT", false);
		Object opSetTypeObj = builtInsDesign.lookupName("operationsT", false);

		AcmePropertyType opPropType, opSetPropType;

		if (opTypeObj instanceof AcmePropertyType) {
			opPropType = (AcmePropertyType) opTypeObj;
		}
		else {
			return ops; // again, we don't have an operationsType to check against...
		};

		if (opSetTypeObj instanceof AcmePropertyType) {
			opSetPropType = (AcmePropertyType) opSetTypeObj;
		}
		else {
			return ops; // again, we don't have an operationSetType to check against...
		};

		// walk through all of the properties, looking for the operations-typed
		// properties.  add them to ops as they are found.
		AcmeEnumeration props = this.getProperties();

		while (props.moreItems()) {
			AcmeProperty curProp = (AcmeProperty) props.getNextItem();
			if (curProp.getType() == opPropType) {
				ops.insertItem(curProp);
			}
			else if (curProp.getType() == opSetPropType) {
				AcmeSet opSet = (AcmeSet) curProp.getValue();
				AcmeEnumeration setVals = opSet.getValues();
				while (setVals.moreItems()) {
					ops.insertItem(setVals.getNextItem());
				};
			};
		};

		return ops;
	};

	/**
	 * implement the acmeLib.Editable interface.
	 * It also creates a workshop for editing this element
	 */
	public void edit() {
		//    if (this.editor == null) {
		//        this.editor = new ElementWorkshopFrame();
		//        editor.editSystemDetails(this);
		//    };
		//    editor.setVisible(true);
		//    editor.toFront();
	};

	public void view() {
		System.out.println("Attempted to view: " + this.getName());
		System.out.println("  but AcmeElement.view() is not yet implemented.");
	};

};
