
package acmeLib.elements;

import acmeLib.core.io.AcmeOutputStream;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.InstanceRefResolutionError;
import acmeLib.typeManagement.AcmeTypeManager;


/**
 *  Class: AcmeUnresolvedAttachment
<BR>
 *  Purpose: Class that represents an attachment within an Acme system
<BR>
 *          whose attached port and role are named but have not yet been
<BR>
 *          resolved to point to an actual pair of AcmePort, AcmeRole objects.
<BR>
 *          This is a temporary class to hold an attachment during parsing
<BR>
 *          but before the names of the system are resolved.
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *   1/15/98   RTM      Created initial file.
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeUnresolvedAttachment extends AcmeAttachment {

  //
  // Protected members
  //
  protected String outerElt1, innerElt1;
  protected String outerElt2, innerElt2;

  //
  // Public members
  //
  /**
   * This is a constructor for this class
   * @param outerElt1 name of this.outerElt1
   * @param outerElt2 name of this.outerElt2
   * @param innerElt1 name of this.innerElt1
   * @param innerElt2 name of this.innerElt2
   */
  public AcmeUnresolvedAttachment (String outerElt1, String innerElt1,
                                   String outerElt2, String innerElt2)
  {
     super(null, null);
     this.outerElt1 = outerElt1;
     this.innerElt1 = innerElt1;
     this.outerElt2 = outerElt2;
     this.innerElt2 = innerElt2;
  };


  /**
   * This is a  copy constructor to create a new unresolved attachment from either
   * a resolved attachment or an unresolved attachment.
   * @param oldAttachment the attachment used to create a new copy
   */
  public AcmeUnresolvedAttachment (AcmeAttachment oldAttachment) {
    super(null, null);
    if (oldAttachment == null)
        return;

    if (oldAttachment instanceof AcmeUnresolvedAttachment) {
        this.outerElt1 = ((AcmeUnresolvedAttachment) oldAttachment).outerElt1;
        this.outerElt2 = ((AcmeUnresolvedAttachment) oldAttachment).outerElt2;
        this.innerElt1 = ((AcmeUnresolvedAttachment) oldAttachment).innerElt1;
        this.innerElt2 = ((AcmeUnresolvedAttachment) oldAttachment).innerElt2;
    } else {
        AcmePort oldPort = oldAttachment.getPort();
        if (oldPort != null) {
            if (oldPort.getParent() != null)
                this.outerElt1 = oldPort.getParent().getName();
            this.innerElt1 = oldPort.getName();
        };

        AcmeRole oldRole = oldAttachment.getRole();
        if (oldRole != null) {
            if (oldRole.getParent() != null)
                this.outerElt2 = oldRole.getParent().getName();
            this.innerElt2 = oldRole.getName();
        }
    }

    this.setParent(oldAttachment.getParent());
    this.setParentDesign(oldAttachment.getParentDesign());
    oldAttachment.copyProperties(this);
  }


  /**
   * This method creates a copy of this attachment
   * @return a copy of this attachment
   */
  public AcmeElement copy() {
    AcmeUnresolvedAttachment new_attachment = new AcmeUnresolvedAttachment(this);
    this.copyProperties(new_attachment);
    return new_attachment;
  };


  /**
   * This method checks whether two attachments are equal
   * @param test attachment to be tested
   * @return whether "this" attachments and test are equal
   */
  public boolean equals(Object test) {
    AcmeUnresolvedAttachment testAttach;
    boolean areEqual = false;

    if (!(test instanceof AcmeUnresolvedAttachment)) {
        return false;
    } else {
        testAttach = (AcmeUnresolvedAttachment) test;
    };

    if (outerElt1 != null) {
        if (outerElt1.equals(testAttach.outerElt1)) {
            if (innerElt1 != null) {
                if (innerElt1.equals(testAttach.innerElt1)) {
                    if (outerElt2 != null) {
                        if (outerElt2.equals(testAttach.outerElt2)) {
                            if (innerElt2 != null) {
                                if (innerElt2.equals(testAttach.innerElt2)) {
                                    areEqual = true;
                                };
                            };
                        };
                    };
                };
            };
        }
        else if (outerElt1.equals(testAttach.outerElt2)) {
            // the pair (x,y) == the pair (y,x), check for switched places
            if (innerElt1 != null) {
                if (innerElt1.equals(testAttach.innerElt2)) {
                    if (outerElt2 != null) {
                        if (outerElt2.equals(testAttach.outerElt1)) {
                            if (innerElt2 != null) {
                                if (innerElt2.equals(testAttach.innerElt1)) {
                                    areEqual = true;
                                };
                            };
                        };
                    };
                };
            };
        };
    };

    return areEqual;
  };


  /**
   * This method converts an attachment to a string
   * @return the string representation of an attachment
   */
  public String toString() {
      String ret = "Attachment ";
      if (outerElt1 != null) ret += outerElt1 + ".";
      ret += innerElt1 + " to ";
      if (outerElt2 != null) ret += outerElt2 + ".";
      ret += innerElt2 + ";";
      return ret;
//    return outerElt1 + "." + innerElt1 + " to "
//               + outerElt2 + "." + innerElt2
//               + " { Property status : string = \"unresolved\" }; ";
  };

  /**
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs while writing
   */
  public void exportAcme (AcmeOutputStream s) throws java.io.IOException {
  	s.writeln (this.toString());
  };


  /**
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs while writing
   */
public void exportArmani(AcmeOutputStream s) throws java.io.IOException {
    s.beginRegion(this);
    s.write(this.toString());
    s.endRegion(this);
    s.writeln();
};

  /**
   * This method looks up the names of the port and the role, creates a new AcmeAttachment
   * and adds it to the parent system.
   */
  public void resolveInstanceReferences() {
      AcmeSystem parent_sys = (AcmeSystem) this.getParent();

      boolean insideFamilyDef = (parent_sys instanceof AcmeElementType);
      AcmeAttachment newAttachment = null;
      AcmeComponent comp = null;
      AcmeConnector conn = null;
      AcmePort resolvedPort = null;
      AcmeRole resolvedRole = null;
      boolean unableToResolvePort = false;
      boolean unableToResolveRole = false;
      String unresolvedPortName = null;
      String unresolvedRoleName = null;

      // check if we have a component or a connector as the first element.
      comp = parent_sys.getComponent(outerElt1);
      if (comp != null) {
          // the first element is a component, lookup portname in compname
          resolvedPort = comp.getPort(innerElt1);
          if (resolvedPort == null) {
            // flag an error.
            unableToResolvePort = true;
            unresolvedPortName = outerElt1 + "." + innerElt1;
          };

          // lookup connname and rolename in connname
          conn = parent_sys.getConnector(outerElt2);
          if (conn != null) {
              resolvedRole = conn.getRole(innerElt2);
              if (resolvedRole == null) {
                // flag an error.
                unableToResolveRole = true;
                unresolvedRoleName = outerElt2 + "." + innerElt2;
              };
          } else {
            // else unable to look up conn, and thus the role. flag an error.
            unableToResolveRole = true;
            unresolvedRoleName = outerElt2 + "." + innerElt2;
          };
      } else {
          // else the first element should be a connector,
          // lookup connname and rolename in connname
          conn = parent_sys.getConnector(outerElt1);
          if (conn != null) {
              resolvedRole = conn.getRole(innerElt1);
              if (resolvedRole == null) {
                // flag an error.
                unableToResolveRole = true;
                unresolvedRoleName = outerElt1 + "." + innerElt1;
              };
          } else {
            // else unable to look up conn, and thus the role. flag an error.
            unableToResolveRole = true;
            unresolvedRoleName = outerElt1 + "." + innerElt1;
          };

          // second set of tokens should refer to component/port pair
          comp = parent_sys.getComponent(outerElt2);
          if (comp != null) {
              resolvedPort = comp.getPort(innerElt2);
              if (resolvedPort == null) {
                // flag an error.
                unableToResolvePort = true;
                unresolvedPortName = outerElt2 + "." + innerElt2;
              };
          } else {
            // else unable to look up conn, and thus the role. flag an error.
            unableToResolvePort = true;
            unresolvedPortName = outerElt2 + "." + innerElt2;
          };
      };


      // if we haven't resolved an appropriate port/role pair then
      // look in family supertypes if this is in a family, else
      // flag an error and continue
      if (unableToResolvePort && insideFamilyDef) {
          Object foundObj = parent_sys.lookupName(outerElt1, true);
          if ((foundObj instanceof AcmeComponent)
                && !(foundObj instanceof AcmeElementType))
          {
                comp = (AcmeComponent) foundObj;
                resolvedPort = comp.getPort(innerElt1);
                if (resolvedPort != null) {
                    // unflag error.
                    unableToResolvePort = false;
                };
          } else {
                foundObj = parent_sys.lookupName(outerElt2, true);
                if ((foundObj instanceof AcmeComponent)
                      && !(foundObj instanceof AcmeElementType))
                {
                    comp = (AcmeComponent) foundObj;
                    resolvedPort = comp.getPort(innerElt1);
                    if (resolvedPort != null) {
                        // unflag error.
                        unableToResolvePort = false;
                    };
                };
          };
      };


      if (unableToResolveRole && insideFamilyDef) {
          Object foundObj = parent_sys.lookupName(outerElt1, true);
          if ((foundObj instanceof AcmeConnector)
                && !(foundObj instanceof AcmeElementType))
          {
                conn = (AcmeConnector) foundObj;
                resolvedRole = conn.getRole(innerElt1);
                if (resolvedRole != null) {
                    // unflag error.
                    unableToResolveRole = false;
                };
          } else {
                foundObj = parent_sys.lookupName(outerElt2, true);
                if ((foundObj instanceof AcmeConnector)
                      && !(foundObj instanceof AcmeElementType))
                {
                    conn = (AcmeConnector) foundObj;
                    resolvedRole = conn.getRole(innerElt2);
                    if (resolvedRole != null) {
                        // unflag error.
                        unableToResolveRole = false;
                    };
                };
          };
      };


      // if we still haven't resolved a port or a role, announce an error.
      if (unableToResolveRole || unableToResolvePort) {
        if (unableToResolveRole) {
            InstanceRefResolutionError roleError =
                new InstanceRefResolutionError(this.getParent(),
                        "unable to resolve role reference " + unresolvedRoleName
                        + " in attachment " + this.toString(),
                        unresolvedRoleName);
            roleError.setUnresolvedRef(this);
            AcmeErrorManager.errorDetected(roleError);
        };

        if (unableToResolvePort) {
            InstanceRefResolutionError portError =
                new InstanceRefResolutionError(this.getParent(),
                        "unable to resolve port reference " + unresolvedPortName
                        + " in attachment " + this.toString(),
                        unresolvedPortName);
            portError.setUnresolvedRef(this);
            AcmeErrorManager.errorDetected(portError);
        };

      };


      if ((resolvedPort != null) && (resolvedRole != null))
      {
          // we have a resolved port and role pair.  ask the parent
          // if it already has an attachment with this port-role pair.
          // If so, simply attempt to unify the properties of this attachment
          // with the properties of the existing attachment.  Normally,
          // this unification would be done in the instantiate() method,
          // but instantiate() is called before resolveInstanceReferences(),
          // and the instance references need to be resolved after
          // instantiation...  so this requests that the TypeManager merge
          // the properties if it is an "extended" attachment specification.

          newAttachment = parent_sys.getAttachment(resolvedPort, resolvedRole);
          if (newAttachment != null) {
              // then parent_sys has an existing attachment, merge the
              // new attachment's properties with the existing attachment.
              AcmeTypeManager.unifyProperties(this, newAttachment);
          } else {
              // else we can safely create a new attachment, add it to the
              // parent system, and copy the properties of this unresolved
              // attachment to the new attachment.
              newAttachment = new AcmeAttachment(resolvedPort, resolvedRole);
              newAttachment.setParent(parent_sys);
              newAttachment.setParentDesign(parent_sys.getParentDesign());
              this.copyProperties(newAttachment);
              // set the creation order of the new attachment to the creation
              // order of the unresolved entity, so that they will be exported
              // in the proper order.
              newAttachment.setCreationOrder(this.getCreationOrder());
              parent_sys.addAttachment(newAttachment);
          };

          // clean up, removing this unresolved attachment from the parent.
          parent_sys.removeAttachment(this);
      };
   }

};

