/**
 * Class: AcmeError<br>
 * Purpose: AcmeError is the base class for all of the errors that can be<br>
 *          discovered in the process of parsing an Acme/Armani description.<br>
 *          It stores a vector of "causes" indicating other errors that<br>
 *          propogated to create this error.  It also includes a reference<br>
 *          to the source of the error and an error message.<br>
 *<br>
 *          Note:AcmeError is _completely_ independent of the java.lang.Error<br>
 *               class tree.  These are indications that an error was found<br>
 *               in an Acme/Armani spec, it's not a throwable.<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    2/11/98  RTM       created initial class<br>
 *    12/8/98  RTM       added implementation of methods for integrating with
 *                        AcmeStudio error handling.
 * @author Bob Monroe
 */

/** Magic registration chant to register this class with the Windows repository

    %javareg /register /class:acmeLib.errorManagement.AcmeError
             /clsid:{A8C8B820-8F0C-11d2-BC3D-00609737B51A}
             /progid:acmeLib.errorManagement.AcmeError.java.1
             /typelib:AcmeError.tlb
 **/




package acmeLib.errorManagement;

import java.util.HashSet;
import java.util.Set;

import acmeLib.*;
import acmeLib.core.container.*;
import acmeLib.designRules.*;
import acmeLib.elements.*;
import acmeLib.properties.*;
import acmeLib.representations.*;

public class AcmeError extends AcmeObject {

  //
  // protected members
  //

  /**
   * the source of this error.  what constitutes a source can be
   * fairly liberally interpreted by subclasses.
   */
    protected Object source = null;


  /**
   * caches the fully qualified name of the source, if appropriate.
   **/
    protected String qualifiedSourceName = null;

  /**
   * a vector of other errors that "caused" this error.  This is useful
   * primarily for tracking how errors propogate.
   */
    protected AcmeVector causedBy = null;

  /**
   * an error message
   */
    protected String errorMessage = null;

  /**
   * a flag to indicate whether this is a warning (1) or an error (2).
   *  Defaults to ERROR.
   **/
    protected int msgType = AcmeError.ERROR;

  /**
   * a string that contains a comma-separated list of the fully qualified names
   * of entities that may also be relevant to the message.
   **/
    protected String seeAlsoStr = "";  // default to empty string rather than null

  /**
   * how many spaces should each indenting add
   */
    protected static int standardIndent = 4;

    //
    // public members
    //

  /**
   * public constant to represent a warning message rather than an error message
   **/
   public static final int WARNING = 1;

  /**
   * public constant to represent an error message rather than a warning message
   **/
   public static final int ERROR = 2;

  /**
   * default constructor.  Assume this is an error and not a heuristic.
   */
    public AcmeError() {
        this.source = null;
        this.causedBy = new AcmeVector(2,2);
        this.errorMessage = "Unspecified AcmeError";
    };

  /**
   * This is the constructor for this class
   * @param errorSource Source of error
   * @param msg Error message
   */
    public AcmeError(Object errorSource, String msg) {
        this.source = errorSource;
        this.errorMessage = msg;
        this.causedBy = new AcmeVector(2,2);
    };


    // accessors/settors
  /**
   * Get the error message
   * @return The error message
   */
    public String getErrorMessage() { return this.errorMessage; };
  /**
   * Set the error message
   * @param newMsg The error message
   */
    public void setErrorMessage(String newMsg) { this.errorMessage = newMsg; };

  /**
   * Get the source
   * @return The source object
   */
    public Object getSource() { return this.source; };
  /**
   * Set the source
   * @param newSource The source object
   */
    public void setSource(Object newSource) { this.source = newSource; };

    // manipulate the causedBy vector
  /**
   * Get the causedBy vector
   * @return The AcmeVector
   */
    public AcmeVector getCausedByVector() { return this.causedBy; };
  /**
   * Set the causedBy vector
   * @param newCauses The AcmeVector
   */
    public void setCausedByVector(AcmeVector newCauses) { this.causedBy = newCauses; };

  /**
   * Add to the causedBy vector
   * @param newCause The item added to the vector
   */
    public void addCause(AcmeError newCause) { this.causedBy.addItem(newCause); };
  /**
   * Get causes
   * @return The enumeration of the causedBy vector
   */
    public AcmeEnumeration getCauses() { return this.causedBy.enumerate(); };

  /**
   * output methods.
   * @return output string
   */
    public String toString() {
        Set visitedErrors = new HashSet ();
        visitedErrors.add(this);
        return this.toString(0, visitedErrors);
    }

  /**
   * helper method that indents the requested number of spaces before
   * converting this error to a string.
   * @param indentSpaces spaces to be indented
   * @return output string
   */
    protected String toString(int indentSpaces, Set visitedErrors) {
        // write out the error message indented properly.
        StringBuffer msgBuf = new StringBuffer();
        for (int i = 0; i < indentSpaces; i++) {
            msgBuf.append(" ");
        };
        if (indentSpaces > 0) {
            // then this is a causing error
            msgBuf.append("caused by --> ");
        };
        msgBuf.append(this.getErrorMessage());
        msgBuf.append("\n");

        // append a new line indented further for each
        for (int i = 0; i < causedBy.getSize(); i++) {
            AcmeError nextCause = (AcmeError) causedBy.getItemAt(i);
            if (!visitedErrors.contains (nextCause)) {
                visitedErrors.add (nextCause);
                msgBuf.append(nextCause.toString(indentSpaces + standardIndent, visitedErrors));
            }
        };

        return msgBuf.toString();
    };



    ///////////////////////////////////////////////////////////////////
    //
    // Implement the methods that the AcmeStudio uses to interact with
    // error messages via qualified names and messages rather than
    // direct object references.
    //
    ///////////////////////////////////////////////////////////////////

    /**
     * @return an integer value representing the type of this message.
     *         NOTE == 1, WARNING == 2, ERROR == 3.
     **/
    public int getMsgType() { return this.msgType; }

    /**
     * @param newMsgType an integer value representing the type of this
     *         message.  NOTE == 1, WARNING == 2, ERROR == 3.
     **/
    public void setMsgType(int newMsgType)  { this.msgType = newMsgType; };



    /**
     * @return the fully qualified name of the source of this message.
     **/
    public String getMsgSource() {
        if (this.qualifiedSourceName == null && this.source != null) {
            this.qualifiedSourceName = findQualifiedName(this.source);
        };
        return this.qualifiedSourceName;
    };


    /**
     * set the name of the MsgSource
     * @param qualifiedSourceName the fully qualified name of the source
     *        of this message.
     **/
    public void setMsgSource(String newQualifiedSourceName) {
        this.qualifiedSourceName = newQualifiedSourceName;
    };




    /**
     * @return the text of this message
     **/
    public String getMsgText() { return this.getErrorMessage(); };

    /**
     * @param newMsgText the text to which this message will be set
     **/
    public void setMsgText(String newMsgText) {
        this.setErrorMessage(newMsgText);
    };



    /**
     * @return a comma-separated list of qualified names.  These names refer
     *         to other entities of interest in the message.  The meaning of
     *         these names is not defined in the toplevel interface.  Classes
     *         that implement this interface can apply meaning to them as
     *         appropriate.
     **/
    public String seeAlso() { return this.seeAlsoStr; };


    /**
     * @param namesToSeeAlso a comma-separated list of qualified names.  that
     *         refer to other entities of interest in the message.  The meaning
     *         of these names is not defined in the toplevel interface.
     *         Classes that implement this interface can apply meaning to
     *         them as appropriate.
     **/
    public void setSeeAlso(String namesToSeeAlso) {
        this.seeAlsoStr = namesToSeeAlso;
    };



    /**
     * @return the number of children AcmeErrors of this node.
     **/
    public int getNumChildrenNodes() { return this.causedBy.getSize(); };

    /**
     * @return the i'th child of this error message
     **/
    public AcmeError getChildNode(int childNum) {
        if (this.causedBy.getItemAt(childNum) instanceof AcmeError) {
            return (AcmeError) this.causedBy.getItemAt(childNum);
        }
            return null;
    };

    /**
     * Add a child AcmeError as the last child of this node.
     * @param newChild child node to add.
     **/
    public void addChildNode(AcmeError newChild) {
        if ((newChild instanceof AcmeError)
            && (newChild instanceof AcmeObject))
        {
            this.causedBy.addItem((AcmeObject) newChild);
        } else {
            // quietly do nothing rather than throw an exception?
            this.causedBy.addItem(new AcmeError(this.source, "Internal error?"));
        };
    };


    /**
     * removes the i'th child node from this message
     * @param i index of child node to remove
     **/
    public void removeChildNode(int i) {
        this.causedBy.removeItemAt(i);
    };


    /**
     * helper function that attempts to retrieve the fully qualified name
     * of any Object.
     * @param nameOf Object that you want to retrieve the name of.
     **/
    protected String findQualifiedName(Object obj) {
        String qualName = null;
        if (obj instanceof AcmeElement) {
            return ((AcmeElement) obj).getQualifiedName();
        } else if (obj instanceof AcmeProperty) {
            return ((AcmeProperty) obj).getQualifiedName();
        } else if (obj instanceof AcmePropertyType) {
            return ((AcmePropertyType) obj).getQualifiedName();
        } else if (obj instanceof DesignAnalysis) {
            return ((DesignAnalysis) obj).getQualifiedName();
        } else if (obj instanceof AcmeRepresentation) {
            return ((AcmeRepresentation) obj).getQualifiedName();
        }

        return "Error.  Unable to lookup qualified name of " + obj.toString();
    };



};