package acmeLib.errorManagement;

import java.util.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.designRules.DesignAnalysis;
import acmeLib.designRules.DesignRule;
import acmeLib.designRules.Scope;
import acmeLib.elements.AcmeDesign;
import acmeLib.elements.AcmeElement;
import acmeLib.properties.AcmeProperty;
import acmeLib.typeManagement.AcmeTypeManager;

/**
 * Class: AcmeErrorManager<br>
 * Purpose: AcmeErrorManager keeps track of the outstanding errors that<br>
 *          the AcmeLib (or Armani) discover while processing Acme/Armani<br>
 *          specifications. Includes query methods for listing outstanding<br>
 *          errors (and warnings)<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    2/11/98   RTM      created initial class<br>
 *<br>
 * @author Bob Monroe
 */

public class AcmeErrorManager extends AcmeObject {

  /**
   * static storage for detected errors.
   * RTM 2/11/98 -- This is temporary, experimental storage.
   */
    protected static AcmeVector errorVault = new AcmeVector();
    
    /**
     * Keeps a map of the threads that have generated errors
     */
    protected static Map threadMap = new HashMap ();

  /**
   * flag that indicates whether errors should be logged.
   * true = log errors, false = don't log errors.
   **/
    protected static boolean logErrors = true;

  /**
   * @return true if error logging is currently turned on, else return false
   **/
    public static boolean getLogErrors() { return logErrors; };

  /**
   * turn error logging on (true) or off (false).
   * @param doLogErrors true to turn error logging on, false turns error logging off.
   **/
     public static void setLogErrors(boolean doErrorLogging) {
        logErrors = doErrorLogging;
     };

  /**
   * trivial constructor.  everything of interest in this class is static.
   */
    public AcmeErrorManager() {};

  /**
   * RTM 2/11/98 -- this is a stub method that currently just dumps
   *                the discovered error to stderr.  Eventually, this
   *                will store the error in some sort of data structure
   *                and possibly announce an event to the rest of the world.
   * @param detectedError The AcmeError detected
   */
    public static void errorDetected(AcmeError detectedError) {
        if (detectedError == null) {
            return;
        } else if (logErrors) {
            connectThread ();
            AcmeVector errorVault = (AcmeVector )threadMap.get (Thread.currentThread ());
            errorVault.addItem(detectedError);
            announceErrorOccurred (detectedError);
        };
    };


    protected static void connectThread () {
        Thread currentThread = Thread.currentThread();
        if (!threadMap.containsKey (currentThread)) {
            AcmeVector errorVault = new AcmeVector ();
            threadMap.put (currentThread, errorVault);
        }   
    }

    protected static AcmeVector getErrorVault () {
        connectThread ();
        return (AcmeVector )threadMap.get (Thread.currentThread ());
    }  
      

  /**
   * This is a simple implementation of a method that clears an error out
   * after it has (somehow) been shown to be repaired.  The AcmeErrorManager
   * class performs no checking that it has actually been repaired, it just
   * searches for the actual error node in the errorVault and removes
   * it if it finds it.
   * @param repairedError The repaired AcmeError
   */
    public static void errorRepaired(AcmeError repairedError) {
        // RTM 2/12/98 -- This is a very simple preliminary implementation,
        //                I'm not sure how this will eventually work in the
        //                final version, or how (of if) it will deal with errors
        //                that might be embedded in the causal error tree.
        synchronized (errorVault) {
            getErrorVault ().removeItem(repairedError);
        }
    };


    public  static void errorRepaired (AcmeObject obj, AcmeError error) {
        if (logErrors)
            announceErrorFixed (obj, error);
    }

  /**
   * The enumeration of the errors
   * @return The AcmeEnumeration of errors
   */
    public static AcmeEnumeration getErrors() {
        AcmeEnumeration enum = null;
        synchronized (errorVault) {
            enum = getErrorVault ().enumerate();
        }
        return enum;
    };

  /**
   * Returns the number of errors stored in the error vault.
   **/
   public static int getNumErrorsReported() {
       int size = 0;
       synchronized (errorVault) {
            size = getErrorVault ().getSize();
       }
       return size;
   };
   
  /**
   * Remove all errors
   */
    public static void clearAllErrors() {
        synchronized (errorVault) {
            getErrorVault ().removeAllItems();
        }
    };

  /**
   * Writes all errors in the ErrorVault to the PrintStream passed in
   **/
   public static void printAllErrors(java.io.PrintStream ps) {
        AcmeEnumeration errs = null;
        synchronized (errorVault) {
            errs = getErrorVault ().enumerate();
        }
        while(errs.moreItems()) {
            Object nextObj = errs.getNextItem();
            if (nextObj instanceof AcmeError) {
                ps.println("Error: " + ((AcmeError) nextObj).toString());
            } else {
                ps.println("Encountered unrecognized error.  <internal error>");
            };
        };
   };


    private static Set errorListeners = new HashSet ();

    public static void disconnectThread () {
        threadMap.remove (Thread.currentThread ());  
        AcmeTypeManager.typecheckingErrorStack.disconnect(); 
    }

    public static void addErrorListener (IAcmeErrorListener listener) {
        errorListeners.add (listener);
    }
    
    public static void removeErrorListener (IAcmeErrorListener listener) {
        errorListeners.remove (listener);
    }
    
    protected static void announceErrorOccurred (AcmeError typecheckError) {
        Object erroneousObj = typecheckError.getSource();
        AcmeDesign des = findTargetDesign(erroneousObj);
        Iterator i = errorListeners.iterator();
        while (i.hasNext ()) {
            ((IAcmeErrorListener )i.next ()).errorDetected(des, typecheckError);
        }       
    }
    
    protected static void announceErrorFixed (AcmeObject obj, AcmeError error) {
        AcmeDesign des = findTargetDesign (obj);
        Iterator i = errorListeners.iterator ();
        while (i.hasNext ()) {
        	if (error != null) {
            	((IAcmeErrorListener )i.next ()).clearError(des, obj, error);
        	}
        	else {
				((IAcmeErrorListener )i.next ()).clearErrors(des, obj);
        		
        	}
        }
    }

    private static AcmeDesign findTargetDesign(Object erroneousObj) {
           AcmeDesign des = null;
            if (erroneousObj instanceof AcmeElement) {
                des = ((AcmeElement )erroneousObj).getParentDesign ();
            }
            else if (erroneousObj instanceof AcmeProperty) {
                Scope scope = ((AcmeProperty )erroneousObj).getScope ();
                if (scope instanceof AcmeElement) {
                    des = ((AcmeElement )scope).getParentDesign ();
                }
                else {
                    des = ((AcmeElement )((AcmeProperty )scope).getScope()).getParentDesign();
                }
                
            }
            else if (erroneousObj instanceof DesignRule) {
                des = ((AcmeElement ) ((DesignRule )erroneousObj).getDesignRuleScope()).getParentDesign();
            }
            else if (erroneousObj instanceof DesignAnalysis) {
                DesignAnalysis da = (DesignAnalysis )erroneousObj;
                des = ((AcmeElement )da.getScope()).getParentDesign ();
            }
            
        return des;
    }
};