package acmeLib.errorManagement;

import java.util.*;

import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeVector;

/**
 * Class: CausalErrorStack<br>
 * Purpose: CausalErrorStack keeps track of the outstanding errors that<br>
 *          are discovered during processing (initially only during typechecking).<br>
 *          The basic idea is that any given error can be caused by multiple<br>
 *          other errors.  This stack keeps the<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    2/11/98   RTM      created initial class<br>
 * @author Bob Monroe
 */

public class CausalErrorStack extends AcmeObject {

    //
    // Protected members
    //

  /**
   * stores the vector of current errors.
   */
//    protected AcmeVector causalErrors = null;


    protected Map _threadMap = new HashMap ();
    
    public void connect () {
        Thread thread = Thread.currentThread();
        if (!_threadMap.containsKey (thread)) {
            AcmeVector causalErrors = new AcmeVector (2,1);   
            _threadMap.put (thread, causalErrors);
        }   
    }
    
    public void disconnect () {
        _threadMap.remove (Thread.currentThread());
    }

  /**
   * flag that indicates whether the error stack is disabled.  If it is disabled
   * then it won't push or pop any error messages.
   */
    protected boolean disabled = false;

    //
    // public members
    //
  /**
   * This is the constuctor of this class
   */
    public CausalErrorStack() {
        // create a new vector, trading storage size for speed of growth.
        //this.causalErrors = new AcmeVector(2,1);
    };


  /**
   * disable the error stack
   **/
    public void disable() { this.disabled = true; };

  /**
   * enable the error stack
   **/
    public void enable() { this.disabled = false; };

  /**
   * check whether the error stack is disabled
   **/
    public boolean getIsDisabled() { return this.disabled; };

  /**
   * Get the causal error vector
   * @return The AcmeVector of the causal errors
   */
    public synchronized AcmeVector getCausalErrors() { 
        connect ();
        return (AcmeVector )_threadMap.get (Thread.currentThread ()); 
    };
  /**
   * Set the causal error vector
   * @param newErrorsVector The new causal error vector
   */
    public synchronized void setCausalErrors(AcmeVector newErrorsVector) {
        _threadMap.put (Thread.currentThread (), newErrorsVector);
    };

  /**
   * push a new causal error onto the top of the error stack
   * @param newError The AcmeError
   */
    public synchronized void pushCausalError(AcmeError newError) {
        if (this.disabled == false) {
            getCausalErrors().addItem(newError);
        };
    };


  /**
   * retrieve the top error on the stack, if there is one, without
   * removing it.  If the stack is empty then return null
   * @return The AcmeError
   */
    public synchronized AcmeError peekCausalError() {
        AcmeError topError = null;
        try {
            topError = (AcmeError)
                    getCausalErrors ().getItemAt(getCausalErrors ().getSize() - 1);
        }
        catch (IndexOutOfBoundsException obExc) {
            // return null if the array/stack is empty
            topError = null;
        }
        catch (ClassCastException castExc) {
            // return null if the top item of the stack isn't an AcmeError
            topError = null;
        };

        return topError;
    };


  /**
   * retrieve the top error on the stack, if there is one, and
   * remove it.  If the stack is empty then return null
   * @return The AcmeError
   */
    public synchronized AcmeError popCausalError() {
        AcmeError topError = null;
        try {
            if (this.disabled == false) {
                topError = (AcmeError)
                        getCausalErrors ().getItemAt(getCausalErrors ().getSize() - 1);
                getCausalErrors ().removeItemAt(getCausalErrors ().getSize() - 1);
            };
        }
        catch (IndexOutOfBoundsException obExc) {
            // return null if the array/stack is empty
            topError = null;
        }
        catch (ClassCastException castExc) {
            // return null if the top item of the stack isn't an AcmeError
            topError = null;
        };

        return topError;
    };


  /**
   * cancelError is used when an element typechecks and the pre-emptive
   * error that was placed on the error stack should be removed.
   * If the requested error is not the top of the stack this should
   * be an internal error, as it should not be possible to cancel
   * an error that has causal error children.  For now, just wipe out
   * the entire stack above the error who's removal is requested.
   * @param errToCancel error to cancel
   */
    public synchronized void cancelError(AcmeError errToCancel) {
        AcmeError topError = this.popCausalError();
        if (errToCancel == topError) {
            return;     // success.  we've cancelled the top error.
        }
            // this probably shouldn't happen, but if it does cancel the
            // requested error and all errors above it on the stack.
            int index = getCausalErrors ().lastIndexOf(errToCancel);
            if (index == -1) {
                return;     // errToCancel is not in the stack
            }
                // else remove the last instance of errToCancel and and all
                // errors above that in the stack.
                try {
                    while(getCausalErrors () != null) {
                        // repeat until we pass the array's last index, which
                        // will cause it to throw an out of bounds exception.
                        // catch the exception and continue with the vector
                        // truncated appropriately.
                        getCausalErrors ().removeItemAt(index);
                    };
                } catch (ArrayIndexOutOfBoundsException obExc) {
                    // do nothing, just fall through
                };
    };


  /**
   * The parameter parentError is the overall error caused by all of the
   * causalErrors currently on the stack above it (there may be no errors
   * above it, though.)  This method pops all of the errors from the stack
   * until it reaches parentError, and adds each of the popped errors to
   * the parentError as causes.  It should leave parentError on the top
   * of the stack at the end of the method.
   * @param parentError The parent error
   */
    public void packageErrorCauses(AcmeError parentError) {
        if (parentError == null)
            return;     // don't mess the stack up...

 
        AcmeError curError = this.popCausalError();
        while ((curError != parentError) && (curError != null)) {
            parentError.addCause(curError);
            curError = this.popCausalError();
        };

        // we either just popped the parent error, or ran into an
        // internal error looking for it.  In either case, put
        // parentError on the top of the stack and return.
        this.pushCausalError(parentError);
    };
};



/** RTM 2/15/98 -- Algorithm under consideration for maintaining a causal
                   error audit trail.

  In each typecheck*() method:
    - at the beginning of this method push an appropriate typechecking error
        on the error stack.  (do this only for public methods, not protected
        helper methods).
    - each error that occurs in calls made by this method or any method that
        this method calls will be pushed onto the top of the error stack.
    - at the end of the method:
        - if the entity typechecks, then it pops the self-error that it pushed
           at the beginning of the method off of the error stack and discards it.
           There should be no additional errors on the error stack because if
           there were any errors added by a called routine, then this containing
           entity would also fail to typecheck.
           - return typechecks = true.
        - if the entity does _not_ typecheck, then pop all of the error objects
           above the error added at the beginning of this method off of the
           stack and add them as causes to the error added at the beginning
           of the method.  This should leave the beginning error on the top
           of the stack, with a full tree of causing errors hanging from it.
           - return typechecks = false

   With this algorithm, the AcmeTypeManager does not announce any errors to
   the AcmeErrorManager.  It simply returns whether the entity typechecks.
   It is the calling method/object's responsibility to pull any errors off of
   the type manager's error stack and to then announce the appropriate error
   to the AcmeErrorManager.
*/