/**
 * Class:   IMsgNode<br>
 * Purpose: The IMsgNode interface provides a standard way to get error,
 *          warning, or other messages from the Java Acme/ArmaniLib.
 *          It is particularly designed to make the AcmeError tree more
 *          accessible to COM clients.
 *
 *
 * Signatures (without the javaDoc comments):

    public static final int NOTE    = 1;
    public static final int WARNING = 2;
    public static final int ERROR   = 3;
    public int getMsgType();
    public void setMsgType(int newMsgType);
    public String getMsgSource();
    public void setMsgSource(String qualifiedSourceName);
    public String getMsgText();
    public void setMsgText(String newMsgText);
    public String seeAlso();
    public void setSeeAlso(String namesToSeeAlso);
    public int getNumChildrenNodes();
    public IMsgNode getChildNode(int childNum);
    public void addChildNode(IMsgNode newChild);
    public void removeChildNode(int i);

 *
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    12/8/98  RTM       created initial class<br>
 * @author Bob Monroe
 */


/** Magic registration chant to register this class with the Windows repository

    RTM 12/8/98 -- Note that I don't seem to be able to create a typeLib from
    the interface itself.  As a result, I've created a type library for the
    AcmeError class, which is currently the toplevel class that actually
    implements this class.  This may need to be modified later to deal
    with non-error type messages (such as update request messages)

    %javareg /register /class:acmeLib.errorManagement.IMsgNode
             /clsid:{DE025010-8F0A-11d2-BC3D-00609737B51A}
             /progid:acmeLib.errorManagement.IMsgNode.java.1
 **/



package acmeLib.errorManagement;


public interface IMsgNode {

    // Constants to denote whether the message is a note, warning, or error.
    public static final int NOTE    = 1;
    public static final int WARNING = 2;
    public static final int ERROR   = 3;



    /**
     * @return an integer value representing the type of this message.
     *         NOTE == 1, WARNING == 2, ERROR == 3.
     **/
    public int getMsgType();

    /**
     * @param newMsgType an integer value representing the type of this
     *         message.  NOTE == 1, WARNING == 2, ERROR == 3.
     **/
    public void setMsgType(int newMsgType);



    /**
     * @return the fully qualified name of the source of this message.
     **/
    public String getMsgSource();

    /**
     * set the name of the MsgSource
     * @param qualifiedSourceName the fully qualified name of the source
     *        of this message.
     **/
    public void setMsgSource(String qualifiedSourceName);




    /**
     * @return the text of this message
     **/
    public String getMsgText();

    /**
     * @param newMsgText the text to which this message will be set
     **/
    public void setMsgText(String newMsgText);



    /**
     * @return a comma-separated list of qualified names.  These names refer
     *         to other entities of interest in the message.  The meaning of
     *         these names is not defined in the toplevel interface.  Classes
     *         that implement this interface can apply meaning to them as
     *         appropriate.
     **/
    public String seeAlso();

    /**
     * @param namesToSeeAlso a comma-separated list of qualified names.  that
     *         refer to other entities of interest in the message.  The meaning
     *         of these names is not defined in the toplevel interface.
     *         Classes that implement this interface can apply meaning to
     *         them as appropriate.
     **/
    public void setSeeAlso(String namesToSeeAlso);



    /**
     * @return the number of children IMsgNode's of this message node.
     **/
    public int getNumChildrenNodes();

    /**
     * @return the i'th IMsgNode child of this message
     **/
    public IMsgNode getChildNode(int childNum);

    /**
     * Add a child IMsgNode as the ith + 1 child of this node.
     * @param newChild child node to add.
     **/
    public void addChildNode(IMsgNode newChild);


    /**
     * removes the i'th child node from this message
     * @param i index of child node to remove
     **/
    public void removeChildNode(int i);

};