package acmeLib.properties;

import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeDuplicateNameException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.exceptions.AcmeKeywordNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.LanguageConfig;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.designRules.Scope;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeFamily;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.TypeRefResolutionError;
import acmeLib.typeManagement.IInheritable;

/**
 *  Class: AcmeProperty
BR>  <BR>
 *
 *  Purpose: base class for all Properties in ACME
BR><BR>
 *
 *  Description: A AcmeProperty class has a name, a type, a value and a set of meta-properties
BR>
 *  associated with it.  It includes public methods for accessing these
BR>
 *  values.  In order for a property to be valid it must be assigned a
BR>
 *  non-NULL name; the type and value may be null.  A NULL value means that
BR>
 *  the value is undefined.    A NULL value assigned to the type is
BR>
 *  interpreted as "AnyType".
BR> <BR>
 *
 *  When a value is assigned to a AcmeProperty with <methodname> setValue()
BR>
 *  </methodname>, or when a type and value are supplied to the
BR>
 *  constructor, the value must be consistent with the type, or an
BR>
 *  AcmeTypeViolation exception is generated.
BR> <BR>
 *
 *  Revision History:
BR>
 *    Date:     Person    Change
BR>
 *     ?/?/97    DK      Created C++ version
BR>
 *     4/22/97   RTM     Translated to Java
BR> <BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeProperty
	extends acmeLib.AcmeObject
	implements
		acmeLib.designRules.Scope,
		acmeLib.core.io.Exportable,
		acmeLib.core.io.ArmaniExportable,
		AcmePropertyBearer,
		IInheritable {

	//
	// protected members
	//
	protected String name;
	protected AcmePropertyType type;
	protected AcmePropertyValue val;
	protected AcmeMap metaPropMap;
	protected acmeLib.designRules.Scope scope;
	protected AcmeVector inheritedFrom;

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		boolean eq = false;
		if (obj instanceof AcmeProperty) {
			AcmeProperty prop = (AcmeProperty) obj;
			return name.equals(prop.getName())
				&& (type == prop.getType() || (type != null && type.equals(prop.getType())))
				&& (val == prop.getValue() || (val != null && val.equals(prop.getValue())));
		}
		return eq;
	}

	//
	// public methods
	//

	// Constructors
	/**
	 * This is a constructor for this class
	 */
	public AcmeProperty() {
		metaPropMap = new AcmeMap();
		this.setType(ConstantTypes.UNSPECIFIED_TYPE);
		this.name = null;
		this.scope = null;
	};

	/**
	 * This is a constructor for this class
	 * @param type the value to be assigned to type
	 */
	public AcmeProperty(AcmePropertyType type) {
		metaPropMap = new AcmeMap();
		this.name = null;
		this.setType(type);
		this.scope = null;
	};

	/**
	 * This is a constructor for this class
	 * @param type the value to be assigned to type
	 * @param prop_name the value to be assigned to name
	 * @param val the value to be assigned to val
	 */
	public AcmeProperty(String prop_name, AcmePropertyType type, AcmePropertyValue val) {
		metaPropMap = new AcmeMap();
		this.name = prop_name;
		this.setType(type);
		this.setValue(val);
		this.scope = null;
	};

	// accessors and setters.
	/**
	 * This method returns the scope of this property
	 * @return the scope of this property
	 */
	public acmeLib.designRules.Scope getScope() {
		return this.scope;
	};
	/**
	 * This method sets the scope of this property
	 * @param newScope the value to be assigned to this property's scope
	 */
	public void setScope(acmeLib.designRules.Scope newScope) {
		this.scope = newScope;
	}

	/**
	 * This method adds a Metaproperty to this property
	 * @param p the property to be added
	 * @exception AcmeIllegalNameException thrown when the property's name is illegal
	 */
	public void addProperty(AcmeProperty p) throws AcmeIllegalNameException {
		if (LanguageConfig.CHECK_FOR_KEYWORDS) {
			if (LanguageConfig.isAcmeKeyword(p.getName()))
				throw new AcmeKeywordNameException("Illegal keyword \"" + p.getName() + "\" used for property name.");
		};

		if (p.getName() == null) {
			throw new AcmeIllegalNameException("Error.  Attempted to add a " + "property with null name.");
		}

		else if (metaPropMap.lookupItem(p.getName()) != null) {
			throw new AcmeDuplicateNameException(
				"Attempted to redefine identifier " + p.getName() + " as an added property.");
		};

		// RTM 10/97: currently not allowing duplicate names interferes with the
		//            approach that I take to instantiate types.
		metaPropMap.addItem(p.getName(), p);
		p.setScope(this);
	};

	/**
	 * This method returns the metaProperty of this property
	 * @param name the name of the property to be returned
	 * @return the metaProperty of this property
	 */
	public AcmeProperty getMetaProperty(String name) {
		AcmeProperty p = (AcmeProperty) metaPropMap.lookupItem(name);
		if (p == null)
			return null;
		else
			return p;
	};

	/**
	 * This method removes a metaProperty from this property
	 * @param prop the property to be removed
	 */
	public void removeProperty(AcmeProperty prop) {
		this.metaPropMap.removeItem(prop.getName());
	};

	/**
	 * This method removes a metaProperty from this property
	 * @param name the name of the property to be removed
	 */
	public void removeProperty(String name) {
		AcmeProperty prop = (AcmeProperty) metaPropMap.lookupItem(name);
		if (prop != null) {
			prop.setScope(null);
		};
		metaPropMap.removeItem(name);
	};

	/**
	 * This method returns a metaProperty
	 * @return an enumeration of all metaProperties
	 */
	public AcmeEnumeration getProperties() {
		return metaPropMap.enumerate();
	};

	/**
	 * This method returns the number of metaProperites
	 * @return number of metaProperites
	 */
	public int getNumberOfMetaProperties() {
		return metaPropMap.getSize();
	};

	// property type accessors and settors
	/**
	 * This method returns the type of this property
	 * @return the type of this property
	 */
	public AcmePropertyType getType() {
		return type;
	};
	/**
	 * This method sets the type of this property
	 * @param t the value to be assigned to type
	 */
	public void setType(AcmePropertyType type) {
		this.type = type;
	};

	/**
	 * This method returns the name of this propery
	 * @return the name of this propery
	 */
	public String getName() {
		return name;
	};

	/**
	 * Retrieves the fully qualified, period delimited, name of the AcmeElement.
	 * This name identifies an object in the structure tree that can be reached
	 * by taking a unique path from the root design.
	 * @return the fully qualified name of the AcmeElement
	 **/
	public String getQualifiedName() {
		StringBuffer nameBuf = new StringBuffer();
		if (this.getScope() != null) {
			String parentName = "";

			if (this.getScope() instanceof AcmeElement) {
				parentName = ((AcmeElement) this.getScope()).getQualifiedName();
			}

			else if (this.getScope() instanceof AcmeProperty) {
				parentName = ((AcmeProperty) this.getScope()).getQualifiedName();
			};

			if (!(parentName.equals(""))) {
				nameBuf.append(parentName + ".");
			};

			nameBuf.append(this.getName());

		}
		else {
			nameBuf.append(this.getName());
		};

		return nameBuf.toString();
	};

	/**
	 * This method sets the name of this property
	 * @param newName value to be assigned to the name of the property
	 * @exception AcmeIllegalNameException thrown when newName is illegal
	 */
	public void setName(String newName) throws AcmeIllegalNameException {
		String oldName = this.name;
		this.name = newName;
		if (this.getScope() == null) {
			return;
		}
		else if (this.getScope() instanceof AcmeElement) {
			((AcmeElement) this.getScope()).childNameChanged(oldName, this);
		}
		else if (this.getScope() instanceof AcmeProperty) {
			((AcmeProperty) this.getScope()).childNameChanged(oldName, this);
		}
	};

	/**
	 * If one of this property's meta-properties changed its name, we
	 * need to update that meta property's entry in our meta property map.
	 * @param oldName the old name of the property
	 * @param obj the new property
	 * @exception AcmeIllegalNameException thrown when the property's name is illegal
	 */
	public void childNameChanged(String oldName, AcmeObject obj) throws AcmeIllegalNameException {
		// check for oldName in the following order: property, representation,
		if (obj instanceof AcmeProperty) {
			String newName = ((AcmeProperty) obj).getName();
			if (newName == null) {
				throw new AcmeIllegalNameException("Error.  Attempted to set a " + "meta-property name to null.");
			}
			else if (!newName.equals(oldName)) {
				this.removeProperty(oldName);
				this.addProperty((AcmeProperty) obj);
			};
			// else the name hasn't really changed, just return.
		};
	};

	// Generic Value interface
	/**
	 * This method returns the value of this property
	 * @return the value of this property
	 */
	public AcmePropertyValue getValue() {
		return this.val;
	};
	/**
	 * This method sets the value of this property
	 * @param value the value of this property
	 */
	public void setValue(AcmePropertyValue val) {
		this.val = val;
	};

	/**
	 * copy method creates a new property with a deep-copied value
	 * and set of meta properties, and a shallow-copied type pointer.
	 * @return a copy of this property
	 */
	public AcmeProperty copy() {
		AcmeProperty new_property;

		if (this.val != null) {
			new_property = new AcmeProperty(new String(this.name), this.type, this.val.copy());
		}
		else {
			new_property = new AcmeProperty(new String(this.name), this.type, null);
		}

		new_property.setScope(this.getScope());

		// copy the meta-properties of this property into the new object
		AcmeEnumeration e = getProperties();
		AcmeProperty old_meta_property;
		AcmeProperty new_meta_property;
		while (e.moreItems()) {
			old_meta_property = (AcmeProperty) e.getNextItem();
			if (old_meta_property == null) {
				continue; // don't add a null meta-property
			}
			else if (old_meta_property.getValue() == null) {
				new_meta_property = new AcmeProperty(old_meta_property.getName(), old_meta_property.getType(), null);
			}
			else {
				new_meta_property =
					new AcmeProperty(
						old_meta_property.getName(),
						old_meta_property.getType(),
						old_meta_property.getValue().copy());
			};
			try {
				new_property.addProperty(new_meta_property);
			}
			catch (AcmeIllegalNameException ainExc) {
				// if we're copying an existing (meta) property then there
				// should be no exceptions thrown...
			};
		}

		return new_property;
	};

	/**
	 * This is a generic unparsing method that works for both Acme and Armani.
	 * @param s AcmeOutputStream
	 * @param exportPropertyKeyword boolean
	 * @exception java.io.IOException when error occurs while writing
	 */
	public void export(AcmeOutputStream s, boolean exportPropertyKeyword) throws java.io.IOException {
		exportComment(s);
		s.beginRegion(this);
		if (exportPropertyKeyword) {
			s.write("Property ");
		};

		s.write(this.getName());

		if (this.getType() != null) {
			if (this.getType() != ConstantTypes.UNSPECIFIED_TYPE) {
				s.write(" : " + this.getType().getAcmeIdentifierString());
			};
			// else the type is undeclared so don't declare one...
		};

		if (this.getValue() != null) {
			s.write(" = ");
			this.getValue().exportAcme(s);
		}

		if (metaPropMap.getSize() > 0) {
			boolean haveWrittenBrace = false;

			AcmeEnumeration e = getProperties();
			e.sortByCreationOrder();
			AcmeProperty p;
			while (e.moreItems()) {
				p = (AcmeProperty) e.getNextItem();
				if (LanguageConfig.shouldExport(p)) {
					if (!haveWrittenBrace) {
						s.write(" << ");
						haveWrittenBrace = true;
					}

					p.export(s, false);
					s.write(" ");
				}
			};
			if (haveWrittenBrace)
				s.write(" >>");
		};
		if (exportPropertyKeyword)
			s.writeln(";");
		else
			s.write(";");
		s.endRegion(this);
	};

	// both exportAcme and exportArmani use the same generic export
	// method because there is no difference between property exporting
	// in Acme and Armani.
	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs while writing
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException {
		this.export(s, true);
	};

	/**
	  * @param s AcmeOutputStream
	  * @exception java.io.IOException when error occurs while writing
	  */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException {
		this.export(s, true);
	};

	/**
	 * Do all of the post-parsing processing of the design that is supposed
	 * to happen after parsing, in preparation for, but not including,
	 * typechecking.
	 */
	public void doPostParsingProcessing() {
		this.resolveTypeReferences();
	};

	/**
	 * This method resolves the property type object that this property refers to.
	 */
	public void resolveTypeReferences() {
		String lookupError = null;
		// first try to resolve the types specified in the property value
		// (this is probably only relevant for Records as of 3/4/98).
		if (this.getValue() != null) {
			this.getValue().resolveTypeReferences(this.getScope());
		};

		if (this.getType() instanceof AcmeUnresolvedPropertyType) {
			Object foundObject = null;
			if (this.getScope() == null) {
				// if scope is null then we can't resolve names,
				// announce an error and return.
				lookupError = "Unable to resolve property typename: " + this.getType().getTypeName();
				TypeRefResolutionError badScopeError =
					new TypeRefResolutionError(this, lookupError, this.getType().getTypeName());
				badScopeError.setUnresolvedRef(this);
				AcmeErrorManager.errorDetected(badScopeError);
				return;
			}
			else {
				if (this.getType().getQualifier() != null) {
					Object o = getScope().lookupName(getType().getQualifier(), true);
					if (o instanceof AcmeFamily) {
						foundObject = ((AcmeFamily) o).getType(((AcmeUnresolvedPropertyType )getType()).getUnqualifiedTypeName());
						if (foundObject instanceof AcmePropertyType) {
							foundObject = new AcmePropertyTypeRef ((AcmeFamily )o, (AcmePropertyType )foundObject);
						}
					}
				}
				else {
					foundObject = this.getScope().lookupName(this.getType().getTypeName(), true);
				}
			}

			if (foundObject instanceof AcmePropertyType) {
				this.setType((AcmePropertyType) foundObject);
				this.resolveSubTypeReferences ((AcmePropertyType )foundObject, this.getScope ());
			}
			else {
				// Announce an error and return
				lookupError = "Unable to resolve property typename: " + this.getType().getTypeName();
				TypeRefResolutionError notPropertyTypeError =
					new TypeRefResolutionError(this, lookupError, this.getType().getTypeName());
				notPropertyTypeError.setUnresolvedRef(this);
				AcmeErrorManager.errorDetected(notPropertyTypeError);
				AcmePropertyType unresolvedType = new AcmeUnresolvedPropertyType(this.getType().getTypeName());
				this.setType(unresolvedType);
				return;
			};
		};

		
		
		// now that we have a resolved property type, call resolve names
		// on the property type object, to resolve any unresolved names
		// in complex property types.
		if (this.getType() != null) {
			this.getType().resolveTypeReferences(this.getScope());
		}
		else {
			// Announce an error and return
			lookupError = "Unable to resolve property typename in property: " + this.getName();
			TypeRefResolutionError notPropertyTypeError = new TypeRefResolutionError(this, lookupError, null);
			notPropertyTypeError.setUnresolvedRef(this);
			AcmeErrorManager.errorDetected(notPropertyTypeError);
			AcmePropertyType unresolvedType = new AcmeUnresolvedPropertyType(this.getType().getTypeName());
			this.setType(unresolvedType);
			return;
		};

		
		
		// Resolve type references in values
		if (this.getValue () != null) {
			this.getValue ().assignType (this.getType ());
			resolveTypeReferencesInValue (this.getValue (), this.getScope ());
		}
		
		
		// recurse into this property's meta properties.
		AcmeEnumeration e = this.getProperties();
		while (e.moreItems()) {
			((AcmeProperty) e.getNextItem()).resolveTypeReferences();
		};

		// finally, set the type of this property's value to correspond
		// to the type of the property as a whole.
		if ((this.getType() instanceof IntType) && (this.getValue() instanceof AcmeInt)) {
			((AcmeInt) this.getValue()).setType((IntType) this.getType());
		}
		else if ((this.getType() instanceof FloatType) && (this.getValue() instanceof AcmeFloat)) {
			((AcmeFloat) this.getValue()).setType((FloatType) this.getType());
		}
		else if ((this.getType() instanceof StringType) && (this.getValue() instanceof AcmeString)) {
			((AcmeString) this.getValue()).setType((StringType) this.getType());
		}
		else if ((this.getType() instanceof BooleanType) && (this.getValue() instanceof AcmeBoolean)) {
			((AcmeBoolean) this.getValue()).setType((BooleanType) this.getType());
		}
		else if ((this.getType() instanceof SetType) && (this.getValue() instanceof AcmeSet)) {
			((AcmeSet) this.getValue()).setType((SetType) this.getType());
		}
		else if ((this.getType() instanceof SequenceType) && (this.getValue() instanceof AcmeSequence)) {
			((AcmeSequence) this.getValue()).setType((SequenceType) this.getType());
		}
		else if (this.getType() == ConstantTypes.UNSPECIFIED_TYPE) {
			if (this.getValue() instanceof AcmeSet) {
				AcmeSet thisSet = (AcmeSet) this.getValue();
				thisSet.setSetType(ConstantTypes.UNSPECIFIED_TYPE);
				thisSet.setType(ConstantTypes.ANY_SET_TYPE);
			};
		}
	};

	/**
	 * @param value
	 */
	private void resolveTypeReferencesInValue(AcmePropertyValue value, Scope scope) {
		
		if (value instanceof AcmeSet) {
			AcmeSet set = (AcmeSet)value;
			if (set.getSetType () instanceof AcmeUnresolvedPropertyType) {
				AcmeUnresolvedPropertyType urt = (AcmeUnresolvedPropertyType )set.getSetType ();
				Object foundObject = scope.lookupName(urt.getTypeName(), true);
				if (foundObject instanceof AcmePropertyType) {
					resolveSubTypeReferences((AcmePropertyType )foundObject, scope);
					set.setSetType ((AcmePropertyType )foundObject);
				}
			}
			AcmeEnumeration values = set.getValues();
			while (values.moreItems ()) {
				AcmePropertyValue v = (AcmePropertyValue )values.getNextItem ();
				resolveTypeReferencesInValue(v, scope);
			}
		}
		else if (value instanceof AcmeSequence) {
			AcmeSequence seq = (AcmeSequence )value;
			AcmeEnumeration values = seq.getValues ();
			while (values.moreItems ()) {
				AcmePropertyValue v = (AcmePropertyValue )values.getNextItem ();
				resolveTypeReferencesInValue(v, scope);
			}
		}
		else if (value instanceof AcmeRecord) {
			AcmeRecord rec = (AcmeRecord )value;
			AcmeEnumeration fields = rec.getFields ();
			while (fields.moreItems ()) {
				AcmeRecordField field = (AcmeRecordField )fields.getNextItem ();
				if (field.getType () instanceof AcmeUnresolvedPropertyType) {
					AcmeUnresolvedPropertyType urt = (AcmeUnresolvedPropertyType )field.getType ();
					Object foundObject = scope.lookupName(urt.getTypeName(), true);
					if (foundObject instanceof AcmePropertyType) {
						resolveSubTypeReferences((AcmePropertyType )foundObject, scope);
						field.setType ((AcmePropertyType )foundObject);
					}
					
				}
				resolveTypeReferencesInValue(field.getValue(), scope);
			}
		}
			
		
	}

	/**
	 * 
	 */
	private void resolveSubTypeReferences(AcmePropertyType type, Scope scope) {
		if (type instanceof SetType) {
			SetType st = (SetType )type;
			AcmePropertyType constituentType = (AcmePropertyType )st.getSetType ();
			if (constituentType instanceof AcmeUnresolvedPropertyType) {
				Object candidate = scope.lookupName(constituentType.getTypeName(), true);
				if (candidate == null || !(candidate instanceof AcmePropertyType)) {
					
				}
				else {
					AcmePropertyType theType = (AcmePropertyType )constituentType;
					resolveSubTypeReferences(theType, scope);
					st.setSetType(theType);
				}
			}
			else {
				resolveSubTypeReferences(constituentType, scope);
			}
		}
		else if (type instanceof SequenceType) {
			SequenceType st = (SequenceType )type;
			AcmePropertyType constituentType = (AcmePropertyType )st.getSequenceType();
			if (constituentType instanceof AcmeUnresolvedPropertyType) {
				Object candidate = scope.lookupName(constituentType.getTypeName (), true);
				if (candidate == null || !(candidate instanceof AcmePropertyType)) {
					
				}
				else {
					AcmePropertyType theType = (AcmePropertyType )constituentType;
					resolveSubTypeReferences(theType, scope);
					st.setSequenceType(theType);
				}
			}
			else
				resolveSubTypeReferences(constituentType, scope);
		}
		else if (type instanceof RecordType) {
			RecordType rt = (RecordType )type;
			AcmeEnumeration fields = rt.getFields();
			while (fields.moreItems ()) {
				AcmeRecordField field = (AcmeRecordField )fields.getNextItem ();
				AcmePropertyType fT = field.getType ();
				if (fT instanceof AcmeUnresolvedPropertyType) {
					Object cand = scope.lookupName(fT.getTypeName (), true);
					if (cand == null || !(cand instanceof AcmePropertyType)) {
						
					}
					else {
						AcmePropertyType theType = (AcmePropertyType )cand;
						resolveSubTypeReferences(theType, scope);
						field.setType (theType);
					}
				}
				else {
					resolveSubTypeReferences(fT, scope);
				}
				
			}
		}
	}

	/**
	 * This method implements the acmeLib.designRules.Scope interface to map names
	 * to property substructure and meta-properties.  Subclasses of
	 * @param name the name of the property
	 * @param bubbleUp boolean
	 * @return the Object corresponding to "name"
	 */
	public Object lookupName(String name, boolean bubbleUp) {
		Object retObj = null;
		if (name.equalsIgnoreCase("value"))
			retObj = this.getValue();
		if ((retObj == null) && (name.equalsIgnoreCase("name")))
			retObj = this.getName();
		if ((retObj == null) && (name.equalsIgnoreCase("type")))
			retObj = this.getType();
		if ((retObj == null) && (name.equalsIgnoreCase("ref")))
			retObj = this; // return a reference to this property, rather than its value.
		if ((retObj == null) && (this.getValue() != null))
			retObj = this.getValue().lookupName(name, false);
		if (retObj == null)
			retObj = this.getMetaProperty(name);

		if ((this.getScope() != null) && (bubbleUp) && (retObj == null)) {
			retObj = this.getScope().lookupName(name, bubbleUp);
		};

		return retObj;
	};

	/* (non-Javadoc)
	 * @see acmeLib.properties.AcmePropertyBearer#addProperty(java.lang.String, acmeLib.properties.AcmePropertyType, acmeLib.properties.AcmePropertyValue)
	 */
	public void addProperty(String name, AcmePropertyType t, AcmePropertyValue val) throws AcmeIllegalNameException {
		AcmeProperty prop = new AcmeProperty(name, t, val);
		addProperty(prop);

	}

	/* (non-Javadoc)
	 * @see acmeLib.properties.AcmePropertyBearer#getNumberOfProperties()
	 */
	public int getNumberOfProperties() {
		return metaPropMap.getSize();
	}

	/* (non-Javadoc)
	 * @see acmeLib.properties.AcmePropertyBearer#getProperty(java.lang.String)
	 */
	public AcmeProperty getProperty(String name) {
		return (AcmeProperty) metaPropMap.lookupItem(name);
	}

	/**
	 * Adds a property that this property is based on. If the property is
	 * unified from a property of a supertype or instantiated type, then the 
	 * property will point to the inherited property.
	 * @param typeProp The super-property that this property is inherited from
	 */
	public void addInheritedSource(IInheritable typeProp) {
		if (typeProp instanceof AcmeProperty) {
			AcmeProperty prop = (AcmeProperty) typeProp;

			if (inheritedFrom == null)
				inheritedFrom = new AcmeVector(1);
			if (!inheritedFrom.contains(prop)) {
				inheritedFrom.addItem(prop);
			}
		}
	}

	/**
	 * Returns true if this property is the same as an inherited property. 
	 * For this to be true, the name, type, value, and meta-properties all need
	 * to be the same.
	 * @return whether the property is the same as one of its inherited
	 */
	public boolean isSameAsInherited() {
		boolean same = false;
		if (inheritedFrom == null)
			return same;
		AcmeEnumeration inhs = inheritedFrom.enumerate();
		while (inhs.moreItems() && !same) {
			AcmeProperty prop = (AcmeProperty) inhs.getNextItem();
			same = prop.equals(this);
		}
		return same;
	}

	/* (non-Javadoc)
	 * @see acmeLib.typeManagement.IInheritable#getInheritedSources()
	 */
	public AcmeEnumeration getInheritedSources() {
		if (inheritedFrom == null)
			return new AcmeEnumeration();
		else
			return inheritedFrom.enumerate();
	}

	/* (non-Javadoc)
	 * @see acmeLib.typeManagement.IInheritable#isInherited()
	 */
	public boolean isInherited() {
		return inheritedFrom != null && inheritedFrom.getSize() != 0;
	}

};
