
package acmeLib.properties;

import acmeLib.*;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.io.*;
import acmeLib.core.exceptions.AcmeUnnamedPropertyException;
import acmeLib.designRules.Scope;
import acmeLib.elements.*;

import java.io.IOException;



/**
 *  Class: AcmePropertyType
<BR><BR>
 *
 *  Purpose: abstract base class for all AcmeProperty type classes in Acme
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person  Change<BR>
 *    ?/?/97    DK      Created C++ version<BR>
 *    4/22/97   RTM     Translated to Java<BR>
 *    6/16/97   RTM     merged acmeLib.properties.propertyType package<BR>
 *                      into acmeLib.properties.<BR>
 *    6/19/97   RTM     Added copy() and copyMetaProperties() methods<BR>
 *    8/6/97    RTM     removed deprecated getCategoryName method<BR><BR>
 *    12/7/98   RTM     Added the getQualifiedType() method<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public abstract class AcmePropertyType extends AcmeObject
                              implements AcmeType, Exportable, ArmaniExportable
{

    //
    // protected members
    //

  /**
   *  The scope in which this type was declared, if it is not an inlineTypeSpec.
   */
    protected Scope declarationScope = null;

  /**
   * Optional Name
   */
    protected String  typeName;    // Optional Name
    protected String qualifier; // Optional qualifying name
    protected AcmeMap typeMetaPropMap = null;
  /**
   * True if this is an inline type
   */
    protected boolean inlineTypeSpec = false;  // true if this is an inline type

  /**
   * If this is a property type declaration it is possible that it is aliasing
   * another type, but giving it a new name.  This is not the normal case, so
   * set isAlias to false and aliasedType to null by default.
   * if this property type declaration is an alias they will be set to true
   * and the aliased AcmePropertyType object, respectively.
   */
    protected boolean isAlias = false;
    protected AcmePropertyType aliasedType = null;

    //
    // protected members
    //

  /**
   * This method does a deep copy of the meta properties from "this" to dest_obj
   * note that dest_obj is passed by reference and the metaProperties
   * from "this" are copied into it.
   * @param destObj the PropertyType that gets the meta properties form "this'
   */

    protected void copyMetaProperties(AcmePropertyType destObj) {

        // copy the meta-properties of this property into the new object
        AcmeEnumeration e = getMetaProperties();
        AcmeProperty old_meta_property;
        AcmeProperty new_meta_property;
        while (e.moreItems()) {
            old_meta_property = (AcmeProperty) e.getNextItem();
            new_meta_property =
               new AcmeProperty(new String(old_meta_property.getName()),
                            old_meta_property.getType(),
                            old_meta_property.getValue().copy());
            try {
                destObj.addMetaProperty(new_meta_property);
            } catch (AcmeUnnamedPropertyException exception) {
                // if this throws an exception simply don't add the meta-prop
            };
        }
    };


    //
    // public members
    //

 /**
   * This is a constructor for this class
   */
    public AcmePropertyType() {
        this.setTypeName(null);
        typeMetaPropMap = new AcmeMap();
        inlineTypeSpec = false;  // true if this is an inline type
    };

 /**
   * This is a constructor for this class
   * @param name is the value to be assigned to the PropertyType's name
   */
    public AcmePropertyType(String name) {
        typeMetaPropMap = new AcmeMap();
        this.setTypeName(name);
        inlineTypeSpec = false;  // true if this is an inline type
    };

 /**
   * This method returns the name of this propery type
   * @return the name of this propery type
   */
	public String getTypeName () { return this.typeName; };
 /**
   * This method sets the name of this property type
   * @param newName value to be assigned to the name of the property type
   */
	public void setTypeName (String name) { this.typeName = name; };

	public void setQualifier (String q) {this.qualifier = q;}
	
	public String getQualifier () {return qualifier;}


  /**
   * Retrieves the fully qualified, period delimited, name of the property type.
   * This name identifies an object in the structure tree that can be reached
   * by taking a unique path from the root design.
   * @return the fully qualified name of the AcmeElement
   **/
    public String getQualifiedName() {
        StringBuffer nameBuf = new StringBuffer();

        if(this.getDeclarationScope() != null) {
            String parentName = "";

            if (this.getDeclarationScope() instanceof AcmeElement) {
                AcmeElement scopeElt = (AcmeElement) this.getDeclarationScope();
                parentName = scopeElt.getQualifiedName();
            };

            if (parentName != null) {
                if (!(parentName.equals(""))) {
                    nameBuf.append(parentName + ".");
                };
                nameBuf.append(this.getTypeName());
            };
        } else {
            nameBuf.append(this.getTypeName());
        };

        return nameBuf.toString();
    };





  /**
   * @return this property type's inlineTypeSpec
   */
    public boolean getInlineTypeSpec() { return this.inlineTypeSpec; };
  /**
   * @param new_val the value to be assigned to  this property type's inlineTypeSpec
   */
    public void setInlineTypeSpec(boolean new_val) {
        this.inlineTypeSpec = new_val;
    }

  /**
   * @return this property type's declarationScope
   */
    public Scope getDeclarationScope() { return this.declarationScope; };
 /**
   * @param newScope the value to be assigned to  this property type's declarationScope
   */
    public void setDeclarationScope(Scope newScope) {
        this.declarationScope = newScope;
    };


    // alias information getters/setters.
  /**
   * @return this property type's isAlias
   */
    public boolean getIsAlias() { return this.isAlias; };
  /**
   * @param newIsAlias the value to be assigned to this property type's isAlias
   */
    public void setIsAlias(boolean newIsAlias) { this.isAlias = newIsAlias; };

  /**
   * @return this property type's aliasedType
   */
    public AcmePropertyType getAliasedType() { return this.aliasedType; };
  /**
   * @param newAliasedType the value to be assigned to this property tyoe's aliasedType
   */
    public void setAliasedType(AcmePropertyType newAliasedType) {
        this.aliasedType = newAliasedType;
    };


  /**
   * copy is an abstract method that each AcmePropertyType subclass
   * needs to overide to do a deep copy of the property type.
   * @return null
   */
    public AcmePropertyType copy() { return null; }

  /**
   * This method exports this property type as a declaration.  This is an abstract
   * method that all subclasses need to override.
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs while writing
   */
	public abstract void exportAcme(AcmeOutputStream s) throws IOException;

  /**
   * This method exports this property type as a declaration.  This is an abstract
   * method that all subclasses need to override.
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs while writing
   */
	public abstract void exportArmani(AcmeOutputStream s) throws IOException;

  /**
   * This method gets this property type's string identifier expression.
   * @return this property type's string identifier expression
   */
    public abstract String getAcmeIdentifierString();


  /**
   * This method add a meta-property to this property type
   * @param p property to be added to this property type
   * @exception AcmeUnnamedPropertyException thrown when the property has no name
   */
	public void addMetaProperty (AcmeProperty p)
      throws AcmeUnnamedPropertyException {
        if (p.getName() == null) {
            String errMsg = "Error: Attempted to add a meta-property with no name";
            throw new AcmeUnnamedPropertyException(errMsg);
        };

    	typeMetaPropMap.addItem (p.getName(), p);
    };

  /**
   * This method returns a meta-property corresponding to "name"
   * @param name the name of the meta-property to be returned
   * @return meta-property corresponding to "name"
   */
	public AcmeProperty getMetaProperty(String name) {
	    AcmeProperty p = (AcmeProperty) typeMetaPropMap.lookupItem(name);
    	if (p == null)
    	    return null;
    	else
    	    return p;
    };

  /**
   * This method returns an enumeration of all meta-properties in this property type
   * @return enumeration of all meta-properties in this property type
   */
	public AcmeEnumeration getMetaProperties() {
	    return typeMetaPropMap.enumerate();
	};

  /** subclasses of AcmePropertyType should implement typechecking.
   *  AcmePropertyType is a quasi-abstract base class.
   *  @param val the PropertyValue to be typechecked
   *  @return true
   */
    public abstract boolean typecheck (AcmePropertyValue val); 


    /**
     * This method recursively resolves the unresolved property type references
     * in any compound property type definitions.  This is an empty
     * quasi-abstract method in this class, implemented by subclasses
     * that need it.
     * @param scope acmeLib.designRules.Scope
     */
    public void resolveTypeReferences(acmeLib.designRules.Scope scope) { };
    
    public abstract String getDefinitionAsString ();

};
