
package acmeLib.properties;

import java.io.IOException;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.typeManagement.AcmeTypeManager;


/**
 *  Class: AcmeRecord<BR><BR>
 *
 *  Purpose:  A Record value.  Like all ACMEPropertyValue subclasses, an<BR>
 *          ACMERecord has a type.  It also contains a collection of<BR>
 *          fields which are simply attribute value pairs composed of a<BR>
 *          name and a AcmePropertyValue.<BR><BR>
 *
 *  Revision History:<BR>
 *    Date:     Person    Change<BR>
 *    ?/?/97    DK      Created C++ version<BR>
 *    4/30/97   RTM     Translated to Java<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeRecord extends AcmePropertyValue {

  //
  // protected members
  //
  protected RecordType type;
  protected AcmeMap recordFieldMap;

  //
  // public members
  //
  /**
   * This is a constructor for this class
   */
  public AcmeRecord () {
    this.type = null;
    this.recordFieldMap = new AcmeMap();
  };

  /**
   * This is a constructor for this class
   * @param t the value assigned to this record's type
   */
  public AcmeRecord (RecordType t) {
    this.type = t;
    this.recordFieldMap = new AcmeMap();
  };

 /**
   * This method returns a copy of this record
   * @return a copy of this record
   */
  public AcmePropertyValue copy() {
    AcmeRecord new_rec = new AcmeRecord(this.type);

    AcmeEnumeration fields_enum = recordFieldMap.enumerate();
    while (fields_enum.moreItems()) {
        new_rec.addField(
            ((AcmeRecordField) fields_enum.getNextItem()).copy() );
    };

    new_rec.setIsDefaultValue(this.getIsDefaultValue());
    return new_rec;
  };


  // Accessors/settors
 /**
   * This method sets the type of this record
   * @param t the value to be assigned to record
   */
  public void setType (RecordType t) { this.type = t; };
 /**
   * This method returns the type of this record
   * @return the type of this record
   */
  public AcmePropertyType getType () { return this.type; };


 /**
   * This method checks whether two recodrs are equal
   * @param test_value the record to be compared
   * @return whether "this" and test_value are equal
   */
  public boolean equals(Object test_value) {
    if (!(test_value instanceof AcmeRecord))
        return false;
    AcmeRecord test_record = (AcmeRecord) test_value;

    // make sure that there are the same number of fields in the record.
    AcmeEnumeration this_fields_enum = this.getFields();
    AcmeEnumeration test_fields_enum = test_record.getFields();
    if (this_fields_enum.getSize() != test_fields_enum.getSize())
        return false;

    // compare the record fields (type and value) of each of this's
    // record fields to the corresponding field in the test_record
    AcmeRecordField this_record_field, test_record_field;
    AcmePropertyType this_record_field_type, test_record_field_type;
    AcmePropertyValue this_record_field_value, test_record_field_value;
    while (this_fields_enum.moreItems()) {
        this_record_field = (AcmeRecordField) this_fields_enum.getNextItem();
        if (this_record_field == null)
            continue;   // some error in this's field return method
        test_record_field = test_record.getField(this_record_field.getName());
        if (test_record_field == null)
            return false;  // field exists in this but not test instance.
        this_record_field_type = this_record_field.getType();
        test_record_field_type = test_record_field.getType ();
        if (this_record_field_type instanceof AcmePropertyTypeRef) {
			this_record_field_type = ((AcmePropertyTypeRef) this_record_field_type).getReferentType ();	
		}
		if (test_record_field_type instanceof AcmePropertyTypeRef) {
			test_record_field_type = ((AcmePropertyTypeRef) test_record_field_type).getReferentType ();	
		}
        if (this_record_field_type != null) {
            if (!AcmeTypeManager.isLogicalSubtype(test_record_field.getType(),
                                  this_record_field_type, false))
            {
                return false;
            };
        };
        this_record_field_value = this_record_field.getValue();
        if (this_record_field_value != null) {
            if (!this_record_field_value.equals(test_record_field.getValue()))
                return false;
        };
    };  // end while moreItems();

    // if it makes it this far then test_record and this are equal
    return true;
  };


  //
  // Record Field manipulation methods
  //

  /**
   * This method returns this record's field corresponding to field_name
   * @param field_name the name of the field to be returned
   * @return the record's field corresponding to field_name
   */
  public AcmeRecordField getField (String field_name) {
    return (AcmeRecordField) recordFieldMap.lookupItem (field_name);
  }

  /**
   * This method sets this record's field
   * @param name the name of the field to be set
   * @param val the value of the field to be set
   */
  public void setField (String name, AcmePropertyValue val) {
	AcmeRecordField field_to_set;
	field_to_set = (AcmeRecordField) recordFieldMap.lookupItem (name);
    // if the field wasn't found, then add this as a new field, else set it
    if (field_to_set == null) {
        this.addField(name,val);
    } else {
        field_to_set.setValue(val);
    }
  };

  /**
   * This method adds a field to this record
   * @param filed the field to be added to this record
   */
  public void addField (AcmeRecordField field) {
    field.setParent(this);
    recordFieldMap.addItem (field.getName(), field);
  };

  /**
   * This method adds a field to this record
   * @param name the name of the field to be added to this record
   * @param val the value of the field to be added to this record
   */
  public void addField (String name, AcmePropertyValue val) {
    AcmeRecordField newField = new AcmeRecordField (name, null, val);
    newField.setParent(this);
    recordFieldMap.addItem (name, newField);
  };

 /**
   * This method adds a field to this record
   * @param name the name of the field to be added to this record
   * @param val the value of the field to be added to this record
   * @param type the type of the field to be added to this record
   */
  public void addField (String name, AcmePropertyType type, AcmePropertyValue val) {
    AcmeRecordField newField = new AcmeRecordField (name, type, val);
    newField.setParent(this);
    recordFieldMap.addItem (name, newField);
  };

  /**
   * This method returns this record's field corresponding to name
   * @param name the name of the field to be returned
   * @return the field corresponding to name
   */
  public AcmePropertyValue getFieldValue (String name) {
    AcmeRecordField field =
            (AcmeRecordField) recordFieldMap.lookupItem (name);
    if (field == null)
        return null;
    else
        return field.getValue();
  };

  /**
   * remove the field with the name "fieldName" from this record.
   * @param fieldName the name of the field to remove from this record
   **/
  public void removeField(String fieldName) {
    AcmeRecordField fieldToRemove = (AcmeRecordField)
                            this.recordFieldMap.lookupItem(fieldName);
    if (fieldToRemove != null) {
        if (fieldToRemove.getParent() == this) {
            fieldToRemove.setParent(null);
        };
        this.recordFieldMap.removeItem(fieldName);
    };
  }


  /**
   * remove the requested record field from this record, iff it is currently
   * a field of the record.
   * @param fieldToRemove the field to remove from this record
   **/
  public void removeField(AcmeRecordField fieldToRemove) {
    if (recordFieldMap.contains(fieldToRemove)) {
        this.recordFieldMap.removeItem(fieldToRemove.getName());
        if (fieldToRemove.getParent() == this) {
            fieldToRemove.setParent(null);
        };
    };
  }


  /**
   * This method returns an enumeration of all the fields of this record
   * @return an enumeration of all the fields of this record
   */
  public AcmeEnumeration getFields() { return recordFieldMap.enumerate(); };


  /**
   * If one of this record's fields changed its name, we
   * need to update that field's entry in our recordFieldMap.
   * @param oldName the old name of the property
   * @param obj the new property
   * @exception AcmeIllegalNameException thrown when the property's name is illegal
   **/
  public void childNameChanged(String oldName, AcmeRecordField field)
      throws AcmeIllegalNameException
  {
      if (field.getName() == null) {
          throw new AcmeIllegalNameException("Error.  Attempted to set a "
                      + "record field name to null.");
      } else if (!field.getName().equals(oldName)) {
          this.removeField(oldName);
          this.addField(field);
      };
      // else the name hasn't really changed, just return.
  };


 /**
   * This method converts AcmeRecord to String
   * @return the String value of this record
   */
  public String toString() {
    try {
        AcmeOutputStream s = new AcmeOutputStream ();
        exportAcme (s);
        s.flush ();
        String ret = s.toString();
        s.close ();
        return ret;
    }
    catch (IOException e) {
    }
    return "";
  };


    public void exportAcme (AcmeOutputStream s) throws IOException {
        s.beginRegion(this);
        AcmeEnumeration e = getFields ();
        e.sortByCreationOrder ();
        AcmeRecordField f;
        s.writeln ("[");
        s.indent();
        while (e.moreItems ()) {
            f = (AcmeRecordField )e.getNextItem ();
            f.exportAcme (s);
            if (e.moreItems ()) {
                s.writeln("; ");
            }
        }
        s.unIndent ();
        s.write (" ]");
        s.endRegion(s);
    }
  /**
   * This method looks up a field named "name" in this record.  Return the field if
   * there is a field by that name, else return null
   * @param name the name of the record
   * @param bubbleUp boolean
   * @return the field corresponding to name
   */
  public Object lookupName(String name, boolean bubbleUp) {
    return this.getField(name);
  };

  /**
   * This method resolves the property type object for all of this record's fields.
   * @param scope acmeLib.designRules.Scope
   */
  public void resolveTypeReferences(acmeLib.designRules.Scope scope) {
    // iterate through all of the fields, resolving their typenames.
    AcmeEnumeration e = this.getFields();
    AcmeRecordField f = null;
    while (e.moreItems()) {
        f = (AcmeRecordField) e.getNextItem();
        f.resolveTypeReferences(scope);
    };
  }

/* (non-Javadoc)
 * @see acmeLib.properties.AcmePropertyValue#assignType(acmeLib.properties.AcmePropertyType)
 */
public void assignType(AcmePropertyType type) {
	if (type instanceof RecordType) setType ((RecordType )type);
};


};
