package acmeLib.properties;

import java.io.IOException;

import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeFamily;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.TypeRefResolutionError;

/**
 *  Class: AcmeRecordField
BR><BR>
 *
 *  Purpose: a type description for a record field
BR><BR>
 *
 *  Revision History:
BR>
 *    Date:     Person    Change
BR>
 *    ?/?/97    DK      Created C++ version
BR>
 *    6/1/97    RTM     Translated to Java
BR>
 *    6/19/97   RTM     Added copy() method
BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeRecordField extends acmeLib.AcmeObject {

	//
	// protected members
	//
	protected String name;
	protected AcmePropertyType type;
	protected AcmePropertyValue value;
	protected AcmeRecord parent;

	//
	// public members
	//
	/**
	 * This is a constructor for this class
	 * @param name is the value to be assigned to the record field's name
	 * @param t the value to be assigned to the record field's type
	 * @param val the value to be assigned to the record field's value
	 */
	public AcmeRecordField(String name, AcmePropertyType t, AcmePropertyValue val) {
		this.name = name;
		this.setType(t);
		value = val;
		parent = null;
	};

	/**
	 * This method sets the name of this record field and alerts its parent
	 * record that the name has been changed.
	 * @param name value to be assigned to the name of the record field
	 * @exception acmeLib.core.exceptions.AcmeIllegalNameException if the
	 * parameter "name" is illegal
	 **/
	public void setName(String name) throws AcmeIllegalNameException {
		String oldName = this.name;
		this.name = name;
		if (this.getParent() != null) {
			this.getParent().childNameChanged(oldName, this);
		};
	};

	/**
	 * This method returns the name of this record field
	 * @return the name of this record field
	 */
	public String getName() {
		return this.name;
	};

	/**
	  * This method sets the type of this record field
	  * @param t the value to be assigned to type
	  */
	public void setType(AcmePropertyType t) {
		type = t;
	};
	/**
	  * This method returns the type of this record field
	  * @return the type of this record field
	  */
	public AcmePropertyType getType() {
		return this.type;
	};

	/**
	 * This method returns the value of this record field
	 * @return the value of this record field
	 */
	public AcmePropertyValue getValue() {
		return value;
	};
	/**
	 * This method sets the value of this record field
	 * @param val the value of this record field
	 */
	public void setValue(AcmePropertyValue val) {
		this.value = val;
	};

	/**
	 * This method returns the record that contains this record field
	 * @return the parent record
	 */
	public AcmeRecord getParent() {
		return parent;
	};

	/**
	 * This method sets the record that contains this record field
	 * @param val the parent of this record field
	 */
	public void setParent(AcmeRecord newParent) {
		this.parent = newParent;
	};

	/**
	 * This method returns a deep-copy of "this" RecordField
	 * @return a copy of this record field
	 */

	public AcmeRecordField copy() {
		AcmePropertyType type;
		AcmePropertyValue val;

		if (this.getType() == null)
			type = null;
		else
			type = this.getType();

		if (this.getValue() == null)
			val = null;
		else
			val = this.getValue().copy();

		AcmeRecordField newField = new AcmeRecordField(this.getName(), type, val);

		return newField;
	}

	/**
	 * This method converts AcmeRecordField to String
	 * @return the String value of this record field
	 */
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append(this.getName());
		if (this.getType() != null) {
			if (this.getType() != ConstantTypes.UNSPECIFIED_TYPE) {
				buf.append(" : " + this.getType().getAcmeIdentifierString());
			};
			// else the type is undeclared so don't declare one...
		};

		if (getValue() != null) {
			buf.append(" = " + this.getValue().toString());
		};

		return buf.toString();
	};

	public void exportAcme(AcmeOutputStream s) throws IOException {
		s.beginRegion(this);
		s.write(this.getName());
		if (this.getType() != null) {
			if (this.getType() != ConstantTypes.UNSPECIFIED_TYPE) {
				s.write(" : " + this.getType().getAcmeIdentifierString());
			}
		}
		if (this.getValue() != null) {
			s.write(" = ");
			this.getValue().exportAcme(s);
		}
		s.endRegion(this);

	}

	/**
	 * This method resolves the property type object that this record field type
	 * stores (e.g. resolves "foo" and bar of a [ x : foo; y : bar]
	 * record type definition).  Recursively calls resolveTypeReferences on the
	 * field value object also.
	 * @param scope acmeLib.designRules.Scope
	 */
	public void resolveTypeReferences(acmeLib.designRules.Scope scope) {
		String lookupError = null;
		if (this.getType() instanceof AcmeUnresolvedPropertyType) {
			Object foundObject = null;
			if (scope == null) {
				// if scope is null then we can't resolve names,
				// announce an error and return.
				lookupError = "Unable to resolve record field typename: " + this.getType().getTypeName();
				TypeRefResolutionError badScopeError =
					new TypeRefResolutionError(this, lookupError, this.getType().getTypeName());
				badScopeError.setUnresolvedRef(this);
				AcmeErrorManager.errorDetected(badScopeError);
				return;
			}
			else {
				if (this.getType().getQualifier() != null) {
					Object o = scope.lookupName(getType().getQualifier(), true);
					if (o instanceof AcmeFamily) {
						foundObject =
							((AcmeFamily) o).getType(((AcmeUnresolvedPropertyType) getType()).getUnqualifiedTypeName());
						if (foundObject instanceof AcmePropertyType) {
							foundObject = new AcmePropertyTypeRef((AcmeFamily) o, (AcmePropertyType) foundObject);
						}
					}
				}
				else {
					foundObject = scope.lookupName(this.getType().getTypeName(), true);
				}
			};

			if (foundObject instanceof AcmePropertyType) {
				this.setType((AcmePropertyType) foundObject);
			}
			else {
				// Announce an error and return
				lookupError = "Unable to resolve record field typename: " + this.getType().getTypeName();
				TypeRefResolutionError notPropertyTypeError =
					new TypeRefResolutionError(this, lookupError, this.getType().getTypeName());
				notPropertyTypeError.setUnresolvedRef(this);
				AcmeErrorManager.errorDetected(notPropertyTypeError);
				AcmePropertyType unresolvedType = new AcmeUnresolvedPropertyType(this.getType().getTypeName());
				this.setType(unresolvedType);
				return;
			};
		};

		// now that we have a resolved property type, resolve the typenames
		// on the property type object, to resolve any unresolved names
		// in complex property types.
		if (this.getType() != null) {
			this.getType().resolveTypeReferences(scope);
		}
		else {
			// Announce an error and return
			lookupError = "Unable to resolve record field typename: " + this.getType().getTypeName();
			TypeRefResolutionError notPropertyTypeError =
				new TypeRefResolutionError(this, lookupError, this.getType().getTypeName());
			notPropertyTypeError.setUnresolvedRef(this);
			AcmeErrorManager.errorDetected(notPropertyTypeError);
			AcmePropertyType unresolvedType = new AcmeUnresolvedPropertyType(this.getType().getTypeName());
			this.setType(unresolvedType);
			return;
		};

		// recursively resolve all types that might be referenced in the
		// property value.
		if (this.getValue() != null) {
			this.getValue().resolveTypeReferences(scope);
		};
	};

};
