
package acmeLib.properties;

import java.io.IOException;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.io.AcmeOutputStream;


/** 
 *  Class: AcmeSequence
<BR><BR>
 *
 *  Purpose:The Sequence class implements an ACME sequence.  It includes
<BR>
 *          methods for accessing its list of AcmePropertyValues.  A sequence
<BR>
 *          may be assigned a type.
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    4/30/97   RTM     Translated to Java
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeSequence extends AcmePropertyValue {

  //
  // protected members
  //
  protected SequenceType type;
  protected AcmeVector sequenceValues;   // Vector of AcmePropertyValues

  //
  // public members
  //
 /**
  * This is a constructor for this class 
  */
  public AcmeSequence() {
    this.setType(null);
    sequenceValues = new AcmeVector();
  };

 /**
   * This is a constructor for this class 
   * @param t the value to be assigned to the sequence's type       
   */
  public AcmeSequence (SequenceType t) {
    this.setType(t);
    sequenceValues = new AcmeVector();
  };

  /**
   * This method performs a deep copy -- copy all of the values from this sequence
   * into new objects in the new sequence
   * @return a copy of this sequence
   */
  public AcmePropertyValue copy () {
	AcmeSequence new_seq = new AcmeSequence((SequenceType )this.getType());

    // walk over the elements in this sequence, creating a new AcmeVector
    AcmeEnumeration e = sequenceValues.enumerate();
	while (e.moreItems()) {
	    try {
	        new_seq.addValue( ((AcmePropertyValue) e.getNextItem()).copy() );
	    } catch (ClassCastException exception) {
	        // can't copy this, just continue
	    };
	}

    new_seq.setIsDefaultValue(this.getIsDefaultValue());
    return new_seq;
  };

/** 
   * This method checks whether two sequences are equal
   * @param test_value the sequence to be compared
   * @return whether "this" and test_value are equal
   */
  public boolean equals(Object test_value) {
    // first check whether test_value and this are the same types of sequences
    if (!(test_value instanceof AcmeSequence))
        return false;

    // else test_value and "this" are the same types of sequences,
    // by comparing their values.
    AcmeEnumeration these_values = this.getValues();
    AcmeEnumeration test_values = ((AcmeSequence) test_value).getValues();
    AcmePropertyValue this_prop_val, test_prop_val;
    while (these_values.moreItems()) {
        if (!(test_values.moreItems()))
            return false;   // this has more values than test
        else {
            this_prop_val = (AcmePropertyValue) these_values.getNextItem();
            test_prop_val = (AcmePropertyValue) test_values.getNextItem();
            if (!(this_prop_val.equals(test_prop_val)))
                return false;
        };
    };

    if (test_values.moreItems())
        return false;   // test has more values than this.

    return true;    // if it gets here then test_value and this are equal
  };


  /**
   * This method sets the type of this sequence
   * @param t the value to be assigned to type
   */
  public void setType (SequenceType t) { this.type = t; };
  /**
   * This method returns the type of this sequence
   * @return the type of this sequence
   */
  public AcmePropertyType getType() { return this.type; };

  /**
   * This method returns the size of the sequence
   * @return the size of the sequence
   */
  public int getSize() { return sequenceValues.getSize(); };

  /**
   * This method returns the value of this sequence
   * @param index the property value at position index
   * @return the value of this sequence
   */
  public AcmePropertyValue getValue (int index) {
    return (AcmePropertyValue) sequenceValues.getItemAt(index);
  }
 /**
   * This method adds value to this sequence 
   * @param value the value to be added to this sequence
   */ 
  public void addValue (AcmePropertyValue value) {
    sequenceValues.addItem(value);
  };
  
  public void removeValue (AcmePropertyValue value) {
  	sequenceValues.removeItem (value);
  }

  /**
   * This method inserts a value in the sequence at a given index
   * @param value the value to be inserted into the sequence
   * @param index the position at which value will be inserted
   */
  public void insertValueAt(AcmePropertyValue value, int index) {
    sequenceValues.insertItemAt(value,index);
  };

  /**
   * This method returns an enumeration of all the values in this sequence
   * @return an enumeration of all the values in this sequence
   */
  public AcmeEnumeration getValues () {
  	return sequenceValues.enumerate();
  };

  /**
   * This method converts AcmeSequence to String
   * @return the String value of this sequence
   */
  public String toString() {
    String ret = "";
    try {
        AcmeOutputStream s = new AcmeOutputStream ();
        exportAcme (s);
        s.flush ();
        ret = s.toString();
        s.close ();
        
    }
    catch (IOException e) {
    }
    return ret;
  };
  
  public void exportAcme (AcmeOutputStream s) throws IOException {
      s.beginRegion(this);
      AcmeEnumeration e = sequenceValues.enumerate ();
      s.write ("< ");
      AcmePropertyValue item;
      while (e.moreItems ()) {
        item = (AcmePropertyValue )e.getNextItem();
        item.exportAcme(s);
        if (e.moreItems ()) s.write (", ");
        
      }
      s.write (" >");
      s.endRegion (this);
  }

/**
 * @param value
 * @param i
 */
public void removeValueAt(int i) {
	sequenceValues.removeItemAt(i);
}

/* (non-Javadoc)
 * @see acmeLib.properties.AcmePropertyValue#assignType(acmeLib.properties.AcmePropertyType)
 */
public void assignType(AcmePropertyType type) {
	if (type instanceof SequenceType) setType ((SequenceType )type);
}

};

