
package acmeLib.properties;

import acmeLib.designRules.DesignRuleEvalException;
import acmeLib.designRules.ExpressionNode;


/** 
 *  Class:  AcmeString
<BR><BR>
 *
 *  Purpose: An ACME string value.  When a string value is assigned it is
<BR>
 *          assigned by value (that is, the string is copied into the
<BR>
 *          AcmeString).
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    4/30/97   RTM     Translated to Java
<BR>
 *    8/7/97    RTM     added massage* methods to deal with embedded quotes
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class AcmeString extends AcmePropertyValue {

  //
  // protected members
  //
  protected StringType type;
  protected String val;

  //
  // public members
  //
/**
   * This is a constructor for this class 
   * @param type the value to be assigned to the string's type 
   * @param val the value to be assigned to the string's val      
   */
  public AcmeString (StringType type, String val) {
    this.setValue(val);
    this.setType(type);
  };

/**
   * This is a constructor for this class 
   * @param val the value to be assigned to the string's val   
   */
  public AcmeString (String val) {
    this.setValue(val);
    this.setType(null);
  };

 /** 
   * This method performs copy of this string
   * @return the copy of this string
   */
  public AcmePropertyValue copy() {
    AcmeString copyVal = new AcmeString(this.type, this.val);
    copyVal.setIsDefaultValue(this.getIsDefaultValue());
    return copyVal;
  };

/** 
   * This method checks whether two strings are equal
   * @param test_value the string to be compared
   * @return whether "this" and test_value are equal
   */
  public boolean equals(Object test_value) {
    if (test_value instanceof AcmeString) {
        if (((AcmeString)test_value).getValue() != null) {
            return (((AcmeString)test_value).getValue().equals(this.val));
        }
            return (this.getValue() == null);
    } else if (test_value instanceof ExpressionNode) {
        try {
            if (this.getValue() == null) {
                return (((ExpressionNode) test_value).evaluateAsString() == null);
            }
                return this.getValue().equals(
                            ((ExpressionNode) test_value).evaluateAsString());
        } catch (DesignRuleEvalException drExc) {
        };
    }
        return false;
  };


 /**
   * This method sets the type of this string
   * @param t the value to be assigned to type
   */
  public void setType (StringType t) { this.type = t; };
 /**
   * This method returns the type of this string
   * @return the type of this string
   */
  public AcmePropertyType getType () { return this.type; };

/**
   * This method returns the value of this string
   * @return the value of this string
   */
  public String getValue () { return val; };
 /**
   * This method sets the  value of this string
   * @param value the value to be set
   */ 
  public void setValue (String value) { this.val = value; };


  /**
   * The toString method writes out the string with all internal
   * quotes backslashed, and non-backslashed quotes are added
   * to the beginning and the end of the string.
   * @return the string representation of AcmeString
   */
  public String toString() {
    return massageForExport(this.getValue());
  };


  /**
   * massageForInternalStorage is a function that removes any leading
   * and trailing quotes that are not backslashed ( \" ).  It then removes
   * the backslashes from all internal quotes in the string.  This method
   * massages a string sent from the JavaCC-based AcmeParser class
   *
   * This method's converse, massageForExport(String s) adds backslashes
   * to all internal quotes and then adds leading and trailing quotes.
   *
   * NOTE: this is a function.  it does not adjust the internal representation
   *       of an AcmeString.  It simply massages the string passed in as
   *       an argument
   * @param s the string to be manipulated
   * @return the changed string
   */
  public static String massageForInternalStorage(String s) {

        // first trim the leading and trailing quotes, if they exist
        int start_index = 0;
        int end_index = s.length();

        if ( s.startsWith("\"") && s.endsWith("\"") && !s.endsWith("\\\"") ) {
            start_index = 1;
            end_index = s.length() - 1;
        };

        String prepped_str = new String (s.substring(start_index, end_index));

        // now that we have prepared the string, convert all backslashed
        // quotes to non-backslashed quotes.

        StringBuffer result_buf = new StringBuffer();
        int str_end = prepped_str.length();
        int cur_index = 0;
        int next_slash_quote_index = 0;
        boolean end_reached = false;

        while (!end_reached) {
            // find the next occurence of the substring \"
            next_slash_quote_index = prepped_str.indexOf("\\\"", cur_index);

            if (next_slash_quote_index == -1) {
                // there were no more backslashed quotes found
                result_buf.append(prepped_str.substring(cur_index, str_end));
                end_reached = true;
            } else {
                // append everything up to the backslash to the results buffer
                // and move the cur_index to the following quote.
                result_buf.append(prepped_str.substring(cur_index,
                                            next_slash_quote_index));
                cur_index = next_slash_quote_index + 1;
            };

            if (cur_index > str_end)
                end_reached = true;
        }

        return result_buf.toString();
  };


  /**
   * massageForExport adds leading and trailing quotes. It then removes
   * the backslashes from all internal quotes in the string.  This method
   * massages an internally stored string that has been formatted with
   * the method massageForInternalStorage
   *
   * NOTE: this is a function.  it does not adjust the internal representation
   *       of an AcmeString.  It simply massages the string passed in as
   *       an argument
   * @param s the string to be manipulated
   * @return the changed string
   */
  public static String massageForExport(String s) {

        StringBuffer result_buf = new StringBuffer();
        // add the initial quote to the result string
        result_buf.append("\"");

        // search the string s for quotes, when found add a backslash
        // and a quote to the string buffer.
        int cur_index = 0;
        int next_quote_index = 0;
        int str_end = s.length();
        boolean end_reached = false;

        while (!end_reached) {
            next_quote_index = s.indexOf("\"", cur_index);

            if (next_quote_index == -1) {
                // there were no quotes found
                result_buf.append(s.substring(cur_index,str_end));
                end_reached = true;
            } else if (next_quote_index == 0) {
                result_buf.append("\\\"");
                cur_index = next_quote_index + 1;
            } else {
                result_buf.append(s.substring(cur_index, next_quote_index));
                result_buf.append("\\\"");
                cur_index = next_quote_index + 1;
            };

            if (cur_index > str_end)
                end_reached = true;
        }

        // add the final quote to the result string
        result_buf.append("\"");

        return result_buf.toString();
  }

/* (non-Javadoc)
 * @see acmeLib.properties.AcmePropertyValue#assignType(acmeLib.properties.AcmePropertyType)
 */
public void assignType(AcmePropertyType type) {
	if (type instanceof StringType) setType ((StringType )type);
};

};

