
package acmeLib.properties;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.io.AcmeOutputStream;


/**
 *  Class: EnumType
<BR><BR>
 *
 *  Purpose: the type object for an enumerated type declaration
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *     ?/?/97    DK      Created C++ version
<BR>
 *    6/1/97    RTM     Translated to Java
<BR>
 *    6/19/97   RTM     Added copy() method
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class EnumType extends AcmePropertyType
                      implements acmeLib.designRules.Scope
{

  //
  // private methods
  //
  private AcmeVector vals;

  //
  // public methods
  //
 /**
   * This is the constructor for this class
   * @param name the name of the enum type to be created
   */
  public EnumType (String name) {
    setTypeName(name);
    vals = new AcmeVector();
  };


  /**
   * This method adds a value to the enumType vector
   * @param val the value to be added to the enumType vector
   */
  public void addValue (String val) {
    // RTM: how to deal with attempting to add an enum value multiple times?
    //      currently I just quietly ignore the request to add it.
    //      should this be an exception or error?
    if (!(this.hasValue(val)))
        vals.addItem (new EnumValue (val));
  };

  /**
   * This method checks whether this enumeration type has val
   * @param val the value to be tested for
   * @return whether this enumeration type has val
   */
  public boolean hasValue (String val) {
    // RTM: this is inefficient, and an AcmeMap might be more appropriate,
    //      but AcmeMaps don't allow me to keep the order of the elements,
    //      which might be important to tools, though it is not currently
    //      specified that the ordering has any meaning in Acme.
    AcmeEnumeration e = vals.enumerate();
    EnumValue test_val;
    boolean has_val = false;
    while (e.moreItems() && !has_val) {
        test_val = (EnumValue) e.getNextItem();
        if (test_val.value.equals(val)) has_val = true;
    }
    return has_val;
  };


  /**
   * This method implements the acmeLib.designRules.Scope interface to look up
   * a name in the list of enumerated type names. Return an AcmeEnum
   * (an enum-typed property value) object instead of the internal
   * EnumValue storage object, to facilitate comparison with property
   * values.
   * @param val the name to be looked up
   * @param bubbleUp boolean
   * @return an enum-typed property value
   */
  public Object lookupName(String val, boolean bubbleUp) {
    AcmeEnumeration e = vals.enumerate();
    EnumValue test_val = null;
    while (e.moreItems()) {
        test_val = (EnumValue) e.getNextItem();
        if (test_val.value.equals(val)) {
            return new AcmeEnum(this, test_val.value);
        };
    }
    // there's nowhere to bubble up to, so if we don't find the name
    // just return null.
    return null;
  }

  /**
   * This method returns an enumeration of the values of this enumeration type
   * @return an enumeration of the values of this enumeration type
   */
  public AcmeEnumeration getValues() { return vals.enumerate(); };


  /**
   * This method returns a deep-copy of "this" AcmePropertyType object
   * @return a copy of this enumeration type
   */
  public AcmePropertyType copy() {
    EnumType newType = new EnumType(this.getTypeName());
    this.copyMetaProperties(newType);

    // load the vals list of the original type into the new type/
    AcmeEnumeration e = getValues();
    while (e.moreItems()) {
        EnumValue enum_val = (EnumValue) e.getNextItem();
        newType.addValue(enum_val.value);
    };

    return newType;
  };


  /**
   * @return the enumeration type's IdentifierString
   */
  public String getAcmeIdentifierString() {
    StringBuffer buf = new StringBuffer();
    if (getInlineTypeSpec()) {
        buf.append("Enum {");
        AcmeEnumeration e = this.getValues();
        while (e.moreItems()) {
            buf.append(e.getNextItem().toString());
            if (e.moreItems())
                buf.append(", ");
        };
        buf.append(" }");
    } else if (getTypeName() != null) {
        buf.append(getTypeName());
    } else {
        buf.append("Enum");
    };

    return buf.toString();
  };

  /**
   * This is a helper function that implements "export*" for all *'s
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs while writing
   */
  protected void export(AcmeOutputStream s) throws java.io.IOException {
    exportComment (s);
    s.beginRegion(this);
    if (this.getIsAlias() == true) {
        s.write ("Property Type ");
        s.write (getTypeName() + " = ");
        if (this.getAliasedType() != null) {
            s.writeln(" = " + this.getAliasedType().getTypeName() + ";");
        } else {
            s.writeln(" = Enum{};  /* warning: aliased type with null alias value */");
        };
    } else {
        s.write("Property Type " + getTypeName() + " = Enum");
        AcmeEnumeration e = this.getValues();
        if (!e.moreItems()) {
            s.writeln(";");
        } else {
            s.write(" { ");
            while (e.moreItems()) {
                s.write(e.getNextItem().toString());;
                if (e.moreItems()) s.write (", ");
            }
            s.writeln(" }; ");
        };
    };
    s.endRegion (this);
  };

  /**
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs while writing
   */
  public void exportAcme (AcmeOutputStream s) throws java.io.IOException {
    this.export(s);
  };

  /**
   * @param s AcmeOutputStream
   * @exception java.io.IOException when error occurs while writing
   */
  public void exportArmani(AcmeOutputStream s) throws java.io.IOException {
    this.export(s);
  }


/* (non-Javadoc)
 * @see acmeLib.properties.AcmePropertyType#typecheck(acmeLib.properties.AcmePropertyValue)
 */
public boolean typecheck(AcmePropertyValue val) {
    return true;
}


/* (non-Javadoc)
 * @see acmeLib.properties.AcmePropertyType#getDefinitionAsString()
 */
public String getDefinitionAsString() {
    if (getIsAlias () && getAliasedType() != null) return getAliasedType().getTypeName();
    else {
        String retStr = "Enum";
        AcmeEnumeration e = this.getValues();
        if (e.moreItems ()) {
            retStr += "{";
            while (e.moreItems()) {
                retStr += e.getNextItem ().toString ();
                if (e.moreItems ()) retStr += ", ";
                   
            }   
            retStr += "}";
        }   
        return retStr;
    }  
};


};
