
package acmeLib.properties;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeMap;
import acmeLib.core.io.AcmeOutputStream;


/**
 *  Class:  RecordType<BR><BR>
 *
 *  Purpose:the type object for a record <BR><BR>
 *
 *  Revision History:<BR>
 *    Date:     Person    Change<BR>
 *    ?/?/97    DK      Created C++ version<BR>
 *    6/1/97    RTM     Translated to Java<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class RecordType extends AcmePropertyType
                        implements acmeLib.designRules.Scope
{

    //
    // private members
    //
    private AcmeMap fields;

    //
    // public members
    //
 /**
   * This is the constructor for this class
   * @param name the name of the record type being created
   */
    public RecordType (String name) {
        setTypeName(name);
        fields = new AcmeMap();
    };

  /**
   * This method returns the fields of this record
   * @return the fields of this record
   */
    public AcmeEnumeration getFields() { return fields.enumerate(); };

  /**
   * This method returns a specific field of this record
   * @param fieldName the name of the field to be returned
   * @return the field corresponding to fieldName
   */
    public AcmeRecordField getField(String fieldName) {
        return (AcmeRecordField) fields.lookupItem(fieldName);
    };

  /**
   * This method adds a field to a record
   * @param name the name of the field to be added
   * @param t the type of the field to be added
   * @param val the value of the field to be added
   */
	public void addField (String name, AcmePropertyType t, AcmePropertyValue val) {
	    fields.addItem (name, new AcmeRecordField (name, t, val));
    };

  /**
   * This method adds a field to a record
   * @param f the field to be added to the record
   */
	public void addField (AcmeRecordField f) {
    	fields.addItem (f.getName(), f);
    };


    public void removeField (AcmeRecordField f) {
        fields.removeItem(f.getName ());   
    }
    
    public void removeField (String name) {
        fields.removeItem (name);   
    }

  /**
   * This method returns a deep-copy of "this" RecordType object
   * @return a copy of this RecordType object
   */
    public AcmePropertyType copy() {
      RecordType newType = new RecordType(this.getTypeName());

      AcmeEnumeration e = this.getFields();
      while(e.moreItems()) {
        AcmeRecordField f = (AcmeRecordField) e.getNextItem();
        newType.addField(f.copy());
      };

      this.copyMetaProperties(newType);
      return newType;
    };


   /**
    * @return the record type's IdentifierString
    */
    public String getAcmeIdentifierString() {
        StringBuffer buf = new StringBuffer();
        if ((getInlineTypeSpec() == false) || (this.getTypeName() == null)) {
            buf.append(getTypeName());
        } else {
            AcmeEnumeration e = getFields();
            e.sortByCreationOrder();
            AcmeRecordField f;
            AcmePropertyType t;
            buf.append("Record [");
            while (e.moreItems()) {
              f = (AcmeRecordField) e.getNextItem();
              buf.append(f.getName());
              t = f.getType();
              if (t != null) {
                  buf.append(" : " + t.getAcmeIdentifierString());
              };
              buf.append("; ");
            }
            buf.append("]");
        }

        return buf.toString();
    };


  /**
   * Outputs this type to a stream as
   *   property type <name> = record [
   *      field1;
   *      field2;
   *   ];
   * @param s AcmeOutputStream
   * @exception java.io.IOException when an error occurs while writing
   */
	protected void export(AcmeOutputStream s) throws java.io.IOException {
        exportComment (s);
        s.beginRegion (this);
        if (this.getIsAlias() == true) {
            s.write ("Property Type ");
            s.write (getTypeName() + " = ");
            if (this.getAliasedType() != null) {
                s.writeln(" = " + this.getAliasedType().getTypeName() + ";");
            } else {
                s.writeln(" = Record[];  /* warning: aliased type with null alias value */");
            };
        } else {
    	    AcmeEnumeration e = getFields();
            e.sortByCreationOrder();
    	    AcmeRecordField f;

    	    s.write ("Property Type ");
    	    s.write (getTypeName());
    	    s.writeln (" = Record [");
            s.indent();
    	    while (e.moreItems()) {
    		    f = (AcmeRecordField) e.getNextItem();
    		    f.exportAcme(s);
                s.writeln ("; ");
            };
            s.unIndent ();
    	    s.writeln ("];");
        };
        s.endRegion (this);
    };

  /**
   * @param s AcmeOutputStream
   * @exception java.io.IOException when an error occurs while writing
   */
    public void exportAcme (AcmeOutputStream s) throws java.io.IOException {
        this.export(s);
    };

  /**
   * @param s AcmeOutputStream
   * @exception java.io.IOException when an error occurs while writing
   */
    public void exportArmani(AcmeOutputStream s) throws java.io.IOException {
        this.export(s);
    };


    /**
     * This method resolves the property type object for all of this record's fields.
     * @param scope acmeLib.designRules.Scope
     */
    public void resolveTypeReferences(acmeLib.designRules.Scope scope) {
        // iterate through all of the fields, resolving their typenames.
        AcmeEnumeration e = this.getFields();
        AcmeRecordField f = null;
        while (e.moreItems()) {
            f = (AcmeRecordField) e.getNextItem();
            f.resolveTypeReferences(scope);
        };
    };



  /**
   * Implements the acmeLib.designRules.Scope interface to map names
   * to record fields.
   * @param name the name of the property
   * @param bubbleUp boolean
   * @return the Object corresponding to "name"
   */
    public Object lookupName(String name, boolean bubbleUp)
    {
        return getField(name);
    }

/* (non-Javadoc)
 * @see acmeLib.properties.AcmePropertyType#typecheck(acmeLib.properties.AcmePropertyValue)
 */
public boolean typecheck(AcmePropertyValue val) {
    return true;
}

/* (non-Javadoc)
 * @see acmeLib.properties.AcmePropertyType#getDefinitionAsString()
 */
public String getDefinitionAsString() {
    if (getIsAlias() && getAliasedType() != null) return getAliasedType().getTypeName();
    else {          
        AcmeEnumeration e = getFields();
        e.sortByCreationOrder();
        AcmeRecordField f;

        String retStr = "Record [";
        while (e.moreItems()) {
            f = (AcmeRecordField) e.getNextItem();
            retStr += f.toString () + "; ";
        }
        retStr += "]";
        return retStr;
        
    }
};

};
