package acmeLib.properties;

import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmeFamily;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.TypeRefResolutionError;

/**
 *  Class:  SequenceType
<BR><BR>
 *
 *  Purpose:the type object for a sequence type declaration
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    ?/?/97    DK      Created C++ version
<BR>
 *    6/1/97    RTM     Translated to Java
<BR>
 *    6/19/97   RTM     Added copy() method
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class SequenceType extends AcmePropertyType {

	//
	// private members
	//
	private AcmePropertyType type;

	//
	// public members
	//
	/**
	  * This is the constructor for this class
	  * @param name the name of the sequence type being created
	  */
	public SequenceType(String name) {
		setTypeName(name);
		this.setSequenceType(ConstantTypes.UNSPECIFIED_TYPE);
	};

	/**
	   * This is the constructor for this class
	   * @param name the name of the sequence type being created
	   * @param type the type of the sequence type being created
	   */
	public SequenceType(String name, AcmePropertyType type) {
		if (name == null || name == "")
			name = "sequence<" + type.getTypeName() + ">";
		setTypeName(name);
		this.type = type;
	};

	/**
	 * This method returns the type of this sequence type
	 * @return the type of this sequence type
	 */
	public AcmePropertyType getSequenceType() {
		return this.type;
	};
	/**
	 * This method sets the type of this sequence type
	 * @param newType the value to be assigned to type
	 */
	public void setSequenceType(AcmePropertyType newType) {
		this.type = newType;
	};

	/**
	 * This method returns a deep-copy of "this" AcmePropertyType object
	 * @return a copy of this SequenceType object
	 */
	public AcmePropertyType copy() {
		SequenceType newType = new SequenceType(this.getTypeName(), this.getSequenceType().copy());
		this.copyMetaProperties(newType);
		return newType;
	};

	/**
	 * @return the sequence type's IdentifierString
	 */
	public String getAcmeIdentifierString() {
		StringBuffer buf = new StringBuffer();
		if (getInlineTypeSpec()) {
			buf.append("Sequence<");
			if (type instanceof AcmePropertyType) {
				buf.append(((AcmePropertyType) type).getAcmeIdentifierString());
			}
			else if (type instanceof AcmeElementType) {
				buf.append(type.getTypeName());
			};
			buf.append(">");
		}
		else if (getTypeName() != null) {
			buf.append(getTypeName());
		}
		else {
			buf.append("Sequence<>");
		};

		return buf.toString();
	};

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs while writing
	 */
	protected void export(AcmeOutputStream s) throws java.io.IOException {
		exportComment(s);
		s.beginRegion(this);
		if (this.getIsAlias() == true) {
			s.write("Property Type ");
			s.write(getTypeName() + " = ");
			if (this.getAliasedType() != null) {
				s.writeln(" = " + this.getAliasedType().getTypeName() + ";");
			}
			else {
				s.writeln(" = Sequence<>;  /* warning: aliased type with null alias value */");
			};
		}
		else {
			s.write("Property Type ");
			s.write(getTypeName());
			s.write(" = Sequence<");

			if (this.getSequenceType() instanceof AcmePropertyType) {
				s.write(((AcmePropertyType) this.getSequenceType()).getAcmeIdentifierString());
			}
			else if (this.getSequenceType() instanceof AcmeElementType) {
				s.write(this.getSequenceType().getTypeName());
			};

			s.writeln(">;");
		};
		s.endRegion(this);
	};

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs while writing
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException {
		this.export(s);
	};

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs while writing
	 */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException {
		this.export(s);
	};

	/**
	 * This method resolves the property type object that this sequence property type
	 * stores (e.g. resolves "foo" of a sequence<foo> type definition)
	 * @param scope acmeLib.designRules.Scope
	 */
	public void resolveTypeReferences(acmeLib.designRules.Scope scope) {
		String lookupError = null;
		if (this.getSequenceType() instanceof AcmeUnresolvedPropertyType) {
			Object foundObject = null;
			if (scope == null) {
				// if scope is null then we can't resolve names,
				// announce an error and return.
				lookupError = "Unable to resolve sequence property typename: " + this.getSequenceType().getTypeName();
				TypeRefResolutionError badScopeError =
					new TypeRefResolutionError(this, lookupError, this.getSequenceType().getTypeName());
				badScopeError.setUnresolvedRef(this);
				AcmeErrorManager.errorDetected(badScopeError);
				return;
			}
			else {
				if (this.getSequenceType().getQualifier() != null) {
					Object o = scope.lookupName(getSequenceType().getQualifier(), true);
					if (o instanceof AcmeFamily) {
						foundObject =
							((AcmeFamily) o).getType(
								((AcmeUnresolvedPropertyType) getSequenceType()).getUnqualifiedTypeName());
						if (foundObject instanceof AcmePropertyType) {
							foundObject = new AcmePropertyTypeRef((AcmeFamily) o, (AcmePropertyType) foundObject);
						}
					}
				}
				else {
					foundObject = scope.lookupName(this.getSequenceType().getTypeName(), true);
				}

			};

			if (foundObject instanceof AcmePropertyType) {
				this.setSequenceType((AcmePropertyType) foundObject);
			}
			else {
				// Announce an error and return
				lookupError = "Unable to resolve sequence property typename: " + this.getSequenceType().getTypeName();
				TypeRefResolutionError notPropertyTypeError =
					new TypeRefResolutionError(this, lookupError, this.getSequenceType().getTypeName());
				notPropertyTypeError.setUnresolvedRef(this);
				AcmeErrorManager.errorDetected(notPropertyTypeError);
				AcmePropertyType unresolvedType = new AcmeUnresolvedPropertyType(this.getSequenceType().getTypeName());
				this.setSequenceType(unresolvedType);
				return;
			};
		};

		// now that we have a resolved property type, call resolve names
		// on the property type object, to resolve any unresolved names
		// in complex property types.
		if (this.getSequenceType() != null) {
			this.getSequenceType().resolveTypeReferences(scope);
		}
		else {
			// Announce an error and return
			lookupError = "Unable to resolve sequence property typename: " + this.getSequenceType().getTypeName();
			TypeRefResolutionError notPropertyTypeError =
				new TypeRefResolutionError(this, lookupError, this.getSequenceType().getTypeName());
			notPropertyTypeError.setUnresolvedRef(this);
			AcmeErrorManager.errorDetected(notPropertyTypeError);
			AcmePropertyType unresolvedType = new AcmeUnresolvedPropertyType(this.getSequenceType().getTypeName());
			this.setSequenceType(unresolvedType);
			return;
		};
	};

	/* (non-Javadoc)
	 * @see acmeLib.AcmeType#getTypeName()
	 */
	public String getTypeName() {
		String name = super.getTypeName();
		if (name == null) {
			name = "sequence <" + this.getSequenceType().getTypeName() + ">";
		}
		return name;
	}

	/* (non-Javadoc)
	 * @see acmeLib.properties.AcmePropertyType#typecheck(acmeLib.properties.AcmePropertyValue)
	 */
	public boolean typecheck(AcmePropertyValue val) {
		return true;
	}

	/* (non-Javadoc)
	 * @see acmeLib.properties.AcmePropertyType#getDefinitionAsString()
	 */
	public String getDefinitionAsString() {
		if (getIsAlias() && getAliasedType() != null)
			return getAliasedType().getTypeName();
		else {
			String retStr = "Sequence<";

			if (this.getSequenceType() instanceof AcmePropertyType) {
				retStr += ((AcmePropertyType) this.getSequenceType()).getAcmeIdentifierString();
			}
			else if (this.getSequenceType() instanceof AcmeElementType) {
				retStr += this.getSequenceType().getTypeName();
			};
			return retStr;
		}
	}

};
