package acmeLib.properties;

import acmeLib.AcmeType;
import acmeLib.AcmeUnresolvedType;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.elements.AcmeFamily;
import acmeLib.elements.AcmeUnresolvedElementType;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.TypeRefResolutionError;

/**
 *  Class: SetType
<BR><BR>
 *
 *  Purpose:the type object for a set type declaration
<BR><BR>
 *
 *  Revision History:
<BR>
 *    Date:     Person    Change
<BR>
 *    6/3/97    RTM     Created in Java
<BR>
 *    6/19/97   RTM     Added copy() method
<BR><BR>
 *
 *  @author Bob Monroe<BR>
 */
public class SetType extends AcmePropertyType {

	//
	// protected members
	//
	protected AcmeType type;

	//
	// public members
	//
	/**
	  * This is the constructor for this class
	  * @param name the name of the set type being created
	  */
	public SetType(String name) {
		setTypeName(name);
		this.setSetType(ConstantTypes.UNSPECIFIED_TYPE);
	};

	/**
	  * This is the constructor for this class
	  * @param name the name of the set type being created
	  * @param type the type of the set type being created
	  */
	public SetType(String name, AcmeType type) {
		if (name == null || name == "")
			name = "set<" + type.getTypeName() + ">";
		setTypeName(name);
		this.type = type;
	};

	/**
	 * This method returns the type of this set type
	 * @return the type of this set type
	 */
	public AcmeType getSetType() {
		return this.type;
	};
	/**
	  * This method sets the type of this set type
	  * @param t the value to be assigned to type
	  */
	public void setSetType(AcmeType t) {
		this.type = t;
	};

	/**
	* This method returns a deep-copy of "this" AcmePropertyType object
	* @return a copy of this SetType object
	*/
	public AcmePropertyType copy() {
		SetType newType;
		if (this.getSetType() instanceof AcmePropertyType)
			newType = new SetType(this.getTypeName(), ((AcmePropertyType) this.getSetType()).copy());
		else if (this.getSetType() instanceof AcmeElementType)
			newType = new SetType(this.getTypeName(), ((AcmeElementType) this.getSetType()).copyType());
		else {
			// else we have an unknown type of type, return an unresolved type
			// with the appropriate name.
			return new AcmeUnresolvedPropertyType(this.getTypeName());
		};

		this.copyMetaProperties(newType);
		return newType;
	};

	/**
	 * @return the set type's IdentifierString
	 */
	public String getAcmeIdentifierString() {
		StringBuffer buf = new StringBuffer();
		if (getInlineTypeSpec()) {
			buf.append("Set{");
			if (type instanceof AcmePropertyType) {
				buf.append(((AcmePropertyType) type).getAcmeIdentifierString());
			}
			else if (type instanceof AcmeElementType) {
				buf.append(type.getTypeName());
			};
			buf.append("}");
		}
		else if (getTypeName() != null) {
			buf.append(getTypeName());
		}
		else {
			buf.append("Set{}");
		};

		return buf.toString();
	};

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs while writing
	 */
	protected void export(AcmeOutputStream s) throws java.io.IOException {
		exportComment(s);
		s.beginRegion(this);
		if (this.getIsAlias() == true) {
			s.write("Property Type ");
			s.write(getTypeName() + " = ");
			if (this.getAliasedType() != null) {
				s.writeln(" = " + this.getAliasedType().getTypeName() + ";");
			}
			else {
				s.writeln(" = Set{};  /* warning: aliased type with null alias value */");
			};
		}
		else {
			s.write("Property Type ");
			s.write(getTypeName());
			s.write(" = Set{");

			if (this.getSetType() instanceof AcmePropertyType) {
				s.write(((AcmePropertyType) this.getSetType()).getAcmeIdentifierString());
			}
			else if (this.getSetType() instanceof AcmeElementType) {
				s.write(this.getSetType().getTypeName());
			};

			s.writeln("};");
		};
		s.endRegion(this);
	};

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs while writing
	 */
	public void exportAcme(AcmeOutputStream s) throws java.io.IOException {
		this.export(s);
	};

	/**
	 * @param s AcmeOutputStream
	 * @exception java.io.IOException when error occurs while writing
	 */
	public void exportArmani(AcmeOutputStream s) throws java.io.IOException {
		this.export(s);
	};

	/**
	 * This method resolves the type object that this set property type
	 * stores (e.g. resolves "foo" of a set{foo} type definition)
	 * @param scope acmeLib.designRules.Scope
	 */
	public void resolveTypeReferences(acmeLib.designRules.Scope scope) {
		String lookupError = null;
		if ((this.getSetType() instanceof AcmeUnresolvedPropertyType)
			|| (this.getSetType() instanceof AcmeUnresolvedElementType)
			|| (this.getSetType() instanceof AcmeUnresolvedType)) {
			Object foundObject = null;
			if (scope == null) {
				// if scope is null then we can't resolve names,
				// announce an error and return.
				lookupError = "Unable to resolve set contents typename: " + this.getSetType().getTypeName();
				TypeRefResolutionError badScopeError =
					new TypeRefResolutionError(this, lookupError, this.getSetType().getTypeName());
				badScopeError.setUnresolvedRef(this);
				AcmeErrorManager.errorDetected(badScopeError);
				return;
			}
			else {
				String q = null;
				if (getSetType() instanceof AcmeUnresolvedPropertyType) {
					q = ((AcmeUnresolvedPropertyType) getSetType()).getQualifier();
				}
				else if (getSetType() instanceof AcmeUnresolvedElementType) {
					q = ((AcmeUnresolvedElementType) getSetType()).getQualifier();
				}
				if (q != null) {
					Object o = scope.lookupName(q, true);
					if (o instanceof AcmeFamily) {
						foundObject =
							((AcmeFamily) o).getType(
								((AcmeUnresolvedPropertyType) getSetType()).getUnqualifiedTypeName());
						if (foundObject instanceof AcmePropertyType) {
							foundObject = new AcmePropertyTypeRef((AcmeFamily) o, (AcmePropertyType) foundObject);
						}
					}
				}
				else {
					foundObject = scope.lookupName(this.getSetType().getTypeName(), true);
				}
			};

			if (foundObject instanceof AcmeType) {
				this.setSetType((AcmeType) foundObject);
			}
			else {
				// Announce an error and return
				lookupError = "Unable to resolve set contents typename: " + this.getSetType().getTypeName();
				TypeRefResolutionError notPropertyTypeError =
					new TypeRefResolutionError(this, lookupError, this.getSetType().getTypeName());
				notPropertyTypeError.setUnresolvedRef(this);
				AcmeErrorManager.errorDetected(notPropertyTypeError);
				AcmePropertyType unresolvedType = new AcmeUnresolvedPropertyType(this.getSetType().getTypeName());
				this.setSetType(unresolvedType);
				return;
			};
		}

		// now that we have a resolved set contents type, call resolve names
		// on the property or element type object, to resolve any unresolved
		// names in complex property types.
		if (this.getSetType() != null) {
			if (this.getSetType() instanceof AcmeElementType) {
				((AcmeElement) this.getSetType()).resolveTypeReferences();
			}
			else if (this.getSetType() instanceof AcmePropertyType) {
				((AcmePropertyType) this.getSetType()).resolveTypeReferences(scope);
			};
		}
		else {
			// Announce an error and return
			lookupError = "Unable to resolve set contents typename: " + this.getSetType().getTypeName();
			TypeRefResolutionError notPropertyTypeError =
				new TypeRefResolutionError(this, lookupError, this.getSetType().getTypeName());
			notPropertyTypeError.setUnresolvedRef(this);
			AcmeErrorManager.errorDetected(notPropertyTypeError);
			AcmePropertyType unresolvedType = new AcmeUnresolvedPropertyType(this.getSetType().getTypeName());
			this.setSetType(unresolvedType);
			return;
		};
	};

	/* (non-Javadoc)
	 * @see acmeLib.AcmeType#getTypeName()
	 */
	public String getTypeName() {
		if (this.typeName == null) {
			return "set {" + this.getSetType().getTypeName() + "}";
		}
		return super.getTypeName();
	}

	/* (non-Javadoc)
	 * @see acmeLib.properties.AcmePropertyType#typecheck(acmeLib.properties.AcmePropertyValue)
	 */
	public boolean typecheck(AcmePropertyValue val) {
		return true;
	}

	/* (non-Javadoc)
	 * @see acmeLib.properties.AcmePropertyType#getDefinitionAsString()
	 */
	public String getDefinitionAsString() {
		if (getIsAlias() && getAliasedType() != null)
			return getAliasedType().getTypeName();
		else {
			String retStr = "Set{";

			if (this.getSetType() instanceof AcmePropertyType) {
				retStr += ((AcmePropertyType) this.getSetType()).getAcmeIdentifierString();
			}
			else if (this.getSetType() instanceof AcmeElementType) {
				retStr += this.getSetType().getTypeName();
			};
			return retStr;
		}
	}

};
