package acmeLib.representations;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeElementType;
import acmeLib.properties.AcmeProperty;

/**
 * Class: AcmeBinding<br>
 * Purpose: Stores a "binding" from one element to another element<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    ?/?/97    DK      Created C++ version<br>
 *    6/05/97   RTM     Translated to Java (stubs)<br>
 *    6/20/97   RTM     converted from Binding to AcmeBinding.  updated to<br>
 *                       work with simpler, generic, representation model<br>
 *<br>
 * @author Bob Monroe
 */

public class AcmeBinding extends AcmeElement
                         implements acmeLib.designRules.Scope,
                                    acmeLib.core.io.Exportable,
                                    acmeLib.core.io.ArmaniExportable
{

    //
    // protected members
    //
    protected AcmeElement inner = null;
    protected AcmeElement outer = null;
    protected AcmeRepresentation parentRep = null;

  /**
   * flags that can be set to suppress the exporting of qualifiers
   * these are used by the parser and unparser to make the exported
   * Acme closer to the imported Acme.  default to true.
   */
    protected boolean exportInnerQualifier = true;
    protected boolean exportOuterQualifier = true;


    //
    // public members
    //
  /**
   * This is the constructor for this class
   * @param outer AcmeElement
   * @param inner AcmeElement
   */
    public AcmeBinding (AcmeElement outer, AcmeElement inner) {
        super();
    	this.setInner(inner);
	    this.setOuter(outer);
    };

  /**
   * Set the inner AcmeElement to the new AcmeElement
   * @param inner_binding The new inner AcmeElement
   */
    public void setInner (AcmeElement inner_binding) {
	this.inner = inner_binding;
    };

  /**
   * Set the outer AcmeElement to the new AcmeElement
   * @param outer_binding The new outer AcmeElement
   */
	public void setOuter (AcmeElement outer_binding) {
	    this.outer = outer_binding;
	};

  /**
   * Return the inner AcmeElement
   * @return The inner AcmeElement
   */
	public AcmeElement getInner() { return this.inner; };
  /**
   * Return the outer AcmeElement
   * @return The outerAcmeElement
   */
	public AcmeElement getOuter() { return this.outer; };

  /**
   * Return the exportInnerQualifier
   * @return ExportInnerQualifier
   */
    public boolean getExportInnerQualifier() { return this.exportInnerQualifier; };
  /**
   * Set the exportInnerQualifier to the new exportInnerQualifier
   * @param exportIt The new exportInnerQualifier
   */
    public void setExportInnerQualifier(boolean exportIt) {
        this.exportInnerQualifier = exportIt;
    };

  /**
   * Return the exportOuterQualifier
   * @return ExportOuterQualifier
   */
    public boolean getExportOuterQualifier() { return this.exportOuterQualifier; };
  /**
   * Set the exportOuterQualifier to the new exportOuterQualifier
   * @param exportIt The new exportOuterQualifier
   */
    public void setExportOuterQualifier(boolean exportIt) {
        this.exportOuterQualifier = exportIt;
    };

  /**
   * Return the parent AcmeRepresentation
   * @return AcmeRepresentation
 */
    public AcmeRepresentation getParentRep() { return this.parentRep; };
  /**
   * Set the parent AcmeRepresentation to the new AcmeRepresentation
   * @param newRep The new AcmePresentation
   */
    public void setParentRep(AcmeRepresentation newRep) {
        this.parentRep = newRep;
    };

  /**
   * Return a copy of this AcmeBinding
   * @return a copy of this AcmeBinding
   */
    public AcmeElement copy() {
        AcmeBinding newBinding =
                        new AcmeBinding(this.getOuter(), this.getInner());
        newBinding.setParentRep(this.getParentRep());
        newBinding.setExportInnerQualifier(this.getExportInnerQualifier());
        newBinding.setExportOuterQualifier(this.getExportOuterQualifier());
        this.copyProperties(newBinding);
        return newBinding;
    };

  /**
   * Make a copy of this AcmeBinding as AcmeUnresolvedBinding
   * @return a copy of AcmeUnresolvedBinding
   */
    public AcmeUnresolvedBinding copyAsUnresolvedBinding() {
        AcmeUnresolvedBinding newBinding = new AcmeUnresolvedBinding(this);
        newBinding.setParentRep(this.getParentRep());
        this.copyProperties(newBinding);
        return newBinding;
    };


  /**
   * Test to see if the test object is equal to this AcmeBinding
   * @param testObj Object to be tested
   * @return true if equal
   */
    public boolean equals(Object testObj) {
        AcmeBinding testBinding = null;
        if (testObj != null && testObj instanceof AcmeBinding) {
            testBinding = (AcmeBinding) testObj;
        } else {
            return false;
        };

        boolean eqInner = false;
        boolean eqOuter = false;
        if (testBinding.inner == null || this.inner == null) {
            // compare refernce equality
            if (testBinding.inner == this.inner) {
                eqInner = true;
            }
        } else {
            eqInner = testBinding.inner.getQualifiedName().equals (this.inner.getQualifiedName ());
        }
        if (testBinding.outer == null || this.outer == null) {
            // compare reference equality
            if (testBinding.outer == this.outer) {
                eqOuter = true;
            }
        } else {
            eqOuter = testBinding.outer.getQualifiedName ().equals (this.outer.getQualifiedName ());
        }

        return eqInner && eqOuter;
    }


  /**
   * export information of this AcmeBinding
   * @param s AcmeOutputStream
   * @param exportLand ExportLanguage
   * @exception java.io.IOException For I/O exception
   */
	protected void export (AcmeOutputStream s, ExportLanguage exportLang)
	    throws java.io.IOException
    {
    exportComment (s);
    s.beginRegion (this);
	if (this.getExportInnerQualifier ()) {
	    s.write (inner.getParent ().getName ());
	    s.write (".");
	}
	s.write (inner.getName ());
	s.write (" to ");

	    if (!(outer.getParent() instanceof AcmeElementType)
	          && (this.getExportOuterQualifier()))
        {
    	    s.write (outer.getParent().getName());
    	    s.write (".");
    	};
	    s.write (outer.getName());
 //    	s.write (" to ");

//         if (this.getExportInnerQualifier()) {
//     	    s.write (inner.getParent().getName());
//         	s.write (".");
//         };
// 	    s.write (inner.getName());

    	// export properties

	    if (this.propMap.getSize()==0) {
		    s.writeln (";");
    		return;
	    } else {
        	AcmeEnumeration e = this.getProperties();
        	e.sortByCreationOrder();
        	AcmeProperty p;

    	    s.writeln ("{");
        	s.indent();
    	    while (e.moreItems()) {
	    	    s.indent();
    		    p = (AcmeProperty) e.getNextItem();
    		    if (exportLang == ExportLanguage.ARMANI) {
            		p.exportArmani(s);
            	} else {
            	    // export Acme by default.
            	    p.exportAcme(s);
            	};
    	    	s.writeln ("");
	    	    s.unIndent();
        	};
	        s.unIndent();
        	s.writeln ("};");
	    };
        s.endRegion (this);
    };


  /**
   * export Acme
   * @param s AcmeOutputStream
   * @exception java.io.IOException For I/O exception
   */
	public void exportAcme (AcmeOutputStream s) throws java.io.IOException {
        this.export(s, ExportLanguage.ACME);
    };

  /**
   * export Armani
   * @param s AcmeOutputStream
   * @exception java.io.IOException For I/O exception
   */
	public void exportArmani (AcmeOutputStream s) throws java.io.IOException {
        this.export(s, ExportLanguage.ARMANI);
    };


  /**
   * Resolve type references for this AcmeBinding
   */
    public void resolveTypeReferences() {
        AcmeEnumeration e = this.getProperties();
        while (e.moreItems()) {
            ((AcmeProperty) e.getNextItem()).resolveTypeReferences();
        };
    };


  /**
   * this is a fully resolved AcmeBinding, so do nothing.  Any
   * subclasses that hold unresolved bindings should override this method
   */
    public void resolveInstanceReferences() {
    };



  /**
   * implement the acmeLib.designRules.Scope interface.
   * @param name name of the object
   * @param bubbleUp Goes up to the parent Representation to find the object
   * @return the object in the propMap
   */
    public Object lookupName(String name, boolean bubbleUp) {
        Object retObj = propMap.lookupItem(name);

        if ((retObj == null) && (bubbleUp)) {
            if (this.getParentRep() != null) {
                retObj = this.getParentRep().lookupName(name, bubbleUp);
            };
        };

        return retObj;
    };

};


