package acmeLib.representations;

import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.filters.AcmeObjectFilter;
import acmeLib.core.globals.ExportLanguage;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmePort;
import acmeLib.elements.AcmeRole;

/**
 * Class: BindingsMap<br>
 * Purpose: special kind of abstraction map that holds name bindings<br>
 *          between pairs of ports, or pairs of roles.<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    ?/?/97    DK      Created C++ version<br>
 *    6/05/97   RTM     Translated to Java (stubs)<br>
 *<br>
 * @author Bob Monroe
 */

public class AcmeBindingsMap extends AcmeAbstractionMap {

    //
    // private members
    //
	private AcmeVector bindings;

    //
    // public members
    //
  /**
   * This is the constructor for this class
   */
    public AcmeBindingsMap() {
        bindings = new AcmeVector();
    }

  /**
   * Add an Acme binding
   * @param b the new AcmeBinding
   */
	public void addBinding (AcmeBinding b) {
	    // maintain set-semantics for the bindings map.
	    if (!bindings.contains(b)) {
    	    bindings.addItem (b);
        };
    };

  /**
   * Add an new AcmeBinding
   * @param outer_port outer port for the new binding
   * @param inner_port inner port for the new binding
   */
	public void addBinding (AcmePort outer_port, AcmePort inner_port) {
	    // maintain set-semantics for the bindings map.
	    AcmeBinding newBinding = new AcmeBinding (outer_port, inner_port);
        if (!bindings.contains(newBinding)) {
        	bindings.addItem(newBinding);
        };

    };

  /**
   * Add an new AcmeBinding
   * @param outer_role outer role for the new binding
   * @param inner_role inner role for the new binding
   */
	public void addBinding (AcmeRole outer_role, AcmeRole inner_role) {
	    // maintain set-semantics for the bindings map.
	    AcmeBinding newBinding = new AcmeBinding (outer_role, inner_role);
        if (!bindings.contains(newBinding)) {
        	bindings.addItem(newBinding);
        };
    };


  /**
   * Remove the Acme binding b, if b is in stored in this bindings map.  Quietly
   * return without changing the map if the binding b is not an element of
   * this bindings map.
   * @param b the new AcmeBinding
   */
	public void removeBinding (AcmeBinding b) {
	    bindings.removeItem(b);
    };

  /**
   * Remove the AcmeBinding that binds outer_port to inner_port from this
   * bindings map.  If this map doesn't contain such a binding then quietly
   * return successfully.
   * @param outer_port outer port for the new binding
   * @param inner_port inner port for the new binding
   */
	public void removeBinding (AcmePort outer_port, AcmePort inner_port) {
        AcmeEnumeration e = this.getBindings();
        AcmeBinding curBinding = null;
        AcmeBinding bindingToRemove = getBinding(outer_port, inner_port);

        if (bindingToRemove != null) {
            this.removeBinding(bindingToRemove);
        };
    };


  /**
   * Remove the AcmeBinding that binds outer_role to inner_role from this
   * bindings map.  If this map doesn't contain such a binding then quietly
   * return successfully.
   * @param outer_role outer role for the new binding
   * @param inner_role inner role for the new binding
   */
	public void removeBinding (AcmeRole outer_role, AcmeRole inner_role) {
        AcmeEnumeration e = this.getBindings();
        AcmeBinding curBinding = null;
        AcmeBinding bindingToRemove = getBinding(outer_role, inner_role);

        if (bindingToRemove != null) {
            this.removeBinding(bindingToRemove);
        };
    };



  /**
   * replace the oldBinding in position i with a new Binding, maintaining
   * set semantics for the set of all bindings.
   * @param oldBinding the old binding to be replaced
   * @param newBinding the new binding
   */
    public void replaceBinding (AcmeBinding oldBinding, AcmeBinding newBinding)
    {
        int i = bindings.indexOf(oldBinding);
        if (i > -1) {
            bindings.setItemAt(newBinding, i);
        } else {
            // else oldBinding isn't in this binding map, just add the newBinding
            bindings.addItem(newBinding);
        };
    }

  /**
   * Get the bindings of this binding map
   * @return AcmeEnumeration of the bindings
   */
    public AcmeBinding getBinding (AcmeElement outer, AcmeElement inner)  {
        AcmeEnumeration e = getBindings();
        AcmeBinding curBinding = null;
        AcmeBinding foundBinding = null;
        while (e.moreItems() && (foundBinding == null)) {
            curBinding = (AcmeBinding) e.getNextItem();
            if ((curBinding.getInner() == inner)
                && (curBinding.getOuter() == outer)) {

                    foundBinding = curBinding;
            };
        };
        return foundBinding;
    };

  /**
   * Get the bindings of this binding map
   * @return AcmeEnumeration of the bindings
   */
	public AcmeEnumeration getBindings()  {
	    return bindings.enumerate();
    };
    
	/**
	 * @param port
	 * @return
	 */
	public AcmeEnumeration getBindings(final AcmeElement port) {
		
		AcmeObjectFilter filter = new AcmeObjectFilter () {

			public boolean test(AcmeObject obj) {
				if (obj instanceof AcmeBinding) {
					AcmeBinding binding = (AcmeBinding) obj;
					return binding.getInner ().equals (port) ||
						binding.getOuter ().equals (port);
				}
				return false;
			}
			
		};
		AcmeEnumeration e = bindings.enumerate(filter);
		return e;
	}

  /** Get the bindings
   * @param e The AcmeEnumeration
   * @return AcmeEnumeration
   */
	public AcmeEnumeration getBindings(AcmeEnumeration e)  {
	    return bindings.enumerate(e);
    };

  /**
   * copy all of the bindings in this bindings map, returning a new map object
   * @return new map object
   */
    public AcmeAbstractionMap copy() {
        AcmeBindingsMap newMap = new AcmeBindingsMap();
        for (int i = 0; i < this.bindings.getSize(); i++) {
            newMap.addBinding((AcmeBinding)
                              ((AcmeBinding) bindings.getItemAt(i)).copy());
        };
        return newMap;
    };



  /**
   * make a new bindings map that is filled with unresolved bindings
   * names.  The abstraction map will need to be resolved subsequently
   * for it to be generally useful.
   * @return new map object
   */
    public AcmeAbstractionMap copyAsUnresolvedAbsMap() {
        AcmeBindingsMap newMap = new AcmeBindingsMap();
        for (int i = 0; i < this.bindings.getSize(); i++) {
            newMap.addBinding(new AcmeUnresolvedBinding(
                                    (AcmeBinding) bindings.getItemAt(i)));
        };
        return newMap;
    };


  /**
   * reset all of the bindings stored in this map so that they have their
   * parentRep fields point to newParentRep.
   * @param newParentRep The new parent AcmeRepresentation
   */
    public void setParentRep(AcmeRepresentation newParentRep) {
        for (int i = 0; i < this.bindings.getSize(); i++) {
            AcmeBinding curBinding = (AcmeBinding) bindings.getItemAt(i);
            curBinding.setParentRep(newParentRep);
        };
    };


  /**
   * export the information about this binding map
   * @param s AcmeOutputStream
   * @param exportLang ExportLanguage
   * @exception java.io.IOException For I/O exception
   */
    protected void export(AcmeOutputStream s, ExportLanguage exportLang)
        throws java.io.IOException
    {
        s.beginRegion (this);
        if (bindings.getSize() == 0)
    	    s.writeln ("Bindings {};");
        else {
            s.writeln ("Bindings {");
    	    s.indent();
            AcmeBinding cur_binding;
            AcmeEnumeration e = bindings.enumerate();
            e.sortByCreationOrder();
            while (e.moreItems()) {
                cur_binding = (AcmeBinding) e.getNextItem();
                if (exportLang == ExportLanguage.ARMANI) {
                    cur_binding.exportArmani(s);
                } else {
                    // export Acme by default...
                    cur_binding.exportAcme(s);
                };
            };
	        s.unIndent();
    	    s.writeln ("}");
    	};
        s.endRegion (this);
    };


  /**
   * export Acme
   * @param s AcmeOutputStream
   * @exception java.io.IOException For I/O exception
   */
    public void exportAcme(AcmeOutputStream s) throws java.io.IOException {
        this.export(s, ExportLanguage.ACME);
    };

  /**
   * export Armani
   * @param s AcmeOutputStream
   * @exception java.io.IOException For I/O exception
   */
    public void exportArmani(AcmeOutputStream s) throws java.io.IOException {
        this.export(s, ExportLanguage.ARMANI);
    };

  /**
   * Resolve the type references
   */
    public void resolveTypeReferences() {
        AcmeEnumeration e = this.getBindings();
        while (e.moreItems()) {
            ((AcmeBinding) e.getNextItem()).resolveTypeReferences();
        };
    };



  /**
   * resolve instance references
   */
    public void resolveInstanceReferences() {
        AcmeEnumeration e = this.getBindings();
        while (e.moreItems()) {
            ((AcmeBinding) e.getNextItem()).resolveInstanceReferences();
        };

        // clean up the resolved bindings...
        this.removeDuplicates();
    };


  /**
   * Walk over the set of bindings, removing any duplicates that may have
   * been added through various stages of instantiation and name resolution...
   */
    public void removeDuplicates() {
        AcmeBinding curBinding = null;
        for (int i = 0; i < this.bindings.getSize(); i++) {
            curBinding = (AcmeBinding) this.bindings.getItemAt(i);
            for (int j = i + 1; j < this.bindings.getSize(); j++) {
                if (curBinding.equals(this.bindings.getItemAt(j))) {
                    this.bindings.removeItemAt(j);
                    j--;  // decrement to account for "left-shifted" vector.
                }
            };
        };
    }


};
