package acmeLib.representations;

import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeException;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmeSystem;


/**
 * Class: AcmeRepresentation<br>
 * Purpose: base class ACME representations<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    ?/?/97    DK      Created C++ version<br>
 *    4/24/97   RTM     Translated to Java (stubs)<br>
 *    8/6/97    RTM     removed deprecated getCategoryName method<br>
 *<br>
 *@author Bob Monroe
 */

public class AcmeRepresentation extends acmeLib.AcmeObject
                                implements acmeLib.designRules.Scope,
                                           acmeLib.core.io.Exportable,
                                           acmeLib.core.io.ArmaniExportable
{

  //
  // protected members
  //
  protected AcmeSystem systemDesc;
  protected AcmeElement parent;
  protected AcmeAbstractionMap absMap;
  protected String repName;

/**
 * flag to indicate rep is anonymous.  if a rep has no explicit name it
 * is given the name of its system, so that it can be stored by name. If
 * it is anonymous, though, the name will not be written out on export.
 * this is a bit ugly.
 */
  protected boolean repNameIsAnonymous = false;
private AcmeVector inheritedFrom;


  //
  // public members
  //


/**
 * This is the constructor for this class
 */
  public AcmeRepresentation() {
    systemDesc = null;
    parent = null;
    absMap = null;
    repName = null;
  };

/**
 * This is the constructor for this class
 * @param desc AcmeSystem
 * @param parent The parent AcmeElement
 * @param new_map The new AcmeAbstractionMap
 * @param name The name of the representation
 */
  public AcmeRepresentation (AcmeSystem desc, AcmeElement parent,
                             AcmeAbstractionMap new_map, String name)
  {
    this.parent = parent;
    this.setSystem(desc);
    absMap = new_map;
    if (name != null) {
        this.repName = name;
    } else if (desc != null) {
        this.repName = desc.getName();
        repNameIsAnonymous = true;
    };
  };


/**
 * Get the system
 * @return AcmeSystem
 */
  public AcmeSystem getSystem() { return systemDesc; };

/**
 * Set the system to the new AcmeSystem
 * @param sys The new AcmeSystem
 */
  public void setSystem (AcmeSystem sys) {
    this.systemDesc = sys;
    if (sys != null) {
        sys.setParentRepresentation(this);
        sys.setParent(this.getParent());
    };
  };


/**
 * Get the name of this representation
 * @return the name of this representation
 */
  public String getName() { return this.repName; };


/**
 * set the name of this representation
 * @param The new name of this representation
 * @exception AcmeIllegalNameException If the name is illegal
 */
  public void setName(String newName) throws AcmeIllegalNameException
  {
    String oldName = this.repName;
    this.repName = newName;
    if (this.getParent() != null) {
        this.getParent().childNameChanged(oldName, this);
    };
  };



  /**
   * Retrieves the fully qualified, period delimited, name of the AcmeElement.
   * This name identifies an object in the structure tree that can be reached
   * by taking a unique path from the root design.
   * @return the fully qualified name of the AcmeElement
   **/
  public String getQualifiedName() {
    StringBuffer nameBuf = new StringBuffer();

    if(this.getParent() != null) {
        String parentName = this.getParent().getQualifiedName();
        if (parentName != null) {
            if (getRepNameIsAnonymous() == false) {
                if (!(parentName.equals(""))) {
                    nameBuf.append(parentName + ".");
                };
                nameBuf.append(this.getName());
            } else {
                nameBuf.append(parentName);
            };
        };
    } else if (getRepNameIsAnonymous() == false) {
        nameBuf.append(this.getName());
    };

    return nameBuf.toString();
  };





/**
 *  The only named child of a representation is it's system.  If this
 *  changes we need to update the name of the representation itself, if
 *  the rep has an anonymous name.
 *  @param oldName string with the system's old name
 *  @param obj the child object whose name has changed.
 *  @exception AcmeIllegalNameException
 */
  public void childNameChanged(String oldName, AcmeObject obj)
        throws AcmeIllegalNameException
  {
    if (this.getRepNameIsAnonymous() && (obj instanceof AcmeSystem)) {
        this.setName(((AcmeSystem) obj).getName());
    };
  };


/**
 * Check if representation name is anonymous
 * @return true if anonymous
 */
  public boolean getRepNameIsAnonymous() { return this.repNameIsAnonymous; };
/**
 * Set the repNameIsAnonymous for this representation
 * @param repIsAnonymous true if anonymous
 */
  public void setRepNameIsAnonymous(boolean repIsAnonymous) {
    this.repNameIsAnonymous = repIsAnonymous;
  };

/**
 * Get the parent of this Representation
 * @return the parent
 */
  public AcmeElement getParent() { return parent; };
/**
 * Set the parent of this Representation
 * @param p The new parent
 */
  public void setParent (AcmeElement p) { parent = p; };

/**
 * Get the AcmeAbstractionMap of this representation
 * @return The AcmeAbstractionMap
 */
  public AcmeAbstractionMap getAbstractionMap() { return this.absMap; };
/**
 * Set the AcmeAbstraction to a new one
 * @param map The new AcmeAbstractionMap
 */
  public void setAbstractionMap(AcmeAbstractionMap map) {
    this.absMap = map;
  };


/**
 * Get the scope of this representation
 * @return the scope
 */
  public acmeLib.designRules.Scope getScope() {
    return this.getParent();
  };


  /**
   * returns a deep-copy of this representation, pointing to the same parent
   * @return The copy of this AcmeRepresentation
   **/
  public AcmeRepresentation copy() {
    AcmeSystem new_sys_description = null;
    AcmeAbstractionMap new_map = null;

    if (this.getSystem() != null)
        new_sys_description = (AcmeSystem) this.getSystem().copy();

    if (this.getAbstractionMap() != null)
        new_map = this.getAbstractionMap().copy();

    AcmeRepresentation new_rep =
        new AcmeRepresentation(new_sys_description, this.getParent(),
                               new_map, this.getName());

    if (new_sys_description != null) {
        new_sys_description.setParentRepresentation(new_rep);
    };

    if (new_map != null)
        new_map.setParentRep(new_rep);

    new_rep.setRepNameIsAnonymous(this.getRepNameIsAnonymous());

    return new_rep;
  };



/**
 * Export information in Acme
 * @param s AcmeOutputStream
 * @exception java.io.IOException For I/O exception
 * @exception AcmeException For AcmeExcpetion
 */
  public void exportAcme(AcmeOutputStream s)
    throws java.io.IOException, AcmeException
  {
    if (this.getName() != null) {
        s.writeln("Representation " + this.getName() + " = {");
    } else {
        s.writeln("Representation {");
    };

    s.indent();
    if (this.getSystem() == null)
        s.writeln(" System __no_system_description__ = {}; ");
    else {
	    this.getSystem().exportAcme(s);
    }

    // export the abstraction map
    if (this.getAbstractionMap() != null)
        this.getAbstractionMap().exportAcme(s);

    s.unIndent();
    s.writeln("};");
  };


/**
 * Export information in Armani
 * @param s AcmeOutputStream
 * @exception java.io.IOException For I/O exception
 * @exception AcmeException For AcmeException
 */
  public void exportArmani(AcmeOutputStream s)
    throws java.io.IOException, AcmeException
  {
    exportComment (s);
    s.beginRegion (this);
    if (this.getName() != null) {
        s.writeln("Representation " + this.getName() + " = {");
    } else {
        s.writeln("Representation {");
    };

    s.indent();
    if (this.getSystem() == null)
        s.writeln(" System __no_system_description__ = {}; ");
    else {
	    this.getSystem().exportArmani(s);
    }

    // export the abstraction map
    if (this.getAbstractionMap() != null)
        this.getAbstractionMap().exportArmani(s);

    s.unIndent();
    s.writeln("};");
    s.endRegion (this);
  };



/**
 * implement the acmeLib.designRules.Scope interface.
 * Representations only look to see if the name is the name of the
 * representation or the system used in this representation (since
 * bindings don't have names).  If it is the system's or representation's
 * name it returns the system, otherwise it bubbles the search up to this
 * rep's parent, if the bubbleUp flag is true.
 * @param name name of the object
 * @param bubbleUp Goes up to the parent Representation to find the object
 * @return The object
 */
  public Object lookupName(String name, boolean bubbleUp) {

    if (this.getSystem() != null) {
        if (this.getSystem().getName() != null)
            if (this.getSystem().getName().equalsIgnoreCase(name))
                return this.getSystem();
    };

    if (this.getName() != null) {
        if (this.getName().equals(name)) {
            return this;
        };
    };


    if ((bubbleUp) && (this.getParent() != null))
        return this.getParent().lookupName(name, bubbleUp);
    else
        return null;
  };


/**
 * do all of the post-parsing processing of the design that is supposed
 * to happen after parsing, in preparation for, but not including,
 * typechecking.
 */
  public void doPostParsingProcessing()
  {
    this.resolveTypeReferences();
    this.instantiateElements();
    this.resolveInstanceReferences();
  };



/**
 * resolve all unresolved typenames in this rep's system and that system's
 * substructure.
 */
  public void resolveTypeReferences() {

    if (this.getSystem() != null) {
        this.getSystem().resolveTypeReferences();
    };

    if (this.getAbstractionMap() != null) {
        this.getAbstractionMap().resolveTypeReferences();
    };

  };


/**
 * resolve all unresolved instance references in this rep's system, that
 * system's substructure, and the rep's bindings.
 */
  public void resolveInstanceReferences() {
    if (this.getSystem() != null) {
        this.getSystem().resolveInstanceReferences();
    };

    // walk through and resolve all of the unresolved bindings in this
    // rep's abstraction map (it's probably a BindingsMap object in Armani)
    if (this.getAbstractionMap() != null)
        this.getAbstractionMap().resolveInstanceReferences();
  };



/**
 * instantiateElements() simply passes the request to this rep's system.
 */
  public void instantiateElements() {
    if (this.getSystem() != null)
        this.getSystem().instantiateElements();
  }

/**
 * @param protoRep
 */
public void addInheritedSource(AcmeRepresentation protoRep) {
	if (inheritedFrom == null) 
		inheritedFrom = new AcmeVector ();
	if (!inheritedFrom.contains (protoRep)) {
		inheritedFrom.addItem (protoRep);
	}
};


};

