package acmeLib.representations;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeElement;
import acmeLib.elements.AcmePort;
import acmeLib.elements.AcmeRole;
import acmeLib.errorManagement.AcmeError;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.InstanceRefResolutionError;
import acmeLib.properties.AcmeProperty;

/**
 * Class: AcmeUnresolvedBinding<br>
 * Purpose: Class that represents a binding within an Acme representation<br>
 *          whose attached ports or roles are named but have not yet been<br>
 *          resolved to point to an actual pair of AcmePort or AcmeRole objects.<br>
 *          This is a temporary class to hold an binding during parsing<br>
 *          but before the instance references within a system are resolved.<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    1/24/98   RTM      Created initial file.<br>
 *<br>
 * @author Bob Monroe
 */

public class AcmeUnresolvedBinding extends AcmeBinding
                                   implements acmeLib.core.io.Exportable,
                                              acmeLib.core.io.ArmaniExportable

{

  //
  // Protected members
  //
  protected String firstQualifier, firstId;
  protected String secondQualifier, secondId;

  //
  // Public members
  //
/**
 * This is the constructor for this class
 * @param firstQualifier The first qualifier
 * @param firstId The first ID
 * @param secondQualifier The second qualifier
 * @param secondId The second ID
 */
  public AcmeUnresolvedBinding (String firstQualifier, String firstId,
                                String secondQualifier, String secondId)
  {
     super(null, null);
     this.firstQualifier = firstQualifier;
     this.firstId = firstId;
     this.secondQualifier = secondQualifier;
     this.secondId = secondId;
  };


/**
 * copy constructor to create a new unresolved binding from either
 * a resolved binding or an unresolved binding.
 * @param oldBinding The old binding
 */
  public AcmeUnresolvedBinding (AcmeBinding oldBinding) {
    super(null, null);
    if (oldBinding == null)
        return;

    if (oldBinding instanceof AcmeUnresolvedBinding) {
        this.firstQualifier = ((AcmeUnresolvedBinding) oldBinding).firstQualifier;
        this.secondQualifier = ((AcmeUnresolvedBinding) oldBinding).secondQualifier;
        this.firstId = ((AcmeUnresolvedBinding) oldBinding).firstId;
        this.secondId = ((AcmeUnresolvedBinding) oldBinding).secondId;
    } else {
        // make the first qualifier null and give first id the name of the
        // outer element the old binding is bound to.  This way the name
        // of the outer port/role is specified, but it can be the port/role
        // of whatever element happens to contain the appropriate port/role
        // when this binding's instance names are resolved.
        this.firstQualifier = null;
        AcmeElement oldOuter = oldBinding.getOuter();
        if (oldOuter != null) {
            this.firstId = oldOuter.getName();
        };

        // the second qualifier and id refer to the inner element.  These
        // have to be associated properly with both the containing component
        // or connector and the specific port/role, so set both of their
        // values.
        AcmeElement oldInner = oldBinding.getInner();
        if (oldInner != null) {
            if (oldInner.getParent() != null)
                this.secondQualifier = oldInner.getParent().getName();
            this.secondId = oldInner.getName();
        }
    }

    // copy the old binding's properties to the newly created binding.
    oldBinding.copyProperties(this);

    // explicitly do not set the parent or parentDesign of the new unresolved
    // binding in the constructor, the client creating an instance with
    // the copy constructor needs to set this appropriately.
  }



/**
 * Copy this binding
 * @return The copy of this binding
 */
  public AcmeElement copy() {
    AcmeUnresolvedBinding newBinding =
        new AcmeUnresolvedBinding(this.firstQualifier, this.firstId,
                                  this.secondQualifier, this.secondId);
    this.copyProperties(newBinding);
    return newBinding;
  };


/**
 * equality checks only that the two bindings refer to the same (port,port)
 * or (role,role) pairs.  It doesn't consider properties because its relation
 * is the unique-identification for a binding.
 * @param test The object to be tested
 * @return true if the objects are equal
 */
  public boolean equals(Object test) {
    AcmeUnresolvedBinding testBinding;
    if (test instanceof AcmeUnresolvedBinding) {
        testBinding = (AcmeUnresolvedBinding) test;
    } else {
        return false;
    };

    boolean areEqualAsIs = true;
    boolean areEqualFlipped = true;

    // first try the case where firstRef(this) == firstRef(testBinding) and
    // secondRef(this) == secondRef(that), then try the case where
    // firstRef(this) == secondRef(testBinding) and secondRef(this) == firstRef(testBinding).
    // either of these cases should describe equality.
    if (this.firstQualifier == null) {
        if (testBinding.firstQualifier != null) {
            areEqualAsIs = false;
        };
    } else if (!this.firstQualifier.equals(testBinding.firstQualifier)) {
        areEqualAsIs = false;
    };

    if (this.firstId == null) {
        if (testBinding.firstId != null) {
            areEqualAsIs = false;
        };
    } else if (!this.firstId.equals(testBinding.firstId)) {
        areEqualAsIs = false;
    };

    if (this.secondQualifier == null) {
        if (testBinding.secondQualifier != null) {
            areEqualAsIs = false;
        };
    } else if (!this.secondQualifier.equals(testBinding.secondQualifier)) {
        areEqualAsIs = false;
    };

    if (this.secondId == null) {
        if (testBinding.secondId != null) {
            areEqualAsIs = false;
        };
    } else if (!this.secondId.equals(testBinding.secondId)) {
        areEqualAsIs = false;
    };


    // try the names flipped
    if (this.firstQualifier == null) {
        if (testBinding.secondQualifier != null) {
            areEqualFlipped = false;
        };
    } else if (!this.firstQualifier.equals(testBinding.secondQualifier)) {
        areEqualFlipped = false;
    };

    if (this.firstId == null) {
        if (testBinding.secondId != null) {
            areEqualFlipped = false;
        };
    } else if (!this.firstId.equals(testBinding.secondId)) {
        areEqualFlipped = false;
    };

    if (this.secondQualifier == null) {
        if (testBinding.firstQualifier != null) {
            areEqualFlipped = false;
        };
    } else if (!this.secondQualifier.equals(testBinding.firstQualifier)) {
        areEqualFlipped = false;
    };

    if (this.secondId == null) {
        if (testBinding.firstId != null) {
            areEqualFlipped = false;
        };
    } else if (!this.secondId.equals(testBinding.firstId)) {
        areEqualFlipped = false;
    };

    return (areEqualAsIs || areEqualFlipped);
  };


/**
 * Make a copy of this binding as unresolved binding
 * @return AcmeUnresolvedBinding
 */
  public AcmeUnresolvedBinding copyAsUnresolvedBinding() {
    return new AcmeUnresolvedBinding(this);
  };


/**
 * Output string for the binding
 * @return The string
 */
  public String toString() {
    StringBuffer buf = new StringBuffer();
    if (firstQualifier != null) buf.append (firstQualifier + ".");
    buf.append (firstId + " to ");
    if (secondQualifier != null) buf.append (secondQualifier + ".");
    buf.append(secondId);
//    buf.append("{ property resolved : boolean = false; ");
    AcmeEnumeration e = this.getProperties();
    if (e.getSize () > 1) {
    	buf.append (" { ");
	    
	    e.sortByCreationOrder();
	    while (e.moreItems()) {
	        AcmeProperty curProp = (AcmeProperty) e.getNextItem();
	        try {
	            AcmeOutputStream s = new AcmeOutputStream();
	            curProp.exportAcme(s);
	            s.flush();
	            buf.append(s.toString());
	        } catch (java.io.IOException ioExc) {
	            buf.append("/* error, caught java.io.IOException exporting binding property */");
	        };
	    };
	    buf.append(" }");
    }
    buf.append (";");

    return buf.toString();
  };

/**
 * export Acme
 * @param s AcmeOutputStream
 * @exception java.io.IOException For I/O exception
 */
  public void exportAcme (AcmeOutputStream s) throws java.io.IOException {
  	s.writeln(this.toString());
  };

/**
 * export Armani
 * @param s AcmeOutputStream
 * @exception java.io.IOException For I/O exception
 */
  public void exportArmani (AcmeOutputStream s) throws java.io.IOException {
      s.beginRegion (this);
  	s.writeln(this.toString());
    s.endRegion (this);
  };


/**
 * resolve the names of the instances that this binding refers to.
 */
  public void resolveInstanceReferences() {
    AcmeElement innerEltSystem = null;
    AcmeElement outerEltParent = null;
    AcmeElement innerEltParent = null;
    AcmeElement outerElt = null;
    AcmeElement innerElt = null;

    // first look up the parent rep, the system that we are hunting in,
    // and the element that is the parent of this binding's rep.
    outerEltParent = this.getParentRep().getParent();
    innerEltSystem = this.getParentRep().getSystem();

    if ((outerEltParent == null) || (innerEltSystem == null)) {
        // unable to resolve without properly specified parents, reps,
        // systems, etc.  do nothing, fall through to the end where
        // errors are handled.
    } else if ((firstQualifier == null) && (secondQualifier == null)) {
        // unable to resolve without any qualifiers.  do nothing but
        // fall through to the end where errors are handled.
    } else if ((firstQualifier != null) && (secondQualifier != null)) {
        // find out which of the qualifiers refers to outer element and
        // which refers to inner
        if (firstQualifier.equalsIgnoreCase(outerEltParent.getName())) {
            // then the first item is the outer one, lookup its port or role
            outerElt = (AcmeElement) outerEltParent.lookupName(firstId, false);

            // the second qualifier and id is the inner one, look it up
            innerEltParent = (AcmeElement) innerEltSystem.lookupName(secondQualifier, false);
            if (innerEltParent != null)
                innerElt = (AcmeElement) innerEltParent.lookupName(secondId, false);

        } else {
            // else first qualifier must refer to the inner element
            innerEltParent = (AcmeElement) innerEltSystem.lookupName(firstQualifier, false);
            if (innerEltParent != null)
                innerElt = (AcmeElement) innerEltParent.lookupName(firstId, false);

            if (secondQualifier.equalsIgnoreCase(outerEltParent.getName()))
                outerElt = (AcmeElement) outerEltParent.lookupName(secondId, false);
        };

    } else if ((firstQualifier == null) && (secondQualifier != null)) {
        // then the (implicit) first qualifier must refer to the outer parent
        // element and the second qualifier refers to a component or connector
        // in the representation system.
        outerElt = (AcmeElement) outerEltParent.lookupName(firstId, false);
        this.setExportOuterQualifier(false);

        // the second qualifier and id must be the inner one, look it up.
        innerEltParent = (AcmeElement) innerEltSystem.lookupName(secondQualifier, false);
        if (innerEltParent != null)
            innerElt = (AcmeElement) innerEltParent.lookupName(secondId, false);

    } else if ((firstQualifier != null) && (secondQualifier == null)) {
        // then the (implicit) second qualifier must refer to the outer parent
        // element and the first qualifier refers to a component or connector
        // in the representation system.
        outerElt = (AcmeElement) outerEltParent.lookupName(secondId, false);
        this.setExportOuterQualifier(false);

        // the first qualifier and id must be the inner one, look it up.
        innerEltParent = (AcmeElement) innerEltSystem.lookupName(firstQualifier, false);
        if (innerEltParent != null)
            innerElt = (AcmeElement) innerEltParent.lookupName(firstId, false);
    };

    // if this is a valid binding we will now have innerElt and outerElt set.
    // if either of these is null raise an error and don't create a
    // new binding.  If they are both non-null check that they are either
    // both ports or both roles.  If that is satisfied then create a new
    // resolved binding, add it to the parent representation, and remove
    // this unresolvedBinding from the parent rep's set of bindings
    if ((outerElt == null) || (innerElt == null)) {
        InstanceRefResolutionError innerRefError =
            new InstanceRefResolutionError(this.getParentRep(),
                    "unable to resolve binding " + this.toString(),
                    this.toString());
        innerRefError.setUnresolvedRef(this);
        AcmeErrorManager.errorDetected(innerRefError);
    } else if (((innerElt instanceof AcmePort) && (outerElt instanceof AcmePort))
        || ((innerElt instanceof AcmeRole) && (outerElt instanceof AcmeRole)))
    {
            // create the new binding and copy this' bindings info,
            // properties, and creationOrder into it.
            AcmeBinding newBinding = new AcmeBinding(outerElt, innerElt);
            newBinding.setParentRep(this.getParentRep());
            newBinding.setExportOuterQualifier(this.getExportOuterQualifier());
            newBinding.setExportInnerQualifier(this.getExportInnerQualifier());
            newBinding.setCreationOrder(this.getCreationOrder());

            // reuse this unresolved binding's property map, but reset the
            // scope of all of the toplevel properties.
            AcmeEnumeration props = this.getProperties();
            while (props.moreItems()) {
                AcmeProperty curProp = (AcmeProperty) props.getNextItem();
                curProp.setScope(newBinding);
                try { newBinding.addProperty(curProp); }
                catch (AcmeIllegalNameException aknExc) {
                    // do nothing, its a copied property so it shouldn't
                    // have a keyword that was previously accepted...
                };
            };

            // and replace this with the new binding in the parent rep.
            AcmeAbstractionMap absMap = this.getParentRep().getAbstractionMap();
            if (absMap instanceof AcmeBindingsMap)
                ((AcmeBindingsMap) absMap).replaceBinding(this, newBinding);
            else {
                // else this representation doesn't have bindings.  very strange.
                String repName = "<Unknown Rep>";
                if (this.getParentRep() != null)
                    repName = this.getParentRep().getName();
                AcmeError noBindingsMapError = new AcmeError(this,
                    "Internal error -- representation : " + repName
                    + " has no bindings map to add AcmeUnresolvedBinding <"
                    + this.toString() + "> to.");
            };
    } else {
        // throw an error, we can't make a valid binding from this
        // unresolved binding.
        InstanceRefResolutionError typeMismatchError =
            new InstanceRefResolutionError(this.getParentRep(),
                    "unable to resolve binding " + this.toString()
                    + " because the resolved elements are neither a port/port"
                    + " nor a role/role pair.", this.toString());
        typeMismatchError.setUnresolvedRef(this);
        AcmeErrorManager.errorDetected(typeMismatchError);
    };

  };  // end resolveInstanceReferences()

};
