/*
 * Created on Apr 12, 2005
 */
package acmeLib.tools;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

/**
 * A custom ClassLoader to allow user-defined external analyses to be loaded
 * properly.  This loader overcomes the problem of multiple classpath space
 * in Eclipse.
 *   
 * @author Shang-Wen Cheng (zensoul@cs.cmu.edu)
 */
public class ExternalAnalysesClassLoader extends ClassLoader {
    private String _path;

	/**
	 * @param parent  Parent class loader
	 */
	public ExternalAnalysesClassLoader(ClassLoader parent) {
        super(parent);
	}

    public void setPath (String path) {
        if (path == null)
            throw new IllegalArgumentException("Null external analysis path");

        _path = path;
    }
    public String getPath () {
        return _path;
    }

	protected Class findClass(String name)
	throws ClassNotFoundException {
        Class c = null;

        // Convert class name argument to filename
		// Convert package names into subdirectories
		String filename = name.replace('.', File.separatorChar) + ".class";

		try {
			// Load class data from file and save in byte array
            byte data[] = loadClassData(filename);

			// Convert byte array to Class
            c = defineClass(name, data, 0, data.length);

			// If failed, throw exception
            if (c == null) {
                throw new ClassNotFoundException(name);
            }
		} catch (IOException e) {
			throw new ClassNotFoundException("Error reading file: "
					+ filename);
		}

		// Return class just created
		return c;
	}

	private byte[] loadClassData(String filename) throws IOException {

		// Create a file object relative to directory provided
		File f = new File(_path, filename);

		// Get size of class file
		int size = (int) f.length();

		// Reserve space to read
		byte buff[] = new byte[size];

		// Get stream to read from
		FileInputStream fis = new FileInputStream(f);
		DataInputStream dis = new DataInputStream(fis);

		// Read in data
		dis.readFully(buff);

		// close stream
		dis.close();

		// return data
		return buff;
	}

}
