// Class: GraphWalkers
// Purpose: defines a set of functions for walking over a graph of AcmeElements
//          and possibly their properties.  These functions can be parameterized
//          by passing objects that implement the various Walker*Operation
//          interfaces to perform an operation at each step in the walk.  This
//          is basically an aid to Armani and Acme tool writers.
//
// Revision History:
//    Date:     Person     Change
//    5/26/98    RTM        * created initial class

package acmeLib.tools;

import acmeLib.AcmeObject;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.elements.*;
import acmeLib.representations.AcmeAbstractionMap;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;

public class GraphWalker {

    //
    // protected methods can be defined as helpers for the public methods.
    //

    protected static void walkOverCommonElts (AcmeElement startElement, WalkerElementOperation op, int walkOrder) {
		AcmeEnumeration e = startElement.getRepresentations();
		while(e.moreItems()) {
			AcmeRepresentation rep = (AcmeRepresentation) e.getNextItem();
			if (rep.getSystem() != null) {
				WalkOverAllElements(rep.getSystem(), op, walkOrder);
				AcmeAbstractionMap amap = rep.getAbstractionMap();
				if (amap instanceof AcmeBindingsMap) {
					AcmeBindingsMap bmap = (AcmeBindingsMap) amap;
					AcmeEnumeration bindings = bmap.getBindings();
					while (bindings.moreItems()) {
						AcmeBinding binding = (AcmeBinding )bindings.getNextItem();
						op.doElementOp(binding);
					}
					
				}
			};
		};    	
    }


    //
    // class constants
    //

    // constants that specify how the structure tree should be traversed
    // (bottom-up or top-down).  With bottom-up, children nodes are processed
    // before their parents, with top-down parents are processed before their
    // children.
    public static final int WALK_BOTTOM_UP = 1;
    public static final int WALK_TOP_DOWN  = 2;


    //
    // public static methods.
    //

    /** Walk over a component and its substructure */
    public static void WalkOverAllElements(AcmeComponent startComponent,
                                           WalkerElementOperation op,
                                           int walkOrder)
    {
        if ((startComponent == null) || (op == null)) {
            return;
        };

        if (walkOrder == WALK_TOP_DOWN) {
            op.doElementOp(startComponent);
        };

        AcmeEnumeration e = startComponent.getPorts();
        while(e.moreItems()) {
            AcmePort childPort = (AcmePort) e.getNextItem();
            WalkOverAllElements(childPort, op, walkOrder);
        };

		walkOverCommonElts(startComponent, op, walkOrder);

        if (walkOrder == WALK_BOTTOM_UP) {
            op.doElementOp(startComponent);
        };
    };


    /** Walk over a connector and its substructure */
    public static void WalkOverAllElements(AcmeConnector startConnector,
                                           WalkerElementOperation op,
                                           int walkOrder)
    {
        if ((startConnector == null) || (op == null)) {
            return;
        };

        if (walkOrder == WALK_TOP_DOWN) {
            op.doElementOp(startConnector);
        };

        AcmeEnumeration e = startConnector.getRoles();
        while(e.moreItems()) {
            AcmeRole childRole = (AcmeRole) e.getNextItem();
            WalkOverAllElements(childRole, op, walkOrder);
        };

		walkOverCommonElts(startConnector, op, walkOrder);

        if (walkOrder == WALK_BOTTOM_UP) {
            op.doElementOp(startConnector);
        };
    };


    /** Walk over a role and its substructure */
    public static void WalkOverAllElements(AcmeRole startRole,
                                           WalkerElementOperation op,
                                           int walkOrder)
    {
        if ((startRole == null) || (op == null)) {
            return;
        };

        if (walkOrder == WALK_TOP_DOWN) {
            op.doElementOp(startRole);
        };

		walkOverCommonElts(startRole, op, walkOrder);


        if (walkOrder == WALK_BOTTOM_UP) {
            op.doElementOp(startRole);
        };
    };



    /** Walk over a port and its substructure */
    public static void WalkOverAllElements(AcmePort startPort,
                                           WalkerElementOperation op,
                                           int walkOrder)
    {
        if ((startPort == null) || (op == null)) {
            return;
        };

        if (walkOrder == WALK_TOP_DOWN) {
            op.doElementOp(startPort);
        };

		walkOverCommonElts(startPort, op, walkOrder);

        if (walkOrder == WALK_BOTTOM_UP) {
            op.doElementOp(startPort);
        };
    };



    /** Walk over a system and its substructure */
    public static void WalkOverAllElements(AcmeSystem startSystem,
                                           WalkerElementOperation op,
                                           int walkOrder)
    {
        if ((startSystem == null) || (op == null)) {
            return;
        };

        if (walkOrder == WALK_TOP_DOWN) {
            op.doElementOp(startSystem);
        };

        AcmeEnumeration e = startSystem.getComponents();
        while(e.moreItems()) {
            AcmeComponent childComp = (AcmeComponent) e.getNextItem();
            WalkOverAllElements(childComp, op, walkOrder);
        };

        e = startSystem.getConnectors();
        while(e.moreItems()) {
            AcmeConnector childConn = (AcmeConnector) e.getNextItem();
            WalkOverAllElements(childConn, op, walkOrder);
        };

        e = startSystem.getRoles();
        while(e.moreItems()) {
            AcmeRole childRole = (AcmeRole) e.getNextItem();
            WalkOverAllElements(childRole, op, walkOrder);
        };

        e = startSystem.getPorts();
        while(e.moreItems()) {
            AcmePort childPort = (AcmePort) e.getNextItem();
            WalkOverAllElements(childPort, op, walkOrder);
        };

		walkOverCommonElts(startSystem, op, walkOrder);


        if (walkOrder == WALK_BOTTOM_UP) {
            op.doElementOp(startSystem);
        };
    };



    /** Walk over a family, its types, analysis, and substructure */
    public static void WalkOverAllElements(AcmeFamily startFamily,
                                           WalkerElementOperation op,
                                           int walkOrder)
    {
        if ((startFamily == null) || (op == null)) {
            return;
        };

        if (walkOrder == WALK_TOP_DOWN) {
            op.doElementOp(startFamily);
        };

        // first walk over all of the element types defined in the family
        AcmeEnumeration e = startFamily.getTypes();
        while (e.moreItems()) {
            AcmeObject nextItem = e.getNextItem();
            if (nextItem instanceof AcmeComponent) {
                WalkOverAllElements((AcmeComponent) nextItem, op, walkOrder);
            } else if (nextItem instanceof AcmeConnector) {
                WalkOverAllElements((AcmeConnector) nextItem, op, walkOrder);
            } else if (nextItem instanceof AcmePort) {
                WalkOverAllElements((AcmePort) nextItem, op, walkOrder);
            } else if (nextItem instanceof AcmeRole) {
                WalkOverAllElements((AcmeRole) nextItem, op, walkOrder);
            };
        };

        // walk over the (default) structure defined in this family
        e = startFamily.getComponents();
        while(e.moreItems()) {
            AcmeComponent childComp = (AcmeComponent) e.getNextItem();
            WalkOverAllElements(childComp, op, walkOrder);
        };

        e = startFamily.getConnectors();
        while(e.moreItems()) {
            AcmeConnector childConn = (AcmeConnector) e.getNextItem();
            WalkOverAllElements(childConn, op, walkOrder);
        };

        e = startFamily.getRoles();
        while(e.moreItems()) {
            AcmeRole childRole = (AcmeRole) e.getNextItem();
            WalkOverAllElements(childRole, op, walkOrder);
        };

        e = startFamily.getPorts();
        while(e.moreItems()) {
            AcmePort childPort = (AcmePort) e.getNextItem();
            WalkOverAllElements(childPort, op, walkOrder);
        };

        e = startFamily.getRepresentations();
        while(e.moreItems()) {
            AcmeRepresentation rep = (AcmeRepresentation) e.getNextItem();
            if (rep.getSystem() != null) {
                WalkOverAllElements(rep.getSystem(), op, walkOrder);
            };
        };


        if (walkOrder == WALK_BOTTOM_UP) {
            op.doElementOp(startFamily);
        };
    };



    /** Walk over a design and its substructure */
    public static void WalkOverAllElements(AcmeDesign startDesign,
                                           WalkerElementOperation op,
                                           int walkOrder)
    {
        if ((startDesign == null) || (op == null)) {
            return;
        };

        // for top-down operation look at imported designs designs first...
        if (walkOrder == WALK_TOP_DOWN) {
            // first do our operation on this design
            op.doElementOp(startDesign);

            // then walk over all of the imported designs before walking down
            // this design's tree.
            int numDesigns = startDesign.getImportedDesigns().getSize();
            for (int i = 0; i < numDesigns; i++) {
                AcmeImportedDesignRecord curDesignRec =
                    (AcmeImportedDesignRecord) startDesign.getImportedDesigns().getItemAt(i);
                if (curDesignRec.getResolvedDesign() != null) {
                    WalkOverAllElements(curDesignRec.getResolvedDesign(), op, walkOrder);
                } else {
                    // error?  just quietly ignore it for now...
                };
            };
        };

        // call walkOverAllElements recursively on all types defined in
        // this design.  Only walk over element types that have been fully
        // resolved (that is, no property types or unresolved references).
        AcmeEnumeration e = startDesign.getTypes();
        while (e.moreItems()) {
            AcmeObject nextItem = e.getNextItem();
            if (nextItem instanceof AcmeComponent) {
                WalkOverAllElements((AcmeComponent) nextItem, op, walkOrder);
            } else if (nextItem instanceof AcmeConnector) {
                WalkOverAllElements((AcmeConnector) nextItem, op, walkOrder);
            } else if (nextItem instanceof AcmePort) {
                WalkOverAllElements((AcmePort) nextItem, op, walkOrder);
            } else if (nextItem instanceof AcmeRole) {
                WalkOverAllElements((AcmeRole) nextItem, op, walkOrder);
            } else if (nextItem instanceof AcmeFamily) {
                WalkOverAllElements((AcmeFamily) nextItem, op, walkOrder);
            };
        };


        // call walkOverAllElements recursively on all families defined in
        // this design.  Only walk over families that have been fully
        // resolved (that is, no property types or unresolved references).
        e = startDesign.getFamilies();
        while (e.moreItems()) {
            AcmeObject nextItem = e.getNextItem();
            if (nextItem instanceof AcmeFamily) {
                WalkOverAllElements((AcmeFamily) nextItem, op, walkOrder);
            };
        };

        // call walkOverAllElements recursively on all systems defined in
        // this design.
        e = startDesign.getSystems();
        while (e.moreItems()) {
            WalkOverAllElements((AcmeSystem) e.getNextItem(), op, walkOrder);
        };

        if (walkOrder == WALK_BOTTOM_UP) {
            // do our operation on this design
            op.doElementOp(startDesign);
            // then walk over all of the imported designs before walking up the
            // design trees of the imported designs.
            int numDesigns = startDesign.getImportedDesigns().getSize();
            for (int i = 0; i < numDesigns; i++) {
                AcmeImportedDesignRecord curDesignRec =
                    (AcmeImportedDesignRecord) startDesign.getImportedDesigns().getItemAt(i);
                if (curDesignRec.getResolvedDesign() != null) {
                    WalkOverAllElements(curDesignRec.getResolvedDesign(), op, walkOrder);
                } else {
                    // error?  just quietly ignore it for now...
                };
            };
        };

    };

};


