/**
 *  Class: Acme2Armani
 *
 *  Purpose: Class to convert Acme code with Armani design rules and design
 *           analyses appropriately encoded as properties into an Armani
 *           design that can be typechecked and constraint-checked by the
 *           Armani design checker.
 *
 *  Revision History: <BR>
 *    Date:     Person    Change <BR>
 *    6/15/98   RTM        Created initial class <BR>
 *
 *  @author Bob Monroe
 */

package acmeLib.tools.converters;

import java.io.FileOutputStream;

import acmeLib.acmeParser.AcmeParser;
import acmeLib.armaniParser.ArmaniParser;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeDesign;
import acmeLib.tools.GraphWalker;

public class Acme2Armani {

    //
    // Protected class members
    //

    /** pause processing until "enter" is pressed on the keyboard */
    protected static void pause_for_input() {
        System.out.println("Please press enter to continue");
        try {
            System.in.read();
        } catch (java.io.IOException e) {
            return;
        }
    };

    //
    // Public class members
    //

    /** Main reads in an acme file, converts it to Armani, and writes it out
     *  as an Armani file.
     **/
    public static void main(String args[]) {
        String ParserID = new String("Acme2Armani translator (release 0.1)");
        AcmeParser parser;
        java.io.OutputStream outputFileStream = null;
        java.io.InputStream inputFileStream = null;
        String canonicalInputFilename = null;

        if (args.length == 0) {
            System.out.println(ParserID + ":  Reading from standard input . . .");
            inputFileStream = System.in;
            canonicalInputFilename = "System.in";
            outputFileStream = System.out;
        } else if (args.length == 1) {
            System.out.println(ParserID + ":  Reading from file " +
                             args[0] + " . . .");
          try {
            inputFileStream = new java.io.FileInputStream(args[0]);
            java.io.File inputFile = new java.io.File(args[0]);
            try {
                canonicalInputFilename = inputFile.getCanonicalPath();
            } catch (java.io.IOException ioExc) {
                canonicalInputFilename = args[0];
            };
            outputFileStream = System.out;
          } catch (java.io.FileNotFoundException e) {
                System.out.println(ParserID + ":  File " + args[0] + " not accessible.");
                return;
          }
        } else if (args.length == 2) {
            System.out.println(ParserID + ":  Reading from file "
                               + args[0] + " . . .\n"
                               + "Writing to file " + args[1] + " . . .");
            try {
               inputFileStream = new java.io.FileInputStream(args[0]);
               java.io.File inputFile = new java.io.File(args[0]);
               try {
                    canonicalInputFilename = inputFile.getCanonicalPath();
               } catch (java.io.IOException ioExc) {
                    canonicalInputFilename = args[0];
               };
               outputFileStream = new FileOutputStream(args[1]);
            } catch (java.io.IOException e) {
                System.out.println(ParserID + ":  File " + args[1] + " not accessible.");
                return;
            }

        } else {
            System.out.println("Acme2Armani translator (release 0.1): " +
                             "Usage is one of:");
            System.out.println("         java Acme2Armani");
            System.out.println("            -- reads Acme from stdin and writes Armani to stdout");
            System.out.println("OR");
            System.out.println("         java Acme2Armani <inputfile>");
            System.out.println("            -- reads Acme from <inputfile> and writes Armani to stdout");
            System.out.println("OR");
            System.out.println("         java Acme2Armani <inputfile> <outputfile>");
            System.out.println("            -- reads Acme from <inputfile> and writes Armani to <outputfile>");
            return;
        }


        // since we are parsing Acme, make sure that we have the language
        // config flags set right, even if they might have been previously
        // setup to parse Armani.  After parsing set them for Armani
        acmeLib.core.globals.LanguageConfig.IS_PURE_ACME = true;
        acmeLib.core.globals.LanguageConfig.IS_ARMANI = false;

        // initialize the design and the parser
        parser = new AcmeParser(inputFileStream);
        AcmeParser.initDesign();

        // turn parser tracing and debugging off
        AcmeParser.disable_tracing();

        // parse the Acme design
        AcmeDesign targetDesign = null;
        try {
            targetDesign = AcmeParser.parse_AcmeDesign();
        }
        catch (acmeLib.acmeParser.ParseError e) {
            System.out.println(ParserID +
                         ":  Encountered errors during Acme specification parse.");
            e.printStackTrace();
            AcmeParser.printWarningMsgs();
            AcmeParser.printErrorMsgs();
            pause_for_input();
        }

        // convert it to Armani
        Acme2Armani.translate(targetDesign);

        // set the language configuration to Armani after the translation.
        acmeLib.core.globals.LanguageConfig.IS_PURE_ACME = false;
        acmeLib.core.globals.LanguageConfig.IS_ARMANI = true;

        // write the design out as Armani
        AcmeOutputStream s = new AcmeOutputStream();
        try {
            targetDesign.exportArmani(s);

            if (outputFileStream == null) {
                s.writeTo(System.out);
                s.flush();
            } else {
                s.writeTo(outputFileStream);
                s.flush();
            };
        } catch (Exception ex) {
            System.out.println("Caught Exception in AcmeDesign.exportArmani");
            ex.printStackTrace();
            pause_for_input();
        };


        // close the files we opened
        try {
            if (outputFileStream != null) {
                outputFileStream.close();
            };
        } catch (java.io.IOException ioExc) {};

        try {
            if (inputFileStream != null) {
                inputFileStream.close();
            };
        } catch (java.io.IOException ioExc) {};

    };   // end main method




    /** Converts the Armani analyses, invariants, and heuristics, that are
     *  encoded as Acme properties to their native Armani representations.
     *  @param design The AcmeDesign that is to be translated from Acme to Armani
     **/
    public static void translate(AcmeDesign design) {

        // reinitialize the Armani parser and load the built-ins into
        // this Acme design.
        ArmaniParser.setToplevelDesign(null);
        ArmaniParser.initDesign(design);

        // walk over the Acme description and translate it into Armani.
        Acme2ArmaniWalker newOp = new Acme2ArmaniWalker();
        GraphWalker.WalkOverAllElements(design, newOp, GraphWalker.WALK_TOP_DOWN);

        // reify all of the references and types in the design.
        design.doPostParsingProcessing();
    };


}