/**
 *  Class: Acme2ArmaniWalker
 *
 *  Purpose: Implements a graph walker operation that converts Acme code
 *           with Armani design rules and design analyses appropriately
 *           encoded as properties into an Armani design that can be
 *           typechecked and constraint-checked by the Armani design
 *           checker.
 *<BR><BR>
 *           This class is a helper class for the Acme2Armani class.
 *
 *  Revision History: <BR>
 *    Date:     Person    Change <BR>
 *    6/16/98   RTM        Created initial class <BR>
 *
 *  @author Bob Monroe
 */

package acmeLib.tools.converters;

import acmeLib.armaniParser.ArmaniParser;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.core.globals.ConversionConstants;
import acmeLib.designRules.DesignRule;
import acmeLib.elements.*;
import acmeLib.errorManagement.AcmeErrorManager;
import acmeLib.errorManagement.TranslationError;
import acmeLib.properties.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;

public class Acme2ArmaniWalker implements acmeLib.tools.WalkerElementOperation
{
    //
    // Protected class members
    //

    /** Checks whether the property has a valid type assignment for Armani.
     *  Recursively repeats this for all of prop's meta-properties.
     *  @param prop The property to make Armani-safe.
     **/
    protected void makePropValidArmani(AcmeProperty prop) {
        if ((prop.getType() == ConstantTypes.UNSPECIFIED_TYPE)
            || (prop.getType() == ConstantTypes.EXPLICIT_ANY_TYPE)
            || (prop.getType() == null))
        {
            this.assignDefaultPropType(prop);
        };

        AcmeEnumeration metaProps = prop.getProperties();
        while(metaProps.moreItems()) {
            AcmeProperty curMetaProp = (AcmeProperty) metaProps.getNextItem();
            makePropValidArmani(curMetaProp);
        };
    };


    /** if prop.getType() is null or "any" assign the minimal base type
     *  to prop that will support prop's value.
     *  @param prop  The property to which a default property is assigned
     **/
    protected void assignDefaultPropType(AcmeProperty prop) {
        if (prop == null) { return; };

        AcmePropertyValue propVal = prop.getValue();
        if (propVal instanceof AcmeInt) {
            prop.setType(acmeLib.core.globals.ConstantTypes.INT_TYPE);
        } else if (propVal instanceof AcmeFloat) {
            prop.setType(acmeLib.core.globals.ConstantTypes.FLOAT_TYPE);
        } else if (propVal instanceof AcmeBoolean) {
            prop.setType(acmeLib.core.globals.ConstantTypes.BOOLEAN_TYPE);
        } else if (propVal instanceof AcmeString) {
            prop.setType(acmeLib.core.globals.ConstantTypes.STRING_TYPE);
        } else if (propVal instanceof AcmeSet) {
            prop.setType(acmeLib.core.globals.ConstantTypes.ANY_SET_TYPE);
        } else if (propVal instanceof AcmeSequence) {
            prop.setType(acmeLib.core.globals.ConstantTypes.ANY_SEQUENCE_TYPE);
        } else if (propVal instanceof AcmeRecord) {
            prop.setType(acmeLib.core.globals.ConstantTypes.ANY_RECORD_TYPE);
        } else if (propVal instanceof AcmeEnum) {
            prop.setType(acmeLib.core.globals.ConstantTypes.EMPTY_ENUM_TYPE);
        } else if (propVal == null) {
            // RTM 6/16/98 -- this is a tricky case because all Armani properties
            //                must be typed.  If we can't infer a type then
            //                throw a conversion error and set the property's
            //                type to the default property type.
            String errMsg = "Translation Error:  Property " + prop.getName()
                            + " has neither a value nor a valid Armani Type"
                            + " specification.  Unable to correctly infer"
                            + " Armani Type.  Setting property type to default"
                            + " of String.";
            TranslationError newTransError = new TranslationError(prop, errMsg);
            AcmeErrorManager.errorDetected(newTransError);
        };
    };

    //
    // Public class members
    //

    /** Convert Acme properties that encode Armani analyses and design rules
     *  into their appropriate Armani constructs.  Also, for all properties
     *  that have null type objects apply an appropriate type object to them.
     *  Finally, if the element has any attachments or representations be
     *  sure to convert any of their properties to Armani-legal properties also.
     *
     *  @param elt The AcmeElement to do the conversion on.
     **/
    public void doElementOp(AcmeElement elt) {
        AcmeEnumeration props = elt.getProperties();
        boolean isDesignRule, isAnalysis;
        String conversionStr;
        boolean doConversion = false;
        boolean deleteProp = false;

        while (props.moreItems()) {
            AcmeProperty curProp = (AcmeProperty) props.getNextItem();
            makePropValidArmani(curProp);

            // check for an invariant or heuristic encoded as an Acme property
            if ((ConversionConstants.ARMANI_INVARIANT_TYPE_NAME_AS_ACME_PROP.equals(
                    curProp.getType().getTypeName()))
             || (ConversionConstants.ARMANI_HEURISTIC_TYPE_NAME_AS_ACME_PROP.equals(
                        curProp.getType().getTypeName())))
            {
                // first check that if this is a design rule that should only
                // be defined in a type definition. only do the conversion
                // if elt is indeed the appropriate type.
                AcmeProperty typeNameMetaProp = curProp.getMetaProperty(
                                ConversionConstants.DECLARED_IN_TYPENAME);

                String eltTypeName = null;
                if (elt instanceof AcmeElementType) {
                    eltTypeName = ((AcmeElementType) elt).getTypeName();
                };

                if (typeNameMetaProp == null) {
                    // then this is an instance-based design rule
                    doConversion = true;
                    deleteProp = true;
                }
                else if ((typeNameMetaProp.getValue() instanceof AcmeString)
                         && (eltTypeName != null))
                {
                    // then this is a type-based design rule.  only do the
                    // conversion if this is the type in which the design rule
                    // was originally declared.
                    if (eltTypeName.equals(((AcmeString) typeNameMetaProp.getValue()).getValue()))
                    {
                        doConversion = true;
                        deleteProp = true;
                    }
                    else {
                        // else this is a type-based design rule, but it should
                        // not be associated with this type.
                        doConversion = false;
                        deleteProp = true;
                    };
                }
                else {
                    // else this is a type-based design rule in an instance.
                    doConversion = false;
                    deleteProp = true;
                };

                if (doConversion && (curProp.getValue() instanceof AcmeString)) {
                    try {

                        // get the string encoding the design rule and have
                        // the ArmaniParser convert it into an Armani
                        // DesignRule object.

                        conversionStr = ((AcmeString) curProp.getValue()).getValue();

                        java.io.ByteArrayInputStream drInputStream =
                            new java.io.ByteArrayInputStream(conversionStr.getBytes());
                        if (ArmaniParser.token_source == null)
                            new ArmaniParser(drInputStream);
                        else 
                            ArmaniParser.ReInit(drInputStream);
                        DesignRule newDesignRule = ArmaniParser.parse_DesignRule(elt);

                        if (newDesignRule.getIsInvariant()) {
                            elt.addInvariant(newDesignRule);
                        } else {
                            elt.addHeuristic(newDesignRule);
                        };

                    } catch (acmeLib.armaniParser.ParseError pExc) {
                        // if we can't parse the design rule then throw a
                        // translation error indicating as much, but don't
                        // delete the property.
                        String errMsg = "Translation Error:  Property " + curProp.getName()
                                        + " has the Acme design-rule type encoding but"
                                        + " the ArmaniParser was unable to parse its value: "
                                        + pExc.getMessage();

                        TranslationError newTransError = new TranslationError(curProp, errMsg);
                        AcmeErrorManager.errorDetected(newTransError);

                        // keep the prop for user-translation debugging reference
                        deleteProp = false;
                    };
                };

                // remove the original property if previous actions require it.
                if (deleteProp) {
                    elt.removeProperty(curProp.getName());
                };
            }

            // else check for a Analysis-encoded property
            else if (ConversionConstants.ARMANI_ANALYSIS_TYPE_NAME_AS_ACME_PROP.equals(
                        curProp.getType().getTypeName()))
            {
                // first check that if this is an analysis that should only
                // be defined in a type/family definition and not a system def.
                // only do the conversion if this is the family in which it
                // is defined.
                AcmeProperty typeNameMetaProp = curProp.getMetaProperty(
                                ConversionConstants.DECLARED_IN_TYPENAME);

                String eltTypeName = null;
                if (elt instanceof AcmeElementType) {
                    eltTypeName = ((AcmeElementType) elt).getTypeName();
                };

                if (typeNameMetaProp == null) {
                    // then this is an instance-based design analysis
                    doConversion = true;
                    deleteProp = true;
                }
                else if ((typeNameMetaProp.getValue() instanceof AcmeString)
                         && (eltTypeName != null))
                {
                    // then this is a type-based analysis.  only do the
                    // conversion if this is the type/family in which the
                    // design analysis was originally declared.
                    if (eltTypeName.equals(((AcmeString) typeNameMetaProp.getValue()).getValue()))
                    {
                        doConversion = true;
                        deleteProp = true;
                    }
                    else {
                        // else this is a type-based design analysis, but it
                        // should not be associated with this type.
                        doConversion = false;
                        deleteProp = true;
                    };
                }
                else {
                    // else this is a type-based design analysis in an instance
                    // which should not be converted.
                    doConversion = false;
                    deleteProp = true;
                };

                if (doConversion && (curProp.getValue() instanceof AcmeString)) {
                    try {

                        // get the string encoding the design analysis and have
                        // the ArmaniParser convert it into an Armani
                        // DesignAnalysis object.

                        conversionStr = ((AcmeString) curProp.getValue()).getValue();
                        java.io.ByteArrayInputStream daInputStream =
                            new java.io.ByteArrayInputStream(conversionStr.getBytes());
                        ArmaniParser.ReInit(daInputStream);
                        if (elt instanceof AcmeDesign) {
                            ((AcmeDesign) elt).addDesignAnalysis(
                                    ArmaniParser.DesignAnalysisDeclaration(elt));
                        }
                        else if (elt instanceof AcmeSystem) {
                            ((AcmeSystem) elt).addDesignAnalysis(
                                    ArmaniParser.DesignAnalysisDeclaration(elt));
                        }
                        else {
                            // else we're trying to add an analysis to an element
                            // that doesn't support analyses, so announce a
                            // translation error.
                            String errMsg = "Translation Error:  Element "
                                    + elt.getName()
                                    + " has an Acme DesignAnalysis type encoded"
                                    + " property but it is not a Design, System,"
                                    + " or Family.  Keeping property but not"
                                    + " translating it to a design analysis.";

                            TranslationError newTransError =
                                        new TranslationError(curProp, errMsg);
                            AcmeErrorManager.errorDetected(newTransError);

                            // keep the prop for user-translation debugging reference
                            deleteProp = false;
                        };

                    }
                    catch (acmeLib.armaniParser.ParseError pExc) {
                        // if we can't parse the design rule then throw a
                        // translation error indicating as much, but don't
                        // delete the property.
                        String errMsg = "Translation Error:  Property " + curProp.getName()
                                        + " has the Acme design-rule type encoding but"
                                        + " the ArmaniParser was unable to parse its value: "
                                        + pExc.getMessage();

                        TranslationError newTransError = new TranslationError(curProp, errMsg);
                        AcmeErrorManager.errorDetected(newTransError);

                        // keep the prop for user-translation debugging reference
                        deleteProp = false;
                    }
                    catch (AcmeIllegalNameException aineExc) {
                        // else we're trying to add an analysis to an element
                        // that has a name collision problem.
                        String errMsg = "Translation Error: Adding DesignAnalysis"
                                + " to element (" + elt.getName() + ")"
                                + " caused a name collision.  Property "
                                + curProp.getName() + " preserved as a property"
                                + " but not translated into a design analysis."
                                + " AcmeIllegalNameException message = "
                                + aineExc.getMessage();

                        TranslationError newTransError =
                                    new TranslationError(curProp, errMsg);
                        AcmeErrorManager.errorDetected(newTransError);

                        // keep the prop for user-translation debugging reference
                        deleteProp = false;
                    };
                }

                // remove the original property if previous actions require it.
                if (deleteProp) {
                    elt.removeProperty(curProp.getName());
                };

            }


        };

        // check whether the element has any representations that could have
        // bindings with properties, if it does then check that all of
        // those properties have legal Armani types and provide them if they're
        // missing.
        AcmeEnumeration reps = elt.getRepresentations();
        AcmeEnumeration bindingsEnum = null;
        while (reps.moreItems()) {
            AcmeRepresentation curRep = (AcmeRepresentation) reps.getNextItem();
            if (curRep.getAbstractionMap() instanceof AcmeBindingsMap) {
                AcmeBindingsMap bindMap = (AcmeBindingsMap) curRep.getAbstractionMap();
                bindingsEnum = bindMap.getBindings();
                while (bindingsEnum.moreItems()) {
                    AcmeBinding curBinding = (AcmeBinding) bindingsEnum.getNextItem();
                    AcmeEnumeration bindProps = curBinding.getProperties();
                    while(bindProps.moreItems()) {
                        makePropValidArmani((AcmeProperty) bindProps.getNextItem());
                    };
                };
            };
        };

    };
}