/**
 *  Class: Armani2Acme
 *
 *  Purpose: Class to translate an Armani file into an Acme file with
 *           all of the Armani-specific constructs encoded as Acme properties.
 *
 *  Revision History: <BR>
 *    Date:     Person    Change <BR>
 *    6/16/98   RTM        Created initial class <BR>
 *
 *  @author Bob Monroe
 */

package acmeLib.tools.converters;

import java.io.FileOutputStream;

import acmeLib.armaniParser.ArmaniParser;
import acmeLib.core.io.AcmeOutputStream;
import acmeLib.elements.AcmeDesign;

public class Armani2Acme {

    //
    // Protected class members
    //

    /** pause processing until "enter" is pressed */
    protected static void pause_for_input() {
        System.out.println("Please press enter to continue");
        try {
            System.in.read();
        } catch (java.io.IOException e) {
            return;
        }
    };

    //
    // Public class members
    //

    /** Main reads in an Armani file, converts it to Acme, and writes it out
     *  as an Acme file.
     **/
    public static void main(String args[]) {
        String ParserID = new String("Armani2Acme translator (release 0.1)");
        ArmaniParser parser;
        java.io.OutputStream outputFileStream = null;
        java.io.InputStream inputFileStream = null;
        String canonicalInputFilename = null;

        if (args.length == 0) {
            System.out.println(ParserID + ":  Reading from standard input . . .");
            inputFileStream = System.in;
            canonicalInputFilename = "System.in";
            outputFileStream = System.out;
        } else if (args.length == 1) {
            System.out.println(ParserID + ":  Reading from file " +
                             args[0] + " . . .");
          try {
            inputFileStream = new java.io.FileInputStream(args[0]);
            java.io.File inputFile = new java.io.File(args[0]);
            try {
                canonicalInputFilename = inputFile.getCanonicalPath();
            } catch (java.io.IOException ioExc) {
                canonicalInputFilename = args[0];
            };
            outputFileStream = System.out;
          } catch (java.io.FileNotFoundException e) {
                System.out.println(ParserID + ":  File " + args[0] + " not accessible.");
                return;
          }
        } else if (args.length == 2) {
            System.out.println(ParserID + ":  Reading from file "
                               + args[0] + " . . .\n"
                               + "Writing to file " + args[1] + " . . .");
            try {
               inputFileStream = new java.io.FileInputStream(args[0]);
               java.io.File inputFile = new java.io.File(args[0]);
               try {
                    canonicalInputFilename = inputFile.getCanonicalPath();
               } catch (java.io.IOException ioExc) {
                    canonicalInputFilename = args[0];
               };
               outputFileStream = new FileOutputStream(args[1]);
            } catch (java.io.IOException e) {
                System.out.println(ParserID + ":  File " + args[1] + " not accessible.");
                return;
            }

        } else {
            System.out.println("Armani2Acme translator (release 0.1): " +
                             "Usage is one of:");
            System.out.println("         java Armani2Acme");
            System.out.println("            -- reads Acme from stdin and writes Armani to stdout");
            System.out.println("OR");
            System.out.println("         java Armani2Acme <inputfile>");
            System.out.println("            -- reads Acme from <inputfile> and writes Armani to stdout");
            System.out.println("OR");
            System.out.println("         java Armani2Acme <inputfile> <outputfile>");
            System.out.println("            -- reads Acme from <inputfile> and writes Armani to <outputfile>");
            return;
        }


        // since we are parsing Armani, make sure that we have the language
        // config flags set right, even if they might have been previously
        // setup to parse Armani.  After parsing set them for Acme before
        // exporting
        acmeLib.core.globals.LanguageConfig.IS_PURE_ACME = false;
        acmeLib.core.globals.LanguageConfig.IS_ARMANI = true;

        // initialize the design and the parser
        parser = new ArmaniParser(inputFileStream);
        ArmaniParser.initDesign();

        // it's possible that the initDesign() method will ReInit the
        // parser to parse the builtInAnalysisDesign, so ReInit the
        // parser with the calculated input stream.
        ArmaniParser.ReInit(inputFileStream);

        // turn parser tracing and debugging off
//        parser.disable_tracing();
        ArmaniParser.enable_tracing();

        // parse the Acme design
        AcmeDesign targetDesign = null;
        try {
            targetDesign = ArmaniParser.parse_AcmeDesign();
            targetDesign.doPostParsingProcessing();
        }
        catch (acmeLib.armaniParser.ParseError e) {
            System.out.println(ParserID +
                         ":  Encountered errors during Armani parse.");
            e.printStackTrace();
            ArmaniParser.printWarningMsgs();
            ArmaniParser.printErrorMsgs();
            pause_for_input();
        }

        // write the design out as Acme
        AcmeOutputStream s = new AcmeOutputStream();
        try {
            targetDesign.exportAcme(s);

            if (outputFileStream == null) {
                s.writeTo(System.out);
                s.flush();
            } else {
                s.writeTo(outputFileStream);
                s.flush();
            };
            pause_for_input();
        } catch (Exception ex) {
            System.out.println("Caught Exception in AcmeDesign.exportAcme");
            ex.printStackTrace();
            pause_for_input();
        };

        // set the language configuration to Armani after the translation.
        acmeLib.core.globals.LanguageConfig.IS_PURE_ACME = false;
        acmeLib.core.globals.LanguageConfig.IS_ARMANI = true;


        // close the files we opened
        try {
            if (outputFileStream != null) {
                outputFileStream.close();
            };
        } catch (java.io.IOException ioExc) {};

        try {
            if (inputFileStream != null) {
                inputFileStream.close();
            };
        } catch (java.io.IOException ioExc) {};

    };   // end main method

}