// Class: ArcAnalyzer
// Purpose: Tool that walks over an ARC-enabled system and evaluates all
//          of the possible combinations of ARC components to determine
//          which of these combinations are valid and which are not.
//          After the evaluation is done, it annotates the system with
//          the property called "legalCombinations" which is a set of sets
//          of legal ARC configurations.  (Details of this property should
//          be worked out by junli and bmonroe).
//
// Revision History:
//    Date:     Person     Change
//    11/6/98    RTM        * created initial class

package acmeLib.tools.junli;

import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.elements.*;
import acmeLib.properties.AcmeProperty;
import acmeLib.properties.AcmeSet;
import acmeLib.properties.SetType;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.typeManagement.AcmeTypeManager;

public class ArcAnalyzer {

    // define constants for the three variants of ARC component modules that
    // can be "plugged-in" to the "outer" component.
    protected static final int SAFETY_VARIANT = 1;
    protected static final int BASELINE_VARIANT = 2;
    protected static final int COMPLEX_VARIANT = 3;

    // define constants for the expected names for the variants available
    // in an ARC.
    protected static final String safetyVariantName = "safety";
    protected static final String complexVariantName = "complex";
    protected static final String baselineVariantName = "baseline";

    // pointer to the ARC style that all methods can access and use
    protected AcmeFamily arcStyle = null;

    // pointer to the ARC_Component component type that stores the
    // structure required of all ARC's.
    protected static AcmeComponentType ARC_ComponentType = null;

    // pointer to the ReplacementUnit component type that stores the
    // structure required of all ARC's.
    protected AcmeComponentType ReplacementUnitType = null;

    // Property that contains the set of valid configurations that have
    // been found.
    protected AcmeSet validConfigSet = null;
    protected SetType validConfigSetType = null;


    // default constructor
    public ArcAnalyzer() {
        this.validConfigSet = new AcmeSet();
    };


    // entry method for iterating over all possible combinations of ARC
    // components, changing the configuration to check for the legality
    // of each combination, and adding the configuration set to the system's
    // "legalCombinations" property if the combination is legal.
    public void evalAllArcConfigurations(AcmeSystem sys) {

        AcmeSystem originalSys = sys;
        String sysNameToRestore = sys.getName();

        // initialize this call to the routine by reseting the validConfigSet
        // and making a copy of the system to test.
        this.validConfigSet = new AcmeSet();
        AcmeSystem testSys = (AcmeSystem) sys.copy();
        testSys.setParentDesign(originalSys.getParentDesign());

        // look up the arcStyle, ARC_Component, and ReplacementUnit types and
        // set them for the system in question.  This hard-codes in the names
        // as follows:
        //  - arcStyle is a family called "ARC"
        //  - ARC_ComponentType is a component type called "ARC_Component"
        //      defined in the ARC style/family
        //  - ReplacementUnitType is a component type called "ReplacementUnit"
        //      defined in the ARC style/family.
        // These names can be changed as needed in the style definition or in
        // this code.
        Object foundObj = sys.lookupName("ARC", true);
        if (foundObj instanceof AcmeFamily) {
            this.arcStyle = (AcmeFamily) foundObj;
        } else {
            this.arcStyle = null;
            System.err.println("Warning -- unable to resolve style \"ARC\"");
        };

        foundObj = sys.lookupName("ARC_Component", true);
        if (foundObj instanceof AcmeComponentType) {
            ARC_ComponentType = (AcmeComponentType) foundObj;
        } else {
            ARC_ComponentType = null;
            System.err.println("Warning -- unable to resolve Component Type \"ARC_Component\"");
        };

        foundObj = sys.lookupName("ReplacementUnit", true);
        if (foundObj instanceof AcmeComponentType) {
            this.ReplacementUnitType = (AcmeComponentType) foundObj;
        } else {
            this.ReplacementUnitType = null;
            System.err.println("Warning -- unable to resolve Component Type \"ReplacementUnit\"");
        };


        // make the initial enumeration of ARC's to try in different configurations
        AcmeEnumeration compsEnum = testSys.getComponents();
        AcmeEnumeration arcsEnum = new AcmeEnumeration(compsEnum.getSize());

        while (compsEnum.moreItems()) {
            // if the current component is an ARC then add it to arcsEnum
            AcmeComponent curComp = (AcmeComponent) compsEnum.getNextItem();
            if (compIsArc(curComp)) {
                arcsEnum.insertItem(curComp);
            };
        };

        // send the copied system and the enum filled with ARC's for evaluation.
        evalArcConfiguration(testSys, arcsEnum);

// debug -- begin -- print out the testSys after it's been analyzed
// dumpSys(testSys);
// debug -- end --

        // set the original system's validConfigSet property based on the
        // calculated combinations.
        AcmeProperty validConfigProp = originalSys.getProperty("validConfigSet");
        if (validConfigProp == null) {
            // initialize the property and property type for storing the
            // valid configurations.  This should be configured to work with
            // the names and types used in the final version of the ARC family
            // specification
            validConfigProp = new AcmeProperty();
            try {
                validConfigProp.setName("validConfigProp");
            } catch (AcmeIllegalNameException ainExc) {
                // do nothing, this is a valid name...
            };

            // lookup the configNameSet type.  if we can't find it then
            // create a new one locally.
            // RTM 11/9/98 -- This lookup isn't yet implemented properly.
            SetType defaultSetType = new SetType("configNameSet",
                          new SetType("string", ConstantTypes.STRING_TYPE));
            defaultSetType.setInlineTypeSpec(true);
            validConfigProp.setType(defaultSetType);
        };


    };



    // this is the method that is recursively called for all configurations
    // and combinations of the ARC's.  It works recursively by modifying the
    // system passed to it by the calling method.  This fits into the entire
    // class' analysis by working on a copy of the system in question that
    // it modifies many different ways.
    public void evalArcConfiguration(AcmeSystem curSys,
                                     AcmeEnumeration arcsToTest)
    {
        if (!arcsToTest.moreItems()) {
            System.out.println(" -- no ARC's to test in acmeLib.tools.junli.ArcAnalyzer.evalArcConfiguration.");
            return;
        };

        // the current component we will be replacing in this iteration
        // with the ARC variants
        AcmeComponent originalComp = (AcmeComponent) arcsToTest.getNextItem();

        if (originalComp == null) {
            System.out.println("Error in acmeLib.tools.junli.ArcAnalyzer.evalArcConfiguration."
                               + "\n -- originalComp is null.");
            return;
        };


        // store the ARC representation system for so that it won't need
        // to be repeatedly looked up.
        AcmeRepresentation tempRep = originalComp.getRepresentation("ARC_Sub");
        if (tempRep == null) {
            // RTM 11/11/98 -- Should this run the typechecking "as-is" and
            // memo-ize accordingly, or should it just return?
            // for now just quietly return.  This isn't an ARC.
            return;
        };

        AcmeSystem repSys = tempRep.getSystem();
        if (repSys == null) {
            // RTM 11/11/98 -- Should this run the typechecking "as-is" and
            // memo-ize accordingly, or should it just return?
            // for now just quietly return.  This isn't an ARC.
            return;
        };


        // Stores the attachments to originalComp that we will be removing
        // from the system for checking, but will want to be able to reattach
        // appropriately for each of the alternative component ARC variants
        // that we'll be replacing originalComp with.
        AcmeEnumeration originalCompAttachments = new AcmeEnumeration();

        // fill the originalCompAttachments with all of the attachments that
        // are attached to originalComp.
        AcmeEnumeration originalCompPorts = originalComp.getPorts();
        while (originalCompPorts.moreItems()) {
            AcmePort curPort = (AcmePort) originalCompPorts.getNextItem();
            AcmeEnumeration curPortAttachments = curSys.getAttachments(curPort);
            originalCompAttachments.concatenate(curPortAttachments);
        };

        // create three copies of the enumerations of the remaining
        // combinations of ARC variants.
        AcmeEnumeration safetyBranchArcs = new AcmeEnumeration(arcsToTest);
        AcmeEnumeration complexBranchArcs = new AcmeEnumeration(arcsToTest);
        AcmeEnumeration baselineBranchArcs = new AcmeEnumeration(arcsToTest);


// Debug -- avoid recursion...
//safetyBranchArcs.getNextItem();
//complexBranchArcs.getNextItem();
//baselineBranchArcs.getNextItem();
// end debug

        // replace the originalComp with each of the three variants, typecheck
        // the system with the replacement made, and then recursively continue
        // checking for all combinations of ARC variants that include this
        // configuration.
        AcmeComponent replacementComp = null;

        // for the safety variant
        try {
            AcmeEnumeration safetyVariantAttachments = new AcmeEnumeration();
            replacementComp = replaceCompWithVariant(originalComp, curSys,
                                                     SAFETY_VARIANT,
                                                     originalCompAttachments,
                                                     safetyVariantAttachments);
// debug
if (replacementComp == null) {
    System.out.println("== After replaceCompWithVariant, replacementComp is null "
                       + "for SAFETY_VARIANT where "
                       + "originalComp = " + originalComp.getName()
                       + " and curSys = " + curSys.getName());
} else {
    System.out.println("==!! After replaceCompWithVariant, replacementComp is --NOT-- null "
                       + "for SAFETY_VARIANT where "
                       + "originalComp = " + originalComp.getName()
                       + " and curSys = " + curSys.getName());
}
// end debug
            // typecheck this configuration
            if (curSys.typecheck()) {
                // RTM 11/9/98 -- note that this should be replaced with a
                // memo-izing action to add this configuration to the
                // "legal configuration set".  For the moment it just prints
                // a message out.
                System.out.println("+++System " + curSys.getName()
                    + " typechecks with component " + originalComp.getName()
                    + " replaced with its safety variant");
            } else {
                System.out.println("  --- System " + curSys.getName()
                    + " DOES NOT TYPECHECK with component "
                    + originalComp.getName()
                    + " replaced with its safety variant");
            };

            // recurse for other configurations if there are any unchecked
            // ARC's left
            if (safetyBranchArcs.moreItems()) {
                evalArcConfiguration(curSys, safetyBranchArcs);
            };

            // put the last variant that we tried back into the original
            // component
            curSys.removeComponent(replacementComp);
            restoreVariantToCompRep(repSys, replacementComp, SAFETY_VARIANT,
                                    safetyVariantAttachments);

        } catch (VariantReplacementException vrExc) {
// debug
dumpSys(curSys);
// end debug
            System.err.println("Unable to replace component "
                               + originalComp.getName() + " in system "
                               + curSys.getName() + " with a safety variant\n"
                               + "Exception message = " + vrExc.getMessage());
        } catch (AcmeIllegalNameException ainExc) {
            System.err.println("Unable to replace component "
                               + originalComp.getName() + " in system "
                               + curSys.getName() + " with a safety variant\n"
                               + "Caught an illegal name exception. Message = "
                               + ainExc.getMessage());
        };


        // for the complex variant
        try {
            AcmeEnumeration complexVariantAttachments = new AcmeEnumeration();
            replacementComp = replaceCompWithVariant(originalComp, curSys,
                                                     COMPLEX_VARIANT,
                                                     originalCompAttachments,
                                                     complexVariantAttachments);
            // typecheck this configuration
            if (curSys.typecheck()) {
                // RTM 11/9/98 -- note that this should be replaced with a
                // memo-izing action to add this configuration to the
                // "legal configuration set".  For the moment it just prints
                // a message out.
                System.out.println("+++System " + curSys.getName()
                    + " typechecks with component " + originalComp.getName()
                    + " replaced with its complex variant");
            } else {
                System.out.println("  --- System " + curSys.getName()
                    + " DOES NOT TYPECHECK with component "
                    + originalComp.getName()
                    + " replaced with its complex variant");
            };

            // recurse for other configurations if there are any unchecked
            // ARC's left
            if (complexBranchArcs.moreItems()) {
                evalArcConfiguration(curSys, complexBranchArcs);
            };

            // put the last variant that we tried back into the original
            // component
            curSys.removeComponent(replacementComp);
            restoreVariantToCompRep(repSys, replacementComp, COMPLEX_VARIANT,
                                    complexVariantAttachments);

        } catch (VariantReplacementException vrExc) {
            System.err.println("Unable to replace component "
                               + originalComp.getName() + " in system "
                               + curSys.getName() + " with a complex variant\n"
                               + "Exception message = " + vrExc.getMessage());
        } catch (AcmeIllegalNameException ainExc) {
            System.err.println("Unable to replace component "
                               + originalComp.getName() + " in system "
                               + curSys.getName() + " with a complex variant\n"
                               + "Caught an illegal name exception. Message = "
                               + ainExc.getMessage());
        };



        // for the baseline variant
        try {
            AcmeEnumeration baselineVariantAttachments = new AcmeEnumeration();
            replacementComp = replaceCompWithVariant(originalComp, curSys,
                                                     BASELINE_VARIANT,
                                                     originalCompAttachments,
                                                     baselineVariantAttachments);
            // typecheck this configuration
            if (curSys.typecheck()) {
                // RTM 11/9/98 -- note that this should be replaced with a
                // memo-izing action to add this configuration to the
                // "legal configuration set".  For the moment it just prints
                // a message out.
                System.out.println("+++System " + curSys.getName()
                    + " typechecks with component " + originalComp.getName()
                    + " replaced with its baseline variant");
            } else {
                System.out.println("  --- System " + curSys.getName()
                    + " DOES NOT TYPECHECK with component "
                    + originalComp.getName()
                    + " replaced with its baseline variant");
            };

            // recurse for other configurations if there are any unchecked
            // ARC's left
            if (baselineBranchArcs.moreItems()) {
                evalArcConfiguration(curSys, baselineBranchArcs);
            };

            // put the last variant that we tried back into the original
            // component
            curSys.removeComponent(replacementComp);
            restoreVariantToCompRep(repSys, replacementComp, BASELINE_VARIANT,
                                    baselineVariantAttachments);

        } catch (VariantReplacementException vrExc) {
            System.err.println("Unable to replace component "
                               + originalComp.getName() + " in system "
                               + curSys.getName() + " with a baseline variant\n"
                               + "Exception message = " + vrExc.getMessage());
        } catch (AcmeIllegalNameException ainExc) {
            System.err.println("Unable to replace component "
                               + originalComp.getName() + " in system "
                               + curSys.getName() + " with a baseline variant\n"
                               + "Caught an illegal name exception. Message = "
                               + ainExc.getMessage());
        };



        // restore the originalComp to the curSys
        try {
            curSys.addComponent(originalComp);

// debug
System.out.println("Successfully returned component " + originalComp.getName()
                    + " to system " + curSys.getName());
dumpSys(curSys);
// end debug

        } catch (AcmeIllegalNameException ainExc) {
            // do nothing, this was a legal name just moments before...
            System.out.println("Caught AcmeIllegalNameException restoring originalComp"
                                + originalComp.getName() + " in system "
                                + curSys.getName());
        };

        while (originalCompAttachments.moreItems()) {
            curSys.addAttachment(
                    (AcmeAttachment) originalCompAttachments.getNextItem());
        };
    }



    /** helper method that replaces originalComp in sys with the requested
     * variant of originalComp, removes all of the old attachments of
     * originalComp, and attempts to hook up replacement attachments
     * returns the new component given to replace the original component.
     * originalCompAttachmentsEnum will be returned reset.
     * @param originalComp The component that will be replaced with a variant
     * @param parentSys The parent system in which the originalComp lives
     *                   and the variant that will replace it lives.
     * @param replaceWithVariant flag describing which variant should replace
     *                            the originalComp.  Valid values are SAFETY_VARIANT,
     *                            BASELINE_VARIANT, and COMPLEX_VARIANT, all of
     *                            which are constants defined in this class (ArcAnalyzer).
     * @param originalCompAttachmentsEnum lists the attachments to the originalComp
     *                                      that the replacement should try to
     *                                      replicate.
     * @param originalVariantAttachmentsEnum returns the set of attachments that
     *                                        the variant had before it was removed
     *                                        so that they can be restored later.
     *                                        This should be passed as an empty
     *                                        AcmeEnumeration that the method will
     *                                        fill up and return full.
     * @returns Reference to the variant component that replaced originalComp
     *           in parentSys.
     **/
    protected AcmeComponent replaceCompWithVariant(
                                AcmeComponent originalComp,
                                AcmeSystem parentSys,
                                int replaceWithVariant,
                                AcmeEnumeration originalCompAttachmentsEnum,
                                AcmeEnumeration originalVariantAttachmentsEnum)
                                throws VariantReplacementException,
                                       AcmeIllegalNameException
    {
        AcmeComponent newComp = null;

        // stores the variant name for whatever's actually being substituted.
        String thisVariantName = null;

        AcmeRepresentation tempRep = originalComp.getRepresentation("ARC_Sub");
        if (tempRep == null) {
            throw new VariantReplacementException("Error, unable to find \"ARC_Sub\" representation");
        };

        AcmeSystem repSys = tempRep.getSystem();

        if (replaceWithVariant == SAFETY_VARIANT) {
            newComp = repSys.getComponent(safetyVariantName);
            thisVariantName = safetyVariantName;
        } else if (replaceWithVariant == BASELINE_VARIANT) {
            newComp = repSys.getComponent(baselineVariantName);
            thisVariantName = baselineVariantName;
        } else if (replaceWithVariant == COMPLEX_VARIANT) {
            newComp = repSys.getComponent(complexVariantName);
            thisVariantName = complexVariantName;
        };

        // now we want to replace the original component with the newComp
        // throw an exception at this point if we don't have a newComp because
        // we won't have anything to replace it with.
        if (newComp == null) {
            throw new VariantReplacementException(
                    "Error, unable to find replacement component "
                    + thisVariantName + " in repSys " + repSys.getName());
        };


        // store all of the original attachments to the variant in the ARC
        // representation so that they can be replaced when the variant is
        // replaced with the original.
        AcmeEnumeration variantCompPorts = newComp.getPorts();
        while (variantCompPorts.moreItems()) {
            AcmePort curPort = (AcmePort) variantCompPorts.getNextItem();
            AcmeEnumeration curPortAttachments = repSys.getAttachments(curPort);
            originalVariantAttachmentsEnum.concatenate(curPortAttachments);
        };


        // replace the originalComponent with one of its variants.
        parentSys.removeComponent(originalComp);
        newComp.setName(originalComp.getName() + "-" + thisVariantName);
        parentSys.addComponent(newComp);


        // now look through all of the old attachments between the originalComp
        // and its various connectors and try to replicate them with the
        // replacementComp.
        AcmeRole curTmpRole = null;
        AcmePort curTmpPort = null;
        AcmePort curOldPort = null;
        String curOldPortName = null;
        while(originalCompAttachmentsEnum.moreItems()) {
            AcmeAttachment curOldAttachment = (AcmeAttachment)
                                originalCompAttachmentsEnum.getNextItem();
            curTmpRole = curOldAttachment.getRole();
            curOldPort = curOldAttachment.getPort();
            curTmpPort = newComp.getPort(curOldPort.getName());

            if ((curTmpPort != null) && (curTmpRole != null)) {
                parentSys.addAttachment(new AcmeAttachment(curTmpPort, curTmpRole));
            } else {
                throw new VariantReplacementException(
                        "Error.  unable to create attachment "
                        + curOldAttachment.getPortNameAsString() + " to "
                        + curOldAttachment.getRoleNameAsString());
            };
        };

        // reset the attachments enumeration so that it can be reused...
        originalCompAttachmentsEnum.reset();

        return newComp;
    };


    /**
     * Helper method that restores the variant components to originalComp
     * in parentSys with the requested variant passed as variantComp.
     * It then restores all of the old attachments of the variant to the
     * other variants or substructure in the variant ARC representation.
     * @param variantRepSys The system to which the variant component will be added
     * @param variantComp The variant component to be restored as a variant
     *                     of originalComp
     * @param restoreVariant flag describing which variant should replace
     *                            the originalComp.  Valid values are SAFETY_VARIANT,
     *                            BASELINE_VARIANT, and COMPLEX_VARIANT, all of
     *                            which are constants defined in this class (ArcAnalyzer).
     * @param originalCompAttachmentsEnum lists the attachments to the originalComp
     *                                      that the replacement should try to
     *                                      replicate.
     * @param originalVariantAttachmentsEnum the set of attachments that the
     *                                        variant had before it was removed.
     *                                        These will be restored.
     **/
    protected void restoreVariantToCompRep(
                                AcmeSystem variantRepSys,
                                AcmeComponent variantComp,
                                int restoreVariant,
                                AcmeEnumeration originalVariantAttachmentsEnum)
    {

// debug (begin)
System.out.println(">>> Entering restoreVariantToCompRep with: ");
System.out.println(">>>     variantRepSys = " + variantRepSys.getName());
System.out.println(">>>     variantComp = " + variantComp.getName());
System.out.println(">>>     restoreVariant = " + restoreVariant);
// debug (end)

        // try to replace the appropriate variant component.  If it failse then
        // throw a variantReplacementException
        String variantName = null;
        try {
            // reset the name of the variantComp to its variant name
            if (restoreVariant == SAFETY_VARIANT) {
                variantComp.setName(safetyVariantName);
                variantName = safetyVariantName;
            } else if (restoreVariant == COMPLEX_VARIANT) {
                variantComp.setName(complexVariantName);
                variantName = complexVariantName;
            } else if (restoreVariant == BASELINE_VARIANT) {
                variantComp.setName(baselineVariantName);
                variantName = baselineVariantName;
            };

            // try to add the variant back to the variantRepSys if it's
            // not already in there.  If it's trying to overwrite a name
            // we'll catch the exception later and quietly leave variantRepSys
            // as-is.

            if (variantRepSys.getComponent(variantComp.getName()) == null) {
                variantRepSys.addComponent(variantComp);
            };


            // try to replace all of the originalVariantAttachments to variantRepSys
            while (originalVariantAttachmentsEnum.moreItems()) {
                variantRepSys.addAttachment((AcmeAttachment)
                            originalVariantAttachmentsEnum.getNextItem());
            };
        }
        catch (AcmeIllegalNameException ainExc) {
            // do nothing, we're trying to overwrite something that's
            // already there.  just write out a warning message.
            System.err.println("Warning, caught AcmeIllegalNameException trying "
                                + "to restore variant " + variantName
                                + " to system representation: "
                                + variantRepSys.getName());
        };

    };



    // helper method to check whether a given component is an ARC.
    // Returns true if the component is an ARC, else returns false.
    public static boolean compIsArc(AcmeComponent comp) {
        boolean arcTypeFound = false;
        AcmeEnumeration e = comp.getDeclaredTypes();
        while (e.moreItems()) {
            AcmeComponentType curType = (AcmeComponentType) e.getNextItem();
            if (AcmeTypeManager.isDeclaredSubtype(curType, ARC_ComponentType))
            {
                arcTypeFound = true;
                break;
            };
        };

        return arcTypeFound;
    };


    // debugging method to dump out systems in question to stdout
    public void dumpSys(AcmeSystem sys) {
        acmeLib.core.io.AcmeOutputStream s = new acmeLib.core.io.AcmeOutputStream();
        try {
            sys.exportArmani(s);
            s.flush();
            s.writeTo(System.out);
        } catch (java.io.IOException jioExc) {
            System.out.println("--- caught java.io.IOException dumping System: "
                               + jioExc.getMessage());
        } catch (acmeLib.core.exceptions.AcmeException aExc) {
            System.out.println("--- caught AcmeException dumping System: "
                                + aExc.getMessage());
        };
    };

};


class VariantReplacementException extends Exception {

    public VariantReplacementException() { super(); };
    public VariantReplacementException(String message) { super(message); };
};