package acmeLib.typeManagement;

import java.util.HashSet;
import java.util.Set;

import acmeLib.AcmeObject;
import acmeLib.AcmeType;
import acmeLib.AcmeTypeParameter;
import acmeLib.AcmeUnresolvedType;
import acmeLib.core.container.AcmeEnumeration;
import acmeLib.core.container.AcmeVector;
import acmeLib.core.exceptions.AcmeIllegalNameException;
import acmeLib.core.exceptions.AcmePropertyExtensionFailedException;
import acmeLib.core.globals.ConstantTypes;
import acmeLib.designRules.*;
import acmeLib.elements.*;
import acmeLib.errorManagement.*;
import acmeLib.properties.*;
import acmeLib.representations.AcmeBinding;
import acmeLib.representations.AcmeBindingsMap;
import acmeLib.representations.AcmeRepresentation;
import acmeLib.representations.AcmeUnresolvedBinding;

/**
 * Class: AcmeTypeManager<br>
 * Purpose: AcmeTypeManager manages the types defined in a design.<br>
 *          Includes methods for looking up types based on their scope.<br>
 *          Makes use of local/family and global typespace.<br>
 *<br>
 * Revision History:<br>
 *    Date:   Person    Change<br>
 *    4/25/97   RTM     Stub created for compilation of the rest of AcmeLib.<br>
 *    6/04/97   RTM     Translated May 30th, 1997 C++ version to Java<br>
 *    6/16/97   RTM     stubbed out for compilation of other changes<br>
 *    7/01/97   RTM     Added typechecking for elements, properties,<br>
 *                       and representations.<br>
 *    10/28/97  RTM     Made typechecking methods static to avoid having<br>
 *                       to pass a typemanager object around through method<br>
 *                       invocations.  These methods are also functions<br>
 *                       that don't rely on any internal object state.<br>
 *                       Also made instantiation methods static.<br>
 *    2/6/98    RTM     Complete renovation to deal with typechecking supertypes,<br>
 *                       declared types, design analyses, and design rules.<br>
 *    1/3/03    SWC     Added a condition to isLogicalSubtype() to allow a
 *                       subtype of AcmeElementType and a supertype of "Type"
 *                       (for the specific case of a "Type" parameter type
 *                       supplied with an argument of an AcmeElementType) to
 *                       pass as logical subtype.
 * @author Bob Monroe
 **/

public class AcmeTypeManager extends AcmeObject {

	//
	// Constructor
	//

	/**
	 * trivial constructor.  nothing to initialize.
	 */
	public AcmeTypeManager() {
	};

	/**
	 * static stack that stores the causal chain of errors that can be
	 * detected during typechecking.
	 */
	public static CausalErrorStack typecheckingErrorStack = new CausalErrorStack();

	//
	// RTM -- danger here in making these static, as multiple users and/or
	//        threads sharing a type manager can get into race conditions.
	//        for now, it's pretty much one thread per type manager/jvm so it
	//        shouldn't be a big problem.  There's a time bomb here, though.

	/**
	 * flag to indicate if error reporting is enabled, true by default.
	 **/
	protected static boolean errorReportingEnabled = true;

	/**
	 * returns true if error reporting is enabled, else false.
	 **/
	public static boolean errorReportingIsEnabled() {
		return errorReportingEnabled;
	};

	/**
	 * enable error reporting and recording
	 **/
	public static void enableErrorReporting() {
		errorReportingEnabled = true;
		AcmeErrorManager.setLogErrors(true);
		typecheckingErrorStack.enable();
	};

	/**
	 * disable error reporting and recording
	 **/
	public static void disableErrorReporting() {
		errorReportingEnabled = false;
		AcmeErrorManager.setLogErrors(false);
		typecheckingErrorStack.disable();
	};

	/**
	 *
	 * Unification methods  (both public and protected)
	 *
	 *
	 *
	 * Type/instance unification methods.  Each of these methods takes an
	 * element type object and an empty (except, possibly, for a name)
	 * element instance as paramaters.  It then fills in/merges the
	 * structure required by the type and all of its supertypes with the
	 * structure that already exists in the object instance.
	 *
	 *
	 * generic representation initialization/unification method
	 * @param protoTypeElt Prototype Element
	 * @param targetElt Target Element
	 */
	protected static void unifyRepresentations(AcmeElement protoTypeElt, AcmeElement targetElt) {
		unifyRepresentations(protoTypeElt, targetElt, true);
	}

	protected static boolean unifyRepresentations(
		AcmeElement protoTypeElt,
		AcmeElement targetElt,
		boolean makeChanges) {
		boolean ok = true;
		AcmeEnumeration e = protoTypeElt.getRepresentations();
		while (e.moreItems()) {
			AcmeRepresentation protoRep = (AcmeRepresentation) e.getNextItem();
			if (protoRep == null)
				continue;

			AcmeRepresentation targetRep = targetElt.getRepresentation(protoRep.getName());

			if (targetRep == null && makeChanges) {
				// then this rep doesn't exist in target object, just add
				// a copy of the prototype rep to the target.
				AcmeRepresentation newRep = protoRep.copy();
				newRep.setParent(targetElt);
				if (newRep.getSystem() != null) {
					newRep.getSystem().setParentRepresentation(newRep);
					newRep.getSystem().setParent(targetElt);
				}
				try {
					targetElt.addRepresentation(newRep);
					// call unifySystem(newRep,newRep) to set all of the default
					// values appropriately.
					newRep.addInheritedSource(protoRep);
					unifySystem(protoRep.getSystem(), newRep.getSystem());
				}
				catch (AcmeIllegalNameException ainExc) {
					// then this is an attempt to add an illegally named rep.
					// this is probably an attempt to add a duplicate-named rep.
					String errMsg =
						"Attempted to add an illegally named rep: "
							+ newRep.getName()
							+ " to element: "
							+ targetElt.getName()
							+ ".  Message = "
							+ ainExc.getMessage();
					InstantiationConflictError newError =
						new InstantiationConflictError(targetElt, errMsg, targetRep, protoRep);
					AcmeErrorManager.errorDetected(newError);
				}

			}
			else {
				ok &= unifySystem(protoRep.getSystem(), targetRep.getSystem(), makeChanges);

				// unify the bindings of this rep by creating new unresolved
				// bindings objects from the existing bindings, and adding the
				// new unresolved bindings to the targetElt's representation.
				AcmeBindingsMap protoBindingsMap = (AcmeBindingsMap) protoRep.getAbstractionMap();
				AcmeBindingsMap targetBindingsMap = (AcmeBindingsMap) targetRep.getAbstractionMap();

				if (protoBindingsMap == null) {
					continue; // there are no bindings to retrieve
				}
				else if (targetBindingsMap == null && makeChanges) {
					targetBindingsMap = (AcmeBindingsMap) protoBindingsMap.copy();
					targetRep.setAbstractionMap(targetBindingsMap);
				}
				else {
					ok &= unifyBindingsMaps(protoBindingsMap, targetBindingsMap, makeChanges);
					if (makeChanges) {
						targetRep.setAbstractionMap(targetBindingsMap);
					}
				}

				if (makeChanges) {
					// update the parentRep for all of the new/unified bindings
					AcmeEnumeration bindingsEnum = ((AcmeBindingsMap) targetRep.getAbstractionMap()).getBindings();
					while (bindingsEnum.moreItems()) {
						AcmeBinding curBinding = (AcmeBinding) bindingsEnum.getNextItem();
						curBinding.setParentRep(targetRep);
					}
				}
			}

		}
		return ok;
	}

	/**
	 * generic representation initialization/unification method
	 * @param protoMap Prototype Map
	 * @param targetMap Target Map
	 */
	protected static void unifyBindingsMaps(AcmeBindingsMap protoMap, AcmeBindingsMap targetMap) {
		unifyBindingsMaps(protoMap, targetMap, true);
	}

	protected static boolean unifyBindingsMaps(
		AcmeBindingsMap protoMap,
		AcmeBindingsMap targetMap,
		boolean makeChanges) {
		boolean ok = true;
		if ((targetMap == null) || (protoMap == null)) {
			return ok;
		}

		AcmeEnumeration protoBindingsEnum = protoMap.getBindings();
		AcmeEnumeration targetBindingsEnum = targetMap.getBindings();
		while (protoBindingsEnum.moreItems()) {
			AcmeUnresolvedBinding curProtoBinding =
				new AcmeUnresolvedBinding((AcmeBinding) protoBindingsEnum.getNextItem());
			if (curProtoBinding == null) {
				continue;
			}

			// if this is a unification of an existing binding
			// then unify the properties, if not, create a fresh copy.
			targetBindingsEnum.reset();
			AcmeBinding targetBinding = null;
			while (targetBindingsEnum.moreItems() && (targetBinding == null)) {
				AcmeBinding curTargetBinding = (AcmeBinding) targetBindingsEnum.getNextItem();
				if (curProtoBinding.equals(curTargetBinding)) {
					targetBinding = curTargetBinding;
				}
			}

			if (targetBinding == null && makeChanges) {
				targetBinding = new AcmeUnresolvedBinding(curProtoBinding);
				targetMap.addBinding(targetBinding);
			}
			else {
				ok &= unifyProperties(curProtoBinding, targetBinding, makeChanges);
			}

		}
		return ok;
	}

	/**
	 * generic property initialization/unification method for all of
	 * the properties of an AcmeElement prototype and target.
	 * @param protoTypeElt prototype element
	 * @param targetElt target element
	 */
	public static void unifyProperties(AcmeElement protoTypeElt, AcmeElement targetElt) {
		unifyProperties(protoTypeElt, targetElt, true);
	}

	public static boolean unifyProperties(AcmeElement protoTypeElt, AcmeElement targetElt, boolean makeChanges) {
		boolean ok = true;
		AcmeEnumeration e = protoTypeElt.getProperties();

		while (e.moreItems()) {
			AcmeProperty typeProp = (AcmeProperty) e.getNextItem();
			AcmeProperty objProp = targetElt.getProperty(typeProp.getName());
			if (objProp != null) {
				// then the target object already has a property of this name.
				// attempt to unify the type's property structure with the target
				// object's property structure.
				if ((objProp.getType() == null) || (objProp.getType() == ConstantTypes.UNSPECIFIED_TYPE)) {
					// then add the protoTypeElt object's property type to this property name
					if (makeChanges
						&& (typeProp.getType() != ConstantTypes.UNSPECIFIED_TYPE)
						&& (typeProp.getType() != null)) {
						objProp.setType(typeProp.getType());
					}
					// else keep objProp's old property type.
				}
				else if (!isLogicalSubtype(objProp.getType(), typeProp.getType(), false)) {
					// then this is an attempt to put in a different type with
					// the same name, which is an error.
					String errMsg =
						"attempted to assign two property types ("
							+ objProp.getType().getTypeName()
							+ " from "
							+ protoTypeElt.getName()
							+ " and "
							+ typeProp.getType().getTypeName()
							+ " from "
							+ targetElt.getName()
							+ ")"
							+ " to the single property name "
							+ typeProp.getName();
					InstantiationConflictError newError =
						new InstantiationConflictError(targetElt, errMsg, objProp, typeProp);
					AcmeErrorManager.errorDetected(newError);
					ok = false;
				}
				else {
					// else the target object property and the type property have
					// the same or complimentary types already, so modify nothing.
				}

				// Attempt to unify the values of the type property and the
				// target element property.
				if (typeProp.getValue() != null) {
					if ((objProp.getValue() == null) || (objProp.getValue().getIsDefaultValue())) {
						if (makeChanges) {
							objProp.setValue(typeProp.getValue().copy());
							objProp.addInheritedSource(typeProp);
						}
					}
					else if ((objProp.getValue() != null) && typeProp.getValue().getIsDefaultValue()) {
						// then leave the object property's value as/is, because an
						// explicitly defined value overrules a default value.
					}
					else if (!objProp.getValue().equals(typeProp.getValue())) {
						// then we can't unify the values of the type's property
						// and the object's property, which is an error.
						String errMsg =
							"attempted to assign two non-default property values"
								+ " ("
								+ typeProp.getValue().toString()
								+ " from "
								+ protoTypeElt.getName()
								+ " and "
								+ objProp.getValue().toString()
								+ "from "
								+ targetElt.getName()
								+ ")"
								+ " to the single property name "
								+ typeProp.getName();
						InstantiationConflictError newError =
							new InstantiationConflictError(targetElt, errMsg, objProp, typeProp);
						AcmeErrorManager.errorDetected(newError);
						ok = false;
					}
					else {
						// else the target object property and the type property both
						// have the same value already, so modify nothing.
					}
				}
				else if (
					((objProp.getValue() != null) && (objProp.getValue().getIsDefaultValue()))
						|| (objProp.getValue() == null)) {
					// try to put a default value into the instance, and mark it
					// as a default value.
					AcmeProperty defaultProp = typeProp.getMetaProperty("default");
					if (defaultProp == null) {
						// RTM 2/10/98 -- then instantiate a default empty value, eventually.
					}

					// else see if the default property value satisfies the property's
					// type.  if it does then make it the default value, else announce
					// an error.
					else if (AcmeTypeManager.typecheckProperty(typeProp.getType(), defaultProp.getValue())) {
						if (defaultProp.getValue() != null) {
							if (makeChanges) {
								objProp.setValue(defaultProp.getValue().copy());
								objProp.getValue().setIsDefaultValue(true);
							}
						}
						else {
							// RTM 2/10/98 -- else instantiate a default empty value,
							//                (eventually).
						}
					}
					else {
						// else default type doesn't satisfy required property type
						String errMsg = "Default type does not satisfy property type";
						InstantiationConflictError newError =
							new InstantiationConflictError(targetElt, errMsg, objProp, typeProp);
						AcmeErrorManager.errorDetected(newError);
						ok = false;
					}
				}
				else {
					//  else leave the current value of objProp alone.
				}

			}
			else { // else for if (objProp != null)
				// else the property does not already exist in the target element.
				// Assign a new property to the target element
				AcmeProperty newProp = null;
				try {
					if (typeProp.getValue() == null) {
						if (makeChanges) {
							newProp = new AcmeProperty(typeProp.getName(), typeProp.getType(), null);
							newProp.addInheritedSource(typeProp);
							targetElt.addProperty(newProp);
						}
						// try to put a default value into the instance, and mark it
						// as a default value.
						AcmeProperty defaultProp = typeProp.getMetaProperty("default");
						if (defaultProp == null) {
							// RTM 2/10/98 -- then instantiate a default empty value, eventually.
						}
						else if (AcmeTypeManager.typecheckProperty(typeProp.getType(), defaultProp.getValue())) {
							if (defaultProp.getValue() != null) {
								if (makeChanges) {
									newProp.setValue(defaultProp.getValue().copy());
									newProp.getValue().setIsDefaultValue(true);
								}
							}
							else {
								// RTM 2/10/98 -- else instantiate a default empty value,
								//                (eventually).
							}
						}
						else {
							// else default type doesn't satisfy required property type
							String errMsg = "Default type does not satisfy property type";
							InstantiationConflictError newError =
								new InstantiationConflictError(targetElt, errMsg, objProp, typeProp);
							AcmeErrorManager.errorDetected(newError);
							ok = false;
						}

					}
					else {
						if (makeChanges) {
							if (typeProp.getValue() == null) {
								newProp = new AcmeProperty(typeProp.getName(), typeProp.getType(), null);
							}
							else {
								newProp =
									new AcmeProperty(
										typeProp.getName(),
										typeProp.getType(),
										typeProp.getValue().copy());
							}
							targetElt.addProperty(newProp);
						}
					}

				}
				catch (AcmeIllegalNameException ainExc) {
					// typeProp has an illegal name.  odd.
					String errMsg =
						"Attempted to add property with illegal name: "
							+ newProp.getName()
							+ " to target element: "
							+ targetElt.getName()
							+ " at instantiation.  Message = "
							+ ainExc.getMessage();
					InstantiationConflictError newError =
						new InstantiationConflictError(targetElt, errMsg, newProp, typeProp);
					AcmeErrorManager.errorDetected(newError);
					ok = false;
				}
			}

			// at this point, we should have a prototype object property and a
			// target element property.  Now unify the meta-properties of
			// the unified property.
			if (makeChanges)
				ok &= unifyMetaProperties(typeProp, targetElt.getProperty(typeProp.getName()), makeChanges);

		}
		return ok;
	}

	/**
	 * generic property initialization/unification method that merges
	 * a single addition to the target property.
	 * @param extensionProp Extension Property
	 * @param targetProp Target Property
	 * @exception AcmePropertyExtensionFailedException Property extension failure
	 */
	public static void extendProperty(AcmeProperty extensionProp, AcmeProperty targetProp)
		throws AcmePropertyExtensionFailedException {
		if ((extensionProp == null) || (targetProp == null)) {
			return;
		}

		if (targetProp.getName() != null) {
			if (!targetProp.getName().equals(extensionProp.getName())) {
				String errMsg =
					"Unable to extend property name "
						+ targetProp.getName()
						+ " with a property named: "
						+ extensionProp.getName()
						+ ".";
				throw new AcmePropertyExtensionFailedException(errMsg);
			}
		}
		else {
			String errMsg = "Unable to extend property with a null name.";
			throw new AcmePropertyExtensionFailedException(errMsg);
		}

		// else the names are equal, extend the existing property with
		// extensionProp's type and value, if/as appropriate.
		if ((extensionProp.getType() == null) || (extensionProp.getType() == ConstantTypes.UNSPECIFIED_TYPE)) {
			// then the extension adds no type information, ignore type unification.
		}
		else if ((targetProp.getType() == null) || (targetProp.getType() == ConstantTypes.UNSPECIFIED_TYPE)) {
			// then add the protoTypeElt object's property type to this property name
			if ((extensionProp.getType() != ConstantTypes.UNSPECIFIED_TYPE) && (extensionProp.getType() != null)) {
				targetProp.setType(extensionProp.getType());
			}
			// else keep targetProp's old property type.
		}
		else if (
			!((isLogicalSubtype(extensionProp.getType(), targetProp.getType(), false))
				&& (isLogicalSubtype(targetProp.getType(), extensionProp.getType(), false)))) {
			// this checks that extensionProp's type and targetProp's type
			// are structurally equivalent.  If true, then this is an attempt
			// to associate a different type with the same name, which is
			// an error.
			StringBuffer typesStringBuf = new StringBuffer("(");
			typesStringBuf.append(targetProp.getType().getTypeName() + ", ");
			typesStringBuf.append(extensionProp.getType().getTypeName() + ")");
			String errMsg =
				"attempted to assign two property types "
					+ typesStringBuf.toString()
					+ " to the single property name "
					+ extensionProp.getName();
			throw new AcmePropertyExtensionFailedException(errMsg);
		}
		else {
			// else the target object property and the type property have
			// the same or complimentary types already, so modify nothing.
		}

		// Attempt to unify the values of the type property and the
		// target element property.
		if (extensionProp.getValue() != null) {
			if ((targetProp.getValue() == null) || (targetProp.getValue().getIsDefaultValue())) {
				targetProp.setValue(extensionProp.getValue().copy());
			}
			else if (!targetProp.getValue().equals(extensionProp.getValue())) {
				// then we can't unify the values of the type's property
				// and the object's property, which is an error.
				StringBuffer valuesStringBuf = new StringBuffer("(");
				valuesStringBuf.append(targetProp.getValue().toString() + ", ");
				if (extensionProp.getValue() == null) {
					valuesStringBuf.append("null");
				}
				else {
					valuesStringBuf.append(extensionProp.getValue().toString());
				}
				valuesStringBuf.append(")");

				String errMsg =
					"attempted to assign two unequal non-default "
						+ "property values "
						+ valuesStringBuf.toString()
						+ " to the single property name "
						+ extensionProp.getName();
				throw new AcmePropertyExtensionFailedException(errMsg);
			}
			else {
				// else the target object property and the type property both
				// have the same value already, so modify nothing.
			}
		}
		else {
			//  else leave the current value of targetProp alone.
		}

		// finally, unify the property's meta-props.
		unifyMetaProperties(extensionProp, targetProp);
	}

	/**
	 * helper method to properly instantiate, initialize, and unify
	 * meta properties from a prototype property into a target property
	 * @param prototypeProp Prototype Property
	 * @param targetProp Target Property
	 */
	protected static void unifyMetaProperties(AcmeProperty prototypeProp, AcmeProperty targetProp) {
		unifyMetaProperties(prototypeProp, targetProp, true);
	}

	protected static boolean unifyMetaProperties(
		AcmeProperty prototypeProp,
		AcmeProperty targetProp,
		boolean makeChanges) {
		boolean ok = true;
		// Unify the meta-properties of the prototype with the meta-properties
		// of the targetProp.  Note that it is ok for the values and types of
		// the meta-properties to differ in the target and prototype properties,
		// as meta-properties are not considered in typechecking.  In this
		// case, the type and value of the meta-property the target element
		// is preserved and the prototype's type and value are ignored.
		// Effectively, instances can ignore and over-ride the meta-properties
		// of their types if they choose to do so.

		if ((prototypeProp == null) || (targetProp == null)) {
			return false; // can't unify null parent properties...
		}

		AcmeEnumeration metaProps = prototypeProp.getProperties();
		AcmeProperty prototypeMetaProp;
		AcmeProperty targetMetaProp;
		while (metaProps.moreItems()) {
			prototypeMetaProp = (AcmeProperty) metaProps.getNextItem();
			if (prototypeMetaProp == null) {
				continue; // don't add a null meta-property
			}

			targetMetaProp = targetProp.getMetaProperty(prototypeMetaProp.getName());
			if (targetMetaProp == null) {
				if (makeChanges) {
					try {
						AcmeProperty pCopy = prototypeMetaProp.copy();
						targetProp.addProperty(pCopy);
						pCopy.addInheritedSource(prototypeMetaProp);
					}
					catch (AcmeIllegalNameException ainExc) {
						AcmeError newError =
							new AcmeError(
								targetMetaProp,
								"Attempted to add an illegally named meta-property "
									+ "to property "
									+ targetProp.getName()
									+ ".  message = "
									+ ainExc.getMessage());
						AcmeErrorManager.errorDetected(newError);
					}
				}
				// repeat recursively for any meta-meta-*properties
				ok &= unifyMetaProperties(prototypeMetaProp, targetMetaProp, makeChanges);
				continue;
			}

			if (targetMetaProp.getValue() == null) {
				if (prototypeMetaProp.getValue() != null) {
					if (makeChanges) {
						AcmePropertyValue newPropVal = prototypeMetaProp.getValue().copy();
						newPropVal.setIsDefaultValue(true);
						targetMetaProp.setValue(newPropVal);
					}
				}
				// if targetMetaProp has a value then don't change it.
			}
			else {
				if (targetMetaProp.getValue().getIsDefaultValue() && makeChanges) {
					// override default values with a new value (which is
					// also set as a default value so that it can be
					// subsequently overriden.
					if (prototypeMetaProp.getValue() != null) {
						AcmePropertyValue newPropVal = prototypeMetaProp.getValue().copy();
						newPropVal.setIsDefaultValue(true);
						targetMetaProp.setValue(newPropVal);
					}
				}
			}

			if (targetMetaProp.getType() == null) {
				if (prototypeMetaProp.getType() != null) {
					if (makeChanges) {
						targetMetaProp.setType(prototypeMetaProp.getType());
					}
				}
				// if targetMetaProp has a type then don't change it.
			}

			// repeat recursively for any meta-meta-*properties
			ok &= unifyMetaProperties(prototypeMetaProp, targetMetaProp, makeChanges);
		}
		return ok;
	}

	// RTM 2/15/98 -- Currently, the only method that announces an
	//                InstantiationConflictError is unifyProperties.
	//                Is this really the only place that such a conflict
	//                can occur?  should instantiation errors also leave
	//                an audit trail to determine both the source of the
	//                error and the things that the error affects?

	/**
	*
	* element/type instantiation and unification methods.  These methods
	* implement the unification algorithm specified in the Armani LRM.  They
	* take in a prototype element and an element instance and try to add/merge
	* all of the substructure of the type with the element instance.
	* The result is that targetComp (or targetConn, etc.) has the
	* value of the unification of the type and the target instance's values.
	*
	**/

	/**
	 * Unify Element is the unification method for unifying a generic
	 * ElementProtoType with any kind of element.
	 * @param protoTypeElt Prototype Element
	 * @param targetElt Target Element
	 */
	public static void unifyElementWithPrototype(AcmeElement protoTypeElt, AcmeElement targetElt) {
		unifyElementWithPrototype(protoTypeElt, targetElt, true);
	}

	protected static boolean unifyElementWithPrototype(
		AcmeElement protoTypeElt,
		AcmeElement targetElt,
		boolean makeChanges) {
		boolean ok = true;
		if ((protoTypeElt == null) || (targetElt == null))
			return ok;

		// Fill in super type structure (recurse for each supertype)
		if (protoTypeElt instanceof AcmeElementProtoType) {
			AcmeEnumeration superTypesEnum = ((AcmeElementProtoType) protoTypeElt).getSuperTypes().enumerate();
			while (superTypesEnum.moreItems()) {
				AcmeElementProtoType superType = (AcmeElementProtoType) superTypesEnum.getNextItem();
				ok &= unifyElementWithPrototype(superType, targetElt, makeChanges);
			}
		}

		// All we need to unify with generic elements are properties and reps.
		ok &= unifyProperties(protoTypeElt, targetElt, makeChanges);
		ok &= unifyRepresentations(protoTypeElt, targetElt, makeChanges);

		return ok;
	}

	/*
	* Unify Component
	* @param protoTypeComp Prototype Component
	* @param targetComp Target Component
	*/
	public static void unifyComponent(AcmeComponent protoTypeComp, AcmeComponent targetComp) {
		unifyComponent(protoTypeComp, targetComp, true);
	}

	public static boolean unifyComponent(AcmeComponent protoTypeComp, AcmeComponent targetComp, boolean makeChanges) {

		boolean ok = true;
		if ((protoTypeComp == null) || (targetComp == null))
			return ok;

		// Fill in super type structure (recurse for each supertype)
		if (protoTypeComp instanceof AcmeComponentType) {
			AcmeEnumeration superTypesEnum = ((AcmeComponentType) protoTypeComp).getSuperTypes().enumerate();
			while (superTypesEnum.moreItems()) {
				AcmeElementType superType = (AcmeElementType) superTypesEnum.getNextItem();
				if (superType instanceof AcmeElementTypeRef) {
					superType = ((AcmeElementTypeRef) superType).getReferentType();
				}
				if (superType instanceof AcmeComponentType) {
					ok &= unifyComponent((AcmeComponentType) superType, targetComp, makeChanges);
				}
				else if (superType instanceof AcmeElement) {
					ok &= unifyElementWithPrototype((AcmeElement) superType, targetComp, makeChanges);
				}
			}
		}

		// first unify the properties of these components
		ok &= unifyProperties(protoTypeComp, targetComp, makeChanges);

		// then unify their representations
		ok &= unifyRepresentations(protoTypeComp, targetComp, makeChanges);

		// then unify the ports of the components.
		AcmeEnumeration ports = protoTypeComp.getPorts();
		AcmePort protoPort = null;
		AcmePort targetPort = null;
		while (ports.moreItems()) {
			protoPort = (AcmePort) ports.getNextItem();
			targetPort = targetComp.getPort(protoPort.getName());
			if (targetPort == null) {
				// then there isn't a port with this name in the
				// current component, make a copy of the prototype and add
				// the copy to the targetComp.
				targetPort = (AcmePort) protoPort.copy();
				ok &= unifyPort(protoPort, targetPort, makeChanges);
				if (makeChanges) {
					try {
						targetComp.addPort(targetPort);
						targetPort.addInheritedSource(protoPort);
					}
					catch (AcmeIllegalNameException ainExc) {
						AcmeError newError =
							new AcmeError(
								targetPort,
								"Attempted to add an illegally named port "
									+ "to component "
									+ targetComp.getName()
									+ ".  message = "
									+ ainExc.getMessage());
						AcmeErrorManager.errorDetected(newError);
					}
				}
			}
			else {
				// else we've found a port with the same name as the
				// port we are currently trying to add to the component that
				// already exists in the targetComp.  unifyPorts uses the
				// unification algorithm described in the Armani LRM to
				// attempt to merge the structure of the prototype
				// port with the structure of the existing port.
				ok &= unifyPort(protoPort, targetPort, makeChanges);
			}
		}
		return ok;
	}

	/*
	* Unify Group
	* @param protoTypeComp Prototype Group
	* @param targetComp Target Group
	*/
	public static void unifyGroup(AcmeGroup protoTypeComp, AcmeGroup targetComp) {
		unifyGroup(protoTypeComp, targetComp, true);

	}

	public static boolean unifyGroup(AcmeGroup protoTypeComp, AcmeGroup targetComp, boolean makeChanges) {
		boolean ok = true;
		if ((protoTypeComp == null) || (targetComp == null))
			return ok;

		// Fill in super type structure (recurse for each supertype)
		if (protoTypeComp instanceof AcmeGroupType) {
			AcmeEnumeration superTypesEnum = ((AcmeGroupType) protoTypeComp).getSuperTypes().enumerate();
			while (superTypesEnum.moreItems()) {
				AcmeElementType superType = (AcmeElementType) superTypesEnum.getNextItem();
				if (superType instanceof AcmeElementTypeRef) {
					superType = ((AcmeElementTypeRef) superType).getReferentType();
				}
				if (superType instanceof AcmeGroupType) {
					ok &= unifyGroup((AcmeGroupType) superType, targetComp, makeChanges);
				}
				else if (superType instanceof AcmeElement) {
					ok &= unifyElementWithPrototype((AcmeElement) superType, targetComp, makeChanges);
				}
			}
		}

		// first unify the properties of these components
		ok &= unifyProperties(protoTypeComp, targetComp, makeChanges);

		return ok;
	}

	/**
	 * Unify Connector
	 * @param protoTypeConn Prototype Connector
	 * @param targetConn Target Connector
	 */
	public static void unifyConnector(AcmeConnector protoTypeConn, AcmeConnector targetConn) {
		unifyConnector(protoTypeConn, targetConn, true);
	}
	public static boolean unifyConnector(AcmeConnector protoTypeConn, AcmeConnector targetConn, boolean makeChanges) {
		boolean ok = true;
		if ((protoTypeConn == null) || (targetConn == null))
			return ok;

		// Fill in super type structure (recurse for each supertype)
		if (protoTypeConn instanceof AcmeConnectorType) {
			AcmeEnumeration superTypesEnum = ((AcmeConnectorType) protoTypeConn).getSuperTypes().enumerate();
			while (superTypesEnum.moreItems()) {
				AcmeElementType superType = (AcmeElementType) superTypesEnum.getNextItem();
				if (superType instanceof AcmeElementTypeRef) {
					superType = ((AcmeElementTypeRef) superType).getReferentType();
				}
				if (superType instanceof AcmeConnectorType) {
					ok &= unifyConnector((AcmeConnectorType) superType, targetConn, makeChanges);
				}
				else if (superType instanceof AcmeElement) {
					ok &= unifyElementWithPrototype((AcmeElement) superType, targetConn, makeChanges);
				}
			}
		}

		// first unify the properties of these connectors
		ok &= unifyProperties(protoTypeConn, targetConn, makeChanges);

		// then unify their representations
		ok &= unifyRepresentations(protoTypeConn, targetConn, makeChanges);

		// then unify the roles of the connectors
		AcmeEnumeration roles = protoTypeConn.getRoles();
		AcmeRole protoRole = null;
		AcmeRole targetRole = null;
		while (roles.moreItems()) {
			protoRole = (AcmeRole) roles.getNextItem();
			targetRole = targetConn.getRole(protoRole.getName());
			if (targetRole == null) {
				// then there isn't a role with this name in the
				// current system, make a copy of the prototype and add
				// the copy to the targetConn.
				targetRole = (AcmeRole) protoRole.copy();
				ok &= unifyRole(protoRole, targetRole, makeChanges);
				if (makeChanges) {
					try {
						targetConn.addRole(targetRole);
						targetRole.addInheritedSource(protoRole);
					}
					catch (AcmeIllegalNameException ainExc) {
						AcmeError newError =
							new AcmeError(
								targetRole,
								"Attempted to add an illegally named role "
									+ "to connector "
									+ targetConn.getName()
									+ ".  message = "
									+ ainExc.getMessage());
						AcmeErrorManager.errorDetected(newError);
					}
				}
			}
			else {
				// else we've found a role with the same name as the
				// role we are currently trying to add to the connector that
				// already exists in the targetConn.  unifyRoles uses the
				// unification algorithm described in the Armani LRM to
				// attempt to merge the structure of the prototype
				// role with the structure of the existing role.
				ok &= unifyRole(protoRole, targetRole, makeChanges);
			}
		}
		return ok;
	}

	/**
	 * Unify port
	 * @param protoTypePort Prototype Port
	 * @param targetPort Target Port
	 */
	public static void unifyPort(AcmePort protoTypePort, AcmePort targetPort) {
		unifyPort(protoTypePort, targetPort, true);
	}
	public static boolean unifyPort(AcmePort protoTypePort, AcmePort targetPort, boolean makeChanges) {
		boolean ok = true;
		if ((protoTypePort == null) || (targetPort == null))
			return ok;

		// Fill in super type structure (recurse for each supertype)
		if (protoTypePort instanceof AcmePortType) {
			AcmeEnumeration superTypesEnum = ((AcmePortType) protoTypePort).getSuperTypes().enumerate();
			while (superTypesEnum.moreItems()) {
				AcmeElementType superType = (AcmeElementType) superTypesEnum.getNextItem();
				if (superType instanceof AcmeElementTypeRef) {
					superType = ((AcmeElementTypeRef) superType).getReferentType();
				}
				if (superType instanceof AcmePortType) {
					ok &= unifyPort((AcmePortType) superType, targetPort, makeChanges);
				}
				else if (superType instanceof AcmeElement) {
					ok &= unifyElementWithPrototype((AcmeElement) superType, targetPort, makeChanges);
				}
			}
		}

		// All we need to unify with ports are properties and reps.
		ok &= unifyProperties(protoTypePort, targetPort, makeChanges);
		ok &= unifyRepresentations(protoTypePort, targetPort, makeChanges);

		return ok;
	}

	/**
	 * Unify role
	 * @param protoTypeRole Prototype Role
	 * @param targetRole Target Role
	 */
	public static void unifyRole(AcmeRole protoTypeRole, AcmeRole targetRole) {
		unifyRole(protoTypeRole, targetRole, true);
	}

	public static boolean unifyRole(AcmeRole protoTypeRole, AcmeRole targetRole, boolean makeChanges) {
		boolean ok = true;
		if ((protoTypeRole == null) || (targetRole == null))
			return ok;

		// Fill in super type structure (recurse for each supertype)
		if (protoTypeRole instanceof AcmeRoleType) {
			AcmeEnumeration superTypesEnum = ((AcmeRoleType) protoTypeRole).getSuperTypes().enumerate();
			while (superTypesEnum.moreItems()) {
				AcmeElementType superType = (AcmeElementType) superTypesEnum.getNextItem();
				if (superType instanceof AcmeElementTypeRef) {
					superType = ((AcmeElementTypeRef) superType).getReferentType();
				}
				if (superType instanceof AcmeRoleType) {
					ok &= unifyRole((AcmeRoleType) superType, targetRole, makeChanges);
				}
				else if (superType instanceof AcmeElement) {
					ok &= unifyElementWithPrototype((AcmeElement) superType, targetRole, makeChanges);
				}
			}
		}

		// All we need to unify with roles are properties and reps.
		ok &= unifyProperties(protoTypeRole, targetRole, makeChanges);
		ok &= unifyRepresentations(protoTypeRole, targetRole, makeChanges);

		return ok;
	}

	/**
	 * unify an existing system object with another system or family.
	 * @param protoTypeSys Prototype System
	 * @param targetSys Target System
	 */
	public static void unifySystem(AcmeSystem protoTypeSys, AcmeSystem targetSys) {
		unifySystem(protoTypeSys, targetSys, true);

	}
	public static boolean unifySystem(AcmeSystem protoTypeSys, AcmeSystem targetSys, boolean makeChanges) {
		boolean ok = true;
		if ((protoTypeSys == null) || (targetSys == null))
			return ok;

		if (protoTypeSys instanceof AcmeFamily) {
			// then fill in super type structure (recurse for each supertype)
			AcmeEnumeration superTypesEnum = ((AcmeFamily) protoTypeSys).getSuperTypes().enumerate();
			while (superTypesEnum.moreItems()) {
				AcmeElementType superType = (AcmeElementType) superTypesEnum.getNextItem();
				if (superType instanceof AcmeElementTypeRef) {
					superType = ((AcmeElementTypeRef) superType).getReferentType();
				}
				if (superType instanceof AcmeFamily) {
					ok &= unifySystem((AcmeFamily) superType, targetSys, makeChanges);
				}
				else if (superType instanceof AcmeElement) {
					ok &= unifyElementWithPrototype((AcmeElement) superType, targetSys, makeChanges);
				}

			}
		}

		// add the appropriate properties and representations
		ok &= unifyProperties(protoTypeSys, targetSys, makeChanges);
		ok &= unifyRepresentations(protoTypeSys, targetSys, makeChanges);

		// make copies of all of the components, connectors, ports, roles,
		// and attachments declared in the family and unify them with the target
		// system's substructure.

		// instantiate the family's components
		AcmeEnumeration e = protoTypeSys.getComponents();
		AcmeComponent protoComp = null;
		AcmeComponent newComp = null;
		while (e.moreItems()) {
			try {
				protoComp = (AcmeComponent) e.getNextItem();
				newComp = targetSys.getComponent(protoComp.getName());
				if (newComp == null) {
					// then there isn't a component with this name in the
					// current system, make a copy of the prototype and add
					// the copy.
					newComp = (AcmeComponent) protoComp.copy();
					if (makeChanges)
						targetSys.addComponent(newComp);
				}

				// if newComp already existed then we need to merge it
				// properly with protoComp.  If we created a new copy then we
				// need to instantiate/unify substructure properly.
				ok &= unifyComponent(protoComp, newComp, makeChanges);

			}
			catch (AcmeIllegalNameException ainExc) {
				AcmeError newError =
					new AcmeError(
						newComp,
						"Attempted to add an illegally named component: "
							+ newComp.getName()
							+ " to system "
							+ targetSys.getName()
							+ ".  message = "
							+ ainExc.getMessage());
				AcmeErrorManager.errorDetected(newError);
			}

		}

		// on leaving the previous block, we should have all of the components
		// of this family properly instantiated in the system instance.
		// repeat for the entities named in the comments below.

		// instantiate the family's connectors
		e = protoTypeSys.getConnectors();
		AcmeConnector protoConn = null;
		AcmeConnector newConn = null;
		while (e.moreItems()) {
			try {
				protoConn = (AcmeConnector) e.getNextItem();
				newConn = targetSys.getConnector(protoConn.getName());
				if (newConn == null) {
					// then there isn't a connector with this name in the
					// current system, make a copy of the prototype and add
					// the copy.
					newConn = (AcmeConnector) protoConn.copy();
					if (makeChanges)
						targetSys.addConnector(newConn);
				}

				// if newConn already existed then we need to merge it
				// properly with protoConn.  If we created a new copy then we
				// need to instantiate/unify substructure properly.
				ok &= unifyConnector(protoConn, newConn, makeChanges);

			}
			catch (AcmeIllegalNameException ainExc) {
				AcmeError newError =
					new AcmeError(
						newConn,
						"Attempted to add an illegally named connector: "
							+ newConn.getName()
							+ " to system "
							+ targetSys.getName()
							+ ".  message = "
							+ ainExc.getMessage());
				AcmeErrorManager.errorDetected(newError);
			}
		}

		// unify the ports of the system (if it has any).
		AcmeEnumeration ports = protoTypeSys.getPorts();
		AcmePort protoPort = null;
		AcmePort targetPort = null;
		while (ports.moreItems()) {
			try {
				protoPort = (AcmePort) ports.getNextItem();
				targetPort = targetSys.getPort(protoPort.getName());
				if (targetPort == null) {
					// then there isn't a port with this name in the
					// current system, make a copy of the prototype and add
					// the copy to the targetSys.
					targetPort = (AcmePort) protoPort.copy();
					if (makeChanges)
						targetSys.addPort(targetPort);
				}

				// if targetPort already existed then we need to merge it
				// properly with protoPort.  If we created a new copy then we
				// need to instantiate/unify substructure properly.
				ok &= unifyPort(protoPort, targetPort, makeChanges);

			}
			catch (AcmeIllegalNameException ainExc) {
				AcmeError newError =
					new AcmeError(
						targetPort,
						"Attempted to add an illegally named port: "
							+ targetPort.getName()
							+ " to system "
							+ targetSys.getName()
							+ ".  message = "
							+ ainExc.getMessage());
				AcmeErrorManager.errorDetected(newError);
			}

		}

		// unify the system roles (if it has any)
		AcmeEnumeration roles = protoTypeSys.getRoles();
		AcmeRole protoRole = null;
		AcmeRole targetRole = null;
		while (roles.moreItems()) {
			try {
				protoRole = (AcmeRole) roles.getNextItem();
				targetRole = targetSys.getRole(protoRole.getName());
				if (targetRole == null) {
					// then there isn't a role with this name in the
					// current system, make a copy of the prototype and add
					// the copy to the targetSys.
					targetRole = (AcmeRole) protoRole.copy();
					if (makeChanges)
						targetSys.addRole(targetRole);
				}

				// if targetRole already existed then we need to merge it
				// properly with protoRole.  If we created a new copy then we
				// need to instantiate/unify substructure properly.
				ok &= unifyRole(protoRole, targetRole, makeChanges);

			}
			catch (AcmeIllegalNameException ainExc) {
				AcmeError newError =
					new AcmeError(
						targetRole,
						"Attempted to add an illegally named role: "
							+ targetRole.getName()
							+ " to system "
							+ targetSys.getName()
							+ ".  message = "
							+ ainExc.getMessage());
				AcmeErrorManager.errorDetected(newError);
			}
		}

		// add the family's attachments to the target system.
		// because we want to make the new attachments attach ports and roles
		// in the newly unified system, rather than the prototype system, we
		// can't just copy the attachments.  Rather, we need to create unresolved
		// attachments that store the name and scope of the attached elements,
		// but that can be later resolved to ports and roles in the instance
		// object.
		if (makeChanges) {
			e = protoTypeSys.getAttachments();
			while (e.moreItems()) {
				AcmeUnresolvedAttachment newAttachment = new AcmeUnresolvedAttachment((AcmeAttachment) e.getNextItem());
				newAttachment.setParent(targetSys);
				newAttachment.setParentDesign(targetSys.getParentDesign());
				targetSys.addAttachment(newAttachment);
			}
		}
		return ok;
	}

	public static void unifyMaps(AcmeMapInstance map1, AcmeMapInstance map2) {
		System.out.println("unify map not implemented");
	}

	//
	// Typechecking methods  (both public and protected)
	//

	/**
	 * typecheckSystemStructure is a protected helper method used for
	 * typechecking representations.  The method determines whether
	 * instance_sys's structure satisfies the required structure specified
	 * by prototype_sys. It also checks all properties and recursively
	 * checks all elements in the systems and any representations that they
	 * have for conformance of the instance system's substructure to
	 * the prototype system's substructure.  It does _not_ check for internal
	 * consistency of the instance system and its substructure.
	 * @param prototype_sys ProtoType AcmeSystem
	 * @param instance_sys Instance AcmeSystem
	 * @return true if instance satisfies the structure specified in type
	 */
	protected static boolean typecheckSystemStructure(AcmeSystem prototype_sys, AcmeSystem instance_sys) {
		boolean typechecks = true;

		// first compare the substructure common to all elements -- properties,
		// and representations.
		AcmeFamily proto_fam = (AcmeFamily) prototype_sys.copyAsPrototype("tempFamily");

		if (satisfiesTypeProperties(proto_fam, instance_sys) == false)
			typechecks = false;
		if (satisfiesTypeRepresentations(proto_fam, instance_sys) == false)
			typechecks = false;

		AcmeEnumeration comps, conns, ports, roles, attachments;

		// check that all of the components that prototype_sys requires are
		// existant in instance_sys.
		comps = prototype_sys.getComponents();
		while (comps.moreItems()) {
			AcmeComponent proto_comp = (AcmeComponent) comps.getNextItem();
			if (proto_comp == null)
				continue;
			AcmeComponent instance_comp = instance_sys.getComponent(proto_comp.getName());
			if (instance_comp == null)
				typechecks = false;
			else {
				AcmeComponentType proto_comp_type = (AcmeComponentType) proto_comp.copyAsPrototype("tempCompType");
				if (satisfiesComponentType(proto_comp_type, instance_comp) == false)
					typechecks = false;
				else
					proto_comp_type = null;
			}
		} // end components while loop

		// check that all of the connectors that prototype_sys requires are
		// existant in instance_sys.
		conns = prototype_sys.getConnectors();
		while (conns.moreItems()) {
			AcmeConnector proto_conn = (AcmeConnector) conns.getNextItem();
			if (proto_conn == null)
				continue;
			AcmeConnector instance_conn = instance_sys.getConnector(proto_conn.getName());
			if (instance_conn == null)
				typechecks = false;
			else {
				AcmeConnectorType proto_conn_type = (AcmeConnectorType) proto_conn.copyAsPrototype("tempConnType");
				if (satisfiesConnectorType(proto_conn_type, instance_conn) == false)
					typechecks = false;
				else
					proto_conn_type = null;
			}
		} // end connectors while loop

		// check that all of the ports that prototype_sys requires are
		// existant in instance_sys.
		ports = prototype_sys.getPorts();
		while (ports.moreItems()) {
			AcmePort proto_port = (AcmePort) ports.getNextItem();
			if (proto_port == null)
				continue;
			AcmePort instance_port = instance_sys.getPort(proto_port.getName());
			if (instance_port == null)
				typechecks = false;
			else {
				AcmePortType proto_port_type = (AcmePortType) proto_port.copyAsPrototype("tempPortType");
				if (satisfiesPortType(proto_port_type, instance_port) == false)
					typechecks = false;
				else
					proto_port_type = null;
			}
		} // end ports while loop

		// check that all of the roles that prototype_sys requires are
		// existant in instance_sys.
		roles = prototype_sys.getRoles();
		while (roles.moreItems()) {
			AcmeRole proto_role = (AcmeRole) roles.getNextItem();
			if (proto_role == null)
				continue;
			AcmeRole instance_role = instance_sys.getRole(proto_role.getName());
			if (instance_role == null)
				typechecks = false;
			else {
				AcmeRoleType proto_role_type = (AcmeRoleType) proto_role.copyAsPrototype("tempRoleType");
				if (satisfiesRoleType(proto_role_type, instance_role) == false)
					typechecks = false;
				else
					proto_role_type = null;
			}
		} // end roles while loop

		// RTM 2/15/98 -- It doesn't look like attachments work properly

		// check that the attachments required by prototype_sys are
		// attached in instance_sys.  This comparison is done purely
		// based on the names of the ports, roles, and their parents.
		// It's not clear how to do a better typecheck here.
		attachments = prototype_sys.getAttachments();
		AcmeEnumeration instanceAttachments = instance_sys.getAttachments();

		while (attachments.moreItems()) {
			AcmeAttachment proto_attachment = (AcmeAttachment) attachments.getNextItem();
			String protoPortStr = proto_attachment.getPortNameAsString();
			String protoRoleStr = proto_attachment.getRoleNameAsString();
			if (proto_attachment == null) {
				continue;
			}

			instanceAttachments.reset();
			boolean foundInstanceMatch = false;
			while (instanceAttachments.moreItems()) {
				AcmeAttachment curInstAttach = (AcmeAttachment) instanceAttachments.getNextItem();
				if (protoPortStr != null) {
					if (protoPortStr.equals(curInstAttach.getPortNameAsString())) {
						if (protoRoleStr != null) {
							if (protoRoleStr.equals(curInstAttach.getRoleNameAsString())) {
								foundInstanceMatch = true;
								break;
							}
						}
					}
				}
			}

			if (foundInstanceMatch == false) {
				typechecks = false;
			}
		} // end attachments while loop

		// if we get to here and typechecks has not been set to false then
		// the instance system satisfies the structural type predicate of
		// the prototype system, and typechecks will still be set to true.
		// otherwise, a type-error was encountered and typechecks should
		// have been set to false.
		return typechecks;
	}

	/**
	 * The satisfiesType* methods are helper functions that do a lot of
	 * the standard typechecking that all elements need to do regardless
	 * of their category (component, connector, port, role, or system).
	 * Examples include checking properties, representations, and design
	 * rules.  These are protected because they are intended to be helper
	 * methods called from the public methods for determining whether
	 * an element instance satisfies an element type.  These methods
	 * do _not_ check that the instance's properties that are not required
	 * by the type are internally-consistent.
	 *
	 * check that instance has all of the properties required by type
	 * @param prototype Prototype AcmeElement
	 * @param instance Instance AcmeElement
	 * @return true if instance satisfies the structure specified in prototype
	 */
	protected static boolean satisfiesTypeProperties(AcmeElement prototype, AcmeElement instance) {
		boolean typechecks = true;
		AcmeEnumeration e = prototype.getProperties();
		AcmeProperty type_cur_property, instance_cur_property;
		AcmePropertyType property_type;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (prototype == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		while (e.moreItems()) {
			type_cur_property = (AcmeProperty) e.getNextItem();

			// find the corresponding property in instance
			instance_cur_property = instance.getProperty(type_cur_property.getName());
			if (instance_cur_property == null) {
				typechecks = false;
                // Where does the property come from?
                AcmeElementType origin = null;
                if (prototype instanceof AcmeElementType) {
                    origin = (AcmeElementType )prototype;
                }
                else {
                    // This means that the property is defined in the component type that defines this port
                    if (prototype.getParent() instanceof AcmeElementType) {
                        origin = (AcmeElementType )prototype.getParent ();
                    }
                    else {
                        origin = ConstantTypes.UNRESOLVABLE_ELEMENT_TYPE;
                    }
                }
				ElementTypeSatisfactionError eltTypeErr =
					new ElementTypeSatisfactionError(
						instance,
						instance.getName()
							+ " does not satisfy type "
							+ origin.getTypeName()
							+ " because it does not provide "
							+ "the property "
							+ type_cur_property.getName()
							+ " required by the type.",
						instance,
						origin);
				typecheckingErrorStack.pushCausalError(eltTypeErr);
			}

			// see if a property type is required, and if so whether the instance
			// property's type is a subtype of the type's property type.
			else if (type_cur_property.getType() != null) {
				if (!(isLogicalSubtype(instance_cur_property.getType(), type_cur_property.getType(), false))) {
					typechecks = false;
					ElementTypeSatisfactionError eltTypeErr =
						new ElementTypeSatisfactionError(
							instance,
							instance.getName()
								+ " does not satisfy type "
								+ prototype.getName()
								+ " because its property "
								+ type_cur_property.getName()
								+ "'s type is not "
								+ "a legal subtype of "
								+ type_cur_property.getType().getTypeName());
					typecheckingErrorStack.pushCausalError(eltTypeErr);
				}
			}

			// see if a property value is required, and if so whether they match
			if (type_cur_property.getValue() != null) {
				if (instance_cur_property == null) {
					typechecks = false;
					ElementTypeSatisfactionError eltTypeErr =
						new ElementTypeSatisfactionError(
							instance,
							instance.getName()
								+ " does not satisfy type "
								+ prototype.getName()
								+ " because its property "
								+ type_cur_property.getName()
								+ "'s value must be "
								+ type_cur_property.getValue().toString()
								+ ", but no value is provided in the instance.");
					typecheckingErrorStack.pushCausalError(eltTypeErr);

				}
				else if (!(type_cur_property.getValue().equals(instance_cur_property.getValue()))) {
					if (!type_cur_property.getValue().getIsDefaultValue()) {
						typechecks = false;
						ElementTypeSatisfactionError eltTypeErr =
							new ElementTypeSatisfactionError(
								instance,
								instance.getName()
									+ " does not satisfy type "
									+ prototype.getName()
									+ " because its property "
									+ type_cur_property.getName()
									+ "'s value must be "
									+ type_cur_property.getValue().toString()
									+ ", but instead it's value is "
									+ instance_cur_property.getValue().toString());
						typecheckingErrorStack.pushCausalError(eltTypeErr);
					}
				}
			}

			// check that this property in the instance typechecks with
			// its own type specification.
			if (instance_cur_property != null) {
				property_type = instance_cur_property.getType();
				if (property_type != null) {
					if (typecheckProperty(property_type, instance_cur_property.getValue()) == false) {
						typechecks = false;
						ElementTypeSatisfactionError eltTypeErr =
							new ElementTypeSatisfactionError(
								instance,
								instance.getName()
									+ " does not satisfy type "
									+ prototype.getName()
									+ "'s property "
									+ instance_cur_property.getName()
									+ "'s value does not "
									+ "satisfy its type "
									+ property_type.getTypeName());
						typecheckingErrorStack.pushCausalError(eltTypeErr);
					}
				}
			}
		}

		return typechecks;
	}

	/**
	 * Check that the representations of an element satisfy the representations
	 * required by the reference elementType passed in as a parameter
	 * @param prototype Prototype AcmeElement
	 * @param instance Instance AcmeElement
	 * @return true if instance satisfies the structure specified in prototype
	 */
	protected static boolean satisfiesTypeRepresentations(AcmeElement prototype, AcmeElement instance) {
		// RTM 2/9/98 -- Add error-handling here for appropriately registering
		//               exactly what it is that doesn't typecheck.

		// try to prove that the instance's representations do not satisfy
		// the requirements of the prototype element's representations.
		// if we can not prove that they don't typecheck then return true to
		// indicate that they do typecheck, else return false.
		AcmeEnumeration type_reps = prototype.getRepresentations();

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (prototype == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		while (type_reps.moreItems()) {
			AcmeRepresentation type_rep = (AcmeRepresentation) type_reps.getNextItem();
			if (type_rep == null)
				continue;
			else if (type_rep.getSystem() == null)
				continue;
			else {
				AcmeRepresentation instance_rep;
				instance_rep = instance.getRepresentation(type_rep.getName());
				if (instance_rep == null) {
					// there is no rep with the appropriate name.  announce error and bail
					ElementTypeSatisfactionError eltTypeErr =
						new ElementTypeSatisfactionError(
							instance,
							instance.getName()
								+ " does not satisfy type "
								+ prototype.getName()
								+ " because it does not provide "
								+ "the representation "
								+ type_rep.getName()
								+ " required by the type.",
							instance,
							(AcmeElementType) prototype);
					typecheckingErrorStack.pushCausalError(eltTypeErr);

					return false;

				}
				else {
					AcmeSystem prototype_sys, instance_sys;
					prototype_sys = type_rep.getSystem();
					instance_sys = instance_rep.getSystem();
					if (typecheckSystemStructure(prototype_sys, instance_sys) == false) {
						// then the representation doesn't have the required structure.
						ElementTypeSatisfactionError eltTypeErr =
							new ElementTypeSatisfactionError(
								instance,
								instance.getName()
									+ " does not satisfy type "
									+ prototype.getName()
									+ " because its representation "
									+ type_rep.getName()
									+ " does not meet the type's "
									+ " requirements..",
								instance,
								(AcmeElementType) prototype);
						typecheckingErrorStack.pushCausalError(eltTypeErr);
						return false;
					}
				}
			}
		}

		// if the test gets to here it means that we haven't been able to
		// prove that the representations do _not_ typecheck, therefore
		// assume they do typecheck.
		return true;
	}

	/**
	 * check whether instance satisfies each of the design rules (both
	 * invariants and heuristics) declared in type.
	 * @param prototype Prototype AcmeElement
	 * @param instance Instance AcmeElement
	 * @return true if instance satisfies the structure specified in prototype
	 */
	protected static boolean satisfiesTypeDesignRules(AcmeElement prototype, AcmeElement instance) {
		boolean typechecks = true;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (prototype == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		// evaluate all of the design invariants of the prototype, in the
		// context of the instance.
		AcmeEnumeration e = prototype.getInvariants(false);
		while (e.moreItems()) {
			DesignRule curInvariant = (DesignRule) e.getNextItem();
			Scope prototypeScope = curInvariant.getDesignRuleScope();

			// if the invariant hasn't been typechecked then do so...
			if (curInvariant.getWasTypeChecked() == false) {
				curInvariant.setDesignRuleScope(instance);
				if (curInvariant.typecheck() == false) {
					typechecks = false;
					curInvariant.setDesignRuleScope(prototypeScope);
					continue;
				}
				curInvariant.setDesignRuleScope(prototypeScope);
			}
			else if (curInvariant.getIsTypeCorrect() == false) {
				typechecks = false;
				continue;
				// don't even try to evaluate if it doesn't typecheck.
			}

			// set the scope of the prototype's design rule to the instance
			// and evaluate it.  do this in a try block to insure that the
			// scope of the type's design rule is definitely set back to prototype.
			try {
				curInvariant.setDesignRuleScope(instance);
				if ((curInvariant.evaluate() == false) && curInvariant.getIsInvariant()) {
					typechecks = false;
					if (!(prototype instanceof AcmeElementType)) {
						if (prototype.getParent() instanceof AcmeElementType) {
							prototype = prototype.getParent ();
						}
						else {
							prototype = null;
						}
					}
					ElementTypeSatisfactionError eltTypeErr =
						new ElementTypeSatisfactionError(
							instance,
							instance.getName()
								+ " does not satisfy type "
								+ prototype.getName()
								+ "'s invariant expression \""
								+ curInvariant.toString()
								+ "\". ",
							instance,
							(AcmeElementType) prototype);
					typecheckingErrorStack.pushCausalError(eltTypeErr);
					AcmeErrorManager.errorDetected(eltTypeErr);
				}
			}
			catch (DesignRuleEvalException drException) {
				curInvariant.setDesignRuleScope(prototypeScope);
				typechecks = false;
				ElementTypeSatisfactionError eltTypeErr =
					new ElementTypeSatisfactionError(
						instance,
						"Evaluation of type "
							+ prototype.getName()
							+ "'s invariant expression \""
							+ curInvariant.toString()
							+ "\" caused an evaluation exception.",
						instance,
						(AcmeElementType) prototype);
				typecheckingErrorStack.pushCausalError(eltTypeErr);

			}
			finally {
				curInvariant.setDesignRuleScope(prototypeScope);
			}
		}

		// evaluate all of the design heuristics of the prototype, in the
		// context of the instance.  Note that heuristics that are not
		// satisfied just announce an anomalous event, they don't cause
		// the instance not to typecheck.
		e = prototype.getHeuristics(false);
		while (e.moreItems()) {
			DesignRule curHeuristic = (DesignRule) e.getNextItem();
			Scope prototypeScope = curHeuristic.getDesignRuleScope();

			// if the invariant hasn't been typechecked then do so...
			if (curHeuristic.getWasTypeChecked() == false) {
				if (curHeuristic.typecheck() == false) {
					typechecks = false;
					continue;
				}
			}
			else if (curHeuristic.getIsTypeCorrect() == false) {
				typechecks = false;
				continue;
				// don't even try to evaluate if it doesn't typecheck.
			}

			// set the scope of the prototype's design rule to the instance
			// and evaluate it.  do this in a try block to insure that the
			// scope of the type's design rule is definitely set back to prototype.
			try {
				curHeuristic.setDesignRuleScope(instance);
				if (curHeuristic.evaluate() == false) {
					String warningMsg =
						"Warning.  Heuristic violation discovered. "
							+ "Instance \""
							+ instance.getName()
							+ "\" does not satisfy type \""
							+ prototype.getName()
							+ "'s\" heuristic: "
							+ curHeuristic.getDesignRuleExpression().toString();
					HeuristicViolationError newWarning =
						new HeuristicViolationError(instance, warningMsg, curHeuristic, instance);
					AcmeErrorManager.errorDetected(newWarning);
				}
			}
			catch (DesignRuleEvalException drException) {
				curHeuristic.setDesignRuleScope(prototypeScope);
			}
			finally {
				curHeuristic.setDesignRuleScope(prototypeScope);
			}
		}

		return typechecks;
	}

	/** 
	 * The declaresType (type, instance) method checks that the instance
	 * declares the type passed in. It is more restrictive than satisfiesType
	 * which checks the structure. declaresType implements name equivalence 
	 * type checking, satisfiesType implements structural type checking
	 * 
	 * For elements other than AcmeElements, this method calls satisfiesType
	 * @param type The type to check the declaration of
	 * @param instance The instance to check
	 * @return true if the element declares the type
	 */
	public static boolean declaresType(AcmeType type, Object instance) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType();
		}
		if ((type instanceof AcmeElementType) && (instance instanceof AcmeElement)) {
			AcmeElementType t = (AcmeElementType) type;
			AcmeElement el = (AcmeElement) instance;
			return el.declaresType(t);
		}
		else {
			return satisfiesType(type, instance);
		}

	}

	/**
	 * The satisifies*Type(type, instance) methods check that the instance
	 * contains all of the structure, properties, and substructure required
	 * by the type.  It recursively checks this for all of the superTypes of
	 * type, as well.  These methods make use of the satisfiesType*() helper
	 * methods.  satisfies*Type(...) methods do _not_ check that the
	 * instance-based properties and substructure of "instance" are internally
	 * type-correct.  It only checks that the requirements of "type" are
	 * present in "instance".
	 *
	 * RTM 2/16/98 -- Currently, the satisfies*Type(...) methods do not
	 *                participate in the detailed error tracking with
	 *                AcmeError objects and the AcmeErrorManager.  Any
	 *                errors discovered by these methods should, in fact,
	 *                be recorded by the calling method, but, if I was
	 *                so inclined, I could add the detailed error handling
	 *                to these methods to make it more clear exactly _why_
	 *                an instance doesn't satisfy its type.  I should
	 *                probably do this eventually, but for now it is not
	 *                critical.
	 *
	 *
	 * satisfiesType(type, object) checks for a basic type-object match,
	 * and if it finds one it calls the appropriate satisfies*Type method
	 * to check whether object satisfies type.  else it returns false
	 * immediately.
	 * @param type The AcmeType
	 * @param obj The Object
	 * @return true if obj satisfies the structure specified in type
	 */
	public static boolean satisfiesType(AcmeType type, Object obj) {
		boolean typechecks = false;
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType();
		}
		else if (type instanceof AcmePropertyTypeRef) {
			type = ((AcmePropertyTypeRef) type).getReferentType();
		}
		// first check for any
		if (type instanceof AnyType) {
			return true; // everything satisfies Any.
		}

		// special case: any property object satisfies a property reference type.
		else if ((type instanceof PropertyReferenceType) && (obj instanceof AcmeProperty)) {
			return true;
		}

		// then check for Acme and Armani-specific types & objects
		else if ((type instanceof AcmeElementType) && (obj instanceof AcmeElement)) {
			return satisfiesElementType((AcmeElementType) type, (AcmeElement) obj);
		}
		// BRS: Added checking for representations to do the type checking
		else if ((type instanceof AcmeFamily) && (obj instanceof AcmeSystem)) {
			return satisfiesFamily((AcmeFamily) type, ((AcmeSystem) obj));
		}
		else if ((type instanceof AcmePropertyType) && (obj instanceof AcmeProperty)) {
			return typecheckProperty(type, ((AcmeProperty) obj).getValue());
		}
		else if ((type instanceof AcmePropertyType) && (obj instanceof AcmePropertyValue)) {
			return typecheckProperty((AcmePropertyType) type, (AcmePropertyValue) obj);
		}
		else if (obj instanceof ExpressionNode) {
			return isLogicalSubtype(((ExpressionNode) obj).getType(), type, false);
		}

		// then check for Java objects that can satisfy Acme/Armani types.
		else if (obj instanceof Boolean) {
			return (type instanceof BooleanType);
		}
		else if (obj instanceof Float) {
			return (type instanceof FloatType);
		}
		else if (obj instanceof Integer) {
			return (type instanceof IntType);
		}
		else if (obj instanceof String) {
			return (type instanceof StringType);
		}

		return typechecks;
	}

	/**
	 * satisfiesElementType(type,instance) is a generic dispatch method that
	 * calls the appropriate category's satisfies* method if the category
	 * of the element matches the category of the type.
	 * @param type The AcmeElementType
	 * @param instance The AcmeElement
	 * @return true if instance satisfies the structure specified in type
	 */
	public static boolean satisfiesElementType(AcmeElementType type, AcmeElement instance) {
		if (type instanceof AcmeElementTypeRef) {
			type = ((AcmeElementTypeRef) type).getReferentType();
		}
		// every element satisfies ELEMENT_BASE_TYPE
		if ((type == ConstantTypes.ELEMENT_BASE_TYPE) || (type == null)) {
			return true;
		}

		if (type == instance) {
			return true;
		}

		if ((type instanceof AcmeComponentType) && (instance instanceof AcmeComponent)) {
			return satisfiesComponentType((AcmeComponentType) type, (AcmeComponent) instance);
		}
		else if ((type instanceof AcmeConnectorType) && (instance instanceof AcmeConnector)) {
			return satisfiesConnectorType((AcmeConnectorType) type, (AcmeConnector) instance);
		}
		else if ((type instanceof AcmePortType) && (instance instanceof AcmePort)) {
			return satisfiesPortType((AcmePortType) type, (AcmePort) instance);
		}
		else if ((type instanceof AcmeRoleType) && (instance instanceof AcmeRole)) {
			return satisfiesRoleType((AcmeRoleType) type, (AcmeRole) instance);
		}
		else if ((type instanceof AcmeFamily) && (instance instanceof AcmeSystem)) {
			return satisfiesFamily((AcmeFamily) type, (AcmeSystem) instance);
		}
		else if ((type instanceof AcmeElementProtoType)) {
			return satisfiesEltProtoType((AcmeElementProtoType) type, instance);
		}
		else {
			// else we don't have a legal element/element-type pair.
			return false;
		}
	}

	/**
	 * satisfiesEltProtoType(type,instance) determines whether instance
	 * satifies all of the requirements of type.  It only checks that the
	 * requirements of type are satisfied by instance.  It does not check
	 * instance for any kind of internal type-consistency. Returns true
	 * if instance satisfies type, else false.
	 * This method checks any instance of an AcmeElement against a generic
	 * AcmeElementPrototype.
	 *
	 * @param type The AcmeElementProtoType to check against
	 * @param instance The AcmeElement to check
	 * @return true if instance satisfies the structure specified in type
	 */
	public static boolean satisfiesEltProtoType(AcmeElementProtoType type, AcmeElement instance) {
		boolean typechecks = true;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (type == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		// recursively check that instance satisfies all of type's supertypes.
		AcmeEnumeration super_types_enum = type.getSuperTypes().enumerate();
		while (super_types_enum.moreItems()) {
			AcmeElementProtoType super_type = (AcmeElementProtoType) super_types_enum.getNextItem();
			if (satisfiesEltProtoType(super_type, instance) == false) {
				return false;
			}
		}

		// then check that this instance satisfies all of the properties,
		// representations, and design rules required by type.
		if ((satisfiesTypeProperties(type, instance)
			&& satisfiesTypeRepresentations(type, instance)
			&& satisfiesTypeDesignRules(type, instance))
			== false) {
			return false;
		}

		return typechecks;
	}

	/**
	 * satisfiesFamily(type,instance) determines whether instance satifies
	 * all of the requirements of type.  It only checks that the requirements
	 * of type are satisfied by instance.  It does not check instance for
	 * any kind of internal type-consistency. Returns true if instance
	 * satisfies type, else false.
	 * @param type The AcmeFamily
	 * @param instance The AcmeSystem
	 * @return true if instance satisfies the structure specified in type
	 */
	public static boolean satisfiesFamily(AcmeFamily type, AcmeSystem instance) {
		boolean typechecks = true;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (type == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		// recursively check that instance satisfies all of type's supertypes.
		AcmeEnumeration super_types_enum = type.getSuperTypes().enumerate();
		while (super_types_enum.moreItems()) {
			AcmeElementType super_type = (AcmeElementType) super_types_enum.getNextItem();
			if (super_type instanceof AcmeFamily) {
				if (satisfiesFamily((AcmeFamily) super_type, instance) == false) {
					return false;
				}
			}
			else if (super_type instanceof AcmeElementProtoType) {
				if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance) == false) {
					typechecks = false;
				}
			}
		}

		// then check that this instance satisfies all of the properties,
		// representations, and design rules required by type.
		if ((satisfiesTypeProperties(type, instance)
			&& satisfiesTypeRepresentations(type, instance)
			&& satisfiesTypeDesignRules(type, instance))
			== false) {
			typechecks = false;
		}

		// check that the instance contains all of the components and
		// component substructure required by the type.
		AcmeEnumeration e = type.getComponents();
		while (e.moreItems()) {
			AcmeComponent type_comp = (AcmeComponent) e.getNextItem();
			if (type_comp == null)
				continue;
			AcmeComponent instance_comp = instance.getComponent(type_comp.getName());
			if (instance_comp == null) {
				typechecks = false;
				continue;
			}

			// make sure that the instance component satisifies all of the
			// type_comp's types
			AcmeEnumeration typeEnum = type_comp.getDeclaredTypes();
			while (typeEnum.moreItems()) {
				AcmeElementType super_type = (AcmeElementType) typeEnum.getNextItem();
				if (super_type instanceof AcmeElementTypeRef) {
					super_type = ((AcmeElementTypeRef) super_type).getReferentType();
				}
				if (super_type instanceof AcmeComponentType) {
					if (satisfiesComponentType((AcmeComponentType) super_type, instance_comp) == false) {
						typechecks = false;
					}
				}
				else if (super_type instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance_comp) == false) {
						typechecks = false;
					}
				}
			}

			// check that instance_comp satisfies all of the requirements
			// of type_comp in this family (in addition to type_comp's types)
			AcmeComponentType prototype_comp = (AcmeComponentType) type_comp.copyAsPrototype("tempCompType");
			if (satisfiesComponentType(prototype_comp, instance_comp) == false)
				typechecks = false;

			// make sure that the properties, reps and design rules of the
			// type's component are included in the instance's role.
			if (satisfiesTypeProperties(type_comp, instance_comp) == false)
				typechecks = false;
			if (satisfiesTypeRepresentations(type_comp, instance_comp) == false)
				typechecks = false;
			if (satisfiesTypeDesignRules(type_comp, instance_comp) == false)
				typechecks = false;
		}

		// check that the instance contains all of the connectors and
		// connector substructure required by the type.
		e = type.getConnectors();
		while (e.moreItems()) {
			AcmeConnector type_conn = (AcmeConnector) e.getNextItem();
			if (type_conn == null)
				continue;
			AcmeConnector instance_conn = instance.getConnector(type_conn.getName());
			if (instance_conn == null) {
				typechecks = false;
				continue;
			}

			// make sure that the instance connector satisifies all of the
			// type_conn's types
			AcmeEnumeration typeEnum = type_conn.getDeclaredTypes();
			while (typeEnum.moreItems()) {
				AcmeElementType super_type = (AcmeElementType) typeEnum.getNextItem();
				if (super_type instanceof AcmeElementTypeRef) {
					super_type = ((AcmeElementTypeRef) super_type).getReferentType();
				}
				if (super_type instanceof AcmeConnectorType) {
					if (satisfiesConnectorType((AcmeConnectorType) super_type, instance_conn) == false) {
						typechecks = false;
					}
				}
				else if (super_type instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance_conn) == false) {
						typechecks = false;
					}
				}
			}

			// check that instance_conn satisfies all of the requirements
			// of type_conn in this family (in addition to type_conn's types)
			AcmeConnectorType prototype_conn = (AcmeConnectorType) type_conn.copyAsPrototype("tempConnType");
			if (satisfiesConnectorType(prototype_conn, instance_conn) == false)
				typechecks = false;

			// make sure that the properties, reps and design rules of the
			// type's connector are included in the instance's role.
			if (satisfiesTypeProperties(type_conn, instance_conn) == false)
				typechecks = false;
			if (satisfiesTypeRepresentations(type_conn, instance_conn) == false)
				typechecks = false;
			if (satisfiesTypeDesignRules(type_conn, instance_conn) == false)
				typechecks = false;
		}

		// if we're dealing with Acme then we need to check for ports and
		// roles also...
		if (acmeLib.core.globals.LanguageConfig.IS_PURE_ACME == true) {

			// check that the instance contains all of the ports and
			// port substructure required by the type.
			e = type.getPorts();
			while (e.moreItems()) {
				AcmePort type_port = (AcmePort) e.getNextItem();
				if (type_port == null)
					continue;
				AcmePort instance_port = instance.getPort(type_port.getName());
				if (instance_port == null) {
					typechecks = false;
					continue;
				}

				// make sure that the instance connector satisifies all of the
				// type_port's types
				AcmeEnumeration typeEnum = type_port.getDeclaredTypes();
				while (typeEnum.moreItems()) {
					AcmeElementType super_type = (AcmeElementType) typeEnum.getNextItem();
					if (super_type instanceof AcmeElementTypeRef) {
						super_type = ((AcmeElementTypeRef) super_type).getReferentType();
					}
					if (super_type instanceof AcmePortType) {
						if (satisfiesPortType((AcmePortType) super_type, instance_port) == false) {
							typechecks = false;
						}
					}
					else if (super_type instanceof AcmeElementProtoType) {
						if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance_port) == false) {
							typechecks = false;
						}
					}
				}

				// check that instance_port satisfies all of the requirements
				// of type_port in this family (in addition to type_port's types)
				AcmePortType prototype_port = (AcmePortType) type_port.copyAsPrototype("tempPortType");
				if (satisfiesPortType(prototype_port, instance_port) == false)
					typechecks = false;

				// make sure that the properties, reps and design rules of the
				// type's port are included in the instance's port.
				if (satisfiesTypeProperties(type_port, instance_port) == false)
					typechecks = false;
				if (satisfiesTypeRepresentations(type_port, instance_port) == false)
					typechecks = false;
				if (satisfiesTypeDesignRules(type_port, instance_port) == false)
					typechecks = false;
			}

			// check that the instance contains all of the roles and
			// role substructure required by the type.
			e = type.getRoles();
			while (e.moreItems()) {
				AcmeRole type_role = (AcmeRole) e.getNextItem();
				if (type_role == null)
					continue;
				AcmeRole instance_role = instance.getRole(type_role.getName());
				if (instance_role == null) {
					typechecks = false;
					continue;
				}

				// make sure that the instance connector satisifies all of the
				// type_role's types
				AcmeEnumeration typeEnum = type_role.getDeclaredTypes();
				while (typeEnum.moreItems()) {
					AcmeElementType super_type = (AcmeElementType) typeEnum.getNextItem();
					if (super_type instanceof AcmeElementTypeRef) {
						super_type = ((AcmeElementTypeRef) super_type).getReferentType();
					}
					if (super_type instanceof AcmeRoleType) {
						if (satisfiesRoleType((AcmeRoleType) super_type, instance_role) == false) {
							typechecks = false;
						}
					}
					else if (super_type instanceof AcmeElementProtoType) {
						if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance_role) == false) {
							typechecks = false;
						}
					}
				}

				// check that instance_role satisfies all of the requirements
				// of type_role in this family (in addition to type_role's types)
				AcmeRoleType prototype_role = (AcmeRoleType) type_role.copyAsPrototype("tempRoleType");
				if (satisfiesRoleType(prototype_role, instance_role) == false)
					typechecks = false;

				// make sure that the properties, reps and design rules of the
				// type's role are included in the instance's role.
				if (satisfiesTypeProperties(type_role, instance_role) == false)
					typechecks = false;
				if (satisfiesTypeRepresentations(type_role, instance_role) == false)
					typechecks = false;
				if (satisfiesTypeDesignRules(type_role, instance_role) == false)
					typechecks = false;
			}

		} // end if IS_PURE_ACME

		// check to make sure that all of the attachments required by this
		// family exist in the instance being checked.  Compare the simple
		// string names of the attached entities (and their parents) for
		// equality, since the relationship between the actual attached objects
		// being compared should be that the instance's attachment objects
		// are copies of the type's attachment objects.
		AcmeEnumeration typeAttachments = type.getAttachments();
		AcmeEnumeration instanceAttachments = instance.getAttachments();
		String typePortName, instancePortName, typeRoleName, instanceRoleName;
		AcmeAttachment typeAttachment = null;
		AcmeAttachment instanceAttachment = null;
		while (typeAttachments.moreItems()) {
			typeAttachment = (AcmeAttachment) typeAttachments.getNextItem();
			instanceAttachments.reset();
			boolean matchFound = false;
			while (instanceAttachments.moreItems() && !matchFound) {
				instanceAttachment = (AcmeAttachment) instanceAttachments.getNextItem();
				instancePortName = instanceAttachment.getPortNameAsString();
				instanceRoleName = instanceAttachment.getRoleNameAsString();
				typePortName = typeAttachment.getPortNameAsString();
				typeRoleName = typeAttachment.getRoleNameAsString();
				if ((instanceRoleName == null)
					|| (instancePortName == null)
					|| (typeRoleName == null)
					|| (typePortName == null)) {
					continue;
				}
				else if (instanceRoleName.equals(typeRoleName) && instancePortName.equals(typePortName)) {
					matchFound = true;
					// check whether the attachment's properties match,
					// (but only if we're dealing with PURE_ACME
					if (acmeLib.core.globals.LanguageConfig.IS_PURE_ACME == true) {
						if (!satisfiesTypeProperties(typeAttachment, instanceAttachment)) {
							typechecks = false;
							String errMsg =
								"Properties of system instance "
									+ instance.getName()
									+ "'s attachment: \""
									+ typeAttachment.getPortNameAsString()
									+ " to "
									+ typeAttachment.getRoleNameAsString()
									+ "\" do not typecheck. ";
							ElementTypeSatisfactionError newETError =
								new ElementTypeSatisfactionError(instance, errMsg, instance, type);
							typecheckingErrorStack.pushCausalError(newETError);
						}
					}
				}
			}

			if (matchFound == false) {
				typechecks = false;
				String errMsg =
					"Family "
						+ type.getName()
						+ " requires attachment: \""
						+ typeAttachment.getPortNameAsString()
						+ " to "
						+ typeAttachment.getRoleNameAsString()
						+ "\", which is not found in system "
						+ instance.getName();

				ElementTypeSatisfactionError newETError =
					new ElementTypeSatisfactionError(instance, errMsg, instance, type);
				typecheckingErrorStack.pushCausalError(newETError);
			}
		}

		// typechecks should now hold the typechecking status of this
		// (type,instance) pair.
		return typechecks;
	}

	/**
	 * satisfiesMapType(type,instance) determines whether instance
	 * satifies all of the requirements of type.  It only checks that the
	 * requirements of type are satisfied by instance.  It does not check
	 * instance for any kind of internal type-consistency. Returns true
	 * if instance satisfies type, else false.
	 * @param type The AcmeMapType
	 * @param instance The AcmeMapInstance
	 * @return true if instance satisfies the constrains specified by type
	 */
	public static boolean satisfiesMapType(AcmeMapType type, AcmeMapInstance instance) {
		boolean typechecks = true;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (type == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		// recursively check that instance satisfies all of type's supertypes.
		AcmeEnumeration super_types_enum = type.getSuperTypes().enumerate();
		while (super_types_enum.moreItems()) {
			AcmeElementType super_type = (AcmeElementType) super_types_enum.getNextItem();
			if (super_type instanceof AcmeElementTypeRef) {
				super_type = ((AcmeElementTypeRef) super_type).getReferentType();
			}
			if (super_type instanceof AcmeMapType) {
				if (satisfiesMapType((AcmeMapType) super_type, instance) == false) {
					typechecks = false;
				}
			}
			else if (super_type instanceof AcmeElementProtoType) {
				if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance) == false) {
					typechecks = false;
				}
			}
		}

		// confirm that all of map's maplets typecheck.  Any errors discovered
		// while typechecking the maplets will be put properly at the top
		// of the typecheckingErrorStack in the typecheckPort method.
		AcmeEnumeration e = instance.getMaplets();
		while (e.moreItems()) {
			AcmeMaplet cur_maplet = (AcmeMaplet) e.getNextItem();
			if (typecheckMaplet(cur_maplet) == false)
				typechecks = false;
		}
		return typechecks;
	}

	/**
	 * satisfiesComponentType(type,instance) determines whether instance
	 * satifies all of the requirements of type.  It only checks that the
	 * requirements of type are satisfied by instance.  It does not check
	 * instance for any kind of internal type-consistency. Returns true
	 * if instance satisfies type, else false.
	 * @param type The AcmeComponentType
	 * @param instance The AcmeComponent
	 * @return true if instance satisfies the structure specified in type
	 */
	public static boolean satisfiesComponentType(AcmeComponentType type, AcmeComponent instance) {
		boolean typechecks = true;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (type == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		// recursively check that instance satisfies all of type's supertypes.
		AcmeEnumeration super_types_enum = type.getSuperTypes().enumerate();
		while (super_types_enum.moreItems()) {
			AcmeElementType super_type = (AcmeElementType) super_types_enum.getNextItem();
			if (super_type instanceof AcmeComponentType) {
				if (satisfiesComponentType((AcmeComponentType) super_type, instance) == false) {
					typechecks = false;
				}
			}
			else if (super_type instanceof AcmeElementProtoType) {
				if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance) == false) {
					typechecks = false;
				}
			}
		}

		// then check that this instance satisfies all of the properties,
		// representations, and design rules required by type.
		if ((satisfiesTypeProperties(type, instance)
			&& satisfiesTypeRepresentations(type, instance)
			&& satisfiesTypeDesignRules(type, instance))
			== false) {
			typechecks = false;
		}

		// check that the instance contains all of the ports and port
		// substructure required by the type.
		AcmeEnumeration e = type.getPorts();
		while (e.moreItems()) {
			AcmePort type_port = (AcmePort) e.getNextItem();
			if (type_port == null)
				continue;
			AcmePort instance_port = instance.getPort(type_port.getName());
			if (instance_port == null) {
				typechecks = false;
				continue;
			}

			// make sure that the instance port satisifies all of the
			// type_port's types
			AcmeEnumeration typeEnum = type_port.getDeclaredTypes();
			while (typeEnum.moreItems()) {
				AcmeElementType super_type = (AcmeElementType) typeEnum.getNextItem();
				if (super_type instanceof AcmeElementTypeRef) {
					super_type = ((AcmeElementTypeRef) super_type).getReferentType();
				}
				if (super_type instanceof AcmePortType) {
					if (satisfiesPortType((AcmePortType) super_type, instance_port) == false) {
						typechecks = false;
					}
				}
				else if (super_type instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance_port) == false) {
						typechecks = false;
					}
				}
			}

			// make sure that the properties, reps and design rules of the
			// type's port are included in the instance's role.
			if (satisfiesTypeProperties(type_port, instance_port) == false)
				typechecks = false;
			if (satisfiesTypeRepresentations(type_port, instance_port) == false)
				typechecks = false;
			if (satisfiesTypeDesignRules(type_port, instance_port) == false)
				typechecks = false;
		}

		return typechecks;
	}

	/**
	 * satisfiesConnectorType(type,instance) determines whether instance
	 * satifies all of the requirements of type.  It only checks that the
	 * requirements of type are satisfied by instance.  It does not check
	 * instance for any kind of internal type-consistency. Returns true
	 * if instance satisfies type, else false.
	 * @param type The AcmeConnectorType
	 * @param instance The AcmeConnector
	 * @return true if instance satisfies the structure specified in type
	 */
	public static boolean satisfiesConnectorType(AcmeConnectorType type, AcmeConnector instance) {
		boolean typechecks = true;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (type == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		// recursively check that instance satisfies all of type's supertypes.
		AcmeEnumeration super_types_enum = type.getSuperTypes().enumerate();
		while (super_types_enum.moreItems()) {
			AcmeElementType super_type = (AcmeElementType) super_types_enum.getNextItem();
			if (super_type instanceof AcmeElementTypeRef) {
				super_type = ((AcmeElementTypeRef) super_type).getReferentType();
			}
			if (super_type instanceof AcmeConnectorType) {
				if (satisfiesConnectorType((AcmeConnectorType) super_type, instance) == false) {
					typechecks = false;
				}
			}
			else if (super_type instanceof AcmeElementProtoType) {
				if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance) == false) {
					typechecks = false;
				}
			}
		}

		// then check that this instance satisfies all of the properties,
		// representations, and design rules required by type.
		if ((satisfiesTypeProperties(type, instance)
			&& satisfiesTypeRepresentations(type, instance)
			&& satisfiesTypeDesignRules(type, instance))
			== false) {
			typechecks = false;
		}

		// check that the instance contains all of the roles and role
		// substructure required by the type.
		AcmeEnumeration e = type.getRoles();
		while (e.moreItems()) {
			AcmeRole type_role = (AcmeRole) e.getNextItem();
			if (type_role == null)
				continue;
			AcmeRole instance_role = instance.getRole(type_role.getName());
			if (instance_role == null) {
				typechecks = false;
				continue;
			}
			// make sure that the instance role satisifies all of the
			// type_role's types
			AcmeEnumeration typeEnum = type_role.getDeclaredTypes();
			while (typeEnum.moreItems()) {
				AcmeElementType super_type = (AcmeElementType) typeEnum.getNextItem();
				if (super_type instanceof AcmeElementTypeRef) {
					super_type = ((AcmeElementTypeRef) super_type).getReferentType();
				}
				if (super_type instanceof AcmeRoleType) {
					if (satisfiesRoleType((AcmeRoleType) super_type, instance_role) == false) {
						typechecks = false;
					}
				}
				else if (super_type instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance_role) == false) {
						typechecks = false;
					}
				}
			}

			// make sure that the properties, reps and design rules of the
			// type's role are included in the instance's role.
			if (satisfiesTypeProperties(type_role, instance_role) == false)
				typechecks = false;
			if (satisfiesTypeRepresentations(type_role, instance_role) == false)
				typechecks = false;
			if (satisfiesTypeDesignRules(type_role, instance_role) == false)
				typechecks = false;
		}

		return typechecks;
	}

	/**
	 * satisfiesPortType(type,instance) determines whether instance
	 * satifies all of the requirements of type.  It only checks that the
	 * requirements of type are satisfied by instance.  It does not check
	 * instance for any kind of internal type-consistency. Returns true
	 * if instance satisfies type, else false.
	 * @param type The AcmePortType
	 * @param instance The AcmePort
	 * @return true if instance satisfies the structure specified in type
	 */
	public static boolean satisfiesPortType(AcmePortType type, AcmePort instance) {
		boolean typechecks = true;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (type == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		// recursively check that instance satisfies all of type's supertypes.
		AcmeEnumeration super_types_enum = type.getSuperTypes().enumerate();
		while (super_types_enum.moreItems()) {
			AcmeElementType super_type = (AcmeElementType) super_types_enum.getNextItem();
			if (super_type instanceof AcmeElementTypeRef) {
				super_type = ((AcmeElementTypeRef) super_type).getReferentType();
			}
			if (super_type instanceof AcmePortType) {
				if (satisfiesPortType((AcmePortType) super_type, instance) == false) {
					typechecks = false;
				}
			}
			else if (super_type instanceof AcmeElementProtoType) {
				if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance) == false) {
					typechecks = false;
				}
			}
		}

		// then check that this instance satisfies all of the properties,
		// representations, and design rules required by type.
		if ((satisfiesTypeProperties(type, instance)
			&& satisfiesTypeRepresentations(type, instance)
			&& satisfiesTypeDesignRules(type, instance))
			== false) {
			typechecks = false;
		}
		return typechecks;
	}

	/**
	 * satisfiesRoleType(type,instance) determines whether instance
	 * satifies all of the requirements of type.  It only checks that the
	 * requirements of type are satisfied by instance.  It does not check
	 * instance for any kind of internal type-consistency. Returns true
	 * if instance satisfies type, else false.
	 * @param type The AcmeRoleType
	 * @param instance The AcmeRole
	 * @return true if instance satisfies the structure specified in type
	 */
	public static boolean satisfiesRoleType(AcmeRoleType type, AcmeRole instance) {
		boolean typechecks = true;

		// a null type can't be type incorrect and a null instance can't be
		// type correct.
		if (type == null) {
			return true;
		}
		else if (instance == null) {
			return false;
		}

		// recursively check that instance satisfies all of type's supertypes.
		AcmeEnumeration super_types_enum = type.getSuperTypes().enumerate();
		while (super_types_enum.moreItems()) {
			AcmeElementType super_type = (AcmeElementType) super_types_enum.getNextItem();
			if (super_type instanceof AcmeElementTypeRef) {
				super_type = ((AcmeElementTypeRef) super_type).getReferentType();
			}
			if (super_type instanceof AcmeRoleType) {
				if (satisfiesRoleType((AcmeRoleType) super_type, instance) == false) {
					typechecks = false;
				}
			}
			else if (super_type instanceof AcmeElementProtoType) {
				if (satisfiesEltProtoType((AcmeElementProtoType) super_type, instance) == false) {
					typechecks = false;
				}
			}
		}

		// then check that this instance satisfies all of the properties,
		// representations, and design rules required by type.
		if ((satisfiesTypeProperties(type, instance)
			&& satisfiesTypeRepresentations(type, instance)
			&& satisfiesTypeDesignRules(type, instance))
			== false) {
			typechecks = false;
		}

		return typechecks;
	}

	/**
	 * the typecheck*(...) instance typechecking methods take a single
	 * element instance as a parameter and confirm that the instance
	 * satisfies all of the requirements of all of its types.  They also
	 * confirm that all of the properties, and representations of the instance
	 * are type correct, and that all of the instance's invariants and
	 * heuristics are satisfied.
	 *
	 *
	 * helper method to perform typechecking of generic element properties,
	 * representations, and design rules.  Only checks instances of these
	 * items for self-type-consistency.  Doesn't care about declared types.
	 * @param elt The AcmeElement
	 * @return true if type check is ok
	 */
	public static boolean typecheckCommonElementItems(AcmeElement elt) {
		// assume this element typechecks unless we can show that it doesn't
		// it's important that this variable can only be set to _false_ and
		// not to _true_, or else a flagged type-incorrectness could be
		// subsequently overriden.
		boolean typechecks = true;

		// Note that because this is a helper method it only adds errors to
		// the type stack as they are found, rather than adding an initial
		// "top-level" error that can be later removed, as the public instance
		// typechecking methods do.

		// check the maps of the instance
		AcmeEnumeration eMaps = elt.getMaps();
		while (eMaps.moreItems()) {
			AcmeMapInstance map = (AcmeMapInstance) eMaps.getNextItem();
			if (!map.typecheck()) {
				typechecks = false;
				AcmeError error = new AcmeError(map, "Map " + map.getName() + "doesn't satisfy its declared type");
				typecheckingErrorStack.pushCausalError(error);
			}
		}

		// check the properties of the instance to make sure that they
		// typecheck with their own property type specifications.
		AcmeEnumeration e = elt.getProperties();
		while (e.moreItems()) {
			AcmeProperty instance_cur_property = (AcmeProperty) e.getNextItem();
			AcmePropertyType property_type = instance_cur_property.getType();
			if (property_type != null) {
				if (typecheckProperty(property_type, instance_cur_property.getValue()) == false) {
					typechecks = false;
					PropertyTypecheckError propTypeError =
						new PropertyTypecheckError(
							elt,
							"Property "
								+ instance_cur_property.getName()
								+ " does not satisfy its declared type: "
								+ property_type.getTypeName(),
							instance_cur_property);
					typecheckingErrorStack.pushCausalError(propTypeError);
					AcmeErrorManager.errorDetected(propTypeError);
				}
			}
		}

		// Check that the instance-specific representations of this element
		// are internally consistent.
		AcmeEnumeration elt_reps = elt.getRepresentations();
		while (elt_reps.moreItems()) {
			AcmeRepresentation elt_rep = (AcmeRepresentation) elt_reps.getNextItem();
			if (elt_rep == null)
				continue;
			if (elt_rep.getSystem() == null)
				continue;
			else {
				boolean repTypechecks = true;
				AcmeSystem elt_sys = elt_rep.getSystem();

				// assume a type error in typechecking the rep, cancel it
				// if the rep does indeed typecheck.  This will leave an
				// appropriate causal error trail if there is indeed a type error
				// in the rep.
				RepTypecheckError repError = new RepTypecheckError();
				typecheckingErrorStack.pushCausalError(repError);

				repTypechecks = typecheckSystem(elt_sys);

				// typecheck the properties on all of the bindings
				if (elt_rep.getAbstractionMap() instanceof AcmeBindingsMap) {
					AcmeEnumeration bindings = ((AcmeBindingsMap) elt_rep.getAbstractionMap()).getBindings();
					while (bindings.moreItems()) {
						AcmeBinding curBinding = (AcmeBinding) bindings.getNextItem();
						AcmeEnumeration props = curBinding.getProperties();
						while (props.moreItems()) {
							if (!typecheckProperty((AcmeProperty) props.getNextItem())) {
								repTypechecks = false;
							}
						}
					}
				}

				if (repTypechecks) {
					typecheckingErrorStack.cancelError(repError);
					AcmeErrorManager.errorRepaired(elt_rep, repError);
				}
				else {
					typechecks = false;
					repError.setSource(elt_rep);
					repError.setErrorRep(elt_rep);
					repError.setErrorMessage("Representation does not typecheck");
					typecheckingErrorStack.packageErrorCauses(repError);
					AcmeErrorManager.errorDetected(repError);
				}
			}
		}

		// Check that the instance-specific invariants and heuristics of
		// this element are type-correct and evaluate to true.

		// first check the invariants.
		e = elt.getInvariants(false);
		while (e.moreItems()) {
			DesignRule dr = (DesignRule) e.getNextItem();
			if (dr.typecheck() == false) {
				typechecks = false;
				ExpressionTypecheckError exprTypecheckError =
					new ExpressionTypecheckError(
						elt,
						"Element "
							+ elt.getName()
							+ "'s invariant expression \""
							+ dr.toString()
							+ "\" does not typecheck. ",
						dr.getDesignRuleExpression());
				typecheckingErrorStack.packageErrorCauses(exprTypecheckError);
				AcmeErrorManager.errorDetected(exprTypecheckError);
			}
			else {
				try {
					if (dr.evaluate() == false) {
						typechecks = false;
						InvariantViolationError invariantEvalError =
							new InvariantViolationError(
								elt,
								"Element " + elt.getName() + "'s invariant: " + dr.toString() + " does not hold.",
								dr,
								elt);
						typecheckingErrorStack.pushCausalError(invariantEvalError);
						AcmeErrorManager.errorDetected(invariantEvalError);
					}
				}
				catch (DesignRuleEvalException dre) {
					typechecks = false;
					ExpressionTypecheckError exceptionError =
						new ExpressionTypecheckError(
							elt,
							"Element "
								+ elt.getName()
								+ "'s invariant expression \""
								+ dr.toString()
								+ "\" caused an evaluation exception: "
								+ dre.getMessage(),
							dr.getDesignRuleExpression());
					typecheckingErrorStack.pushCausalError(exceptionError);
					AcmeErrorManager.errorDetected(exceptionError);
				}
			}
		}

		// then check the heuristics.
		e = elt.getHeuristics(false);
		while (e.moreItems()) {
			DesignRule dr = (DesignRule) e.getNextItem();
			if (dr.typecheck() == false) {
				typechecks = false;
				ExpressionTypecheckError exprTypecheckError =
					new ExpressionTypecheckError(
						elt,
						elt.getName() + "'s Heuristic expression \"" + dr.toString() + "\" does not typecheck. ",
						dr.getDesignRuleExpression());
				typecheckingErrorStack.packageErrorCauses(exprTypecheckError);
				AcmeErrorManager.errorDetected(exprTypecheckError);
			}
			else {
				try {
					if (dr.evaluate() == false) {
						// An unsatisfied heuristic doesn't make the method
						// type-incorrect.  Announce that the heuristic
						// is not satisfied directly, rather than as part
						// of the type-checking process.
						HeuristicViolationError heuristicEvalError =
							new HeuristicViolationError(
								elt,
								"Element "
									+ elt.getName()
									+ "'s heuristic: "
									+ dr.getDesignRuleExpression().toString()
									+ " evaluates to false.",
								dr,
								elt);
						AcmeErrorManager.errorDetected(heuristicEvalError);
					}

				}
				catch (DesignRuleEvalException dre) {
					typechecks = false;
					ExpressionTypecheckError exceptionError =
						new ExpressionTypecheckError(
							elt,
							elt.getName()
								+ "'s Heuristic expression \""
								+ dr.toString()
								+ "\" caused an evaluation exception: "
								+ dre.getMessage(),
							dr.getDesignRuleExpression());
					typecheckingErrorStack.pushCausalError(exceptionError);
					AcmeErrorManager.errorDetected(exceptionError);
				}
			}
		}

		// if we get to here then we weren't able to show that the instance's
		// representations or properties are type-incorrect, so assume they
		// are type correct and return true.
		return typechecks;
	}

	/**
	 * confirm that the instance "sys" satisfies all of its declared families
	 * and element types.
	 * @param sys The AcmeSystem tested
	 * @return true if sys satisfies all declared families
	 */
	public static boolean typecheckSystem(AcmeSystem sys) {
		boolean typechecks = true;
		ElementTypecheckError eltError = new ElementTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (sys == null) {
			typechecks = false;
			eltError.setErrorMessage("Component to typecheck is a null reference");
		}
		else {
			// confirm that sys satisfies all of its declared families and element types.
			AcmeEnumeration e = sys.getDeclaredTypes();
			while (e.moreItems()) {
				AcmeElementType super_type = (AcmeElementType) e.getNextItem();
				if (super_type instanceof AcmeFamily) {
					if (satisfiesFamily((AcmeFamily) super_type, sys) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								sys,
								"System "
									+ sys.getName()
									+ " does not satisfy "
									+ "its declared type (family): "
									+ super_type.getTypeName()
									+ ". ",
								sys,
								super_type);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}
					else {
						AcmeErrorManager.errorRepaired(sys, null);
					}
				}
				else if (super_type instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) super_type, sys) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								sys,
								"System "
									+ sys.getName()
									+ " does not satisfy "
									+ "its declared type: "
									+ super_type.getTypeName()
									+ ". ",
								sys,
								super_type);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}
					else {
						AcmeErrorManager.errorRepaired(sys, null);
					}
				}
			}

			// confirm that all of sys's components typecheck.  Any errors
			// discovered while typechecking the components will be put
			// properly at the top of the typecheckingErrorStack in the
			// typecheckComponent method.
			e = sys.getComponents();
			while (e.moreItems()) {
				AcmeComponent cur_comp = (AcmeComponent) e.getNextItem();
				if (typecheckComponent(cur_comp) == false)
					typechecks = false;
			}

			// confirm that all of sys's connectors typecheck.  Any errors
			// discovered while typechecking the connectors will be put
			// properly at the top of the typecheckingErrorStack in the
			// typecheckConnector method.
			e = sys.getConnectors();
			while (e.moreItems()) {
				AcmeConnector cur_conn = (AcmeConnector) e.getNextItem();
				if (typecheckConnector(cur_conn) == false)
					typechecks = false;
			}

			// if we're dealing with PURE_ACME then a system can have ports
			// and roles and the attachments can have properties that need
			// to be typechecked, so typecheck those also.
			if (acmeLib.core.globals.LanguageConfig.IS_PURE_ACME == true) {
				e = sys.getPorts();
				while (e.moreItems()) {
					AcmePort cur_port = (AcmePort) e.getNextItem();
					if (typecheckPort(cur_port) == false)
						typechecks = false;
				}

				e = sys.getRoles();
				while (e.moreItems()) {
					AcmeRole cur_role = (AcmeRole) e.getNextItem();
					if (typecheckRole(cur_role) == false)
						typechecks = false;
				}

				e = sys.getAttachments();
				while (e.moreItems()) {
					AcmeAttachment cur_attach = (AcmeAttachment) e.getNextItem();
					AcmeEnumeration props = cur_attach.getProperties();
					while (props.moreItems()) {
						AcmeProperty prop = (AcmeProperty) props.getNextItem();
						if (typecheckProperty(prop) == false) {
							typechecks = false;
							PropertyTypecheckError newPTError =
								new PropertyTypecheckError(
									cur_attach,
									"Property \""
										+ prop.getName()
										+ "\" of attachment \""
										+ cur_attach.getPortNameAsString()
										+ " to "
										+ cur_attach.getRoleNameAsString()
										+ "\" does not typecheck.",
									prop);
							typecheckingErrorStack.pushCausalError(newPTError);
							AcmeErrorManager.errorDetected(newPTError);
						}
						else {
							AcmeErrorManager.errorRepaired(cur_attach, null);
						}
					}
				}
			}

			// then confirm that the system's properties, representations, and
			// design rules are internally consistent.  Any generated errors
			// will be added in the typecheckCommonElementItems method.
			if (typecheckCommonElementItems(sys) == false)
				typechecks = false;

			// confirm that all of the design analyses in this system typecheck.
			e = sys.getDesignAnalyses();
			while (e.moreItems()) {
				DesignAnalysis da = (DesignAnalysis) e.getNextItem();
				if (da.typecheck() == false) {
					typechecks = false;
					ExpressionTypecheckError exprTypecheckError =
						new ExpressionTypecheckError(
							da,
							sys.getName()
								+ "'s design analysis declaration for \""
								+ da.getName()
								+ "\" does not typecheck. ",
							da.getDesignRuleExpression());
					typecheckingErrorStack.pushCausalError(exprTypecheckError);
					AcmeErrorManager.errorDetected(exprTypecheckError);

				}
				else {
					AcmeErrorManager.errorRepaired(da, null);
				}
			}

		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(sys, eltError);
		}
		else {
			eltError.setElement(sys);
			eltError.setSource(sys);
			if (sys != null)
				eltError.setErrorMessage("System " + sys.getName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
			AcmeErrorManager.errorDetected(eltError);
		}

		return typechecks;
	}

	public static boolean typecheckMapType(AcmeMapType maptype) {
		//        System.out.println("type checking map type not implemented");
		return true;
	}

	public static boolean typecheckMaplet(AcmeMaplet maplet) {
		boolean typechecks = true;
		ElementTypecheckError eltError = new ElementTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (maplet == null) {
			typechecks = false;
			eltError.setErrorMessage("Maplet to typecheck is a null reference");
			return typechecks;
		}
		AcmeMapInstance parentMap = (AcmeMapInstance) maplet.getParent();
		AcmeMapType parentType = (AcmeMapType) parentMap.getDeclaredTypes().getNextItem();
		AcmeElementType sourceType = ConstantTypes.ELEMENT_BASE_TYPE;
		AcmeElementType targetType = ConstantTypes.ELEMENT_BASE_TYPE;
		if (parentType != null) {
			sourceType = parentType.getSignature().getSource();
			targetType = parentType.getSignature().getDomain();
		}
		if (!satisfiesType(sourceType, maplet.getSource())) {
			System.out.println(
				"source type doesn't match " + sourceType.getTypeName() + " " + maplet.getSource().getQualifiedName());
			typechecks = false;
		}
		if (!satisfiesType(targetType, maplet.getTarget())) {
			System.out.println(
				"domain type doesn't match " + targetType.getTypeName() + " " + maplet.getTarget().getQualifiedName());
			typechecks = false;
		}

		// then confirm that the map instance's properties, representations, and
		// design rules are internally consistent.  Any generated errors
		// will be added in the typecheckCommonElementItems method.
		if (typecheckCommonElementItems(maplet) == false) {
			typechecks = false;
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
		}
		else {
			eltError.setElement(maplet);
			eltError.setSource(maplet);
			if (maplet != null)
				eltError.setErrorMessage("Maplet " + maplet.getName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the instance "map" satisfies all of its declared types.
	 * and that all of its substructure is internally type-correct.
	 * @param map The AcmeMapInstance tested
	 * @return true if map satisfies all of its declared types
	 */
	public static boolean typecheckMap(AcmeMapInstance map) {
		boolean typechecks = true;
		ElementTypecheckError eltError = new ElementTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (map == null) {
			typechecks = false;
			eltError.setErrorMessage("Map instance to typecheck is a null reference");
		}
		else {
			// confirm that all maplets type check.
			AcmeEnumeration e = map.getMaplets();
			while (e.moreItems()) {
				// confirm that all of map's maplets typecheck.  Any errors discovered
				// while typechecking the maplets will be put properly at the top
				// of the typecheckingErrorStack in the typecheckPort method.
				AcmeMaplet cur_maplet = (AcmeMaplet) e.getNextItem();
				if (typecheckMaplet(cur_maplet) == false) {
					typechecks = false;
				}
			}
		}

		// then confirm that the map instance's properties, representations, and
		// design rules are internally consistent.  Any generated errors
		// will be added in the typecheckCommonElementItems method.
		if (typecheckCommonElementItems(map) == false) {
			typechecks = false;
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
		}
		else {
			eltError.setElement(map);
			eltError.setSource(map);
			if (map != null)
				eltError.setErrorMessage("Map instance " + map.getName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the instance "comp" satisfies all of its declared types.
	 * and that all of its substructure is internally type-correct.
	 * @param comp The Acme Component tested
	 * @return true if comp satisfies all of its declared types
	 */
	public static boolean typecheckComponent(AcmeComponent comp) {
		boolean typechecks = true;
		ElementTypecheckError eltError = new ElementTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		String elementStr = "Component ";
		if (comp == null) {
			typechecks = false;
			eltError.setErrorMessage("Component to typecheck is a null reference");
		}
		else {
			// confirm that comp satisfies all of its declared types.
			AcmeEnumeration e = comp.getDeclaredTypes();
			while (e.moreItems()) {
				AcmeType nextType = (AcmeType) e.getNextItem();
				if (nextType instanceof AcmeElementTypeRef) {
					nextType = ((AcmeElementTypeRef) nextType).getReferentType();
				}
				if (nextType instanceof AcmeComponentType) {
					AcmeComponentType cur_type = (AcmeComponentType) nextType;
					if (satisfiesComponentType(cur_type, comp) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								comp,
								elementStr
									+ comp.getName()
									+ " does not satisfy "
									+ " its declared type: "
									+ cur_type.getTypeName()
									+ ". ",
								comp,
								cur_type);
//						typecheckingErrorStack.packageErrorCauses(newTypeError);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}

				}
				else if (nextType instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) nextType, comp) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								comp,
								elementStr
									+ comp.getName()
									+ " does not satisfy "
									+ " its declared type: "
									+ nextType.getTypeName()
									+ ". ",
								comp,
								(AcmeElementProtoType) nextType);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}

				}
				else if (nextType instanceof AcmeUnresolvedElementType) {
					// type is not yet resolved, announce an error.
					typechecks = false;
					ElementTypeSatisfactionError newTypeError =
						new ElementTypeSatisfactionError(
							comp,
							elementStr
								+ comp.getName()
								+ "'s declared type "
								+ ((AcmeElementType) nextType).getTypeName()
								+ " is unresolved.",
							comp,
							(AcmeUnresolvedElementType) nextType);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
				else if (nextType instanceof AcmeType) {
					// else this is an unknown, but illegal type.  announce error
					// type is not yet resolved, announce an error.
					typechecks = false;
					ElementTypeSatisfactionError newTypeError =
						new ElementTypeSatisfactionError(
							comp,
							elementStr
								+ comp.getName()
								+ "'s declared type "
								+ ((AcmeType) nextType).getTypeName()
								+ " has not been successfully resolved.");
					newTypeError.setElement(comp);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
				else {
					// else this is an unknown non-type entity. announce error
					typechecks = false;
					ElementTypecheckError newTypeError =
						new ElementTypecheckError(
							comp,
							elementStr
								+ comp.getName()
								+ " declares a completely unrecognizable type."
								+ " The typechecker is just plain baffled.",
							comp);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);
				}
			}

			// confirm that all instantiating types are actually component types
			// or element types
			e = comp.getInstantiatedTypes();
			while (e.moreItems()) {
				AcmeType nextType = (AcmeType) e.getNextItem();
				if (nextType instanceof AcmeElementTypeRef) {
					nextType = ((AcmeElementTypeRef) nextType).getReferentType();
				}
				if (!(nextType instanceof AcmeComponentType) && !(nextType instanceof AcmeElementProtoType)) {
					typechecks = false;
					TypeInstantiationError newTypeError =
						new TypeInstantiationError(
							comp,
							"Instantiation Error.  Type: "
								+ nextType.getTypeName()
								+ " is neither a component nor an element type.  Therefore it "
								+ "can not be used to instantiate component: ",
							nextType);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);
				}
			}

			// confirm that all of comp's ports typecheck.  Any errors discovered
			// while typechecking the ports will be put properly at the top
			// of the typecheckingErrorStack in the typecheckPort method.
			e = comp.getPorts();
			while (e.moreItems()) {
				AcmePort cur_port = (AcmePort) e.getNextItem();
				if (typecheckPort(cur_port) == false)
					typechecks = false;
			}

			// then confirm that the component's properties, representations, and
			// design rules are internally consistent.  Any generated errors
			// will be added in the typecheckCommonElementItems method.
			if (typecheckCommonElementItems(comp) == false)
				typechecks = false;
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(comp, eltError);
		}
		else {
			eltError.setElement(comp);
			eltError.setSource(comp);
			if (comp != null)
				eltError.setErrorMessage(elementStr + comp.getName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the instance "conn" satisfies all of its declared types.
	 * and that all of its substructure is internally type-correct.
	 * @param conn The AcmeConnector tested
	 * @return true if conn satisfies all of it declared types
	 */
	public static boolean typecheckConnector(AcmeConnector conn) {
		boolean typechecks = true;
		ElementTypecheckError eltError = new ElementTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (conn == null) {
			typechecks = false;
			eltError.setErrorMessage("Connector to typecheck is a null reference");
		}
		else {
			// confirm that conn satisfies all of its declared types.
			AcmeEnumeration e = conn.getDeclaredTypes();
			while (e.moreItems()) {
				AcmeType nextType = (AcmeType) e.getNextItem();
				if (nextType instanceof AcmeElementTypeRef) {
					nextType = ((AcmeElementTypeRef) nextType).getReferentType();
				}
				if (nextType instanceof AcmeConnectorType) {
					AcmeConnectorType cur_type = (AcmeConnectorType) nextType;
					if (satisfiesConnectorType(cur_type, conn) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								conn,
								"Connector "
									+ conn.getName()
									+ " does not satisfy "
									+ " its declared type: "
									+ cur_type.getTypeName()
									+ ". ",
								conn,
								cur_type);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}

				}
				else if (nextType instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) nextType, conn) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								conn,
								"Connector "
									+ conn.getName()
									+ " does not satisfy "
									+ " its declared type: "
									+ nextType.getTypeName()
									+ ". ",
								conn,
								(AcmeElementProtoType) nextType);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}

				}
				else if (nextType instanceof AcmeUnresolvedElementType) {
					// type is not yet resolved, announce an error.
					typechecks = false;
					ElementTypeSatisfactionError newTypeError =
						new ElementTypeSatisfactionError(
							conn,
							"Connector "
								+ conn.getName()
								+ "'s declared type "
								+ ((AcmeElementType) nextType).getTypeName()
								+ " is unresolved.",
							conn,
							(AcmeUnresolvedElementType) nextType);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
				else if (nextType instanceof AcmeType) {
					// else this is an unknown, but illegal type.  announce error
					// type is not yet resolved, announce an error.
					typechecks = false;
					ElementTypeSatisfactionError newTypeError =
						new ElementTypeSatisfactionError(
							conn,
							"Connector "
								+ conn.getName()
								+ "'s declared type "
								+ ((AcmeType) nextType).getTypeName()
								+ " has not been successfully resolved.");
					newTypeError.setElement(conn);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
				else {
					// else this is an unknown non-type entity. announce error
					typechecks = false;
					ElementTypecheckError newTypeError =
						new ElementTypecheckError(
							conn,
							"Connector "
								+ conn.getName()
								+ " declares a completely unrecognizable type."
								+ " The typechecker is just plain baffled.",
							conn);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
			}
			// confirm that all instantiating types are actually connector types
			// or element types
			e = conn.getInstantiatedTypes();
			while (e.moreItems()) {
				AcmeType nextType = (AcmeType) e.getNextItem();
				if (nextType instanceof AcmeElementTypeRef) {
					nextType = ((AcmeElementTypeRef) nextType).getReferentType();
				}
				if (!(nextType instanceof AcmeConnectorType) && !(nextType instanceof AcmeElementProtoType)) {
					typechecks = false;
					TypeInstantiationError newTypeError =
						new TypeInstantiationError(
							conn,
							"Instantiation Error.  Type: "
								+ nextType.getTypeName()
								+ " is neither a connector nor an element type.  Therefore it "
								+ "can not be used to instantiate component: ",
							nextType);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);
				}
			}
			// confirm that all of conn's roles typecheck.  Any errors discovered
			// while typechecking the ports will be put properly at the top
			// of the typecheckingErrorStack in the typecheckRole method.
			e = conn.getRoles();
			while (e.moreItems()) {
				AcmeRole cur_role = (AcmeRole) e.getNextItem();
				if (typecheckRole(cur_role) == false)
					typechecks = false;
			}

			// then confirm that the connector's properties, representations, and
			// design rules are internally consistent.  Any generated errors
			// will be added in the typecheckCommonElementItems method.
			if (typecheckCommonElementItems(conn) == false)
				typechecks = false;
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(conn, eltError);
		}
		else {
			eltError.setElement(conn);
			eltError.setSource(conn);
			if (conn != null)
				eltError.setErrorMessage("Connector " + conn.getName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the instance "port" satisfies all of its declared types.
	 * and that all of its substructure is internally type-correct.
	 * @param port The AcmePort tested
	 * @return true if port satisfies all of its declared types
	 */
	public static boolean typecheckPort(AcmePort port) {
		boolean typechecks = true;
		ElementTypecheckError eltError = new ElementTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (port == null) {
			typechecks = false;
			eltError.setErrorMessage("Port to typecheck is null reference");
		}
		else {
			// confirm that port satisfies all of its declared types.
			AcmeEnumeration e = port.getDeclaredTypes();
			while (e.moreItems()) {
				AcmeType nextType = (AcmeType) e.getNextItem();
				if (nextType instanceof AcmeElementTypeRef) {
					nextType = ((AcmeElementTypeRef) nextType).getReferentType();
				}
				if (nextType instanceof AcmePortType) {
					AcmePortType cur_type = (AcmePortType) nextType;
					if (satisfiesPortType(cur_type, port) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								port,
								"Port "
									+ port.getName()
									+ " does not satisfy "
									+ " its declared type: "
									+ cur_type.getTypeName()
									+ ". ",
								port,
								cur_type);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}

				}
				else if (nextType instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) nextType, port) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								port,
								"Port "
									+ port.getName()
									+ " does not satisfy "
									+ " its declared type: "
									+ nextType.getTypeName()
									+ ". ",
								port,
								(AcmeElementProtoType) nextType);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}

				}
				else if (nextType instanceof AcmeUnresolvedElementType) {
					// type is not yet resolved, announce an error.
					typechecks = false;
					ElementTypeSatisfactionError newTypeError =
						new ElementTypeSatisfactionError(
							port,
							"Port "
								+ port.getName()
								+ "'s declared type "
								+ ((AcmeElementType) nextType).getTypeName()
								+ " is unresolved.",
							port,
							(AcmeUnresolvedElementType) nextType);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
				else if (nextType instanceof AcmeType) {
					// else this is an unknown, but illegal type.  announce error
					// type is not yet resolved, announce an error.
					typechecks = false;
					ElementTypeSatisfactionError newTypeError =
						new ElementTypeSatisfactionError(
							port,
							"Port "
								+ port.getName()
								+ "'s declared type "
								+ ((AcmeType) nextType).getTypeName()
								+ " has not been successfully resolved.");
					newTypeError.setElement(port);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
				else {
					// else this is an unknown non-type entity. announce error
					typechecks = false;
					ElementTypecheckError newTypeError =
						new ElementTypecheckError(
							port,
							"Port "
								+ port.getName()
								+ " declares a completely unrecognizable type."
								+ " The typechecker is just plain baffled.",
							port);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
			}
			// confirm that all instantiating types are actually component types
			// or element types
			e = port.getInstantiatedTypes();
			while (e.moreItems()) {
				AcmeType nextType = (AcmeType) e.getNextItem();
				if (nextType instanceof AcmeElementTypeRef) {
					nextType = ((AcmeElementTypeRef) nextType).getReferentType();
				}
				if (!(nextType instanceof AcmePortType) && !(nextType instanceof AcmeElementProtoType)) {
					typechecks = false;
					TypeInstantiationError newTypeError =
						new TypeInstantiationError(
							port,
							"Instantiation Error.  Type: "
								+ nextType.getTypeName()
								+ " is neither a port nor an element type.  Therefore it "
								+ "can not be used to instantiate component: ",
							nextType);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);
				}
			}
			// then confirm that the port's properties, representations, and
			// design rules are internally consistent.  Any generated errors
			// will be added in the typecheckCommonElementItems method.
			if (typecheckCommonElementItems(port) == false)
				typechecks = false;
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(port, eltError);
		}
		else {
			eltError.setElement(port);
			eltError.setSource(port);
			if (port != null)
				eltError.setErrorMessage("Port " + port.getName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the instance "role" satisfies all of its declared types.
	 * and that all of its substructure is internally type-correct.
	 * @param role The AcmeRole tested
	 * @return true if role satisfies all of its declared types
	 */
	public static boolean typecheckRole(AcmeRole role) {
		boolean typechecks = true;
		ElementTypecheckError eltError = new ElementTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (role == null) {
			typechecks = false;
			eltError.setErrorMessage("Role to typecheck is a null reference");
		}
		else {
			// confirm that role satisfies all of its declared types.
			AcmeEnumeration e = role.getDeclaredTypes();
			while (e.moreItems()) {
				AcmeType nextType = (AcmeType) e.getNextItem();
				if (nextType instanceof AcmeElementTypeRef) {
					nextType = ((AcmeElementTypeRef) nextType).getReferentType();
				}
				if (nextType instanceof AcmeRoleType) {
					AcmeRoleType cur_type = (AcmeRoleType) nextType;
					if (satisfiesRoleType(cur_type, role) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								role,
								"Role "
									+ role.getName()
									+ " does not satisfy "
									+ " its declared type: "
									+ cur_type.getTypeName()
									+ ". ",
								role,
								cur_type);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}

				}
				else if (nextType instanceof AcmeElementProtoType) {
					if (satisfiesEltProtoType((AcmeElementProtoType) nextType, role) == false) {
						typechecks = false;
						ElementTypeSatisfactionError newTypeError =
							new ElementTypeSatisfactionError(
								role,
								"Role "
									+ role.getName()
									+ " does not satisfy "
									+ " its declared type: "
									+ nextType.getTypeName()
									+ ". ",
								role,
								(AcmeElementProtoType) nextType);
						typecheckingErrorStack.pushCausalError(newTypeError);
						AcmeErrorManager.errorDetected(newTypeError);
					}

				}
				else if (nextType instanceof AcmeUnresolvedElementType) {
					// type is not yet resolved, announce an error.
					typechecks = false;
					ElementTypeSatisfactionError newTypeError =
						new ElementTypeSatisfactionError(
							role,
							"Role "
								+ role.getName()
								+ "'s declared type "
								+ ((AcmeElementType) nextType).getTypeName()
								+ " is unresolved.",
							role,
							(AcmeUnresolvedElementType) nextType);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
				else if (nextType instanceof AcmeType) {
					// else this is an unknown, but illegal type.  announce error
					// type is not yet resolved, announce an error.
					typechecks = false;
					ElementTypeSatisfactionError newTypeError =
						new ElementTypeSatisfactionError(
							role,
							"Role "
								+ role.getName()
								+ "'s declared type "
								+ ((AcmeType) nextType).getTypeName()
								+ " has not been successfully resolved.");
					newTypeError.setElement(role);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
				else {
					// else this is an unknown non-type entity. announce error
					typechecks = false;
					ElementTypecheckError newTypeError =
						new ElementTypecheckError(
							role,
							"Role "
								+ role.getName()
								+ " declares a completely unrecognizable type."
								+ " The typechecker is just plain baffled.",
							role);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);

				}
			}
			//confirm that all instantiating types are actually role types
			// or element types
			e = role.getInstantiatedTypes();
			while (e.moreItems()) {
				AcmeType nextType = (AcmeType) e.getNextItem();
				if (nextType instanceof AcmeElementTypeRef) {
					nextType = ((AcmeElementTypeRef) nextType).getReferentType();
				}
				if (!(nextType instanceof AcmeRoleType) && !(nextType instanceof AcmeElementProtoType)) {
					typechecks = false;
					TypeInstantiationError newTypeError =
						new TypeInstantiationError(
							role,
							"Instantiation Error.  Type: "
								+ nextType.getTypeName()
								+ " is neither a role nor an element type.  Therefore it "
								+ "can not be used to instantiate component: ",
							nextType);
					typecheckingErrorStack.pushCausalError(newTypeError);
					AcmeErrorManager.errorDetected(newTypeError);
				}
			}
			// then confirm that the role's properties, representations, and
			// design rules are internally consistent.  Any generated errors
			// will be added in the typecheckCommonElementItems method.
			if (typecheckCommonElementItems(role) == false)
				typechecks = false;
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(role, eltError);
		}
		else {
			eltError.setElement(role);
			eltError.setSource(role);
			if (role != null)
				eltError.setErrorMessage("Role " + role.getName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * the typecheck*Type(...) element type typechecking methods take a
	 * single element type as a parameter and confirm that the properties
	 * and substructure of the instance are type consistent, and that all
	 * of the instance's invariants and heuristics expressions typecheck.
	 * These methods also attempt to create an instance of the type in
	 * question using the unify* methods (this should create the same
	 * minimal instance that would be created with a "new <typename>"
	 * request.  If it is unable to unify itself with its super types
	 * then the type itself is not type-correct.
	 *
	 *
	 *
	 * helper function that confirms all design rules typecheck and all
	 * properties that require values are internally type-consistent.
	 * As a helper method, this only pushes any specific errors that it
	 * finds on an existing error stack.  It leaves the packaging and
	 * aggregating of errors up to the public methods that call it.
	 * @param type The AcmeElementType tested
	 * @return true if all design rules typecheck and properties are consistent
	 */
	protected static boolean typecheckElementType(AcmeElementType type) {
		boolean typechecks = true;

		typechecks = checkForCyclesInSupertypes (type);
		if (!typechecks) {
			String errMsg = "The type \"" + type.getTypeName() + "\" contains cycles in the type hierarchy.";
			ElementTypecheckError error = new ElementTypecheckError (type, errMsg);
			typecheckingErrorStack.pushCausalError(error);
			AcmeErrorManager.errorDetected(error);
		}
		
		// Don't check constraints on types
		// check invariants
		AcmeEnumeration e = ((AcmeElement) type).getInvariants(false);
//		while (e.moreItems()) {
//			DesignRule invariant = (DesignRule) e.getNextItem();
//			if (invariant.getIsInvariant() == false) {
//				typechecks = false;
//				AcmeError newInternalError =
//					new AcmeError(
//						type,
//						"Heuristic stored in element type "
//							+ type.getTypeName()
//							+ " invariants list (internal error): "
//							+ invariant.toString());
//				typecheckingErrorStack.pushCausalError(newInternalError);
//				AcmeErrorManager.errorDetected(newInternalError);
//			}
//
//			if (invariant.typecheck() == false) {
//				typechecks = false;
//				ExpressionTypecheckError newError =
//					new ExpressionTypecheckError(
//						type,
//						"invariant does not typecheck: " + invariant.toString(),
//						invariant.getDesignRuleExpression());
//				typecheckingErrorStack.pushCausalError(newError);
//				// invariant should add its own error detected
//			}
//		}
//
//		// check heuristics
//		e = ((AcmeElement) type).getHeuristics(false);
//		while (e.moreItems()) {
//			DesignRule heuristic = (DesignRule) e.getNextItem();
//			if (heuristic.getIsInvariant() == true) {
//				typechecks = false;
//				AcmeError newInternalError =
//					new AcmeError(
//						type,
//						"Invariant stored in element type "
//							+ type.getTypeName()
//							+ " heuristics list (internal error): "
//							+ heuristic.toString());
//				typecheckingErrorStack.pushCausalError(newInternalError);
//				AcmeErrorManager.errorDetected(newInternalError);
//			}
//
//			if (heuristic.typecheck() == false) {
//				typechecks = false;
//				ExpressionTypecheckError newError =
//					new ExpressionTypecheckError(
//						type,
//						"Heuristic does not typecheck: " + heuristic.toString(),
//						heuristic.getDesignRuleExpression());
//				typecheckingErrorStack.pushCausalError(newError);
//			}
//		}

		// check properties
		e = ((AcmeElement) type).getProperties();
		while (e.moreItems()) {
			AcmeProperty prop = (AcmeProperty) e.getNextItem();
			if (typecheckProperty(prop) == false) {
				typechecks = false;
				PropertyTypecheckError propTypeError =
					new PropertyTypecheckError(type, "Property " + prop.getName() + " does not typecheck.", prop);
				typecheckingErrorStack.pushCausalError(propTypeError);
				AcmeErrorManager.errorDetected(propTypeError);
			}
		}

		// Check that the representations of this element type
		// are internally consistent.
		AcmeEnumeration type_reps = ((AcmeElement) type).getRepresentations();
		while (type_reps.moreItems()) {
			AcmeRepresentation type_rep = (AcmeRepresentation) type_reps.getNextItem();
			if (type_rep == null)
				continue;
			if (type_rep.getSystem() == null)
				continue;
			else {
				AcmeSystem rep_sys = type_rep.getSystem();

				// assume a type error in typechecking the rep, cancel it
				// if the rep does indeed typecheck.  This will leave an
				// appropriate causal error trail if there is indeed a type error
				// in the rep.
				RepTypecheckError repError = new RepTypecheckError();
				typecheckingErrorStack.pushCausalError(repError);

				if (typecheckSystem(rep_sys)) {
					typecheckingErrorStack.cancelError(repError);
					AcmeErrorManager.errorRepaired(rep_sys, repError);
				}
				else {
					typechecks = false;
					repError.setSource(type_rep);
					repError.setErrorRep(type_rep);
					repError.setErrorMessage("Representation does not typecheck");
					typecheckingErrorStack.packageErrorCauses(repError);
				}
			}
		}

		return typechecks;
	}

	/**
	 * @param type
	 * @return
	 */
	private static boolean checkForCyclesInSupertypes(AcmeElementType type) {
		Set cycleSet = new HashSet ();
		cycleSet.add (type);
		return checkForCycles (type, cycleSet);
	}
	
	private static boolean checkForCycles (AcmeElementType type, Set cycleSet) {
		boolean typechecks = true;
		AcmeVector superTypes = type.getSuperTypes ();
		if (superTypes != null && superTypes.getSize() > 0) {
			AcmeEnumeration enum = superTypes.enumerate();
			while (enum.moreItems () && typechecks) {
				AcmeElementType  et = (AcmeElementType  )enum.getNextItem ();
				Set newCycleSet = new HashSet (cycleSet);
				if (newCycleSet.contains (et)) typechecks = false;
				else {
					newCycleSet.add(et);
					typechecks = checkForCycles (et, newCycleSet);
				}
			}
		}
		return typechecks;
	}

	/**
	 * helper method that sorts the types declared in the family so that
	 * all substructure that is used within a type declarations that has
	 * a type declared, that type will already have been typechecked before
	 * the enclosing type declaration is typechecked.  This is primarily
	 * an issue for
	 *
	 * @param type The AcmeFamily tested
	 * @return true if type is internally type-consistent and all of its supertypes are also families
	 */
	protected static AcmeEnumeration getSortedFamilyTypeDecls(AcmeFamily type) {
		return null;
	}

	/**
	 * confirm that the AcmeFamily "type" is internally type-consistent
	 * and that all of its supertypes are also families.
	 * @param type The AcmeFamily tested
	 * @return true if type is internally type-consistent and all of its supertypes are also families
	 */
	public static boolean typecheckFamily(AcmeFamily type) {
		boolean typechecks = true;
		boolean curItemTypechecks = false;
		ElementTypeTypecheckError eltError = new ElementTypeTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (type == null) {
			typechecks = false;
			eltError.setErrorMessage("Family to typecheck is a null reference");

		}
		else {

			// check that all of this family's supertypes are legal
			// family supertypes
			AcmeVector superTypes = type.getSuperTypes();
			for (int i = 0; i < superTypes.getSize(); i++) {
				try {
					AcmeElementType superType = (AcmeElementType) superTypes.getItemAt(i);
					if (!(superType instanceof AcmeFamily) && !(superType instanceof AcmeElementProtoType)) {
						IllegalSubtypeError newError =
							new IllegalSubtypeError(
								type,
								"Element type " + superType.getTypeName() + " can not be a Family supertype",
								type,
								superType);
						typecheckingErrorStack.pushCausalError(newError);
						AcmeErrorManager.errorDetected(newError);
					}

					// RTM 2/9/98 -- Add support here for checking that all of the
					//               supertypes and this type's extensions can be
					//               legally instantiated/unified.
					boolean tc = true;
					if (superType instanceof AcmeElementProtoType) {
						tc = unifyElementWithPrototype((AcmeElement) superType, (AcmeElement) type, false);
					}
					else
						tc = unifySystem((AcmeSystem) superType, (AcmeSystem) type, false);
					if (!tc) {
						ElementTypeTypecheckError error =
							new ElementTypeTypecheckError(
								type,
								"Could not unify with supertype " + superType.getTypeName());
						typecheckingErrorStack.pushCausalError(error);
						AcmeErrorManager.errorDetected(error);
					}
					typechecks &= tc;
				}
				catch (ClassCastException exc) {
					typechecks = false;
					IllegalSubtypeError newError =
						new IllegalSubtypeError(type, "Illegal family supertype", type, null);
					typecheckingErrorStack.pushCausalError(newError);
					AcmeErrorManager.errorDetected(newError);
				}
			}

			// confirm that all of the design analyses in this family typecheck.
			AcmeEnumeration e = type.getDesignAnalyses();
			while (e.moreItems()) {
				DesignAnalysis da = (DesignAnalysis) e.getNextItem();
				if (da.typecheck() == false) {
					typechecks = false;
					ExpressionTypecheckError exprTypecheckError =
						new ExpressionTypecheckError(
							da,
							type.getTypeName()
								+ "'s design analysis declaration for \""
								+ da.getName()
								+ "\" does not typecheck. ",
							da.getDesignRuleExpression());
					typecheckingErrorStack.pushCausalError(exprTypecheckError);
				}
			}

			// typecheck all of the type definitions defined in the
			// family's typespace.
			AcmeEnumeration types = type.getTypes();
			while (types.moreItems()) {
				AcmeObject curType = types.getNextItem();
				if (curType instanceof AcmePortType) {
					if (AcmeTypeManager.typecheckPortType((AcmePortType) curType) == false) {
						typechecks = false;
					}
				}
				else if (curType instanceof AcmeRoleType) {
					if (AcmeTypeManager.typecheckRoleType((AcmeRoleType) curType) == false) {
						typechecks = false;
					}
				}
				else if (curType instanceof AcmeComponentType) {
					if (AcmeTypeManager.typecheckComponentType((AcmeComponentType) curType) == false) {
						typechecks = false;
					}
				}
				else if (curType instanceof AcmeConnectorType) {
					if (AcmeTypeManager.typecheckConnectorType((AcmeConnectorType) curType) == false) {
						typechecks = false;
					}
				}
				else if (curType instanceof AcmeElementProtoType) {
					if (AcmeTypeManager.typecheckElementProtoType((AcmeElementProtoType) curType) == false) {
						typechecks = false;
					}
				}
			}

			// confirm that all of the family's components and connectors
			// typecheck.  Any errors discovered while typechecking them
			// will be put properly at the top of the typecheckingErrorStack
			// in the typecheck{Component,Connector} methods.
			e = type.getComponents();
			while (e.moreItems()) {
				AcmeComponent cur_comp = (AcmeComponent) e.getNextItem();
				if (typecheckComponent(cur_comp) == false)
					typechecks = false;
			}

			e = type.getConnectors();
			while (e.moreItems()) {
				AcmeConnector cur_conn = (AcmeConnector) e.getNextItem();
				if (typecheckConnector(cur_conn) == false)
					typechecks = false;
			}

			// if we're dealing with PURE_ACME then a system can have ports
			// and roles and the attachments can have properties that need
			// to be typechecked, so typecheck those also.
			if (acmeLib.core.globals.LanguageConfig.IS_PURE_ACME == true) {
				e = type.getPorts();
				while (e.moreItems()) {
					AcmePort cur_port = (AcmePort) e.getNextItem();
					if (typecheckPort(cur_port) == false)
						typechecks = false;
				}

				e = type.getRoles();
				while (e.moreItems()) {
					AcmeRole cur_role = (AcmeRole) e.getNextItem();
					if (typecheckRole(cur_role) == false)
						typechecks = false;
				}

				e = type.getAttachments();
				while (e.moreItems()) {
					AcmeAttachment cur_attach = (AcmeAttachment) e.getNextItem();
					AcmeEnumeration props = cur_attach.getProperties();
					while (props.moreItems()) {
						AcmeProperty prop = (AcmeProperty) props.getNextItem();
						if (typecheckProperty(prop) == false) {
							typechecks = false;
							PropertyTypecheckError newPTError =
								new PropertyTypecheckError(
									cur_attach,
									"Property \""
										+ prop.getName()
										+ "\" of attachment \""
										+ cur_attach.getPortNameAsString()
										+ " to "
										+ cur_attach.getRoleNameAsString()
										+ "\" does not typecheck.",
									prop);
							typecheckingErrorStack.pushCausalError(newPTError);
							AcmeErrorManager.errorDetected(newPTError);
						}
					}
				}
			}

			curItemTypechecks = typecheckElementType(type);
			if (curItemTypechecks == false) {
				// RTM 4/2/98 -- throw more detailed error message.
				typechecks = false;
			}

		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(type, eltError);
		}
		else {
			eltError.setElementType(type);
			eltError.setSource(type);
			if (type != null)
				eltError.setErrorMessage("Family " + type.getTypeName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the AcmeComponentType "type" is internally type-consistent
	 * and that all of its supertypes are also AcmeComponentTypes.
	 * @param type The AcmeComponentType tested
	 * @return true if type is internally type-consistent and all its supertypes are also AcmeComponentTypes
	 */
	public static boolean typecheckComponentType(AcmeComponentType type) {
		boolean typechecks = true;
		ElementTypeTypecheckError eltError = new ElementTypeTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (type == null) {
			typechecks = false;
			eltError.setErrorMessage("Component Type to typecheck is a null reference");
		}
		else {
			typechecks = typecheckElementType(type);

			// confirm that all of the type's roles typecheck.  Any errors
			// discovered while typechecking the ports will be put properly at
			// the top of the typecheckingErrorStack in the typecheckRole method.
			AcmeEnumeration e = type.getPorts();
			while (e.moreItems()) {
				AcmePort cur_port = (AcmePort) e.getNextItem();
				if (typecheckPort(cur_port) == false)
					typechecks = false;
			}
			if (typechecks) {
			AcmeVector superTypes = type.getSuperTypes();
			for (int i = 0; i < superTypes.getSize(); i++) {
				try {
					AcmeElementType superType = (AcmeElementType) superTypes.getItemAt(i);
					if (!(superType instanceof AcmeComponentType)
						&& !(superType instanceof AcmeElementProtoType)
						&& (superType instanceof AcmeElementTypeRef
							&& (!(((AcmeElementTypeRef) superType).getReferentType() instanceof AcmeElementProtoType)
								&& !(((AcmeElementTypeRef) superType).getReferentType() instanceof AcmeComponentType)))) {

						IllegalSubtypeError newError =
							new IllegalSubtypeError(
								type,
								"Element type " + superType.getTypeName() + " can not be a Component supertype",
								type,
								superType);
						typecheckingErrorStack.pushCausalError(newError);
						AcmeErrorManager.errorDetected(newError);
					}

					// RTM 2/9/98 -- Add support here for checking that all of the
					//               supertypes and this type's extensions can be
					//               legally instantiated/unified.
					boolean tc = true;
					if (superType instanceof AcmeElementProtoType) {
						tc = unifyElementWithPrototype((AcmeElement) superType, (AcmeElement) type, false);
					}
					else if (superType instanceof AcmeElementTypeRef) {
						AcmeElementTypeRef ref = (AcmeElementTypeRef) superType;
						tc = unifyComponent((AcmeComponent) ref.getReferentType(), (AcmeComponent) type, false);
					}
					else {
						tc = unifyComponent((AcmeComponent) superType, (AcmeComponent) type, false);
					}
					if (!tc) {
						ElementTypeTypecheckError error =
							new ElementTypeTypecheckError(
								type,
								"Could not unify with supertype " + superType.getTypeName());
						typecheckingErrorStack.pushCausalError(error);
						AcmeErrorManager.errorDetected(error);
					}
					typechecks &= tc;
				}
				catch (ClassCastException exc) {
					typechecks = false;
					IllegalSubtypeError newError =
						new IllegalSubtypeError(type, "Illegal component supertype", type, null);
					typecheckingErrorStack.pushCausalError(newError);
					AcmeErrorManager.errorDetected(newError);
				}
			}
		}
		}
		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(type, eltError);
		}
		else {
			eltError.setElementType(type);
			eltError.setSource(type);
			if (type != null)
				eltError.setErrorMessage("Component Type " + type.getTypeName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	   * confirm that the AcmeConnectorType "type" is internally type-consistent
	   * and that all of its supertypes are also AcmeConnectorTypes.
	   * @param type The AcmeConnectorType tested
	   * @return true if type is internally type-consistent and all its supertypes are also AcmeConnectorTypes
	   */
	public static boolean typecheckConnectorType(AcmeConnectorType type) {
		boolean typechecks = true;
		ElementTypeTypecheckError eltError = new ElementTypeTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (type == null) {
			typechecks = false;
			eltError.setErrorMessage("Connector Type to typecheck is a null reference");
		}
		else {
			typechecks = typecheckElementType(type);

			// confirm that all of the type's roles typecheck.  Any errors
			// discovered while typechecking the ports will be put properly at
			// the top of the typecheckingErrorStack in the typecheckRole method.
			AcmeEnumeration e = type.getRoles();
			while (e.moreItems()) {
				AcmeRole cur_role = (AcmeRole) e.getNextItem();
				if (typecheckRole(cur_role) == false)
					typechecks = false;
			}

			AcmeVector superTypes = type.getSuperTypes();
			for (int i = 0; i < superTypes.getSize(); i++) {
				try {
					AcmeElementType superType = (AcmeElementType) superTypes.getItemAt(i);
					if (!(superType instanceof AcmeConnectorType)
						&& !(superType instanceof AcmeElementProtoType)
						&& !(superType instanceof AcmeElementProtoType)
						&& (superType instanceof AcmeElementTypeRef
							&& (!(((AcmeElementTypeRef) superType).getReferentType() instanceof AcmeElementProtoType)
								&& !(((AcmeElementTypeRef) superType).getReferentType() instanceof AcmeConnectorType)))) {
						IllegalSubtypeError newError =
							new IllegalSubtypeError(
								type,
								"Element type " + superType.getTypeName() + " can not be a Connector supertype",
								type,
								superType);
						typecheckingErrorStack.pushCausalError(newError);
						AcmeErrorManager.errorDetected(newError);
					}

					// RTM 2/9/98 -- Add support here for checking that all of the
					//               supertypes and this type's extensions can be
					//               legally instantiated/unified.
					boolean tc = true;
					if (superType instanceof AcmeElementProtoType) {
						tc = unifyElementWithPrototype((AcmeElement) superType, (AcmeElement) type, false);
					}
					else if (superType instanceof AcmeElementTypeRef) {
						AcmeElementTypeRef ref = (AcmeElementTypeRef) superType;
						tc = unifyConnector((AcmeConnector) ref.getReferentType(), (AcmeConnector) type, false);
					}
					else
						tc = unifyConnector((AcmeConnector) superType, (AcmeConnector) type, false);
					if (!tc) {
						ElementTypeTypecheckError error =
							new ElementTypeTypecheckError(
								type,
								"Could not unify with supertype " + superType.getTypeName());
						typecheckingErrorStack.pushCausalError(error);
						AcmeErrorManager.errorDetected(error);
					}
					typechecks &= tc;
				}
				catch (ClassCastException exc) {
					typechecks = false;
					IllegalSubtypeError newError =
						new IllegalSubtypeError(type, "Illegal connector supertype", type, null);
					typecheckingErrorStack.pushCausalError(newError);
					AcmeErrorManager.errorDetected(newError);
				}
			}
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(type, eltError);
		}
		else {
			eltError.setElementType(type);
			eltError.setSource(type);
			if (type != null)
				eltError.setErrorMessage("Connector Type " + type.getTypeName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the AcmePortType "type" is internally type-consistent
	 * and that all of its supertypes are also AcmePortTypes.
	 * @param type The AcmePortType tested
	 * @return true if type is internally type-consistent and all its supertypes are also AcmePortTypes
	 */
	public static boolean typecheckPortType(AcmePortType type) {
		boolean typechecks = true;
		ElementTypeTypecheckError eltError = new ElementTypeTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (type == null) {
			typechecks = false;
			eltError.setErrorMessage("Port Type to typecheck is a null reference");
		}
		else {
			typechecks = typecheckElementType(type);

			AcmeVector superTypes = type.getSuperTypes();
			for (int i = 0; i < superTypes.getSize(); i++) {
				try {
					AcmeElementType superType = (AcmeElementType) superTypes.getItemAt(i);
					if (!(superType instanceof AcmePortType)
						&& !(superType instanceof AcmeElementProtoType)
						&& (superType instanceof AcmeElementTypeRef
							&& (!(((AcmeElementTypeRef) superType).getReferentType() instanceof AcmeElementProtoType)
								&& !(((AcmeElementTypeRef) superType).getReferentType() instanceof AcmePortType)))) {
						IllegalSubtypeError newError =
							new IllegalSubtypeError(
								type,
								"Element type " + superType.getTypeName() + " can not be a Port supertype",
								type,
								superType);
						typecheckingErrorStack.pushCausalError(newError);
						AcmeErrorManager.errorDetected(newError);
					}

					// RTM 2/9/98 -- Add support here for checking that all of the
					//               supertypes and this type's extensions can be
					//               legally instantiated/unified.
					boolean tc = true;
					if (superType instanceof AcmeElementProtoType) {
						tc = unifyElementWithPrototype((AcmeElement) superType, (AcmeElement) type, false);
					}
					else if (superType instanceof AcmeElementTypeRef) {
						AcmeElementTypeRef ref = (AcmeElementTypeRef) superType;
						tc = unifyPort((AcmePort) ref.getReferentType(), (AcmePort) type, false);
					}
					else
						tc = unifyPort((AcmePort) superType, (AcmePort) type, false);
					if (!tc) {
						ElementTypeTypecheckError error =
							new ElementTypeTypecheckError(
								type,
								"Could not unify with supertype " + superType.getTypeName());
						typecheckingErrorStack.pushCausalError(error);
						AcmeErrorManager.errorDetected(error);
					}
					typechecks &= tc;
				}
				catch (ClassCastException e) {
					typechecks = false;
					IllegalSubtypeError newError = new IllegalSubtypeError(type, "Illegal port supertype", type, null);
					typecheckingErrorStack.pushCausalError(newError);
					AcmeErrorManager.errorDetected(newError);
				}
			}
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(type, eltError);
		}
		else {
			eltError.setElementType(type);
			eltError.setSource(type);
			if (type != null)
				eltError.setErrorMessage("Port Type " + type.getTypeName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the AcmeRoleType "type" is internally type-consistent
	 * and that all of its supertypes are also AcmeRoleTypes.
	 * @param type The AcmeRoleType tested
	 * @return true if type is internally type-consistent and all its supertypes are also AcmeRoleTypes
	 */
	public static boolean typecheckRoleType(AcmeRoleType type) {
		boolean typechecks = true;
		ElementTypeTypecheckError eltError = new ElementTypeTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (type == null) {
			typechecks = false;
			eltError.setErrorMessage("Role Type to typecheck is a null reference");
		}
		else {
			typechecks = typecheckElementType(type);

			AcmeVector superTypes = type.getSuperTypes();
			for (int i = 0; i < superTypes.getSize(); i++) {
				try {
					AcmeElementType superType = (AcmeElementType) superTypes.getItemAt(i);
					if (!(superType instanceof AcmeRoleType)
						&& !(superType instanceof AcmeElementProtoType)
						&& !(superType instanceof AcmeElementProtoType)
						&& (superType instanceof AcmeElementTypeRef
							&& (!(((AcmeElementTypeRef) superType).getReferentType() instanceof AcmeElementProtoType)
								&& !(((AcmeElementTypeRef) superType).getReferentType() instanceof AcmeRoleType)))) {
						IllegalSubtypeError newError =
							new IllegalSubtypeError(
								type,
								"Element type " + superType.getTypeName() + " can not be a Role supertype",
								type,
								superType);
						typecheckingErrorStack.pushCausalError(newError);
						AcmeErrorManager.errorDetected(newError);
					}

					// RTM 2/9/98 -- Add support here for checking that all of the
					//               supertypes and this type's extensions can be
					//               legally instantiated/unified.
					boolean tc = true;
					if (superType instanceof AcmeElementProtoType) {
						tc = unifyElementWithPrototype((AcmeElement) superType, (AcmeElement) type, false);
					}
					else if (superType instanceof AcmeElementTypeRef) {
						AcmeElementTypeRef ref = (AcmeElementTypeRef) superType;
						tc = unifyRole((AcmeRole) ref.getReferentType(), (AcmeRole) type, false);
					}
					else
						tc = unifyRole((AcmeRole) superType, (AcmeRole) type, false);
					if (!tc) {
						ElementTypeTypecheckError error =
							new ElementTypeTypecheckError(
								type,
								"Could not unify with supertype " + superType.getTypeName());
						typecheckingErrorStack.pushCausalError(error);
						AcmeErrorManager.errorDetected(error);
					}
					typechecks &= tc;
				}
				catch (ClassCastException e) {
					typechecks = false;
					IllegalSubtypeError newError = new IllegalSubtypeError(type, "Illegal role supertype", type, null);
					typecheckingErrorStack.pushCausalError(newError);
					AcmeErrorManager.errorDetected(newError);
				}
			}
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(type, eltError);
		}
		else {
			eltError.setElementType(type);
			eltError.setSource(type);
			if (type != null)
				eltError.setErrorMessage("Role Type " + type.getTypeName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	/**
	 * confirm that the AcmeElementPrototType "type" is internally type-consistent
	 * and that all of its supertypes are also AcmeElementPrototTypes.
	 * @param type The AcmePortType tested
	 * @return true if type is internally type-consistent and all its supertypes are
	 *    also AcmeElementPrototTypes
	 */
	public static boolean typecheckElementProtoType(AcmeElementProtoType type) {
		boolean typechecks = true;
		ElementTypeTypecheckError eltError = new ElementTypeTypecheckError();
		typecheckingErrorStack.pushCausalError(eltError);

		if (type == null) {
			typechecks = false;
			eltError.setErrorMessage("Element Type to typecheck is a null reference");
		}
		else {
			typechecks = typecheckElementType(type);

			AcmeVector superTypes = type.getSuperTypes();
			for (int i = 0; i < superTypes.getSize(); i++) {
				try {
					AcmeElementType superType = (AcmeElementType) superTypes.getItemAt(i);
					if (!(superType instanceof AcmeElementProtoType)) {
						IllegalSubtypeError newError =
							new IllegalSubtypeError(
								type,
								"Element type " + superType.getTypeName() + " can not be a Element Type supertype",
								type,
								superType);
						typecheckingErrorStack.pushCausalError(newError);
						AcmeErrorManager.errorDetected(newError);
					}

					// RTM 2/9/98 -- Add support here for checking that all of the
					//               supertypes and this type's extensions can be
					//               legally instantiated/unified.

				}
				catch (ClassCastException e) {
					typechecks = false;
					IllegalSubtypeError newError =
						new IllegalSubtypeError(type, "Illegal Element Type supertype", type, null);
					typecheckingErrorStack.pushCausalError(newError);
					AcmeErrorManager.errorDetected(newError);
				}
			}
		}

		// clean up the error stack and return the result of the typechecking.
		if (typechecks) {
			typecheckingErrorStack.cancelError(eltError);
			AcmeErrorManager.errorRepaired(type, eltError);
		}
		else {
			eltError.setElementType(type);
			eltError.setSource(type);
			if (type != null)
				eltError.setErrorMessage("Element Type " + type.getTypeName() + " does not typecheck");
			typecheckingErrorStack.packageErrorCauses(eltError);
		}

		return typechecks;
	}

	//
	// Typechecking utility methods
	//

	/**
	 * check that a property's value satisfies it's type
	 * @param p The AcmeProperty tested
	 * @return true if the property's value satisfies it's type
	 */
	public static boolean typecheckProperty(AcmeProperty p) {
		return typecheckProperty(p.getType(), p.getValue());
	}

	/**
	 * check that a property's value satisfies it's type
	 * @param t The AcmePropertyType tested
	 * @param val The AcmePropertyValue tested
	 * @return true if the property's value satisfies it's type
	 */
	public static boolean typecheckProperty(AcmeType t, AcmePropertyValue val) {
		// unlike the other typechecking methods, this one tries to
		// prove that val is of type t, and returns false if it is not.
		//
		// first check for the easy cases, where we have no type, primitive
		// types that have matching values or the type "any"
		if (t instanceof AcmePropertyTypeRef) {
			t = ((AcmePropertyTypeRef) t).getReferentType();
		}

		if (t == null)
			return true; // null type means "any"
		else if (val == null)
			return true; // null value satisfies any type.

		else if (t instanceof AnyType)
			return true; // all values satisfy type "any"

		else if (t instanceof BooleanType) {
			return (val instanceof AcmeBoolean);
		}
		else if (t instanceof IntType) {
			if (val instanceof AcmeInt)
				return true;
			else
				return false;
		}
		else if (t instanceof FloatType) {
			if (val instanceof AcmeFloat)
				return true;
			else if (val instanceof AcmeInt)
				return true;
			else
				return false;
		}
		else if (t instanceof StringType) {
			return (val instanceof AcmeString);
		}
		else if (t instanceof EnumType) {
			if (!(val instanceof AcmeEnum))
				return false;
			else {
				// Make sure the value has the the type of the enum
				
				
				boolean	ok	= (((EnumType) t).hasValue(((AcmeEnum) val).getValue()));
				if (ok) {
					((AcmeEnum )val).setType ((EnumType )t);
				}
				return ok;
			}
		}
		else if (t instanceof SetType) {
			if (!(val instanceof AcmeSet))
				return false;
			else if (
				(((SetType) t).getSetType() == null)
					|| (((SetType) t).getSetType() instanceof AnyType)
					|| (((SetType) t).getSetType() == ConstantTypes.UNSPECIFIED_TYPE)) {
				return true; // all set elements satisfy the set type
			}
			else {
				// check that all of the elements in the set satisfy setType
				AcmePropertyType prop_type = (AcmePropertyType) ((SetType) t).getSetType();
				((AcmeSet )val).setType((SetType )t);
				AcmeEnumeration set_items = ((AcmeSet) val).getValues();
				while (set_items.moreItems()) {
					AcmePropertyValue cur_val = (AcmePropertyValue) set_items.getNextItem();
					if (typecheckProperty(prop_type, cur_val) == false)
						return false;
				}
				return true; // all of the set's elements typecheck.
			}
		}
		else if (t instanceof SequenceType) {
			if (!(val instanceof AcmeSequence))
				return false;
			else if (
				(((SequenceType) t).getSequenceType() == null)
					|| (((SequenceType) t).getSequenceType() instanceof AnyType)
					|| (((SequenceType) t).getSequenceType() == ConstantTypes.UNSPECIFIED_TYPE)) {
				return true; // all sequence elements satisfy the type "any"
			}
			else {
				// check that each element in the sequence satisfies sequenceType
				((AcmeSequence )val).setType ((SequenceType )t);
				AcmePropertyType prop_type = ((SequenceType) t).getSequenceType();
				AcmeEnumeration seq_items = ((AcmeSequence) val).getValues();
				while (seq_items.moreItems()) {
					AcmePropertyValue cur_val = (AcmePropertyValue) seq_items.getNextItem();
					if (typecheckProperty(prop_type, cur_val) == false)
						return false;
				}
				return true; // all of the sequence's elements typecheck.
			}
		}
		else if (t instanceof RecordType) {
			if (!(val instanceof AcmeRecord))
				return false;
			else {
				// if val doesn't have all of the fields required by
				// RecordType then it doesn't typechecks
				AcmeRecordField type_field, val_field;
				AcmePropertyType type_field_type, val_field_type;
				AcmePropertyValue type_field_value, val_field_value;
				AcmeRecord val_record = (AcmeRecord) val;
				AcmeEnumeration fields = ((RecordType) t).getFields();
				while (fields.moreItems()) {
					type_field = (AcmeRecordField) fields.getNextItem();
					val_field = val_record.getField(type_field.getName());
					if (val_field == null)
						return false;
					else {
						type_field_type = type_field.getType();
						val_field_type = val_field.getType();
						if ((val_field_type != null) && (val_field_type != ConstantTypes.UNSPECIFIED_TYPE)) {
							if (!isLogicalSubtype(val_field_type, type_field_type, false))
								return false;
						}
						else if (typecheckProperty(type_field_type, val_field.getValue()) == false) {
							return false;
						}
						type_field_value = type_field.getValue();
						if (type_field_value != null) {
							if (!type_field_value.equals(val_field.getValue()))
								return false;
						}
					}
				}

				// if we get to here, then the instance has all of the fields
				// required by the type.  Now check that each of the instance's
				// fields' type/value pairs are internally consistent.
				fields = null; // mark for gc
				fields = val_record.getFields();

				while (fields.moreItems()) {
					val_field = (AcmeRecordField) fields.getNextItem();
					val_field_type = val_field.getType();
					val_field_value = val_field.getValue();
					if (typecheckProperty(val_field_type, val_field_value) == false)
						return false; // value doesn't typecheck with type.
				}
			}

			// if we get to here, within this "if a record" block, then we have
			// been unable to show that this record does not typecheck, in effect
			// proving that it does typecheck (just within the record tests)
			return true;
		}

		return false; // Assume it doesn't typecheck if I can't prove
		// that it does by now.
	}

	/**
	 * isLogicalSubtype(subType, superType) determines whether the set of
	 * entities defined by the type "subType" is a logical subset of the set
	 * of entities defined by the type "superType".  If so, then sub is a
	 * subtype of super, regardless of whether they are declared to have an
	 * explicit sub/super type relationship.  Returns true if subType is a
	 * logical subtype of superType.
	 * @param subType The subtype
	 * @param superType The supertype
	 * @return true if subType is a logical subtype of superType
	 */
	public static boolean isLogicalSubtype(AcmeType subType, AcmeType superType, boolean unresolvedOK) {
		// a null type is not a legal subtype or supertype of anything.
		if ((superType == null) || (subType == null))
			return false;

		if (subType == superType)
			return true;
		if (subType instanceof AcmeElementTypeRef) {
			subType = ((AcmeElementTypeRef) subType).getReferentType();

		}
		if (superType instanceof AcmeElementTypeRef) {
			superType = ((AcmeElementTypeRef) superType).getReferentType();

		}
		if (subType instanceof AcmePropertyTypeRef) {
			subType = ((AcmePropertyTypeRef) subType).getReferentType();
		}
		if (superType instanceof AcmePropertyTypeRef) {
			superType = ((AcmePropertyTypeRef) superType).getReferentType();
		}
		// every element type is a subtype of ELEMENT_BASE_TYPE
		if (superType == ConstantTypes.ELEMENT_BASE_TYPE) {
			return (subType instanceof AcmeElementType || subType instanceof AcmeUnresolvedType);
		}

		// and all types are subtypes of "any"
		if (superType instanceof AnyType) {
			return true;
		}

		// check property types
		if (subType instanceof AcmePropertyType) {
			if (superType instanceof AcmeElementType)
				return false;
			else if (superType instanceof PropertyReferenceType) {
				// all property types can be PropertyReferenceTypes.
				// RTM 12/1/98 -- This may prove to cause problems distinguishing
				//                property value types from property reference types
				//                but it doesn't seem to present a problem currently...
				return true;
			}
			else if (subType instanceof BooleanType)
				return (superType instanceof BooleanType);
			else if (subType instanceof IntType)
				return ((superType instanceof IntType) || (superType instanceof FloatType));
			else if (subType instanceof FloatType)
				return (superType instanceof FloatType);
			else if (subType instanceof StringType)
				return (superType instanceof StringType);

			else if (subType instanceof EnumType) {
				// enum types have to have exactly the same fields in exactly
				// the same order to be valid sub/super types.
				if (!(superType instanceof EnumType)) {
					return false;
				}
				else {
					// see if the contents of the EnumTypes are identical
					// (satisfying structural equivalence).
					AcmeEnumeration subVals = ((EnumType) subType).getValues();
					AcmeEnumeration superVals = ((EnumType) superType).getValues();
					String curSubVal, curSuperVal;
					while (subVals.moreItems() && superVals.moreItems()) {
						curSubVal = ((EnumValue) subVals.getNextItem()).value;
						curSuperVal = ((EnumValue) superVals.getNextItem()).value;
						if (!curSubVal.equals(curSuperVal)) {
							return false;
						}
					}
					// we haven't found a mismatch yet.  Check whether we have
					// exhausted both enumerations.  If there are values left
					// over in one of the enumerations then the types are
					// different because they have a different number of values.
					return !(subVals.moreItems() || superVals.moreItems());
				}
			}
			else if (subType instanceof SetType) {
				if (!(superType instanceof SetType)) {
					return false;
				}
				else {
					// if the set supertype specifies a type for its elements
					// check that the subtype's element type is an appropriate
					// subtype.  Note that a "null" value for the setType indicates
					// it is a set that can hold any kind of object.  All sets
					// are subtypes of type Set{}.
					AcmeType subTypeSetType = ((SetType) subType).getSetType();
					AcmeType superTypeSetType = ((SetType) superType).getSetType();

					if (superTypeSetType == null) {
						return true; // All sets are subtypes of type Set{}
					}
					else {
						return isLogicalSubtype(subTypeSetType, superTypeSetType, unresolvedOK);
					}
				}
			}
			else if (subType instanceof SequenceType) {
				if (!(superType instanceof SequenceType))
					return false;
				else
					return isLogicalSubtype(
						((SequenceType) subType).getSequenceType(),
						((SequenceType) superType).getSequenceType(),
						unresolvedOK);
			}
			else if (subType instanceof RecordType) {
				if (!(superType instanceof RecordType)) {
					return false;
				}
				AcmeEnumeration superFields = ((RecordType) superType).getFields();
				AcmeRecordField subField, superField;
				while (superFields.moreItems()) {
					superField = (AcmeRecordField) superFields.getNextItem();
					if (superField == null)
						continue; // some error in this's field return method
					subField = ((RecordType) subType).getField(superField.getName());
					if (subField == null)
						return false;
					// field exists in this but not test instance.
					if (subField != null) {
						if (!AcmeTypeManager
							.isLogicalSubtype(subField.getType(), superField.getType(), unresolvedOK)) {
							return false;
						}
					}
				} // end while moreItems();

				// if we get to here then we haven't been able to show that
				// subType is _not_ a subtype of superType, so return true.
				return true;
			}
			else {
				// else we have an unrecognized property type.  print
				// out a warning message and return false.
				// RTM 2/2/98 -- add better error handling here.
				System.err.println("Warning: Encountered unrecognized property type " + subType.getTypeName());
				return false;
			}

		}
		// end of subType is a property type clause, check for element types.
		else if (subType instanceof AcmeElementType) {
			// For now (2/2/98), we are only dealing with explicitly
			// declared subtypes, so check whether subType is an ancestor
			// of superType in the type hierarchy of explicitly declared
			// Armani element types.  Return true if it is, else return false.
			if (!(superType instanceof AcmeElementType)) {
				// SWC 1/03/03 -- the "Type" formal parameter is not an
				//   AcmeElementType, so the original implementation here
				//   would not allow a "ComponenT" to be a logical subtype
				//   of "Type".  This change below allows such a case to pass.
				if (superType instanceof AcmeTypeParameter
					&& ((AcmeTypeParameter) superType).getTypeName().equalsIgnoreCase("Type")) {
					return true;
				}
				else {
					return false;
				}
			}
			else if (isDeclaredSubtype((AcmeElementType) subType, (AcmeElementType) superType)) {
				return true;
			}
			else {
				// RTM 2/2/98 -- consider looking for implicit subtypes
				//               here.  ie subtypes that will work even
				//               though they are not explicitly specified.
				return false;
			}
		}

		// else we're dealing with a type object that is neither a property
		// nor an element type.  flag a warning and return false.
		else {
			// RTM 2/2/98 -- add better error handling here.
			System.err.println(
				"Warning: Encountered unrecognized type "
					+ subType.getTypeName()
					+ ", which is neither a property nor element type");
			return unresolvedOK;
		}

	} // end isLogicalSubtype

	/**
	 * isDeclaredSubtype returns true iff the parameter subType is an
	 * explicitly declared subtype of the parameter superType.
	 * @param subType subtype
	 * @param superType supertype
	 * @return true iff subType is explicitly declared subtype of superType
	 */
	public static boolean isDeclaredSubtype(AcmeElementType subType, AcmeElementType superType) {
		// first check for equality.
		if (subType == superType)
			return true;
		if (subType instanceof AcmeElementTypeRef) {
			subType = ((AcmeElementTypeRef) subType).getReferentType();

		}
		if (superType instanceof AcmeElementTypeRef) {
			superType = ((AcmeElementTypeRef) superType).getReferentType();

		}
		// if they're not equal then try to walk up the subtype's inheritance
		// tree to see if it finds the supertype.
		AcmeVector superTypesV = subType.getSuperTypes();

		if (superTypesV.contains((AcmeObject) superType))
			return true;

		// else walk up the type hierarchy rooted at subType looking for
		// superType.  This method uses that fact that subTyping in Armani
		// is transitive.  That is, A < B and B < C -> A < C
		// (Where X < Y indicates that X is a subtype of Y),
		boolean superTypeFound = false;
		for (int i = 0; i < superTypesV.getSize(); i++) {
			AcmeElementType intermediateSuperType = (AcmeElementType) superTypesV.getItemAt(i);
			superTypeFound = isDeclaredSubtype(intermediateSuperType, superType);
			if (superTypeFound) {
				return true;
			}
		}

		// if we exhaust the list of superTypes without finding the required
		// type, then return false.
		return false;

	}

	/**
	 * inferLeastCommonSupertype walks over the vector of types passed
	 * as a parameter and computes the most specific type for which all of
	 * the types in the param vector are subtypes.  If it can not find
	 * such a type then it returns "any" in the form of
	 * ConstantTypes.UNSPECIFIED_TYPE.
	 * @param types The AcmeVector
	 * @return the most specific type for which all types in vector are subtypes
	 */
	public static AcmeType inferLeastCommonSupertype(AcmeVector types) {

		boolean allPropTypes = true;
		boolean allEltTypes = true;
		int typesSize = types.getSize();
		AcmeType returnType = ConstantTypes.UNSPECIFIED_TYPE;

		// four primary cases: an empty set, all property types,
		// all element types, or a mix of property and element types.
		if (types.getSize() == 0) {
			// can't infer a type from an empty set, return UNSPECIFIED_TYPE
			return ConstantTypes.UNSPECIFIED_TYPE;
		}

		// check for the all property/all element type vectors
		for (int i = 0; i < typesSize; i++) {
			if (!(types.getItemAt(i) instanceof AcmePropertyType))
				allPropTypes = false;
			else if (!(types.getItemAt(i) instanceof AcmeElementType))
				allEltTypes = false;
		}

		if (allPropTypes == true) {
			// then find the leastCommon _property_ Supertype.
			// all elements in the types vector have been shown to be
			// AcmePropertyTypes, and we know that we have at least 1 element.

			AcmeType lcs = null; // lcs == least common supertype
			AcmeType curType = null;

			// get the first type and set it as the base type for comparison
			lcs = (AcmeType) types.getItemAt(0);

			// then walk through the rest of the set elements, setting the
			// current type to the greatest-common-supertype of the two.
			for (int i = 1; i < types.getSize(); i++) {
				curType = (AcmeType) types.getItemAt(i);

				// check whether curType is a subtype of lcs.
				// if not, then check whether lcs is a subType of curType.
				// update lcs if and as appropriate.
				if (AcmeTypeManager.isLogicalSubtype(curType, lcs, false)) {
					// do nothing, lcs is set properly.
				}
				else if (AcmeTypeManager.isLogicalSubtype(lcs, curType, false)) {
					lcs = curType;
				}
				else {
					// else there is not sub/supertype relationship between
					// curType and nextType, so there is no legal lcs.
					lcs = ConstantTypes.UNSPECIFIED_TYPE;
					break;
				}
			}

			// we've found our lcs, set the returnType appropriately.
			returnType = lcs;
		}
		else if (allEltTypes == true) {
			// then find the leastCommon _element_ Supertype.

			// RTM 3/6/98 -- I haven't yet figured out how to do this when
			//               dealing with elements that declare multiple types,
			//               each with an arbitrary supertype structure.
			//               for now, just make sure that they're all components,
			//               connectors, ports, roles, or systems and set
			//               the type to that top-level type.  Perhaps this
			//               should be updated to actualy calculate the
			//               least-common-type.
			//
			//               In the meantime, just walk over all of the entries
			//               in the QuantifiedExpressionNode.typecheck() method
			//               to make sure that each of the entities satisfies
			//               the declared element type of the quantified variable.

			// begin with all flags true.  The checking algorithm can only
			// set them to false.  If one remains true, return that category's
			// ConstantType, else return UNRESOLVABLE_TYPE.
			boolean allComps = true;
			boolean allConns = true;
			boolean allPorts = true;
			boolean allRoles = true;
			boolean allFams = true;
			boolean allEltProtos = true;

			for (int i = 0; i < typesSize; i++) {
				if (types.getItemAt(i) instanceof AcmeComponentType) {
					allConns = false;
					allPorts = false;
					allRoles = false;
					allFams = false;
					allEltProtos = false;
				}
				else if (types.getItemAt(i) instanceof AcmeConnectorType) {
					allComps = false;
					allPorts = false;
					allRoles = false;
					allFams = false;
					allEltProtos = false;
				}
				else if (types.getItemAt(i) instanceof AcmePortType) {
					allComps = false;
					allConns = false;
					allRoles = false;
					allFams = false;
					allEltProtos = false;
				}
				else if (types.getItemAt(i) instanceof AcmeRoleType) {
					allComps = false;
					allConns = false;
					allPorts = false;
					allFams = false;
					allEltProtos = false;
				}
				else if (types.getItemAt(i) instanceof AcmeFamily) {
					allComps = false;
					allConns = false;
					allPorts = false;
					allRoles = false;
					allEltProtos = false;
				}
				else if (types.getItemAt(i) instanceof AcmeElementProtoType) {
					allComps = false;
					allConns = false;
					allPorts = false;
					allRoles = false;
					allFams = false;
				}
			}

			// no more than one of the all* flags should still be true,
			// set the return type appropriately.
			if (allComps == true)
				returnType = ConstantTypes.COMPONENT_BASE_TYPE;
			else if (allConns == true)
				returnType = ConstantTypes.CONNECTOR_BASE_TYPE;
			else if (allPorts == true)
				returnType = ConstantTypes.PORT_BASE_TYPE;
			else if (allRoles == true)
				returnType = ConstantTypes.ROLE_BASE_TYPE;
			else if (allFams == true)
				returnType = ConstantTypes.SYSTEM_BASE_TYPE;
			else if (allEltProtos == true)
				returnType = ConstantTypes.ELEMENT_BASE_TYPE;
		}

		// just fall through if its a mix of property and element types,
		// because returnType is initially set with the default error return
		// value.

		// returnType should now be set to the leastCommonSupertype
		return returnType;
	}

	/**
	 * overloaded method that simply converts the enumeration passed
	 * in into a vector and uses the method of the same name with a
	 * vector argument.
	 * @param typesEnum The types enumeration
	 * @return the most specific type for which all types in enum are subtypes
	 */
	public static AcmeType inferLeastCommonSupertype(AcmeEnumeration typesEnum) {
		AcmeVector typeVec = new AcmeVector(typesEnum);
		return AcmeTypeManager.inferLeastCommonSupertype(typeVec);
	}
	
	public static AcmeType inferGreatestCommonSupertype (AcmeVector types) {
		boolean allEltTypes = true;
		int typesSzie = types.getSize ();
		AcmeType returnType = ConstantTypes.UNSPECIFIED_TYPE;
		
		if (types.getSize () == 0) {
			return returnType;
		}
		else if (types.getSize () == 1) {
			return (AcmeType )types.getItemAt(0);
		}
		//TODO: Do hard work if there is more than one type
		return ConstantTypes.UNSPECIFIED_TYPE;
	}

} /* end AcmeTypeManager class description */